#include "pollard_rho_log_z_p.h"

#include "const_return_codes.h"
#include "common_utils.h"
#include <stdlib.h>
#include <iostream>


PollardRhoLogP::PollardRhoLogP() :ASupportingLogPAlg ()
{
	pollard = (pollard_entry*) Utils::AllocationMachine((void**)&pollard, DEFAULT_POLLARD_ENTRIES * sizeof(pollard_entry));
	pollard_maximal_index = DEFAULT_POLLARD_ENTRIES-1;
	for (int i=0; i<=pollard_maximal_index; i++) {
		mpz_init(pollard[i].value);
		mpz_init(pollard[i].basecount);
		mpz_init(pollard[i].valuecount);
		pollard[i].inited = 1;
	}
	mpz_init(fx1);
	mpz_init(bx1);
	mpz_init(vx1);
	mpz_init(fx2);
	mpz_init(bx2);
	mpz_init(vx2);

	restarts = 0;
	max_restarts = 30;
}

PollardRhoLogP::~PollardRhoLogP() {
	if (pollard != NULL) {
		for (int i=0; i<=pollard_maximal_index; i++) {
			if (pollard[i].inited) {
				mpz_clear(pollard[i].value);
				mpz_clear(pollard[i].basecount);
				mpz_clear(pollard[i].valuecount);
				pollard[i].inited = 0;
			}
		}
		free(pollard);
		pollard = NULL;
	}
	mpz_clear(fx1);
	mpz_clear(bx1);
	mpz_clear(vx1);
	mpz_clear(fx2);
	mpz_clear(bx2);
	mpz_clear(vx2);
}

/*!
 An overloaded method providing interface for DLP solving.
*/
int PollardRhoLogP::LogP( mpz_t aBase, mpz_t aValue, mpz_t aExponent) {
    int index = 0;
    int max = 100000;
    int rep_unit = 0;
    int current_rep_unit_index = -1;
    int next_rep_unit_index = 0;
    mpz_set_ui(fx1,1);        
    mpz_set_ui(bx1,0);        
    mpz_set_ui(vx1,0);        
    int local_restarts = 0;
    while (index < max) {    	
	// Next iteration of f(x)	
	switch (mpz_get_ui(fx1) % 3) {
	case 0:		mpz_powm_ui(fx1,fx1,2,Modulus);
			mpz_mul_ui(bx1,bx1,2);
			mpz_mul_ui(vx1,vx1,2);
			mpz_mod(vx1,vx1,Order);
			mpz_mod(bx1,bx1,Order);
			break;
	case 1:		mpz_mul(fx1,fx1,aBase);
			mpz_mod(fx1,fx1,Modulus);
			mpz_add_ui(bx1,bx1,1);
			break;
	case 2:		mpz_mul(fx1,fx1,aValue);
			mpz_mod(fx1,fx1,Modulus);
			mpz_add_ui(vx1,vx1,1);
			break;
	}
	// gmp_printf("x_%d = %Zd (%Zd, %Zd). Rep_unit = %d\n",index,fx1,bx1,vx1,rep_unit);
	// Counting f(x) - f(j_0). Has no sense for iteration #0
	if ((index != 0) && (index >= start_from) && (rep_unit != 0)) {
		if (0 == mpz_cmp(pollard[current_rep_unit_index].value,fx1)) {
//			gmp_printf("\nBase: %Zd, Value: %Zd, Modulus: %Zd, Order: %Zd",aBase,aValue,Modulus,Order);
//			gmp_printf("\nBx1: %Zd, Bx2: %Zd",bx1,pollard[current_rep_unit_index].basecount);
//			gmp_printf("\nVx1: %Zd, Vx2: %Zd",vx1,pollard[current_rep_unit_index].valuecount);
//			gmp_printf("\nValue 1: %Zd, Value2: %Zd",fx1,pollard[current_rep_unit_index].value);
//			gmp_printf("\nBase^Bx1*Value^Vx1 ~ Base^Bx2*Value^Vx2 modulo Modulus");
//			gmp_printf("\n%Zd^%Zd*%Zd^%Zd ~ %Zd^%Zd*%Zd^%Zd modulo %Zd\t\t=>",aBase,bx1,aValue,vx1,aBase,pollard[current_rep_unit_index].basecount,aValue,pollard[current_rep_unit_index].valuecount,Modulus);
			mpz_sub(bx1,  pollard[current_rep_unit_index].basecount, bx1);
			mpz_sub(vx1, vx1, pollard[current_rep_unit_index].valuecount);
//			gmp_printf("\nBase^(Bx2-Bx1) ~ Value^(Vx1-Vx2) modulo Modulus");
//			gmp_printf("\n%Zd^%Zd ~ %Zd^%Zd modulo %Zd",aBase,bx1,aValue,vx1,Modulus);
			// compute inverse: may not exist
			if (!mpz_invert(vx1,vx1,Order))  {
			// not comprime => restart
				local_restarts++;
				if (local_restarts > max_restarts) {
					restarts += local_restarts;
					return ConstRC::NotFactorized;
				};
				mpz_urandomm(bx1,RandomState,Order);
				mpz_urandomm(vx1,RandomState,Order);
				mpz_set(fx1,aBase);
				mpz_powm(fx1,fx1,bx1,Modulus);
				// array pollard is not used at the moment - we can use it for temporary value	
				mpz_set(pollard[0].basecount,aValue);
				mpz_powm(pollard[0].basecount,pollard[0].basecount,vx1,Modulus);
				mpz_mul(fx1,fx1,pollard[0].basecount);
				index = 0;
				rep_unit = 0;
				current_rep_unit_index = -1;
				next_rep_unit_index = 0;
			} else {
	//			gmp_printf("\n(Vx1-Vx2)^(-1) = %Zd",vx1);
				mpz_mul(vx1,vx1,bx1);
				mpz_mod(aExponent,vx1,Order);
				restarts += local_restarts;
				return ConstRC::Ok;
			};
		}
	}	
	if (index == rep_unit) {
		// gmp_printf("Index = %d, rep_unit = %d, rep_unit_index = %d. Setting as %Zd.\n",index,rep_unit,next_rep_unit_index,aux_1);
		mpz_set(pollard[next_rep_unit_index].value,fx1);				
		mpz_set(pollard[next_rep_unit_index].basecount,bx1);				
		mpz_set(pollard[next_rep_unit_index].valuecount,vx1);				
		start_from = ((rep_unit+1) >> 1)* 3;
		rep_unit <<= 1;
		rep_unit += 1;
		next_rep_unit_index++;
		current_rep_unit_index++;
		
	} 
	index++;
    }
	return ConstRC::NotFactorized;
}

int PollardRhoLogP::Test() {
	mpz_t modulo, base, exponent, value, result, limit, check;
	mpz_init(modulo);
	mpz_init(base);
	mpz_init(exponent);
	mpz_init(value);
	mpz_init(result);
	mpz_init(check);
	mpz_init(limit);
	mpz_set_str(limit,"100000",10);
	int tests = 100;
	int success = 0;
	int failure = 0;
	int error = 0;
	for(int i = 1;i<=tests; i++) {
		mpz_urandomm(modulo,RandomState,limit);
		mpz_nextprime(modulo,modulo);
		SetModulus(modulo);
		mpz_sub_ui(base,modulo,1);
		SetOrder(base);
		mpz_urandomm(exponent,RandomState,base);
		mpz_urandomm(base,RandomState,modulo);
		mpz_powm(value,base,exponent,modulo);
		int rvalue = LogP(base, value, result);
		if (rvalue == ConstRC::Ok) {
			mpz_powm(check,base,result, modulo);
			if (mpz_cmp(value,check)) {
				//gmp_printf("\n%Zd^%Zd %% %Zd = %Zd, ... %Zd ??",base,exponent,modulo,value,result);
//				gmp_fprintf(stderr,"E");
				error++;
			} else {
				success++;
//				gmp_fprintf(stderr,"S");
			};
		} else {
			failure++;
//			gmp_fprintf(stderr,"F");
		};
	};	
	std::cout << "\nTests: " << tests << "   Success: " << success << "    Failure: "<<failure << "    Error: "<< error << "   Restarts: "<< restarts;
	mpz_clear(modulo);
	mpz_clear(base);
	mpz_clear(exponent);
	mpz_clear(value);
	mpz_clear(result);
	mpz_clear(limit);
	mpz_clear(check);
};
