#ifndef __LINE_SIEVER__
#define __LINE_SIEVER__
#include "definitions.h"
#include "gmp.h"
#include "definitions.h"
#include "types.h"
#include "abstract_sieve.h"

#ifdef QS_LINE_SIEVING
#include "quadratic_sieve.h"
// Forward declaration
class QuadraticSieve;
#endif

#ifdef GNFS_LINE_SIEVING
#include "number_field_sieve.h"
#include "hermite_matrix_class.h"
// Forward declaration
class NumberFieldSieve;
class HermiteMatrix;
#endif

#if defined(__GNUC__)
#define PREFETCH(x)  __builtin_prefetch(x)
#else
#define PREFETCH(x) /* nothing */
#endif

enum internal_sieve_state { EBeforeInit = 0, EBeingInited, EWaitingForFBSetup, EReadyForSieving, EKSSieving, ENotReady, EBeingDestroyed };

/*! 
 * \brief Running factor base for QS
 
   Type taken and slightly redefined from Chris Papadopoulos' line siever.

   As the size of the sieving block can well exceed 64K, the 'h' value must be 32-bit.	

*/
typedef struct {
  main_sieving_type	 p;       /* prime */
  main_sieving_type 	 h;       /* offset within block */
#ifdef GNFS_LINE_SIEVING
   main_sieving_type 	c_p;	/* for algebraic primes only */
#endif
  log_type 	      logp;     /* log of prime */
} ks_update_t;

/*!
 * \brief updates for QS bucket sieving
 */
typedef struct {
  ks_update_t	updates[CACHE_UPDATES];
  int		used;	
} ks_cache_t;

/*!
 *\brief Linkes list of lists for bucket sieving in QS
	Type taken from Chris Papadopoulos' line siever.

	This is a type suitable for construction of a linked list of lists, where each member
	list is of reasonable size (to fit into cache).
``
	Since the size of ks_update_t has grown, MAX_UPDATES should be lower than original 1000.
	Currently, we use 500 (see definitions.h)
*/
typedef struct ks_script_t {
  struct ks_script_t *next;
  int	 num_used;
  double align_me;
  ks_update_t list[MAX_UPDATES];
} ks_script_t;

/* the blocks are indexed via a hashtable. Each block
   also maintains a cache of update structures that are
   packed together to conserve TLB resources */

typedef struct {
  ks_script_t *script;  /* linked list for this block */
  int cache_off;     /* offset into the cache array where block cache starts */
  int cache_used;    /* number of entries currently used in block's cache */
} hashtable_t;


/**
	\short Class for line sieving.

	Differences from the Texas Tech implementation:
	1) block size is variable
	2) caching of updates for a hashtable is different from the original.


*/
class LineSiever {

public:
	LineSiever();
	~LineSiever();
	
	/**
		\short Initializes the data structures.

		This method will initialize all the data structures except the factor base. All allocations will be done. 

		Returns OK in case of success, or error codes defined in definitions.h if any
		error occurs. The most probable error code is NOT_ENOUGH_MEMORY, if the current
		memory is insufficient for the demands of the line siever. This is a critical
 		error; the siever cannot be used if the structures are not all allocated well.
	*/
	int Init(line_sieve_size aLineSieveSize, int aLineLowerBound, int aLineUpperBound, int aFactorBaseUpperBound);

	/**
		\short Partial reset of the siever.

		This method will partially reset the siever. The factor base and interval sizes are not changed; the intended use is in situations, when the block size is changed
		from one value to another. In such case, the corresponding re-allocations will
		be done.

		Returns OK in case of success, various error codes for failures. NOT_ENOUGH_MEMORY is the most outstanding one.		
		
	*/
	int ResetBlocks(line_sieve_size aNewLineSieveSize);

#ifdef QS_LINE_SIEVING
	void SetQsFactorBase(qs_fb_type* aQsFactorBase); ///< Sets the pointer to the factor base. The factor base itself is allocated and deleted in QuadraticSieve class.

	void SetQsInstance(QuadraticSieve* aQsInstance); ///< Sets the pointer to the external quadratic sieve instance. The instance itself is allocated and deleted elsewhere.
#endif
#ifdef GNFS_LINE_SIEVING
	void SetNfsInstance(NumberFieldSieve* aNfsInstance); ///< Sets the pointer to the external number field sieve instance. The instance itself is allocated and deleted elsewhere.
	void SetNfsIntegralFactorBase(nfs_fb_type* aNfsIntegralFactorBase); ///< Sets the pointer to the integral factor base. The factor base itself is allocated and deleted in NumberFieldSieve class.
	void SetNfsAlgebraicFactorBase(nfs_fb_type* aNfsAlgebraicFactorBase); ///< Sets the pointer to the algebraic factor base. The factor base itself is allocated and deleted in NumberFieldSieve class.
	void SetNfsCompressEvenIntervals(bool aValue); ///< Sets the variable determining whether even intervals should be compressed (to avoid pairs (a,b) divisible by 2).
#endif
	
	int PerformSieving();

	int RunSieve();
	

private:

#ifdef GNFS_LINE_SIEVING
	bool  TestNfsIntegral(nfs_fb_type aInfo, main_sieving_type aUpdateIndex);
	bool  TestNfsAlgebraic(nfs_fb_type aInfo, main_sieving_type aUpdateIndex);
	void  PrintNfsElementInfo(nfs_fb_type aInfo);
	
#endif
	void  PrintUpdateInfo(ks_update_t aInfo);
	/**
		\short Common method for Init and ResetBlocks; reallocates the internal structures.

		This method will be called by both Init and ResetBlocks to (re)allocate blocks
		and hashtables used in the sieve. It expects to have all relevant parameters
		already set to reasonable values.

		Returns OK in case of success, various error codes for failures (like NOT_ENOUGH_MEMORY).

	*/
	int InternalReset();

	/**
		\short Deletes the existing hashtable and its internal structures.

		This method will delete the current hashtable and scripts in it.
	
	*/
	void DeleteHashtable(hashtable_t* aHashtable, int aMax);

	int  SetupHashtable(hashtable_t** aHashtable, ks_cache_t** aCache);

	/**
		\short Adds a new script to the linked list in the hashtable.

		Will allocate (using malloc) a new script for the linked list in the hashtable.
		Returns OK on success, NOT_ENOUGH_MEMORY otherwise.
	
	*/
	int  AddScript(ks_script_t* aLast);

	/**
		\short Add a new update to an indicated hashtable.

		This method adds the indicated values to the hashtable.
	
	*/
	int  AddToTable(hashtable_t* aHashtable, ks_cache_t* aCache, main_sieving_type aPrime, main_sieving_type aRoot, main_sieving_type aCp, log_type aLog);

	int  AddToTable(hashtable_t* aHashtable, ks_cache_t* aCache, main_sieving_type aPrime, main_sieving_type aRoot, main_sieving_type aCp, log_type aLog, int aHashtableIndex);

	int  FlushUpdateCache(hashtable_t* aHashtable, ks_cache_t* aCache);


	int SieveLine();	
	int BlockSetSieve(int aBlockMax, int aSign, bool aInitNext);
	
	/**
		\short This method will fill the hashtable(s) for sieving with largish primes.

		This method will fill the hashtable(s) for sieving with largish primes.
		In case of the quadratic sieve, only one hashtable is used; in NFS, two of them;
		that is why the hashtable parameter is used.

		The direction argument distinguishes positive and negative parts of the line.
		In the negative part, update of the sieving indices for a new line is also done,
		because the corresponding parts of the factor base are in cache at that moment.

		The aIsAlgebraic serves for easier distinguishing of algebraic and other.
	*/
	int FillHashtable(hashtable_t* aHashtable, ks_cache_t* aCache, int aDirection, bool aIsAlgebraic);


	void  ClearHashtables(hashtable_t* aTableList, ks_cache_t* aCacheList);
	void  ZeroizeLastHashtable(hashtable_t* aTableList);
	void  SwapHashtable(hashtable_t* aTableList, int aIndex);

	bool  TestQs(main_sieving_type aUpdateIndex, main_sieving_type aPrime);
	bool  TestQs(main_sieving_type aUpdateIndex, main_sieving_type aPrime, bool aPrintOK);
	bool  TestNfs(main_sieving_type aUpdateIndex, main_sieving_type aPrime, bool aPrintOK);

	void  PrintHashtables(hashtable_t* aTableList);

#ifdef QS_LINE_SIEVING
	int   SieveWithMiddlePrimes(qs_fb_type* aStart, qs_fb_type* aEnd, log_type* aBlock, int aSign, bool aInit);
	int   SetupRootShort();
	void  DetermineLoAndHiBlocks(main_sieving_type aRootAbs, int& loBlockIndex, int& hiBlockIndex);
	void  PrintDivisorsOfA();
#endif
#ifdef GNFS_LINE_SIEVING
	int   SieveWithMiddlePrimes(nfs_fb_type* aStart, nfs_fb_type* aEnd, log_type* aBlock, int aSign, bool aInit);
	inline void UpdateNfsRoot(nfs_fb_type* aEntry);
	bool   TestCoprimality();
	bool   CheckTwoInCompressedCase();
#endif
	int   SieveWithLargishPrimes(ks_script_t* aScript, log_type* aBlock);
	int   ShiftLargishPrimes(ks_script_t* aScript, hashtable_t* aTableList, ks_cache_t* aCacheList);
	int   FindAndFactorizeCandidates(hashtable_t* aRationalHashtable, hashtable_t* aAlgebraicHashtable);
	int   Factor(int aOffset,hashtable_t* aRationalHashtable, hashtable_t* aAlgebraicHashtable);
#ifdef GNFS_LINE_SIEVING
	int   FactorAlgebraic(main_sieving_type aOffset,main_sieving_type aAuxiliaryOffset, hashtable_t* aAlgebraicHashtable);
	int   AddAlgebraicDivisor(main_sieving_type& aDivisor, main_sieving_type& aExponent, main_sieving_type& aCp,int aFlags, int& aLastWrittenIndex);
#endif

	void  WalkTheBlock();
	inline bool  IsDivisorOfA(main_sieving_type& aPrime);
#ifdef QS_LINE_SIEVING
	void  WalkTheQsFactorBase();
	inline void CalculateNewRoots(qs_fb_type* aEntry, main_sieving_type inverse_a);
	// Rozmyslet problem casteho testovani subtract!s
	inline void UpdateSiqsRoots(qs_fb_type* aEntry);
	//inline void UpdateSiqsRoots(qs_fb_type* aEntry);	
	bool  InsertSiqsDivisor(main_sieving_type& aPrime, int& aExponent);
#endif

	void  ResetRelation();
	void  SetupRelation(int aX);

	int   AddIntegralDivisor(main_sieving_type& aDivisor, int& aExponent, int& aLastWrittenFactor);

	void  SearchForPrimeInHashtable(hashtable_t* aHashtable, main_sieving_type aPrime, main_sieving_type aCp);
	
private:
	// Internal parameters of the line siever. Cannot be manipulated (set) from the outside, save by special methods as Init()
	line_sieve_size 	 iLineSieveSize; /// Size of the block as enum.
	int			 iBlockSize; /// Size of the block as an integer. Should be equal to iLineSieveSize
	int			 iFactorBaseUpperBound; /// The upper bound of the factor base.
	log_type*		 iBlock; /// The sieving block itself (for integers)

	// Current sieving properties.
	int			 iLineLowerBound; ///< must be < 0; determines M_0 such that we sieve through [-M_0, -1]. Is set directly by Init().
	int			 iLineUpperBound; ///< must be > 0; determines M_1 such that we sieve through [1, M_1]. Is set directly by Init().
	int			 iPositiveBlocks; ///< Is equal to number of positive blocks (basically, upper integral part of M_1/ iBlockSize)
	int			 iNegativeBlocks; ///< Is equal to number of negative blocks (basically, upper integral part of -M_0/ iBlockSize)
	int			 iBlockSetSize; ///< Size of the set of blocks determined by upper integral part of iFactorBaseUpperBound/ iBlockSize. Denoted by K in the accompanying material.

	internal_sieve_state 	 iInternalSieveState; ///< Internal variable for checking the sieve state.

	int			 iAllocatedHashtableSize;
	hashtable_t*		 iHashtable; ///< Internal hashtable for sieving with largish primes.
	
	ks_cache_t*		 iCaches; ///< Caches for writing into the internal hashtable.



#ifdef GNFS_LINE_SIEVING	
	hashtable_t*		 iAlgebraicHashtable; ///< Internal hashtable for sieving with largish prime ideals. Not needed in quadratic sieve.
	int			 iAlgebraicAllocatedHashtableSize;
	log_type		 iInitAlgebraicsBlockValue;
	log_type*		 iAlgebraicBlock; ///< Internal sieving block for algebraic blocks.

	ks_cache_t*		 iAlgebraicCaches; ///< Caches for writing into the algebraic hashtable.
#endif

#ifdef QS_LINE_SIEVING
	qs_fb_type*		 iQsFactorBase; ///< A pointer to the factor base. The factor base itself is allocated and deleted in QuadraticSieve class.
	qs_fb_type*		 iQsFirstMiddlePrime;
	qs_fb_type*		 iQsLastMiddlePrime; // Must be different from FirstLargishPrime-1, since there are instances where no largish primes exist.
	qs_fb_type*		 iQsFirstLargishPrime;

	QuadraticSieve*		 iQsInstance; ///< the sieving instance for MPQS/SIQS. Needed for communication about polynomial re-generation etc.
	qs_mode 		 iQsMode; // Mpqs vs. Siqs
	bool			 iQsMajorPolyChange; // Siqs ... 'a' was changed (once every 2^{s-1} polys)
	bool			 iQsCounterSinceMPC; ///< Counter of polynomials since the last major poly change.
	quad_polynomial		 iQsCurrentPolynomial; ///< the current polynomial to be sieved with
	quad_polynomial	 	 iQsNextPolynomial; ///< the next polynomial to be sieved with
	int			 iQsSubtract; ///< for Siqs root recalc only
	int			 iQsNu;

	qs_relation*		 iQsRelation;

	main_sieving_type*	 iQsCurrentDivisorsOfA;
	main_sieving_type	 iLeastCurrentDivisorOfA;	
	main_sieving_type	 iGreatestCurrentDivisorOfA;

	main_sieving_type*	 iQsNextDivisorsOfA;
	main_sieving_type	 iLeastNextDivisorOfA;	
	main_sieving_type	 iGreatestNextDivisorOfA;

	int			 iQsSiqsSValue;
	bool			 iQsUseRootShort;
	// Roots of the current polynomial
	main_sieving_type	 iQsRoot1;
	main_sieving_type	 iQsRoot2;
	main_sieving_type	 iQsNextRoot1;
	main_sieving_type	 iQsNextRoot2;

	int			 iQsPositiveLowestBlockIndex;
	int			 iQsPositiveHighestBlockIndex;
	int 			 iQsPositiveLowestBlockSetIndex;
	int 			 iQsPositiveHighestBlockSetIndex;
	int			 iQsPositiveStartSetIndex;
	int			 iQsPositiveEndSetIndex;

	int			 iQsNegativeLowestBlockIndex;
	int			 iQsNegativeHighestBlockIndex;
	int 			 iQsNegativeLowestBlockSetIndex;
	int 			 iQsNegativeHighestBlockSetIndex;
	int			 iQsNegativeStartSetIndex;
	int			 iQsNegativeEndSetIndex;
#endif // QS_LINE_SIEVING

#ifdef GNFS_LINE_SIEVING
	nfs_fb_type*		 iNfsIntegralFactorBase; ///< A pointer to the integral factor base. The factor base itself is allocated and deleted in the NumberFieldSieve class.
	nfs_fb_type*		 iNfsFirstIntegralMiddlePrime;
	nfs_fb_type*		 iNfsLastIntegralMiddlePrime; // Must be different from FirstLargishPrime-1, since there are instances where no largish primes exist.
	nfs_fb_type*		 iNfsFirstIntegralLargishPrime;

	nfs_fb_type*		 iNfsAlgebraicFactorBase; ///< A pointer to the algebraic factor base. The factor base itself is allocated and deleted in the NumberFieldSieve class.
	nfs_fb_type*		 iNfsFirstAlgebraicMiddlePrime;
	nfs_fb_type*		 iNfsLastAlgebraicMiddlePrime; // Must be different from FirstLargishPrime-1, since there are instances where no largish primes exist.
	nfs_fb_type*		 iNfsFirstAlgebraicLargishPrime;

	NumberFieldSieve*        iNfsInstance; ///< the sieving instance for NFS. Needed for communication regarding new smooth / partial values etc.

	nfs_relation*		 iNfsRelation;
	log_type		 iInitAlgebraicBlockValue;
	mpz_t			 iM;
	HermiteMatrix* 		 iIntegralBase;
	bool			 iNfsCompressEvenIntervals;
	bool			 iNfsOddToEven;
	bool			 iNfsEvenToOdd;
	bool			 iIsCompressedNow;

	bool			 iShiftingAlgebraic;

#endif // GNFS_LINE_SIEVING


	log_type		 iInitIntegralBlockValue;

	mpz_t			 aux_1;
	mpz_t		         aux_2;
	mpz_t			 aux_3;
	mpz_t			 aux_4;
	mpz_t			 aux_5;
	int			 inited;
	int			 iAccepted;
	int			 iRejected;
	int			 iTotalDivrem;
	
	bool			 iSievingAlgebraic;
	bool			 iSievingIntegral;
	 			
	int			 iTimeMessage;

	bool			 iSievingNegative;
	int			 iCurrentBlockSetIndex;	
	int			 iCurrentBlockIndex;
	bool			 iAssertionMode;
	bool			 iCheckAllMode;
	int			 iMaxLines;

	int			 iSievingTimeStart;
	int			 iSievingTimeEnd;

	int			 iSievingClockStart;
	int			 iSievingClockDiff;
	double		 	 iSievingClockTotal;
	int		         iAuxStart;
	double			 iBlockSieveTotal;
	double			 iFillHashtableTotal;
	double			 iFillHashtableNeg;
	double			 iRootsUpdateTotal;
	double			 iFindAndFactorizeTotal;

	mpz_t			 iFactoredValue; // kN
	bool			 iFVInited;
	
	bool			 iUseHashingForMiddlePrimes;
	bool			 iUseMirrorTableForLargishPrimes;

	float			 iIntermediateIntegers;
	float			 iIntermediateIntegersStart;
	float			 iIntermediateAlgebraic;
	float			 iIntermediateAlgebraicStart;
	float			 iLocalPerform;
	float			 iLocalStart;
	float			 iDiff;



public:
	void			SetAssertionMode(bool aValue);
	void 			SetInitBlockValue(log_type aValue);
	void			SetMaxLines(int aValue);
	void			SetUnlimitedSieving();
#ifdef QS_LINE_SIEVING
	void			SetQsUseRootShort(bool aValue);
#endif
#ifdef GNFS_LINE_SIEVING
	void			SetInitAlgebraicBlockValue(log_type aValue);
#endif
	line_sieve_size		GetLineSieveSize()  const; ///< Returns line sieve size (size of the block) as enum.
	int			GetLineLowerBound() const; 
	int			GetLineUpperBound() const;
	internal_sieve_state 	GetSieveState() const;
	
	
};
#endif // __LINE_SIEVER


