#ifndef __QS_PARAMETERS_H__
#define __QS_PARAMETERS_H__

#include "gmp.h"
#include "types.h"
#include "quadratic_sieve.h"
#include <string>

#include "job_parameters.h"

const 	       int KMaxMSD     = 32;



/*!
 * \brief contains the information necessary for a QS run. 
 *
 * The parameters
 are most often given from the command line, but they can be acquired from other
 sources as well.

 It serves as a communication channel between the user interface and the engine of QS.
*/
class QSParameters : public JobParameters
{
public:	// Construction and destruction
	
	// The default constructor
	QSParameters();
	// The copy constructor
	QSParameters(const QSParameters& aSrc);
	// The constructor that copies info from current instance
	QSParameters(const QuadraticSieve& aInstance);

	virtual ~QSParameters();

private:
	void InitDefaults();

public:
	// Inherited from JobParameters
	virtual const string  GetStringParameter(const char* aName) const;
	virtual int      GetIntegerParameter(const char* aName) const;	
	virtual double   GetDoubleParameter(const char* aName) const;	
	virtual bool     GetBoolParameter(const char* aName) const;
	virtual int     GetMpzParameter(const char* aName, mpz_t aResult) const;

	// Are the current QS parameters complete? (can a job be run with them?)
	// Complete job must contain:
	// - positive modulus 		>= KMinimalModulus (2^32-1)
	// - positive FB upper bound 	>= KMinimalFBBound (256)
	// - positive sieving interval  >= KMinimalSievingInterval (2048)
	virtual bool	 Complete() 	    const;
	virtual string	 GetFailureReason() const;

	virtual QSParameters*	CreateCopy() const;

	virtual	void	Print() const;
	virtual void	PrintInBatch(const char *aPrefix) const;

public:

	// Getters
	bool	Is_AssertionMode() const;
	qs_mode	Get_QSMode() const;
	bool	Is_InfoMode() const;
	bool	Is_DebugMode() const;
	check_mode	Get_CheckMode() const;
	bool	Is_LtMode() const;
	siev_i_mode	Get_SievingIntervalMode() const;
	int	Get_CheckLevel() const;
	int	Get_MemblockSize() const;
	int	Get_TimeMessageInterval() const;
	bool	Is_DiMode() const;
	linalg_type	Get_LinalgType() const;
	bool	Is_WpList() const;
	variations_types	Get_UsingVariations() const;
	bool	Is_RankCalculationMode() const;
	bool	Is_ValueDistributionMode() const;
	print_sort	Get_PrintSort() const;
	bool	Is_ExtendedK() const;
	sgt_mode Get_SingletonMode() const;
	power_mode Get_PowerMode() const;
	factor_alg Get_FactorAlg() const;
	void	Get_Modulus(mpz_t aTarget) const;
	int	Get_MaxIter() 	   const;
	int	Get_SievingIntervalLength() const;
	int	Get_UpperIntBound() const;
	bool	Is_RootShort() const;
	string	Get_DirectoryName() const;
	int	Get_LoadFlags() const;
	int	Get_CompressionLevel() const;	
	int	Get_ThresholdOptimizer_Level() const;
	string Get_FrontendPipe() const;
	machine_specific_generation_type Get_MachineSpecificGeneration() const;


	// Distribution-related getters

	bool		Machine_Specific_Divisors_Given() const;
	int		Machine_Specific_Divisors_Count() const;
	int		Get_MSD(int aIndex) const;
	bool		MSD_Interval_Given() const;
	int		MSD_Lower_Bound() const;
	int		MSD_Upper_Bound() const;

	// Setters
	void	Set_AssertionMode(bool aValue);
	void	Set_QSMode(qs_mode aValue);
	void	Set_InfoMode(bool aValue);
	void	Set_DebugMode(bool aValue);
	void	Set_CheckMode(check_mode aValue);
	void	Set_LtMode(bool aValue);
	void	Set_SievingIntervalMode(siev_i_mode aValue);
	void	Set_CheckLevel(int aValue);
	void	Set_MemblockSize(int aValue);
	void	Set_TimeMessageInterval(int aValue);
	void	Set_DiMode(bool aValue);
	void	Set_LinalgType(linalg_type aValue);
	void	Set_WpList(bool aValue);
	void	Set_UsingVariations(variations_types aValue);
	void	Set_RankCalculationMode(bool aValue);
	void	Set_ValueDistributionMode(bool aValue);
	void	Set_PrintSort(print_sort aValue);
	void	Set_ExtendedK(bool aValue);
	void	Set_SingletonMode(sgt_mode aMode);
	void	Set_PowerMode(power_mode aMode);
	void	Set_FactorAlg(factor_alg aAlg);
	void	Set_Modulus(const mpz_t aValue);
	void	Set_MaxIter(int aValue);
	void	Set_SievingIntervalLength(int aValue);	
	void	Set_UpperIntBound(int aValue);
	void	Set_RootShort(bool aValue);
	void	Set_DirectoryName(const string& aValue);
	void	Set_DirectoryName(const char*  aValue);
	void	Set_LoadFlags(int aValue);
	void	Set_Flag(int aFlag);
	void	Set_CompressionLevel(int aValue);
	// 0 ... threshold optimizer off.
	void	Set_ThresholdOptimizer_Level(int aValue);
	void 	Set_FrontendPipe(const char* aValue);
	void	Set_MachineSpecificGeneration(machine_specific_generation_type aValue);

	// Distribution-related setters

	void    Set_Machine_Specific_Divisors_Given(bool aState);	
	void	Add_MSD(int aMachineSpecificDivisor);
	void	Set_MSD_Interval_Given(bool aState);
	void 	Set_MSD_Lower_Bound(int aLower);
	void 	Set_MSD_Upper_Bound(int aUpper);

public:
	const static unsigned int  KMinModulus;
	const static char*	   KModulusMustBePositive;
	const static char*    	   KModulusTooSmall;
	const static unsigned int  KMinFBBound;
	const static char*    	   KFBBoundTooSmall;
	const static unsigned int  KMinSieveInt;
	const static char*    	   KSieveIntTooSmall;
	const static char*	   KSeparator;
	const static char*	   KFinisher;

private:

	bool		iAssertionMode; ///
	qs_mode		iQsMode;	///
	bool		iInfoMode;	///
	bool		iDebugMode;
	check_mode 	iCheckMode;	///
	bool		iLtMode;	///
	siev_i_mode     iSievingIntervalMode; ///
	int		iCheckLevel;	///
	int 		iMemblockSize;	///
	int    		iTimeMessageInterval; ///
	bool		iDiMode;	///
	linalg_type 	iLinalgType;	///
	bool		iWpList;	///
	variations_types 	iUseVariations;	///
	bool	        iRankCalculationMode; ///
	bool		iValueDistributionMode; ///
	print_sort	iPrintingSort;	///
	bool		iUseExtendedK;  ///
	sgt_mode	iSingletonMode; ///
	power_mode	iPowerMode;     /// 
	factor_alg	iFactorAlg;     ///
	mpz_t 		iN;		///
	int		iMaxIter;	///
	unsigned int	iSievingIntervalLength;	///
	unsigned int	iUpperIntBound;	///
	bool		iUseRootShort;	///
	string		iDirectoryName; ///
	int		iLoadFlags;	///
	int 		iCompressionLevel; ///


	string		iFailureReason;
	
	int		iMachineSpecificDivisors[KMaxMSD];
	int		iMSDGiven;
	int		iMSDIntervalGiven;
	int		iLowerMSDBound;
	int		iUpperMSDBound;
	machine_specific_generation_type iMachineSpecificGeneration;

	
	int		iThresholdOptimizerLevel;

	string		iFrontendPipe;
};


#endif

