/*! \file
 * \brief class #QuadraticSieve header file
*/

#ifndef __QUADRATIC_SIEVE__
#define __QUADRATIC_SIEVE__
#include "types.h"
#include "factor_alg.h"
#include "pollard_rho.h"
#include "abstract_sieve.h"
#include "lanczos_class.h"
#include "bit_matrix_class.h"
#include "sparse_matrix_class.h"
#include "line_siever.h"
#include <string.h>
#include "ttr_class.h"
#include "abstract_connector.h"
#include "node_info.h"
#include "distributed_algorithm.h"
#include "connectivity_framework_definitions.h"
#include "threshold_optimizer.h"
#include <pthread.h>
#include "abstract_supporting_factor_alg.h"
#include "lp_hashtable_common.h"


#ifndef WIN32
   #include <inttypes.h>
#endif

class	QSParameters; // FWD declaration

// These are small primes read from primefile. As this reading is an old C function,
// it reads the contents into a global variable.
extern "C" long *small_primes;
extern "C" long length_of_small_primes;



// Forward declaration
class LineSiever;

// Types for the "most critical" data structures - sieving matrix, logarithm table etc. 
// Useful in performance tuning.

/*!
	\brief The core class for MPQS/SIQS algorithm.
	
	This class serves as an implementation class for QuadraticSieve class. It is by far the most complicated class in the MPQS/SIQS program.
	
	This class is very rich in member variables. These variables represent parameters of the sieving proces. Their access modifier is 'protected'.
	
		The core of the MPQS program lies in business methods of quadratic_sieve.cpp. Since the algorithm
	is pretty complicated, there are lots of business methods, each taking care of a specific MPQS-related task.
	
	The approach chosen in programming these business methods is the following:
	
	-# The methods are named accordingly to what they perform
	-# The methods are hierarchically sorted according to the phase of the algorithm. Basically, they are
	   divided into first-, second- and third-level methods. There are very few first-level methods, namely 5;
	   each of them calls several second-level methods in given order, and these methods call the third-level
	   methods. It is usually the third-level methods which perform the "real" operations. Sometimes, these
	   third-level methods call another auxiliary methods for help.
	-# Each "fatal" error, like allocation failure or calculation error, propagates to the top level and forces
	   the whole calculation to abort.
	
	Aside from MPQS-related methods, there is one more method in quadratic_sieve.cpp, which takes care of
	TTR invocation:
	
	int QuadraticSieve::RunTTR(testing_mode aMode) 
	
	This method is called iff user specifies that he wants to perform Thorough Test Routine (by -t,-tfast,-tstrict).


*/
class QuadraticSieve : public AbstractSieve, public virtual DistributedAlgorithm {
     public:
	// CONSTRUCTORS AND DESTRUCTORS
	QuadraticSieve();
	QuadraticSieve(AbstractSieve* aSieve);
	~QuadraticSieve();
	int  SetupParameters(const JobParameters& aParameters);
	int  SetupParameters(const QSParameters& aParameters);	

	// GETTERS AND SETTERS
	void Set_UsingVariations(variations_types aValue);
	void Set_M(mpz_t aValue);
	void Set_Centre(mpz_t aValue);
	void Set_Root(mpz_t aValue);
	void Set_Acc(mpz_t aValue);
	void Set_MPF(mpz_t aValue);
	void Set_BLMode(bool aValue);
	void Set_LIMode(bool aValue);
	void Set_LTMode(bool aValue);
	void Set_WPList(bool aValue);
	void Set_CheckMode(check_mode aMode);
	void Set_CheckLevel(int aValue);
	void Set_NrPolynomials(long aValue);
	void Set_UpperFBBound(long aValue);
	void Set_DIMode(bool aValue);
	void Set_PrintingSort(print_sort aValue);	
	void Set_QSMode(qs_mode aValue);
	void Set_S(int aValue);
	void Set_TimeMessageInterval(int aValue);
	void Set_SingletonMode(sgt_mode aValue);
	void Set_PowerMode(power_mode aValue);
	void Set_FactoringAlgorithm(factor_alg aValue);
	void Set_SievingIntervalMode(siev_i_mode aValue);
	void Set_ValDistMode(bool aMode);
	void Set_ExtK(bool aValue);
	void Set_MemblockSize(int aValue);
	void Set_MemblockSizeLSS(line_sieve_size aValue);
	void Set_UseRootShort(bool aValue);
	void Set_DirectoryName(const char* const aValue);
	void Set_FactorBase(main_sieving_type* aFB, long aMaxFBIndex, long aFBAllocated);
	void Set_InterruptNow(bool avalue);
	void Set_IdealAValue(mpz_t aValue);
	void Set_ContiniThreshold(log_type aValue);
	void Set_ErrorFactor(log_type aValue);
	void Set_SmoothsFound(long aValue);
	void Set_1PartialsFound(long aValue);
	void Set_2PartialsFound(long aValue);
	void Set_XPartialsFound(long aValue);
	void Set_RelationProcessingPhase(relproc_phase aPhase);
	void Set_RelationProcessingFlags(int aValue);
	void Set_Polynomials(quad_polynomial* aPolynomials, long& aNumberOfPolynomials, long& aAllocatedPolynomials, long& aIndexInPolynomial, long& aCurrentPolynomial);
	void Set_Nu(int aValue);
	void Set_Subtract(int aValue);
	

	variations_types Get_UsingVariations() const;
	void Get_M(mpz_t aRop) const;
	int  Get_Integer_M() const;
	void Get_Centre(mpz_t aRop) const;
	void Get_Root(mpz_t aRop) const;
	void Get_Acc(mpz_t aRop) const;
	void Get_MPF(mpz_t aRop) const;
	bool Get_BLMode() const;
	bool Get_LIMode() const;
	bool Get_LTMode() const;
	bool Get_WPMode() const;
	void Get_kN(mpz_t aRop) const;
	check_mode Get_CheckMode() const;
	int  Get_CheckLevel() const;
	long Get_NrPolynomials() const;
	long Get_UpperFBBound() const;
	bool Get_DIMode() const;
	print_sort Get_PrintingSort() const;
	qs_mode Get_QSMode() const;
	int Get_S() const;
	int Get_TimeMessageInterval() const;
	sgt_mode Get_SingletonMode() const;
	power_mode Get_PowerMode() const;
	factor_alg Get_FactoringAlgorithm() const;
	siev_i_mode Get_SievingIntervalMode() const;
	bool Get_ValDistMode() const;
	bool Get_ExtK() const;
	int  Get_Subtract() const; // SIQS only
	int  Get_Nu() const; // SIQS only
	int  CopyADivisors(main_sieving_type* aArray) const;
	bool Get_UseRootShort() const;
	char* Get_DirectoryName() const;
	main_sieving_type* Get_FactorBase(long& aMaxFBIndex, long& aFbAllocated);
	long  Get_MaxFBIndex() const;
	bool  Get_Interruptable() const;
	bool  Get_InterruptNow() const;
	void  Get_IdealAValue(mpz_t aTarget) const;
	log_type Get_ContiniThreshold() const;
 	log_type Get_ErrorFactor() const;
	long Get_SmoothsFound() const;
	long Get_1PartialsFound() const;
	long Get_2PartialsFound() const;
	long Get_XPartialsFound() const;
	relproc_phase Get_RelationProcessingPhase() const;
	int Get_RelationProcessingFlags() const;
	LineSiever* Get_AdvancedSieverInstance() const;
	line_sieve_size Get_MemblockSize() const;
	void	Set_RelationProcessingFlag(int aFlag);
	void	Clear_RelationProcessingFlag(int aFlag);	
	quad_polynomial* Get_Polynomials(long& aNumberOfPolynomials, long& aAllocatedPolynomials, long& aIndexInPolynomial, long& aCurrentPolynomial) const;

	/// Getters and setters for distribution-related parameters
	bool  		Get_IsDistributed() const;
	distribution    Get_DistributionMachineType() const;
	const char*		Get_SharedURL() const;	
	const char*		Get_MachineName() const;	
	int		Get_CommitInterval() const;
	const main_sieving_type* Get_MachineSpecificDivisors() const;
	unsigned int  	Get_MSDCount() const;
	main_sieving_type  Get_LowerMachineBound() const;
	main_sieving_type  Get_UpperMachineBound() const;

	machine_specific_generation_type Get_MachineSpecificGeneration() const;
	void Set_MachineSpecificGeneration(machine_specific_generation_type aValue);

	void		Set_DistributionMachineType(distribution aValue);
	void		Set_CommitInterval(int aValue);
	// A method that sets msd's from an external source (like job.xml), typical for nodes of a distributed computation.
	void 		Set_MachineSpecificDivisors(main_sieving_type* aDivisors, int aLength);
	void		Set_MSDCount(int aCount);
	void		Set_LowerMachineBound(main_sieving_type aPrime);	
	void		Set_UpperMachineBound(main_sieving_type aPrime);	

	// From DistributedAlgorithm
	virtual Batch	CreateNewJobs(unsigned int aBatchSize = KDefaultBatchSize);
	virtual void	RunNodeInstance();
	virtual bool	NodeInstanceRunning() const;
	virtual bool	HasDataToSend() const;	
	virtual void	ClearFreshData();
	virtual unsigned int DataToSend() const;	
	virtual const char* DataUnit() const;
	virtual void	LockDataMutex();
	virtual void	UnlockDataMutex();
	virtual bool	IsDistributedPhaseFinished() const;
	virtual int	RunNondistributedPhase();
	int DeleteParameters(const JobParameters *aParameters);

	virtual bool IsComputationFinished();
	virtual void SetComputationFinished(bool aFlag);

	virtual void	InterruptNodeNow();
	virtual void    ResetCenter();

	// Useful for XmlQsSerializator and distribution
	virtual int	CloseDataFile(data_file_type aType);
	virtual FILE*	OpenDataFile(data_file_type aType, int aMode);

/*	bool		iIsDistributed;
	distribution	iDistributionMachineType;
	char*		iSharedURL;
	int		iAllocatedURLSpace;
	char*		iMachineName;
	int		iAllocatedMachineNameSpace;
	int		iCommitInterval;	*/


    // executive methods for QS. For return values, see definitions.h
    	int  Reload();
        int  RunMPQS();
	int  RunTTR(testing_mode aMode);
	void PolyCopy(quad_polynomial* aTarget);
	int  RootCopy(main_sieving_type& aTarget1, main_sieving_type& aTarget2);
	int  GenerateNewPolynomial();
	void IncrementPolynomialCounter();
	void DestroyPolynomial(quad_polynomial* aPoly);

	// A demo method
	int  DemonstrateBatches();

	/*static int Mulm(INT64 arg1, INT64 arg2, int mod);
	static int Mulm(INT64 arg1, int arg2, int mod);
	static int Mulm(int arg1, int arg2, int mod);
        static int Powm(int arg, int exp, int mod);*/

// New sieving
	main_sieving_type  Evaluate(main_sieving_type aX, main_sieving_type aP);
	main_sieving_type  Evaluate(quad_polynomial* aPoly,main_sieving_type aX, main_sieving_type aP);
	main_sieving_type  EvaluateWithPrint(quad_polynomial* aPoly,main_sieving_type aX, main_sieving_type aP);
	//main_sieving_type  Evaluate(quad_polynomial* aPoly,main_sieving_type aX);
	static void PrintPolynomial(quad_polynomial* aPoly);
	static void PrintPolynomialMod(quad_polynomial* aPoly, main_sieving_type aP);
	
// Relation-related methods
	qs_relation* InitRelation();
	void ClearRelation(qs_relation* aRelation);

	void PrintTimeInfo(int aWhich, double& aInProcess, double& aTotal);
	void PrintEfficiencyInfo(int aBigDiv, int aAccepted, int aRejected, int aLines, double aTotalTime);

	void AllocPolynomial(quad_polynomial& aPolynomial);
	void ClearPolynomial(quad_polynomial& aPolynomial);

	mpz_t*	Get_B2Block();
	void    Clear_B2Block();
	void 	Set_B2Block(mpz_t* aB2Block, int* aB2_Inited);
	int     Get_Solve() const;
	void    Set_Solve(int aValue);

	int     Revert(long aSmooths, long aPartials);
	

	integer_solutions& Get_Solutions();
	void    Set_Solutions(integer_solutions aSolutions);
	void    Clear_Solutions();
	// Tests integrity of the solution arrays
	bool    Verify_Solutions();
	/*!	For sieving, solutions of current polynomial modulo all primes
	 * of the factor base are needed. These solutions
	occupy some memory space; this memory space is allocated in this
	procedure.  The solutions are contained in member variable
	#int_sols_of_cur_poly respectively 3 arrays in this
	struct-variable: 
	- sol_1
	- sol_2 (this expresses the fact that Q(x) mod
	p may have, and usually has, 2 solutions)
	- usable, which expresses whether solutions at
	given indices are valid or no (sometimes, Q(x) reduces to linear
	equation and then it has at most 1 solution mod p; therefore, the
	second solution is invalid).  
	
	If any of the allocations fails, the
	method returns NOT_ENOUGH_MEMORY, else OK. */
	int 	AllocSolutions();
	
	main_sieving_type* Get_DivisorsOfA();
	void		   Set_DivisorsOfA(main_sieving_type* aDivisorsOfA);

	

	/*!
		This method initializes the necessary data for the user,
		if the user wishes to look at the details of a saved
		factorization job, but does not wish to rerun the job again.
	*/
	int InitForList(const char* aDirectoryName);

	void WaitForNodeSetup();
	int  RunCenter();

	const char* AlgorithmName() const;
	JobParameters* CreateNewParameters() const;

	ThresholdOptimizer&	GetOptimizer() const;

	int 	GetOptimizationPolicy() const;
	void	SetOptimizationPolicy(int aValue);

private:
	// GENERIC METHODS FOR QSIEVE - 1st level (being called from RunMPQS)
	
	/// Performs operations needed for setting up the algorithm environment.	
	/*! In detail:
	 * - checks whether factored number and sieving interval are set (via CheckInitConsistency())
	 * - checks whether factored number is composite (via CheckPrimality())
	 * - fills array #small_primes with small primes (via ReadPrimesFromFile())
	 * - checks that #N is not divisible by prime of #small_primes (via CheckDivisibility())
	 * - computes #k, #value_d and #root (via CalculateKDRoot())
	 * - sets #maximal_partial_factor and #hash_root (via AllocVariationsAndHashes())
	 * - define factor base (via DefineFactorBase())
	 * - if sieving should be performed with prime powers (see #ESieveWithPrimePowers] then also
	 *   - prepares structures for sieving with prime powers (via DefinePrimePowers())
	 *   - calculates square root of kN mod prime powers used for sieving (via CalculateRootsOfkNForPrimePowers())
	 * - if Variations should be used (see #EDoNotUseVariations) prepares structures for Variations 
	 *   	(via SetFactoringStructures())
	 * - allocates #roots_of_kN_mod_p (via AllocRootsOfkN())
	 * - populates #roots_of_kN_mod_p (via CalculateRootsOfkN())
	 * - allocates space for solutions of polynomial modulo primes in factorbase (via AllocSolutions())
	 * - allocates matrices for sieving (via AllocMatrices())
	 * - if #QSMode is ESiqs, initialize
	 *   - (via SetSValue())
	 *   - (via InitiateSiqsStructures())
	 * - allocates structures for sieving polynomials (via AllocPolynomials())
	 * - sets threshold (via CheckEnough())
	 */ 
	int InitEnvironment();
        int PerformSieving();
	int ProcessRelations();
	int LinearAlgebraStep();
	int Factor();
	// METHODS FOR QSIEVE - 2nd level (being called from 1st level functions)
	// Methods of InitEnvironment
	
	/// Checks whether #N (factored number) and #M (sieving interval) have been defined. 
	/*! If no, it returns CALLED_TOO_EARLY, otherwise OK.*/
	int CheckInitConsistency();
	
	int InitFileNames();
	
	/// Checks whether factored number #N is composite number or no.
	/*! Checks whether N is a composite number or no, by calling several
	 * (MILLER_RABIN_ROUNDS_STRICT) rounds of mpz_probab_prime_p test. 
	 * Returns either OK, when N is composite, or PRIME_NUMBER, if it is
	 * prime.*/
	static int CheckPrimality(mpz_t aArg);
	/// Fills array #small_primes with small primes
	/*! Tries to read primes from file (#prime_numbers_file_name) to memory. If it does
	 * not find the file in the
	current working directory, it tries to generate it from scratch.
	Returns ERROR_INIT_ENVIRONMENT in case of any problem, or OK when
	everything went smoothly.	*/	
	int ReadPrimesFromFile();
	/// Tries to find divisor of #N among primes from #small_primes
	/*! 	
	Tries to find a nontrivial divisor of #N among elements of #small primes
	 by trial division. If it finds
	one, it returns DIVISIBLE_BY_SMALL_PRIME, otherwise it returns OK.*/
	int CheckDivisibility();
	/// Computes #k, #value_d and  #root
	/*! Calculates three useful values:
	 * 
	 - small multiplier #k, which is calculated as N mod 8, and used to alter the factorized value to allow
			inclusion of 2 into the FB.
	 -  #root - value of lower integer part of square root of kN.

	It returns only OK value.	*/
	int CalculateKRoot();
	/// Allocates stuctures for sieving polynomials
	/*!	In this method, structures for holding sieving polynomials are created. Their size is determined differently 
	in MPQS and in SIQS mode: in MPQS, space for DEFAULT_POLYNOMIALS_ALLOCATED is allocated, while in SIQS, the
	number of polynomials kept in memory is (1 << (s-1)). Therefore, this method must be called after determination
	of number 's'.
	During allocation of polynomial structures, all involved mpz_t's are inited as well.

	Return codes:
	NOT_ENOUGH_MEMORY if any allocation fails
	OK otherwise.*/
	int AllocPolynomials();
        /// sets #maximal_partial_factor and #hash_root 	
	/*! 	
	This method is a historic relict from earlier versions of MPQS/SIQS. It
	only performs the following operations: 
	- sets #maximal_partial_factor to 128*maximal element of factor base 
	- sets #hash_root to NULL.
	
	It returns only OK value.*/
	int AllocVariationsAndHashes();
	int SetFactoringStructures();
	/// defines (and allocates) the factor base array
	/*! 
	At first, one cannot determine the size of FB exactly before actually
	defining it. So, the method allocates rich space for the factor base by
	malloc-ing a relatively large space (large enough to contain ALL primes
	smaller than B). Then, it puts -1 and 2 into the FB on the first two
	indices. After that, the method searches the array of primes defined by
	ReadPrimesFromFile(); and if kN is quadratic residue modulo a prime,
	that prime is inserted into the FB.  Once the FB is generated, the
	method generates other auxiliary data structures whose size depends on
	size of the factor base, which are arrays:
	
	- #a_inverses
	- #exact_binary_logarithms_of_factor_base
	- #binary_logarithms_of_factor_base_times_log_factor
	- #is_this_prime_divisor_of_a

	Each of these allocations may fail, in which case the method returns NOT_ENOUGH_MEMORY.
	In rough logarithmic mode, not all primes are sieved with; the sieving begins at some position in the FB,
	say 100 or 200. If the factor base generated is smaller than this threshold, an error TOO_SMALL_FACTORBASE
	is returned.
	Moreover, several integer values are updated:	    
	- #factor_base_allocated
	- #maximal_factor_base_index
	- #original_maximal_fb_index (this value is needed at deallocation time in destructor, 
			since maximal_factor_base_index changes when singletons are removed)
	- #contini_threshold
	- #accepted
	- #margin	
	
	Aside from the already mentioned return values, OK is returned when everything went well.	*/
	

	int DefineFactorBase();	
	/// allocates #roots_of_kN_mod_p
	/*! allocates space for square roots of kN modulo primes of the factor base. If this allocation fails,
	NOT_ENOUGH_MEMORY is returned; otherwise OK.*/
	int AllocRootsOfkN();
	/// populates #roots_of_kN_mod_p with square roots of kN modulo elements of #factor_base
	/*! calculates square roots of kN modulo primes of the factor base (excluding -1 and 2).      
	For this purpose, a faster variant of Shanks-Tonelli, using only integer arithmetics, is used.
	The only return value is OK.*/
	int CalculateRootsOfkN();

	/*! 	In this method, three matrices needed for sieving process are allocated. The first one is the sieving matrix,
	which must contain 1 entry per each integer in the sieving interval. The other ones are "relation matrices",
	which contain so far found relations; they are always used for keeping the smooth and partial values in memory
	during the linear algebra phase. Moreover, if no intermediate files are used for keeping so far found relations,
	the relations found during sieving are kept in those matrices.
	Only the struct-s which make the matrices themselves are allocated, but no content; this is just a few bytes
	of memory, so a NOT_ENOUGH_MEMORY should not happen.  However, the possibility is still checked for.
	\todo Tenhle komentar chce zrejme updateovat, nesedi
	*/
	int AllocMatrices();
	int OpenFiles(int aMode);
	FILE* OpenFile(const string& aName, int aMode);
	/// for SIQS mode. 
	/*! 	This method is only called in SIQS sieving mode.
	In order to determine 's', it divides the set of all 'a' divisors into
	"machine specific divisors", "even-indexed divisors" and "odd-indexed
	divisors", as described by Carrier-Wagstaff (only the procedure is more
	complicate, to avoid some practical traps). If the number of
	"odd-indexed divisors" is lower than 3, the Carrier-Wagstaff method
	would not work well; the calculation is therefore aborted and return
	code UNSUITABLE_FOR_SIQS is returned.  Otherwise, OK is returned. 
	\todo vymyslet strucny komentar
	*/
	int SetSValue();
	/*! 	This method is only called in SIQS sieving mode.
	It takes care of allocation of data structures used in SIQS, namely arrays
	- ** B2, plus initialization of all its mpz_t members
	- ** B2_inited
	- ** Ba_inv
	- ** divisors_of_a
	- ** indices_of_divisors_of_a

	If any of the allocations fail, the method returns NOT_ENOUGH_MEMORY.
	Furthermore, this method takes care of the NexKSB-algorithm structures, as used by the Carrier-Wagstaff SIQS
	algorithm. It creates an auxiliary instance of NexKSB environment variable, and uses this instance to generate
	all triples of even-indexed primes (for details on even- and odd- indexed primes see the Carrier-Wagstaff article
	'Implementing the Hypercube Quadratic Sieve ....'). Then it sorts the outcoming set of triples (it should be 
	almost sorted already) by binary logarithms of their products. After that, the set of even-indexed triples is
	determined; now, the bounds of the odd-indexed interval must be determined so that the method would not generate
	many similar sets of 'a' divisors (see thesis of M.Kechlibar). This determination is performed by calling a third-
	level method; then, the NexKSB-structures are reset to serve the main purpose of generation of the odd-indexed
	n-tuples.
	Finally, if sieving with prime powers is enabled, corresponding structures Ba_inv are allocated as well.
	Return codes:
	anything thrown by NexKSB_Init();
	NOT_ENOUGH_MEMORY if any allocation fails.
	\todo vymyslet strucny komentar
	OK otherwise.*/
	int InitiateSiqsStructures();
	int TrueIndexEven(int nexksb_index);
	int TrueIndexOdd(int nexksb_index);	
	void CopyNexKSBToEvens();
	void DetermineTriplesRange();
	void Swap(void** array, int index_1, int index_2);
        void Quicksort_Triples(int** array, int left, int right);
	int  Binary_Search_Triples(int** array, int left, int right, int what);
	void AdjustLowerAndUpperOddBound();
	int  CalculateLogOfOdd();
	int  CalculateLogOfSequence(int start_index, int length, int pause);
	void PrintInformation();
	// Methods of PerformSieving
	int QuadraticMachine();
	int PerformSievingByThisPolynomial();
	
	int DeleteMainMatrix();
	// METHODS FOR QSIEVE - 3rd+ level (being called from 2nd level functions)
           // all these methods are connected with solution of quadratic equations modulo p (possibly modulo p^k)
	int  SolveQuadraticEquation();
	void ComputeAInverses();
	int  ShiftSolutionsForSiqs();
	void CreateNewMpqsPolynomial();
	void CreateNewSiqsPolynomial();
	void CreateNewAForSiqs();
	void RecalculateBForSiqs();
	void CalculateSolutionForADivisor(main_sieving_type aIndex, main_sieving_type aPrimeP, mpz_t aCoeffB, main_sieving_type aMinusC);
public:
	// returns x such that Q(x) = 0 for p | a.
	main_sieving_type  CalculateSolutionForADivisor(quad_polynomial* aPoly, main_sieving_type aPrimeP);
private:
	//int  SolverSingle(long p, long i);
	//int  SolverMulti(long p);
	void ToLeastAbsoluteValue(mpz_t rop, mpz_t op, mpz_t modulus);
	int  MaxPowerOfTwo(long p);
          // all these methods are connected with main sieving loop	
	int  RefillMainMatrix();		
	void AssertQuadPoly(mpz_t aArg, long aModulus);
	void AssertQuadPoly(long aArg, long aModulus);
	void AssertSpecialQuadPoly(long aArg, long aModulus);
	int  SievingLoopExact();
	int  SievingLoopRough();
	int  SievingLoopLogarithmic();
	int  SortSmoothValuesExact();
	int  SortSmoothValuesRough();
	int  SortSmoothValuesLogarithmic();
public:
	int  SaveCandidate(qs_relation* candidate, int& accepted, int& rejected);

	int  SaveCandidate(qs_relation* candidate, TCandidateType type);

	int  ReadRelationFromFile(FILE* aFr, qs_relation* aElement);

	/*! 	This method serves for determination whether there has been enough smooth and partial values collected. 
	It returns ENOUGH if true, NOT_ENOUGH otherwise.
	Besides this, it also determines the value of threshold, which is the amount of smooth and partial values
	suitable for running the linear algebra phase.*/
	int  CheckEnough();

	float GetSecondsInSieving();

private:
          // all these methods are connected with storage of smooth and partial values
	void PrintMainMatrix();
	void PrintRelationMatrix(relation_matrix* aSource);
	void PrintRelation(qs_relation* aElement);
	void PrintRelationToFile(FILE*& aFw, qs_relation* aElement, long aIndex);
	int  PrintRelationToMatrix(relation_matrix* aTarget, qs_relation* aElement, long aIndex);
	int  CheckNulity(qs_relation* aElement);
        // 2nd level methods of ProcessRelations
	int  RemoveLargeSingletons();
	int  AddSmoothValuesToRelationMatrix();
	int  DeleteRelationMatrix(relation_matrix* aTarget);
	int  CheckAndDeleteDuplicates();
	void DeleteRowsFromRelationMatrix(relation_matrix *aTarget, long* aListOfRowsToDelete, long aMaxIndexInList);
	int  CheckAndDeleteSingletons();
	int  RemoveUnusablePrimesFromFB(long* aListOfPrimesToDelete, long aMaxIndexInList);
	void AssertSmooths();
	void RefillByConstant(long* aArray, long aMaxIndex, long aConstant);
	void RefillByConstantMST(main_sieving_type* aArray, main_sieving_type aMaxIndex, main_sieving_type aConstant);
	int  DeleteDuplicatesFromSortedArray(long* aArray, long aMaxIndex);
        // 2nd level methods of LinearAlgebraStep
	int  ToBitMatrix();
	int  SolveSystem();
	// 3+ level methods
	int  CloseFiles();
	int  CloseFile(const string& aName);
	
	int  CombineTwoRelations(qs_relation* aTarget, qs_relation* aRow1, qs_relation* aRow2, int aType);
	int  CombineTwoRelations(qs_relation* aTarget, qs_relation* aRow1, qs_relation* aRow2);
          // all these methods cooperate with CheckAndDeleteDuplicates
	void CalculateHash(mpz_t aResult, qs_relation* aRow);
	int  AddHashToTree(mpz_t aValue, long aIndex);
	bool FindHash(int* aLast,ROW_HASH** aPred, mpz_t aValue);
	void DeleteHashes(ROW_HASH* aVar);
	void SwapTwoRows(relation_matrix* aTarget, long aInd_1, long aInd_2);
	long IsInList(long* aList, long aMaxIndex, long aValue);
public:
	int  AssertRelationConsistency(qs_relation* aElement, const char* aMessage, int aValue);
	void DetailedAssertion(qs_relation* aElement);
private:
	// 2nd level methods of Factor
	int  FindFactors();
	int  TryKthDependence(int aK);
	void PrintFactors();
	void PrintApology();
	// Methods for "Next k-Subset in a n-Set" algorithm.

	int  NEXKSB_Init(nexksb_type n, nexksb_type k);
	void NEXKSB_Shift();
	void NEXKSB_Next();	
	void NEXKSB_Delete();

	// added in version 1.04
	void mpz_add_si(mpz_t aRop, mpz_t aOp1, long aOp2);
	void mpz_sub_si(mpz_t aRop, mpz_t aOp1, long aOp2);
	int ShanksTonelliIntegral(long n, long p); // added in 1.05
	int ShanksTonelliIntegral(long n, long p, long i, long k);
	int ShanksTonelliMultiexpMpz(mpz_t solution, mpz_t n_in, mpz_t p, unsigned int exp);

	void PrintMpzToFile(FILE* aFile, int base, mpz_t aValue);
	
	// Enhanced Carrier-Wagstaff methods. Automatic MSD generator for case that MSDs are not externally specified.
	void SetMachineSpecificDivisors();
	
	// Sieving with prime powers. Added in 1.15
	/// allocates data structures needed for sieving with prime powers and sets some variables
	/*! 	This method is called if and only if sieving with prime powers is enabled. It allocates data structures 
	needed for sieving with prime powers and sets several member variables. This means that allocation failure
	may happen.
	The following arrays are allocated:
	- * #prime_powers
	-  for each element of #prime_powers, the arrays 
	  - ** prime_powers[i].powers
	  - ** prime_powers[i].roots_of_kN_mod_powers
	  - ** prime_powers[i].a_inverses
	  - ** prime_powers[i].solutions_1
	  - ** prime_powers[i].solutions_2
	  - ** prime_powers[i].usable	  
	  - In case of SIQS, also
	    ** prime_powers[i].Ba_inv
	  
	If any of the allocation fails, NOT_ENOUGH_MEMORY is returned, else OK. */
	int  DefinePrimePowers();
	/// calculates square root of kN mod all prime powers used for sieving.
	/*! This method calculates square root of kN mod all prime powers used for sieving. It is invoked
	if and only if sieving with prime powers is enabled.
	For this purpose, Shanks-Tonelli algorithm is used. If the Shanks-Tonelli fails at some prime power,
	it is considered a fatal failure and the computation ends.
	Returns: either error code of Shanks-Tonelli, or OK.*/
	int  CalculateRootsOfkNForPrimePowers();	
	int  PerformSievingWithPrimePowersLogarithmic();	
	
	// Double Large Prime Variation
	void ResetHashtable();
	void ResetHashtable_2();
	void TrialDivideByADivisors(qs_relation* aCandidate, int* aOne);
	int  FactorizeCandidate(qs_relation* aCandidate);
	int  RunPollardRho(qs_relation* aCandidate);
	int  RelationToPartialCycle(qs_relation* aRelation);
	void SortRelation(qs_relation* aRelation);
	
	void EnlargeHashtable_1();
	int  FindPrimeInHashtable_2(large_prime_type aPrime, bool* aPresent);
        void AddToCycleCountStructures(qs_relation* aRelation);
	int  CollisionResolution_2(int aHash, large_prime_type aPrime);
	int  FindAndResetRoot(int aIndex);
	int  CountRoots();

public:
	void PrintHeader();
	void PrintValues();
	nexksb_environment* NEXKSB_Get() const;
	int  NEXKSB_Set(nexksb_type n, nexksb_type k, nexksb_type h, nexksb_type m, nexksb_type* subset, nexksb_type finished);

private:
	void CountRootsAndComponentSizes();
	int  RereadRelations(detection_type* aField, int aMax);
	int  ConstructCycles();
	int  PutToConstructionEntry(cycle_construction_entry *aCycle,int aRelationIndex);
	void FindCycleSizes(cycle_construction_entry *aCycles, int aMaxIndex);
	int  AddDivisor(int* aMaxIndex,main_sieving_type* aSmallPrime,large_prime_type* aLargePrime, int aExponent);
	int  SmoothenCycle(qs_relation* aElement);
	int  GetSuitableHashtableSize(int aAmount);
	void PrintGraphInformation();
	void RemoveDuplicitePaths(cycle_construction_entry* aEntry);

	int  GetTrueIndexInFB(main_sieving_type* aP) const;
	int  RecalculateAuxiliarySiqsValues();
	int  RefillInverses_Of_A(bool aCalculate);
	
	int  ReloadState();
	int  ReloadRelations();
	int  ReadInt(int* aTarget, int aNameLength, FILE* aFile);
	int  ReadLong(long* aTarget, int aNameLength, FILE* aFile);
	int  ReadDouble(double* aTarget, int aNameLength, FILE* aFile);
	int  ReadMpz(mpz_t aTarget, int aNameLength, FILE* aFile);
	
	// Autocheck methods
	int  AddToScore();
	double  CountNewScore();
	void ResetMainMatrix();
	int  FindOptimalCheckInterval();
	void AddDistributionInfo(qs_relation* aCandidate, int aType);
	void PrintDistributionInfo();
	
	// New sieving
	int  InitAdvancedFactorBase();

	void ComputeC(quad_polynomial* aPoly);

	void AllocateTexts();

	int  ReloadSieverState();
	int  ReconstructHashtable();

	bool Check_LastSolutions();

	void FreeBa_inv();
	int  AllocateBa_inv();
	
 protected:	
	/// This variable holds the bound of the sieving interval [-M,M].
	mpz_t M;        
	/// This variable holds the integral part of the square root of kN.
  	mpz_t root;	 
	/// Maximal value of remaining. In EXACT_MODE = 1.
	//mpz_t accepted;
	/// 	This variable holds the maximal value of P accepted in single LPV. It is set as 128*B, where B is the upper bound of factor base numbers.
  	mpz_t maximal_partial_factor; 
	/*!
	This variable holds the ideal value of A in MPQS/SIQS. It is used only to calculate the corresponding logarithm, which is further used
	in determination of optimal SIQS values.
	*/
	mpz_t ideal_a_value;	// Ideal 'a' value for MPQS
	///	determines the threshold separating tested and untested values in the sieving interval. 
	/*! If a value of Q(x) gathers logarithms
	greater or equal to contini_threshold, it will be tested for smoothness, otherwise no.*/
	log_type 	contini_threshold; 
	/*!
	This variable determines the error factor taken into consideration in computation of contini_threshold. Unlike other building blocks of
	contini_threshold, this one can be directly influenced by the user by setting -checklevelX option from the command line. The greater
	is the error factor, the more values are let into the smoothness testing round.
	*/
	log_type 	error_factor; // Determined at construction time by SIEVE_ROUNDOFF_ERROR and changed by "checklevel".
	int integer_M; ///< This is just the integral value of M.

	/// This tells whether double or single LPV is used or only smooth values are gathered.
	variations_types         using_variations;
	/// This tells whether all values in [-M,M] will be tested for smoothness (-checkall parameter) or no (default), or whether the checking threshold will be automatically optimized (-checkauto).
	check_mode       checking_mode;
	factor_alg       factoring_algorithm;
	siev_i_mode      sieving_interval_mode;
	
	/// The least index in FB, where rough sieving will start from.
	long 		 rough_factorbase_index;
	/// A positive number determining how many extra (plus) smooth values will be generated.
	long 		 margin;
	/// A nonnegative number determining how many smooth values have been already found.
	long 		 so_far_smooths_found;
	/// A nonnegative number determining how many partially smooth values have been found.
  	long 		 so_far_1_partials_found;
	/// A nonnegative number determining how many 2-partially smooth values have been found.
  	long 		 so_far_2_partials_found;
	/// A nonnegative indexing variable for writing down 1- and 2-partials into file
	long             so_far_x_partials_found;
	
	/// A block of quadratic polynomials. Its size is (maximal_index_in_polynomial+1).
	quad_polynomial* polynomials;
	long             number_of_polynomials;	/// how many polynomials will be used during sieving
	long             allocated_polynomials; /// space for how many polynomials will be allocated (default 128, used in MPQS; in SIQS, 2^(s-1)
	long		 index_in_polynomial; /// from 0 to allocated_polynomials-1, cyclic
	long             current_polynomial; /// from 0 to number_of_polynomials-1, linear.

	/// The maximal number allowable into the factor base.
	long             upper_bound_on_factor_base;
	/// The factorization base. Is signed, as -1 always comes into FB
	main_sieving_type*	 factor_base;	
	/// The new factorization base. Used in the newer sieving methods.
	qs_fb_type*	 advanced_factor_base;
	/// The maximal meaningful index in the factor base field
	long 	 	 maximal_factor_base_index;
	/// the original allocation size of FB, before singletons were removed
	long             original_maximal_fb_index; 
	/// Obsolete. If -ad mode of Large Prime Variation is used, this is the maximal index in factor base enriched by the large primes from the LPV.
	long             enriched_maximal_factor_base_index;
	
	///	 determines number of entries allocated in time of factor_base allocation.
	long             factor_base_allocated;
	long             div_allocate; // number of divisors allocated at main matrix for each row
	/// added in 1.05: roots of kN mod p
	main_sieving_type*  roots_of_kN_mod_p;
	
	// added in 1.09 and used only in SIQS:
	/*!
		When calculating with logarithms, we often do not manipulate their float values directly, but use integers "close to them". It
	is too error-prone to calculate only with integer parts of logarithms; so we, at first, multiply the float logarithms by a given
	constant, called LOG_FACTOR, and take the integer part of this product. This approach is much more exact.
	So, this variable contains the ideal logarithm of 'A', multiplied by LOG_FACTOR.
	*/
	int                ideal_log_of_a_times_triples_log_factor;
	/*!
		This is the sum of logarithm of all machine-specific divisors (they are constant during the multiplication) of A (see Carrier-
	Wagstaff) multiplied by LOG_FACTOR.

	*/
	int  		   machine_divisors_log_times_triples_log_factor;
	/*!
	This is the number of divisors of 'A' in SIQS mode.
	*/	
	int		   s; // a = q_1...q_s
	/*!
	This is number of machine-specific divisors (see Carrier-Wagstaff).
	*/
	int                machine_specific_divisors_number;
	/*!
	These are the machine-specific divisors. We expect at most 3 of them.

	This array is not yet saved to the siever state xml file. 
	*/
	main_sieving_type  machine_specific_divisors[3];
	/// True if the machine specific divisors were specified externally - typical for nodes. Is set to true by the setter method which specifies the external msdivisors from the outside.
	bool		   iMSDSpecifiedExternally;

	main_sieving_type  iLowerMachineBound; 
	int		   iLowerMachineBoundIndex; /// Its index in the FB.

	main_sieving_type  iUpperMachineBound; 	    /// Prime number.
	int		   iUpperMachineBoundIndex; /// Its index in the FB.

	/*!
	These are indices of machine-specific divisors in the factor base. We expect at most 3 of them.

	This array is not yet saved to the siever state xml file. 
	*/
	main_sieving_type  machine_specific_divisor_indices[3];
	/*!
	This is the floating point version of the sum of machine specific divisors.

	This array is not yet saved to the siever state xml file. 
	*/
	float              mspecdivisors_log;
	/*!
	These are the divisors of 'A'. There are altogether 's' of them. The allocated space for this field corresponds to this fact.

	This array is saved to the siever state xml file. 
	*/
	main_sieving_type* divisors_of_a;
	/*!
	These are the indices of divisors of 'A' in the factor base. There are altogether 's' of them. The allocated space for this field 
	corresponds to this fact.

	This array is not saved to the siever state xml file, but rather recalculated
	during reload.
	*/
	main_sieving_type* indices_of_divisors_of_a;
	/*!
	These are values of B_1 to B_s, used in SIQS calculation of new b_i's. There are altogether 's' of them. The allocated space for this field 
	corresponds to this fact.	

	This array is saved to the siever state xml file. 
	*/
	mpz_t*		   B2; // 2B_1 ... 2B_s - useful in calculating SIQS's new b_i's
	/*!
	This is the 'initialization status field' for the previous field. Its entries give us information whether we should call mpz_clear on
	corresponding indices in B2 or no.

	This array is not saved to the siever state xml file, but rather recalculated
	during reload.
	*/
	int*		   B2_inited;// has the corresponding mpz been inited or no?
	/*!
	This is a two-dimensional structure for SIQS, having as many rows as the factor base has elements, and 's' entries for each row. The entries
	corresponding to current 'A' divisors are meaningless, others are used in SIQS polynomial generation.

	This array is not saved to the siever state xml file, but rather recalculated
	during reload.
	*/
	main_sieving_type** Ba_inv;   // has as many indices as factorbase, but some entries are meaningless (those corresponding to q_i's)
	/*!

	These values are saved to the siever state xml file. 
	*/
	int                nu;       ///< 	These are two auxiliary variables used in Gray code formula, which generates new b's in SIQS.
	int                subtract; ///< 	These are two auxiliary variables used in Gray code formula, which generates new b's in SIQS.
	/// Array with values 1 or 0 for each element in factor base
	/*! 	This is a field as big as the factor base is, having 1's on
	 * entries which correspond to current divisors of 'A' coefficient and
	 * 0's on others;
	it has meaning only in SIQS.

	This array is not saved to the siever state xml file, but rather recalculated
	during reload.

	Existence of this array is a speed/memory compromise, and since the array is quite
	large (up to megabytes), it could be possibly removed in need.
	*/
	main_sieving_type* is_this_prime_divisor_of_a; // used in "profi" implementation of SIQS
	/*! If 'a' is the leading coefficient of Q(x), this array contains (2a)^{-1} mod p for all p in FB except p = -1

	This array is not saved to the siever state xml file, but rather recalculated
	during reload.
	*/
	main_sieving_type* a_inverses;
	/// The integer closest to the first real root of Q(x)
	main_sieving_type  root_1; 
	/// The integer closest to the second real root of Q(x)
	main_sieving_type  root_2;
	
	// The following two attributes are only set by CheckEnough() method.
	long             threshold;    ///< Contains the minimal number of smooth/partial values, that must be collected.
	long             collected;    ///< Collected number of smooth/partial values.
	long             nulities;     ///< Number of partial combinations, whose all exponents are zero mod 2.

	/*!
	This field is as long as the factor base and contains integral parts of binary logarithms of factor base primes multiplied by LOG_FACTOR.
	*/
	log_type* 	 binary_logarithms_of_factor_base_times_log_factor;
	
	/*!
		This block of variables is used in SIQS for Carrier-Wagstaff method of 'a' generation.
	Short comments: two sets of primes are created. One of them consists of primes with odd indices in FB, where the indices start
	at odd_index_start and end at odd_index_end. The other one consists of primes with even indices in [even_index_start,even_index_end].
	Triples from the second set are selected, and (s-3-#mach_spec_divs)-tuples from the first set are found in such a way, so that 'A'
	consisting of these factors would be very close to the ideal 'A' value.

	*/
	float*           exact_binary_logarithms_of_factor_base;
	int		 triple_count;
	int**		 triples; // 2D array of integers: logarithm of product and three corresponding primes as indices from 1 to CW_EVEN_INDEX_SIZE
	int              even_index_start;
	int              even_index_center;
	int              even_index_end;
	int		 minimal_even_log;
	int		 maximal_even_log;
	int              odd_index_start;
	int              odd_index_center;
	int              odd_index_end;
	int              number_of_odd_index_divisors;
	int              odd_index_size;
	int              odds_count;
	int              ideal_divisor_of_a;	
	int              ideal_divisor_of_a_closest_value;
	int              ideal_divisor_of_a_index;
	int              checklevel;
	bool             extK;

	// Automatic saving of sieving
	time_t		 lastSaved;	

	/*!
		Root of a binary tree, which serves as detector of duplicate rows
	*/	
	ROW_HASH*        hash_root;
	/// A pointer to file containing so far found x-partials; used during sieving phase.
	FILE* 		 partials_file;
	/// A pointer to file containing so far found smooth values.
  	FILE* 		 smooth_values_file;
	/// A pointer to file containing all x-partials (after the sieving phase).
	FILE*            all_partials_file;
	/// A pointer to file containing those x-partials which are used in some cycles. These are precisely x-partials without all singletons
	FILE*            usable_partials_file;
	/// A pointer to a temporary file. Used extensively in x-partial processing (large singleton removal).
	FILE*            temporary_file;
	/// Filename.
	string            partial_values_file_name;
	/// Filename.
	string            smooth_values_file_name;
	/// Filename.
	string            all_partial_values_file_name;
	/// Filename.
	string            usable_partial_values_file_name;
	/// Filename.
	string            temporary_file_name;
	/// Name of file with small primes.
	string            prime_numbers_file_name;

	string		 iGenerateNewPolyText;
	string		 iQuadraticMachineText;
	string		 iPerformSievingByPolyText;
	string	         iSievingLoopText;
	string		 iRefillMainMatrixText;
	string	         iSortSmoothText;
	string		 iDivisorsOfAText;
	string		 iSmallFactorText;
	string		 iCountCyclesText;
	string		 iBlockSieveText;
	string		 iFillHashtableText;
	string		 iFillHashtableNegText;
	string		 iRootsUpdateText;
	string		 iFindAndFactorText;
	string		 iUnknownPartText;
	
	int              open;    ///< This int contains flags determining open/close state of files.
	bool             lanczos_time_message_mode; ///< Shall we have detailed info on Lanczos time structure?
	bool             without_primes_listing; ///< Shall we list primes in FB on screen?
	bool             density_info_mode;   ///< TRUE if density of matrices is to be calculated
	bool             value_distribution_mode;  ///< TRUE if distribution of values over the sieving interval will be calculated
	print_sort       printing_sort;    ///< shall we print out number of total tries and successes in SortSmoothValues?

	/// determines whether MPQS or SIQS is being run. 
	/*! This is set according to use of \a -s or \a -q command line options.*/
	qs_mode		 QSMode;
	/*!
		Determines whether singletons will be removed. If no, the Singleton Gap manifests.
	*/
	sgt_mode         SGTMode;
	
	/*!
		Determines whether we sieve with prime powers (e.g. 9, 25 ...).
	*/
	power_mode       power_sieving_mode; // new in 1.15
	
	
	int              solve; ///< An auxiliary true/false variable for polynomial solutions.
	nexksb_environment* nx_env; ///< Structure for Next k-subset algorithm, used in Carrier-Wagstaff SIQS
	int              time_message_interval; ///< If compiled with according symbolic constant defined, this variable determines the interval (in #'s of polynomials) in which info about percentual speed is printed out.
	int              log_base; ///< So far unused.

	Lanczos*         lanczos; ///< A pointer to an instance of Lanczos algorithm, used in linear algebra phase.

	LineSiever*	 iLineSiever; ///< A pointer to an instance of a line siever for newer MPQS/SIQS modes. The corresponding LineSiever must be compiled with QS_SIEVING_MODE symbolic constant defined (and NFS_SIEVING_MODE undefined).
	line_sieve_size  iLineSieveSize;
	bool 		 iFirstLineSieverPolynomial;
	bool		 iUseRootShort;

	sieve_matrix*    main_matrix; ///< A pointer to the sieving interval struct.
	relation_matrix*   smooth_values; ///< A pointer to a collection of smooth values, used in linear algebra phase.

	BitMatrix*       class_result_bit;	///< Auxiliary matrix used in linear algebra phase.
	BitMatrix*       class_transpose_bit; ///< Auxiliary matrix used in linear algebra phase.
	SparseMatrix*    class_result_sparse; ///< Auxiliary matrix used in linear algebra phase.
	SparseMatrix*    class_transpose_sparse; ///< Auxiliary matrix used in linear algebra phase.
	NormalMatrix*    class_normal_result_bit; ///< Auxiliary matrix used in linear algebra phase.
	NormalMatrix*    class_normal_transpose_bit; ///< Auxiliary matrix used in linear algebra phase.
	NormalMatrix*    class_normal_result_sparse;///< Auxiliary matrix used in linear algebra phase.
	NormalMatrix*    class_normal_transpose_sparse; ///< Auxiliary matrix used in linear algebra phase.

	BitMatrix*       class_dependencies; ///< Dependency matrix generated by Lanczos (or other matrix solving algorithm).

	ASupportingFactorAlg *fact_alg; ///< A pointer to a factoring algorithm used during Double LPV to factor remaining values.


	/*! auxiliary variables. They are used for intermediate operations in member functions. Their
	'globality' reduces number of mpz_init-s greatly, thus saving time (each call of mpz_t is
	in fact a memory allocation and consumes time).
	
	Before introducing these auxiliary variables, just startup of QS (init_environment) took 6 million 	allocations;
	 after, this was reduced to a thousand.
	 */
	mpz_t            aux_1;
	mpz_t            aux_2;
	mpz_t            aux_3;
	mpz_t            aux_4;
	mpz_t            aux_5;
	mpz_t            aux_a;
	mpz_t            aux_b;
	mpz_t            aux_c;
	mpz_t            aux_d;
	mpz_t            aux_p;
	mpz_t            aux_p_exp;
	mpz_t            aux_a_inv;
	mpz_t            aux_k;
	mpz_t            aux_l;
	int              aux_inited;
	// Shanks-Tonelli auxiliary variables.
	mpz_t            st_aux_a;
	mpz_t            st_aux_a_inverse;
	mpz_t            st_aux_b;
	mpz_t            st_aux_r;
	mpz_t            st_aux_root;
	mpz_t            st_aux_n;
	mpz_t            st_aux_p_minus_1;
	mpz_t            st_aux_p;
	mpz_t            st_aux_1;
	mpz_t            st_aux_b_exp_2i;
	mpz_t            st_aux_non;
	int              st_aux_inited;

	// Auxiliary variables for WithinIntervalBlock
	mpz_t            wib_diff;
	mpz_t            wib_ratio;
	mpz_t            wib_mul_p;
	mpz_t            wib_arg;
	mpz_t            wib_arg_2;
	mpz_t            wib_arg_plus;
	int              wib_inited;

	// new variables in 1.04
	/// Solutions of current polynomial modulo primes in FB, starting with first odd prime.
	/*! These are used in extent sieving.*/	
	integer_solutions int_sols_of_cur_poly;	
	
	/// Structure used for sieving with prime powers. Sieving with prime powers. Added in 1.15
	prime_power_structure*  prime_powers;
	int		allocated_prime_powers;
	int		prime_power_limit; ///< this is the upper bound for size of prime power used in prime power sieving. (for example: 10*M, where [-M,M] is the sieving interval). It is definitely lower than 2^31, so no multiple-precision arithmetic must be used.
	int		primes_with_powers; ///< number of FB elements, whose higher powers are really used in sieving
	int		powers_together;    ///< number of all prime powers really used in sieving
	
	// Double large prime variation
	mpz_t           minimal_candidate_value; ///< minimal remaining value for 2-partial, usually B^2
	mpz_t           maximal_candidate_value; ///< maximal remaining value for 2-partial, usually (128B)^(128B)

	bool		factoring_structures_inited;
	
	/// Hashtable for counting fundamental cycles.
	lp_hashtable_common<CCommonHashEntry>*  cycle_table_1;
	
	/// Hashtable for building fundamental cycles
	lp_hashtable_common<CRelationHashEntry>* cycle_table_2;
	//hashtable_entry_type_2*  cycle_table_2;
	// int                    table_2_maximal_index;	
	// unsigned int           mask_2;
	
	/// Array serving for finding roots
	unsigned int*         intermediate_roots;
        int                    max_ir_index;

	
	int 		       components; ///< Parameter of current graph in Double LPV
	int 		       edges; ///< Parameter of current graph in Double LPV
	int                    vertices; ///< Parameter of current graph in Double LPV
	int                    cycles; ///< Parameter of current graph in Double LPV
	/// Total number of independent cycles
	int                    independent_cycles;        
	
	/// List of roots of components found
	hashtable_entry_type_1* root_list;
	int		       rl_allocated;
	int			rl_max_index;

	/// Structure for combining relations
	intermediate_array_element*  intermediate_array;
	int			ia_allocated;

	/// Structure for automatic contini threshold adjustment
	checkauto_struct        autocheck;
	score_info              checkinfo;
	int			kN_decimal_length;
	
	/// Structure for flexible sieving interval
	checkauto_struct        flexicheck;
	
	
	/// This string is used for reading and writing mpzs from/to files. That is because under win32 the standard IO provided by GMP causes crashes.
	char*   auxCharForMpz;
	
	// check structures
	double total_sieve;
	double total_refill;
	double total_inner;
	double total_solv;
	double total_sort;
	double total_divisors_of_a;
	double total_double_factors;
	double total_cyclecount;
	double total_poly;
	double total_psbtp;
	double total_perform;
	int    total_short_divisions;
	int    total_hits;
	int    total_divrem;
	int    total_tested;
	int    accepted_nr;
	int    rejected_nr;
	
	/// These auxiliary variables are used to determine smooth value distribution (if specified by \a -valdist )
	int    smooths_distribution[MAX_DISTANCE+1];
	int    part_1_distribution[MAX_DISTANCE+1];
	int    part_2_distribution[MAX_DISTANCE+1];

	long   iStartSievingClock;
	long   iFinishSievingClock;

	char*  iDirectoryName;

	int    iLoadFlags;
	
	bool	iIsInterruptable;
	bool	iInterruptNow;
	bool iComputationFinished;

	relproc_phase 	iRelationProcessingPhase;
	int	iRelationProcessingFlags;

	/// Auxiliary variables for distributed sieving
	distribution	iDistributionMachineType;
	int		iCommitInterval;	
	bool		iCenterRunInterrupted;

	string		iMachineName;
	string		iSharedURL;

	/// Already created jobs for nodes.
	mutable 	vector<QSParameters>	iJobsForNodes;
	bool		iNodeInstanceRunning;

	/// The mutex for distributed sieving.
	pthread_mutex_t iDistributedMutex;
	bool		iDistributedMutexInited;

	unsigned int	iRelationsSinceLastDataTransmission;
	unsigned int	iRelationsSent;

	/// Init environment already run.
	bool		iInitEnvironmentAlreadyRun;

	int			iOptimizationPolicy;
	ThresholdOptimizer*	iOptimizer;

	machine_specific_generation_type iMachineSpecificGeneration;
};

#endif

