#ifndef _TYPES_H_
#define _TYPES_H_

#include <gmp.h>
#include "definitions.h"
#include "types_common.h"



/*!
  \file 
  \brief Aggregated declarations of types used throughout the MPQS/SIQS source.
  
  This file aggregates (both primitive and composite) types used in MPQS/SIQS. 

*/

typedef 	 int  isol_type; 		///< Used for storage of integer solutions
typedef          int  main_sieving_type; 	///< Used for representation of prime numbers
typedef unsigned char log_type;          	///< Used for representation of logarithms.
typedef 	 int  nexksb_type;

enum relproc_phase { EBeforeRelationCollection = 0,
	 ECollectingRelations,
	 ERemovingLargeSingletons,
	 EConstructingCycles,
	 EDeletingDuplicates,
	 ERemovingSmallSingletons,
	 ERelationsReady };
enum qs_mode     { EMpqs = 0,
	 ESiqs,
	 EOldMpqs,
	 EOldSiqs };
enum sgt_mode    { EDeleteSingletons = 0,
	 EDoNotDeleteSingletons  };
enum power_mode  { EDoNotSieveWithPrimePowers = 0,
	 ESieveWithPrimePowers };
enum factor_alg  { EPollardRho = 0,
	 ECFRAC,
	 EPMinus1,
	 EEllipticCurve };
enum siev_i_mode { EFixed = 0,
	 EFlexible };
enum autostates  { EBeforeStart = 0,
	 EStartUp,
	 EStartDown,
	 EDecide,
	 EGoUp,
	 EGoDown,
	 EFinished };
enum test_type   { ETestUndefined = 0,
	 ETestMemblock,
	 ETestCheckLevel,
	 ETestIntegralFBSize,
	 ETestAlgebraicFBSize,
	 ETestSievingIntervalSize };

/// Allowed overlapping of machine specific divisors
enum machine_specific_generation_type { 
	MSGNoRestrict = 0, ///< Any overlapping possible - will generate just one set for all of the machines
	MSGOneDiffer, ///< Machine specific divisors sets will differ in at least one %divisor
	MSGTwoDiffer, ///< Machine specific divisors sets will differ in at least two divisors (default)
	MSGThreeDiffer///< Machine specific divisors sets will differ in at least three divisors
};

enum qs_serialize_action { EQsJob = 0, EQsFactorBase, EQsSieving, EQsSieverState, EQsResult };

enum data_file_type { ESmooth, EPartial };

typedef long log_walking_type; // Used for going through the block of logs.

/*!
 * \brief encloses data used for the "Next k-subset in a n-set" algorithm.
 */
typedef struct {
	nexksb_type  n;
	nexksb_type  k;
	nexksb_type  h;
	nexksb_type  m;
	nexksb_type* subset;
	nexksb_type  finished;
} nexksb_environment;

/*!
 * \brief  represents a quadratic polynomial
 */
typedef struct { 
	mpz_t a; ///< Quadratic coefficient
	mpz_t b; ///< Linear coefficient
	mpz_t c; ///< Absolute coefficient
	mpz_t d; ///< Square root of the quadratic coefficient
	mpz_t d2; 
	int inited;
} quad_polynomial;

/*!
 * \brief represents solutions of a quadratic polynomial mod p, p^2, p^3 ... where p is a prime
 */
typedef struct {
        isol_type* usable; ///< Field of flags telling whether a solution is usable (consistent) or no.
	isol_type* sol_1; 
	isol_type* sol_2;
	isol_type  allocated;
} integer_solutions;

/*!
 * \brief is the base type for the "first" hashtable used in both large prime variations. 
 *
 * Intent of this hashtable is to give exact information
	about number of collected fundamental cycles throughout the runtime. As this hashtable needs to be large (to hold many large primes), its 
	elements must be as small as possible.

*/
typedef struct {
	large_prime_type value;	
	large_prime_type ancestor;
} hashtable_entry_type_1;

/*!
 * \brief is the base type for the "second" hashtable used in both large prime variations during the cycle construction phase. 
 *
 * As this
	hashtable needs only to contain information about large primes taking part in at least one of the fundamental cycles, its elements can be
	larger; however, only one extra field is needed. 
	The most significant bit of the relation field is used to denote the iteration parity (see Lenstra, Manasse: Factoring with Two Large Primes).

*/
typedef struct {
	large_prime_type value;	
	unsigned int ancestor;
	unsigned int relation;
} hashtable_entry_type_2;
/*! 
 * \brief helps in keeping track of rows encountered in the "duplicate deletion phase" of the MPQS/SIQS, large prime variation.
 *
	All the rows which had been chosen to the sieving matrix are iterated through, and their hashes are put into a binary tree sorted 
	by absolute value of the hash_value member. If a row with already existing hash is met, this row is marked as a duplicate row and
	it is not allowed to proceed further.
	Duplicates are very sparse, with one exception: if two SIQS polynomial series differ only by a single different factor in A, duplicates
	happen often. This can happen if the Carrier-Wagstaff method of dividing the potential factors into odd-indexed and even-indexed fails.
 */
typedef struct row_hash {
	mpz_t hash_value;   /// Usually, we calculate hash as a very long number.
	int hash_inited;    /// Flag whether mpz_init(hash_value) has been done.
	int twice;          /// 0 if only once, N-1 if N-times.
	long first_row_index;
	struct row_hash *left;
	struct row_hash *right;
} ROW_HASH;

/*!
 * \brief used when combining two relations (partial or smooth) into one, mainly during the cycle construction in ProcessRelations.

*/
typedef struct {
	main_sieving_type    p;
	large_prime_type large_prime;
	unsigned int    exponent;
} intermediate_array_element;

/*! 
 * \brief used to contain data related to sieving with large prime powers. 
 *
 * Each of the primes, whose powers are
	to be used in sieving, has a variable of this type.
	
	The members "prime" and "index_in_factor_base" are self-explanatory. In array "powers", the powers of the prime are stored; the maximal
	index can be determined from the "allocated_powers" member. The "roots_of_kN_mod_powers" member array is self-explanatory; the "a_inverses"
	array contains inverses of 2A of current polynomial mod powers; the "Ba_inv" member array is used in SIQS polynomial changes.
	
	The arrays "solutions_1" and "solutions_2" contain solutions of Q(x) modulo "powers", with the indices matching. The "usable" array determines
	whether these solutions are valid.
	
	To explain the meaning of the last members, one must realize the following problem: let us say that we want to sieve with prime powers up
	to 10000. But we do not want to sieve with prime powers lesser than, say, 100, since this would take long. So, in case of small primes like 3,
	not all powers from the series 3,9,27,81,243 ... will be used; the first ones are too small to sieve with them. Also, when we start sieving
	with a larger power like 243, we must use logarithm value corresponding not to 3, but to this precise power.
	
	Now we see the reason why use the member variables. The least_index_to_sieve_with tells us with which power we should start sieving. The
	*_least_logarithm_to_use are equal to the logarithm of this "start power", which would be, in the example before, equal to log(243). The
	*_standard_logarithm_to_use express the logs used in further steps (further powers) and are equal to log p.

 */
typedef struct {
	main_sieving_type  prime;
	int 		   index_in_factor_base;
	main_sieving_type* powers;
	int		   allocated_powers;
	main_sieving_type* roots_of_kN_mod_powers;
	main_sieving_type* a_inverses;
	main_sieving_type** Ba_inv;
	isol_type*	   solutions_1;
	isol_type*	   solutions_2;
	int*		   usable;
	int                least_index_to_sieve_with; //  in some cases, the zero-index power is too small to be sieved with (3^2 etc)
	double             exact_least_logarithm_to_use; // in such case, the logarithm used in first round of sieving 
	double		   exact_standard_logarithm_to_use; // = equal to log p; 
	log_type	   binary_least_logarithm_to_use;
	log_type	   binary_standard_logarithm_to_use;		
} prime_power_structure;

typedef enum { EAutodetectSmoothness, ESmoothCandidate, EPartialCandidate } TCandidateType;


/*!
 * \brief contains information about the sieving interval. 
 *
 * The \a accumulated_logarithm array has one entry of log_type per each
	sieving interval entry, and contains logarithms collected during sieving.
	The \a maximal_index member contains the maximal index in \a accumulated_logarithm array with a meaningful entry.
	The \a assigned member contains the amount of allocated log_type's in array \a accumulated_logarithm. It is rather duplicate with \a maximal_index,	since in normal conditions \a assigned should be equal to \a maximal_index+1.
	The \a lower_arg and \a upper_arg contain the least and the greatest X in the sieving interval. So, the entry \a accumulated_logarithm[0] should
	correspond to \a lower_arg and the entry \a accumulated_logarithm[maximal_index] should correspond to \a upper_arg.

*/
typedef struct {
	log_type* accumulated_logarithm;
	long maximal_index; ///< This is maximal index with real elements
 	long assigned;
  	main_sieving_type lower_arg; ///< The lowest and the largest x
  	main_sieving_type upper_arg;
} sieve_matrix;

/*!
 * \brief This structure is used to contain the found relations. 

The array "matrix" contains the relations, while the "long" members determine
	the size of the array. In normal conditions, assigned_rows == maximal_row_index+1, so one of the entries is redundant.

*/
typedef struct {
	qs_relation* matrix;
	long maximal_row_index; ///< This is maximal index with real elements
 	long assigned_rows;
} relation_matrix;

/*!
 * \brief  autochecking (automatic search for the optimal contini threshold value and sieving interval length)
 */
typedef struct {
	autostates state;
	double     score_start;
	double     score_up;
	double     score_down;
	double     score_last;
	double     score_now;	
	int        collected_cycles;
	int        next_check_moment;
	//directions direction;
} checkauto_struct;

/*!
 * \brief Information on score
 */
typedef struct {
	int        partials_1[CHECKAUTO_MAXINDEX+1];
	int        partials_2[CHECKAUTO_MAXINDEX+1];
	int        index;
} score_info;

/*!
 * \brief new approach to line sieving
 */
typedef struct {
	main_sieving_type   p; /// The prime value
	main_sieving_type   root_1; /// The first offset s.t. Q(x) = 0. Note that root_1 != x.
	main_sieving_type   next_1;
	main_sieving_type   root_2; /// The second offset s.t. Q(x) = 0. Note that root_2 != x.
	main_sieving_type   next_2;
	main_sieving_type   sqrt_kn;
	main_sieving_type*  ba_inv;
	log_type	    log_p;
} qs_fb_type;
#endif



