#include "xml_qs_serializator.h"
#include "xml_qs_definitions.h"
#include "definitions.h"
#include "quadratic_sieve.h"
#include <boost/filesystem/path.hpp>
#include "constants.h"
#include "log.h"

// Enumeration enum qs_serialize_action is defined there.
#include "types.h"

#include <typeinfo>   // For std::bad_cast
#include "boost_version.h"

XmlQsSerializator::XmlQsSerializator() : AbstractSerializator::AbstractSerializator(NULL,NULL,0)
{
	this->iQsInstance = NULL;
}

XmlQsSerializator::XmlQsSerializator(const char* aDirectoryName, int aCompressionLevel) : AbstractSerializator::AbstractSerializator(NULL,aDirectoryName,aCompressionLevel)
{
	this->iQsInstance = NULL;
}

XmlQsSerializator::XmlQsSerializator(AbstractSieve* aSieveInstance, const char* aDirectoryName, int aCompressionLevel) : AbstractSerializator::AbstractSerializator(aSieveInstance,aDirectoryName,aCompressionLevel)
{
	this->iQsInstance = (QuadraticSieve*) aSieveInstance;
}

XmlQsSerializator::~XmlQsSerializator()
{

}

int XmlQsSerializator::SerializeJob(xmlTextWriterPtr& aWriter, const JobParameters& aParameters)
{
 	try
  	{
		const QSParameters& qp = dynamic_cast<const QSParameters&>(aParameters);
		return SerializeJob(aWriter, qp);
	}
	catch (const std::bad_cast & e)
  	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","SerializeJob (Parameters)",0,0,"Unknown subtype of JobParameters, doing nothing");
		return GENERAL_ERROR;
  	}
}

void 	XmlQsSerializator::RegisterInstance(Algorithm*	aAlgorithm)
{
 	try
  	{
		QuadraticSieve* qs = dynamic_cast<QuadraticSieve*>(aAlgorithm);
		iQsInstance = qs;
		AbstractSerializator::RegisterInstance(qs);
	}
	catch (const std::bad_cast & e)
  	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","RegisterInstance",0,0,"Unknown subtype of aAlgorithm");
		exit(1);
  	}

}

void	XmlQsSerializator::RegisterInstance(AbstractSieve* aSieveInstance)
{
	AbstractSerializator::RegisterInstance(aSieveInstance);
 	try
  	{
		QuadraticSieve* qs = dynamic_cast<QuadraticSieve*>(aSieveInstance);
		iQsInstance = qs;
	}
	catch (const std::bad_cast & e)
  	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","RegisterInstance",0,0,"Unknown subtype of AbstractSieve");
		exit(1);
  	}
}

Algorithm* XmlQsSerializator::CreateInstance(const JobParameters& aParameters) const
{
	try
  	{
		const QSParameters& qp = dynamic_cast<const QSParameters&>(aParameters);
		return CreateInstance(qp);
	} catch (const std::bad_cast & e)
  	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","CreateInstance (Parameters)",0,0,"Unknown subtype of JobParameters, doing nothing");
		return NULL;
  	}
}

Algorithm* XmlQsSerializator::CreateInstance(const QSParameters& aParameters) const
{
	QuadraticSieve* qs_instance = new QuadraticSieve();
	if (qs_instance->SetupParameters(aParameters) == ConstRC::Ok)
	{
		return qs_instance;
	} else
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","CreateInstance (QSParameters)",0,0,"Could not correctly setup parameters for a QS Instance.");
		delete qs_instance;
		return NULL;
	}
}


int XmlQsSerializator::DeserializeJob(xmlTextReaderPtr& aReader, JobParameters& aParameters)
{
 	try
  	{
		QSParameters& qp = dynamic_cast<QSParameters&>(aParameters);
		return DeserializeJob(aReader, qp);
	}
	catch (const std::bad_cast & e)
  	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","DeserializeJob (Parameters)",0,0,"Unknown subtype of JobParameters, doing nothing");
		return GENERAL_ERROR;
  	}
}

int XmlQsSerializator::DeserializeJob(const char* aPath, JobParameters& aParameters)
{
	if (!aPath)
	{
		return GENERAL_ERROR;
	}
	xmlTextReaderPtr ptr = xmlNewTextReaderFilename(aPath);
	int rc = DeserializeJob(ptr, aParameters);
	xmlFreeTextReader(ptr);
	return rc;
}

int XmlQsSerializator::DeserializeData(const char* aPath)
{
	if (!aPath)
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize 		data",0,0,"Path is null, aborting.");
		return GENERAL_ERROR;
	}
	if (!iQsInstance)
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize data",0,0,"QS Instance is null, aborting.");
		return GENERAL_ERROR;
	}
	//printf("XMLQSSerializator::DeserializeData: path %s\n",aPath);
	xmlTextReaderPtr ptr = xmlNewTextReaderFilename(aPath);
	xmlChar* name = xmlTextReaderName(ptr);
	int type = xmlTextReaderNodeType(ptr);
	int rc = 0;
	bool freshSmooth = FALSE;
	bool freshPartial = FALSE;
	boost::filesystem::path message_path(aPath);
#if BOOST_MINOR <= 45
	string smoothName(message_path.leaf());
	string partName(message_path.leaf());
#else
	string smoothName(message_path.filename().string());
	string partName(message_path.filename().string());
#endif
	smoothName += ".smooth";
	partName += ".partial";
	while (TRUE) {
		rc = xmlTextReaderRead(ptr);
		if (rc == 0)
		{
        		break;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Data",rc,0,"Could not deserialize Data.");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
         	type = xmlTextReaderNodeType(ptr);
 		name = xmlTextReaderName(ptr);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST SMOOTH_NAME))
			{
				if (!xmlTextReaderIsEmptyElement(ptr))
				{
					freshSmooth = TRUE;
				}

			}
			if (xmlStrEqual(name,BAD_CAST PARTIAL_NAME))
			{
				if (!xmlTextReaderIsEmptyElement(ptr))
				{
					freshPartial = TRUE;
				}
			}
		}
		if (type == 3)
		{
			// text
			if (freshSmooth)
			{
				xmlNodePtr	ptr1 = xmlTextReaderExpand(ptr);
				xmlChar*  content  = ptr1->content;
				int       contLen  = xmlStrlen(content);
				FILE* f = fopen(smoothName.c_str(),"wb");
				fwrite(content, 1, contLen, f );
				fflush(f);
				fclose(f);
			}
			if (freshPartial)
			{
				xmlNodePtr	ptr1 = xmlTextReaderExpand(ptr);
				xmlChar*  content  = ptr1->content;
				int       contLen  = xmlStrlen(content);
				FILE* f = fopen(partName.c_str(),"wb");
				fwrite(content, 1, contLen, f );
				fflush(f);
				fclose(f);
			}
		}
		if (type == 15)
		{
			if (xmlStrEqual(name,BAD_CAST SMOOTH_NAME))
			{
				freshSmooth = FALSE;
			}
			if (xmlStrEqual(name,BAD_CAST PARTIAL_NAME))
			{
				freshSmooth = FALSE;
			}
		}
	}
	xmlFreeTextReader(ptr);
	remove(aPath); // remove serialized file



	qs_relation* candidate = iQsInstance->InitRelation();
	FILE* f1 = NULL;
	int accepted = 0;
	int rejected = 0;

	// Smooths
	f1 = fopen(smoothName.c_str(),"r");
	// Note that the smooths file need not exist (message might have contained 0 smooths)
	if (f1)
	{
		while (iQsInstance->ReadRelationFromFile(f1,candidate) == ConstRC::Ok)
		{
			iQsInstance->SaveCandidate(candidate, ESmoothCandidate);
		}
		fclose(f1);
		remove(smoothName.c_str());

	}
	iQsInstance->ClearRelation(candidate);


	candidate = iQsInstance->InitRelation();
	// Partials
	f1 = fopen(partName.c_str(),"r");
	// Note that the partials file need not exist (message might have contained 0 partials)
	if (f1)
	{
		while (iQsInstance->ReadRelationFromFile(f1,candidate) == ConstRC::Ok)
		{
			int ret = iQsInstance->AssertRelationConsistency(candidate,"Received relation",512);

			iQsInstance->SaveCandidate(candidate, EPartialCandidate);
		}
		fclose(f1);
		remove(partName.c_str());
	}
	iQsInstance->ClearRelation(candidate);

	candidate = NULL;
	//printf("Warning: we do not call AssertRelationConsistency, as it depends on sequence number of a polynomial.\n");
	printf("XML QS Serializator: accepted %5d, rejected %5d.\n",accepted, rejected);
	iQsInstance->PrintHeader();
	iQsInstance->PrintValues();
	printf("\n");
	return ConstRC::Ok;
}

int XmlQsSerializator::Serialize(int aAction)
{
	if (!iQsInstance)
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize",aAction,0,"QS Instance is null, aborting.");
		return GENERAL_ERROR;
	}
	this->iCompressionLevel = iQsInstance->GetDataCompressionLevel();
	//printf("Serialize: iCompressionLevel is %d.\n",this->iCompressionLevel);
	switch (aAction)
	{
		case EQsJob:
			return SerializeJob();
		case EQsFactorBase:
			return SerializeFactorBase();
		case EQsSieving:
			return SerializeSieving();
		case EQsSieverState:
			return SerializeSieverState();
		case EQsResult:
			return SerializeResult();
		default:
			CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize",aAction,0,"Unknown serialization request, doing nothing");
			break;
	}
	return ConstRC::Ok;
}

int XmlQsSerializator::Deserialize(int aAction)
{
	if (!iQsInstance)
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize",aAction,0,"QS Instance is null, aborting.");
		return GENERAL_ERROR;
	}
	switch (aAction)
	{
		case EQsJob:
			return DeserializeJob();
		case EQsFactorBase:
			return DeserializeFactorBase();
		case EQsSieving:
			return DeserializeSieving();
		case EQsSieverState:
			return DeserializeSieverState();
		default:
			CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize",aAction,0,"Unknown deserialization request, doing nothing");
			break;
	}
	return ConstRC::Ok;
}

int XmlQsSerializator::ReadSerializatorData(xmlTextReaderPtr& aReader, const char* aMessage)
{
	if (!xmlTextReaderHasAttributes(aReader))
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Read Serializator data",0,0,"No attributes of serializator, aborting");
		return GENERAL_ERROR;
	}
	xmlChar* ser = xmlTextReaderGetAttribute(aReader, BAD_CAST NAME);
	if (!ser)
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Read Serializator data",0,0,"No serializator specified, aborting");
		return GENERAL_ERROR;
	}
	if (!xmlStrEqual(ser,BAD_CAST SERIALIZATOR_VAL))
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Read Serializator data",0,0,"Wrong serializator type");
		return GENERAL_ERROR;
	}
	xmlChar* version = xmlTextReaderGetAttribute(aReader, BAD_CAST VERSION_NAME);
	if (!version)
	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","Read Serializator data",0,0,"No version of serializator specified, aborting");
		return GENERAL_ERROR;
	}
	double ver;
	// CommonXMLHelper::ParseDecimalDouble is in xml_helper.h
	int rc = CommonXMLHelper::ParseDecimalDouble(version,ver);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}
	if (iInfoMode)
	{
		printf("%s: Data serialized by %s, version %1.1f.\n",aMessage,ser,ver);
	}
	if (ver > 1.0)
	{
		printf("Warning: version of the XML data source newer than can be reliably processed.\n");
	}
	return ConstRC::Ok;
}

int XmlQsSerializator::SerializeResult() {
	int rc;
	xmlTextWriterPtr writer = xmlNewTextWriterFilename("result.xml", iCompressionLevel);
	if (writer == NULL) {
    	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Result",0,0,"XML Text Writer is null, aborting.");
	return GENERAL_ERROR;
  	}
rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
	if (rc < 0) {
        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Result",rc,0,"Could not write the document start.");
	return GENERAL_ERROR;
	}
rc = xmlTextWriterStartElement(writer, BAD_CAST RESULT_DATA);
	if (rc < 0) {
        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Result",rc,0,"Could not write the serializator tag.");
    	return GENERAL_ERROR;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(writer,SERIALIZATOR_VAL, SERIALIZATOR_VERSION);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeTime(writer, START_TIME, iQsInstance->GetStartUnixTime());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeTime(writer, STOP_TIME, time(NULL));
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_t auxMpz;
	mpz_init(auxMpz);
	iQsInstance->Get_N(auxMpz);
	rc = CommonXMLHelper::SerializeMpz(writer,MODULUS,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	iQsInstance->Get_Factor1(auxMpz);
	rc = CommonXMLHelper::SerializeMpz(writer,FACTOR,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	iQsInstance->Get_Factor2(auxMpz);
	rc = CommonXMLHelper::SerializeMpz(writer,FACTOR,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = xmlTextWriterEndDocument(writer);
	if (rc < 0) {
		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
	}
	xmlFreeTextWriter(writer);
	mpz_clear(auxMpz);
	return ConstRC::Ok;
}



// see http://xmlsoft.org/examples/testWriter.c
int XmlQsSerializator::SerializeJob()
{
	Remove(JOB_FILE);
	// JOB_FILE defined in xml_definitions.h
	// Auto ... extension depends on current compression level
	CreateFileNameAuto(JOB_FILE);
 	int rc;
    	xmlTextWriterPtr writer;
	/* Create a new XmlWriter for uri */
    	writer = xmlNewTextWriterFilename(iAuxName.c_str(), iCompressionLevel);
    	if (writer == NULL) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",0,0,"XML Text Writer is null, aborting.");
		return GENERAL_ERROR;
      	}
	rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not write the document start.");
		return GENERAL_ERROR;
    	}
	rc = xmlTextWriterStartElement(writer, BAD_CAST JOB_DATA);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not write the serializator tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(writer,SERIALIZATOR_VAL, SERIALIZATOR_VERSION);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeTime(writer);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_t auxMpz;
	mpz_init(auxMpz);
	iQsInstance->Get_N(auxMpz);
	rc = CommonXMLHelper::SerializeMpz(writer,MODULUS,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,UPPER_FB_BOUND,iQsInstance->Get_UpperFBBound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	iQsInstance->Get_M(auxMpz);
	rc = CommonXMLHelper::SerializeMpz(writer,SIEVING_INTERVAL,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedInt(writer,QS_MODE,iQsInstance->Get_QSMode());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,MEMBLOCK_SIZE,iQsInstance->Get_MemblockSize());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,VARIATIONS_NAME,iQsInstance->Get_UsingVariations());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,INFO_MODE_NAME,iQsInstance->GetInfoMode());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,DEBUG_MODE_NAME,iQsInstance->GetDebugMode());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,DISTRIBUTED,(unsigned int) iQsInstance->Get_IsDistributed());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	if (iQsInstance->Get_IsDistributed())
	{
		distribution val = iQsInstance->Get_DistributionMachineType();
		const char* text = NULL;
		switch (val)
		{
			case ECenter: text = CENTER;
				      break;
			case ENode:   text = NODE;
				      break;
			default:      CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Invalid distribution machine type.");
					return GENERAL_ERROR;
		}
		rc = CommonXMLHelper::SerializeString(writer,DISTRIBUTION_MACHINE_TYPE, text);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeUnsignedInt(writer,COMMIT_INTERVAL,(unsigned int) iQsInstance->Get_CommitInterval());
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		if (val == ENode)
		{
			// Save node specific information.
			rc = SerializeNodeInfo(writer);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}
	rc = CommonXMLHelper::SerializeSignedInt(writer,THRESHOLD_OPTIMIZATION_POLICY_NAME, iQsInstance->GetOptimizationPolicy());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedInt(writer,MACHINE_SPECIFIC_GENERATION, iQsInstance->Get_MachineSpecificGeneration());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(writer,FRONTEND_PIPE,iQsInstance->GetFrontendPipe().c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	/*
	These values are typically not known at time of job serialization
	rc = SerializeSignedInt(writer,SMALL_MULTIPLIER,iQsInstance->Get_k());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	iQsInstance->Get_kN(auxMpz);
	rc = SerializeMpz(writer,MULTIPLIED_MODULUS,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}*/
    	rc = xmlTextWriterEndDocument(writer);
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}
	xmlFreeTextWriter(writer);
	mpz_clear(auxMpz);
	printf("SerializeJob finished.\n");
	return ConstRC::Ok;
}

int XmlQsSerializator::SerializeJob(xmlTextWriterPtr& aWriter, const QSParameters& aParameters)
{
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST JOB_DATA);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not write the serializator tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(aWriter,SERIALIZATOR_VAL, SERIALIZATOR_VERSION);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeTime(aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_t auxMpz;
	mpz_init(auxMpz);
	aParameters.Get_Modulus(auxMpz);
	rc = CommonXMLHelper::SerializeMpz(aWriter,MODULUS,auxMpz);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_clear(auxMpz);
	rc = CommonXMLHelper::SerializeSignedLong(aWriter,UPPER_FB_BOUND,aParameters.Get_UpperIntBound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_t siLen;
	mpz_init_set_ui(siLen,aParameters.Get_SievingIntervalLength());
	rc = CommonXMLHelper::SerializeMpz(aWriter,SIEVING_INTERVAL,siLen);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_clear(siLen);
	rc = CommonXMLHelper::SerializeSignedInt(aWriter,QS_MODE,aParameters.Get_QSMode());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,MEMBLOCK_SIZE,aParameters.Get_MemblockSize());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,INFO_MODE_NAME,aParameters.Is_InfoMode());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,DEBUG_MODE_NAME,aParameters.Is_DebugMode());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,VARIATIONS_NAME,aParameters.Get_UsingVariations());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,THRESHOLD_OPTIMIZATION_POLICY_NAME, aParameters.Get_ThresholdOptimizer_Level());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,MACHINE_SPECIFIC_GENERATION, aParameters.Get_MachineSpecificGeneration());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter,FRONTEND_PIPE,aParameters.Get_FrontendPipe().c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	if (aParameters.Get_DistributionMachineType())
	{
		distribution val = aParameters.Get_DistributionMachineType();
		const char* text = NULL;
		switch (val)
		{
			case ECenter: text = CENTER;
				      break;
			case ENode:   text = NODE;
				      break;
			default:      CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Invalid distribution machine type.");
					return GENERAL_ERROR;
		}
		rc = CommonXMLHelper::SerializeString(aWriter,DISTRIBUTION_MACHINE_TYPE, text);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,COMMIT_INTERVAL,(unsigned int) aParameters.Get_CommitInterval());
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		if (val == ENode)
		{
			// Save node specific information.
			rc = SerializeNodeInfo(aWriter, &aParameters);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}
	rc = xmlTextWriterEndElement(aWriter);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not write the node specific info tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return ConstRC::Ok;
	}
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Job",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}
	return ConstRC::Ok;
}



int	XmlQsSerializator::SerializeNodeInfo(xmlTextWriterPtr& aWriter, const QSParameters* aParameters)
{
// 	printf("XML QS Serializator: QS instance = %p, Parameters = %p\n",iQsInstance, aParameters);

	if (!iQsInstance && !aParameters)
	{
  		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Node Info",0,0,"Nothing to serialize!");
        	return GENERAL_ERROR;
	}
/*	if (aParameters)
	{
		printf("XML QS Serializator: Serialize node info.\n");
		aParameters->Print();
	}*/
	int rc =  xmlTextWriterStartElement(aWriter, BAD_CAST NODE_SPECIFIC_INFORMATION_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Node Info",rc,0,"Could not write the node specific info tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return ConstRC::Ok;
	}
	rc =  xmlTextWriterStartElement(aWriter, BAD_CAST MACHINE_SPECIFIC_DIVISORS_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Node Info",rc,0,"Could not write the node specific info tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return ConstRC::Ok;
	}
	// Count of MSDs
	unsigned int count = 0;
	if (aParameters)
	{
		count = aParameters->Machine_Specific_Divisors_Count();

	}  else if (iQsInstance)
	{
		count = iQsInstance->Get_MSDCount();
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,COUNT_NAME,count);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	int i;
	int size = count*sizeof(main_sieving_type);
	main_sieving_type* msdField = NULL;
	msdField = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &msdField, size);

 	if (aParameters)
	{
		for (i=0; i< count; i++)
		{
			msdField[i] = aParameters->Get_MSD(i);
		}
	}
	else if (iQsInstance)
	{
		memcpy(msdField, iQsInstance->Get_MachineSpecificDivisors(), size);
	}

	for (i=0; i< count; i++)
	{
		rc = CommonXMLHelper::SerializeSignedLong(aWriter,QS_FB_ELEMENT,(signed long) msdField[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	// End of MACHINE_SPECIFIC_DIVISORS_NAME
	rc = xmlTextWriterEndElement(aWriter);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Node Info",rc,0,"Could not write the node specific info tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return ConstRC::Ok;
	}
	free(msdField);
	msdField = NULL;

	int lowerBound = 0;
	if (aParameters)
	{
		lowerBound = aParameters->MSD_Lower_Bound();
	} else if (iQsInstance)
	{
		lowerBound = iQsInstance->Get_LowerMachineBound();
	}
	rc = CommonXMLHelper::SerializeSignedLong(aWriter,MACHINE_LOWER_BOUND_NAME,(signed long) lowerBound);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	int upperBound = 0;
	if (aParameters)
	{
		upperBound = aParameters->MSD_Upper_Bound();
	}
	else if (iQsInstance)
	{
		upperBound = iQsInstance->Get_UpperMachineBound();
	}
	rc = CommonXMLHelper::SerializeSignedLong(aWriter,MACHINE_UPPER_BOUND_NAME,(signed long) upperBound);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	// End of NODE_SPECIFIC_INFORMATION_NAME
	rc = xmlTextWriterEndElement(aWriter);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Node Info",rc,0,"Could not write the node specific info tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return ConstRC::Ok;
	}
	return ConstRC::Ok;
}

int	XmlQsSerializator::DeserializeNodeInfo(xmlTextReaderPtr& aReader, QSParameters* aParameters)
{
	int type;
	unsigned int count = 0;
	int cnter = 0;
	main_sieving_type msds[10];
	signed long lowerBound;
	signed long upperBound;
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	type = xmlTextReaderNodeType(aReader);
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Node Info",rc,0,"Could not deserialize Node Info, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Node Info",rc,0,"Could not deserialize Node Info.");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{

			break;
		}
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//fprintf(stderr, "Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{

			if (xmlStrEqual(name,BAD_CAST COUNT_NAME))
			{
				rc = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_NODE_INFO,count,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				if (count > 10)
				{
					CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Node Info",count,10,"Count larger than 10.");
					return GENERAL_ERROR;
				}
				if (iInfoMode) {
					printf("%s: Total machine specific divisors %d.\n",DESERIALIZE_NODE_INFO,count);
				}
				//printf("Poly index: %d.\n",index);
			}
			if (xmlStrEqual(name,BAD_CAST MACHINE_SPECIFIC_DIVISORS_NAME))
			{
				if (iInfoMode)
				{
					printf("%s: Reading machine specific divisors \n",DESERIALIZE_NODE_INFO);
				}
			}
			if (xmlStrEqual(name,BAD_CAST QS_FB_ELEMENT))
			{
				signed long var;
				rc = CommonXMLHelper::DeserializeSignedLong(aReader,DESERIALIZE_NODE_INFO,var ,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				if (iInfoMode)
				{
					if (!cnter)
					{
						printf("%s: Divisor list: ",DESERIALIZE_NODE_INFO);
					}
				}
				msds[cnter++] = (main_sieving_type) var;
				if (iInfoMode)
				{
					printf("%4ld ",var);
				}
			}
			if (xmlStrEqual(name,BAD_CAST MACHINE_LOWER_BOUND_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedLong(aReader,DESERIALIZE_NODE_INFO,lowerBound,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				if (iInfoMode) {
					printf("%s: Lower machine bound %5ld.\n",DESERIALIZE_NODE_INFO,lowerBound);
				}
			}
			if (xmlStrEqual(name,BAD_CAST MACHINE_UPPER_BOUND_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedLong(aReader,DESERIALIZE_NODE_INFO,upperBound,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				if (iInfoMode) {
					printf("%s: Upper machine bound %5ld.\n",DESERIALIZE_NODE_INFO,upperBound);
				}
			}
		} // end of if type == 1
		if (type == 15)
		{
			if (xmlStrEqual(name,BAD_CAST MACHINE_SPECIFIC_DIVISORS_NAME))
			{
				printf("\n");
			}
		}
	} // end of while(TRUE)
	if (iQsInstance)
	{
		iQsInstance->Set_MachineSpecificDivisors(msds, count);
		iQsInstance->Set_MSDCount(count);
		iQsInstance->Set_LowerMachineBound(lowerBound);
		iQsInstance->Set_UpperMachineBound(upperBound);
	}
	if (aParameters)
	{
	}
	//printf("Read integer %d with log %d, sieving index %d, m %d, inversion %d, flags %u\n",aTarget.prime, aTarget.log, aTarget.sieving_index,aTarget.m, aTarget.inversion_of_m,aTarget.flags);
	return ConstRC::Ok;
}


int XmlQsSerializator::DeserializeJob()
{
	CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Job",0,0,"Support of direct DeserializeJob was discontinued after problems with indirection in SetupParameters. Use the method variant with parameters.");
	return NOT_SUPPORTED;
	/*if (iInfoMode)
	{
		printf("\n%s: Running...\n",DESERIALIZE_JOB);
	}
	CreateFileNamePlain(JOB_FILE);
	xmlTextReaderPtr reader;
    	reader = xmlNewTextReaderFilename(iAuxName.c_str());
	if (reader == NULL) {
		//printf("\n%s: Could not read the uncompressed file. Trying to read compressed...\n",DESERIALIZE_JOB);
		CreateFileNameCompressed(JOB_FILE);
		reader = xmlNewTextReaderFilename(iAuxName.c_str());
		if (reader == NULL)
		{
        		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Job",0,0,"XML Text Reader is null, aborting.");
			return GENERAL_ERROR;
		}
      	}
	QSParameters parameters;
	int ret = DeserializeJob(reader,parameters);
	if ((ret == ConstRC::Ok) && (iQsInstance))
	{
		iQsInstance->SetupParameters(parameters);
	}
        xmlFreeTextReader(reader);
	return ret;*/
}



int XmlQsSerializator::DeserializeJob(QSParameters& aParameters)
{
	if (iInfoMode)
	{
		printf("\n%s: Running...\n",DESERIALIZE_JOB);
	}
	CreateFileNamePlain(JOB_FILE);
	xmlTextReaderPtr reader;
    	reader = xmlNewTextReaderFilename(iAuxName.c_str());
	if (reader == NULL) {
		//printf("\n%s: Could not read the uncompressed file. Trying to read compressed...\n",DESERIALIZE_JOB);
		CreateFileNameCompressed(JOB_FILE);
		reader = xmlNewTextReaderFilename(iAuxName.c_str());
		if (reader == NULL)
		{
        		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Job",0,0,"XML Text Reader is null, aborting.");
			return GENERAL_ERROR;
		}
      	}
	int ret = DeserializeJob(reader,aParameters);
        xmlFreeTextReader(reader);
	return ret;
}

int XmlQsSerializator::DeserializeJob(JobParameters& aParameters)
{
	try
  	{
		QSParameters& qp = dynamic_cast<QSParameters&>(aParameters);
		return DeserializeJob(qp);
	}
	catch (const std::bad_cast & e)
  	{
		CLog::LogError(WHEREARG,"XML Qs Serializator","DeserializeJob (Parameters)",0,0,"Unknown subtype of JobParameters, doing nothing");
		return GENERAL_ERROR;
  	}

}


/**
	This method loads job parameters from the XML reader supplied, and
	sets up the provided aParameters. If iQsInstance is non-NULL, it also
	sets up the iQsInstance.

*/

int XmlQsSerializator::DeserializeJob(xmlTextReaderPtr& aReader, QSParameters& aParameters)
{
    	int ret = xmlTextReaderRead(aReader);
	int type = 0;
	xmlChar* name = 0;
	mpz_t auxMpz;
	mpz_init(auxMpz);
        while (ret == 1) {
        	type = xmlTextReaderNodeType(aReader);
		name = xmlTextReaderName(aReader);
		//fprintf(stderr,"Read type %d of name %s.\n",type,name);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST SERIALIZATOR_NAME))
			{
				ret = ReadSerializatorData(aReader,DESERIALIZE_JOB);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST TIME_NAME))
			{
				ret = CommonXMLHelper::DeserializeTime(aReader,DESERIALIZE_JOB,iInfoMode);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST MODULUS))
			{
				ret = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_JOB,auxMpz,iInfoMode);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_Modulus(auxMpz);
			}
			if (xmlStrEqual(name,BAD_CAST UPPER_FB_BOUND))
			{
				long target;
				ret = CommonXMLHelper::DeserializeSignedLong(aReader,DESERIALIZE_JOB,target,iInfoMode);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_UpperIntBound(target);
			}
			if (xmlStrEqual(name,BAD_CAST SIEVING_INTERVAL))
			{
				ret = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_JOB,auxMpz,iInfoMode);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_SievingIntervalLength(mpz_get_ui(auxMpz));
			}
			if (xmlStrEqual(name,BAD_CAST THRESHOLD_OPTIMIZATION_POLICY_NAME))
			{
				unsigned int policy;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,policy,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_ThresholdOptimizer_Level(policy);
			}
			if (xmlStrEqual(name,BAD_CAST QS_MODE))
			{
				int mode;
				ret = CommonXMLHelper::DeserializeSignedInt(aReader,DESERIALIZE_JOB,mode,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				qs_mode mode1 = (qs_mode) mode;
				if (iInfoMode)
				{
					printf("%s: The sieving mode used is ",DESERIALIZE_JOB);
					switch (mode1)
					{
						case EOldSiqs: printf("classical SIQS");
						break;
						case EOldMpqs: printf("classical MPQS");
						break;
						case ESiqs: printf("new line siever with SIQS");
						break;
						case EMpqs: printf("new line siever with MPQS");
						break;
						default: printf("not known, error");
						break;
					}
					printf(".\n");
				}
				aParameters.Set_QSMode(mode1);
			}
			if (xmlStrEqual(name,BAD_CAST INFO_MODE_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_InfoMode(target);
				printf("%s: Info Mode set.\n",DESERIALIZE_JOB);

				/*string target;
				ret =  CommonXMLHelper::DeserializeString(aReader, DESERIALIZE_JOB, target, FALSE);
				if (!strcmp(target.c_str(),NONE_NAME))
				{
					aParameters.Set_InfoMode(FALSE);
					aParameters.Set_DebugMode(FALSE);
				} else if (!strcmp(target.c_str(),INFO_NAME))
				{
					aParameters.Set_InfoMode(TRUE);
					aParameters.Set_DebugMode(FALSE);
				} else if (!strcmp(target.c_str(),DEBUG_NAME))
				{
					aParameters.Set_InfoMode(TRUE);
					aParameters.Set_DebugMode(TRUE);
				} else
				{
					printf("XML QS Serializator: Deserialize Job. Unknown info mode: %s\n",target.c_str());
				}*/
			}
			if (xmlStrEqual(name,BAD_CAST DEBUG_MODE_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_DebugMode(target);
			}
			if (xmlStrEqual(name,BAD_CAST VARIATIONS_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_UsingVariations((variations_types) target);
				switch (target)
				{
					case EDoNotUseVariations:
						printf("%s: Collecting smooth values only.\n",DESERIALIZE_JOB);
						break;
					case ELargePrimeV:
						printf("%s: Using single large prime variation.\n",DESERIALIZE_JOB);
						break;
					case EDoubleLargePrimeV:
						printf("%s: Using double large prime variation.\n",DESERIALIZE_JOB);
						break;
					default:
						printf("%s: Unknown vari mode, error?\n",DESERIALIZE_JOB);
						break;
				}
			}
			/*if (xmlStrEqual(name,BAD_CAST FACTALG_NAME))
			{
				unsigned int target;
				ret = DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_FactoringAlgorithm((factor_alg) target);
				if (aParameters.Get_UsingVariations() == EDoubleLargePrimeV)
				{
					switch (target)
					{
						case EPollardRho:
							printf("%s: Pollard rho is used in double LPV.\n",DESERIALIZE_JOB);
							break;
						case ECFRAC:
							printf("%s: CFRAC is used in double LPV.\n",DESERIALIZE_JOB);
							break;
						case EPMinus1:
							printf("%s: p-1 is used in double LPV.\n",DESERIALIZE_JOB);
							break;
						case EEllipticCurve:
							printf("%s: Elliptic curve factoring is used in double LPV.\n",DESERIALIZE_JOB);
							break;
						default:
							printf("%s: Unknown algorithm  is used in double LPV, error?\n",DESERIALIZE_JOB);
							break;
					}
				}
			}
			*/
			if (xmlStrEqual(name,BAD_CAST MEMBLOCK_SIZE))
			{
				unsigned int val;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,val,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_MemblockSize(val);
			}
			if (xmlStrEqual(name, BAD_CAST DISTRIBUTION_MACHINE_TYPE))
			{
				char* auxChar = (char*) malloc(100*sizeof(char));
				ret = CommonXMLHelper::DeserializeString(aReader, DESERIALIZE_JOB,auxChar,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				if (strstr(auxChar,NODE))
				{
					aParameters.Set_DistributionMachineType(ENode);
				}
				if (strstr(auxChar,CENTER))
				{
					aParameters.Set_DistributionMachineType(ECenter);
				}
				printf("%s: This machine is %s.\n",DESERIALIZE_JOB,auxChar);
				free (auxChar);
			}
			if (xmlStrEqual(name, BAD_CAST FRONTEND_PIPE))
			{
				char* auxChar = (char*) malloc(300*sizeof(char));
				ret = CommonXMLHelper::DeserializeString(aReader, DESERIALIZE_JOB,auxChar,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_FrontendPipe(auxChar);
				fprintf(stderr, "%s: Frontend pipe: %s\n",DESERIALIZE_JOB,auxChar);
				free (auxChar);
			}
			if (xmlStrEqual(name,BAD_CAST MACHINE_SPECIFIC_GENERATION))
			{
				unsigned int val;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,val,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_MachineSpecificGeneration(static_cast<machine_specific_generation_type>(val));
				printf("%s: Machine specific generation is %d.\n",DESERIALIZE_JOB,val);
			}
			if (xmlStrEqual(name,BAD_CAST COMMIT_INTERVAL))
			{
				unsigned int val;
				ret = CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_JOB,val,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				aParameters.Set_CommitInterval(val);
				printf("%s: Commit interval is %d.\n",DESERIALIZE_JOB,val);
			}
			if (xmlStrEqual(name, BAD_CAST NODE_SPECIFIC_INFORMATION_NAME))
			{
				ret = DeserializeNodeInfo(aReader);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
		}
            	ret = xmlTextReaderRead(aReader);
        }
        if (ret != 0) {
        	printf("%s : failed to parse\n", iAuxName.c_str());
        }
	mpz_clear(auxMpz);
	if (aParameters.Is_InfoMode())
	{
		aParameters.Print();
	}
	return ConstRC::Ok;
}


int XmlQsSerializator::SerializeFactorBase()
{
	Remove(FB_FILE);
	// FB_FILE defined in xml_definitions.h
	// Auto.... extension depends on current compression level
	CreateFileNameAuto(FB_FILE);
 	int rc;
    	xmlTextWriterPtr writer;
	/* Create a new XmlWriter for uri */
    	writer = xmlNewTextWriterFilename(iAuxName.c_str(), iCompressionLevel);
    	if (writer == NULL) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",0,0,"XML Text Writer is null, aborting.");
		return GENERAL_ERROR;
      	}
	rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",rc,0,"Could not write the document start.");
		return GENERAL_ERROR;
    	}
	rc = xmlTextWriterStartElement(writer, BAD_CAST FACTOR_BASE_DATA);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",rc,0,"Could not write the serializator tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(writer,SERIALIZATOR_VAL, SERIALIZATOR_VERSION);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	long maxIndex = 0;
	long allocated = 0;
	main_sieving_type* fb = iQsInstance->Get_FactorBase(maxIndex,allocated);
	if (fb)
	{
		rc = CommonXMLHelper::SerializeSignedLong(writer,ALLOCATED_NAME,allocated);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeSignedLong(writer,MAX_INDEX_NAME,maxIndex);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		int i;
		for (i=0; i<= maxIndex; i++)
		{
			if (fb[i] == 0)
			{
				printf("A strange error @ writing fb ! FB[%d] is 0.\n",i);
			}
			rc = CommonXMLHelper::SerializeSignedLong(writer,QS_FB_ELEMENT,fb[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}
    	rc = xmlTextWriterEndDocument(writer);
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}
	xmlFreeTextWriter(writer);
	return ConstRC::Ok;
}

int XmlQsSerializator::SerializeSieving()
{
	Remove(SIEVING_FILE);
	// SIEVING_FILE defined in xml_definitions.h
	// Auto.... extension depends on current compression level
	CreateFileNameAuto(SIEVING_FILE);
 	int rc;
    	xmlTextWriterPtr writer;
	/* Create a new XmlWriter for uri */
    	writer = xmlNewTextWriterFilename(iAuxName.c_str(), iCompressionLevel);
    	if (writer == NULL) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",0,0,"XML Text Writer is null, aborting.");
		return GENERAL_ERROR;
      	}
	rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",rc,0,"Could not write the document start.");
		return GENERAL_ERROR;
    	}
	rc = xmlTextWriterStartElement(writer, BAD_CAST SIEVING_DATA);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Factor Base",rc,0,"Could not write the serializator tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(writer,SERIALIZATOR_VAL, SERIALIZATOR_VERSION);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	sieving_phase phase = iQsInstance->GetSievingPhase();
	rc = CommonXMLHelper::SerializeSignedInt(writer,SIEVING_STATUS,phase);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedInt(writer,LINE_NUMBER,iQsInstance->GetLineNr());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_t aux;
	mpz_init(aux);
	iQsInstance->Get_Root(aux);
	rc = CommonXMLHelper::SerializeMpz(writer,SQRT_OF_KN_NAME,aux);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	iQsInstance->Get_MPF(aux);
	rc = CommonXMLHelper::SerializeMpz(writer,MAXIMAL_PARTIAL_FACTOR_NAME,aux);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	iQsInstance->Get_IdealAValue(aux);
	rc = CommonXMLHelper::SerializeMpz(writer,IDEAL_A_VALUE_NAME,aux);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,CONTINI_THRESHOLD_NAME,iQsInstance->Get_ContiniThreshold());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,ERROR_FACTOR_NAME,iQsInstance->Get_ErrorFactor());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,VARIATIONS_NAME,iQsInstance->Get_UsingVariations());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,FACTALG_NAME,iQsInstance->Get_FactoringAlgorithm());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,SO_FAR_SMOOTHS_FOUND_NAME,iQsInstance->Get_SmoothsFound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,SO_FAR_1_PARTIALS_FOUND_NAME,iQsInstance->Get_1PartialsFound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,SO_FAR_2_PARTIALS_FOUND_NAME,iQsInstance->Get_2PartialsFound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,SO_FAR_X_PARTIALS_FOUND_NAME,iQsInstance->Get_XPartialsFound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(writer,RELATION_PROCESSING_PHASE_NAME,iQsInstance->Get_RelationProcessingPhase());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedInt(writer,RELATION_PROCESSING_FLAGS_NAME,iQsInstance->Get_RelationProcessingFlags());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_clear(aux);
   	rc = xmlTextWriterEndDocument(writer);
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Sieving",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}
	xmlFreeTextWriter(writer);
	return ConstRC::Ok;
}

int XmlQsSerializator::DeserializeSieving()
{
	if (iInfoMode)
	{
		printf("\n%s: Running...\n",DESERIALIZE_SIEVING);
	}
	CreateFileNamePlain(SIEVING_FILE);
	xmlTextReaderPtr reader;
    	int ret, overallResult = ConstRC::Ok;
    	reader = xmlNewTextReaderFilename(iAuxName.c_str());
	if (reader == NULL) {
		CreateFileNameCompressed(SIEVING_FILE);
		reader = xmlNewTextReaderFilename(iAuxName.c_str());
		if (reader == NULL)
		{
        		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Sieving",0,0,"XML Text Reader is null, aborting.");
			return GENERAL_ERROR;
		}
      	}
	ret = xmlTextReaderRead(reader);
	int type = 0;
	xmlChar* name = 0;
	mpz_t aux;
	mpz_init(aux);
	bool contThreshRead = FALSE;
	bool errorFactRead  = FALSE;

        while (ret == 1) {
        	type = xmlTextReaderNodeType(reader);
		name = xmlTextReaderName(reader);
		//printf("Read type %d of name %s.\n",type,name);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST SERIALIZATOR_NAME))
			{
				ret = ReadSerializatorData(reader,DESERIALIZE_SIEVING);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST SIEVING_STATUS))
			{
				int target;
				ret = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				sieving_phase phase = (sieving_phase) target;
				switch (phase)
				{
					case ESievingNotStarted:
						printf("%s: Sieving not started.\n",DESERIALIZE_SIEVING);
						break;
					case ESievingRunning:
						printf("%s: Sieving running.\n",DESERIALIZE_SIEVING);
						break;
					case ESievingFinished:
						printf("%s: Sieving finished.\n",DESERIALIZE_SIEVING);
						break;
					default:
						printf("%s: Sieving status unknown, error?\n",DESERIALIZE_SIEVING);
						break;
				}
				iQsInstance->SetSievingPhase(phase);
			}
			if (xmlStrEqual(name,BAD_CAST LINE_NUMBER))
			{
				int target;
				ret = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->SetLineNr(target);
				if (iInfoMode)
				{
					printf("%s: Last line # is %d.\n",DESERIALIZE_SIEVING,target);
				}
			}
			if (xmlStrEqual(name,BAD_CAST SQRT_OF_KN_NAME))
			{
				ret = CommonXMLHelper::DeserializeMpz(reader,DESERIALIZE_SIEVING,aux,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_Root(aux);
			}
			if (xmlStrEqual(name,BAD_CAST MAXIMAL_PARTIAL_FACTOR_NAME))
			{
				ret = CommonXMLHelper::DeserializeMpz(reader,DESERIALIZE_SIEVING,aux,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_MPF(aux);
				if (iInfoMode)
				{
					gmp_printf("%s: Maximal acceptable partial factor is %Zd (%d bits).\n",DESERIALIZE_SIEVING,aux,mpz_sizeinbase(aux,2));
				}
			}
			if (xmlStrEqual(name,BAD_CAST IDEAL_A_VALUE_NAME))
			{
				ret = CommonXMLHelper::DeserializeMpz(reader,DESERIALIZE_SIEVING,aux,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_IdealAValue(aux);
				if (iInfoMode)
				{
					switch (iQsInstance->Get_QSMode())
					{
						case ESiqs:
						case EOldSiqs:
							gmp_printf("%s: Ideal value of 'a' in a SIQS polynomial is %Zd (%d bits).\n",DESERIALIZE_SIEVING,aux,mpz_sizeinbase(aux,2));
							break;
						default:
							break;
					}

				}
			}
			if (xmlStrEqual(name,BAD_CAST CONTINI_THRESHOLD_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_ContiniThreshold((log_type) target);
				contThreshRead = TRUE;
			}
			if (xmlStrEqual(name,BAD_CAST ERROR_FACTOR_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_ErrorFactor((log_type) target);
				errorFactRead = TRUE;
			}
			if (xmlStrEqual(name,BAD_CAST VARIATIONS_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_UsingVariations((variations_types) target);
				switch (target)
				{
					case EDoNotUseVariations:
						printf("%s: Collecting smooth values only.\n",DESERIALIZE_SIEVING);
						break;
					case ELargePrimeV:
						printf("%s: Using single large prime variation.\n",DESERIALIZE_SIEVING);
						break;
					case EDoubleLargePrimeV:
						printf("%s: Using double large prime variation.\n",DESERIALIZE_SIEVING);
						break;
					default:
						printf("%s: Unknown vari mode, error?\n",DESERIALIZE_SIEVING);
						break;
				}
			}
			if (xmlStrEqual(name,BAD_CAST FACTALG_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_FactoringAlgorithm((factor_alg) target);
				if (iQsInstance->Get_UsingVariations() == EDoubleLargePrimeV)
				{
					switch (target)
					{
						case EPollardRho:
							printf("%s: Pollard rho is used in double LPV.\n",DESERIALIZE_SIEVING);
							break;
						case ECFRAC:
							printf("%s: CFRAC is used in double LPV.\n",DESERIALIZE_SIEVING);
							break;
						case EPMinus1:
							printf("%s: p-1 is used in double LPV.\n",DESERIALIZE_SIEVING);
							break;
						case EEllipticCurve:
							printf("%s: Elliptic curve factoring is used in double LPV.\n",DESERIALIZE_SIEVING);
							break;
						default:
							printf("%s: Unknown algorithm  is used in double LPV, error?\n",DESERIALIZE_SIEVING);
							break;
					}
				}
			}
			if (xmlStrEqual(name,BAD_CAST SO_FAR_SMOOTHS_FOUND_NAME))
			{
				long target;
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_SmoothsFound(target);
				if (iInfoMode)
				{
					printf("%s: Smooth values found so far : %ld\n",DESERIALIZE_SIEVING,target);
				}
			}
			if (xmlStrEqual(name,BAD_CAST SO_FAR_1_PARTIALS_FOUND_NAME))
			{
				long target;
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_1PartialsFound(target);
				if (iInfoMode)
				{
					if (iQsInstance->Get_UsingVariations() != EDoNotUseVariations)
					{
						printf("%s: 1-partial values found so far : %ld\n",DESERIALIZE_SIEVING,target);
					}
				}
			}
			if (xmlStrEqual(name,BAD_CAST SO_FAR_2_PARTIALS_FOUND_NAME))
			{
				long target;
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_2PartialsFound(target);
				if (iInfoMode)
				{
					if (iQsInstance->Get_UsingVariations() == EDoubleLargePrimeV)
					{
						printf("%s: 2-partial values found so far : %ld\n",DESERIALIZE_SIEVING,target);
					}
				}
			}
			if (xmlStrEqual(name,BAD_CAST SO_FAR_X_PARTIALS_FOUND_NAME))
			{
				long target;
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_XPartialsFound(target);
			}
			if (xmlStrEqual(name,BAD_CAST RELATION_PROCESSING_PHASE_NAME))
			{
				unsigned int target;
				ret = CommonXMLHelper::DeserializeUnsignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				relproc_phase phase = (relproc_phase) target;
				iQsInstance->Set_RelationProcessingPhase(phase);
				bool cannotReload = FALSE;
				switch (phase)
				{
					case EBeforeRelationCollection:
						cannotReload = TRUE;
						break;
					case ECollectingRelations:
						break;
					case ERemovingLargeSingletons:
						cannotReload = TRUE;
						break;
					case EConstructingCycles:
						cannotReload = TRUE;
						break;
					case EDeletingDuplicates:
						cannotReload = TRUE;
						break;
					case ERemovingSmallSingletons:
						cannotReload = TRUE;
						break;
					case ERelationsReady:
						break;
					default:
						cannotReload = TRUE;
						break;
				}
				if (cannotReload)
				{
					overallResult = GENERAL_ERROR;
					CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize sieving",0,0,"Cannot reload this phase yet, aborting");
				}
			}
			if (xmlStrEqual(name,BAD_CAST RELATION_PROCESSING_FLAGS_NAME))
			{
				int target;
				ret = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVING,target,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				iQsInstance->Set_RelationProcessingFlags(target);
				printf("%s: Relation processing flags: 0x%x.\n",DESERIALIZE_SIEVING,target);
				if (target & SIEVING_RUNNING_FLAG)
				{
					printf("%s: Sieving running.\n",DESERIALIZE_SIEVING);
				}
				if (target & RELATION_COLLECTION_FINISHED_FLAG)
				{
					printf("%s: Relation collection finished.\n",DESERIALIZE_SIEVING);
				}
				if (target & RELATIONS_READY_FLAG)
				{
					printf("%s: Relation processing finished.\n",DESERIALIZE_SIEVING);
				}
			}
		}
		if (contThreshRead & errorFactRead)
		{
			if (iInfoMode)
			{
				printf("%s: Contini threshold 0x%x, error factor makes 0x%x (%d) of it.\n",DESERIALIZE_SIEVING,iQsInstance->Get_ContiniThreshold(),iQsInstance->Get_ErrorFactor(),iQsInstance->Get_ErrorFactor());
			}
			contThreshRead = FALSE;
			errorFactRead  = FALSE;
		}
            	ret = xmlTextReaderRead(reader);
        }
	xmlFreeTextReader(reader);
        if (ret != 0) {
        	printf("%s : failed to parse\n", iAuxName.c_str());
        }
	mpz_clear(aux);
	return overallResult;
}

int XmlQsSerializator::DeserializeFactorBase()
{
	if (iInfoMode)
	{
		printf("\n%s: Running...\n",DESERIALIZE_FB);
	}
	CreateFileNamePlain(FB_FILE);
	xmlTextReaderPtr reader;
    	int ret;
    	reader = xmlNewTextReaderFilename(iAuxName.c_str());
	if (reader == NULL) {
		CreateFileNameCompressed(FB_FILE);
		reader = xmlNewTextReaderFilename(iAuxName.c_str());
		if (reader == NULL)
		{
        		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Factor Base",0,0,"XML Text Reader is null, aborting.");
			return GENERAL_ERROR;
		}
      	}
	ret = xmlTextReaderRead(reader);
	int type = 0;
	xmlChar* name = 0;
	mpz_t auxMpz;
	mpz_init(auxMpz);
	long running_index = 0;
	long allocated = 0;
	long max_index = -1;
	long element;
	main_sieving_type* fb = NULL;
        while (ret == 1) {
        	type = xmlTextReaderNodeType(reader);
		name = xmlTextReaderName(reader);
		//printf("Read type %d of name %s.\n",type,name);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST SERIALIZATOR_NAME))
			{
				ret = ReadSerializatorData(reader,DESERIALIZE_FB);
				if (ret != ConstRC::Ok)
				{
					if (fb) free(fb);
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST ALLOCATED_NAME))
			{
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_FB,allocated,FALSE);
				if (ret != ConstRC::Ok)
				{
					if (fb) free(fb);
					return ret;
				}
				fb = (main_sieving_type*) malloc(allocated*sizeof(main_sieving_type));
				if (!fb)
				{
					CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Factor Base",allocated,0,"Could not allocate the factor base, aborting");
					if (fb) free(fb);
					return NOT_ENOUGH_MEMORY;
				}
				memset(fb,0,allocated*sizeof(main_sieving_type));
			}
			if (xmlStrEqual(name,BAD_CAST MAX_INDEX_NAME))
			{
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_FB,max_index,FALSE);
				if (ret != ConstRC::Ok)
				{
					if (fb) free(fb);
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST QS_FB_ELEMENT))
			{
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_FB,element,FALSE);
				if (ret != ConstRC::Ok)
				{
					if (fb) free(fb);
					return ret;
				}
				if (element == 0)
				{
					printf("Strange error @ load fb: element is 0 for running index %ld.\n",running_index);
				}
				fb[running_index++] = (main_sieving_type) element;
			}
		}
            	ret = xmlTextReaderRead(reader);
        }
	printf("%s: The factor base has %ld elements and allocated size %ld.\n",DESERIALIZE_FB,(max_index+1),allocated);
	iQsInstance->Set_FactorBase(fb,max_index,allocated);
        xmlFreeTextReader(reader);
        if (ret != 0) {
        	printf("%s : failed to parse\n", iAuxName.c_str());
        }
	mpz_clear(auxMpz);
	return ConstRC::Ok;
}

int  XmlQsSerializator::SerializeSieverState()
{
	Remove(SIEVER_STATE_FILE);
	// JOB_FILE defined in xml_definitions.h
	// Auto ... extension depends on current compression level
	CreateFileNameAuto(SIEVER_STATE_FILE);
 	int rc;
    	xmlTextWriterPtr writer;
	/* Create a new XmlWriter for uri */
    	writer = xmlNewTextWriterFilename(iAuxName.c_str(), iCompressionLevel);
    	if (writer == NULL) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",0,0,"XML Text Writer is null, aborting.");
		return GENERAL_ERROR;
      	}
	rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the document start.");
		return GENERAL_ERROR;
    	}
	rc = xmlTextWriterStartElement(writer, BAD_CAST SIEVER_STATUS_DATA);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the serializator tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(writer,SERIALIZATOR_VAL, SERIALIZATOR_VERSION);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeTime(writer);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	mpz_t auxMpz;
	mpz_init(auxMpz);
	// Now, we will save the block of polynomials.
	int s = iQsInstance->Get_S();
	long nrPol, allocPol, indPol, curPol;
	quad_polynomial* poly = iQsInstance->Get_Polynomials(nrPol,allocPol,indPol,curPol);
	rc = CommonXMLHelper::SerializeSignedInt(writer,SIQS_S_VALUE,s);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedInt(writer,SOLVE_VALUE,iQsInstance->Get_Solve());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,SO_FAR_SMOOTHS_FOUND_NAME,iQsInstance->Get_SmoothsFound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,SO_FAR_X_PARTIALS_FOUND_NAME,iQsInstance->Get_XPartialsFound());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,ALLOCATED_POLY_NAME,allocPol);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,NR_POLY_NAME,nrPol);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,IND_POLY_NAME,indPol);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeSignedLong(writer,CUR_POLY_NAME,curPol);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	int i, max = allocPol;
	quad_polynomial* one_poly;
	for (i=0; i< max; i++)
	{
		one_poly = &(poly[i]);
		rc = xmlTextWriterStartElement(writer, BAD_CAST POLYNOMIAL_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        		return GENERAL_ERROR;
    		}
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeSignedInt(writer,INDEX_NAME,i);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeMpz(writer,A_NAME,one_poly->a);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeMpz(writer,B_NAME,one_poly->b);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeMpz(writer,C_NAME,one_poly->c);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeMpz(writer,D_NAME,one_poly->d);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeMpz(writer,D2_NAME,one_poly->d2);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CommonXMLHelper::SerializeSignedInt(writer,INITED_NAME,one_poly->inited);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = xmlTextWriterEndElement(writer);
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	qs_mode mode = iQsInstance->Get_QSMode();
	bool isSiqs = ((mode == ESiqs) || (mode == EOldSiqs));
	bool isOptimizedSiever = ((mode == ESiqs) || (mode == EMpqs));
	if (isSiqs)
	{
		// Serialize B2 array
		mpz_t* b2 = iQsInstance->Get_B2Block();
		if (b2)
		{
			rc = xmlTextWriterStartElement(writer, BAD_CAST B2_ARRAY_NAME);
    			if (rc < 0) {
	        		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        			return GENERAL_ERROR;
    			}
			rc = CommonXMLHelper::NewLine(writer,FALSE);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			max = iQsInstance->Get_S();
			for (i=0; i< max; i++)
			{
				rc = CommonXMLHelper::SerializeMpz(writer,B2_NAME,b2[i]);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			rc = xmlTextWriterEndElement(writer);
			rc = CommonXMLHelper::NewLine(writer,FALSE);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
		// Serialize divisors of 'a'
		main_sieving_type* divisors_of_a = iQsInstance->Get_DivisorsOfA();
		if (divisors_of_a)
		{
			rc = xmlTextWriterStartElement(writer, BAD_CAST DIVISORS_OF_A_ARRAY_NAME);
    			if (rc < 0) {
	        		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        			return GENERAL_ERROR;
    			}
			rc = CommonXMLHelper::NewLine(writer,FALSE);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			max = iQsInstance->Get_S();
			for (i=0; i< max; i++)
			{
				rc = CommonXMLHelper::SerializeSignedInt(writer, DIVISOR_OF_A_NAME,divisors_of_a[i]);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			rc = xmlTextWriterEndElement(writer);
			rc = CommonXMLHelper::NewLine(writer,FALSE);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
		// Serialize NEXKSB Environment
		nexksb_environment* env = iQsInstance->NEXKSB_Get();
		if (env)
		{
			rc = xmlTextWriterStartElement(writer, BAD_CAST NEXKSB_NAME);
    			if (rc < 0) {
	        		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        			return GENERAL_ERROR;
    			}
			rc = CommonXMLHelper::SerializeSignedInt(writer, NEXKSB_K_NAME,env->k);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CommonXMLHelper::SerializeSignedInt(writer, NEXKSB_N_NAME,env->n);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CommonXMLHelper::SerializeSignedInt(writer, NEXKSB_H_NAME,env->h);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}

			rc = CommonXMLHelper::SerializeSignedInt(writer, NEXKSB_N_NAME,env->m);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CommonXMLHelper::SerializeSignedInt(writer, NEXKSB_FINISHED_NAME,env->finished);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = xmlTextWriterStartElement(writer, BAD_CAST NEXKSB_SUBSET_ARRAY_NAME);
    			if (rc < 0) {
	        		CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        			return GENERAL_ERROR;
    			}
			for (i=0; i<= env->k; i++)
			{
				rc = CommonXMLHelper::SerializeSignedInt(writer, NEXKSB_SUBSET_ELEMENT_NAME,env->subset[i]);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			rc = xmlTextWriterEndElement(writer);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = xmlTextWriterEndElement(writer);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CommonXMLHelper::NewLine(writer,FALSE);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
		// Their indices in a FB will NOT be serialized, but recalculated during
		//   the reload.
		// Ba_inv will not be serialized either.
		int nu = iQsInstance->Get_Nu();
		rc = CommonXMLHelper::SerializeSignedInt(writer,NU_NAME,nu);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		int subtract = iQsInstance->Get_Subtract();
		rc = CommonXMLHelper::SerializeSignedInt(writer,SUBTRACT_NAME,subtract);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	if (!isOptimizedSiever)
	{
		// Serialize sol_1, sol_2 and usable arrays.
		integer_solutions sols = iQsInstance->Get_Solutions();
		rc = CommonXMLHelper::SerializeSignedInt(writer,ALLOC_SOLUTIONS_NAME,sols.allocated);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		// Usable field
		rc = xmlTextWriterStartElement(writer, BAD_CAST USABLE_ARRAY_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        		return GENERAL_ERROR;
    		}
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		max = iQsInstance->Get_MaxFBIndex();
		for (i=0; i<=max; i++){
			rc = CommonXMLHelper::SerializeSignedInt(writer, USABLE_NAME,sols.usable[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
		rc = xmlTextWriterEndElement(writer);
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		// Solutions 1 field
		rc = xmlTextWriterStartElement(writer, BAD_CAST SOL_1_ARRAY_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        		return GENERAL_ERROR;
    		}
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		for (i=0; i<=max; i++){
			rc = CommonXMLHelper::SerializeSignedInt(writer, SOL_1_NAME,sols.sol_1[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
		rc = xmlTextWriterEndElement(writer);
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		// Solutions 2 field
		rc = xmlTextWriterStartElement(writer, BAD_CAST SOL_2_ARRAY_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not write the polynomial tag.");
        		return GENERAL_ERROR;
    		}
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		for (i=0; i<= max; i++){
			rc = CommonXMLHelper::SerializeSignedInt(writer, SOL_2_NAME,sols.sol_2[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
		rc = xmlTextWriterEndElement(writer);
		rc = CommonXMLHelper::NewLine(writer,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

	} else
	{
		// Serialize advanced factor base.
		// Serialize hashtables.
	}
    	rc = xmlTextWriterEndDocument(writer);
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Siever State",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}

	xmlFreeTextWriter(writer);
	mpz_clear(auxMpz);
	return ConstRC::Ok;
}

int  XmlQsSerializator::DeserializePolynomial(xmlTextReaderPtr& aReader, quad_polynomial*  aArray, int aMaxIndex)
{
	int type;
	int index;
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	type = xmlTextReaderNodeType(aReader);
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Polynomial",rc,0,"Could not deserialize polynomial, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Polynomial",rc,0,"Could not deserialize polynomial.");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{
			//printf("Deserialize integer for sieving finished.");
			break;
		}
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		if (type == 1)
		{
			//printf("Name: %s.\n",name);
			if (xmlStrEqual(name,BAD_CAST INDEX_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(aReader,DESERIALIZE_POLYNOMIAL,index,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				if ((index < 0) || (index > aMaxIndex))
				{
					CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Polynomial",index,aMaxIndex,"Index out of acceptable bounds.");
					return GENERAL_ERROR;
				}
				//printf("Poly index: %d.\n",index);
			}
			if (xmlStrEqual(name,BAD_CAST A_NAME))
			{
				rc = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_POLYNOMIAL,aArray[index].a,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST B_NAME))
			{
				rc = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_POLYNOMIAL,aArray[index].b,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST C_NAME))
			{
				rc = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_POLYNOMIAL,aArray[index].c,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST D_NAME))
			{
				rc = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_POLYNOMIAL,aArray[index].d,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST D2_NAME))
			{
				rc = CommonXMLHelper::DeserializeMpz(aReader,DESERIALIZE_POLYNOMIAL,aArray[index].d2,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
		} // end of if type == 1
	} // end of while(TRUE)
	//printf("Read integer %d with log %d, sieving index %d, m %d, inversion %d, flags %u\n",aTarget.prime, aTarget.log, aTarget.sieving_index,aTarget.m, aTarget.inversion_of_m,aTarget.flags);
	return ConstRC::Ok;
}

int  XmlQsSerializator::DeserializeSieverState()
{
	if (iInfoMode)
	{
		printf("\n%s: Running...\n",DESERIALIZE_SIEVER_STATE);
	}
	int counter = 0, b2_counter = 0, a_counter = 0;
	int sol_1_counter = 0, sol_2_counter = 0, usable_counter = 0;
	int env_counter = 0;
	xmlTextReaderPtr reader;
    	int ret;
	CreateFileNamePlain(SIEVER_STATE_FILE);
	reader = xmlNewTextReaderFilename(iAuxName.c_str());
	nexksb_environment env;
	env.subset = NULL;
	if (reader == NULL) {
		CreateFileNameCompressed(SIEVER_STATE_FILE);
		reader = xmlNewTextReaderFilename(iAuxName.c_str());
		if (reader == NULL)
		{
        		CLog::LogError(WHEREARG,"XML Qs Serializator","Deserialize Siever state",counter,0,"XML Text Reader is null, aborting.");
			return GENERAL_ERROR;
		}
      	}
	ret = xmlTextReaderRead(reader);
	int type = 0;
	xmlChar* name = 0;
	int s = -1,rc = -1;
	long nrPol = -1, allocPol = -1, indPol = -1, curPol = -1;
	long x_partials = -1;
	long smooths = -1;
	quad_polynomial* poly = NULL;
	mpz_t* b2 = NULL;
	int*   b2_inited = NULL;
	main_sieving_type* divisors_of_a = NULL;
	integer_solutions sols;
	int max_fb_index = iQsInstance->Get_MaxFBIndex();
	bool isSiqs = ((iQsInstance->Get_QSMode() == ESiqs) || (iQsInstance->Get_QSMode() == EOldSiqs));
        while (ret == 1) {
        	type = xmlTextReaderNodeType(reader);
		name = xmlTextReaderName(reader);
		//printf("Read type %d of name %s.\n",type,name);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST SERIALIZATOR_NAME))
			{
				ret = ReadSerializatorData(reader,DESERIALIZE_SIEVER_STATE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST SIQS_S_VALUE))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,s,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				iQsInstance->Set_S(s);
			}
			if (xmlStrEqual(name,BAD_CAST SOLVE_VALUE))
			{
				int r1;
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,r1,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				iQsInstance->Set_Solve(r1);
			}
			if (xmlStrEqual(name,BAD_CAST NEXKSB_K_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,env.k,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST NEXKSB_N_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,env.n,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST NEXKSB_H_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,env.h,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}

			if (xmlStrEqual(name,BAD_CAST NEXKSB_M_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,env.m,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST NEXKSB_FINISHED_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,env.finished,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST NEXKSB_SUBSET_ARRAY_NAME))
			{
				if (env.subset == NULL)
				{
					env.subset = (nexksb_type*) malloc(sizeof(nexksb_type)*(env.k+1));
					memset(env.subset,0,sizeof(nexksb_type)*(env.k+1));
				}
			}
			if (xmlStrEqual(name,BAD_CAST NEXKSB_SUBSET_ELEMENT_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,env.subset[env_counter],FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				env_counter++;
			}
			if (xmlStrEqual(name,BAD_CAST ALLOCATED_POLY_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVER_STATE,allocPol,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				poly = (quad_polynomial*) AbstractSieve::AllocationMachine((void**) & poly,allocPol*sizeof(quad_polynomial));
				int i;
				for (i=0; i< allocPol; i++)
				{
					iQsInstance->AllocPolynomial(poly[i]);
				}
				printf("%s: Allocated polynomials %ld\n",DESERIALIZE_SIEVER_STATE,allocPol);
			}
			if (xmlStrEqual(name,BAD_CAST NR_POLY_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVER_STATE,nrPol,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				printf("%s: Number of polynomials %ld\n",DESERIALIZE_SIEVER_STATE,nrPol);
			}
			if (xmlStrEqual(name,BAD_CAST IND_POLY_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_POLYNOMIAL,indPol,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				//printf("%s: Index in polynomial %d\n",DESERIALIZE_SIEVER_STATE,indPol);
			}
			if (xmlStrEqual(name,BAD_CAST CUR_POLY_NAME))
			{
				rc = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_POLYNOMIAL,curPol,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				printf("%s: Current polynomial %ld\n",DESERIALIZE_SIEVER_STATE,curPol);
			}
			if (xmlStrEqual(name,BAD_CAST POLYNOMIAL_NAME))
			{
				ret = DeserializePolynomial(reader,poly,allocPol-1);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				counter++;
			}
			if (xmlStrEqual(name,BAD_CAST SO_FAR_X_PARTIALS_FOUND_NAME))
			{
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVING,x_partials,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
			if (xmlStrEqual(name,BAD_CAST SO_FAR_SMOOTHS_FOUND_NAME))
			{
				ret = CommonXMLHelper::DeserializeSignedLong(reader,DESERIALIZE_SIEVING,smooths,FALSE);
				if (ret != ConstRC::Ok)
				{
					return ret;
 				}
			}
			// Init values are counter = 0, allocPol = -1, so they do not match
			if (counter == allocPol)
			{
				printf("\r%s: Altogether %d polynomials read. \n",DESERIALIZE_SIEVER_STATE,counter);
				iQsInstance->Set_Polynomials(poly,nrPol,allocPol,indPol,curPol);
				counter = 0; // Not to write the message again

			}
			if (xmlStrEqual(name,BAD_CAST B2_ARRAY_NAME))
			{
				if (!b2)
				{
					//printf("Allocating b2 for %d elements.",s);
					b2 = (mpz_t*) AbstractSieve::AllocationMachine((void**) &b2,s*sizeof(mpz_t));
					b2_inited = (int*) AbstractSieve::AllocationMachine((void**) &b2_inited,s*sizeof(int));
					memset(b2_inited,0,s*sizeof(int));
					for (int i=0; i< s; i++)
					{
						mpz_init(b2[i]);
						b2_inited[i] = 1;
					}
				}
			}
			if (xmlStrEqual(name,BAD_CAST B2_NAME))
			{
				//printf("Reading element b2[%d].\n",b2_counter);
				rc = CommonXMLHelper::DeserializeMpz(reader,DESERIALIZE_SIEVER_STATE,b2[b2_counter++],FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST DIVISORS_OF_A_ARRAY_NAME))
			{
				if (!divisors_of_a)
				{
					//printf("\nAllocating divisors of a for %d elements.\n",s);
					divisors_of_a = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &divisors_of_a,s*sizeof(main_sieving_type));
					memset(divisors_of_a,0,s*sizeof(main_sieving_type));
				}
			}
			if (xmlStrEqual(name,BAD_CAST DIVISOR_OF_A_NAME))
			{
				//printf("Reading element adiv[%d].\n",a_counter);
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,divisors_of_a[a_counter++],FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				//printf("Read element %d.\n",divisors_of_a[a_counter-1]);
			}
			if (xmlStrEqual(name,BAD_CAST NU_NAME))
			{
				int nu;
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,nu,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				iQsInstance->Set_Nu(nu);
			}
			if (xmlStrEqual(name,BAD_CAST SUBTRACT_NAME))
			{
				int subtract;
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,subtract,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				iQsInstance->Set_Subtract(subtract);
			}
			// Init values are counter = 0, s = -1, so they do not match
			if (b2_counter == s)
			{
				if (isSiqs)
				{
					printf("\r%s: Altogether %d b*2s read. \n",DESERIALIZE_SIEVER_STATE,b2_counter);
					iQsInstance->Set_B2Block(b2,b2_inited);
					b2_counter = 0;  // Not to write the message again
				}
			}
			// Init values are counter = 0, s = -1, so they do not match
			if (a_counter == s)
			{
				if (isSiqs)
				{
					printf("\r%s: Altogether %d divisors of 'a' read.\n",DESERIALIZE_SIEVER_STATE,a_counter);
					iQsInstance->Set_DivisorsOfA(divisors_of_a);
					//printf("\r%s: Saved. \n",DESERIALIZE_SIEVER_STATE,a_counter);
					a_counter = 0;  // Not to write the message again
				}
			}
			// Deserialize usable, sol_1 and sol_2...
			if (xmlStrEqual(name,BAD_CAST ALLOC_SOLUTIONS_NAME))
			{
				// We will allocate the solutions
				iQsInstance->Clear_Solutions();
				iQsInstance->AllocSolutions();
				// Returns sols by value ...
				sols = iQsInstance->Get_Solutions();
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,sols.allocated,FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST USABLE_NAME))
			{
				//printf("Reading element b2[%d].\n",b2_counter);
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,sols.usable[usable_counter++],FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST SOL_1_NAME))
			{
				//printf("Reading element b2[%d].\n",b2_counter);
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,sols.sol_1[sol_1_counter++],FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST SOL_2_NAME))
			{
				//printf("Reading element b2[%d].\n",b2_counter);
				rc = CommonXMLHelper::DeserializeSignedInt(reader,DESERIALIZE_SIEVER_STATE,sols.sol_2[sol_2_counter++],FALSE);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (sol_1_counter == (max_fb_index+1))
			{
				printf("\r%s: Altogether %d sol_1's read. \n",DESERIALIZE_SIEVER_STATE,sol_1_counter);
				//printf("\r%s: Saved. \n",DESERIALIZE_SIEVER_STATE,a_counter);
				sol_1_counter = 0;  // Not to write the message again
			}
			if (sol_2_counter == (max_fb_index+1))
			{
				printf("\r%s: Altogether %d sol_2's read. \n",DESERIALIZE_SIEVER_STATE,sol_2_counter);
				//printf("\r%s: Saved. \n",DESERIALIZE_SIEVER_STATE,a_counter);
				sol_2_counter = 0;  // Not to write the message again
			}
			if ((x_partials != -1) && (smooths != -1))
			{
				iQsInstance->Revert(smooths,x_partials);
				x_partials = -1;
				smooths = -1;
			}
		}
       		ret = xmlTextReaderRead(reader);
		/*if ((ret == 0) || (ret == -1))
		{
			break;
		}*/
	}
	if (isSiqs)
	{
		// Will init the environment if necessary
		iQsInstance->NEXKSB_Set(env.n,env.k,env.h,env.m,env.subset,env.finished);
	}
	xmlFreeTextReader(reader);
	return ConstRC::Ok;
}

const char* 	XmlQsSerializator::AlgorithmName()       const
{
	return "MPQS/SIQS";
}

JobParameters*  XmlQsSerializator::CreateNewParameters() const
{
	return new QSParameters();
}

// Has no real meaning in Xml Qs Serializator
int	XmlQsSerializator::SetupParameters(const JobParameters& /*aParameters*/)
{
	return ConstRC::Ok;
}

int 	XmlQsSerializator::SerializeData(xmlTextWriterPtr& aWriter) ///< Used in parallelization to send fresh data from Node to Center.
{
	//printf("XMLQS SerializatorL this = %p: instance = %p\n",this,iQsInstance);
	if (iQsInstance)
	{
		// The data mutex has already been locked by the Node call.
		int rc = xmlTextWriterStartElement(aWriter, BAD_CAST RELATIONS_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Data",rc,0,"Could not write the relations tag.");
        		return GENERAL_ERROR;
    		}
		rc = CommonXMLHelper::NewLine(aWriter,FALSE);

		rc = xmlTextWriterStartElement(aWriter, BAD_CAST SMOOTH_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Data",rc,0,"Could not write the smooth tag.");
        		return GENERAL_ERROR;
    		}
		// There must not be NewLine after SMOOTH_NAME
		//rc = CommonXMLHelper::NewLine(aWriter,FALSE);

		// Useful for XmlQsSerializator and distribution
		int resClose1 = iQsInstance->CloseDataFile(ESmooth);
		if (resClose1 == ConstRC::Ok)
		{
			FILE* smFile = iQsInstance->OpenDataFile(ESmooth, READ);
			if (smFile)
			{
				CopyTextFileToWriter(aWriter, smFile);
				// The smooths have been copied.
				int resClose2 = iQsInstance->CloseDataFile(ESmooth);
				// Now, we erase the file by reopening it for writing
				iQsInstance->OpenDataFile(ESmooth, WRITE);
			}
		} else
		{
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Data",rc,0,"Could not close the smooth file.");
		}
		rc = xmlTextWriterEndElement(aWriter); // smooth
		rc = CommonXMLHelper::NewLine(aWriter,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		rc = xmlTextWriterStartElement(aWriter, BAD_CAST PARTIAL_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Data",rc,0,"Could not write the partial tag.");
        		return GENERAL_ERROR;
    		}
		// There must not be NewLine after PARTIAL_NAME
		//rc = CommonXMLHelper::NewLine(aWriter,FALSE);
		int resClose2 = iQsInstance->CloseDataFile(EPartial);
		if (resClose2 == ConstRC::Ok)
		{
			FILE* paFile = iQsInstance->OpenDataFile(EPartial, READ);
			if (paFile)
			{
				CopyTextFileToWriter(aWriter, paFile);
				// The partials have been copied.
				int resClose2 = iQsInstance->CloseDataFile(EPartial);
				// Now, we erase the file by reopening it for writing
				iQsInstance->OpenDataFile(EPartial, WRITE);
			}
		} else
		{
	        	CLog::LogError(WHEREARG,"XML Qs Serializator","Serialize Data",rc,0,"Could not close the partial file.");
		}

		rc = xmlTextWriterEndElement(aWriter); // partial
		rc = CommonXMLHelper::NewLine(aWriter,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		rc = xmlTextWriterEndElement(aWriter); // relations
		rc = CommonXMLHelper::NewLine(aWriter,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		return ConstRC::Ok;


	} else
	{
		exit(1);
	}
}

void	XmlQsSerializator::PrintInstance() const
{
	printf("*************** XmlQsSerializator: %p, instance %p, qs %p.\n",this,iAlgorithmInstance, iQsInstance);
}


