#include <cmath>
#include <cstring> 
#include <iostream>
#include <fstream>
#include "CL/cl.h"
#include "gmp.h"
#include "ECM.h"
#include <time.h>       /* clock_t, clock, CLOCKS_PER_SEC */

#define SUCCESS 0
#define FAILURE 1

using namespace std;

/*!
Convert number x to Montgomery representation, i.e. to x*R mod N
*/
int ECM::toMontgomery(mpz_t input, mpz_t * output)
{
	mpz_t rm;
	mpz_init_set_ui(rm, 1);
	mpz_mul_2exp(rm, rm, r * 32);
	mpz_mul(*output, input, rm);
	mpz_mod(*output, *output, actualN);
	mpz_clear(rm);
	return SUCCESS;
}

/*!
Convert number x from Montgomery representation, i.e. to x*R^(-1) mod N
*/
int ECM::fromMontgomery(mpz_t input, mpz_t * output)
{
	mpz_t rm, inv, x;
	mpz_init_set_ui(rm, 1);
	mpz_mul_2exp(rm, rm, r * 32);
	mpz_init(inv);
	mpz_init(x);
	mpz_invert(inv, rm, actualN); //set inv to inverse of rm mod N, allways 0 < inv <actualN
	mpz_mul(*output, input, inv);
	mpz_mod(*output, *output, actualN);
	mpz_clear(rm);
	mpz_clear(inv);
	mpz_clear(x);
	return SUCCESS;
}

/*!
Converts number from mpz_t representation to my_mpz reprezentation, which is used at kernel
*/
int ECM::mpzToMyMpz(mpz_t input, my_mpz * output)
{
	mpz_t quot, aux, base, rem;
	mpz_init(rem);
	//	output->numCoeff = 0;
	mpz_init(quot);
	mpz_init_set_ui(base, MAX);
	mpz_add_ui(base, base, 1); //base is MAXINT + 1
	mpz_init_set(aux, input); //aux is number we want to change 
	unsigned int remint;
	//unsigned long int mpz_tdiv_qr_ui (mpz_t q, const mpz_t n, unsigned long int d)  n = q*d + r
	for (int i = 0; i < DELKA; i++)
	{
		mpz_tdiv_r(rem, aux, base);
		mpz_tdiv_q(quot, aux, base);
		//rem = mpz_tdiv_q_ui(quot, aux, base); 
		mpz_set(aux, quot);
		remint = mpz_get_ui(rem);
		output->coeff[i] = remint;
		//if(rem != 0)
		{
			//			(output->numCoeff)++;
		}
	}
	mpz_clear(quot);
	mpz_clear(aux);
	mpz_clear(base);
	mpz_clear(rem);
	return SUCCESS;
}

/*!
Computes N' which is used in formulas for Montgomery's multiplication
*/
int ECM::computeNdashed()
{
	mpz_t rm, Ndashm, var1;
	mpz_init(Ndashm); mpz_init(var1);
	mpz_init_set_ui(rm, 1);
	mpz_mul_2exp(rm, rm, r * 32);
	mpz_invert(Ndashm, actualN, rm); //set Ndashm to inverse of N mod rm, allways 0 < Ndashm < rm
	mpz_neg(Ndashm, Ndashm);
	mpz_mod(Ndashm, Ndashm, rm);
	//mpz_out_str(NULL, 10, Ndashm);
	N_dashed = new my_mpz;
	mpzToMyMpz(Ndashm, N_dashed);
	return SUCCESS;
}

/*!
Get all primes smaller then given bound B1
Sieve of Erastothenes is used
*/
void ECM::getPrimes() 
{
	bool * isComposite = new bool[B1 + 1];
	memset(isComposite, 0, sizeof(bool) * (B1 + 1)); //get false into the array
	primesLength = B1 - 1;
	for (unsigned m = 2; m <= sqrt(double(B1)); m++)
	{
		if (!isComposite[m]) //then it is prime number
		{
			//do something, because it is prime number?
			for (unsigned k = m + m; k <= B1; k += m)
			{
				if (!(isComposite[k]))
					primesLength--;
				//set all his multiples as composite numbers
				isComposite[k] = true;
			}
		}
	}
	primes = new unsigned int[primesLength];
	unsigned int i = 0;
	//cout << "\n";
	unsigned int k;
	for (unsigned m = 2; m < B1; m++)
		//copy only prime numbers into array primes
		if (!isComposite[m])
		{
			//save p_i^k for k maximal such that p_i^k < B1
			k = (log2(B1) / log2(m));
			primes[i] = pow(m, k);
			i++;
		}
	delete[] isComposite;
}

/*!
Generates random numbers for factoring for testing purposes.
These numbers are products of just two primes. The length of
numbers depends on parameter given by user.
*/
int ECM::generateN()
{
	N = new mpz_t[quantity];
	gmp_randstate_t state; //TODO taky jako promnn ECM?
	gmp_randinit_default(state);
#ifdef RANDOM_N
	long seed = time(NULL);
	gmp_randseed_ui(state, seed);
#endif
	for (int i = 0; i < quantity; i++)
	{
		mpz_t number1, number2;
		mpz_init(number1);
		mpz_init(number2);
		//generate random number, it's length depends on size
		mpz_urandomb(number1, state, size);
		mpz_urandomb(number2, state, size);
		//get next prime number, so N will be product of 2 numbers only
		mpz_nextprime(number1, number1);
		mpz_nextprime(number2, number2);
		mpz_mul(number1, number1, number2); 
		mpz_init_set(N[i], number1);
		mpz_clear(number1);
		mpz_clear(number2);
	}
	return 0;
}

/*!
Function for loading kernel file - convert the kernel file into a string
*/
int ECM::convertToString(const char *filename, std::string& s)
{
	size_t size;
	char*  str;
	std::fstream f(filename, (std::fstream::in | std::fstream::binary));

	if (f.is_open())
	{
		size_t fileSize;
		f.seekg(0, std::fstream::end);
		size = fileSize = (size_t)f.tellg();
		f.seekg(0, std::fstream::beg);
		str = new char[size + 1];
		if (!str)
		{
			f.close();
			return 0;
		}

		f.read(str, fileSize);
		f.close();
		str[size] = '\0';
		s = str;
		delete[] str;
		return 0;
	}
	cout << "Error: failed to open file\n:" << filename << "\n";
	return FAILURE;
}
