#include <iostream>
#include <cstring>
#include <string>
#include "CL/cl.h"
#include "gmp.h"
#include "ECM.h"
#include "ECM_parameters.h"
#include "ECM_const.h"

#define SUCCESS 0
#define FAILURE 1

using namespace std;

/*!
Set parameters of ECM (numbers for factorization, B1 and number of curves) 
in the class ECM when param. were read from console
*/
int ECM::setup(ECMparameters * parameters)
{
	//set test to true, if testing ECM 
	if (parameters->test)
	{
		test = true;
		size = parameters->size;
		B1 = parameters->B1;
		NumCurves = parameters->curves;
		quantity = parameters->quantity;
		
	}
	else
	{
		test = false;
		N = parameters->N;
		quantity = parameters->quantity;
		B1 = parameters->B1;
		NumCurves = parameters->curves;
	}
	return 0;
}
/*!
Setup the bound B1 and number of curves that should be use.
Parameters will be used for all numbers until they are changed.
*/
int ECM::setup(int aB1, int aNumCurves)
{
	B1 = aB1;
	getPrimes();
	NumCurves = aNumCurves;
	return 0;
}

/*!
Setup the number, the bound B1 and number of curves that should be use.
Parameters will be used for all numbers until they are changed.
*/
int ECM::setup(mpz_t aN, int aB1, int aNumCurves)
{
	mpz_set(*N,aN);
	setup(aB1, aNumCurves);
	return 0;
}




/*!
* OpenCL related initialisations.
* Choose platform, device. 
* Set up Context, Device list, Command Queue, Memory buffers
* Build CL kernel program executable from source
*/
int ECM::setupCL()
{
	/*Step1: Getting platforms and choose an available one.*/
	cl_uint numPlatforms;	//the NO. of platforms
	cl_platform_id platform = NULL;	//the chosen platform
	char* info;
	size_t infoSize;
	cl_int	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	if (status != CL_SUCCESS)
	{
		cout << "Error: Getting platforms!" << "\n";
		return FAILURE;
	}

	if (numPlatforms == 0)
	{
		cout << "No platform avaiable.";
		exit(0);
	}

	/*Choose platform. */
#ifdef USER_CHOICE /* Choose the platform by hand */
	cl_platform_id* platforms = (cl_platform_id*)malloc(numPlatforms* sizeof(cl_platform_id));
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	for (int i = 0; i < numPlatforms; i++)
	{
		// get platform attribute value size
		clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 0, NULL, &infoSize);
		info = (char*)malloc(infoSize);
		// get platform attribute value
		clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, infoSize, info, NULL);

		cout << "platforma " << i << " " << info << "\n";
		free(info);
	}
	cout << "Enter the number of platform you want to choose: ";
	int platforma;
	/*if (test)
	platforma = 1;
	else*/
	cin >> platforma;
	platform = platforms[platforma];
	free(platforms);
#endif
#ifdef NVIDIA_CUDA
	cl_platform_id* platforms = (cl_platform_id*)malloc(numPlatforms* sizeof(cl_platform_id));
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	if (numPlatforms == 1)
	{
		platform = platforms[0];
		free(platforms);
	}
	/*Find the first with GPU device if has one. Otherwise use the first platform.*/
	else
	{
		for (int i = 0; i < numPlatforms; ++i)
		{
			clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 0, NULL, &infoSize);
			info = (char*)malloc(infoSize);
			// get platform attribute value
			clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, infoSize, info, NULL);
			if ( strcmp(info, "NVIDIA CUDA") == 0) 
			{
				platform = platforms[i];
				free(platforms);
				break;
			}
			free(info);
		}
		if (platform == NULL) //platform with GPU not found - take the first one
		{
			platform = platforms[0];
			free(platforms);
		}
	}
	// get platform attribute value size
	clGetPlatformInfo(platform, CL_PLATFORM_NAME, 0, NULL, &infoSize);
	info = (char*)malloc(infoSize);
	// get platform attribute value
	clGetPlatformInfo(platform, CL_PLATFORM_NAME, infoSize, info, NULL);
	cout << "Chosen platform:  " << info << endl;
	free(info);
#endif
#if !defined(NVIDIA_CUDA) && !defined(USER_CHOICE) 
	cl_platform_id* platforms = (cl_platform_id*)malloc(numPlatforms* sizeof(cl_platform_id));
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	if (numPlatforms == 1)
	{
		platform = platforms[0];
		free(platforms);
	}
	/*Find the first with GPU device if has one. Otherwise use the first platform.*/
	else
	{
		for (unsigned i = 0; i < numPlatforms; ++i)
		{
			numDevices = 0;
			status = clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 0, NULL, &numDevices);
			if (numDevices != 0) //GPU found
			{
				platform = platforms[i];
				free(platforms);
				break;
			}
		}
		if (platform == NULL) //platform with GPU not found - take the first one
		{
			platform = platforms[0];
			free(platforms);
		}
	}
	// get platform attribute value size
	clGetPlatformInfo(platform, CL_PLATFORM_NAME, 0, NULL, &infoSize);
	info = (char*)malloc(infoSize);
	// get platform attribute value
	clGetPlatformInfo(platform, CL_PLATFORM_NAME, infoSize, info, NULL);
	cout << "Chosen platform:  " << info << endl;
	free(info);

#endif
	/* Platform chosen */
	/*Choose device */
	status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &numDevices);
	if (numDevices == 0)	
	{
		cout << "No GPU device available." << endl;
		cout << "Choose CPU as default device." << endl;
		status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, 0, NULL, &numDevices);
		//TODO oetit, kdyby platforma nebyla ani GPU ani CPU
		devices = (cl_device_id*)malloc(numDevices * sizeof(cl_device_id));
		status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, numDevices, devices, NULL);
	}
	else
	{
		cout << "GPU found \n";
		status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &numDevices);
		devices = (cl_device_id*)malloc(numDevices * sizeof(cl_device_id));
		status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, numDevices, devices, NULL);

		/****************************************GET INFO*******************************/
		// Get platform name
		char platformVendor[1024];
		status = clGetPlatformInfo(platform, CL_PLATFORM_VENDOR, sizeof(platformVendor), platformVendor, NULL);

		std::cout << "Selected Platform Vendor : " << platformVendor << "status: " << status << endl;

		// Get number of devices available 
		cl_uint deviceCount = 0;
		status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &deviceCount);

		cl_device_id* deviceIds = (cl_device_id*)malloc(sizeof(cl_device_id) * deviceCount);

		// Get device ids
		status = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, deviceCount, deviceIds, NULL);
		cout << "clGetDeviceIDs status: " << status << endl;
		if (status < 0)
		{
			cout << "Couldn't get device ID" << endl;
		}

#ifdef PRINT_DEVICES_INFO
		
		for (cl_uint i = 0; i < deviceCount; ++i)
		{
			char deviceName[1024];
			status = clGetDeviceInfo(deviceIds[i], CL_DEVICE_NAME, sizeof(deviceName), deviceName, NULL);
			//get device mem size
			cl_ulong mem;
			status = clGetDeviceInfo(deviceIds[i], CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &mem, NULL);

			std::cout << "Device " << i << " : " << deviceName << " Device ID is " << deviceIds[i] << " av. memory is: " << mem << "\n";
			status = clGetDeviceInfo(deviceIds[i], CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof(cl_ulong), &mem, NULL);
			cout << "Max mem alloc size is :" << mem << "\n";
			status = clGetDeviceInfo(deviceIds[i], CL_DEVICE_LOCAL_MEM_SIZE, sizeof(cl_ulong), &mem, NULL);
			cout << "Local mem size is: " << mem << "\n";
		
		}
#endif

		free(deviceIds);
	}


	/*Step 3: Create context.*/
	context = clCreateContext(NULL, 1, devices, NULL, NULL, &status);
	cout << "Create context status: " << status << "\n";
	if (status < 0)
	{
		cout << "Couldn't create a context";
		exit(1);
	}

	/*Step 4: Creating command queue associate with the context.*/
	commandQueue = clCreateCommandQueue(context, devices[0], CL_QUEUE_PROFILING_ENABLE, &status);
	cout << "Create command queue status: " << status << "\n";
	if (status < 0)
	{
		cout << "Couldn't create a command queue";
		exit(1);
	}

	/*Step 5: Create program object */
	const char *filename = "ECM_Kernels.cl"; /*TODO  full path for profiling: C:\\Users\\Adla\\Documents\\MFF\\faktorizace\\ECM\\ECM\\  */

	string sourceStr;
	status = convertToString(filename, sourceStr);
	const char *source = sourceStr.c_str();
	size_t sourceSize[] = { strlen(source) };
	program = clCreateProgramWithSource(context, 1, &source, sourceSize, &status);
	cout << "Create program status: " << status << "\n";
	if (status < 0)
	{
		cout << "Couldn't create program";
		exit(1);
	}

	/*Step 6: Build program. */
	status = clBuildProgram(program, 1, devices, NULL, NULL, NULL); //"-g -s C:\\Users\\Adla\\Documents\\MFF\\faktorizace\\ECM\\ECM\\ECM_kernels.cl"
	cout << "Build program status: " << status << "\n";
	if (status < 0) {                                       //originally: if (status == CL_BUILD_PROGRAM_FAILURE) {
		cout << "Couldn't build program. Log:" << endl;
		// Determine the size of the log
		size_t log_size;
		clGetProgramBuildInfo(program, devices[0], CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);

		// Allocate memory for the log
		char *log = (char *)malloc(log_size);

		// Get the log
		clGetProgramBuildInfo(program, devices[0], CL_PROGRAM_BUILD_LOG, log_size, log, NULL);

		// Print the log
		printf("%s\n", log);
		exit(1);
	}
}

