#include "abstract_connector.h"
#include "abstract_sieve.h"
#include <string.h>
#include <stdlib.h>
#include <algorithm>
#include "log.h"


/*!
 * \brief Compares messages
 *
 * taken from http://www.codeproject.com/KB/stl/stdsort.aspx
 */
struct SCompareMessagesByCounter
{
     bool operator()(AbstractMessage* const& aArg1, AbstractMessage* const& aArg2)
     {
	  // The comparison must be >, not >=
	  // See also http://groups.google.com/group/comp.lang.c++/browse_thread/thread/2e1268fd88c40b5e
          return (aArg1->Get_Counter() > aArg2->Get_Counter());
     }
};


// Forward declaration
class AbstractSieve;

AbstractConnector::AbstractConnector()
{
	Init();
}

AbstractConnector::~AbstractConnector()
{
	int i, size = iIncomingQueue.size();
	for (i=0; i< size; i++)
	{
		delete iIncomingQueue[i];
	}	
	delete iURL;
	iURL = NULL;
}

void AbstractConnector::Init()
{
	iURL 	       = NULL;
	iSortByCounter = TRUE;
	iConnectorType = EUnknownConnector;
}

AbstractConnector::AbstractConnector(const char* aURL, const char* aIdentification)
{
	Init();
	iURL = new URL(aURL);
	iURLString.assign(aURL);
	iIdentification.assign(aIdentification);
}

AbstractConnector::AbstractConnector(const string& aURL, const char* aIdentification)
{
	Init();
	iURL = new URL(aURL);
	iURLString.assign(aURL);
	iIdentification.assign(aIdentification);
}

AbstractConnector::AbstractConnector(const string& aURL, const string& aIdentification)
{
	Init();
	iURL = new URL(aURL);
	iURLString.assign(aURL);
	iIdentification.assign(aIdentification);
}

connector_type AbstractConnector::Get_ConnectorType() const
{
	return iConnectorType;
}

void 	AbstractConnector::Set_Identification(const char* aValue)
{
	iIdentification.assign(aValue);
}

const char*   AbstractConnector::Get_Identification() const
{
	return iIdentification.c_str();
}

URL*  AbstractConnector::Get_URL() const
{
	return iURL;
}

string AbstractConnector::Get_URLString() const
{
	return iURLString;
}

vector<AbstractMessage*>* AbstractConnector::Sort(bool aFilterType, message_type aType)
{	
	int size = iIncomingQueue.size();
	// Empty queue
	if (size == 0) {		
		return NULL;
	}

	int i;
	//printf("Abstract Connector::Sort. Size = %d\n",size);
	vector<AbstractMessage*>* field = new vector<AbstractMessage*>();
	for (i=0; i< size; i++)
	{
		AbstractMessage* ptr = iIncomingQueue[i];	
		if (aFilterType)
		{
			if (ptr->Get_Type() == aType) 
			{
				field->push_back(ptr);
			}
		} else
		{	// We take all
			field->push_back(ptr);
		}
	}
	/*printf("Abstract Connector::Sort. Field done = %d. Calling qsort\n",field->size());
	AbstractMessage* begin = *field->begin();
	
	printf("Abstract Connector::Sort Begin = %p, counter = %d.\n",begin,begin->Get_Counter());*/
	//printf("Abstract Connector::Sort End   = %p.\n",field->end());
	// Now, calling qsort
	if (field->size() > 1)
	{
		if (iSortByCounter)
		{
			// Taken from http://www.codeproject.com/KB/stl/stdsort.aspx
			std::sort(field->begin(),field->end(),SCompareMessagesByCounter());
			// Sort by counter	
			
		}
		else {
			printf("Abstract connector: sort by other means needed");			
			// Sort by time?			
		}
	}
	return field;	
}

void	AbstractConnector::Set_SortByCounter(bool aValue)
{
	this->iSortByCounter = aValue;
}

bool	AbstractConnector::Get_SortByCounter() const
{
	return this->iSortByCounter;
}

vector<AbstractMessage*>*	AbstractConnector::ReceiveAll(message_type aType)
{		
	ReceiveNow();
	vector<AbstractMessage*>* result = Sort(aType);
	DeleteFromIncomingQueue(result);
	return result;
}

vector<AbstractMessage*>*	AbstractConnector::ReceiveAll()
{
	ReceiveNow();
	vector<AbstractMessage*>* result = Sort(FALSE);
	DeleteFromIncomingQueue(result);
	return result;
}

vector<AbstractMessage*>*	AbstractConnector::ReceiveOne(message_type aType)
{		
	ReceiveNow();	
	return Sort(aType);
}

int	AbstractConnector::Send(AbstractMessage* aMessage)
{
	bool pushedToPending = FALSE;
	if (aMessage == NULL)
	{
		return BAD_ARGUMENT;
	}
	if (aMessage->Has_Timeout() || aMessage->Has_Delay())
	{
		// Put to the vector of ready messages
		iPendingQueue.push_back(aMessage);
		pushedToPending = TRUE;
	}
	// Set the time of the sending.
	if (aMessage->Get_Delay() == NO_DELAY_SEND)
	{
		aMessage->Set_CurrentTime();
		DoSend(aMessage);
	} 
        else // a Message with delay. We will not send it immediately
	{
		aMessage->Set_FutureTime(aMessage->Get_Delay());
	}
	if (!pushedToPending)
	{
		if (aMessage->Can_Destroy())
		{
			delete aMessage;		
		}
	}
	return ConstRC::Ok;
}

bool	AbstractConnector::IsMine(const AbstractMessage* aMessage) const
{
	if (!aMessage)
	{
		printf("AbstractConnector::IsMine: error. Message is NULL\n");
		return FALSE;
	}
	const char* target = aMessage->Get_TargetId();
	if (!target)
	{		
		printf("AbstractConnector::IsMine: error. Target NULL\n");
		return FALSE;
	}
	if (iIdentification == target)
	{		
		return TRUE;
	}	
	printf("AbstractConnector::IsMine: error. Target != ident :  %s vs. %s\n",target, iIdentification.c_str());
	aMessage->PrintHeaderInfo();
	return FALSE;
}


int	AbstractConnector::ResendPending()
{
	if (iPendingQueue.empty())
	{
		return 0;
	}
	int count = 0;
	AbstractMessage* mess;
	time_t lastSent;
	time_t currentTime;
	int diff;
	bool hasTimeout;
	bool hasDelay;
	vector<AbstractMessage*>::iterator iter = iPendingQueue.end();
	vector<AbstractMessage*>::iterator start = iPendingQueue.begin();	
	do 
	{
		iter--;
		mess = (*iter);
		hasTimeout = mess->Has_Timeout();
		hasDelay   = mess->Has_Delay();
		lastSent = mess->Get_Time();
		currentTime = time(NULL);
		diff = currentTime-lastSent;
		if (diff < 0)
		{
			// A future message
// 			printf("AbstractConnector::Resend pending. We have a future message in the queue, %d seconds to send it remain.\n",diff);
		} 
 		else // diff >= 0 
		{
			// At first, we take care of delayed messages
			if (hasDelay)
			{
				// It is time to send it
				if (diff >= 0)
				{				
					mess->Set_CurrentTime();
					// This message is no more delayed.
					mess->Set_Delay(NO_DELAY_SEND);
					DoSend(mess);
					if (!hasTimeout)
					{
						// remove from the pending queue and delete the pointer
						iPendingQueue.erase(iter);
 						if (mess->Can_Destroy())
						{		
							delete mess;
						}
					}
					continue; // go to next message
				}
			}
			else if (hasTimeout) // not a depal
			{
				if (((unsigned int) diff) > mess->Get_Timeout())
				{
					printf("AbstractConnector: Resend pending. Resent message with counter %d, since %d seconds have elapsed from its last sending.\n",mess->Get_Counter(),diff);
					mess->Set_CurrentTime();
					DoSend(mess);
				}
			}
		}
		
	} while (iter != start);
	return count;
}

int	AbstractConnector::DeleteFromPending(const char* aTargetId, unsigned int aCounter)
{
	if (iPendingQueue.empty())
	{
		return 0;
	}
	if (!aTargetId)
	{
		return 0;
	}
	// end() gives a position behind the vector, not the last element.
	vector<AbstractMessage*>::iterator iter = iPendingQueue.end();
	vector<AbstractMessage*>::iterator start = iPendingQueue.begin();
	int count = 0;
	do 
	{
		iter--;
		if ((*iter)->Get_Counter() == aCounter)
		{
			const char* target = (*iter)->Get_TargetId();
			if ((target != NULL) && (!strcmp(target,aTargetId)))
			{
				delete *iter; // Destroy the object AbstractMessage*
			// Do something
				iPendingQueue.erase(iter);
				count++;
// 				printf("AbstractConnector::DeleteFromPending. Removed 1 message from pending\n");
			}
		}
	} while (iter != start);
	return count;
}

int	AbstractConnector::DeleteFromPending_Type(message_type aType)
{
	if (iPendingQueue.empty())
	{
		return 0;
	}
	// end() gives a position behind the vector, not the last element.
	vector<AbstractMessage*>::iterator iter = iPendingQueue.end();
	vector<AbstractMessage*>::iterator start = iPendingQueue.begin();
	int count = 0;
	do 
	{
		iter--;
		if ((*iter)->Get_Type() == aType)
		{
			delete *iter; // Destroy the object AbstractMessage*
			// Do something
			iPendingQueue.erase(iter);
			count++;

		}
	} while (iter != start);
	return count;
}

int	AbstractConnector::DeleteFromIncomingQueue(vector<AbstractMessage*>* aList)
{
	if (!aList)
	{
// 		printf("AbstractConnector::DeleteFromIncomingQueue: vector is NULL\n");
		return 0;
	}
	if (aList->empty())
	{
// 		printf("AbstractConnector::DeleteFromIncomingQueue: vector is empty\n");
		return 0;
	}
	vector<AbstractMessage*>::iterator iter = aList->end();
	vector<AbstractMessage*>::iterator start = aList->begin();	
	int count = 0;
	do 
	{
		iter--;
// 		printf("AbstractConnector::DeleteFromIncomingQueue: will remove message:\n");
		//(*iter)->PrintHeaderInfo();
		count += DeleteFromIncomingQueue(*iter);
	} while (iter != start);
// 	printf("AbstractConnector::DeleteFromIncomingQueue: finished erasing from vector\n");
	return count;

}

int	AbstractConnector::DeleteFromIncomingQueue(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return 0;
	}
	// end() gives a position behind the vector, not the last element.
	vector<AbstractMessage*>::iterator iter = iIncomingQueue.end();
	vector<AbstractMessage*>::iterator start = iIncomingQueue.begin();	
	do 
	{
		iter--;
		if ((*iter) == aMessage)
		{
// 			printf("AbstractConnector::DeleteFromIncomingQueue: Found message in list, erasing...\n");
			iIncomingQueue.erase(iter);
			return 1;

		}
	} while (iter != start);
	return 0;
}

int	AbstractConnector::ClearPending()
{
	int i, size = iPendingQueue.size();
	AbstractMessage* mess;
	for (i= size-1; i>=0; i--)
	{
		mess = iPendingQueue.at(i);
		delete mess;
	}
	iPendingQueue.clear();
	return size;
}

int	AbstractConnector::AddReceivedMessage(AbstractMessage* aMessage)
{
	int result = 0;
	if (aMessage)
	{
		if (IsMine(aMessage))
		{
			iIncomingQueue.push_back(aMessage);
		} else	
		{
			CLog::LogError(WHEREARG,"Abstract connector::AddReceivedMessage","this message is not mine!",aMessage->Get_Counter(),0,"Check the target");
			if (aMessage->Can_Destroy())
			{
				delete aMessage;
				result |= MESSAGE_INSTANCE_DELETED;
			}
		}
	}
	return result;
}

int	AbstractConnector::GetState() const
{
	if (!iURL)
	{
		return KNullPointer;
	}
	if (iURLString.empty())
	{
		return KUrlEmpty;
	}
	if (iURL->GetParsingResult() != EOk)
	{
		return KUrlMalformed;
	}
	return KOk;
}

string	AbstractConnector::GetStateExplanation(int aState) const
{
	string res;
	switch (aState)
	{
		case KOk: 
			res = OkChar;
			break;
		case KGeneralInitError:
			res = General;
			break;
		case KNullPointer:
			res = NullPtr;
			break;
		case KUrlEmpty:
			res = UrlEmpty;
			break;
		case KUrlMalformed:
			res = Url;
			res += "'";
			res += iURLString;
			res += "'";
			res += IsMalformed;
			break;
	}
	return res;
}
