#ifndef ABSTRACT_CONNECTOR_H
#define ABSTRACT_CONNECTOR_H

#include "abstract_message.h"
#include "url.h"

#define DEFAULT_INCOMING_QUEUE_SIZE  40

#include <string>
#include <vector>

using namespace std;

// Forward declaration
class AbstractMessage;

// Possible error states
const int KOk = 0;
const int KGeneralInitError = -1;
const int KNullPointer	    = -2;
const int KUrlEmpty	    = -3;
const int KUrlMalformed	    = -4;

#define OkChar   "O.K."
#define General  "General Init Error"
#define NullPtr  "Null Pointer"
#define UrlEmpty "URL is empty"
#define Url	 "URL "
#define IsMalformed " is malformed"

/*!
 *
 *  \brief A basic connection class for communication between nodes/centers of a distributed process.
 * 
 * This abstract class serves as the base for all communication classes
 * in distributed computing. It contains as much common code as possible,
 * while leaving all the details of the concrete communication protocols 
 * (IMAP, SMTP, filesystem comm...) to its subclasses.
 *
 * The connector must know the id (name) of the current running instance.
 * Missing an id is a reason for failure of the connection.
 *
 * The main functionality of a Connector instance is to send and receive
 * AbstractMessages. This includes resending of messages that were sent already,
 * but for which no reply was received in given time interval (called "timeout").
 * 
 * For each message that is sent, the connector finds out whether it has 
 * a "timeout" set, or no. If no "timeout" is set (more precisely, the value
 * of "timeout" member in the AbstractMessage corresponds to the value NO_TIMEOUT),
 * the message is "fired and forgotten". If a timeout is set, the sent message
 * is placed into a list of "pending messages" This is usually the case of
 * messages for which we require a reply from the other side. If no reply
 * is received within the timeout interval, the message is re-sent, and the
 * timeout starts running again from the moment of resending.
 *
 * The method that resends the timeouted messages is "ResendPending"
 *
 * The abstract connector does not run in a separate thread, and invoking
 * the Receive() and ResendPending() methods periodically is a responsibility 
 * of its owner. 
 * This may change with introduction of threading, but so far it is far from
 * realization.
 *
 * In the future, the AbstractConnector class will support sorting of the received 
 * messages by given parameters. Currently, the message sequence is undefined.
 *
 * The abstract methods that need to be implemented by subclasses are the following:
 *
 * DoSend(AbstractMessage* aMessage), which sends a single message to its target.
 * 
 *         Concrete implementations of DoSend do not have to actually send the message
 *	     before the method returns; sometimes, it is more reasonable to send the
 *         messages in bursts (batches), say, 10 at a time (for example, sending
 *	     a lot of mail messages over SMTP protocol is more effective if we do
 *	     not perform the handshake and connect every time). So, instead, the
 *	     messages may be put into an outgoing queue and resent later. However,
 *	     the interval between calling DoSend() and the actual sending of the message
 *	     should be reasonable; most important, it should not exceed about a third
 * 	     of the minimal timeout value (defined as a symbolic constant). This all
 *	     is the responsibility of the concrete implementation; it is well possible
 *	     that yet another abstract subclass (ThreadedBufferedConnector) will be
 *	     created for that purpose.
 *		
 *	     However, if a connector intends to send the message later, it must take
 *	     some precautions. Normally, a message without timeout would be destroyed
 * 	     (the instance deleted) in the AbstractConnector::Send(AbstractMessage) method
 *	     immediately after calling the virtual DoSend() method. This would lead
 *	     to a memory error, since the sending thread would later access an already
 *	     destroyed AbstractMessage instance. That is why AbstractMessage class has
 * 	     a bool methods "Can_Destroy()" and "Set_CanDestroy(bool)". The default
 *	     is True (a message can be destroyed), and AbstractConnector will destroy
 * 	     all non-timeouted messages that return Can_Destroy() as True. It is 
 *         responsibility of the more concrete subclass of AbstractConnector to
 *	     set this property to False for messages that are not sent already, but
 *	     rather queued for further sending.
 *
 *
 * ReceiveNow(), which obtains freshly received messages from the connector.
 *
 * 		This method must fetch all freshly received messages from the connector,
 * 		for further use within the system. Note that in multi-threaded environment,
 *		only messages that have already been completely received by the connector
 *		count as "received". If the connector is in the middle of receiving a message
 *		when ReceiveNow() is called, it should not return the incomplete message
 *		to the caller.
 *
 * 		This method must always call the method "AddReceivedMessage(AbstractMessage*)
 *		 of the abstract connector, which checks whether the message belongs to this
 *		recipient (comparing the ids of this connector and the target), 
 *          and if it is true, adds it to the incoming queue; if it is not
 *		intended for this recipient, and can be destroyed (Can_Destroy()) is true),
 *		the pointer is deleted.
 *
 *		This should NOT be the only way
 *          of distinguishing the correct messages. It should only be used as 
 * 		a "last ditch " attempt to sort out any possible remaining messages
 *		for another recipient. The subclasses should make honest attempt
 *		to identify the correct messages before they fetch and deserialized them,
 *		for example from the filename or from the e-mail subject. This saves
 *		system resources.
 *
 *		For example, more running distributed processes may share a single IMAP
 *		account for the incoming traffic from nodes. Each process should be able
 *		to recognize "its own" messages from their headers only (for example,
 *		from specially formed Subjects), and fetch only those messages from the
 *		mail server.
 *
 *		If the method AddReceivedMessage() detects a foreign message, it is usually
 *		a programming error in the concrete subclass of the connector.
 *
 */
class AbstractConnector {

	public: 		
		AbstractConnector();
		AbstractConnector(const char*   aURL, const char* aIdentification);
		AbstractConnector(const string& aURL, const char* aIdentification);
		AbstractConnector(const string& aURL, const string& aIdentification);

		virtual ~AbstractConnector();

		void    Set_URL(const char* aURL);
		void	Set_URL(const string& aURL);
		URL*    Get_URL() const;
		string  Get_URLString() const;


		// Returns TRUE if this connector can be used to receive messages
		virtual bool SupportsReceive() const = 0;
		// Returns TRUE if this connector can be used to send messages
		virtual bool SupportsSend() const = 0;

		void 	Set_Identification(const char* aValue);
		void	Set_SortByCounter(bool aValue);

		const char* 	Get_Identification() const;
		bool	Get_SortByCounter() const;

		/*! This method will identify the current instance of the
		 * connector without need to call typeid() or dynamic_cast
             */
		connector_type Get_ConnectorType() const;

		/*!
		 * 	This method will send the defined message. It checks the 
             	 *    timeout of the message, it puts the message to the pending
		 *    list (if it has a timeout), and invokes "DoSend" on the 
             	 *    message.
		 */
		int	Send(AbstractMessage* aMessage);

		// Returns either EOk, or one of the more specific error states.
		virtual int	GetState() 	      		const;
		// Returns human-readable explanation of the error state.
		virtual string	GetStateExplanation(int aState) const;

		/*!
			This method will check all pending messages
			for timeout, send them again if timeouted.
	
			It will also send "delayed" messages (those that
			need to be sent in the future, not immediately).
                 */
		int	ResendPending();

		int	DeleteFromPending(const char* aTargetId, unsigned int aCounter);
		int	DeleteFromPending_Type(message_type aType);
		
		int	ClearPending();

		/*!
		 *	This method will check whether some messages
                 *      of this type exist.
		 *  	If there is none, it returns NULL
		 * 	If there is one or more, it returns a list of them,
		 * 	sorted by their counter. The pointers to the returned messages are deleted from
		 *      the current incoming queue. It is the responsibility
 		 *      of the calling code to destroy them when they are
                 *      no longer needed.
		 * 
		 */
		virtual vector<AbstractMessage*>*	ReceiveAll(message_type aType);

		virtual vector<AbstractMessage*>*	ReceiveAll();
		
		/*!
 		 *	Same as ReceiveAll, but it does not return all the
		 * 	messages. Only the oldest message of that type is
 		 *	returned (basically, a one-entry field).
		 *
		 * 
		 *	For some types of messages, the returned list may be
		 * 	sorted by value of message->Get_Counter() instead
		 * 	of time.
		 */
		virtual vector<AbstractMessage*>*	ReceiveOne(message_type aType);

	private:

		bool	IsMine(const AbstractMessage* aMessage) const;
		int	DeleteFromIncomingQueue(vector<AbstractMessage*>* aList);
		int	DeleteFromIncomingQueue(AbstractMessage* aMessage);

	protected:
	
		

		/*!
		 * 	This method will send the defined message.     
		 *	Pure virtual in the abstract class.
		 */
		virtual int			DoSend(AbstractMessage* aMessage) = 0;


		/*!
		 * Will check for incoming messages NOW. Usually called
		 * from ReceiveAll, ReceiveOne prior to Sort().
		 *  
		 * This method is not intended to create new threads.
		 * All concrete subclasses of this class MUST overload this
		 * method, even if it should be empty.
		 *
		 * 
		 */
		virtual int			ReceiveNow() = 0;
		virtual vector<AbstractMessage*>* Sort(bool aFilterType = TRUE, message_type aType = EUnknown);	
		virtual void			Init();	
	
		virtual int			AddReceivedMessage(AbstractMessage* aMessage);

	protected:
		// Member variables
		URL*			 iURL;
		string			 iURLString;
		
		string			 iIdentification; ///< The distinguishing parameter (most often, a node/center name).
		vector<AbstractMessage*> iPendingQueue; ///< A list of messages that have been sent, their timeout has not yet run out, and no acknowledgement/reply has been received for them. Messages with timeout value NO_TIMEOUT never come to this list; they are "fired and forgotten".
		bool			 iSortByCounter;
		connector_type		 iConnectorType;

	private:
		vector<AbstractMessage*> iIncomingQueue;		
};
#endif
