/*
 * abstract_integer_matrix.cpp
 *
 *      Author: jero
 */

#include "abstract_integer_matrix.h"
#include <iostream>
#include "log.h"

AbstractIntegerMatrix::AbstractIntegerMatrix() {
	mpz_init(tempadd);
	mpz_init(tempmul);
	mpz_init(tempval);
	mpz_init(tempval2);
	mpz_set_ui(tempadd,0);
	mpz_set_ui(tempmul,0);
	mpz_set_ui(tempval,0);
	mpz_set_ui(tempval2,0);
}

AbstractIntegerMatrix::~AbstractIntegerMatrix() {
	mpz_clear(tempadd);
	mpz_clear(tempmul);
	mpz_clear(tempval);
	mpz_clear(tempval2);
}

int AbstractIntegerMatrix::Copy(AbstractIntegerMatrix *aSource) {
	for(long row=0; row <=GetMaximalRowIndex(); row++)
		for (long column=0; column <= GetMaximalColumnIndex(); column++) {
			integer_matrix_type value;
			aSource->GetValue(row, column, value);
			PutValue(row, column, value);
	};
	ConstRC::Ok;
}

int AbstractIntegerMatrix::PartialCopy(AbstractIntegerMatrix* aSource,
		long aMaxRow, long aMaxColumn) {
	for (long row = 0; row <=aMaxRow; row++)
		for (long column=0; column <= aMaxColumn; column++) {
			integer_matrix_type value;
			aSource->GetValue(row, column, value);
			PutValue(row, column, value);
	};
	ConstRC::Ok;
}

/**
 * Iterates through nonzero items on a row. Parameters
 * \param aRow row index not needed and ignored if aRowIndex initialized by previous call
 * \param aRowIndex internal row index, should be set to -1 if not known for this row, otherwise used from previous call
 * \param aColumn column index (not internal sparse index, but full index)
 * \param aValue value
 */
long AbstractIntegerMatrix::IterateRow(long aRow,long & aRowIndex, long aColumnIndex, long & aColumn, integer_matrix_type & aValue) {
	if (aRow == GetMaximalRowIndex()) {
		aColumn = -1;
		CLog::LogError(WHEREARG,"IntegerSparseMatrix","IterateRow",aRowIndex,aRow,"Bad row index.");
		return ConstRC::ExcessiveRowIndex;
	}
	return ConstRC::NotSupported;		
};

/*!
	  This method prints the calling instance onto the screen. It is suitable for printing matrices with entries smaller
	  than 3 decimal digits.

	  This method does not change any data.

*/
void AbstractIntegerMatrix::PrintToScreen(){
	printf("\n");
	printf("Abstract integer matrix: %ld x %ld\n",this->GetMaximalRowIndex()+1,this->GetMaximalColumnIndex()+1);
	for (long i=0; i<=this->GetMaximalRowIndex(); i++){
		printf("\nRow %4ld:   ",i);
		for (long j=0; j<=this->GetMaximalColumnIndex(); j++){
			integer_matrix_type value;
			int rv = GetValue(i,j, value);
			printf("%3u",value);
		}
	}
	printf("\n\n");
}


int AbstractIntegerMatrix::Gauss(long aLimit, long aRowIndices[], long & aRank, bool aPartial) {
	// PrintToScreen();
	long limit = aLimit;
	// if given aRowIndices, initialize it to identity permutation
	if (aRowIndices) {
		for(long i = 0;i<= GetMaximalRowIndex(); i++) {
			aRowIndices[i]=i;
		};
	};
	if (limit >= GetMaximalColumnIndex()) limit = GetMaximalColumnIndex();
	int independent_rows = 0;
	cout << endl;
	for (long column_loop= 0; column_loop <= limit; column_loop++) {
		// PrintToScreen();
		long column = limit-column_loop; // right-to-left
	    // long column = column_loop; // left-to-right
		// in case of zero on the main diagonal
		// TODO: check coprimeness as well
		long pivot_row = ChoosePivotRow(independent_rows,column);
		cout << "\rComputing rank (Gauss), progress " << column_loop << "/" << limit << ", pivot:" << pivot_row<< "     ";
		// cout << "\t Pivot:" << pivot_row;
		if (pivot_row == -1) continue;
		if (pivot_row != independent_rows)  {
			SwapRows(pivot_row,independent_rows);
			// if given aRowIndices, we should keep information on swapping the columns
			if (aRowIndices) {
				long swap = aRowIndices[pivot_row];
				aRowIndices[pivot_row] = aRowIndices[independent_rows];
				aRowIndices[independent_rows] = swap;
			}
		};
		GaussColumn(independent_rows,column,aPartial);
		independent_rows++;
	}
	aRank = independent_rows;
	return ConstRC::Ok;
}

long AbstractIntegerMatrix::ChoosePivotRow(long aRow, long aColumn) {
	while ((aRow<GetMaximalRowIndex()) && (!IsInvertible(aRow,aColumn))) aRow++;
	if (IsZero(aRow,aColumn)) aRow = -1;
	return aRow;
}

int AbstractIntegerMatrix::ReorderRows(long aLimit, long aRowIndices[]) {
	long current_row_indices[GetMaximalRowIndex()];
	for(long i = 0;i<= GetMaximalRowIndex(); i++) {
		current_row_indices[i]=i;
	};
	for (long i = 0; i<= aLimit; i++ ) {
		if (i != aRowIndices[i]) {
			/* for (long ind=0;ind<=aLimit;ind++) cout << " | " << current_row_indices[ind] ;
			PrintToScreen();*/
			SwapRows(i, current_row_indices[aRowIndices[i]]);
			long swapped = current_row_indices[i];
			current_row_indices[i]=current_row_indices[aRowIndices[i]];
			current_row_indices[aRowIndices[i]] = swapped;
		};
	};
};

bool AbstractIntegerMatrix::IsRowZero(long aRow, long aLimit) {
	if ((aLimit>GetMaximalColumnIndex()) || (aLimit < 0)) aLimit = GetMaximalColumnIndex();
	for (long index=0; index <= aLimit; index ++)
		if (!IsZero(aRow,index)) return false;
	return true;
}

int AbstractIntegerMatrix::GetRegularSubmatrixModQ(long aLimit, long aModulo, long & aRank) {
	if (aLimit>GetMaximalColumnIndex()) aLimit = GetMaximalColumnIndex();
	AbstractIntegerMatrix * temporary_matrix = this->clone();
	temporary_matrix->SetMaximalRowIndex(GetMaximalRowIndex());
	temporary_matrix->SetMaximalColumnIndex(aLimit);
	temporary_matrix->Allocate();
	temporary_matrix->PartialCopy(this, this->GetMaximalRowIndex(),aLimit);
	temporary_matrix->SetModulo(aModulo);
	temporary_matrix->ApplyModulo();
	// temporary_matrix->PrintToScreen();
	cout << "\nNonzero 1:" <<temporary_matrix->CountNonZeroItems();

	long * row_indices  = new long[temporary_matrix->GetMaximalRowIndex()+1];
	temporary_matrix->Gauss(aLimit, row_indices, aRank, true);
	cout << "\nNonzero 2:" <<temporary_matrix->CountNonZeroItems();
	// temporary_matrix->ApplyModulo();
	//temporary_matrix->PrintToScreen();
	// cout << "Rank:"<< aRank << endl;
	// for (long row = 0; row<=GetMaximalRowIndex(); row++) {
	//	cout << " |" << row_indices[row];
	// };
	ReorderRows(aLimit, row_indices);

//	delete row_indices;
//	delete temporary_matrix;
}

int AbstractIntegerMatrix::RowProduct(long & aResult, long aRow1, long aRow2) 
{
	aResult=0;
	if ((aRow1 < 0) || (aRow2 < 0))	return ConstRC::GeneralError;
	if (aRow1 > this->maximal_row_index) return ConstRC::GeneralError;
	if (aRow2 > this->maximal_row_index) return ConstRC::GeneralError;
	for(int column=0; column <= this->maximal_column_index; column++) {
		integer_matrix_type value1, value2;
		GetValue(aRow1, column, value1);
		GetValue(aRow2, column, value2);
		aResult += value1* value2;
	};
	return ConstRC::Ok;
}

int AbstractIntegerMatrix::LLL()
{
	int rc;
	AbstractIntegerMatrix * gramova = clone();
	gramova->SetMaximalRowIndex(maximal_row_index);
	gramova->SetMaximalColumnIndex(maximal_row_index);
	AbstractIntegerMatrix * relative = clone();
	relative->SetMaximalRowIndex(maximal_row_index);
	relative->SetMaximalColumnIndex(maximal_row_index);
	AbstractIntegerMatrix * lambda = clone();
	lambda->SetMaximalRowIndex(maximal_row_index);
	lambda->SetMaximalColumnIndex(maximal_row_index);

	integer_matrix_type value;
	GetValue(0,0,value);
	lambda->PutValue(0,0,value);

	for(int row1=0; row1 <= this->maximal_row_index; row1++) {
		for(int row2=0; row2 <= this->maximal_row_index; row2++) {
			long product;
			if (ConstRC::Ok == (rc = RowProduct(product, row1, row2))) {
				gramova->PutValue(row1,row2, product);
			} else { return rc; };

		};
	};
	// Inicializace
	int k = 2;
	int kmax = 1;
	long * d = new long[this->maximal_row_index+2];
	d[0]=1;
	this->RowProduct(d[1],0,0);
	long u;
	for(int column=0; column <= this->maximal_row_index; column++) {
		for(int row=0; row <= this->maximal_row_index; row++) {
				relative->PutValue(row,column, row == column ? 1 : 0);

		};
	};

	do
	{
		// cout << "Jeden krok: "<<k <<"   "<<kmax;
		// cout<< "\nD: "; for (int ou=0;ou<=3;ou++) { cout << d[ou] <<", ";};
		// this->PrintToScreen();
		// gramova->PrintToScreen();
		// relative->PrintToScreen();
		// 2.
		if (k>kmax) {
			kmax=k;
			for (int j=1;j<=k;j++) {
				// cout << " KMAX "<< k <<" " <<j<<"   |";
				rc = this->RowProduct(u,k-1,j-1);
				if (ConstRC::Ok != rc) {return rc;};
				// cout << "{u:" << u <<",";
				for(int i=1;i<j;i++) {
					// cout << u <<",";
					integer_matrix_type value1, value2;
					lambda->GetValue(k-1,i-1, value1);
					lambda->GetValue(j-1,i-1,value2);
					u = (d[i]*u - value1*value2)/d[i-1];
				};
				// cout << " u:" << u <<"}";
				if (j<k) { lambda->PutValue(k-1,j-1,u);};
				if (j==k) { d[k]=u; /*cout <<"(d["<<k<<"]="<<u<<")";*/};
			};
		};
		// lambda->PrintToScreen();
		bool cycle3 = false;
		do {
			//REDI(k,k-1)
			// cout << "REDI (" << k<<","<<k-1<<")";
			// cout << "Qéčko:" << abs(2*lambda->GetXY(k-1,k-2)) << " > ";
			// cout << d[k-1] << " | ";
			integer_matrix_type aux_value;
			lambda->GetValue(k-1,k-2,aux_value);
			if (abs(2*aux_value)>d[k-1]) {
				// cout << "REDI: " << k << "    " << k-1 << "\n";
				// lambda->PrintToScreen();
				long q = (aux_value*2+d[k-1])/(2*d[k-1]);
				// cout << "q=" << q <<";";
				relative->SubtractRowMultiple(k-1,k-2,q);
				this->SubtractRowMultiple(k-1,k-2,q);
				lambda->GetValue(k-1,k-2,value);
				lambda->PutValue(k-1,k-2,value-q*d[k-1]);
				for ( int i=1;i<=k-2;i++) {
					integer_matrix_type aux_value1, aux_value2;
					lambda->GetValue(k-1,i-1,aux_value1);
					lambda->GetValue(k-2,i-1,aux_value2);
					lambda->PutValue(k-1,i-1,aux_value1-q*aux_value2);
				};
				
				// this->PrintToScreen();
			};
			// lambda->PrintToScreen();
			// cout<< "\nD: "; for (int ou=0;ou<=3;ou++) { cout << d[ou] <<", ";};
			// cout << 4*d[k]*d[k-2] << " < " <<3*d[k-1]*d[k-1] - lambda->GetXY(k-1,k-2)* lambda->GetXY(k-1,k-2);
			integer_matrix_type aux_value1, aux_value2;
			lambda->GetValue(k-1,k-2, aux_value1);
			lambda->GetValue(k-1,k-2, aux_value2);
			if (4*d[k]*d[k-2] < 3*d[k-1]*d[k-1] - aux_value1*aux_value2) {
				cycle3 = true;
				//SWAP(k)
				//cout << "SWAP: " << k << "  "<< k-1 << " \n";
				relative->SwapRows(k-1,k-2);
				this->SwapRows(k-1,k-2);
				//lambda->PrintToScreen();
				if (k>2) {
					for (int j=1;j<=k-2;j++) {
						integer_matrix_type temp, temp2;
						lambda->GetValue(k-1,j-1,temp);
						lambda->GetValue(k-2,j-1, temp2);
						lambda->PutValue(k-1,j-1, temp2);
						lambda->PutValue(k-2,j-1, temp);
					};
				};
				integer_matrix_type tlambda;
				lambda->GetValue(k-1,k-2, tlambda);
				integer_matrix_type tbe = (d[k-2]*d[k]+tlambda*tlambda)/d[k-1];
				for (int i = k+1;i<=kmax;i++) {
					integer_matrix_type tt;
					lambda->GetValue(i-1,k-1,tt);
					integer_matrix_type aux_value1;
					lambda->GetValue(i-1,k-2, aux_value1);
					lambda->PutValue((d[k]*aux_value1-tlambda*tt)/d[k-1],i-1,k-1);
					lambda->GetValue(i-1,k-1,aux_value1);
					lambda->PutValue((tbe*tt+tlambda*aux_value1)/d[k],i-1,k-2);
				}
				//cout << "<--tlambda->>>" << tlambda << " tbe--->" << tbe <<"<";
				d[k-1] = tbe;
				k = max(2,k-1);
				//lambda->PrintToScreen();
			} else { 
				//cout << "NESWAP: " << k << "  "<< k-1 << " \n";
				cycle3 = false;
			};
			//this->PrintToScreen();
			// gramova->PrintToScreen();
			// relative->PrintToScreen();
		} while (cycle3);
		for (int l=k-2;l>0;l--) {
			//REDI(k,k-1)
			//cout << "REDI2 (" << k<<","<<k-1<<")";
			integer_matrix_type aux_value;
			lambda->GetValue(k-1,l-1,aux_value);
			if (abs(2*aux_value)>d[l]) {
				lambda->GetValue(k-1,l-1, aux_value);
				long q = (aux_value*2+d[l])/(2*d[l]);
				//cout << "q=" << q <<";";
				relative->SubtractRowMultiple(k-1,l-1,q);
				this->SubtractRowMultiple(k-1,l-1,q);
				lambda->GetValue(k-1,l-1,aux_value);
				lambda->PutValue(aux_value-q*d[l],k-1,l-1);
				for ( int i=1;i<=l-1;i++) {
					integer_matrix_type aux_value1, aux_value2;
					lambda->GetValue(k-1,i-1,aux_value1);
					lambda->GetValue(l-1,i,aux_value);
					lambda->PutValue(aux_value1-q*aux_value2,k-1,i-1);
				};
			};
			//this->PrintToScreen();
		};
		k++;
	} while (k <= this->maximal_row_index+1);


	delete gramova;
	return ConstRC::Ok;
}
