#ifndef __ABSTRACT_MATRIX__
#define __ABSTRACT_MATRIX__

#include "mserializable.h"
#include "base_parameters.h"

typedef unsigned int matrix_type;

/*!
	\brief  The most abstract type of linear algebra matrix.
	
	This abstract class is meant to be a common denominator for all matrix types used in this program. By this, its inner structures are determined.
	It is reasonable to expect all 'real' matrix types to have some number of rows and some number of columns.
	That is why 
			long maximal_row_index;
			long maximal_column_index;
	are declared as member variables of the AbstractMatrix class.
	Names of the variables are chosen to reflect the fact that they determine the maximal index of a row, or
	a column, in the array of all rows or columns. 
	In C/C++, the first element of an array of length N is indexed by 0, and the last one by N-1; we hold to this
	model, and so a 32 x 64 matrix will have maximal_row_index equal to 31 and maximal_column_index equal to 63.
	A matrix with no rows and columns will have both maximal_row_index and maximal_column_index equal to -1.

	Before we describe properties of concrete matrix classes, let us define some terminology.
	
	This snippet of code
	
		NormalMatrix* nm;
		
	is a declared, but uninstantiated matrix.
	
	Now, 
	
		NormalMatrix* nm = new NormalMatrix(3,4);
		
	is an instantiated matrix, but unallocated - by this we mean that there is an instance of class type NormalMatrix,
	but that its huge inner structure (array) containing the data itself has not been allocated yet.
	You can call some getters and setters onto an instantiated, but unallocated matrix, but you cannot perform any
	arithmetic operations like addition and multiplication with this matrix (one exception is when this matrix is a target
	of such an operation, see the next paragraph). Such matrix literally has no contents, except for
	(maybe) information about its maximal row and column index.
	
	Finally,
	
		NormalMatrix* nm = new NormalMatrix(3,4);
		
		nm->Allocate(); 
		
	is an instantiated allocated matrix. You can do any mathematics with this matrix, since its data has already been
	allocated. The allocation can be either performed explicitely, by calling nm->Allocate(), or it can be performed
	within methods of NormalMatrix*. For example, if you call an addition of two already allocated matrices, with a 
	target matrix which has been instantiated, but not allocated yet, the Add() method will take care of the allocation
	of the target itself, without the user noticing.
	
	Now, another important thing must be mentioned. Mathematical operations like addition or multiplication generally 
	come in two flavours. The first type is something like
	
		NormalMatrix*	MultiplyInternal(NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	
		This type of method always instantiates AND allocates the target matrix for the operation. Unless specified 
		otherwise, the two operands can be identical (the same pointer). Of course, calling such function is a potential
		risk of memory leak; use it only if you know what you are doing.
		
	The other type is something like
	
		int	MultiplyInternal(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2);
		
		This type of method:
			- at first checks whether the aTarget/aOperand1/aOperand2 have been already instantiated. If one of them 
			  is NULL, it returns with an error value
			  	NULL_POINTER_SUPPLIED
			  defined in definitions.h.
			  At this place, I was in doubt whether return an error value, or to instantiate the matrix itself. Two
			  reasons made me choose this approach:
			       -# the function prototype would have to use double pointers. That is because change in pointer
				  aTarget would not propagate up to the caller: the pointer is just a local copy of the pointer
				  in the caller. Double pointers, like the ones in AllocationMachine(), remedy this, but they
				  are ugly.
			       -# calling an operation with an uninstantiated matrix is at least strange and the programmer
				  should be told about it. An uninstantiated matrix does not have dimensions etc., and usage
				  of it means that the programmer probably either did not care about correctness of the code,
				  or that he/she has a little mess in variables. 
				  Of course, if such programmer does not bother to check the return codes of the method calls,
				  he/she is doomed anyway.
			- then it optionally allocates the matrices. This concerns mainly the target matrix.
			- finally, it performs the requested operations.
			
		One great limitation of this type of method is the following:		
			- you must NEVER EVER use one of the operands as a target matrix!
			  Any call like Multiply(A,A,B) will end up in unpredictable results; the only predictable effect is that
			  the output will be incorrect.
			  
		I know that this is not a particularly good approach; but it works for me, and I did not feel the need to write
		code to remove this limitation. Feel free to add such functionality!




*/
class AbstractMatrix : public virtual MSerializable
{
public:
	AbstractMatrix();
	virtual ~AbstractMatrix();

	virtual AbstractMatrix * clone() = 0; ///< Virtual method for dynamic cloning of matrix type.

// GETTERS AND SETTERS
	long GetMaximalRowIndex() const; ///< Returns maximal row index.
	long GetMaximalColumnIndex() const;///< Returns maximal column index.
	void SetMaximalRowIndex(long aArg);///< Sets maximal row index.
	void SetMaximalColumnIndex(long aArg);///<Set maximal column index.
	species_of_matrices GetSpecies() const;

// Pure virtual methods
	virtual int  Allocate()  = 0; ///< This method will be used for allocation of dynamic internal structures of a subclass.
	virtual int  Randomize() = 0; ///< This method will be used for initialization of the matrix by random elements.
	virtual int  Zeroize() = 0; ///< This method will be used for initialization of the matrix by 0s.
	//virtual bool Equals(AbstractMatrix* aMatrix) = 0;
	virtual void PrintToScreen() = 0; ///< This method will be used for display of the matrix.
	virtual int  PutOne(long aRow, long aColumn) = 0; ///< This method will be used to put number 1 to the entry indexed by aRow and aColumn.
	virtual int  PutZero(long aRow, long aColumn) = 0; ///< This method will be used to put number 0 to the entry indexed by aRow and aColumn.
	virtual int  IsOne(long aRow, long aColumn) = 0; ///< This method will respond whether there is a 1 at the entry.
	virtual int  IsZero(long aRow, long aColumn) = 0; ///< This method will respond whether there is a 0 at the entry.
	virtual long  CountNonZeroItems(); ///< Count number of nonzero cells in the matrix.
	// Swaps etc.
	virtual long MaxNonZeroItemsPerRow(); ///< Find maximum number of cells on row.
	virtual int  ZeroizeRow(long aRow); ///< This method will be used to put 0s into a row given by index aRow.
	virtual int  ZeroizeRow(long* aRowList, long aListMaxIndex); ///< 			This method will be used to put 0s into rows, whose indices are given by array aRowList.			The aListMaxIndex parameter gives the final index in aRowList array, in order to prevent			buffer overflow error.

	virtual int  SwapRows(long aRow1, long aRow2) = 0; ///< 			This method will be used to swap rows with indices given by aRow1 and aRow2 parameters.

	virtual int  AddRows(long aTarget, long aSource) = 0; ///< 			This method will be used to add row with index aSource to row with index aTarget.

	virtual int  ZeroizeColumn(long aColumn) = 0; ///< 			This method will be used to put 0s into a column given by index aColumn.
	virtual int  ZeroizeColumn(long* aColumnList, long aListMaxIndex) = 0; ///<  This method will be used to put 0s into columns, whose indices are given by array aColumnList.			The aListMaxIndex parameter gives the final index in aColumnList array, in order to prevent			buffer overflow error.
			
	virtual int  SwapColumns(long aColumn1, long aColumn2) = 0; ///< This method will be used to swap columns with indices given by aColumn1 and aColumn2 parameters.
	virtual int  AddColumns(long aTarget, long aSource) = 0; ///< This method will be used to add column with index aSource to column with index aTarget.

	virtual int  AddToRow(long aRow,AbstractMatrix* aOperand, long aRow2) = 0; ///< This method will be used to add rows from two different matrices of the same type.

	virtual int	   IsZero(); ///< This method will tell if the matrix is zero.
	virtual int 	   Copy(AbstractMatrix *aSource); ///< This method will copy contents of aSource into current instance.
	virtual int	   PerformColumnMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask); ///< This method will serve as wrappers for multiple column zeroizing for bit matrices of width		at most 32. This is especially desirable in case of the Lanczos block algorithm, where late 			binding of virtual methods will promote "one source for various data types" programming paradigma.

	virtual int      PerformRowMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask);
	// Mathematical operations
	virtual int  	   Transpose(AbstractMatrix* aTarget);
	// AbstractMatrix* Transpose() = 0;
	virtual int  	   Add(AbstractMatrix* aTarget, AbstractMatrix* aOperand2);
	// AbstractMatrix* Add(AbstractMatrix* aOperand2) = 0;
	virtual int		MultiplyInternal(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	// AbstractMatrix*	MultiplyInternal(AbstractMatrix* aOperand1, AbstractMatrix* aOperand2) = 0;
	virtual int		MultiplyInternalTransposed(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	void SetInfoMode(int aMode);
	void SetAssertionMode(int aMode);
	int  GetInfoMode() const;
	int  GetAssertionMode() const;
//  Save and load methods
	virtual int  Save(char *aName) = 0;
	virtual int  Load(char *aName) = 0;
	virtual int  Compare(AbstractMatrix* aOperand);


	// De/Serialization methods
public:
	
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

protected:
	
	virtual int WriteHeader(xmlTextWriterPtr aWriter) const;
	virtual int WriteBody(xmlTextWriterPtr aWriter) const;
	virtual int WriteData(xmlTextWriterPtr aWriter) const = 0;

	virtual int ReadHeader(xmlTextReaderPtr aReader);
	virtual int ReadBody(xmlTextReaderPtr aReader);
	virtual int ReadData(xmlTextReaderPtr aReader) = 0;

//#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
public:
	// only for time measurement
	static int start; ///< An auxiliary variable for time measurement.
	static int zero_start; ///< An auxiliary variable for time measurement.
	static int diff; ///< An auxiliary variable for time measurement.
	static double total_big_multiply; ///< An auxiliary variable for time measurement.
	static double total_word_multiply; ///< An auxiliary variable for time measurement.
	static double total_word_transpose_multiply; ///< An auxiliary variable for time measurement.
	static double total_transpose;///< An auxiliary variable for time measurement.
	static double total_add; ///< An auxiliary variable for time measurement.
	static double total_zeroize; ///< An auxiliary variable for time measurement.
	static double total_copy; ///< An auxiliary variable for time measurement.
	static double total_iszero; ///< An auxiliary variable for time measurement.
//#endif

protected:
	long maximal_row_index;
	long maximal_column_index;
	int info_mode; ///<  Nonzero value of info mode means that some extra information about method run may be printed on screen
	int assertion_mode; ///< Nonzero value of assertion mode means that some extra	tests on invariants may be performed.
	species_of_matrices species; ///< 	Each matrix also has a "species", that means it belongs to a concrete matrix class. In order to be able	to effectively handle the arithmetics with different matrix classes, we keep the species information in this variable.
};
#endif

