#include "abstract_message.h"
#include "abstract_sieve.h"
#include <string.h>
#include <stdlib.h>

#include "xml_definitions.h"
#include "xml_message_definitions.h"
#include "common_xml_helper.h"
#include "log.h"

#define DESERIALIZE_HEADER "Deserialize Header"
#define DESERIALIZE_BODY "Deserialize Body"

// Forward declaration
class AbstractSieve;

AbstractMessage::AbstractMessage()
{
	iReplyTo = REPLIES_TO_NOTHING;
	iAskAgain = MINIMAL_ASK_AGAIN;
	Set_CanDestroy(TRUE);
	Init();
	iDelayFirstSend_Seconds = NO_DELAY_SEND;
}

AbstractMessage::AbstractMessage(const char* aSenderAddress, const char* aSenderURL, const char* aSubject, message_type aType, unsigned int aFlags, unsigned int aCounter, const char* aSenderId, const char* aTargetId, const char* aData, unsigned int aTimeout)
{
	iReplyTo = REPLIES_TO_NOTHING;
	iAskAgain = MINIMAL_ASK_AGAIN;
	Set_CanDestroy(TRUE);
	Init(aSenderAddress,aSenderURL,aSubject,aType,aFlags,aCounter,aSenderId,aTargetId,aData, aTimeout);	
	iDelayFirstSend_Seconds = NO_DELAY_SEND;
}

int AbstractMessage::Init()
{
	Set_Counter((unsigned int) 0);
	Set_Flags((unsigned int) 0);
	Set_CurrentTime();
	Set_Type(EUnknown);
	Set_Timeout(MINIMAL_TIMEOUT_IN_SECONDS);
	return 0;
}

int AbstractMessage::Init(const char* aSenderAddress, const char* aSenderURL, const char* aSubject, message_type aType, unsigned int aFlags, unsigned int aCounter, const char* aSenderId, const char* aTargetId, const char* aData, unsigned int aTimeout)
{
	Init();
	Set_SenderAddress(aSenderAddress);
	Set_SenderURL(aSenderURL);
	Set_Subject(aSubject);
	Set_Counter(aCounter);
	Set_Flags(aFlags);
	Set_SenderId(aSenderId);
	Set_TargetId(aTargetId);
	Set_Data(aData);
	Set_Type(aType);
	Set_Timeout(aTimeout);
	return 0;
}


AbstractMessage::~AbstractMessage()
{
	if (!iCanDestroy)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","dtor",iCounter,0,"Somebody tried to destroy an instance that should not be destroyed.");
	}
}

void 	AbstractMessage::Set_SenderAddress(const char* aValue)
{
	if (aValue)
	{
		iSenderAddress.assign(aValue);
	}
}

void 	AbstractMessage::Set_SenderURL(const char* aValue)
{
	if (aValue)
	{
		iSenderURL.assign(aValue);
	}
}

void 	AbstractMessage::Set_TargetURL(const char* aValue)
{
	if (aValue)
	{
		iTargetURL.assign(aValue);
	}
}


void	AbstractMessage::Set_Counter(unsigned int aCounter)
{
	iCounter = aCounter;
}

void	AbstractMessage::Set_Subject(const char* aValue)
{
	iSubject.assign(aValue);
}

void	AbstractMessage::Set_SenderId(const char* aValue)
{
	iSenderId.assign(aValue);
}

void	AbstractMessage::Set_TargetId(const char* aValue)
{
	iTargetId.assign(aValue);
}

bool	AbstractMessage::Has_Delay() const
{
	return (iDelayFirstSend_Seconds != NO_DELAY_SEND);
}

// This setter sets the whole flags field.
void	AbstractMessage::Set_Flags(unsigned int aFlags)
{
	iFlags = aFlags;
}

void	AbstractMessage::Set_Type(message_type aType)
{
	iType = aType;
	const char* description = "";
	switch (iType)
	{
		case EReady: 
			description = MESSAGE_READY_DESCRIPTION;
			break;
		case ENoJob:
			description = MESSAGE_NOJOB_DESCRIPTION;
			break;
		case EJobParameters:
			description =  MESSAGE_JOB_PARAMETERS_DESCRIPTION;
			break;
		case EAck:
			description = MESSAGE_ACK_DESCRIPTION;
			break;
		case ELocked:
			description = MESSAGE_LOCKED_DESCRIPTION;
			break;
		case ENodeBusy:
			description = MESSAGE_NODE_BUSY_DESCRIPTION;
			break;
		case EData:
			description = MESSAGE_DATA_DESCRIPTION;
			break;
		case EAlive:
			description = MESSAGE_ALIVE_DESCRIPTION;
			break;	
		case EQuit:
			description = MESSAGE_QUIT_DESCRIPTION;
			break;	
		default:
			break;
	}
	iDescription.assign(description);
}

// This setter sets one bit flag
void	AbstractMessage::Set_Flag(unsigned int aFlag)
{
	iFlags |= aFlag;
}

// This setter clears one bit flag
void	AbstractMessage::Clear_Flag(unsigned int aFlag)
{
	iFlags = (~aFlag) & iFlags;
}

void	AbstractMessage::Set_Data(const char* aValue)
{
	iData.assign(aValue);
}
void	AbstractMessage::Set_CanDestroy(bool aValue)
{
	iCanDestroy  = aValue;
}

void	AbstractMessage::Set_ReplyTo(unsigned int aValue)
{
	iReplyTo = aValue;
}

void	AbstractMessage::Set_Delay(unsigned int aValue)
{
	if (aValue == NO_DELAY_SEND)
	{
		iDelayFirstSend_Seconds = aValue;
		return;
	}
	if (aValue < MINIMAL_DELAY_SEND)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","Set_Delay",aValue,MINIMAL_DELAY_SEND,"The value is too low, setting to minimal possible.");
		iDelayFirstSend_Seconds = MINIMAL_DELAY_SEND;
		return;
	}
	if (aValue > MAXIMAL_DELAY_SEND)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","Set_Delay",aValue,MAXIMAL_DELAY_SEND,"The value is too low, setting to minimal possible.");
		iDelayFirstSend_Seconds = MAXIMAL_DELAY_SEND;
		return;
	}
	iDelayFirstSend_Seconds = aValue;
}

const char*	AbstractMessage::Get_SenderAddress() const
{
	return iSenderAddress.c_str();
}

const string&	AbstractMessage::Get_SenderURL_String() const
{
	return iSenderURL;
}

const string&	AbstractMessage::Get_TargetURL_String() const
{
	return iTargetURL;
}
const char*	AbstractMessage::Get_SenderURL() const
{
	return iSenderURL.c_str();
}

const char*	AbstractMessage::Get_Subject() const
{
	return iSubject.c_str();
}

const char*	AbstractMessage::Get_SenderId() const
{
	return iSenderId.c_str();
}

const char*	AbstractMessage::Get_TargetId() const
{
	return iTargetId.c_str();
}

unsigned int AbstractMessage::Get_Flags() const
{
	return iFlags;
}

bool	AbstractMessage::Is_Flag_Set(unsigned int aFlag) const
{
	return (iFlags & aFlag);
}

const char*	AbstractMessage::Get_Data() const
{
	return iData.c_str();
}

string	AbstractMessage::Get_QuitReason() const
{
	return iQuitReason;
}

bool 	AbstractMessage::Can_Destroy() const
{
	return iCanDestroy;
}

void	AbstractMessage::Set_CurrentTime()
{
	iTime = time(NULL);
}

void	AbstractMessage::Set_FutureTime(unsigned int aPlusSeconds)
{
	iTime = time(NULL) + aPlusSeconds;
}

void	AbstractMessage::Set_Time(time_t aTime)
{
	iTime = aTime;
}

time_t	AbstractMessage::Get_Time() const
{
	return iTime;
}

message_type AbstractMessage::Get_Type() const
{
	return iType;
}

unsigned int AbstractMessage::Get_Counter() const
{
	return iCounter;
}

unsigned int AbstractMessage::Get_ReplyTo() const
{
	return iReplyTo;
}

unsigned int AbstractMessage::Get_Delay() const
{
	return iDelayFirstSend_Seconds;
}

int     AbstractMessage::Deserialize()
{	
	//iPath_Body.assign(aPathBody);
	xmlTextReaderPtr reader = CreateInputStream();
	int rc = Deserialize(reader);	
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	xmlFreeTextReader(reader);
	rc = FinishDeserialization();
	return rc;
}

int     AbstractMessage::Serialize()
{	
	xmlTextWriterPtr writer = CreateOutputStream();
	int rc =  Serialize(writer);	
	if (rc != ConstRC::Ok) {
		return rc;
	}
	xmlFreeTextWriter(writer);
	rc = FinishSerialization();
	return rc;
}


int     AbstractMessage::Deserialize(xmlTextReaderPtr& aReader)
{	
	xmlChar* name;
	int type;
	int rc = xmlTextReaderRead(aReader);		
	while (rc == 1) {
        	type = xmlTextReaderNodeType(aReader);
		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{		
			
			if (xmlStrEqual(name,BAD_CAST MESSAGE_HEADER_NAME))
			{
				rc = ReadHeader(aReader);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			if (xmlStrEqual(name,BAD_CAST MESSAGE_BODY_NAME))
			{
				rc = ReadBody(aReader);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
		} // if (type == 1)
		rc = xmlTextReaderRead(aReader);
	} // end of while(TRUE)
	
	return ConstRC::Ok;
}

int	AbstractMessage::ReadHeader(xmlTextReaderPtr& aReader)
{
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	bool hasSender = FALSE;
	bool hasTarget = FALSE;
	bool hasCounter = FALSE;
	bool hasType = FALSE;
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadHeader",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadHeader",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{		
			
			if (xmlStrEqual(name,BAD_CAST SENDER_NAME))
			{
				ReadSenderInfo(aReader);
				bool hasSenderId      = !iSenderId.empty();
				// bool hasSenderAddress = !iSenderAddress.empty();
				// bool hasSenderURL     = !iSenderURL.empty();	
				if (hasSenderId)
				{
					hasSender = TRUE;	
				}
			}
			if (xmlStrEqual(name,BAD_CAST TARGET_NAME))
			{
				ReadTargetInfo(aReader);
				if (!iTargetId.empty())
				{
					hasTarget = TRUE;	
				}
			}
			if (xmlStrEqual(name,BAD_CAST COUNTER_NAME))
			{
				CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_HEADER,iCounter,FALSE);
				hasCounter = TRUE;
			}
			if (xmlStrEqual(name,BAD_CAST REPLY_TO_NAME))
			{
				CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_HEADER,iReplyTo,FALSE);
				hasCounter = TRUE;
			}
			if (xmlStrEqual(name,BAD_CAST MESSAGE_TYPE_NAME))
			{
				unsigned int target;
				CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_HEADER,target,FALSE);
				iType = (message_type) target;			
				hasType = TRUE;
			}
			if (xmlStrEqual(name,BAD_CAST MESSAGE_DESCRIPTION_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"Description read as ",iDescription,FALSE);
			}
		} // if (type == 1)
	} // end of while(TRUE)
	int ret = ConstRC::Ok;
	if (!hasSender)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadHeader",rc,0,"The message header does not contain sufficient sender information");
		ret = GENERAL_ERROR;
	}
	if (!hasTarget)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadHeader",rc,0,"The message header does not contain sufficient target information");
		ret = GENERAL_ERROR;
	}
	if (!hasCounter)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadHeader",rc,0,"The message header does not contain counter");
		ret = GENERAL_ERROR;
	}
	if (!hasType)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadHeader",rc,0,"The message header does not contain message type");
		ret = GENERAL_ERROR;
	}
// 	PrintHeaderInfo();
	return ret;
}

void	AbstractMessage::PrintHeaderInfo() const
{
	printf("Message summary:\n");	
	printf("%15s: %s\n","Sender ID",iSenderId.c_str());
	printf("%15s: %s\n","Sender Address",iSenderAddress.c_str());
	printf("%15s: %s\n","Sender URL",iSenderURL.c_str());
	printf("%15s: %s\n","Target ID",iTargetId.c_str());
	printf("%15s: %d\n","Counter",iCounter);
	const char* type = "Unknown";
	switch (iType)
	{
		case EReady: type = "Ready";
			break;
		case ENoJob:
			type = "No Job";
			break;
		case EJobParameters:
			type = "Job Parameters";
			break;
		case EAck:
			type = "Ack";
			break;
		case ELocked:
			type = "Locked";
			break;
		case ENodeBusy:
			type = "Node busy";
			break;
		case EData:
			type = "Data";
			break;
		case EAlive:
			type = "Alive";
			break;	
		case EQuit:
			type = "Quit";
			break;	
		default:
			type = "Unknown!";
			break;
	}
	printf("%15s: %s\n","Type",type);
	printf("%15s: %s\n","Description",iDescription.c_str());
}



int	AbstractMessage::ReadSenderInfo(xmlTextReaderPtr& aReader)
{
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	iSenderId.clear();
	iSenderAddress.clear();
	iSenderURL.clear();	
	#ifdef READSENDERINFOCONNECTIONCHECK	
		bool hasAddress = FALSE;
		bool hasConnector=FALSE;
	#endif
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadPartyInfo",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadPartyInfo",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST ID_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"Id read as ",iSenderId,FALSE);
			}
			if (xmlStrEqual(name,BAD_CAST ADDRESS_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"Address read as ",iSenderAddress,FALSE);
	#ifdef READSENDERINFOCONNECTIONCHECK	
				hasAddress = TRUE;
	#endif
			}			
			if (xmlStrEqual(name,BAD_CAST CONNECTOR_NAME))
			{
	#ifdef READSENDERINFOCONNECTIONCHECK	
				hasConnector = TRUE;
	#endif
			}
			if (xmlStrEqual(name,BAD_CAST URL_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"URL read as ",iSenderURL,FALSE);
			}			
		}
	} // while (TRUE)
#ifdef READSENDERINFOCONNECTIONCHECK	
	if (!(hasAddress ^ hasConnector))
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadSenderInfo",rc,0,"Illegal connection parameters read");
       	return GENERAL_ERROR;
	}
#endif
	return ConstRC::Ok;
}

int	AbstractMessage::ReadTargetInfo(xmlTextReaderPtr& aReader)
{
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);	
	iTargetId.clear();	
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadTargetInfo",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadTargetInfo",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{
			if (xmlStrEqual(name,BAD_CAST ID_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"Id read as ",iTargetId,FALSE);
			}			
		}
	} // while (TRUE)
	
	return ConstRC::Ok;
}
		

const vector<string>&	AbstractMessage::Get_SupportedAlgorithms() const
{
	return iSupportedAlgorithms;
}

int	AbstractMessage::Serialize(xmlTextWriterPtr& aWriter) const
{
	int rc = xmlTextWriterStartDocument(aWriter, NULL, XML_ENCODING, NULL);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not write the document start.");
		return GENERAL_ERROR;        	
    	}
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST SIEVING_MESSAGE_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not write the serializator tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	// Now, header will be created
	rc = CreateHeader(aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CreateBody(aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = xmlTextWriterEndElement(aWriter); // SIEVING_MESSAGE_NAME
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not end the message tag.");
        	return GENERAL_ERROR;
    	}
    	rc = xmlTextWriterEndDocument(aWriter);
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}
	return ConstRC::Ok;
}

int AbstractMessage::CreateHeader(xmlTextWriterPtr& aWriter) const
{
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST MESSAGE_HEADER_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not write the message tag.");
        	return GENERAL_ERROR;
    	}	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::WriteSerializatorData(aWriter, ABSTRACT_MESSAGE_NAME, ABSTRACT_MESSAGE_VERSION);
	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not write the serializator info.");
        	return GENERAL_ERROR;
    	}	
	/* Sender */
	rc = WriteSenderInfo(aWriter);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}
	/* Target */
	rc = WriteTargetInfo(aWriter);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}
	/* Counter */
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,COUNTER_NAME,iCounter);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}

	/* Reply-to */
	if (iReplyTo != REPLIES_TO_NOTHING)
	{
		rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,REPLY_TO_NAME,iReplyTo);
		if (rc != ConstRC::Ok)
		{
			return GENERAL_ERROR;
		}
	}
	
	/* Time */
	rc = CommonXMLHelper::SerializeTime(aWriter);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}
	/* Message Type */
	rc = WriteMessageType( aWriter);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}

	rc = CreateSpecificHeaderFields(aWriter);
	if (rc < 0)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not write the specific header fields.");
        	return GENERAL_ERROR;
	}
	rc = xmlTextWriterEndElement(aWriter); // MESSAGE_HEADER
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","Serialize Message",rc,0,"Could not end the message tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	return ConstRC::Ok;
}
int	AbstractMessage::WriteMessageType(xmlTextWriterPtr& aWriter) const
{
	
	int rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,MESSAGE_TYPE_NAME,iType);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, MESSAGE_DESCRIPTION_NAME, iDescription.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
	}
	return ConstRC::Ok;
}

int	AbstractMessage::WriteSenderInfo(xmlTextWriterPtr& aWriter) const
{	
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST SENDER_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the sender start tag.");
        	return GENERAL_ERROR;
    	}	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, ID_NAME, iSenderId.c_str());
	//int rc = xmlTextWriterStartElement(aWriter, BAD_CAST ID_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the id.");
        	return GENERAL_ERROR;
    	}	
	// In Sender field, either "Address" or "Connector" must be set. They cannot
	// be both set at the same time.
	bool hasSenderAddress = !iSenderAddress.empty();
	bool hasSenderURL     = !iSenderURL.empty();
	// Precisely one of those must be set. If not ...
	if (!(hasSenderAddress ^ hasSenderURL))
	{
		 CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Illegal sender parameters. Either no connection info is set, or too much");
        	return GENERAL_ERROR;
	}
	if (hasSenderAddress)
	{
		rc = CommonXMLHelper::SerializeString(aWriter, ADDRESS_NAME, iSenderAddress.c_str());
		//int rc = xmlTextWriterStartElement(aWriter, BAD_CAST ID_NAME);
    		if (rc < 0) {
	        	CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the address.");
        		return GENERAL_ERROR;
		}
    	}
//	if (hasSenderURL)
//	{
//		int rc = xmlTextWriterStartElement(aWriter, BAD_CAST CONNECTOR_NAME);
//	    	if (rc < 0) {
//		        CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the connector.");
//	        	return GENERAL_ERROR;
//    		}	
//		rc = CommonXMLHelper::NewLine(aWriter,FALSE);
//		if (rc < 0) {
//		        CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the new line after connector.");
//	        	return GENERAL_ERROR;
//    		}	
//		rc = CommonXMLHelper::SerializeString(aWriter, URL_NAME, iSenderURL.c_str());	
//    		if (rc < 0) {
//	        	CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the address.");
//        		return GENERAL_ERROR;
//		}
//		rc = xmlTextWriterEndElement(aWriter); // connector
//		if (rc != ConstRC::Ok)
//		{
//			return rc;
//		}
//		rc = CommonXMLHelper::NewLine(aWriter,FALSE);
//		if (rc != ConstRC::Ok)
//		{
//			return rc;
//		}
//    	}	
	rc = xmlTextWriterEndElement(aWriter); // /sender
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","WriteSenderInfo",rc,0,"Could not write the party end tag.");
        	return GENERAL_ERROR;
    	}	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	return ConstRC::Ok;
}

int	AbstractMessage::WriteTargetInfo(xmlTextWriterPtr& aWriter) const
{	
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST TARGET_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","WriteTargetInfo",rc,0,"Could not write the target start tag.");
        	return GENERAL_ERROR;
    	}	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, ID_NAME, iTargetId.c_str());
	//int rc = xmlTextWriterStartElement(aWriter, BAD_CAST ID_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","WriteTargetInfo",rc,0,"Could not write the id.");
        	return GENERAL_ERROR;
    	}	
	rc = xmlTextWriterEndElement(aWriter); // trget
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","WriteTargetInfo",rc,0,"Could not write the party end tag.");
        	return GENERAL_ERROR;
    	}	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	return ConstRC::Ok;
}


int AbstractMessage::CreateBody(xmlTextWriterPtr& aWriter) const
{
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST MESSAGE_BODY_NAME);
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","CreateBody ",rc,0,"Could not write the body tag.");
        	return GENERAL_ERROR;
    	}	
	rc = CreateSpecificBodyFields(aWriter);
	if (rc < 0)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","CreateBody ",rc,0,"Could not write the specific body fields.");
        	return GENERAL_ERROR;
	}

	rc = xmlTextWriterEndElement(aWriter); // MESSAGE_BODY_NAME
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","CreateBody Message",rc,0,"Could not end the body tag.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	return ConstRC::Ok;
}

int AbstractMessage::CreateSpecificHeaderFields(xmlTextWriterPtr& aWriter) const
{	
	return ConstRC::Ok;
}

int AbstractMessage::CreateSpecificBodyFields(xmlTextWriterPtr& aWriter) const
{	
	switch (iType)
	{
		case EReady:
			return CreateBody_ReadyMessage(aWriter);
		case ENoJob:
			return CreateBody_NoJobMessage(aWriter);
		case EJobParameters:
			return CreateBody_JobParametersMessage(aWriter);
		case EAck:
			return CreateBody_AckMessage(aWriter);
		case ELocked:
			return CreateBody_LockedMessage(aWriter);
		case ENodeBusy:
			return CreateBody_NodeBusyMessage(aWriter);
		case EQuit:
			return CreateBody_QuitMessage(aWriter);
		case EData:
			return CreateBody_DataMessage(aWriter);
		case EAlive:
			return CreateBody_AliveMessage(aWriter);
		default:
			CLog::LogError(WHEREARG,"AbstractMessage","CreateSpecificBodyFields",iType,0,"Unsupported message type.");
        		return GENERAL_ERROR;			

	}
	return ConstRC::Ok;
}

int	AbstractMessage::ReadBody(xmlTextReaderPtr& aReader)
{
	switch (iType)
	{
		case EReady:
			return ReadBody_ReadyMessage(aReader);
		case ENoJob:
			return ReadBody_NoJobMessage(aReader);
		case EJobParameters:
			return ReadBody_JobParametersMessage(aReader);
		case EAck:
			return ReadBody_AckMessage(aReader);
		case ELocked:
			return ReadBody_LockedMessage(aReader);
		case ENodeBusy:
			return ReadBody_NodeBusyMessage(aReader);
		case EQuit:
			return ReadBody_QuitMessage(aReader);
		case EData:
			return ReadBody_DataMessage(aReader);
		case EAlive:
			return ReadBody_AliveMessage(aReader);
		default:
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody",iType,0,"Unsupported message type.");
        		return GENERAL_ERROR;			
	}	
	return ConstRC::Ok;
}


int 	AbstractMessage::CreateBody_ReadyMessage(xmlTextWriterPtr& aWriter) const
{
	int rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	// list supported algorithms
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST SUPPORTED_ALGORITHMS_NAME);	
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	int i, max = iSupportedAlgorithms.size();
	for (i=0; i< max; i++)
	{
		rc = CommonXMLHelper::SerializeString(aWriter,ALGORITHM_NAME, iSupportedAlgorithms.at(i).c_str());
		if (rc != ConstRC::Ok)
		{
			return rc;	
		}			
	}
	rc = xmlTextWriterEndElement(aWriter); // SUPPORTED_ALGORITHMS_NAME
    	if (rc < 0) {
	       	CLog::LogError(WHEREARG,"AbstractMessage","CreateBody_ReadyMessage",rc,0,"Could not end a tag in the body.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	return ConstRC::Ok;
}

string	AbstractMessage::Get_BodyPath() const
{
	return iFullPath_Body;
}

int 	AbstractMessage::ReadBody_ReadyMessage(xmlTextReaderPtr& aReader)
{
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	iSupportedAlgorithms.clear();
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_ReadyMessage",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_ReadyMessage",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{		
			if (xmlStrEqual(name,BAD_CAST ALGORITHM_NAME))
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Algorithm read as ",s1,FALSE);				
				iSupportedAlgorithms.push_back(s1);
			}
		} // if (type == 1)
	} // end of while(TRUE)
	int ret = ConstRC::Ok;
	if (!iSupportedAlgorithms.size())
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_ReadyMessage",rc,0,"The message body does not contain any supported algorithms.");
		ret = GENERAL_ERROR;
	}	
	return ret;
}


int AbstractMessage::CreateBody_NoJobMessage(xmlTextWriterPtr& aWriter) const
{
	int rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,ASK_AGAIN_NAME,iAskAgain);
	if (rc != ConstRC::Ok)
	{
		return GENERAL_ERROR;
	}
	// list supported algorithms
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST SUPPORTED_ALGORITHMS_NAME);	
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	int i, max = iSupportedAlgorithms.size();
	for (i=0; i< max; i++)
	{
		rc = CommonXMLHelper::SerializeString(aWriter,ALGORITHM_NAME, iSupportedAlgorithms.at(i).c_str());
		if (rc != ConstRC::Ok)
		{
			return rc;	
		}			
	}
	rc = xmlTextWriterEndElement(aWriter); // SUPPORTED_ALGORITHMS_NAME
  	if (rc < 0) {
  	     	CLog::LogError(WHEREARG,"AbstractMessage","CreateSpecificBodyFields",rc,0,"Could not end a tag in the body.");
        		return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	// list unsupported algorithms
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST UNSUPPORTED_ALGORITHMS_NAME);	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	max = iUnsupportedAlgorithms.size();
	for (i=0; i< max; i++)
	{
		rc = CommonXMLHelper::SerializeString(aWriter,ALGORITHM_NAME, iUnsupportedAlgorithms.at(i).c_str());
		if (rc != ConstRC::Ok)
		{
			return rc;	
		}			
	}
	rc = xmlTextWriterEndElement(aWriter); // UNSUPPORTED_ALGORITHMS_NAME
    	if (rc < 0) {
	       	CLog::LogError(WHEREARG,"AbstractMessage","CreateBody_NoJob",rc,0,"Could not end a tag in the body.");
        	return GENERAL_ERROR;
    	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	return ConstRC::Ok;
}

int AbstractMessage::ReadBody_NoJobMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok;
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	iSupportedAlgorithms.clear();
	iUnsupportedAlgorithms.clear();
	bool supportedActive = FALSE;
	bool unsupportedActive = FALSE;
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_NoJob",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_NoJob",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == XML_READER_TYPE_ELEMENT)
		{		
			if (xmlStrEqual(name,BAD_CAST SUPPORTED_ALGORITHMS_NAME))
			{
				supportedActive = TRUE;
				unsupportedActive = FALSE;
			}
			if (xmlStrEqual(name,BAD_CAST UNSUPPORTED_ALGORITHMS_NAME))
			{
				supportedActive = FALSE;
				unsupportedActive = TRUE;
			}
			if (xmlStrEqual(name,BAD_CAST ALGORITHM_NAME))
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Algorithm read as ",s1,FALSE);
				if (supportedActive)
				{
					iSupportedAlgorithms.push_back(s1);
				}
				if (unsupportedActive)
				{
					iUnsupportedAlgorithms.push_back(s1);
				}
			}
			if (xmlStrEqual(name,BAD_CAST ASK_AGAIN_NAME))
			{
				CommonXMLHelper::DeserializeUnsignedInt(aReader,"Ask again read as",iAskAgain,FALSE);
			}
		} // if (type == XML_READER_TYPE_ELEMENT)
		if (type == XML_READER_TYPE_END_ELEMENT)
		{
			if (xmlStrEqual(name,BAD_CAST SUPPORTED_ALGORITHMS_NAME))
			{
				supportedActive = FALSE;
// 				printf("AbstractMessage::ReadBody_NoJob: closing supportedActive\n");
			}
			if (xmlStrEqual(name,BAD_CAST UNSUPPORTED_ALGORITHMS_NAME))
			{
				unsupportedActive = FALSE;
// 				printf("AbstractMessage::ReadBody_NoJob: closing unsupportedActive\n");
			}
		}
	} // end of while(TRUE)
	if (!iSupportedAlgorithms.size())
	{
		CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_ReadyMessage",rc,0,"The message body does not contain any supported algorithms.");
		ret = GENERAL_ERROR;
	}	

	return ret;
}


int AbstractMessage::CreateBody_JobParametersMessage(xmlTextWriterPtr& aWriter) const
{
	int ret = ConstRC::Ok;
	int rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, JOB_ID_NAME, iJobId.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
	}	
	if (iSerializator && iJobParameters)
	{
		//printf("Abstract messagre creates body of job parameters\n");
		//iJobParameters->Print();
		iSerializator->SerializeJob(aWriter, *iJobParameters);
	}
	return ret;
}


int AbstractMessage::ReadBody_JobParametersMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok, rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	//printf("Abstract Message: Read Body Job Parameters. Ready to go ..."); 
	while (TRUE)
	{
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadJobParameters",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_JobParameters",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}		
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		}		
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		if (xmlStrEqual(name, BAD_CAST JOB_ID_NAME))
		{
			if (type == 1)
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Job-id read as ",s1,FALSE);
				iJobId.assign(s1);			
			}
		}
		if (xmlStrEqual(name, BAD_CAST JOB_DATA_NAME))
		{
			if (type == 1) 
			{ 	// start element
				//printf("########## Will dump body now #########\n\n\n");
				//printf("The path is : %s\n",iFullPath_Body.c_str());
				xmlNodePtr	ptr = xmlTextReaderExpand(aReader);
				FILE* f = fopen(iFullPath_Body.c_str(),"wb");
				xmlDocPtr 	doc = xmlNewDoc(BAD_CAST "1.0");
				xmlDocSetRootElement(doc, ptr);
				xmlDocDump(f, doc);
				fflush(f);	
				fclose(f);					
				return ConstRC::Ok;
			} 
		}
	}	
	return ret;
}

/*int	AbstractMessage::CreateBodyFile()
{
	iBodyWriter = xmlNewTextWriterFilename(iFullPath_Body.c_str(),0);
	return 0;
}

int	AbstractMessage::AddStartElementToBodyFile(xmlTextReaderPtr& aReader)
{
	xmlChar* name = xmlTextReaderName(aReader);
	xmlTextWriterStartElement(name);
	int i, attrCount = xmlTextReaderAttributeCount(aReader);
	int rc;
	for (i=0; i< attrCount; i++)
	{
		rc = xmlTextWriterWriteFormatAttribute(aWriter, BAD_CAST TYPE,"%s",BAD_CAST MPZ);
	}
}

int	AbstractMessage::AddEndElementToBodyFile()
{
	int rc = xmlTextWriterEndElement(iBodyWriter);
	if (rc < 0) {
        	CLog::LogError(WHEREARG,"Abstract Message","Close Body File",rc,0,"Could not write end element.");
		return GENERAL_ERROR;
    	}
	return rc;
}

int	AbstractMessage::AddTextToBodyFile()
{
}

int	AbstractMessage::AddCommentToBodyFile()
{
}

int	AbstractMessage::CloseBodyFile()
{
	int rc = xmlTextWriterEndDocument(iBodyWriter);
    	if (rc < 0) {
        	CLog::LogError(WHEREARG,"Abstract Message","Close Body File",rc,0,"Could not close the document.");
		return GENERAL_ERROR;
    	}
	xmlFreeTextWriter(iBodyWriter);
	return rc;
}
*/
int AbstractMessage::CreateBody_AckMessage(xmlTextWriterPtr& aWriter) const
{
	// Ack message has no body.	
	return ConstRC::Ok;
}

int AbstractMessage::ReadBody_AckMessage(xmlTextReaderPtr& aReader)
{
	// Ack message has no body.
	return ConstRC::Ok;
}


int AbstractMessage::CreateBody_LockedMessage(xmlTextWriterPtr& aWriter) const
{
	int ret = ConstRC::Ok;
	// The body is empty
	//printf("AbstractMessage::CreateBody_Locked not implemented yet\n");
	return ret;
}

int AbstractMessage::ReadBody_LockedMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok;
	// Nothing to read, the body is empty
	//printf("AbstractMessage::ReadBody_Locked not implemented yet\n");
	return ret;
}

int AbstractMessage::CreateBody_NodeBusyMessage(xmlTextWriterPtr& aWriter) const
{
	int ret = ConstRC::Ok;
	// The body is empty
	//printf("AbstractMessage::CreateBody_NodeBusy not implemented yet\n");
	return ret;
}

int AbstractMessage::CreateBody_AliveMessage(xmlTextWriterPtr& aWriter) const
{
	int rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, JOB_ID_NAME, iJobId.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
 	}		
	return ConstRC::Ok;
}

int AbstractMessage::ReadBody_NodeBusyMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok;
	// The body is empty
	//printf("AbstractMessage::ReadBody_NodeBusy not implemented yet\n");
	return ret;
}


int AbstractMessage::CreateBody_QuitMessage(xmlTextWriterPtr& aWriter) const
{
	int ret = ConstRC::Ok;
	int rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, JOB_ID_NAME, iJobId.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
	}	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	rc = CommonXMLHelper::SerializeString(aWriter, QUIT_REASON_NAME, iQuitReason.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
	}	
	return ret;
}

int AbstractMessage::ReadBody_QuitMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok, rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);	
	while (TRUE)
	{
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadQuit",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadQuit",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}		
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		}		
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		if (xmlStrEqual(name, BAD_CAST JOB_ID_NAME))
		{
			if (type == 1)
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Job-id read as ",s1,FALSE);
				iJobId.assign(s1);			
			}
		}
		if (xmlStrEqual(name, BAD_CAST QUIT_REASON_NAME))
		{
			if (type == 1)
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Quit reason read as ",s1,FALSE);
				iQuitReason.assign(s1);			
			}
		}
	}	
	return ret;
}

int AbstractMessage::CreateBody_DataMessage(xmlTextWriterPtr& aWriter) const
{
	int ret = ConstRC::Ok;
	int rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, JOB_ID_NAME, iJobId.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
	}	
	rc = CommonXMLHelper::SerializeUnsignedInt(aWriter,DATA_COUNT_NAME,iDataCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CommonXMLHelper::SerializeString(aWriter, DATA_UNIT_NAME, iDataUnit.c_str());
	if (rc != ConstRC::Ok)
	{
		return rc;	
	}	
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST DATA_NAME);	
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	if (iSerializator)
	{
		iSerializator->SerializeData(aWriter);
	}	
	rc = xmlTextWriterEndElement(aWriter); // DATA_NAME
    	if (rc < 0) {
	        CLog::LogError(WHEREARG,"AbstractMessage","CreateBody_DataMessage",rc,0,"Could not end the data tag.");
        	return GENERAL_ERROR;
    	}
	return ret;
}

int AbstractMessage::ReadBody_DataMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok, rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	//printf("Abstract Message: Read Body Data Parameters. Ready to go ..."); 
	while (TRUE)
	{
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadData",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadBody_DataParameters",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}		
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		}	
	
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		if (xmlStrEqual(name, BAD_CAST JOB_ID_NAME))
		{
			if (type == 1)
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Job-id read as ",s1,FALSE);
				iJobId.assign(s1);			
			}
		}
		if (xmlStrEqual(name,BAD_CAST DATA_COUNT_NAME))
		{
			CommonXMLHelper::DeserializeUnsignedInt(aReader,DESERIALIZE_BODY,iDataCount,FALSE);
			
		}
		if (xmlStrEqual(name, BAD_CAST DATA_UNIT_NAME))
		{
			if (type == 1)
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Data unit read as ",s1,FALSE);
				iDataUnit.assign(s1);			
			}
		}
		if (xmlStrEqual(name, BAD_CAST DATA_NAME))
		{
			if (type == 1) 
			{ 	// start element
				xmlNodePtr	ptr = xmlTextReaderExpand(aReader);
				FILE* f = fopen(iFullPath_Body.c_str(),"wb");
				xmlDocPtr 	doc = xmlNewDoc(BAD_CAST "1.0");
				xmlDocSetRootElement(doc, ptr);
				xmlDocDump(f, doc);
				fflush(f);	
				fclose(f);				
				return ConstRC::Ok;
			} 
		}
	}	
	return ret;
}

int AbstractMessage::ReadBody_AliveMessage(xmlTextReaderPtr& aReader)
{
	int ret = ConstRC::Ok, rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	//printf("Abstract Message: Read Body Alive. Ready to go ..."); 
	while (TRUE)
	{
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadAlive",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"AbstractMessage","ReadAlive",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}		
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		}		
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		if (xmlStrEqual(name, BAD_CAST JOB_ID_NAME))
		{
			if (type == 1)
			{
				string s1;
				CommonXMLHelper::DeserializeString(aReader,"Job-id read as ",s1,FALSE);
				iJobId.assign(s1);			
			}
		}
	}	
	return ret;

}

void	AbstractMessage::AddSupportedAlgorithm(const char* aAlgorithmName)
{
	if (aAlgorithmName)
	{
		int i, max = iSupportedAlgorithms.size();
		for (i=0; i< max; i++)
		{
			string& s1 = iSupportedAlgorithms.at(i);
			if (s1 == aAlgorithmName)
			{
				return;
			}
		}
		//string s1 = aAlgorithmName;
		iSupportedAlgorithms.push_back(aAlgorithmName);
	}
}

void	AbstractMessage::AddUnsupportedAlgorithm(const char* aAlgorithmName)
{
	if (aAlgorithmName)
	{
		int i, max = iUnsupportedAlgorithms.size();
		for (i=0; i< max; i++)
		{
			string& s1 = iUnsupportedAlgorithms.at(i);
			if (s1 == aAlgorithmName)
			{
				return;
			}
		}
		//string s1 = aAlgorithmName;
		iUnsupportedAlgorithms.push_back(aAlgorithmName);
	}
}

void	AbstractMessage::AddJobParameters(const JobParameters* aParameters, AbstractSerializator* aSerializator)
{
	 if (!aParameters)
	 {
		CLog::LogError(WHEREARG,"AbstractMessage","AddJobParameters",0,0,"Parameters are NULL.");
		return;
	 }
	 if (!aSerializator)
	 {
		CLog::LogError(WHEREARG,"AbstractMessage","AddJobParameters",0,0,"Serializator is NULL.");
		return;
	 }
	 iJobParameters = aParameters;
	 iSerializator  = aSerializator;
	 //printf("Abstract message prints out parameters.\n");
	 //iJobParameters->Print();
}

void	AbstractMessage::AddData(AbstractSerializator* aSerializator)
{
	 if (!aSerializator)
	 {
		CLog::LogError(WHEREARG,"AbstractMessage","AddData",0,0,"Serializator is NULL.");
		return;
	 }	 
	 iSerializator  = aSerializator;
}

void	AbstractMessage::Set_Timeout(unsigned int aValueInSeconds)
{
	if ((aValueInSeconds == NO_TIMEOUT) || (aValueInSeconds >= MINIMAL_TIMEOUT_IN_SECONDS))
	{
		iTimeoutInSeconds = aValueInSeconds;
	} else
	{
		CLog::LogError(WHEREARG,"AbstractMessage","Set_Timeout",aValueInSeconds,MINIMAL_TIMEOUT_IN_SECONDS,"The timeout is too low, setting to minimal possible.");
		iTimeoutInSeconds = MINIMAL_TIMEOUT_IN_SECONDS;
	}	
}

unsigned int	AbstractMessage::Get_Timeout() const
{
	return iTimeoutInSeconds;
}

bool		AbstractMessage::Has_Timeout() const
{
	return (iTimeoutInSeconds != NO_TIMEOUT);
}

void	AbstractMessage::Set_AskAgain(unsigned int aValue)
{
	if (aValue < MINIMAL_ASK_AGAIN)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","Set_AskAgain",aValue,MINIMAL_ASK_AGAIN,"The value is too low, setting to minimal possible.");
		iAskAgain = MINIMAL_ASK_AGAIN;
		return;
	}
	if (aValue > MAXIMAL_ASK_AGAIN)
	{
		CLog::LogError(WHEREARG,"AbstractMessage","Set_AskAgain",aValue,MAXIMAL_ASK_AGAIN,"The value is too big, setting to maximal possible.");
		iAskAgain = MAXIMAL_ASK_AGAIN;
		return;
	}
	iAskAgain = aValue;
}

unsigned int AbstractMessage::Get_AskAgain() const
{
	return iAskAgain;
}

void	AbstractMessage::Set_JobId(const string& aValue)
{
	iJobId.assign(aValue);
}

string AbstractMessage::Get_JobId() const
{
	return iJobId;
}

void	AbstractMessage::Set_DataCount(unsigned int aValue)
{
	iDataCount = aValue;
}

void	AbstractMessage::Set_DataUnit(const char* aValue)
{
	if (aValue)
	{
		iDataUnit.assign(aValue);
	} else
	{
		iDataUnit.assign("");
	}
}

void	AbstractMessage::Set_DataUnit(const string& aValue)
{
	iDataUnit.assign(aValue);
}

unsigned int	AbstractMessage::Get_DataCount() const
{
	return iDataCount;
}

string		AbstractMessage::Get_DataUnit() const
{
	return iDataUnit;
}

void	AbstractMessage::AddQuitReason(const char* aQuitReason)
{
	if (aQuitReason)
	{
		iQuitReason.assign(aQuitReason);
	} else
	{
		iQuitReason.assign("unknown");
	}
}
 ///< Only used in the Quit message.


