#ifndef ABSTRACT_MESSAGE_H
#define ABSTRACT_MESSAGE_H

#include <time.h>
#include "types_common.h"
#include "connectivity_framework_definitions.h"

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <libxml/xmlreader.h>

#include <string>
#include <vector>

#define BODY_EXTENSION	".body"

#define REPLIES_TO_NOTHING 0

#define MINIMAL_ASK_AGAIN 45
#define MAXIMAL_ASK_AGAIN 86400

// the default value
#define NO_DELAY_SEND 	   0  
#define MINIMAL_DELAY_SEND 15
#define MAXIMAL_DELAY_SEND 3600

class JobParameters;
class AbstractSerializator;

using namespace std;

/*!
 *
 * \brief A class for containing messages between nodes/centers of a distributed process.
 *
 * This abstract base class for a message contains the common functionality
 * for messaging in the distributed computing process.
 *
 * An AbstractMessage is expected to be able to serialize itself to XML
 * format and deserialize itself again from that format. This de/serialization
 * is done using the libxml2 library and the xml_helper module of this distribution.
 *
 * The concrete subclasses of this class are responsible for connection
 * between this data object in the memory, and a "concrete" object on some
 * kind of information medium. The "concrete" object can be a file, an e-mail
 * message etc. Mainly, the concrete subclasses need to open the connection,
 * and flush (close) it afterwards; the latter is especially important for
 * writing of the messages to the medium.
 *
 * Each message has a timeout. "Timeout" means the time interval that elapses
 * between the sending of the message, and the first attempt to re-send it,
 * if no adequate reply has been received from the other party. The "timeout"
 * can be set to NO_TIMEOUT, which means that this message will never be resent again.
 * This is especially useful for messages like "Ack".
 *
 * The timeout is given in seconds and must have a minimal value. The minimal value
 * is MINIMAL_TIMEOUT_IN_SECONDS, and this constant should be reasonably set.
 * The method Set_Timeout() will take care of setting the timeout in allowed
 * bounds. However, for some types of messages (like e-mail messages), even
 * MINIMAL_TIMEOUT_IN_SECONDS can be too small; that is why the method Set_Timeout
 * is declared as virtual, and concrete subclasses are allowed to overload it.
 *
 * Currently, MINIMAL_TIMEOUT_IN_SECONDS is set to 90 seconds. For e-mail messages,
 * it should be much higher; 900 is probably the smallest reasonable value.
 *
 * The message is logically divided into a header and a body. The header is always
 * present and non-empty. As a minimum, it contains the following information:
 *  		- version of the message. Current version is 1.0. This value should
 *		  help the reading code distinguish messages that it may not be able
 *            to process completely, and tell the user to update his/her version
 *           of the software
 *		- identification of the sender
 *		- identification of the target
 *		- the counter value
 *		- the time of creation
 *		- the message type
 * If some of these required values is missing, or is nonsensical (empty sender etc.),
 * the Deserialization process should complain loudly (on the screen) to the user. 
 * Currently, it does not check time and message version, but all other necessary
 * values are checked.
 * As a rule of a thumb, the message header should contain only information that 
 * is relevant to all possible message types (metadata).
 *
 * The message also has body, which can be empty (but the element tag in the XML,
 * albeit empty, should be present anyway), or it can contain various data, often
 * very complicated and big. In some messages, the body is defined in generality,
 * independent of any specific protocols and processes; in other, it is
 * algorithm-dependent. If the body is process-independent, it is processed within
 * this class as well.
 *
 * For example: an Ack message contains the counter of the message that it replies
 * to in its body. This is generic for all environments.
 * On the other hand, the body of a Data message is almost totally algorithm-dependent.
 * That is why it is left alone in this class.
 *
 */
class AbstractMessage {
	public:
		AbstractMessage();
		AbstractMessage(const char* aSenderAddress, const char* aSenderURL, const char* aSubject, message_type aType, unsigned int aFlags, unsigned int aCounter, const char* aSenderId, const char* aTargetId, const char* aData, unsigned int aTimeout);
		virtual ~AbstractMessage();

		void 	Set_SenderAddress(const char* aValue);
		void 	Set_SenderURL(const char* aValue);
		void 	Set_TargetURL(const char* aValue);
		void	Set_Subject(const char* aValue);
		void	Set_SenderId(const char* aValue);
		void	Set_TargetId(const char* aValue);
		void	Set_Counter(unsigned int aValue);
		// This setter sets the whole flags field.
		void	Set_Flags(unsigned int aFlags);
		// This setter sets one bit flag
		void	Set_Flag(unsigned int aFlag);
		// This setter clears one bit flag
		void	Clear_Flag(unsigned int aFlag);
		void	Set_Data(const char* aValue);
		void	Set_Time(time_t aTime);
		void	Set_CurrentTime();
		void	Set_FutureTime(unsigned int aPlusSeconds);
		void	Set_Type(message_type aType);
		virtual void	Set_Timeout(unsigned int aValueInSeconds);
		void	Set_CanDestroy(bool aValue);
		void	Set_ReplyTo(unsigned int aValue);
		void	Set_AskAgain(unsigned int aValue);
		void	Set_Delay(unsigned int aValue);
		void	Set_JobId(const string& aValue);
		void	Set_DataCount(unsigned int aValue);
		void	Set_DataUnit(const char* aValue);
		void	Set_DataUnit(const string& aValue);

		const char*	Get_SenderAddress() const;
		const char*	Get_SenderURL() const;
		const string&	Get_SenderURL_String() const;
		const string&	Get_TargetURL_String() const;
		const char*	Get_Subject() const;
		const char*	Get_SenderId() const;
		const char*	Get_TargetId() const;
		unsigned int Get_Counter() const;
		unsigned int Get_Flags() const;
		bool	Is_Flag_Set(unsigned int aFlag) const;
		const char*	Get_Data() const;
		string		Get_QuitReason() const;
		time_t	Get_Time() const;
		message_type	Get_Type() const;
		unsigned int	Get_Timeout() const;
		bool		Has_Timeout() const;
	        bool		Can_Destroy() const;
		unsigned int 	Get_ReplyTo() const;
		unsigned int	Get_AskAgain() const;
		unsigned int	Get_Delay() const;
		bool		Has_Delay() const;
		string		Get_BodyPath() const;
		string		Get_JobId() const;
		unsigned int	Get_DataCount() const;
		string		Get_DataUnit() const;

	
		// aPathBody ... path to the file where the process should
		// dump the body of the message, if required		
		int	Deserialize();
		int	Serialize();

	public:
		// Fields specific for the concrete message types
		void	AddSupportedAlgorithm(const char* aAlgorithmName); ///< Only used in the Ready and NoJob message.
		void	AddUnsupportedAlgorithm(const char* aAlgorithmName);  ///< Only used in the NoJob message
		
		void	AddJobParameters(const JobParameters* aParameters, AbstractSerializator* aSerializator); ///< Only used in the JobParameters message.

		void	AddData(AbstractSerializator* aSerializator); ///< Only used in the Data message.

		void	AddQuitReason(const char* aQuitReason); ///< Only used in the Quit message.

		const vector<string>&		Get_SupportedAlgorithms() const;



		/*!
		 * This method will mark the message as processed
		 * (it has been fetched by the connector into the 
		 * incoming queue).
 		 */
		virtual int			MarkAsProcessed() = 0;
		/*! This method will mark the message as a one whose
		 * processing has been finished (it is no longer
		 * useful to the system).
		 */
		virtual int			MarkAsFinished(bool aKeepMessage) = 0;
		

	protected:
		// Serialization-specific methods
	
		// These are intended to create the basic fields in header and body.
		// These methods shouldn't probably be overloaded in derived classes.
		virtual int 		CreateHeader(xmlTextWriterPtr& aWriter) const;
		virtual int		CreateBody(xmlTextWriterPtr& aWriter) const;

		virtual int		ReadHeader(xmlTextReaderPtr& aReader);
		virtual int		ReadBody(xmlTextReaderPtr& aReader);
		int			ReadSenderInfo(xmlTextReaderPtr& aReader);

		int			ReadTargetInfo(xmlTextReaderPtr& aReader);

		virtual int		ReadBody_ReadyMessage(xmlTextReaderPtr& aReader);
		virtual int ReadBody_NoJobMessage(xmlTextReaderPtr& aReader);
		
		// This message will copy the data within the body to a file 
		// on the filesystem, without any real parsing. The distributed
		// algorithm itself will then parse the data.
		virtual int ReadBody_JobParametersMessage(xmlTextReaderPtr& aReader);
		
		virtual int ReadBody_AckMessage(xmlTextReaderPtr& aReader);
		virtual int ReadBody_LockedMessage(xmlTextReaderPtr& aReader);
		virtual int ReadBody_NodeBusyMessage(xmlTextReaderPtr& aReader);
		virtual int ReadBody_QuitMessage(xmlTextReaderPtr& aReader);
		virtual int ReadBody_DataMessage(xmlTextReaderPtr& aReader);
		virtual int ReadBody_AliveMessage(xmlTextReaderPtr& aReader);

		// These methods are called from the CreateHeader and CreateBody
		// methods. These methods should be overloaded in derived classes,
		// to provide additional functionality.
		virtual int		CreateSpecificHeaderFields(xmlTextWriterPtr& aWriter) const;
		virtual int		CreateSpecificBodyFields(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_ReadyMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_NoJobMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_JobParametersMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_AckMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_LockedMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_NodeBusyMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_QuitMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_DataMessage(xmlTextWriterPtr& aWriter) const;

		virtual int CreateBody_AliveMessage(xmlTextWriterPtr& aWriter) const;

		// Writes down sender / target
		virtual	int		WriteSenderInfo(xmlTextWriterPtr& aWriter) const;
		virtual	int		WriteTargetInfo(xmlTextWriterPtr& aWriter) const;

		/* Writes down message type and description */
		virtual int		WriteMessageType(xmlTextWriterPtr& aWriter) const;

		// Undefined
		virtual xmlTextWriterPtr	CreateOutputStream() = 0;
		virtual xmlTextReaderPtr	CreateInputStream() = 0;

		// Closes files etc. Undefined in abstract message
		virtual int			FinishDeserialization() = 0;
		virtual int			FinishSerialization() = 0;



		// De/serialization from/to XML.
		int			Deserialize(xmlTextReaderPtr& aReader);
		int			Serialize(xmlTextWriterPtr& aWriter) const;		

		int	Init();
		int	Init(const char* aSenderAddress, const char* aSenderURL, const char* aSubject, message_type aType, unsigned int aFlags, unsigned int aCounter, const char* aSenderId, const char* aTargetId, const char* aData, unsigned int aTimeout);

		// Deserialization-specific methods.

	public:
		// Auxiliary
		void	PrintHeaderInfo() const;
		
	protected:
		
		string  iSubject;		
		string  iSenderId;
		string  iSenderAddress;		
		string	iSenderURL;		
		string  iTargetId;
		string  iTargetURL;

		unsigned int iCounter;
		unsigned int iFlags;
		unsigned int iTimeoutInSeconds;		
		string iData;
		time_t iTime;	
		message_type iType;
		string	iDescription;
		

		// The path to the body of the message (relevant only for JobParameters
		// and Data messages)
		string	iFullPath_Body;

		vector<string> 	iSupportedAlgorithms;
		vector<string>  iUnsupportedAlgorithms;
		// These pointers are not owned by the message,
		// and therefore will not be destroyed in the destructor.
		const JobParameters*	iJobParameters;
		AbstractSerializator*   iSerializator;

		string	iJobId;

		unsigned int 	iDataCount;
		string  	iDataUnit;

		string		iQuitReason;


		bool	iCanDestroy;

		unsigned int iReplyTo;
		unsigned int iAskAgain;
		
		unsigned int iDelayFirstSend_Seconds; ///< If this is not 0, the message will not be sent immediately, but put to the Pending queue, and sent for the first time after this # of seconds. The lowest possible value is MINIMAL_DELAY_SEND, the highest possible value is MAXIMAL_DELAY_SEND.
};
#endif
