/*
 * abstract_mpz_matrix.cpp
 *
 *      Author: jero
 */

#include "abstract_mpz_matrix.h"
#include "gmp.h"
#include <iostream>

AbstractMpzMatrix::AbstractMpzMatrix() {
	modular=false;
	mpz_init(modulo);
	mpz_set_ui(modulo,0);
}

AbstractMpzMatrix::~AbstractMpzMatrix() {
	mpz_clear(modulo);
}

void AbstractMpzMatrix::PrintToScreen() {
	gmp_printf("\n(%ld x %ld)",GetMaximalRowIndex()+1, GetMaximalColumnIndex()+1);
	for (long row = 0;row <=maximal_row_index; row++) {
		gmp_printf("\n");
		for (long column = 0; column <= maximal_column_index; column++)
			gmp_printf("|%Zd\t",*GetValueDirect(row,column));
	}
}

int AbstractMpzMatrix::SetModulo(mpz_t aValue) {
	modular = true;
	mpz_set(modulo, aValue);
	return ConstRC::Ok;
}

int AbstractMpzMatrix::SetModulo(int aValue) {
	if (0!=aValue) {
		modular = true;
		mpz_set_ui(modulo,aValue);
	} else {
		modular = false;
	}

	return ConstRC::Ok;
}


int AbstractMpzMatrix::Randomize() {
	//time_t seconds;
	//time(&seconds);
	//srand((unsigned int) seconds);
	for (long row=0;row <=maximal_row_index; row ++) {
		for(long column = 0; column <= maximal_column_index; column ++) {
			mpz_t * pointer = GetValueDirect(row,column);
			mpz_set_ui(*pointer,random());
			if (modular) mpz_mod(*pointer, * pointer, modulo);
		}
	}
	return ConstRC::Ok;
}

int AbstractMpzMatrix::SwapRows(long aRow1, long aRow2) {
	for(long column = 0; column <= maximal_column_index; column++) {
			mpz_swap(*GetValueDirect(aRow1,column), *GetValueDirect(aRow2, column));
	}
	return ConstRC::Ok;
}

int AbstractMpzMatrix::MultiplyRow(long aRow, mpz_t aValue) {
	for (long column = 0; column <= GetMaximalColumnIndex(); column ++) {
		mpz_t * pointer = GetValueDirect(aRow,column);
		mpz_mul(*pointer, *pointer, aValue);
		if (modular) mpz_mod(*pointer, * pointer, modulo);
	}
	return ConstRC::Ok;
}

int AbstractMpzMatrix::SubtractRowMultiple(long aTargetRow, long aSourceRow,
		mpz_t aMultiple) {
	// aMultiple may be direct pointer to the pivot, we want to change it last
	for (long column = GetMaximalColumnIndex();column >= 0;column--) {
		mpz_mul(tempmul,aMultiple,*GetValueDirect(aSourceRow, column));
		mpz_t *pointer = GetValueDirect(aTargetRow,column);
		mpz_sub(*pointer, *pointer, tempmul);
		if (modular) mpz_mod(*pointer, * pointer, modulo);
	}
	return ConstRC::Ok;
}

bool AbstractMpzMatrix::IsInvertible(long aRow, long aColumn) {
	if (!modular) return !IsZero(aRow,aColumn);
	if (mpz_invert(tempmul,*GetValueDirect(aRow,aColumn), tempmul))
		return true;
	else
		return false;
	return ConstRC::Ok;
}

int AbstractMpzMatrix::GaussColumn(long aRow, long aColumn, bool aPartial) {
	GetValueMpz(aRow,aColumn,tempval);
	mpz_invert(tempval2,tempval,modulo);
	MultiplyRow(aRow, tempval2);
	long start;
	if (aPartial)
		start = aRow+1;
	else
		start = 0;
	for(long row = start;row<=GetMaximalRowIndex();row++) {
		if(row == aRow) continue;
		if (IsZero(row, aColumn)) continue;
		SubtractRowMultiple(row, aRow,*GetValueDirect(row,aColumn));
	}
	return ConstRC::Ok;
}

int AbstractMpzMatrix::GetModulo() {
	return mpz_get_si(modulo);
}

int AbstractMpzMatrix::SubtractRowMultiple(long aTargetRow, long aSourceRow,
		integer_matrix_type aMul) {
	mpz_t multiplicator;
	mpz_init(multiplicator);
	mpz_set_si(multiplicator,aMul);
	int rc = SubtractRowMultiple(aTargetRow,aSourceRow,multiplicator);
	mpz_clear(multiplicator);
	return rc;
}

int AbstractMpzMatrix::MultiplyRow(long aRow, integer_matrix_type aValue) {
	mpz_t multiplier;
	mpz_init(multiplier);
	mpz_set_si(multiplier,aValue);
	int rc = MultiplyRow(aRow,multiplier);
	mpz_clear(multiplier);
	return rc;
}

int AbstractMpzMatrix::RowProduct(mpz_t & aResult, long aRow1, long aRow2)
{
	if ((aRow1 < 0) || (aRow2 < 0))	return ConstRC::GeneralError;
	if (aRow1 > this->maximal_row_index) return ConstRC::GeneralError;
	if (aRow2 > this->maximal_row_index) return ConstRC::GeneralError;
	mpz_set_ui(aResult,0);
	for(int column=0; column <= this->maximal_column_index; column++) {
		integer_matrix_type value1, value2;
		GetValue(aRow1, column, value1);
		GetValue(aRow2, column, value2);
		mpz_mul(tempmul,*GetValueDirect(aRow1,column),*GetValueDirect(aRow2,column));
		mpz_add(aResult,aResult,tempmul);
	};
	return ConstRC::Ok;
}

int AbstractMpzMatrix::LLL() {

	int rc;
	AbstractMpzMatrix * gramova = (AbstractMpzMatrix *) clone();
	gramova->SetMaximalRowIndex(maximal_row_index);
	gramova->SetMaximalColumnIndex(maximal_row_index);
	gramova->Allocate();
	AbstractMpzMatrix * relative =  (AbstractMpzMatrix *) clone();
	relative->SetMaximalRowIndex(maximal_row_index);
	relative->SetMaximalColumnIndex(maximal_row_index);
	relative->Allocate();
	AbstractMpzMatrix * lambda =  (AbstractMpzMatrix *) clone();
	lambda->SetMaximalRowIndex(maximal_row_index);
	lambda->SetMaximalColumnIndex(maximal_row_index);
	lambda->Allocate();
	integer_matrix_type value;
	lambda->PutValueMpz(0,0,*GetValueDirect(0,0));

	for(int row1=0; row1 <= this->maximal_row_index; row1++) {
		for(int row2=0; row2 <= this->maximal_row_index; row2++) {
			if (ConstRC::Ok == (rc = RowProduct(tempval, row1, row2))) {
				gramova->PutValueMpz(row1,row2, tempval);
			} else { return rc; };

		};
	};
	// Inicializace
	int k = 2;
	int kmax = 1;
	mpz_t * d = new mpz_t[this->maximal_row_index+2];
	for (long index = 0;index <= maximal_row_index+1; index++)
		mpz_init(d[index]);
	mpz_set_ui(d[0],1);
	this->RowProduct(d[1],0,0);
	mpz_t u;
	mpz_init(u);
	mpz_t q;
	mpz_init(q);
	for(int column=0; column <= this->maximal_row_index; column++) {
		for(int row=0; row <= this->maximal_row_index; row++) {
			relative->PutValue(row,column, row == column ? 1 : 0);

		};
	};

	do
	{
		// cout << "Jeden krok: "<<k <<"   "<<kmax;
		// cout<< "\nD: "; for (int ou=0;ou<=3;ou++) { cout << d[ou] <<", ";};
		// this->PrintToScreen();
		// gramova->PrintToScreen();
		// relative->PrintToScreen();
		// 2.
		if (k>kmax) {
			kmax=k;
			for (int j=1;j<=k;j++) {
				// cout << " KMAX "<< k <<" " <<j<<"   |";
				rc = this->RowProduct(u,k-1,j-1);
				if (ConstRC::Ok != rc) {return rc;};
				// cout << "{u:" << u <<",";
				for(int i=1;i<j;i++) {
					// cout << u <<",";
					/*lambda->GetValue(k-1,i-1, value1);
					lambda->GetValue(j-1,i-1,value2);
					u = (d[i]*u - value1*value2)/d[i-1];*/
					mpz_mul(tempval,*lambda->GetValueDirect(k-1,i-1),*lambda->GetValueDirect(j-1,i-1));
					mpz_mul(tempval2,d[i],u);
					mpz_sub(tempadd, tempval2, tempval);
					mpz_div(u,tempadd,d[i-1]);
				};
				// cout << " u:" << u <<"}";
				if (j<k) { lambda->PutValueMpz(k-1,j-1,u);};
				if (j==k) { mpz_set(d[k],u); /*cout <<"(d["<<k<<"]="<<u<<")";*/};
			};
		};
		// lambda->PrintToScreen();
		bool cycle3 = false;
		do {
			//REDI(k,k-1)
			// cout << "REDI (" << k<<","<<k-1<<")";
			// cout << "Qéčko:" << abs(2*lambda->GetXY(k-1,k-2)) << " > ";
			// cout << d[k-1] << " | ";
			mpz_mul_ui(tempval,*lambda->GetValueDirect(k-1,k-2),2);
			mpz_abs(tempval,tempval);
			if (mpz_cmp(tempval, d[k-1]) >0) {
				// cout << "REDI: " << k << "    " << k-1 << "\n";
				// lambda->PrintToScreen();
				mpz_mul_ui(tempval,*lambda->GetValueDirect(k-1,k-2),2);
				mpz_add(tempval,tempval,d[k-1]);
				mpz_mul_ui(tempval2,d[k-1],2);
				mpz_div(q,tempval,tempval2);
				// cout << "q=" << q <<";";
				relative->SubtractRowMultiple(k-1,k-2,q);
				this->SubtractRowMultiple(k-1,k-2,q);
				mpz_mul(tempval,q,d[k-1]);
				mpz_sub(tempval,*lambda->GetValueDirect(k-1,k-2), tempval);
				lambda->PutValueMpz(k-1,k-2,tempval);
				for ( int i=1;i<=k-2;i++) {
					mpz_mul(tempval,q,*lambda->GetValueDirect(k-2,i-1));
					mpz_sub(tempval,*lambda->GetValueDirect(k-1,i-1),tempval);
					lambda->PutValueMpz(k-1,i-1,tempval);
				};

				// this->PrintToScreen();
			};
			// lambda->PrintToScreen();
			// cout<< "\nD: "; for (int ou=0;ou<=3;ou++) {  gmp_printf("%Zd,",d[ou]);};
			// cout << 4*d[k]*d[k-2] << " < " <<3*d[k-1]*d[k-1] - lambda->GetXY(k-1,k-2)* lambda->GetXY(k-1,k-2);
			mpz_mul(tempval,d[k-1],d[k-1]);
			mpz_mul_ui(tempval,tempval,3);
			mpz_mul(tempval2,*lambda->GetValueDirect(k-1,k-2),*lambda->GetValueDirect(k-1,k-2));
			mpz_sub(tempadd,tempval,tempval2);
			mpz_mul(tempmul,d[k],d[k-2]);
			mpz_mul_ui(tempmul,tempmul,4);
			if (mpz_cmp(tempmul,tempadd)<0) {
				cycle3 = true;
				//SWAP(k)
				//cout << "SWAP: " << k << "  "<< k-1 << " \n";
				relative->SwapRows(k-1,k-2);
				this->SwapRows(k-1,k-2);
				//lambda->PrintToScreen();
				if (k>2) {
					for (int j=1;j<=k-2;j++) {
						integer_matrix_type temp, temp2;
						mpz_set(tempval,*lambda->GetValueDirect(k-1,j-1));
						mpz_set(*lambda->GetValueDirect(k-1,j-1),*lambda->GetValueDirect(k-2,j-1));
						lambda->PutValueMpz(k-2,j-1, tempval);

					};
				};
				// tempmul = tlambda
				// tempadd = tbe
				mpz_set(tempmul,*lambda->GetValueDirect(k-1,k-2));
				// cerr<< "\nJuD: "; for (int ou=0;ou<=3;ou++) {  gmp_fprintf(stderr,"%Zd,",d[ou]);};
				mpz_mul(tempval,d[k-2],d[k]);
				mpz_mul(tempval2, tempmul, tempmul);
				mpz_add(tempval, tempval, tempval2);
				mpz_div(tempadd,tempval, d[k-1]);
				for (int i = k+1;i<=kmax;i++) {
					integer_matrix_type tt;
					lambda->GetValue(i-1,k-1,tt);
					integer_matrix_type aux_value1;
					lambda->GetValue(i-1,k-2, aux_value1);
					mpz_mul(tempval, d[k], *lambda->GetValueDirect(i-1,k-2));
					mpz_mul(tempval2,tempmul, *lambda->GetValueDirect(i-1,k-1));
					mpz_sub(tempval, tempval,tempval2);
					mpz_div(tempval,tempval,d[k-1]);
					lambda->PutValueMpz(i-1,k-1, tempval);
					mpz_mul(tempval, tempadd, *lambda->GetValueDirect(i-1,k-1));
					mpz_mul(tempval2,tempmul, *lambda->GetValueDirect(i-1,k-1));
					mpz_add(tempval, tempval, tempval2);
					mpz_div(tempval, tempval, d[k]);
					lambda->PutValueMpz(i-1,k-2, tempval);
				}
				//cout << "<--tlambda->>>" << tlambda << " tbe--->" << tbe <<"<";
				mpz_set(d[k-1],tempadd);
				k = max(2,k-1);
				//lambda->PrintToScreen();
			} else {
				//cout << "NESWAP: " << k << "  "<< k-1 << " \n";
				cycle3 = false;
			};
			//this->PrintToScreen();
			// gramova->PrintToScreen();
			// relative->PrintToScreen();
		} while (cycle3);
		for (int l=k-2;l>0;l--) {
			//REDI(k,k-1)
			//cout << "REDI2 (" << k<<","<<k-1<<")";
			mpz_set(tempmul,*lambda->GetValueDirect(k-1,l-1));
			mpz_mul_ui(tempval,tempmul,2);
			mpz_abs(tempval, tempval);
			if (mpz_cmp(tempval, d[l])> 0) {
				mpz_mul_ui(tempval, *lambda->GetValueDirect(k-1,l-1),2);
				mpz_add(tempval,tempval,d[l]);
				mpz_div(tempval, tempval, d[l]);
				mpz_div_ui(q,tempval,2);
				//cout << "q=" << q <<";";
				relative->SubtractRowMultiple(k-1,l-1,q);
				this->SubtractRowMultiple(k-1,l-1,q);
				mpz_set(tempval,*lambda->GetValueDirect(k-1,l-1));
				mpz_mul(tempval2,q,d[l]);
				mpz_sub(tempval, tempval, tempval2);
				lambda->PutValueMpz(k-1,l-1, tempval);
				for ( int i=1;i<=l-1;i++) {
					integer_matrix_type aux_value1, aux_value2;
					mpz_mul(tempval,q,*lambda->GetValueDirect(l-1,i));
					mpz_sub(tempval, *lambda->GetValueDirect(k-1,i-1), tempval);
					lambda->PutValueMpz(k-1,i-1,tempval);
				};
			};
			//this->PrintToScreen();
		};
		k++;
	} while (k <= this->maximal_row_index+1);


	delete gramova;
	return ConstRC::Ok;

}

int AbstractMpzMatrix::Test() {
	time_t seconds;
	time(&seconds);
	srand((unsigned int) seconds);
	for(int n = 0; n < maximal_column_index*maximal_row_index/2; n++) {
		integer_matrix_type value = rand();
		long column = rand() % maximal_column_index;
		long row = rand() % maximal_row_index;
		integer_matrix_type readvalue;
		PutValue(row, column, value);
		GetValue(row, column, readvalue);
		if (readvalue != value) {
			return ConstRC::GeneralError;
		};
	};
	mpz_t value, readvalue;
	mpz_init(value);
	mpz_init(readvalue);
	for(int n = 0; n < maximal_column_index*maximal_row_index/2; n++) {
			mpz_set_ui(value,rand());
			mpz_mul_ui(value,value,rand());
			mpz_mul_ui(value,value,rand());
			mpz_add_ui(value,value,rand());
			long column = rand() % maximal_column_index;
			long row = rand() % maximal_row_index;
			PutValueMpz(row, column, value);
			GetValueMpz(row, column, readvalue);
			if (0 != mpz_cmp(readvalue,value)) {
				return ConstRC::GeneralError;
			};
	};
	mpz_clear(value);
	mpz_clear(readvalue);
	return ConstRC::Ok;
}
