/*!
 * \file 
 * \brief Definitions for class AbstractSerializator
 * 
 */
#ifndef __ABSTRACT_SERIALIZATOR__
#define __ABSTRACT_SERIALIZATOR__

#include <string>
# include "constants.h"
#include "common_definitions.h"
#include "job_parameters.h"
#include "algorithm.h"

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <libxml/xmlreader.h>

using namespace std;

/*!  * \brief Abstract common ancestor for serializable clases
 *
 *
 */
class AbstractSerializator : public virtual Algorithm {
     public:

	/// constructors
	AbstractSerializator();
	AbstractSerializator(Algorithm *aAlgorithmInstance, const char* aDirectoryName, int aCompressionLevel);

	virtual ~AbstractSerializator();
	
	int	Get_CompressionLevel() const;
	void	Set_CompressionLevel(int aCompressionLevel);
	void  	Set_DirectoryName(const string& aDirectoryName);
	
	virtual int Serialize(int aAction) = 0;
	virtual int Deserialize(int aAction) = 0;	

	virtual int SerializeResult() = 0;
	virtual int SerializeJob(xmlTextWriterPtr& aWriter, const JobParameters& aParameters) = 0;
	// Loads Job data from the indicated XML reader to aParameters.
	virtual int DeserializeJob(xmlTextReaderPtr& aReader, JobParameters& aParameters) = 0;
	// Loads Job data from the default file to aParameters.
	virtual int DeserializeJob(JobParameters& aParameters) = 0;

	virtual int DeserializeJob(const char* aPath, JobParameters& aParameters) = 0;		
	virtual int SerializeData(xmlTextWriterPtr& aWriter) = 0; ///< Used in parallelization to send fresh data from Node to Center.

	virtual int DeserializeData(const char* aPath) = 0;

	virtual Algorithm*	CreateInstance(const JobParameters& aParameters) const = 0;

	// From Algorithm
	virtual const char* 	AlgorithmName()       const = 0;
	virtual JobParameters*  CreateNewParameters() const = 0;


	static void	AppendPathSeparator(string& aPath);

     
	/**
		This function does not return reference, but a copy of 
		the string. This is because we do not want the user
		to mess with internal member of this class.

	*/
	string    CreateFileNamePlain(const char* aFile);

	void    CreateFileNameCompressed(const char* aFile);
	
	void	CreateFileNameAuto(const char* aFile);

	// Removes both XXX.xml and XXX.xml.z from the current directory
	void    Remove(const char* aFile);

	FILE*	OpenFile(const char* aMode);

	virtual void 	RegisterInstance(Algorithm*	aAlgorithm) = 0;

	virtual void	PrintInstance() const;

protected:
	virtual int	CopyTextFileToWriter(xmlTextWriterPtr& aWriter, FILE* aFile);

protected:

	// The pointer does not belong to Serializator and will not be deleted by the local destructor
	Algorithm *iAlgorithmInstance;
	string	       iDirectoryName;
	int	       iCompressionLevel;
	string		iAuxName;
	bool		iInfoMode;
};

#endif

