/*!\file
 * \brief implementation of AbstractSieve
 */
#include "abstract_sieve.h"
#include <errno.h>
#include <string.h>
#include "log.h"

using namespace std;


const string	 AbstractSieve::iGenerateNewPolyText = GEN_NEW_POLY_TEXT;
const string	 AbstractSieve::iQuadraticMachineText = QUAD_MACHINE_TEXT;
const string	 AbstractSieve::iPerformSievingByPolyText = PERF_SIEVING_TEXT;
const string	 AbstractSieve::iSievingLoopText = SIEVING_LOOP_TEXT;
const string	 AbstractSieve::iRefillMainMatrixText = REFILL_MAINM_TEXT;
const string	 AbstractSieve::iSortSmoothText = SORT_SMOOTHS_TEXT;
const string	 AbstractSieve::iDivisorsOfAText = DIVISORS_OFA_TEXT;
const string	 AbstractSieve::iSmallFactorText = SMALL_FACTOR_TEXT;
const string	 AbstractSieve::iCountCyclesText = COUNT_CYCLES_TEXT;
const string	 AbstractSieve::iBlockSieveText = BLOCK_SIEVES_TEXT;
const string	 AbstractSieve::iFillHashtableText = FILL_HSTABLE_TEXT;
const string	 AbstractSieve::iFillHashtableNegText = FILL_HST_NEG_TEXT;
const string	 AbstractSieve::iRootsUpdateText = ROOTS_UPDATE_TEXT;
const string	 AbstractSieve::iFindAndFactorText = FIND_FACTORZ_TEXT;
const string	 AbstractSieve::iUnknownPartText = UNKNOWN_PART_TEXT;

/*
 The following input causes segfault at transpose_sparse_bit_matrix.c, line 677 (as of Mar 6, 2004)

 valgrind ./a.out -q -m10000 -f250000 -N18421584111341447874487112377 -l -p5 -e16 -w -v -uf -it -i
*/

// CONSTRUCTORS AND DESTRUCTORS
/*!
	  Both of the constructors initialize the member variables according to the following rule:
	  	- the boolean variables are initialized to defaults, which are \a FALSE
		- the linalg_type variable matrix_type is set to \a EBitMatrix;
		- the linalg_algo variable algorithm_type is set to \a EBlockLanczos
		- the mpz_t variables are initialized using mpz_init and their status is marked in the \a inited variable by setting the corresponding
		  bits to 1; they are set to defaults given by appropriate symbolic constants defined in definitions.h
		- the variable \a k is set to default given by  appropriate symbolic constant defined in definitions.h

*/
AbstractSieve::AbstractSieve(){
	this->inited = 0; // Status of initialization of N and kN
	this->k      = INIT_VALUE_k;
	this->info_mode           = FALSE;
	this->debug_mode	  = FALSE;
	this->assertion_mode      = FALSE;
	this->rank_calculation_mode=FALSE;
	this->matrix_type         = EBitMatrix;
	this->algorithm_type      = EBlockLanczos;
	this->frontend = false;
	this->frontend_pipe = NULL;
	mpz_init_set_si(this->N,INIT_VALUE_N);
	mpz_init(this->kN);
	mpz_init_set_ui(this->factor_1,INIT_VALUE_FACTOR);
	mpz_init_set_ui(this->factor_2,INIT_VALUE_FACTOR);
	mpz_init_set_ui(this->start_time_main_process,clock());  // This variable contains time of start of a process
	mpz_init(start_time_subprocess_level_1); // This variable contains time of start of a level-1 subprocess
	mpz_init(start_time_subprocess_level_2); // This variable contains time of start of a level-2 subprocess
	this->start_unix_time=time(NULL);

	this->inited |= (N_INITED | kN_INITED | F1_INITED | F2_INITED | STIME_MAIN_INITED | STIME_L1_INITED | STIME_L2_INITED);
	this->iSerializator  = NULL;
	this->iDataCompressionLevel = 0;
	this->iSievingPhase = ESievingNotStarted;
}

/*!
	  The second constructor moreover sets the value \a N to the value given by argument.
*/
AbstractSieve::AbstractSieve(mpz_t N){
	this->inited = 0; // Status of initialization of N and kN
	this->k      = INIT_VALUE_k;
	this->info_mode           = FALSE;
	this->debug_mode	  = FALSE;
	this->assertion_mode      = FALSE;
	this->rank_calculation_mode=FALSE;
	this->matrix_type         = EBitMatrix;
	this->algorithm_type      = EBlockLanczos;
	this->frontend = false;
	this->frontend_pipe = NULL;
	mpz_init(this->N);
	mpz_init(this->kN);
	mpz_set(this->N,N);
	mpz_init_set_ui(this->factor_1,INIT_VALUE_FACTOR);
	mpz_init_set_ui(this->factor_2,INIT_VALUE_FACTOR);
	mpz_init_set_ui(this->start_time_main_process,clock());  // This variable contains time of start of a process
	mpz_init(start_time_subprocess_level_1); // This variable contains time of start of a level-1 subprocess
	mpz_init(start_time_subprocess_level_2); // This variable contains time of start of a level-2 subprocess
	this->start_unix_time=time(NULL);

	this->inited |= (N_INITED | kN_INITED | F1_INITED | F2_INITED | STIME_MAIN_INITED | STIME_L1_INITED | STIME_L2_INITED);

	this->iSerializator  = NULL;
	this->iDataCompressionLevel = 0;
	this->iSievingPhase = ESievingNotStarted;
	this->iLineNr = 0;
}

/*!
	  The destructor tries to mpz_clear() all the mpz_t member variables, whose corresponding bit in "inited" is set to 1, and after each successful
	  clearance sets the corresponding bit to 0.

	  The destructor is virtual, which is a recommended programming practice in base classes.

*/
AbstractSieve::~AbstractSieve(){

	delete this->iSerializator;
	this->iSerializator = NULL;

        if (this->inited & N_INITED){
		mpz_clear(this->N);
		this->inited &= (~N_INITED);
	}
	if (this->inited & kN_INITED){
		mpz_clear(this->kN);
		this->inited &= (~kN_INITED);
	}
	if (this->inited & F1_INITED){
		mpz_clear(this->factor_1);
		this->inited &= (~F1_INITED);
	}
	if (this->inited & F2_INITED){
		mpz_clear(this->factor_2);
		this->inited &= (~F2_INITED);
	}
	if (this->inited & STIME_MAIN_INITED){
		mpz_clear(this->start_time_main_process);
		this->inited &= ~STIME_MAIN_INITED;
	}
	if (this->inited & STIME_L1_INITED){
		mpz_clear(this->start_time_subprocess_level_1);
		this->inited &= ~STIME_L1_INITED;
	}
	if (this->inited & STIME_L2_INITED){
		mpz_clear(this->start_time_subprocess_level_2);
		this->inited &= ~STIME_L2_INITED;
	}

	if (this->inited) {
		CLog::LogError(WHEREARG,"Abstract sieve destructor","something was not deleted",15,this->inited,"Check destructors and table of symbolic constants for this value.");
	}

	if (this->frontend)
		fclose(this->frontend_pipe);


}

// GETTERS AND SETTERS
void AbstractSieve::Set_N(mpz_t aValue){
	if (!(this->inited & N_INITED)){
		mpz_init(this->N);
	}
	mpz_set(this->N,aValue);
}

void AbstractSieve::Set_k(int aValue){
	this->k = aValue;
}

void AbstractSieve::SetAssertionMode(bool aValue){
	this->assertion_mode = aValue;
}

void AbstractSieve::SetInfoMode(bool aValue){
	this->info_mode = aValue;
}

void AbstractSieve::SetDebugMode(bool aValue){
	this->debug_mode = aValue;
}

void AbstractSieve::SetFrontendPipe(string aName){
	this->FrontendPipe = aName;
	this->frontend = true;
	if (NULL == (this->frontend_pipe=fopen(aName.c_str(),"w"))) {
		CLog::LogError(WHEREARG, "Opening frontend pipe","Unable to open frontend pipe - check path",0,0,aName.c_str());
		this->frontend = false;
	}
}

void AbstractSieve::SetRankCalculationMode(bool aValue){
	this->rank_calculation_mode = aValue;
}

void AbstractSieve::SetMatrixType(linalg_type aValue){
	this->matrix_type = aValue;
}

void AbstractSieve::SetLinalgAlgorithm(linalg_algo aValue){
	this->algorithm_type = aValue;
}

void AbstractSieve::SetDataCompressionLevel(int aDataCompressionLevel)
{
	this->iDataCompressionLevel = aDataCompressionLevel;
}

void AbstractSieve::Get_N(mpz_t aRop) const {
	if (this->inited & N_INITED)
	mpz_set(aRop,this->N); else mpz_set_ui(aRop,0);
}

void AbstractSieve::Get_Factor1(mpz_t aRop) const {
	if (this->inited & F1_INITED){
		mpz_set(aRop,this->factor_1);
	}
}

void AbstractSieve::Get_Factor2(mpz_t aRop) const {
	if (this->inited & F2_INITED){
		mpz_set(aRop,this->factor_2);
	}
}

int AbstractSieve::Get_k() const {
	return (this->k);
}

bool AbstractSieve::GetAssertionMode() const {
	return (this->assertion_mode);
}

bool AbstractSieve::GetInfoMode() const {
	return (this->info_mode);
}

string AbstractSieve::GetFrontendPipe() const {
	return (this->FrontendPipe);
}

bool AbstractSieve::GetDebugMode() const {
	return (this->debug_mode);
}

bool AbstractSieve::GetRankCalculationMode() const {
	return (this->rank_calculation_mode);
}

linalg_type AbstractSieve::GetMatrixType() const {
	return (this->matrix_type);
}

linalg_algo AbstractSieve::GetLinalgAlgorithm() const {
	return (this->algorithm_type);
}

int AbstractSieve::GetDataCompressionLevel() const {
	return (this->iDataCompressionLevel);
}


// STATIC UTILS
/*!
	  This is a static utility to print a message on screen. This message has an optional number. If \a aNumber is specified as \a NO_NUMBER symbolic constant, no number is printed.

*/
void AbstractSieve::NormalMessage(const char* aMessage, int aNumber){
	if (aMessage != NULL) {
		printf("\n");
		printf("%s",aMessage);
		if (aNumber != NO_NUMBER){
			printf(" Number: %d.",aNumber);
		}
		fflush(stdout);
	}
}

// a directory name must end with PATH_SEPARATOR
bool  AbstractSieve::TestDirectoryAccess(const char* aDirectoryName)
{
	if (!aDirectoryName)
	{
		CLog::LogError(WHEREARG,"Abstract sieve","Test directory access",0,0,"You cannot specify directory name as null.");
		return FALSE;
	}
	FILE* f;
	int len = (strlen(aDirectoryName)+20)*sizeof(char);
	char* aux = (char*) malloc(len);
	memset(aux,0,len);
	strcat(aux,aDirectoryName);
	strcat(aux,"tmp");
	f = fopen(aux,"w");
	if (!f)
	{
		int err = errno;
		CLog::LogError(WHEREARG,"Abstract sieve","Test directory access",err,0,"Could not create file in indicated directory");
		printf("Name of tested file: %s.\n",aux);
		if (err == ENOENT)
		{
			CLog::LogError(WHEREARG,"Abstract sieve","Test directory access",err,0,"Directory does not exist. Please create it first, and assign sufficient rights for write access");
		}
		if (err == EACCES)
		{
			CLog::LogError(WHEREARG,"Abstract sieve","Test directory access",err,0,"Directory does not have sufficient rights. Please assign sufficient rights for write access");
		}
		free(aux);
		return FALSE;
	}
	fclose(f);
	remove(aux);
	free(aux);
	return TRUE;
}

/*!
	  A static utility, which gets number of seconds, recalculates it into days, hours and minutes, and prints out the result. Useful when
	  printing out information about duration of processes.

*/
void AbstractSieve::TimeFormat(int aSeconds){
	int days  = aSeconds / 86400;
	aSeconds -= days*86400;
	int hours = aSeconds / 3600;
	aSeconds -= hours*3600;
	int minutes = aSeconds/60;
	aSeconds -= minutes*60;
	printf("%0d d %0d h %0d m %0d sec.", days,hours,minutes,aSeconds);
}

/*!
	  This utility prints out message about duration of a specified process or subprocess on a given level. It cannot be static, since it
	  uses the non-static member variables.

*/
void AbstractSieve::TimeMessage(const char* aProcess, int aLevel){
	mpz_t aux_1, aux_2;
	mpz_init(aux_1);
	mpz_init(aux_2);
	printf("Process ");
	if (aProcess != NULL){
		printf("%s ",aProcess);
	}
	printf("took ");
	mpz_set_ui(aux_1,time(NULL));
	switch (aLevel){
		case MAIN_LEVEL: mpz_sub(aux_1,aux_1,this->start_time_main_process);
				  break;
		case LEVEL_1:	  mpz_sub(aux_1,aux_1,this->start_time_subprocess_level_1);
				  break;
		case LEVEL_2:    mpz_sub(aux_1,aux_1,this->start_time_subprocess_level_2);
				  break;
	}
	TimeFormat(mpz_get_ui(aux_1));
	printf("\n");
	mpz_clear(aux_1);
	mpz_clear(aux_2);
}

/*!
	  This is a static utility to print an error message on screen. The error message is composed of information stored in the argument \a aContext.

*/
void AbstractSieve::ErrorMessage(error_context* aContext) {
	fprintf(stderr,"\n");
	if (aContext->process != NULL){
		fprintf(stderr,"Process %s: ",aContext->process);
	} else {
		fprintf(stderr,"Unknown process: ");
	}
	if (aContext->problem != NULL){
		fprintf(stderr,"had following problem: %s. ",aContext->problem);
	} else {
		fprintf(stderr,"had unknown problem. ");
	}
	fprintf(stderr,"Error code was %d, error label was %d ",aContext->errorCode,aContext->label);
	if (aContext->sourceFile != NULL){
		fprintf(stderr,"(file %s). ",aContext->sourceFile);
	} else fprintf(stderr," (unknown file).");
	if (aContext->explanation != NULL){
		fprintf(stderr,"%s.\n",aContext->explanation);
	}
}

/*!
	  This is a static utility ensuring proper allocation or reallocation of the given pointer. The pointer is double-starred, since otherwise
	  the memory allocation would influence only a local copy of the pointer. If allocation or reallocation was unsuccessful, the machine
	  returns \a NULL.
	  AllocationMachine is the central point of memory management in MPQS/SIQS. All calls for memory allocation, except for mpz_init(), should  go through this method.

*/
void* AbstractSieve::AllocationMachine(void** aPointer, long aSize){
	void* poi;
	if (*aPointer == NULL){
		poi = malloc(aSize);
		if (poi == NULL) {
			return NULL;
		}
		else {
			return (poi);
		}
	} else {
		poi = *aPointer;
		poi = realloc(poi, aSize);
		if (poi == NULL) {
			return NULL;
		}
		else {
			return (poi);
		}
	}
}

/*!
	  A variant of the preceding utility, which in case of failure prints out an error message with given context.

*/
void* AbstractSieve::AllocationMachine(void** aPointer, long aSize, error_context* aContext){
	void* poi;
	if (*aPointer == NULL){
		poi = malloc(aSize);
		if (poi == NULL) {
			ErrorMessage(aContext);
			return NULL;
		}
		else {
			return (poi);
		}
	} else {
		poi = *aPointer;
		poi = realloc(poi, aSize);
		if (poi == NULL) {
			ErrorMessage(aContext);
			return NULL;
		}
		else {
			return (poi);
		}
	}
}

/*!
	  This is a static utility which calculates and returns the absolute value of the input parameter.
*/
int  AbstractSieve::AbsInt(int aValue){
	if (aValue >=0) return aValue;
	else return ((-1)*aValue);
}

void AbstractSieve::PrintTimeInfo(int aWhich, double& aInProcess, double& aTotal)
{
	string text = iUnknownPartText;
	switch  (aWhich)
	{
		case GEN_NEW_POLY:
			text = iGenerateNewPolyText;
			break;
		case QUAD_MACHINE:
			text = iQuadraticMachineText;
			break;
		case PERF_SIEVING:
			text = iPerformSievingByPolyText;
			break;
		case SIEVING_LOOP:
			text = iSievingLoopText;
			break;
		case REFILL_MAINM:
			text = iRefillMainMatrixText;
			break;
		case SORT_SMOOTHS:
			text = iSortSmoothText;
			break;
		case DIVISORS_OFA:
			text = iDivisorsOfAText;
			break;
		case SMALL_FACTOR:
			text = iSmallFactorText;
			break;
		case COUNT_CYCLES:
			text = iCountCyclesText;
			break;
		case BLOCK_SIEVES:
			text = iBlockSieveText;
			break;
		case FILL_HSTABLE:
			text = iFillHashtableText;
			break;
		case FILL_HST_NEG:
			text = iFillHashtableNegText;
			break;
		case ROOTS_UPDATE:
			text = iRootsUpdateText;
			break;
		case FIND_FACTORZ:
			text = iFindAndFactorText;
			break;
		default:
			text = iUnknownPartText;
			break;
	}
	float abs = aInProcess/(double) CLOCKS_PER_SEC;
	float pct = 0;
	if (aTotal)
	{
		pct = 100*(aInProcess/(double) aTotal);
	}
	printf("%s %12.4f sec (%6.4f %%)\n",text.c_str(),abs,pct);
}

AbstractSerializator* AbstractSieve::GetSerializator() const
{
	return iSerializator;
}

sieving_phase AbstractSieve::GetSievingPhase() const
{
	return iSievingPhase;
}

void AbstractSieve::SetSievingPhase(sieving_phase aValue)
{
	iSievingPhase = aValue;
}

void AbstractSieve::SetLineNr(int aValue)
{
	iLineNr = aValue;
}

int  AbstractSieve::GetLineNr() const
{
	return iLineNr;
}

int  AbstractSieve::GetStartUnixTime() const
{
	return start_unix_time;
}
