/*!
 * \file
 * \brief Definitions for class AbstractSieve.
 *
 */
#ifndef __ABSTRACT_SIEVE__
#define __ABSTRACT_SIEVE__

#include "common_definitions.h"
#include <gmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "ttr_class.h"
#include "abstract_serializator.h"
#include "algorithm.h"
#include "aux_structures.h"

// C++ STD Library includes
#include <string>

using namespace std;

// Forward declaration
class AbstractSerializator;

// general structures used in quadratic sieving

enum sieving_phase { ESievingNotStarted = 0, ESievingRunning = 100, ESievingFinished = 200 };

/*!
This enumeration typedef is used in QuadraticSieve to distinguish three modes of sieving:
	-	EExactDivisionMode, which means sieving with division (instead of logarithm additions) and with all prime powers.
	-	ERoughDivisionMode, which means sieving with division (instead of logarithm additions) and without prime powers.
	-	ERoughLogarithmicMode, which means sieving with logarithmic approach and without prime powers.
	This is, more or less, a heritage of the past. In old versions of the MPQS, programmed in C, those three modes were really
	supported, but since then the Division modes have proved themselves so slow (or rather unusably slow), that their support was removed
	utterly, and the corresponding methods are empty, returning only NOT_SUPPORTED. See documentation of QuadraticSieve for details.
*/
enum sieving_mode   { EExactDivisionMode = 1, ERoughDivisionMode, ERoughLogarithmicMode };

/*!
This enumeration typedef is used to distinguish modes of relation gathering:
	-	EDoNotUseVariations, which means 'collect smooth values only';
	-	ELargePrimeV, which means 'collect also 1-partials';
	-	EDoubleLargePrimeV, which means 'collect also 1- and 2-partials';
        -       ETrippleLargePrimeV, which means 'collect also 1- and 2-partials and 3-partials' relations; -- NOT SUPPORTED
*/
enum variations_types { EDoNotUseVariations = 0, ELargePrimeV = 1, EDoubleLargePrimeV = 2, ETrippleLargePrimeV = 3 };

/*!
 Type of B matrix used in Linear Algebra part
*/
enum linalg_type    { ESparseBitMatrix = 0, EBitMatrix, EBCMatrix, ENormalMatrix };

/*!
This enumeration typedef is used to determine the algorithm for solution of the linear algebra system. The items are self-explaining.

*/
enum linalg_algo    { EBlockLanczos = 0, EBlockWiedemann };

/*!
	This enumeration typedef is used to determine mode of checking the values in the sieving interval.
	If all values should be checked for smoothness (extremely slow), the "ECheckAll" variant is used.
	If the check threshold should be readjusted automatically, the "ECheckAuto" variant is used.

*/
enum check_mode     { ECheckNormal = 0, ECheckAll, ECheckAuto };

/*!
   	This enumeration typedef is used to determine whether accepted/rejected statistics will be printed out after each polynomial change.
	It should be replaced by a bool value, since there is no logical third possibility.

*/
enum print_sort     { EDoNotPrintSort = 0, EPrintSort };

/*! \brief  The most abstract class, usable for all types of sieving, not only QS

  This is the most abstract class, usable for all types of sieving, not only QS. Therefore it does not really have "executive" methods, only constructors, destructors, getters, setters and a few static functions, usable in many contexts (error messages, safe memory alloc/realloc).



 */
class AbstractSieve : public virtual Algorithm {
     public:
	/// constructor
	AbstractSieve();
	/*!
	 * @param N factored number
	 */
	AbstractSieve(mpz_t N);
	virtual ~AbstractSieve();

	// GETTERS AND SETTERS
	/// Sets AbstractSieve::N (factored number)
	void Set_N(mpz_t aValue);
	/// Sets AbstractSieve::k ("small multiplier")
	void Set_k(int aValue);
	/// Sets AbstractSieve::assertion_mode
	void SetAssertionMode(bool aValue);
	/// Sets AbstractSieve::info_mode
	void SetInfoMode(bool aValue);

	void SetDebugMode(bool aValue);
	/// Sets AbstractSieve::frontend_pipe
	void SetFrontendPipe(string aName);
	///  Sets AbstractSieve::rank_calculation_mode
	void SetRankCalculationMode(bool aValue);
	///  Sets AbstractSieve::matrix_type
	void SetMatrixType(linalg_type aValue);
	/// Sets AbstractSieve::algorithm_type
	void SetLinalgAlgorithm(linalg_algo aValue);
	void SetDataCompressionLevel(int aDataCompressionLevel);
	void SetSievingPhase(sieving_phase aValue);
	void SetLineNr(int aValue);

	void Get_N(mpz_t aRop) const;
	void Get_Factor1(mpz_t aRop) const;
	void Get_Factor2(mpz_t aRop) const;
	int  Get_k() const;
	bool GetAssertionMode() const;
	bool GetInfoMode() const;
	string GetFrontendPipe () const;
	bool GetDebugMode() const;
	bool GetRankCalculationMode() const;
	AbstractSerializator* GetSerializator() const;
	linalg_type GetMatrixType() const;
	linalg_algo GetLinalgAlgorithm() const;
	int  GetDataCompressionLevel() const;
	sieving_phase GetSievingPhase() const;
	int  GetLineNr() const;

	int GetStartUnixTime() const;

	/// Prints \a aMessage and optionally \a aNumber.
	static void NormalMessage(const char* aMessage, int aNumber);
	/// Prints information about an error
	static void ErrorMessage(error_context* aContext);

	static void* AllocationMachine(void** aPointer, long aSize);
	static void* AllocationMachine(void** aPointer, long aSize, error_context* aContext);
	static void TimeFormat(int aSeconds);
 	static int  AbsInt(int aValue);

	static bool  TestDirectoryAccess(const char* aDirectoryName);

	void TimeMessage(const char* aProcess, int aLevel); // uses mpz_t timers, cannot be static
	static inline void AddTime(double& aAggregator, int& aStart)
	{
		int diff = clock() - aStart;
		if (diff > 0){
			aAggregator += diff;
		}
	};

	void PrintTimeInfo(int aWhich, double& aInProcess, double& aTotal);


     protected:
	/// factored number
	mpz_t  N;
	/// factored number multiplied by k. Is determined from N and k, of course.
	mpz_t kN;
	/// If sieving succeeded, it contains one factor of N, else 0
	mpz_t factor_1;
	/// If sieving succeeded, it contains one factor of N, else 0
	mpz_t factor_2;
	/// Start time (unix time)
	int start_unix_time;
	/// This variable contains time of start of a process
	mpz_t start_time_main_process;
	/// This variable contains time of start of a level-1 subprocess
	mpz_t start_time_subprocess_level_1;
	/// This variable contains time of start of a level-2 subprocess
	mpz_t start_time_subprocess_level_2;
	/// Status of initialization of mpz_t variables N and kN
	long   inited;
	/// "small multiplier"
	int    k;
	/// Will some more information about run be displayed?
	bool   info_mode;
	bool   debug_mode;
        /// Will assertions be run?
	/*! 	  determines whether assertions will be executed. If set,
	 * several invariants are tested at each iteration, which means
	 * slowdown, but gives
	  a chance to find possible bugs. The default value is false, and it is
	  reset to true by quadratic_sieve.cpp, if the user had specified \a -a
	  option on the command line.*/
	bool   assertion_mode;

	/// Will ranks of A and B be calculated and displayed?
	/*! 	  determines whether ranks of matrices A and B will be
	 * calculated (at the end of block Lanczos run). This is an expensive
	 * operation both
	  in time and memory. The default value is false and it is reset to
	  true by quadratic_sieve.cpp, if the user had specified  \a -ranks option
	  on the command line. */
	bool   rank_calculation_mode;

	/// Should be formatted output for some GUI frontend printed to AbstractSieve::frontend_pipe ?
	bool frontend;

	/// File stream used by GUI frontends.
	FILE *frontend_pipe;
	/// Frontend pipe filename
	string FrontendPipe;

	/// Type of matrix B used in Linear Algebra part
	/*! determines whether the sieving matrix B is input as a SparseMatrix
	 * or a BitMatrix (this is possibly expandable to other AbstractMatrix-
	  based matrix classes, with some programming effort). The typedef is
	  located in abstract_sieve.h. The default value is ESparseBitMatrix,
	  but there is a constructor setting this variable explicitely. It is
	  set to EBitMatrix, if the user had specified \a -bit option on the
	  command	  line. */
	linalg_type   matrix_type;

	/// Type of algorithm used in Linear Algebra part
	linalg_algo   algorithm_type;

	/*char*		 iGenerateNewPolyText;
	char*		 iQuadraticMachineText;
	char*		 iPerformSievingByPolyText;
	char*	         iSievingLoopText;
	char*		 iRefillMainMatrixText;
	char*	         iSortSmoothText;
	char*		 iDivisorsOfAText;
	char*		 iSmallFactorText;
	char*		 iCountCyclesText;
	char*		 iBlockSieveText;
	char*		 iFillHashtableText;
	char*		 iFillHashtableNegText;
	char*		 iRootsUpdateText;
	char*		 iFindAndFactorText;
	char*		 iUnknownPartText;*/
	static const string	 iGenerateNewPolyText;
	static const string	 iQuadraticMachineText;
	static const string	 iPerformSievingByPolyText;
	static const string      iSievingLoopText;
	static const string	 iRefillMainMatrixText;
	static const string      iSortSmoothText;
	static const string	 iDivisorsOfAText;
	static const string	 iSmallFactorText;
	static const string	 iCountCyclesText;
	static const string	 iBlockSieveText;
	static const string	 iFillHashtableText;
	static const string	 iFillHashtableNegText;
	static const string	 iRootsUpdateText;
	static const string	 iFindAndFactorText;
	static const string	 iUnknownPartText;

	AbstractSerializator* iSerializator;

	int		 iDataCompressionLevel;
	sieving_phase	 iSievingPhase;
	int		 iLineNr;
};


#endif
