#include "abstract_supporting_factor_alg.h"
#include "const_return_codes.h"

#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	#include <iostream>
	#include <typeinfo>
#endif

ASupportingFactorAlg::ASupportingFactorAlg() {
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	SFAS_computed.resize(500);
	SFAS_success.resize(500);
	SFAS_notfit.resize(500);
	for(std::vector<long>::iterator iter = SFAS_computed.begin(); iter != SFAS_computed.end();iter++) *iter = 0;
	for(std::vector<long>::iterator iter = SFAS_success.begin(); iter != SFAS_success.end();iter++) *iter = 0;
	for(std::vector<long>::iterator iter = SFAS_notfit.begin(); iter != SFAS_notfit.end();iter++) *iter = 0;
#endif
	mpz_init(factor1);
	mpz_init(factor2);
}

ASupportingFactorAlg::~ASupportingFactorAlg() {
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	PrintStatistics();
#endif
	mpz_clear(factor1);
	mpz_clear(factor2);
}

int ASupportingFactorAlg::FactorWithStatistics(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) {
	int result = Factor(aModulus, aFactor1, aFactor2);
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	int bits = mpz_sizeinbase(aModulus,2);
	SFAS_computed[bits]++;
	if (0 == SFAS_computed[bits] % 1000) PrintStatistics();
	if (result == ConstRC::Factorized) {
		SFAS_success[bits]++;
	}
#endif
	return result;
}

int ASupportingFactorAlg::Factor(mpz_t aModulus) {
	return FactorWithStatistics(aModulus, factor1, factor2);
}


int ASupportingFactorAlg::Factor(mpz_t aModulus, unsigned long & aFactor1, unsigned long & aFactor2){
	int result = Factor(aModulus);
	if (result == ConstRC::Factorized) {
		if (mpz_fits_ulong_p(factor1)) {
			aFactor1 = mpz_get_ui(factor1);			
		} else {
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
			int bits = mpz_sizeinbase(aModulus,2);
			SFAS_notfit[bits]++;
#endif
			result = ConstRC::FactorTooLarge;
		}
		if (mpz_fits_ulong_p(factor2)) {
			aFactor2= mpz_get_ui(factor2);						
		} else {
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
			int bits = mpz_sizeinbase(aModulus,2);
			SFAS_notfit[bits]++;
#endif
			result = ConstRC::FactorTooLarge;			
		}
	}
	return result;
}

/*!
 * Factorizes a whole array of numbers. All input arrays are considered
 * allocated and all of their mpz_t already initialized.
 *
 * \param aCount - amount of numbers for factorization
 * \param aComposite - numbers for factorization
 * \param aFactors1 - first array of factors
 * \param aFactors2 - second array of factors
 * \return Count of successful factorizations
*/
int ASupportingFactorAlg::FactorArray(int aCount, mpz_t * aComposite, mpz_t * aFactors1, mpz_t * aFactors2) {
	int success = 0;
	for (int i = 0; i < aCount; i++) {
		if (ConstRC::Ok == Factor(aComposite[i], aFactors1[i], aFactors2[i])) 
				success++;
	};
	return success;
}

/*!
 * \brief Another interface for factorization
 *
 * All mpz_t are considered already initialized.
 * \param [in] aModulus Number to factorize
 * \param [out] aFactor1 First factor
 * \param [out] aFactor1 Second factor
 * \return 
 * - ConstRC::Ok - computation successful
 * - ConstRC::NotFactorized - computation unsuccessful
 *
*/

int ASupportingFactorAlg::Factor(factoring_environment* aEnv) {
        int result = Factor(aEnv->modulus,aEnv->factor_1,aEnv->factor_2);
	if (result == ConstRC::Factorized) {
		if (mpz_fits_uint_p(aEnv->factor_1)) {
			aEnv->f_1_int = mpz_get_ui(aEnv->factor_1);
			aEnv->flags |= FACTOR_1_FITS_IN_UINT;
		} else {
			aEnv->flags &= ~FACTOR_1_FITS_IN_UINT;
		}
		if (mpz_fits_uint_p(aEnv->factor_2)) {
			aEnv->f_2_int = mpz_get_ui(aEnv->factor_2);
			aEnv->flags |= FACTOR_2_FITS_IN_UINT;
		} else {
			aEnv->flags &= ~FACTOR_2_FITS_IN_UINT;
		}
	}
        return result;
}

int ASupportingFactorAlg::Factor(mpz_t aModulus, unsigned int & aFactor1, unsigned int & aFactor2){
	int result = Factor(aModulus);
	if (result == ConstRC::Factorized) {
		if (mpz_fits_uint_p(factor1)) {
			aFactor1= mpz_get_ui(factor1);			
		} else {
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
			int bits = mpz_sizeinbase(aModulus,2);
			SFAS_notfit[bits]++;
#endif
			return ConstRC::FactorTooLarge;
		}
		if (mpz_fits_uint_p(factor2)) {
			aFactor2= mpz_get_ui(factor2);						
		} else {
#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
			int bits = mpz_sizeinbase(aModulus,2);
			SFAS_notfit[bits]++;
#endif
			return ConstRC::FactorTooLarge;			
		}
	}
	return result;
};

#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	int ASupportingFactorAlg::PrintStatistics() {
		std::cout << "ASupportingFactorAlg statistics (object: " << typeid(*this).name() <<") :" << std::endl;
		long sum_total = 0;
		long sum_success = 0;
		int minimum_bits=0;
	        int maximum_bits=0;	
		long sum_notfits = 0;
		for(unsigned int index = 0; index < SFAS_computed.size(); index ++) {
			if (0 != SFAS_computed[index]) {
				if (! minimum_bits) minimum_bits = index;
				maximum_bits = index;
			};
			sum_total += SFAS_computed[index];
			sum_success += SFAS_success[index];
			sum_notfits += SFAS_notfit[index];
		};
		std::cout << "Total factorizations:" << sum_total << "\t bit range: " << minimum_bits << "-" << maximum_bits << std::endl;
		std::cout << "Total successes     :" << sum_success << "\t relative:" << 100 * sum_success / (float)sum_total << "%" << std::endl;
		std::cout << "Total notfits       :" << sum_notfits << std::endl;
		std::cout << "Bit                 :";
		for(int index = minimum_bits; index <= maximum_bits; index ++) {
			std::cout << "\t" << index;
		};
		std::cout << std::endl;
		std::cout << "Computed            :";
		for(int index = minimum_bits; index <= maximum_bits; index ++) {
			std::cout << "\t" << SFAS_computed[index];
		};
		std::cout << std::endl;
		std::cout << "Successes           :";
		for(int index = minimum_bits; index <= maximum_bits; index ++) {
			std::cout << "\t" << SFAS_success[index];
		};
		std::cout << std::endl;
		std::cout << "Notfits             :";
		for(int index = minimum_bits; index <= maximum_bits; index ++) {
			std::cout << "\t" << SFAS_notfit[index];
		};
		std::cout << std::endl;
		return ConstRC::Ok;
	};
#endif

/*!
 This method returns binary logarithm of the argument, and -1 if argument has been 0. Results
	for negative values are strange.

*/
int ASupportingFactorAlg::BinLog(int aArg) {
	int i, max = 8*sizeof(int);
	for (i=0; i< max; i++) {
		if (!(aArg >> i)) {
			break;
		}
	}
	return (i-1);
};
