#define SUPPORTING_FACTORING_ALG_STATISTICS
#ifndef __ABSTRACT_SUPPORTING_FACTOR_ALG_H_
#define __ABSTRACT_SUPPORTING_FACTOR_ALG_H_

#include "gmp.h"
#include "enums_common.h"
#include "factor_alg.h"
#include <vector>

#define FACTOR_1_FITS_IN_UINT 0x1
#define FACTOR_2_FITS_IN_UINT 0x2

/*!
\brief An abstract class representing a supporting factoring algorithm.
   
  This class is an abstract base class for factoring algorithms. Its main purpose is to serve for factorization
  of products of two (relatively large) primes, which are often encountered in Multi Large Prime Variation
  of the MPQS/SIQS/NFS.
  For this purpose, the overloaded method 
  
       Factor
       
  of each extension class should return one of the three values (defined in const_return_codes.h):
  
       ConstRC::NotFactorized - if factorization of was unsuccessful

       ConstRC::Ok - if factorization was successful
  
*/

class ASupportingFactorAlg
{
private:
	supporting_factor_alg_types iAlgType;
	mpz_t factor1, factor2;

#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	std::vector<long> SFAS_computed;
	std::vector<long> SFAS_success;
	std::vector<long> SFAS_notfit;
#endif

public:
	ASupportingFactorAlg();
	virtual ~ASupportingFactorAlg();
	virtual int Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) = 0;
	virtual int FactorArray(int aCount, mpz_t * aComposite, mpz_t * aFactors1, mpz_t * aFactors2);
	virtual int FactorWithStatistics(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2);
	virtual int Factor(mpz_t aModulus);
	virtual int Factor(mpz_t aModulus, unsigned long & aFactor1, unsigned long & aFactor2);
	virtual int Factor(mpz_t aModulus, unsigned int & aFactor1, unsigned int & aFactor2);
	virtual int Factor(factoring_environment* aEnv);

// Getters and Setters

public:
	supporting_factor_alg_types Get_AlgorithmType() { return this->iAlgType; }

#ifdef SUPPORTING_FACTORING_ALG_STATISTICS
	int PrintStatistics(); ///< Print statistics
#endif 

protected:
	void Set_AlgorithmType(supporting_factor_alg_types aType) { this->iAlgType = aType; }
	/*!
		This method returns binary logarithm of the argument, and -1 if argument has been 1. Results
		for negative values are strange.
	*/
	int BinLog(int aArg);
};

#endif // __ABSTRACT_SUPPORTING_FACTOR_ALG_H_
