#include "algorithm_m.h"
#include "integer_sparse_matrix.h"
#include "integer_class.h"
#include "wiedemann_z_p.h"
#include "log.h"
#include <cmath>
#include <iostream>

AlgorithmM::AlgorithmM() {
	mpz_init(E2TT);
}

AlgorithmM::~AlgorithmM() {
	mpz_clear(E2TT);
}

int AlgorithmM::SetMatrix(AbstractIntegerMatrix * aMatrix) {
   main_matrix = aMatrix;
}

int AlgorithmM::Compute(long aLimit) {
	if (!main_matrix) {
		CLog::LogError(WHEREARG,"AlgorithmM","Compute",0,0,"Main matrix not set.");
		return ConstRC::GeneralError;
	}
	ComputeBoundaries();
	long * primes;
	long primecount;
	CIntegerClass::GetPrimes(0,ETlogT,primecount,primes);
	cout << "\nPrimecount: " <<primecount;

	// Compute rank

	// main_matrix->PrintToScreen();
	main_matrix->GetRegularSubmatrixModQ(aLimit, 257, rank);
	cout <<"\nRank:" << rank << endl;
	// main_matrix->PrintToScreen();
	IntegerSparseMatrix * regular_matrix = new IntegerSparseMatrix;
	regular_matrix->SetMaximalRowIndex(rank-1);
	regular_matrix->SetMaximalColumnIndex(rank-1);
	regular_matrix->Allocate();
	regular_matrix->PartialCopy(main_matrix,rank-1, rank-1);

	MpzMatrix * determinants = new MpzMatrix();
	determinants->SetMaximalRowIndex(primecount-1);
	determinants->SetMaximalColumnIndex(1);
	determinants->Allocate();
	for(long detrow=0;detrow <=primecount-1;detrow ++) {
		determinants->PutValue(detrow,0,0);
		determinants->PutValue(detrow,1,-3);
	}

	MpzMatrix * current_determinants = new MpzMatrix();
	current_determinants->SetMaximalRowIndex(primecount-1);
	current_determinants->SetMaximalColumnIndex(1);
	current_determinants->Allocate();

	MpzMatrix * good_primes = new MpzMatrix();
	good_primes->SetMaximalRowIndex(primecount-1);
	good_primes->SetMaximalColumnIndex(0);
	good_primes->Allocate();

	MpzMatrix * partial_coeffs = new MpzMatrix();
	partial_coeffs->SetMaximalColumnIndex(rank-1);
	partial_coeffs->SetMaximalRowIndex(primecount-1);
	partial_coeffs->Allocate();

	MpzMatrix * coeffs = new MpzMatrix();
	coeffs->SetMaximalColumnIndex(0);
	coeffs->SetMaximalRowIndex(rank-1);
	coeffs->Allocate();

	//main_matrix->PrintToScreen();
	mpz_t unsucceeded;
	mpz_init(unsucceeded);
	mpz_t partial_determinant;
	mpz_init(partial_determinant);
	for (long row=rank;row<=main_matrix->GetMaximalRowIndex();row++) {
		if (main_matrix->IsRowZero(row,aLimit)) {
			cout << "Row " << row << " already zero, continue."<< endl;
			continue;
		}
		// main_matrix->PrintToScreen();
		long succeeded = 0;
		mpz_set_ui(unsucceeded,1);
		bool determinant_failed = false;
		for (long primeindex = 0; primeindex < primecount;primeindex++) {
			// cout << "Primeindex:" <<primeindex << "\tPrime:" << primes[primeindex];
			cout << "\rLooping Wiedemann: " << primeindex+1 << "/" << primecount;
			WiedemannZP * wiedemann = new WiedemannZP();
			wiedemann->SetModulo(primes[primeindex]);
			wiedemann->SetMainMatrix(regular_matrix);
			wiedemann->SetLeftMatrix(main_matrix, row);

			wiedemann->GenerateSequence();
			wiedemann->ComputePolynomial();

			bool wied_success;
			int rc = wiedemann->GetDeterminant(partial_determinant);
			if (rc != ConstRC::Ok) {
				mpz_set_ui(partial_determinant,0);
				wied_success = false;
			} else {
				wied_success = wiedemann->Compute(partial_coeffs, succeeded);
			}
			if(!wied_success) {
				cout << "\nFail:" <<primes[primeindex];
				mpz_mul_ui(unsucceeded,unsucceeded,primes[primeindex]);
			} else {
				cout << "\nSuccess:" << primes[primeindex];
				determinants->PutValueMpz(succeeded,0,partial_determinant);
				good_primes->PutValue(succeeded,0,primes[primeindex]);
				succeeded++;
			};
			delete wiedemann;
		};
		cout << "\nSucc:" <<succeeded << "/" << primecount<< endl;
		good_primes->PrintToScreen();
		partial_coeffs->PrintToScreen();
		if (mpz_cmp(E2TT,unsucceeded) == -1) {
			cout << "Iterating wiedemann unsuccessful, row " << row <<endl;
			gmp_printf(" %Zd < %Zd\n", E2TT,unsucceeded);
			row--;
			continue;
		};
		determinants->PrintToScreen();

		good_primes->ColumnCRT(0,succeeded,determinants,0,partial_determinant, NULL, true);
		if (mpz_cmp(E2TT,partial_determinant) == -1) {
			cout << "Searching determinant unsuccessful, row " << row <<endl;
			gmp_printf(" %Zd < %Zd\n", E2TT,partial_determinant);
			row--;
			continue;
		};
		gmp_printf("\nDet:%Zd",partial_determinant);
		gmp_printf("\nETT:%Zd",E2TT);
		for(long index = 0;index < rank; index++) {
			good_primes->ColumnCRT(0,succeeded,partial_coeffs,index, unsucceeded, &partial_determinant, true);
			coeffs->PutValueMpz(index,0,unsucceeded);
			if (mpz_cmp(E2TT,unsucceeded) == -1) {
				cout << "Searching coefficient unsuccessful:" <<endl;
				gmp_printf("Coeff %d: %Zd < %Zd\n", index, E2TT,unsucceeded);
				determinant_failed = true;
				row--;
				break;
			};
		}
		if (determinant_failed) continue;
		cout << "\nSubtracting row " << row;
		coeffs->PrintToScreen();
		main_matrix->MultiplyRow(row,partial_determinant);
		for(long index = 0; index < rank;index ++){
			main_matrix->SubtractRowMultiple(row,index,*coeffs->GetValueDirect(index,0));

		}
	};
	main_matrix->PrintToScreen();


	mpz_clear(partial_determinant);
	mpz_clear(unsucceeded);
	delete partial_coeffs;
	delete determinants;
	delete primes;
}

int AlgorithmM::ComputeBoundaries() {
	// ETlogT
	long max_on_row = main_matrix->MaxNonZeroItemsPerRow();
	double tempd = main_matrix->GetMaximalColumnIndex()+1;
	tempd = max_on_row*tempd*log(tempd);
	cout << "MaxNonZeroItemsPerRow:" << max_on_row <<endl;
	cout << "Columns:" << main_matrix->GetMaximalColumnIndex()+1 <<endl;
	cout << "ETlogT:" << tempd <<endl;
	ETlogT = tempd;

	mpz_set_ui(E2TT,sqrt(max_on_row));
	mpz_mul_ui(E2TT,E2TT,main_matrix->GetMaximalColumnIndex()+1);
	mpz_pow_ui(E2TT,E2TT,main_matrix->GetMaximalColumnIndex()+1);
	gmp_printf("(E^½*T)^T: %Zd\n", E2TT);

}


int AlgorithmM::Test() {
	cout << "AlgorithmM::Test()";
    IntegerSparseMatrix * test_matrix = new IntegerSparseMatrix();
    test_matrix->SetMaximalRowIndex(15);
    test_matrix->SetMaximalColumnIndex(9);
    // test_matrix->SetModulo(257));
    test_matrix->Allocate();
    test_matrix->Randomize();
    //test_matrix->ZeroizeRow(4);
    SetMatrix(test_matrix);
    Compute(test_matrix->GetMaximalColumnIndex());
    delete(test_matrix);
}


