/*
 * base_int_matrix.cpp
 *      Author: jero
 */

#include "base_int_matrix.h"
#include "const_return_codes.h"
#include "log.h"
#include <iostream>

BaseIntMatrix::BaseIntMatrix() {
	modulo = 0;
}

BaseIntMatrix::~BaseIntMatrix() {
}


/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  matrix with pseudorandom values, gained from standard randomization in C - calling rand(). The seed is determined
	  at each call of Randomize() by the current clock() value.

	  This method overwrites any previous elements in matrix data array.

	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>
*/
int  BaseIntMatrix::Randomize(){
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok){
		return ret;
	}
	long i, j;
	srand(time(NULL));
	for (i=0; i<= this->GetMaximalRowIndex(); i++){
		for (j=0; j<= this->GetMaximalColumnIndex(); j++){
			integer_matrix_type value = (integer_matrix_type) rand();
			if (modulo) value %= modulo;
			PutValue(i,j,value);
		//	cout << "("<<i<<","<<j<<","<<value <<")";
		}
	}
	return ConstRC::Ok;
}

int BaseIntMatrix::SetModulo(int aValue) {
	modulo = aValue;
	return ConstRC::Ok;
}

int BaseIntMatrix::SetModulo(mpz_t aValue) {
	modulo = mpz_get_si(aValue);
	return ConstRC::Ok;
}

int BaseIntMatrix::MultiplyRow(long aRow, integer_matrix_type aMul) {
	if ((aRow<0) || (aRow>GetMaximalRowIndex())) {
		CLog::LogError(WHEREARG,"BaseIntMatrix","MultiplyRow",aRow,GetMaximalRowIndex(),"Bad row index.");
		return ConstRC::ExcessiveRowIndex;
	}
	for(long column = 0; column <=GetMaximalColumnIndex(); column++) {
		integer_matrix_type value;
		GetValue(aRow, column, value);
		value *= aMul;
		if (modulo) value %= modulo;
		PutValue(aRow, column, value);
	};
	return ConstRC::Ok;
}

int BaseIntMatrix::SubtractRowMultiple(long aTargetRow, long aSourceRow, integer_matrix_type aMul) {
	if ((aTargetRow<0) || (aTargetRow>GetMaximalRowIndex())) {
		CLog::LogError(WHEREARG,"BaseIntMatrix","SubtractRowMultiple",aTargetRow,GetMaximalRowIndex(),"Bad target row index.");
		return ConstRC::ExcessiveRowIndex;
	}
	if ((aSourceRow<0) || (aSourceRow>GetMaximalRowIndex())) {
		CLog::LogError(WHEREARG,"BaseIntMatrix","SubtractRowMultiple",aSourceRow,GetMaximalRowIndex(),"Bad source row index.");
		return ConstRC::ExcessiveRowIndex;
	}
	for (long column=0; column <= GetMaximalColumnIndex(); column++){
		integer_matrix_type value, value2;
		GetValue(aTargetRow, column, value);
		GetValue(aSourceRow, column, value2);
		value -=value2 * aMul;
		PutValue(aTargetRow, column, value);
	};
	return ConstRC::Ok;
};

int BaseIntMatrix::GetModulo() {
	return modulo;
}

int BaseIntMatrix::ApplyModulo() {
	if (0 == modulo) return ConstRC::Ok;
	for (long row = 0; row <= GetMaximalRowIndex(); row ++)
		for (long column = 0; column <= GetMaximalColumnIndex(); column++) {
			integer_matrix_type value;
			GetValue(row, column, value);
			value %= modulo;
			if (value < 0) value+=modulo;
			PutValue(row, column, value);
		}
	return ConstRC::Ok;
}

int BaseIntMatrix::GaussColumn(long aRow, long aColumn, bool aPartial) {
	if ((aRow<0) || (aRow>GetMaximalRowIndex())) {
		CLog::LogError(WHEREARG,"BaseIntMatrix","GaussColumn",aRow,GetMaximalRowIndex(),"Bad row index.");
		return ConstRC::ExcessiveRowIndex;
	}
	if ((aColumn<0) || (aColumn>GetMaximalColumnIndex())) {
		CLog::LogError(WHEREARG,"BaseIntMatrix","GaussColumn",aColumn,GetMaximalColumnIndex(),"Bad column index.");
		return ConstRC::ExcessiveColumnIndex;
	}
	mpz_set_si(tempadd,modulo);
	integer_matrix_type member;
	GetValue(aRow,aColumn,member);
	mpz_set_si(tempmul,member);
	mpz_invert(tempmul,tempmul, tempadd);
	member = mpz_get_si(tempmul);
	MultiplyRow(aRow, member);
	long start;
	if (aPartial)
		start = aRow+1;
	else
		start = 0;
	for (long row = start; row <= GetMaximalRowIndex(); row++) {
			if(row == aRow) continue;
			GetValue(row, aColumn, member);
			mpz_set_si(tempmul, member);
			// TODO: SubstractRowMultiple(long,long,long)
			if (!IsZero(row, aColumn))	SubtractRowMultiple(row,aRow,mpz_get_si(tempmul));
	};
	return ConstRC::Ok;
}
;

int BaseIntMatrix::SubtractRowMultiple(long aTargetRow, long aSourceRow,
		mpz_t aMul) {
	return SubtractRowMultiple(aTargetRow,aSourceRow, mpz_get_si(aMul));
}

int BaseIntMatrix::MultiplyRow(long aRow, mpz_t aMul) {
	return MultiplyRow(aRow,mpz_get_si(aMul));
}

bool BaseIntMatrix::IsInvertible(long aRow, long aColumn) {
	if (!modulo) return !IsZero(aRow,aColumn);
	mpz_set_ui(tempmul,modulo);
	integer_matrix_type value;
	GetValue(aRow,aColumn,value);
	mpz_set_si(tempadd,value);
	if (mpz_invert(tempadd,tempadd, tempmul))
		return true;
	else
		return false;
}

bool BaseIntMatrix::Equals(BaseIntMatrix* aMatrix) {
	long rowmax;
	long colmax;
	if (rowmax = GetMaximalRowIndex() != aMatrix->GetMaximalRowIndex()) return false;
	if (colmax = GetMaximalColumnIndex() != aMatrix->GetMaximalColumnIndex()) return false;
	for (long row = 0;row<=rowmax;row ++)
		for(long col = 0;col<=colmax;col++) {
			integer_matrix_type val1,val2;
			GetValue(row,col,val1);
			aMatrix->GetValue(row,col,val2);
			if (modulo) {
				val1 %= modulo;
				val2 %= modulo;
			}
			if (val1 != val2) return false;
		}
	return true;
}
