#include "base_parameters.h"
#include "const_return_codes.h"
#include "common_xml_helper.h"
#include "log.h"

using namespace std;

typedef map<std::string, std::string>::const_iterator ParametersIter;
typedef set<string>::const_iterator NonSerialParamIter;

CBaseParameters::CBaseParameters() : MSerializable(FILE_NAME_BASE_PARAMETERS), iWRMutex(DEFAULT_WRMUTEX_MAX_READERS)
{
}
CBaseParameters::CBaseParameters(const CBaseParameters& aOperand) : MSerializable(aOperand), 
	iWRMutex(aOperand.iWRMutex)
{
	for (ParametersIter it = aOperand.iParameters.begin(); it != aOperand.iParameters.end(); it++)
	{
		AddParameter((*it).first, (*it).second);
	}

	for (NonSerialParamIter it = aOperand.iNonSerializableParam.begin(); it != aOperand.iNonSerializableParam.end(); it++)
	{
		this->iNonSerializableParam.insert((*it));
	}
}
int CBaseParameters::Copy2(CBaseParameters &aOperand)
{
	for (ParametersIter it = this->iParameters.begin(); it != this->iParameters.end(); it++)
	{
		aOperand.SetParameter((*it).first, (*it).second);
	}

	for (NonSerialParamIter it = this->iNonSerializableParam.begin(); it != this->iNonSerializableParam.end(); it++)
	{
		aOperand.iNonSerializableParam.insert((*it));
	}

	return ConstRC::Ok;
}
int CBaseParameters::AddNonSerializableParam(const string& aName)
{
	this->iWRMutex.LockWrite();
	this->iNonSerializableParam.insert(aName);
	this->iWRMutex.UnlockWrite();

	return ConstRC::Ok;
}
int CBaseParameters::SetParameter(const string& aName, const string& aValue)
{
	this->iWRMutex.LockWrite();

	if (this->iParameters.count(aName) == 0)
	{
		// new entry
		this->iParameters.insert(make_pair(aName, aValue));
	}
	else
	{
		// set new value
		this->iParameters[aName] = aValue;
	}

	this->iWRMutex.UnlockWrite();

	return ConstRC::Ok;
}

int CBaseParameters::AddParameter(const string& aName, const string& aValue)
{
	this->iWRMutex.LockWrite();

	if (this->iParameters.count(aName) == 0)
	{
		// new entry
		this->iParameters.insert(make_pair(aName, aValue));
		this->iWRMutex.UnlockWrite();
		return ConstRC::Ok;
	}
	else
	{
		this->iWRMutex.UnlockWrite();
		// parameter already exists!!!
		return ConstRC::EntryExists;
	}
}

bool CBaseParameters::ContainsParameter(const string& aName) const
{
	this->iWRMutex.LockRead();
	bool ret = this->iParameters.count(aName) > 0;
	this->iWRMutex.UnlockRead();
	return ret;
}

int CBaseParameters::GetStringParameter(const string& aName, string& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = (*it).second;
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetStringParameter(const string& aName, const string& aDefaultValue, string& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		aResult = aDefaultValue;
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = (*it).second;
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetBoolParameter(const string& aName, bool& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = (*it).second == TRUE_STRING;
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetBoolParameter(const string& aName, const bool& aDefaultValue, bool& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		aResult = aDefaultValue;
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = (*it).second == TRUE_STRING;
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetIntegerParameter(const string& aName, int& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = atoi((*it).second.c_str());
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetIntegerParameter(const string& aName, const int& aDefaultValue, int& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		aResult = aDefaultValue;
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = atoi((*it).second.c_str());
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetLongParameter(const string& aName, long& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = atol((*it).second.c_str());
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetLongParameter(const string& aName, const long& aDefaultValue, long& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		aResult = aDefaultValue;
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = atol((*it).second.c_str());
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}

int CBaseParameters::GetDoubleParameter(const string& aName, double& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = atof((*it).second.c_str());
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}
int CBaseParameters::GetDoubleParameter(const string& aName, const double& aDefaultValue, double& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		aResult = aDefaultValue;
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		aResult = atof((*it).second.c_str());
		this->iWRMutex.UnlockRead();
		return ConstRC::Ok;
	}
}


int CBaseParameters::GetMpzParameter(const string& aName, mpz_t aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		if (mpz_set_str(aResult, (*it).second.c_str(), DEFAULT_NUMBER_BASE) == 0)
		{
			this->iWRMutex.UnlockRead();
			return ConstRC::Ok;
		}
		
		this->iWRMutex.UnlockRead();
		return ConstRC::CannotConvert;
	}
}
int CBaseParameters::GetMpzParameter(const string& aName, const mpz_t aDefaultValue, mpz_t aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		mpz_set(aResult, aDefaultValue);
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		if (mpz_set_str(aResult, (*it).second.c_str(), DEFAULT_NUMBER_BASE) == 0)
		{
			this->iWRMutex.UnlockRead();
			return ConstRC::Ok;
		}

		this->iWRMutex.UnlockRead();		
		return ConstRC::CannotConvert;
	}
}
int CBaseParameters::GetPolynomialParameter(const string& aName, Polynomial *& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		this->iWRMutex.UnlockRead();
		return ConstRC::NotFound;
	}
	else
	{
		if (aResult != NULL)
		{
			delete aResult;
			aResult = NULL;
		}
		aResult = Polynomial::Parse((*it).second);

		if (aResult == NULL)
		{
			this->iWRMutex.UnlockRead();
			return ConstRC::CannotConvert;
		}
		else
		{
			this->iWRMutex.UnlockRead();
			return ConstRC::Ok;
		}
	}
}
int CBaseParameters::GetPolynomialParameter(const string& aName, const Polynomial * aDefaultValue, Polynomial *& aResult) const
{
	this->iWRMutex.LockRead();
	ParametersIter it = this->iParameters.find(aName);

	if (it == this->iParameters.end())
	{
		if (aResult != NULL)
		{
			delete aResult;
			aResult = NULL;
		}
		this->iWRMutex.UnlockRead();
		aResult = aDefaultValue->Copy();
		if (aResult == NULL)
		{
			return ConstRC::GeneralError;
		}	
	
		return ConstRC::NotFound;
	}
	else
	{
		if (aResult != NULL)
		{
			delete aResult;
			aResult = NULL;
		}
		aResult = Polynomial::Parse((*it).second);

		if (aResult == NULL)
		{
			this->iWRMutex.UnlockRead();
			return ConstRC::CannotConvert;
		}
		else
		{
			this->iWRMutex.UnlockRead();
			return ConstRC::Ok;
		}
	}
}


int CBaseParameters::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CBaseParameters::Serialize(const CBaseParameters& aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_PARAM_ELEMENT_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CBaseParameters","Serialize Parameters",rc,0,"Could not write the parameter tag.");
    	return ConstRC::GeneralError;
    }
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	this->iWRMutex.LockRead();
	// we write all parameters
	for (ParametersIter it = this->iParameters.begin(); it != this->iParameters.end(); it++)
	{
		if (this->iNonSerializableParam.find((*it).first) == this->iNonSerializableParam.end())
		{
			CommonXMLHelper::SerializeParameterEntry(aWriter, XML_SERIAL_PARAMETER_ENTRY_NAME, (*it).first.c_str(), (*it).second.c_str());
		}
	}	
	this->iWRMutex.UnlockRead();

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CBaseParameters","Serialize Parameters",rc,0,"Could not end the parameter tag.");
        return ConstRC::GeneralError;
    }
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;    
}
int CBaseParameters::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CBaseParameters::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_PARAM_ELEMENT_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CBaseParameters","Deserialize Parameters",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CBaseParameters","Deserialize Parameters",0,0,"Main element name is wrong.");
		xmlFree(name);
		return ConstRC::GeneralError; 
	}

	this->iWRMutex.LockWrite();
	// Clear all existing parameters
	this->iParameters.clear();

	int depth = xmlTextReaderDepth(aReader);
	string paramName;
	string paramValue;

	while ((rc = CommonXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		if (NULL != name) xmlFree(name);
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PARAMETER_ENTRY_NAME))
		{
			rc = CommonXMLHelper::DeserializeParameterEntry(aReader, "Param entry read as ", paramName, paramValue, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			// insert new parameter
			this->iParameters.insert(make_pair(paramName, paramValue));
			continue;
		}
	}
	if (NULL != name) xmlFree(name);

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	this->iWRMutex.UnlockWrite();

	return ConstRC::Ok;
}

// print parameters with prefix
void CBaseParameters::Print(FILE *aFile, const char *aPrefix) const {
	std::map<std::string, std::string>::const_iterator it;

	for (it = iParameters.begin(); it != iParameters.end(); it++) fprintf(aFile, "%s%s:\t%s\n", aPrefix, it->first.c_str(), it->second.c_str());

	return;
}

