#ifndef __BASE_PARAMETERS_H__
#define __BASE_PARAMETERS_H__

#include <string>
#include <map>
#include <iterator>
#include <set>
#include "common_definitions.h"
#include "mserializable.h"
#include "write_read_mutex.h"
#include "polynom_class.h"

/*!
 * \brief Parameter deserialization
 */
class CBaseParameters : public virtual MSerializable
{
protected:
	///Container for parameters
	std::map<std::string, std::string> iParameters;
	///Write/Read mutex for iParameters - thread safe
	mutable CWriteReadMutex iWRMutex;
	///Set of parameter names which shouldn't be serialized
	std::set<std::string> iNonSerializableParam;

public:

	CBaseParameters();
	CBaseParameters(const CBaseParameters& aOperand);
	virtual ~CBaseParameters() {};

	int Copy2(CBaseParameters& aOperand);

	///Add new parameter, if parameter already exists nothing happends 
	int AddParameter(const std::string& aName, const std::string& aValue);

	///Add or Set parameter, add new or set value of existing parameter
	int SetParameter(const std::string& aName, const std::string& aValue);
	
	///We don't want all parameters to serialize so it's enough to give the name
	int AddNonSerializableParam(const std::string& aName);
	
	bool ContainsParameter(const std::string& aName) const;

	int GetStringParameter(const std::string& aName, std::string& aResult) const;
	int GetStringParameter(const std::string& aName, const std::string& aDefaultValue, std::string& aResult) const;

	int GetIntegerParameter(const std::string& aName, int& aResult) const;
	int GetIntegerParameter(const std::string& aName, const int& aDefaultValue, int& aResult) const;

	int GetLongParameter(const std::string& aName, long& aResult) const;
	int GetLongParameter(const std::string& aName, const long& aDefaultValue, long& aResult) const;

	int GetDoubleParameter(const std::string& aName, double& aResult) const;
	int GetDoubleParameter(const std::string& aName, const double& aDefaultValue, double& aResult) const;

	int GetBoolParameter(const std::string& aName, bool& aResult) const;
	int GetBoolParameter(const std::string& aName, const bool& aDefaultValue, bool& aResult) const;

	int GetMpzParameter(const std::string& aName, mpz_t aResult) const;
	int GetMpzParameter(const std::string& aName, const mpz_t aDefaultValue, mpz_t aResult) const;

	int GetPolynomialParameter(const std::string& aName, Polynomial *& aResult) const;
	int GetPolynomialParameter(const std::string& aName, const Polynomial * aDefaultValue, Polynomial *& aResult) const;

	// De/Serialization methods
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

	// print parameters with prefix
	void Print(FILE *aFile, const char *aPrefix) const;

};


#endif
