#include "basic_hashtable.h"
#include "const_return_codes.h"
#include "const_element_flags.h"
#include "const_hashtables.h"
#include "common_utils.h"
#include "nfs_utils.h"
#include "log.h"
#include <iostream>

CBasicHashtable::CBasicHashtable()
{
	this->Hashtable = NULL;
	this->AllocatedSize = 0;
	this->Mask = 0;
	this->filled = 0;
}

CBasicHashtable::~CBasicHashtable()
{
	DeleteHashtable();
}

void CBasicHashtable::DeleteHashtable()
{
	if (this->Hashtable != NULL)
	{
		free(this->Hashtable);
		this->Hashtable = NULL;
		this->AllocatedSize = 0;
		this->Mask = 0;
	}
}

int CBasicHashtable::SetupHashtable(long aAmount)
{
	if (this->Hashtable != NULL)
	{
		CLog::LogError(WHEREARG,"CBasicHashtable","SetupHashtable",0,0,"Hashtable is not NULL.");
		return ConstRC::BadArgument;
	}

	this->AllocatedSize = Utils::GetSuitableHashtableSize(aAmount);
	this->Hashtable = (CBasicHashEntry *) Utils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(CBasicHashEntry));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CBasicHashtable","SetupHashtable", 0, this->AllocatedSize * sizeof(CBasicHashEntry),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	ClearHashtable();

	return ConstRC::Ok;
}

int CBasicHashtable::ClearHashtable()
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CBasicHashtable","ClearHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	memset(Hashtable, 0, AllocatedSize*sizeof(CBasicHashEntry));

	for(CBasicHashEntry *index = this->Hashtable; index < this->Hashtable+AllocatedSize; index++) index->SetMask();

	this->filled = 0;

	return ConstRC::Ok;
}

/*!
	This method tries to find large prime aPrime in hashtable, and indicates
	the result of this search in two values: value of aIndex, which is an index
	corresponding to the real or expected position of the prime in the hashtable,
	and value of aPresent, which indicates whether the prime has already been
	in the hashtable or not. In the second case, the value of aIndex has meaning
	"if you want to insert aPrime, this index is the one where it should be inserted".

*/
int CBasicHashtable::FindPrimeInHashtable(large_prime_type aPrime, bool & aPresent, long & aIndex, bool & aEnlarged)
{
	aEnlarged = false;
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CBasicHashtable","FindPrimeInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aIndex = 0;
		return ConstRC::BadArgument;
	}

	long hash = (aPrime >> 1) & this->Mask;			
	int rc = CollisionResolution(hash, aPrime, aIndex);
	while (rc == ConstRC::NeedEnlargement) {
		aEnlarged = true;
		// Print();
		if (EnlargeHashtable() != ConstRC::Ok) {
			aPresent = false;
			aIndex = 0;
			return ConstRC::GeneralError;
		}
		// Print();
		long hash = (aPrime >> 1) & this->Mask;			
		rc = CollisionResolution(hash, aPrime, aIndex);
	};
	if (rc != ConstRC::Ok)
	{
		aPresent = false;
		aIndex = 0;
		return ConstRC::GeneralError;
	}

    	if (this->Hashtable[aIndex].IsMasked()) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

// 	std::cout << "(" << aPrime << "," << aPresent << "," << aIndex << ")";

	return ConstRC::Ok;
}

/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where aPrime can be stored.

*/
int CBasicHashtable::CollisionResolution(long aHash, large_prime_type aPrime, long & aIndex) const
{
	aIndex = aHash;
	long iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CBasicHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	CBasicHashEntry * aux = this->Hashtable + aIndex;
	CBasicHashEntry * end = this->Hashtable + this->AllocatedSize;

	while ((aux->GetValue() != aPrime) && (aux->GetValue() != ConstHT::FULL_MASK_NFS)) 
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"CBasicHashtable","CollisionResolution",iter,this->AllocatedSize,"The hash table has overflown.");
				/* for (int i = 0;i<this->AllocatedSize;i++) {
					std::cerr << "[" << i <<", "<<Hashtable[i].GetValue() <<"]\t";
				}; */
				return ConstRC::GeneralError;
			}
			aux = this->Hashtable;
		}
	}

	aIndex = aux - this->Hashtable;

	if (ConstHT::MAX_FULLNESS < (double) this->filled / this->AllocatedSize) {
		return ConstRC::NeedEnlargement;
	} else {
		return ConstRC::Ok;
	};
}

int CBasicHashtable::AddToTable(large_prime_type aPrime)
{
	bool present = false;
	long index;
	bool enlarged;

	if (FindPrimeInHashtable(aPrime, present, index, enlarged) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CBasicHashtable","AddToTable",0,0,"Find prime in hashtable failed.");
		return ConstRC::GeneralError;
	}

	if (!present) 
	{
		// This prime has not been in hashtable yet.					
		this->Hashtable[index].SetValue(aPrime);
		this->Added();
	}

	return ConstRC::Ok;
}

int CBasicHashtable::EnlargeHashtable() {
	int new_maximal_index = (this->AllocatedSize)*2-1;
	int new_mask = new_maximal_index;
	CBasicHashEntry *new_table = NULL;

	if ((new_table = (CBasicHashEntry *) Utils::AllocationMachine((void**) &new_table, (new_maximal_index+1)* sizeof(CBasicHashEntry))) == NULL) {
		CLog::LogError(WHEREARG,"CBasicHashtable","EnlargeHashtable", 0, (new_maximal_index+1) * sizeof(CBasicHashEntry),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}

	memset(new_table, 0, (new_maximal_index+1)*sizeof(CBasicHashEntry));

	for (int i=0; i <= new_maximal_index; i++) {
		new_table[i].SetValue(ConstHT::FULL_MASK_NFS);
	}

	// first rearrange values
	for (int j = 0;j<this->AllocatedSize;j++) {
//		std::cerr << "-(" << j << ")\t";
		if (ConstHT::FULL_MASK_NFS != Hashtable[j].GetValue()) {
			large_prime_type tempPrime = Hashtable[j].GetValue();
			int j2 = ( tempPrime >> 1) & new_mask;			
//				fprintf(stderr,"\nRearranging values: %d : %d : %d ",j, Hashtable[j].GetValue(), j2);
			int iter2 = 0;
			while (new_table[j2].GetValue() != ConstHT::FULL_MASK_NFS) {
				j2++;
				iter2++;
				if (j2 > new_maximal_index) {
					j2 = 0;
				};
				if (iter2 > new_maximal_index) {
					fprintf(stderr, "Overflow!");
					abort();
				};
			};
//				fprintf(stderr,"  [%d %d]  ",iter2, j2);
			new_table[j2].SetValue(Hashtable[j].GetValue());
		};
	}

	this->AllocatedSize = new_maximal_index+1;
	this->Mask = new_mask;			
	free(Hashtable);
	Hashtable= new_table;

	return ConstRC::Ok;
}

void CBasicHashtable::Print() {
	std::cout<<"\nVelikost: "<<AllocatedSize;
	int count = 0;
	int values = 0;
	for(int i = 0; i <AllocatedSize; i++) {
		if (! Hashtable[i].IsMasked()) {
			count++;
			values+=Hashtable[i].GetValue();
		}
	}
	std::cout << "\ncount: " << count << "values: " << values;

	return;
};

