#include <cstring>

#include "bc_matrix_class.h"
#include "common_utils.h" // because of allocation machine
#include "const_return_codes.h"
#include "log.h"

/*!
	Used for bit shift constants: K_SHIFTS[i] = (1 << i), for i=0...(bits_in_m_t - 1). It is
	initialized at construction of first instance of BCMatrix, and is NULL before; but it can be
	forcedly initialized with no instance of BCMatrix in existence by calling static method DefineKShifts().
	
	Before using K_SHIFTS constants (they are public), make sure that they have been initialized, otherwise a null pointer exception occurs.

*/

matrix_type* BCMatrix::K_SHIFTS = NULL;
long BCMatrix::bits_in_m_t = 8*sizeof(matrix_type);

/*!
	The default constructor does not take any parameters and constructs an instance of a "generic" BC matrix,
	with unknown dimensions. The dimensions may be later set by appropriate setter methods.

*/

BCMatrix::BCMatrix() : MSerializable(FILE_NAME_BC_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->maximal_row_index = -1;
	this->maximal_column_index = -1;
	this->matrix = NULL;
	bits_in_m_t = 8*sizeof(matrix_type);
	this->max_allocated_column_index = 0;
        if (!this->K_SHIFTS) { DefineKShifts(); }
	this->species = SBCMatrix;
}

/*!
	The second constructor constructs an instance of a BC matrix with known dimensions. Beware that the parameters
	taken mean the actual number of rows and columns, and not their maximal indices; so, if we want to construct
	a matrix of dimensions 17x32, we call
	
		BCMatrix* bm = new BCMatrix(17,32);
		
	Now, we have an instance of a BC matrix; its member variables will be set to:
	
		bm->maximal_row_index = 16;
		
		bm->maximal_column_index = 31;
		
		bm->maximal_allocated_row_index = -1; // no initialization of the data array yet!
		
		bm->maximal_allocated_column_index = -1; // no initialization of the data array yet!
		
	Both of the constructors initialize the data array to NULL pointer, and a real allocation is performed later	- at the time of need. This programming pattern is called lazy initialization, and helps to reduce runtime memory requirements.
	
	Both of the constructors check whether the static member variable K_SHIFTS is equal to zero, and if true, they call DefineKShifts() method in order to initialize the variable.

*/

BCMatrix::BCMatrix(long aRows,long aColumns) : MSerializable(FILE_NAME_BC_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->maximal_row_index = aRows - 1;
	this->maximal_column_index = aColumns - 1;
	this->matrix = NULL;
	bits_in_m_t = 8*sizeof(matrix_type);
	this->max_allocated_column_index = 0;
        if (!K_SHIFTS) { DefineKShifts(); }
	this->species = SBCMatrix;
}


/*!
	The destructor frees all the allocated memory, that means it dealocates the memory used for the field matrix.

*/

BCMatrix::~BCMatrix(){
	if (max_allocated_row_index >= 0){
		free(this->matrix);
		this->matrix = NULL;
	}
}

BCMatrix * BCMatrix::clone() {
	BCMatrix * new_matrix = new BCMatrix(GetMaxAllocatedRowIndex()+1, GetMaximalColumnIndex()+1);
	return new_matrix;	
}

/*!
	  This method is used for definition of the K_SHIFTS static array, which is a public member variable of the class.
	  Its length depends on length of matrix_type, and so it cannot be pre-hardcoded into the source code.
	  
	  There is a subtle and not-easily-soluble problem concerning the run of the method. The right place to run this
	  method are the constructors of BCMatrix(), and the constructors must always run O.K. That is why the DefineKShifts()
	  method returns no return code; it would be of no use, since the constructor has no means of telling its caller
	  that it has encountered an error.
	  
	  However, an allocation error MAY theoretically occur in DefineKShifts(), since we really do allocate memory space
	  for the static array. Possible as it is, I still consider this improbable, because of the following facts:
	    - the method is called only once in all program run, at the time of instantiation of the first instance of BitMatrix
	    - the heap space allocated is very small (typically 32*4 or 64*8 bytes, so lesser than 0.5 kB), and, in "normal"
	      systems, there should always be such a small portion of memory available.

*/

void BCMatrix::DefineKShifts() {
        K_SHIFTS = (matrix_type *) Utils::AllocationMachine((void **) &K_SHIFTS,(bits_in_m_t)*sizeof(matrix_type));
        if (K_SHIFTS == NULL) {
                CLog::LogError(WHEREARG,"BCMatrixClass","error allocating K_SHIFTS",(bits_in_m_t)*sizeof(matrix_type),2001,"There was not enough memory to allocate K_SHIFTS");
        }
        int i;
        for (i=0; i< bits_in_m_t; i++){
                K_SHIFTS[i] = (1 << i);
        }
}

/*!
	  This method takes care of allocation of the internal array of matrix_type, which contains the
	  matrix entries. First of all it tests whether maximal_column_index does not exceed the possible
	  number of columns. Then it tests the need for allocation by evaluating
	  <CENTER>
	  	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1))
	  </CENTER>
	  condition.
	  
	  If there is decision to run the allocation job, the matrix is zeroized as well.

	  Return codes:
	  <TABLE border=0>
	  <TR><TD>	ConstRC::Ok	</TD><TD>	   everything all right</TD></TR>
	  <TR><TD>  ConstRC::NotEnoughMemory </TD><TD> unsuccessful allocation of the matrix</TD></TR>
	<TR><TD>	ConstRC::SizeMismatch   </TD><TD> the number of columns too high</TD></TR>
	</TABLE>
	  
*/

int BCMatrix::Allocate(){	 
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){		
		this->matrix = (matrix_type *) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(matrix_type));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"BCMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate 1-byte column matrix row");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->matrix,0,(this->maximal_row_index+1)*sizeof(matrix_type));
		this->max_allocated_row_index=maximal_row_index;
	}
	return ConstRC::Ok;
}

/*!
	  This method at first ensures that the matrix is allocated, by calling Allocate(), and then refills the matrix
	  with random values.

	  The random values are gained from standard randomization in C - calling rand(). The seed is determined
	  at each call of Randomize() by the current clock() value.

	  This method overwrites any previous elements in matrix data array.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory  </TD><TD>thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		</TABLE>
*/

int BCMatrix::Randomize(){
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok){
		return ret;
	}
	long i; // j;
	matrix_type seed = (matrix_type) clock();
	srand(seed);
//	srand(100000);
	matrix_type mask = 0;
	int diff = 32 - (this->bits_in_m_t - (maximal_column_index+1));
	if (diff != 32) {diff %= 32; }
	for (i=0; i< diff; i++){
		mask |= K_SHIFTS[i];
	}
	for (i=0; i<= this->max_allocated_row_index; i++)
	{
		this->matrix[i] = (matrix_type) (rand() & mask);
	if(diff==32) this->matrix[i] ^= K_SHIFTS[31]*(rand() & 1);
	}
	return ConstRC::Ok;
}

/*!
	  This method at first ensures that the sizes of the calling instance and aSource match and that the calling instance 
	  is allocated, and then copies the entries of aSource into the calling instance, overwriting any previous elements
	  in matrix data array.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::SizeMismatch     </TD><TD> the dimensions of the calling instance and aSource do not match</TD></TR>
		</TABLE>
		
 	  If the symbolic constant
	        MATRIX_OPERATIONS_TIME_MESSAGE
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_copy.
*/

int  BCMatrix::Copy(BCMatrix* aSource){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        zero_start = clock();
#endif
	if (this->maximal_row_index != aSource->maximal_row_index)  		return ConstRC::SizeMismatch;
	if (this->maximal_column_index != aSource->maximal_column_index)  	return ConstRC::SizeMismatch;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok){
		return ret;
	}
	long i;
	for (i=0; i<= this->max_allocated_row_index; i++){
		this->matrix[i] = aSource->matrix[i];
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        diff = clock() - zero_start;
        if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values
                total_copy += (double) diff;
        }
#endif
	return ConstRC::Ok;
}

/*!
	  This method is used to treat the copying of matrices that the compiler thinks are abstract. The method checks
	  the species of aSource and if it is a BCMatrix, it calls
	  	Copy((BCMatrix*) aSource);
	  otherwise it calls the general abstract method.

	  Return codes: the method returns the return code of the launched copy method.
*/

int  BCMatrix::Copy(AbstractMatrix* aSource){
        if (aSource->GetSpecies()==this->GetSpecies()) return this->Copy((BCMatrix*) aSource);
        else return AbstractMatrix::Copy(aSource);
}

/*!
	  This method performs comparison of the calling instance and of aMatrix. It does not call Allocate() to ensure 
	  allocation of anything.

	  The principles for equality are the following:
	  <UL>
	  	<LI> if aMatrix is NULL, then the matrices are not equal
		<LI> if the respective dimensions maximal_row_index and maximal_column_index differ, then the matrices are not equal
		<LI> if the respective dimensions max_allocated_row_index and max_allocated_column_index differ, then the matrices are not equal		
		<LI> if any entry on any position differs, then the matrices are not equal.
	  </UL>
	  The comparison ends as soon as any of the previous conditions is met.
	  This method does not change any data of any operand.
	  
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>TRUE 		 </TD><TD>     the matrices are equal</TD></TR>
		<TR><TD>FALSE		  </TD><TD>    the matrices are not equal</TD></TR>
		</TABLE>
*/

bool BCMatrix::Equals(BCMatrix* aMatrix){
	if (!aMatrix)		return FALSE; //ConstRC::NullPointerSupplied;
	// performs a "true" comparison of two matrices. They must be equal up to allocation (have the same number of
	// elements allocated
	if (this->max_allocated_row_index != aMatrix->max_allocated_row_index)		return FALSE;
	if (this->maximal_row_index != aMatrix->maximal_row_index)			return FALSE;
	if (this->maximal_column_index != aMatrix->maximal_column_index)		return FALSE;
	long i;
	for (i=0; i<= this->max_allocated_row_index; i++){
		if (this->matrix[i] != aMatrix->matrix[i])		return FALSE;
	}
	return TRUE;
}

/*!
	  This method at first ensures that the matrix is allocated, by calling Allocate(), and then refills the matrix
	  with 0 values, using memset() operation.

	  This method overwrites any previous elements in matrix data array.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		   </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory  </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.	</TD></TR>
		</TABLE>
*/

int  BCMatrix::Zeroize(){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        zero_start = clock();
#endif
	long i = this->max_allocated_row_index;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	if (this->max_allocated_row_index == i){
	// zeroizing is needed
		memset(this->matrix,0,(this->max_allocated_row_index+1)*sizeof(matrix_type));
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        diff = clock() - zero_start;
        if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values
                total_zeroize += (double) diff;
        }
#endif
	return ConstRC::Ok;
}

/*!
	  This method prints the calling instance onto the screen. It does not change any data.
*/

void BCMatrix::PrintToScreen(){
	long i, k;
	printf("\n");
	printf("Byte column matrix: %ld x %ld\n",this->maximal_row_index+1,this->maximal_column_index+1);
	for (i=0; i<=this->max_allocated_row_index; i++){
		printf("\nRow %4ld:   ",i);
		for (k=0; k<=this->maximal_column_index; k++){
			if (this->matrix[i] & K_SHIFTS[k]) printf("1 "); else printf("0 ");
		}
	}
	printf("\n\n");
}

/*!
  Puts 1 into the cell given by the coordinates.

  Return Codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		   </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::ExcessiveRowIndex  </TD><TD> if the row index is outside the matrix</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex  </TD><TD> if the column index is outside the matrix</TD></TR>
		</TABLE>
*/

int  BCMatrix::PutOne(long aRow, long aColumn){
	int ret;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	this->matrix[aRow] |= K_SHIFTS[aColumn];
	return ConstRC::Ok;
}

/*!
  Puts 0 into the cell given by the coordinates.

  Return Codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		   </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::ExcessiveRowIndex  </TD><TD> if the row index is outside the matrix</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex  </TD><TD> if the column index is outside the matrix</TD></TR>
		</TABLE>
*/

int  BCMatrix::PutZero(long aRow, long aColumn){
	int ret;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	this->matrix[aRow] &= ~K_SHIFTS[aColumn];
	return ConstRC::Ok;
}

/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then sets the given
	  row to be 0.

	  This method checks validity of aRow index.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		   </TD><TD> everything all right</TD></TR>

		<TR><TD>ConstRC::NotEnoughMemory  </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>

		<TR><TD>ConstRC::ExcessiveRowIndex </TD><TD>if aRow &gt; maximal_row_index</TD></TR>

		<TR><TD>ConstRC::NegativeIndex	   </TD><TD> if aRow &lt; 0</TD></TR>
		</TABLE>
*/

int  BCMatrix::ZeroizeRow(long aRow){
	int ret;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;	
	this->matrix[aRow]=0;
	return ConstRC::Ok;
}

/*!
	  This method reads numbers from aRowList array[0 ... aListMaxIndex] and zeroes out rows with those indices.
	  Before this, it checks whether the aRowList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). It checks for validity of all indices and aListMaxIndex.
	  
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveRowIndex</TD><TD>if aRow &gt; maximal_row_index</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied </TD><TD> if aRowList is NULL</TD></TR>
		<TR><TD>ConstRC::NegativeIndex	  </TD><TD> if aListMaxIndex &lt; 0 or any of the indices in list < 0</TD></TR>
		</TABLE>
*/

int  BCMatrix::ZeroizeRow(long* aRowList, long aListMaxIndex){
	if (!aRowList)		return ConstRC::NullPointerSupplied;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long j;
	for (j=0; j<= aListMaxIndex; j++){
		if (aRowList[j] > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		this->matrix[aRowList[j]]=0;
	}
	return ConstRC::Ok;
}

/*!
  	  This method ensures swap of rows with indices aRow1 and aRow2. If the indices are equal, it does nothing.
	  If they are distinct, it ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices
	  
	  The swap is performed by three XORs.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveRowIndex</TD><TD>if aRow1 or aRow2 &gt; maximal_row_index</TD></TR>
		<TR><TD>ConstRC::NegativeIndex	  </TD><TD> if aRow1 or aRow2 &lt; 0 </TD></TR>
		</TABLE>
*/

int  BCMatrix::SwapRows(long aRow1, long aRow2){
	if (aRow1 != aRow2){
		if (aRow1 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		if (aRow2 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		int ret;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		this->matrix[aRow2] ^= this->matrix[aRow1];
		this->matrix[aRow1] ^= this->matrix[aRow2];
		this->matrix[aRow2] ^= this->matrix[aRow1];
		}
	return ConstRC::Ok;
}

/*!
	  This method adds row with index aSource to the row with index aTarget. The indices may be equal, in which
	  case it zeroes out the row by calling ZeroizeRow(aTarget) - multiplication by 2 is equivalent to zero in GF(2).
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.
	  
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>

		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>

		<TR><TD>ConstRC::ExcessiveRowIndex</TD><TD>if aTarget or aSource > maximal_row_index</TD></TR>

		<TR><TD>ConstRC::NegativeIndex	  </TD><TD> if aTarget or aSource < 0 </TD></TR>
		</TABLE>
*/

int  BCMatrix::AddRows(long aTarget, long aSource){
	int ret;
	if (aTarget != aSource){
		if (aTarget > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		if (aSource > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		this->matrix[aTarget] ^= this->matrix[aSource];
		}
	else { // adding a row to itself is zeroizing in GF(2)
		ret = ZeroizeRow(aTarget);
		if (ret != ConstRC::Ok) return ret;
	}
	return ConstRC::Ok;
}

/*!
	  This method adds row with index aRow to the row with index aRow2 from the matrix aOperand.
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.
	  
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>

		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>

		<TR><TD>ConstRC::ExcessiveRowIndex</TD><TD>if aRow or aRow2 > corresponding maximal_row_index</TD></TR>

		<TR><TD>ConstRC::NegativeIndex	  </TD><TD> if aRow or aRow2 < 0 </TD></TR>
		</TABLE>
*/


int  BCMatrix::AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2){
        int ret;
        BCMatrix *aO;
        if ((aRow < 0) || (aRow2 < 0))                             return ConstRC::NegativeIndex;
        if (aRow > this->maximal_row_index)       	           return ConstRC::ExcessiveRowIndex;
        if (aRow2 > aOperand->GetMaximalRowIndex())                return ConstRC::ExcessiveRowIndex;
	if (aOperand->GetSpecies() != SBCMatrix)	     return ConstRC::GeneralError;
        aO = (BCMatrix*) aOperand;
        ret = aO->Allocate();
        if (ret != ConstRC::Ok)  return ret;
        this->matrix[aRow] ^= aO->matrix[aRow2];
        return ConstRC::Ok;
}

/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  column at given index aColumn with zeros (by masking the corresponding bit out).
	  The index is checked for validity.
	  
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex</TD><TD>if aColumn &gt; maximal_column_index</TD></TR>
		<TR><TD>ConstRC::NegativeIndex	  </TD><TD> if aColumn &lt; 0	</TD></TR>
		</TABLE>
*/

int  BCMatrix::ZeroizeColumn(long aColumn){
	int ret,i;
	if (aColumn > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	for (i=0; i<=this->max_allocated_row_index; i++){
		this->matrix[i] &= ~K_SHIFTS[aColumn];
	}
	return ConstRC::Ok;
}

/*!
	  This method reads numbers from aColumnList array[0 ... aListMaxIndex] and zeroes out columns with those indices.
	  Before this, it checks whether the aColumnList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). The indices are checked for validity.
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex</TD><TD>if aColumn &gt; maximal_column_index</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied </TD><TD> if aColumnList is NULL</TD></TR>
		<TR><TD>ConstRC::NegativeIndex    </TD><TD> if either aListMaxIndex or any of the indices is < 0</TD></TR>
		</TABLE>
*/

int  BCMatrix::ZeroizeColumn(long* aColumnList, long aListMaxIndex){
	if (!aColumnList)		return ConstRC::NullPointerSupplied;
	int ret,i,j;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	matrix_type mask = ~0;
	for (j=0; j<= aListMaxIndex; j++){
		if (aColumnList[j] > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		mask &= ~K_SHIFTS[aColumnList[j]];
	}
	for (i=0; i<=this->max_allocated_row_index; i++){
		this->matrix[i] &= mask;
	}
	return ConstRC::Ok;
}


/*!
	  This method is used to perform the column mask of matrices that the compiler thinks are abstract. The method
	  checks the species of both matrices and if it they are BCMatrices, it calls
		int PerformColumnMask(BCMatrix* aTarget, BCMatrix* aOperand, matrix_type aMask);
	  otherwise it returns ConstRC::NotSupported.

	  Return codes: 
	  <TABLE border=0>
	 <TR><TD colspan=2> the method returns the return code of the launched method or</TD></TR>
		<TR><TD>ConstRC::NotSupported	</TD><TD>if the matrices are not BCMatrices.</TD></TR>
		</TABLE>
*/

int  BCMatrix::PerformColumnMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask){
	if(aTarget->GetSpecies()!=SBCMatrix) return ConstRC::GeneralError;
	if(aOperand->GetSpecies()!=SBCMatrix) return ConstRC::GeneralError;
	return this->PerformColumnMask((BCMatrix*)aTarget, (BCMatrix*) aOperand, aMask);
}

int  BCMatrix::PerformRowMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask)
{
	if(aTarget->GetSpecies()!=SBCMatrix) return ConstRC::GeneralError;
	if(aOperand->GetSpecies()!=SBCMatrix) return ConstRC::GeneralError;
	return this->PerformRowMask((BCMatrix*)aTarget, (BCMatrix*) aOperand, aMask);
}

/*!
	  The aTarget and aOperand matrices must be of the same dimensions. Proper allocation of both aTarget and aOperand
	  is ensured by call of Allocate(). The operation is masking of the aOperand entries with aMask, which is equivalent
	  to zeroizing of those columns, whose indices are given by 0 bits in aMask. In turn, this is equivalent to
	  multiplying of aOperand by a bits_in_m_t x bits_in_m_t matrix, which has zeros outside the main diagonal, and 1s
	  on these indices of main diagonal, where aMask & K_SHIFTS[index] != 0.

	  This is an auxiliary method for the block Lanczos method.
	  
	  Return values:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok 		</TD><TD>if everything went all right</TD></TR>
		<TR><TD>ConstRC::SizeMismatch   </TD><TD>if the requirements on sizes of matrices aOperand and aTarget are not met</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD>if Allocate() fails.</TD></TR>
		</TABLE>
*/

int  BCMatrix::PerformColumnMask(BCMatrix* aTarget, BCMatrix* aOperand, matrix_type aMask){
        if (!aOperand || !aTarget)                      {
                return ConstRC::NullPointerSupplied;
        }
        int ret;
        ret = aTarget->Allocate();
        if (ret != ConstRC::Ok) {
                return ret;
        }
        ret = aOperand->Allocate();
        if (ret != ConstRC::Ok) {
                return ret;
        }
        if (aTarget->maximal_row_index != aOperand->maximal_row_index)          {
                return ConstRC::SizeMismatch;
        }
        if (aTarget->maximal_column_index != aOperand->maximal_column_index){
                return ConstRC::SizeMismatch;
        }
        if (aTarget->max_allocated_column_index) {
                return ConstRC::SizeMismatch;
        }
        int i;
        for (i=0; i<=aTarget->max_allocated_row_index; i++){
                aTarget->matrix[i] = aOperand->matrix[i] & aMask;
        }
        return ConstRC::Ok;
}

int  BCMatrix::PerformRowMask(BCMatrix* aTarget, BCMatrix* aOperand, matrix_type aMask)
{
	if (!aOperand || !aTarget)   			{
		return ConstRC::NullPointerSupplied;
	}	
	int ret;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok) {
		return ret;
	}
	ret = aOperand->Allocate();
	if (ret != ConstRC::Ok) {
		return ret;
	}
	if (aTarget->maximal_row_index != aOperand->maximal_row_index)		{
		return ConstRC::SizeMismatch;
	}
	if (aTarget->maximal_column_index != aOperand->maximal_column_index){
		return ConstRC::SizeMismatch;
	}
	if (aTarget->max_allocated_column_index) {
		return ConstRC::SizeMismatch;
	}
	if ((aOperand->maximal_row_index+1) != bits_in_m_t)
	{
		return ConstRC::SizeMismatch;
	}
	int i;
	aTarget->Copy(aOperand);
	for (i=0; i< bits_in_m_t; i++){
		if (!(aMask & K_SHIFTS[i]))
		{
			aTarget->ZeroizeRow(i);
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures swap of columns with indices aColumn1 and aColumn2. If the indices are equal, it does nothing.
	  If they are distinct, it ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.

	  The swap is performed iff the corresponding bits differ, by |= operation.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex</TD><TD>if aColumn1 or aColumn2 &gt; maximal_column_index</TD></TR>
		<TR><TD>ConstRC::NegativeIndex    </TD><TD> if aColumn1 or aColumn2 &lt; 0</TD></TR>
		</TABLE>
*/

int  BCMatrix::SwapColumns(long aColumn1, long aColumn2){
	if (aColumn1 != aColumn2){
		if (aColumn1 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		if (aColumn2 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		int ret;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		int bit_1, bit_2;
		for (k=0; k<=this->max_allocated_row_index; k++){
			// do the bits need to be swapped?
			bit_1 = (this->matrix[k] >> aColumn1) & 1;
			bit_2 = (this->matrix[k] >> aColumn2) & 1;
			if (bit_1 != bit_2){
				// we set both the bits to the other value
				this->matrix[k] ^= K_SHIFTS[aColumn1];
				this->matrix[k] ^= K_SHIFTS[aColumn2];
			}
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method adds column with index aSource to the column with index aTarget. The indices may be equal, in which
	  case it just multiplies the contents of the column by 2.

	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  - </TD><TD>  everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory - </TD><TD>  thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex- </TD><TD> if aTarget or aSource > maximal_column_index</TD></TR>
		<TR><TD>ConstRC::NegativeIndex    - </TD><TD>  if aTarget or aSource < 0</TD></TR>
		</TABLE>
*/

int  BCMatrix::AddColumns(long aTarget, long aSource){
	int ret;
	if (aTarget != aSource){
		if (aTarget > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		if (aSource > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		for (k=0; k<=this->max_allocated_row_index; k++){
			 this->matrix[k] ^= (((this->matrix[k] >> aSource) & 1) * K_SHIFTS[aTarget]); // either 0 or power of 2
		}
	} else { // adding a column to itself is zeroizing in GF(2)
		ret = ZeroizeColumn(aTarget);
		if (ret != ConstRC::Ok) return ret;
	}
	return ConstRC::Ok;
}

/*!
	  This method adds column with index aColumn2 from aSource matrix to the column with index aColumn1 in aTarget
	  matrix. The method checks validity of indices.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex</TD><TD>if aTarget or aSource > maximal_column_index</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied </TD><TD>if aTarget and/or aSource are NULL.</TD></TR>
		<TR><TD>ConstRC::NegativeIndex    </TD><TD> if aColumn1 or aColumn2 < 0</TD></TR>
		</TABLE>
*/	

int  BCMatrix::AddColumnsTwoMatrices(BCMatrix* aTarget, BCMatrix* aSource, long aColumn1, long aColumn2){

	if (aColumn1 > aTarget->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aColumn2 > aSource->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (!aTarget->matrix || !aSource->matrix)		     return ConstRC::NullPointerSupplied;

	long k;
	for (k=0; k<=this->max_allocated_row_index; k++){
		aTarget->matrix[k] ^= (((aSource->matrix[k] >> aColumn2) & 1) * K_SHIFTS[aColumn1]);
	}
	return ConstRC::Ok;
}

/*
int  BCMatrix::Transpose(BitMatrix* aTarget){
        start = clock();
        if (!aTarget)                                                   return ConstRC::NullPointerSupplied;
        if (aTarget->GetMaximalRowIndex() != this->maximal_column_index)   return ConstRC::SizeMismatch;
        if (aTarget->GetMaximalColumnIndex() != this->maximal_row_index)   return ConstRC::SizeMismatch;
        int ret = aTarget->Allocate();
        if (ret != ConstRC::Ok)          return ret;
        long i, j, x;
        // One by one, we read columns of the original matrix
        for (i=0; i<=this->maximal_column_index; i++)
        {
                x = K_SHIFTS[i];
                for (j=0; j<=this->maximal_row_index; j++){
                        if (this->matrix[j] & x){
                                aTarget->PutOne(i,j);
                        }
                }
        }
        diff = clock() - start;
        if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values
                total_transpose += (double) diff;
        }
        return ConstRC::Ok;
}

BitMatrix* BCMatrix::Transpose(){
        BitMatrix* result = new BitMatrix(this->maximal_column_index+1, this->maximal_row_index+1);
        if (!result) return NULL;
        if (this->Transpose(result) != ConstRC::Ok){
                delete result;
                return NULL;
        }
        return result;
}
*/

/*!
	  This method ensures allocation state, dimension requirements etc., and then adds aOperand2 to the calling instance
	  and places the result into matrix aTarget. Both the calling instance and aOperand2 must NOT be equal to aTarget.
	  
	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::SizeMismatch     </TD><TD> the dimensions do not match</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied </TD><TD> if aOperand2 or aTarget is NULL</TD></TR>
		<TR><TD>ConstRC::BadArgument      </TD><TD>  if aTarget is equal to any of the operands.</TD></TR>
		</TABLE>

 	  If the symbolic constant
	        MATRIX_OPERATIONS_TIME_MESSAGE
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_add.
*/

int BCMatrix::Add(BCMatrix* aTarget, BCMatrix* aOperand2){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        start = clock();
#endif
	if (!aTarget)  								return ConstRC::NullPointerSupplied;
	if (aTarget->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (!aOperand2)								return ConstRC::NullPointerSupplied;
	if (aOperand2->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aOperand2->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	int ret;
	long i;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	for (i=0; i<=aTarget->max_allocated_row_index;i++){
		aTarget->matrix[i] = this->matrix[i] ^ aOperand2->matrix[i];
		}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        diff = clock() - start;
        if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values
                total_add += (double) diff;
        }
#endif
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the addition operation, and then performs
	  the addition by calling 
	  	int Add(BCMatrix* aTarget)
	  If result has been allocated, but Add did not finish well, the result is deleted.
	  
	  Returns:
	  <TABLE border=0>
	  	<TR><TD>pointer to result of operation, if everything went all right</TD></TR>
		<TR><TD>NULL in case of any problem (size mismatches, allocation problems)</TD></TR>
		</TABLE>
*/	

BCMatrix* BCMatrix::Add(BCMatrix* aOperand2){
	BCMatrix* result = new BCMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if ((!result) || (this->Add(result,aOperand2) != ConstRC::Ok)) return NULL; else return result;
}

/*!
	  This method is used to treat the adding of matrices that the compiler thinks are abstract. The method checks
	  the species of aTarget and aOperand2 and if they are BCMatrices, it calls
	  	Add((BCMatrix*) aTarget, (BCMatrix*) aOperand2);
	  otherwise it calls the general abstract method.

	  Return codes: the method returns the return code of the launched method.
*/

int  BCMatrix::Add(AbstractMatrix* aTarget, AbstractMatrix* aOperand2){
        if((aOperand2->GetSpecies()==SBCMatrix)&&(aTarget->GetSpecies()==SBCMatrix))
          return this->Add((BCMatrix*)aTarget, (BCMatrix*) aOperand2);
        return AbstractMatrix::Add(aTarget, aOperand2);
}

/*!
	  This method is used to treat the multiplying of matrices that the compiler thinks are abstract. The method checks
	  the species of aTarget, aOperand1 and aOperand2 and if they are BCMatrices, it calls
	  	MultiplyInternal((BCMatrix*) aTarget, (BCMatrix*) aOperand1, (BCMatrix*) aOperand2);
	  otherwise it calls the general abstract method.

	  Return codes: the method returns the return code of the launched method.
*/

int  BCMatrix::MultiplyInternal(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2){
        if((aOperand1->GetSpecies()==SBCMatrix)&&(aOperand2->GetSpecies()==SBCMatrix)&&(aTarget->GetSpecies()==SBCMatrix))
          return this->MultiplyInternal((BCMatrix*)aTarget, (BCMatrix*) aOperand1, (BCMatrix*) aOperand2);
        return AbstractMatrix::MultiplyInternal(aTarget, aOperand1, aOperand2);
}

/*!
	  This method is used to treat the multiplying of matrices that the compiler thinks are abstract. The method checks
	  the species of aTarget, aOperand1 and aOperand2 and if they are BCMatrices, it calls
	  	MultiplyInternalTransposed((BCMatrix*) aTarget, (BCMatrix*) aOperand1, (BCMatrix*) aOperand2);
	  otherwise it calls the general abstract method.

	  Return codes: the method returns the return code of the launched copy method.
*/

int  BCMatrix::MultiplyInternalTransposed(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2){
        if((aOperand1->GetSpecies()==SBCMatrix)&&(aOperand2->GetSpecies()==SBCMatrix)&&(aTarget->GetSpecies()==SBCMatrix))
          return this->MultiplyInternalTransposed((BCMatrix*)aTarget, (BCMatrix*) aOperand1, (BCMatrix*) aOperand2);
        return AbstractMatrix::MultiplyInternalTransposed(aTarget, aOperand1, aOperand2);
}


/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget. If the matrix aOperand1 has more than 1000 rows, it calls
		MultiplyInternalCoppersmith(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2);
	  instead.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		  </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::SizeMismatch     </TD><TD> the dimensions do not match</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied </TD><TD> if aOperand1, aOperand2 or aTarget is NULL</TD></TR>
		</TABLE>
 	  If the symbolic constant
	        MATRIX_OPERATIONS_TIME_MESSAGE
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_word_multiply.
*/

int  BCMatrix::MultiplyInternal(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        start = clock();
#endif
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if (aTarget->maximal_row_index != aOperand1->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_column_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	aTarget->Zeroize();
	if(aOperand1->maximal_row_index>1000) MultiplyInternalCoppersmith(aTarget,aOperand1,aOperand2);
	else {
	for(i=0; i<=aOperand1->maximal_row_index; i++)
		for(j=0; j<=aOperand1->maximal_column_index; j++)
			if (aOperand1->matrix[i] & K_SHIFTS[j])
				aTarget->matrix[i] ^= aOperand2->matrix[j];
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        diff = clock() - start;
        if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values
                total_word_multiply += (double) diff;
        }
#endif
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternal(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.

	  Returns:
	  <TABLE border=0>
	  	<TR><TD>pointer to result of operation, if everything went all right</TD></TR>
		<TR><TD>NULL in case of any problem (size mismatches, allocation problems)	</TD></TR>		  
		</TABLE>

*/

BCMatrix* BCMatrix::MultiplyInternal(BCMatrix *aOperand1,BCMatrix* aOperand2){
	BCMatrix* result = new BCMatrix(aOperand1->maximal_row_index+1, aOperand2->maximal_column_index+1);
	if ((!result) || (this->MultiplyInternal(result,aOperand1,aOperand2) != ConstRC::Ok)) return NULL; else return result;
}

/*!
	  This method performs the multiplication of two BCMatrices which is about 8 times faster (if aOperand1
	  has more than about 1000 rows) than the classical MultiplyInternal. This method performs multiplication
	  aOperand1 x aOperand2, saving the
	  result into aTarget. Both aOperand1 and aOperand2 must NOT be equal to aTarget.

	  Return codes: ConstRC::Ok
*/

int  BCMatrix::MultiplyInternalCoppersmith(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2){
	int bytes,k; // ret
	long i,j;
	BCMatrix *C;
	bytes=this->bits_in_m_t/8;
	C=new BCMatrix(bytes*256,bits_in_m_t);
	C->Allocate();
	C->Zeroize();
	for(j=0;j<256;j++)
		for(k=0;k<bytes;k++)
		{
			for(i=0;i<8;i++)
			if(j & K_SHIFTS[i]) C->matrix[k+bytes*j] ^= aOperand2->matrix[i+k*8];
		}
	for(j=0;j<=aTarget->maximal_row_index;j++)
	{
		i=aOperand1->matrix[j];
		for(k=0;k<bytes;k++)
		{
			aTarget->matrix[j] ^= C->matrix[k+bytes*(i & 255)];
			i=i >> 8;
		}
	}
	delete C;
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication (aOperand1)^T x (aOperand2), saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget. If the matrix aOperand1 has more than 1000 rows, it calls
		MultiplyInternalCoppersmith(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2);
	  instead.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok		   </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::NotEnoughMemory  </TD><TD> thrown from Allocate() call; there was not enough memory to allocate the requested space.</TD></TR>
		<TR><TD>ConstRC::SizeMismatch      </TD><TD> the dimensions do not match</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied  </TD><TD> if aOperand1, aOperand2 or aTarget is NULL</TD></TR>
		<TR><TD>ConstRC::BadArgument       </TD><TD>  if aTarget is equal to any of the operands.</TD></TR>
		</TABLE>

 	  If the symbolic constant
	        MATRIX_OPERATIONS_TIME_MESSAGE
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_multiply_transpose_word.
*/

int  BCMatrix::MultiplyInternalTransposed(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        start = clock();
#endif
	if (!aOperand1 || !aOperand2 || !aTarget)   			{
		return ConstRC::NullPointerSupplied;
	}
	if (aTarget->maximal_row_index != aOperand1->maximal_column_index)		{
		return ConstRC::SizeMismatch;
	}
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index){
		return ConstRC::SizeMismatch;
	}
	if (aOperand1->maximal_row_index != aOperand2->maximal_row_index){
		return ConstRC::SizeMismatch;
	}
	int ret;
	long i,j;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	aTarget->Zeroize();
	if(aOperand1->maximal_row_index>1000) MultiplyInternalTransposedCoppersmith(aTarget,aOperand1,aOperand2);
	else {
	for(i=0; i<=aOperand1->maximal_column_index; i++)
		for(j=0; j<=aOperand1->maximal_row_index; j++)
			if (aOperand1->matrix[j] & K_SHIFTS[i])
				aTarget->matrix[i] ^= aOperand2->matrix[j];
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
        diff = clock() - start;
        if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values
                total_word_transpose_multiply += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternalTransposed(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.

	  Returns:
	  <TABLE border=0>
	  	<TR><TD>pointer to result of operation, if everything went all right</TD></TR>
		<TR><TD>NULL in case of any problem (size mismatches, allocation problems)	</TD></TR>		  
		</TABLE>

*/


BCMatrix* BCMatrix::MultiplyInternalTransposed(BCMatrix *aOperand1,BCMatrix* aOperand2){
	BCMatrix* result = new BCMatrix(aOperand1->maximal_column_index+1, aOperand2->maximal_column_index+1);
	if ((!result) || (this->MultiplyInternalTransposed(result,aOperand1,aOperand2) != ConstRC::Ok)) return NULL; else return result;
}

/*!
	  This method performs the multiplication of two BCMatrices which is about 8 times faster (if aOperand1
	  has more than about 1000 rows) than the classical MultiplyInternal. This method performs multiplication
	  aOperand1^T x aOperand2, saving the
	  result into aTarget. Both aOperand1 and aOperand2 must NOT be equal to aTarget.

	  Return codes: ConstRC::Ok
*/

int  BCMatrix::MultiplyInternalTransposedCoppersmith(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2){
	int k, l, h = 0; // ret
	long i,j,bytes;
	BCMatrix *C;
	bytes=this->bits_in_m_t/8;
	C=new BCMatrix(bytes*256,this->bits_in_m_t);
	C->Allocate();
	C->Zeroize();
	for(j=0; j<=aOperand1->maximal_row_index; j++)
	{
		i=aOperand1->matrix[j];
		for(k=0; k<bytes; k++)
		{
			C->matrix[k*256+(i & 255)] ^= aOperand2->matrix[j];
			i = i >> 8;
		}
	}
	for(k=0;k<bytes;k++)
		for(l=0;l<8;l++)
		{
			for(j=0;j<256;j++)
			if(j & K_SHIFTS[l]) aTarget->matrix[h] ^= C->matrix[k*256+j];
			h++;
		}
	delete(C);
	return ConstRC::Ok;
}


/*!
	  This method instantiates a new NormalMatrix instance of the same dimensions as the caller instance of BCMatrix; then,
	  it allocates the internal matrix of the NormalMatrix instance and converts the matrix represented by caller BCMatrix
	  instance into this freshly allocated NormalMatrix instance.

	  Return values:
	  <TABLE border=0>
	  	<TR><TD>pointer to NormalMatrix* instance if everything went all right</TD></TR>
		<TR><TD>NULL if some of the allocations was unsuccessful.</TD></TR>
		</TABLE>
*/

NormalMatrix* BCMatrix::ToNormalMatrix(){
	NormalMatrix* aResult = new NormalMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i, j; //, index;
	for (i=0; i<=this->maximal_row_index; i++){
		for (j=0; j<=this->maximal_column_index; j++){
			if (this->matrix[i] & K_SHIFTS[j]) aResult->PutOne(i,j);
		}
	}
	return aResult;
}

/*!
	  This method instantiates a new BCMatrix instance of the same dimensions as the parameter aMatrix; then,
	  it allocates the internal matrix of the BCMatrix instance and converts the matrix represented by aMatrix
	  instance into this freshly allocated BitMatrix instance. The rule is that odd entries of aMatrix are turned to bit 1
	  and even entries are turned to bit 0.

	  Return values:
	  <TABLE border=0>
	  	<TR><TD>pointer to BitMatrix* instance if everything went all right</TD></TR>
		<TR><TD>NULL if some of the allocations was unsuccessful.</TD></TR>
		</TABLE>
*/

BCMatrix* BCMatrix::ToBCMatrix(NormalMatrix* aMatrix){
	long max_r = aMatrix->GetMaximalRowIndex();
	long max_c = aMatrix->GetMaximalColumnIndex();
	BCMatrix* aResult = new BCMatrix(max_r+1,max_c+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j;
	integer_matrix_type read;

	for (i=0; i<=max_r; i++){
		for (j=0; j<=max_c; j++){
			read = aMatrix->GetXY(i,j);
			if (read == ConstRC::GeneralError){
				delete aResult;
				return NULL;
			}
			if (read % 2){
				aResult->PutOne(i,j);
			}
		}
	}
	return aResult;
}

/*!
  Returns whether the cell given by the coordinates contains 1.

  Return Codes:
	  <TABLE border=0>
	  	<TR><TD>TRUE or FALSE		   </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::ExcessiveRowIndex  </TD><TD> if the row index is outside the matrix</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex  </TD><TD> if the column index is outside the matrix</TD></TR>
		</TABLE>
*/

int  BCMatrix::IsOne(long aRow, long aColumn){
	if (aRow > this->maximal_row_index) 		return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index) 	return ConstRC::ExcessiveColumnIndex;
	if ((this->matrix[aRow] & K_SHIFTS[aColumn]))	{
		return TRUE;
	}
	else return FALSE;
}

/*!
  Returns whether the cell given by the coordinates contains 1.

  Return Codes:
	  <TABLE border=0>
	  	<TR><TD>TRUE or FALSE		   </TD><TD> everything all right</TD></TR>
		<TR><TD>ConstRC::ExcessiveRowIndex  </TD><TD> if the row index is outside the matrix</TD></TR>
		<TR><TD>ConstRC::ExcessiveColumnIndex  </TD><TD> if the column index is outside the matrix</TD></TR>
		</TABLE>
*/

int  BCMatrix::IsZero(long aRow, long aColumn){
	if (aRow > this->maximal_row_index) 		return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index) 	return ConstRC::GeneralError;
	if (this->matrix[aRow] & K_SHIFTS[aColumn])	return FALSE;
	else return TRUE;
}

/*!
	  This method tests whether the calling instance is a zero matrix. It does not call Allocate() to ensure 
	  allocation of anything.

	  The comparison ends as soon as any nonzero matrix_type is met.
	  This method does not change any data of any operand.

	  Return codes:
	  <TABLE border=0>
	  	<TR><TD>TRUE 		   </TD><TD> the matrix is a zero matrix</TD></TR>
		<TR><TD>FALSE		   </TD><TD> the matrix is not a zero matrix</TD></TR>
		</TABLE>
*/

int  BCMatrix::IsZero(){
	long i;
	for (i=0; i<= this->max_allocated_row_index; i++){
		if (this->matrix[i]) return FALSE;
	}
	return TRUE;
}

/*!
	  This method uses linear combinations of rows in order to turn the caller instance into a semidiagonal matrix, with
	  all the nonzero entries above or on the main diagonal. It is designed to run on square matrices only.

	  Return values:
	  <TABLE border=0>
	  	<TR><TD>ConstRC::Ok  </TD><TD>if everything went all right</TD></TR>
		<TR><TD>ConstRC::NullPointerSupplied  </TD><TD>if the calling matrix has not been allocated yet.</TD></TR>
		</TABLE>
*/

int BCMatrix::ToSemidiagonalShape(){
	long i,j;
	if (!this->matrix)   	return ConstRC::NullPointerSupplied;
	// use on square matrices only
	long shift = 0;
	for (i=0; i<=this->maximal_column_index	; i++){
		if ((i<= this->maximal_row_index) && (i+shift <= this->maximal_column_index)){
			for (j=i; j<= this->maximal_row_index; j++){
				if (this->IsZero(i,i+shift) && this->IsOne(j,i+shift)){
					this->SwapRows(i,j);
					break;
				}
			}
			if (this->IsOne(i,i+shift)){
				for (j=i+1; j<=this->maximal_row_index; j++){
					if (this->IsOne(j,i+shift)){
						this->AddRows(j,i);
					}
				}
			} else {
				shift++;
				i--;
			}

		}
	}
	return ConstRC::Ok;
}

/*!
  The method returns the maximal allocated row index
*/

long BCMatrix::GetMaxAllocatedRowIndex()  const {
	return this->max_allocated_row_index;
}

/*!
  The method returns the maximal allocated column index
*/

long BCMatrix::GetMaxAllocatedColumnIndex() const {
	return this->max_allocated_column_index;
}

/*!
	  This method creates a copy of the caller instance, performs ToSemidiagonalShape() on this copy and then counts number
	  of nonzero rows. This number is equal to the rank of the caller instance.
	  Possible errors may result from calls of Copy() and ToSemidiagonalShape().
	  This method does not change the contents of the caller instance.

	  Return values:
	  <TABLE border=0>
	  	<TR><TD colspan=2>nonnegative integer equal to the rank of the caller instance if everything went all right</TD></TR>
		<TR><TD>ConstRC::GeneralError </TD><TD> otherwise</TD></TR>
		</TABLE>
*/	

int  BCMatrix::CalculateRank() {
	BCMatrix* result = new BCMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	result->Copy(this);
	result->ToSemidiagonalShape();
	long i, rank = 0;
	for (i=0; i<= result->maximal_row_index; i++){
		if (result->matrix[i])
			rank++;
			else break;
	}
	delete result;
	return rank;
}


/*!
	  This method calls ToSemidiagonalShape() on the caller instance, thereby altering its contents. Then it counts number
	  of nonzero rows. This number is equal to the rank of the caller instance.
	  This approach is usable in situation, when we want to compute rank of an unused matrix, without need to allocate more
	  memory. For example, it is used in optional rank computations in Lanczos method.

	  Return values:
	  <TABLE border=0>
	  	<TR><TD colspan=2>nonnegative integer equal to the rank of the caller instance if everything went all right</TD></TR>
		<TR><TD>ConstRC::GeneralError </TD><TD> otherwise</TD></TR>
		</TABLE>
*/

int  BCMatrix::CalculateRank(bool aOnMyself) {
	this->ToSemidiagonalShape();
	long i, rank = 0;
	for (i=0; i<= this->maximal_row_index; i++){
		if (this->matrix[i]){
			rank++;
			break;
		}
	}
	return rank;
}

/*!
  This method returns the content of the given row
*/

matrix_type BCMatrix::GetRow(long aRow) {
        if ((aRow < 0) || (this->matrix == NULL))       return ConstRC::GeneralError;
        if (aRow > this->maximal_row_index)             return ConstRC::GeneralError;
        return this->matrix[aRow];
}

/*!
  This methods sets the given row to the given value
*/

int BCMatrix::SetRow(long aRow, matrix_type aValue) {
        if ((aRow < 0)  || (this->matrix == NULL))       return ConstRC::GeneralError;
        if (aRow > this->maximal_row_index)             return ConstRC::SizeMismatch;
        this->matrix[aRow] = aValue;
	return ConstRC::Ok;
}

/*!
  This methos adds the given value to the given row
*/

int BCMatrix::AddToRow(long aRow, matrix_type aValue) {
        if ((aRow < 0)  || (this->matrix == NULL))       return ConstRC::GeneralError;
        if (aRow > this->maximal_row_index)             return ConstRC::SizeMismatch;
        this->matrix[aRow] ^= aValue;
	return ConstRC::Ok;
}

int BCMatrix::Save(char *aName)
{
	FILE *fw;
	fw=fopen(aName,"w");
	if (!fw) return ConstRC::GeneralError;
	fwrite(&(this->maximal_row_index),1,sizeof(long),fw);
	fwrite(&(this->maximal_column_index),1,sizeof(long),fw);
	fwrite(this->matrix,this->maximal_row_index+1,sizeof(matrix_type),fw);
	fclose(fw);
	return ConstRC::Ok;
}

int BCMatrix::Load(char *aName)
{
	FILE *fr;
	if (this->matrix) return ConstRC::GeneralError;
	fr=fopen(aName,"r");
	if (!fr) return ConstRC::GeneralError;
	fread(&(this->maximal_row_index),1,sizeof(long),fr);
	fread(&(this->maximal_column_index),1,sizeof(long),fr);
	this->Allocate();
	fread(this->matrix,this->maximal_row_index+1,sizeof(matrix_type),fr);
	fclose(fr);
	return ConstRC::Ok;
}

//int BCMatrix::Serialize(const CBaseParameters &aParam) const
//{
//	return MSerializable::Serialize(aParam);
//}
//
//int BCMatrix::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
//{
//	return AbstractMatrix::Serialize(aParam, aWriter);
//}
//
//int BCMatrix::Deserialize(const CBaseParameters &aParam)
//{
//	return MSerializable::Deserialize(aParam);
//}
//
//int BCMatrix::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
//{
//	return AbstractMatrix::Deserialize(aParam, aReader);
//}

int BCMatrix::WriteData(xmlTextWriterPtr aWriter) const
{
	// write all matrix data
	return CommonXMLHelper::SerializeBinaryData(aWriter, XML_SERIAL_MATRIX_NAME, this->matrix, this->maximal_row_index + 1, sizeof(matrix_type));
}

int BCMatrix::ReadData(xmlTextReaderPtr aReader)
{
	// we know max_row_index and max_column_index
	Allocate();

	xmlChar * name = xmlTextReaderName(aReader);

	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MATRIX_NAME) == 0)
	{
		CLog::LogError(WHEREARG,"BCMatrix","ReadData",0,0,"Could not find the matrix parameter tag.");
		return ConstRC::GeneralError;
	}

	return CommonXMLHelper::DeserializeBinaryDataIn(aReader, "Deserialize matrix...", this->matrix, this->maximal_row_index + 1, sizeof(matrix_type), FALSE);
}
