#ifndef __BC_MATRIX_CLASS__
#define __BC_MATRIX_CLASS__

//defines 1-byte column matrices

#include "normal_matrix_class.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>


/*!
  \brief A class for memory-savvy representation of not-too-sparse matrices with limited width over GF(2).

	The word BCMatrix is an abbreviation for Byte Column Matrix, saying that the number of columns of these matrices
	is limited by the length of byte. Actually, the number of columns is limited by the length of word, not byte,
	but the abbreviation WCMatrix would not be nice.
	
	These matrices are in fact a special kind of BitMatrix type, the only difference is that they are represented
	as an one dimensional field - each element of the field corresponds to a row of the matrix. Each row can
	have at most the length of unsigned integer (typically 32) elements and these elements are stored as the bits
	of the integer representing the given row.

	The advantage of these matrices with respect to Bit Matrices is that they are treated linearly and hence the
	work with them is faster than the work with Bit Matrices of the same size (about 25%). Many algorithms
	(like the Block Lanczos Algorithm) are block ones and they utilize matrices of a fixed width (typically 32)
	and this is exactly where the BCMatrices are used.

*/

class BCMatrix : public AbstractMatrix {
public:
	BCMatrix();
	BCMatrix(long aRows, long aColumns);
	~BCMatrix();

	virtual BCMatrix * clone();

	long GetMaxAllocatedRowIndex()    const;
	long GetMaxAllocatedColumnIndex()    const;
	int  Allocate();
	int  Randomize();
	int  Zeroize();
	int  Copy(BCMatrix* aSource);
	int  Copy(AbstractMatrix* aSource);
	bool Equals(BCMatrix* aMatrix);
	void PrintToScreen();
	int  PutOne(long aRow, long aColumn);
	int  PutZero(long aRow, long aColumn);
	int  IsOne(long aRow, long aColumn);
	int  IsZero(long aRow, long aColumn);
	int  IsZero();
	// Swaps etc.
	int  ZeroizeRow(long aRow);
	int  ZeroizeRow(long* aRowList, long aListMaxIndex);
	int  SwapRows(long aRow1, long aRow2);
	int  AddRows(long aTarget, long aSource);
	int  SetRow(long aRow, matrix_type aValue);
	matrix_type  GetRow(long aRow);
	int  AddToRow(long aRow, matrix_type aValue);
	int  AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2);
	int  ZeroizeColumn(long aColumn);
	int  ZeroizeColumn(long* aColumnList, long aListMaxIndex);
	int  PerformColumnMask(BCMatrix *aTarget, BCMatrix* aOperand, matrix_type aMask);
	int  PerformColumnMask(AbstractMatrix *aTarget, AbstractMatrix* aOperand, matrix_type aMask);
	int  PerformRowMask(BCMatrix* aTarget, BCMatrix* aOperand, matrix_type aMask);
	int  PerformRowMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask);
	int  SwapColumns(long aColumn1, long aColumn2);
        int  AddColumns(long aTarget, long aSource);
	// Mathematical operations
	int  	   Add(BCMatrix* aTarget, BCMatrix* aOperand2);
        int        Add(AbstractMatrix* aTarget, AbstractMatrix* aOperand2);
	BCMatrix* Add(BCMatrix* aOperand2);
	int	MultiplyInternal(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2);
	int	MultiplyInternalCoppersmith(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2);
	//int	MultiplyInternalBig(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2);
        int     MultiplyInternal(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
        int     MultiplyInternalTransposed(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	BCMatrix*	MultiplyInternal(BCMatrix* aOperand1, BCMatrix* aOperand2);
	BCMatrix*	MultiplyInternalTransposed(BCMatrix* aOperand1, BCMatrix* aOperand2);
	//BCMatrix*	MultiplyInternalBig(BCMatrix* aOperand1, BCMatrix* aOperand2);
	int 	MultiplyInternalTransposed(BCMatrix *aTarget, BCMatrix* Operand1, BCMatrix* aOperand2);
	int 	MultiplyInternalTransposedCoppersmith(BCMatrix *aTarget, BCMatrix* Operand1, BCMatrix* aOperand2);
#ifdef RANK_COMPUTATIONS
	//int	MultiplyInternalRank(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2);
	//BCMatrix*	MultiplyInternalRank(BCMatrix* aOperand1, BCMatrix* aOperand2);
#endif
	// Type conversions
	NormalMatrix*     ToNormalMatrix();
	static BCMatrix* ToBCMatrix(NormalMatrix* aMatrix);
	// Others
	int CalculateRank();
	int CalculateRank(bool aOnMyself);
	int ToSemidiagonalShape();
	//int ToSemidiagonalShapeForLanczos(BCMatrix* aMirror);
	int AddColumnsTwoMatrices(BCMatrix* aTarget, BCMatrix* aSource, long aColumn1, long aColumn2);
	static matrix_type* K_SHIFTS;
	int Save(char *aName);
	int Load(char *aName);

	// De/Serialization methods
	//int Serialize(const CBaseParameters& aParam) const;
	//int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	//int Deserialize(const CBaseParameters& aParam);
	//int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
protected:
	int WriteData(xmlTextWriterPtr aWriter) const;
	int ReadData(xmlTextReaderPtr aReader);

protected:
	static void DefineKShifts();
	long max_allocated_row_index;
	long max_allocated_column_index;
	matrix_type* matrix;
	static long bits_in_m_t;///< 	Common to all instances, reflects the number of bits in matrix_type. This value	is needed for type-independent implementation of the BCMatrix class.

};

#endif

