# include "buffered_file_writer.h"
# include "constants.h"
# include "const_return_codes.h"
# include "log.h"
# include <stdlib.h>

/*!
Start routine of writting thread.
After successfull writting, both items and buffer are freed.
*/
void *WriteItems(void *arg) {
	FILE *file;
	CBufferedFileWriterItem **items_buffer, **items_buffer_ptr;
	int filled, i;

	if (arg == NULL) return NULL; // used for iWrittingThread initialization

	file = ((struct WrittingThreadArgs *) arg)->file;
	items_buffer = ((struct WrittingThreadArgs *) arg)->itemsBuffer;
	filled = ((struct WrittingThreadArgs *) arg)->filled;

	free(arg);

	items_buffer_ptr = items_buffer;
	for (i = 0; i < filled; i++) {
		if ((*(items_buffer_ptr++))->PrintToFile(file) != ConstRC::Ok) CLog::LogError(WHEREARG, "CBufferedFileWriter", "WriteItems", 0, 0, "Write item failed.");
	}

	items_buffer_ptr = items_buffer;
	for (i = 0; i < filled; i++) (*(items_buffer_ptr++))->FreeItem();

	free(items_buffer);

	return NULL;
}

CBufferedFileWriter::CBufferedFileWriter() {

	iFile = NULL;
	iItemsBuffer = NULL;
	iFileBuffer = NULL;
	if (pthread_create(&iWrittingThread, NULL, WriteItems, NULL) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "CBufferedFileWriter", 0, 0, "Create writting thread failed.");
		return;
	}

	return;
}

CBufferedFileWriter::~CBufferedFileWriter() {

	if (iFile != NULL) Close();

	return;
}

int CBufferedFileWriter::Flush() {
	struct WrittingThreadArgs *write_args;

	if (iFile == NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Flush", 0, 0, "File is NULL.");
		return ConstRC::GeneralError;
	}

	if (pthread_join(iWrittingThread, NULL) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Flush", 0, 0, "Join writting thread failed.");
		return ConstRC::GeneralError;
	}

	write_args = (struct WrittingThreadArgs *)malloc(sizeof(struct WrittingThreadArgs));

	write_args->file = iFile;
	write_args->itemsBuffer = iItemsBuffer;
	write_args->filled = iItemsBufferFilled;

	if (pthread_create(&iWrittingThread, NULL, WriteItems, write_args) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Flush", 0, 0, "Create writting thread failed.");
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}

/*!
Opens file for write/append. Allocates items buffer and file buffer.
Sets file buffer to be used by stream output operations.
*/
int CBufferedFileWriter::Open(const char *aFilename, bool aAppend, int aItemsBufferSize, int aFileBufferSize) {

	if (iFile != NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "File is not NULL.");
		return ConstRC::GeneralError;
	}

	if (iItemsBuffer != NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "Items buffer is not NULL.");
		return ConstRC::GeneralError;
	}

	if (iFileBuffer != NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "File buffer is not NULL.");
		return ConstRC::GeneralError;
	}

	if (aItemsBufferSize <= 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "Bad items buffer size.");
		return ConstRC::GeneralError;
	}

	if (aFileBufferSize <= 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "Bad file buffer size.");
		return ConstRC::GeneralError;
	}

	if (aAppend) iFile = fopen(aFilename, "a");
	else iFile = fopen(aFilename, "w");

	if (iFile == NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "fopen failed.");
		return ConstRC::GeneralError;
	}

	if ((iFileBuffer = (char *)malloc(aFileBufferSize*sizeof(char))) == NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "malloc failed.");
		fclose(iFile);
		iFile = NULL;
		return ConstRC::GeneralError;
	}

	if (setvbuf(iFile, iFileBuffer, _IOFBF, aFileBufferSize) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "setvbuf failed.");
		fclose(iFile);
		iFile = NULL;
		free(iFileBuffer);
		iFileBuffer = NULL;
		return ConstRC::GeneralError;
	}

	if ((iItemsBuffer = (CBufferedFileWriterItem **)malloc(aItemsBufferSize*sizeof(CBufferedFileWriterItem *))) == NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Open", 0, 0, "malloc failed.");
		fclose(iFile);
		iFile = NULL;
		free(iFileBuffer);
		iFileBuffer = NULL;
		return ConstRC::GeneralError;
	}
	iItemsBufferSize = aItemsBufferSize;
	iItemsBufferFilled = 0;

	return ConstRC::Ok;
}

int CBufferedFileWriter::Close() {

	if (iFile == NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Close", 0, 0, "File is NULL.");
		return ConstRC::GeneralError;
	}

	if (Flush() != ConstRC::Ok) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Close", 0, 0, "Flush buffer failed.");
		return ConstRC::GeneralError;
	}

	if (pthread_join(iWrittingThread, NULL) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Close", 0, 0, "Join writting thread failed.");
		return ConstRC::GeneralError;
	}

	iItemsBuffer = NULL;


	if (fflush(iFile) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Close", 0, 0, "fflush failed.");
		return ConstRC::GeneralError;
	}

	if (fclose(iFile) != 0) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Close", 0, 0, "fclose failed.");
		return ConstRC::GeneralError;
	}

	iFile = NULL;

	free(iFileBuffer);
	iFileBuffer = NULL;

	return ConstRC::Ok;
}

int CBufferedFileWriter::Write(CBufferedFileWriterItem &aItem) {

	if (iFile == NULL) {
		CLog::LogError(WHEREARG, "CBufferedFileWriter", "Write", 0, 0, "File is NULL.");
		return ConstRC::GeneralError;
	}

	if (iItemsBufferFilled == iItemsBufferSize) {
		if (Flush() != ConstRC::Ok) {
			CLog::LogError(WHEREARG, "CBufferedFileWriter", "Write", 0, 0, "Flush failed.");
			return ConstRC::GeneralError;
		}

		if ((iItemsBuffer = (CBufferedFileWriterItem **)malloc(iItemsBufferSize*sizeof(CBufferedFileWriterItem *))) == NULL) {
			CLog::LogError(WHEREARG, "CBufferedFileWriter", "Write", 0, 0, "malloc failed.");
			return ConstRC::GeneralError;
		}
		iItemsBufferFilled = 0;
	}

	iItemsBuffer[iItemsBufferFilled++] = aItem.CopyItem();

	return ConstRC::Ok;
}

