#include "center.h"
#include "abstract_connector.h"
#include "abstract_message.h"
#include "filesystem_message.h"
#include "node_info.h"
#include "distributed_algorithm.h" 
#include "abstract_serializator.h"
#include "job_parameters.h"
#include "job_info.h"
#include "crc32.h"
#include "xml_message_definitions.h"

// The library for cross-platform management of filesystem operations.
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include "boost/progress.hpp"
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>
#include <string.h>

#include "boost_version.h"

namespace fs = boost::filesystem;

#define DEFAULT_ASK_AGAIN_IN_SECONDS 60


Center::Center(AbstractSerializator& aSerializator, const string& aWorkingDirectoryName) : Receiver(aSerializator)
{
	iWorkingDirectoryName.assign(aWorkingDirectoryName);
	iBreakNow = FALSE;
	iRunningJob = NULL;
	iCounter = 1;

	return;
}



Center::~Center()
{
	
	int n = 0, size = iNodes.size();
	while (n < size)
	{
		delete this->iNodes[n];		
		n++;
	}

}

/// This method will register a node into the list of currently active
/// nodes, using the content of the message supplied (EReady message).
int	Center::RegisterNode(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return ConstRC::Ok;
	}
	NodeInfo*	info = new NodeInfo(aMessage);
	int i, size = iNodes.size();
	for (i=0; i< size; i++)
	{
		if (info->Identical(iNodes[i]))
		{	
			delete info; // To prevent memory leak
			return NODE_ALREADY_REGISTERED;
		}
	}
	iNodes.push_back(info);
	printf("RegisterNode::New node registered. Check should be performed!");
	return ConstRC::Ok;
}

void	Center::UnregisterNode(NodeInfo* aInfo)
{
	if (aInfo)
	{			
		ExpungeCommunicationInfo(aInfo->GetId());
	}
	printf("RegisterNode::Unregister node not implemented yet.");
}

int 	Center::Run()
{
	int wakeCounter = 0, ret;

	ret = CheckInitConsistency();
	if (ret != ConstRC::Ok)
	{
		return ret;
	}
	CreateExtraConnectors();
	string str1;

	do {
		if ((wakeCounter) % 12 == 0)
		{
			PrintAllNodes();
		}
		if ((wakeCounter) % 15 == 0)
		{
			UpdateFromJobsInDir();
		}
		// check running job
		CheckRunningJob();
		// Select a new job if a previous one was finished		
		if (!IsSomeJobRunning())
		{
			SelectNextJobToRun();
			PrepareFirstBatch();
		}
		// Defined as a macro in definitions.h
		SLEEP(1); // Sleep for 1 second		
		FormatTime(str1,wakeCounter);
		printf("\rCenter woke up (%lu jobs to do, %lu jobs completed) ... %s     ",iJobsToDo.size(),iCompletedJobs.size(),str1.c_str());
		FetchNewMessages();
		wakeCounter++;
		fflush(stdout);		
	} while (!iBreakNow);
	return ConstRC::Ok;
}

void	Center::PrepareFirstBatch() 
{
	if (iRunningJob)
	{
		// Init the algorithm first.
		const JobParameters& params = iRunningJob->Get_Parameters();
		if (!iDistAlg)
		{
			Algorithm* alg1 = iSerializator.CreateInstance(params);
			if (alg1)
			{
				// Try to upcast the alg to distributed algorithm
 				try
  				{
					iDistAlg = dynamic_cast<DistributedAlgorithm*>(alg1);
				} 
				catch (std::bad_cast e)
  				{    	
					printf("Center::Prepare First Batch. Could not cast to distributed algorithm\n");
					return;
  				}				
			}
		} 
                else // iDistAlg available
		{
			iDistAlg->SetupParameters(params);
		}
		PrepareNextBatch();
	}
	else
	{
		printf("Center::Prepare First Batch. No RunningJob available\n");
	}
}

int	Center::PrepareNextBatch()
{
	if (!AvailableParsForNodes())
	{
		if ((iRunningJob) && (iDistAlg))
		{
			Batch newBatch = iDistAlg->CreateNewJobs();
			int i,size = newBatch.size();
			fprintf(stdout, "\nnew batch of parameters:\n\n");
			for (i=0; i< size; i++)
			{
				iAvailableParametersForNodes.push_back(newBatch[i]);
			}
			fprintf(stdout, "\tlast job in batch:");
			newBatch[size-1]->PrintInBatch("\t");
			fprintf(stdout, "\n");
			return size;
		} else
		{
			printf("Center::Prepare Next Batch. Not ready: RunningJob = %p, DistAlg = %p\n",iRunningJob,iDistAlg);
		}
	}
	return 0;
}

string	Center::CreateJobId() const
{
	string val;
	val.assign(GetMachineName());
	val += '_';
	if (iDistAlg)
	{
		const char* c1 = iDistAlg->AlgorithmName();
		if (c1)
		{
			val += c1;
		} else
		{
			val += "NULL";
		}
	}
	if (iRunningJob)
	{
		TUint32 crc 	    = iRunningJob->Get_Crc();
		TInfoSource  source = iRunningJob->Get_InfoSource();
		TPriority    prior  = iRunningJob->Get_Priority();
		char  aux[128];
		memset(aux, 0, 128);
		const char* sourceStr = NULL;
		switch (source)
		{
			case ECommandLine:
				sourceStr = "CmdLine";
				break;
			case EFile:
				sourceStr = "File";
				break;
			default:
				sourceStr = "Unknown";
				break;
		}
		const char* priorStr = NULL;
		switch (prior)
		{
			case ELow:
				priorStr = "Low";
				break;
			case EStandard:
				priorStr = "Standard";
				break;
			case EAbsolute:
				priorStr = "Absolute";
				break;
			default:
				priorStr = "Unknown";
				break;
		}
		sprintf(aux,"%08x_%s_%s",crc,sourceStr, priorStr);
		val += aux;
		iRunningJob->Set_JobId(val);
	}
	return val;
}

int	Center::AvailableParsForNodes() const
{
	return iAvailableParametersForNodes.size();
}

bool	Center::IsSomeJobRunning() const
{
	return (iRunningJob != NULL);	
}

void Center::CheckRunningJob() {
	if (iRunningJob == NULL) return;

	if (iDistAlg == NULL) return;

	if (!(iDistAlg->IsComputationFinished())) return;

	for (unsigned i = 0; i < iJobsToDo.size(); i++) {
		if (iRunningJob == iJobsToDo[i]) {
			iCompletedJobs.push_back(iRunningJob);
			iJobsToDo.erase(iJobsToDo.begin()+i);
			iRunningJob = NULL;
			return;
		}
	}

	fprintf(stdout, "\n\tRunning job was not found in jobs to do!\n");

	return;
}

void	Center::SelectNextJobToRun()
{
	int i, max;
	iRunningJob = NULL; 
	if ((max = iJobsToDo.size()) == 0) 
	{ 		
		return;
	}
	for (i=0; i< max; i++)
	{
		JobInfo* info = iJobsToDo[i];
		if (!info->Is_Running())
		{
			if (info->Has_AbsolutePriority())
			{	
				iRunningJob = info; 			
				break; // the whole cycle.
			} else if (!iRunningJob)
			{
				iRunningJob = info;
				// Can still be superseded by a job with absolute priority.
			}
		}
	}
	if (iRunningJob)
	{		
		CreateJobId(); // Create Id for the job
		if (iDistAlg)
		{
			iDistAlg->ResetCenter();			
		}
		iRunningJob->Set_Running();
		iRunningJob->Set_StartedNow();
	}	
}

void	Center::Break()
{
	iBreakNow = TRUE;
}

void	Center::ProcessReceivedMessage(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return;
	}
	
	bool res = CheckMessageSequenceConsistency(aMessage);
	if (res)
	{
		Receiver::ProcessReceivedMessage(aMessage);
		UpdateCommunicationInfo(aMessage);		
		ProcessMessageByType(aMessage);
	} else
	{
		printf("Center: did not know how to process inconsistent situation.\n");
	}
}

bool	Center::ProcessMessageByType(AbstractMessage* aMessage)
{
	if (Receiver::ProcessMessageByType(aMessage))
	{
		// We still need to update the "last received" value in the appropriate NodeInfo.
		NodeInfo* info = FindNodeInfo(aMessage);
		if (info)
		{
			info->Update_LastReceived();
		}
		return TRUE;
	}
	if (!aMessage)
	{
		return FALSE;
	}
	switch (aMessage->Get_Type())
	{
		case EReady:
				return ProcessReadyMessage(aMessage);	
		case ELocked:
				return ProcessLockedMessage(aMessage);
		case ENodeBusy:
				return ProcessNodeBusyMessage(aMessage);
		case EAlive:
				return ProcessAliveMessage(aMessage);		
		case EData:
				return ProcessDataMessage(aMessage);
		default:        printf("Center: process received message by type: not fully implemented yet.");
				return FALSE;
				break;
	}
}

bool	Center::ProcessReadyMessage(AbstractMessage* aMessage)
{
// 	printf("Center: this is a ready message from node %s.\n",aMessage->Get_SenderId());				
	
	bool replyNoJob = TRUE;
	bool replyHasJob = FALSE;
	const vector<string>& algs = aMessage->Get_SupportedAlgorithms();
	int i, max = algs.size();
	vector<string> unsupported;
	vector<string> supported;
	for (i=0; i< max; i++)
	{
// 		printf("Center: node supports algorithm %s\n",algs[i].c_str());
		if (iDistAlg != NULL)
		{
			const char* algName = iDistAlg->AlgorithmName();
			if (algName)
			{
				if (!strcmp(algName,algs[i].c_str()))
				{
// 					printf("Center supports this algorithm.\n");
					string s2 = algs[i].c_str();
					supported.push_back(s2);					
				}
			} else // No name for the sieve
			{
				string s2 = algs[i].c_str();
				unsupported.push_back(s2);
			}
		}
		else // No alg instance available
		{
			string s2 = algs[i].c_str();
			unsupported.push_back(s2);
		}
	} // for
	if (supported.size() > 0)
	{
		/*
		int j, max = supported.size();
		printf("Center::Received Ready Message from a Node. Supported algs: ");
		for (j=0; j< max; j++)
		{
			cout << supported[j] << " ";
		}		
		cout << '\n';*/
		if (!AvailableParsForNodes())
		{
			PrepareNextBatch();			
		}
		// If we either had some un-assigned params
		// or PrepareNextBatch generated some ...
		if (AvailableParsForNodes())
		{
			replyNoJob = FALSE;
			replyHasJob = TRUE;
			//printf("Center::We can send some job parameters back to the Node\n");
		}

	}
	// Now we have processed the algorithms
	// This will, if necessary, create new NodeInfo and push it into iNodes vector
	bool	fresh = FALSE;
	NodeInfo* info = CreateOrFindNodeInfo(aMessage, fresh);
	if (fresh)
	{
		int ret;
		ret = info->CreateConnector(GetMachineName());
		if (CONNECTOR_NOT_READY == ret) {
			PrintErrorMessage("Connector not ready.",cerr);
			aMessage->MarkAsFinished(GetKeepMessages());
			DeleteNodeInfo(aMessage);
			return CONNECTOR_NOT_READY;
		}
	}
	info->Update_LastReceived();
	unsigned int replyCounter = aMessage->Get_Counter();
	if (replyNoJob)
	{
		Send_NoJobToNode(info,supported,unsupported,replyCounter,DEFAULT_ASK_AGAIN_IN_SECONDS);
		printf("Center::ProcessReadyMessage. Hard-coded reply time.\n");
	}
	if (replyHasJob)
	{
		Send_JobToNode(info, replyCounter);
	}	
	aMessage->MarkAsFinished(GetKeepMessages());
	return TRUE;
}

bool	Center::ProcessAliveMessage(AbstractMessage* aMessage)
{
	bool	fresh = FALSE;
	NodeInfo* info = CreateOrFindNodeInfo(aMessage, fresh);
	if (fresh)
	{
		// This is extremely strange!
		printf("Center::ProcessAliveMessage: somehow, we did not find the appropriate NodeInfo in iNodes!");
	}
	info->Update_LastReceived();	
	Send_AckToNode(info, aMessage->Get_Counter());
	aMessage->MarkAsFinished(GetKeepMessages());
	return TRUE;
}

bool	Center::ProcessDataMessage(AbstractMessage* aMessage)
{
	bool	fresh = FALSE;
	NodeInfo* info = CreateOrFindNodeInfo(aMessage, fresh);
	if (fresh)
	{
		// This is extremely strange!
		printf("Center::ProcessDataMessage: somehow, we did not find the appropriate NodeInfo in iNodes!");
	}
	info->Update_LastReceived();	
	info->IncrementDataReceived(aMessage->Get_DataCount());
	info->SetDataUnit(aMessage->Get_DataUnit());
	Send_AckToNode(info, aMessage->Get_Counter());

	// 
	string bodypath = aMessage->Get_BodyPath();
	//printf("Center::ProcessDatamessage. Body Path: %s\n",bodypath.c_str());
	iSerializator.RegisterInstance(iDistAlg);
	iSerializator.DeserializeData(bodypath.c_str());			
	aMessage->MarkAsFinished(GetKeepMessages());
	if (iDistAlg->IsDistributedPhaseFinished())
	{
		if (iAvailableParametersForNodes.size()) iAvailableParametersForNodes.clear();

		if (!(iDistAlg->IsComputationFinished())) {
			printf("Center::ProcessDataMessage: the distributed phase is finished!\n");
			iQuitReason.assign(ENOUGH_DATA_HARVESTED);
			// Send QUIT to all the nodes.
			vector<NodeInfo*>::iterator iter;
			NodeInfo* node = NULL;
			for (iter = iNodes.begin(); iter != iNodes.end(); iter++)
			{
				node = *iter;
				Send_QuitToNode(node);
			}
			printf("Center::ProcessDataMessage: now running the non-distributed phase.\n");
			iDistAlg->RunNondistributedPhase();
		}
	}
	return TRUE;
}

bool	Center::ProcessLockedMessage(AbstractMessage* aMessage)
{
	// There should be some NodeInfo in iNodes for this message.
	// If there is none, it is strange to say the least.
	bool	fresh = FALSE;
	NodeInfo* info = CreateOrFindNodeInfo(aMessage, fresh);
	if (fresh)
	{
		// This is extremely strange!
		printf("Center::ProcessLockedMessage: somehow, we did not find the appropriate NodeInfo in iNodes!");
	}
	info->Update_LastReceived();	
	// We will find the appropriate pending parameters and move them to
	// the 'assigned' vector.
	const char* senderId = aMessage->Get_SenderId();	
	
	vector<JobParameters*>::iterator iter;
	JobParameters* read = NULL;
	bool found = FALSE;
	for (iter = iPendingParametersForNodes.begin(); iter != iPendingParametersForNodes.end(); iter++)
	{
		read = *iter;
		string machineId = read->Get_MachineName();
		const char* mIdC = machineId.c_str();
		if (senderId && mIdC && !(strcmp(senderId, mIdC)))
		{
			// This is the right one
			found = TRUE;
			iPendingParametersForNodes.erase(iter);
			iAssignedParametersForNodes.push_back(read);
			break;
		}
	}
	if (!found)
	{
		// This is extremely strange!
		printf("Center::ProcessLockedMessage: somehow, we did not find the appropriate JobParameters in PendingParameters!");
	}
	else
	{
		Send_AckToNode(info, aMessage->Get_Counter());
	}
	info->Set_Locked();
	aMessage->MarkAsFinished(GetKeepMessages());
	return TRUE;
}

bool	Center::ProcessNodeBusyMessage(AbstractMessage* aMessage)
{
	// There should be some NodeInfo in iNodes for this message.
	// If there is none, it is strange to say the least.
	bool	fresh = FALSE;
	NodeInfo* info = CreateOrFindNodeInfo(aMessage, fresh);
	if (fresh)
	{
		// This is extremely strange!
		printf("Center::ProcessNodeBusyMessage: somehow, we did not find the appropriate NodeInfo in iNodes!");
	}
	info->Update_LastReceived();
	// We will find the appropriate pending parameters and move them to
	// the 'available' vector.
	const char* senderId = aMessage->Get_SenderId();	
	
	vector<JobParameters*>::iterator iter;
	JobParameters* read = NULL;
	bool found = FALSE;
	for (iter = iPendingParametersForNodes.begin(); iter != iPendingParametersForNodes.end(); iter++)
	{
		read = *iter;
		string machineId = read->Get_MachineName();
		const char* mIdC = machineId.c_str();
		if (senderId && mIdC && !(strcmp(senderId, mIdC)))
		{
			// This is the right one
			found = TRUE;
			iPendingParametersForNodes.erase(iter);
			iAvailableParametersForNodes.push_back(read);
			break;
		}
	}
	if (!found)
	{
		// This is extremely strange!
		printf("Center::ProcessNodeBusyMessage: somehow, we did not find the appropriate JobParameters in PendingParameters!");
	}
	else
	{
		Send_AckToNode(info, aMessage->Get_Counter());
	}
	// This node is not anymore important for us.
	DeleteNodeInfo(aMessage);
	aMessage->MarkAsFinished(GetKeepMessages());
	return TRUE;
}

void	Center::PrintAllNodes()
{
	vector<NodeInfo*>::const_iterator iter;
	NodeInfo* read = NULL;
	int size = iNodes.size();
	
	printf("\n\t%3d Nodes available%c\n",size, ( size ? ':' : '.'));
	
	for (iter = iNodes.begin(); iter != iNodes.end(); iter++)
	{
    		read = *iter;			
		read->PrintInfo();
		
	}

}

NodeInfo*	Center::FindNodeInfo(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		PrintErrorMessage("Center::FindNodeInfo. Message is NULL.",cerr);	
		return NULL;
	}
	const char* senderId = aMessage->Get_SenderId();	
	vector<NodeInfo*>::const_iterator iter;
	NodeInfo* read = NULL;
	for (iter = iNodes.begin(); iter != iNodes.end(); iter++)
	{
    		read = *iter;	
		const char* nodeId = read->GetId();
		if (senderId && nodeId && !strcmp(senderId, nodeId))
		{
			// This is the right node info
			
			return read;
		}
	}
	return NULL;
}

// This will try to find NodeInfo instance (in iNodes) for the given message,
// and if none is found, a new instance will be created.
NodeInfo*	Center::CreateOrFindNodeInfo(AbstractMessage* aMessage, bool& aFreshlyCreated)
{
	if (!aMessage)
	{
		PrintErrorMessage("Center::CreateOrFindNodeInfo. Message is NULL.",cerr);	
		return NULL;
	}
	NodeInfo* existingInfo = FindNodeInfo(aMessage);
	if (existingInfo)
	{
		aFreshlyCreated = FALSE;
		return existingInfo;
	}
	// No info found
	aFreshlyCreated = TRUE;
	NodeInfo* newInfo = new NodeInfo(aMessage);
	iNodes.push_back(newInfo);
	return newInfo;
}

// This will try to remove the instance of NodeInfo for this message from iNodes
void	Center::DeleteNodeInfo(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		PrintErrorMessage("Center::CreateOrFindNodeInfo. Message is NULL.",cerr);	
		return;
	}
	const char* senderId = aMessage->Get_SenderId();	
	vector<NodeInfo*>::iterator iter;
	NodeInfo* read = NULL;
	for (iter = iNodes.begin(); iter != iNodes.end(); iter++)
	{
    		read = *iter;	
		const char* nodeId = read->GetId();
		if (senderId && nodeId && !strcmp(senderId, nodeId))
		{
			// This is the right node info. We're gonna delete it now, yeah.
			iNodes.erase(iter);
			return;
		}
	}

}

void	Center::RemoveFromPending(const char* aTargetName, unsigned int aCounter)
{
	if (!aTargetName)
	{
	return;
	}
	int i, max = iNodes.size();
	NodeInfo* ptr = NULL;
	for (i=0; i< max; i++)
	{
		ptr = iNodes[i];
		if (ptr)
		{
			const char* nodeName = ptr->GetId();
			if (nodeName && (!strcmp(nodeName,aTargetName)))
			{
// 				printf("Center::RemoveFromPending. Found node name \"%s\" as required.\n",nodeName);
				if (ptr->HasConnector())
				{
					AbstractConnector& ctr = ptr->GetConnector();
					RemoveFromPending_Connector(&ctr,aTargetName,aCounter);
				}
			}
		}
	}

}

bool	Center::VerifyNodeInfoIntegrity(NodeInfo* aInfo)
{
	if (!aInfo)
	{
		PrintErrorMessage("Center::VerifyNodeInfoIntegrity. Node info is NULL.",cerr);
		return FALSE;
	}
	if (!aInfo->HasConnector())
	{
		PrintErrorMessage("Center::VerifyNodeInfoIntegrity. Node info has no connector.",cerr);
		return FALSE;
	}
	AbstractConnector& connector = aInfo->GetConnector();
	switch (connector.Get_ConnectorType())
	{
		case EFileConnector:
			break;
		default:
			PrintErrorMessage("Center::VerifyNodeInfoIntegrity. Connector not supported.",cerr);
			return FALSE;
	}
	return TRUE;
}

void	Center::Send_JobToNode(NodeInfo* aInfo, unsigned int aReplyCounter)
{
	if (!VerifyNodeInfoIntegrity(aInfo))
	{
		return;
	}
	AbstractConnector& connector = aInfo->GetConnector();
	URL*		   url = connector.Get_URL();
	AbstractMessage*  message = NULL;
	switch (connector.Get_ConnectorType())
	{
	case EFileConnector:
		{				
			if (url)
			{
				message = new FilesystemMessage(url->GetFullPath().c_str(),"Job_Parameters",EJobParameters, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",120);
				printf("Center::Send_JobToNode: hardcoded  as %d\n",120);
			}			
		}
		break;
	default:  PrintErrorMessage("Send_JobToNode: Connector not supported.",cerr);
	}
	if (message)
	{
		printf("Center::Send Job to Node. Available params count = %lu. Pending params count = %lu. Assigned params count = %lu\n",iAvailableParametersForNodes.size(), iPendingParametersForNodes.size(),iAssignedParametersForNodes.size());
		JobParameters* params = iAvailableParametersForNodes[0];
		//params->Print();
		params->Set_MachineName(aInfo->GetId());
		message->AddJobParameters(params, &iSerializator);
		message->Set_JobId(iRunningJob->Get_JobId());
		message->Set_ReplyTo(aReplyCounter);
		iAvailableParametersForNodes.erase(iAvailableParametersForNodes.begin());
		iPendingParametersForNodes.push_back(params);
		//params->Print();		
		connector.Send(message);
	}
}

void	Center::Send_QuitToNode(NodeInfo* aInfo)
{
	if (!VerifyNodeInfoIntegrity(aInfo))
	{
		return;
	}
	AbstractConnector& connector = aInfo->GetConnector();
	URL*		   url = connector.Get_URL();
	AbstractMessage*  message = NULL;
	switch (connector.Get_ConnectorType())
	{
	case EFileConnector:
		{				
			if (url)
			{
				message = new FilesystemMessage(url->GetFullPath().c_str(),"Quit",EQuit, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",120);
				printf("Center::Send_QuitToNode: hardcoded  as %d\n",120);
			}			
		}
		break;
	default:  PrintErrorMessage("Send_QuitToNode: Connector not supported.",cerr);
	}
	if (message)
	{
		message->Set_JobId(iRunningJob->Get_JobId());
		message->AddQuitReason(iQuitReason.c_str());
		//params->Print();		
		connector.Send(message);
	}
}


void	Center::Send_AckToNode(NodeInfo* aInfo, unsigned int aReplyCounter)
{
	Send_SimpleMessageToNode(aInfo, aReplyCounter, EAck, "Confirmation");
}

void	Center::Send_SimpleMessageToNode(NodeInfo* aInfo, unsigned int aInReplyTo, message_type aType, const char* aSubject)
{
	if (!VerifyNodeInfoIntegrity(aInfo))
	{
		return;
	}
	AbstractConnector& connector = aInfo->GetConnector();
	URL*		   url = connector.Get_URL();
	AbstractMessage*  message = NULL;
	switch (connector.Get_ConnectorType())
	{
	case EFileConnector:
		{				
			if (url)
			{
				message = new FilesystemMessage(url->GetFullPath().c_str(),aSubject,aType, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",120);
				printf("Center::Send_SimpleToNode: hardcoded timeout as %d\n",120);
			}			
		}
		break;
	default:  PrintErrorMessage("Send_QuitToNode: Connector not supported.",cerr);
	}
	if (message)
	{
		message->Set_ReplyTo(aInReplyTo);
		connector.Send(message);
	}
}



void	Center::Send_NoJobToNode(NodeInfo* aInfo, vector<string>& aSupported, vector<string>& aUnsupported, unsigned int aReplyCounter, unsigned int aTimeToAskAgain)
{
	if (!VerifyNodeInfoIntegrity(aInfo))
	{
		return;
	}
	AbstractConnector& connector = aInfo->GetConnector();
	URL*		   url = connector.Get_URL();
	AbstractMessage*  message = NULL;
	switch (connector.Get_ConnectorType())
	{
	case EFileConnector:
		{				
			if (url)
			{
				message = new FilesystemMessage(url->GetFullPath().c_str(),"Waiting",ENoJob, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",120);
				printf("Center::Send_NoJobToNode: hardcoded timeout as %d\n",120);
			}			
		}
		break;
	default:
		{
			PrintErrorMessage("Center::Send_NoJobToNode Cannot send message to a center with unknown connector type.",cerr);
		}
		break;
	}
	if (message != NULL)
	{
		int i, max = aSupported.size();
		for (i=0; i< max; i++)
		{
			message->AddSupportedAlgorithm(aSupported[i].c_str());		
		}
		max = aUnsupported.size();
		for (i=0; i< max; i++)
		{
			message->AddUnsupportedAlgorithm(aUnsupported[i].c_str());	
		}
		message->Set_ReplyTo(aReplyCounter);
		message->Set_AskAgain(aTimeToAskAgain);
		// The method "send" will take care of deleting the message eventually
		connector.Send(message);
// 		PrintInfoMessage(aInfo->GetId(),cout);
// 		PrintInfoMessage("Sent ENoJob over the file connector.",cout);
	}	
}

bool	Center::CheckMessageSequenceConsistency(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return FALSE;
	}
	bool isOk = Receiver::CheckMessageSequenceConsistency(aMessage);
	if (!isOk)
	{
		// Some mismatch in message sequence
		printf("Center::Message Sequence Consistency. Some kind of error, the newly received message has lower counter than some previous.\n");
		//aMessage->PrintHeaderInfo();
	}
	return isOk;
}

void	Center::ProcessCmdLineJobParameters(const JobParameters& aParameters)
{
	if (aParameters.Complete())
	{
		JobParameters* copy = aParameters.CreateCopy();
		string	      cmdline = "Command Line";
		JobInfo*      info = new JobInfo(copy,cmdline,ECommandLine,EAbsolute);
		AddJobInfo(info);
	} 
	else
	{
		printf("Center::ProcessCmdLineJobParameters. Command line does not specify a complete job.\n");
	}
}

void	Center::UpdateFromJobsInDir()
{
	fs::path full_path(iWorkingDirectoryName);

  	if ( !fs::exists( full_path ) )
  	{
#if BOOST_MINOR <= 45
    		std::cout << "\nCenter:: Jobs dir not found Not found: " << full_path.native_file_string() << std::endl;
#else
    		std::cout << "\nCenter:: Jobs dir not found Not found: " << full_path.string() << std::endl;
#endif
    		return;
  	}
// 	printf("Center: update Jobs from directory \"%s\"\n",iWorkingDirectoryName.c_str());
	// At first, we check the current job*.xml files in the directory
	// for recent additions and changes.
	if ( fs::is_directory( full_path ) )
  	{
		fs::directory_iterator end_iter;
    		for ( fs::directory_iterator dir_itr( full_path );         		dir_itr != end_iter; ++dir_itr )
    		{
      			try
      			{
				// We are interested in files only, not in dirs
				if ( !(fs::is_directory( *dir_itr ) ))
        			{
#if BOOST_MAJOR == 1
   #if BOOST_MINOR <= 34
					string filename = dir_itr->leaf();
   #elif BOOST_MINOR <= 45
					string filename = dir_itr->filename();
   #else
					string filename = dir_itr->path().filename().string();
   #endif
#else
   #error "Unknown boost version"
#endif
					
					if (StartsWithJob(filename) && EndsWithXmlExtension(filename) && !(IsResult(filename)))
					{
						Crc32 crc(filename);
						// This seems to be a job file
						string fullPath(dir_itr->path().root_path().string());
						fullPath.append(dir_itr->path().relative_path().string());				
						ProcessJobFromFile(fullPath,filename);
						// Now, we test whether this job file has been already processed.
						
					}
				}
      			}
      			catch ( const std::exception & ex )
      			{        			
#if BOOST_MINOR <= 45
        			std::cout << (*dir_itr).path().file_string() /*->path().leaf()*/ << " " << ex.what() << std::endl;
#else
        			std::cout << (*dir_itr).path().string() /*->path().leaf()*/ << " " << ex.what() << std::endl;
#endif
      			}
    		}
		// Then, we check the ToDo queue and Completed queue for missing files.
		vector<JobInfo*>::iterator iter;
		JobInfo* read = NULL;
		string fileName;
		fs::path fullPathWithName(full_path);
		for (iter = iJobsToDo.begin(); iter != iJobsToDo.end(); iter++)
		{
    			read = *iter;		
    			if ((read) && (read->Is_FileBased()))
			{
				fullPathWithName = full_path;
				fs::path fileName(read->Get_FileNameFull());
				fullPathWithName /= fileName;
// 				printf("Searching for existence of file %s\n",fullPathWithName.string().c_str());
				if (!fs::exists(fullPathWithName))
				{
					printf("Center: File \"%s\" with job ceased to exist. Will remove the job from the queue.\n",fullPathWithName.string().c_str());
					iJobsToDo.erase(iter);
					iter--;
				}
			}
		}
		for (iter = iCompletedJobs.begin(); iter != iCompletedJobs.end(); iter++)
		{
			read = *iter;
			if ((read) && (read->Is_FileBased()))
			{
				fullPathWithName = full_path;
				fs::path fileName(read->Get_FileNameFull());
				fullPathWithName /= fileName;
// 				printf("Searching for existence of file %s\n",fullPathWithName.string().c_str());
				if (!fs::exists(fullPathWithName))
				{
					printf("Center: File \"%s\" with job ceased to exist. Will remove the job from the queue.\n",fullPathWithName.string().c_str());
					iCompletedJobs.erase(iter);
					iter--;
				}
			}
		}
		
	}
}

bool 	Center::StartsWithJob(const string& aFilename) const
{
	return (aFilename.find("job") == 0);
}

bool	Center::EndsWithXmlExtension(const string& aFilename) const
{
	int size = aFilename.length();
	if (size < 6) return FALSE;
	int indof = aFilename.rfind(".xml");
// 	printf("EndsWithXMLExtension: %s, %d\n",aFilename.c_str(),indof);
	if (indof == (size - 4)) return TRUE;
	int indof2 = aFilename.rfind(".xml.z");
// 	printf("EndsWithXMLExtension: %s, %d\n",aFilename.c_str(),indof2);
	return (indof2 == (size - 6));
}

bool	Center::IsResult(const string& aFilename) const
{
	int size = aFilename.length();
	if (size < 11) return FALSE;
	int indof = aFilename.rfind("-result.xml");
	return (indof == (size - 11));
}

string  Center::ConstructResultName(const string& aFilename) const
{
	string newfilename(aFilename);
	if (aFilename.length() < 4)
	{
		newfilename = "job-result.xml";
	}
	else
	{
		newfilename.insert(aFilename.length()-4,"-result");
	}
	return newfilename;
}

void	Center::AddJobInfo(JobInfo* aInfo)
{
	if (!aInfo)
	{
		return;
	}	
	if (!aInfo->Get_Parameters().Complete())
	{
		return;
	}
	// Jobs with absolute priority will be inserted into the queue in a special way
	if (aInfo->Has_AbsolutePriority() && (iJobsToDo.size() > 0))
	{		
		JobInfo* read = NULL;
		vector<JobInfo*>::iterator iter;
		bool inserted = FALSE;
		for (iter = iJobsToDo.begin(); iter != iJobsToDo.end(); iter++)
		{
    			read = *iter;
			// The absolute priority job will get in front of this waiting job
    			if (read && read->Is_Waiting() && (!read->Has_AbsolutePriority()))
			{
				iJobsToDo.insert(iter,aInfo);
				inserted = TRUE;
				break;			
			}
		}
		if (inserted)
		{
			return;
		} else
		{
			// fall to the push_back branch
		}
	}
	iJobsToDo.push_back(aInfo);
}

bool	Center::ProcessJobFromFile(const string& aFullPathAndName, const string& aFileName)
{
	JobInfo* read = NULL;
	JobInfo* found = NULL;
	vector<JobInfo*>::iterator iter;
	vector<JobInfo*>::iterator foundWhere;
	bool foundInToDo = FALSE;
	// At first, we search through the toDO jobs.
// 	printf("Now looking for %s in ToDo jobs.\n",aFileName.c_str());
	for (iter = iJobsToDo.begin(); iter != iJobsToDo.end(); iter++)
	{
    		read = *iter;		
    		if (read)
		{
// 			printf("In ToDo jobs: read a job from filename %s\n",read->Get_FileNameFull().c_str());
			if (read->Is_FromFile(aFileName))
			{
				if (read->Is_Running())
				{
					//printf("Center::ProcessJobFromFile. This job is currently running.\n");
					// We will silently ignore any possible change
					// in a running job.
					return FALSE;
				} else if (read->Is_Waiting())
				{
					foundInToDo = TRUE;
					found = read;
					foundWhere = iter;
					break;
				} else 
				{
					printf("Center::ProcessJobFromFile. Unknown state of the job.\n");
				}
			}
		}
	}

	bool foundInCompleted = FALSE;	
	if (!foundInToDo)
	{	
		for (iter = iCompletedJobs.begin(); iter != iCompletedJobs.end(); iter++)
		{
    			read = *iter;		
    			if ((read) && (read->Is_FromFile(aFileName)))
			{
				if (!read->Is_Finished())
				{
					// Serious error!
					PrintErrorMessage("Center::ProcessJobFromFile. Unfinished job in CompletedJobs",cerr);
					return FALSE;
				}
				foundInCompleted = TRUE;
				foundWhere = iter;
				found = read;
				break;
			}
		}	
	}

	if ((foundInCompleted || foundInToDo) && (found))
	{
		// We will check the CRC
		TUint32 crcOld = found->Get_Crc();
		Crc32 crc(aFullPathAndName);
		TUint32 crcNew = crc.Count();
		if (crcOld != crcNew)
		{
			printf("\rCenter: Detected a change in job file \"%s\". CRC has changed. Reloading the job ...\n",aFullPathAndName.c_str());
			JobParameters* params = GetJobParametersFromFile(aFullPathAndName);			
			found->Set_Parameters(params);
			found->Set_Crc(crcNew);
			if (foundInCompleted)
			{
				iCompletedJobs.erase(foundWhere);
				AddJobInfo(found);
				//printf("Center::ProcessJobFromFile. Completed job was  put into the ToDo queue again\n");
			} 
			if (params->Complete())
			{
				printf("Center: Job reloaded OK.\n");
			} else
			{
				printf("Center: Job reloading failed, reason: %s\n",params->GetFailureReason().c_str());
			}
	
		}
		else
		{
			//printf("Center::ProcessJobFromFile. The file has been already processed before, and CRC has not changed\n");
		}
	} else
	{	
		printf("\rCenter: Detected a newly created job file \"%s\" ...\n",aFullPathAndName.c_str());
		JobParameters* params = GetJobParametersFromFile(aFullPathAndName);	
		if (params)
		{	
			Crc32 crc(aFullPathAndName);
			TUint32 crcNew = crc.Count();
			JobInfo* newJob = new JobInfo(params,aFileName, crcNew);
			AddJobInfo(newJob);			
			if (params->Complete())
			{
				printf("Center: Job loaded OK.\n");
				params->Print();
			} else
			{
				printf("Center: Job loading failed, reason: %s\n",params->GetFailureReason().c_str());
			}
			return TRUE;
		} else
		{
			return FALSE;
		}				
		
	}
	return FALSE;
}

JobParameters*  Center::GetJobParametersFromFile(const string& aFullPathAndName)
{
	xmlTextReaderPtr reader = xmlNewTextReaderFilename(aFullPathAndName.c_str());
	if (!reader)
	{
		printf("Center::GetJobParametersFromFile. Could not create XML TEXT Reader Ptr from  file %s\n",aFullPathAndName.c_str());
		return NULL;
	}
	JobParameters* params = iSerializator.CreateNewParameters();		
	if (params)
	{
		iSerializator.DeserializeJob(reader,*params);
	}
	xmlFreeTextReader(reader);
	return params;
}
