#ifndef __CENTER_H__
#define __CENTER_H__

class NodeInfo; 	 // Forward declaration
class AbstractMessage;	 // Forward declaration
class AbstractConnector; // Forward declaration
class ConnectionInfo;	 // Forward declaration

#include <vector>
#include "receiver.h"


using namespace std;

class AbstractSerializator;
class JobInfo;
class JobParameters;

/*!
 * \brief Center class for distributed computing
 */
class Center : public Receiver {

public:

	// the serializator serves as reader / writer of the parameters
	// working directory is the directory where job*.xml files are located. 
	// It is NOT the directory that is used to communicate with nodes over filesystem.
	// Once again: it is NOT the directory for communication with the nodes.
	Center(AbstractSerializator& aSerializator, const string& aWorkingDirectoryName);
	virtual ~Center();

	void	ProcessCmdLineJobParameters(const JobParameters& aParameters);	

	int	Run();
	void	Break();

protected:
	virtual void	ProcessReceivedMessage(AbstractMessage* aMessage);

	virtual void	RemoveFromPending(const char* aTargetName, unsigned int aCounter);

	virtual void	UpdateFromJobsInDir();

	virtual void	AddJobInfo(JobInfo* aInfo);

	/// This method will take a file and do the following:
	/// - try to find the JobInfo corresponding to this file in JobsToDo and CompletedJobs
	/// - if not found, JobInfo will be added to JobsToDo
	/// - if found, crc32 will be tested. 
	/// - if crc32 does not match, the file was altered
	///	- if the job has not been started yet, the JobInfo in JobsToDo will be
	///	  changed
	///	- if the job has been finished (is in CompletedJobs), the JobInfo will
	///	  be changed and moved to JobsToDo.	
	virtual bool	ProcessJobFromFile(const string& aFullPathAndName, const string& aFileName);

	JobParameters*  GetJobParametersFromFile(const string& aFullPathAndName);

protected:
	/// This method will register a node into the list of currently active
	/// nodes, using the content of the message supplied (EReady message).
	int	RegisterNode(AbstractMessage* aMessage);
	void	UnregisterNode(NodeInfo* aInfo);

	virtual bool	ProcessMessageByType(AbstractMessage* aMessage);

	bool	ProcessReadyMessage(AbstractMessage* aMessage);
	bool	ProcessLockedMessage(AbstractMessage* aMessage);
	bool	ProcessNodeBusyMessage(AbstractMessage* aMessage);
	bool	ProcessAliveMessage(AbstractMessage* aMessage);
	bool	ProcessDataMessage(AbstractMessage* aMessage);
	
	virtual bool	CheckMessageSequenceConsistency(AbstractMessage* aMessage);

	void	Send_JobToNode(NodeInfo* aInfo, unsigned int aReplyCounter);
	void	Send_NoJobToNode(NodeInfo* aInfo, vector<string>& aSupported, vector<string>& aUnsupported, unsigned int aReplyCounter, unsigned int aTimeToAskAgain);
	
	void	Send_QuitToNode(NodeInfo* aInfo);
	
	void	Send_AckToNode(NodeInfo* aInfo, unsigned int aReplyCounter);
	// The execution method for Send_Ack and other simple messages
	void	Send_SimpleMessageToNode(NodeInfo* aInfo, unsigned int aInReplyTo, message_type aType, const char* aSubject);

	




	// Checks whether this info is non-NULL, has a connector and the
	// connector is of supported type.
	bool	VerifyNodeInfoIntegrity(NodeInfo* aInfo);

	bool 	StartsWithJob(const string& aFilename) const;
	// Returns True if the file ends with .xml or .xml.z extensions
	bool	EndsWithXmlExtension(const string& aFilename) const;
	bool	IsResult(const string& aFilename) const;
	string  ConstructResultName(const string& aFilename) const;

	bool		IsSomeJobRunning() const;
	void		SelectNextJobToRun();
	void		PrepareFirstBatch();
	int		PrepareNextBatch();
	// How many parameters for Nodes are still available?
	int		AvailableParsForNodes() const;

	// This will try to find NodeInfo instance (in iNodes) for the given message,
	// and if none is found, a new instance will be created.
	// The status of creation is given by aFreshlyCreated
	NodeInfo*	CreateOrFindNodeInfo(AbstractMessage* aMessage, bool& aFreshlyCreated);
	// Will try to find NodeInfo in iNodes, but does not create new if unsuccessful.
	NodeInfo*	FindNodeInfo(AbstractMessage* aMessage);
	// This will try to remove the instance of NodeInfo for this message from iNodes
	void		DeleteNodeInfo(AbstractMessage* aMessage);

	void		PrintAllNodes();

	string		CreateJobId() const;

	void CheckRunningJob();

protected:

	vector<NodeInfo*>	iNodes; ///< An array of currently connected nodes.

	vector<JobInfo*>	iJobsToDo;  ///< An array of jobs loaded from the command line or from XML files that are waiting to be processed.

	mutable JobInfo*	iRunningJob;

	vector<JobInfo*>	iCompletedJobs;  ///< An array of jobs loaded from the command line or from XML files that were already completed.

	/// The parameters that haven't been yet assigned to individual nodes.
	vector<JobParameters*>  iAvailableParametersForNodes;
	vector<JobParameters*>	iPendingParametersForNodes;
	vector<JobParameters*>	iAssignedParametersForNodes;
		
	bool			iBreakNow;

	unsigned int    	iCounter;

	string			iWorkingDirectoryName;

	string			iQuitReason;
};

#endif

