
#include "center_info.h"
#include <stdlib.h>
#include <string.h>
#include "abstract_sieve.h"
#include <iostream>
#include "common_xml_helper.h"
#include "xml_message_definitions.h"

#include "log.h"

CenterInfo::CenterInfo(const char* aId, connector_type aConnection, const char* aAddress, unsigned int aTimeout) : ConnectorInfo(aAddress,aConnection)
{
	iOwnsConnector = TRUE;
 	iConnector = NULL;
	SetId(aId);
	SetTimeout(aTimeout);
}

CenterInfo::CenterInfo(string& aId, connector_type aConnection, string& aAddress, unsigned int aTimeout) : ConnectorInfo(aAddress,aConnection)
{
	iOwnsConnector = TRUE;
	iConnector = NULL;
	iId.assign(aId);	
	SetTimeout(aTimeout);
}


CenterInfo::~CenterInfo()
{	
}

const char*	CenterInfo::GetId() 	const
{
	return iId.c_str();
}


unsigned int	CenterInfo::GetTimeout() 	const
{
	return iTimeout;
}
	
void	CenterInfo::SetId(const char* aId)
{
	if (!aId)
	{
		iId.assign("");
		return;
	}
	iId.assign(aId);
}


void	CenterInfo::SetTimeout(unsigned int aTimeout)
{
	if (aTimeout >= MINIMAL_TIMEOUT_IN_SECONDS)
	{
		iTimeout = aTimeout;
	} else
	{
		CLog::LogError(WHEREARG,"CenterInfo","SetTimeout",aTimeout,MINIMAL_TIMEOUT_IN_SECONDS,"The timeout is too low, setting to minimal possible.");
		iTimeout = MINIMAL_TIMEOUT_IN_SECONDS;
	}	
}

void	CenterInfo::Print() const
{
	cout << "------------------------ \n";
	cout << "Center Info:\n";
	cout << "  Id:      " << iId << '\n';
	ConnectorInfo::Print(2);
	cout << "  Timeout: " << iTimeout << " sec\n";	
	if (iFailureReason.size() > 0)
	{
	cout << "  Failure Reason:  " << iFailureReason << "\n";
	}
}

bool		CenterInfo::IsEqual(const CenterInfo* aOtherInstance) const
{
	bool isEqualConnection = ConnectorInfo::IsEqual(aOtherInstance);
	if (!isEqualConnection)
	{
		return FALSE;
	}
	// If connections are identical, we will check for IDs. If Ids are
	// identical, the center infos are identical (timeout may vary)
	int sameIds = iId.compare(aOtherInstance->iId);
	if (sameIds == 0) return TRUE; // 0 ... equal strings

	return FALSE;
}

CenterInfo*	CenterInfo::Parse(const char* info)
{
	string str(info);
	// At first, check for correct formatting
	// Zero'th char must be '['
	// last char must be ']'
	int len = str.length();
	if (len < 6)
	{
		printf("CenterInfo.Parse: Center info must be longer than 5 chars!");
		return NULL;
	}
	if (str[0] != '[')
	{
		printf("CenterInfo.Parse: Center info must start with '[' char!");
		return NULL;
	}
	if (str[len-1] != ']')
	{
		printf("CenterInfo.Parse: Center info must end with ']' char!");
		return NULL;
	}
	str.erase(len-1,1); // removal of the last ']'
	str.erase(0,1);	// removal of the first '['
	
	// At first, we detect the id of the center
	unsigned int pos1 = str.find(',',0);
	if (pos1 == string::npos)
	{
		printf("CenterInfo.Parse: Not enough , separators!");
		return NULL;
	}
	string id(str,0,pos1);
// 	printf("Id read as %s\n",id.c_str());
	if (id.empty())
	{
		printf("CenterInfo.Parse: Center info must have nonempty id!");
		return NULL;
	}
	str.erase(0,pos1+1); // the id and the , removed

	// Then, we detect the preferred connection
	pos1 = str.find(',',0);
	if (pos1 == string::npos)
	{
		printf("CenterInfo.Parse: Not enough , separators!");
		return NULL;
	}
	string connection(str,0,pos1);
// 	printf("Connection read as %s\n",connection.c_str());
	if (connection.empty())
	{
		printf("CenterInfo.Parse: Center info must have nonempty connection!");
		return NULL;
	}
	connector_type ct = EFileConnector;
	int conn = atoi (connection.c_str());
	if (conn == 0)
	{
		// The "connection" string was not a number.
		// We will try to parse it.
		bool conv = ParseConnectorType(connection,ct);
		// Error will be printed by the ParseConnectionType call
		if (!conv)
		{		
			return NULL;
		}
	} 
        else 
	{
		// outside legal values for connection
		if ((conn < 0)  || (conn >= ESupremum))
		{
			PrintAllowedConnTypes();
			return NULL;
		} else
		{
			// A valid cast
			ct = static_cast<connector_type>(conn);
		}	
		
	}
	str.erase(0,pos1+1);	// the connection and the , removed
	
	// Then, we detect the address
	pos1 = str.find(',',0);
	if (pos1 == string::npos)
	{
		printf("CenterInfo.Parse: Not enough , separators!");
		return NULL;
	}
	string address(str,0,pos1);
// 	printf("Address read as %s\n",address.c_str());
	if (address.empty())
	{
		printf("CenterInfo.Parse: Center info must have nonempty address!");
		return NULL;
	}
	str.erase(0,pos1+1);	// the address and the , removed

	// Then, we detect the timeout
	pos1 = str.find(',',0);
	string timeoutstr;
	if (pos1 == string::npos)
	{
		timeoutstr.assign(str);
	}
	else {
		timeoutstr.assign(str,0,pos1);
	}
	unsigned int timeout = atoi (timeoutstr.c_str());
	if (timeout == 0)
	{
		printf("CenterInfo.Parse: Invalid timeout value: %s !",timeoutstr.c_str());
		return NULL;
	}
	
	CenterInfo* ctr = new CenterInfo(id,ct,address,timeout);
	return ctr;
	
}


CenterInfo*	CenterInfo::Deserialize(xmlTextReaderPtr& aReader)
{
	printf("CenterInfo::Deserialize. Center Info has not yet been fully migrated into the new framework for connectors and addresses. Namely, it should support the \"address or connector\" mode in the future.\n");
	CenterInfo* target = new CenterInfo("",EFileConnector,"",120);
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"CenterInfo","Deserialize",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		delete target;
			return NULL;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"CenterInfo","Deserialize",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		delete target;
			return NULL;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
// 		printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{		
			
			if (xmlStrEqual(name,BAD_CAST CONNECTOR_NAME))
			{				
				target->ConnectorInfo::Deserialize(aReader);
			}
			if (xmlStrEqual(name,BAD_CAST ID_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"Center id read as ",target->iId,FALSE);					
			}
			if (xmlStrEqual(name,BAD_CAST TIMEOUT_NAME))
			{
				CommonXMLHelper::DeserializeUnsignedInt(aReader,"Timeout",target->iTimeout,FALSE);
			}
		} // if (type == 1)
	} // end of while(TRUE)
	return target;

}

