#include "cfrac_class.h"

#include "const_return_codes.h"
#include "integer_class.h"

/*!
  	The default constructor sets maximal factor base length to 256
*/

CFRAC::CFRAC(long aUpperFBBound) : ASupportingFactorAlg()
{
	Set_AlgorithmType(EFACFRAC);

	int i=0;
        mpz_init(this->kN);
        mpz_init(this->aux2);
        mpz_init(this->An);
        mpz_init(this->An1);
        mpz_init(this->An2);
        mpz_init(this->AS);
        mpz_init(this->R);
        mpz_init(this->N);
        mpz_init(this->Q0);
	this->prime_base = (long *)malloc(sizeof(long)*1024);
        this->A = (mpz_t *)malloc(sizeof(mpz_t)*1024);
	for(i=0;i<128;i++) mpz_init(this->A[i]);
        this->Q = (unsigned long *)malloc(sizeof(long)*1024);
        this->smallB = new BitMatrix(256,256);
	this->smallB->Allocate();
        this->smallC = new BitMatrix(256,256);
	this->smallC->Allocate();
        this->bigB = new BitMatrix(1024,1024);
	this->bigB->Allocate();
        this->bigC = new BitMatrix(1024,1024);
	this->bigC->Allocate();
	prime_base[0]=-1;
	prime_base[1]=2;
	prime_base[2]=3;
	this->max_fb_length=256;
	this->small_primes = NULL;
	unsigned long count;
	CIntegerClass::GetPrimes(1, aUpperFBBound, count, this->small_primes);
}

CFRAC::~CFRAC(){
	int i;
	mpz_clear(this->kN);
	mpz_clear(this->aux2);
	mpz_clear(this->An);
	mpz_clear(this->An1);
	mpz_clear(this->An2);
	mpz_clear(this->AS);
	mpz_clear(this->R);
	mpz_clear(this->N);
	mpz_clear(this->Q0);
	free(this->prime_base);
	for(i=0;i<=1024;i++) mpz_clear(this->A[i]);
	free(this->A);
	free(this->Q);
	free(this->B);
	free(this->C);
}

/*!
  	Tests whether all allocations were successful
*/

int CFRAC::TestAllocation(){
	if (!Q) return ConstRC::GeneralError;
	if (smallB->GetMaxAllocatedRowIndex()==-1) return ConstRC::GeneralError;
	if (bigB->GetMaxAllocatedRowIndex()==-1) return ConstRC::GeneralError;
	if (smallC->GetMaxAllocatedRowIndex()==-1) return ConstRC::GeneralError;
	if (bigC->GetMaxAllocatedRowIndex()==-1) return ConstRC::GeneralError;
	return ConstRC::Ok;
}

/*!
  	Gets the maximal factor base length
*/

unsigned int CFRAC::GetMaxFBLength(){
	return this->max_fb_length;
}

/*!
  	Sets the maximal factor base length
*/
int  CFRAC::SetMaxFBLength(unsigned int aLength){
	if (aLength<3){
		this->max_fb_length=3;
	}
		else{
		if (aLength>1024) this->max_fb_length=1024;
			else this->max_fb_length=aLength;
	}
	return ConstRC::Ok;
}

/*!
  	Computes the continued fractions until the bound is reached or a square is found
*/

bool CFRAC::ContinuedFractions(unsigned int aBound){
	bool bit;
	unsigned int j;
	unsigned long aux;
	for(;rel<aBound;n++){
		this->rn1 = this->rn;
//		rn=gPn;
//		for(qn=0;(qn<4)&&(rn>=Qn);qn++) rn-=Qn;
//		printf("%d = %d %d + %d\n",gPn,qn,Qn,rn);
//		if (rn>Qn){
			this->qn = this->gPn / this->Qn;
			this->rn = this->gPn - this->Qn * this->qn;
//		}
		mpz_set(this->An2,this->An1);
		mpz_set(this->An1,this->An);
		mpz_mul_ui(this->An,this->An,this->qn);
		mpz_add(this->An,this->An,this->An2);
		if (qn<5){
			while(mpz_cmp(this->N,this->An)<=0) mpz_sub(this->An,this->An,this->N);
		} else {
			mpz_fdiv_r(this->An,this->An,this->N);
		}
		//gmp_printf("%2u: %6u  %6u  %6u  %6u  %11Zu\n",n,gPn,Qn,qn,rn,An1);fflush(stdout);
		aux=this->Qn;
		bit=0;
		while(!(aux & 1)){
			bit=!bit;
			aux=aux>>1;
		}
		if (bit) this->B->PutOne(this->rel,1);
		for(j=2;j<this->fb_length;j++){
			bit=0;
			while(!(aux % this->prime_base[j])){
				aux/=this->prime_base[j];
				bit=!bit;
			}
			if (bit) this->B->PutOne(this->rel,j);
		}
		if (aux>1) {
			this->B->ZeroizeRow(this->rel);
			bit=0;
		}
		else{
			//for(j=0;j<fb_length;j++) printf("%d",B->IsOne(this->rel,j));
			if (this->n % 2) this->B->PutOne(this->rel,0);
			mpz_set(this->A[this->rel],this->An1);
			this->Q[this->rel]=this->Qn;
			this->rel++;
			bit=1;
		}
		this->gPn = (this->g<<1) - this->rn;
		this->Qn2 = this->Qn1;
		this->Qn1 = this->Qn;
		this->Qn = this->Qn2 + this->qn * (this->rn - this->rn1);
	//	if ((bit)&&(this->B->IsZeroRow(this->rel-1))) return TRUE;
	}
	return false;
}

/*!
	Expands a relation into a congruence Q<SUP>2</SUP>=A<SUP>2</SUP> mod N and tries
	to compute a factor
*/

unsigned long CFRAC::ExpandRelation(unsigned int aRow){
	unsigned int j;
	mpz_set_ui(this->AS,1);
//        mpz_set_ui(this->Qb,1);
        mpz_set_ui(this->R,1);
        for(j=0;j<this->rel;j++){
        	if(C->IsOne(aRow,j)) {
//			gmp_printf("%d %Zd\n",Q[j],R);fflush(stdout);
                	mpz_mul(this->AS,this->AS,this->A[j]);
                	mpz_fdiv_r(this->AS,this->AS,this->N);
                	mpz_mul_si(this->R,this->R,this->Q[j]);
//			mpz_set_ui(this->aux2,this->Q[j]);
//			mpz_gcd(this->X,this->R,this->aux2);
//			mpz_mul(this->Qb,this->X,this->Qb);
//			while (mpz_cmp(this->N,this->Qb)<=0) mpz_sub(this->Qb,this->Qb,this->N);
//			mpz_fdiv_q(this->R,this->R,this->X);
//			mpz_fdiv_q(this->aux2,this->aux2,this->X);
//			mpz_mul(this->R,this->R,this->aux2);
        	}
        }
//	mpz_sqrt(this->X,this->R);
//	mpz_mul(this->Qb,this->Qb,this->X);
//	while (mpz_cmp(this->N,this->Qb)<=0) mpz_sub(this->Qb,this->Qb,this->N);
        mpz_sqrt(this->aux2,this->R);
        mpz_fdiv_r(this->aux2,this->aux2,this->N);
        mpz_sub(this->aux2,this->AS,this->aux2);
        mpz_gcd(this->R,this->aux2,this->N);
//	gmp_printf("%d: %Zd - %Zd = ? .... %Zd\n",aRow,AS,aux2,R);fflush(stdout);
        if((mpz_cmp_ui(this->R,1)>0)&&(mpz_cmp(this->R,this->N)<0)){
        	mpz_cdiv_q(this->aux2,this->N,this->R);
        	return(mpz_get_ui(this->R));
	} else return 0;
}

/*!
  	Computes a small coefficient k and performs CFRAC for kN. If a factor is not found
	for a factor base, increases its lenght.
*/

unsigned long CFRAC::Factorize(mpz_t aNumber){
	unsigned int i,h;
	unsigned long aux;
	k=mpz_fdiv_ui(aNumber,4);
	mpz_mul_ui(this->kN,aNumber,k);
	mpz_set(this->N,aNumber);
	mpz_sqrtrem(this->An,this->Q0,this->kN);
//	gmp_printf("%Zd = %Zd^2 + %Zd\n",kN,An,Q0);
	Qn=mpz_get_ui(this->Q0);
	this->g=mpz_get_ui(this->An);
	if (!Qn) return(g);
	this->fb_length=0;
	for(aux=this->g;aux;aux=aux>>1) this->fb_length++;
	this->fb_length*=2;
	this->fb_length-=3;
	i=3;
	h=1;
	while(fb_length <= this->max_fb_length){
	if(fb_length<=256){
		this->B=this->smallB;
		this->C=this->smallC;
	} else {
		this->B=this->bigB;
		this->C=this->bigC;
	}
	Qn=mpz_get_ui(this->Q0);
	for(;i<this->fb_length;i++){
		aux=0;
		while(!aux){
			mpz_set_ui(this->aux2,small_primes[++h]);
			aux=mpz_jacobi(this->kN,this->aux2);	
		}
		this->prime_base[i]=small_primes[h];
//		printf("%d ",prime_base[i]);fflush(stdout);
	}
//	printf(" %d\n",fb_length);fflush(stdout);
	this->Qn1=1;
	mpz_set_ui(this->An1,1);
	this->gPn=this->g<<1;
	this->qn=this->g;
	this->rn=0;
	this->rel=0;
	this->n=1;
	this->B->Zeroize();
	this->C->Zeroize();
	for(i=0;(signed) i<=this->C->GetMaximalRowIndex();i++) this->C->PutOne(i,i);
	while(rel<fb_length){
		if(n==1){
			while (this->rel<this->fb_length/2) {
				if (this->ContinuedFractions(this->fb_length/2)){
					aux=ExpandRelation(rel);
					if (aux){
						return aux;
					} else rel--;
				}	
			}
		}
		else {
			while (this->rel<this->fb_length) {
				if (this->ContinuedFractions(this->fb_length)){
					aux=ExpandRelation(rel);
					if (aux){
						return aux;
					} else rel--;
				}
			}
		}
		//B->PrintToScreen();
		this->B->ToSemidiagonalShapeForCFRAC(this->C,rel,fb_length);
		//C->PrintToScreen();
		//B->PrintToScreen();
		for(i=this->rel-1;i>0;i--){
			if(this->B->IsZeroRow(i)){
				//for(j=0;j<fb_length;j++) printf("%d ",C->IsOne(i,j));
				aux=ExpandRelation(i);
				if (aux) {
					return aux;
				}
			}
		}
	h++;
	}
	i=fb_length;
	this->fb_length*=2;
	mpz_set_ui(An,g);
	
	}
	return 0;
}

/*!
  	Constructor with preset maximal factor base length
*/

CFRAC::CFRAC(long aUpperFBBound, unsigned int aMax)
{
	Set_AlgorithmType(EFACFRAC);

	int i=0;
        mpz_init(this->kN);
        mpz_init(this->aux2);
        mpz_init(this->An);
        mpz_init(this->An1);
        mpz_init(this->An2);
        mpz_init(this->AS);
        mpz_init(this->R);
        mpz_init(this->N);
        mpz_init(this->Q0);
	this->prime_base = (long *)malloc(sizeof(long)*1024);
        this->A = (mpz_t *)malloc(sizeof(mpz_t)*1024);
	for(i=0;i<128;i++) mpz_init(this->A[i]);
        this->Q = (unsigned long *)malloc(sizeof(long)*1024);
        this->smallB = new BitMatrix(256,256);
	this->smallB->Allocate();
        this->smallC = new BitMatrix(256,256);
	this->smallC->Allocate();
        this->bigB = new BitMatrix(1024,1024);
	this->bigB->Allocate();
        this->bigC = new BitMatrix(1024,1024);
	this->bigC->Allocate();
	prime_base[0]=-1;
	prime_base[1]=2;
	prime_base[2]=3;
	this->max_fb_length=aMax;
	this->small_primes = NULL;
	unsigned long count;
	CIntegerClass::GetPrimes(1, aUpperFBBound, count, this->small_primes);
}

/*!
 An overloaded method from FactoringAlgorithm, providing interface for factorization.
*/

int CFRAC::Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) {
	mpz_set_ui(aFactor1,this->Factorize(aModulus));
	if (mpz_cmp_ui(aFactor1,0)==0) return ConstRC::NotFactorized;
	mpz_divexact(aFactor2,aModulus,aFactor1);
	return ConstRC::Ok;
}

