#ifndef __CFRAC_CLASS_H_
#define __CFRAC_CLASS_H_

#include "abstract_supporting_factor_alg.h"

#include "bit_matrix_class.h"

/*!
	\brief This class implements the Continued Fractions (CFRAC) factoring algorithm

	This class implements the Continued Fractions (CFRAC) factoring algorithm for the purpose
	of the LPV. It supposes that the factored number has two factors and therefore both of the factors have to
	fit into long int. The description of the algorithm is in
	M. A. Morrison, J. Brillhart: <I> A method of factoring and the factorisation of F</I><SUB>7</SUB>,
	Math. Comp., vol. 29, no. 129, (1975), 183-205
*/

class CFRAC : public ASupportingFactorAlg
{
public:
	CFRAC(long aUpperFBBound);
	CFRAC(long aUpperFBBound, unsigned int aMax);
	~CFRAC();
	int Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2);
	
	unsigned long  Factorize(mpz_t aNumber);
	int  TestAllocation();
	unsigned int   GetMaxFBLength();
	int  SetMaxFBLength(unsigned int aLength);
	
protected:
	bool ContinuedFractions(unsigned int aBound);
	unsigned long ExpandRelation(unsigned int aRow);

	unsigned long gPn;///<	The integer g+P<SUB>n</SUB>
	unsigned long gPn1;///<	The integer g+P<SUB>n-1</SUB>
	unsigned long Qn;///<	The integer Q<SUB>n</SUB>
	unsigned long Qn1;///<	The integer Q<SUB>n-1</SUB>
	unsigned long Qn2;///<	The integer Q<SUB>n-2</SUB>
	unsigned long qn;///<	The integer q<SUB>n</SUB>
	unsigned long rn;///<	The integer r<SUB>n</SUB>
	unsigned long rn1;///<	The integer r<SUB>n-1</SUB>
	unsigned long g;///<	The integer g
	unsigned long *Q;///<	The array of factored Q<SUB>n</SUB>
        unsigned int k;///<	The small coefficient k
	unsigned long fb_length;///<	The actual factor base size
	unsigned long n;///<	The counter n
	unsigned long rel;///<	The number of relations found
	unsigned long max_fb_length;///<	The maximal possible factor base size
	long*	small_primes;///<The array of small primes
    long *prime_base;///<	The factor base
    BitMatrix *bigB;///<	The big version of the matrix of relations
	BitMatrix *bigC;///<	The big version of the matrix of row dependencies
	BitMatrix *smallB;///<	The small version of the matrix of relations
	BitMatrix *smallC;///<	The small version of the matrix of row dependencies
	BitMatrix *B;///<	The currently used version of the matrix of relations
	BitMatrix *C;///<	The currently used version of the matrix of row dependencies
        mpz_t aux2;///<		An auxiliary variable
	mpz_t An;///<	The integer A<SUB>n</SUB>
	mpz_t An1;///<	The integer A<SUB>n-1</SUB>
	mpz_t An2;///<	The integer A<SUB>n-2</SUB>
	mpz_t kN;///<	The number kN
	mpz_t R;///<	The multiple of all the chosen Q<SUB>n</SUB>
	mpz_t AS;///<	The multiple of all the chosen A<SUB>n</SUB>
	mpz_t *A;///<	The array of all A<SUB>n</SUB>
	mpz_t N;///<	The number N
	mpz_t Q0;///<	The integer Q<sub>0</SUB>
};

#endif // __CFRAC_CLASS_H_
