#include "common_utils.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "polynom_class.h"
#include <sstream>
#include <fstream>
#include "const_return_codes.h"
#include "log.h"
#include "const_hashtables.h"


// for creating directory - portable way (WIN and UNIX) 
#ifdef WIN32
#include <direct.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#endif

using namespace std;

void Utils::Wait()
{
	char i;
	printf("\n\nPress Enter\n\n");
	scanf("%c",&i);
}

/*!
	  This is a static utility to print an error message on screen. The error message is composed of information stored in the argument \a aContext.

*/
void Utils::ErrorMessage(error_context* aContext) {
	printf("\n");
	if (aContext->process != NULL){
		printf("Process %s: ",aContext->process);
	} else {
		printf("Unknown process: ");
	}
	if (aContext->problem != NULL){
		printf("had following problem: %s. ",aContext->problem);
	} else {
		printf("had unknown problem. ");
	}
	printf("Error code was %d, error label was %d ",aContext->errorCode,aContext->label);
	if (aContext->sourceFile != NULL){
		printf("(file %s). ",aContext->sourceFile);
	} else printf(" (unknown file).");
	if (aContext->explanation != NULL){
		printf("%s.\n",aContext->explanation);
	}
}

/*!
	  This is a static utility to print an error message on screen. The error message is composed of information stored in the arguments.

*/
void Utils::ErrorMessage(const char* aProcess, const char* aProblem, int aErrorCode, int aLabel, const char* aSourceFile, const char* aExplanation){
	printf("\n");
	if (aProcess != NULL){
		printf("Process %s: ",aProcess);
	} else {
		printf("Unknown process: ");
	}
	if (aProblem != NULL){
		printf("had following problem: %s. ",aProblem);
	} else {
		printf("had unknown problem. ");
	}
	printf("Error code was %d, error label was %d ",aErrorCode,aLabel);
	if (aSourceFile != NULL){
		printf("(file %s). ",aSourceFile);
	} else printf(" (unknown file).");
	if (aExplanation != NULL){
		printf("%s.\n",aExplanation);
	}
}

/*!
	  This is a static utility to print an error message on screen. The error message is composed of information stored in the arguments. Frequently called using #WHEREARG macro in place of the first two parameters.

*/
void Utils::ErrorMessage(int aLineNumber, const char* aSourceFile, const char* aProcess, const char* aProblem, int aErrorCode, int aLabel, const char* aExplanation){
	printf("\n");
	if (aProcess != NULL)
	{
		printf("Process %s: ",aProcess);
	} 
	else 
	{
		printf("Unknown process: ");
	}
	if (aProblem != NULL)
	{
		printf("had following problem: %s. ",aProblem);
	} 
	else 
	{
		printf("had unknown problem. ");
	}
	printf("Error code was %d, error label was %d ",aErrorCode,aLabel);
	if (aSourceFile != NULL)
	{
		printf("(file %s). ",aSourceFile);
	} 
	else printf(" (unknown file).");
	
	if (aExplanation != NULL)
	{
		printf(" %s.",aExplanation);
	}

	printf(" Source line number: %d.\n",aLineNumber);
}



/*!
	  This is a static utility to print a message on screen. This message has an optional number. If \a aNumber is specified as \a NO_NUMBER symbolic constant, no number is printed.

*/
void Utils::NormalMessage(const char* aMessage, int aNumber)
{
	if (aMessage != NULL) 
	{
		printf("\n");
		printf("%s",aMessage);
		if (aNumber != NO_NUMBER)
		{
			printf(" Number: %d.", aNumber);
		}
		printf("\n");
		fflush(stdout);
	}
}
/*!
	  A static utility, which gets number of seconds, recalculates it into days, hours and minutes, and prints out the result. Useful when
	  printing out information about duration of processes.	

*/
void Utils::TimeFormat(int aSeconds)
{	
	int days  = aSeconds / 86400;
	aSeconds -= days*86400;
	int hours = aSeconds / 3600;
	aSeconds -= hours*3600;	
	int minutes = aSeconds/60;
	aSeconds -= minutes*60;	
	printf("%0d d %0d h %0d m %0d sec.", days,hours,minutes,aSeconds);
}
/*!
	  This is a static utility ensuring proper allocation or reallocation of the given pointer. The pointer is double-starred, since otherwise
	  the memory allocation would influence only a local copy of the pointer. If allocation or reallocation was unsuccessful, the machine 
	  returns \a NULL.
	  AllocationMachine is the central point of memory management in MPQS/SIQS. All calls for memory allocation, except for mpz_init(), should  go through this method.

*/
void* Utils::AllocationMachine(void** aPointer, long aSize)
{
	void* poi;
	if (*aPointer == NULL)
	{
		poi = malloc(aSize);
		return(poi);
	} 
	else 
	{
		poi = *aPointer;
		poi = realloc(poi, aSize);
		return(poi);
	}
}
/*!
	  A variant of the preceding utility, which in case of failure prints out an error message with given context.

*/
void* Utils::AllocationMachine(void** aPointer, long aSize, error_context* aContext)
{
	void* poi;
	if (*aPointer == NULL)
	{
		poi = malloc(aSize);
	} 
	else 
	{
		poi = *aPointer;
		poi = realloc(poi, aSize);
	}
	if (poi == NULL) 
	{
		ErrorMessage(aContext);
		return NULL;
	}
	else 
	{
		return (poi);
	}
}

string Utils::GetDirectoryFromPath(const char *aPath)
{
	if (aPath == NULL)
	{
		return "";
	}

	string str(aPath);
	size_t index = str.find_last_of(PATH_SEPARATOR);
	if (index == str.npos)
	{
		return "";
	}

	return str.substr(0, index);
}

/*!
	Returns the natural logarithm of aArg.
*/
double Utils::Ln(mpf_t aArg)
{
	long e;
	double d = mpf_get_d_2exp(&e, aArg);
	return log((double) fabs(d))+e*log((double) 2);
}
double Utils::Round(double aArg)
{
	return ((ceil(aArg)-aArg) <= (aArg-floor(aArg))) ? ceil(aArg) : floor(aArg);
}
/*!
	aResult is the closest power of 2 of aValue.
	Return value is log_{2} aResult.
*/
int Utils::GetClosestPowerOf2(long &aResult, long aValue)
{
	int i = 0;
	aResult = aValue;
	while ((aResult = aResult >> 1))
	{
		i++;
	}

	aResult = 1 << (i - 1); 

	if (aResult & aValue)
	{
		aResult = aResult << 2;
		i++;
	}
	else
	{
		aResult = aResult << 1;
	}

	return i;
}
/*!
	aResult is the closest bigger power of 2 of aValue.
	Return value is log_{2} of aResult.
*/
int Utils::GetClosestBiggerPowerOf2(long &aResult, long aValue)
{
	int i = 0;
	aResult = aValue;
	while ((aResult = aResult >> 1))
	{
		i++;
	}

	aResult = 1 << (i + 1); 

	return i;
}
/** This function is here because Win32 executables do not like mpz_out_str with
 * statically linked libc. (In my opinion: not even with dynamically linked,
 *  at least I had no success).
 */
int Utils::PrintMpzToFile(FILE* aFile, int base, mpz_t aValue) 
{		
	char * str = mpz_get_str(NULL, base, aValue);		
	if (str != NULL)
	{
		fprintf(aFile,"%s",str);
		free(str);
		return ConstRC::Ok;
	}
	else
	{
		return ConstRC::GeneralError;
	}
}
/*!
 * \brief Compare two "long" types.
 *
 * This function is used in qsort calls operating
 * on arrays of long types.
 *
 */	
int Utils::CompareLongs(const void* aArg1, const void* aArg2) 
{
	return (*((long*) aArg1) - *((long*) aArg2));
}
/*!
 * \brief Compare two "int" types.
 *
 * This function is used in qsort calls operating
 * on arrays of int types.
 *
 */	
int Utils::CompareInts(const void* aArg1, const void* aArg2) 
{
	return (*((int*) aArg1) - *((int*) aArg2));
}
/*!
 * \brief Compare two "double" types.
 *
 * This function is used in qsort calls operating
 * on arrays of double types.
 *
 */	
int Utils::CompareDoubles(const void* aArg1, const void* aArg2) 
{
	return (*((double*) aArg1) - *((double*) aArg2));
}

/*
int Utils::IncorrectLegendre(int c_p, int prime, Polynomial* polynomial) 
{	
	mpz_t temp, temp2;
	mpz_init(temp);
	mpz_init(temp2);
	mpz_set_ui(temp,c_p);
	polynomial->Evaluate(temp,temp);
	mpz_set_ui(temp2,prime);
	int legendre = mpz_legendre(temp,temp2);
	mpz_clear(temp);
	mpz_clear(temp2);

	return legendre;
}*/

string Utils::ConvertToString(bool aValue)
{
	return aValue ? TRUE_STRING : FALSE_STRING;
}
string Utils::ConvertToString(int aValue)
{
	stringstream stream;
	stream << aValue;

	return stream.str();
}
string Utils::ConvertToString(long aValue)
{
	stringstream stream;
	stream << aValue;

	return stream.str();
}
string Utils::ConvertToString(double aValue)
{
	stringstream stream;
	stream << aValue;

	return stream.str();
}
string Utils::ConvertToString(const mpz_t aValue)
{
	char * strChar = mpz_get_str(NULL, DEFAULT_NUMBER_BASE, aValue);
	if (strChar == NULL)
	{
		CLog::LogError(WHEREARG,"Utils","ConvertToString",0,0,"Could not convert mpz_t to string.");
		return "";
	}

	string str(strChar);
	free(strChar);
	strChar = NULL;

	return str;
}
int Utils::GCD(const main_sieving_type &aX, const main_sieving_type &aY, main_sieving_type &aResult)
{
	if ((aX < 0) || (aY <= 0))
	{
		aResult = 0;
		CLog::LogError(WHEREARG,"Utils","GCD",aX,aY,"Incorrect X or Y.");
		return ConstRC::GeneralError;
	}
	main_sieving_type x = aX;
	main_sieving_type y = aY;
	if (aX <= aY)
	{
		x = aY;
		y = aX;
	}

	while (y)
	{
		aResult = y;
		y = x % y;
		x = aResult;
	}

	return ConstRC::Ok;
}

int Utils::LGCD(const long &aX, const long &aY, long &aResult)
{
	if ((aX < 0) || (aY <= 0))
	{
		aResult = 0;
		CLog::LogError(WHEREARG,"Utils","GCD",aX,aY,"Incorrect X or Y.");
		return ConstRC::GeneralError;
	}
	long x = aX;
	long y = aY;
	if (aX <= aY)
	{
		x = aY;
		y = aX;
	}

	while (y)
	{
		aResult = y;
		y = x % y;
		x = aResult;
	}

	return ConstRC::Ok;
}

/*!
	Tonelli-Shanks algorithm for square root modulo p
	more Cohen - A Course in Computational Algebraic Number Theory - algorithm 1.5.1 
	same notation
*/
int Utils::TonelliShanks(mpz_t x, mpz_t a, mpz_t p)
{
	mpz_t b,n,y,q,t,power;
	gmp_randstate_t state;
	int e,r,m;
	
	mpz_init(b);
	mpz_init(n);
	mpz_init(y);
	mpz_init(q);
	mpz_init(t);
	mpz_init(power);
	
	gmp_randinit_mt(state);
	
	do
	{	
		mpz_urandomm(n, state, p);
	}
	while (mpz_legendre(n, p) != -1);
	
	// p-1 = q * 2^e
	mpz_sub_ui(t, p, 1);
	for (e = 1; mpz_divisible_2exp_p(t, e); e++);
	e--;
	mpz_fdiv_q_2exp(q, t, e);
	
	mpz_powm(y, n, q, p);
	r = e;
	mpz_cdiv_q_ui(t, q, 2);
	mpz_powm(x, a, t, p);
	mpz_powm(b, a, q, p);
		
	while (mpz_cmp_si(b, 1))
	{
		mpz_set_si(power, 1);
		m = 0;
		do
		{
			mpz_mul_si(power, power, 2);
			m++;
			mpz_powm(t, b, power, p);
		}
		while (mpz_cmp_si(t, 1));
				
		mpz_ui_pow_ui(power, 2, r-m-1);
		mpz_powm(t, y, power, p);
		mpz_powm_ui(y, t, 2, p);
		r = m;
		mpz_mul(x, x, t);
		mpz_mod(x, x, p);
		mpz_mul(b, b, y);
		mpz_mod(b, b, p);
	}
	
	
	mpz_clear(b);
	mpz_clear(n);
	mpz_clear(y);
	mpz_clear(q);
	mpz_clear(t);		
	mpz_clear(power);
	gmp_randclear(state);
	
	return ConstRC::Ok;
}
int Utils::PreviousPrime(mpz_t rop, mpz_t op)
{
	if (mpz_divisible_2exp_p(op, 1))
	{
		mpz_sub_ui(rop, op, 1);
	}
	else
	{
		mpz_sub_ui(rop, op, 2);
	}
	
	while (mpz_probab_prime_p(rop, 7) == 0)
	{
		mpz_sub_ui(rop, rop, 2);
	}	
	
	return ConstRC::Ok;
}

int Utils::WriteDataToFile(string aFileName, void * aData, long aCount, int aSize)
{
	FILE * fw = NULL;
	fw = fopen(aFileName.c_str(), FILE_MODE_WRITE_BINARY);
	if (fw == NULL) 
	{
		return ConstRC::GeneralError;
	}

	fwrite(&aCount, 1, sizeof(long), fw);
	fwrite(&aSize, 1, sizeof(int), fw);
	fwrite(aData, aCount, aSize, fw);

	fclose(fw);
	fw = NULL;

	return ConstRC::Ok;
}
int Utils::ReadDataFromFile(string aFileName, void ** aData, long& aCount, int& aSize)
{
	FILE * fr = NULL;
	if (*aData != NULL)
	{
		return ConstRC::BadArgument;
	}

	fr = fopen(aFileName.c_str(), FILE_MODE_READ_BINARY);
	if (fr == NULL)
	{
		return ConstRC::GeneralError;
	}

	fread(&aCount, 1, sizeof(long), fr);
	fread(&aSize, 1, sizeof(int), fr);

	*aData = Utils::AllocationMachine(aData, aCount * aSize);
	if (*aData == NULL)
	{
		return ConstRC::NotEnoughMemory;
	}

	fread(*aData, aCount, aSize, fr);
	
	fclose(fr);
	fr = NULL;

	return ConstRC::Ok;
}

int Utils::CreateDirectory(std::string aPath)
{
	int rc = ConstRC::Ok;
#ifdef WIN32
	rc = _mkdir(aPath.c_str());
#else
	rc = mkdir(aPath.c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
#endif

	if (rc == -1)
	{
		if (errno == EEXIST)
		{
			rc = ConstRC::AlreadyExists;
		}
		else
		{
			CLog::LogError(WHEREARG,"Utils","CreateDirectory",0,errno,strerror(errno));
			rc = ConstRC::GeneralError;
		}
	}
	else
	{
		rc = ConstRC::Ok;
	}

	return rc;
}

/*!
	This method will assess the reasonable hashtable size. 
	We want this size to be a power of 2.

	Generally, the table should be occupied from less than 80 per cent;
	if our assessment of fullness exceeds this, we multiply the hashtable
	size by 2.	

*/
relation_index_type Utils::GetSuitableHashtableSize(relation_index_type aAmount)
{
	relation_index_type orig = aAmount;
	aAmount = ConstHT::MIN_SIZE;
	double ratio = orig / (double) aAmount;
	// We do not want the table to be 'too full'
	while (ratio > ConstHT::MAX_FULLNESS) 
	{
		ratio = orig / (double) aAmount;
		aAmount <<=1;
	}

	printf("Original %ld, ratio %0.2f, Suitable size: %ld, 0x%lx.\n",orig,ratio,aAmount,aAmount);
	return aAmount;
}

