#ifndef __XML_HELPER__
#define __XML_HELPER__

#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>
#include <libxml/chvalid.h>
#include "common_definitions.h"
#include "xml_definitions.h"
#include "constants.h"
#include "const_return_codes.h"
#include "common_utils.h"
#include "enums_common.h"
#include <gmp.h>
#include <time.h>
#include <string>

using namespace std;

/*!
 * \brief XML helper methods common for NFS and QS
 */
class CommonXMLHelper
{
private:
	typedef unsigned char byte;

	static endian_types iMachineEndianType;
	static endian_types Endian();

	static const char* HEX_NUMBERS;
	static const int HEX_NUMBERS_LENGTH = 16;
	static const int MAX_DEPTH = 100000;
public:
	static int MoveToNextNode(xmlTextReaderPtr& aReader, int aMinDepth = -1, int aMaxDepth = CommonXMLHelper::MAX_DEPTH);

	static int ParseDecimalDouble(xmlChar* aString, double& aTarget);
	static int ParseDecimalMpz(xmlChar* aString, mpz_t aTarget);
	static int ParseDecimalInt(xmlChar* aString, int& aTarget);
	static int ParseDecimalUnsignedIntOrLong(xmlChar* aString, unsigned int& aTarget);
	static int ParseDecimalLong(xmlChar* aString, long& aTarget);
	static int ParseDecimalUnsignedIntOrLong(xmlChar* aString, unsigned long& aTarget);

	static int NewLine(xmlTextWriterPtr& aWriter, bool aIndent);

	static int DeserializeTime(xmlTextReaderPtr& aReader, const char* aMessage, bool aInfoMode);
	static int SerializeTime(xmlTextWriterPtr& aWriter);
	static int SerializeTime(xmlTextWriterPtr& aWriter, const char* aTag, time_t aTime);

	static int SerializeMpz(xmlTextWriterPtr& aWriter, const char* aTagName, /*const char* aAttributeName,*/ const mpz_t aValue);
	static int DeserializeMpz(xmlTextReaderPtr& aReader, const char* aMessage, mpz_t aTarget, bool aInfoMode);

	static int SerializeSignedInt(xmlTextWriterPtr& aWriter, const char* aTagName, /*const char* aAttributeName,*/ int aValue);
	static int DeserializeSignedInt(xmlTextReaderPtr& aReader, const char* aMessage, int& aTarget, bool aInfoMode);

	static int SerializeUnsignedInt(xmlTextWriterPtr& aWriter, const char* aTagName, /*const char* aAttributeName,*/ unsigned int aValue);
	static int DeserializeUnsignedInt(xmlTextReaderPtr& aReader, const char* aMessage, unsigned int& aTarget, bool aInfoMode);

	static int SerializeSignedLong(xmlTextWriterPtr& aWriter, const char* aTagName, /*const char* aAttributeName,*/ long aValue);
	static int DeserializeSignedLong(xmlTextReaderPtr& aReader, const char* aMessage, long& aTarget, bool aInfoMode);

	static int DeserializeUnsignedLong(xmlTextReaderPtr& aReader, const char* aMessage, unsigned long& aTarget, bool aInfoMode);
	static int SerializeUnsignedLong(xmlTextWriterPtr& aWriter, const char* aTagName, /*const char* aAttributeName,*/ unsigned long aValue);

	static int SerializePolynomialCoefficient(xmlTextWriterPtr& aWriter, mpz_t aValue, int aDegree);
	static int DeserializePolynomialCoefficient(xmlTextReaderPtr& aReader, const char* aMessage, mpz_t aTargetValue, int& aTargetDegree, bool aInfoMode);


	static int SerializeDouble(xmlTextWriterPtr& aWriter, const char* aTagName, /*const char* aAttributeName,*/ double aValue);
	static int DeserializeDouble(xmlTextReaderPtr& aReader, const char* aMessage, double& aTarget, bool aInfoMode);


	static int SerializeString(xmlTextWriterPtr& aWriter, const char* aTagName, const char* aValue);
	static int DeserializeString(xmlTextReaderPtr& aReader, const char* aMessage, char*& aTarget, bool aInfoMode);

	static int SerializeString(xmlTextWriterPtr& aWriter, const char* aTagName, const std::string& aValue);
	static int DeserializeString(xmlTextReaderPtr& aReader, const char* aMessage, string& aTarget, bool aInfoMode);

	static int WriteSerializatorData(xmlTextWriterPtr& aWriter, const char* aSerializatorNameVal, const char* aSerializatorVersion);

	static int SerializeBinaryData(xmlTextWriterPtr& aWriter, const char* aTagName, const void* aValue,  const long& aValueUnitCount, const long& aValueUnitByteLength);
	static int DeserializeBinaryDataIn(xmlTextReaderPtr& aReader, const char* aMessage, void* aTarget, const long& aTargetUnitCount, const long& aTargetUnitByteLength, bool aInfoMode);
	static int DeserializeBinaryDataOut(xmlTextReaderPtr &aReader, const char *aMessage, void** aTarget, long& aTargetUnitCount, long& aTargetUnitByteLength, bool aInfoMode);

	static int ConvertToHEX(const void * aField, const endian_types &aHexEndianType, char * aResult, const long &aUnitCount, const long &aUnitByteLength, const long &aResultCharLength);
	static int ConvertFromHEX(const char *aHex, const endian_types &aHexEndianType, void *aResult, const long &aHexCharLength, const long &aResultUnitCount, const long &aResultUnitByteLength);

	static int SerializeParameterEntry(xmlTextWriterPtr& aWriter, const char* aTagName, const char* aName, const char* aValue);
	static int DeserializeParameterEntry(xmlTextReaderPtr& aReader, const char* aMessage, char*& aTargetName, char*& aTargetValue, bool aInfoMode);
	static int DeserializeParameterEntry(xmlTextReaderPtr& aReader, const char* aMessage, std::string& aTargetName, std::string& aTargetValue, bool aInfoMode);

	static int ParseEndianType(xmlChar* aString, endian_types& aTarget); 

	static int SerializePrimitiveRoot(xmlTextWriterPtr& aWriter, const char* aTagName, const int& aRoot, const int& aPrime);
	static int DeserializePrimitiveRoot(xmlTextReaderPtr& aReader, const char* aMessage, int& aRoot, int& aPrime, bool aInfoMode);

	static int SerializePowerModP(xmlTextWriterPtr& aWriter, const char* aTagName, const int& aPrime, const int& aNumber, const int& aPower);
	static int DeserializePowerModP(xmlTextReaderPtr& aReader, const char* aMessage, int& aPrime, int& aNumber, int& aPower, bool aInfoMode);
	};

#endif

