#include "connector_info.h"
#include "xml_message_definitions.h"
#include "abstract_sieve.h"
#include "abstract_message.h"
#include <iostream>
#include "log.h"

#include "filesystem_connector.h"
#include "xml_message_definitions.h"


ConnectorInfo::ConnectorInfo(string& aUrlString, connector_type aConnectorType)
{
	iOwnsConnector = FALSE;
	SetUrlString(aUrlString);
	SetConnectorType(aConnectorType);
}

ConnectorInfo::ConnectorInfo(const char* aUrlString, connector_type aConnectorType)
{
	iOwnsConnector = FALSE;
	SetUrlString(aUrlString);
	SetConnectorType(aConnectorType);
}

ConnectorInfo::ConnectorInfo(const ConnectorInfo& aSrc)
{
	iOwnsConnector = FALSE;
	SetUrlString(aSrc.iUrlString);
	SetConnectorType(aSrc.iConnectorType);
}

ConnectorInfo::~ConnectorInfo()
{
	if (iOwnsConnector)
	{
		delete iConnector;
		iConnector = NULL;
	}
}

const char*	ConnectorInfo::GetUrlString_Char()	const
{
	return iUrlString.c_str();
}

const string&	ConnectorInfo::GetUrlString() const
{
	return iUrlString;
}

connector_type	ConnectorInfo::GetConnectorType() const
{
	return iConnectorType;
}

void    ConnectorInfo::SetUrlString(const string& aUrlString)
{
	iUrlString.assign(aUrlString);
}

void	ConnectorInfo::SetUrlString(const char* aUrlString)
{
	if (!aUrlString)
	{
		iUrlString.assign("");
		return;
	}
	iUrlString.assign(aUrlString);
}

void	ConnectorInfo::SetConnectorType(connector_type aConnectorType)
{
	iConnectorType = aConnectorType;
}

void 	ConnectorInfo::Load(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return;
	}
	string URL = aMessage->Get_SenderURL_String();
	if (URL.empty())
	{ 
		URL = aMessage->Get_TargetURL_String();
	};
	if (!URL.empty())
	{
		iUrlString.assign(URL);
		iConnectorType = DetectConnectorTypeFromUrl(URL);
	}
}

connector_type	ConnectorInfo::DetectConnectorTypeFromUrl(const string& aUrl)
{
	URL url(aUrl);
	if (!url.HasScheme())
	{		
		return EUnknownConnector;
	}
	string str1 = url.GetScheme();
	if (str1 == "file")
	{		
		return EFileConnector;
	}
	return EUnknownConnector;
}


void 	ConnectorInfo::Print(unsigned int indent) const
{
	string indented;	
	if (indent < 25)
	{
		unsigned int i;
		for (i=0; i< indent; i++)
		{
			indented.append(" ");		
		}
	}
	const char* connection = "(\?\?\?)";
	switch (iConnectorType)
	{
		case EFileConnector:
			connection = "(file)";
			break;
		default: 
			CLog::LogError(WHEREARG,"ConnectorInfo","Print",iConnectorType,0,"Unsupported connector type");
	}
	cout << indented << "URL String: " << iUrlString << '\n';
	cout << indented << "Connect   : " << iConnectorType << " " << connection <<'\n';
}

bool	ConnectorInfo::IsEqual(const ConnectorInfo* aOtherInstance) const
{
	if (!aOtherInstance) return FALSE;
	// Now we know that the pointer is not NULL
	bool sameConns = (GetConnectorType() == aOtherInstance->GetConnectorType());
	if (!sameConns) return FALSE;
	int sameAddresses = iUrlString.compare(aOtherInstance->iUrlString);
	if (sameAddresses == 0) return TRUE; // 0 ... equal strings
	return FALSE;
}

ConnectorInfo* ConnectorInfo::Parse(const char* info)
{
	string str(info);
	// At first, check for correct formatting
	// Zero'th char must be '['
	// last char must be ']'
	unsigned int pos1, len  = str.length();
	if (len < 6)
	{
		printf("ConnectorInfo.Parse: ConnectorType info must be longer than 5 chars!\n");
		cout << str << "\n";
		return NULL;
	}
	if (str[0] != '[')
	{
		printf("ConnectorInfo.Parse: ConnectorType info must start with '[' char!");
		return NULL;
	}
	if (str[len-1] != ']')
	{
		printf("ConnectorInfo.Parse: ConnectorType info must end with ']' char!");
		return NULL;
	}
	str.erase(len-1,1); // removal of the last ']'
	str.erase(0,1);	// removal of the first '['

	// Now, we detect the connector type. It can be either given explicitely
	// [file,/url/to/file/], or embedded as the url scheme [file:///url/to/file]
	pos1 = str.find(',',0);
	bool mustHaveScheme = FALSE;
	connector_type ct = EFileConnector;
	if (pos1 == string::npos)
	{
		mustHaveScheme = TRUE;
		// This is not an error; however, the URL that follows, must have scheme
	} else
	{
		string connection(str,0,pos1);
	// 	printf("ConnectorType read as %s\n",connection.c_str());
		if (connection.empty())
		{
			printf("ConnectorInfo.Parse: ConnectorType info must have nonempty connection!");
			return NULL;
		}	
		
		int conn = atoi (connection.c_str());
		if (conn == 0)
		{
			// The "connection" string was not a number.
			// We will try to parse it.
			bool conv = ParseConnectorType(connection,ct);
			// Error will be printed by the ParseConnectorType call
			if (!conv)
			{		
				return NULL;
			}
		} 
        	else 
		{
			// outside legal values for connection
			if ((conn < 0)  || (conn >= ESupremum))
			{
				PrintAllowedConnTypes();
				return NULL;
			} else
			{
				// A valid cast
				ct = static_cast<connector_type>(conn);
			}	
			
		}
		str.erase(0,pos1+1);	// the connection and the , removed
	}
	// Then, we detect the url
	pos1 = str.find(',',0);
	string addressstr;
	if (pos1 == string::npos)
	{
		addressstr.assign(str);
	}
	else {
		addressstr.assign(str,0,pos1);
	}	
	// Scheme detection in the current URL
	pos1 = str.find(':',0);	
	if (pos1 == string::npos)
	{
		if (mustHaveScheme)
		{
			printf("ConnectorInfo.Parse: Connector type was not determined!");
			return NULL;
		}
		else
		{
			AddSchemeToUrlString(addressstr,ct);
		}
	}
	else
	{
		string scheme;
		scheme.assign(addressstr,0,pos1);
		bool res = ParseConnectorType(scheme,ct);
		if (!res)
		{
			return NULL;
		}
	}
		
	ConnectorInfo* inst = new ConnectorInfo(addressstr,ct);
	return inst;
}

bool	ConnectorInfo::ParseConnectorType(const string& aValue, connector_type& aType)
{
	if ((aValue == "FILE") || (aValue == "file"))
	{
		aType = EFileConnector;
		return TRUE;
	}
	printf("ConnectorInfo.ParseConnectorTypeType: Invalid connection value: \"%s\"!",aValue.c_str());
	PrintAllowedConnTypes();
	return FALSE;
}
	
void	ConnectorInfo::PrintAllowedConnTypes()
{
	printf("\n");
	printf(" ______________________________________________\n");
	printf("|           Allowed connector  types           |\n");
	printf("|______________________________________________|\n");
	printf("|   Number             |    Text name          |\n");
	printf("|______________________|_______________________|\n");
	printf("|   1                  |    file	       |\n");
	printf("|______________________|_______________________|\n");
}

// This method is usually called from the subclasses
int	ConnectorInfo::Deserialize(xmlTextReaderPtr& aReader)
{
	int rc;
	int currDepth = xmlTextReaderDepth(aReader);
	xmlChar* name = xmlTextReaderName(aReader);
	int type = xmlTextReaderNodeType(aReader);
	SetConnectorType(EUnknownConnector);
	xmlChar* connectorType = xmlTextReaderGetAttribute(aReader, BAD_CAST TYPE_NAME);	
	if (connectorType != NULL)
	{					
		if (xmlStrEqual(connectorType, BAD_CAST FILE_NAME))
		{			
			SetConnectorType(EFileConnector);			
		}					
	}
	while (TRUE) {
		rc = xmlTextReaderRead(aReader);
		if (rc == 0)
		{
			CLog::LogError(WHEREARG,"ConnectorInfo","Deserialize",rc,0,"Could not deserialize tag, no more nodes");
			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"ConnectorInfo","Deserialize",rc,0,"Could not deserialize tag, no more nodes");			printf("Name: %s.\n",name);
        		return GENERAL_ERROR;
		}
		if (xmlTextReaderDepth(aReader) == currDepth)
		{			
			break;
		} 
         	type = xmlTextReaderNodeType(aReader);
 		name = xmlTextReaderName(aReader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{					
			if (xmlStrEqual(name,BAD_CAST URL_NAME))
			{
				CommonXMLHelper::DeserializeString(aReader,"Url read as ",iUrlString,FALSE);
				size_t pos;
				if ((pos = iUrlString.find(':',0)) == string::npos)
				{
					AddSchemeToUrlString(iUrlString,iConnectorType);
				} else
				{
					connector_type t1;
					bool supported = ParseConnectorType(iUrlString.substr(0,pos),t1);
					if (!supported)
					{
						SetConnectorType(EUnknownConnector);
					}
					else
					{				
						if ((GetConnectorType() != EUnknownConnector) && (t1 != GetConnectorType()))
						{
						CLog::LogError(WHEREARG,"ConnectorInfo","Deserialize",rc,0,"Connector type mismatch");
						} else
						{
							SetConnectorType(t1);
						}
				
					}
				}
			}
		} // if (type == 1)
	} // end of while(TRUE)
	return ConstRC::Ok;
}

void	ConnectorInfo::SetExternalConnector(AbstractConnector* aConnector)
{
	if ((iConnector) && (iOwnsConnector))
	{
		delete iConnector;
		iConnector = NULL;
	}
	iOwnsConnector = FALSE;
	iConnector = aConnector;
}

void	ConnectorInfo::AddSchemeToUrlString(string& aTarget, connector_type aType)
{
	string urlwithscheme;
	switch (aType)
	{
		case EFileConnector:
			urlwithscheme.assign("file");
		break;
		default: CLog::LogError(WHEREARG,"ConnectorInfo","AddSchemeToUrlString",aType,0,"Unsupported connector type");
	}
	urlwithscheme.append("://");
	urlwithscheme.append(aTarget);
	aTarget.assign(urlwithscheme);
}

const char*	ConnectorInfo::GetFailureReason() const
{
	return iFailureReason.c_str();
}

int	ConnectorInfo::CreateConnector(const string& aOutgoingIdentification)
{
	FilesystemConnector* connector = NULL;
	int res = ConstRC::Ok;
	switch (GetConnectorType())
	{
		case EFileConnector:
			connector = new FilesystemConnector(GetUrlString(),aOutgoingIdentification);
			iOwnsConnector = TRUE;
			res = connector->IsReady();
			SetFailureReason(res); // manages also "ConstRC::Ok"
			if (res == ConstRC::Ok)
			{	
				iConnector = connector;
				return ConstRC::Ok;
			}
			else
			{				
				delete connector;
				return CONNECTOR_NOT_READY;
			}
			break; // unreachable by now, but for future code changes
		default:
			SetFailureReason(INCORRECT_CONNECTOR_VALUE);
			return INCORRECT_CONNECTOR_VALUE;
	}
	return ConstRC::Ok;
}

void	ConnectorInfo::SetFailureReason(const string& aReason)
{
	iFailureReason = aReason;
}

void	ConnectorInfo::SetFailureReason(int aCode)
{
	if (INCORRECT_CONNECTOR_VALUE == aCode) {
		iFailureReason = "Unsupported connection method";
		return;
	};
	if (ADDRESS_NOT_SET == aCode) { 
		iFailureReason = "Address not set";
		return;
	};
	if (IDENTIFICATION_NOT_SET == aCode) {
		iFailureReason = "Identification not set";
		return;
	};
	if (NULL_POINTER_SUPPLIED == aCode) {
		iFailureReason = "Null pointer supplied";
		return;
	};
	if (DIRECTORY_DOES_NOT_EXIST == aCode) {
		iFailureReason = "Directory does not exist";
		return;
	};
	if (DIRECTORY_ACCESS_DENIED == aCode) {
		iFailureReason = "Access denied to the directory";
		return;
	};
	if (GENERAL_FILESYSTEM_ERROR == aCode) {
		iFailureReason = "General filesystem error";
		return;
	};
	if (ConstRC::Ok == aCode) {
		iFailureReason.clear();
		return;
	};
	iFailureReason = "Unknown error";
	return;
}

bool			ConnectorInfo::HasConnector() const
{	
	return (iConnector != NULL);
}

AbstractConnector&	ConnectorInfo::GetConnector()
{
	return *iConnector;
}

