#ifndef DISTRIBUTED_ALGORITHM_H
#define DISTRIBUTED_ALGORITHM_H

#include "algorithm.h"
#include <vector>

typedef vector<JobParameters*> Batch;

const unsigned int	KDefaultBatchSize = 8;

/*!
 * \brief  All algorithms that need to be distributed must implement  this interface.
 *
 *
 */
class DistributedAlgorithm : public virtual Algorithm
{
public: // New functionality
	
	/// This method will create a batch of new JobParameters for new Nodes.
	virtual Batch	CreateNewJobs(unsigned int aBatchSize = KDefaultBatchSize) = 0;

	virtual void	RunNodeInstance() = 0;	

	/// Inquiry about the status: is an instance already running?
	virtual bool	NodeInstanceRunning() const = 0;

	/// Inquiry about the status: has this algorithm some fresh data to send?
	virtual bool	HasDataToSend() const = 0;

	/// How much data to send?
	virtual unsigned int DataToSend() const = 0;

	/// What is the unit of "data"?
	virtual const char* DataUnit() const = 0;

	/// This will lock the data mutex for data exchange (Node->Center).
	virtual void	LockDataMutex() = 0;
	/// This will unlock the data mutex for data exchange (Node->Center).
	virtual void	UnlockDataMutex() = 0;
	virtual void	ClearFreshData() = 0;

	// We expect that the algorithm has a distributed phase
	// and a non-distributed phase that comes after the distributed phase.	
	virtual bool	IsDistributedPhaseFinished() const = 0;
	virtual int	RunNondistributedPhase() = 0;	
	virtual void	InterruptNodeNow() = 0;

	virtual bool IsComputationFinished() = 0;
	virtual void SetComputationFinished(bool aFlag) = 0;

	virtual void    ResetCenter() = 0;

	// should call the appropriate destructor for the subclass of JobParameters
        virtual int DeleteParameters(const JobParameters *aParameters) = 0;


public: // From Algorithm
	/// Returns a non-NULL name of this algorithm.
	virtual const char* 	AlgorithmName()       const = 0;
	/** Is intended to create a fresh, empty (or with default values)
	 instance of JobParameters appropriate for this algorithm.
	 This instance will probably be a direct subclass of JobParameters. */
	virtual JobParameters*  CreateNewParameters() const = 0;
};

#endif
