#ifndef _oldECM_H_
#define _oldECM_H_
#include "abstract_supporting_factor_alg.h"


/*!
  \brief This class implements the Elliptic Curve Method of factoring (ECM)

  For theoretic backgroung of the algorithm, see for example:
  Cohen, H. : An course in computational algebraic number theory, Springer 1996
*/

class ell_curve_ff;


/*!
 *\brief Class representing single point of an elliptic curve group.
 *\author Jan Jeronym Zvanovec, jero@email.cz
 *
 * 
 */

class ell_point {
	friend class oldECM;
	public:
	/// Constructs elliptic curve point (mainly allocates memory for coordinates)
	ell_point();
	/// Frees memory.
	~ell_point();
	/// True if the points are the same.
	bool equals(ell_point * Point);
	/// Prints point
	int Print();



	/// Pointer to the elliptic curve to which this point belongs.
	ell_curve_ff *curve;
	/// First coordinate of the point.
	mpz_t x;
	/// Second coordinate of the point.
	mpz_t y;
	/// Third coordinate of the point.
	mpz_t z;
	protected:
	/// temporary coordinate
	mpz_t t1;
	/// temporary coordinate
	mpz_t t2;
	/// temporary coordinate
	mpz_t t3;
	private:
	};

/**
 *\brief Basic class for computing in elliptic curves groups
 *\author Jan Jeronym Zvanovec, jero@email.cz
 *
 * Works using class ell_point::
 */

class ell_curve_ff {
	public:
		/// Constructor needing characteristics of the field (cannot be changed)
		ell_curve_ff(mpz_t afield);
		/// frees memory used by private objects
		~ell_curve_ff();
		/// Creates a point on the curve.
		ell_point *create_point();
		/// Copy point coordinates
		int copy_point(ell_point *aResult, ell_point *aSource);
		/// Adds two points on elliptic curve
		int add_points (ell_point *aResult, ell_point *aA, ell_point *aB);
		/// "Multiplies" point by integer (correspond's to powering in multiplicative group]
		int mul_point (ell_point* aResult, ell_point *aA, unsigned int mul);
		/// print point coordinates
		int print_point(ell_point aA);
		/// Sets the only parameter (see ell_curve_ff::a_4) of the elliptic curve over given field (see ell_curve_ff::field)
		int SetA(int aA);

	private:
		/// Characteristics of the field of the elliptic curve.
		mpz_t field;
		/// Result: divisor
		mpz_t result;
		/// Parameter defining the elliptic curve.
		/** At the moment, only the simplest subset of elliptic curves
		 * is implemented. Curves belonging to this subset are
		 * distinguished only by this single parameter.
		 */
		int a_4;
		
	};
/**
 *\brief Parallel Inverse Modulo N -- because multiplication
 * is faster than GCD
 *\author Jan Zvanovec, jero@email.cz
 *
 * Clever trick, can be quickly understood by reading the source or see Cohen,
 * H.: A course in computational algebraic number theory
 *
 * Works using class ell_point::
 */
class parallel_Inverse {
	public:
		parallel_Inverse(mpz_t modN,int newcount);
		parallel_Inverse(int newcount);
		~parallel_Inverse();

		int GetCount() {return count;};///< Get current degree of parallelization 
		int SetCount(int newcount);///< Set degree of parallelization

		int SetN(mpz_t newN);///< Set modulus
		int GetN(mpz_t result);///< Get modulus

		int Input(mpz_t InputA,int index); ///< Set input with index index
		int Output(mpz_t OutputB,int index);///< Get input with index index
	
		int Reset();

		int Put(mpz_t InputA);///< Add input
		int Get(mpz_t OutputA);///< Get input

		int Compute();///< Perform actual computing
		
	private:
		mpz_t N;///< Common modulus
		int count;///< degree of parallelization
		
		int stackindex;///< index for pushing/poping

		mpz_t *Interfaces;///< Inputs/Outputs
		mpz_t *Temps;///< Temporary field
		mpz_t MainInverse;

};


/*!
          \brief This class implements the elliptic curve factorization method
	  
*/
class oldECM: public ASupportingFactorAlg{
public:
	oldECM(int aNumCurves);
	~oldECM();
	int Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2);
	int Factor(qs_relation* aCandidate);
private:
	int NumCurves;
	parallel_Inverse *Inverse;
	ell_point *Points;
	ell_point *TempPoints;
	mpz_t Field;
	int * exponents;
	
	int b_1;
 	int max_exponent_index;
	
	mpz_t	fact_1;
	mpz_t	fact_2;
	int InitialisePoint(ell_point *aPoint);
	int AddPoints1 (ell_point *Result, ell_point *aA, ell_point *aB, unsigned a4);
	int AddPoints2 (ell_point *Result, ell_point *aA, ell_point *aB);
	int copy_point(ell_point *aResult, ell_point *aSource);
		
	int Factoring(mpz_t aN);
};


#endif

