#include "dirutil.h"
#include "abstract_sieve.h"
#include "filesystem_message.h"
#include "filesystem_connector.h"
#include <string.h>
#include "log.h"

#ifndef WIN32
#include <unistd.h>
#else
#ifndef _UNISTD_H
#define _UNISTD_H    1

/* This file intended to serve as a drop-in replacement for 
 *  unistd.h on Windows
 *  Please add functionality as neeeded 
 */

#include <stdlib.h>
#include <io.h>

/* Values for the second argument to access.
   These may be OR'd together.  */
#define R_OK    4       /* Test for read permission.  */
#define W_OK    2       /* Test for write permission.  */
//#define   X_OK    1       /* execute permission - unsupported in windows*/
#define F_OK    0       /* Test for existence.  */

#define access _access
#define open _open
#define read _read
#define close _close

#define STDIN_FILENO 0
#define STDOUT_FILENO 1
#define STDERR_FILENO 2

#endif /* unistd.h  */
#endif

// The library for cross-platform management of filesystem operations.
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/progress.hpp>

// Auxiliary definitions of boost versions
#include "boost_version.h"

namespace fs = boost::filesystem;
		
FilesystemConnector::FilesystemConnector() : AbstractConnector()
{
	iConnectorType = EFileConnector;
}

FilesystemConnector::FilesystemConnector(const char*   aURL, const char* aIdentification) : AbstractConnector(aURL,aIdentification)
{
	iConnectorType = EFileConnector;	
}

FilesystemConnector::FilesystemConnector(const string& aURL, const char* aIdentification) : AbstractConnector(aURL,aIdentification)
{
	iConnectorType = EFileConnector;	
}

FilesystemConnector::FilesystemConnector(const string& aURL, const string& aIdentification) : AbstractConnector(aURL,aIdentification)
{
	iConnectorType = EFileConnector;
}


// Returns TRUE if this connector can be used to receive messages
bool FilesystemConnector::SupportsReceive() const
{
	return TRUE;
}

// Returns TRUE if this connector can be used to send messages
bool FilesystemConnector::SupportsSend() const
{
	return TRUE;
}

int FilesystemConnector::IsReady()
{
	if ((!iURL) || (!iURL->HasFullPath()))
	{
		CLog::LogError(WHEREARG,"Filesystem connector","URL path not set",0,0,"Check constructor call.");
		return ADDRESS_NOT_SET;
	}
	return DirUtil::TestDirectoryAccess(iURL->GetFullPath());
}

using namespace fs;

int FilesystemConnector::ReceiveNow()
{
	if (!iURL->HasFullPath())
	{
		
		CLog::LogError(WHEREARG,"Filesystem connector::ReceiveNow","URL path not set",0,0,"Check constructor call.");
		return 0;
	}
	string pathString = iURL->GetFullPath();
	path full_path(pathString);

  	if ( !exists( full_path ) )
  	{
		CLog::LogError(WHEREARG,"Filesystem connector::ReceiveNow","receive dir does not exist ",0,0,"Check existence of the path");
#if BOOST_MINOR <= 45
		std::cout << full_path.native_file_string() << std::endl;
#else
		std::cout << full_path.string() << std::endl;
#endif
		return 0;	
	}

	unsigned long err_count = 0;	
//  	//printf("FilesystemConnector: full path is %s\n",iURL->GetFullPath().c_str());
	// Reading 
	if (is_directory( full_path ) )
  	{
// 		printf("FC: this path is a directory. Iterating through.\n");
		boost::filesystem::directory_iterator end_iter;
		boost::filesystem::directory_iterator dir_itr(full_path);
    		for (; dir_itr != end_iter; ++dir_itr )
    		{
      			try
      			{
				// We are interested in files only, not in dirs
				if ( !(is_directory( *dir_itr ) ))
        			{
					/*std::string filename = dir_itr->leaf();
						iFilenameCopy.assign(filename);**/		
					//iFilename.assign(dir_itr->leaf());
#if BOOST_MAJOR == 1
   #if BOOST_MINOR <= 34
					iFilename = dir_itr->leaf();
   #elif BOOST_MINOR  <= 45
				iFilename = dir_itr->filename();
   #else
					iFilename = dir_itr->path().filename().string();
   #endif
#else // BOOST_MAJOR != 1
   #error "Unknown boost version"
#endif
					
					// If the filename is well-formed...
					const char* chars = iFilename.c_str();
					//printf("\n---------");
 					//printf("\nFilename: %s. Ptr to chars = %p. Ptr to leaf chars = %p\n",chars,chars,dir_itr->leaf().c_str());
					/*if (chars == dir_itr->leaf().c_str())
					{
						printf("SHALLOW COPY!\n");
						exit(-1);
					}*/
					if (Check_Correctness(chars,TRUE))
					{
						Get_Sender(chars);
						Get_Counter(chars);	
						Get_Target(chars);
						Get_Extension(chars);
// 						printf("\n---------");
//  						printf("Sender = %s, target = %s, counter string = %s, counter read = %d extension = %s\n",iSenderAux.c_str(),iTargetAux.c_str(),iCounterAux.c_str(),iCounterRead,iExtensionAux.c_str());
						if ((iTargetAux == iIdentification) && (iExtensionAux == MESSAGE_EXTENSION))
						{
							// this message is for us.
//  							printf("File \"%s\" is a message for this receiver. Target Aux = %, Extension aux = %s\n",filename.c_str());
							string fullPath = iURL->GetFullPath();
							AbstractSerializator::AppendPathSeparator(fullPath);
							fullPath += iFilename;
							int ret =  ProcessMessageFromFile(fullPath);
							if (ConstRC::Ok != ret )
								CLog::LogError(WHEREARG,"Filesystem connector::ReceiveNow","Message processing unsuccessful.",1,0,"");
							
						}
					}
		
				}
      			}
      			catch ( const std::exception & ex )
      			{
        			++err_count;
        			std::cout << (*dir_itr) << " " << ex.what() << std::endl;
      			}
    		}
	} 
	else 
	{
		CLog::LogError(WHEREARG,"Filesystem connector::ReceiveNow","receive path is not a directory",0,0,"Check the path");
#if BOOST_MINOR <= 45
		std::cout << full_path.native_file_string() << std::endl;
#else
		std::cout << full_path.string() << std::endl;
#endif

		return 0;
	}
	return 0;
}

int FilesystemConnector::ProcessMessageFromFile(string& aPath)
{
	FilesystemMessage* messageInstance = new FilesystemMessage(aPath);
	messageInstance->Deserialize();	
// 	printf("FS Connector: process message from file %s\n",aPath.c_str());
// 	messageInstance->PrintHeaderInfo();
	if (messageInstance->Get_TargetURL_String() == "") {
		messageInstance->Set_TargetURL(iURL->GetCanonicalForm().c_str());
	}
	int result = AddReceivedMessage(messageInstance);
	//printf("FilesystemConnector: renamed %s to processed.\n",aPath.c_str());
	if (! (result & MESSAGE_INSTANCE_DELETED))
	{
		messageInstance->MarkAsProcessed();
	}
	return ConstRC::Ok;
}

// Partly copied from http://www.boost.org/libs/filesystem/example/simple_ls.cpp
void FilesystemConnector::Dir(const char* aPath)
{
/*	fs::path full_path( fs::initial_path<fs::path>() );  	

    	full_path = fs::system_complete( fs::path( iReceiveAddress, fs::native ) );*/
	
	return;
	fs::path full_path(aPath);

	unsigned long file_count = 0;
  	unsigned long dir_count = 0;
  	unsigned long other_count = 0;
  	unsigned long err_count = 0;

  	if ( !fs::exists( full_path ) )
  	{
#if BOOST_MINOR <= 45
    		std::cout << "\nNot found: " << full_path.native_file_string() << std::endl;
#else
    		std::cout << "\nNot found: " << full_path.string() << std::endl;
#endif

    		return;
  	}
	// Reading 
	if ( fs::is_directory( full_path ) )
  	{
#if BOOST_MINOR <= 45
 		std::cout << "\nIn directory: " << full_path.native_directory_string() << "\n\n";
#else
 		std::cout << "\nIn directory: " << full_path.string() << "\n\n";
#endif
    		fs::directory_iterator end_iter;
    		for ( fs::directory_iterator dir_itr( full_path );         		dir_itr != end_iter; ++dir_itr )
    		{
      			try
      			{
        			if ( fs::is_directory( *dir_itr ) )
        			{
          				++dir_count;
          				std::cout << (*dir_itr)<< " [directory]\n";
        			}
        			else /*if ( fs::is_regular( dir_itr->path() ) )*/
        			{
          				++file_count;
          				std::cout << (*dir_itr) << "\n";
					
        			}
        			/*else
        			{
          				++other_count;
          				std::cout << dir_itr->path().leaf() << " [other]\n";
        			}*/

      			}
      			catch ( const std::exception & ex )
      			{
        			++err_count;
        			std::cout << (*dir_itr) << " " << ex.what() << std::endl;
      			}
    		}
    		std::cout << "\n" << file_count << " files\n"
              << dir_count << " directories\n"
              << other_count << " others\n"
              << err_count << " errors\n";
	} else {
    		std::cout << "\nFound: " << full_path << "\n";  
	}
	return;
}

bool	FilesystemConnector::Check_Correctness(const char* aFilename, bool aVerbose)
{
	//printf("Filename = %s\n",aFilename);
	int i, length = strlen(aFilename);
	if (length < 10) // a_b_000001
	{
		if (aVerbose)
		{	
			CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","Length of filename is too small",0,length,"Check the filename");
		}
		return FALSE;
	}
	iFirst_ = -1;
	iSecond_= -1;	
	iDot_   = -1;
	for (i=0; i< length; i++)
	{
		if (aFilename[i] == '_')
		{
			if (iFirst_ == -1)
			{
				iFirst_ = i;
			} else if (iSecond_ == -1)
			{
				iSecond_ = i;
			} else {
				// More than two _s found. Incorrect!
				if (aVerbose)
				{
					CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","More than two _s found",0,i,"Check the filename");
				}
				return FALSE;
			}			
		}
		if (aFilename[i] == '.')
		{
			iDot_ = i;
		}
	}
	if ((iFirst_ == -1) || (iSecond_ == -1))
	{
		if (aVerbose)
		{
			printf("Filename: %s\n",aFilename);
			CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","Less than two _s found",0,0,"Check the filename");
		}
		return FALSE;
	}
	if (iFirst_ == 0)
	{
		// Empty target
		if (aVerbose)
		{
			CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","Target is empty",0,0,"Check the filename");
		}
		return FALSE;
	}
	if (iSecond_ == (length-1))
	{
		if (aVerbose)
		{
			CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","Counter is empty",0,0,"Check the filename");
		}
		return FALSE;
	}
	if ((iFirst_+1) == iSecond_)
	{
		if (aVerbose)
		{
			CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","Sender is empty",0,0,"Check the filename");
		}
		return FALSE;
	}
	const char* chr = &(aFilename[iSecond_+1]); // a correct address, since iSecond < length-1
	if (!atoi(chr))
	{
		if (aVerbose)
		{
			CLog::LogError(WHEREARG,"Filesystem connector::Check_Correctness","Counter is not a decimal number, or is equal to zero",0,0,"Check the filename");
		}
		return FALSE;
	}
	return TRUE;
}

const char*	FilesystemConnector::Get_Target(const char* aFilename)
{
	iTargetAux.assign("");
	if (!aFilename)
	{
		return NULL;
	}
	int targetLength = iFirst_;
	//printf("Get_Target: ptr = %p, aFilename = %s\n",aFilename, aFilename);
	//printf("Get_Target: targetLength = %d",iFirst_);
	iTargetAux.assign(aFilename,targetLength);
	return iTargetAux.c_str();
}

const char*	FilesystemConnector::Get_Sender(const char* aFilename)
{
	iSenderAux.assign("");
	if (!aFilename)
	{
		return NULL;
	}
	int senderLength = iSecond_ - iFirst_ - 1; 
	iSenderAux.assign(&(aFilename[iFirst_+1]),senderLength);
	return iSenderAux.c_str();
}

int	FilesystemConnector::Get_Counter(const char* aFilename)
{
	if (!aFilename)
	{
		return 0;
	}
	int counterLength = 0;
	if (iDot_ == -1)
	{
		counterLength = strlen(aFilename) - iSecond_ - 1;
	} else
	{
		counterLength = iDot_ - iSecond_ - 1;
	}	
	char aux[10];
	memset(aux,0,10*sizeof(char));
	strncpy(aux,&(aFilename[iSecond_+1]),counterLength);
	iCounterAux = aux;
	return (iCounterRead  = atoi(aux));
}

const char*	FilesystemConnector::Get_Extension(const char* aFilename)
{
	iExtensionAux.assign("");
	if (!aFilename)
	{	
		return NULL;
	}
	const char* ptr = strrchr(aFilename,'.');
	if (ptr == NULL)
	{
		return NULL;
	}
	// We accept only "short" extensions.
	if (strlen(ptr) > 6)
	{
		return NULL;
	}
	iExtensionAux.assign(ptr);	
	return iExtensionAux.c_str();
}

int 	FilesystemConnector::DoSend(AbstractMessage* aMessage)
{
	if (aMessage)
	{
		aMessage->Serialize();
		// Serialize will just write the message to the disk.
	}
	return ConstRC::Ok;
}

// Returns either EOk, or one of the more specific error states.
int	FilesystemConnector::GetState() const
{
	int state = AbstractConnector::GetState();
	if (state != KOk)
	{
		return state;
	}
	// We will test existence of the path and writability

	// Existence (non-NULL) ness of the URL is already ensured
	// by the previous call of "AbstractConnector::GetState"
	if (!iURL->HasFullPath())
	{
		return KDirectoryNameEmpty;		
	}
	fs::path full_path(iURL->GetFullPath());
  	if ( !fs::exists( full_path ) )
  	{
		return KDirectoryDoesNotExist;	
	}
	if ( !fs::is_directory( full_path ) )
	{
		return KIsNotDirectory;
	}
	// Now we know that the path exists and is a directory
	int result;
	result = access(iURL->GetFullPath().c_str(),R_OK);
	if (result == -1)
	{
		return KDirectoryReadProtected;
	}
	result = access(iURL->GetFullPath().c_str(),W_OK);
	if (result == -1)
	{
		return KDirectoryWriteProtected;
	}
	return KOk;
}

// Returns human-readable explanation of the error state.
string	FilesystemConnector::GetStateExplanation(int aState) const
{
	string res = AbstractConnector::GetStateExplanation(aState);
	if (!res.empty())
	{
		return res;
	}
	string path = iURL->GetFullPath();
	switch (aState)
	{
		case KDirectoryDoesNotExist:
			res = String_Directory;
			res += "'";
			res += path;
			res += "'";
			res += String_DoesNotExist;
			break;
		case KDirectoryWriteProtected:
			res = String_Directory;
			res += "'";
			res += path;
			res += "'";
			res += String_IsWriteProtected;
			break;
		case KDirectoryNameEmpty:
			res = String_DirectoryNameEmpty;
			break;
		case KDiskFull:
			res = String_DiskFull;
			break;
		case KIsNotDirectory:
			res = String_Path;
			res += "'";
			res += path;
			res += "'";
			res += String_IsNotADirectory;
			break;
		case KDirectoryReadProtected:
			res = String_Directory;
			res += "'";
			res += path;
			res += "'";
			res += String_IsReadProtected;
			break;
	}
	return res;
}
