#ifndef FILESYSTEM_CONNECTOR_H
#define FILESYSTEM_CONNECTOR_H

#include "abstract_connector.h"

#include <string>

using namespace std;

const int KDirectoryDoesNotExist 	= -100;
const int KDirectoryWriteProtected	= -101;
const int KDirectoryNameEmpty		= -102;
const int KDiskFull			= -103;
const int KIsNotDirectory		= -104;
const int KDirectoryReadProtected	= -105;

#define String_Directory		"Directory "
#define String_DoesNotExist		" does not exist."
#define String_IsWriteProtected    	" is protected against writing."
#define String_IsReadProtected    	" is protected against reading."
#define String_DiskFull		"Disk is full."
#define String_Path			"Path "
#define String_IsNotADirectory		" is not a directory."
#define String_DirectoryNameEmpty	"Directory name is empty."


/*!
 \brief A communication class for messaging that uses files on a filesystem.

 FilesystemConnector fetches incoming messages from a directory indicated
 by the iReceiveAddress, and writes outgoing messages to a directory indicated
 by their own iTargetAddress. Those directories need not be the same, they
 can be on different machines etc.

 
 The connector creates instances of FilesystemMessages during the fetching 
 process. FilesystemMessages are subclasses of AbstractMessages. 

 During writing of the outgoing messages, the connector tacitly assumes that
 these messages are FilesystemMessages. This should be checked in the future,
 and the connector should refuse to send other AbstractMessages.

 This class makes heavy use of "iReceiveAddress", but "iSendAddress" is
 usually senseless. Target address for a message is taken from the message
 itself.

 The ReceiveNow() process lists the whole content of the receive directory,
 and fetches the files that belong to it by comparing the file names with
 the general pattern. This is a bit problematic from the performance point
 of view; it should be corrected before going on. For example, the target id
 should be the first in the filename, so that the pattern for matching files
 could be easier (like target_*.msg), and messages that are already processed
 and useless should be moved to another ("historical") directory, so that
 the working directory does not fill with too much garbage.

*/
class FilesystemConnector : public AbstractConnector {

	public:		
		FilesystemConnector();

		FilesystemConnector(const char*   aURL, const char* aIdentification);

		FilesystemConnector(const string& aURL, const char* aIdentification);

		FilesystemConnector(const string& aURL, const string& aIdentification);

		int IsReady();

		static void Dir(const char* aPath);

		// Returns TRUE if this connector can be used to receive messages
		virtual bool SupportsReceive() const;
		// Returns TRUE if this connector can be used to send messages
		virtual bool SupportsSend() const;

		// Returns either EOk, or one of the more specific error states.
		virtual int	GetState() const;
		// Returns human-readable explanation of the error state.
		virtual string	GetStateExplanation(int aState) const;
		

	protected:
		virtual int DoSend(AbstractMessage* aMessage);

		int ReceiveNow();

		int ProcessMessageFromFile(string& aPath);

		/*!
		*   Parses the filename for presence of exactly two _ chars.
 		*   All sections determined by them must be nonempty.
 		*   The last section of the string must be a 6-digit decimal numeric value greater than 0.
		*/
		bool	Check_Correctness(const char* aFilename, bool aVerbose);
		const char*	Get_Sender(const char* aFilename);
		const char*	Get_Target(const char* aFilename);
		const char*	Get_Extension(const char* aFilename);
		int	Get_Counter(const char* aFilename);

	private:
		string	iSenderAux;		
		string	iTargetAux;		
		string	iCounterAux;
		string  iExtensionAux;
		string	iFilename;
		int	iCounterRead;
		int	iFirst_;
		int	iSecond_;
		int	iDot_;
		
};
#endif


