#include "filesystem_message.h"
#include "abstract_sieve.h"
#include <errno.h>
#include <string.h>
#include <iostream>
#ifdef ARCHIVE_SUPPORT
	#include <archive.h>
	#include <archive_entry.h>
#endif
#include <cstdio>
#include <sys/stat.h>
#include <sys/types.h>

#ifndef WIN32
#include <unistd.h>
#else
#ifndef _UNISTD_H
#define _UNISTD_H    1

/* This file intended to serve as a drop-in replacement for 
 *  unistd.h on Windows
 *  Please add functionality as neeeded 
 */

#include <stdlib.h>
#include <io.h>

/* Values for the second argument to access.
   These may be OR'd together.  */
#define R_OK    4       /* Test for read permission.  */
#define W_OK    2       /* Test for write permission.  */
//#define   X_OK    1       /* execute permission - unsupported in windows*/
#define F_OK    0       /* Test for existence.  */

#define access _access
#define open _open
#define read _read
#define close _close

#define STDIN_FILENO 0
#define STDOUT_FILENO 1
#define STDERR_FILENO 2

#endif /* unistd.h  */
#endif

#include <fcntl.h>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include "boost_version.h"
#include "log.h"

FilesystemMessage::FilesystemMessage()
{
	AbstractMessage::Init();
// 	printf("FSMessage: Set sender connector type as %d.\n",iSenderConnectorType);
}

FilesystemMessage::FilesystemMessage(const string& aFullPath)
{
	AbstractMessage::Init();
	iFullPath.assign(aFullPath);	
	ConstructBodyPath();
	// 	printf("FSMessage: Set sender connector type as %d.\n",iSenderConnectorType);
}

FilesystemMessage::FilesystemMessage(const char* aDirectory, const char* aSubject, message_type aType, unsigned int aFlags, unsigned int aCounter,  const char* aSenderId, const char* aTargetId, const char* aData, unsigned int aTimeout)
{
	string URL = "file://";
	if (aDirectory)
	{
		iDirectory.assign(aDirectory);
		URL += aDirectory;
	}
	AbstractMessage::Init(NULL,URL.c_str(),aSubject,aType,aFlags,aCounter,aSenderId,aTargetId,aData,aTimeout);	
// 	printf("FSMessage: Set sender connector type as %d.\n",iSenderConnectorType);
}

FilesystemMessage::~FilesystemMessage()
{
}

void    FilesystemMessage::ConstructBodyPath()
{
	string extension = DetectExtension();
	boost::filesystem::path body_path(iFullPath);
#if BOOST_MINOR <= 45
	iFullPath_Body.assign(body_path.leaf());
#else
	iFullPath_Body.assign(body_path.filename().string());
#endif
	if (extension.length() > 0)
	{
		iFullPath_Body.erase(iFullPath_Body.length()-extension.length());	
	}
	iFullPath_Body.append(BODY_EXTENSION);
}

void    FilesystemMessage::CreateFileName(const char* aExtension)
{
	if (iFullPath.size() > 0)
	{
		//printf("FilesystemMessage::CreateFileName. FullPath already set: %s",iFullPath.c_str());
		return;
	}
	iFullPath.append(iDirectory);
	// Appends / or \ if necessary
	AbstractSerializator::AppendPathSeparator(iFullPath);
	iFullPath.append(iTargetId);
	iFullPath.append("_");
	iFullPath.append(iSenderId);
	char aux[10];
	memset(aux,0,10*sizeof(char));
	sprintf(aux,"_%06d",iCounter);
	iFullPath.append(aux);
	if (aExtension)
	{
		iFullPath.append(aExtension);
	}
	//printf("FilesystemMessage: created file name as %s\n",iFullPath.c_str());
}	

int	FilesystemMessage::MarkAsProcessed()
{
	ChangeExtension(PROCESSED_EXTENSION);
	return ConstRC::Ok;
}

int	FilesystemMessage::MarkAsFinished(bool aKeepMessage)
{
	if (aKeepMessage) {
		// archive old messages
#ifdef ARCHIVE_SUPPORT
		ChangeExtension(FINISHED_EXTENSION);
		MoveToSubdirectory();
#else
		CLog::LogError(WHEREARG,"FilesystemMessage","MarkAsFinished",0,0,"Keeping option not compiled in, either set deleting or recompile with ARCHIVE_SUPPORT");
#endif
	}
	else {
		// delete message
		remove(iFullPath.c_str());
	}

	return 0;
}

void	FilesystemMessage::ChangeExtension(const char* aNewExtension)
{
	if (aNewExtension == NULL)
	{
		return;
	}

	string extension = DetectExtension();
	if (extension == aNewExtension)
	{
		// Nothing to be done.
		return;
	}
	string oldName;
	oldName += iFullPath;
	string newName;
	newName += iFullPath;
	if (extension.length() > 0)
	{
		newName.erase(newName.length()-extension.length());
		//printf("FilesystemMessage: erased old extension. Path is now %s.\n",newName.c_str());		
	}
	newName += aNewExtension;
	//printf("FilesystemMessage: renaming from %s to %s.\n",oldName.c_str(),newName.c_str());
	// Rename the file - an old function from the C stdlib.
	int res = rename(oldName.c_str(),newName.c_str());
	if (res == -1)
	{
		CLog::LogError(WHEREARG,"FilesystemMessage","ChangeExtension",res,errno,"Could not rename the file.");
		return;        
	}
	else
	{
		iFullPath = newName;	
	}
}

string  FilesystemMessage::DetectExtension()
{
	string s1;
	size_t index = iFullPath.find_last_of('.');
	if (index == string::npos)
	{
		return s1; // empty string
	} 
	s1 = iFullPath.substr(index);
	//printf("FilesystemMessage: Extension detected as %s.\n",s1.c_str());
	return s1;
}


xmlTextWriterPtr	FilesystemMessage::CreateOutputStream()
{
	CreateFileName(NULL);
	return xmlNewTextWriterFilename(iFullPath.c_str(),0);
}

xmlTextReaderPtr	FilesystemMessage::CreateInputStream()
{
	CreateFileName(NULL);
	return xmlNewTextReaderFilename(iFullPath.c_str());
}

int			FilesystemMessage::FinishDeserialization()
{
	return ConstRC::Ok;
}

int			FilesystemMessage::FinishSerialization()
{
	//printf("Filesystem Message : Finish serialization\n");
	//printf("Filesystem Message : Finish serialization, calling change extension\n");	
	ChangeExtension(); // to .msg

	return ConstRC::Ok;
}

#ifdef ARCHIVE_SUPPORT
void	FilesystemMessage::MoveToSubdirectory()
{
	using namespace boost::filesystem;
	boost::filesystem::path message_path(iFullPath);
	string dir_path(message_path.parent_path().string());

	dir_path.append("/");
	dir_path.append(iTargetId);
	dir_path.append("_");
	dir_path.append(iSenderId);
	if (!is_directory(dir_path)) {
		create_directory(dir_path);
	};
	string in_dir_path(dir_path);
	in_dir_path.append("/");
#if BOOST_MINOR <= 45
	in_dir_path.append(message_path.leaf());
#else
	in_dir_path.append(message_path.filename().string());
#endif
	if ((!exists (in_dir_path))&& exists(message_path))  {
			rename(message_path, in_dir_path);
	};

	int cetnost = 10;
	if (0 == iCounter%cetnost) {
		struct archive_entry *Entry;
		struct stat Stats;
		struct archive *MessageArchive;	
		char buff[8192];
		char aux[10];
		int len;
		int fd;

		string ArchivPath;
		string MessagePath;
		string BodyPath;
		ArchivPath.append(iTargetId);
		ArchivPath.append("_");
		ArchivPath.append(iSenderId);
		memset(aux,0,10*sizeof(char));
		sprintf(aux,"_%06d",iCounter);
		ArchivPath.append(aux);
		ArchivPath.append(".tar.gz");

		MessageArchive = archive_write_new();
			archive_write_add_filter_gzip(MessageArchive);
			archive_write_set_format_pax_restricted(MessageArchive);
			archive_write_open_filename(MessageArchive, ArchivPath.c_str());
			
		directory_iterator end; for(directory_iterator iter(dir_path) ; iter != end; ++iter) {
			string MessagePath (iter->path().string());
			if (!boost::filesystem::exists(MessagePath)){
				continue;
			};
		     stat(MessagePath.c_str(), &Stats);
		      Entry = archive_entry_new();
		      archive_entry_set_pathname(Entry, MessagePath.c_str());
		      archive_entry_set_size(Entry, Stats.st_size);
		      archive_entry_set_filetype(Entry, AE_IFREG);
		      archive_entry_set_perm(Entry, 0644);
		      archive_write_header(MessageArchive, Entry);
		      fd = open(MessagePath.c_str(), O_RDONLY);
		      len = read(fd, buff, sizeof(buff));
		      while ( len > 0 ) {
			  archive_write_data(MessageArchive, buff, len);
			  len = read(fd, buff, sizeof(buff));
		      }
		      close(fd);
		      remove (MessagePath.c_str());
		archive_entry_free(Entry);
		};
		archive_write_close(MessageArchive);
		archive_write_free(MessageArchive);
	}
}
#endif

