#include "integer_class.h"

#include <stdlib.h>
#include <stdio.h>
#include <gmp.h>
#include "common_definitions.h"
#include <iostream> 
#include <math.h>
#include "const_return_codes.h"
#include "log.h"
#include "common_utils.h"
#include "const_filenames.h"
#include "integer_class.h"
#include "primitive_root.h"


short int CIntegerClass::Sgn(long aOperand){
	if (aOperand)
	{
		return ((aOperand > 0) ? 1 : -1);
	}
	
	return 0;
}

long CIntegerClass::Max(long a,long b)
{
	return a < b ? b : a;
}

long CIntegerClass::Min(long a,long b)
{
	return a > b ? b : a;
}

/// return random integer in interval [0..aRange-1]
long CIntegerClass::RandomNumber(long aRange)
{
	return (long) ((float)aRange * rand() / (RAND_MAX + 1.0));
}

/// return gcd(aOperand1,aOperand2) = aComb1*aOperand1 + aComb2*aOperand2
long CIntegerClass::ExtendedEuclid(long *aComb1, long *aComb2, long aOperand1, long aOperand2)
{
	long u,d,v1,v3,t1,t3,q;
	d = abs(aOperand1);
	u = CIntegerClass::Sgn(aOperand1);
	if (!aOperand2)
	{
		*aComb1 = u;
		*aComb2 = 0;
		return d;
	}
	v1 = 0;
	v3 = abs(aOperand2);
	while (v3) 
	{
		q = d / v3;
		t3 = d - q * v3;
		t1 = u - q * v1;
		u = v1;
		d = v3;
		v1 = t1;
		v3 = t3;
	}

	*aComb2 = (d - aOperand1 * u) / aOperand2;
	*aComb1 = u;
	return d;
}

long CIntegerClass::LeastCommonMultiple(long aOperand1, long aOperand2){
	long aux;
	return(aOperand1*aOperand2/ExtendedEuclid(&aux,&aux,aOperand1,aOperand2));
}

long CIntegerClass::InverseModN(long aOperand, long aModulus){
	long a,b;
	if (ExtendedEuclid(&a,&b,aOperand,aModulus)!=1)
	return 0;
	else return a;
}

long CIntegerClass::InverseModNPositive(long aOperand, long aModulus){
	long a,b;
	if (ExtendedEuclid(&a,&b,aOperand,aModulus)!=1)
	return 0;
	else
		if (a < 0)
			a += aModulus;
		return a;
}

long CIntegerClass::PowerModN(long aOperand, unsigned int aPower, long aModulus){
	int i,l;
	INT64 s;
	if (aPower==0) return 1;
	if (aOperand>=0) s=aOperand;
	else s=aModulus+aOperand%aModulus;
	if (s>unsigned (aModulus)) s%=aModulus;
	INT64 res=1;
	i=aPower;
	for(l=0;i;l++) i=i>>1;
	for(i=0;i<l;i++){
			if (i) {
		s*=s;
		if (s>0x7fffffff) s%=aModulus;
			}
			if (aPower & 1<<i) {
		res*=s;
		if(res>0x7fffffff) res%=aModulus;
			}
		}
	if(res>unsigned(aModulus)) res%=aModulus;
	return (long) res;
}

short int CIntegerClass::JacobiSymbol(long aOperand, long aModulus){
	const short int tab2[8]={0,1,0,-1,0,-1,0,1};
	short int v,k;
	long r;
	if (!aModulus) {
	if (abs(aOperand)==1) return 1;
	else return 0;
	}
	if (!(1 & (aOperand | aModulus))) return 0;
	for(v=0;!(aModulus & 1);v++) aModulus=aModulus>>1;
	if (v & 1) {
	k=tab2[aOperand&7];
	} else k=1;
	if (aModulus<0) {
	aModulus=-aModulus;
	if (aOperand<0) k=-k;
	}
	while (aOperand) {
	for(v=0;!(aOperand & 1);v++) aOperand=aOperand>>1;
	if (v & 1) {
		k=k*tab2[aModulus&7];
	}
	if (aOperand & aModulus & 2) k=-k;
	r=abs(aOperand);
	aOperand=aModulus%r;
	aModulus=r;
	}
	if (aModulus==1) return k;
	else return 0;
}
/*!
	This method is used to calculate arg1*arg2 mod mod, where arg1 and arg2 are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.

*/
int CIntegerClass::Mulm(INT64 aArg1, INT64 aArg2, int aMod)
{
	INT64 inter =  aArg1* aArg2;
	return (inter % aMod);	
}

/*!
	This method is used to calculate arg1*arg2 mod mod, where arg1 and arg2 are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.

*/
int CIntegerClass::Mulm(INT64 aArg1, int aArg2, int aMod)
{
	INT64 inter =  aArg1* (INT64) aArg2;
	return (inter % aMod);	
}

/*!
	This method is used to calculate arg1*arg2 mod mod, where arg1 and arg2 are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.

*/
int CIntegerClass::Mulm(int aArg1, int aArg2, int aMod)
{
	INT64 inter = (INT64) aArg1* (INT64) aArg2;	
	return (inter % aMod);
}

/*!
	This method is used to calculate arg1^exp mod mod, where arg1 is are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.
	A standard binary algorithm is used for calculation of powers.

*/
int CIntegerClass::Powm(int aArg, int aExp, int aMod)
{
	if (!aExp) return 0;
	if (aExp == 1) return (aArg % aMod);	
	INT64 inter = (INT64) aArg;	
	// nejprve zjistime maximalni jednickovy bit v cisle exp
	int i = -1;	
	while (aExp >> (++i)){} 
	int max = i - 1;  // nejvyssi signifikantni bit.
	int MASK = (1 << (max - 1)); // zacneme od druheho nejvyssiho signifikatniho bitu
	while (MASK > 0)
	{
		inter = CIntegerClass::Mulm(inter, inter, aMod);
		if (aExp & MASK) 
		{
			inter = CIntegerClass::Mulm(inter, aArg, aMod);
		}
		MASK = MASK >> 1;
	}

	return (inter % aMod);
}
int CIntegerClass::IPow(int aBase, int aExp)
{
	int result = 1;
	while (aExp)
	{
		if (aExp & 1)
			result *= aBase;
		aExp >>= 1;
		aBase *= aBase;
	}

	return result;
}
double CIntegerClass::Trunc(double aArg)
{
	if (aArg > 0)
	{
		double val = aArg;
		val = floor(val);
		return val;
	}
	else
	{
		double val = aArg;
		val = ceil(val);
		return val;
	}
}
long CIntegerClass::SquareRootModPrime(long aOperand, long aModulus){
	long i;
	if (aOperand>=aModulus){
	aOperand%=aModulus;
	} else if (aOperand<0) {
	aOperand=aModulus+(aOperand % aModulus);
	}
	if ((aModulus & 3) == 3){
	return(PowerModN(aOperand,(aModulus+1)>>2,aModulus));
	}
	if ((aModulus & 7) == 5){
	i=PowerModN(aOperand,(aModulus-1)>>2,aModulus);
	if (i==1) {
		return (PowerModN(aOperand,(aModulus+3)>>3,aModulus));
	} else {
		i=PowerModN(aOperand<<2,(aModulus-5)>>3,aModulus);
		INT64 aux;
		aux = aOperand;
		aux *= i;
		return ((aux<<1) % aModulus);
	}
	}
	//   Shanks - Tonelli
	int m,r,e,k=1;
	long n=aModulus>>1;
	INT64 a,b,x,q,y,t;
	q=aModulus-1;
	a=aOperand;
	for(e=0;!(q&1);e++) q=q>>1;
	while (k==1) {
	k=JacobiSymbol(--n,aModulus);
	}
	n=PowerModN(n,q,aModulus);
	y=n;
	r=e;
	x=PowerModN(a,(q-1)>>1,aModulus);
	b=(a*PowerModN(x,2,aModulus))%aModulus;
	x=(a*x)%aModulus;
	while (b!=1) {
	q=b;
	for(m=0;q!=1;m++) {
		q=(q*q) % aModulus;
	}
//	printf("x=%ld y=%ld z=%ld a=%ld b=%ld m=%d\n",x,y,z,a,b,m);
	if (m==r) return 0;
	t=PowerModN(y,1<<(r-m-1),aModulus);
	y=(t*t) % aModulus;
	r=m;
	x=(x*t) % aModulus;
	b=(b*y) % aModulus;
	}
	return (long) x;
}
int CIntegerClass::GetPrimes(const long& aMinBound, const long& aMaxBound, unsigned long& aCount, long*& aPrimes)
{
	int rc;
	if (aMinBound >= aMaxBound || aMaxBound <= 0)
	{
		return ConstRC::BadArgument;
	}
	aCount = 0;
	aPrimes = NULL;
	std::vector<long> primes;
	aCount = (long)(aMaxBound/log((double)aMaxBound));
	if (aMinBound > 1)
	{
		aCount -= (long)(aMinBound/log((double)aMinBound));
	}
	primes.reserve(aCount);
	aCount = 0;

	// control if exists
	xmlTextReaderPtr reader = xmlNewTextReaderFilename(ConstFN::FULL_PRIMES.c_str());

	if (reader == NULL || aMaxBound > CIntegerClass::GetPrimesMaxBound(reader))
	{
		// we create new file with primes because we don't have enought
		rc = CIntegerClass::GenerateNewPrimesFile(aMaxBound, reader);
		if (reader != NULL)
		{
			xmlFreeTextReader(reader);
		}
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		reader = xmlNewTextReaderFilename(ConstFN::FULL_PRIMES.c_str());
	}

	// move reader to first <prime> of the file
	xmlChar* name = xmlTextReaderName(reader);
	while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME) == 0)
	{
		if (NULL != name) xmlFree(name);
		rc = CommonXMLHelper::MoveToNextNode(reader);
		if (rc != ConstRC::Ok)
		{
			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		name = xmlTextReaderName(reader);
	}

	long auxPrime = 0;
	do 
	{
		if (NULL != name) xmlFree(name);
		name = NULL;
		rc = CommonXMLHelper::DeserializeSignedLong(reader, "Read as signed long ", auxPrime, FALSE);
		if (rc != ConstRC::Ok)
		{
			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		if (aMinBound <= auxPrime)
		{
			if (auxPrime >= aMaxBound)
			{
				// too big
				break;
			}

			aCount++;
			primes.push_back(auxPrime);
		}

		// next node
		rc = CommonXMLHelper::MoveToNextNode(reader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::EndOfFile)
			{
				// all primes
				break;
			}

			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		name = xmlTextReaderName(reader);
	}
	while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME));
	if (NULL != name) 
		xmlFree(name);

	if (aCount <= 0 || aCount != primes.size())
	{
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimes",0,0,"Primes not generated.");
		return ConstRC::GeneralError;
	}

	aPrimes = (long*)Utils::AllocationMachine((void**)&aPrimes, aCount * sizeof(long));
	if (aPrimes == NULL)
	{
		xmlFreeTextReader(reader);
		return ConstRC::NotEnoughMemory;
	}

	for (unsigned long i = 0; i < aCount; i++)
	{
		aPrimes[i] = primes[i];
	}

	xmlFreeTextReader(reader);
	return ConstRC::Ok;
}
int CIntegerClass::GetPrimesCount(const long& aMinBound, const long& aMaxBound, long& aCount)
{
	int rc;
	if (aMinBound >= aMaxBound || aMaxBound <= 0)
	{
		return ConstRC::BadArgument;
	}

	// control if exists
	xmlTextReaderPtr reader = xmlNewTextReaderFilename(ConstFN::FULL_PRIMES.c_str());

	if (reader == NULL || aMaxBound > CIntegerClass::GetPrimesMaxBound(reader))
	{
		// we create new file with primes because we don't have enought
		rc = CIntegerClass::GenerateNewPrimesFile(aMaxBound, reader);
		if (reader != NULL)
		{
			xmlFreeTextReader(reader);
		}
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		reader = xmlNewTextReaderFilename(ConstFN::FULL_PRIMES.c_str());
	}

	// move reader to first <prime> of the file
	xmlChar* name = xmlTextReaderName(reader);
	while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME) == 0)
	{
		if (NULL != name) xmlFree(name);
		rc = CommonXMLHelper::MoveToNextNode(reader);
		if (rc != ConstRC::Ok)
		{
			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		name = xmlTextReaderName(reader);
	}
	long auxPrime = 0;
	aCount = 0;
	do 
	{
		if (NULL != name) xmlFree(name);
		name = NULL;
		rc = CommonXMLHelper::DeserializeSignedLong(reader, "Read as signed long ", auxPrime, FALSE);
		if (rc != ConstRC::Ok)
		{
			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		if (aMinBound <= auxPrime)
		{
			if (auxPrime >= aMaxBound)
			{
				// too big
				break;
			}

			aCount++;
		}

		// next node
		rc = CommonXMLHelper::MoveToNextNode(reader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::EndOfFile)
			{
				// all primes
				break;
			}

			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		name = xmlTextReaderName(reader);
	}
	while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME));
	if (NULL != name)
		xmlFree(name);

	xmlFreeTextReader(reader);
	return ConstRC::Ok;
}
int CIntegerClass::GenerateNewPrimesFile(const long& aMaxBound, xmlTextReaderPtr& aReader)
{
	xmlTextWriterPtr writer = xmlNewTextWriterFilename(ConstFN::FULL_PRIMES_TEMP.c_str(), 0);
	int rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimesFile",rc,0,"Could not write the document start.");
		return ConstRC::GeneralError;        	
	}

	// main element <primes>
	rc = xmlTextWriterStartElement(writer, BAD_CAST XML_SERIAL_PRIMES_NAME);
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimesFile",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// write whole header
	rc = CIntegerClass::WritePrimesHeader(aMaxBound, writer);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// element <body>
	rc = xmlTextWriterStartElement(writer, BAD_CAST XML_SERIAL_BODY_NAME);
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimesFile",rc,0,"Could not write the body parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	long auxPrime = 1;

	// if we have reader (old primes file) we copy old prime to new file
	if (aReader != NULL)
	{
		// move reader to first <prime> of the file
		xmlChar* name = xmlTextReaderName(aReader);
		while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME) == 0)
		{
			if(NULL != name) {
				xmlFree(name);
				name = NULL;
			};
			rc = CommonXMLHelper::MoveToNextNode(aReader);
			if (rc != ConstRC::Ok)
			{
				xmlFreeTextWriter(writer);
				return rc;
			}
			name = xmlTextReaderName(aReader);
		}

		do 
		{
			if(NULL != name) {
				xmlFree(name);
				name = NULL;
			};
			rc = CommonXMLHelper::DeserializeSignedLong(aReader, "Read as signed long ", auxPrime, FALSE);
			if (rc != ConstRC::Ok)
			{
				xmlFreeTextWriter(writer);
				return rc;
			}
			
			rc = CommonXMLHelper::SerializeSignedLong(writer, XML_SERIAL_PRIME_NAME, auxPrime);
			if (rc != ConstRC::Ok)
			{
				xmlFreeTextWriter(writer);
				return rc;
			}

			// next node
			rc = CommonXMLHelper::MoveToNextNode(aReader);
			if (rc != ConstRC::Ok)
			{
				if (rc == ConstRC::EndOfFile)
				{
					// all primes
					break;
				}

				xmlFreeTextWriter(writer);
				return rc;
			}
			name = xmlTextReaderName(aReader);
		}
		while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME));
		if(NULL != name) {
				xmlFree(name);
				name = NULL;
			};
}

	mpz_t tested;
	mpz_init(tested);
	mpz_set_si(tested, auxPrime);

	while (TRUE)
	{
		mpz_nextprime(tested, tested);
		if (mpz_cmp_si(tested, aMaxBound) >= 0)
		{
			// too big
			break;
		}
		// additional control
		if (mpz_probab_prime_p(tested, 30)) 
		{
			auxPrime = mpz_get_si(tested);
			rc = CommonXMLHelper::SerializeSignedLong(writer, XML_SERIAL_PRIME_NAME, auxPrime);
			if (rc != ConstRC::Ok)
			{
				mpz_clear(tested);
				xmlFreeTextWriter(writer);
				return rc;
			}
		}
	}

	rc = xmlTextWriterEndElement(writer); // XML_SERIAL_BODY_NAME
	if (rc < 0) 
	{
		mpz_clear(tested);
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimesFile",rc,0,"Could not end the body parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		mpz_clear(tested);
		xmlFreeTextWriter(writer);
		return rc;
	}

	rc = xmlTextWriterEndElement(writer); // XML_SERIAL_PRIMES_NAME
	if (rc < 0) 
	{
		mpz_clear(tested);
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimesFile",rc,0,"Could not end the main parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		mpz_clear(tested);
		xmlFreeTextWriter(writer);
		return rc;
	}

	mpz_clear(tested);
	xmlFreeTextWriter(writer);

	if (aReader != NULL)
	{
		// we delete old file
		xmlFreeTextReader(aReader);
		aReader = NULL;
		remove(ConstFN::FULL_PRIMES.c_str());
	}

	rename(ConstFN::FULL_PRIMES_TEMP.c_str(), ConstFN::FULL_PRIMES.c_str());

	return ConstRC::Ok;
}
long CIntegerClass::GetPrimesMaxBound(xmlTextReaderPtr aReader)
{
	if (aReader == NULL)
	{
		return 0;
	}

	long bound = 0;
	// control main node
	int rc = CommonXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		// something wrong - we remove the prime file and create it again later
		remove(ConstFN::FULL_PRIMES.c_str());
		return bound;
	}

	xmlChar* name = xmlTextReaderName(aReader);
	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_PRIMES_NAME) == 0)
	{
		// something wrong - we remove the prime file and create it again later
		remove(ConstFN::FULL_PRIMES.c_str());
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimesMaxBound",0,0,"Main element name is wrong.");
		xmlFree(name);
		return bound;
	}

	if (NULL != name) xmlFree(name);
	// control header node
	rc = CommonXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		// something wrong - we remove the prime file and create it again later
		remove(ConstFN::FULL_PRIMES.c_str());
		return bound;
	}

	name = xmlTextReaderName(aReader);
	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_HEADER_NAME) == 0)
	{
		// something wrong - we remove the prime file and create it again later
		remove(ConstFN::FULL_PRIMES.c_str());
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimesMaxBound",0,0,"Header element name is wrong.");
		if(NULL != name) xmlFree(name);
		return bound;
	}	
	if(NULL != name) xmlFree(name);

	// get max bound
	rc = CommonXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		// something wrong - we remove the prime file and create it again later
		remove(ConstFN::FULL_PRIMES.c_str());
		return bound;
	}

	name = xmlTextReaderName(aReader);
	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MAX_BOUND_NAME) == 0)
	{
		// something wrong - we remove the prime file and create it again later
		remove(ConstFN::FULL_PRIMES.c_str());
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimesMaxBound",0,0,"MaxBound element name is wrong.");
		xmlFree(name);
		return bound;
	}

	CommonXMLHelper::DeserializeSignedLong(aReader, "Long read as ", bound, FALSE);
	if (NULL != name) xmlFree(name);

	return bound;
}
int CIntegerClass::WritePrimesHeader(const long& aMaxBound, xmlTextWriterPtr aWriter)
{
	// <header> element
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_HEADER_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CIntegerClass","WritePrimesHeader",rc,0,"Could not write the header parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CommonXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_MAX_BOUND_NAME, aMaxBound);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_HEADER_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CIntegerClass","WritePrimesHeader",rc,0,"Could not end the header parameter tag.");
		return ConstRC::GeneralError;
	}

	return CommonXMLHelper::NewLine(aWriter,FALSE);
}
long CIntegerClass::GetPrimitiveRootsMaxBound(xmlTextReaderPtr aReader)
{
	if (aReader == NULL)
	{
		return 0;
	}

	long bound = 0;
	// control main node
	int rc = CommonXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		// something wrong - we remove the primitive root file and create it again later
		remove(ConstFN::FULL_ROOT_DATA.c_str());
		return bound;
	}

	xmlChar* name = xmlTextReaderName(aReader);
	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_ROOT_DATA_NAME) == 0)
	{
		// something wrong - we remove the primitive root file and create it again later
		remove(ConstFN::FULL_ROOT_DATA.c_str());
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimitiveRootsMaxBound",0,0,"Main element name is wrong.");
		if (NULL != name) xmlFree(name);
		return bound;
	}
	if (NULL != name) xmlFree(name);

	// control header node
	rc = CommonXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		// something wrong - we remove the primitive root file and create it again later
		remove(ConstFN::FULL_ROOT_DATA.c_str());
		return bound;
	}

	name = xmlTextReaderName(aReader);
	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_HEADER_NAME) == 0)
	{
		// something wrong - we remove the primitive root file and create it again later
		remove(ConstFN::FULL_ROOT_DATA.c_str());
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimitiveRootsMaxBound",0,0,"Header element name is wrong.");
		if (NULL != name) xmlFree(name);
		return bound;
	}	
	if (NULL != name) xmlFree(name);

	// get max bound
	rc = CommonXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		// something wrong - we remove the primitive root file and create it again later
		remove(ConstFN::FULL_ROOT_DATA.c_str());
		return bound;
	}

	name = xmlTextReaderName(aReader);
	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MAX_BOUND_NAME) == 0)
	{
		// something wrong - we remove the primitive root file and create it again later
		remove(ConstFN::FULL_ROOT_DATA.c_str());
		CLog::LogError(WHEREARG,"CIntegerClass","GetPrimitiveRootsMaxBound",0,0,"MaxBound element name is wrong.");
		if (NULL != name) xmlFree(name);
		return bound;
	}
	if (NULL != name) xmlFree(name);

	CommonXMLHelper::DeserializeSignedLong(aReader, "Long read as ", bound, FALSE);

	return bound;
}
int CIntegerClass::WritePrimitiveRootsHeader(const long& aMaxBound, xmlTextWriterPtr aWriter)
{
	// <header> element
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_HEADER_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CIntegerClass","WritePrimitiveRootsHeader",rc,0,"Could not write the header parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CommonXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_MAX_BOUND_NAME, aMaxBound);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_HEADER_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CIntegerClass","WritePrimitiveRootsHeader",rc,0,"Could not end the header parameter tag.");
		return ConstRC::GeneralError;
	}

	return CommonXMLHelper::NewLine(aWriter,FALSE);
}
int CIntegerClass::GenerateNewPrimitiveRootsFile(const long& aMaxBound)
{
	xmlTextWriterPtr writer = xmlNewTextWriterFilename(ConstFN::FULL_ROOT_DATA.c_str(), 0);
	int rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not write the document start.");
		return ConstRC::GeneralError;        	
	}

	// main element <root-data>
	rc = xmlTextWriterStartElement(writer, BAD_CAST XML_SERIAL_ROOT_DATA_NAME);
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// write whole header
	rc = CIntegerClass::WritePrimitiveRootsHeader(aMaxBound, writer);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// at first generate primitive roots

	// element <primitive-roots>
	rc = xmlTextWriterStartElement(writer, BAD_CAST XML_SERIAL_PRIMITIVE_ROOTS_NAME);
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	std::vector<CNumber> field(aMaxBound + 1);	

	// generate primitive roots
	CPrimitiveRoot * generator = new CPrimitiveRoot(aMaxBound);
	generator->CPrimitiveRootGenerator(field);

	int i, j;
	for (i = 2; i <= aMaxBound; i++)
	{
		if (field[i].Prime())
		{
			CommonXMLHelper::SerializePrimitiveRoot(writer, XML_SERIAL_PRIMITIVE_ROOT_NAME, field[i].Primit(), field[i].Value());
		}
	}

	rc = xmlTextWriterEndElement(writer); // XML_SERIAL_PRIMITIVE_ROOTS_NAME
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not end the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// element <powers>
	rc = xmlTextWriterStartElement(writer, BAD_CAST XML_SERIAL_POWERS_NAME);
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// generate powers

	int index;
	int base;
	int prime;
	std::vector<int> powers(aMaxBound + 1);
	for (i = 2; i <= aMaxBound; i++)
	{
		if (field[i].Prime())
		{
			base = field[i].Primit();
			prime = field[i].Value();
			index = base;
			for (j = 0; j < prime - 1; j++)		//filling of array of all numbers with their exponent
			{
				powers[index] = j + 1;
				index = (index * base) % prime;
			}
			for (j = 1; j < prime; j++)	
			{
				CommonXMLHelper::SerializePowerModP(writer, XML_SERIAL_POWER_NAME, prime, j, powers[j]);
			}
		}
	}

	rc = xmlTextWriterEndElement(writer); // XML_SERIAL_POWERS_NAME
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not end the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	rc = xmlTextWriterEndElement(writer); // XML_SERIAL_ROOT_DATA_NAME
	if (rc < 0) 
	{
		xmlFreeTextWriter(writer);
		CLog::LogError(WHEREARG,"CIntegerClass","GenerateNewPrimitiveRootsFile",rc,0,"Could not end the main parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	xmlFreeTextWriter(writer);

	return ConstRC::Ok;
}
int CIntegerClass::GetPrimitiveRootsAndPowers(int * aSetP, int aSetPSize, int aMaxBound, int * aPrimitiveRoots, int ** aPowers)
{
	int rc;
	// control if exists
	xmlTextReaderPtr reader = xmlNewTextReaderFilename(ConstFN::FULL_ROOT_DATA.c_str());

	if (reader == NULL || aMaxBound > CIntegerClass::GetPrimitiveRootsMaxBound(reader))
	{
		// we create new file with primitive roots because we don't have enought
		rc = CIntegerClass::GenerateNewPrimitiveRootsFile(aMaxBound);
		if (reader != NULL)
		{
			xmlFreeTextReader(reader);
		}
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		reader = xmlNewTextReaderFilename(ConstFN::FULL_ROOT_DATA.c_str());
	}

	// move reader to <primitive-roots> of the file
	xmlChar* name = xmlTextReaderName(reader);
	while (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIMITIVE_ROOTS_NAME) == 0)
	{
		if (NULL != name) xmlFree(name);
		rc = CommonXMLHelper::MoveToNextNode(reader);
		if (rc != ConstRC::Ok)
		{
			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		name = xmlTextReaderName(reader);
	}

	int index = 0;
	int prime, root, power, number;
	int * currSetP = aSetP;
	int * currPrimitiveElements = aPrimitiveRoots;

	// get primitive roots
	int depth = xmlTextReaderDepth(reader);
	while ((rc = CommonXMLHelper::MoveToNextNode(reader, depth)) == ConstRC::Ok)
	{
		if (NULL != name) xmlFree(name);
		name = xmlTextReaderName(reader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIMITIVE_ROOT_NAME))
		{
			rc = CommonXMLHelper::DeserializePrimitiveRoot(reader, "Deserialize primitive root...", root, prime, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			if (*currSetP == prime)
			{
				*currPrimitiveElements = root;
				currSetP++;
				currPrimitiveElements++;
				index++;
			}
			if (index >= aSetPSize)
			{
				// to much
				break;
			}
			continue;
		}
	}
	if (NULL != name) xmlFree(name);

	// move reader to <powers> of the file
	name = xmlTextReaderName(reader);
	while (xmlStrEqual(name, BAD_CAST XML_SERIAL_POWERS_NAME) == 0)
	{
		if (NULL != name) xmlFree(name);
		rc = CommonXMLHelper::MoveToNextNode(reader);
		if (rc != ConstRC::Ok)
		{
			// something wrong
			xmlFreeTextReader(reader);
			return rc;
		}
		name = xmlTextReaderName(reader);
	}

	index = 0;
	currSetP = aSetP;

	// get powers
	depth = xmlTextReaderDepth(reader);
	while ((rc = CommonXMLHelper::MoveToNextNode(reader, depth)) == ConstRC::Ok)
	{
		if (NULL != name) xmlFree(name);
		name = xmlTextReaderName(reader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_POWER_NAME))
		{
			rc = CommonXMLHelper::DeserializePowerModP(reader, "Deserialize power mod p...", prime, number, power, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			// move to next row (next r) if prime is bigger than current r from P
			if (*currSetP < prime)
			{
				currSetP++;
				index++;
			}
			if (*currSetP == prime)
			{
				aPowers[index][number] = power;
			}
			if (index >= aSetPSize)
			{
				// to much
				break;
			}
			continue;
		}
	}
	if (NULL != name) xmlFree(name);
	
	xmlFreeTextReader(reader);
	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;

}
