#include <cstring>
#include "integer_matrix_class.h"
#include "nfs_utils.h" // because of allocation machine
#include "integer_class.h"
#include "const_return_codes.h"
#include "xml_helper.h"
#include "log.h"

IntegerMatrix::IntegerMatrix() : MSerializable(FILE_NAME_NORMAL_MATRIX),NormalMatrix()
{
}


IntegerMatrix::IntegerMatrix(long aRows, long aColumns) : MSerializable(FILE_NAME_NORMAL_MATRIX),NormalMatrix(aRows, aColumns)
{
}

/*!
	Converts the matrix into the Hermite normal form.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>
*/

int IntegerMatrix::ToHermiteNormalForm(){
    long i,j,k,a,b,c=maximal_column_index;
    INT64 d,e,f;
    integer_matrix_type aux;
    for(i=maximal_row_index;i>=0;i--){
	for(j=0;j<c;j++) if (matrix[i][j]) {
	    if(!matrix[i][j+1]) SwapColumns(j,j+1);
	    else {
		d=CIntegerClass::ExtendedEuclid(&a,&b,matrix[i][j],matrix[i][j+1]);
		e=matrix[i][j+1]/d;
		f=matrix[i][j]/d;
		for(k=0;k<=i;k++){
		    aux=matrix[k][j];
		    matrix[k][j]=matrix[k][j]*e-matrix[k][j+1]*f;
		    matrix[k][j+1]=aux*(INT64)a+matrix[k][j+1]*(INT64)b;
		}
	    }
	}
	if (matrix[i][c]){
	    if(matrix[i][c]<0) {
		for(k=0;k<=i;k++) matrix[k][c] = -matrix[k][c];
	    }
	    for(j=c+1;j<=maximal_column_index;j++){
		d=matrix[i][j] / matrix[i][c];
		if (matrix[i][c]*d>matrix[i][j]) d--;
		for(k=0;k<=i;k++){
		    matrix[k][j] -= matrix[k][c] * d;
		}
	    }
	    c--;
	}
    }
    return ConstRC::Ok;
}

/*!
	Converts the matrix into the Hermite normal form when a multiple of the determinant is known.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>
*/

int IntegerMatrix::ToHermiteNormalForm(long aModulus){
    long R,i,j,k,a,b,c=maximal_column_index;
    INT64 d,e,f;
    R=aModulus;
    integer_matrix_type aux;
    for(i=maximal_row_index;i>=0;i--){
	if (!matrix[i][c]) matrix[i][c]=R;
	for(j=0;j<c;j++) if (matrix[i][j]) {
	    if(!matrix[i][j+1]) SwapColumns(j,j+1);
	    else {
		d=CIntegerClass::ExtendedEuclid(&a,&b,matrix[i][j],matrix[i][j+1]);
		e=matrix[i][j+1]/d;
		f=matrix[i][j]/d;
		for(k=0;k<=i;k++){
		    aux=matrix[k][j];
		    matrix[k][j]=(matrix[k][j]*e-matrix[k][j+1]*f)%R;
		    if (matrix[k][j]<<1 > R) matrix[k][j]-=R;
		    if ((-matrix[k][j])<<1 >= R) matrix[k][j]+=R;
		    matrix[k][j+1]=(aux*(INT64)a+matrix[k][j+1]*(INT64)b)%R;
		    if (matrix[k][j+1]<<1 > R) matrix[k][j+1]-=R;
		    if ((-matrix[k][j+1])<<1 >= R) matrix[k][j+1]+=R;
		}
	    }
//	    PrintToScreen();
	}
	d=CIntegerClass::ExtendedEuclid(&a,&b,matrix[i][c],R);
//	printf("%ld %ld %ld %ld\n",a,b,R,d);
	if (!a) a=1;
        for(k=0;k<=i;k++){
//	    printf("(%ld * %lld)mod %ld=",a,matrix[i][k],R);
	    matrix[k][c]=((INT64)a*matrix[k][c])%R;
	    if (matrix[i][k]<0) matrix[i][k]+=R;
//	    printf("%ld\n",matrix[i][k]);
	}
//	    PrintToScreen();
	if (!matrix[i][c]) matrix[i][c]=R;
        for(j=c+1;j<=maximal_column_index;j++){
	   e=matrix[i][j] / matrix[i][c];
	   if(matrix[i][c]*e>matrix[i][j]) e--;
	   for(k=0;k<=i;k++){
		matrix[k][j] -= (matrix[k][c] * e);
		matrix[k][j] %= R;
	   }
        }
        c--;
        R/=d;
    }
    return ConstRC::Ok;
}

/*!
	Returns the kernel of given homomorphism having the dimension aDimension. The computation is made modulo aModulus

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int IntegerMatrix::KernelModPrime(IntegerMatrix *aTarget, long *aDimension, long aModulus, bool aOnMyself){
    if (aModulus<=0)   return ConstRC::GeneralError;
    if (aTarget->maximal_row_index!=this->maximal_column_index)	    return ConstRC::SizeMismatch;
    if (aTarget->maximal_column_index!=this->maximal_column_index)  return ConstRC::SizeMismatch;
    IntegerMatrix *M;
    int ret;
    if (aOnMyself) M=this;
    else {
	M=new IntegerMatrix(this->maximal_row_index+1,this->maximal_column_index+1);
	ret = M->Allocate();
	if (ret!=ConstRC::Ok) return ret;
	M->Copy(this);
    }
    M->Modulo(aModulus);
    long *c,*d,r,k,i,j,s;
    INT64 aux,q;
    c = (long*)malloc((maximal_row_index+1)*sizeof(long));
    if (c==NULL){
	CLog::LogError(WHEREARG,"IntegerMatrixClass","error allocating auxiliary array",(this->max_allocated_column_index+1),501,"There was not enough memory to allocate an auxiliary array");
	return ConstRC::NotEnoughMemory;
    }
    d = (long*)malloc((maximal_column_index+1)*sizeof(long));
    if (d==NULL){
	CLog::LogError(WHEREARG,"IntegerMatrixClass","error allocating auxiliary array",(this->max_allocated_column_index+1),502,"There was not enough memory to allocate an auxiliary array");
	free(c);
	return ConstRC::NotEnoughMemory;
    }
    r=0;
    for(i=0;i<=maximal_row_index;i++) c[i]=-1;
    for(k=0;k<=maximal_column_index;k++){
	for(j=0;j<=maximal_row_index;j++)
	    if((M->matrix[j][k]!=0)&&(c[j]==-1)) break;
	if (j<=maximal_row_index) {
	    q=-CIntegerClass::InverseModN(M->matrix[j][k],aModulus);
	    M->matrix[j][k]=-1;
	    for(s=k+1;s<=maximal_column_index;s++){
		aux = M->matrix[j][s];
		aux *= q;
		M->matrix[j][s]= aux % aModulus;
	    }
	    for(i=0;i<=maximal_row_index;i++) if(i!=j) {
		q=M->matrix[i][k];
		M->matrix[i][k]=0;
		for(s=k+1;s<=maximal_column_index;s++){
		    aux = M->matrix[i][s];
		    aux += (INT64)q*(INT64)M->matrix[j][s];
		    M->matrix[i][s]= aux % aModulus;
		}
	    }
	    c[j]=k;
	    d[k]=j;
	} else {
	    r++;
	    d[k]=-1;
	}
    }
    aTarget->Zeroize();
    s=0;
    for(k=0;k<=maximal_column_index;k++) if(d[k]==-1){
	for(i=0;i<=maximal_column_index;i++){
	    if (d[i]>=0) aTarget->matrix[i][s]=M->matrix[d[i]][k];
	    else if (i==k) aTarget->matrix[i][s]=1;
	}
	s++;
    }
    *aDimension=r;
    if (!aOnMyself) delete(M);
    free(c);
    free(d);
    return ConstRC::Ok;
}

/*!
	Returns a base of the image of given homomorphism having the dimension aDimension. The computation is made modulo aModulus

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int IntegerMatrix::ImageModPrime(IntegerMatrix *aTarget, long *aDimension, long aModulus){
    if (aModulus<=0)   return ConstRC::GeneralError;
    if (aTarget->maximal_row_index!=this->maximal_row_index)	    return ConstRC::SizeMismatch;
    int ret;
    IntegerMatrix *M;
    {
	M=new IntegerMatrix(this->maximal_row_index+1,this->maximal_column_index+1);
	ret = M->Allocate();
	if (ret!=ConstRC::Ok) return ret;
	M->Copy(this);
    }
    M->Modulo(aModulus);
    long *c,r,k,i,j,s;
    INT64 aux,q;
    c = (long*)malloc((maximal_row_index+1)*sizeof(long));
    if (c==NULL){
	CLog::LogError(WHEREARG,"IntegerMatrixClass","error allocating auxiliary array",(this->max_allocated_column_index+1),501,"There was not enough memory to allocate an auxiliary array");
	return ConstRC::NotEnoughMemory;
    }
    r=0;
    for(i=0;i<=maximal_row_index;i++) c[i]=-1;
    for(k=0;k<=maximal_column_index;k++){
	for(j=0;j<=maximal_row_index;j++)
	    if((M->matrix[j][k]!=0)&&(c[j]==-1)) break;
	if (j<=maximal_row_index) {
	    q=-CIntegerClass::InverseModN(M->matrix[j][k],aModulus);
	    M->matrix[j][k]=-1;
	    for(s=k+1;s<=maximal_column_index;s++){
		aux = M->matrix[j][s];
		aux *= q;
		M->matrix[j][s]= aux % aModulus;
	    }
	    for(i=0;i<=maximal_row_index;i++) if(i!=j) {
		q=M->matrix[i][k];
		M->matrix[i][k]=0;
		for(s=k+1;s<=maximal_column_index;s++){
		    aux = M->matrix[i][s];
		    aux += q*M->matrix[j][s];
		    M->matrix[i][s] = aux % aModulus;
		}
	    }
	    c[j]=k;
	} else {
	    r++;
	}
    }
    M->Copy(this);
    aTarget->Zeroize();
    s=0;
    for(k=0;k<=maximal_row_index;k++) if(c[k]!=-1){
	for(i=0;i<=aTarget->maximal_column_index;i++){
	    aTarget->matrix[i][s]=M->matrix[i][c[k]];
	}
	s++;
    }
    *aDimension=s;
    delete(M);
    free(c);
    return ConstRC::Ok;
}
/*!
	Returns a supplement - the first aDimension vectors are extented into a base of the space. The computation is made modulo aModulus

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
		<tr><td>ConstRC::GeneralError </td><td> the vectors given are not linearly independent</td></tr>
	  </table>
*/

int IntegerMatrix::SupplementModPrime(IntegerMatrix *aTarget, long aDimension, long aModulus){
    if (aModulus<=0)   return ConstRC::GeneralError;
    if (aTarget->maximal_row_index!=this->maximal_row_index)	    return ConstRC::SizeMismatch;
    if (aTarget->maximal_column_index!=this->maximal_row_index)	    return ConstRC::SizeMismatch;
    IntegerMatrix *M;
    int ret;
    M=new IntegerMatrix(this->maximal_row_index+1,this->maximal_column_index+1);
    ret = M->Allocate();
    if (ret!=ConstRC::Ok) return ret;
    M->Copy(this);
    M->Modulo(aModulus);
    long k,i,j,q,s,l=0;
    INT64 aux;
    ret = aTarget->Copy(this);
    if (ret!=ConstRC::Ok) return ret;
    for(k=0;k<=maximal_row_index;k++){
	for(s=k-l;s<aDimension;s++){
	    if(M->matrix[k][s]!=0) break;
	}
	if (s<aDimension) {
	    q=-CIntegerClass::InverseModN(M->matrix[k][s],aModulus);
	    for(i=0;i<=maximal_row_index;i++){
		aux=q*M->matrix[i][s];
		if (k!=s) M->matrix[i][s]=M->matrix[i][k];
		M->matrix[i][k] = aux % aModulus;
	    }
	    for(j=k+1;j<aDimension;j++){
		for(i=0;i<=maximal_row_index;i++) if(i!=k) {
		    aux = M->matrix[i][j];
		    aux += (INT64)M->matrix[i][k]*(INT64)M->matrix[k][j];
		    M->matrix[i][j]= aux % aModulus;
		}
		M->matrix[k][j]=0;
	    }
	} else {
	    if (l+aDimension>maximal_column_index) return ConstRC::GeneralError;
	    aTarget->matrix[k][aDimension+l++]=1;
	}
    }
    delete(M);
    return ConstRC::Ok;
}

/*!
	Returns the inverse matrix of the matrix (modulo aPrime) which is supposed to be upper triangular.
	The function checks sizes of matrices

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
	  </table>
*/

int IntegerMatrix::InverseFromTriangularModPrime(IntegerMatrix *aTarget, long aPrime){
    if (maximal_row_index!=maximal_column_index)    return ConstRC::SizeMismatch;
    if (maximal_column_index!=aTarget->maximal_column_index) return ConstRC::SizeMismatch;
    if (maximal_row_index!=aTarget->maximal_row_index)	return ConstRC::SizeMismatch;
    int ret=aTarget->Zeroize();
    if (ret!=ConstRC::Ok) return ret;
    long i,j,k;
    INT64 a,b;
    for(i=maximal_row_index;i>=0;i--){
	aTarget->matrix[i][i]=CIntegerClass::InverseModN(matrix[i][i],aPrime);
    }
    for(i=maximal_row_index;i>=0;i--){
	for(j=0;j<i;j++){
	    a=-matrix[j][i];
	    a *= aTarget->matrix[j][j];
	    for(k=i;k<=maximal_column_index;k++){
		b = a * aTarget->matrix[i][k];
		b += aTarget->matrix[j][k];
		aTarget->matrix[j][k]= b % aPrime;
	    }
	}
    }
    return ConstRC::Ok;
}

/*!
	Returns the inverse image of the given homomorphism (which is supposed to be regular). The computation is made modulo aPrime

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int IntegerMatrix::InverseImageOfMatrixModPrime(IntegerMatrix *aTarget, IntegerMatrix *aImage, long aPrime, bool aOnMyself)
{
    if (aPrime<=0)   return ConstRC::GeneralError;
    if (aTarget->maximal_row_index!=this->maximal_column_index)	    return ConstRC::SizeMismatch;
    if (aTarget->maximal_column_index!=aImage->maximal_column_index)	return ConstRC::SizeMismatch;
    if (this->maximal_row_index!=this->maximal_column_index)	    return ConstRC::SizeMismatch;
    IntegerMatrix *M;
    int ret;
    if (aOnMyself) {
	M=this;
    }
    else {
	M = new IntegerMatrix(this->maximal_row_index+1,this->maximal_column_index+1);
	ret = M->Allocate();
	if (ret!=ConstRC::Ok) return ret;
	M->Copy(this);
    }
    M->Modulo(aPrime);
    long k,i,j,l;
    INT64 a,q,c;
    ret = aTarget->Copy(aImage);
    for(j=0;j<=maximal_column_index;j++){
	for(i=j;(i<=maximal_row_index)&&(!M->matrix[i][j]);i++);
	if (i>maximal_row_index) {
		if (!aOnMyself) delete(M);	
		return ConstRC::GeneralError;
	}
	if (i!=j){
	    M->SwapRows(i,j);
	    aTarget->SwapRows(i,j);
	}
	q = CIntegerClass::InverseModN(M->matrix[j][j],aPrime);
	for(l=j+1;l<=maximal_column_index;l++){
	    a = q * M->matrix[j][l];
	    M->matrix[j][l] = a % aPrime;
	}
	for(l=0;l<=aTarget->maximal_column_index;l++){
	    a = q * aTarget->matrix[j][l];
	    aTarget->matrix[j][l] = a % aPrime;
	}
	M->matrix[j][j]=1;
	for(k=maximal_row_index;k>j;k--){
	    c = M->matrix[k][j];
	    for(l=maximal_column_index;l>=j;l--){
		a = M->matrix[k][l];
		a -= c*M->matrix[j][l];
		M->matrix[k][l] = a % aPrime;
	    }
	    for(l=aTarget->maximal_column_index;l>=0;l--){
		a = aTarget->matrix[k][l];
		a-= c*aTarget->matrix[j][l];
		aTarget->matrix[k][l] = a%aPrime;
	    }
	}
    }
    for(j=maximal_column_index;j>=0;j--){
	for(i=0;i<j;i++){
	    c = -M->matrix[i][j];
	    for(l=0;l<=aTarget->maximal_column_index;l++){
		a = c*aTarget->matrix[j][l];
		a+= aTarget->matrix[i][l];
		aTarget->matrix[i][l] = a % aPrime;
	    }
	}
    }
    if (!aOnMyself) delete(M);
    return ConstRC::Ok;
}
	
/*!
	Multiplies the matrix by a column vector from the right. The vectors are supposed to be
	sufficiently allocated.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> a vector is NULL</td></tr>
	  </table>
*/


int IntegerMatrix::MultiplyVector(long *aTarget, long *aSource){
    long i,j;
    if ((!aTarget)||(!aSource)) return ConstRC::NullPointerSupplied;
    integer_matrix_type c;
    for(i=0;i<=maximal_row_index;i++){
	c=0;
	for(j=0;j<=maximal_column_index;j++)
	    if(aSource[j]){
	    c+=matrix[i][j]*aSource[j];
	}
	aTarget[i]=(long) c;
    }
    return ConstRC::Ok;
}

/*!
	Multiplies the matrix by a vector from the left. The vectors are supposed to be
	sufficiently allocated.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> a vector is NULL</td></tr>
	  </table>
*/

int IntegerMatrix::VectorMultiply(long *aTarget, long *aSource){
    long i,j;
    if ((!aTarget)||(!aSource)) return ConstRC::NullPointerSupplied;
    integer_matrix_type c;
    for(i=0;i<=maximal_column_index;i++){
	c=0;
	for(j=0;j<=maximal_row_index;j++)
	    if(aSource[j]){
	    c+=matrix[j][i]*aSource[j];
	}
	aTarget[i]=(long) c;
    }
    return ConstRC::Ok;
}
//
//int IntegerMatrix::Serialize(const CBaseParameters &aParam) const
//{
//	return MSerializable::Serialize(aParam);
//}
//
//int IntegerMatrix::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
//{
//	return AbstractMatrix::Serialize(aParam, aWriter);
//}
//
//int IntegerMatrix::Deserialize(const CBaseParameters &aParam)
//{
//	return MSerializable::Deserialize(aParam);
//}
//
//int IntegerMatrix::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
//{
//	return AbstractMatrix::Deserialize(aParam, aReader);
//}
		
int IntegerMatrix::WriteData(xmlTextWriterPtr aWriter) const
{
	int rc;
	// write all matrix data
	for (long i = 0; i <= maximal_row_index; i++)
	{
		rc = CXMLHelper::SerializeBinaryData(aWriter, XML_SERIAL_MATRIX_ROW_NAME, this->matrix[i], this->maximal_column_index + 1, sizeof(integer_matrix_type));  
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

int IntegerMatrix::ReadData(xmlTextReaderPtr aReader)
{
	int rc;
	xmlChar * name;

	// we know max_row_index and max_column_index
	Allocate();
	long i;
	// we deserialize every row and get max nonzero columns in row and nonzero indexies
    for (i = 0; i <= this->maximal_row_index; i++)
	{
		name = xmlTextReaderName(aReader);

		if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MATRIX_ROW_NAME) == 0)
		{
			break;
		}

		rc = CXMLHelper::DeserializeBinaryDataIn(aReader, "Deserialize matrix row...", this->matrix[i], this->maximal_column_index + 1, sizeof(integer_matrix_type), FALSE);
        if (rc != ConstRC::Ok)
		{
			return rc;
		}

		rc = CXMLHelper::MoveToNextNode(aReader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::EndOfFile)
			{
				return ConstRC::Ok;
			}
			return rc;
		}
    }

	return ConstRC::Ok;
}
