#include "integer_sparse_matrix.h"
#include "common_utils.h"
#include <cstdlib>
#include <time.h>
#include "log.h"
#include <iostream>


/*!
	The default constructor does not take any parameters and constructs an instance of a "generic" bit matrix,
	with unknown dimensions. The dimensions may be later set by appropriate setter methods.

*/
IntegerSparseMatrix::IntegerSparseMatrix() : MSerializable(FILE_NAME_INTEGER_SPARSE_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->maximal_row_index = -1;
	this->maximal_column_index = -1;
	this->matrix = NULL;
	mpz_init(tempSum);
	mpz_init(tempMul);
}

/*!
The second constructor constructs an instance of a sparse integer matrix with known dimensions. Beware that the parameters
	taken mean the actual number of rows and columns, and not their maximal indices; so, if we want to construct
	a matrix of dimensions 17x32, we call

		IntegerSparseMatrix* sm = new IntegerSparseMatrix(17,32);

	Now, we have an instance of a sparse integer matrix; its member variables will be set to:

		sm->maximal_row_index = 16;

		sm->maximal_column_index = 31;

		sm->maximal_allocated_row_index = -1; // no initialization of the data array yet!

	Both of the constructors initialize the data array to NULL pointer, and a real allocation is performed later
	- at the time of need. This programming pattern is called lazy initialization, and helps to reduce runtime
	memory requirements.

*/
IntegerSparseMatrix::IntegerSparseMatrix(long aRows, long aColumns) : MSerializable(FILE_NAME_INTEGER_SPARSE_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->maximal_row_index = aRows - 1;
	this->maximal_column_index = aColumns - 1;
	this->matrix = NULL;
	mpz_init(tempSum);
	mpz_init(tempMul);
}

/*!

The destructor performs "cleaning up", in this case deallocation of the data array.
*/
IntegerSparseMatrix::~IntegerSparseMatrix(){
	if (max_allocated_row_index >= 0){
		long i;
		for (i=0; i<=max_allocated_row_index; i++){
			free(this->matrix[i]);
			this->matrix[i] = NULL;
		}
		max_allocated_row_index = -1;
		free(this->matrix);
		free(this->auxRow);
		this->matrix = NULL;
		mpz_clear(tempSum);
		mpz_clear(tempMul);
	}
}

/*!
	  This method takes care of allocation of the internal two-dimensional array of sparse_matrix_type, which contains the
	  matrix entries. It tests the need for allocation by evaluating
	  	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1))
	  condition.

	  If there is decision to run the allocation job, also zeroizing of the result matrix takes place. There are
	  (this->maximal_row_index+1) rows allocated (all rows), each of them with INTEGER_ALLOCATED(0) entries. The entry on index
	  0 is initialized as the row index, the entry on index 1 is initialized as 0.

	  Return codes:

	  	ConstRC::Ok		  - everything all right

		ConstRC::NotEnoughMemory - unsuccessful allocation of the rows and/or columns

*/
int IntegerSparseMatrix::Allocate(){
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){
		long i;
		long max;
		this->matrix = (integer_matrix_type **) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(integer_matrix_type**));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"IntegerSparseMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate sparse matrix rows");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->matrix,0,(this->maximal_row_index+1)*sizeof(integer_matrix_type*));
		this->max_allocated_row_index = this->maximal_row_index;
		max = INTEGER_ALLOCATED(0);
		for (i=0; i<= maximal_row_index; i++){
			this->matrix[i] = (integer_matrix_type *) Utils::AllocationMachine(NULL,max*sizeof(integer_matrix_type));
			if (!this->matrix){
				CLog::LogError(WHEREARG,"IntegerSparseMatrixClass","error allocating columns",max,201,"There was not enough memory to allocate sparse matrix columns");
				return ConstRC::NotEnoughMemory;
			}
			memset(this->matrix[i],0,max*sizeof(integer_matrix_type));
			this->matrix[i][0] = i;
			this->matrix[i][1] = -1;
			// cerr << "\n:" << max <<":"<< i;
		}
		this->auxRow = (integer_matrix_type *) Utils::AllocationMachine(NULL,maximal_row_index+3*sizeof(integer_matrix_type));
	}
	return ConstRC::Ok;
}

IntegerSparseMatrix * IntegerSparseMatrix::clone() {
	IntegerSparseMatrix * new_matrix = new IntegerSparseMatrix();
	return new_matrix;
}

/*!
	Search row index in sparse matrix representation. Return the actual index (if exists) or the future succeeding index.

*/
long IntegerSparseMatrix::BinarySearchRow(long aRow) {
	if ((aRow > this->maximal_row_index) ||(aRow < 0))	return ConstRC::ExcessiveRowIndex;
	long upper,lower, intermediate, max_index = this->max_allocated_row_index;
	lower = 0;
	upper = max_index;
	// binary search
	while ((upper-lower)>1){
		intermediate = lower+(upper-lower)/2;
		if (this->matrix[intermediate][0] <= aRow){
			lower = intermediate;
		}
		else{
			upper = intermediate;
		}
	}
	if (this->matrix[lower][0] == aRow){
		return lower;
	}
	else{
		return upper;
	}
}

long IntegerSparseMatrix::BinarySearchColumn(long aRowIndex, long aColumn) {
	// cout << "<+"<<aRowIndex <<"+"<<aColumn<<"+>";
	if (aRowIndex > this->max_allocated_row_index)		return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)		return ConstRC::ExcessiveColumnIndex;
	long	 k,p,s;
	if (this->matrix[aRowIndex][1] == -1) return 0;
	int max_index = this->matrix[aRowIndex][1]*2+2;
	p = 2;
	k = max_index;
	while ((k-p)>2){
		// cout << "(" << p << "-" << k <<")";
		s = p+((k-p)/4)*2;
		if (this->matrix[aRowIndex][s] <= aColumn){
			p = s;
		}
		else{
			k = s;
		}
	}
	if (this->matrix[aRowIndex][k] == aColumn)
		return (k/2)-1;
	if (this->matrix[aRowIndex][p] >= aColumn){
		return (p/2)-1;
	}
	else  {
		return (k/2)-1;
	}
}

int IntegerSparseMatrix::PutValue(long aRow, long aColumn, integer_matrix_type aValue){
	if ((aRow < 0) || (aColumn < 0))		return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)		return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)	return ConstRC::ExcessiveColumnIndex;
	// CheckMon();
	// cout << "\n>("<<aRow<<","<<aColumn<<","<<aValue<<")";
	// PrintStructure();
	long i,j;
	this->Allocate();
	i=0;
	long l;
	long max_index = this->max_allocated_row_index;
	i = BinarySearchRow(aRow);
	// cout << "||"<< i << "|| ";
	//  now, (*operand).matrix[i][0]>= row
   	// this if solves the case, when the row is 	already allocated
	if ((i <= max_index) && this->matrix[i][0] == aRow){
		j = BinarySearchColumn(i,aColumn);
		// cout << ""<< j << "|| ";
		if ((j == this->matrix[i][1]) && (this->matrix[i][2*j+2] < aColumn)) { j++; };
	   	// this if solves the case, when there is already a value on the denoted place
		if ((j<= (this->matrix[i][1])) && (this->matrix[i][2*j+2] == aColumn)){ // we are there!
			this->matrix[i][2*j+3] = aValue;
			//	CheckMon();
			return ConstRC::Ok;
		}
		// In other cases, there is no need to insert "0"
		if (0 == aValue) return ConstRC::Ok;
		// need to reallocate the row
		if (INTEGER_ALLOCATED(this->matrix[i][1]+1) > INTEGER_ALLOCATED(this->matrix[i][1])) {
				// cout << "ALLOCATION - PutValue("<<aRow <<", "<<aColumn<<","<< aValue <<"): " << this->matrix[i][1] << " (" <<INTEGER_ALLOCATED(this->matrix[i][1])<< ") → " << this->matrix[i][1]+1 << " ("<<INTEGER_ALLOCATED(this->matrix[i][1]+1) <<")\n";
				this->matrix[i] = (integer_matrix_type*) Utils::AllocationMachine((void **) &this->matrix[i],INTEGER_ALLOCATED(this->matrix[i][1]+1)*sizeof(integer_matrix_type));
				if (!this->matrix[i])	{
					return ConstRC::NotEnoughMemory;
				}
				for (l=INTEGER_ALLOCATED(this->matrix[i][1]); l<INTEGER_ALLOCATED(this->matrix[i][1]+1); l++){
					this->matrix[i][l]=0;
				}
		}

		for (long k = (this->matrix[i][1]*2+5); k>2*j+3; k--){
			// cout << "(" << k << "←" << k-2 <<")";
			this->matrix[i][k] = this->matrix[i][k-2];
		}
		this->matrix[i][2*j+2] = aColumn;
		this->matrix[i][2*j+3] = aValue;
		this->matrix[i][1]++;
		// PrintStructure();
		return ConstRC::Ok;
	}
	// otherwise, we must insert a new row
	cerr << "ALOKUJI - PutValue() - new row";
	this->matrix = (integer_matrix_type **) Utils::AllocationMachine((void **) &this->matrix,(++(this->max_allocated_row_index))*sizeof(integer_matrix_type*));
	if (!this->matrix){
		return ConstRC::NotEnoughMemory;
	}
	this->matrix[this->max_allocated_row_index] = (integer_matrix_type *) Utils::AllocationMachine((void **) &this->matrix[this->max_allocated_row_index],INTEGER_ALLOCATED(0)*sizeof(integer_matrix_type));
	if (!this->matrix[this->max_allocated_row_index]){
		return ConstRC::NotEnoughMemory;
	}
	for (j=this->max_allocated_row_index; j>i; j--){
		this->matrix[j] = (integer_matrix_type*) Utils::AllocationMachine((void **) &this->matrix[j], INTEGER_ALLOCATED(this->matrix[j-1][1])*sizeof(integer_matrix_type));
		if (!this->matrix[j]){
			return ConstRC::NotEnoughMemory;
		}		memcpy(this->matrix[j],this->matrix[j-1],INTEGER_ALLOCATED(this->matrix[j-1][1])*sizeof(integer_matrix_type));
	}
	memset(this->matrix[i],0,INTEGER_ALLOCATED(1));
	this->matrix[i][0] = aRow;
	this->matrix[i][1] = 0;
	this->matrix[i][2] = aColumn;
	this->matrix[i][3] = aValue;
	return ConstRC::Ok;
}

int IntegerSparseMatrix::GetValue(long aRow, long aColumn, integer_matrix_type & aValue){
	// cout << "\n>("<<aRow<<","<<aColumn<<","<<aValue<<")";
	if (aRow > this->maximal_row_index)		return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index)	return ConstRC::GeneralError;
	long i,j;
	i = BinarySearchRow(aRow);
	if (this->matrix[i][0] != aRow) return FALSE;
	j = BinarySearchColumn(i,aColumn);
	if ((j <= this->matrix[i][1]) && (this->matrix[i][j*2+2] == aColumn)) {
		aValue  = this->matrix[i][j*2+3];
		return TRUE;
	}
	/*for (j=2; j< (this->matrix[i][1]+2); j++){

		if (this->matrix[i][j] == aColumn){
			return TRUE;
		}
	}*/
	aValue = 0;
	return FALSE;
}

int IntegerSparseMatrix::Zeroize() {
	return ConstRC::NotSupported;
}

void IntegerSparseMatrix::PrintToScreen() {
	cout << "\n(" << maximal_row_index+1 << "x" << maximal_column_index+1 <<")";
	int rowpos = 0;
	long row = 0;
	while (row <= GetMaximalRowIndex()) {
		cout << endl;
		if (matrix[rowpos][0]>row) {
			for (int colind=0;colind<= maximal_column_index;colind++)
				cout <<"|0\t";
		} else {
			long column = 0;
			int colpos = 2;
			while (column <= GetMaximalColumnIndex()) {
				if ((matrix[rowpos][1] == -1) || (matrix[rowpos][1]*2+2 < colpos) || (matrix[rowpos][colpos]>column)) {
					cout <<"|0\t";
				} else {
					cout << "|"<<matrix[rowpos][colpos+1] <<"\t";
					colpos+=2;
				}
				column++;
			}
			rowpos++;
		}
		row++;
	}
}

int IntegerSparseMatrix::PutOne(long aRow, long aColumn) {
	PutValue(aRow,aColumn,1);
}

int IntegerSparseMatrix::PutZero(long aRow, long aColumn) {
	PutValue(aRow, aColumn, 0);
	return ConstRC::Ok;
}

int IntegerSparseMatrix::IsOne(long aRow, long aColumn) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::IsZero(long aRow, long aColumn) {
	integer_matrix_type value;
	GetValue(aRow, aColumn, value);
	//  cerr << "ZERO (" <<aRow <<","<<aColumn<< "): " << value;
	// PrintStructure();
	if (modulo)	return (value % modulo == 0);
		else return (value == 0);
}

long  IntegerSparseMatrix::CountNonZeroItemsOnRow(long aRow) {
	if ((aRow < 0) || (aRow > GetMaximalRowIndex())) {
		return -1;
	}
	long count = 0;
	for(long column = 0; column <= matrix[aRow][1]; column++) {
		if (matrix[aRow][(column+1)*2+1] != 0) count++;
	};
	return count;
}
/**
 * Iterates through nonzero items on a sparse row. Parameters
 * \param aRow row index not needed and ignored if aRowIndex initialized by previous call
 * \param aRowIndex internal row index, should be set to -1 if not known for this row, otherwise used from previous call
 * \param aColumn column index (not internal sparse index, but full index)
 * \param aValue value
 */
long IntegerSparseMatrix::IterateRow(long aRow,long & aRowIndex, long aColumnIndex, long & aColumn, integer_matrix_type & aValue) {
	// cout << "IR:" << aRow <<":" << aRowIndex << ":" << aColumnIndex <<":" << aColumn << ":" << aValue << endl;
	if (aRowIndex < 0)
		aRowIndex = BinarySearchRow(aRow);
	if (aRowIndex == ConstRC::ExcessiveRowIndex) {
		aColumn = -1;
		CLog::LogError(WHEREARG,"IntegerSparseMatrix","IterateRow",aRowIndex,aRow,"Bad row index.");
		return ConstRC::ExcessiveRowIndex;
	}
	if ((aColumnIndex < 0) || (matrix[aRowIndex][1] < aColumnIndex)) {
		aColumn = -1;
		CLog::LogError(WHEREARG,"IntegerSparseMatrix","IterateRow",aColumnIndex,matrix[aRowIndex][1],"Bad column index.");
		return ConstRC::ExcessiveColumnIndex;
	}
	aColumn = matrix[aRowIndex][aColumnIndex*2+2];
	aValue = matrix[aRowIndex][aColumnIndex*2+3];
	// cout << "IR:" << aRow <<":" << aRowIndex << ":" << aColumnIndex <<":" << aColumn << ":" << aValue << endl;
	return ConstRC::Ok;
}

long  IntegerSparseMatrix::CountNonZeroItems() {
	long count = 0;
	for (long rowpos = 0;rowpos <= max_allocated_row_index; rowpos++) {
		long rowindex = matrix[rowpos][0];
		long columncount = matrix[rowpos][1];
		for(long column = 0; column <= columncount; column++) {
			if (matrix[rowpos][(column+1)*2+1] != 0) count++;
		};
	};
	return count;
}

long IntegerSparseMatrix::CountNeededMemory() {
	long count = 0;
	for (long rowpos = 0;rowpos <= max_allocated_row_index; rowpos++) {
		long rowindex = matrix[rowpos][0];
		long columncount = matrix[rowpos][1];
		if (columncount!= -1) count+= INTEGER_ALLOCATED(columncount);
	}
	return count;
}

int IntegerSparseMatrix::ZeroizeRow(long aRow) {
	long rowindex = BinarySearchRow(aRow);
	if ((rowindex <= max_allocated_row_index) && this->matrix[rowindex][0] == aRow)
		this->matrix[rowindex][1]=-1;
	return ConstRC::Ok;
}

int IntegerSparseMatrix::MultiplyRow(long aRow, integer_matrix_type aMul) {
	long rowindex = BinarySearchRow(aRow);
	if ((rowindex <= max_allocated_row_index) && this->matrix[rowindex][0] == aRow) {
		for (long column=0;column<=matrix[rowindex][1];column++) {
			matrix[rowindex][column*2+3] = matrix[rowindex][column*2+3] * aMul;
			if (modulo) matrix[rowindex][column*2+3] % modulo;
		}
		}
}

int IntegerSparseMatrix::SwapRows(long aRow1, long aRow2) {
	// cout << "SWAP(" << aRow1 << "," << aRow2<<")"<<endl;
	long row1 = BinarySearchRow(aRow1);
	long row2 = BinarySearchRow(aRow2);
	integer_matrix_type * swappointer;
	swappointer = matrix[row1];
	matrix[row1] = matrix[row2];
	matrix[row2] = swappointer;

	integer_matrix_type swapindex;
	swapindex = matrix[row1][0];
	matrix[row1][0] = matrix[row2][0];
	matrix[row2][0] = swapindex;
}

int IntegerSparseMatrix::AddRows(long aTarget, long aSource) {
	if ((aTarget>GetMaximalRowIndex()) || (aTarget <0)) return ConstRC::ExcessiveRowIndex;
	if ((aSource>GetMaximalRowIndex()) || (aSource <0)) return ConstRC::ExcessiveRowIndex;

	long sourcemax = matrix[aSource][1]*2+2;
	long targetmax = matrix[aTarget][1]*2+2;

	// result cannot have more nonnull items then operands together
	integer_matrix_type * newRow =
				(integer_matrix_type *) Utils::AllocationMachine(NULL,INTEGER_ALLOCATED(matrix[aSource][1]+matrix[aTarget][1])*sizeof(integer_matrix_type));

    long sourcepointer;
    if (0 == sourcemax) sourcepointer = -1; else sourcepointer = 2;
    long targetpointer;
    if (0 == targetmax) targetpointer = -1; else targetpointer = 2;
    long resultpointer = 0;
    while ((sourcepointer > 0) || (targetpointer > 0 )) {
    	resultpointer +=2;
    	enum {source_row, target_row, add_rows} choose_row;
    	if (targetpointer == -1) choose_row = source_row;
    	else if (sourcepointer == -1) choose_row = target_row;
    	else {
    		if (matrix[aSource][sourcepointer] == matrix[aTarget][targetpointer]) choose_row = add_rows;
    		else if (matrix[aSource][sourcepointer] > matrix[aTarget][targetpointer]) choose_row = target_row;
    		else choose_row = source_row;
    	}
    	if (choose_row == target_row) {
    		newRow[resultpointer] = matrix[aTarget][targetpointer];
    		newRow[resultpointer+1] = matrix[aTarget][targetpointer+1];
    		if (targetpointer != targetmax)
    			targetpointer +=2;
    		else
    			targetpointer = -1;
    	};
    	if (choose_row == source_row) {
    		newRow[resultpointer] = matrix[aSource][sourcepointer];
    		newRow[resultpointer+1] = matrix[aSource][sourcepointer+1];
    		if (sourcepointer != sourcemax)
    			sourcepointer +=2;
    		else
    			sourcepointer = -1;
    	};
    	if (choose_row == add_rows){
    		newRow[resultpointer] = matrix[aSource][sourcepointer];
    		newRow[resultpointer+1] = matrix[aSource][sourcepointer+1]+matrix[aTarget][targetpointer+1];

    		// if item is null, we can omit it
    		if (0 == newRow[resultpointer+1]) resultpointer -= 2;

    		if (targetpointer != targetmax)
    			targetpointer +=2;
    		else
    			targetpointer = -1;
    		if (sourcepointer != sourcemax)
    			sourcepointer +=2;
    		else
    			sourcepointer = -1;
    	}
    }
    newRow[0] = matrix[aTarget][0];
    newRow[1] = resultpointer/2 -1;
    free(matrix[aTarget]);
    matrix[aTarget]=newRow;
}

int IntegerSparseMatrix::ZeroizeColumn(long aColumn) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::ZeroizeColumn(long * aColumnList, long aListMaxIndex) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::SwapColumns(long aColumn1, long aColumn2) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::AddColumns(long aTarget, long aSource) {
	return ConstRC::NotSupported;
}
int IntegerSparseMatrix::AddToRow(long aRow, AbstractMatrix* aOperand,
		long aRow2) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::AddToRow(long aRow, AbstractIntegerMatrix* aOperand,
		long aRow2) {
	//TODO: Rewrite using sparsity
	long max_index1 = this->GetMaximalColumnIndex();
	long max_index2 = aOperand->GetMaximalColumnIndex();
	long max_index;
	if (max_index1 <= max_index2)
		max_index = max_index1;
	else
		max_index = max_index2;
	for(long column = 0; column <= max_index; column++) {
		integer_matrix_type value;
		aOperand->GetValue(aRow, column, value);
		PutValue(aRow, column, value);
	}
	return ConstRC::Ok;
}

int IntegerSparseMatrix::Save(char* aName) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::Load(char* aName) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::WriteData(xmlTextWriterPtr aWriter) const {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::ReadData(xmlTextReaderPtr aReader) {
	return ConstRC::NotSupported;
}

int IntegerSparseMatrix::Test() {
	time_t seconds;
	time(&seconds);
	srand((unsigned int) seconds);
	for(int n = 0; n < maximal_column_index*maximal_row_index/2; n++) {
		int value = rand();
		int column = rand() % maximal_column_index;
		int row = rand() % maximal_row_index;
		integer_matrix_type readvalue;
		PutValue(row,column, value);
		GetValue(row, column, readvalue);
		if (readvalue != value) {
			return ConstRC::GeneralError;
		};
	};
	return ConstRC::Ok;
};

int IntegerSparseMatrix::SubtractRowMultiple(long aTarget, long aSource,
		integer_matrix_type aMul) {
     	// CheckMon();
	// cout << "SubtractRowMultiple(" << aTarget <<"," << aSource <<"," <<aMul <<")";
	// PrintStructure();
	if ((aTarget>GetMaximalRowIndex()) || (aTarget <0)) return ConstRC::ExcessiveRowIndex;
	if ((aSource>GetMaximalRowIndex()) || (aSource <0)) return ConstRC::ExcessiveRowIndex;

	long sourcemax = matrix[aSource][1]*2+2;
	long targetmax = matrix[aTarget][1]*2+2;

	// result cannot have more nonnull items then operands together
	integer_matrix_type * newRow =
				(integer_matrix_type *) Utils::AllocationMachine(NULL,INTEGER_ALLOCATED(matrix[aSource][1]+matrix[aTarget][1]+2)*sizeof(integer_matrix_type));

    long sourcepointer;
    if (0 == sourcemax) sourcepointer = -1; else sourcepointer = 2;
    long targetpointer;
    if (0 == targetmax) targetpointer = -1; else targetpointer = 2;
    long resultpointer = 0;
    while ((sourcepointer > 0) || (targetpointer > 0 )) {
    	resultpointer +=2;
    	enum {source_row, target_row, add_rows} choose_row;
    	if (targetpointer < 1) choose_row = source_row;
    	else if (sourcepointer < 1) choose_row = target_row;
    	else {
    		if (matrix[aSource][sourcepointer] == matrix[aTarget][targetpointer]) choose_row = add_rows;
    		else if (matrix[aSource][sourcepointer] > matrix[aTarget][targetpointer]) choose_row = target_row;
    		else choose_row = source_row;
    	}
    	if (choose_row == target_row) {
    		newRow[resultpointer] = matrix[aTarget][targetpointer];
    		if (modulo) {
    			newRow[resultpointer+1] = matrix[aTarget][targetpointer+1] % modulo;
    		} else {
    			newRow[resultpointer+1] = matrix[aTarget][targetpointer+1];
    		}
    		if (targetpointer != targetmax)
    			targetpointer +=2;
    		else
    			targetpointer = -1;
    	};
    	if (choose_row == source_row) {
    		newRow[resultpointer] = matrix[aSource][sourcepointer];
    		if (modulo) {
        		newRow[resultpointer+1] = (modulo - ((matrix[aSource][sourcepointer+1] % modulo) * aMul ) % modulo) % modulo;
    		} else {
    			newRow[resultpointer+1] = - matrix[aSource][sourcepointer+1] * aMul;
    		}
    		if (sourcepointer != sourcemax)
    			sourcepointer +=2;
    		else
    			sourcepointer = -1;
    	};
    	if (choose_row == add_rows){
    		newRow[resultpointer] = matrix[aSource][sourcepointer];
    		if (modulo) {
    			newRow[resultpointer+1] = (modulo - (((matrix[aSource][sourcepointer+1] % modulo) * aMul ) % modulo)
    					+ matrix[aTarget][targetpointer+1]) % modulo;
    		} else {
    			newRow[resultpointer+1] =-matrix[aSource][sourcepointer+1]  * aMul + matrix[aTarget][targetpointer+1];
    		}
    		// if item is null, we can omit it
    		if (0 == newRow[resultpointer+1]) resultpointer -= 2;

    		if (targetpointer != targetmax)
    			targetpointer +=2;
    		else
    			targetpointer = -1;
    		if (sourcepointer != sourcemax)
    			sourcepointer +=2;
    		else
    			sourcepointer = -1;
    	}
    }
    newRow[0] = matrix[aTarget][0];
    newRow[1] = resultpointer/2 -1;
    free(matrix[aTarget]);
    matrix[aTarget]=newRow;
    // CheckMon();
}

int IntegerSparseMatrix::Copy(IntegerSparseMatrix* aSource) {
  if (max_allocated_row_index < aSource->max_allocated_row_index) return ConstRC::GeneralError; //TODO: reallocate
  for (long row_index = 0; row_index<= aSource->max_allocated_row_index; row_index++) {
	  matrix[row_index][0] = aSource->matrix[row_index][0];
	  matrix[row_index][1] = aSource->matrix[row_index][1];
	  for(long index = 2;index <= 2*(matrix[row_index][1]+1)+1; index++) {
		  matrix[row_index][index] = aSource->matrix[row_index][index];
	  }
  };
  return ConstRC::Ok;
}

int IntegerSparseMatrix::Transpose(IntegerSparseMatrix* aSource) {
	if (GetMaximalColumnIndex() != aSource->GetMaximalRowIndex() || GetMaximalRowIndex() != aSource->GetMaximalColumnIndex())
		return ConstRC::GeneralError;
	for (long row = 0; row <= aSource->GetMaximalRowIndex();row++)
	{
		long columnmax = aSource->CountNonZeroItemsOnRow(row);
		long rowindex=-1;
		long column;
		integer_matrix_type value;
		for (long columnindex = 0; columnindex < columnmax; columnindex++) {
			aSource->IterateRow(row,rowindex,columnindex,column,value);
			PutValue(column,row,value);
		}
	}
}

int IntegerSparseMatrix::Transpose(AbstractMatrix* aSource) {
	IntegerSparseMatrix * ISM = dynamic_cast<IntegerSparseMatrix *>(aSource);
	if (ISM) {
		return Transpose(ISM);
	}
	return ConstRC::GeneralError;
}

long IntegerSparseMatrix::MaxNonZeroItemsPerRow() {
	long maximum = -1;
	for(long row =0; row <= GetMaximalRowIndex(); row++) {
		if (matrix[row][1] > maximum) maximum = matrix[row][1];
	}
	if (maximum == -1) return 0;
	else return maximum+1;
}

int IntegerSparseMatrix::ApplyModulo() {
	for(long row =0; row <= GetMaximalRowIndex(); row++) {
		long items = matrix[row][1];
		for (long item = 3; item <= 2*items+3; item+=2) {
			matrix[row][item] %= modulo;
		}
	}
}

int IntegerSparseMatrix::PrintStructure(long aLimit) {
	if ((aLimit == -1) || (aLimit > max_allocated_row_index)) aLimit = max_allocated_row_index;
	for (long rowpos = 0;rowpos <= aLimit; rowpos++) {
		long rowindex = matrix[rowpos][0];
		long columncount = matrix[rowpos][1];
		cout << endl << "R:" << rowindex << " C:" << columncount << " ";
		for(long column = 0; column <= columncount; column++) {
			cout << "[" << matrix[rowpos][(column+1)*2] << "," << matrix[rowpos][(column+1)*2+1]<<"]";
		};
	};
};

int IntegerSparseMatrix::SparseNonzeroRows(long & aNonzero) {
	return maximal_row_index+1;
}

int IntegerSparseMatrix::SparseNonzeroColumns(long aRowIndex, long & aRow, long & aNonzero) {
	aRow = matrix[aRowIndex][0];
	aNonzero = matrix[aRowIndex][1]+1;
	return ConstRC::Ok;
}

int IntegerSparseMatrix::SparseGetValue(long aRowIndex, long aColumnIndex,
		long & aColumn, integer_matrix_type & aValue) {
	aColumn = matrix[aRowIndex][aColumnIndex*2+2];
	aValue = matrix[aRowIndex][aColumnIndex*2+2+1];
	return ConstRC::Ok;
}

int IntegerSparseMatrix::CheckMon() {
	for (long rowpos = 0;rowpos <= max_allocated_row_index; rowpos++) {
		long rowindex = matrix[rowpos][0];
		long columncount = matrix[rowpos][1];
		if (columncount>1) {
			for(long column = 1; column <= columncount; column++) {
				if (matrix[rowpos][column*2]>=matrix[rowpos][column*2+2]) {
					cout << "CheckMon: ("<<rowindex<<","<< column <<")";
					PrintStructure();
					abort();
				};
			};
		}
	};
}
