#ifndef _LP_HASHTABLE_COMMON_
#define _LP_HASHTABLE_COMMON_

#include <vector>
//#include "relation_part.h"
#include "const_return_codes.h"
#include "const_element_flags.h"
#include "const_debug.h"
// #include "const_rel_processing_phase.h"
#include "common_utils.h"
#include "log.h"
#include "const_hashtables.h"
#include <iostream>
#include <typeinfo>
#include <cassert>


class CRelationPart;

/*!
 * \brief Base hash entry supporting value and mask (and void ancestor and fingerprint)
 */
class CCommonHashEntry {
protected:
	large_prime_type value;	
	large_prime_type ancestor;
public:	
	void SetOne() {value = 1; ancestor = 0;};
	void SetMask() 	{value = ConstHT::FULL_MASK_NFS; };
	bool IsMasked() {return value == ConstHT::FULL_MASK_NFS;};

	large_prime_type GetValue() const {return value;};
	void SetValue(large_prime_type aValue) {value = aValue;};

	large_prime_type GetAncestor() const {return ancestor;};	
	void SetAncestor(large_prime_type aAncestor) {ancestor = aAncestor;};

	large_prime_type GetRootFingerprint() const { return 0;};
	void SetRootFingerprint(large_prime_type aFingerPrint) {};

	large_prime_type GetRelation() const { return 0;};
	void SetRelation(large_prime_type aRelation) {};

};

/*!
 * \brief Advanced hash entry with relation index.
 */
class CRelationHashEntry : public CCommonHashEntry {
private:
	large_prime_type relation;
public:
	large_prime_type GetRelation() const { return relation;};
	void SetRelation(large_prime_type aRelation) {relation = aRelation;};
};


/*!
 * \brief Template class for hashtables
 * */
template <class HashType> class lp_hashtable_common {
protected:
	unsigned long filled;
public:
	///Hashtable data type
	HashType *Hashtable;
	///Number of the allocated hashtable entries
	long AllocatedSize;
	unsigned long	Mask;

	lp_hashtable_common();
	~lp_hashtable_common();

	void DeleteHashtable();
	int ClearHashtable();
	int SetupHashtable(long aAmount);
	int SetupHashtableFB(long aFBSize);

	long FindAndResetRoot(long aIndex, std::vector<long> & aIntermediateRoots);
	int CountRoots(std::vector<HashType> & aRootsList) const;
	int AddToTable(large_prime_type aPrime, large_prime_type aRoot, unsigned long aRootMask, unsigned long aIndex);
	int FindPrimeInHashtable(large_prime_type aPrime, large_prime_type aRootFingerprint, bool & aPresent, long & aIndex, bool & aEnlarged) ;
	int FindSingletonRelations(detection_type * aDetectionField, int & aMaxDetectionField, int& aCurrentlyRemovedVertices);

	void Added() { filled++;};

	void Print();
private:
	int CollisionResolution(long aHash, large_prime_type aPrime, large_prime_type aRootFingerprint, long & aIndex) const;
	
	int EnlargeHashtable();
	lp_hashtable_common(const lp_hashtable_common& aOperand) {}
};

template <class HashType>
lp_hashtable_common<HashType>::lp_hashtable_common()
{
	this->Hashtable = NULL;
	this->AllocatedSize = 0;
	this->Mask = 0;
	this->filled = 0;
}

template <class HashType>
lp_hashtable_common<HashType>::~lp_hashtable_common()
{
	DeleteHashtable();
}

template <class HashType>
void lp_hashtable_common<HashType>::DeleteHashtable()
{
	if (this->Hashtable != NULL)
	{
		free(this->Hashtable);
		this->Hashtable = NULL;
		this->AllocatedSize = 0;
		this->Mask = 0;
	}
}

template <class HashType>
int lp_hashtable_common<HashType>::SetupHashtable(long aAmount)
{
	if (this->Hashtable != NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","SetupHashtable",0,0,"Hashtable is not NULL.");
		return ConstRC::BadArgument;
	}

	this->AllocatedSize = Utils::GetSuitableHashtableSize(aAmount);
	this->Hashtable = (HashType *) Utils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(HashType));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","SetupHashtable", 0, this->AllocatedSize * sizeof(HashType),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	ClearHashtable();

	return ConstRC::Ok;
}

template <class HashType>
int lp_hashtable_common<HashType>::SetupHashtableFB(long aFBSize)
{
	if (this->Hashtable != NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","SetupHashtableFB",0,0,"Hashtable is not NULL.");
		return ConstRC::BadArgument;
	}

	this->AllocatedSize = ConstHT::MIN_LP_TABLE_SIZE;
	//  this->AllocatedSize = 8;
	if (aFBSize > 10000) 
	{ 
		this->AllocatedSize <<= 1;			// 0.25 M entries
	}
	if (aFBSize > 20000) 
	{ 
		this->AllocatedSize <<= 1;			// 0.5 M entries
	}
	if (aFBSize > 40000) 
	{ 
		this->AllocatedSize <<= 1;			// 1 M entries
	}	
	if (aFBSize > 80000) 
	{ 
		this->AllocatedSize <<= 1;			// 2 M entries
	}	
	if (aFBSize > 160000) 
	{
		this->AllocatedSize <<= 1;			// 4 M entries
	}
	if (aFBSize > 320000) 
	{
		this->AllocatedSize <<= 1;			// 8 M entries
	}
	this->Hashtable = (HashType*) Utils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(HashType));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","SetupHashtableFB", 0, this->AllocatedSize * sizeof(HashType),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	ClearHashtable();

	return ConstRC::Ok;
}

template <class HashType>
int lp_hashtable_common<HashType>::ClearHashtable()
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","ClearHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	memset(Hashtable, 0, AllocatedSize*sizeof(HashType));

	// Special case of i = 0: PRIME = 1, root of its own component.
	this->Hashtable->SetOne();
	for(HashType *index = this->Hashtable+1; index < this->Hashtable+AllocatedSize; index ++)
	{
		index->SetMask();

	};
	this->filled=1;
	return ConstRC::Ok;
}
/*!
	A method called from AddToCycleCountStructures (from sieving classes), whose aim is to find
	the root of a component including element at aIndex, and to reset all
	intermediate roots to a new, correct value (see Lenstra, Manasse). Returns
	the index of the root in hashtable.
*/
template <class HashType>
long lp_hashtable_common<HashType>::FindAndResetRoot(long aIndex, std::vector<long> & aIntermediateRoots) 
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","FindAndResetRoot",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	long ancestor = this->Hashtable[aIndex].GetAncestor();
	long index = aIndex;
	aIntermediateRoots.clear();
	//printf("Hledame predka pro tento prvek index = %d, index predka = %d\n", aIndex, ancestor);
	while (ancestor != index) 
	{
		aIntermediateRoots.push_back(index);
		index = ancestor;
		ancestor = this->Hashtable[index].GetAncestor();
		//printf("index = %d, ancestor = %d, iterace: %d\n", index, ancestor, aIntermediateRoots.size());
	}	
	long root = index;
	int max = (int)aIntermediateRoots.size();
	for (int i = 0; i < max; i++) 
	{
		this->Hashtable[aIntermediateRoots[i]].SetAncestor(root);
        //printf("Prenastavujeme koren indexu %d na %d.\n", aIntermediateRoots[i], root);
	}
	return root;
}

template <class HashType>
int lp_hashtable_common<HashType>::CountRoots(std::vector < HashType > & aRootsList) const 
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","CountRoots",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	HashType * aux = this->Hashtable;

	for (long i = 0; i < this->AllocatedSize; i++, aux++) 
	{
		if ((signed)aux->GetAncestor() == i) 
		{
			//printf("Found component root: %d.\n", aux->value);

			HashType next;
			next.SetValue(aux->GetValue());
			next.SetRootFingerprint(aux->GetRootFingerprint());
			next.SetAncestor(i);

			aRootsList.push_back(next);
		}
	}

	return ConstRC::Ok;
}
/*!
	This method tries to find large prime aPrime in hashtable, and indicates
	the result of this search in two values: the return value, which is an index
	corresponding to the real or expected position of the prime in the hashtable,
	and value of aPresent, which indicates whether the prime has already been
	in the hashtable or not. In the latter case, the return value has meaning
	"if you want to insert aPrime, this index is the one where it should be inserted".

*/
template <class HashType>
int lp_hashtable_common<HashType>::FindPrimeInHashtable(large_prime_type aPrime, large_prime_type aRootFingerprint, bool & aPresent, long & aIndex, bool & aEnlarged) 
{
	aEnlarged = false;
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","FindPrimeInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aIndex = 0;
		return ConstRC::BadArgument;
	}

	long hash = (aPrime >> 1) & this->Mask;			
	int rc = CollisionResolution(hash, aPrime, aRootFingerprint, aIndex);
	while (rc == ConstRC::NeedEnlargement) {
		aEnlarged = true;
		// Print();
		if (EnlargeHashtable() != ConstRC::Ok) {
			aPresent = false;
			aIndex = 0;
			return ConstRC::GeneralError;
		}
		// Print();
		long hash = (aPrime >> 1) & this->Mask;			
		rc = CollisionResolution(hash, aPrime, aRootFingerprint, aIndex);
	};	
	if (rc != ConstRC::Ok)
	{
		aPresent = false;
		aIndex = 0;
		return ConstRC::GeneralError;
	}

    	if (this->Hashtable[aIndex].IsMasked()) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

// 	std::cout << "(" << aPrime << "," << aPresent << "," << aIndex << ")";

	return ConstRC::Ok;
}
/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where aPrime can be stored.

*/
template <class HashType>
int lp_hashtable_common<HashType>::CollisionResolution(long aHash, large_prime_type aPrime, large_prime_type aRootFingerprint, long & aIndex) const
{
	aIndex = aHash;
	long iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	HashType * aux = this->Hashtable + aIndex;
	HashType * end = this->Hashtable + this->AllocatedSize;

	while (((aux->GetValue() != aPrime) || (aux->GetRootFingerprint() != aRootFingerprint)) && (aux->GetValue() != ConstHT::FULL_MASK_NFS)) 
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"lp_hashtable_common","CollisionResolution",iter,this->AllocatedSize,"The hash table has overflown.");
				/* for (int i = 0;i<this->AllocatedSize;i++) {
					std::cerr << "[" << i <<", "<<Hashtable[i].GetValue() <<", "<< Hashtable[i].GetAncestor()<<"]\t";
				}; */
				return ConstRC::GeneralError;
			}
			aux = this->Hashtable;
		}
	}

	aIndex = aux - this->Hashtable;

	if (ConstHT::MAX_FULLNESS < (double) this->filled / this->AllocatedSize) {
		return ConstRC::NeedEnlargement;
	} else {
		return ConstRC::Ok;
	};
}

template <class HashType>
int lp_hashtable_common<HashType>::AddToTable(large_prime_type aPrime, large_prime_type aRoot, unsigned long aRootMask, unsigned long aIndex)
{

#ifdef DEBUG_REL_PRO_PHASE_ASSERTION_MODE_ON
	if (root >= ConstSPhase::SECOND_ELEMENT_MASK_NFS)
	{
		NFSUtils::NormalMessage("Error PutToHashtable: root is too large !!!");
	}
#endif

	aRoot |= aRootMask;

	bool present = false;
	long index;
	bool enlarged;
	if (FindPrimeInHashtable(aPrime, aRoot, present, index, enlarged) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","AddToTable",0,0,"Find prime in hashtable failed.");
		return ConstRC::GeneralError;
	}

	if (!present) 
	{
		// This prime has not been in hashtable yet.
		this->Hashtable[index].SetValue(aPrime);
		this->Hashtable[index].SetRootFingerprint(aRoot);
		// TODO: if used for removing large singletons, ancestors should not be recomputed while enlarging hashtable
		this->Hashtable[index].SetAncestor(aIndex);
		this->Added();
	} 
	else 
	{
		// We mark this prime as nonsingleton by bit-masking the "ancestor" field.
		this->Hashtable[index].SetAncestor(this->Hashtable[index].GetAncestor() | ConstHT::NONSINGLETON_MASK_NFS);
	}

	return ConstRC::Ok;
}

template <class HashType>
int lp_hashtable_common<HashType>::FindSingletonRelations(detection_type * aDetectionField, int & aMaxDetectionField, int& aCurrentlyRemovedVertices)
{
	if (this->Hashtable == NULL || aDetectionField == NULL)
	{
		CLog::LogError(WHEREARG,"lp_hashtable_common","FindSingletonRelations",0,0,"Hashtable or DetectionField is NULL.");
		return ConstRC::BadArgument;
	}

	// first item is (formal) 1, so we start from second when looking for singletons
	HashType * aux = this->Hashtable+1;
	HashType * end = this->Hashtable + this->AllocatedSize;

	int edge;
	do
	{
		// There is a prime on this position ...			
		if (!aux->IsMasked()) 
		{
			// This is a singleton prime
			edge = aux->GetAncestor() & (~ConstHT::NONSINGLETON_MASK_NFS);
			// Singleton: has no "NONSINGLETON_MASK_NFS"
			if (aux->GetAncestor() == (unsigned int) edge) 
			{
				//printf("Index %d contains singleton.\n", aux - this->Hashtable);					
				//printf("There is a singleton at edge %d.\n",edge);
				aCurrentlyRemovedVertices++;
					
				// mark as singleton so it is not going to be counted in CountRoots
				aux->SetAncestor(aux->GetAncestor() | ConstHT::NONSINGLETON_MASK_NFS);

				assert(edge <= aMaxDetectionField);

				if (aDetectionField[edge] == ConstHT::NONSINGLETON) 
				{
					aDetectionField[edge] = ConstHT::FRESH_SINGLETON;
					//printf("Turning %d from nonsingleton to fresh singleton.\n",edge);
				}
			} 
			//else 
			//{					
				//printf("Index %d contains nonsingleton, edge is %d.\n",aux - this->Hashtable,edge);
				//aDetectionField[edge] = ConstHT::NONSINGLETON;
				//printf("Turning %d to nonsingleton.\n",edge);
			//}
		}
	}
	while (++aux < end);

	return ConstRC::Ok;
}

template <class HashType>
int lp_hashtable_common<HashType>::EnlargeHashtable() {
	int new_maximal_index = (this->AllocatedSize)*2-1;
	int new_mask = new_maximal_index;
	// std::cerr << "\nEnlarge:" << AllocatedSize << " -> " << new_maximal_index;

	HashType *new_table = NULL;
	if ((new_table = (HashType*) Utils::AllocationMachine((void**) &new_table, (new_maximal_index+1)* sizeof(HashType))) == NULL) {
		CLog::LogError(WHEREARG,"lp_hashtable_common","EnlargeHashtable", 0, (new_maximal_index+1) * sizeof(HashType),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}

	memset(new_table, 0, (new_maximal_index+1)*sizeof(HashType));

	new_table[0].SetOne();
	new_table[0].SetRootFingerprint(Hashtable[0].GetRootFingerprint());
	new_table[0].SetRelation(Hashtable[0].GetRelation());
	for (int i=1; i <= new_maximal_index; i++) {
		new_table[i].SetValue(ConstHT::FULL_MASK_NFS);
	};

	// first rearrange values
	for (int j = 1;j<this->AllocatedSize;j++) {
//		std::cerr << "-(" << j << ")\t";
		if (ConstHT::FULL_MASK_NFS != Hashtable[j].GetValue()) {
			large_prime_type tempPrime = Hashtable[j].GetValue();
			int j2 = ( tempPrime >> 1) & new_mask;			
//				fprintf(stderr,"\nRearranging values: %d : %d : %d : %d ",j, Hashtable[j].GetValue(), j2, Hashtable[j].GetAncestor());
			int iter2 = 0;
			while (new_table[j2].GetValue() != ConstHT::FULL_MASK_NFS) {
				j2++;
				iter2++;
				if (j2 > new_maximal_index) {
					j2 = 0;
				};
				if (iter2 > new_maximal_index) {
//						fprintf(stderr,"[ %d ]  ",j3);
					fprintf(stderr, "Overflow!");
					abort();
				};
			};
//				fprintf(stderr,"  [%d %d]  ",iter2, j2);
			new_table[j2].SetValue(Hashtable[j].GetValue());
			new_table[j2].SetRootFingerprint(Hashtable[j].GetRootFingerprint());
			new_table[j2].SetRelation(Hashtable[j].GetRelation());
			new_table[j2].SetAncestor(Hashtable[j].GetAncestor());
		};
	}
	// then rearrange ancestors
	for (int j = 1;j<=new_maximal_index;j++) {
//		std::cerr << "(" << j << ")\t";
		if (new_table[j].GetValue() != ConstHT::FULL_MASK_NFS) {
//					fprintf(stderr,"\nRearranging ancestors: %d : %d",j, new_table[j].GetAncestor());
			int tagged_ancestor = new_table[j].GetAncestor();
			int ancestor_mask = tagged_ancestor & ConstHT::NONSINGLETON_MASK_NFS;
			int ancestor = tagged_ancestor & ~ ConstHT::NONSINGLETON_MASK_NFS;
			large_prime_type tempPrime= Hashtable[ancestor].GetValue();	
			large_prime_type tempRF = Hashtable[ancestor].GetRootFingerprint();	
			int j2 = (tempPrime >> 1) & new_mask;			
			int iter2=0;
//					fprintf(stderr,"\n .%d (%lu)", j2, new_table[j2].value);
			while (((new_table[j2].GetValue() != tempPrime) || (new_table[j2].GetRootFingerprint() != tempRF)) && (new_table[j2].GetValue() != ConstHT::FULL_MASK_NFS)) {
//						fprintf(stderr," .%d (%lu)", j2, new_table[j2].value);
				j2++;
				iter2++;
				if (j2 > new_maximal_index) {
					j2 = 0;
				};
				if (iter2 > new_maximal_index) {
//							fprintf(stderr,"[ %d ]  ",j2);
					fprintf(stderr, "Overflow!");
					abort();
				};
			};
//			fprintf(stderr,"  [%d %d]  ",iter2, j2);
			if (new_table[j2].GetValue() == tempPrime) {
				new_table[j].SetAncestor(j2 | ancestor_mask);
			} else {
				CLog::LogError(WHEREARG,"lp_hashtable_common","EnlargeHashtable",0,tempPrime,"Prime lost during enlarging hashtable!");
				abort();
			};
		};
	}
//			fprintf(stderr,"  [%d %d]  ",iter2, j2);
	this->AllocatedSize = new_maximal_index+1;
	this->Mask = new_mask;			
	free(Hashtable);
	Hashtable= new_table;

	return ConstRC::Ok;
}

template <class HashType>
void lp_hashtable_common<HashType>::Print() {
	std::cout<<"\nVelikost: "<<AllocatedSize;
	int count = 0;
	int values = 0;
	int sa = 0;
	int rfp = 0;
	for(int i = 0; i <AllocatedSize; i++) {
		if (! Hashtable[i].IsMasked()) {
			count++;
			values+=Hashtable[i].GetValue();
			rfp += Hashtable[i].GetRootFingerprint();
			if (Hashtable[i].GetAncestor() == i) {
				sa++;
			};
		};
	};
	std::cout << "\ncount: " << count << "values: " << values << "sa: " << sa << "rfp: " << rfp;
/*	for(int i =0;i<AllocatedSize;i++) {
		if (0 == i%8) std::cout <<"\n";
		if (Hashtable[i].IsMasked()) {
			std::cout << "X"<< "\t" << Hashtable[i].GetAncestor()<<"\t|";	
		}
		else
		{
			std::cout << Hashtable[i].GetValue()<< "\t" << Hashtable[i].GetAncestor()<<"\t|";	
		}
	}; */
};
#endif // _LP_HASHTABLE_COMMON_
