/*
 * mpz_matrix.cpp
 *
*      Author: jero
 */

#include "mpz_matrix.hpp"
#include "gmp.h"
#include "const_return_codes.h"
#include "log.h"
#include <iostream>

MpzMatrix::MpzMatrix() //: MSerializable(FILE_NAME_MPZ_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->matrix = NULL;
	this->species = SNormalMatrix;
}


MpzMatrix::~MpzMatrix() {
	if (max_allocated_row_index >= 0){
		for (long row=0; row<=max_allocated_row_index; row++){
			if (this->matrix[row]){
				for(long column = 0;column <= max_allocated_column_index;column++) {
					mpz_clear(matrix[row][column]);
				}
				free(this->matrix[row]);
				this->matrix[row] = NULL;
			}
		}
		max_allocated_row_index = -1;
		free(this->matrix);
		this->matrix = NULL;
	}
}


int MpzMatrix::Allocate() {
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){
    	this->matrix = (mpz_t **) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(mpz_t*));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"NormalMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate normal matrix rows");
			return ConstRC::NotEnoughMemory;
		}
		for (long row = 0; row <= maximal_row_index; row++) matrix[row]=NULL;
		this->max_allocated_row_index = this->maximal_row_index;
		this->max_allocated_column_index = this->maximal_column_index;
		for (long row=0; row<= maximal_row_index; row++){
			this->matrix[row] = (mpz_t *) Utils::AllocationMachine((void **) &(this->matrix[row]),(this->max_allocated_column_index+1)*sizeof(mpz_t));
			if (!this->matrix[row]){
				CLog::LogError(WHEREARG,"NormalMatrixClass","error allocating columns",(this->max_allocated_column_index+1),201,"There was not enough memory to allocate normal matrix columns");
				return ConstRC::NotEnoughMemory;
			};
			for(long column = 0;column <= maximal_column_index;column++) {
				mpz_init(matrix[row][column]);
			}
		}
	}
	return ConstRC::Ok;
}

MpzMatrix * MpzMatrix::clone() {
	MpzMatrix * new_matrix = new MpzMatrix();
	return new_matrix;
}


int MpzMatrix::Zeroize() {
	for (long row=0;row <=maximal_row_index; row ++) {
		for(long column = 0; column <= maximal_column_index; column ++) {
			mpz_set_ui(matrix[row][column],0);
		}
	}
	return ConstRC::Ok;
}

int MpzMatrix::PutOne(long aRow, long aColumn) {
	mpz_set_ui(matrix[aRow][aColumn],1);
	return ConstRC::Ok;
}

int MpzMatrix::PutZero(long aRow, long aColumn) {
	mpz_set_ui(matrix[aRow][aColumn],0);
	return ConstRC::Ok;
}

int MpzMatrix::IsOne(long aRow, long aColumn) {
	if(mpz_cmp_ui(matrix[aRow][aColumn],1) == 0)
		return TRUE;
	else
		return FALSE;
}

int MpzMatrix::IsZero(long aRow, long aColumn) {
	if(mpz_cmp_ui(matrix[aRow][aColumn],0) == 0)
		return TRUE;
	else
		return FALSE;
}

int MpzMatrix::ZeroizeRow(long aRow) {
	for(long column = 0; column <= maximal_column_index; column++) {
		mpz_set_ui(matrix[aRow][column],0);
	}
	return ConstRC::Ok;
}

int MpzMatrix::SwapRows(long aRow1, long aRow2) {
	mpz_t * pointer;
	pointer = matrix[aRow1];
	matrix[aRow1] = matrix[aRow2];
	matrix[aRow2] = pointer;
	return ConstRC::Ok;
}

int MpzMatrix::AddRows(long aTarget, long aSource) {
	for(long column = 0; column <= maximal_column_index; column++) {
			mpz_add(matrix[aTarget][column],matrix[aTarget][column], matrix[aSource][column]);
	}
	return ConstRC::Ok;
}

int MpzMatrix::ZeroizeColumn(long aColumn) {
	for (long row = 0; row <= maximal_row_index; row ++) {
		mpz_set_ui(matrix[row][aColumn],0);
	}
	return ConstRC::Ok;
}

int MpzMatrix::ZeroizeColumn(long * aColumnList, long aListMaxIndex) {
	return ConstRC::NotSupported;
}

int MpzMatrix::SwapColumns(long aColumn1, long aColumn2) {
	for (long row = 0; row <= maximal_row_index; row ++) {
		mpz_swap(matrix[row][aColumn1],matrix[row][aColumn2]);
	}
	return ConstRC::Ok;
}

int MpzMatrix::AddColumns(long aTarget, long aSource) {
	for (long row = 0; row <= maximal_column_index; row ++) {
		mpz_add(matrix[row][aTarget],matrix[row][aTarget],matrix[row][aSource]);
	}
	return ConstRC::Ok;
}

int MpzMatrix::AddToRow(long aRow, AbstractMatrix* aOperand, long aRow2) {
	MpzMatrix * MM = dynamic_cast<MpzMatrix*>(aOperand);
	if (MM) {
		return AddToRow(aRow,MM,aRow2);
	}
	return ConstRC::NotSupported;
}

int MpzMatrix::AddToRow(long aRow, AbstractIntegerMatrix* aOperand, long aRow2) {
	if (aRow>GetMaximalRowIndex()) return ConstRC::GeneralError;
	if (aRow2>aOperand->GetMaximalRowIndex()) return ConstRC::GeneralError;
	if (aRow<0) return ConstRC::GeneralError;
	if (aRow2<0) return ConstRC::GeneralError;

	long max_index1 = this->GetMaximalColumnIndex();
	long max_index2 = aOperand->GetMaximalColumnIndex();
	long max_index;
	if (max_index1 <= max_index2)
		max_index = max_index1;
	else
		max_index = max_index2;
	for (long column = 0; column <= max_index; column++) {
		integer_matrix_type tempIntvalue;
		aOperand->GetValue(aRow2,column,tempIntvalue);
		// TODO: mpz_add_si ?!
		mpz_add_ui(matrix[aRow][column], matrix[aRow][column], tempIntvalue);
		if (modular) mpz_mod(matrix[aRow][column], matrix[aRow][column],modulo);
	}
	return ConstRC::Ok;
}


int MpzMatrix::AddToRow(long aRow, MpzMatrix* aOperand, long aRow2) {
	if (GetMaximalColumnIndex() != aOperand->GetMaximalColumnIndex()) return ConstRC::GeneralError;
	if (aRow>GetMaximalRowIndex()) return ConstRC::GeneralError;
	if (aRow2>aOperand->GetMaximalRowIndex()) return ConstRC::GeneralError;
	for (long column = 0; column <= GetMaximalColumnIndex(); column++) {
		aOperand->GetValueMpz(aRow2,column,tempval);
		mpz_add(matrix[aRow][column], matrix[aRow][column], tempval);
		if (modular) mpz_mod(matrix[aRow][column], matrix[aRow][column],modulo);
	}
	return ConstRC::Ok;
}


int MpzMatrix::Save(char* aName) {
	return ConstRC::NotSupported;
}

int MpzMatrix::Load(char* aName) {
	return ConstRC::NotSupported;
}

int MpzMatrix::WriteData(xmlTextWriterPtr aWriter) const {
	return ConstRC::NotSupported;
}

int MpzMatrix::ReadData(xmlTextReaderPtr aReader) {
	return ConstRC::NotSupported;
}

int MpzMatrix::GetValueMpz(long aRow, long aColumn, mpz_t aValue) {
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->maximal_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index) return ConstRC::GeneralError;
	mpz_set(aValue,this->matrix[aRow][aColumn]);
	return ConstRC::Ok;
}

mpz_t* MpzMatrix::GetValueDirect(long aRow, long aColumn) {
	return &(this->matrix[aRow][aColumn]);
}

int MpzMatrix::PutValueMpz(long aRow, long aColumn, mpz_t aValue) {
	if ((aRow < 0) || (aRow > this->maximal_row_index))	{
		CLog::LogError(WHEREARG,"MpzMatrix","PutValueMpz",aRow,GetMaximalRowIndex(),"Bad row index.");
		abort();
		return ConstRC::ExcessiveRowIndex;
	}
	if ((aColumn < 0) || (aColumn > this->maximal_column_index))	{
		CLog::LogError(WHEREARG,"MpzMatrix","PutValueMpz",aColumn,GetMaximalColumnIndex(),"Bad column index.");
		abort();
		return ConstRC::ExcessiveColumnIndex;
	}
	if (modular)
			mpz_mod(matrix[aRow][aColumn], aValue, modulo);
		else
			mpz_set(this->matrix[aRow][aColumn],aValue);
	return ConstRC::Ok;
}

int MpzMatrix::PutValue(long aRow, long aColumn, integer_matrix_type aValue) {
	if ((aRow < 0) || (aRow > this->maximal_row_index))	{
		CLog::LogError(WHEREARG,"MpzMatrix","PutValue",aRow,GetMaximalRowIndex(),"Bad row index.");
		abort();
		return ConstRC::ExcessiveRowIndex;
	}
	if ((aColumn < 0) || (aColumn > this->maximal_column_index))	{
		CLog::LogError(WHEREARG,"MpzMatrix","PutValue",aColumn,GetMaximalColumnIndex(),"Bad column index.");
		abort();
		return ConstRC::ExcessiveColumnIndex;
	}
	mpz_set_si(matrix[aRow][aColumn],aValue);
	return ConstRC::Ok;
}

int MpzMatrix::ApplyModulo() {
	for (long row=0;row <=maximal_row_index; row ++) {
		for(long column = 0; column <= maximal_column_index; column ++) {
			mpz_mod(matrix[row][column], matrix[row][column], modulo);
		}
	}
}

bool MpzMatrix::Equals(MpzMatrix * aMatrix) {
	long rowmax;
	long colmax;
	if (rowmax = GetMaximalRowIndex() != aMatrix->GetMaximalRowIndex()) return false;
	if (colmax = GetMaximalColumnIndex() != aMatrix->GetMaximalColumnIndex()) return false;
	mpz_t val1, val2;
	mpz_init(val1);
	mpz_init(val2);
	for (long row = 0;row<=rowmax;row ++)
		for(long col = 0;col<=colmax;col++) {
			GetValueMpz(row,col,val1);
			aMatrix->GetValueMpz(row,col,val2);
			if (modular) {
				mpz_mod(val1,val1,modulo);
				mpz_mod(val1,val1,modulo);
			}
			if (mpz_cmp(val1,val2)==0) {
				mpz_clear(val1);
				mpz_clear(val2);
				return false;
			};
		}
	mpz_clear(val1);
	mpz_clear(val2);
	return true;
}

int MpzMatrix::ColumnCRT(long aColumnIndex, long aRows, MpzMatrix* aValues, long aValColIndex,
		mpz_t aResult, mpz_t * aMultiplier, bool allow_signed) {
	if (aRows > GetMaximalRowIndex()+1) return ConstRC::GeneralError;
	if (aRows > aValues->GetMaximalRowIndex()+1) return ConstRC::GeneralError;
	if (aColumnIndex > GetMaximalColumnIndex()) return ConstRC::GeneralError;
	if (aValColIndex > aValues->GetMaximalColumnIndex()) return ConstRC::GeneralError;
	mpz_t sum;
	mpz_t full_product;
	mpz_t partial_product;
	mpz_t inverse;
	mpz_t aux1;
	mpz_init(sum);
	mpz_init(full_product);
	mpz_init(partial_product);
	mpz_init(inverse);
	mpz_init(aux1);

	mpz_set_ui(full_product,1);
	mpz_set_ui(sum,0);
	for(long index=0;index<aRows;index++)
		mpz_mul(full_product,full_product,matrix[index][aColumnIndex]);

	// gmp_printf("\nFull product: %Zd",full_product);

	for(long index=0;index<aRows;index++) {
		if (mpz_cmp_ui(matrix[index][aColumnIndex],0) == 0) {
			PrintToScreen();
		};
		mpz_div(partial_product,full_product,matrix[index][aColumnIndex]);
		// gmp_printf("\nPartial product: %Zd",partial_product);

		mpz_invert(inverse,partial_product,matrix[index][aColumnIndex]);
		mpz_mul(aux1,inverse,*aValues->GetValueDirect(index,aValColIndex));
		if (aMultiplier)
			mpz_mul(aux1,aux1, *aMultiplier);
		mpz_mod(aux1,aux1,matrix[index][aColumnIndex]);
		mpz_mul(aux1,aux1,partial_product);
		mpz_add(sum,sum, aux1);
	}
	mpz_mod(aResult,sum,full_product);
	if (allow_signed) {
		mpz_div_ui(sum,full_product,2);
		if(mpz_cmp(aResult,sum)==1) {
			mpz_sub(aResult,aResult,full_product);
		};
	};
	mpz_clear(sum);
	mpz_clear(full_product);
	mpz_clear(partial_product);
	mpz_clear(inverse);
	mpz_clear(aux1);
	return ConstRC::Ok;
}

int MpzMatrix::GetValue(long aRow, long aColumn, integer_matrix_type & aValue) {
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->maximal_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index) return ConstRC::GeneralError;
	if (mpz_fits_slong_p(matrix[aRow][aColumn])) {
		aValue = mpz_get_si(matrix[aRow][aColumn]);
		return ConstRC::Ok;
	}
	return ConstRC::GeneralError;
}

int MpzMatrix::Add(MpzMatrix & aOperand1, MpzMatrix & 	aOperand2) {

};

int MpzMatrix::Multiply(MpzMatrix & aOperand1, MpzMatrix & 	aOperand2) {
	if (aOperand1.GetMaximalColumnIndex() != aOperand2.GetMaximalRowIndex()) {
		return ConstRC::GeneralError;
	}
	for (long row = 0; row <= GetMaximalRowIndex(); row ++) {
		for (long column = 0; column <= GetMaximalColumnIndex(); column ++) {
			mpz_set_ui(tempadd,0);
			for (long internal = 0; internal <= aOperand2.GetMaximalRowIndex(); internal++) {
				aOperand2.GetValueMpz(internal, column, tempval);
				aOperand1.GetValueMpz(row,internal,tempval2);
				mpz_mul(tempmul,tempval2,tempval);
				mpz_add(tempadd,tempadd,tempmul);
				if (modular) mpz_mod(tempadd, tempadd,modulo);
			};
			PutValueMpz(row,column, tempadd);
		}
	};
	return ConstRC::Ok;
};

int MpzMatrix::Multiply(IntegerSparseMatrix & aOperand1, MpzMatrix & aOperand2) {
	if (aOperand1.GetMaximalColumnIndex() != aOperand2.GetMaximalRowIndex()) {
		CLog::LogError(WHEREARG,"MpzMatrix","Multiply",aOperand1.GetMaximalColumnIndex(),aOperand2.GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::GeneralError;
	}
	for (long row = 0; row <= maximal_row_index; row ++) {
		for (int column  = 0; column <= maximal_column_index; column++){
			mpz_set_ui(tempadd,0);
			long index;
			long maxindex = aOperand1.CountNonZeroItemsOnRow(row);
			long rowindex = -1;
			for (int index = 0; index < maxindex; index++) {
				integer_matrix_type value;
				long common;
				if (int rc = aOperand1.IterateRow(row,rowindex,index,common,value) != ConstRC::Ok)
					return rc;
				aOperand2.GetValueMpz(common, column, tempmul);
				mpz_mul_ui(tempmul,tempmul,value);
				mpz_add(tempadd,tempadd,tempmul);
				if (modular) mpz_mod(tempadd, tempadd,modulo);
			}
			PutValueMpz(row,column, tempadd);
		}
	}
	return ConstRC::Ok;
}

int MpzMatrix::Multiply(AbstractIntegerMatrix & aOperand1, MpzMatrix & aOperand2) {
	IntegerSparseMatrix * ISM = dynamic_cast<IntegerSparseMatrix *>(&aOperand1);
	if (ISM) {
		return Multiply(* ISM,aOperand2);
	}
	if (aOperand1.GetMaximalColumnIndex() != aOperand2.GetMaximalRowIndex()) {
		cerr << "|" << aOperand1.GetMaximalColumnIndex() << " x " << aOperand2.GetMaximalRowIndex();
		return ConstRC::GeneralError;
	}
	for (long row = 0; row <= maximal_row_index; row ++) {
		for (int column  = 0; column <= maximal_column_index; column++){
			mpz_set_ui(tempadd,0);
			for (int common = 0; common <= aOperand2.maximal_row_index; common++) {
				integer_matrix_type value;
				aOperand1.GetValue(row,common,value);
				aOperand2.GetValueMpz(common, column, tempmul);
				mpz_mul_ui(tempmul,tempmul,value);
				mpz_add(tempadd,tempadd,tempmul);
				if (modular) mpz_mod(tempadd, tempadd,modulo);
			}
			PutValueMpz(row,column, tempadd);
		}
	}
	return ConstRC::Ok;
}

int MpzMatrix::Multiply(MpzMatrix & aOperand1, AbstractIntegerMatrix & aOperand2) {
	if (aOperand1.GetMaximalColumnIndex() != aOperand2.GetMaximalRowIndex()) {
		cerr << "|" << aOperand1.GetMaximalColumnIndex() << " x " << aOperand2.GetMaximalRowIndex();
		return ConstRC::GeneralError;
	}
	for (long row = 0; row <= maximal_row_index; row ++) {
		for (int column  = 0; column <= maximal_column_index; column++){
			mpz_set_ui(tempadd,0);
			for (int common = 0; common <= aOperand1.maximal_column_index; common++) {
				integer_matrix_type value;
				aOperand1.GetValueMpz(row,common,tempmul);
				aOperand2.GetValue(common, column, value);
				mpz_mul_ui(tempmul,tempmul,value);
				mpz_add(tempadd,tempadd,tempmul);
				if (modular) mpz_mod(tempadd, tempadd,modulo);
			}
			PutValueMpz(row,column, tempadd);
		}
	}
	return ConstRC::Ok;
}

int MpzMatrix::MultiplyTransposed(MpzMatrix & aOperand1, AbstractIntegerMatrix & aOperand2) {
	IntegerSparseMatrix * ISM = dynamic_cast<IntegerSparseMatrix *>(&aOperand2);
	if (ISM) {
		return MultiplyTransposed(aOperand1, * ISM);
	}
	return ConstRC::NotSupported;
}

int MpzMatrix::MultiplyTransposed(MpzMatrix & aOperand1, IntegerSparseMatrix & aOperand2) {
	if (aOperand1.GetMaximalColumnIndex() != aOperand2.GetMaximalColumnIndex()) {
		return ConstRC::GeneralError;
	}
	for (long row = 0; row <= maximal_row_index; row ++) {
		for (int row2  = 0; row2 <= maximal_column_index; row2++){
			mpz_set_ui(tempadd,0);
			long index;
			long maxindex = aOperand2.CountNonZeroItemsOnRow(row2);
			long row2index = -1;
			for (int index = 0; index < maxindex; index++) {
				integer_matrix_type value;
				long common;
				aOperand2.IterateRow(row2,row2index,index,common,value);
				aOperand1.GetValueMpz(row, common, tempmul);
				mpz_mul_ui(tempmul,tempmul,value);
				mpz_add(tempadd,tempadd,tempmul);
				if (modular) mpz_mod(tempadd, tempadd,modulo);
			}
			PutValueMpz(row,row2, tempadd);
		}
	}
	return ConstRC::Ok;
}

int  MpzMatrix::Multiply(long aValue) {
	for (long row=0;row <=maximal_row_index; row ++) {
		for(long column = 0; column <= maximal_column_index; column ++) {
			mpz_mul_si(matrix[row][column],matrix[row][column], aValue);
			if (modular) mpz_mod(matrix[row][column], matrix[row][column],modulo);
		}
	}
	return ConstRC::Ok;
}

int  MpzMatrix::Multiply(mpz_t aValue) {
	for (long row=0;row <=maximal_row_index; row ++) {
		for(long column = 0; column <= maximal_column_index; column ++) {
			mpz_mul(matrix[row][column],matrix[row][column], aValue);
			if (modular) mpz_mod(matrix[row][column], matrix[row][column],modulo);
		}
	}
	return ConstRC::Ok;
}
