#include "mserializable.h"

#include "base_parameters.h"
#include <time.h>
#include <sstream>
#include <fstream>
#include <iostream>
#include "const_return_codes.h"
#include "log.h"
#include "common_definitions.h"

using namespace std;

// GETTERS/SETTERS

void MSerializable::Set_DefaultFileName(string aDefaultFileName)
{ 
	this->iDefaultFileNameMutex.LockWrite();
	this->iDefaultFileName = aDefaultFileName; 
	this->iDefaultFileNameMutex.UnlockWrite();
}
string MSerializable::Get_DefaultFileName() const
{ 
	this->iDefaultFileNameMutex.LockRead();
	string temp = this->iDefaultFileName;
	this->iDefaultFileNameMutex.UnlockRead();
	return temp; 
}

int MSerializable::Serialize(const CBaseParameters& aParam) const
{
	int compressionLevel;

	// get parameters for serialization - with default values
	aParam.GetIntegerParameter(PARAM_COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, compressionLevel);
	
	// get file name with path
	string defaultFileName = Get_DefaultFileName();
	string fileNameFull = MSerializable::CreateFileNameAuto(aParam, defaultFileName);

	xmlTextWriterPtr writer = xmlNewTextWriterFilename(fileNameFull.c_str(), compressionLevel);
	// we write only start of document now
	int rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
    if (rc < 0) 
	{
	    CLog::LogError(WHEREARG,"MSerializable","Serialize",rc,0,"Could not write the document start.");
		xmlFreeTextWriter(writer);
		return ConstRC::GeneralError;        	
    }
	// do rest work 
	rc =  Serialize(aParam, writer);	
	if (rc != ConstRC::Ok) 
	{
		xmlFreeTextWriter(writer);
		return rc;
	}
	// end the document
	rc = xmlTextWriterEndDocument(writer);

	xmlFreeTextWriter(writer);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"MSerializable","Serialize",rc,0,"Could not close the document.");
		return ConstRC::GeneralError;
    }

	return ConstRC::Ok;
}


int MSerializable::Deserialize(const CBaseParameters &aParam)
{
	// get file name with path
	string defaultFileName = Get_DefaultFileName();
	string fileNameFull = MSerializable::CreateFileNameAuto(aParam, defaultFileName);

	xmlTextReaderPtr reader = xmlNewTextReaderFilename(fileNameFull.c_str());
	if (reader == NULL)
	{
		CLog::LogError(WHEREARG,"MSerializable","Deserialize",0,0,"Could not create xml reader.");
		return ConstRC::GeneralError;
	}
	int rc = CommonXMLHelper::MoveToNextNode(reader);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextReader(reader);
		if (rc == ConstRC::EndOfFile)
		{
			return ConstRC::Ok;
		}
		return rc;
	}
	
	rc = Deserialize(aParam, reader);	
	
	xmlFreeTextReader(reader);

	return rc;
}




string MSerializable::CreateFileNameAuto(const CBaseParameters& aParam, const string& aFileNameDefault)
{
	string identifier;
	string directoryName;
	string fileName;
	int compressionLevel;

	if (aParam.ContainsParameter(PARAM_SERIALIZATION_FULL_FILENAME))
	{
		// we have already full name of the file
		aParam.GetStringParameter(PARAM_SERIALIZATION_FULL_FILENAME, fileName);
		if (fileName.length() > 0)
		{
			return fileName;
		}
	}

	// get parameters for serialization - with default values
	aParam.GetIntegerParameter(PARAM_COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, compressionLevel);
	aParam.GetStringParameter(PARAM_SERIALIZATION_IDENTIFIER, "", identifier);
	aParam.GetStringParameter(PARAM_SERIALIZATION_DIRECTORY, "", directoryName);
	aParam.GetStringParameter(PARAM_SERIALIZATION_FILE, aFileNameDefault, fileName);

	// control path separator
	MSerializable::AdaptPathSeparator(directoryName);
	
	// get file name with path
	if (compressionLevel == 0)
	{
		return MSerializable::CreateFileNamePlain(directoryName, fileName, identifier, XML_FILE_EXTENSION);
	}
	else
	{
		return MSerializable::CreateFileNameCompressed(directoryName, fileName, identifier, XML_FILE_EXTENSION);
	}
}

string MSerializable::CreateFileNameAuto(const CBaseParameters& aParam, const char* aFileNameDefault)
{
	string fileName;

	if (aParam.ContainsParameter(PARAM_SERIALIZATION_FULL_FILENAME))
	{
		// we have already full name of the file
		aParam.GetStringParameter(PARAM_SERIALIZATION_FULL_FILENAME, fileName);
		if (fileName.length() > 0)
		{
			return fileName;
		}
	}

	int compressionLevel;
	string identifier;
	string directoryName;
	string fileNameDefault(aFileNameDefault);

	// get parameters for serialization - with default values
	aParam.GetIntegerParameter(PARAM_COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, compressionLevel);
	aParam.GetStringParameter(PARAM_SERIALIZATION_IDENTIFIER, "", identifier);
	aParam.GetStringParameter(PARAM_SERIALIZATION_DIRECTORY, "", directoryName);
	aParam.GetStringParameter(PARAM_SERIALIZATION_FILE, fileNameDefault, fileName);
	
	// control path separator
	MSerializable::AdaptPathSeparator(directoryName);

	// get file name with path
	if (compressionLevel == 0)
	{
		return MSerializable::CreateFileNamePlain(directoryName, fileName, identifier, XML_FILE_EXTENSION);
	}
	else
	{
		return MSerializable::CreateFileNameCompressed(directoryName, fileName, identifier, XML_FILE_EXTENSION);
	}
}


string MSerializable::CreateNewIdentifier()
{
	time_t rawtime;
	tm * timeinfo;
	char buffer [SERIALIZATION_IDENTIFIER_DATETIME_LENGTH + 1]; // for final 0
	
	// random initialization
	srand(0);

	// local datetime
	time(&rawtime);
	timeinfo = localtime(&rawtime);
	strftime (buffer,SERIALIZATION_IDENTIFIER_DATETIME_LENGTH + 1,SERIALIZATION_IDENTIFIER_DATETIME_FORMAT,timeinfo);
	string str(buffer);

	//str.append(FILE_NAME_SEPARATOR);
	//
	//// three random numbers
	//stringstream stream;
	//int digit = rand() % 10;
	//stream << digit;
	//digit = rand() % 10;
	//stream << digit;
	//digit = rand() % 10;
	//stream << digit;

	//str.append(stream.str());

	return str;
}
void MSerializable::AppendPathSeparator(string& aPath)
{
	if (aPath.length() > 0)
	{
		char last = aPath[aPath.length()-1];
		// PATH_SEPARATOR is defined in definitions.h
		if (last != PATH_SEPARATOR)
		{				
			aPath += PATH_SEPARATOR;				
		}
	} 
}
void MSerializable::AdaptPathSeparator(string& aPath)
{
	if (aPath.find(PATH_SEPARATOR_OTHER) != string::npos)
	{
		for (unsigned int i = 0; i < aPath.length(); i++)
		{
			if (aPath[i] == PATH_SEPARATOR_OTHER)
			{
				aPath[i] = PATH_SEPARATOR;
			}
		}
	}
}
/**
	This function does not return reference, but a copy of 
	the string.
*/
string MSerializable::CreateFileNamePlain(const char* aDirectoryName, const char* aFileName, const char* aIdentifier, const char* aExtension)
{
	string str;
	if (aDirectoryName)
	{
		str.append(aDirectoryName);
		AppendPathSeparator(str);
	}
	if (aFileName)
	{
		str.append(aFileName);
	}
	if (aIdentifier)
	{
		str.append(FILE_NAME_SEPARATOR);
		str.append(aIdentifier);
	}
	if (aExtension)
	{
		str.append(aExtension);
	}
	//printf("MSerializable: Create fn plain: %s",str.c_str());

	return str;
}	

/**
	This function does not return reference, but a copy of 
	the string.
*/
string MSerializable::CreateFileNamePlain(const string& aDirectoryName, const string& aFileName, const string& aIdentifier, const string& aExtension)
{
	string str;
	if (aDirectoryName.length() > 0)
	{
		str.append(aDirectoryName);
		AppendPathSeparator(str);
	}
	if (aFileName.length() > 0)
	{
		str.append(aFileName);
	}
	if (aIdentifier.length() > 0)
	{
		str.append(FILE_NAME_SEPARATOR);
		str.append(aIdentifier);
	}
	if (aExtension.length() > 0)
	{
		str.append(aExtension);
	}
	//printf("MSerializable: Create fn plain: %s",str.c_str());

	return str;
}

string MSerializable::CreateFileNameCompressed(const char* aDirectoryName, const char* aFileName, const char* aIdentifier, const char* aExtension)
{
	string str = CreateFileNamePlain(aDirectoryName, aFileName, aIdentifier, aExtension);
	str.append(COMPRESSED_FILE_EXTENSION);
	//printf("MSerializable: Create fn compressed: %s",str.c_str());

	return str;
}

string MSerializable::CreateFileNameCompressed(const string& aDirectoryName, const string& aFileName, const string& aIdentifier, const string& aExtension)
{
	string str = CreateFileNamePlain(aDirectoryName, aFileName, aIdentifier, aExtension);
	str.append(COMPRESSED_FILE_EXTENSION);
	//printf("MSerializable: Create fn compressed: %s",str.c_str());

	return str;
}

	// Removes both XXX.xml and XXX.xml.z from the directory
void MSerializable::RemoveFile(const char* aDirectoryName, const char* aFileName, const char* aIdentifier, const char* aExtension)
{
	if (aFileName == NULL) return;
	// .xml file
	string str = CreateFileNamePlain(aDirectoryName, aFileName, aIdentifier, aExtension);
	//printf("MSerializable: removing file %s\n",str.c_str());
	remove(str.c_str());
	// .xml.z file
	str = CreateFileNameCompressed(aDirectoryName, aFileName, aIdentifier, aExtension);
	remove(str.c_str());
}
	// Removes both XXX.xml and XXX.xml.z from the directory
void MSerializable::RemoveFile(const string& aDirectoryName, const string& aFileName, const string& aIdentifier, const string& aExtension)
{
	if (aFileName.length() == 0) return;
	// .xml file
	string str = CreateFileNamePlain(aDirectoryName, aFileName, aIdentifier, aExtension);
	//printf("MSerializable: removing file %s\n",str.c_str());
	remove(str.c_str());
	// .xml.z file
	str = CreateFileNameCompressed(aDirectoryName, aFileName, aIdentifier, aExtension);
	remove(str.c_str());
}
FILE* MSerializable::OpenFile(const char* aMode, const char* aFilePath)
{
	FILE* result = fopen(aFilePath,aMode);
	if (!result)
	{
		printf("MSerializable - Open File, Could not open file.\n");
	}
	return result;
}
bool MSerializable::ExistFile(const std::string& aFullFileName)
{
	FILE* result = fopen(aFullFileName.c_str(), "r");
	
	if (result != NULL)
	{
		fclose(result);
		return true;
	}

	return false;
}
int MSerializable::MCopyFile(const string& aInFile, const string& aOutFile)
{
	char buff[IO_BUFF_SIZE];
	int readBytes = 1;

	ifstream inFileStream(aInFile.c_str(), ios::in|ios::binary);
	if (!inFileStream)
	{
		return ConstRC::GeneralError;
	}

	remove(aOutFile.c_str()); // if exists
	ifstream tmpStream(aOutFile.c_str());
	if (tmpStream)
	{
		return ConstRC::GeneralError;
	}
	tmpStream.close();

	ofstream outFileStream(aOutFile.c_str(), ios::out|ios::binary);
	if(!outFileStream)
	{
		return ConstRC::GeneralError;
	}

	while (readBytes != 0)
	{
		inFileStream.read((char*)buff, IO_BUFF_SIZE);
		readBytes = inFileStream.gcount();
		outFileStream.write((char*)buff, readBytes);
	}

	return ConstRC::Ok;
}
int MSerializable::RenameFile(const string& aOldFullFileName, const string& aNewFullFileName, bool aRemoveIfExists)
{
	if (aRemoveIfExists)
	{
		remove(aNewFullFileName.c_str());
	}

	return rename(aOldFullFileName.c_str(), aNewFullFileName.c_str()) == 0 ? ConstRC::Ok : ConstRC::GeneralError;
}
int MSerializable::PrepareXMLReader(const CBaseParameters& aParam, const string& aElementName, xmlTextReaderPtr& aReader)
{
	if (aReader != NULL)
	{
		xmlFreeTextReader(aReader);
		aReader = NULL;
	}

	// get file name with path
	string fileNameFull = MSerializable::CreateFileNameAuto(aParam, FILE_NAME_MSERIALIZABLE);

	aReader = xmlNewTextReaderFilename(fileNameFull.c_str());
	if (aReader == NULL)
	{
		CLog::LogError(WHEREARG,"MSerializable","PrepareXMLReader",0,0,"Could not create xml reader.");
		return ConstRC::GeneralError;
	}
	int rc;	
	xmlChar* name;

	while ((rc = CommonXMLHelper::MoveToNextNode(aReader)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST aElementName.c_str()))
		{
			rc = ConstRC::Ok;
			break;
		}
	}
	
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextReader(aReader);
		aReader = NULL;
		rc = ConstRC::GeneralError;
	}

	return rc;
}
