#include "node.h"
#include "abstract_connector.h"
#include "abstract_serializator.h"
#include "filesystem_connector.h"
#include "filesystem_message.h"
#include "distributed_algorithm.h"
#include "center_info.h"
#include <iostream>
#include <string.h>
#include "xml_message_definitions.h"

//const unsigned int Node::KDataHarvestDefaultIntervalSecs = 60;
const unsigned int Node::KDataHarvestDefaultIntervalSecs = 10;
const unsigned int Node::KNodeInstanceCheckIntervalSecs = 3;

Node::Node(AbstractSerializator& aSerializator) : Receiver(aSerializator,NULL), iCurrentlyLockedCenter(NULL), iLocked(FALSE), iCounter(1), iDataHarvestIntervalSecs(KDataHarvestDefaultIntervalSecs)
{
	iBreakNow = FALSE;
}

void	Node::RegisterCenter(CenterInfo* aInfo)
{
	int res;
	if (aInfo)
	{
		// At first, we check whether the info is not a duplicite
		// of some existing center/failed center
		int i, max = iCenters.size();

		CenterInfo* previous;
		for (i=0; i< max; i++)
		{
			previous = iCenters.at(i);
			if (previous)
			{
				if (aInfo->IsEqual(previous))
				{
				iDupliciteCenters.push_back(aInfo);
				return;
				}
			}
		}
		max = iFailedCenters.size();
		for (i=0; i< max; i++)
		{
			previous = iFailedCenters.at(i);
			if (previous)
			{
				if (aInfo->IsEqual(previous))
				{
				iDupliciteCenters.push_back(aInfo);
				return;
				}
			}
		}
		// Now, we will walk through the existing connectors and 
		// if no connector for this 
		AbstractConnector* conn = FindConnectorByURL(aInfo->GetUrlString());	
		if (!conn)
		{
			conn = AddConnector(aInfo);			
			if (conn)
			{
				res = ConstRC::Ok;
			} else
			{
				res = GENERAL_ERROR;
				string r2 = COULD_NOT_CREATE_CONNECTOR_FOR_URL;
				r2 += "'";
				r2 += aInfo->GetUrlString();
				r2 += "'";
				aInfo->SetFailureReason(r2);
			}
		}
		aInfo->SetExternalConnector(conn);		
		if (res == ConstRC::Ok)
		{			
			int state = conn->GetState();			
			if (state == KOk)	
			{
				iCenters.push_back(aInfo);			
			} 
			else
			{	
				aInfo->SetFailureReason(conn->GetStateExplanation(state));
				iFailedCenters.push_back(aInfo);			
			}
		} else
		{
			iFailedCenters.push_back(aInfo);
		}		
	}
}

/*
	Parses and registers string of center infos.
	Method is used for parsing "centers" value from xml configuration.
*/
int Node::ParseCenters(const char *str, const char *separ) {
	char *s, *ptr;
	CenterInfo *ci;

	s = (char *)malloc((strlen(str)+1)*sizeof(char));
	strcpy(s, str);

	ptr = strtok(s, separ);
	while (ptr) {
		ci = CenterInfo::Parse(ptr);
		if (!ci) {
			free(s);
			return ConstRC::GeneralError;
		}
		RegisterCenter(ci);
		ptr = strtok(NULL, separ);
	}

	free(s);

	return ConstRC::Ok;
}

Node::~Node()
{	
	int i, size = iCenters.size();
	for (i=0; i< size; i++)
	{
		delete iCenters.at(i);
	}
	size = iFailedCenters.size();
	for (i=0; i<size; i++)
	{
		delete iFailedCenters.at(i);
	}
	size = iDupliciteCenters.size();
	for (i=0; i< size; i++)
	{
		delete iDupliciteCenters.at(i);
	}
}

int	Node::CheckInitConsistency()
{
	int ret = Receiver::CheckInitConsistency();
	if (ret != ConstRC::Ok)
	{
		return ret;
	}
	int size = iCenters.size();
	if (size <= 0)
	{
		printf("Node.CheckInitConsistency: No center list available. Run aborted.\n");
		PrintCenters(); // print failed centers too.
		return NO_CENTERS_DEFINED;
	}
	return ConstRC::Ok;
}

void	Node::PrintCenters()
{
	int i, size = iCenters.size();
	printf("\nNode: %d center(s) available.\n",size);
	for (i=0; i< size; i++)
	{
		iCenters.at(i)->Print();
	}
	size = iFailedCenters.size();
	if (size > 0)
	{
		printf("\nNode: %d center(s) FAILED during init. List follows.\n",size);
		for (i=0; i< size; i++)
		{
			iFailedCenters.at(i)->Print();
		}
	}
	size = iDupliciteCenters.size();
	if (size > 0)
	{
		printf("\nNode: %d duplicite center(s) detected, only the first occurence will be used. Command line takes precedence over local XML center definitions.\n",size);
	}
}

int	Node::Run()
{
	int wakeCounter = 0, ret;
	ret = CheckInitConsistency();
	if (ret != ConstRC::Ok)
	{
		return ret;
	}
	PrintCenters();
	// Send "ready" messages etc.
	Unlock();
	// The main loop. Exit from this loop will mean exit of the node process.
	do {
		// Defined as a macro in definitions.h
		SLEEP(1); // Sleep for 1 second		
		if (!(iDistAlg->NodeInstanceRunning())) printf("\rNode woke up ... %7d sec   ",wakeCounter);
		wakeCounter++;
		// In case that the distributed algorithm is running now ...
		if (iLocked && iCurrentlyLockedCenter)
		{
			// ... and the periodic interval for sending data is now
			if (!(wakeCounter % iDataHarvestIntervalSecs)) {
				// Ask the algorithm about quantity of data ready to be sent.
				if (iDistAlg->HasDataToSend())
				{				
					iDistAlg->LockDataMutex();			
					SendFreshDataMessageToCenter(iCurrentlyLockedCenter);
					iDistAlg->ClearFreshData();
					iDistAlg->UnlockDataMutex();
				} 
				else // No data to send.
				{
					// We just send an "Alive" message to the center,
					// so that it knows that we're alive.
					SendAliveMessageToCenter(iCurrentlyLockedCenter);
				}
			}

			// ... and node instance check interval expired
			if (!(wakeCounter % KNodeInstanceCheckIntervalSecs)) {
				if ((!(iDistAlg->NodeInstanceRunning())) && (!(iDistAlg->HasDataToSend()))) Unlock();
			}
		}
		FetchNewMessages();
		ResendPendingMessages();
		fflush(stdout);
	} while (!iBreakNow);
	return ConstRC::Ok;
}

void	Node::Break()
{
	iBreakNow = TRUE;
}

void	Node::LoadCenterList(const char* aDirectoryName, const char* aFileName)
{
	if ((!aDirectoryName) || (!aFileName))
	{
		printf("Node.LoadCenterList from dir: could not load info from a NULL-named directory/file.\n");
		return;
	}	
	int counter = 0;
	string str1 = aDirectoryName;
	AbstractSerializator::AppendPathSeparator(str1);
	str1 += aFileName;	
	xmlTextReaderPtr reader  = xmlNewTextReaderFilename(str1.c_str());
	int type;
	xmlChar* name;
	int rc = xmlTextReaderRead(reader);		
	while (rc == 1) {		
        	type = xmlTextReaderNodeType(reader);
		name = xmlTextReaderName(reader);
		//printf("Name: %s, type: %d.\n",name,type);
		if (type == 1)
		{				
			if (xmlStrEqual(name,BAD_CAST CENTER_INFO_NAME))
			{
				CenterInfo* info = CenterInfo::Deserialize(reader);
				if (info)
				{
				RegisterCenter(info);
				counter++;
				}				
			}
		} // if (type == 1)
		rc = xmlTextReaderRead(reader);
	} // end of while(TRUE)
	xmlFreeTextReader(reader);
	if (counter)
	{
	printf("Node: %d center(s) read from XML file \"%s\".\n",counter,str1.c_str());
	}
}

void	Node::Unlock()
{
	iLocked = FALSE;	
	iCurrentlyLockedCenter = NULL;
	int i, max = iCenters.size();
	CenterInfo* ci;
	// At first, we clear the pending messages in all center connectors.
	for (i=0; i< max; i++)
	{
		ci = iCenters.at(i);
		if (ci && ci->HasConnector())
		{
			// TODO: Delete all messages from all connectors?
			// I am not sure whether this is going to work as expected. Will not fresh Ready messages get deleted
			ci->GetConnector().ClearPending();
		}
	}
	// Then, we will resend the ready messages. I am not sure
	// whether these two cycles would work if joined into one.
	// The ClearPending() could erase fresh Ready messages from previous
	// iterations of the cycle.
	// That is why the two cycles are kept separate.
	printf("Unlock Node: altogether %d centers ready\n",max);
	for (i=0; i< max; i++)
	{
		ci = iCenters.at(i);
		if (ci)
		{
			if (ci->HasConnector())
			{
			printf("     Connector with URL %s\n",ci->GetConnector().Get_URLString().c_str());
			}			
			SendReadyMessageToCenter(ci);
		}
	}
}

void	Node::SendNodeBusyMessageToCenter(CenterInfo* aInfo, unsigned int aInReplyTo)
{
	SendSimpleMessageToCenter(aInfo, aInReplyTo, ENodeBusy, "Node busy");
}

void	Node::SendLockedMessageToCenter(CenterInfo* aInfo, unsigned int aInReplyTo)
{
	SendSimpleMessageToCenter(aInfo, aInReplyTo, ELocked, "Locked");
}

void	Node::SendAckMessageToCenter(CenterInfo* aInfo, unsigned int aInReplyTo)
{
	SendSimpleMessageToCenter(aInfo, aInReplyTo, EAck, "Confirmation");
}

void	Node::SendAliveMessageToCenter(CenterInfo* aInfo)
{
	SendSimpleMessageToCenter(aInfo, 0, EAlive, "Alive");
}

void	Node::SendSimpleMessageToCenter(CenterInfo* aInfo, unsigned int aInReplyTo, message_type aType, const char* aSubject)
{
	if (!aInfo)
	{
		return;
	}
	if (aInfo->HasConnector())
	{
		AbstractConnector& connector = aInfo->GetConnector();
		URL* url = connector.Get_URL();
		switch (connector.Get_ConnectorType())
		{
		case EFileConnector:
			{
				if (url)
				{
					// Ack messages have no timeout
					FilesystemMessage* message = new FilesystemMessage(url->GetFullPath().c_str(),aSubject,aType, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",NO_TIMEOUT);
					if (aType == EAlive)
					{
						message->Set_JobId(iJobId);
					} else
					{
						// All those messages except Alive
						// are replies.
						message->Set_ReplyTo(aInReplyTo);
					}
				// The method "send" will take care of deleting the message eventually
					connector.Send(message);
// 				PrintInfoMessage(aInfo->GetId(),cout);
// 				PrintInfoMessage("Sent Ready over the file connector.",cout);	
				}
				
			}
			break;
		default:
			{
				PrintErrorMessage("Cannot send Simple message to a center with unknown connector type.",cerr);
			}
			break;
		}
	} else
	{
		PrintErrorMessage("Node::Cannot send Simple message to a center with no connector.",cerr);
	}
}

void    Node::SendReadyMessageToCenter(CenterInfo* aInfo, unsigned int aDelay)
{	
	if (!aInfo)
	{
		return;
	}
	if (aInfo->HasConnector())
	{
		AbstractConnector& connector = aInfo->GetConnector();
		URL* url = connector.Get_URL();
		switch (connector.Get_ConnectorType())
		{
		case EFileConnector:
			{
				if (url)
				{
					FilesystemMessage* message = new FilesystemMessage(url->GetFullPath().c_str(),"Waiting",EReady, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",aInfo->GetTimeout());
					if (iDistAlg)
					{
						message->AddSupportedAlgorithm(iDistAlg->AlgorithmName());
					}
					else
					{
						message->AddSupportedAlgorithm("No algorithm registered");
					}
					message->Set_Delay(aDelay);
					// The method "send" will take care of deleting the message eventually
					connector.Send(message);
				}
// 				PrintInfoMessage(aInfo->GetId(),cout);
// 				PrintInfoMessage("Sent Ready over the file connector.",cout);
				
			}
			break;
		default:
			{
				PrintErrorMessage("Cannot send Ready to a center with unknown connector type.",cerr);
			}
			break;
		}
	} else
	{
		PrintErrorMessage("Node::Cannot send Ready to a center with no connector.",cerr);
	}
}

void    Node::SendFreshDataMessageToCenter(CenterInfo* aInfo)
{	
	if (!aInfo)
	{
		return;
	}
	//printf("\n\n\n\n\nSend Fresh Data Message To Center\n\n\n\n\n");
	//iSerializator.PrintInstance();
	if (aInfo->HasConnector())
	{
		AbstractConnector& connector = aInfo->GetConnector();
		URL* url = connector.Get_URL();
		switch (connector.Get_ConnectorType())
		{
		case EFileConnector:
			{
				if (url)
				{
					FilesystemMessage* message = new FilesystemMessage(url->GetFullPath().c_str(),"Data",EData, 0, iCounter++, connector.Get_Identification(), aInfo->GetId(), "",aInfo->GetTimeout());
					message->Set_JobId(iJobId);
					message->AddData(&iSerializator);
					message->Set_DataCount(iDistAlg->DataToSend());
					message->Set_DataUnit(iDistAlg->DataUnit());
					// The method "send" will take care of deleting the message eventually
					connector.Send(message);
				}
// 				PrintInfoMessage(aInfo->GetId(),cout);
// 				PrintInfoMessage("Sent Ready over the file connector.",cout);
				
			}
			break;
		default:
			{
				PrintErrorMessage("Cannot send Data to a center with unknown connector type.",cerr);
			}
			break;
		}
	} else
	{
		PrintErrorMessage("Node::Cannot send Data to a center with no connector.",cerr);
	}
}


const char*	Node::GetThisClassName()			const
{
	return "Node";
}

void	Node::ProcessReceivedMessage(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return;
	}
	bool res = CheckMessageSequenceConsistency(aMessage);
	if (res)
	{
		Receiver::ProcessReceivedMessage(aMessage);		
		UpdateCommunicationInfo(aMessage);		
		ProcessMessageByType(aMessage);
	} else
	{
		printf("Node: did not know how to process inconsistent situation.\n");
	}
	
}

bool	Node::ProcessMessageByType(AbstractMessage* aMessage)
{
	if (Receiver::ProcessMessageByType(aMessage))
	{
		return TRUE;
	}
	if (!aMessage)
	{
		return FALSE;
	}
	switch (aMessage->Get_Type())
	{
		case ENoJob:
				return ProcessNoJobMessage(aMessage);			
		case EJobParameters:
				return ProcessJobParametersMessage(aMessage);
		case EQuit:
				return ProcessQuitMessage(aMessage);
		default:        printf("Node: process received message by type not fully implemented yet");
				break;
	}
	return FALSE;
}

bool	Node::ProcessQuitMessage(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return FALSE;
	}
	unsigned int counter  = aMessage->Get_Counter();
	CenterInfo* ptr = FindCenterByName(aMessage->Get_SenderId());
	
	//printf("Node::ProcessNoJobMessage. Received no job from center \"%s\", should ask again in %d seconds.\n",aMessage->Get_SenderId(),aMessage->Get_AskAgain());
	aMessage->MarkAsFinished(GetKeepMessages());
	string jobId = aMessage->Get_JobId();	
	string quitR = aMessage->Get_QuitReason();
	if (jobId == iJobId)
	{
		printf("Node::ProcessQuitMessage. Received quit message for current JobId (%s)\n    Reason stated as: \"%s\"",jobId.c_str(),quitR.c_str());
		iDistAlg->InterruptNodeNow();
		iLocked = FALSE;
	} else
	{
		printf("Node::ProcessQuitMessage. Received quit message for wrong JobId (%s). Current JobId is %s.\n    Reason stated as: \"%s\"\n",jobId.c_str(), iJobId.c_str(),quitR.c_str());
	}
	if (ptr)
	{
		// Send ack
		SendAckMessageToCenter(ptr,counter);
	}
	return TRUE;
}


bool	Node::ProcessNoJobMessage(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return FALSE;
	}
	unsigned int askAgain = aMessage->Get_AskAgain();
	unsigned int counter  = aMessage->Get_Counter();
	CenterInfo* ptr = FindCenterByName(aMessage->Get_SenderId());
	
	//printf("Node::ProcessNoJobMessage. Received no job from center \"%s\", should ask again in %d seconds.\n",aMessage->Get_SenderId(),aMessage->Get_AskAgain());
	aMessage->MarkAsFinished(GetKeepMessages());	

	if (ptr)
	{
		// Send ack
		SendAckMessageToCenter(ptr,counter);
		// Send future ready	
		SendReadyMessageToCenter(ptr,askAgain);
	}
	return TRUE;
}

void	Node::RunDistNode()
{
	iDistAlg->RunNodeInstance();
}

// C-function for pthread_create. We MAY NOT use a C++ member function for this purpose
void* ExecuteNodeThread(void* aParameter)
{
 	Node* node;
   	node = (Node*) aParameter;
   	//printf("Hello World! It's me, thread for node %p!\n", node);
	node->RunDistNode();
//	pthread_exit(NULL);
	return NULL;
}

void	Node::CreateNodeThread()
{
	
	int rc = pthread_create(&threads[0], NULL, ExecuteNodeThread, (void *)this);
	if (rc){
         	printf("ERROR; return code from pthread_create() is %d\n", rc);
         	exit(-1);
      	}
	
}

bool	Node::ProcessJobParametersMessage(AbstractMessage* aMessage)
{
	if (!iDistAlg)
	{
		printf("Node::ProcessJobParametersMessage:  No distribute algorithm.\n");
		return FALSE;
	}
	if (!aMessage)
	{
		return FALSE;
	}
	CenterInfo* ptr = FindCenterByName(aMessage->Get_SenderId());
	unsigned int counter  = aMessage->Get_Counter();
	aMessage->MarkAsFinished(GetKeepMessages());	
	// Node already busy.
	if (iDistAlg->NodeInstanceRunning())
	{
		SendNodeBusyMessageToCenter(ptr,counter);
		return FALSE;
	}
	iJobId = aMessage->Get_JobId();	
	// Node will lock now.
	iLocked = TRUE;
	JobParameters*  params = iDistAlg->CreateNewParameters();
	string bodypath = aMessage->Get_BodyPath();
	iSerializator.DeserializeJob(bodypath.c_str(),*params);
	iDistAlg->SetupParameters(*params);
	fprintf(stdout, "\n");
	params->Print();
	aMessage->MarkAsFinished(GetKeepMessages());	
	iDistAlg->DeleteParameters(params);

	if (ptr)
	{		
		// Send locked
		SendLockedMessageToCenter(ptr, counter);		
		iCurrentlyLockedCenter = ptr;
		
	}
	CreateNodeThread();	
	return TRUE;
}

CenterInfo*	Node::FindCenterByName(const char* aName) const
{
	int i, max = iCenters.size();
	CenterInfo* ptr = NULL;
	for (i=0; i< max; i++)
	{
		ptr = iCenters[i];
		if (ptr)
		{
			const char* centerName = ptr->GetId();
			if (centerName && (!strcmp(centerName,aName)))
			{
				return ptr;
			}
		}
	}
	return NULL;
}

void	Node::ResendPendingMessages()
{
	if (!iLocked)
	{
		int i, max = iCenters.size();
		CenterInfo* ci;
		// At first, we clear the pending messages in all center connectors.
		for (i=0; i< max; i++)
		{
			ci = iCenters.at(i);
			if (ci && ci->HasConnector())
			{			
				ci->GetConnector().ResendPending();
			}
		}
	} else
	{
		// iLocked. Will only communicate with the "chosen" center.
	}
}


bool	Node::CheckMessageSequenceConsistency(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return FALSE;
	}
	bool isOk = Receiver::CheckMessageSequenceConsistency(aMessage);
	if (!isOk)
	{
		// Some mismatch in message sequence
		printf("Node::Message Sequence Consistency. Some kind of error, the newly received message has lower counter than some previous.\n");
// 		aMessage->PrintHeaderInfo();
	}
	return isOk;
}

void	Node::RemoveFromPending(const char* aTargetName, unsigned int aCounter)
{
	if (!aTargetName)
	{
		return;
	}	
	CenterInfo* ptr = FindCenterByName(aTargetName);
	if (ptr)
	{
		if (ptr->HasConnector())
		{
			AbstractConnector& ctr = ptr->GetConnector();
			RemoveFromPending_Connector(&ctr,aTargetName,aCounter);
		}			
	}
	
}
