#include <cstring>
#include "normal_matrix_class.h"
#include "common_utils.h" // because of allocation machine
#include "const_return_codes.h"
#include "log.h"

long NormalMatrix::bits_in_n_m_t = 8*sizeof(integer_matrix_type);

/*!
	The default constructor does not take any parameters and constructs an instance of a "generic" normal matrix,
	with unknown dimensions. The dimensions may be later set by appropriate setter methods.

*/
NormalMatrix::NormalMatrix() : MSerializable(FILE_NAME_NORMAL_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->matrix = NULL;
	this->species = SNormalMatrix;
}


/*!
	The second constructor constructs an instance of a normal matrix with known dimensions. Beware that the parameters
	taken mean the actual number of rows and columns, and not their maximal indices; so, if we want to construct
	a matrix of dimensions 17x32, we call
	
		NormalMatrix* nm = new NormalMatrix(17,32);
		
	Now, we have an instance of a normal matrix; its member variables will be set to:
	
		nm->maximal_row_index = 16;
		
		nm->maximal_column_index = 31;
		
		nm->maximal_allocated_row_index = -1; // no initialization of the data array yet!
		
		nm->maximal_allocated_column_index = -1; // no initialization of the data array yet!
		
	Both of the constructors initialize the data array to NULL pointer, and a real allocation is performed later	- at the time of need. This programming pattern is called lazy initialization, and helps to reduce runtime memory requirements.

*/
NormalMatrix::NormalMatrix(long aRows, long aColumns) : MSerializable(FILE_NAME_NORMAL_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->maximal_row_index = aRows - 1;
	this->maximal_column_index = aColumns - 1;
	this->matrix = NULL;	
	this->species = SNormalMatrix;
}


/*!

	The destructor performs "cleaning up", in this case deallocation of the data array. Its decisions to deallocate
	are based on maximal_allocated_row_index, so it is safe to call it twice. However, I do not see any reason to
	call destructor explicitely; just use the standard C++ pattern
	
		delete nm;
		
	which invokes the destructor implicitely.

*/
NormalMatrix::~NormalMatrix(){
	if (max_allocated_row_index >= 0){
		long i;
		for (i=0; i<=max_allocated_row_index; i++){
			if (this->matrix[i]){
				free(this->matrix[i]);
				this->matrix[i] = NULL;
			}
		}
		max_allocated_row_index = -1;
		free(this->matrix);
		this->matrix = NULL;
	}
}

/*!
	  This method checks whether the calling instance has already been allocated; it uses maximal_allocated_row_index
	  to determine this. If it has not been allocated, but the maximal_row_index is set to 0 or more (that means:
	  if dimensions of this matrix have already been determined), it allocates the rows and columns of the matrix.
	  
	  During allocation, it zeroes out all the elements of the matrix, by using memset() function
	  This function is used to ensure proper allocation of target (and other) matrices in arithmetic functions.
	  Return codes:
	  
	  	ConstRC::Ok 		  - everything all right
		
		ConstRC::NotEnoughMemory - there was not enough memory to allocate the requested space.

*/
int  NormalMatrix::Allocate(){
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){
		long i;
		this->matrix = (integer_matrix_type **) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(integer_matrix_type*));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"NormalMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate normal matrix rows");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->matrix,0,(this->maximal_row_index+1)*sizeof(integer_matrix_type*));
		this->max_allocated_row_index = this->maximal_row_index;
		this->max_allocated_column_index = this->maximal_column_index;
		for (i=0; i<= maximal_row_index; i++){
			this->matrix[i] = (integer_matrix_type *) Utils::AllocationMachine((void **) &(this->matrix[i]),(this->max_allocated_column_index+1)*sizeof(integer_matrix_type));
			if (!this->matrix[i]){
				CLog::LogError(WHEREARG,"NormalMatrixClass","error allocating columns",(this->max_allocated_column_index+1),201,"There was not enough memory to allocate normal matrix columns");
				return ConstRC::NotEnoughMemory;
			}
			memset(this->matrix[i],0,(this->max_allocated_column_index+1)*sizeof(integer_matrix_type));
		}
	}
	return ConstRC::Ok;
}

NormalMatrix * NormalMatrix::clone() {
	NormalMatrix * new_matrix = new NormalMatrix(GetMaximalRowIndex()+1, GetMaximalColumnIndex()+1);
	return new_matrix;
};

/*!
	The getter method GetXY returns the value on aRow and aColumn indices. It only works if the matrix** array
	has already been allocated. If aRow and/or aColumn exceeds max_allocated_row_index or max_allocated_column_index,
	or if they are below zero, the return value is 
		ConstRC::GeneralError
	defined in definitions.h. Remember, this error may happen twice: either if you input excessive indices, or
	if you try to call the getter on an instantiated, but internally unallocated matrix.
	Two things are here to be discussed:
	- the philosophy of the matrix arithmetics. I have taken the stand to consider unallocated matrix
	  invalid, and return error values, if someone tries to access the matrix data by row and column indices.
	  That is because I feel that reading data and writing data are two very close operations, and one does not
	  want to write data into an unallocated matrix.
	- the philosophy of error return codes. In this situation, any integer value is a legitimate output;
	  let us say that ConstRC::GeneralError = 0xffffffff; what if GetXY returns 0xffffffff? Does that mean that GetXY
	  found an error in its parameters, or that the corresponding value in the matrix is equal to 0xffffffff?
	  Of course, one cannot tell the difference.
	  For this reason, I recommend to use the NormalMatrix class for representation of "reasonable" matrices only,
	  say mod p, where p is lesser than 2^32.

*/
integer_matrix_type NormalMatrix::GetXY(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	return (this->matrix[aRow][aColumn]);
}

/*!
	This function returns TRUE if the number on aRow and aColumn is equal to 1.
	Otherwise it returns FALSE.
*/

int NormalMatrix::IsOne(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	return (this->matrix[aRow][aColumn]==1);
}

/*!
	This function returns TRUE if the number on aRow and aColumn is equal to 0.
	Otherwise it returns FALSE.
*/

int NormalMatrix::IsZero(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	return (this->matrix[aRow][aColumn]==0);
}
/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()). Then it switches its behaviour
	  according to the action taken by Allocate().
	     - If Allocate() really allocated the matrix, it is zeroized already; no need to zeroize it again.
	     - Otherwise the matrix is filled with zeros using memset().
	     
	  This method overwrites any previous elements in matrix data array.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
*/
int  NormalMatrix::Zeroize(){
	long i = this->max_allocated_row_index;
	long j = this->max_allocated_column_index;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	// This strangely looking condition in fact checks whether Allocate() was needed or no. In case that
	// Allocate() run the allocation job, it changed values of this->max_allocated_column_index and
	// this->max_allocated_row_index. Zeroizing is a part of allocation job; we do not want to perform
	// it twice. So we run this zeroizing if and only if it has not been already run by Allocate()
	if ((this->max_allocated_row_index == i) && (this->max_allocated_column_index == j)){
		// zeroizing is needed
		for (i=0; i<= this->max_allocated_row_index; i++){
			memset(this->matrix[i],0,(this->max_allocated_column_index+1)*sizeof(integer_matrix_type));
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method performs comparison of the calling instance and of aMatrix. It does not call Allocate() to ensure 
	  allocation of anything.
	  
	  The principles for equality are the following:
	  	- if aMatrix is NULL, then the matrices are not equal
		- if the respective dimensions maximal_row_index and maximal_column_index differ, then the matrices are not equal
		- if the allocated arrays are not of the same size, then the matrices are not equal (discutable)
		- if any entry on any position differs, then the matrices are not equal.
		
	  The comparison ends as soon as any of the previous conditions is met.
	  
	  This method does not change any data of any operand.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>TRUE 		</td><td>    the matrices are equal</td></tr>
		<tr><td>FALSE		</td><td>    the matrices are not equal</td></tr>
	  </table>
*/
bool NormalMatrix::Equals(NormalMatrix* aMatrix){
	if (!aMatrix)		return FALSE; // ConstRC::NullPointerSupplied
	// performs a "true" comparison of two matrices. They must be equal up to allocation (have the same number of
	// elements allocated
	if (this->max_allocated_row_index != aMatrix->max_allocated_row_index)		return FALSE;
	if (this->max_allocated_column_index != aMatrix->max_allocated_column_index)	return FALSE;
	if (this->maximal_row_index != aMatrix->maximal_row_index)			return FALSE;
	if (this->maximal_column_index != aMatrix->maximal_column_index)		return FALSE;
	long i,j;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			if (this->matrix[i][j] != aMatrix->matrix[i][j])		return FALSE;
		}
	}
	return TRUE;
}

/*!
	  This method prints the calling instance onto the screen. It is suitable for printing matrices with entries smaller
	  than 3 decimal digits.
	  
	  This method does not change any data.

*/
void NormalMatrix::PrintToScreen(){
	long i, j;
	printf("\n");
	printf("Normal matrix: %ld x %ld\n",this->maximal_row_index+1,this->maximal_column_index+1);
	for (i=0; i<=this->max_allocated_row_index; i++){
		printf("\nRow %4ld:   ",i);
		for (j=0; j<=this->max_allocated_column_index; j++){
			printf("%3u",this->matrix[i][j]);
		}
	}
	printf("\n\n");
}

/*!	This method sets the number on aRow and aColumn to be 1.
	The method can detect row- and column- overflow and underflow; it
	returns error values
		ConstRC::ExcessiveRowIndex,
		ConstRC::ExcessiveColumnIndex,
		ConstRC::NegativeIndex,
	defined in definitions.h.
	
	If operation has been done, they return value
	
		ConstRC::Ok
		
	defined in definitions.h
*/
int  NormalMatrix::PutOne(long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	this->matrix[aRow][aColumn] = 1;
	return ConstRC::Ok;
}

int  NormalMatrix::PutZero(long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	this->matrix[aRow][aColumn] = 0;
	return ConstRC::Ok;
}

/*!	This method puts the number aNumber on aRow and aColumn.
	The method can detect row- and column- overflow and underflow; it
	returns error values
		ConstRC::ExcessiveRowIndex,
		ConstRC::ExcessiveColumnIndex,
		ConstRC::NegativeIndex,
	defined in definitions.h.
	
	If operation has been done, they return value
		ConstRC::Ok
	defined in definitions.h
*/

int  NormalMatrix::PutNumber(integer_matrix_type aNumber, long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	this->matrix[aRow][aColumn] = aNumber;
	return ConstRC::Ok;
}

int NormalMatrix::PutValue(long aRow, long aColumn,
		integer_matrix_type aValue) {
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	this->matrix[aRow][aColumn] = aValue;
	return ConstRC::Ok;
}

int NormalMatrix::GetValue(long aRow, long aColumn,
		integer_matrix_type& aValue) {
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	aValue = this->matrix[aRow][aColumn];
	return ConstRC::Ok;
}

/*!
	  This method recalculates the current data of the calling matrix modulo argument. In an instantiated & unallocated
	  matrix it does simply nothing.
	  
	  So far, this method has no return codes. If aModulus < 2, nothing is done and the method returns immediately.

*/
void NormalMatrix::Modulo(long aModulus){
	long i,j;
	if (aModulus < 2) return;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			this->matrix[i][j] %= aModulus;
		}
	}
}

/*!
	This methods performs copying of the matrix aSource into the called matrix.

	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch</td><td>if some sizes are not equal</td></tr>
	  </table>
*/

int  NormalMatrix::Copy(NormalMatrix* aSource){
        if (this->maximal_row_index != aSource->maximal_row_index)              return ConstRC::SizeMismatch;
        if (this->maximal_column_index != aSource->maximal_column_index)        return ConstRC::SizeMismatch;
        int ret;
        ret = this->Allocate();
        if (ret != ConstRC::Ok){
                return ret;
        }
        long i, j;
        for (i=0; i<= this->max_allocated_row_index; i++){
                for (j=0; j<= this->max_allocated_column_index; j++){
                        this->matrix[i][j] = aSource->matrix[i][j];
                }
        }
        return ConstRC::Ok;
}

/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  row at given index aRow with zeros (using memset).
	  
	  Checking whether 0 <= aRow <= maximal_row_index is enabled.

 	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if aRow > maximal_row_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aRow < 0</td></tr>
	  </table>
*/
int  NormalMatrix::ZeroizeRow(long aRow){
	int ret;
	if (aRow < 0)	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	memset(this->matrix[aRow],0,(this->max_allocated_column_index+1)*sizeof(integer_matrix_type));
	return ConstRC::Ok;
}

/*!
	  This method reads numbers from aRowList array[0 ... aListMaxIndex] and zeroes out rows with those indices.
	  
	  Before this, it checks whether the aRowList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). If aListMaxIndex < 0, an error is returned.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>  everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if any of the indices in aRowList > maximal_row_index</td></tr>
		<tr><td>ConstRC::NullPointerSupplied</td><td> if aRowList is NULL</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td>  if aListMaxIndex < 0 or any of the indices om aRowList is negative</td></tr>
	  </table>
*/
int  NormalMatrix::ZeroizeRow(long* aRowList, long aListMaxIndex){
	if (!aRowList)		return ConstRC::NullPointerSupplied;
	if (aListMaxIndex < 0)  return ConstRC::NegativeIndex;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long j;
	long index;
	for (j=0; j<= aListMaxIndex; j++){
		index = aRowList[j];
		if (index < 0)					     return ConstRC::NegativeIndex;
		if (index > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		memset(this->matrix[index],0,(this->max_allocated_column_index+1)*sizeof(integer_matrix_type));
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures swap of rows with indices aRow1 and aRow2. If the indices are equal, it does nothing.
	  If they are distinct, it ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are between 0 and maximal_row_index.
	  The swap is performed by three XORs.
	  
	  Return codes:
 	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if aRow1 or aRow2 > maximal_row_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aRow1 or aRow2 < 0</td></tr>
	  </table>
*/
int  NormalMatrix::SwapRows(long aRow1, long aRow2){
	if ((aRow1 < 0) || (aRow2 < 0))	return ConstRC::NegativeIndex;
	if (aRow1 != aRow2){
		if (aRow1 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		if (aRow2 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		int ret;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		for (k=0; k<=this->max_allocated_column_index; k++){
			this->matrix[aRow2][k] ^= this->matrix[aRow1][k];
			this->matrix[aRow1][k] ^= this->matrix[aRow2][k];
			this->matrix[aRow2][k] ^= this->matrix[aRow1][k];
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method adds row with index aSource to the row with index aTarget. The indices may be equal, in which
	  case it just multiplies the contents of the row by 2.
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are between 0 and maximal_row_index.
	  
	  Return codes:
  	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if aTarget or aSource > maximal_row_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aTarget or aSource < 0</td></tr>
	  </table>
*/
int  NormalMatrix::AddRows(long aTarget, long aSource){
	int ret;
	if ((aTarget < 0) || (aSource < 0))		     return ConstRC::NegativeIndex;
	if (aTarget > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aSource > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	for (k=0; k<=this->max_allocated_column_index; k++){
		this->matrix[aTarget][k] += this->matrix[aSource][k];
	}
	return ConstRC::Ok;
}

/*!
	  This method adds to the row with index aRow the row with index aRow2 of the matrix aOperand.
	  The method ensures the allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		 </td><td>  everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if aRow or aRow2 > maximal_row_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aRow or aRow2 < 0</td></tr>
	  </table>
*/
int  NormalMatrix::AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2){
        int ret;
        NormalMatrix *aO;
        if ((aRow < 0) || (aRow2 < 0))                               return ConstRC::NegativeIndex;
        if (aRow > this->maximal_row_index)                  return ConstRC::ExcessiveRowIndex;
        if (aRow2 > aOperand->GetMaximalRowIndex())                 return ConstRC::ExcessiveRowIndex;
	if (aOperand->GetSpecies() != SNormalMatrix)	     return ConstRC::GeneralError;
        aO = (NormalMatrix*) aOperand;
        ret = aO->Allocate();
        if (ret != ConstRC::Ok)  return ret;
        long k;
        for (k=0; k<=this->maximal_column_index; k++){
                this->matrix[aRow][k] ^= aO->matrix[aRow2][k];
        }
        return ConstRC::Ok;
}

/*!

	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  column at given index aColumn with zeros (using memset).
	  The method checks whether aColumn is between 0 and maximal_column_index.	  
	  
	  Return codes:
 	   <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveColumnIndex</td><td>if aColumn > maximal_column_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aColumn < 0</td></tr>
	  </table>
*/
int  NormalMatrix::ZeroizeColumn(long aColumn){
	int ret;
	if (aColumn < 0)    	     		return ConstRC::NegativeIndex;
	if (aColumn > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i;
	for (i=0; i<=this->max_allocated_row_index; i++){
		this->matrix[i][aColumn] = 0;
	}
	return ConstRC::Ok;
}

/*!
This method reads numbers from aColumnList array[0 ... aListMaxIndex] and zeroes out columns with those indices.
	  Before this, it checks whether the aColumnList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). 
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveColumnIndex</td><td>if any column index in list > maximal_column_index</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aColumnList is NULL</td></tr>
		<tr><td>ConstRC::NegativeIndex	</td><td> if any column index in list < 0 or aListMaxIndex < 0.</td></tr>
	  </table>
*/
int  NormalMatrix::ZeroizeColumn(long* aColumnList, long aListMaxIndex){
	if (!aColumnList)		return ConstRC::NullPointerSupplied;
	if (aListMaxIndex < 0)		return ConstRC::NegativeIndex;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i, j, index;
	for (j=0; j<= aListMaxIndex; j++){
		index = aColumnList[j];
		if (index < 0)					     return ConstRC::NegativeIndex;
		if (index > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		for (i=0; i<=this->max_allocated_row_index; i++){
			this->matrix[i][index] = 0;
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures swap of columns with indices aColumn1 and aColumn2. If the indices are equal, it does nothing.
	  If they are distinct, it ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  The swap is performed by three XORs.
	  
	  Return codes:
	 <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveColumnIndex</td><td>if aColumn1 or aColumn2 > maximal_column_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aColumn1 or aColumn2 < 0</td></tr>
	  </table>
*/
int  NormalMatrix::SwapColumns(long aColumn1, long aColumn2){
	if ((aColumn1 < 0) || (aColumn2 < 0))	return ConstRC::NegativeIndex;
	if (aColumn1 != aColumn2){
		if (aColumn1 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		if (aColumn2 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		int ret;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		for (k=0; k<=this->max_allocated_row_index; k++){
			this->matrix[k][aColumn1] ^= this->matrix[k][aColumn2];
			this->matrix[k][aColumn2] ^= this->matrix[k][aColumn1];
			this->matrix[k][aColumn1] ^= this->matrix[k][aColumn2];
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method adds column with index aSource to the column with index aTarget. The indices may be equal, in which
	  case it just multiplies the contents of the column by 2.
	  
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  
	  Return codes:
	    <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveColumnIndex</td><td>if aTarget or aSource > maximal_column_index</td></tr>
		<tr><td>ConstRC::NegativeIndex	</td><td> if aTarget or aSource < 0</td></tr>
	  </table>
*/
int  NormalMatrix::AddColumns(long aTarget, long aSource){
	int ret;
	if ((aTarget < 0) || (aSource < 0))		     return ConstRC::NegativeIndex;
	if (aTarget > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aSource > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	for (k=0; k<=this->max_allocated_row_index; k++){
		this->matrix[k][aTarget] += this->matrix[k][aSource]; // either 0 or power of 2
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures allocation state, dimension requirements etc., and then transposes the calling instance into
	  the matrix aTarget. The calling instance must NOT be equal to aTarget.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		 </td><td>  everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions of the calling instance and aTarget do not match (M x N vs. N x M)</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if the calling instance is equal to aTarget</td></tr>
	  </table>
*/
int  NormalMatrix::Transpose(NormalMatrix* aTarget){
	if (!aTarget)  							return ConstRC::NullPointerSupplied;
	if (this == aTarget)						return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	long i, j;
	for (i=0; i<=this->max_allocated_row_index; i++)
	{
		for (j=0; j<=this->max_allocated_column_index; j++){
			aTarget->matrix[j][i] = this->matrix[i][j];
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the transposition operation, and then performs
	  the transposition by calling 
	  	int Transpose(NormalMatrix* aTarget)
	  If result has been allocated, but Transpose did not finish well, the result is deleted again.
	  
	  Returns:
	  <table border=0>
	  	<tr><td>pointer to result of operation if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>
*/
NormalMatrix* NormalMatrix::Transpose(){
	NormalMatrix* result = new NormalMatrix(this->maximal_column_index+1, this->maximal_row_index+1);
	if (!result) return NULL;
	if (this->Transpose(result) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method ensures allocation state, dimension requirements etc., and then adds aOperand2 to the calling instance
	  and places the result into matrix aTarget. Both the calling instance and aOperand2 must NOT be equal to aTarget.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand2 or aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if aOperand2 == this or aTarget == this.</td></tr>
	  </table>
*/
int  NormalMatrix::Add(NormalMatrix* aTarget, NormalMatrix* aOperand2){
	if (!aTarget)  								return ConstRC::NullPointerSupplied;
	if ((aTarget == this) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (!aOperand2)								return ConstRC::NullPointerSupplied;
	if (aOperand2->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aOperand2->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	for (i=0; i<=aTarget->max_allocated_row_index;i++){
		for (j=0; j<= aTarget->max_allocated_column_index;j++){
			aTarget->matrix[i][j] = this->matrix[i][j] + aOperand2->matrix[i][j];
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the addition operation, and then performs
	  the addition by calling 
	  	int Add(NormalMatrix* aTarget)
	  If result has been allocated, but Add did not finish well, the result is deleted again.
	  
	  Returns:
	  <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>
*/
NormalMatrix* NormalMatrix::Add(NormalMatrix* aOperand2){
	NormalMatrix* result = new NormalMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!result) return NULL;
	if (this->Add(result,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	 <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand1, aOperand2 or aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if aOperand1 or aOperand2 == aTarget.</td></tr>
	  </table>
*/
int  NormalMatrix::MultiplyInternal(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2){
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand1) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_column_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k;
	integer_matrix_type intermediate;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	for (i=0; i<= aTarget->maximal_row_index; i++){
		for (j=0; j<= aTarget->maximal_column_index; j++){
			intermediate = 0;
			for (k=0; k<= aOperand1->maximal_column_index; k++){
				intermediate += aOperand1->matrix[i][k] * aOperand2->matrix[k][j];
			}
			aTarget->matrix[i][j] = intermediate;
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1^T x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand1, aOperand2 or aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if aOperand1 or aOperand2 == aTarget.</td></tr>
	  </table>
*/
int  NormalMatrix::MultiplyInternalTransposed(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2){
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand1) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_column_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_row_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k;
	integer_matrix_type intermediate;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	for (i=0; i<= aTarget->maximal_row_index; i++){
		for (j=0; j<= aTarget->maximal_column_index; j++){
			intermediate = 0;
			for (k=0; k<= aOperand1->maximal_row_index; k++){
				intermediate += aOperand1->matrix[k][i] * aOperand2->matrix[k][j];
			}
			aTarget->matrix[i][j] = intermediate;
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x (aOperand2)^T, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand1, aOperand2 or aTarget is NULL</td></tr>
	  </table>
*/
int  NormalMatrix::MultiplyInternalWithTransposition(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2){
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand1) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k;
	integer_matrix_type intermediate;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	for (i=0; i<= aTarget->maximal_row_index; i++){
		for (j=0; j<= aTarget->maximal_column_index; j++){
			intermediate = 0;
			for (k=0; k<= aOperand1->maximal_column_index; k++){
				intermediate += aOperand1->matrix[i][k] * aOperand2->matrix[j][k];
			}
			aTarget->matrix[i][j] = intermediate;
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternal(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.
	  
	  Returns:
	   <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>
*/
NormalMatrix* NormalMatrix::MultiplyInternal(NormalMatrix* aOperand1, NormalMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	NormalMatrix* result = new NormalMatrix(aOperand1->maximal_row_index+1, aOperand2->maximal_column_index+1);
	if (!result) return NULL;
	if (MultiplyInternal(result,aOperand1,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternalTransposed(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.
	  
	  Returns:
	  <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>

*/
NormalMatrix* NormalMatrix::MultiplyInternalTransposed(NormalMatrix* aOperand1, NormalMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	NormalMatrix* result = new NormalMatrix(aOperand1->maximal_column_index+1, aOperand2->maximal_column_index+1);
	if (!result) return NULL;
	if (MultiplyInternalTransposed(result,aOperand1,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method allocates a new matrix for placement of the result of the operation, and then performs the transposed
	  multiplication by calling 
	  
	  	int MultiplyInternalWithTransposition(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2)
		
	  If result has been allocated, but MultiplyInternalWT did not finish well, the result is deleted.
	  
	  Returns:
	 <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>
*/
NormalMatrix* NormalMatrix::MultiplyInternalWithTransposition(NormalMatrix* aOperand1, NormalMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	NormalMatrix* result = new NormalMatrix(aOperand1->maximal_row_index+1, aOperand2->maximal_row_index+1);
	if (!result) return NULL;
	if (MultiplyInternalWithTransposition(result,aOperand1,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;	
}

int NormalMatrix::Save(char *aName)
{
        FILE *fw;
        long i;
        fw=fopen(aName,"w");
        if (!fw) return ConstRC::GeneralError;
        fwrite(&(this->maximal_row_index),1,sizeof(long),fw);
        fwrite(&(this->maximal_column_index),1,sizeof(long),fw);
        for(i=0;i<=maximal_row_index;i++){
                fwrite(this->matrix[i],this->maximal_column_index+1,sizeof(matrix_type),fw);
        }
        fclose(fw);
	return ConstRC::Ok;
}



int NormalMatrix::Load(char *aName)
{
        FILE *fr;
        long i;
        if (this->matrix) return ConstRC::GeneralError;
        fr=fopen(aName,"r");
        if (!fr) return ConstRC::GeneralError;
        fread(&(this->maximal_row_index),1,sizeof(long),fr);
        fread(&(this->maximal_column_index),1,sizeof(long),fr);
        this->Allocate();
        for(i=0;i<=maximal_row_index;i++){
                fread(this->matrix[i],this->maximal_column_index+1,sizeof(matrix_type),fr);
        }
        fclose(fr);
	return ConstRC::Ok;
}

int NormalMatrix::PerformColumnMask(NormalMatrix* aTarget, NormalMatrix* aOperand, matrix_type aMask) {	

        if (!aOperand || !aTarget)                      {
                return ConstRC::NullPointerSupplied;
        }
        int ret;
        ret = aTarget->Allocate();
        if (ret != ConstRC::Ok) {
                return ret;
        }
        ret = aOperand->Allocate();
        if (ret != ConstRC::Ok) {
                return ret;
        }
        if (aTarget->maximal_row_index != aOperand->maximal_row_index)          {
                return ConstRC::SizeMismatch;
        }
        if (aTarget->maximal_column_index != aOperand->maximal_column_index){
                return ConstRC::SizeMismatch;
        }
	int bitsize = sizeof(aMask)*8;
        if (aTarget->maximal_column_index != (bitsize-1)) {
                return ConstRC::SizeMismatch;
        }
        int i, j;
	aTarget->Zeroize();
        for (i=0; i<=aTarget->max_allocated_row_index; i++){
		for (j=0; j< bitsize; j++)
		{
			if (aMask & (1 << j)) 
			{
				aTarget->matrix[i][j] = aOperand->matrix[i][j];
			}
		}
        }
        return ConstRC::Ok;
}

int  NormalMatrix::PerformColumnMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask){
	if(aTarget->GetSpecies()!=SNormalMatrix) return ConstRC::GeneralError;
	if(aOperand->GetSpecies()!=SNormalMatrix) return ConstRC::GeneralError;
	return this->PerformColumnMask((NormalMatrix*)aTarget, (NormalMatrix*) aOperand, aMask);
}

//int NormalMatrix::Serialize(const CBaseParameters &aParam) const
//{
//	return MSerializable::Serialize(aParam);
//}
//
//int NormalMatrix::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
//{
//	return AbstractMatrix::Serialize(aParam, aWriter);
//}
//
//int NormalMatrix::Deserialize(const CBaseParameters &aParam)
//{
//	return MSerializable::Deserialize(aParam);
//}
//
//int NormalMatrix::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
//{
//	return AbstractMatrix::Deserialize(aParam, aReader);
//}

int NormalMatrix::WriteData(xmlTextWriterPtr aWriter) const
{
	int rc;
	// write all matrix data
	for (long i = 0; i <= maximal_row_index; i++)
	{
		rc = CommonXMLHelper::SerializeBinaryData(aWriter, XML_SERIAL_MATRIX_ROW_NAME, this->matrix[i], this->maximal_column_index + 1, sizeof(integer_matrix_type));  
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

int NormalMatrix::ReadData(xmlTextReaderPtr aReader)
{
	int rc;
	xmlChar * name;

	// we know max_row_index and max_column_index
	Allocate();
	long i;
	// we deserialize every row and get max nonzero columns in row and nonzero indexies
    for (i = 0; i <= this->maximal_row_index; i++)
	{
		name = xmlTextReaderName(aReader);

		if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MATRIX_ROW_NAME) == 0)
		{
			break;
		}

		rc = CommonXMLHelper::DeserializeBinaryDataIn(aReader, "Deserialize matrix row...", this->matrix[i], this->maximal_column_index + 1, sizeof(integer_matrix_type), FALSE);
        if (rc != ConstRC::Ok)
		{
			return rc;
		}

		rc = CommonXMLHelper::MoveToNextNode(aReader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::EndOfFile)
			{
				return ConstRC::Ok;
			}
			return rc;
		}
    }

	return ConstRC::Ok;
}
