#ifndef __NORMAL_MATRIX_CLASS__
#define __NORMAL_MATRIX_CLASS__

#include "base_int_matrix.h"
#include "common_definitions.h"
#include <stdlib.h>

/*!
		\brief A class for representation of matrices over integers.
	
	This class implements matrices and matrix operations over the ring of integers. The representation is given by two-dimensional
	array of custom ordinal type normal_matrix_type, which is in the i386 version defined as unsigned int.
	Such representation is suitable only for smaller matrices and is very unsuitable for sparse matrices mod 2; each
	entry, even zero, occupies several (in this case 4) bytes of memory. One can easily see that representation
	of a 10 000 x 10 000 matrix would need 400 MB of memory.
	There are two main reasons for existence of this class:
		- in future, there may be need to process other types of matrices than only GF(2)-based
		- it is programatically easy to write methods implementing matrix arithmetics in this representation.
		  That is why this class can serve as a means of verifying correctness of matrix operations in other
		  matrix representations, and it really does in the TTR part.

	The double array 
	
		normal_matrix_type** matrix
		
	is the one that contains the matrix data. It is initialized by the constructor to a NULL pointer. The variables
	max_allocated_row_index and max_allocated_column_index reflect the dimensions of the allocated double array;
	for example, if max_allocated_row_index = 3 and max_allocated_column_index = 5, the matrix pointer has been
	initialized to point to an array of size 4x6.
	
	A word of explanation is here at place. The inherited variables
	
		long maximal_row_index;
		
		long maximal_column_index;
		
	which are defined in the superclass AbstractMatrix, express the "intended" dimensions of the matrix, e.g. how
	many rows and columns it "should have". On the other hand, 
	
		long max_allocated_row_index;
		
		long max_allocated_column_index;
		
	are connected with the real physical size of 
	
		normal_matrix_type** matrix;
		
	and tell us about the size of memory already allocated for the representation of this matrix.
	Generally, only two non-pathologic situations should happen:
		- maximal_row_index and maximal_column_index have reasonable values, like 0 and 5, while
		  max_allocated_row_index = max_allocated_column_index = -1.
		  This means precisely that this instance of normal_matrix will one day contain a matrix 1x6, but
		  the initialization of the representation array has not yet been done. This is a correct situation,
		  because we do not want to allocate large bunches of memory until the matrix is really used 
		  to contain data or perform operations. For example:
		  
			NormalMatrix* nm = new NormalMatrix(1,6);
			
			... // memory-intensive operation with other matrices
			
			nm->Allocate();
			
		  is a construction which enables us to declare and instantiate the matrix nm at the beginning
		  of a code block, while leaving the allocation of the array to the time of need.
		- maximal_row_index = maximal_allocated_row_index
		  and
		  maximal_column_index = maximal_allocated_column_index
		  This means that the matrix is ready for any operations, including reading of data.
		  
	There are no getters and setters for the *_allocated_* member variables. That is because these member
	variables are business of the internal logic and the user should not be interested in reading them, let
	alone changing them manually.
		  

*/
class NormalMatrix : public BaseIntMatrix {
public:
	NormalMatrix();
	NormalMatrix(long aRows, long aColumns);
	~NormalMatrix();
	int  Allocate();

	virtual NormalMatrix * clone();

	int Copy(NormalMatrix *aMatrix);
	int  Zeroize();
	bool Equals(NormalMatrix* aMatrix);
	void PrintToScreen();
	int  PutOne(long aRow, long aColumn);
	int  PutZero(long aRow, long aColumn);
	int  IsOne(long aRow, long aColumn);
	int  IsZero(long aRow, long aColumn);

	int  PutNumber(integer_matrix_type aNumber, long aRow, long aColumn);
	void Modulo(long aModulus);
	integer_matrix_type GetXY(long aRow, long aColumn);

	virtual int PutValue(long aRow, long aColumn, integer_matrix_type aValue);
	virtual int GetValue(long aRow, long aColumn, integer_matrix_type & aValue);

	// Swaps etc.
	int  ZeroizeRow(long aRow);
	int  ZeroizeRow(long* aRowList, long aListMaxIndex);
	int  SwapRows(long aRow1, long aRow2);
	int  AddRows(long aTarget, long aSource);
	int  AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2);
	int  ZeroizeColumn(long aColumn);
	int  ZeroizeColumn(long* aColumnList, long aListMaxIndex);
	int  SwapColumns(long aColumn1, long aColumn2);
	int  AddColumns(long aTarget, long aSource);
	// Mathematical operations
	int  	   Transpose(NormalMatrix* aTarget);
	NormalMatrix* Transpose();
	int  	   Add(NormalMatrix* aTarget, NormalMatrix* aOperand2);
	NormalMatrix* Add(NormalMatrix* aOperand2);
	int	MultiplyInternal(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	int	MultiplyInternalTransposed(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	int	MultiplyInternalWithTransposition(NormalMatrix* aTarget, NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	NormalMatrix*   MultiplyInternalWithTransposition(NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	NormalMatrix*	MultiplyInternal(NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	NormalMatrix*	MultiplyInternalTransposed(NormalMatrix* aOperand1, NormalMatrix* aOperand2);
	int      PerformColumnMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask);
	int      PerformColumnMask(NormalMatrix* aTarget, NormalMatrix* aOperand, matrix_type aMask);
	int Save(char *aName);
	int Load(char *aName);

	// De/Serialization methods
	//int Serialize(const CBaseParameters& aParam) const;
	//int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	//int Deserialize(const CBaseParameters& aParam);
	//int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

protected:
	int WriteData(xmlTextWriterPtr aWriter) const;
	int ReadData(xmlTextReaderPtr aReader);

protected:
	long max_allocated_row_index;
	long max_allocated_column_index;
	integer_matrix_type** matrix; ///< This double array contains the matrix data. It is initialized by the constructor to a NULL pointer.
	static long bits_in_n_m_t; ///< Common to all instances of normal_matrix, reflects the number of bits in normal_matrix_type. This value	is not very interesting for the class itself. It has been made a part of the class for the sake of symmetry with BitMatrix class.

};

#endif

