#include "pollard_p_m_1.h"
#include "const_return_codes.h"
#include "common_utils.h"
#include "integer_class.h"
#include <string.h>

/* \class

 * This implementation expects that the array "small_primes", defined as extern in gmp_tests.h,
 * has already been initialized by the calling process. This is not a problem in case of the
 * quadratic sieve.
 * Further, this implementation expects that small_primes are considerably longer than requested
 * B_1 or B_2, and so it does not check for buffer overflow.
 */

 /*!
 	The default constructor sets 'mode' to optimization level 1 (using extensions of the factor base).
 */
PollardPMinus1::PollardPMinus1(long aUpperFBBound)  : ASupportingFactorAlg()
{
	Set_AlgorithmType(EFAPMinus1);

	inited = 0;
	if ((inited & AUX_1_INITED) == 0){
		mpz_init(aux_1);
		inited |= AUX_1_INITED;
	}
	if ((inited & AUX_2_INITED) == 0){
		mpz_init(aux_2);
		inited |= AUX_2_INITED;
	}
	if ((inited & POLLARD_B_INITED) == 0){
		mpz_init(b);
		inited |= POLLARD_B_INITED;
	}
	if ((inited & FACT_1_INITED) == 0){
		mpz_init(fact_1);
		inited |= FACT_1_INITED;
	}
	if ((inited & FACT_2_INITED) == 0){
		mpz_init(fact_2);
		inited |= FACT_2_INITED;
	}
	this->mode = OPTIMIZATION_LEVEL_1;
	this->max_iters = DEFAULT_MAX_ITERS;
	this->b_1 = DEFAULT_B_1;
	this->b_2 = DEFAULT_B_2;
	this->continued = 0;
	this->factorized = 0;
	this->not_factorized = 0;
	this->exponents = NULL;
	this->exponents = (int*) Utils::AllocationMachine((void**)&this->exponents, this->b_1 *sizeof(int));
	this->small_primes = NULL;
	unsigned long count;
	CIntegerClass::GetPrimes(1, aUpperFBBound, count, this->small_primes);
	this->max_exponent_index = this->b_1 - 1;
	memset(this->exponents,0,(this->b_1)*sizeof(int));
	int i;  long prime;
	int logb_1 = BinLog(this->b_1);	
	for (i=0; i<= this->max_exponent_index; i++) 
	{
		prime = small_primes[i];
		if (prime > this->b_1) 
		{
			break;
		} 
		else 
		{
			// We will find the maximal power of this prime which is lower than bound B_1. This can be done by (log b_1)/(log prime).
			exponents[i] = logb_1 / BinLog(prime);
			//printf(" %d ^ %d, b_1 = %d.\n",prime,exponents[i],this->b_1);
		}
	}
	max_usable_index = i-1;
}

PollardPMinus1::~PollardPMinus1(){
	if ((inited & AUX_1_INITED) != 0){
		mpz_clear(aux_1);
		inited &= ~AUX_1_INITED;
	}
	if ((inited & AUX_2_INITED) != 0){
		mpz_clear(aux_2);
		inited &= ~AUX_2_INITED;
	}	
	if ((inited & POLLARD_B_INITED) != 0){
		mpz_clear(b);
		inited &= ~POLLARD_B_INITED;
	}	
	if ((inited & FACT_1_INITED) != 0){
		mpz_clear(fact_1);
		inited &= ~FACT_1_INITED;
	}
	if ((inited & FACT_2_INITED) != 0){
		mpz_clear(fact_2);
		inited &= ~FACT_2_INITED;
	}
	free(this->exponents);
	this->max_exponent_index = -1;	
}

/*!
 An overloaded method from FactoringAlgorithm, providing interface for factorization.
*/
int PollardPMinus1::Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) 
{
    int index = 0;
    int a = 2, i = 0, j = 0, prime;
    int diff;
    while (index < max_iters) {
	a++;
	// We will count a^R - 1 mod aModulus
	mpz_set_ui(aux_1,a);
	for (i=0; i<= max_usable_index; i++) {	
		mpz_ui_pow_ui(aux_2,small_primes[i],exponents[i]);
		//gmp_printf("Exponent %Zd = %d^%d\n",aux_2,small_primes[i],exponents[i]);
		//gmp_printf(" Soucasny stav (a=%d): %Zd.\n",a,aux_1);
		mpz_powm(aux_1,aux_1,aux_2,aModulus);
	}
	mpz_sub_ui(aux_1,aux_1,1);
	mpz_gcd(aux_2,aux_1,aModulus);
	if (mpz_cmp(aux_2,aModulus) && mpz_cmp_ui(aux_2,1)) 
	{
		this->factorized++;
		mpz_set(aFactor1,aux_2);
		mpz_fdiv_q(aFactor2,aModulus,aFactor1);

		return ConstRC::Factorized;
	} 
	else 
	{
		// At level 0, we do not try to reuse b with higher B_2 bound
		if (mode == OPTIMIZATION_LEVEL_0) {
			continue;
		}
		// At level 1, we try to continue. we have b-1 in aux_1. 
		if (mode == OPTIMIZATION_LEVEL_1) {
			int first_continuation = 1;
			mpz_add_ui(b,aux_1,1);
			// for all primes from B_1 to B_2, we count b^s - 1 ...
			j = max_usable_index+1;
			while ((prime = small_primes[j]) < this->b_2) {
				if (first_continuation) {
					mpz_powm_ui(aux_1,b,prime,aModulus);
					first_continuation = 0;
				} else {
					// We find the difference between last two primes
					// and calculate b^{s_j} b^{s_{j+1} - s_j}
					diff = small_primes[j] - small_primes[j-1];
					mpz_powm_ui(aux_2,b,diff,aModulus);
					mpz_mul(aux_1,aux_1,aux_2);
					mpz_mod(aux_1,aux_1,aModulus);
				}		
				// now, we will test the GCD again
				mpz_sub_ui(aux_2,aux_1,1);
				mpz_gcd(aux_2,aux_2,aModulus);
				if (mpz_cmp(aux_2,aModulus) && mpz_cmp_ui(aux_2,1)) {
					this->factorized++;
					this->continued++;
					mpz_set(aFactor1,aux_2);
					mpz_fdiv_q(aFactor2,aModulus,aFactor1);		
					return ConstRC::Factorized;
				}
				j++;
			}
		}
	}
	index++;
    }
    this->not_factorized++;
	return ConstRC::NotFactorized;
}

