#ifndef __POLLARD_P_M_1_H_
#define __POLLARD_P_M_1_H_

#include "abstract_supporting_factor_alg.h"

#define DEFAULT_B_1             5000
#define DEFAULT_B_2             20000
#define AUX_1_INITED            0x1
#define AUX_2_INITED            0x2
#define FACT_1_INITED            0x1000000
#define FACT_2_INITED            0x2000000
#define POLLARD_B_INITED         0x4000000
#define DEFAULT_MAX_ITERS        16

/*!
	\brief Class implementing Pollard p-1 factoring algorithm.
	
	This class implements Pollard p-1 factoring algorithm in two variants: the first one,
	which tries random choices of 'a' (level 0), and the other one, which, in addition, tries to 
	expand the set of used primes (level 1). Switching between these two variants is performed by
	changing the implicit value of the internal variable "mode".
	
	The implementation expects that the array "small_primes", defined as extern in gmp_tests.h,
 	has already been initialized by the calling process. This is not a problem in case of the
 	quadratic sieve.
 	Further, this implementation expects that small_primes are considerably longer than requested
 	B_1 or B_2, and so it does not check for buffer overflow.
*/
class PollardPMinus1 : public ASupportingFactorAlg 
{

public:
	PollardPMinus1(long aUpperFBBound);
	~PollardPMinus1();

	int Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2);

private:
	mpz_t   aux_1; ///< These are auxiliary variables, used throughout the iteration process.
	mpz_t   aux_2; ///< These are auxiliary variables, used throughout the iteration process.
	mpz_t   b; ///< These are auxiliary variables, used throughout the iteration process.
	mpz_t   fact_1; ///< These are auxiliary variables, used throughout the iteration process.
	mpz_t   fact_2; ///< These are auxiliary variables, used throughout the iteration process.
	int 	max_iters; ///< Number of maximal "tries" (choices of 'a').
	int     mode;    ///< Level of optimization.
	int 	inited; ///< An auxiliary variable showing allocation status of mpz_t's.
	int		b_1; ///< The first bound on prime numbers, used in level 0 and first stage of level 1 optimization.
	int 	b_2; ///< The second bound on prime numbers, used in second stage of level 1 optimization.
	int		factorized; ///< Auxiliary variable for measurement of factorization efficiency.
	int		continued;  ///< Auxiliary variable for measurement of factorization efficiency.
	int		not_factorized; ///< Auxiliary variable for measurement of factorization efficiency.
	long*	small_primes;///<The array of small primes - max is upper FB bound
	int*    exponents; ///< The array of suitable exponents for factor base primes.
	int		max_exponent_index; ///< Length of the array of suitable exponents for factor base primes.
	int		max_usable_index; ///< The maximal index of a prime, which is used during the algorithm.
};

#endif // __POLLARD_P_M_1_H_

