#include "pollard_rho.h"

#include "const_return_codes.h"
#include "common_utils.h"
#include <stdlib.h>

pollard_entry* PollardRho::pollard               = NULL;
int            PollardRho::pollard_maximal_index = -1;
int            PollardRho::inited                = 0;

/*!
 	The default constructor sets 'mode' to optimization level 1 (using Brent's improvement).
*/
PollardRho::PollardRho() :ASupportingFactorAlg ()
{
	Set_AlgorithmType(EFAPollardRho);

	inited = 0;
	if (pollard == NULL) {
		pollard = (pollard_entry*) Utils::AllocationMachine((void**)&pollard, DEFAULT_POLLARD_ENTRIES * sizeof(pollard_entry));
		pollard_maximal_index = DEFAULT_POLLARD_ENTRIES-1;
		int i = 0;
		for (i=0; i<=pollard_maximal_index; i++) {
			mpz_init(pollard[i].value);
			pollard[i].inited = 1;
		}
	}
	if ((inited & AUX_1_INITED) == 0){
		mpz_init(aux_1);
		inited |= AUX_1_INITED;
	}
	if ((inited & AUX_2_INITED) == 0){
		mpz_init(aux_2);
		inited |= AUX_2_INITED;
	}
	if ((inited & LAST_INITED) == 0) {
		mpz_init(last);
		inited |= LAST_INITED;
	}
	if ((inited & FACT_1_INITED) == 0){
		mpz_init(fact_1);
		inited |= FACT_1_INITED;
	}
	if ((inited & FACT_2_INITED) == 0){
		mpz_init(fact_2);
		inited |= FACT_2_INITED;
	}

	this->mode = OPTIMIZATION_LEVEL_1;
	if (this->mode == OPTIMIZATION_LEVEL_2) {
		InitLevel2Structures();		
	}
}

PollardRho::~PollardRho() {
	if (pollard != NULL) {
		int i;
		for (i=0; i<=pollard_maximal_index; i++) {
			if (pollard[i].inited) {
				mpz_clear(pollard[i].value);
				pollard[i].inited = 0;
			}
		}
		free(pollard);
		pollard = NULL;
	}
	if ((inited & AUX_1_INITED) != 0){
		mpz_clear(aux_1);
		inited &= ~AUX_1_INITED;
	}
	if ((inited & AUX_2_INITED) != 0){
		mpz_clear(aux_2);
		inited &= ~AUX_2_INITED;
	}	
	if ((inited & LAST_INITED) != 0){
		mpz_clear(last);
		inited &= ~LAST_INITED;
	}		
	if ((inited & FACT_1_INITED) != 0){
		mpz_clear(fact_1);
		inited &= ~FACT_1_INITED;
	}
	if ((inited & FACT_2_INITED) != 0){
		mpz_clear(fact_2);
		inited &= ~FACT_2_INITED;
	}
	if (this->mode == OPTIMIZATION_LEVEL_2) {
		FreeLevel2Structures();
	}
}

/*!
	A method of the (unfinished) implementation of the Montgomery improvement.
*/
void PollardRho::CountQuad(mpz_t aArg, mpz_t aNext, mpz_t aModulus) {
	mpz_add(aux_2,aArg,rand_value.alpha_1);
	mpz_add(rand_value.val,aNext,rand_value.alpha_4);
	mpz_mul(rand_value.val,rand_value.val,aux_2);	
	mpz_add(rand_value.val,rand_value.val,rand_value.alpha_3);	
	mpz_mod(rand_value.val,rand_value.val,aModulus);
}

/*!
	This method calculates (X-Y) and finds the GCD of the result with aModulus. If it is successful
	in finding some proper nontrivial divisor of aModulus, it sets aFactor1 and aFactor2 to the values
	of the divisors.
*/
int PollardRho::Compare(mpz_t aX, mpz_t aY, mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) {
	mpz_sub(aux_2,aX,aY);
	mpz_gcd(aux_2,aux_2,aModulus);
	if (mpz_cmp(aux_2,aModulus) && mpz_cmp_ui(aux_2,1)) {
		mpz_set(aFactor1,aux_2);
		mpz_fdiv_q(aFactor2,aModulus,aFactor1);
		return ConstRC::Factorized;
	}
	return ConstRC::NotFactorized;
}

/*!
 An overloaded method from FactoringAlgorithm, providing interface for factorization.
*/
int PollardRho::Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) {
    int    index = 0, max = 100000;
    int    rep_unit = 0;
    int    current_rep_unit_index = -1;
    int    next_rep_unit_index = 0;
    mpz_set_ui(aux_1,4);        
    int is_level_2 = (mode == OPTIMIZATION_LEVEL_2);
    while (index < max) {    	
	// Next iteration of f(x)	
	mpz_powm_ui(aux_1,aux_1,2,aModulus);
	mpz_add_ui(aux_1,aux_1,5);
	mpz_mod(aux_1,aux_1,aModulus); 
	//gmp_printf("Index = %d, x_%d = %Zd. Rep_unit = %d\n",index,index,aux_1,rep_unit);
	// Counting f(x) - f(j_0). Has no sense for iteration #0
	if (index != 0) {
		if (mode == OPTIMIZATION_LEVEL_0) {
			if (Compare(pollard[current_rep_unit_index].value,aux_1,aModulus,aFactor1,aFactor2) == ConstRC::Factorized) {
				return ConstRC::Factorized;
			}			
		} else if (mode == OPTIMIZATION_LEVEL_1) {
			if ((index >= start_from) && (rep_unit != 0)) {
				if (Compare(pollard[current_rep_unit_index].value,aux_1,aModulus,aFactor1,aFactor2) == ConstRC::Factorized) {
					return ConstRC::Factorized;
				}
			}
		} else if (is_level_2) {
			printf("  Level 2: Index: %d, start_from: %d, (index-start_from) mod 3 = %d.\n",index,start_from,(index - start_from)%3);
			if ((index >= start_from) && (!((index - start_from)%3))) {
				printf("  Level 2: budeme srovnavat.\n");
				// srovnani s hodnotou kvadratickeho polynomu
				CountQuad(last,aux_1,aModulus);
				if (Compare(rand_value.val,aux_1,aModulus,aFactor1,aFactor2) == ConstRC::Factorized) {
					return ConstRC::Factorized;
				}
			}			
		}
	}	
	if (index == rep_unit) {
		//gmp_printf("Index = %d, rep_unit = %d, rep_unit_index = %d. Setting as %Zd.\n",index,rep_unit,next_rep_unit_index,aux_1);
		mpz_set(pollard[next_rep_unit_index].value,aux_1);				
		if (!is_level_2) start_from = ((rep_unit+1) >> 1)* 3;
		if (is_level_2) {			
			mpz_set(rand_value.new_t_1,aux_1);					
			rand_value.new_t_1_index = index;				
		}
		rep_unit <<= 1;
		rep_unit += 1;
		next_rep_unit_index++;
		current_rep_unit_index++;
		
	} 
	if (is_level_2) {
		PrintLevel2Structures();
		if (index == (rep_unit/2+1)) {
			mpz_set(rand_value.new_t_2,aux_1);	
			rand_value.new_t_2_index = index;				
		}
		if (index == (rep_unit/2+2)) {
			mpz_set(rand_value.new_t_3,aux_1);	
			rand_value.new_t_3_index = index;
			printf("Level 2 shift new -> current\n");			
			// shift
			mpz_set(rand_value.t_1,rand_value.new_t_1);
			mpz_set(rand_value.t_2,rand_value.new_t_2);
			mpz_set(rand_value.t_3,rand_value.new_t_3);
			rand_value.t_1_index = rand_value.new_t_1_index;
			rand_value.t_2_index = rand_value.new_t_2_index;
			rand_value.t_3_index = rand_value.new_t_3_index;
			start_from  = ((rand_value.t_1_index+1) >> 1)* 3 + 2;
			printf("   New start from: %d.\n",start_from);
			// We will determine the alphas needed to get value of g(x)
			mpz_add(rand_value.alpha_1,rand_value.t_2,rand_value.t_3);
			mpz_add(rand_value.alpha_2,rand_value.alpha_1,rand_value.t_1);
			mpz_mul(aux_1,rand_value.t_2,rand_value.t_3);
			mpz_mul(rand_value.alpha_3,rand_value.alpha_1,rand_value.t_1);
			mpz_add(rand_value.alpha_3,rand_value.alpha_3,aux_1);
			mpz_add(rand_value.alpha_4,rand_value.alpha_4,rand_value.t_2);
			mpz_mul(rand_value.alpha_4,rand_value.alpha_4,rand_value.alpha_1);
		}
	}
	index++;
	mpz_set(last,aux_1);
    }
	return ConstRC::NotFactorized;
}

/*!
	A method of the (unfinished) implementation of the Montgomery improvement.
*/
void PollardRho::InitLevel2Structures() {
	mpz_init(rand_value.t_1);
	mpz_init(rand_value.t_2);	
	mpz_init(rand_value.t_3);
	mpz_init(rand_value.alpha_1);
	mpz_init(rand_value.alpha_2);
	mpz_init(rand_value.alpha_3);
	mpz_init(rand_value.alpha_4);
	mpz_init(rand_value.val);
	mpz_init(rand_value.new_t_1);
	mpz_init(rand_value.new_t_2);	
	mpz_init(rand_value.new_t_3);
	rand_value.inited = (T_1_INITED | T_2_INITED | T_3_INITED | A_1_INITED | A_2_INITED | A_3_INITED | A_4_INITED | VAL_INITED | NEW_T_1_INITED | NEW_T_2_INITED | NEW_T_3_INITED);
	start_from = 8;
}

/*!
	A method of the (unfinished) implementation of the Montgomery improvement.
*/
void PollardRho::FreeLevel2Structures() {
	if (rand_value.inited & T_1_INITED) {
		mpz_clear(rand_value.t_1);
		rand_value.inited &= (~T_1_INITED);
	}
	if (rand_value.inited & T_2_INITED) {
		mpz_clear(rand_value.t_2);
		rand_value.inited &= (~T_2_INITED);
	}
	if (rand_value.inited & T_3_INITED) {
		mpz_clear(rand_value.t_3);
		rand_value.inited &= (~T_3_INITED);
	}
	if (rand_value.inited & NEW_T_1_INITED) {
		mpz_clear(rand_value.new_t_1);
		rand_value.inited &= (~NEW_T_1_INITED);
	}
	if (rand_value.inited & NEW_T_2_INITED) {
		mpz_clear(rand_value.new_t_2);
		rand_value.inited &= (~NEW_T_2_INITED);
	}
	if (rand_value.inited & NEW_T_3_INITED) {
		mpz_clear(rand_value.new_t_3);
		rand_value.inited &= (~NEW_T_3_INITED);
	}
	if (rand_value.inited & A_1_INITED) {
		mpz_clear(rand_value.alpha_1);
		rand_value.inited &= (~A_1_INITED);
	}
	if (rand_value.inited & A_2_INITED) {
		mpz_clear(rand_value.alpha_2);
		rand_value.inited &= (~A_2_INITED);
	}
	if (rand_value.inited & A_3_INITED) {
		mpz_clear(rand_value.alpha_3);
		rand_value.inited &= (~A_3_INITED);
	}
	if (rand_value.inited & A_4_INITED) {
		mpz_clear(rand_value.alpha_4);
		rand_value.inited &= (~A_4_INITED);
	}
	if (rand_value.inited & VAL_INITED) {
		mpz_clear(rand_value.val);
		rand_value.inited &= (~VAL_INITED);
	}
}

/*!
	A method of the (unfinished) implementation of the Montgomery improvement.
*/
void PollardRho::PrintLevel2Structures() {
	gmp_printf("t1 (%4d): %Zd.\n",rand_value.t_1_index,rand_value.t_1);
	gmp_printf("t2 (%4d): %Zd.\n",rand_value.t_2_index,rand_value.t_2);
	gmp_printf("t3 (%4d): %Zd.\n",rand_value.t_3_index,rand_value.t_3);
	gmp_printf("new t1 (%4d): %Zd.\n",rand_value.new_t_1_index,rand_value.new_t_1);
	gmp_printf("new t2 (%4d): %Zd.\n",rand_value.new_t_2_index,rand_value.new_t_2);
	gmp_printf("new t3 (%4d): %Zd.\n",rand_value.new_t_3_index,rand_value.new_t_3);
	gmp_printf("a1: %Zd.\n",rand_value.alpha_1);
	gmp_printf("a2: %Zd.\n",rand_value.alpha_2);
	gmp_printf("a3: %Zd.\n",rand_value.alpha_3);
	gmp_printf("a4: %Zd.\n",rand_value.alpha_4);
	gmp_printf("val: %Zd.\n",rand_value.val);
}



