#ifndef __POLLARD_RHO_H_
#define __POLLARD_RHO_H_

#include "abstract_supporting_factor_alg.h"

#define DEFAULT_POLLARD_ENTRIES 100
#define AUX_1_INITED            0x1
#define AUX_2_INITED            0x2
#define LAST_INITED             0x4

#define T_1_INITED              0x10
#define T_2_INITED              0x20
#define T_3_INITED              0x40
#define A_1_INITED              0x100
#define A_2_INITED              0x200
#define A_3_INITED              0x400
#define A_4_INITED              0x800
#define VAL_INITED              0x1000
#define NEW_T_1_INITED              0x10000
#define NEW_T_2_INITED              0x20000
#define NEW_T_3_INITED              0x40000

#define FACT_1_INITED            0x1000000
#define FACT_2_INITED            0x2000000

/*!
 * \brief A basic type for representation of the intermediate values of iterated f(x).
*/
typedef struct {
	mpz_t value;
	int   inited;
} pollard_entry;

/*!
 * \brief This typedef should serve in calculation of the (unfinished) Montgomery improvement of Pollard-rho algorithm

*/
typedef struct {
	mpz_t t_1;
	mpz_t t_2;
	mpz_t t_3;
	int   t_1_index;
	int   t_2_index;
	int   t_3_index;
	int   new_t_1_index;
	int   new_t_2_index;
	int   new_t_3_index;
	mpz_t new_t_1;
	mpz_t new_t_2;
	mpz_t new_t_3;
	mpz_t alpha_1;
	mpz_t alpha_2;
	mpz_t alpha_3;
	mpz_t alpha_4;
	mpz_t val;
	int   inited;
	int   relative_rep_unit;
	int   relative_rep_unit_index;
} randomizing_poly_value;


/*!
 * \brief This class implements the Pollard Rho (a.k.a. Monte Carlo) factoring algorithm.
	  
	This class implements the Pollard Rho (a.k.a. Monte Carlo) factoring algorithm. For Double LPV 
	purposes, this seems to be the most effective factoring algorithm.
	The implementation uses Brent's improvement to the algorithm, resulting in faster (cca 40%) run
	than original Pollard Rho. Another improvement by Montgomery, which should make another 10% per cent improvement, has not been implemented completely, since its description is insufficient.
	The original Pollard Rho method is called "level 0", the Brent's improvement (default) is called 
	"level 1" and the unfinished Montgomery improvement "level 2". Current level of the algorithm
	is set via a member variable.

*/
class PollardRho : public ASupportingFactorAlg 
{
public:
	PollardRho();
	~PollardRho();
	int Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2);

private:
	void InitLevel2Structures();
	void FreeLevel2Structures();
	void PrintLevel2Structures();
	void CountQuad(mpz_t aArg, mpz_t aNext, mpz_t aModulus);
	int Compare(mpz_t aX, mpz_t aY, mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2);


	static 	pollard_entry*  pollard; ///< 	This array is used for storage of necessary iterated f(x) values.

	static  int  		pollard_maximal_index;	///< This variable contains length-related information for the main array.
	mpz_t                   aux_1;	///< An auxiliary variable used throughout the factoring process.
	mpz_t                   aux_2;  ///< An auxiliary variable used throughout the factoring process.
	mpz_t                   fact_1; ///< An auxiliary variable used throughout the factoring process.
	mpz_t                   fact_2; ///< An auxiliary variable used throughout the factoring process.

	mpz_t                   last;   ///< An auxiliary variable for level 2 optimization (unfinished).
	static  int             inited; ///< An auxiliary variable showing allocation status of mpz_t's.
	int                     mode; ///< Level of optimization.
	int                     start_from;///< Optimization level 1 auxiliary variable -n from which iteration to start checks again (see Brent).
	randomizing_poly_value  rand_value; ///< An auxiliary variable for level 2 optimization (unfinished).
};

#endif // __POLLARD_RHO_H_

