#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "polynom_class.h"
#include "base_parameters.h"
#include <sstream>
#include <fstream>
#include "log.h"
#include "integer_class.h"
// #include "nfs_definitions.h"

using namespace std;

int Polynomial::count = 0;
mpz_t Polynomial::Z_E_R_O;
bool Polynomial::Z_E_R_O_allocated = FALSE;
Polynomial *Polynomial::AuxPoly[aux_pols];
bool Polynomial::AuxPolyOccupied[aux_pols];

/*!
	Standard constructor. The polynomial is set to be 0. (More precisely, its degree is set to be -1).
*/

Polynomial::Polynomial() : MSerializable(FILE_NAME_POLYNOMIAL)
{
	this->allocated=-1;
	this->degree=-1;
	initAuxPolynomials();
}

/*!
	Constructor with predefined expected degree. The polynomial is allocated to aDegree
	and set to be 0 by setting its degree to be -1.
*/

Polynomial::Polynomial(int aDegree) : MSerializable(FILE_NAME_POLYNOMIAL)
{
	this->allocated=-1;
	this->Allocate(aDegree);
	this->degree=-1;
	initAuxPolynomials();
}

Polynomial::Polynomial(const Polynomial& aOperand)
{
	this->allocated=-1;
	this->Allocate(aOperand.degree);
	this->degree = aOperand.degree;
	for (int i = 0; i <= this->degree; i++)
	{
		mpz_set(this->coeff[i], aOperand.coeff[i]);
	}
	initAuxPolynomials();
}

void Polynomial::initAuxPolynomials() {
	count++;
	this->AuxPolynomialNumber=-1;
	if (!Z_E_R_O_allocated){
		mpz_init_set_ui(Z_E_R_O,0);
		Z_E_R_O_allocated=1;
		for(int i = 0; i<aux_pols; i++){
		AuxPoly[i] = new Polynomial();
		AuxPolyOccupied[i] = FALSE;
		AuxPoly[i]->AuxPolynomialNumber=i;
		}
	}
};

/*!
	Standard destructor.
*/


Polynomial::~Polynomial(){
	int i;
	for(i=0;i<=allocated;i++) mpz_clear(this->coeff[i]);
	if (allocated>=0) free(coeff);
	count --;
	if (0 == count) {
		mpz_clear(Z_E_R_O);
		Z_E_R_O_allocated=0;
		for(int i=0;i<aux_pols;i++){
			delete AuxPoly[i];
		}
};
}

/*!
	Allocates the coefficients so that the polynomial can be of degree up to aSize.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory</td></tr>
	</table>
*/

int Polynomial::Allocate(int aSize){
	if (allocated>=aSize) return ConstRC::Ok;
	int i=0;
	mpz_t *aux = (mpz_t*) malloc((aSize+1)*sizeof(mpz_t));
	if (!aux) return ConstRC::NotEnoughMemory;
	if (allocated >= 0){
		for(i=0; i<= allocated;i++){
			mpz_init_set(aux[i],this->coeff[i]);
			mpz_clear(this->coeff[i]);
		}
		free(coeff);
	}
	this->coeff=aux;
	this->allocated = aSize;
	for(;i<=aSize;i++) mpz_init(this->coeff[i]);
	return ConstRC::Ok;
}

/*!
	This method assigns a preconstructed auxiliary polynomial for a temporary use.
	If no preconstructed polynomial is at hand, a new one is constructed.
	
	Returns a polynomial.
*/

Polynomial *Polynomial::AssignAuxPoly(){
	int i;
	for(i=0;i<aux_pols;i++) if(!AuxPolyOccupied[i]) break;
	if (i<aux_pols){
	AuxPolyOccupied[i]=TRUE;
	AuxPoly[i]->degree=-1;
	return AuxPoly[i];
	}
	Polynomial *p;
	p = new Polynomial();
	return p;
}

/*!
	Finishes the usage of an auxiliary polynomial. If the polynomial
	is not a preconstructed one, it is deleted.

	Return code: ConstRC::Ok
*/
int Polynomial::Dispose(Polynomial *aPolynomial)
{
	if (aPolynomial->AuxPolynomialNumber>=0) 
	{
		AuxPolyOccupied[aPolynomial->AuxPolynomialNumber]=FALSE;
	}
	else 
	{	
		delete aPolynomial;
	}
	return ConstRC::Ok;
}

/*!
	The degree variable need not be the real degree of the polynomial, i.e. the
	degree-th coefficient can be 0. This method corrects this situation diminishing
	the variable degree so that it shows the real degree of the polynomial.

	Return code: ConstRC::Ok
*/
int Polynomial::Reduce(){
	if (allocated<0) return ConstRC::Ok;
	while ((degree>-1)&&(!mpz_cmp_ui(coeff[degree],0))) degree--;
	return ConstRC::Ok;
}

/*!
	Sets all the coefficients from aFrom to degree to 0.

	Return code: ConstRC::Ok
*/
int Polynomial::Zeroize(int aFrom){
	if (aFrom>degree) return ConstRC::Ok;
	int i;
	for(i=aFrom;i<=degree;i++) mpz_set_ui(coeff[i],0);
	return ConstRC::Ok;
}

/*!
	Returns a newly created copy of the current polynomial.
*/

Polynomial* Polynomial::Copy() const
{
	Polynomial *aux;
	int i;
	aux=new Polynomial(this->degree);
	for(i=0;i<=degree;i++)
		mpz_set(aux->coeff[i],this->coeff[i]);
	aux->degree=this->degree;
	return aux;
}

/*!
	Copies the polynomial into aTarget.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::Copy(Polynomial *aTarget)
{
	int i = aTarget->Allocate(this->degree);
	if (i!=ConstRC::Ok) return i;
	for(i=0;i<=degree;i++)
		mpz_set(aTarget->coeff[i],this->coeff[i]);
	aTarget->degree=this->degree;
	return ConstRC::Ok;
}

/*!
	Prints the polynomial on the screen.

	Return code: ConstRC::Ok
*/

int Polynomial::PrintToScreen()
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
	int i;
	this->Reduce();
	if (degree<0)
	{
		CLog::Log(5000, "0900000", "0");
		return ConstRC::Ok;
	}
	for (i=degree; i >= 0; i--) 
		if (mpz_cmp_ui(coeff[i],0))
		{
			if (i == degree)
			{
				if (mpz_cmp_si(coeff[i],-1)==0)
				{
					if (i) CLog::Log(5000, "0900001", "-"); else CLog::Log(5000, "0900001", "-1");
				} 
				else 
				{
					if (mpz_cmp_si(coeff[i],1)!=0) CLog::Log(5000, "0900001", "%Zd",coeff[i]);
					else if (!i) CLog::Log(5000, "0900001", "1");
				} 
			}
			else 
			{
				if ((mpz_cmp_si(coeff[i],-1)==0)&&(i)) CLog::Log(5000, "", "-");
				else 
				{
					if ((mpz_cmp_si(coeff[i],1)==0)&&(i)) CLog::Log(5000, "", "+");
					else CLog::Log(5000, "", "%+Zd",coeff[i]);
				}
			}
			if (i > 1) CLog::Log(5000, "", "x^%d",i);
			else if (i == 1) CLog::Log(5000, "", "x");
		}
	CLog::Log(5000, "", "\n");
#endif
	return ConstRC::Ok;
}

/*!
	Compare aOperand polynomial with this polynomial.
*/
bool Polynomial::Equals(const Polynomial &aOperand) const
{
	if (this->degree != aOperand.degree) return false;

	for (int i = 0; i <= this->degree; i++)
	{
		if (mpz_cmp(this->coeff[i], aOperand.coeff[i]) != 0) return false;
	}

	return true;
}

/*!
	Sets the aPower-th coefficient to be aNumber.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::Set(int aPower,long aNumber){
	if (aPower<0) return ConstRC::Ok;
	int ret;
	if (aPower>allocated){
		ret=this->Allocate(aPower);
		if(ret!=ConstRC::Ok) return ret;
	}
	if (aPower>degree){
		int i=degree+1;
		degree=aPower;
		this->Zeroize(i);
	}
	mpz_set_si(coeff[aPower],aNumber);
	return ConstRC::Ok;
}

/*!
	Sets the aPower-th coefficient to be aNumber.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::Set(int aPower,mpz_t aNumber){
	if (aPower<0) return ConstRC::Ok;
	int ret;
	if (aPower>allocated){
		ret=this->Allocate(aPower);
		if(ret!=ConstRC::Ok) return ret;
	}
	if (aPower>degree){
		int i=degree+1;
		degree=aPower;
		this->Zeroize(i);
	}
	mpz_set(coeff[aPower],aNumber);
	return ConstRC::Ok;
}

/*!
	Increases the aPower-th coefficient by aNumber.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::Inc(int aPower,long aNumber){
	if (aPower<0) return ConstRC::Ok;
	int ret;
	if (aPower>allocated){
		ret=this->Allocate(aPower);
		if(ret!=ConstRC::Ok) return ret;
	}
	if (aPower>degree){
		int i=degree+1;
		degree=aPower;
		this->Zeroize(i);
	}
	if (aNumber>0) mpz_add_ui(coeff[aPower],coeff[aPower],aNumber);
	else mpz_sub_ui(coeff[aPower],coeff[aPower],-aNumber);
	return ConstRC::Ok;
}

/*!
	Increases the aPower-th coefficient by aNumber.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::Inc(int aPower,mpz_t aNumber){
	if (aPower<0) return ConstRC::Ok;
	int ret;
	if (aPower>allocated){
		ret=this->Allocate(aPower);
		if(ret!=ConstRC::Ok) return ret;
	}
	if (aPower>degree){
		int i=degree+1;
		degree=aPower;
		this->Zeroize(i);
	}
	mpz_add(coeff[aPower],coeff[aPower],aNumber);
	return ConstRC::Ok;
}

/*!
	Decreases the aPower-th coefficient by aNumber.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::Dec(int aPower,long aNumber){
	if (aPower<0) return ConstRC::Ok;
	int ret;
	if (aPower>allocated){
		ret=this->Allocate(aPower);
		if(ret!=ConstRC::Ok) return ret;
	}
	if (aPower>degree){
		int i=degree+1;
		degree=aPower;
		this->Zeroize(i);
	}
	if (aNumber>0) mpz_sub_ui(coeff[aPower],coeff[aPower],aNumber);
	else mpz_add_ui(coeff[aPower],coeff[aPower],-aNumber);
	return ConstRC::Ok;
}

/*!
	Decreases the aPower-th coefficient by aNumber.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::Dec(int aPower,mpz_t aNumber){
	if (aPower<0) return ConstRC::Ok;
	int ret;
	if (aPower>allocated){
		ret=this->Allocate(aPower);
		if(ret!=ConstRC::Ok) return ret;
	}
	if (aPower>degree){
		int i=degree+1;
		degree=aPower;
		this->Zeroize(i);
	}
	mpz_sub(coeff[aPower],coeff[aPower],aNumber);
	return ConstRC::Ok;
}

/*!
	Returns the aPower-th coefficient of the polynomial. Coefficients of index less than 0 or more than degree are considered to be 0.
*/

long Polynomial::Get(int aPower){
	if (aPower<0) return 0;
	if (aPower>degree) return 0;
	return(mpz_get_si(coeff[aPower]));
}

/*!
	Returns the reference to the aPower-th coefficient of the polynomial. Coefficients of index less than 0 or more than degree are considered to be 0.
*/

mpz_t *Polynomial::Get2(int aPower){
	if (aPower<0) return &Z_E_R_O;
	if (aPower>degree) return &Z_E_R_O;
	return(&coeff[aPower]);
}

/*!
	Returns the degree of the polynomial.
*/

int Polynomial::GetDegree(){
	this->Reduce();
	return(degree);
}

/*!
	Sets the degree variable to be aDegree. Useful e.g. for setting a polynomial to be 0.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::SetDegree(int aDegree){
	if (aDegree<-1) aDegree=-1;
	this->degree=aDegree;
	if (aDegree>allocated) return(this->Allocate(aDegree));
	else return ConstRC::Ok;
}

/*!
	Returns the leading coefficient of the polynomial.
*/

long Polynomial::Leading(){
	this->Reduce();
	if (degree<0) return 0;
	return(mpz_get_si(coeff[degree]));
}

/*!
	Returns the reference to the leading coefficient of the polynomial.
*/

mpz_t *Polynomial::Leading2(){
	this->Reduce();
	if (degree<0) return &Z_E_R_O;
	return(&coeff[degree]);
}

/*!
	Returns the content (i.e. the greatest common divisor of the conefficients) of the polynomial.
*/

long Polynomial::Content(mpz_t aContent){
	this->Reduce();
	int i;
	if (degree>=0) mpz_set(aContent,coeff[degree]);
		else mpz_set_ui(aContent,1);
	for(i=0;i<degree;i++)
		if (mpz_cmp_ui(coeff[i],0)){
			mpz_gcd(aContent,aContent,coeff[i]);
		}
	return (mpz_get_si(aContent));
}

bool Polynomial::IsOne(){
	if (GetDegree()!=0) return FALSE;
	if (mpz_cmp_ui(coeff[0],1)) return FALSE;
	return TRUE;
}


/*!
	Performs adding aOperand1+aOperand2=aTarget of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::Add(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	int i,h,k;
	k=CIntegerClass::Max(aOperand1->degree,aOperand2->degree);
	int ret=aTarget->Allocate(k);
	if (ret!=ConstRC::Ok) return ret;
	h=CIntegerClass::Min(aOperand1->degree,aOperand2->degree);
	for(i=0;i<=h;i++)
		mpz_add(aTarget->coeff[i],aOperand1->coeff[i],aOperand2->coeff[i]);
	for(i=h+1;i<=aOperand1->degree;i++)
		mpz_set(aTarget->coeff[i],aOperand1->coeff[i]);
	for(i=h+1;i<=aOperand2->degree;i++)
		mpz_set(aTarget->coeff[i],aOperand2->coeff[i]);
	aTarget->degree=k;
	return ConstRC::Ok;
}

/*!
	Performs adding aOperand1+aOperand2=aTarget of polynomials mod aPrime.
	aOperand1 and aOperand2 must have coeff mod aPrime!

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::AddModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, unsigned long aPrime)
{
	int i;
	int k = max(aOperand1->GetDegree(), aOperand2->GetDegree());
	int ret = aTarget->Allocate(k);
	if (ret != ConstRC::Ok) return ret;
	int h = min(aOperand1->degree, aOperand2->degree);
	
	for (i = 0; i <= h; i++)
	{
		mpz_add(aTarget->coeff[i], aOperand1->coeff[i], aOperand2->coeff[i]);
		if (mpz_cmp_ui(aTarget->coeff[i], aPrime) >= 0)
		{
			mpz_sub_ui(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		}
	}
	for (i = h + 1; i <= aOperand1->degree; i++)
	{
		mpz_set(aTarget->coeff[i], aOperand1->coeff[i]);
	}
	for (i = h + 1; i <= aOperand2->degree; i++)
	{
		mpz_set(aTarget->coeff[i], aOperand2->coeff[i]);
	}

	aTarget->degree = k;
	
	return ConstRC::Ok;
}

/*!
	Performs adding aOperand1+aOperand2=aTarget of polynomials mod aPrime.
	aOperand1 and aOperand2 must have coeff mod aPrime!

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::AddModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime)
{
	int i;
	int k = max(aOperand1->GetDegree(), aOperand2->GetDegree());
	int ret = aTarget->Allocate(k);
	if (ret != ConstRC::Ok) return ret;
	int h = min(aOperand1->degree, aOperand2->degree);
	
	for (i = 0; i <= h; i++)
	{
		mpz_add(aTarget->coeff[i], aOperand1->coeff[i], aOperand2->coeff[i]);
		if (mpz_cmp(aTarget->coeff[i], aPrime) >= 0)
		{
			mpz_sub(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		}
	}
	for (i = h + 1; i <= aOperand1->degree; i++)
	{
		mpz_set(aTarget->coeff[i], aOperand1->coeff[i]);
	}
	for (i = h + 1; i <= aOperand2->degree; i++)
	{
		mpz_set(aTarget->coeff[i], aOperand2->coeff[i]);
	}

	aTarget->degree = k;
	
	return ConstRC::Ok;
}

/*!
	Performs aNumber * aOperand + aTarget = aTarget of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::AddMultiply(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber)
{
	int k = CIntegerClass::Max(aTarget->degree, aOperand->degree);
	int ret = aTarget->Allocate(k);
	if (ret != ConstRC::Ok) 
	{
		return ret;
	}
	int h = aOperand->degree;
	for(int i = 0; i <= h; i++)
	{
		mpz_addmul(aTarget->coeff[i], aOperand->coeff[i], aNumber);
	}

	aTarget->degree = k;

	return ConstRC::Ok;
}

/*!
	Performs adding aOperand1+aOperand2=aTarget of polynomials in GF(aCharacteristic)[x]/aModulo).

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::AddInGF(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo, mpz_t aCharacteristic){
	int ret = Add (aTarget, aOperand1, aOperand2);
	aTarget -> Modulo(aCharacteristic);
	return ret;
}

/*!
	Performs substraction aOperand1-aOperand2=aTarget of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::Substract(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	int i,h,k;
	k=max(aOperand1->GetDegree(), aOperand2->GetDegree());
	int ret=aTarget->Allocate(k);
	if (ret!=ConstRC::Ok) return ret;
	h=CIntegerClass::Min(aOperand1->degree,aOperand2->degree);
	for(i=0;i<=h;i++)
		mpz_sub(aTarget->coeff[i],aOperand1->coeff[i],aOperand2->coeff[i]);
	for(i=h+1;i<=aOperand1->degree;i++)
		mpz_set(aTarget->coeff[i],aOperand1->coeff[i]);
	for(i=h+1;i<=aOperand2->degree;i++)
		mpz_neg(aTarget->coeff[i],aOperand2->coeff[i]);
	aTarget->degree=k;
	return ConstRC::Ok;
}

/*!
	Performs substraction aOperand1-aOperand2=aTarget of polynomials mod aPrime.
	aOperand1 and aOperand2 must have coeff mod aPrime!

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::SubstractModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, unsigned long aPrime)
{
	int i;
	int k = max(aOperand1->GetDegree(), aOperand2->GetDegree());
	int ret = aTarget->Allocate(k);
	if (ret!=ConstRC::Ok) return ret;

	int h = min(aOperand1->degree, aOperand2->degree);
	for (i = 0; i <= h; i++)
	{
		mpz_sub(aTarget->coeff[i], aOperand1->coeff[i], aOperand2->coeff[i]);
		if (mpz_cmp_ui(aTarget->coeff[i], 0) < 0)
		{
			mpz_add_ui(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		}
	}
	for (i = h + 1; i <= aOperand1->degree; i++)
	{
		mpz_set(aTarget->coeff[i], aOperand1->coeff[i]);
	}
	for (i = h + 1; i <= aOperand2->degree; i++)
	{
		if (mpz_cmp_ui(aOperand2->coeff[i], 0) != 0)
		{
			mpz_ui_sub(aTarget->coeff[i], aPrime, aOperand2->coeff[i]);
		}
	}

	aTarget->degree = k;

	return ConstRC::Ok;
}

/*!
	Performs substraction aOperand1-aOperand2=aTarget of polynomials mod aPrime.
	aOperand1 and aOperand2 must have coeff mod aPrime!

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::SubstractModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime)
{
	int i;
	int k = max(aOperand1->GetDegree(), aOperand2->GetDegree());
	int ret = aTarget->Allocate(k);
	if (ret!=ConstRC::Ok) return ret;

	int h = min(aOperand1->degree, aOperand2->degree);
	for (i = 0; i <= h; i++)
	{
		mpz_sub(aTarget->coeff[i], aOperand1->coeff[i], aOperand2->coeff[i]);
		if (mpz_cmp_ui(aTarget->coeff[i], 0) < 0)
		{
			mpz_add(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		}
	}
	for (i = h + 1; i <= aOperand1->degree; i++)
	{
		mpz_set(aTarget->coeff[i], aOperand1->coeff[i]);
	}
	for (i = h + 1; i <= aOperand2->degree; i++)
	{
		if (mpz_cmp_ui(aOperand2->coeff[i], 0) != 0)
		{
			mpz_sub(aTarget->coeff[i], aPrime, aOperand2->coeff[i]);
		}
	}

	aTarget->degree = k;

	return ConstRC::Ok;
}

/*!
	Performs substraction aOperand1-aOperand2=aTarget of polynomials in GF(aCharacteristic)[x]/aModulo).

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::SubstractInGF(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo, mpz_t aCharacteristic){
	int ret = Substract (aTarget, aOperand1, aOperand2);
	aTarget -> Modulo(aCharacteristic);
	return ret;
}

/*!
	Performs multiplication aOperand1*aOperand2=aTarget of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::Multiply(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	if ((aOperand1->degree<0)||(aOperand2->degree<0)){
		aTarget->degree=-1;
		return ConstRC::Ok;
	}
	int h,i,j;
	mpz_t aux;
	mpz_init(aux);
	h = aOperand1->GetDegree()+aOperand2->GetDegree();
	int ret=aTarget->Allocate(h);
	if (ret!=ConstRC::Ok) return ret;
	for(i=0;i<=h;i++){
		mpz_set_ui(aTarget->coeff[i],0);
		for(j=max(i-aOperand2->degree, 0);j<=i;j++)
		{
			if(j>aOperand1->degree) break;
			mpz_mul(aux,aOperand1->coeff[j],aOperand2->coeff[i-j]);
			mpz_add(aTarget->coeff[i],aTarget->coeff[i],aux);
		}
	}
	aTarget->degree=h;
	mpz_clear(aux);
	return ConstRC::Ok;
}

/*!
	Performs multiplication aOperand1*aOperand2=aTarget of polynomials mod aPrime.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, unsigned long aPrime)
{
	if ((aOperand1->degree < 0) || (aOperand2->degree < 0))
	{
		aTarget->degree = -1;
		return ConstRC::Ok;
	}
	int j;
	mpz_t aux;
	mpz_init(aux);
	int h = aOperand1->GetDegree() + aOperand2->GetDegree();
	int ret = aTarget->Allocate(h);
	if (ret != ConstRC::Ok) return ret;

	for (int i = 0; i <= h; i++)
	{
		mpz_set_ui(aTarget->coeff[i], 0);
		for (j = max(i - aOperand2->degree, 0); j <= i; j++)
		{
			if (j > aOperand1->degree) break;
			mpz_mul(aux, aOperand1->coeff[j], aOperand2->coeff[i - j]);
			mpz_add(aTarget->coeff[i], aTarget->coeff[i], aux);
			mpz_mod_ui(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		}
	}

	aTarget->degree = h;
	mpz_clear(aux);

	return ConstRC::Ok;
}

/*!
	Performs multiplication aOperand1*aOperand2=aTarget of polynomials mod aPrime.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime)
{
	if ((aOperand1->degree < 0) || (aOperand2->degree < 0))
	{
		aTarget->degree = -1;
		return ConstRC::Ok;
	}
	int j;
	mpz_t aux;
	mpz_init(aux);
	int h = aOperand1->GetDegree() + aOperand2->GetDegree();
	int ret = aTarget->Allocate(h);
	if (ret != ConstRC::Ok) return ret;

	for (int i = 0; i <= h; i++)
	{
		mpz_set_ui(aTarget->coeff[i], 0);
		for (j = max(i - aOperand2->degree, 0); j <= i; j++)
		{
			if (j > aOperand1->degree) break;
			mpz_mul(aux, aOperand1->coeff[j], aOperand2->coeff[i - j]);
			mpz_add(aTarget->coeff[i], aTarget->coeff[i], aux);
			mpz_mod(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		}
	}

	aTarget->degree = h;
	mpz_clear(aux);

	return ConstRC::Ok;
}

/*!
	Performs multiplication aOperand1*aOperand2=aTarget of polynomials in GF(aCharacteristic)[x]/aModulo).

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::MultiplyInGF(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo, mpz_t aCharacteristic){
	int ret = Multiply (aTarget, aOperand1, aOperand2);
	if (ret!=ConstRC::Ok) return ret;
	ret = aTarget -> ReduceInGF(aModulo,aCharacteristic);
	return ret;
}

/*!
	Multiplies the polynomial by <i>x</i> to the power aPower.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate the polynomial</td></tr>
	</table>
*/

int Polynomial::MultiplyByPowerOfX(int aPower){
	if (aPower<=0) return ConstRC::Ok;
	int ret=this->Allocate(degree+aPower);
	if (ret!=ConstRC::Ok) return ret;
	int i;
	for (i=degree;i>=0;i--)
		mpz_set(coeff[i+aPower],coeff[i]);
	i=degree;
	degree=aPower-1;
	this->Zeroize(0);
	degree=i+aPower;
	return ConstRC::Ok;
}

/*!
	Performs division with remainder aOperand1=aOperand2*aTarget+aRemainder of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/
int Polynomial::Divide(Polynomial *aTarget, Polynomial *aRemainder, Polynomial *aOperand1, Polynomial *aOperand2){
	if (aOperand2->GetDegree()<0) return -1;
	int h,i;
	aOperand1->Reduce();
	Polynomial *aux = AssignAuxPoly();
	int ret=aux->Allocate(aOperand2->degree);
	if (ret!=ConstRC::Ok) return ret;
	h = aOperand1->degree-aOperand2->degree;
	ret=aTarget->Allocate(h);
	if (ret!=ConstRC::Ok) return ret;
	aOperand1->Copy(aRemainder);
	for(i=h;i>=0;i--){
		mpz_fdiv_q(aTarget->coeff[i],aRemainder->coeff[i+aOperand2->degree],aOperand2->coeff[aOperand2->degree]);
		this->Multiply(aux,aOperand2,aTarget->coeff[i]);
		aux->MultiplyByPowerOfX(i);
		this->Substract(aRemainder,aRemainder,aux);
	}
	aTarget->degree=CIntegerClass::Max(h,-1);
	Dispose(aux);
	return ConstRC::Ok;
}

/*!
	Performs division with remainder aOperand1 = aOperand2 * aTarget + aRemainder of polynomials mod aPrime.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/
int Polynomial::DivideModPrime(Polynomial *aTarget, Polynomial *aRemainder, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime)
{
	if (aOperand2->GetDegree() < 0) return -1;
	aOperand1->Reduce();
	Polynomial *aux = AssignAuxPoly();
	int ret = aux->Allocate(aOperand2->degree);
	if (ret != ConstRC::Ok) return ret;
	int h = aOperand1->degree - aOperand2->degree;
	ret = aTarget->Allocate(h);
	if (ret != ConstRC::Ok) return ret;
	aOperand1->Copy(aRemainder);

	mpz_t linv;
	mpz_t coeff;
	mpz_init(linv);
	mpz_init(coeff);
	mpz_invert(linv, *aOperand2->Leading2(), aPrime);

	for (int i = h; i >= 0; i--)
	{
		mpz_mul(aTarget->coeff[i], aRemainder->coeff[i + aOperand2->degree], linv);
		mpz_mod(aTarget->coeff[i], aTarget->coeff[i], aPrime);
		MultiplyModPrime(aux, aOperand2, aTarget->coeff[i], aPrime);
		aux->MultiplyByPowerOfX(i);
		SubstractModPrime(aRemainder, aRemainder, aux, aPrime);
	}

	aTarget->degree = CIntegerClass::Max(h, -1);
	Dispose(aux);

	return ConstRC::Ok;
}
	   
/*!
	Performs number multiplication aOperand*aNumber=aTarget.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::Multiply(Polynomial *aTarget, Polynomial *aOperand, long aNumber){
	int i;
	int ret=aTarget->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	for(i=0;i<=aOperand->degree;i++)
		mpz_mul_si(aTarget->coeff[i],aOperand->coeff[i],aNumber);
	aTarget->degree=aOperand->degree;
	return ConstRC::Ok;
}

/*!
	Performs number multiplication aOperand*aNumber=aTarget.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::Multiply(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber){
	int i;
	int ret=aTarget->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	for(i=0;i<=aOperand->degree;i++)
		mpz_mul(aTarget->coeff[i],aOperand->coeff[i],aNumber);
	aTarget->degree=aOperand->degree;
	return ConstRC::Ok;
}

/*!
	Performs number multiplication aOperand*aNumber=aTarget mod aPrime.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand, unsigned long aNumber, unsigned long aPrime)
{
	int ret = aTarget->Allocate(aOperand->GetDegree());
	if (ret != ConstRC::Ok) return ret;

	for (int i = 0; i <= aOperand->degree; i++)
	{
		mpz_mul_ui(aTarget->coeff[i], aOperand->coeff[i], aNumber);
		mpz_mod_ui(aTarget->coeff[i], aTarget->coeff[i], aPrime);
	}

	aTarget->degree = aOperand->degree;

	return ConstRC::Ok;
}

/*!
	Performs number multiplication aOperand*aNumber=aTarget mod aPrime.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/
int Polynomial::MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber, mpz_t aPrime)
{
	int ret = aTarget->Allocate(aOperand->GetDegree());
	if (ret != ConstRC::Ok) return ret;

	for (int i = 0; i <= aOperand->degree; i++)
	{
		mpz_mul(aTarget->coeff[i], aOperand->coeff[i], aNumber);
		mpz_mod(aTarget->coeff[i], aTarget->coeff[i], aPrime);
	}

	aTarget->degree = aOperand->degree;

	return ConstRC::Ok;
}

/*!
	Performs number division aOperand/aNumber=aTarget.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::Divide(Polynomial *aTarget, Polynomial *aOperand, long aNumber){
	if (aNumber==0) return -1;
	int i;
	long j=abs(aNumber);
	int ret=aTarget->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	for(i=0;i<=aOperand->degree;i++){
		mpz_divexact_ui(aTarget->coeff[i],aOperand->coeff[i],j);
	}
	aTarget->degree=aOperand->degree;
	if (aNumber<0) this->Multiply(aTarget,aTarget,-1);
	return ConstRC::Ok;
}

/*!
	Performs number division aOperand/aNumber=aTarget.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::Divide(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber){
	if (aNumber==0) return -1;
	int i;
	int ret=aTarget->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	for(i=0;i<=aOperand->degree;i++){
		mpz_divexact(aTarget->coeff[i],aOperand->coeff[i],aNumber);
	}
	aTarget->degree=aOperand->degree;
	return ConstRC::Ok;
}

/*!
	Performs powering aOperand^x=aTarget of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::Power(Polynomial *aTarget, Polynomial *aOperand, unsigned int x){
	int i,l;
	Polynomial *aux=NULL,*aux2;
	aux2=AssignAuxPoly();
	int ret=aux2->Allocate(aOperand->degree*x);
	if (ret!=ConstRC::Ok) return ret;
	ret=aTarget->Allocate(aOperand->degree*x);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->degree=0;
	mpz_set_ui(aTarget->coeff[0],1);
	i=x;
	aux=AssignAuxPoly();
	aOperand->Copy(aux);
	for(l=0;i;l++) i=i>>1;
	for(i=0;i<l;i++){
		//aTarget->PrintToScreen();puts("");
		if (i) aux->Multiply(aux,aux2,aux2);
		if (x & 1<<i) {
			this->Multiply(aux2,aTarget,aux);
			aux2->Copy(aTarget);
		}
		aux->Copy(aux2);
	}
	Dispose(aux);
	Dispose(aux2);
	return ConstRC::Ok;
}

/*!
	Performs powering aOperand^x=aTarget (mod aModulo) of polynomials.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::PowerMod(Polynomial *aTarget, Polynomial *aOperand, unsigned int x,Polynomial *aModulo){
	int i,l;
	Polynomial *aux,*aux2,*aux3;
	aux2=AssignAuxPoly();
	int ret=aux2->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	ret=aTarget->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->degree=0;
	mpz_set_ui(aTarget->coeff[0],1);
	i=x;
	aux=AssignAuxPoly();
	aux3=AssignAuxPoly();
	ret=aux->Divide(aux2,aux,aOperand,aModulo);
	if (ret!=ConstRC::Ok) return ret;
//		  aux->PrintToScreen();puts("");
//		  aModulo->PrintToScreen();puts("");
	for(l=0;i;l++) i=i>>1;
	for(i=0;i<l;i++){
		//aTarget->PrintToScreen();puts("");
		if (i) {
			ret=this->Multiply(aux3,aux2,aux2);
			if (ret!=ConstRC::Ok) return ret;
			ret=this->Divide(aux2,aux,aux3,aModulo);
			if (ret!=ConstRC::Ok) return ret;
//		  aux->PrintToScreen();puts("");
		}
		if (x & 1<<i) {
			ret=this->Multiply(aux2,aTarget,aux);
			if (ret!=ConstRC::Ok) return ret;
			ret=this->Divide(aux3,aTarget,aux2,aModulo);
			if (ret!=ConstRC::Ok) return ret;
//		  aTarget->PrintToScreen();puts("");
		}
		ret=aux->Copy(aux2);
		if (ret!=ConstRC::Ok) return ret;
	}
	Dispose(aux);
	Dispose(aux2);
	Dispose(aux3);
	return ConstRC::Ok;
}

/*!
	Performs powering aOperand^x=aTarget (mod aModulo) of polynomials in the field <I>GF</I>(aPrime).

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::PowerModPolyPrime(Polynomial *aTarget, Polynomial *aOperand, unsigned int x,Polynomial *aModulo,long aPrime){
	int i,l;
	long d;
	Polynomial *aux,*aux2,*aux3,*auxMod;
	aux2=AssignAuxPoly();
	auxMod = AssignAuxPoly();
	int ret=aux2->Allocate(aOperand->GetDegree());
	if (ret!=ConstRC::Ok) return ret;
	ret=aTarget->Allocate(aOperand->degree);
	if (ret!=ConstRC::Ok) return ret;
	ret=auxMod->Allocate(aModulo->GetDegree());
	if (ret!=ConstRC::Ok) return ret;
	aModulo->Copy(auxMod);
	d=CIntegerClass::InverseModN(mpz_fdiv_ui(auxMod->coeff[aModulo->degree],aPrime),aPrime);
	if (d!=1){
		auxMod->Multiply(auxMod,auxMod,d);
		auxMod->Modulo(aPrime);
	}
	aTarget->degree=0;
	mpz_set_ui(aTarget->coeff[0],1);
	i=x;
	aux=AssignAuxPoly();
	aux3=AssignAuxPoly();
	ret=aux->Divide(aux2,aux,aOperand,aModulo);
	if (ret!=ConstRC::Ok) return ret;
//		  aux->PrintToScreen();puts("");
//		  aModulo->PrintToScreen();puts("");
	for(l=0;i;l++) i=i>>1;
	for(i=0;i<l;i++){
//		aTarget->PrintToScreen();puts("");
		if (i) {
			ret=this->Multiply(aux3,aux2,aux2);
			if (ret!=ConstRC::Ok) return ret;
			aux3->SymmetricModulo(aPrime);
			ret=this->Divide(aux2,aux,aux3,auxMod);
			if (ret!=ConstRC::Ok) return ret;
			aux->SymmetricModulo(aPrime);
//		  aux->PrintToScreen();puts("");
		}
		if (x & 1<<i) {
			ret=this->Multiply(aux2,aTarget,aux);
			if (ret!=ConstRC::Ok) return ret;
			aux2->SymmetricModulo(aPrime);
			ret=this->Divide(aux3,aTarget,aux2,auxMod);
			if (ret!=ConstRC::Ok) return ret;
			aTarget->SymmetricModulo(aPrime);
//		  aTarget->PrintToScreen();puts("");
		}
		ret=aux->Copy(aux2);
		if (ret!=ConstRC::Ok) return ret;
	}
//	if (d) aTarget->Multiply(aTarget,aTarget,aModulo->Leading());
	Dispose(aux);
	Dispose(aux2);
	Dispose(aux3);
	Dispose(auxMod);
	return ConstRC::Ok;
}

int Polynomial::MultiplyMod(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo){
	int ret;
	Polynomial *prod,*quo;
	prod=AssignAuxPoly();
	quo=AssignAuxPoly();
	ret = this->Multiply(prod,aOperand1,aOperand2);
	if (ret!=ConstRC::Ok) return ret;
	ret = this->Divide(quo,aTarget,prod,aModulo);
	Dispose(prod);
	Dispose(quo);
	return ret;
}

/*!
	Performs powering aOperand1^x=aTarget of polynomials in GF(aCharacteristic)[x]/aModulo).

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::PowerInGF(Polynomial *aTarget, unsigned int x,Polynomial *aModulo,mpz_t aCharacteristic){
	int i,l;
	Polynomial *aux,*aux2,*aux3;
	aux2=AssignAuxPoly();
	int ret=aux2->Allocate(this->degree);
	if (ret!=ConstRC::Ok) return ret;
	ret=aTarget->Allocate(this->degree);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->degree=0;
	mpz_set_ui(aTarget->coeff[0],1);
	i=x;
	aux=AssignAuxPoly();
	aux3=AssignAuxPoly();
	ret=aux->Divide(aux2,aux,this,aModulo);
	if (ret!=ConstRC::Ok) return ret;
//		  aux->PrintToScreen();puts("");
//		  aModulo->PrintToScreen();puts("");
	for(l=0;i;l++) i=i>>1;
	for(i=0;i<l;i++){
//		aTarget->PrintToScreen();puts("");
		if (i) {
			ret=this->Multiply(aux3,aux2,aux2);
			if (ret!=ConstRC::Ok) return ret;
			aux3->SymmetricModulo(aCharacteristic);
			ret=this->Divide(aux2,aux,aux3,aModulo);
			if (ret!=ConstRC::Ok) return ret;
			aux->SymmetricModulo(aCharacteristic);
//		  aux->PrintToScreen();puts("");
		}
		if (x & 1<<i) {
			ret=this->Multiply(aux2,aTarget,aux);
			if (ret!=ConstRC::Ok) return ret;
			aux2->SymmetricModulo(aCharacteristic);
			ret=this->Divide(aux3,aTarget,aux2,aModulo);
			if (ret!=ConstRC::Ok) return ret;
			aTarget->SymmetricModulo(aCharacteristic);
//		  aTarget->PrintToScreen();puts("");
		}
		ret=aux->Copy(aux2);
		if (ret!=ConstRC::Ok) return ret;
	}
	Dispose(aux);
	Dispose(aux2);
	Dispose(aux3);
	return ConstRC::Ok;
}

int Polynomial::PowerInGF(Polynomial *aTarget, mpz_t x,Polynomial *aModulo,mpz_t aCharacteristic){
	long i,l;
	Polynomial *aux,*aux2,*aux3;
	aux2=AssignAuxPoly();
	int ret=aux2->Allocate(this->degree);
	if (ret!=ConstRC::Ok) return ret;
	ret=aTarget->Allocate(this->degree);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->degree=0;
	mpz_set_ui(aTarget->coeff[0],1);
	aux=AssignAuxPoly();
	aux3=AssignAuxPoly();
	ret=aux->Divide(aux2,aux,this,aModulo);
	if (ret!=ConstRC::Ok) return ret;
//		  aux->PrintToScreen();puts("");
//		  aModulo->PrintToScreen();puts("");
	l=mpz_sizeinbase(x,2);
	for(i=0;i<l;i++){
//		aTarget->PrintToScreen();puts("");
		if (i) {
			ret=this->Multiply(aux3,aux2,aux2);
			if (ret!=ConstRC::Ok) return ret;
			aux3->SymmetricModulo(aCharacteristic);
			ret=this->Divide(aux2,aux,aux3,aModulo);
			if (ret!=ConstRC::Ok) return ret;
			aux->SymmetricModulo(aCharacteristic);
//		  aux->PrintToScreen();puts("");
		}
		if (mpz_tstbit(x,i)) {
			ret=this->Multiply(aux2,aTarget,aux);
			if (ret!=ConstRC::Ok) return ret;
			aux2->SymmetricModulo(aCharacteristic);
			ret=this->Divide(aux3,aTarget,aux2,aModulo);
			if (ret!=ConstRC::Ok) return ret;
			aTarget->SymmetricModulo(aCharacteristic);
//		  aTarget->PrintToScreen();puts("");
		}
		ret=aux->Copy(aux2);
		if (ret!=ConstRC::Ok) return ret;
	}
	Dispose(aux);
	Dispose(aux2);
	Dispose(aux3);
	return ConstRC::Ok;
}
/*!
	Returns a the reduced representant of the polynomials in GF(aCharacteristic)[x]/aModulo).

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate aTarget</td></tr>
	</table>
*/

int Polynomial::ReduceInGF(Polynomial *aModulo, mpz_t aCharacteristic){
	Polynomial *aux,*aux2;
	aux = AssignAuxPoly();
	aux2 = AssignAuxPoly();
	int ret = this->Divide(aux,aux2,this,aModulo);
	if (ret!=ConstRC::Ok) return ret;
	ret = aux2->Copy(this);
	if (ret!=ConstRC::Ok) return ret;
	this->Modulo(aCharacteristic);
	Dispose(aux);
	Dispose(aux2);
	return ConstRC::Ok;
}

/*!
	Finds the greatest common divisor of aOperand1 and aOperand2 using the Euclidian algorithm.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::GCD_Euclid(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e;
	int ret;
	mpz_t aux;
	mpz_init(aux);
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
	}
	while(b->degree>=0){
		mpz_pow_ui(aux,*b->Leading2(),a->degree-b->degree+1);
/*		gmp_printf("%Zd(",aux);
		a->PrintToScreen();
		printf(") = (");
		b->PrintToScreen();
		printf(")(");fflush(stdout);
*/		ret=this->Multiply(c,a,aux);
		if (ret!=ConstRC::Ok) return ret;
		ret=this->Divide(a,d,c,b);
		if (ret!=ConstRC::Ok) return ret;
/*		a->PrintToScreen();
		printf(")+(");
		d->PrintToScreen();
		printf(")\n");
*/		e=a;
		a=b;
		b=d;
		d=e;
		b->Reduce();
	}
	a->Content(aux);
	ret=this->Divide(aTarget,a,aux);
	if (ret!=ConstRC::Ok) return ret;
	mpz_clear(aux);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	return ConstRC::Ok;
}

/*!
	Finds the greatest common divisor of aOperand1 and aOperand2 using the reduced algorithm.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::GCD_Reduced(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e;
	int ret;
	mpz_t alpha,beta;
	mpz_init(alpha);
	mpz_init_set_ui(beta,1);
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
	}
	while(b->degree>=0){
		mpz_pow_ui(alpha,*b->Leading2(),a->degree-b->degree+1);
/*		gmp_printf("%Zd(",alpha);
		a->PrintToScreen();
		printf(") = (");
		b->PrintToScreen();
		printf(")(");fflush(stdout);
*/		ret=this->Multiply(c,a,alpha);
		if (ret!=ConstRC::Ok) return ret;
		ret=this->Divide(a,d,c,b);
		if (ret!=ConstRC::Ok) return ret;
/*		a->PrintToScreen();
		gmp_printf(")%+Zd(",beta);
*/		ret=this->Divide(c,d,beta);
		if (ret!=ConstRC::Ok) return ret;
/*		c->PrintToScreen();
		printf(")\n");
*/		mpz_set(beta,alpha);
		e=a;
		a=b;
		b=c;
		c=e;
		b->Reduce();
	}
	a->Content(beta);
	ret=this->Divide(aTarget,a,beta);
	if (ret!=ConstRC::Ok) return ret;
	mpz_clear(alpha);
	mpz_clear(beta);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	return ConstRC::Ok;
}

/*!
	Finds the greatest common divisor of aOperand1 and aOperand2 using the subresultant algorithm.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::GCD_Subresultant(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e;
	mpz_t alpha,beta,psi,aux;
	int delta,ret;
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
	}
	mpz_init(alpha);
	mpz_init_set_si(beta,-1);
	mpz_init_set_si(psi,-1);
	mpz_init(aux);
	delta=a->degree-b->degree;
	if (delta % 2) mpz_set_si(beta,1);
	while(b->degree>=0){
		mpz_pow_ui(alpha,*b->Leading2(),delta+1);
/*		gmp_printf("%Zd(",alpha);
		a->PrintToScreen();
		printf(") = (");
		b->PrintToScreen();
		printf(")(");fflush(stdout);
*/		ret=this->Multiply(c,a,alpha);
		if (ret!=ConstRC::Ok) return ret;
		ret=this->Divide(a,d,c,b);
		if (ret!=ConstRC::Ok) return ret;
/*		a->PrintToScreen();
		gmp_printf(")%+Zd(",beta);
*/		ret=this->Divide(c,d,beta);
		if (ret!=ConstRC::Ok) return ret;
/*		c->PrintToScreen();
		gmp_printf(")   psi=%Zd\n",psi);
*/		mpz_neg(aux,*b->Leading2());
		mpz_pow_ui(aux,aux,delta);
		if(delta){
			mpz_pow_ui(psi,psi,delta-1);
			mpz_divexact(psi,aux,psi);
		} else {
			mpz_mul(psi,aux,psi);
		}
		c->Reduce();
		delta=b->degree-c->degree;
		mpz_pow_ui(aux,psi,delta);
		mpz_mul(beta,*b->Leading2(),aux);
		mpz_mul_si(beta,beta,-1);
		e=a;
		a=b;
		b=c;
		c=e;
	}
	a->Content(beta);
	ret=this->Divide(aTarget,a,beta);
	if (ret!=ConstRC::Ok) return ret;
	mpz_clear(alpha);
	mpz_clear(beta);
	mpz_clear(psi);
	mpz_clear(aux);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	return ConstRC::Ok;
}

/*!
	Finds the greatest common divisor of aOperand1 and aOperand2 using the primitive part algorithm.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::GCD_Content(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e;
	int ret;
	mpz_t aux;
	mpz_init(aux);
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
	}
	while(b->degree>=0){
		mpz_pow_ui(aux,*b->Leading2(),a->degree-b->degree+1);
/*		gmp_printf("%Zd(",aux);
		a->PrintToScreen();
		printf(") = (");
		b->PrintToScreen();
		printf(")(");fflush(stdout);
*/		ret=this->Multiply(c,a,aux);
		if (ret!=ConstRC::Ok) return ret;
		ret=this->Divide(a,d,c,b);
		if (ret!=ConstRC::Ok) return ret;
//		a->PrintToScreen();
		d->Content(aux);
//		gmp_printf(")%+Zd(",aux);
		ret=this->Divide(c,d,aux);
		if (ret!=ConstRC::Ok) return ret;
/*		c->PrintToScreen();
		printf(")\n");
*/		e=a;
		a=b;
		b=c;
		c=e;
		b->Reduce();
	}
	ret=a->Copy(aTarget);
	if (ret!=ConstRC::Ok) return ret;
	mpz_clear(aux);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	return ConstRC::Ok;
}

/*!
	Finds the greatest common divisor of aOperand1 and aOperand2 over <i>GF</i>(<i>p</i>) using the Euclidian algorithm.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::GCD_p(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2,unsigned long p){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e;
	int ret;
	long k;
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
	}
	a->SymmetricModulo(p);
	b->SymmetricModulo(p);
	while(b->degree>=0){
		k=CIntegerClass::InverseModN(b->Leading(),p);
/*		printf("(");
		a->PrintToScreen();
		printf(") = (");
*/		ret=this->Multiply(c,b,k);
		if (ret!=ConstRC::Ok) return ret;
		c->SymmetricModulo(p);
		ret=this->Divide(b,d,a,c);
		if (ret!=ConstRC::Ok) return ret;
/*		c->PrintToScreen();
		printf(")(");fflush(stdout);
		b->PrintToScreen();
		printf(")+(");
		d->PrintToScreen();
		printf(")\n");
*/		e=b;
		b=d;
		d=e;
		e=a;
		a=c;
		c=e;
		b->SymmetricModulo(p);
	}
	int i;
	i=CIntegerClass::InverseModN(a->Leading(),p);
	ret=this->Multiply(aTarget,a,i);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->Modulo(p);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	return ConstRC::Ok;
}

/*!
	Finds the greatest common divisor of aOperand1 and aOperand2 over <i>GF</i>(<i>p</i>) using the Euclidian algorithm,
	along with the combinations needed to achieve the greatest common divisor.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::GCD_Extended_p(Polynomial *aTarget, Polynomial *aComb1, Polynomial *aComb2, Polynomial *aOperand1, Polynomial *aOperand2,unsigned long p){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e,*a1,*a2,*b1,*b2,*aux;
	int ret;
	long k;
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	a1 = AssignAuxPoly();
	a2 = AssignAuxPoly();
	b1 = AssignAuxPoly();
	b2 = AssignAuxPoly();
	aux = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
		a2->degree=-1;
		a1->Set(0,1);
		a1->degree=0;
		b1->degree=-1;
		b2->Set(0,1);
		b2->degree=0;
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
		a1->degree=-1;
		a2->Set(0,1);
		a2->degree=0;
		b2->degree=-1;
		b1->Set(0,1);
		b1->degree=0;
	}
	a->SymmetricModulo(p);
	b->SymmetricModulo(p);
	while(b->degree>=0){
		k=CIntegerClass::InverseModN(b->Leading(),p);
/*		printf("(");
		a->PrintToScreen();
		printf(") = (");
*/		ret=this->Multiply(c,b,k);
		if (ret!=ConstRC::Ok) return ret;
		c->SymmetricModulo(p);
		ret=this->Multiply(b1,b1,k);
		if (ret!=ConstRC::Ok) return ret;
		b1->Modulo(p);
		ret=this->Multiply(b2,b2,k);
		if (ret!=ConstRC::Ok) return ret;
		b2->Modulo(p);
		ret=this->Divide(b,d,a,c);
		if (ret!=ConstRC::Ok) return ret;
		ret = Multiply(aux,b,b1);
		if (ret!=ConstRC::Ok) return ret;
		ret = Substract(a1,a1,aux);
		if (ret!=ConstRC::Ok) return ret;
		ret = Multiply(aux,b,b2);
		if (ret!=ConstRC::Ok) return ret;
		ret = Substract(a2,a2,aux);
		if (ret!=ConstRC::Ok) return ret;
		e=a1;
		a1=b1;
		b1=e;
		e=a2;
		a2=b2;
		b2=e;
/*		c->PrintToScreen();
		printf(")(");fflush(stdout);
		b->PrintToScreen();
		printf(")+(");
		d->PrintToScreen();
		printf(")\n");
*/		e=b;
		b=d;
		d=e;
		e=a;
		a=c;
		c=e;
		b->SymmetricModulo(p);
	}
	k=CIntegerClass::InverseModN(a->Leading(),p);
	ret=this->Multiply(aTarget,a,k);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->Modulo(p);
	ret=this->Multiply(aComb1,a1,k);
	if (ret!=ConstRC::Ok) return ret;
	aComb1->Modulo(p);
	ret=this->Multiply(aComb2,a2,k);
	if (ret!=ConstRC::Ok) return ret;
	aComb2->Modulo(p);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	Dispose(a1);
	Dispose(a2);
	Dispose(b1);
	Dispose(b2);
	Dispose(aux);
	return ConstRC::Ok;
}

int Polynomial::GCD_Extended_p(Polynomial *aTarget, Polynomial *aComb1, Polynomial *aComb2, Polynomial *aOperand1, Polynomial *aOperand2,mpz_t p){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e,*a1,*a2,*b1,*b2,*aux;
	int ret;
	mpz_t k;
	mpz_init(k);
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	a1 = AssignAuxPoly();
	a2 = AssignAuxPoly();
	b1 = AssignAuxPoly();
	b2 = AssignAuxPoly();
	aux = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
		a2->degree=-1;
		a1->Set(0,1);
		a1->degree=0;
		b1->degree=-1;
		b2->Set(0,1);
		b2->degree=0;
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
		a1->degree=-1;
		a2->Set(0,1);
		a2->degree=0;
		b2->degree=-1;
		b1->Set(0,1);
		b1->degree=0;
	}
	a->SymmetricModulo(p);
	b->SymmetricModulo(p);
	while(b->degree>=0){
		mpz_invert(k,*b->Leading2(),p);
/*		printf("(");
		a->PrintToScreen();
		printf(") = (");
*/		ret=this->Multiply(c,b,k);
		if (ret!=ConstRC::Ok) return ret;
		c->SymmetricModulo(p);
		ret=this->Multiply(b1,b1,k);
		if (ret!=ConstRC::Ok) return ret;
		b1->Modulo(p);
		ret=this->Multiply(b2,b2,k);
		if (ret!=ConstRC::Ok) return ret;
		b2->Modulo(p);
		ret=this->Divide(b,d,a,c);
		if (ret!=ConstRC::Ok) return ret;
		ret = Multiply(aux,b,b1);
		if (ret!=ConstRC::Ok) return ret;
		ret = Substract(a1,a1,aux);
		if (ret!=ConstRC::Ok) return ret;
		ret = Multiply(aux,b,b2);
		if (ret!=ConstRC::Ok) return ret;
		ret = Substract(a2,a2,aux);
		if (ret!=ConstRC::Ok) return ret;
		e=a1;
		a1=b1;
		b1=e;
		e=a2;
		a2=b2;
		b2=e;
/*		c->PrintToScreen();
		printf(")(");fflush(stdout);
		b->PrintToScreen();
		printf(")+(");
		d->PrintToScreen();
		printf(")\n");
*/		e=b;
		b=d;
		d=e;
		e=a;
		a=c;
		c=e;
		b->SymmetricModulo(p);
	}
	mpz_invert(k,*a->Leading2(),p);
	ret=this->Multiply(aTarget,a,k);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->Modulo(p);
	ret=this->Multiply(aComb1,a1,k);
	if (ret!=ConstRC::Ok) return ret;
	aComb1->Modulo(p);
	ret=this->Multiply(aComb2,a2,k);
	if (ret!=ConstRC::Ok) return ret;
	aComb2->Modulo(p);
	Dispose(a);
	Dispose(b);
	Dispose(c);
	Dispose(d);
	Dispose(a1);
	Dispose(a2);
	Dispose(b1);
	Dispose(b2);
	Dispose(aux);
	mpz_clear(k);
	return ConstRC::Ok;
}

/*!
	Returns the resultant of aOperand1 and aOperand2 using the subresultant algorithm.

	Return codes:
	<table border=0>
	<tr><td>resultant	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

long Polynomial::Resultant(mpz_t aResult, Polynomial *aOperand1, Polynomial *aOperand2){
	Polynomial *a=NULL,*b=NULL,*c,*d,*e;
	int ret;
	mpz_t alpha,beta,psi,aux;
	int delta,k=1;
	a = AssignAuxPoly();
	b = AssignAuxPoly();
	c = AssignAuxPoly();
	d = AssignAuxPoly();
	if(aOperand1->degree>=aOperand2->degree){
		aOperand1->Copy(a);
		aOperand2->Copy(b);
	} else {
		aOperand1->Copy(b);
		aOperand2->Copy(a);
	}
	mpz_init(alpha);
	mpz_init_set_si(beta,-1);
	mpz_init_set_si(psi,-1);
	mpz_init(aux);
	delta=a->degree-b->degree;
	if (delta % 2) mpz_set_si(beta,1);
	while(b->degree>0){
		mpz_pow_ui(alpha,*b->Leading2(),delta+1);
		mpz_pow_ui(aux,beta,b->degree);
		ret=this->Multiply(c,a,alpha);
		if (ret!=ConstRC::Ok) return ret;
		ret=this->Divide(a,d,c,b);
		if (ret!=ConstRC::Ok) return ret;
		if((delta+1)*b->degree>c->GetDegree()-d->GetDegree()){
			mpz_pow_ui(aux,*b->Leading2(),(delta+1)*b->degree-c->degree+d->degree);
		} else {
			mpz_pow_ui(aux,*b->Leading2(),(-delta-1)*b->degree+c->degree-d->degree);
		}
		ret=this->Divide(c,d,beta);
		if (ret!=ConstRC::Ok) return ret;
//		gmp_printf("%Zd/%Zd\n",N,D);fflush(stdout);
		mpz_neg(aux,*b->Leading2());
		mpz_pow_ui(aux,aux,delta);
		if(delta){
			mpz_pow_ui(psi,psi,delta-1);
			mpz_divexact(psi,aux,psi);
		} else {
			mpz_mul(psi,aux,psi);
		}
		c->Reduce();
		delta=b->degree-c->degree;
		mpz_pow_ui(aux,psi,delta);
		mpz_mul(beta,*b->Leading2(),aux);
		mpz_mul_si(beta,beta,-1);
		e=a;
		a=b;
		b=c;
		c=e;
		k++;
	}
	mpz_clear(alpha);
	mpz_clear(beta);
	Dispose(c);
	Dispose(d);
	if(b->degree<0){
		Dispose(a);
		Dispose(b);
		mpz_clear(aux);
		mpz_clear(psi);
		mpz_set_ui(aResult,0);
		return 0;
	} else {
		if (a->GetDegree()==1){
			mpz_set(aResult,b->coeff[0]);
		} else {
			mpz_neg(aux,b->coeff[0]);
			mpz_pow_ui(aux,aux,a->degree);
			mpz_pow_ui(psi,psi,a->degree-1);
			mpz_divexact(psi,aux,psi);
			if (1&(a->degree+k+1)) mpz_neg(psi,psi);
			mpz_set(aResult,psi);
		}
		Dispose(a);
		Dispose(b);
		mpz_clear(aux);
		mpz_clear(psi);
		return(mpz_get_si(aResult));
	}
}

/*!
	Returns the discriminant of the polynomial.

	Return codes:
	<table border=0>
	<tr><td>discriminant		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

long Polynomial::Discriminant(mpz_t aResult){
	Polynomial *d;
	d = AssignAuxPoly();
	int ret=d->Allocate(this->degree-1);
	if (ret!=ConstRC::Ok) return ret;
	ret=this->Derivate(d,this);
	if (ret!=ConstRC::Ok) return ret;
	this->Resultant(aResult,this,d);
	mpz_divexact(aResult,aResult,*this->Leading2());
	if((this->degree%4)>1) mpz_neg(aResult,aResult);
	Dispose(d);
	return mpz_get_si(aResult);
}

/*!
	Returns the formal derivative of the polynomial.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/
int Polynomial::Derivate(Polynomial *aTarget, Polynomial *aOperand){
	int i;
	int ret=aTarget->Allocate(aOperand->degree-1);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->degree=aOperand->degree-1;
	for(i=0;i<aOperand->degree;i++){
		mpz_mul_ui(aTarget->coeff[i],aOperand->coeff[i+1],i+1);
	}
	return ConstRC::Ok;
}

/*!
	Returns the definite integral of the polynomial on the interval [aLowerLimit, aUpperLimit]

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	</table>
*/
int Polynomial::DefiniteIntegral(mpz_t aResult, Polynomial *aOperand, mpz_t aLowerLimit, mpz_t aUpperLimit)
{
	mpz_set_ui(aResult, 0);
	int degree = aOperand->GetDegree();

	if (degree < 0)
	{
		return ConstRC::GeneralError;
	}
	
	mpz_t powerLower;
	mpz_t powerUpper;
	mpz_t aux;
	mpz_init(powerLower);
	mpz_init(powerUpper);
	mpz_init(aux);

	mpz_pow_ui(powerLower, aLowerLimit, degree + 1);
	mpz_pow_ui(powerUpper, aUpperLimit, degree + 1);

	for (int i = degree; i >= 0; i--)
	{
		if (mpz_cmp_ui(*aOperand->Get2(i), 0) != 0)
		{
			mpz_sub(aux, powerUpper, powerLower);
			mpz_mul(aux, aux, *aOperand->Get2(i));
			mpz_fdiv_q_ui(aux, aux, i + 1);
			mpz_add(aResult, aResult, aux);
		}

		mpz_divexact(powerLower, powerLower, aLowerLimit);
		mpz_divexact(powerUpper, powerUpper, aUpperLimit);
	}

	mpz_clear(aux);
	mpz_clear(powerLower);
	mpz_clear(powerUpper);

	return ConstRC::Ok;
}
/*!
	Recalculates all coefficients modulo p so that they are from 0 to p-1

	Return code: ConstRC::Ok
*/

int Polynomial::Modulo(unsigned long p){
	int i;
	for(i=0;i<=this->degree;i++){
		mpz_set_ui(this->coeff[i],mpz_fdiv_ui(this->coeff[i],p));
	}
	this->Reduce();
	return ConstRC::Ok;
}

/*!
	Recalculates all coefficients modulo p so that they are from 0 to aCharacteristic-1

	Return code: ConstRC::Ok
*/

int Polynomial::Modulo(mpz_t aCharacteristic){
	int i;
	for(i=0;i<=this->degree;i++){
		mpz_fdiv_r(this->coeff[i],this->coeff[i],aCharacteristic);
	}
	this->Reduce();
	return ConstRC::Ok;
}

/*!
	Recalculates all coefficients modulo p so that they are from (-p+1)/2 to p/2

	Return code: ConstRC::Ok
*/

int Polynomial::SymmetricModulo(unsigned long p){
	int i;
	unsigned long j;
	for(i=0;i<=this->degree;i++){
		j=mpz_fdiv_ui(this->coeff[i],p);
		if((j<<1)>p) j=j-p;
		mpz_set_si(this->coeff[i],j);
	}
	this->Reduce();
	return ConstRC::Ok;
}

/*!
	Recalculates all coefficients modulo p so that they are from (-aCharacteristic+1)/2 to aCharacteristic/2.

	Return code: ConstRC::Ok
*/

int Polynomial::SymmetricModulo(mpz_t aCharacteristic){
	int i;
	mpz_t aux;
	mpz_init(aux);
	mpz_fdiv_q_ui(aux,aCharacteristic,2);
	for(i=0;i<=this->degree;i++){
		mpz_fdiv_r(this->coeff[i],this->coeff[i],aCharacteristic);
		if (mpz_cmp(this->coeff[i],aux)>0) mpz_sub(this->coeff[i],this->coeff[i],aCharacteristic);
	}
	this->Reduce();
	mpz_clear(aux);
	return ConstRC::Ok;
}

/*!
	Quickly recomputes coefficients modulo p so that they are no longer from (-p+1)/2 to p/2
	but from 0 to p-1

	Return code: ConstRC::Ok
*/

int Polynomial::SymToModulo(unsigned long p){
	int i;
	for(i=0;i<=this->degree;i++){
		if(mpz_cmp_ui(coeff[i],0)<0) mpz_add_ui(coeff[i],coeff[i],p);
	}
	return ConstRC::Ok;
}

/*!
	Evaluates the value polynomial in aOperand
*/

long Polynomial::Evaluate(mpz_t aResult, mpz_t aOperand){
	int i;
	mpz_set_ui(aResult,0);
	Reduce();
	if (this->degree<0) return 0;
	for(i=this->degree;i>0;i--){
		mpz_add(aResult,aResult,this->coeff[i]);
		mpz_mul(aResult,aResult,aOperand);
	}
	mpz_add(aResult,aResult,this->coeff[i]);
	return(mpz_get_si(aResult));
}

/*!
	Evaluates the non-integer value polynomial in aOperand
*/

double Polynomial::EvaluateDecimal(mpf_t aResult, mpf_t aOperand)
{
	int i;
	mpf_t tmp;
	mpf_init (tmp);
	mpf_set_ui(aResult,0);
	Reduce();

	if (this->degree<0) 
	{
		return 0;
	}

	for(i=this->degree;i>0;i--)
	{
		mpf_set_z(tmp, this->coeff[i]);
		mpf_add(aResult,aResult,tmp);
		mpf_mul(aResult,aResult,aOperand);
	}
	mpf_set_z(tmp, this->coeff[i]);
	mpf_add(aResult,aResult,tmp);

	return(mpf_get_d(aResult));
}

/*!
	Evaluates the value polynomial in aOperand in the field <I>GF</I>(aPrime)
*/

long Polynomial::EvaluateModPrime(long aOperand, long aPrime){
	int i;
	INT64 Result=0;
	INT64 aux;
	Reduce();
	if (this->degree<0) return 0;
	for(i=this->degree;i>0;i--){
		aux=Result+mpz_fdiv_ui(this->coeff[i],aPrime);
		Result = aux % aPrime;
		aux = Result * aOperand;
		Result = aux % aPrime;
	}
	aux = Result + mpz_fdiv_ui(this->coeff[i],aPrime);
	Result = aux % aPrime;
	return(Result);
}

/*!
	Finds all roots of a polynomial in the field <I>GF</I>(<I>p</I>)
	and puts them in the array aRoots (which is supposed to be sufficiently allocated).
	The polynomial has to be a square free product of linear factors.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::RootsModPrimeSpecial(long *aRoots,unsigned long p){
// the polynomial is supposed to be a square free product of linear factors, the array aRoots is supposed to be allocated to a sufficient size
   Reduce();
   if (degree<=0) return ConstRC::Ok;
   if (degree==1){
	   aRoots[0]=p-mpz_fdiv_ui(coeff[0],p);
	   if ((unsigned long)aRoots[0]==p) aRoots[0]=0;
	   return ConstRC::Ok;
   }
   if (((long)degree<<3) > (signed long) p){
	   unsigned long i,j,r;
	   mpz_t aux,R;
	   mpz_init(aux);
	   mpz_init(R);
	   j=0;
	   for(i=0;i<p;i++){
	   mpz_set_ui(aux,i);
	   Evaluate(R,aux);
	   r=mpz_fdiv_ui(R,p);
	   if (!r) aRoots[j++]=i;
	   }
	   mpz_clear(R);
	   mpz_clear(aux);
	   return ConstRC::Ok;
   }
   if (degree==2){
	   mpz_t D,aux;
	   mpz_init(D);
	   mpz_mul(D,coeff[1],coeff[1]);
	   mpz_init(aux);
	   mpz_mul(aux,coeff[0],coeff[2]);
	   mpz_mul_ui(aux,aux,4);
	   mpz_sub(D,D,aux);
	   unsigned long sD=CIntegerClass::SquareRootModPrime(mpz_fdiv_ui(D,p),p);
	   long a1=CIntegerClass::InverseModN(-2*mpz_get_si(coeff[2]),p);
	   mpz_add_ui(aux,coeff[1],sD);
	   mpz_mul_si(aux,aux,a1);
	   aRoots[0]=(mpz_fdiv_ui(aux,p));
	   mpz_sub_ui(aux,coeff[1],sD);
	   mpz_mul_si(aux,aux,a1);
	   aRoots[1]=(mpz_fdiv_ui(aux,p));
	   mpz_clear(D);
	   mpz_clear(aux);
	   return ConstRC::Ok;
   }
   Polynomial *v,*w,*u;
   v = AssignAuxPoly();
   w = AssignAuxPoly();
   u = AssignAuxPoly();
   int ret=v->Set(1,1);
   if (ret!=ConstRC::Ok) return ret;
   while((u->degree<=0)||(u->degree==this->degree)){
	v->Set(0,CIntegerClass::RandomNumber(p));
	ret=v->PowerModPolyPrime(w,v,(p-1)>>1,this,p);
	if (ret!=ConstRC::Ok) return ret;
	w->Dec(0,1);
	ret=this->GCD_p(u,this,w,p);
	if (ret!=ConstRC::Ok) return ret;
   }
   ret=this->Divide(w,v,this,u);
   if (ret!=ConstRC::Ok) return ret;
   u->SymmetricModulo(p);
   ret=u->RootsModPrimeSpecial(aRoots,p);
   if (ret!=ConstRC::Ok) return ret;
   w->SymmetricModulo(p);
   ret=w->RootsModPrimeSpecial(aRoots+u->degree,p);
   Dispose(v);
   Dispose(w);
   Dispose(u);
   return ret;
}

/*!
	Finds all roots of a general polynomial in the field <I>GF</I>(<I>p</I>)
	and puts them in the array aRoots (which is supposed to be sufficiently allocated).

	Return codes:
	<table border=0>
	<tr><td>number of the roots		</td><td>everything all right</td></tr>
	<tr><td>-1		</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/


long Polynomial::RootsModPrimeGeneral(long *aRoots,unsigned long p)
{
	//the aRoots array is supposed to be allocated to a sufficient size
	Polynomial *v, *w, *u;

	v = AssignAuxPoly();
	this->Copy(v);
	v->Modulo(p);

	if (v->degree == -1) return 0;
	if (v->Allocate(this->GetDegree()) != ConstRC::Ok) return -1;
	if (((unsigned long)degree << 3) > p)
	{
		unsigned long i, r;
		long j;
		j = 0;
		for (i = 0; i < p; i++)
		{
			r = v->EvaluateModPrime(i, p);
			if (!r) aRoots[j++] = i;
		}
		
		Dispose(v);
		return j;
	}

	w = AssignAuxPoly();
	u = AssignAuxPoly();

	if (u->Set(1, 1) != ConstRC::Ok) return -1;
	if (u->PowerModPolyPrime(w, u, p, v, p) != ConstRC::Ok) return -1;
	
	w->Dec(1, 1);

	if (GCD_p(u, v, w, p) != ConstRC::Ok) return -1;
	if (u->RootsModPrimeSpecial(aRoots, p) != ConstRC::Ok) return -1;
	
	long r = u->GetDegree();
	
	Dispose(w);
	Dispose(v);
	Dispose(u);
	
	return r;
}

/*!
	Tests if the polynomial is irreducible in <I>GF</I>(aPrime)

	Return codes:
	<table border=0>
	<tr><td>TRUE		</td><td>is irreducible</td></tr>
	<tr><td>FALSE		</td><td>is not irreducible</td></tr>
	<tr><td>-1		</td><td>there was not enough memory</td></tr>
	</table>
*/

int Polynomial::IsIrreducibleModPrime(long aPrime){
	PolynomialSplitting *S;
	int ret;
	S=new PolynomialSplitting();
	S->Characteristic=aPrime;
	Polynomial *P;
	P=AssignAuxPoly();
	this->Copy(P);
	P->Modulo(aPrime);
	if (S->Split(P)!=ConstRC::Ok) return -1;
	ret=((S->Length==1)&&(S->Exponents[0]==1));
	delete S;
	Dispose(P);
	return ret;
}

int Polynomial::ModPolynomial(Polynomial *aTarget, Polynomial *aModulo, long aPrime){
	Polynomial *aux;
	aux = AssignAuxPoly();
	int ret = Divide(aux,aTarget,this,aModulo);
	if (ret!=ConstRC::Ok) return ret;
	aTarget->Modulo(aPrime);
	Dispose(aux);
	return ConstRC::Ok;
}

/*!
	Finds the inverse polynomial in the field <B>Z</B><sub>p</sub>/aModulo<B>Z</B><sub>p</sub>[x].
	The result is set 0 if the polynomial have a common factor.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok	</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory	</td><td>there was not enough memory to allocate a polynomial</td></tr>
	</table>
*/

int Polynomial::InverseModPolynomial(Polynomial *aTarget, Polynomial *aModulo, long aPrime){
	Polynomial *a,*b;
	int ret;
	a=AssignAuxPoly();
	b=AssignAuxPoly();
	ret = GCD_Extended_p(a,aTarget,b,this,aModulo,aPrime);
	if (ret!=ConstRC::Ok) return ret;
	if (a->GetDegree()>0) aTarget->degree=-1;
	Dispose(a);
	Dispose(b);
	return ConstRC::Ok;
}

int Polynomial::InverseInGF(Polynomial *aTarget, Polynomial *aModulo, mpz_t aCharacteristic){
	Polynomial *a,*b;
	int ret;
	a=AssignAuxPoly();
	b=AssignAuxPoly();
	ret = GCD_Extended_p(a,aTarget,b,this,aModulo,aCharacteristic);
	if (ret!=ConstRC::Ok) return ret;
	if (a->GetDegree()>0) aTarget->degree=-1;
	Dispose(a);
	Dispose(b);
	return ConstRC::Ok;
}

int Polynomial::AbsoluteValue(mpz_t aResult){
	mpz_set_ui(aResult,0);
	if (this->GetDegree()==-1) return ConstRC::Ok;
	int i;
	mpz_t aux;
	mpz_init(aux);
	for (i=0;i<=this->degree;i++){
	mpz_abs(aux,this->coeff[i]);
	mpz_add(aResult,aResult,aux);
	}
	mpz_clear(aux);
	return ConstRC::Ok;
}

/*
	For a given value aValue, the method returns a monic polynomial of the degree aDegree
	having a root modulo aValue. The root is returned in aRoot. The method checks
	2^aBitRange candidate polynomials and chooses among them the one with the least
	sum of absolute values of coefficients.

	Return codes: ConstRC::Ok
*/

int Polynomial::GeneratePolynomial(mpz_t aRoot, mpz_t aValue, unsigned int aDegree, unsigned int aBitRange){
	mpz_t Range,To,m,pow,mx,val,aux,ab;
	int i;
	Polynomial *A;
	int good=0;
	mpz_init_set_ui(Range,0);
	if (aBitRange) mpz_setbit(Range,aBitRange-1);
	mpz_init(To);
	mpz_init(m);
	mpz_init(pow);
	mpz_init(mx);
	mpz_init(val);
	mpz_init(aux);
	mpz_init(ab);
	A = AssignAuxPoly();
	A->Set(aDegree,1);
	mpz_root(To,aValue,aDegree);
	mpz_sub(m,To,Range);
	this->degree=-1;
	if (aBitRange) mpz_add_ui(m,m,1);
	mpz_add(To,To,Range);
	for(;mpz_cmp(m,To)<=0;mpz_add_ui(m,m,1))if(!good){
//	gmp_printf("%Zd: ",m);
	mpz_pow_ui(pow,m,aDegree);
	mpz_sub(val,aValue,pow);
	mpz_set_ui(ab,1);
	mpz_fdiv_q_2exp(Range,m,1);
	for(i=aDegree-1;i>=0;i--){
		if (i){
		mpz_divexact(pow,pow,m);
		mpz_fdiv_qr(A->coeff[i],val,val,pow);
		if ((i<(signed)aDegree-1)&&(mpz_cmp(A->coeff[i],Range)>0)) {
			mpz_sub(A->coeff[i],A->coeff[i],m);
			mpz_add_ui(A->coeff[i+1],A->coeff[i+1],1);
			mpz_add_ui(ab,ab,1);
		}
		mpz_abs(aux,A->coeff[i]);
		mpz_add(ab,ab,aux);
		}
		else {
		mpz_set(A->coeff[i],val);
		if (mpz_cmp(A->coeff[i],Range)>0) {
			mpz_sub(A->coeff[i],A->coeff[i],m);
			mpz_add_ui(A->coeff[1],A->coeff[1],1);
			mpz_add_ui(ab,ab,1);
		}
		mpz_abs(aux,A->coeff[i]);
		mpz_add(ab,ab,aux);
		}
		if((this->degree>0)&&(mpz_cmp(ab,mx)>=0)) {
		good=i*aDegree>>1;
		break;
		}
	}
//	A->PrintToScreen();
//	gmp_printf("  %Zd\n",val);
	if(((this->degree==-1)||(mpz_cmp(ab,mx)<0))){
		mpz_set(mx,ab);
		A->Copy(this);
		mpz_set(aRoot,m);
	}
	} else good--;
//    puts("Vitezem je:");
//    this->PrintToScreen();
//    printf(" s bitovou delkou %ld\n",mpz_sizeinbase(mx,2));
	mpz_clear(To);
	mpz_clear(m);
	mpz_clear(pow);
	mpz_clear(mx);
	mpz_clear(val);
	mpz_clear(ab);
	mpz_clear(aux);
	mpz_clear(Range);
	Dispose(A);
	return ConstRC::Ok;
}

void Polynomial::Horner(mpf_t aResult,mpf_t *aPolynomial, int aDegree, mpf_t aAt){
	mpf_t aux;
	int i;
	mpf_init(aux);
	mpf_set(aResult,aPolynomial[aDegree]);
	for(i=aDegree-1;i>=0;i--){
	mpf_mul(aux,aResult,aAt);
	mpf_add(aResult,aux,aPolynomial[i]);
	}
	mpf_clear(aux);
}

int Polynomial::RealRootsReal(double *aRoots, mpf_t *aPolynomial, int aDegree){
	if (aDegree==1) {
	aRoots[0]=-mpf_get_d(aPolynomial[0]);
	return 1;
	}
	if (aDegree==2) {
	mpf_t D,aux;
	mpf_init(D);
	mpf_init(aux);
	mpf_mul(D,aPolynomial[1],aPolynomial[1]);
	mpf_mul(aux,aPolynomial[0],aPolynomial[2]);
	mpf_mul_ui(aux,aux,4);
	mpf_sub(D,D,aux);
	if (mpf_cmp_ui(D,0)<0) {
		mpf_clear(D);
		mpf_clear(aux);
		return 0;
	}
	mpf_sqrt(D,D);
	mpf_sub(aux,D,aPolynomial[1]);
	mpf_div(aux,aux,aPolynomial[2]);
	aRoots[0]=mpf_get_d(aux)/2;
	mpf_add(aux,D,aPolynomial[1]);
	mpf_div(aux,aux,aPolynomial[2]);
	aRoots[1]=mpf_get_d(aux)/(-2);
	mpf_clear(D);
	mpf_clear(aux);
	return 2;
	}
	mpf_t* deriv = (mpf_t*) malloc(aDegree*sizeof(mpf_t));
	int i, r = 0;
	for(i=0;i<aDegree;i++){
	mpf_init(deriv[i]);
	mpf_mul_ui(deriv[i],aPolynomial[i+1],i+1);
	}
	int again=FALSE;
	mpf_t x,dx,y,dy,o;
	mpf_init(x);
	mpf_init(dx);
	mpf_init(y);
	mpf_init(dy);
	mpf_init(o);
	mpf_abs(x,aPolynomial[0]);
	mpf_set(o,x);
	for(again=0;again<2;again++){
	mpf_set_ui(dx,1);
	while ((mpf_cmp_d(dx,0.000001)>0)&&(mpf_cmp(x,o)<=0)) {
	Horner(y,aPolynomial,aDegree,x);
	Horner(dy,deriv,aDegree-1,x);
	mpf_div(dx,y,dy);
	if(mpf_cmp_d(dx,0)>0) mpf_set(o,x);
	mpf_sub(x,x,dx);
	mpf_abs(dx,dx);
//	printf("%f :: %f\n %f ",mpf_get_d(y),mpf_get_d(o),mpf_get_d(x));
	} if (mpf_cmp_d(dx,0.000001)<=0){
	for(i=aDegree;i>0;i--){
	mpf_set(deriv[i-1],aPolynomial[i]);
//	printf("/// %f",mpf_get_d(aPolynomial[i]));
	mpf_mul(y,x,aPolynomial[i]);
	mpf_add(aPolynomial[i-1],aPolynomial[i-1],y);
	}
	aRoots[0]=mpf_get_d(x);
	r=RealRootsReal(aRoots+1,deriv,aDegree-1);
	again=2;
	} else if (!again) {
	mpf_abs(x,aPolynomial[0]);
	mpf_neg(x,x);
	} else r=-1;
	}
	mpf_clear(x);
	mpf_clear(y);
	mpf_clear(dx);
	mpf_clear(dy);
	mpf_clear(o);
	for(i=0;i<aDegree;i++) mpf_clear(deriv[i]);
	free(deriv);
	deriv = NULL;
	return r+1;
}

int Polynomial::RealRoots(double *aRoots){
	mpf_t *polyn;
	int i,r;
	polyn=(mpf_t*) malloc((degree+1)*sizeof(mpf_t));
	mpf_set_default_prec(60);
	for(i=0;i<=degree;i++){
	mpf_init(polyn[i]);
	mpf_set_z(polyn[i],coeff[i]);
	}
	r=RealRootsReal(aRoots,polyn,degree);
	for(i=0;i<=degree;i++) mpf_clear(polyn[i]);
	free(polyn);
	return r;
}



int Polynomial::EuclideNormSqr(mpz_t aResult, Polynomial * aOperand)
{
	if (aOperand != NULL)
	{
		mpz_set_ui(aResult, 0);
		
		int max = aOperand->degree;
		for (int i = 0; i <= max; i++)
		{
			mpz_addmul(aResult, aOperand->coeff[i], aOperand->coeff[i]);
		}
		
		return ConstRC::Ok;
	}
	else
	{
		return ConstRC::GeneralError;		
	}
}

int Polynomial::InnerProduct(mpz_t aResult, Polynomial * aOperand1, Polynomial * aOperand2)
{
	if (aOperand1 != NULL && aOperand2 != NULL && aOperand1->degree == aOperand2->degree)
	{
		mpz_set_ui(aResult, 0);
		
		int max = aOperand1->degree;
		for (int i = 0; i <= max; i++)
		{
			mpz_addmul(aResult, aOperand1->coeff[i], aOperand2->coeff[i]);
		}
		
		return ConstRC::Ok;
	}
	else
	{
		return ConstRC::GeneralError;		
	}
}

bool Polynomial::ControlCoeffSize(mpz_t aMaxSize)
{
	for (int i = 0; i <= this->degree; i++)
	{
		if (mpz_cmpabs(this->coeff[i], aMaxSize) > 0)
		{
			return false;
		}
	}

	return true;
}

int Polynomial::GetMaxCoeffSize(mpz_t aResult)
{
	mpz_set_ui(aResult, 0);
	if (GetDegree() == -1) return ConstRC::Ok;

	for (int i = 0; i <= this->degree; i++)
	{
		if (mpz_cmpabs(aResult, this->coeff[i]) < 0)
		{
			mpz_abs(aResult, this->coeff[i]);
		}
	}

	return ConstRC::Ok;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////


PolynomialSplitting::PolynomialSplitting(){
	Length=0;
	Characteristic=0;
	Factors=NULL;
	Exponents=NULL;
	allocated=0;
}

PolynomialSplitting::~PolynomialSplitting(){
	unsigned int i;
	if(allocated)
	{
		for(i=0;i<allocated;i++)
		if (Factors[i]) delete(Factors[i]);
	free(Factors);
	free(Exponents);
	}
}

/*!
	Allocates the object to contain aNumber of polynomials

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory		</td><td>there was not enough memory</td></tr>
	</table>
*/


int PolynomialSplitting::Allocate(unsigned int aNumber){
	if (aNumber <= allocated) return ConstRC::Ok;
	Polynomial **F=NULL;
	unsigned int *E=NULL, i;
	F=(Polynomial **)malloc(aNumber*sizeof(Polynomial*));
	if (!F) return ConstRC::NotEnoughMemory;
	E=(unsigned int*)malloc(aNumber*sizeof(unsigned int));
	if (!E) return ConstRC::NotEnoughMemory;
	for(i=0;i<allocated;i++){
	F[i]=Factors[i];
	E[i]=Exponents[i];
	}
	for(;i<aNumber;i++){
	F[i]=NULL;
	}
	free(Factors);
	free(Exponents);
	Factors=F;
	Exponents=E;
	allocated=aNumber;
	return ConstRC::Ok;
}

/*!
	Prints the splitting to the screen
*/

int PolynomialSplitting::PrintToScreen()
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 500
	CLog::Log(500, "0900002", "");
	unsigned int i;
	for(i=0;i<Length;i++){
	CLog::Log(500, "", "(");
	Factors[i]->PrintToScreen();
	CLog::Log(500, "", ")");
	if(Exponents[i]>1) CLog::Log(500, "", "^%d",(int)Exponents[i]);
	}
	CLog::Log(500, "", "\n");
#endif
	return ConstRC::Ok;
}

/*!
	The main method - splits aOperand to factors over the given finite field.
*/

int PolynomialSplitting::Split(Polynomial *aOperand){
	int i,ret;
	Polynomial **F=NULL;
	if (aOperand->GetDegree()<=0){
	ret=Allocate(1);
	if (ret!=ConstRC::Ok) return ret;
	if (Factors[0]) free(Factors[0]);
	Factors[0]=aOperand->Copy();
	Exponents[0]=1;
	Length=1;
	return ConstRC::Ok;
	}
	ret=Allocate(aOperand->GetDegree());
	if (ret!=ConstRC::Ok) return ret;
	F=(Polynomial **)malloc((aOperand->GetDegree()+1)*sizeof(Polynomial*));
	if (!F) return ConstRC::NotEnoughMemory;
	for(i=0;i<aOperand->GetDegree();i++) F[i]=NULL;
	ret=this->SplitSquareFree(F,aOperand,1);
	if (ret!=ConstRC::Ok) return ret;
	for(i=0;i<aOperand->GetDegree();i++){
	if (F[i]){
		ret=this->SplitDistantDegree(F[i],i+1);
		if (ret!=ConstRC::Ok) return ret;
		delete(F[i]);
	}
	}
	free(F);
	return ConstRC::Ok;
}

/*!
	Splits the polynomial such that each factor is of different exponent
	in the polynomial, more precisely, the i-th polynomial
	is the multiple of all irreducible factors of exponent i*aPower.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory		</td><td>there was not enough memory</td></tr>
	</table>
*/

int PolynomialSplitting::SplitSquareFree(Polynomial **aTarget,Polynomial *aOperand,unsigned int aPower){
	Polynomial *D,*W,*Y,*C,*aux;
	int i,ret;
	D=new Polynomial(aOperand->GetDegree()-1);
	W=new Polynomial();
	Y=new Polynomial();
	aux=new Polynomial();
	C=aOperand->Copy();
	C->Modulo(Characteristic);
	ret=D->Derivate(D,C);
	if (ret!=ConstRC::Ok) return ret;
	if (D->GetDegree()>=0){
	D->Modulo(Characteristic);
	ret=D->GCD_p(C,aOperand,D,Characteristic);
	if (ret!=ConstRC::Ok) return ret;
	ret=D->Divide(W,aux,aOperand,C);
	if (ret!=ConstRC::Ok) return ret;
	W->Modulo(Characteristic);
//	W->PrintToScreen();puts("");
	for (i=aPower-1;W->GetDegree()>0;i+=aPower){
		ret=W->GCD_p(Y,W,C,Characteristic);
		if (ret!=ConstRC::Ok) return ret;
		W->Modulo(Characteristic);
		ret=W->Divide(D,aux,W,Y);
		if (ret!=ConstRC::Ok) return ret;
		D->Modulo(Characteristic);
//	D->PrintToScreen();puts("");
		if(D->GetDegree()>0) aTarget[i]=D->Copy();
		ret=Y->Copy(W);
		if (ret!=ConstRC::Ok) return ret;
		ret=D->Divide(D,aux,C,Y);
		if (ret!=ConstRC::Ok) return ret;
		D->Modulo(Characteristic);
		ret=D->Copy(C);
		if (ret!=ConstRC::Ok) return ret;
//	W->PrintToScreen();puts("");
	}
	if (C->GetDegree()>0) {
		D->SetDegree(-1);
		ret=D->Allocate(C->GetDegree()/Characteristic);
		if (ret!=ConstRC::Ok) return ret;
		for(i=0;(signed long)(i*Characteristic)<=C->GetDegree();i++)
		D->Set(i,C->Get(i*Characteristic));
		ret=this->SplitSquareFree(aTarget,D,aPower*Characteristic);
		if (ret!=ConstRC::Ok) return ret;
	}
	} else {
	D->SetDegree(-1);
	D->Allocate(C->GetDegree()/Characteristic);
	for(i=0;(signed long)(i*Characteristic)<=C->GetDegree();i++)
		D->Set(i,C->Get(i*Characteristic));
	ret=this->SplitSquareFree(aTarget,D,aPower*Characteristic);
	if (ret!=ConstRC::Ok) return ret;
	}
	delete(D);
	delete(W);
	delete(Y);
	delete(C);
	delete(aux);
	return ConstRC::Ok;
}

/*!
	Splits the polynomial to factors that are multiples of irreducible
	polynomials of the same degree.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory		</td><td>there was not enough memory</td></tr>
	</table>
*/

int PolynomialSplitting::SplitDistantDegree(Polynomial *aOperand, unsigned int aPower){
	Polynomial *W,*aux,*A,*R,*aux2;
	int i,ret;
	W = new Polynomial(Characteristic);
	R = new Polynomial();
	aux2 = new Polynomial();
	aux = new Polynomial(1);
	ret=aux->Set(1,1);
	if (ret!=ConstRC::Ok) return ret;
	A = aOperand->Copy();
	for(i=1;i*2<=A->GetDegree();i++){
	ret=W->PowerModPolyPrime(W,aux,Characteristic,A,Characteristic);
	if (ret!=ConstRC::Ok) return ret;
	W->Modulo(Characteristic);
	ret=W->Copy(aux);
	if (ret!=ConstRC::Ok) return ret;
	aux->Dec(1,1);
	ret=W->GCD_p(R,A,aux,Characteristic);
	if (ret!=ConstRC::Ok) return ret;
	if (R->GetDegree()>0) {
		ret=A->Divide(aux,aux2,A,R);
		if (ret!=ConstRC::Ok) return ret;
		ret=aux->Copy(A);
		if (ret!=ConstRC::Ok) return ret;
		A->Modulo(Characteristic);
		ret=A->Divide(aux2,aux,W,A);
		if (ret!=ConstRC::Ok) return ret;
		ret=this->SplitDD(R,i,aPower,0);
		if (ret!=ConstRC::Ok) return ret;
	} else {
		aux->Inc(1,1);
	}
	}
	if(A->GetDegree()>0) ret=this->SplitDD(A,A->GetDegree(),aPower,0);
	delete(aux);
	delete(aux2);
	delete(R);
	delete(W);
	delete(A);
	return ret;
}

/*!
	Splits a polynomial which is supposed to be a multiple of irreducible
	polynomials of the same degree aDegree and of the same exponent aPower
	to the irreducible factors.

	Return codes:
	<table border=0>
	<tr><td>ConstRC::Ok		</td><td>everything all right</td></tr>
	<tr><td>ConstRC::NotEnoughMemory		</td><td>there was not enough memory</td></tr>
	</table>
*/

int PolynomialSplitting::SplitDD(Polynomial *aOperand, int aDegree, unsigned int aPower,unsigned int aTried){
//	aOperand->PrintToScreen();puts("  :A");
	if (aOperand->GetDegree()==aDegree){
	Factors[this->Length]=aOperand->Copy();
	Factors[this->Length]->SymmetricModulo(Characteristic);
	Exponents[this->Length++]=aPower;
	return ConstRC::Ok;
	}
	Polynomial *g,*v,*w,*aux=NULL,*aux2=NULL;
	int i,j,ret;
	long q=Characteristic;
	if (Characteristic>2){
		for(i=1;i<aDegree;i++) q*=Characteristic;
	q--;
	q/=2;
	} else {
	q=aOperand->GetDegree()-1;
	aux=new Polynomial((2*aDegree-1)*q);
	}
	v = new Polynomial(2*aDegree-1);
	w = new Polynomial((2*aDegree-1));
	g = new Polynomial(aOperand->GetDegree());
	ret=v->Set(1,1);
	if (ret!=ConstRC::Ok) return ret;
	v->Set(0,aTried);
	while (g->GetDegree()<=0){
	v->Inc(0,1);
	for(j=0;v->Get(j)==(signed long)Characteristic;j++){
		v->Set(j,(long)0);
		if(j+1<v->GetDegree()) v->Inc(j+1,1);
		else {
		v->Set(j+1,(long)0);
		v->Set(j+2,1);
		}
	}
//	v->PrintToScreen();puts("  :v");
	if (Characteristic>2) w->PowerModPolyPrime(w,v,q,aOperand,Characteristic);
	else {
		aux2=v->Copy();
		ret=v->Copy(w);
		if (ret!=ConstRC::Ok) return ret;
		for(i=0;i<q;i++){
		ret=aux->Multiply(aux,w,w);
		if (ret!=ConstRC::Ok) return ret;
		ret=aux->Divide(aux2,w,aux,aOperand);
		if (ret!=ConstRC::Ok) return ret;
		w->Modulo(2);
		}
	}	
	w->Dec(0,1);
	w->Modulo(Characteristic);
//	w->PrintToScreen();puts("  :w");
	ret=w->GCD_p(g,aOperand,w,Characteristic);
	if (ret!=ConstRC::Ok) return ret;
//	w->PrintToScreen();puts("  :g\n");
	if(g->GetDegree()==aOperand->GetDegree())
		g->SetDegree(-1);
	}
	j=v->Get(0);
	ret=this->SplitDD(g,aDegree,aPower,j);
	if (ret!=ConstRC::Ok) return ret;
	ret=g->Divide(v,w,aOperand,g);
	if (ret!=ConstRC::Ok) return ret;
	v->Modulo(Characteristic);
	ret=this->SplitDD(v,aDegree,aPower,j);
	if (ret!=ConstRC::Ok) return ret;
	if (Characteristic==2){
	delete(aux);
	delete(aux2);
	}
	delete(g);
	delete(v);
	delete(w);
	return ConstRC::Ok;
}

void Polynomial::Finish() {
	if (Z_E_R_O_allocated) {
	mpz_clear(Z_E_R_O);
	int i;
	for (i=0; i<aux_pols; i++) {
		delete AuxPoly[i];
		AuxPoly[i] = NULL;
	}
	Z_E_R_O_allocated = FALSE;
	}
}

int Polynomial::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int Polynomial::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_POLYNOMIAL_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"Polynomial","Serialize",rc,0,"Could not write the main tag.");
		return ConstRC::GeneralError;
	}
	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	
	rc = CommonXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_DEGREE_NAME, this->degree);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_POLY_COEFFICIENTS_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"Polynomial","Serialize",rc,0,"Could not write the tag.");
		return ConstRC::GeneralError;
	}
	
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	for (int i = this->degree; i >= 0; i--)
	{
		rc = CommonXMLHelper::SerializePolynomialCoefficient(aWriter, this->coeff[i], i);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_POLY_COEFFICIENTS_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"Polynomial","Serialize",rc,0,"Could not end the tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"Polynomial","Serialize",rc,0,"Could not end the main tag.");
		return ConstRC::GeneralError;
	}
	rc = CommonXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;   
}

int Polynomial::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int Polynomial::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_POLYNOMIAL_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"Polynomial","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"Polynomial","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CommonXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_DEGREE_NAME))
		{
			int auxDegree;
			rc = CommonXMLHelper::DeserializeSignedInt(aReader, "Deserialize poly degree...", auxDegree, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}

			SetDegree(auxDegree);
			Zeroize();

			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_POLY_COEFFICIENTS_NAME))
		{
			int elementsDepth = xmlTextReaderDepth(aReader);
			int power;
			mpz_t auxMpz;
			mpz_init(auxMpz);
			for (int i = 0; i <= this->degree; i++)
			{
				if (CommonXMLHelper::MoveToNextNode(aReader, elementsDepth) != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					mpz_clear(auxMpz);
					break;
				}
				rc = CommonXMLHelper::DeserializePolynomialCoefficient(aReader, "Deserialize poly coefficient...", auxMpz, power, FALSE);
				if (rc != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					mpz_clear(auxMpz);
					break;
				}
				Set(power, auxMpz);
			}
			mpz_clear(auxMpz);
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
string Polynomial::ToString(poly_str_formats aFormat)
{
	string str;
	char * temp;

	switch (aFormat)
	{
		case EShortFormat:
			// (x)xxxx(x)xx(x)xxxxx(x)xxx - (degree)coefficient
			for (int i = this->degree; i > -1; i--)
			{
				if (mpz_cmp_si(this->coeff[i], 0) != 0)
				{
					stringstream stream;
					str.append(POLY_DEGREE_SEPARATOR_LEFT);
					stream << i;
					str.append(stream.str());
					str.append(POLY_DEGREE_SEPARATOR_RIGHT);
					temp = mpz_get_str(NULL, 10, this->coeff[i]);
					str.append(temp);
					free(temp);
				}
			}
			break;
		case ELongFormat:
		default:
			// xxx|xxxx|xxx|xx|xxx
			for (int i = this->degree; i > -1; i--)
			{
				temp = mpz_get_str(NULL, 10, this->coeff[i]);
				str.append(temp);
				free(temp);
				if (i > 0)
				{
					str.append(POLY_SEPARATOR);
				}
			}
			break;
	}

	return str;
}
Polynomial * Polynomial::Parse(const string aPolyString)
{
	size_t length = aPolyString.length();
	if (length == 0)
	{
		return NULL;
	}

	Polynomial * poly = NULL;
	mpz_t coeff;
	mpz_init(coeff);
	size_t indexStart = 0;
	size_t indexEnd;
	string temp;
	int deg = 0;

	if (aPolyString.find(POLY_DEGREE_SEPARATOR_LEFT) != aPolyString.npos)
	{
		// Short format
		poly = new Polynomial();
		indexStart = 1;
		do
		{
			indexEnd = aPolyString.find(POLY_DEGREE_SEPARATOR_RIGHT, indexStart);
			temp = aPolyString.substr(indexStart, indexEnd - indexStart);
			deg = atoi(temp.c_str());
			indexStart = indexEnd + 1;
			indexEnd = aPolyString.find(POLY_DEGREE_SEPARATOR_LEFT, indexStart);
			if (indexEnd == aPolyString.npos)
			{
				indexEnd = length;
			}
			temp = aPolyString.substr(indexStart, indexEnd - indexStart);
			mpz_set_str(coeff, temp.c_str(), 10);
			if (poly->Set(deg, coeff) != ConstRC::Ok)
			{
				delete poly;
				poly = NULL;
				break;
			}
			indexStart = indexEnd + 1;
		}
		while (indexStart < length);	
	}
	else
	{
		// Long format
		indexStart = -1;
		while ((indexStart = aPolyString.find(POLY_SEPARATOR, indexStart + 1)) != aPolyString.npos)
		{
			deg++;
		}
		poly = new Polynomial(deg);
		indexStart = 0;
		do 
		{
			indexEnd = aPolyString.find(POLY_SEPARATOR, indexStart);
			if (indexEnd == aPolyString.npos) 
			{
				indexEnd = length;
			}
			temp = aPolyString.substr(indexStart, indexEnd - indexStart);
			mpz_set_str(coeff, temp.c_str(), 10);
			if (poly->Set(deg, coeff) != ConstRC::Ok)
			{
				delete poly;
				poly = NULL;
				break;
			}
			deg--;
			indexStart = indexEnd + 1;
		}
		while (deg > -1);
	}

	mpz_clear(coeff);

	return poly;
}
int Polynomial::Parse2(Polynomial * aPoly, const string aPolyString)
{
	if (aPoly == NULL)
	{
		return ConstRC::GeneralError;
	}

	size_t length = aPolyString.length();
	if (length == 0)
	{
		return ConstRC::GeneralError;
	}

	mpz_t coeff;
	mpz_init(coeff);
	size_t indexStart = 0;
	size_t indexEnd;
	string temp;
	int deg = 0;

	if (aPolyString.find(POLY_DEGREE_SEPARATOR_LEFT) != aPolyString.npos)
	{
		// Short format
		aPoly->SetDegree(-1);
		indexStart = 1;
		do
		{
			indexEnd = aPolyString.find(POLY_DEGREE_SEPARATOR_RIGHT, indexStart);
			temp = aPolyString.substr(indexStart, indexEnd - indexStart);
			deg = atoi(temp.c_str());
			indexStart = indexEnd + 1;
			indexEnd = aPolyString.find(POLY_DEGREE_SEPARATOR_LEFT, indexStart);
			if (indexEnd == aPolyString.npos)
			{
				indexEnd = length;
			}
			temp = aPolyString.substr(indexStart, indexEnd - indexStart);
			mpz_set_str(coeff, temp.c_str(), 10);
			if (aPoly->Set(deg, coeff) != ConstRC::Ok)
			{
				mpz_clear(coeff);
				return ConstRC::GeneralError;
			}
			indexStart = indexEnd + 1;
		}
		while (indexStart < length);	
	}
	else
	{
		// Long format
		indexStart = -1;
		while ((indexStart = aPolyString.find(POLY_SEPARATOR, indexStart + 1)) != aPolyString.npos)
		{
			deg++;
		}
		aPoly->SetDegree(deg);
		indexStart = 0;
		do 
		{
			indexEnd = aPolyString.find(POLY_SEPARATOR, indexStart);
			if (indexEnd == aPolyString.npos) 
			{
				indexEnd = length;
			}
			temp = aPolyString.substr(indexStart, indexEnd - indexStart);
			mpz_set_str(coeff, temp.c_str(), 10);
			if (aPoly->Set(deg, coeff) != ConstRC::Ok)
			{
				mpz_clear(coeff);
				return ConstRC::GeneralError;
			}
			deg--;
			indexStart = indexEnd + 1;
		}
		while (deg > -1);
	}

	mpz_clear(coeff);

	return ConstRC::Ok;
}
