#ifndef __POLYNOM_CLASS__
#define __POLYNOM_CLASS__

#include <gmp.h>
#include "mserializable.h"
#include "enums_common.h"


/*!
\brief	Polynomial class

	This class represents a polynomial on one variable with (arbitrarily large) integer coefficients.
*/


class Polynomial : public virtual MSerializable
{
public:
	Polynomial();
	Polynomial(int aDegree);
	Polynomial(const Polynomial& aOperand);
	virtual ~Polynomial();

	// De/Serialization methods
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

	std::string ToString(poly_str_formats aFormat);
	static int			Parse2(Polynomial * aPoly, const std::string aPolyString);
	static Polynomial * Parse(const std::string aPolyString);

//getters and setters
	int Set(int aPower,long aNumber);
	int Set(int aPower,mpz_t aNumber);
	long Get(int aPower);
	mpz_t *Get2(int aPower);
	int GetDegree();
	int SetDegree(int aDegree);
	long Leading();
	mpz_t *Leading2();
	long Content(mpz_t aContent);
	int Inc(int aPower,long aNumber);
	int Dec(int aPower,long aNumber);
	int Inc(int aPower,mpz_t aNumber);
	int Dec(int aPower,mpz_t aNumber);
	bool IsOne(void);

//aritmetical operations
	int Add(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int AddMultiply(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber);
	int Substract(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int Multiply(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int MultiplyByPowerOfX(int aPower);
	int MultiplyMod(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo);
	int Divide(Polynomial *aTarget, Polynomial *aRemainder, Polynomial *aOperand1, Polynomial *aOperand2);
	int Multiply(Polynomial *aTarget, Polynomial *aOperand, long aNumber);
	int Divide(Polynomial *aTarget, Polynomial *aOperand, long aNumber);
	int Multiply(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber);
	int Divide(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber);
	int Power(Polynomial *aTarget, Polynomial *aOperand, unsigned int x);
	int PowerMod(Polynomial *aTarget, Polynomial *aOperand, unsigned int x,Polynomial *aModulo);
	int PowerModPolyPrime(Polynomial *aTarget, Polynomial *aOperand, unsigned int x,Polynomial *aModulo,long aPrime);
	long Evaluate(mpz_t aResult,mpz_t aOperand);
	double EvaluateDecimal(mpf_t aResult,mpf_t aOperand);
	long EvaluateModPrime(long aOperand,long aPrime);
	int Modulo(unsigned long p);
	int SymmetricModulo(unsigned long p);
	int SymToModulo(unsigned long p);
	int InverseModPolynomial(Polynomial *aTarget, Polynomial *aModulo, long aPrime);

//arithmetic mod prime
	int AddModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, unsigned long aPrime);
	int AddModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime);
	int SubstractModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, unsigned long aPrime);
	int SubstractModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime);
	int MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, unsigned long aPrime);
	int MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime);
	int MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand, unsigned long aNumber, unsigned long aPrime);
	int MultiplyModPrime(Polynomial *aTarget, Polynomial *aOperand, mpz_t aNumber, mpz_t aPrime);
	int DivideModPrime(Polynomial *aTarget, Polynomial *aRemainder, Polynomial *aOperand1, Polynomial *aOperand2, mpz_t aPrime);

//arithmetic in a galois finite field
	int AddInGF(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo, mpz_t aCharacteristic);
	int SubstractInGF(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo, mpz_t aCharacteristic);
	int MultiplyInGF(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2, Polynomial *aModulo, mpz_t aCharacteristic);
	int InverseInGF(Polynomial *aTarget, Polynomial *aModulo, mpz_t aCharacteristic);
	int PowerInGF(Polynomial *aTarget, unsigned int x, Polynomial *aModulo, mpz_t aCharacteristic);
	int PowerInGF(Polynomial *aTarget, mpz_t x, Polynomial *aModulo, mpz_t aCharacteristic);
	int ReduceInGF(Polynomial *aModulo, mpz_t aCharacteristic);
	int Modulo(mpz_t aCharacteristic);
	int SymmetricModulo(mpz_t aCharacteristic);


//greatest common divisor
	int GCD_Euclid(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int GCD_Reduced(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int GCD_Subresultant(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int GCD_Content(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2);
	int GCD_p(Polynomial *aTarget, Polynomial *aOperand1, Polynomial *aOperand2,unsigned long p);
	int GCD_Extended_p(Polynomial *aTarget, Polynomial *aComb1, Polynomial *aComb2, Polynomial *aOperand1, Polynomial *aOperand2,unsigned long p);
	int GCD_Extended_p(Polynomial *aTarget, Polynomial *aComb1, Polynomial *aComb2, Polynomial *aOperand1, Polynomial *aOperand2,mpz_t p);

//resultants
	long Resultant(mpz_t aResult,Polynomial *aOperand1,Polynomial *aOperand2);
	long Discriminant(mpz_t aResult);
	int Derivate(Polynomial *aTarget, Polynomial *aOperand);
	int DefiniteIntegral(mpz_t aResult, Polynomial *aOperand, mpz_t aLowerLimit, mpz_t aUpperLimit);

//factoring
	int RootsModPrimeSpecial(long *aRoots, unsigned long p); // for squarefree products of linear factors only
	long RootsModPrimeGeneral(long *aRoots, unsigned long p); // for general polynomials -- returns # of roots or -1 if error
	int IsIrreducibleModPrime(long aPrime);
	int RealRoots(double *aRoots);
	
//polynomial generating
	int AbsoluteValue(mpz_t aResult);
	int GeneratePolynomial(mpz_t aRoot, mpz_t aValue, unsigned int aDegree, unsigned int aBitRange);
	
//others
	int Reduce();
	int Allocate(int aSize);
	int Zeroize(int aFrom=0);
	Polynomial *Copy() const;
	int Copy(Polynomial *aTarget);
	int PrintToScreen();

	bool Equals(const Polynomial& aOperand) const;

	static void Finish();

	Polynomial *AssignAuxPoly();
	int Dispose(Polynomial *aPolynomial);

	///Control if abs value of poly coeff is lower than aMaxSize
	bool ControlCoeffSize(mpz_t aMaxSize);
	///Return absolute maximal coeff size 
	int GetMaxCoeffSize(mpz_t aResult);

//helper methods for using class Polynomial as mpz_t vector
	static int EuclideNormSqr(mpz_t aResult, Polynomial * aOperand);
	static int InnerProduct(mpz_t aResult, Polynomial * aOperand1, Polynomial * aOperand2);


protected:
	int degree;///<		The degree of the polynomial.
	mpz_t *coeff;///<	The coefficients of the polynomial.
	int allocated;///<		The number of allocated coefficients.
	static mpz_t Z_E_R_O;///<		Static mupltiprecision integer with value 0. Serves as a reference when asking for a coefficient out of the range.
	static int count;
	static bool Z_E_R_O_allocated;///<	Is the Z_E_R_O variable allocated?
	static const int aux_pols=20;///<	The number of auxiliary polynomials.
	static Polynomial *AuxPoly[aux_pols];///<	Auxiliary polynomial for usage inside the methods.
	static bool AuxPolyOccupied[aux_pols];///<	An array of flags showing whether an auxiliary polynomial is being used.
	int AuxPolynomialNumber;///<		For an auxiliary polynomial it's its number in the array, otherwise it is -1.
	int ModPolynomial(Polynomial *aTarget, Polynomial *aModulo, long aPrime);

	int RealRootsReal(double *aRoots, mpf_t *aPolynomial, int aDegree);
	void Horner(mpf_t aResult,mpf_t *aPolynomial, int aDegree, mpf_t aAt);
private:
	void initAuxPolynomials();
};

/*!
\brief	Class of a split polynomial.

	Serves for splitting a polynomial under a <I>GF</I>(<I>p</I>) field.
*/


class PolynomialSplitting 
{
public:
	PolynomialSplitting();
	~PolynomialSplitting();

	int Split(Polynomial *aOperand);
	int Allocate(unsigned int aNumber);
	int PrintToScreen();
protected:
	int SplitSquareFree(Polynomial **aTarget,Polynomial *aOperand,unsigned int aPower);
	int SplitDistantDegree(Polynomial *aOperand,unsigned int aPower);
	int SplitDD(Polynomial *aOperand, int aDegree, unsigned int aPower,unsigned int aTried);
public:
	Polynomial **Factors;///<	The array of the factors of the given polynomial.
	unsigned long Characteristic;///<	Characteristic of the field.
	unsigned int *Exponents;///<	The exponents of the factors.
	unsigned int Length;///<	The length of the decomposition.
protected:
	unsigned int allocated;///<	The number of allocated factors.
};
#endif
