#include "receiver.h"
#include "abstract_sieve.h"
#include "abstract_message.h"
#include "connector_info.h"
#include "filesystem_connector.h"
#include <string.h>
#include <iostream>
#include "distributed_algorithm.h"
# include <typeinfo>   // For std::bad_cast

Receiver::Receiver(AbstractSerializator& aSerializator) : iSerializator(aSerializator)
{
	InitializeDistAlg(aSerializator);
	iKeepMessages = FALSE;
}

Receiver::Receiver(AbstractSerializator& aSerializator, const string& aMachineName) : iSerializator(aSerializator)
{
	SetMachineName(aMachineName);
	InitializeDistAlg(aSerializator);
	iKeepMessages = FALSE;
}

Receiver::Receiver(AbstractSerializator& aSerializator, const char* aMachineName) : iSerializator(aSerializator)
{
	SetMachineName(aMachineName);
	InitializeDistAlg(aSerializator);
	iKeepMessages = FALSE;
}

Receiver::~Receiver()
{
	
	int n = 0, size = iConnectors.size();
	while (n < size)
	{
		delete this->iConnectors[n];	
		n++;
	}
	size = iReceiveConnections.size();
	for (n=0; n< size; n++)
	{	
		delete iReceiveConnections[n];
	}
}

void			Receiver::SetMachineName(const char* aValue)
{
	if (!aValue)
	{
	iMachineName.assign("");
	} 
	else 
	{
	iMachineName.assign(aValue);
	}
}

void			Receiver::SetMachineName(const string& aValue)
{
	iMachineName.assign(aValue);
}

const string&		Receiver::GetMachineName()		const
{
	return iMachineName;
}

void Receiver::SetKeepMessages(bool aValue) {
	iKeepMessages = aValue;
	return;
}

bool Receiver::GetKeepMessages() const {
	return iKeepMessages;
}

void Receiver::InitializeDistAlg(AbstractSerializator &aSerializator) {

        // the case when aSerializator is an instance of DistributedAlgorithm
        try {
                iDistAlg = dynamic_cast<DistributedAlgorithm*>(&aSerializator);
        }
        catch (std::bad_cast e) {
		iDistAlg = NULL;
	}

	return;
}

int	Receiver::CheckInitConsistency()
{
	/*if (iSieveInstance == NULL)
	{
		printf("Receiver.CheckInitConsistency: No sieve instance available. Run aborted.\n");
		return NO_SIEVE_INSTANCE_REGISTERED;
	}*/
	if (Connections() <= 0)
	{
		printf("Receiver.CheckInitConsistency: No receive connections available. Run aborted.\n");
		return NO_CONNECTORS_REGISTERED;
	}
	if (iMachineName.size() == 0)
	{
		printf("Receiver.CheckInitConsistency: Machine name empty - illegal.\n");
		return IDENTIFICATION_NOT_SET;
	}
	return ConstRC::Ok;
}


void	Receiver::FetchNewMessages()
{
	int i, size = Connectors();
	int j, nmSize;
	AbstractConnector*	 	cn = NULL;
	vector<AbstractMessage*>* 	newMessages = NULL;
	AbstractMessage*		message = NULL;
//  	printf("Receiver::Fetch new messages. Altogether %d connectors\n",size);
	for (i=0; i< size; i++)
	{
		cn = GetConnector(i);
		if ((cn) && (cn->GetState() == KOk))
		{
//  			printf("Receiver::Fetch new messages[%d]: receive all\n",i);
			newMessages = cn->ReceiveAll();
			if (newMessages)
			{
				nmSize = newMessages->size();
				if (nmSize > 0)
				{
//  					cout << "Receiver::FetchNewMessages. Received " << nmSize << " messages.\n";
				}
				for (j=0; j < nmSize; j++)
				{
				message = newMessages->at(j);
				if (message)
					{
					ProcessReceivedMessage(message);
					}
				delete message;
				}
			} else
			{
//  				printf("No new messages on this connector\n");
			}
			delete newMessages;
		}
		
	}
}

void	Receiver::ProcessReceivedMessage(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return;
	}
	// Checks whether this message is a reply to some older message,
	// and removes that older message from the queue.
	unsigned int replyTo = aMessage->Get_ReplyTo();
	if (replyTo != REPLIES_TO_NOTHING)
	{
// 		printf("Receiver::ProcessReceivedMessage. This message replies to message with counter %d\n",replyTo);
		RemoveFromPending(aMessage->Get_SenderId(),replyTo);
	}
}


// This method transfers ownership to Receiver
// aInfo pointer is either added to the internal 
// vector of connections, or deleted if equal connection is already present
void	Receiver::RegisterReceiveConnection(ConnectorInfo& aInfo)
{
// 	printf("Register receive connection \n");
	int i, max = iReceiveConnections.size();		
	for (i=0; i< max; i++)
	{
		if (aInfo.IsEqual(iReceiveConnections.at(i)))
		{
			printf("Receiver.RegisterReceiveConnection. A connection was doubly defined. Will be present only once.\n");
			aInfo.Print(2);
			return;
		}
	}
	ConnectorInfo* localInstance = new ConnectorInfo(aInfo);
	// This connection has not been defined yet
	iReceiveConnections.push_back(localInstance);
// 	printf("Receive Connections now have %d elements\n",iReceiveConnections.size());	
}

/*
	Parses and registers string of connector infos.
	Method is used for parsing "connectors" value from xml configuration.
*/
int Receiver::ParseConnectors(const char *str, const char *separ) {
	char *s, *ptr;
	ConnectorInfo *ci;

	s = (char *)malloc((strlen(str)+1)*sizeof(char));
	strcpy(s, str);

	ptr = strtok(s, separ);
	while (ptr) {
		ci = ConnectorInfo::Parse(ptr);
		if (!ci) {
			free(s);
			return ConstRC::GeneralError;
		}
		RegisterReceiveConnection(*ci);
		delete ci;
		ptr = strtok(NULL, separ);
	}

	free(s);

	return ConstRC::Ok;
}

void Receiver::PrintConnectors(FILE *f) {
	for (unsigned i = 0; i < iReceiveConnections.size(); i++) {
		iReceiveConnections.at(i)->Print(4);
	}

	return;
}

void 	Receiver::CreateExtraConnectors()
{
	int i, max = iReceiveConnections.size();		
	for (i=0; i< max; i++)
	{
		ConnectorInfo* info = iReceiveConnections[i];
		AbstractConnector* conn = FindConnectorByURL(info->GetUrlString());	
		if (!conn)
		{
			// FALSE ... do not register into iReceiveConnections (it is already there)
			conn = AddConnector(info, FALSE);			
			if (!conn)
			{						
				string r2 = COULD_NOT_CREATE_CONNECTOR_FOR_URL;
				r2 += "'";
				r2 += info->GetUrlString();
				r2 += "'";
				info->SetFailureReason(r2);
			}
		}
	}
}

bool    Receiver::IsSupportedConnectorType(connector_type aType)
{
	switch (aType)
	{
	case EFileConnector:
		return TRUE;
	default:
		return FALSE;
	}
}

AbstractConnector*	Receiver::AddConnector(ConnectorInfo* aInfo, bool aRegister)
{	
	if (aInfo)
	{
		AbstractConnector* connector = NULL;
		string empty;
		switch (aInfo->GetConnectorType())
		{
		case EFileConnector:
			connector = new FilesystemConnector(aInfo->GetUrlString(),iMachineName);
// 			printf("Receiver::AddConnector: URL String is %s\n",aInfo->GetUrlString().c_str());
			break;
		default:
			break;
		}	
		if (connector)
		{
			iConnectors.push_back(connector);
			if (aRegister)
			{
				if (connector->SupportsReceive())
				{
					ConnectorInfo* copy = new ConnectorInfo(*aInfo);
					RegisterReceiveConnection(*copy);
				}
			}
		}

		return connector;
	}
	return NULL;
}

int	Receiver::Connectors() 			const
{	
	return iConnectors.size();
}

AbstractConnector*	Receiver::GetConnector(int aIndex)	const
{
	if ((aIndex >= 0) && (aIndex < Connectors()))
	{
		return iConnectors.at(aIndex);
	}
	return NULL;
}


AbstractConnector*	Receiver::FindConnectorByURL(const string& aURL) const
{
	int i, max = Connectors();
	for (i=0; i< max; i++)
	{
		AbstractConnector* conn = iConnectors.at(i);
		if (conn)
		{
			if (conn->Get_URLString() == aURL)
			{
				printf("Already found a connector for URL %s on index %d",aURL.c_str(),i);
				return conn;
			}
		}
	}
	return NULL;
}

int			Receiver::Connections() 		const
{
	return iReceiveConnections.size();
}

ConnectorInfo*		Receiver::GetConnectorInfo(int aIndex)	const
{
	if ((aIndex >= 0) && (aIndex < Connections()))
	{
		return iReceiveConnections.at(aIndex);
	}
	return NULL;
}

void	Receiver::RemoveFromPending_Connector(AbstractConnector* aConnector, const char* aTargetName, unsigned int aCounter)
{
	if ((!aConnector) || (!aTargetName))
	{
		return;
	}
	aConnector->DeleteFromPending(aTargetName,aCounter);	
}

void    Receiver::PrintInfoMessage(const char* aMessage, ostream& aTargetStream)
{
	aTargetStream << GetThisClassName() << ": Info: " << aMessage << '\n';
}

void    Receiver::PrintErrorMessage(const char* aMessage,ostream& aTargetStream)
{
	aTargetStream << GetThisClassName() << ": Error: " << aMessage << '\n';
}

const char*	Receiver::GetThisClassName()			const
{
	return "Receiver";
}

void	Receiver::RegisterDistAlg(DistributedAlgorithm* aInstance)
{
	if (aInstance)
	{
		printf("Receiver::RegisterDistAlg. Registering instance %p (name = %s).",aInstance, aInstance->AlgorithmName());
	}
	iDistAlg = aInstance;
}

void	Receiver::ExpungeCommunicationInfo(string& aSenderName)
{
	ExpungeCommunicationInfo(aSenderName.c_str());
}

void	Receiver::ExpungeCommunicationInfo(const char* aSenderName)
{
	printf("Receiver::ExpungeCommunicationInfo: not implemented yet\n");
}

bool	Receiver::CheckMessageSequenceConsistency(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
// 		printf("Receiver::CheckMessageSequenceConsistency:: no message, FALSE\n");
		return FALSE;
	}
	CommunicationInfo info;
	info.iSender  = aMessage->Get_SenderAddress();
	info.iCounter = aMessage->Get_Counter();
	unsigned int i, max = iCommunicationInfo.size();
	for (i=0; i< max; i++)
	{
		CommunicationInfo& read = iCommunicationInfo[i];
		if (read.iSender == info.iSender)
		{
//  			printf("Receiver::found previous communication history with sender %s\n",info.iSender.c_str());
			if (read.iCounter >= info.iCounter)
			{
				printf("Receiver::the last previous message had counter %d, and this has counter %d.\n",read.iCounter,info.iCounter);
 				return FALSE;
			}
		}
	}
	return TRUE;
}

void    Receiver::UpdateCommunicationInfo(AbstractMessage* aMessage)
{	if (!aMessage)
	{
		return;
	}
	CommunicationInfo info;
	info.iSender  = aMessage->Get_SenderId();
	info.iCounter = aMessage->Get_Counter();
	unsigned int i, max = iCommunicationInfo.size();
	bool found = FALSE;
	for (i=0; i< max; i++)
	{
		CommunicationInfo& read = iCommunicationInfo[i];
		if (read.iSender == info.iSender)
		{
			read.iCounter = info.iCounter;
			found = TRUE;
		}
	}
	if (!found)
	{
		iCommunicationInfo.push_back(info);
//  		printf("Receiver::Update Comm Info, new info created for sender %s\n",info.iSender.c_str());
	}
}


bool	Receiver::ProcessMessageByType(AbstractMessage* aMessage)
{
	if (!aMessage)
	{
		return FALSE;
	}
	// Some types of messages, like Ack, can be processed in
	// common code for Node and Center.
	switch (aMessage->Get_Type())
	{	
		case EAck:
// 			printf("Receiver::ProcessMessageByType. Ack message read.\n");
			RemoveFromPending(aMessage->Get_SenderId(),aMessage->Get_ReplyTo());
			aMessage->MarkAsFinished(GetKeepMessages());
			return TRUE;
		default:
			break;
	
	}
	return FALSE;
}

void Receiver::FormatTime(string& aTarget, unsigned int aSeconds)
{
	aTarget.clear();
	if (aSeconds >= 86400)
	{
		char days[25];
		memset(days,0,25);
		int what = aSeconds / 86400;
		if (what != 1)
		{
			sprintf(days,"%3d days ",what);
		} else
		{
			sprintf(days,"%3d day ",what);	
		}
		aTarget += days;
	}
	if (aSeconds >= 3600)
	{
		unsigned int secDay = aSeconds % 86400;
		char hrs[25];
		memset(hrs,0,25);
		int what = secDay / 3600;
		if (what != 1)
		{
			sprintf(hrs,"%2d hrs ",what);
		} else
		{
			sprintf(hrs,"%2d hr ",what);
		}
		aTarget += hrs;
	}
	if (aSeconds >= 60)
	{
		unsigned int secHr = aSeconds % 3600;
		char mins[25];
		memset(mins,0,25);
		int what = secHr / 60;
		if (what != 1)
		{
			sprintf(mins,"%2d mins ",secHr / 60);
		} else
		{
			sprintf(mins,"%2d min ",secHr / 60);
		}		
		aTarget += mins;
	}
	aSeconds %= 60;
	char secs[25];
	memset(secs,0,25);
	if (aSeconds != 1)
	{
		sprintf(secs,"%2d secs ",aSeconds);
	} else
	{
		sprintf(secs,"%2d sec ",aSeconds);
	}

	aTarget += secs;
}
