#ifndef __RECEIVER_H__
#define __RECEIVER_H__


#include "types_common.h"
#include <vector>
#include <string>
#include <iostream>

class ConnectorInfo;
class AbstractConnector;
class AbstractSerializator;
class DistributedAlgorithm;	 // forward declaration
class AbstractMessage;


using namespace std;

/*!
 *\brief sender-counter pair
 */
typedef struct {
	string 		iSender;
	unsigned int    iCounter;
} CommunicationInfo;

#define COULD_NOT_CREATE_CONNECTOR_FOR_URL "Could not create connector for URL "

/*!
 * \brief Receiving connector
 */
class Receiver	
{
public:	
	Receiver(AbstractSerializator& aSerializator);
	Receiver(AbstractSerializator& aSerializator, const char* aMachineName);
	Receiver(AbstractSerializator& aSerializator, const string& aMachineName);
	virtual ~Receiver();

	// This method transfers ownership to Receiver.
	// aInfo pointer is either added to the internal 
	// vector of connections, or deleted if equal connection is already
	// present
	void	RegisterReceiveConnection(ConnectorInfo& aInfo);

	int ParseConnectors(const char *str, const char *separ);

	void PrintConnectors(FILE *f);

	int			Connectors() 			const;
	AbstractConnector*	GetConnector(int aIndex)	const;
	AbstractConnector*	FindConnectorByURL(const string& aURL) const;

	int			Connections() 			const;
	ConnectorInfo*		GetConnectorInfo(int aIndex)	const;

	void			SetMachineName(const char* aValue);
	void			SetMachineName(const string& aValue);
	const string&		GetMachineName()		const;

	virtual const char*	GetThisClassName()			const;

	void    		RegisterDistAlg(DistributedAlgorithm* aInstance);

	virtual int		Run() = 0;

	void			SetKeepMessages(bool aValue);
	bool			GetKeepMessages() const;

protected:
	bool	IsSupportedConnectorType(connector_type aType);
	int	CheckInitConsistency();

	void    PrintInfoMessage(const char* aMessage, ostream& aTargetStream);
	void    PrintErrorMessage(const char* aMessage,ostream& aTargetStream);

	// Goes through all connectors and fetches new messages from the "Received" queue.
	void	FetchNewMessages();	
	virtual void	ProcessReceivedMessage(AbstractMessage* aMessage) = 0;

	void	ExpungeCommunicationInfo(string& aSenderName);
	void	ExpungeCommunicationInfo(const char* aSenderName);

	virtual bool	CheckMessageSequenceConsistency(AbstractMessage* aMessage);

	void    UpdateCommunicationInfo(AbstractMessage* aMessage);

	// returns whether the message has been processed or no.
	virtual bool	ProcessMessageByType(AbstractMessage* aMessage);

	/*!
		This method is invoked when the Receiver receives
		a reply to some message. All outgoing connectors 
		are asked to remove the original message from their
		pending list.
		
		This needs to be overloaded in both Node and Center,
		since their outgoing connectors are specific.

	*/
	virtual void	RemoveFromPending(const char* aTargetName, unsigned int aCounter) = 0;

	/*!
		An auxiliary method for RemoveFrom Pending, containing
		common functionality
	*/
	
	void	RemoveFromPending_Connector(AbstractConnector* aConnector, const char* aTargetName, unsigned int aCounter);

	/*!
		This method will check the iReceiveConnections vector and 
		create extra connectors for URLs that were found in this vector
		and not yet present in iConnectors.
	
		This is useful for manual -conninfo parameters for both Nodes and Centers.
	*/
	void 	CreateExtraConnectors();

	static	void FormatTime(string& aTarget, unsigned int aSeconds);

	// checks the case when aSerializator is an instance of DistributedAlgorithm
	void InitializeDistAlg(AbstractSerializator &aSerializator);

protected:	
	AbstractConnector*	AddConnector(ConnectorInfo* aInfo, bool aRegister = TRUE);

private: // Member variables
	vector<AbstractConnector*>	iConnectors; ///< An array of connectors used by in distributed sieving. 

	vector<ConnectorInfo*>		iReceiveConnections; ///< An array of receive connections	

	vector<CommunicationInfo>	iCommunicationInfo;

	string				iMachineName;

	bool				iKeepMessages;

protected: 
	// Member variables
	DistributedAlgorithm*	   iDistAlg;  ///< This pointer is not owned by the node.

	AbstractSerializator& iSerializator;
};

#endif
