#include "sparse_matrix_class.h"
#include "common_utils.h" // because of allocation machine
#include <time.h>
#include <cstring>
#include "const_return_codes.h"
#include "log.h"

/*
This all generally works only with matrices having at least two nonzero rows.

*/

long SparseMatrix::bits_in_s_m_t = 8*sizeof(sparse_matrix_type);

/*!
	The default constructor does not take any parameters and constructs an instance of a "generic" bit matrix,
	with unknown dimensions. The dimensions may be later set by appropriate setter methods.
	
*/
SparseMatrix::SparseMatrix() : MSerializable(FILE_NAME_SPARSE_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->maximal_row_index = -1;
	this->maximal_column_index = -1;
	this->matrix = NULL;
	this->species = SSparseMatrix;
}

/*!
The second constructor constructs an instance of a bit matrix with known dimensions. Beware that the parameters
	taken mean the actual number of rows and columns, and not their maximal indices; so, if we want to construct
	a matrix of dimensions 17x32, we call
	
		SparseMatrix* sm = new SparseMatrix(17,32);
		
	Now, we have an instance of a bit matrix; its member variables will be set to:
	
		sm->maximal_row_index = 16;
		
		sm->maximal_column_index = 31;
		
		sm->maximal_allocated_row_index = -1; // no initialization of the data array yet!
		
	Both of the constructors initialize the data array to NULL pointer, and a real allocation is performed later
	- at the time of need. This programming pattern is called lazy initialization, and helps to reduce runtime
	memory requirements.
	
*/
SparseMatrix::SparseMatrix(long aRows, long aColumns) : MSerializable(FILE_NAME_SPARSE_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->maximal_row_index = aRows - 1;
	this->maximal_column_index = aColumns - 1;
	this->matrix = NULL;
	this->species = SSparseMatrix;
}

/*!

The destructor performs "cleaning up", in this case deallocation of the data array. Its decisions to deallocate
	are based on maximal_allocated_row_index, so it is safe to call it twice. However, I do not see any reason to
	call destructor explicitely; just use the standard C++ pattern
	
		delete sm;
		
	which invConstRC::Okes the destructor implicitely.
	
*/
SparseMatrix::~SparseMatrix(){
	if (max_allocated_row_index >= 0){
		long i;
		for (i=0; i<=max_allocated_row_index; i++){
			free(this->matrix[i]);
			this->matrix[i] = NULL;
		}
		max_allocated_row_index = -1;
		free(this->matrix);
		this->matrix = NULL;
	}
}

SparseMatrix* SparseMatrix::clone() {
	SparseMatrix * new_matrix = new SparseMatrix(GetMaximalRowIndex()+1, GetMaximalColumnIndex()+1);
	return new_matrix;
}

/*!
	  This method takes care of allocation of the internal two-dimensional array of sparse_matrix_type, which contains the
	  matrix entries. It tests the need for allocation by evaluating
	  	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1))
	  condition.
	  
	  If there is decision to run the allocation job, also zeroizing of the result matrix takes place. There are 
	  (this->maximal_row_index+1) rows allocated (all rows), each of them with ALLOCATED(0) entries. The entry on index
	  0 is initialized as the row index, the entry on index 1 is initialized as 0.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - unsuccessful allocation of the rows and/or columns

*/
int SparseMatrix::Allocate(){
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){
		long i;
		long max;
		this->matrix = (sparse_matrix_type **) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(sparse_matrix_type*));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"SparseMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate sparse matrix rows");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->matrix,0,(this->maximal_row_index+1)*sizeof(sparse_matrix_type*));
		this->max_allocated_row_index = this->maximal_row_index;
		max = ALLOCATED(0);
		for (i=0; i<= maximal_row_index; i++){
			this->matrix[i] = (sparse_matrix_type *) Utils::AllocationMachine((void **) &(this->matrix[i]),max*sizeof(sparse_matrix_type));
			if (!this->matrix){
				CLog::LogError(WHEREARG,"SparseMatrixClass","error allocating columns",max,201,"There was not enough memory to allocate sparse matrix columns");
				return ConstRC::NotEnoughMemory;
			}
			memset(this->matrix[i],0,max*sizeof(sparse_matrix_type));
			this->matrix[i][0] = i;
			this->matrix[i][1] = 0;
		}
	}
	return ConstRC::Ok;
}

sparse_matrix_type SparseMatrix::BinarySearchRow(sparse_matrix_type aRow) {
	if (aRow > this->maximal_row_index)		return ConstRC::ExcessiveRowIndex;
	sparse_matrix_type i,k,p,s, max_index = this->max_allocated_row_index;
	p = 0;
	k = max_index;
	// binary search
	while ((k-p)>1){
		s = p+(k-p)/2;
		if (this->matrix[s][0] <= aRow){
			p = s;
		}
		else{
			k = s;
		}
	}
	if (this->matrix[p][0] == aRow){
		i = p;
	}
	else{
		i = k;
	}
	return i;	
}

sparse_matrix_type SparseMatrix::BinarySearchColumn(sparse_matrix_type aRowIndex, sparse_matrix_type aColumn) {
	if (aRowIndex > this->max_allocated_row_index)		return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)		return ConstRC::ExcessiveColumnIndex;
	sparse_matrix_type i,k,p,s, max_index = this->matrix[aRowIndex][1]+1;
	p = 2;
	k = max_index;
	while ((k-p)>1){
		s = p+(k-p)/2;
		if (this->matrix[aRowIndex][s] <= aColumn){
			p = s;
		}
		else{
			k = s;
		}
	}
	if (this->matrix[aRowIndex][p] == aColumn){
		i = p;
	}
	else{
		i = k;
	}
	return i;	
}

/*!

there are getters and setters for the values in the matrix itself. For historic reasons, the
	setters do not come with Set* prefix, but rather Put* prefix.
	
		int  PutOne(long aRow, long aColumn);		
		
		int  IsOne(long aRow, long aColumn);
		
	The first two methods indeed either put a 1 onto designed position, or test whether there is 1 on such position.
	
	The return codes for PutOne are
	
		ConstRC::Ok 			if everything went all right
		
		ConstRC::ExcessiveRowIndex	in case of aRow > maximal_row_index
		
		ConstRC::ExcessiveColumnIndex  in case of aColumn > maximal_column_index
		
		ConstRC::NotEnoughMemory	thrown by Allocate() call ensuring proper allocation status of the matrix
		
		ConstRC::NegativeIndex          if aRow or aColumn < 0.
		
	The methods check validity of indices.
	
	The second two methods	
	
		int  IsZero(long aRow, long aColumn);
		
		int  PutZero(long aRow, long aColumn);
		
	are not yet implemented and return
	
		ConstRC::NotSupported
		
	to let the user know about this. So far, there has been no need to implement these methods.
*/
int SparseMatrix::PutOne(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))		return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)		return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)	return ConstRC::ExcessiveColumnIndex;
	long i,j;
	this->Allocate();
	i=0;
	long l;
	long max_index = this->max_allocated_row_index;
	i = BinarySearchRow(aRow);
	// now, (*operand).matrix[i][0]>= row
   	// this if solves the case, when the row is already allocated
	if ((i <= max_index) && this->matrix[i][0] == aRow){
		j = BinarySearchColumn(i,aColumn);
		if ((j == this->matrix[i][1]+1) && (this->matrix[i][j] != aColumn)) { j++; }		
		/*j=2;
		while ((j < this->matrix[i][1]+2) && (this->matrix[i][j] < aColumn)){
			j++;
		}*/
	   	// this if solves the case, when there is already a 1 on denoted place
		if ((j< (this->matrix[i][1]+2)) && (this->matrix[i][j] == aColumn)){ // there already is a 1 on that position
			return ConstRC::Ok;
		}
		// need to reallocate the row
		if (ALLOCATED(this->matrix[i][1]+1) > ALLOCATED(this->matrix[i][1])){
				this->matrix[i] = (sparse_matrix_type*) Utils::AllocationMachine((void **) &this->matrix[i],ALLOCATED(this->matrix[i][1]+1)*sizeof(sparse_matrix_type));
				if (!this->matrix[i])	{
					return ConstRC::NotEnoughMemory;
				}
				for (l=ALLOCATED(this->matrix[i][1]); l<ALLOCATED(this->matrix[i][1]+1); l++){
					this->matrix[i][l]=0;
				}
		}

		for (long k = (this->matrix[i][1]+2); k>j; k--){
			this->matrix[i][k] = this->matrix[i][k-1];
		}
		this->matrix[i][j] = aColumn;
		this->matrix[i][1]++;
		return ConstRC::Ok;
	}
	// otherwise, we must insert a new row
	this->matrix = (sparse_matrix_type **) Utils::AllocationMachine((void **) &this->matrix,(++(this->max_allocated_row_index))*sizeof(sparse_matrix_type*));
	if (!this->matrix){
		return ConstRC::NotEnoughMemory;
	}
	this->matrix[this->max_allocated_row_index] = (sparse_matrix_type *) Utils::AllocationMachine((void **) &this->matrix[this->max_allocated_row_index],ALLOCATED(0)*sizeof(sparse_matrix_type));
	if (!this->matrix[this->max_allocated_row_index]){
		return ConstRC::NotEnoughMemory;
	}
	for (j=this->max_allocated_row_index; j>i; j--){
		this->matrix[j] = (sparse_matrix_type*) Utils::AllocationMachine((void **) &this->matrix[j], ALLOCATED(this->matrix[j-1][1])*sizeof(sparse_matrix_type));
		if (!this->matrix[j]){
			return ConstRC::NotEnoughMemory;
		}		memcpy(this->matrix[j],this->matrix[j-1],ALLOCATED(this->matrix[j-1][1])*sizeof(sparse_matrix_type));
	}
	memset(this->matrix[i],0,ALLOCATED(1));
	this->matrix[i][0] = aRow;
	this->matrix[i][1] = 1;
	this->matrix[i][2] = aColumn;
	return ConstRC::Ok;
}

int SparseMatrix::IsOne(long aRow, long aColumn){
	if (aRow > this->maximal_row_index)		return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index)	return ConstRC::GeneralError;
	long i,j;
	i = BinarySearchRow(aRow);
	if (this->matrix[i][0] != aRow) return FALSE;
	j = BinarySearchColumn(i,aColumn);
	if ((j < this->matrix[i][1]+2) && (this->matrix[i][j] == aColumn)) { return TRUE; }		
	/*for (j=2; j< (this->matrix[i][1]+2); j++){
		
		if (this->matrix[i][j] == aColumn){
			return TRUE;
		}
	}*/
	return FALSE;
}

/*!
	  This is a maintenance method. It checks the caller instance for occurrence of zero rows, removes
	  them from the array by shifting the nonzero rows upwards and then deallocates the extra memory space.
	  
	  It should probably be called at the end of each operation which outputs a SparseMatrix; so far, the
	  only use is at the end of the Transpose() method.
	  
	  Return codes:
	  
	  	ConstRC::Ok			if everything went all right
		
		ConstRC::NotEnoughMemory	if reallocation in time of shifting the rows has failed.

*/
int SparseMatrix::Reduce(){
// removes zero rows from matrix
	long i,pom;
	long first_zero_index=0;
	for (i=0;i<= this->max_allocated_row_index; i++)
	{
		if (this->matrix[i][1])
		{
			// has there been a zero row before?
			if (i!=first_zero_index)
			{
				this->matrix[first_zero_index] = (sparse_matrix_type*) Utils::AllocationMachine((void**) &this->matrix[first_zero_index], ALLOCATED(this->matrix[i][1])*sizeof(sparse_matrix_type));
				if (!this->matrix[first_zero_index]){
					return ConstRC::NotEnoughMemory;
				}
				// move the row upwards
				memcpy(this->matrix[first_zero_index],this->matrix[i],ALLOCATED(this->matrix[i][1])*sizeof(sparse_matrix_type));
			}
			first_zero_index++;
		}
	}
	pom = this->max_allocated_row_index;
	this->max_allocated_row_index = first_zero_index-1;
	for (i=first_zero_index; i<=pom; i++)
	{
		free((void *) this->matrix[i]);
		this->matrix[i]=NULL;
	}
	return ConstRC::Ok;
}

/*!
	  The method randomly fills the matrix so that it had about 1% of ones. It allocates the matrix
	  if it has not been allocated before.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - if everything went all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.

*/
int SparseMatrix::Randomize(){
	long i,j,k;
	i=this->Allocate();
	if (i!=ConstRC::Ok) return i;
	this->Zeroize();
	srand((int) clock());
	k=(this->maximal_row_index/12)*maximal_column_index;
	j=(int) ((k/(log((float) k)))*(rand()/(RAND_MAX+1.0)));
	k=(int) ((k/(log((float) k)))*(rand()/(RAND_MAX+1.0)));
	k=(j+k+1) / 2;
	for (i=0;i<k;i++)
	{
		PutOne((int) ((maximal_row_index+1)*(rand()/(RAND_MAX+1.0))),(int) ((maximal_column_index+1)*(rand()/(RAND_MAX+1.0))));
	}
	return ConstRC::Ok;
}

/*!
	  This method is not implemented properly and it returns
	  
	        ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::Zeroize(){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it does nothing.
*/
void SparseMatrix::PrintToScreen(){
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int  SparseMatrix::Copy(SparseMatrix* aSource) {
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	  
	        FALSE
		
	  code to indicate this.

*/
bool SparseMatrix::Equals(SparseMatrix* aMatrix) {
	return FALSE;
}

int SparseMatrix::PutZero(long aRow, long aColumn){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
	  
	  code to indicate this.

*/
int SparseMatrix::IsZero(long aRow, long aColumn){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::ZeroizeRow(long aRow){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	  
	        ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::ZeroizeRow(long* aRowList, long aListMaxIndex){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::SwapRows(long aRow1, long aRow2){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::AddRows(long aTarget, long aSource){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::ZeroizeColumn(long aColumn){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::ZeroizeColumn(long* aColumnList, long aListMaxIndex){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::SwapColumns(long aColumn1, long aColumn2){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::AddColumns(long aTarget, long aSource){
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::Add(SparseMatrix* aTarget, SparseMatrix* aOperand2) {
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
int SparseMatrix::IsZero() {
	return ConstRC::NotSupported;
}

/*!
	  This method is not implemented properly and it returns
	        
		ConstRC::NotSupported
		
	  code to indicate this.

*/
SparseMatrix* SparseMatrix::Add(SparseMatrix* aOperand2) {
	return NULL;
}

/*!
	  This method instantiates a new NormalMatrix instance of the same dimensions as the caller instance of SparseMatrix; then,
	  it allocates the internal matrix of the NormalMatrix instance and converts the matrix represented by caller SparseMatrix
	  instance into this freshly allocated NormalMatrix instance.
	  
	  Return values:
	  
	  	pointer to NormalMatrix* instance if everything went all right
		
		NULL if some of the allocations was unsuccessful.

*/
NormalMatrix* SparseMatrix::ToNormalMatrix(){
	NormalMatrix* aResult = new NormalMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j,max, row;
	for (i=0; i<=this->max_allocated_row_index; i++){
		row = this->matrix[i][0];
		max = this->matrix[i][1]+2;
		for (j=2; j< max; j++){
			aResult->PutOne(row,this->matrix[i][j]);
		}
	}
	return aResult;
}

/*!
	  This method instantiates a new BitMatrix instance of the same dimensions as the caller instance of SparseMatrix; then,
	  it allocates the internal matrix of the BitMatrix instance and converts the matrix represented by caller SparseMatrix
	  instance into this freshly allocated BitMatrix instance.
	  
	  Return values:
	  
	  	pointer to BitMatrix* instance if everything went all right
		
		NULL if some of the allocations was unsuccessful.

*/
BitMatrix* SparseMatrix::ToBitMatrix(){
	BitMatrix* aResult = new BitMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j,max, row;
	for (i=0; i<=this->max_allocated_row_index; i++){
		row = this->matrix[i][0];
		max = this->matrix[i][1]+2;
		for (j=2; j< max; j++){
			aResult->PutOne(row,this->matrix[i][j]);
		}
	}
	return aResult;
}

/*!
	  This method instantiates a new SparseMatrix instance of the same dimensions as the parameter aMatrix; then,
	  it allocates the internal matrix of the SparseMatrix instance and converts the matrix represented by aMatrix
	  instance into this freshly allocated SparseMatrix instance. The rule is that odd entries of aMatrix are turned to bit 1
	  and even entries are turned to bit 0.
	  
	  Return values:
	  
	  	pointer to SparseMatrix* instance if everything went all right
		
		NULL if some of the allocations was unsuccessful or some GetXY() failed.

*/
SparseMatrix* SparseMatrix::ToSparseMatrix(NormalMatrix* aMatrix){
	long max_r = aMatrix->GetMaximalRowIndex();
	long max_c = aMatrix->GetMaximalColumnIndex();
	SparseMatrix* aResult = new SparseMatrix(max_r+1,max_c+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j;
	integer_matrix_type read;

	for (i=0; i<=max_r; i++){
		for (j=0; j<=max_c; j++){
			read = aMatrix->GetXY(i,j);
			if (read == ConstRC::GeneralError){
				delete aResult;
				return NULL;
			}
			if (read % 2){
				aResult->PutOne(i,j);
			}
		}
	}
	return aResult;
}

/*!
	  This method instantiates a new SparseMatrix instance of the same dimensions as the parameter aMatrix; then,
	  it allocates the internal matrix of the SparseMatrix instance and converts the matrix represented by aMatrix
	  instance into this freshly allocated SparseMatrix instance.
	  
	  Return values:
	  
	  	pointer to SparseMatrix* instance if everything went all right
		
		NULL if some of the allocations was unsuccessful or some IsOne() failed.

*/
SparseMatrix* SparseMatrix::ToSparseMatrix(BitMatrix* aMatrix){
	long max_r = aMatrix->GetMaximalRowIndex();
	long max_c = aMatrix->GetMaximalColumnIndex();
	SparseMatrix* aResult = new SparseMatrix(max_r+1,max_c+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j;
	int read = 0;

	for (i=0; i<=max_r; i++){
		for (j=0; j<=max_c; j++){
			read = aMatrix->IsOne(i,j);
			if (read == (int) ConstRC::GeneralError){
				delete aResult;
				return NULL;
			}
			if (read == TRUE){
				aResult->PutOne(i,j);
			}
		}
	}
	return aResult;
}

/*!
	  This method ensures allocation state, dimension requirements etc., and then transposes the calling instance into
	  the matrix aTarget. The calling instance must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions of the calling instance and aTarget do not match (M x N vs. N x M)
		
		ConstRC::NullPointerSupplied - if aTarget is NULL
		
		ConstRC::BadArgument	  - if aTarget == this.

*/
int  SparseMatrix::Transpose(SparseMatrix* aTarget){
	if (!aTarget)  							return ConstRC::NullPointerSupplied;
        if (!aTarget)    
	if (aTarget->maximal_row_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	long i, j, max, row;
	for (i=0; i<= this->max_allocated_row_index; i++){
		max = this->matrix[i][1]+2;
		row = this->matrix[i][0];
		for (j=2; j< max; j++){
			aTarget->PutOne(this->matrix[i][j],row);
		}
	}
	aTarget->Reduce();
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the transposition operation, and then performs
	  the transposition by calling 
	  	int Transpose(SparseMatrix* aTarget)
	  If result has been allocated, but Transpose did not finish well, the result matrix is deleted again.
	  
	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)

*/
SparseMatrix* SparseMatrix::Transpose(){
	SparseMatrix* result = new SparseMatrix(this->maximal_column_index+1, this->maximal_row_index+1);
        if (!result) return NULL;
        if (this->Transpose(result) != ConstRC::Ok){
                delete result;
                return NULL;
        }
        return result;
}

/*!	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument	  - if aOperand2 == aTarget.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_big_multiply.		
*/
int	SparseMatrix::MultiplyInternalBig(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2){
        if (!aOperand1 || !aOperand2 || !aTarget)                               return ConstRC::NullPointerSupplied;
        if (aTarget == aOperand2)                       return ConstRC::BadArgument;
	long i, j, k;
	int ret;
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE	
	start = clock();
#endif	
	if (aOperand1->GetMaximalColumnIndex() != aOperand2->GetMaximalRowIndex())
	{										
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){	
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand1->max_allocated_row_index;
	sparse_matrix_type  row_index;		
	long max_block = aOperand2->GetMaxAllocatedColumnIndex();
	long max_sparse;
	for (i=0; i<=max_row; i++){
		row_index = aOperand1->matrix[i][0];
		max_sparse = aOperand1->matrix[i][1]+1;		
		for (j=2; j<= max_sparse; j++){
			for (k=0; k<=max_block; k++){
				aTarget->SetBlock(row_index,k,aTarget->GetBlock(row_index,k) ^ aOperand2->GetBlock(aOperand1->matrix[i][j],k));			
			}
		}		
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE		
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_big_multiply += (double) diff;
	}
#endif
	return ConstRC::Ok;
}
int	SparseMatrix::MultiplyInternalBig(NormalMatrix* aTarget, SparseMatrix* aOperand1, NormalMatrix* aOperand2){
        if (!aOperand1 || !aOperand2 || !aTarget)                               return ConstRC::NullPointerSupplied;
        if (aTarget == aOperand2)                       return ConstRC::BadArgument;
	long i, j, k;
	int ret;
	if (aOperand1->GetMaximalColumnIndex() != aOperand2->GetMaximalRowIndex())
	{										
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){	
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand1->max_allocated_row_index;
	long max_col = aOperand2->GetMaximalColumnIndex();
	long max_sparse, ind;
	for (i=0; i<=max_row; i++){
		max_sparse = aOperand1->matrix[i][1]+1;		
		for (j=2; j<= max_sparse; j++){
			ind = aOperand1->matrix[i][j];
			for (k=0; k<=max_col; k++){
				if (aOperand2->GetXY(ind,k))
				{
					aTarget->PutValue(i,k,aOperand2->GetXY(ind,k)+aTarget->GetXY(i,k));
				}			
			}
		}		
	}
	return ConstRC::Ok;
}


/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument	  - if aOperand2 == aTarget.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_big_multiply.		
*/
int	SparseMatrix::MultiplyInternalBig(BCMatrix* aTarget, SparseMatrix* aOperand1, BCMatrix* aOperand2){
        if (!aOperand1 || !aOperand2 || !aTarget)                               return ConstRC::NullPointerSupplied;
        if (aTarget == aOperand2)                       return ConstRC::BadArgument;
	long i, j;
	int ret;
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	start = clock();
#endif	
	if (aOperand1->GetMaximalColumnIndex() != aOperand2->GetMaximalRowIndex())
	{										
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){	
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBig",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand1->max_allocated_row_index;
	sparse_matrix_type  row_index;		
	long max_sparse;
	for (i=0; i<=max_row; i++){
		row_index = aOperand1->matrix[i][0];
		max_sparse = aOperand1->matrix[i][1]+1;		
		for (j=2; j<= max_sparse; j++){
			aTarget->AddToRow(row_index, aOperand2->GetRow(aOperand1->matrix[i][j]));			
		}		
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE		
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_big_multiply += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1^T x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument	  - if aOperand2 == aTarget.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_big_multiply.

*/
int	SparseMatrix::MultiplyInternalBigTransposed(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2){
        if (!aOperand1 || !aOperand2 || !aTarget)                               return ConstRC::NullPointerSupplied;
        if (aTarget == aOperand2)                       return ConstRC::BadArgument;
	long i, j, k;
	int ret;
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE	
	start = clock();
#endif
	if (aOperand1->GetMaximalRowIndex() != aOperand2->GetMaximalRowIndex())
	{										
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){	
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand1->max_allocated_row_index;
	sparse_matrix_type  row_index;		
	long max_block = aOperand2->GetMaxAllocatedColumnIndex();
	long max_sparse;
	for (i=0; i<=max_row; i++){
		row_index = aOperand1->matrix[i][0];
		max_sparse = aOperand1->matrix[i][1]+1;		
		for (j=2; j<= max_sparse; j++){
			for (k=0; k<=max_block; k++){
				aTarget->SetBlock(aOperand1->matrix[i][j],k,aTarget->GetBlock(aOperand1->matrix[i][j],k) ^ aOperand2->GetBlock(row_index,k));			
			}
		}		
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE	
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_big_multiply += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1^T x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument	  - if aOperand2 == aTarget.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_big_multiply.
*/
int	SparseMatrix::MultiplyInternalBigTransposed(BCMatrix* aTarget, SparseMatrix* aOperand1, BCMatrix* aOperand2){
        if (!aOperand1 || !aOperand2 || !aTarget)                               return ConstRC::NullPointerSupplied;
        if (aTarget == aOperand2)                       return ConstRC::BadArgument;
	long i, j;
	int ret;
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE	
	start = clock();
#endif
	if (aOperand1->GetMaximalRowIndex() != aOperand2->GetMaximalRowIndex())
	{										
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){	
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand1->max_allocated_row_index;
	sparse_matrix_type  row_index;		
	long max_sparse;
	for (i=0; i<=max_row; i++){
		row_index = aOperand1->matrix[i][0];
		max_sparse = aOperand1->matrix[i][1]+1;		
		for (j=2; j<= max_sparse; j++){
			aTarget->AddToRow(aOperand1->matrix[i][j], aOperand2->GetRow(row_index));			
		}		
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE		
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_big_multiply += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

int SparseMatrix::MultiplyInternalBigTransposed(NormalMatrix* aTarget, SparseMatrix* aOperand1, NormalMatrix* aOperand2)
{
        if (!aOperand1 || !aOperand2 || !aTarget) {
		printf("Null pointer supplied.\n");
	        return ConstRC::NullPointerSupplied;
	}
        if (aTarget == aOperand2) {
		printf("Bad argument supplied.\n");
                return ConstRC::BadArgument;
	}
	integer_matrix_type i, j, k, ind; // sum
	int ret;
	if (aOperand1->GetMaximalRowIndex() != aOperand2->GetMaximalRowIndex())
	{										
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){	
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
		CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigTransposed",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	integer_matrix_type max_row = (integer_matrix_type) aOperand1->GetMaximalRowIndex();
	// sparse_matrix_type  index_aux;
	// long max_block = aOperand2->GetMaximalColumnIndex();
	integer_matrix_type max_sparse;
	
	// For every row ...
	for (i=0; i<=max_row; i++){
		//printf("Row %d.\n");
		max_sparse = (integer_matrix_type) aOperand1->matrix[i][1]+1;				
		for (j=2; j <= max_sparse; j++)
		{
			ind = aOperand1->matrix[i][j];
			for (k=0; k<= max_row; k++)
			{
				aTarget->PutNumber(aTarget->GetXY(ind,k)+aOperand2->GetXY(i,k),ind,k);				
			}
		}				
	}
	return ConstRC::Ok;
}


/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x (aOperand2)^T, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  This method is very slow and implemented in a very suboptimal way, but it is not used in normal factorization jobs.
	  It can be only exploited in rank measurements etc. Feel free to mend it.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument	  - if aOperand2 == aTarget.

*/
int     SparseMatrix::MultiplyInternalBigWithTransposition(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2){
        if (!aOperand1 || !aOperand2)                                   return ConstRC::NullPointerSupplied;
        if (aTarget == aOperand2)                       return ConstRC::BadArgument;
	long i, j, k;
	int parity;
	int ret;
	if (!aTarget) {
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",0,1,"Target not allocated");
		return ConstRC::NullPointerSupplied;
	}
	if (aTarget->GetMaximalRowIndex() != aOperand1->GetMaximalRowIndex()){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",aTarget->GetMaximalRowIndex(),aOperand1->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	if (aTarget->GetMaximalColumnIndex() != aOperand2->GetMaximalRowIndex()){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",aTarget->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}

	if (aOperand1->GetMaximalColumnIndex() != aOperand2->GetMaximalColumnIndex())
	{			CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand2->GetMaxAllocatedRowIndex();
	long max_sparse;
	for (j=0; j<=max_row; j++)
	{
		for (i=0; i<=aOperand1->max_allocated_row_index; i++)
		{
			parity=0;
			max_sparse = aOperand1->matrix[i][1]+1;
			for (k=2; k<= max_sparse; k++)
			{
			// jestlize je 1 na (aOperand1->matrix[i][0],aOperand1->matrix[i][k]) souradnice
			//  a jestlize je 1 na (aOperand1->matrix[i,k],j);
				if (aOperand2->IsOne(j,aOperand1->matrix[i][k]))
				{
					parity++;
				}
			}
			/* zapiseme na prislusne misto bitove matice vysledku */
			if (parity%2)
			{
				aTarget->PutOne(aOperand1->matrix[i][0],j);
			}
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x (aOperand2)^T, saving the result into aTarget.
	  This method is very slow and implemented in a very suboptimal way, but it is not used in normal factorization jobs.
	  It can be only exploited in rank measurements etc. Feel free to mend it. In fact, this method and method 
	  	int     MultiplyInternalBigWithTransposition(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2)
	  are technically identical, except for the aPrintInfo parameter.
	  The aPrintInfo parameter determines whether runtime information about the multiplication will be output or no.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL	

*/
int     SparseMatrix::MultiplyInternalBigWithTransposition(BitMatrix* aTarget, SparseMatrix* aOperand1, SparseMatrix* aOperand2, bool aPrintInfo){
        if (!aOperand1 || !aOperand2)                                   return ConstRC::NullPointerSupplied;
	long i, j, k;
	int parity;
	int ret;
	if (!aTarget) {
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",0,2,"Target not allocated");
		return ConstRC::NullPointerSupplied;
	}
	if (aTarget->GetMaximalRowIndex() != aOperand1->GetMaximalRowIndex()){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",aTarget->GetMaximalRowIndex(),aOperand1->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	if (aTarget->GetMaximalColumnIndex() != aOperand2->GetMaximalRowIndex()){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",aTarget->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	if (aOperand1->GetMaximalColumnIndex() != aOperand2->GetMaximalColumnIndex())
	{			CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",aOperand1->GetMaximalColumnIndex(),aOperand2->GetMaximalRowIndex(),"Size mismatch");
		return ConstRC::SizeMismatch;
	}
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",ret,920,"Some problem with matrix allocation.");
		return ret;
	}
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok){
CLog::LogError(WHEREARG,"SparseMatrixClass","MultiplyInternalBigWithTrans",ret,820,"Some problem with matrix zerozing.");
		return ret;
	}
	long max_row = aOperand2->GetMaxAllocatedRowIndex();
	long max_sparse;
	if (aPrintInfo)	printf("\nBig multiplication, please wait ...\n");
	for (j=0; j<=max_row; j++)
	{
	    if (aPrintInfo){
		if (!(j%100)) { printf("\r");
		      printf("%5ld/%ld   ",j,max_row);
		      fflush(stdout);
		    }
	    }
		for (i=0; i<=aOperand1->max_allocated_row_index; i++)
		{
			parity=0;
			max_sparse = aOperand1->matrix[i][1]+1;
			for (k=2; k<= max_sparse; k++)
			{
				if (aOperand2->IsOne(j,aOperand1->matrix[i][k]))
				{
					parity++;
				}
			}
			/* zapiseme na prislusne misto bitove matice vysledku */
			if (parity%2)
			{
				aTarget->PutOne(aOperand1->matrix[i][0],j);
			}
		}
	}
	if (aPrintInfo)	printf("\nBig multiplication finished.\n");
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication aOperand1 x (aOperand2)^T by calling
	  	int MultiplyInternalBigWithTransposition(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2);
		
	  If result has been allocated, but MultiplyInternal... did not finish well, the result is deleted again.
	  
	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)			  
*/
BitMatrix*      SparseMatrix::MultiplyInternalBigWithTransposition(SparseMatrix* aOperand1, BitMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	BitMatrix* result = new BitMatrix(aOperand1->GetMaximalRowIndex()+1, aOperand2->GetMaximalRowIndex()+1);
        if (!result) return NULL;
        if (this->MultiplyInternalBigWithTransposition(result,aOperand1,aOperand2) != ConstRC::Ok) {
                delete result;
                result = NULL;
        }
        return result;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int	MultiplyInternalBig(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2)

	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)			  

*/
BitMatrix*	SparseMatrix::MultiplyInternalBig(SparseMatrix* aOperand1, BitMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	
	BitMatrix* result = new BitMatrix(aOperand1->GetMaximalRowIndex()+1, aOperand2->GetMaximalColumnIndex()+1);
        if (!result) return NULL;
        if (this->MultiplyInternalBig(result,aOperand1,aOperand2) != ConstRC::Ok) {
                delete result;
                result = NULL;
        }
        return result;
}

/*!	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int	MultiplyInternalBig(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2)
	  
	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)			  
*/
BCMatrix*	SparseMatrix::MultiplyInternalBig(SparseMatrix* aOperand1, BCMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	
	BCMatrix* result = new BCMatrix(aOperand1->GetMaximalRowIndex()+1, aOperand2->GetMaximalColumnIndex()+1);
	if (!result) return NULL; 
	if ((this->MultiplyInternalBig(result,aOperand1,aOperand2) != ConstRC::Ok)) {
		delete result;
		result = NULL;
	}
	return result;
}

/*!	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int	MultiplyInternalBigTransposed(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2)
	  
	  Returns:
	  	
		pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)			  
*/
BitMatrix*	SparseMatrix::MultiplyInternalBigTransposed(SparseMatrix* aOperand1, BitMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	
	BitMatrix* result = new BitMatrix(aOperand1->GetMaximalColumnIndex()+1, aOperand2->GetMaximalColumnIndex()+1);
	if ((!result)) return NULL;
	if ((this->MultiplyInternalBigTransposed(result,aOperand1,aOperand2) != ConstRC::Ok)) {	
		delete result; 
		result = NULL; 
	}
	 return result;
}

/*!
		
	  
	  	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int	MultiplyInternalBigTransposed(BCMatrix* aTarget, BCMatrix* aOperand1, BCMatrix* aOperand2)
	  
	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)			  



*/
BCMatrix*	SparseMatrix::MultiplyInternalBigTransposed(SparseMatrix* aOperand1, BCMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	
	BCMatrix* result = new BCMatrix(aOperand1->GetMaximalColumnIndex()+1, aOperand2->GetMaximalColumnIndex()+1);
	if ((this->MultiplyInternalBigTransposed(result,aOperand1,aOperand2) != ConstRC::Ok)) {	
		delete result; 
		result = NULL; 
	}
 return result;
}

long SparseMatrix::GetMaxAllocatedRowIndex()    const {
	return this->max_allocated_row_index;
}

/*!
	  This method performs internal multiplication of matrices the compiler think are abstract. The matrix
	  aOperand1 must be a SparseMatrix, then if aResult and aOperand2 are BitMatrices the method calls
		MultiplyInternalBig((BitMatrix*) aResult, (SparseMatrix*) aOperand1, (BitMatrix*) aOperand2)
	  and if aResult and aOperand are BCMatrices the method calls
		MultiplyInternalBig((BCMatrix*) aResult, (SparseMatrix*) aOperand1, (BCMatrix*) aOperand2)
	  in all other cases the method fails, because SparseMatrices cannot be treated like ordinary matrices,
	  they require special multiplication methods.
	  
	  Return codes:
	  
	  	ConstRC::GeneralError	  - if the parameters are not suitable
		
		any other code from the launched method

*/
int SparseMatrix::MultiplyInternal(AbstractMatrix* aResult, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2) {
        if ((aResult->GetSpecies()==SBitMatrix)&&(aOperand2->GetSpecies()==SBitMatrix)&&(aOperand1->GetSpecies()==SSparseMatrix))
                return MultiplyInternalBig((BitMatrix*) aResult, (SparseMatrix*) aOperand1, (BitMatrix*) aOperand2);
        if ((aResult->GetSpecies()==SBCMatrix)&&(aOperand2->GetSpecies()==SBCMatrix)&&(aOperand1->GetSpecies()==SSparseMatrix))
                return MultiplyInternalBig((BCMatrix*) aResult, (SparseMatrix*) aOperand1, (BCMatrix*) aOperand2);
	if ((aResult->GetSpecies()==SNormalMatrix)&&(aOperand2->GetSpecies()==SNormalMatrix)&&(aOperand1->GetSpecies()==SSparseMatrix))
                return MultiplyInternalBig((NormalMatrix*) aResult, (SparseMatrix*) aOperand1, (NormalMatrix*) aOperand2);
	
        return ConstRC::GeneralError;
}

/*!
	  This method performs internal multiplication of matrices the compiler think are abstract. The matrix
	  aOperand1 must be a SparseMatrix, then if aResult and aOperand2 are BitMatrices the method calls
		MultiplyInternalBigTransposed((BitMatrix*) aResult, (SparseMatrix*) aOperand1, (BitMatrix*) aOperand2)
	  and if aResult and aOperand are BCMatrices the method calls
		MultiplyInternalBigTransposed((BCMatrix*) aResult, (SparseMatrix*) aOperand1, (BCMatrix*) aOperand2)
	  in all other cases the method fails, because SparseMatrices cannot be treated like ordinary matrices,
	  they require special multiplication methods.
	  
	  Return codes:
	  
	  	ConstRC::GeneralError	  - if the parameters are not suitable
		
		any other code from the launched method

*/
int SparseMatrix::MultiplyInternalTransposed(AbstractMatrix* aResult, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2) {
        if ((aResult->GetSpecies()==SBitMatrix)&&(aOperand2->GetSpecies()==SBitMatrix)&&(aOperand1->GetSpecies()==SSparseMatrix))
                return MultiplyInternalBigTransposed((BitMatrix*) aResult, (SparseMatrix*) aOperand1, (BitMatrix*) aOperand2);
        if ((aResult->GetSpecies()==SBCMatrix)&&(aOperand2->GetSpecies()==SBCMatrix)&&(aOperand1->GetSpecies()==SSparseMatrix))
                return MultiplyInternalBigTransposed((BCMatrix*) aResult, (SparseMatrix*) aOperand1, (BCMatrix*) aOperand2);
	if ((aResult->GetSpecies()==SNormalMatrix) &&(aOperand1->GetSpecies()==SSparseMatrix)&&(aOperand2->GetSpecies()==SNormalMatrix))
                return MultiplyInternalBigTransposed((NormalMatrix*) aResult, (SparseMatrix*) aOperand1, (NormalMatrix*) aOperand2);
        return ConstRC::GeneralError;
}


int SparseMatrix::Save(char* aName)
{
        FILE *fw;
        long i;
        fw=fopen(aName,"w");
        if (!fw) return ConstRC::GeneralError;
        fwrite(&(this->maximal_row_index),1,sizeof(long),fw);
        fwrite(&(this->maximal_column_index),1,sizeof(long),fw);
        for(i=0;i<=maximal_row_index;i++){
                fwrite(this->matrix[i],this->matrix[i][1]+2,sizeof(matrix_type),fw);
        }
        fclose(fw);
	return ConstRC::Ok;
}

int SparseMatrix::Load(char* aName)
{
        FILE *fr;
        long i,j,row,max,col;
        if (this->matrix) return ConstRC::GeneralError;
        fr=fopen(aName,"r");
        if (!fr) return ConstRC::GeneralError;
        fread(&(this->maximal_row_index),1,sizeof(long),fr);
        fread(&(this->maximal_column_index),1,sizeof(long),fr);
        this->Allocate();
        for(i=0;i<=maximal_row_index;i++){
                fread(&row,1,sizeof(matrix_type),fr);
                fread(&max,1,sizeof(matrix_type),fr);
		for(j=0;j<max;j++){
        	        fread(&col,1,sizeof(matrix_type),fr);
			this->PutOne(row,col);
		}
        }
        fclose(fr);
        return ConstRC::Ok;
}

//int SparseMatrix::Serialize(const CBaseParameters &aParam) const
//{
//	return MSerializable::Serialize(aParam);
//}
//
//int SparseMatrix::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
//{
//	return AbstractMatrix::Serialize(aParam, aWriter);
//}
//
//int SparseMatrix::Deserialize(const CBaseParameters &aParam)
//{
//	return MSerializable::Deserialize(aParam);
//}
//
//int SparseMatrix::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
//{
//	return AbstractMatrix::Deserialize(aParam, aReader);
//}

int SparseMatrix::WriteData(xmlTextWriterPtr aWriter) const
{
	int rc;
	// write all matrix data
	for (long i = 0; i <= maximal_row_index; i++)
	{
		rc = CommonXMLHelper::SerializeBinaryData(aWriter, XML_SERIAL_MATRIX_ROW_NAME, this->matrix[i], (this->matrix[i][1]+2), sizeof(sparse_matrix_type));  
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

int SparseMatrix::ReadData(xmlTextReaderPtr aReader)
{
	int rc;
	xmlChar * name = NULL;

	// we know max_row_index and max_column_index
	Allocate();
	sparse_matrix_type * value = NULL;
	sparse_matrix_type * start = NULL;
	sparse_matrix_type row, max, col;
	long length, count, i, j;
	// we deserialize every row and get max nonzero columns in row and nonzero indexies
    for (i = 0; i <= this->maximal_row_index; i++)
	{
		if (NULL != name) {
			xmlFree(name);
			name = NULL;
		};
		name = xmlTextReaderName(aReader);

		if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MATRIX_ROW_NAME) == 0)
		{
			break;
		}

		value = NULL;
		rc = CommonXMLHelper::DeserializeBinaryDataOut(aReader, "Deserialize matrix row...", (void**)&value, count, length, FALSE);
        if (rc != ConstRC::Ok)
		{
			if (NULL != name)
				xmlFree(name);
			return rc;
		}
		start = value;

		row = *value;
		value++;
		max = *value;
		value++;

		for (j = 0; j < max; j++)
		{
			// put one to nonzero indexies
			col = *value;
			value++;
			PutOne(row,col);
		}

		free(start);
		value = NULL;
		start = NULL;

		rc = CommonXMLHelper::MoveToNextNode(aReader);
		if (rc != ConstRC::Ok)
		{
			if (NULL != name)
				xmlFree(name);
			if (rc == ConstRC::EndOfFile)
			{
				return ConstRC::Ok;
			}
			return rc;
		}
    }
	if (NULL != name)
		xmlFree(name);

	return ConstRC::Ok;
}
