#ifndef __SPARSE_MATRIX_CLASS__
#define __SPARSE_MATRIX_CLASS__

#include "bit_matrix_class.h"
#include "bc_matrix_class.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

typedef signed int sparse_matrix_type;

/*!
        \brief A class for memory-savvy representation of very sparse matrices over GF(2).
	
	This class is used for memory-savvy representation of very sparse matrices over GF(2). For this representation, it 
	uses a two-dimensional array of custom type sparse_matrix_type, which is in i386 version defined as signed int. 
	
	The aim of this class is to provide suitable representation of very sparse matrices over GF(2).
	This type is suitable for representation of very sparse bit matrices, say having at most 2 per cent of
	nonzero bits. For matrices which are denser, there is another type defined in this program - see the previous
	section.
	
	As for details of the representation: each row in the matrix has at least two, but generally more entries of 
	sparse_matrix_type:
		- the entry I on index zero gives index of the row in the matrix. That is because some rows may be zero
		  and we do not want them to take up memory space in the matrix.
		- the entry N on index one gives number of the following entries. If it is equal to 0, no entries are
		  available for this row
		- the entries on indices 2 to N+1 give indices of columns, where a nonzero bit is. 
	For example:
		let B be a matrix which has three rows:
		
		[1][3][0][6][7]
		
		[2][1][4]
		
		[5][5][0][2][9][11][12]
		
		This means that in row of index 1, there are 3 nonzero entries in columns indexed by 0, 6 and 7.
		In row of index 2, there is 1 nonzero entry in column 4
		In row of index 5, there are 5 nonzero entries in columns indexed by 0, 2, 9, 11 and 12.
		The rest of rows are zero. Dimensions of the matrix, as usual, are given by inherited protected
		member variables maximal_row_index and maximal_column_index.
		
	Allocation of a SparseMatrix is performed with use of macro ALLOCATED, defined in definitions.h by the
	following formula:
		#define SPARSE_REALLOCATION_STEP   10
		#define ALLOCATED(x) SPARSE_REALLOCATION_STEP*((((x)+2)/SPARSE_REALLOCATION_STEP)+1)
	This means that the number of allocated entries in a SparseMatrix row will be always multiple of 10.
	The actual number is derived from the value on index 1 in the row - so, from the value giving number
	of the following entries. From the formula we see that if a row contains 0 to 7 nonzero entries, the
	allocated size is 10; if it contains 8 to 17 entries, the allocated size is 20 etc. The ALLOCATED macro
	has been designed in this way for the two following reasons:
		- if we want to store X entries, we have to have space for X+2 entries (the number of the row
		  and the number of the following entries are an extra burden).
		- to be sure, the unused entries in the allocated blocks could be filled by -1 or similar
		  "illegal value", which would enable an easy test to prevent accidental running out of the row.
		  At least one -1 should be present at the end of the row.
		  
	So, to accomodate 7 entries, we in fact need 10 fields and that is why this approach has been chosen.
	
	The SPARSE_REALLOCATION_STEP is given by a symbolic constant, because in factorization jobs, another
	value may happen to be optimal (optimal = prevents frequent reallocation, while costing not too much
	memory).
	
	The structure of a SparseMatrix means that transposition is very complicated, a quadratic algorithm indeed.
	In block Lanczos method, one transposition at least must be performed; it is tolerable, but a general slowdown.
	A solution to this problem would be to construct both the sieving matrix and its transpose at the same time.
	If the runtime of the transposition turns out to be a significant problem, it will be done this way.

*/
class SparseMatrix : public AbstractMatrix {
public:
	SparseMatrix();
	SparseMatrix(long aRows, long aColumns);
	~SparseMatrix();

	SparseMatrix* clone();	

	int  Allocate();
	int  Randomize();
	int  Zeroize();
	int  Copy(SparseMatrix* aSource);
	bool Equals(SparseMatrix* aMatrix);
	void PrintToScreen();
	int  PutOne(long aRow, long aColumn);
	int  PutZero(long aRow, long aColumn);
	int  IsOne(long aRow, long aColumn);
	int  IsZero(long aRow, long aColumn);
	int  IsZero();
	// Swaps etc.
	int  ZeroizeRow(long aRow);
	int  ZeroizeRow(long* aRowList, long aListMaxIndex);
	int  SwapRows(long aRow1, long aRow2);
	int  AddRows(long aTarget, long aSource);
	int  AddToRow(long aRow, AbstractMatrix* aOperand, long aRow2);
	int  ZeroizeColumn(long aColumn);
	int  ZeroizeColumn(long* aColumnList, long aListMaxIndex);
	int  SwapColumns(long aColumn1, long aColumn2);
	int  AddColumns(long aTarget, long aSource);

	long GetMaxAllocatedRowIndex()    const;
	// Special function
	int  Reduce();
	// Mathematical operations
	int  	   Transpose(SparseMatrix* aTarget);
	SparseMatrix* Transpose();
	int  	   Add(SparseMatrix* aTarget, SparseMatrix* aOperand2);
	SparseMatrix* Add(SparseMatrix* aOperand2);
	int	MultiplyInternalBig(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2);
	int	MultiplyInternalBig(BCMatrix* aTarget, SparseMatrix* aOperand1, BCMatrix* aOperand2);
	int	MultiplyInternalBig(NormalMatrix* aTarget, SparseMatrix* aOperand1, NormalMatrix* aOperand2);
	int	MultiplyInternal(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	int	MultiplyInternalBigTransposed(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2);
	int	MultiplyInternalBigTransposed(BCMatrix* aTarget, SparseMatrix* aOperand1, BCMatrix* aOperand2);
	int	MultiplyInternalBigTransposed(NormalMatrix* aTarget, SparseMatrix* aOperand1, NormalMatrix* aOperand2);

	int	MultiplyInternalTransposed(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	int     MultiplyInternalBigWithTransposition(BitMatrix* aTarget, SparseMatrix* aOperand1, BitMatrix* aOperand2);
	int     MultiplyInternalBigWithTransposition(BitMatrix* aTarget, SparseMatrix* aOperand1, SparseMatrix* aOperand2, bool aPrintInfo);
	BitMatrix*      MultiplyInternalBigWithTransposition(SparseMatrix* aOperand1, BitMatrix* aOperand2);
	BitMatrix*	MultiplyInternalBig(SparseMatrix* aOperand1, BitMatrix* aOperand2);
	BCMatrix*	MultiplyInternalBig(SparseMatrix* aOperand1, BCMatrix* aOperand2);
	BitMatrix*	MultiplyInternalBigTransposed(SparseMatrix* aOperand1, BitMatrix* aOperand2);
	BCMatrix*	MultiplyInternalBigTransposed(SparseMatrix* aOperand1, BCMatrix* aOperand2);

	// Type conversions
	NormalMatrix*     ToNormalMatrix();
	static SparseMatrix* ToSparseMatrix(NormalMatrix* aMatrix);
	BitMatrix*        ToBitMatrix();
	static SparseMatrix* ToSparseMatrix(BitMatrix* aMatrix);
    int Save(char* aName);
    int Load(char* aName);

	// De/Serialization methods
	//int Serialize(const CBaseParameters& aParam) const;
	//int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	//int Deserialize(const CBaseParameters& aParam);
	//int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

protected:
	int WriteData(xmlTextWriterPtr aWriter) const;
	int ReadData(xmlTextReaderPtr aReader);

private:
	sparse_matrix_type BinarySearchRow(sparse_matrix_type aRow);
	sparse_matrix_type BinarySearchColumn(sparse_matrix_type aRowIndex, sparse_matrix_type aColumn);
protected:
	long max_allocated_row_index;
	sparse_matrix_type** matrix; ///< The double array that contains the matrix data. It is initialized by the constructor to a NULL pointer. 
	static long bits_in_s_m_t; ///< Common to all instances, reflects the number of bits in sparse_matrix_type.
};

#endif

