#include "squfof.h"

#include <cmath>
#include <iostream>
#include <cstdlib>
#include "const_return_codes.h"

#define CHECK_SQUFOF

SQUFOF::SQUFOF()
{
	mpz_init(checkprod);
	Set_AlgorithmType(EFASQUFOF);
}; 

SQUFOF::~SQUFOF() {
	mpz_clear(checkprod);
};

bool SQUFOF::sqrti(long & aTarget, long aSource){
	// 1 is a special case
	if (1 == aSource) {
		aTarget = 1;
		return true;
	};
	long upper = aSource;
	if (upper >= (1<< (sizeof(long)*4-1))) 
		upper = (1<< (sizeof(long)*4-1)) - 1;
	long lower = 0;	
	long root = (lower+upper)/2;
	long square = root*root;
	while (square != aSource && (upper - lower)>1)
	{
		if (square>=aSource) {
			upper = root;
		} else {
			lower = root;
		}
		root = (lower+upper)/2;
		square = root*root;
	}
	aTarget = root;
	//std::cerr<< "\n" << aTarget << "^2 = "<< aSource ;
	if (square == aSource) 
		return true;
	else
		return false;
};

// 
// following should hold for the result:
// result + aMB = 0 % 2*aC
long SQUFOF::Rfunction( long aMB, long aC,long aDiscsquare) {
	long r = -aMB;
	long absc = abs(aC);
	if (aDiscsquare < absc) {
		while (r <= -absc) r += 2*absc;	
		while (r > absc) r -= 2*absc;	
	} else {
		while (r <= aDiscsquare - 2*absc) r += 2*absc;
		while (r > aDiscsquare) r -= 2*absc;
	};

#ifdef CHECK_SQUFOF
	if ((r+aMB) % (2*aC) != 0) {
		//std::cerr << "\nReduce error!";
		//std::cerr << "MB:" << aMB <<" C:" <<aC <<" Dsq:"<<aDiscsquare;
		//std::cerr << " result:"<<r;
	};
#endif
	return r;
};

bool SQUFOF::Reduced(long aA, long aB, long aC, long aDiscsquare, long aDisc) {
	if ((abs(aDiscsquare-2*abs(aA))<aB) && (aB <aDiscsquare)) 
		return true;
	else
		return false;
};

int SQUFOF::Reduce(long & aA, long & aB, long & aC, long aDiscsquare, long aDisc) {
#ifdef CHECK_SQUFOF
	long backupA = aA;
	long backupB = aB;
	long backupC = aC;
#endif 
	long resA, resB, resC;
	resA = aC;
	resB = Rfunction(aB,aC,aDiscsquare);
	resC = resB*resB-aDisc;
	resC = resC / 4;
	resC = resC / aC;

	aA = resA;	
	aB = resB;	
	aC = resC;	

#ifdef CHECK_SQUFOF
	resC = aA;
	resB = Rfunction(aB,aA,aDiscsquare);
	resA = resB*resB-aDisc;
	resA /= 4;
	resA /= aA;

	if ((backupA != resA) || (backupB != resB) || (backupC != resC))	
	{
		std::cerr << "Error!";
		return ConstRC::GeneralError;
	};
#endif 
	return ConstRC::Ok;
};

long SQUFOF::gcd(long  aSource1, long aSource2) {
	while (aSource2 != 0) {
		long rem = aSource1 % aSource2;
		aSource1 = aSource2;
		aSource2 = rem;	
	};
	return aSource1;
};

long SQUFOF::gcd(long  aSource1, long aSource2, long aSource3) {
	long result = gcd(aSource1,aSource2);
	result = gcd(result,aSource3);
	return result;
};

int SQUFOF::Factor(mpz_t aModulus, mpz_t aFactor1, mpz_t aFactor2) {

	//gmp_fprintf(stderr, "\n%Zd\n",aModulus);
	mpz_set_ui(aFactor1,0);
	mpz_set_ui(aFactor2,0);

	long qc0 = 0; // q_0
	long qc = 0; // q
	const long queue_length = 100;
	long queue[queue_length+3];
	long N = mpz_get_ui(aModulus);
	long D;
	long t;
	long q;
	long r;
	long P2;
	if (1 == N % 4) 
		D = 2 * N;
	else 
		D = N;
	long S;
	sqrti(S,D);
	long Q2 = 1;
	long P = S;
	long Q = D - P*P;
	long L = floor(2* sqrt(2 * sqrt(D))); // S == sqrt(D)
	long B = 2 * L;
	// 2
	//std::cerr << "N:"<<N<<" D:"<<D<<" S:"<<S<< " Q2:" <<Q2 <<" P:"<<P<<" Q:" << Q<<" L:"<<L << " ";
	long i;
	for (i = 0; i < B; i++) {
		//std::cerr << std::endl<< "N:"<<N<<" D:"<<D<<" S:"<<S<< " Q2:" <<Q2 <<" P:"<<P<<" Q:" << Q<<" L:"<<L << " ";
		//std::cerr << i;
		// 2a)
		q = (S+P)/Q;
		P2 = q*Q-P;
		// 2b)
		if (Q <= L) {
			if (Q % 2 == 0) {
				queue[qc++] = Q/2;
				queue[qc++] = P % (Q/2);
			} else {
				if (Q <= L/2)  {
					queue[qc++] = Q;
					queue[qc++] = P % Q;
				}
			};
			if (qc>=queue_length) 
				qc = 0;
		};
		// 2c)
		t = Q2 + q*(P-P2);
		Q2 = Q;
		Q = t;
		P = P2;
		// 2d)
		//std::cerr << "\tQ:" << Q;
		if (sqrti(r,Q) && (i % 2 == 0)) {
			long qi = qc0;
			while (qi != qc) {	
				//std::cerr << "\tr:" << r;
				// pair found
				if ((r == queue[qi]) &&  (0 == (P - queue[qi+1]) % r ))
					break;
				

				qi += 2;
				if (qi > queue_length) 
					qi = 0;
			};
			// if pair not found, goto 3)
			// if (P - queue[qi+1] % queue[qi] != 0)  {
			if (qc == qi) {
				t = queue[qi+1];
				if (qi > queue_length)
					qi = 0;
				break;
			};
			if (r>1) {	
				// flush all pairs up to qi (including)
				qc0 += qi+2;
				if (qc0 > queue_length) 
					qc0 = 0;
			} else  {
				return 1;
			};
		};
	};		
	if (i>=B) return ConstRC::NotFactorized;;
	//std::cerr << "\t Q:" << Q <<"\t r:"<<r;
	// 3)
	Q2 = r;
	P = P + r * ( (S - P) / r);
	Q = (D - P * P) / Q2;
	// 4)
	while (1) {
		// 4a)
		q = (S+P)/Q;
		P2 = q * Q - P;
		// 4b)
		if (P == P2) break;
		// 4c)
		t = Q2 + q * (P - P2);
		Q2 = Q;
		Q = t;
		P = P2;
	};
	if (Q % 2 == 0) 
		Q /= 2;
	mpz_set_ui(aFactor1,Q);
	mpz_divexact(aFactor2, aModulus, aFactor1);
	mpz_mul(checkprod, aFactor1, aFactor2);
	if (0==mpz_cmp(aModulus,checkprod))
		return ConstRC::Factorized;
	else
		return ConstRC::NotFactorized;
};

int SQUFOF::Test() {
	SQUFOF squ;
	mpz_t number, factor1, factor2;

	mpz_init(number);
	mpz_init(factor1);
	mpz_init(factor2);

	mpz_set_ui(number,93433);
	mpz_set_ui(number,64777);
	mpz_set_ui(number,29041);
	mpz_set_ui(number,22117019);
	mpz_set_ui(number,84611);
	mpz_set_ui(number,846743);

	long rv =squ.Factor(number,factor1,factor2);
	std::cerr << "RV:" << rv<< std::endl;

	gmp_printf("\n -> %Zd", number);
	gmp_printf("\n %Zd * %Zd",factor1, factor2);
	mpz_mul(factor1, factor1, factor2);
	gmp_printf("\n = %Zd\n\n", factor1);

	mpz_clear(number);
	mpz_clear(factor1);
	mpz_clear(factor2);

	return 0;
}
