#include "threshold_optimizer.h"
#include <stdio.h>
#include <stdlib.h>

#define DEBUG_THRESHOLD_OPTIMIZER 1

ThresholdOptimizer::ThresholdOptimizer(int aOptimizationPolicy)
	: iAccepted(0), iRejected(0), iOptimizationPolicy(aOptimizationPolicy)	
	{
	if ((iOptimizationPolicy > 5) || (iOptimizationPolicy < 0))
		{
#if DEBUG_THRESHOLD_OPTIMIZER
		printf("\t\tThreshold optimizer: bad optimization policy: %d.\n",iOptimizationPolicy);
#endif
		}		
	}

ThresholdOptimizer::~ThresholdOptimizer()
	{
	}


void	ThresholdOptimizer::AnalyzeAndOptimize(log_type& aCurrentThreshold)
	{
#if DEBUG_THRESHOLD_OPTIMIZER
	PrintStatus();
#endif
	if (!HasEnoughInformationToAnalyze())
		{
#if DEBUG_THRESHOLD_OPTIMIZER
		printf("\t\tWe do not have enough information to analyze and optimize the threshold.\n");
#endif
		return;
		}
	// Has enough information 
	if (!iRejected)
	{
		// We have at least KMinAccepted, but no rejected.
		aCurrentThreshold += 3;
#if DEBUG_THRESHOLD_OPTIMIZER
		printf("\t\tSpecial case: rejected == 0");
#endif
	}
	else
	{
		iRatio = (float) iAccepted / (float) iRejected;
#if DEBUG_THRESHOLD_OPTIMIZER
		printf("\t\tRatio now: %4.2f\n",iRatio);
#endif
		signed int move = 0;
		if (iRatio < 0.05)
		{
		// A lot of rejected, just a few of accepted
			move = 3;
		} 
		if ((iRatio >= 0.05) && (iRatio < 0.10))
		{
			move = 2;
		} 
		if ((iRatio >= 0.10) && (iRatio < 0.33))
		{
			move = 1;
		} 
		if ((iRatio >= 0.33) && (iRatio < 1))
		{
			move = 0;
		} 
		if ((iRatio >= 1) && (iRatio < 2))
		{
			move = -1;
		} 
		if ((iRatio >= 2) && (iRatio < 5))
		{
			move = -2;
		} 
		if (iRatio >= 5)
		{
			move = -3;
		}
		bool optimize = TRUE;
		switch (iOptimizationPolicy)
		{
			case 0:
				optimize = FALSE;
				break;
			case 1:
				move -= 2;
				break;	
			case 2:
				move -= 1;
				break;	
			case 3:
				// Nothing		
				break;	
			case 4:
				move += 1;
				break;	
			case 5:
				move += 2;
				break;		
			default:
				exit(10);
				break;
		}
		log_type absMove = 0;
		if (move >= 0) 
			{
			absMove = move;
			}
		else 	
			{
			absMove = (-1)*move;
			}		
		bool	tooLow = FALSE;
		if (aCurrentThreshold < absMove)
		{	
			tooLow = TRUE;
		}
		bool 	tooHigh = FALSE;
		if (move > 0)
			{
			// type overflow
			if ((aCurrentThreshold + move) < (aCurrentThreshold))
				{
				tooHigh = TRUE;
				}
			}

		log_type newThreshold = aCurrentThreshold;
		if (!tooLow && !tooHigh)
		{
			newThreshold += move;
		} else if (tooLow)
			{
				newThreshold = 0;
			} 
		  else if (tooHigh)
			{
				newThreshold = 255; // maximal unsigned char.
			}
		
#if DEBUG_THRESHOLD_OPTIMIZER
		printf("\t\tMove calculated as: %d. Old threshold: %d. New threshold: %d.\n",move, aCurrentThreshold, newThreshold);		
#endif
		// Warning is appropriate in these situations.
		if (tooLow)
		{
#if (!DEBUG_THRESHOLD_OPTIMIZER)
			// If not in debug mode, print the status now.
			// Otherwise, it has already been printed.
			PrintStatus();
#endif
			printf(" ^^^^^ Threshold optimizer: Warning. Threshold already too low. Set to %d.\n",newThreshold);
		}
		if (tooHigh)
		{
#if (!DEBUG_THRESHOLD_OPTIMIZER)
			// If not in debug mode, print the status now.
			// Otherwise, it has already been printed.
			PrintStatus();
#endif
			printf(" ^^^^^ Threshold optimizer: Warning. Threshold already too high. Set to %d.\n",newThreshold);
		}
		if (optimize)
		{
			aCurrentThreshold = newThreshold;	
		}
		ClearStatistics();
		return;	
	}

	}

void	ThresholdOptimizer::PrintStatus() const
	{
	printf("\n ^^^^^ Threshold optimizer status: policy %d ^^^^^\n",iOptimizationPolicy);
	printf("\t\tAccepted: %6d\n",iAccepted);
	printf("\t\tRejected: %6d\n",iRejected);
	if (iRejected)
		{
		printf("\t\tRatio: %4.2f\n",(float) iAccepted / (float) iRejected);
		}
	}

void	ThresholdOptimizer::Reset()	
	{
	ClearStatistics();
	}

void	ThresholdOptimizer::ClearStatistics()	
	{
	iAccepted = 0;
	iRejected = 0;
	iRatio = (float) 0;
	}

void	ThresholdOptimizer::IncrementAccepted(unsigned int aBy)
	{
	iAccepted += aBy;
	}

void	ThresholdOptimizer::IncrementRejected(unsigned int aBy)
	{
	iRejected += aBy;
	}


bool	ThresholdOptimizer::HasEnoughInformationToAnalyze() const
	{
	return ((iAccepted >= KMinAccepted) || (iRejected >= KMinRejected));
	}