#ifndef _THRESHOLD_OPTIMIZER_H_
#define _THRESHOLD_OPTIMIZER_H_

#include "types_common.h" // For log_type

#define THRESHOLD_OPTIMIZER_INTERVAL 500

/**
\class ThresholdOptimizer
\brief Optimizes threshold values during sieving. 

This class analyzes performance of current sieving run, mainly the
ratio of accepted to rejected relations. If the ratio indicates suboptimal
choice of the sieving threshold, the class computes a new value.

*/
class ThresholdOptimizer
{
public:
	// Construction and destruction
	/**
	\brief Constructs the optimizer.
	@param aOptimizationPolicy Either 0, 1, 2, 3, 4, 5. 0 = off. In this case, the optimized does not optimize. Otherwise, the higher is the number, the higher the threshold will be. Value other than 5 results in exit of the program.

	*/
	ThresholdOptimizer(int aOptimizationPolicy = 3);
	~ThresholdOptimizer();

	// Enter new information to analyze
	void	IncrementAccepted(unsigned int aBy = 1);
	void	IncrementRejected(unsigned int aBy = 1);

	/**
	\brief Determines new threshold value.
	@param aCurrentThreshold The current value of threshold. Will be rewritten with the new value.

	For the analysis logic, see the document sieving-threshold-optimization.odt

	*/
	void	AnalyzeAndOptimize(log_type& aCurrentThreshold);	


	void	Reset();

private:
	/**
          Clears the accepted / rejected statistics.
	*/
	void	ClearStatistics();

	/**
	  Returns TRUE if iAccepted >= KMinAccepted or iRejected >= KMinRejected.
	*/
	bool	HasEnoughInformationToAnalyze() const;

	void	PrintStatus() const;

private:
	// Data members
	unsigned int	iAccepted;
	unsigned int	iRejected;
	float		iRatio;

	static const unsigned int KMinAccepted = 10;
	static const unsigned int KMinRejected = 100;

	int 		iOptimizationPolicy;
};

#endif //  _THRESHOLD_OPTIMIZER_H_