#include "ttr_class.h"
#include "abstract_sieve.h"
#include <time.h>
#include "log.h"
//#include <conio.h>

/*!
		The default constructor, which sets the testing mode to EFast. In the current implementation, this
		constructor is never called.
*/
TTR::TTR(){
	test_mode = EFast;
	result    = TRUE;
}

/*!
	The constructor which sets the testing mode to aMode. This constructor is used, when some of the
	following command line options is given:
			
			-t 			
			
			-tfast
			
			-tstrict
			
	The -t option is used to set the normal test mode. There is no -tnormal option.
	
	There is no destructor in the TTR class. Each of the test methods cleans up after itself, and there are no dynamic member variables in the TTR, so there is no need for a destructor.

*/
TTR::TTR(testing_mode aMode){
	test_mode = aMode;
	result    = TRUE;
}

/*!
	  This method is called at the beginning of the test and prints out the version and test mode information.

*/
void TTR::InitialMessage(){
	printf("\n");
	printf("Thorough Test Routine v. %.2f\n",QS_VERSION);
	printf("=============================\n");
	printf("  Mode: ");
	switch(this->test_mode){
		case EFast:	printf("fast.");
				break;
		case ENormal:   printf("normal.");
				break;
		case EStrict:   printf("strict.");
				break;
		default:	printf("undefined.");
				break;
	}
	printf("\n\n");
}

/*!
 This method is called at the end of the test and prints out the overal result, OK or Failed, according to the true / false state of the result member variable.
*/
void TTR::FinalMessage(){
	printf("\n");
	printf("Final result: ");
	if (this->result == TRUE) {
				//textcolor(GREEN);
				//cprintf("OK");
				printf("OK");
			}

			else {
				//textcolor(RED);
				//cprintf("Failed");
				printf("Failed");
			}
	//textcolor(LIGHTGRAY);
	printf("\n");
}

int TTR::TestBitMatrix(){
	int ret;
	printf("  Test BitMatrix. \n");
	ret = TestConversionsToNormalMatrix();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestTranspose();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestAddition();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestRowSwaps();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestRowAdding();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestColumnSwaps();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestColumnAdding();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestMultiplications();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestMultiplicationsTransposed();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestGaussianInverse();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestSemidiagonalShape();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	return ConstRC::Ok;
}

int TTR::TestBCMatrix(){
	int ret;
	printf("\n  Test BCMatrix. \n");
	ret = TestBCConversionsToNormalMatrix();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCTranspose();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCAddition();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCRowSwaps();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCRowAdding();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCColumnSwaps();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCColumnAdding();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCMultiplications();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCMultiplicationsTransposed();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCGaussianInverse();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestBCSemidiagonalShape();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	return ConstRC::Ok;
}

/*!
	  This method generates a bit matrix of dimensions between 2 and 201, and converts it to a normal matrix and back.
	  Any mismatch between the original and the twice-converted matrix is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been any mismatch between the original and the final matrix.
		
*/
int TTR::TestConversionsToNormalMatrix(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing conversions to normal matrix",this->test_mode,678,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int seed = (int) clock();
	srand(seed);
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing bit/normal conversions (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		b_m_2 = b_m_1->ToBitMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing conversions to normal matrix",i,679,"Bad conversion");
			b_m_1->PrintToScreen();
			n_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			this->result = FALSE;
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Bit/normal conversions Ok (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a BC matrix of height between 2 and 201, and converts it to a normal matrix and back.
	  Any mismatch between the original and the twice-converted matrix is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestBCConversionsToNormalMatrix(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC conversions to normal matrix",this->test_mode,678,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int seed = (int) clock();
	srand(seed);
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC/normal conversions (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		b_m_2 = b_m_1->ToBCMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC conversions to normal matrix",i,679,"Bad conversion");
			b_m_1->PrintToScreen();
			n_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			this->result = FALSE;
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    BC/normal conversions Ok (%d).                             ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a bit matrix of dimensions between 2 and 201, and transposes it twice. Conversions to and back from
	  NormalMatrix type are performed at all stages of the computation.
	  
	  If either the transposed matrices differ from each other, or the twice-transposed matrix differs from the original matrix,
	  it is considered fatal; there are 4 tests per each iteration, which means that not only transposition, but also conversion
	  of types is tested.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestTranspose(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	BitMatrix* b_m_3;
	BitMatrix* b_m_4;
	BitMatrix* b_m_5;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transposition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		b_m_2 = b_m_1->Transpose();
		n_m_2 = n_m_1->Transpose();
		b_m_4 = b_m_1->ToBitMatrix(n_m_2);
		if (!b_m_2->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,680,"Bad transposition");
			b_m_2->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		b_m_3 = b_m_2->Transpose();
		n_m_3 = n_m_2->Transpose();
		b_m_5 = b_m_3->ToBitMatrix(n_m_3);
		if (!b_m_1->Equals(b_m_3)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,681,"Bad transposition");
			b_m_1->PrintToScreen();
			b_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		if (!n_m_3->Equals(n_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,682,"Bad transposition");
			n_m_1->PrintToScreen();
			n_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		if (!b_m_1->Equals(b_m_5)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,681,"Bad transposition");
			b_m_1->PrintToScreen();
			b_m_5->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete b_m_5; b_m_5 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Transposition OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a BC matrix of height between 2 and 201, and transposes it twice. Conversions to and back from
	  NormalMatrix type are performed at all stages of the computation.
	  If either the transposed matrices differ from each other, or the twice-transposed matrix differs from the original matrix,
	  it is considered fatal.
	  	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  	
		ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestBCTranspose(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	BCMatrix* b_m_1;
	BitMatrix* b_m_2;
	BitMatrix* b_m_3;
	BitMatrix* b_m_4;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC transposition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		b_m_2 = new BitMatrix(dim_y,dim_x);
		b_m_2->Allocate();
		b_m_1->Transpose(b_m_2);
		n_m_2 = n_m_1->Transpose();
		b_m_4 = b_m_2->ToBitMatrix(n_m_2);
		if (!b_m_2->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC transposition matrix",i,680,"Bad transposition");
			b_m_2->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
	}
	printf("\r");
	printf("    Transposition OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates two bit matrices of matching dimensions between 2 and 201, and adds them together. The same addition
	  is performed on their images in NormalMatrix class; then the NormalMatrix result is converted back to BitMatrix and compared
	  to the BitMatrix addition result. Any difference in results is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrices is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  	
		OK			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestAddition(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	BitMatrix* b_m_3;
	BitMatrix* b_m_4;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing addition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		b_m_2 = new BitMatrix(dim_x, dim_y);
		b_m_2->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		n_m_2 = b_m_2->ToNormalMatrix();
		b_m_3 = b_m_2->Add(b_m_1);
		n_m_3 = n_m_2->Add(n_m_1);
		b_m_4 = b_m_3->ToBitMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing addition of matrices",i,680,"Bad addition");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Addition OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates two BC matrices of matching height between 2 and 201, and adds them together. The same addition
	  is performed on their images in NormalMatrix class; then the NormalMatrix result is converted back to BCMatrix and compared
	  to the BCMatrix addition result. Any difference in results is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrices is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	OK			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestBCAddition(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	BCMatrix* b_m_3;
	BCMatrix* b_m_4;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC addition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		b_m_2 = new BCMatrix(dim_x, dim_y);
		b_m_2->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		n_m_2 = b_m_2->ToNormalMatrix();
		b_m_3 = b_m_2->Add(b_m_1);
		n_m_3 = n_m_2->Add(n_m_1);
		b_m_4 = b_m_3->ToBCMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC addition of matrices",i,680,"Bad addition");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Addition OK (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a bit matrix of dimensions between 2 and 201 and two row indices, and swaps the rows with these indices.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	OK			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestRowSwaps(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing row swaps ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing row swaps (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_x;
		ind_2 = abs(rand()) % dim_x;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->SwapRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing row swaps in BitMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->SwapRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing row swaps in NormalMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBitMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing row swaps in BitMatrix",i,680,"Bad row swap");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Row swap Ok (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a BC matrix of height between 2 and 201 and two row indices, and swaps the rows with these indices.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  	ConstRC::Ok			if everything went well
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestBCRowSwaps(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row swaps ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC row swaps (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_x;
		ind_2 = abs(rand()) % dim_x;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->SwapRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row swaps in BCMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->SwapRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row swaps in NormalMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBCMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row swaps in BCMatrix",i,680,"Bad row swap");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Row swap Ok (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a bit matrix of dimensions between 2 and 201 and two row indices, and adds one of the rows to the other.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
	  
*/
int TTR::TestRowAdding(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing row addition ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing row addition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_x;
		ind_2 = abs(rand()) % dim_x;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->AddRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing row addition in BitMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->AddRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing row addition in NormalMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBitMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing row addition in BitMatrix",i,680,"Bad row addition");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Row addition OK (%d).                        ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a BC matrix of height between 2 and 201 and two row indices, and adds one of the rows to the other.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestBCRowAdding(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row addition ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC row addition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_x;
		ind_2 = abs(rand()) % dim_x;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->AddRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row addition in BitMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->AddRows(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row addition in NormalMatrix",ind_1,ind_2,"Something happened with row indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBCMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC row addition in BitMatrix",i,680,"Bad row addition");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Row addition Ok (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!

	  This method generates a bit matrix of dimensions between 2 and 201 and two column indices, and swaps the columns with these indices.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestColumnSwaps(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing column swaps ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	long seed = (long) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing column swaps (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 20 + 2;
		dim_y = abs(rand()) % 20 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_y;
		ind_2 = abs(rand()) % dim_y;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->SwapColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing column swaps in BitMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->SwapColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing column swaps in NormalMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBitMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing column swaps in BitMatrix",ind_1,ind_2,"Bad column swap");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Column swap OK (%d).                         ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a BC matrix of dimensions between 2 and 201 and two column indices, and swaps the columns with these indices.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestBCColumnSwaps(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column swaps ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	long seed = (long) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC column swaps (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 20 + 2;
		dim_y = abs(rand()) % 20 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_y;
		ind_2 = abs(rand()) % dim_y;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->SwapColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column swaps in BitMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->SwapColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column swaps in NormalMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBCMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column swaps in BitMatrix",ind_1,ind_2,"Bad column swap");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Column swap OK (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a bit matrix of dimensions between 2 and 201 and two column indices, and adds one of the columns to the other.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestColumnAdding(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing column adding ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	long seed = (long) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing column addings (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_y;
		ind_2 = abs(rand()) % dim_y;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->AddColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing column adding in BitMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->AddColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing column adding in NormalMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBitMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing column adding in BitMatrix",ind_1,ind_2,"Bad column adding");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Column adding OK (%d).                         ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a BC matrix of dimensions between 2 and 201 and two column indices, and adds one of the columns to the other.
	  The same operation is mirrored in a NormalMatrix instance.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the BC matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	OK			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch
*/
int TTR::TestBCColumnAdding(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column adding ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y, ind_1, ind_2;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	NormalMatrix* n_m_1;
	int ret;
	long seed = (long) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing column addings (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		ind_1 = abs(rand()) % dim_y;
		ind_2 = abs(rand()) % dim_y;
		n_m_1 = b_m_1->ToNormalMatrix();
		ret = b_m_1->AddColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column adding in BitMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		ret = n_m_1->AddColumns(ind_1,ind_2);
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column adding in NormalMatrix",ind_1,ind_2,"Something happened with column indices");
			b_m_1->PrintToScreen();
			return ret;
		}
		b_m_2 = b_m_2->ToBCMatrix(n_m_1);
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC column adding in BitMatrix",ind_1,ind_2,"Bad column adding");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
	}
	printf("\r");
	printf("    Column adding OK (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
  	  This method generates two bit matrices of matching dimensions between 2 and 201, and multiplies them together. The same operation
	  is performed on their images in NormalMatrix class; then the NormalMatrix result is converted back to BitMatrix and compared
	  to the BitMatrix multiplication result. Any difference in results is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  The number of columns of the bit matrices is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	OK			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix
*/
int TTR::TestMultiplications(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplications ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	BitMatrix* b_m_3;
	BitMatrix* b_m_4;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 200 + 2;
		b_m_1 = new BitMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		b_m_2 = new BitMatrix(dim_y, dim_z);
		b_m_2->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		n_m_2 = b_m_2->ToNormalMatrix();
		b_m_3 = b_m_1->MultiplyInternal(b_m_1,b_m_2);
		n_m_3 = n_m_1->MultiplyInternal(n_m_1,n_m_2);
		b_m_4 = b_m_3->ToBitMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplication of matrices",i,680,"Bad multiplication");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Multiplication Ok (%d).                         ",iterations);
	return ConstRC::Ok;
}

/*!
  	  This method generates two bit matrices of matching dimensions between 2 and 201, and multiplies them together,
	  the first one as transposed. The same operation is performed on their images in NormalMatrix class; 
	  then the NormalMatrix result is converted back to BitMatrix and compared
	  to the BitMatrix multiplication result. Any difference in results is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  The number of columns of the bit matrices is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestMultiplicationsTransposed(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposed multiplications ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	BitMatrix* b_m_3;
	BitMatrix* b_m_4;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transposed multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_x%10 == 0) dim_x = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 200 + 2;
		b_m_1 = new BitMatrix(dim_y, dim_x);
		b_m_1->Randomize();
		b_m_2 = new BitMatrix(dim_y, dim_z);
		b_m_2->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		n_m_2 = b_m_2->ToNormalMatrix();
		b_m_3 = b_m_1->MultiplyInternalTransposed(b_m_1,b_m_2);
		n_m_3 = n_m_1->MultiplyInternalTransposed(n_m_1,n_m_2);
		b_m_4 = b_m_3->ToBitMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposed multiplication of matrices",i,680,"Bad multiplication");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Transposed multiplication OK (%d).                         ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates two BC matrices of matching height between 2 and 201, and multiplies them together. The same operation
	  is performed on their images in NormalMatrix class; then the NormalMatrix result is converted back to BCMatrix and compared
	  to the BCMatrix multiplication result. Any difference in results is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  The number of columns of the BC matrices is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestBCMultiplications(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC multiplications ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	BCMatrix* b_m_3;
	BCMatrix* b_m_4;
	BitMatrix* b_m_5;
	BCMatrix* b_m_6;
	BCMatrix* b_m_7;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	NormalMatrix* n_m_4;
	NormalMatrix* n_m_5;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 2000 + 2;
		dim_y = abs(rand()) % 30 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 30 + 2;
		b_m_1 = new BCMatrix(dim_x, dim_y);
		b_m_1->Randomize();
		b_m_2 = new BCMatrix(dim_y, dim_z);
		b_m_2->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		n_m_2 = b_m_2->ToNormalMatrix();
		b_m_3 = b_m_1->MultiplyInternal(b_m_1,b_m_2);
		n_m_3 = n_m_1->MultiplyInternal(n_m_1,n_m_2);
		b_m_4 = b_m_3->ToBCMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC multiplication of matrices",i,680,"Bad multiplication");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		b_m_5 = new BitMatrix(dim_y,dim_x);
		b_m_5->Randomize();
		n_m_5 = b_m_5->ToNormalMatrix();
		b_m_6 = new BCMatrix(dim_y,dim_y);
		b_m_6->Allocate();
		b_m_5->MultiplyInternal(b_m_6,b_m_5,b_m_1);
		n_m_4 = n_m_5->MultiplyInternal(n_m_5,n_m_1);
		b_m_7 = b_m_6->ToBCMatrix(n_m_4);
		if (!b_m_7->Equals(b_m_6)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC and bit multiplication of matrices",i,680,"Bad multiplication");
			b_m_5->PrintToScreen();
			b_m_1->PrintToScreen();
			b_m_6->PrintToScreen();
			b_m_7->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete b_m_5; b_m_5 = NULL;
		delete b_m_6; b_m_6 = NULL;
		delete b_m_7; b_m_7 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
		delete n_m_4; n_m_4 = NULL;
		delete n_m_5; n_m_5 = NULL;
	}
	printf("\r");
	printf("    Multiplication OK (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
 	  This method generates two BC matrices of matching height between 2 and 201, and multiplies them together,
	  the first one as transposed. The same operation is performed on their images in NormalMatrix class; 
	  then the NormalMatrix result is converted back to BCMatrix and compared
	  to the BCMatrix multiplication result. Any difference in results is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  The number of columns of the BC matrices is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the BC matrix (accessed by BCMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  	
		ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestBCMultiplicationsTransposed(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC multiplications ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BCMatrix* b_m_1;
	BCMatrix* b_m_2;
	BCMatrix* b_m_3;
	BCMatrix* b_m_4;
	BitMatrix* b_m_5;
	BCMatrix* b_m_6;
	BCMatrix* b_m_7;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	NormalMatrix* n_m_3;
	NormalMatrix* n_m_4;
	NormalMatrix* n_m_5;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transposed multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 30 + 2;
		dim_y = abs(rand()) % 2000 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 30 + 2;
		b_m_1 = new BCMatrix(dim_y, dim_x);
		b_m_1->Randomize();
		b_m_2 = new BCMatrix(dim_y, dim_z);
		b_m_2->Randomize();
		n_m_1 = b_m_1->ToNormalMatrix();
		n_m_2 = b_m_2->ToNormalMatrix();
		b_m_3 = b_m_1->MultiplyInternalTransposed(b_m_1,b_m_2);
		n_m_3 = n_m_1->MultiplyInternalTransposed(n_m_1,n_m_2);
		b_m_4 = b_m_3->ToBCMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC multiplication transposed of matrices",i,680,"Bad multiplication");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			b_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		b_m_5 = new BitMatrix(dim_y,2*dim_x);
		b_m_5->Randomize();
		n_m_5 = b_m_5->ToNormalMatrix();
		b_m_6 = new BCMatrix(2*dim_x,dim_x);
		b_m_6->Allocate();
		b_m_5->MultiplyInternalTransposed(b_m_6,b_m_5,b_m_1);
		n_m_4 = n_m_5->MultiplyInternalTransposed(n_m_5,n_m_1);
		b_m_7 = b_m_6->ToBCMatrix(n_m_4);
		if (!b_m_7->Equals(b_m_6)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC and bit multiplication transposed of matrices",i,680,"Bad multiplication");
			b_m_5->PrintToScreen();
			b_m_1->PrintToScreen();
			b_m_6->PrintToScreen();
			b_m_7->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete b_m_4; b_m_4 = NULL;
		delete b_m_5; b_m_5 = NULL;
		delete b_m_6; b_m_6 = NULL;
		delete b_m_7; b_m_7 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
		delete n_m_4; n_m_4 = NULL;
		delete n_m_5; n_m_5 = NULL;
	}
	printf("\r");
	printf("    Transposed multiplication OK (%d).                           ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a square bit matrix of dimension 2-101 and tries to semidiagonalize it. Then, it walks through the matrix
	  and if it notices a nonzero bit below the main diagonal, returns error	
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  The number of columns of the bit matrix is not completely random. With probability 10% it is equal to number of bits
	  in the type composing the bit matrix (accessed by BitMatrix::GetBitsInMT()). That is because we are especially interested
	  in correctness of operations over matrices of this width; such matrices are extensively used in the Block Lanczos
	  algorithm and so the motivation to detect all possible problems with them is stronger.
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					a nonzero bit below the main diagonal.
*/
int TTR::TestSemidiagonalShape(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, j, k, dim_x;
	int ret;
	BitMatrix* b_m_1;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing semidiagonal shape (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 100 + 2;
		if (dim_x%10 == 0) dim_x = BitMatrix::GetBitsInMT();
		b_m_1 = new BitMatrix(dim_x, dim_x);
		b_m_1->Randomize();
		ret = b_m_1->ToSemidiagonalShape();
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing semidiagonalizing matrix",i,680,"Error in execution of method");
		}
		for (j=0; j < dim_x; j++){
			for (k=j+1; k < dim_x; k++){
				if (b_m_1->IsOne(k,j) == TRUE){
					printf("\n");
					CLog::LogError(WHEREARG,"Thorough Test Routine","testing semidiagonalizing matrix",i,690,"Bad semidiagonalization");
					b_m_1->PrintToScreen();
					return GENERAL_ERROR;
				}
			}
		}
		delete b_m_1; b_m_1 = NULL;
	}
	printf("\r");
	printf("    Semidiagonal shape OK (%d).                        ",iterations);
	return ConstRC::Ok;
}

int TTR::TestBCSemidiagonalShape(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, j, k, dim_x;
	int ret;
	BCMatrix* b_m_1;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC semidiagonal shape (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 30 + 2;
		if (dim_x%10 == 0) dim_x = BitMatrix::GetBitsInMT();
		b_m_1 = new BCMatrix(dim_x, dim_x);
		b_m_1->Randomize();
		ret = b_m_1->ToSemidiagonalShape();
		if (ret != ConstRC::Ok){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC semidiagonalizing matrix",i,680,"Error in execution of method");
		}
		for (j=0; j < dim_x; j++){
			for (k=j+1; k < dim_x; k++){
				if (b_m_1->IsOne(k,j) == TRUE){
					printf("\n");
					CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC semidiagonalizing matrix",i,690,"Bad semidiagonalization");
					b_m_1->PrintToScreen();
					return GENERAL_ERROR;
				}
			}
		}
		delete b_m_1; b_m_1 = NULL;
	}
	printf("\r");
	printf("    Semidiagonal shape OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method has no real body, it only returns ConstRC::Ok value. That is because the Gaussian inverse operation has not been implemented
	  in BitMatrix.
	  
	  Return values:
	  
	  	ConstRC::Ok			anytime
*/
int TTR::TestGaussianInverse(){
	return ConstRC::Ok;
}

int TTR::TestBCGaussianInverse(){
	return ConstRC::Ok;
}

int TTR::TestSparseMatrix(){
	int ret;
	printf("\n");
	printf("  Test SparseMatrix. \n");
	ret = TestSparseConversionsToNormalMatrix();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestSparseConversionsToBitMatrix();
		if (ret != ConstRC::Ok){
			this->result = FALSE;
		}
	ret = TestSparseTranspose();
		if (ret != ConstRC::Ok){
			if (ret != NOT_SUPPORTED) this->result = FALSE;
		}
	ret = TestSparseBitMultiplications();
		if (ret != ConstRC::Ok){
			if (ret != NOT_SUPPORTED) this->result = FALSE;
		}
	ret = TestSparseBCMultiplications();
		if (ret != ConstRC::Ok){
			if (ret != NOT_SUPPORTED) this->result = FALSE;
		}
	ret = TestSparseBitMultiplicationWithTransposition();
		if (ret != ConstRC::Ok){
			if (ret != NOT_SUPPORTED) this->result = FALSE;
		}
	ret = TestSparseBCMultiplicationWithTransposition();
		if (ret != ConstRC::Ok){
			if (ret != NOT_SUPPORTED) this->result = FALSE;
		}
	ret = TestSparseBitMultiplicationWithTransposition_II();
		if (ret != ConstRC::Ok){
			if (ret != NOT_SUPPORTED) this->result = FALSE;
		}
	return ConstRC::Ok;
}

/*!
	  This method generates a normal matrix of dimensions between 5 and 204, and converts it to a sparse matrix and back.
	  Any mismatch between the original and the twice-converted matrix is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestSparseConversionsToNormalMatrix(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing conversions to normal matrix",this->test_mode,978,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	SparseMatrix* s_m_1;
	NormalMatrix* n_m_1;
	NormalMatrix* n_m_2;
	int seed = (int) clock();
	srand(seed);
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing sparse/normal conversions (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 5;
		dim_y = abs(rand()) % 200 + 5;
		n_m_1 = new NormalMatrix(dim_x,dim_y);
		n_m_1->Randomize();
		n_m_1->Modulo(2);
		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		n_m_2 = s_m_1->ToNormalMatrix();
		if (!n_m_2->Equals(n_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing conversions of sparse to normal matrix",i,979,"Bad conversion");
			n_m_1->PrintToScreen();
			n_m_2->PrintToScreen();
			this->result = FALSE;
			return GENERAL_ERROR;
		}
		delete s_m_1; s_m_1 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
	}
	printf("\r");
	printf("    Sparse/normal conversions OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a bit matrix of dimensions between 5 and 204, and converts it to a sparse matrix and back.
	  Any mismatch between the original and the twice-converted matrix is considered fatal.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestSparseConversionsToBitMatrix(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing conversions to normal matrix",this->test_mode,978,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	SparseMatrix* s_m_1;
	BitMatrix* b_m_1;
	BitMatrix* b_m_2;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing sparse/bit conversions (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 5;
		dim_y = abs(rand()) % 200 + 5;
		b_m_1 = new BitMatrix(dim_x,dim_y);
		b_m_1->Randomize();
		s_m_1 = SparseMatrix::ToSparseMatrix(b_m_1);
		b_m_2 = s_m_1->ToBitMatrix();
		if (!b_m_2->Equals(b_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing conversions of sparse to bit matrix",i,179,"Bad conversion");
			b_m_1->PrintToScreen();
			b_m_2->PrintToScreen();
			this->result = FALSE;
			return GENERAL_ERROR;
		}
		delete s_m_1; s_m_1 = NULL;
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
	}
	printf("\r");
	printf("    Sparse/bit conversions OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a normal matrix of dimensions between 2 and 201, and transposes it twice. Conversions to and back from
	  SparseMatrix type are performed at all stages of the computation.
	  
	  If either the transposed matrices differ from each other, or the twice-transposed matrix differs from the original matrix,
	  it is considered fatal; there are 4 tests per each iteration, which means that not only transposition, but also conversion
	  of types is tested.
	  
	  Number of iterations:
	  
	  fast mode 		20
	  
	  normal mode		100
	  
	  strict mode 		500
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestSparseTranspose(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 20;
				break;
		case ENormal:   iterations = 100;
				break;
		case EStrict:   iterations = 500;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing transpositions ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 500;
				break;
	}
	int i, dim_x, dim_y;
	SparseMatrix* s_m_1 = NULL;
	SparseMatrix* s_m_2 = NULL;
	SparseMatrix* s_m_3 = NULL;
	NormalMatrix* n_m_1 = NULL;
	NormalMatrix* n_m_2 = NULL;
	NormalMatrix* n_m_3 = NULL;
	NormalMatrix* n_m_4 = NULL;
	NormalMatrix* n_m_5 = NULL;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transposition (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();

		n_m_1 = new NormalMatrix(dim_x, dim_y);
		n_m_1->Randomize();
		n_m_1->Modulo(2);

		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		s_m_2 = s_m_1->Transpose();
		n_m_2 = n_m_1->Transpose();
		n_m_3 = s_m_2->ToNormalMatrix();
		if (!n_m_2->Equals(n_m_3)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,980,"Bad transposition");
			n_m_2->PrintToScreen();
			n_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		s_m_3 = s_m_2->Transpose();
		n_m_4 = n_m_2->Transpose();
		n_m_5 = s_m_3->ToNormalMatrix();
		if (!n_m_5->Equals(n_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,981,"Bad transposition");
			n_m_4->PrintToScreen();
			n_m_5->PrintToScreen();
			return GENERAL_ERROR;
		}
		if (!n_m_4->Equals(n_m_1)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing transposition matrix",i,982,"Bad transposition");
			n_m_1->PrintToScreen();
			n_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete s_m_1; s_m_1 = NULL;
		delete s_m_2; s_m_2 = NULL;
		delete s_m_3; s_m_3 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
		delete n_m_4; n_m_4 = NULL;
		delete n_m_5; n_m_5 = NULL;
	}
	printf("\r");
	printf("    Transposition OK (%d).                        ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a normal and a bit matrix of matching dimensions, where dimensions of the normal matrix are between 2
	  and 251, and dimensions of the bit matrix are: number of rows between 2 and 251, and number of columns between 2 and 
	  BitMatrix::bits_in_m_t+1; the bit matrix is converted into a normal matrix.
	  
	  The normal matrix is converted into a sparse matrix and multiplication of the sparse matrix by the bit matrix takes place,
	  mirrored in operation on normal matrices. At the end, the results are compared and any mismatch is considered fatal.
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestSparseBitMultiplications(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplications ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BitMatrix* b_m_1 = NULL;
	BitMatrix* b_m_2 = NULL;
	BitMatrix* b_m_3 = NULL;
	SparseMatrix* s_m_1 = NULL;
	NormalMatrix* n_m_1 = NULL;
	NormalMatrix* n_m_2 = NULL;
	NormalMatrix* n_m_3 = NULL;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 250 + 2;
		dim_y = abs(rand()) % 250 + 2;
		dim_z = abs(rand()) % 50 + 2;
		if (dim_z%10 == 0) dim_z = BitMatrix::GetBitsInMT();
		n_m_1 = new NormalMatrix(dim_x, dim_y);
		n_m_1->Randomize();
		n_m_1->Modulo(2);
		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		b_m_1 = new BitMatrix(dim_y, dim_z);
		b_m_1->Randomize();
		n_m_2 = b_m_1->ToNormalMatrix();
		n_m_3 = n_m_1->MultiplyInternal(n_m_1,n_m_2);
		b_m_2 = s_m_1->MultiplyInternalBig(s_m_1,b_m_1);
		b_m_3 = b_m_3->ToBitMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_2)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplication of matrices",i,780,"Bad multiplication");
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete s_m_1; s_m_1 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Multiplication OK (%d).                         ",iterations);
	return ConstRC::Ok;
}

/*!
	  This method generates a normal and a BC matrix of matching dimensions, where dimensions of the normal matrix are between 2
	  and 251, and dimensions of the BC matrix are: number of rows between 2 and 251, and number of columns between 2 and 
	  BitMatrix::bits_in_m_t+1; the BC matrix is converted into a normal matrix.
	  
	  The normal matrix is converted into a sparse matrix and multiplication of the sparse matrix by the bit matrix takes place,
	  mirrored in operation on normal matrices. At the end, the results are compared and any mismatch is considered fatal.
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestSparseBCMultiplications(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplications ",this->test_mode,679,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BCMatrix* b_m_1 = NULL;
	BCMatrix* b_m_2 = NULL;
	BCMatrix* b_m_3 = NULL;
	SparseMatrix* s_m_1 = NULL;
	NormalMatrix* n_m_1 = NULL;
	NormalMatrix* n_m_2 = NULL;
	NormalMatrix* n_m_3 = NULL;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing BC multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 250 + 2;
		dim_y = abs(rand()) % 250 + 2;
		dim_z = abs(rand()) % 30 + 2;
		if (dim_z%10 == 0) dim_z = BitMatrix::GetBitsInMT();
		n_m_1 = new NormalMatrix(dim_x, dim_y);
		n_m_1->Randomize();
		n_m_1->Modulo(2);
		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		b_m_1 = new BCMatrix(dim_y, dim_z);
		b_m_1->Randomize();
		n_m_2 = b_m_1->ToNormalMatrix();
		n_m_3 = n_m_1->MultiplyInternal(n_m_1,n_m_2);
		b_m_2 = s_m_1->MultiplyInternalBig(s_m_1,b_m_1);
		b_m_3 = b_m_3->ToBCMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_2)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing BC multiplication of matrices",i,780,"Bad multiplication");
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete s_m_1; s_m_1 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Multiplication BC OK (%d).                          ",iterations);
	return ConstRC::Ok;
}

/*!
	  There are three methods test correctness of MultiplyInternalBigWithTransposition() methods in SparseMatrix class.
	  
	  fast mode 		10
	  
	  normal mode		40
	  
	  strict mode 		80
	  
	  Return values:
	  
	  	ConstRC::Ok			if everything went well
		
		GENERAL_ERROR		if either of the called operations failed (due to lack of memory etc.), or if there has been
					any mismatch between the original and the final matrix.
*/
int TTR::TestSparseBitMultiplicationWithTransposition(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplications with transposition ",this->test_mode,779,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BitMatrix* b_m_1 = NULL;
	BitMatrix* b_m_2 = NULL;
	BitMatrix* b_m_3 = NULL;
	SparseMatrix* s_m_1 = NULL;
	NormalMatrix* n_m_1 = NULL;
	NormalMatrix* n_m_2 = NULL;
	NormalMatrix* n_m_3 = NULL;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transpose multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 200 + 2;
		n_m_1 = new NormalMatrix(dim_x, dim_y);
		n_m_1->Randomize();
		n_m_1->Modulo(2);
		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		b_m_1 = new BitMatrix(dim_z, dim_y);
		b_m_1->Randomize();
		n_m_2 = b_m_1->ToNormalMatrix();
		n_m_3 = n_m_1->MultiplyInternalWithTransposition(n_m_1,n_m_2);
		b_m_2 = s_m_1->MultiplyInternalBigWithTransposition(s_m_1,b_m_1);
		b_m_3 = b_m_3->ToBitMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_2)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplication with transposition",i,788,"Bad multiplication");
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete s_m_1; s_m_1 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Multiplication with transposition OK (%d).                         ",iterations);
	return ConstRC::Ok;
}

int TTR::TestSparseBCMultiplicationWithTransposition(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplications with transposition ",this->test_mode,779,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BCMatrix* b_m_1 = NULL;
	BCMatrix* b_m_2 = NULL;
	BCMatrix* b_m_3 = NULL;
	SparseMatrix* s_m_1 = NULL;
	NormalMatrix* n_m_1 = NULL;
	NormalMatrix* n_m_2 = NULL;
	NormalMatrix* n_m_3 = NULL;
	int seed = (int) clock();
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transpose BC multiplication (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 30 + 2;
		n_m_1 = new NormalMatrix(dim_y, dim_x);
		n_m_1->Randomize();
		n_m_1->Modulo(2);
		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		b_m_1 = new BCMatrix(dim_y, dim_z);
		b_m_1->Randomize();
		n_m_2 = b_m_1->ToNormalMatrix();
		n_m_3 = n_m_1->MultiplyInternalTransposed(n_m_1,n_m_2);
		b_m_2 = s_m_1->MultiplyInternalBigTransposed(s_m_1,b_m_1);
		b_m_3 = b_m_3->ToBCMatrix(n_m_3);
		if (!b_m_3->Equals(b_m_2)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplication with transposition",i,788,"Bad multiplication");
			b_m_2->PrintToScreen();
			b_m_3->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete b_m_2; b_m_2 = NULL;
		delete b_m_3; b_m_3 = NULL;
		delete s_m_1; s_m_1 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
	}
	printf("\r");
	printf("    Multiplication BC with transposition OK (%d).                         ",iterations);
	return ConstRC::Ok;
}

int TTR::TestSparseBitMultiplicationWithTransposition_II(){
	int iterations;
	switch (this->test_mode){
		case EFast:  	iterations = 10;
				break;
		case ENormal:   iterations = 40;
				break;
		case EStrict:   iterations = 80;
				break;
		default:	CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplications with transposition II",this->test_mode,779,"Test mode is in undefined state");
				iterations = 80;
				break;
	}
	int i, dim_x, dim_y, dim_z;
	BitMatrix* b_m_1 = NULL;
	// BitMatrix* b_m_2 = NULL;
	// BitMatrix* b_m_3 = NULL;
	SparseMatrix* s_m_1 = NULL;
	SparseMatrix* s_m_2 = NULL;
	NormalMatrix* n_m_1 = NULL;
	NormalMatrix* n_m_2 = NULL;
	NormalMatrix* n_m_3 = NULL;
	NormalMatrix* n_m_4 = NULL;
	int seed = (int) clock();
	int ret;
	srand(seed);
	printf("\n");
	for (i=0; i< iterations; i++){
		printf("\r");
		printf("    Testing transpose multiplication II (%3d/%d iterations) ...",i,iterations);
		fflush(stdout);
		dim_x = abs(rand()) % 200 + 2;
		dim_y = abs(rand()) % 200 + 2;
		if (dim_y%10 == 0) dim_y = BitMatrix::GetBitsInMT();
		dim_z = abs(rand()) % 200 + 2;
		n_m_1 = new NormalMatrix(dim_x, dim_y);
		n_m_1->Randomize();
		n_m_1->Modulo(2);
		n_m_2 = new NormalMatrix(dim_z, dim_y);
		n_m_2->Randomize();
		n_m_2->Modulo(2);
		s_m_1 = SparseMatrix::ToSparseMatrix(n_m_1);
		s_m_2 = SparseMatrix::ToSparseMatrix(n_m_2);
		b_m_1 = new BitMatrix(dim_x, dim_z);
		ret = s_m_1->MultiplyInternalBigWithTransposition(b_m_1,s_m_1,s_m_2,FALSE);
		if (ret != ConstRC::Ok){
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplication with transposition II",i,777,"Some problem in multiplication");
			return ret;
		}
		n_m_3 = b_m_1->ToNormalMatrix();
		n_m_4 = n_m_1->MultiplyInternalWithTransposition(n_m_1,n_m_2);
		n_m_4->Modulo(2);
		if (!n_m_3->Equals(n_m_4)){
			printf("\n");
			CLog::LogError(WHEREARG,"Thorough Test Routine","testing multiplication with transposition II",i,788,"Bad multiplication");
			n_m_3->PrintToScreen();
			n_m_4->PrintToScreen();
			return GENERAL_ERROR;
		}
		delete b_m_1; b_m_1 = NULL;
		delete s_m_1; s_m_1 = NULL;
		delete s_m_2; s_m_2 = NULL;
		delete n_m_1; n_m_1 = NULL;
		delete n_m_2; n_m_2 = NULL;
		delete n_m_3; n_m_3 = NULL;
		delete n_m_4; n_m_4 = NULL;
	}
	printf("\r");
	printf("    Multiplication with transposition II OK (%d).                         ",iterations);
	return ConstRC::Ok;
}
