#ifndef __TTR_CLASS__
#define __TTR_CLASS__

#include "common_definitions.h"
#include "bit_matrix_class.h"
#include "bc_matrix_class.h"
#include "sparse_matrix_class.h"

enum testing_mode   { EFast = 1, ENormal, EStrict };

/*!
	\brief The Thorough Test Routine for testing correctness of matrix operations.
	
	This class is used for performing tests of correctness of matrix operations. It has been written in order to remove many subtle bugs from implementations of BitMatrix and SparseMatrix classes.
	
	The main principle of TTR is the following:
		- it is easy to write correct implementation of algebraic operations over the NormalMatrix class, which uses array indices in the very same way as "textbook formulas say"
		- it is easy to write correct matrix comparisons in BitMatrix and SparseMatrix (Equals() method)
		- it is relatively easy to write correct conversion procedures BitMatrix <-> NormalMatrix and SparseMatrix  <-> NormalMatrix		  
		- it is quite hard to write correct implementation of algebraic operations over the BitMatrix and SparseMatrix class
		
	So, in testing of correctness, at first random BitMatrix/SparseMatrix instances are allocated ("random" does not mean only randomness of entries, but also of dimensions, with the obvious limitation that two matrices to add must have the same dimensions etc.). These instances will be used as arguments of the algebraic operations.
	
	Now, these instances are converted to NormalMatrix instances, and the requested operations are performed in both NormalMatrix class and the tested class (either BitMatrix or SparseMatrix).
	
	Finally, the result from the NormalMatrix class is converted into the tested class representation and compared to result of the tested class operation entry-by-entry. Any mismatch is considered an error. If there is an error in allocation or similar memory problems, the test is considered failed as well; an environment with unreliable memory allocation routines is definitely not optimal.
	
	There are three modes of TTR: fast, normal and strict. The modes in fact do not differ much from each other, except from number of operations. For example, fast mode will perform 20 certain operations, normal mode	will perform 100 of them and strict mode 500 of them. This is aimed at detection of very subtle errors; while it is improbable that any major bug escaped even the fast mode TTR, there could be bugs which demonstrate	themselves only in rare situations (like "there must be X entries in the Ythe row), and there is a better chance to find them using strict mode TTR.
	
	With help of the TTR, it was possible to remove many problems from the matrix procession stage. However, the reason why TTR is included in release versions of MPQS software, is the following: using TTR, any user can test reliability of the matrix operations on his system. Moreover, a run of TTR gives the user a chance to observe speed of matrix operations.
	
	The number of iterations of TTR in each mode, for each operation (like the above mentioned 20, 100, 500) is hardcoded into the source code, without use of symbolic constants. This is not a particularly good design, but the harm done is small, since these numbers do not have to change often (maybe they do not have to change	at all). Correction of this approach to the symbolic-constant approach is not on the to-do list.

*/
class TTR {
	public:
		TTR();
		TTR(testing_mode aMode);
		void InitialMessage();
		int  TestNormalMatrix(); ///< 	  This method is only declared and has not been implemented. The declaration probably had some purpose before.
		int  TestBitMatrix(); ///<   This method is the caller method for all private test methods concerning operations implemented in the BitMatrix class. So far, the only return value is OK. This return value is not tested anywhere, so has no real meaning. This may	  change in the future.
		int  TestBCMatrix(); ///<   This method is the caller method for all private test methods concerning operations implemented in the BCMatrix class. So far, the only return value is OK. This return value is not tested anywhere, so has no real meaning. This may	  change in the future.
		int  TestSparseMatrix(); ///< This method is the caller method for all private test methods concerning operations implemented in the SparseMatrix class. So far, the only return value is OK. This return value is not tested anywhere, so has no real meaning. This may  change in the future.
		void FinalMessage();

	private:
		int TestConversionsToNormalMatrix();
		int TestTranspose();
		int TestAddition();
		int TestRowSwaps();
		int TestRowAdding();
		int TestColumnSwaps();
		int TestColumnAdding();
		int TestMultiplications();
		int TestMultiplicationsTransposed();
		int TestGaussianInverse();
		int TestSemidiagonalShape();
		int TestBCConversionsToNormalMatrix();
		int TestBCTranspose();
		int TestBCAddition();
		int TestBCRowSwaps();
		int TestBCRowAdding();
		int TestBCColumnSwaps();
		int TestBCColumnAdding();
		int TestBCMultiplications();
		int TestBCMultiplicationsTransposed();
		int TestBCGaussianInverse();
		int TestBCSemidiagonalShape();
		int TestSparseConversionsToNormalMatrix();
		int TestSparseConversionsToBitMatrix();
		int TestSparseTranspose();
		int TestSparseBitMultiplications();
		int TestSparseBCMultiplications();
		int TestSparseBitMultiplicationWithTransposition();
		int TestSparseBCMultiplicationWithTransposition();
		int TestSparseBitMultiplicationWithTransposition_II();
		testing_mode test_mode; ///< 		  This variable determines the mode of the test to run (fast, normal, strict).
		bool result; ///< 		  This variable determines the overall result of the running test. It is set to true at the beginning, and detection of any error turns it into false. So, at the end the result is true if and only if no error in no test has been detected, and no other error (like memory  shortage) has been detected.

};
#endif

