#ifndef _TYPES_COMMON_H_
#define _TYPES_COMMON_H_

#include <gmp.h>
#include "common_definitions.h"


/*!
 * \brief used in the ProcessRelations step during the cycle construction. 
 *
 * Each struct corresponds to one cycle, with the array
	containing indices of edges used to compose this cycle; fields "allocated" and "current max index" give information on this dynamic
	array of indices. Finally, the "running index" field is used in the "real" cycle construction, to determine which edges have already
	been added to the cycle and which have not.

*/
typedef struct {
	int* array;
	int  allocated;
	int  current_max_index;
	int  running_index;
} cycle_construction_entry;



/*!
  \file 
  \brief Aggregated declarations of types used throughout the MPQS/SIQS source.
  
  This file aggregates (both primitive and composite) types used in MPQS/SIQS. 

*/

typedef 	 int  isol_type; 		///< Used for storage of integer solutions
typedef          int  main_sieving_type; 	///< Used for representation of prime numbers
#define LONG_LARGE_PRIME
#ifdef LONG_LARGE_PRIME
	typedef unsigned long large_prime_type;        ///< Used for representation of large primes in single and double LPV
#else
	typedef unsigned int large_prime_type;        ///< Used for representation of large primes in single and double LPV
#endif

typedef unsigned char log_type;          	///< Used for representation of logarithms.
typedef unsigned char detection_type;       ///< Used for detection fields in singleton removal (single-,double-LPV)
typedef 	 int  nexksb_type;

enum distribution { ENone=0, ECenter, ENode, ENotSpecified};

/**
By convention, connector_type enum starts by value 1 and does not allow
any jumps in between the values (so the next ones are 2, 3).

Value "ESupremum" does not define any valid connection type, it is just
the highest value in the enum. It is used to detect incorrect connection types
in the class ConnectionInfo - see method "Parse".
If you add another connection type, ESupremum must be the last one.
*/
enum connector_type { EUnknownConnector = 0, EFileConnector = 1, ESupremum};


/*!
 * 
 * ETerminate ... message from center to node: stop sieving.
 * EStartParams ... message from center to node: start sieving with these parameters
 * ESmoothsFound ... message from node to center: list of smooth values found
 * EPartialsFound ... dtto for partial values
 * EReady ... message from node to center: node is ready to receive EStartParams.
 *
 */
enum message_type { 
	EUnknown = 0,  
	EReady = 1, 
	ENoJob = 2, 
	EJobParameters = 3, 
	EAck = 4, 
	ELocked = 5, 
	ENodeBusy = 6, 
	EQuit = 7, 
	EData = 8, 
	EAlive = 9
};

typedef long log_walking_type; // Used for going through the block of logs.

/*! 
 * \brief short structure intended to host a divisor in a relation. 
 *
 * Each relation has an array of such divisors. One can easily see
	that only two variables are used: prime number p (can be also -1), and exponent, which
	means to which power this prime divides the relation.

 */
typedef struct {
	main_sieving_type p;
	unsigned int exponent;
} sieving_element;

/*!
 * \brief auxiliary structure for holding large primes in relations 
 *
 * (especially partially constructed cycles during ProcessRelations
	in one of the large prime variations).

*/
typedef struct {
	large_prime_type prime;
	unsigned int exponent;
} lp_element;

/*!
 * \brief QS relation
 *
	This structure contains information about a relation. 
	- \a index is an auxiliary indexing variable, used for example in factor construction after the linear algebra step.
	- \a index_of_used_polynomial tells which polynomial has been used in construction of this relation; if this relation resulted from combination
	of two other relations, NO_NUMBER constant is used to indicate this.
	- \a argument is the value of x used in Q(x). Perhaps this could be converted to an integer.
	- \a factorized_value is the value of Q(x).
	- \a remaining is the value of Q(x) divided by all sieving_elements.
	- \a left_side_value is the value of K such that K^2 = L mod kN (so, the left side value is the base of the square).
	- \a right_side_value is the value whose SQUARE must be put together with the sieving_elements; in case of smooth and 1-partial relations, 
	it is equal to the value of \a d in case of MPQS, and to 1 in case of SIQS. If we combine two 1-partial relations sharing the same large
	prime P to one relation, the right_side_value is equal to P, since we must multiply the product of sieving_elements by P^2 to get a correct
	relation.
	- \a lprime1 and \a lprime2 contain the large primes in Double LPV. Either or both may be equal to 1.
	- \a lps, \a max_lp_index and \a assigned_lps are used basically only during the cycle construction step in the ProcessRelation step.
	- \a inited, as usual, expresses the state of initialization of mpz_t members.
	- \a maximal_sieving_element_index tells which index in \a elements array is the largest containing relevant data.
	- \a currently_assigned_sieving_elements tells the allocated size of dynamic array \a elements
	- \a elements is a dynamically allocated array which contains the data about prime divisors of Q(x)
*/
typedef struct {
	unsigned long index;
	signed long index_of_used_polynomial;
	mpz_t argument;
	mpz_t factorized_value;
	mpz_t remaining;
	mpz_t left_side_value;
	mpz_t right_side_value;

	large_prime_type lprime1;
	large_prime_type lprime2;
	
	lp_element* lps;
	signed long max_lp_index;
	unsigned long assigned_lps;
	
	int inited;
	signed long maximal_sieving_element_index;
	unsigned long currently_assigned_sieving_elements;
	sieving_element* elements;

	int components; // From how many components was this relation combined? Makes sense only for cycles. Is not saved to disk (so far)
} qs_relation;

/*
	Used as index of relations - can be quite huge 
	Can be serialized on some places as long!!!
	Can be negative!!!
*/
typedef long	relation_index_type;


#endif



