#include "url.h"
#include <stdlib.h> // atoi
#include <stdio.h>

URL::URL(const string& aUrl) : iPort(KNoPortDetermined)
{
	Parse(aUrl);
}

URL::~URL()
{
	Reset(); // Clears the info
}

string URL::GetScheme()   const
{
	return iScheme;
}

bool   URL::HasScheme()   const
{
	return (!iScheme.empty());
}

string URL::GetUsername() const
{
	return iUsername;
}

bool   URL::HasUsername() const
{
	return (!iUsername.empty());
}

string URL::GetPassword() const
{
	return iPassword;
}

bool   URL::HasPassword() const
{
	return (!iPassword.empty());
}

string URL::GetDomain()   const
{
	return iDomain;
}

bool   URL::HasDomain()   const
{
	return (!iDomain.empty());
}


string URL::GetFullPath() const
{
	return iFullPath;
}

bool URL::HasFullPath() const
{
	return (!iFullPath.empty());
}

string URL::GetPath()     const
{
	return iPath;
}

bool   URL::HasPath()     const
{
	return (!iPath.empty());
}

string URL::GetFilename() const
{
	return iFilename;
}

bool   URL::HasFilename() const
{
	return (!iFilename.empty());
}

string URL::GetExtension() const
{
	return iExtension;
}

bool   URL::HasExtension() const
{
	return (!iExtension.empty());
}

string URL::GetParamString() const
{
	return iParamString;
}

bool   URL::HasParamString() const
{
	return (!iParamString.empty());
}
	
string URL::GetQuery() const
{
	return iQuery;
}

bool   URL::HasQuery() const
{
	return (!iQuery.empty());
}

string URL::GetFragment() const
{
	return iFragment;
}

bool   URL::HasFragment() const
{
	return (!iFragment.empty());
}

int    URL::GetPort() const
{
	return iPort;
}

bool   URL::HasPort() const
{
	return (iPort != KNoPortDetermined);
}

string URL::GetCanonicalForm() const
{
	return iCanonicalForm;
}

void   URL::Parse(const string& aUrl)
{
	Reset(); // Clears all previous information.
	iCanonicalForm = aUrl;
	string copy(aUrl);
	size_t schemeDelimiter = copy.find_first_of(':');
	if (schemeDelimiter == string::npos)
	{
		printf("URL.Parse: no ':' found, scheme missing?\n");
		// Not a fatal problem. This might be an URI. We will
		// not change the TParsingResult to erroneous state.
	}
	else
	{
		iScheme = copy.substr(0,schemeDelimiter);
		copy.erase(0,schemeDelimiter); 
		if ((copy[0] != ':') || (copy[1] != '/') || (copy[2] != '/'))
		{
			printf("URL.Parse: no :// after scheme. Malformed!\n");
			iResult = EMalformedScheme;
			return; // No reason to go on. The URL string is seriously damaged.
		}
		copy.erase(0,3); // :// erased		
	}
	// The 'userinfo' part (username:password@)
	size_t indexOfAt = copy.find_first_of('@');
	size_t indexQuery= copy.find_first_of('?');
	if (indexOfAt == string::npos)
	{
		// No userinfo present
	} 
	else
	{
		// Only @ts present before query are valid userinfo parameters
		if ((indexQuery == string::npos) || (indexOfAt < indexQuery))
		{
			// Userinfo present
			string userinfo = copy.substr(0,indexOfAt);			
			size_t separatorIndex = userinfo.find_first_of(':');
			if (separatorIndex == string::npos)
			{
				// No password present, only username
				iUsername = userinfo;
			} else
			{
				iUsername = userinfo.substr(0,separatorIndex);
				iPassword = userinfo.substr(separatorIndex+1);
			}
			copy.erase(0, userinfo.length()+1); // +1 for the @ sign
		}
	}
	// The 'domain' part
	size_t slashIndex = copy.find_first_of('/');
	if (slashIndex == string::npos)
	{
		// The domain is probably the rest of the URL. Might contain port
		
		size_t queryIndex = copy.find_first_of('?');
		string domainPlusPort;
		if (queryIndex == string::npos)
		{
			// No query behind the URL. There still might be 'fragment'
			size_t fragmentIndex = copy.find_first_of('#');
			if (fragmentIndex == string::npos)
			{
				domainPlusPort = copy;
			} 
                        else
			{		
				domainPlusPort = copy.substr(0,fragmentIndex);
			}			
		} else
		{
			// Query present.
			domainPlusPort = copy.substr(0,queryIndex);
		}
		FindPort(domainPlusPort);
		copy.erase(0,domainPlusPort.length());
	}
	else
	{
		string domainPlusPort = copy.substr(0,slashIndex);
		FindPort(domainPlusPort);
		copy.erase(0,domainPlusPort.length());
	}
	// Now, it remains to separate path, query and fragment
	size_t queryIndex = copy.find_first_of('?');
	string pathPlusFilename;
	if (queryIndex == string::npos)
	{
		// there still might be fragment
		size_t fragIndex = copy.find_first_of('#');
		if (fragIndex == string::npos)
		{			
			pathPlusFilename = copy;			
		} else
		{
			pathPlusFilename = copy.substr(0,fragIndex);
		}
	}
	else
	{
		pathPlusFilename = copy.substr(0,queryIndex);
	}
	copy.erase(0,pathPlusFilename.length());	
	if (!pathPlusFilename.empty())
	{
		iFullPath=pathPlusFilename;
		// Now, process the path plus filename
		// if the path ends with /, filename is empty
		if (pathPlusFilename[pathPlusFilename.length()-1] == '/')
		{
			iPath = pathPlusFilename;
		}
		// filename present? We will detect the last .
		else
		{
			size_t dotIndex = pathPlusFilename.find_last_of('.');		
			if (dotIndex != string::npos)
			{
				iExtension = pathPlusFilename.substr(dotIndex+1);
				pathPlusFilename.erase(dotIndex);
			}
			size_t lastSlashIndex = pathPlusFilename.find_last_of('/');
			if (lastSlashIndex != string::npos)
			{
				iFilename = pathPlusFilename.substr(lastSlashIndex+1);
				pathPlusFilename.erase(lastSlashIndex);
			}
			iPath = pathPlusFilename;
		}
	}
	if (!copy.empty())
	{
		printf("URL.Parse: query and fragment present? Cannot parse them yet");
		iResult = ECannotProcessQueryAndFragment;
	}
}

void   URL::FindPort(const string& aDomain)
{
	size_t portIndex = aDomain.find_first_of(':');
	if (portIndex == string::npos)
	{
		iDomain = aDomain;
	} else
	{
		iDomain = aDomain.substr(0,portIndex);
		string portText = aDomain.substr(portIndex+1);
		iPort   = atoi(portText.c_str());
	}

}

TParsingResult URL::GetParsingResult() const
{
	return iResult;
}

bool   URL::IsOk() const
{
	return (iResult == EOk);
}

void   URL::Reset()
{
	iScheme.clear();
	iUsername.clear();
	iPassword.clear();
	iDomain.clear();
	iPort = KNoPortDetermined;
	iPath.clear();
	iFilename.clear();
	iExtension.clear();
	iParamString.clear();
	iQuery.clear();
	iFragment.clear();	
	iResult = EOk;
}

void   URL::PrintExtended() const
{
	printf("--------------------\n");
	printf("Scheme	 : %s\n",iScheme.c_str());
	printf("Username : %s\n",iUsername.c_str());
	printf("Password : %s\n",iPassword.c_str());
	printf("Domain   : %s\n",iDomain.c_str());
	printf("Port     : %d\n",iPort);
	printf("Path     : %s\n",iPath.c_str());
	printf("Filename : %s\n",iFilename.c_str());
	printf("Extension: %s\n",iExtension.c_str());
	printf("FullPath : %s\n",iFullPath.c_str());
}

void   URL::SelfTest()
{
	URL* url;

	url = new URL("http://ion:popescu@worldview.com:8080/never.seen.php?query=3&login=5#anchor");
	url->PrintExtended();
	delete url;
	
	url = new URL("http://www.absurdare.com/aimyra.do#clock");
	url->PrintExtended();
	delete url;

	url = new URL("file:///usr/local/bin/");
	url->PrintExtended();
	delete url;

	url = new URL("http://grigiskes.lt#nuolaida");
	url->PrintExtended();
	delete url;

	url = new URL("http://kamikaze.jp:143");
	url->PrintExtended();
	delete url;

	url = new URL("smtp://user:password@smtp.user.com:143");
	url->PrintExtended();
	delete url;

	url = new URL("chybna)url_nemela*by+jit#rozparsovat");
	url->PrintExtended();
	delete url;

}

