#ifndef __ABSTRACT_PHASE_H__
#define __ABSTRACT_PHASE_H__

#include "abstract_factor_alg_parameters.h"
#include "phase_functor.h"
#include "time_container.h"
#include "statistic_container.h"
#include "const_return_codes.h"


///Abstract class for factorization algorithm phases
class APhase : public virtual MSerializable
{
private:

	//Members for RunningSerialization - they are initialized only if we use them
	///If all mutexes, .... are initialized
	bool	iRSerialInit;
	///Thread for running serialization
	pthread_t				iRSerialThread;
	///Running serialization thread is defined as JOINABLE - for portability
	pthread_attr_t			iRSerialThreadAttr;
	///If running serialization is used
	bool	iRSerialRunning;
	mutable pthread_mutex_t iRSerialRunningMutex;
	///If now is performed serialization
	bool	iRSerialRun;
	mutable pthread_mutex_t iRSerialRunMutex;
	///Main condition for signaling to the running serialization thread
	mutable pthread_cond_t	iRSerialRunCond;
	///Parameters used for running serialization
	CBaseParameters *		iRSerialParameters;


	

	///If functor field is initialized
	bool	iFunctorsInit;
	mutable pthread_mutex_t iFunctorsInitMutex;
	///Info mode turn on/off
	bool	iInfoMode;
	mutable pthread_mutex_t iInfoModeMutex;
	///Assertion mode turn on/off
	bool	iAssertionMode;
	mutable pthread_mutex_t iAssertionModeMutex;
	///If can be serialized during calculation
	bool	iAutoSerializable;
	mutable pthread_mutex_t iAutoSerializableMutex;
	///Identifier of factorization run - for serialization 
	std::string iSerialIdentifier;
	mutable pthread_mutex_t iSerialIdentifierMutex;
	///Directory used for serialization
	std::string iSerialMainDirectory;
	mutable pthread_mutex_t iSerialMainDirectoryMutex;
	///For this phase serialization
	std::string iSerialPhaseFullFileName;
	mutable pthread_mutex_t iSerialPhaseFullFileNameMutex;
	///For parameters serialization
	std::string iParametersFullFileName;
	mutable pthread_mutex_t iParametersFullFileNameMutex;

	// accepted functor return codes
	std::vector<int> iAcceptedReturnCodes;

	/*!
		Parameters for setup this phase.

		iParameters is private so we can control setting of this pointer. 
		We use inline getter and setter for speed up access.

		AFactorAlgParameters is thread safe so we have to protect only the pointer,
		but there is only one setting of this pointer so we don't use mutex fot it.
	*/
	AFactorAlgParameters*  iParameters;
	///Phase number (from enum)
	int iPhaseNumber;
	mutable pthread_mutex_t iPhaseNumberMutex;
	/*!
		Phase inner state (from enum which is specific for derived classes).
		Each state has at most one corresponding function, which sets this state.
	*/
	int iInnerState;
	mutable pthread_mutex_t iInnerStateMutex;

protected:
	/*!
		Field of functions which correspond to the inner state of the phase.
		We have to use functors because of inheritance and different methods used.

		There is always only one thread which can use this field so we don't need a mutex for it.
	*/
	std::vector<APhaseFunctor*> iPhaseFunctors;

	///For all time informations about run-time - creator must destroy it
	CTimeContainer * iTimes;
	///For all statistic informations - creator must destroy it
	CStatContainer * iStats;

private:
	/*!
		We want only one running RunPhase in tha same time. 
		Because of private inner state and its mutex, we need some method which can check 
		if the phase is running and if not then set up inner state correctly. But we don't know enum for 
		inner states so we just suppose that next state is the good one for us.
	*/
	int SetStarted();

	///We don't want to use copy-constructor
	APhase(const APhase& aOperand) {}

protected:

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	virtual int FillFunctorField() = 0;

	///Dispose functors - call only from destructor
	int DisposeFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	virtual int Reset();

	///Dispose all mutexes - call only from destructor
	virtual int DisposeMutexes();

	// Dispose accepted return values.
	void DisposeAcceptedReturnCodes();

	///Save phase's result
	virtual int SaveResult() = 0;

	///Init all parameters necessary for calculation - get them from iParameters
	virtual int InitParameters();

	// For inner state

	///Return TRUE if the phase is running - according to inner state 
	virtual bool Running() const;
	///Return number of not started state
	virtual int InnerStateNotStarted() const;
	///Return number of finished state
	virtual int InnerStateFinished() const = 0;

	// Running Serialization methods

	///Initialize all used members and create serialization thread
	int StartRunningSerialization(const CBaseParameters & aParameters);
	///Dispose all used members and joint serialization thread
	int EndRunningSerialization();
	///Signal to perform serialization
	int RunRunningSerialization();
	///Prepare members for serialization which is ran from other thread - virtual for future changes
	virtual int PrepareRunningSerialization();

private:
	///Thread method for running serialization
	static void * RunningSerialization(void * aParam);

public:

	///Constructor  - initialization of mutexes
	APhase();

	///Virtual destructor for later use
	virtual ~APhase() {}

	///Start method for phase
	virtual int RunPhase(AFactorAlgParameters*  aParameters);

	///Delete all used files according to enum cleaning_types
	virtual int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	///If the phase has done all work
	bool Finished() const;

	// manage and check return codes
	void AcceptReturnCode(const int rc);
	bool CheckReturnCode(const int rc);

	// Getters and Setters
public:
	bool Get_InfoMode() const;
	int  Set_InfoMode(bool aInfoMode);
	bool Get_AssertionMode() const;
	int  Set_AssertionMode(bool aAssertionMode);
	bool Get_AutoSerializable() const;
	int  Get_PhaseNbr() const;
	std::string Get_SerialIdentifier() const;
	std::string Get_SerialMainDirectory() const;
	std::string Get_SerialPhaseFullFileName() const;
	std::string Get_ParametersFullFileName() const;
	AFactorAlgParameters* Get_Parameters() const { return this->iParameters; }
protected:
	bool Get_FunctorsInit() const;
	int  Set_FunctorsInit(bool aFunctorsInit);
	int  Set_AutoSerializable(bool aAutoSerializable);
	int  Get_InnerState() const;
	int  Set_InnerState(int aInnerState);
	int  Set_PhaseNbr(int aPhaseNumber);
	bool Get_RSerialRunning() const;
	int  Set_Parameters(AFactorAlgParameters* aParameters) { this->iParameters = aParameters; return ConstRC::Ok; }
private:
	int  Set_SerialIdentifier(std::string aSerialIdentifier);
	int  Set_SerialMainDirectory(std::string aSerialMainDirectory);
	int  Set_SerialPhaseFullFileName(std::string aSerialPhaseFullFileName);
	int  Set_ParametersFullFileName(std::string aParametersFullFileName);
	int  Set_RSerialRunning(bool aRunning);

};

#endif // __ABSTRACT_PHASE_H__
