#ifndef __ABSTRACT_POLY_SELECTION_PHASE_H__
#define __ABSTRACT_POLY_SELECTION_PHASE_H__

#include "abstract_phase.h"
#include "polynom_class.h"

///Abstract class for polynomials selection phases
class APolyPhase : public APhase
{
protected:
	long	iSelfInitializingDelay;
	long	iSubintervalsCount;		///<The number of subintervals of the half circle
	mpf_t	iSubintervalLength;		///<The length of the subintervals of the half circle
	long	iFactorialToPrecision;	///<The biggest number such that 1/factorial fits into the precision
	mpf_t	iPiHalf;				///<pi/2
	mpf_t	iLogTwo;				///<ln2
	mpf_t** iRhoTaylor;				///<The Taylor expansion of the rho function
	mpf_t*	iSin;					///<The sin values over the half circle
	mpf_t*	iCos;					///<The cos values over the half circle
	mpf_t	iRhoMax;				///<The maximal value of rho

	///Number to factor
	mpz_t	iN;
	///Common root of polynomials mod N
	mpz_t	iRoot;
	///Number of candidate polynomials for finding the best
	long	iCandidatesCount;
	///Number of tries for computing content of polynomials
	int		iContentTries;
	///Upper bound for range in which is content of polynomial computing
	long	iContentBound;
	///The bound of primes for which alpha estimate is done and the content precisely counted
	int		iAlphaPrimeBound;
	///The bound of primes for which the content is estimated when counted alpha
	int		iAlphaSmallerPrimeBound;
	///The bound where to abort the alpha estimate computation
	double	iAlphaAbort;
	///<The bound of alpha to be accepted when searching non-skewed polynomials
	double	iAlphaBound;

	mutable mpf_t	iAuxMpfCo;
	mutable mpf_t	iAuxMpfPow;
	mutable mpf_t	iAuxMpfC;
	mutable mpf_t	iAuxMpfS;
	

protected:

	///Computes content p of the homogenized polynomial in area iContentBound x iContentBound  
	int ContPHomPoly(mpf_t aResult, Polynomial *aPoly, unsigned int aPrime) const;
	///Computes content p of the polynomial in range 0 ... iContentBound
	int ContPPoly(mpf_t aResult, Polynomial *aPoly, unsigned int aPrime) const;
	///Computes the alpha of the homogenized polynomial aPoly
	int AlphaHomPoly(mpf_t aResult, long * aRoots, Polynomial *aPoly) const;
	///Computes the alpha of the polynomial aPoly
	int AlphaPoly(mpf_t aResult, long * aRoots, Polynomial *aPoly) const;
	///Computes a quick estimate of the alpha of the homogenized polynomial aPoly
	int AlphaEstimateHomPoly(mpf_t aResult, Polynomial *aPoly) const;
	int AlphaEstimateHomPoly(mpf_t aResult, long * aRoots, Polynomial *aPoly) const;
	///Computes a quick estimate of the alpha of the polynomial aPoly
	int AlphaEstimatePoly(mpf_t aResult, long * aRoots, Polynomial *aPoly) const;

	int Goniometric(mpf_t aSin, mpf_t aCos, mpf_t aAngle);
	int GenerateSinCos();
	int GenerateRho();
	int Rho(mpf_t aResult, mpf_t aValue);

	int EvaluateAngle(mpf_t aResult, Polynomial *aPolynomial, mpf_t aAngle);
	int EvaluateAngle(mpf_t aResult, Polynomial *aPolynomial, long aInterval);
	int EvaluateEllipseAngle(mpf_t aResult, Polynomial *aPolynomial, long aInterval, mpf_t aAxis1, mpf_t aAxis2);

	// Phase methods

	///Init all parameters necessary for calculation - get from iParameters
	int InitParameters();
	///Dispose mutexes in current class - call only from destructor
	int DisposeMutexes();
	///Reset all resources which was used and prepare for new start.
	int Reset();

	///Dispose mpz_t and mpf_t members in current class - call only from destructor
	virtual int DisposeGMP();

	///Return number of finished state
	int InnerStateFinished() const;

public:

	APolyPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	// Getters and Setters
protected:

	void Set_Precision(unsigned long aPrecision);
	void Set_SubintervalsCount(long aCount);

	// Getters and Setters
	//void Get_N(mpz_t aResult) const;
	//void Get_Root(mpz_t aResult) const;
	//long Get_SievingIntervalLength() const { return this->iSievingIntervalLength; }
	//long Get_CandidatePolyCount() const { this->iCandidatesCount; }
	//long Get_ContentBound() const { this->iContentBound; }
	//int Get_ContentTries() const { this->iContentTries; }
	//int Get_AlphaPrimeBound() const { this->iAlphaPrimeBound; }
	//int Get_AlphaSmallerPrimeBound() const { this->iAlphaSmallerPrimeBound; }
	//double Get_AlphaAbort() const { this->iAlphaAbort; }

};

#endif // __ABSTRACT_POLY_SELECTION_PHASE_H__
