#include "abstract_rel_processing_phase.h"

#include <cstring>
#include "relation_part.h"
#include "const_rel_processing_phase.h"
#include "const_parameter_names.h"
#include "const_hashtables.h"
#include "test_suite.h"
#include "log.h"


using namespace std;

ARelProcessingPhase::ARelProcessingPhase()
{
	this->iRelationMatrix = NULL;

	int rc;
	if ((rc = pthread_mutex_init(&(this->iRelationMatrixMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","ARelProcessingPhase",rc,0,"Cannot initialize mutex.");
	}
}
int ARelProcessingPhase::DisposeMutexes()
{
	int rc;
	if ((rc = pthread_mutex_destroy(&(this->iRelationMatrixMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ALinearPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}

	return ConstRC::Ok;
}
int ARelProcessingPhase::Reset()
{
	pthread_mutex_lock(&(this->iRelationMatrixMutex));
	if (this->iRelationMatrix != NULL)
	{
		delete this->iRelationMatrix;
	}
	this->iRelationMatrix = NULL;
	pthread_mutex_unlock(&(this->iRelationMatrixMutex));

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iVariationsInfo[i].Reset();
	}

	return ConstRC::Ok;
}
int ARelProcessingPhase::InnerStateFinished() const
{
	return (int)ERPPhaseFinished;
}
int ARelProcessingPhase::CleanUp(AFactorAlgParameters *aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	string identifier;
	string directory = "";
	string file;

	if (aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","CleanUp",0,0,"Didn't find serial identifier.");
		return ConstRC::GeneralError;
	}
	aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, directory, (int)EMain);

	switch (aType)
	{
		case ECleanAllIfNotError:
			if (aError)
			{
				break;
			}
		case ECleanAll:
		case ECleanAllExcCF:
		case ECleanAllExcCFRF:
			// Relation matrix full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::RELATION_MATRIX_FULL_FILENAME, aPhaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::RELATION_MATRIX_FULL_FILENAME, file, aPhaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::RELATION_MATRIX_FILENAME, 
					ConstPN::RELATION_MATRIX_DIRECTORY, ConstPN::RELATION_MATRIX_COMPRESSION, directory,
					ConstRPPhase::RELATION_MATRIX_FILENAME, identifier, aPhaseNBR);
			}
			remove(file.c_str());
			break;
		default:
			break;
	}

	return APhase::CleanUp(aParameters, aPhaseNBR, aType, aError);
}
int ARelProcessingPhase::InitParameters()
{
	int rc = APhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	AFactorAlgParameters * param = Get_Parameters();
	if (param == NULL)
	{
		return ConstRC::GeneralError;
	}

	string auxStr;
	int auxInt;
	int variationType = (int)ConstRPPhase::VARIATIONS_TYPE;
	param->GetIntegerPhaseParameter(ConstPN::VARIATIONS_TYPE, variationType, variationType, (int)EMain);
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = Utils::ConvertToString(i + 1);
		param->GetIntegerPhaseParameter(ConstPN::VARIATIONS_TYPE + auxStr, variationType, auxInt, (int)EMain);
		this->iVariationsInfo[i].Type = (variations_types)auxInt;
	}

	param->GetIntegerPhaseParameter(ConstPN::RELATION_MATRIX_TYPE, (int)ConstRPPhase::RELATION_MATRIX_TYPE, auxInt, Get_PhaseNbr());
	this->iRelationMatrixType = (species_of_matrices)auxInt;

	// Relation matrix full filename
	if (param->ContainsPhaseParameter(ConstPN::RELATION_MATRIX_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::RELATION_MATRIX_FULL_FILENAME, this->iRelationMatrixFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iRelationMatrixFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iRelationMatrixFullFileName, ConstPN::RELATION_MATRIX_FILENAME, 
			ConstPN::RELATION_MATRIX_DIRECTORY, ConstPN::RELATION_MATRIX_COMPRESSION, Get_SerialMainDirectory(),
			ConstRPPhase::RELATION_MATRIX_FILENAME, Get_SerialIdentifier(), Get_PhaseNbr());
	}

	// Relations result full filename
	if (param->ContainsPhaseParameter(ConstPN::RELATIONS_RESULT_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::RELATIONS_RESULT_FULL_FILENAME, this->iRelationsResultFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iRelationsResultFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iRelationsResultFullFileName, ConstPN::RELATIONS_RESULT_FILENAME, 
			ConstPN::RELATIONS_RESULT_DIRECTORY, ConstPN::RELATIONS_RESULT_COMPRESSION, Get_SerialMainDirectory(),
			ConstRPPhase::RELATIONS_RESULT_FILENAME, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}

	return ConstRC::Ok;
}

/*!
	This method which fills the aQuadChars field with suitable prime ideals. 
*/
int ARelProcessingPhase::PrepareQuadraticCharacters(prime_ideal_for_legendre*& aQuadChars, int aQuadCharsCount, ARelation * aRelation, int aPartIndex, Polynomial * aThetaPoly, long aMaxPrime, string aSmoothRelFullFileName) 
{	
	if (aThetaPoly == NULL || aRelation == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters",0,0,"Some argument si NULL");
		return ConstRC::BadArgument;
	}
	
	aQuadChars = (prime_ideal_for_legendre*) Utils::AllocationMachine((void**) &aQuadChars, aQuadCharsCount *sizeof(prime_ideal_for_legendre));
	if (aQuadChars == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters",0,0,"Problem in allocation of the field of QC");
		return ConstRC::NotEnoughMemory;
	}

	// In order to be able to choose prime ideals not used in large prime variation we construct ordered array of prime ideals used in LPV
	// Construction follows.
	// In theory, it would be better to store values in eg. binary search tree but ATM I think it's not worth the effort 
	
	if (Get_InfoMode())
	{
		printf("\n\t\tFilling large prime ideals ....   ");
		fflush(stdout);
	}
	
	set<main_sieving_type> largePrimeIdeals;

	CBaseParameters paramSmooth;
	paramSmooth.AddParameter(ConstPN::SERIALIZATION_FULL_FILENAME, aSmoothRelFullFileName);

	FILE * smoothsFILE = fopen(aSmoothRelFullFileName.c_str(), FILE_MODE_READ);
	if (smoothsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters",0,0,"Could not open file with smooth relations.");
		return ConstRC::CouldNotOpenFile;
	}

	CDivisors * divisors = &aRelation->Get_Parts(aPartIndex)->Large;
	int counter = 0;
	long i;
	while (aRelation->ReadFromFile(smoothsFILE) == ConstRC::Ok) 
	{
		for (i = 0; i <= divisors->MaxElementIndex; i++)
		{
			largePrimeIdeals.insert(divisors->Elements[i].prime);
		}
		counter++;
		if (Get_InfoMode() && !(counter % 10)) 
		{
			printf("\r\t\tFilling large prime ideals ... %10d   ", counter);
			fflush(stdout);
		}
	}

	if (Get_InfoMode())
	{
		printf("\r\t\tFilling large prime ideals finished.                    \n");
		fflush(stdout);
	}

	fclose(smoothsFILE);
	smoothsFILE = NULL;
	
	mpz_t mprime;
	mpz_init(mprime);

	long* roots = NULL;
	roots = (long*) Utils::AllocationMachine((void**) &roots, DEFAULT_ROOTS_COUNT * sizeof(long));
	if (NULL == roots) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters ",0,0,"Could not allocate roots.");
		free(aQuadChars);
		aQuadChars = NULL;
		mpz_clear(mprime);
		return ConstRC::NotEnoughMemory;
	}

	int rootNr;
	unsigned int prime;
	mpz_set_si(mprime, aMaxPrime);

	for (int i = 0; i < aQuadCharsCount; i++)
	{										
		mpz_nextprime(mprime,mprime);
		prime = mpz_get_ui(mprime);							
		
		// We will test whether sieving polynomial has some roots mod p.
		if (Get_InfoMode())
		{
			gmp_printf("\r\t\tSolving poly mod %d  ", prime);
			fflush(stdout);
		}
		if (prime < 0)
		{
			CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters ",0,prime,"We try to count roots of the sieving polynomial mod a negative number.");
			free(aQuadChars);
			aQuadChars = NULL;
			mpz_clear(mprime);
			free(roots);
			roots = NULL;
			return ConstRC::GeneralError;
		}
		rootNr = aThetaPoly->RootsModPrimeGeneral(roots, prime);

		if ((rootNr > 0) && largePrimeIdeals.find(prime) == largePrimeIdeals.end()) 
		{				
			aQuadChars[i].prime = prime;
			aQuadChars[i].c_p = roots[0];
			//printf("\n >%i\t >%i\n",prime,roots[0]);
			if (Get_AssertionMode()) 
			{
				int result = aThetaPoly->EvaluateModPrime(roots[0], prime);
				if (result != 0) 
				{
					printf("Error determining root of polynomial mod %d. Fake root is %ld, result is %d. Total roots %d\n", prime, roots[0], result, rootNr);
				}					
			}
		}
		else 
		{
			i--;
		}
	}

	mpz_clear(mprime);

	free(roots);
	roots = NULL;

	return ConstRC::Ok;
}
/*!
	This method which fills the aQuadChars field with suitable prime ideals. 
*/
int ARelProcessingPhase::PrepareQuadraticCharacters(prime_ideal_for_legendre*& aQuadChars, int aQuadCharsCount, int aPartIndex, Polynomial * aThetaPoly, main_sieving_type aMaxPrime, CFrequencyHashtable & aTable) 
{	
	if (aThetaPoly == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters",0,0,"Some argument si NULL");
		return ConstRC::BadArgument;
	}
	
	aQuadChars = (prime_ideal_for_legendre*) Utils::AllocationMachine((void**) &aQuadChars, aQuadCharsCount *sizeof(prime_ideal_for_legendre));
	if (aQuadChars == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters",0,0,"Problem in allocation of the field of QC");
		return ConstRC::NotEnoughMemory;
	}

	mpz_t mprime;
	mpz_init(mprime);

	long* roots = NULL;
	roots = (long*) Utils::AllocationMachine((void**) &roots, DEFAULT_ROOTS_COUNT * sizeof(long));
	if (NULL == roots) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters ",0,0,"Could not allocate roots.");
		free(aQuadChars);
		aQuadChars = NULL;
		mpz_clear(mprime);
		return ConstRC::NotEnoughMemory;
	}

	if (Get_InfoMode())
	{
		printf("\n\t\tGenerating prime ideals for quadratic characters ....   ");
		fflush(stdout);
	}

	int rc;
	int rootNr;
	unsigned int prime;
	mpz_set_si(mprime, aMaxPrime);
	bool present = false;
	relation_index_type index;

	for (int i = 0; i < aQuadCharsCount; i++)
	{										
		mpz_nextprime(mprime,mprime);
		prime = mpz_get_ui(mprime);							
		
		// We will test whether sieving polynomial has some roots mod p.
		if (Get_InfoMode())
		{
			gmp_printf("\r\t\tSolving poly mod %d  ", prime);
			fflush(stdout);
		}
		if (prime < 0)
		{
			CLog::LogError(WHEREARG,"ARelProcessingPhase","PrepareQuadraticCharacters ",0,prime,"We try to count roots of the sieving polynomial mod a negative number.");
			free(aQuadChars);
			aQuadChars = NULL;
			mpz_clear(mprime);
			free(roots);
			roots = NULL;
			return ConstRC::GeneralError;
		}
		rootNr = aThetaPoly->RootsModPrimeGeneral(roots, prime);

		if (rootNr > 0)
		{
			/*
				Because prime ideal are not always identified by prime and root (c_p) (see filling of factor base) 
				we are looking for suitable primes (prime ideals) which are not generating any of the prime ideal used in factorization.
			*/
			rc = aTable.FindPrimeInHashtable(prime, present, index);
			if (rc != ConstRC::Ok)
			{
				free(aQuadChars);
				aQuadChars = NULL;
				mpz_clear(mprime);
				free(roots);
				roots = NULL;
				return ConstRC::GeneralError;
			}
			if (present == false)
			{				
				aQuadChars[i].prime = prime;
				aQuadChars[i].c_p = roots[0];
				//printf("\n >%i\t >%i",prime,roots[0]);
				if (Get_AssertionMode()) 
				{
					int result = aThetaPoly->EvaluateModPrime(roots[0], prime);
					if (result != 0) 
					{
						printf("Error determining root of polynomial mod %d. Fake root is %ld, result is %d. Total roots %d\n", prime, roots[0], result, rootNr);
					}					
				}
			}
			else
			{
				i--;
			}
		}
		else 
		{
			i--;
		}
	}

	mpz_clear(mprime);

	free(roots);
	roots = NULL;

	return ConstRC::Ok;
}
/*!
	This method computes the required amount of Legendre symbols for the supplied
	relation.
*/
int ARelProcessingPhase::ComputeLegendreSymbols(prime_ideal_for_legendre * aQuadChars, long aQuadCharsCount, Polynomial* aCandidatePoly, AbstractMatrix * aMatrix, int aIndexInMatrix, int aShift) 
{
	if (aQuadChars == NULL || aCandidatePoly == NULL || aMatrix == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","ComputeLegendreSymbols",0,0,"Something is NULL.");
		return ConstRC::BadArgument;
	}

	mpz_t temp, temp2, result;
	int legendre;
	//int legendre2;
	
	mpz_init(temp);
	mpz_init(temp2);
	mpz_init(result);

	//int ok = 0;
	//int notok = 0;	
	for (long i = 0; i < aQuadCharsCount; i++)
	{
		/*
			The function "Evaluate" does not accept two identical parameters.
			That is why we cannot call "Evaluate(temp, temp), but rather
			Evaluate(result, temp).

			This was a serious bug in the previous version of NFS. In such case,
			the polynomial was evaluated in 0 instead.

			Oddly enough, this worked sometimes; how is this possible?
		*/
		mpz_set_si(temp, aQuadChars[i].c_p);
		aCandidatePoly->Evaluate(result, temp);
		mpz_set_si(temp2, aQuadChars[i].prime);

		legendre = mpz_legendre(result, temp2);		
		//legendre2 = Utils::IncorrectLegendre(aQuadChars[i].c_p, aQuadChars[i].prime, aCandidatePoly);
		//if (legendre != legendre2) 
		//{
		//	//printf("L != L2 for %d[%d].\n", aQuadChars[i].prime, aQuadChars[i].c_p);
		//	notok++;
		//} 
		//else 
		//{
		//	ok++;
		//}

		if (-1 == legendre) 
		{
			aMatrix->PutOne(aIndexInMatrix, aShift + i);
		}
		else if (0 == legendre)
		{
			printf("\n!!\t Legendre symbol equals zero\t !!!");
			printf("\n-----------------------------------------------");
			printf("\n Prime = %i, C_p = %i:", aQuadChars[i].prime, aQuadChars[i].c_p);
			aCandidatePoly->PrintToScreen();
			gmp_printf("\n Temp1:%Zd, temp2:%Zd\n", result, temp2);
	
		}
	}	
	
	//printf("Not ConstRC::Ok: %d, ConstRC::Ok: %d.\n", notok, ok);

	mpz_clear(temp);
	mpz_clear(temp2);
	mpz_clear(result);

	return ConstRC::Ok;
}
int ARelProcessingPhase::SerializeQuadCharacters(const CBaseParameters &aParam, prime_ideal_for_legendre *aQuadChars, long aNumQuadChars)
{
	if (aQuadChars == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",0,0,"Quadratic characters are NULL");
		return ConstRC::BadArgument;
	}

	int compressionLevel;

	// get parameters for serialization - with default values
	aParam.GetIntegerParameter(ConstPN::COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, compressionLevel);
	
	// get file name with path
	string fileNameFull = MSerializable::CreateFileNameAuto(aParam, ConstFN::QUAD_CHARS);

	xmlTextWriterPtr writer = xmlNewTextWriterFilename(fileNameFull.c_str(), compressionLevel);
	// we write only start of document now
	int rc = xmlTextWriterStartDocument(writer, NULL, XML_ENCODING, NULL);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",rc,0,"Could not write the document start.");
		xmlFreeTextWriter(writer);
		return ConstRC::GeneralError;        	
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_QUADRATIC_CHARACTERS_NAME, mainTagName);

	rc = xmlTextWriterStartElement(writer, BAD_CAST mainTagName.c_str());
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",rc,0,"Could not write the main tag.");
		xmlFreeTextWriter(writer);
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// serialize number of quadratic characters
	rc = CXMLHelper::SerializeSignedLong(writer, XML_SERIAL_COUNT_NAME, aNumQuadChars);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	rc = xmlTextWriterStartElement(writer, BAD_CAST XML_SERIAL_ELEMENTS_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",rc,0,"Could not write the elements tag.");
		xmlFreeTextWriter(writer);
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// serialize quadratic characters of FB
	for (long i = 0; i < aNumQuadChars; i++)
	{
		rc = CXMLHelper::SerializePrimeIdealForLegendre(writer, XML_SERIAL_ELEMENT_NAME, aQuadChars[i]);
		if (rc != ConstRC::Ok)
		{
			xmlFreeTextWriter(writer);
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(writer); // XML_SERIAL_ELEMENTS_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",rc,0,"Could not end the elements tag.");
		xmlFreeTextWriter(writer);
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	rc = xmlTextWriterEndElement(writer); // mainTagName
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",rc,0,"Could not end the main tag.");
		xmlFreeTextWriter(writer);
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(writer,FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(writer);
		return rc;
	}

	// end the document
	rc = xmlTextWriterEndDocument(writer);

	xmlFreeTextWriter(writer);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","SerializeQuadCharacters",rc,0,"Could not close the document.");
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}
int ARelProcessingPhase::DeserializeQuadCharacters(const CBaseParameters &aParam, prime_ideal_for_legendre *&aQuadChars, long &aNumQuadChars)
{
	// get file name with path
	string fileNameFull = MSerializable::CreateFileNameAuto(aParam, ConstFN::QUAD_CHARS);

	xmlTextReaderPtr reader = xmlNewTextReaderFilename(fileNameFull.c_str());
	if (reader == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","DeserializeQuadCharacters",0,0,"Could not create xml reader.");
		return ConstRC::GeneralError;
	}
	int rc = CXMLHelper::MoveToNextNode(reader);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextReader(reader);
		if (rc == ConstRC::EndOfFile)
		{
			return ConstRC::Ok;
		}
		return rc;
	}
	
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_QUADRATIC_CHARACTERS_NAME, mainTagName);

	xmlChar* name = xmlTextReaderName(reader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","DeserializeQuadCharacters",0,0,"Deserialization main element name not found.");
		xmlFreeTextReader(reader);
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"ARelProcessingPhase","DeserializeQuadCharacters",0,0,"Main element name is wrong.");
		xmlFreeTextReader(reader);
		return ConstRC::GeneralError; 
	}

	aNumQuadChars = 0;

	int depth = xmlTextReaderDepth(reader);

	while ((rc = CXMLHelper::MoveToNextNode(reader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(reader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_COUNT_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(reader, "Deserialize count...", aNumQuadChars, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENTS_NAME))
		{
			if (aNumQuadChars > 0)
			{
				if ((aQuadChars = (prime_ideal_for_legendre*) Utils::AllocationMachine((void **) &(aQuadChars),aNumQuadChars*sizeof(prime_ideal_for_legendre))) == NULL) 
				{
					// Allocation of the quadratic characters has failed!
					CLog::LogError(WHEREARG,"ARelProcessingPhase","DeserializeQuadCharacters",aNumQuadChars*sizeof(prime_ideal_for_legendre),1,"Problem in allocation of quadratic characters.");
					rc = ConstRC::NotEnoughMemory;
					break;
				}
				memset(aQuadChars, 0, aNumQuadChars * sizeof(prime_ideal_for_legendre));

				int elementsDepth = xmlTextReaderDepth(reader);
				prime_ideal_for_legendre * current = aQuadChars;
				for (long i = 0; i < aNumQuadChars; i++)
				{
					if (CXMLHelper::MoveToNextNode(reader, elementsDepth) != ConstRC::Ok)
					{
						rc = ConstRC::GeneralError;
						break;
					}
					rc = CXMLHelper::DeserializePrimeIdealForLegendre(reader, "Deserialize quadratic character...", *current, FALSE);
					if (rc != ConstRC::Ok)
					{
						rc = ConstRC::GeneralError;
						break;
					}
					current++;
				}
			}
			continue;
		}
	}
	
	xmlFreeTextReader(reader);
	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int ARelProcessingPhase::DeleteRelationsArray(ARelation **&aRelations, long & aMaxIndex)
{
	if (aRelations != NULL)
	{
		for (long i = 0; i <= aMaxIndex; i++)
		{	
			if (aRelations[i] != NULL) 
			{
				delete aRelations[i];
				aRelations[i] = NULL;
			}
		}
		free(aRelations);
		aRelations = NULL;
		aMaxIndex = -1;
	}

	return ConstRC::Ok;
}
int ARelProcessingPhase::DetectSingletons(long aDetectionMaxIndex, const long *aDetectionField, int &aDeletionMaxIndex, int &aDeletionAllocated, long *&aDeletionField)
{
	//int twice = 0;
	//int zero = 0;
	for (long j = 0; j <= aDetectionMaxIndex; j++) 
	{
		//if (integral_detection_field[j] == NOT_YET) 
		//{
		//	zero++;
		//}
		//if (integral_detection_field[j] == (signed) TWICE) 
		//{
		//	twice++;
		//}
		if ((aDetectionField[j] != ConstHT::NOT_YET) && (aDetectionField[j] != ConstHT::TWICE)) 
		{
			aDeletionMaxIndex++;
			if (aDeletionMaxIndex == aDeletionAllocated) 
			{
				// Reallocation 
				aDeletionAllocated += ConstHT::IMPLICIT_LIST_FOR_DELETION_RESIZE;
				aDeletionField = (long*) Utils::AllocationMachine((void**) &aDeletionField,(aDeletionAllocated)*sizeof(long));					
				if (aDeletionField == NULL)
				{
					CLog::LogError(WHEREARG,"ARelProcessingPhase","DetectSingletons",0,0,"Not enough memory to allocate data space.");
					aDeletionMaxIndex = -1;
					aDeletionAllocated = 0;
					return ConstRC::NotEnoughMemory;
				}
			}
			aDeletionField[aDeletionMaxIndex] = aDetectionField[j];
		}
	}
	//printf("Vysledek pro integral: zero %d, twice %d, add %d, max %d, diff = %d.\n",zero,twice,zero+twice,aDetectionMaxIndex+1,aDetectionMaxIndex+1-(zero+twice));
		
	return ConstRC::Ok;
}
int ARelProcessingPhase::ControlSingletons(CDivisors &aDivisors, unsigned long aIndex, CFBInfo & aFB, long *aDetectionField, int &aSingletons)
{
	long index;
	long max = aDivisors.MaxElementIndex;
	nfs_sieving_element * element = aDivisors.Elements;
	for (long j = 0; j <= max; j++, element++) 
	{
		if (element->exponent % 2) 
		{					
			index = aFB.FindIndexInFB(element);
			if (index == -1) 
			{
				CLog::LogError(WHEREARG,"ARelProcessingPhase","ControlSingletons",0,0,"we were not able to find prime in integral factor base.");
				return ConstRC::GeneralError;
			}
			if (aDetectionField[index] == ConstHT::NOT_YET) 
			{
				aSingletons++;
				aDetectionField[index] = aIndex;
			} 
			else if ((aDetectionField[index] != ConstHT::NOT_YET) && (aDetectionField[index] != ConstHT::TWICE)) 
			{
				aSingletons--;
				aDetectionField[index] = ConstHT::TWICE;
			}
		}
	}

	return ConstRC::Ok;
}
AbstractMatrix* ARelProcessingPhase::Get_RelationMatrix() const
{
	AbstractMatrix* ret = NULL;

	// we ignore if the phase is running
	pthread_mutex_lock(&(this->iRelationMatrixMutex));
	ret = this->iRelationMatrix;
	pthread_mutex_unlock(&(this->iRelationMatrixMutex));

	return ret;
}
int             ARelProcessingPhase::Set_RelationMatrix(AbstractMatrix* aRelationMatrix)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iRelationMatrixMutex));
		if (this->iRelationMatrix != NULL)
		{
			delete this->iRelationMatrix;
		}
		this->iRelationMatrix = aRelationMatrix;
		pthread_mutex_unlock(&(this->iRelationMatrixMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
