#ifndef __ABSTRACT_REL_PROCESSING_PHASE_H__
#define __ABSTRACT_REL_PROCESSING_PHASE_H__

#include "abstract_phase.h"
#include "structures.h"
#include "nfs_relation.h"
#include <vector>
#include "abstract_matrix.h"
#include "factor_base_info.h"
#include "frequency_hashtable.h"

///Abstract class for relation processing phases
class ARelProcessingPhase : public APhase
{
private:
	///Matrix created from relations
	AbstractMatrix * iRelationMatrix;
	mutable pthread_mutex_t iRelationMatrixMutex;

protected:
	std::string iRelationMatrixFullFileName;
	species_of_matrices iRelationMatrixType;

	///Fullname of file with result relations - these relations are used in next phases
	std::string iRelationsResultFullFileName;

	/*!
		Info about variations used. We need only few things from the list below.
		We use two "number fields" - (we can think about integral part as a number field)
		VariationInfo contains these things:
		 - variations mode = EDoNotUseVariations, ELargePrimeV, EDoubleLargePrimeV, ETrippleLargePrimeV
		 - factorization algorithm for factorizing remainder after FB factorization
		 - bounds - max remainder, min remainder, max factor
		 - leading coeff of the sieving polynomial - just for c_p calculation
	*/
	CVariationsInfo	iVariationsInfo[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

protected:
	virtual int CreateRelationMatrix() = 0;

	///Aux method for deleting complicated structure
	int DeleteRelationsArray(ARelation **& aRelations, long & aMaxIndex);

	///Allocate and generate quadratic characters
	int PrepareQuadraticCharacters(prime_ideal_for_legendre*& aQuadChars, int aQuadCharsCount, ARelation * aRelation, int aPartIndex, Polynomial * aThetaPoly, long aMaxPrime, std::string aSmoothRelFullFileName);
	int PrepareQuadraticCharacters(prime_ideal_for_legendre*& aQuadChars, int aQuadCharsCount, int aPartIndex, Polynomial * aThetaPoly, main_sieving_type aMaxPrime, CFrequencyHashtable & aTable);
	///Compute legendre symbols for a relation represented by the polynomial
	int ComputeLegendreSymbols(prime_ideal_for_legendre* aQuadChars, long aQuadCharsCount, Polynomial* aCandidatePoly, AbstractMatrix * aMatrix, int aIndexInMatrix, int aShift);

	int InitParameters();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();
	int DisposeMutexes();

	///Return number of finished state
	int InnerStateFinished() const;

public:
	ARelProcessingPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	// Getters and Setters
protected:
	AbstractMatrix * Get_RelationMatrix() const;
	int              Set_RelationMatrix(AbstractMatrix * aRelationMatrix);

public:
	// Methods for hashtables
	static int ControlSingletons(CDivisors &aDivisors, unsigned long aIndex, CFBInfo & aFB, long * aDetectionField, int & aSingletons);
	static int DetectSingletons(long aDetectionMaxIndex, const long* aDetectionField, int& aDeletionMaxIndex, int& aDeletionAllocated, long*& aDeletionField);

	static int SerializeQuadCharacters(const CBaseParameters &aParam, prime_ideal_for_legendre * aQuadChars, long aNumQuadChars);
	static int DeserializeQuadCharacters(const CBaseParameters &aParam, prime_ideal_for_legendre *& aQuadChars, long & aNumQuadChars);
};

#endif // __ABSTRACT_REL_PROCESSING_PHASE_H__

