#include "abstract_relation.h"

#include "base_parameters.h"
#include "integer_class.h"
#include "const_return_codes.h"
#include "nfs_utils.h"
#include "log.h"


using namespace std;

ARelation::ARelation()
{
	this->iClassType = ERelAbstract;
	this->iNumberOfRelations = 1;
	this->iParts = NULL;
	this->iPartsCount = 0;

	mpz_init(this->iAuxMpz1);
	mpz_init(this->iAuxMpz2);
}
void ARelation::DisposeMpz()
{
	mpz_clear(this->iAuxMpz1);
	mpz_clear(this->iAuxMpz2);
}

int ARelation::Sort()
{
	if (this->iParts != NULL)
	{
		CRelationPart * part = this->iParts;
		for (int i = 0; i < this->iPartsCount; i++, part++)
		{
			qsort(part->Main.Elements, 
				part->Main.MaxElementIndex + 1,
				sizeof(nfs_sieving_element),
				&NFSUtils::CompareSievingElements);

			qsort(part->Large.Elements, 
				part->Large.MaxElementIndex + 1,
				sizeof(nfs_sieving_element),
				&NFSUtils::CompareSievingElements);
		}
	}

	return ConstRC::Ok;
}

int ARelation::CombineWithRelation(ARelation *aOperand, int aThetaPolyCount, Polynomial ** aThetaPolys)
{
	if (this->iParts != NULL && aOperand->iParts != NULL && 
		aOperand->iPartsCount == this->iPartsCount && 
		aThetaPolys != NULL && aThetaPolyCount == this->iPartsCount)
	{
		int rc;
		CRelationPart * part = this->iParts;
		CRelationPart * opart = aOperand->iParts;
		for (int i = 0; i < this->iPartsCount; i++, part++, opart++, aThetaPolys++)
		{
			rc = part->CombineWithPart(*opart, *aThetaPolys);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		this->iNumberOfRelations += aOperand->iNumberOfRelations;

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"ARelation","CombineWithRelation",0,0,"Operands inconsistency.");				
		return ConstRC::GeneralError;
	}
}
int ARelation::CombineWithRelation(ARelation *aOperand)
{
	if (this->iParts != NULL && aOperand->iParts != NULL && aOperand->iPartsCount == this->iPartsCount)
	{
		int rc;
		CRelationPart * part = this->iParts;
		CRelationPart * opart = aOperand->iParts;
		for (int i = 0; i < this->iPartsCount; i++, part++, opart++)
		{
			rc = part->CombineWithPart(*opart);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		this->iNumberOfRelations += aOperand->iNumberOfRelations;

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"ARelation","CombineWithRelation",0,0,"Operands inconsistency.");				
		return ConstRC::GeneralError;
	}
}
int ARelation::ControlExponents()
{
	if (this->iParts != NULL)
	{
		int rc;
		CRelationPart * part = this->iParts;
		for (int i = 0; i < this->iPartsCount; i++, part++)
		{
			rc = part->Main.ControlExponents();
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = part->Large.ControlExponents();
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}

	return ConstRC::Ok;
}
bool ARelation::Equals(const ARelation & aOperand) const
{
	if (this->iClassType != aOperand.iClassType) return false;
	if (this->Index != aOperand.Index) return false;
	if (this->Type != aOperand.Type) return false;
	if (this->iNumberOfRelations != aOperand.iNumberOfRelations) return false;
	if (this->iPartsCount != aOperand.iPartsCount) return false;

	if (this->iParts == NULL && aOperand.iParts == NULL) return true;
	if (this->iParts == NULL && aOperand.iParts != NULL) return false;
	if (this->iParts != NULL && aOperand.iParts == NULL) return false;

	for (int i = 0; i < this->iPartsCount; i++)
	{
		if (this->iParts[i].Equals(aOperand.iParts[i]) == false) return false;
	}

	return true;
}
int ARelation::Reset()
{
	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].Reset();
	}

	this->Type = 0;
	this->Index = 0;
	this->iNumberOfRelations = 1;

	return ConstRC::Ok;
}
main_sieving_type ARelation::GetRoot(main_sieving_type aA, main_sieving_type aB, main_sieving_type aPrime)
{
	// root = (-a)/b mod p
	main_sieving_type root;
	mpz_set_si(this->iAuxMpz1, aB);
	mpz_set_si(this->iAuxMpz2, aPrime);
	mpz_invert(this->iAuxMpz1, this->iAuxMpz1, this->iAuxMpz2);							

	root = mpz_get_si(this->iAuxMpz1);
	root = CIntegerClass::Mulm(root, aA, aPrime);
	root = aPrime - root;
	if (root < 0) 
	{
		root += aPrime;
	}
	else if (root >= aPrime) 
	{	
		root -= aPrime;
	}

	return root;
}
int ARelation::SerializeParts(xmlTextWriterPtr &aWriter) const
{
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_PARTS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"ARelation","SerializeParts",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }

	rc = xmlTextWriterWriteFormatAttribute(aWriter, BAD_CAST XML_SERIAL_PARTS_COUNT_ATTR,"%d",this->iPartsCount);	
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelation","SerializeParts",rc,0,"Could not write the parts count.");
        return ConstRC::GeneralError;
    }

	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters param;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].Serialize(param, aWriter);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"ARelation","SerializeParts",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;  
}

int ARelation::DeserializeParts(xmlTextReaderPtr &aReader)
{
	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"ARelation","DeserializeParts",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_PARTS_NAME) == 0)
	{
		CLog::LogError(WHEREARG,"ARelation","DeserializeParts",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int count;
	xmlChar* attr = xmlTextReaderGetAttribute(aReader, BAD_CAST XML_SERIAL_PARTS_COUNT_ATTR);  
	if (!attr)
	{
		CLog::LogError(WHEREARG,"ARelation","DeserializeParts",0,0,"Parts count not found, aborting");			
		return ConstRC::GeneralError;	
	}
	rc = CXMLHelper::ParseDecimalInt(attr, count);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ARelation","DeserializeParts",0,0,"Could not parse int");			
		xmlFree(attr);
		return ConstRC::GeneralError;	
	}
	xmlFree(attr);
	attr = NULL;

	if (count != this->iPartsCount)
	{
		CLog::LogError(WHEREARG,"ARelation","DeserializeParts",0,0,"Parts count incorrect, aborting");			
		return ConstRC::GeneralError;
	}

	if (this->iParts == NULL)
	{
		CLog::LogError(WHEREARG,"ARelation","DeserializeParts",0,0,"Parts don't exist, aborting");			
		return ConstRC::GeneralError;
	}

	CBaseParameters param;
	// int depth = xmlTextReaderDepth(aReader);

	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = CXMLHelper::MoveToNextNode(aReader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
			{
				return rc;
			}
			return ConstRC::Ok;
		}

		rc = this->iParts[i].Deserialize(param, aReader);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int ARelation::StartRelationsSerialization(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter)
{
	if (aWriter == NULL)
	{
		int compressionLevel;

		// get parameters for serialization - with default values
		aParam.GetIntegerParameter(PARAM_COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, compressionLevel);
		
		// get file name with path
		string fileNameFull = MSerializable::CreateFileNameAuto(aParam, FILE_NAME_RELATIONS);

		aWriter = xmlNewTextWriterFilename(fileNameFull.c_str(), compressionLevel);
	}

	// we write only start of document now
	int rc = xmlTextWriterStartDocument(aWriter, NULL, XML_ENCODING, NULL);
    if (rc < 0) 
	{
	    CLog::LogError(WHEREARG,"ARelation","StartRelationsSerialization",rc,0,"Could not write the document start.");
		xmlFreeTextWriter(aWriter);
		aWriter = NULL;
		return ConstRC::GeneralError;        	
    }

	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_MAIN_ELEMENT_NAME, XML_SERIAL_ELEMENTS_NAME, mainTagName);

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"ARelation","StartRelationsSerialization",rc,0,"Could not write the main tag.");
    	xmlFreeTextWriter(aWriter);
		aWriter = NULL;
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(aWriter);
		aWriter = NULL;
		return rc;
	}

	return ConstRC::Ok;
}
int ARelation::EndRelationsSerialization(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter)
{
	if (aWriter != NULL)
	{
		int rc = xmlTextWriterEndElement(aWriter); // mainTagName
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"ARelation","EndRelationsSerialization",rc,0,"Could not end the main tag.");
			xmlFreeTextWriter(aWriter);
			aWriter = NULL;
			return ConstRC::GeneralError;
		}

		rc = CXMLHelper::NewLine(aWriter,false);
		if (rc != ConstRC::Ok)
		{
			xmlFreeTextWriter(aWriter);
			aWriter = NULL;
			return rc;
		}

		// end the document
		rc = xmlTextWriterEndDocument(aWriter);

		xmlFreeTextWriter(aWriter);
		aWriter = NULL;
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"ARelation","EndRelationsSerialization",rc,0,"Could not close the document.");
			return ConstRC::GeneralError;
		}
	}

	return ConstRC::Ok;
}
int ARelation::AppendToRelationsSerialization(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter)
{
	xmlTextReaderPtr reader = NULL;
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_MAIN_ELEMENT_NAME, XML_SERIAL_ELEMENTS_NAME, mainTagName);

	// may be main element is 'elements'
	int rc = MSerializable::PrepareXMLReader(aParam, mainTagName, reader);
	if (rc != ConstRC::Ok)
	{
		if (reader != NULL)
		{
			xmlFreeTextReader(reader);
		}
		return rc;
	}

	xmlChar * inner = xmlTextReaderReadInnerXml(reader);
	xmlFreeTextReader(reader);
	reader = NULL;

	ARelation::StartRelationsSerialization(aParam, aWriter);
	rc = xmlTextWriterWriteString(aWriter, inner);
	xmlFree(inner);
	if (rc < 0)
	{
		CLog::LogError(WHEREARG,"ARelation","AppendToRelationsSerialization",rc,0,"Could not write string.");
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}
int ARelation::ReadRelationSerialization(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_MAIN_ELEMENT_NAME, XML_SERIAL_ELEMENTS_NAME, mainTagName);

	// may be main element is 'elements'
	int rc = MSerializable::PrepareXMLReader(aParam, mainTagName, aReader);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// next element is already first relation
	rc = CXMLHelper::MoveToNextNode(aReader);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextReader(aReader);
		aReader = NULL;
		return rc;
	}

	return ConstRC::Ok;
}
