#ifndef __ABSTRACT_RELATION_H__
#define __ABSTRACT_RELATION_H__

#include "mserializable.h"
#include "relation_part.h"
#include "integral_domain.h"
#include "const_debug.h"
#include "enums.h"
#include "xml_helper.h"
/*!
 * \brief Abstract relation class
 */

class ARelation : public virtual MSerializable
{
private:
	relation_types iClassType;

protected:
	///Relation parts (integral part, algebraic part...) - field of CRelationPart
	CRelationPart * iParts;
	///Length of field iParts
	int iPartsCount;
	///Number of relations used for building this relation - for large prime variations
	int iNumberOfRelations;

	mpz_t iAuxMpz1;
	mpz_t iAuxMpz2;

protected:
	// De/Serialization methods
	int SerializeParts(xmlTextWriterPtr& aWriter) const;
	int DeserializeParts(xmlTextReaderPtr& aReader);

	///Dispose used mpz members
	virtual void DisposeMpz();

	main_sieving_type GetRoot(main_sieving_type aA, main_sieving_type aB, main_sieving_type aPrime);

public:
	unsigned long Index;
	/*!
		Type of the relation - it is combination of the part's types
		Type = b1.b2.b3.b4 (bX is byte)
		b1 is the fourth part, b2 is the third part, b3 is the second part and b4 is the first part
	*/
	unsigned int Type;

// CONSTANTS
	///Setup for mpz_probab_prime_p
	static const int PROBAB_PRIME_DEGREE		= 25;
	///Max number of large primes - for regular relations
	static const int MAX_LARGE_PRIMES			= 3;
	///Max number of main divisors
	static const int DEFAULT_DIVISORS_ALLOCATE	= 16;
	///Type of relation is combination of the part's types - so we need to shift part type values
	static const int PART_TYPE_SHIFT			= 8;
	static const int PART_TYPE_MASK				= 0xff;
	///Maximum number of supported relation parts
	static const int MAX_SUPPORTED_PARTS		= 4;
	
	static const unsigned int P1_SMOOTH		= 0x1;
	static const unsigned int P1_LARGE		= 0x2;
	static const unsigned int P1_DOUBLE_LARGE	= 0x4;
	static const unsigned int P1_TRIPPLE_LARGE	= 0x8;
	static const unsigned int P1_PARTIAL		= 0xe;
	static const unsigned int P1_TOO_LARGE		= 0x80;
	static const unsigned int P2_SMOOTH		= 0x100;
	static const unsigned int P2_LARGE		= 0x200;
	static const unsigned int P2_DOUBLE_LARGE	= 0x400;
	static const unsigned int P2_TRIPPLE_LARGE	= 0x800;
	static const unsigned int P2_PARTIAL		= 0xe00;
	static const unsigned int P2_TOO_LARGE		= 0x8000;
	///Mask for determination if some part's remainder is too large
	static const unsigned int TOO_LARGE_MASK	= 0x80808080;
	///Mask for determination if relation is partial
	static const unsigned int PARTIAL_MASK	= 0x0e0e0e0e;

public:
	ARelation();
	virtual ~ARelation() {}

	virtual int Reset();
	virtual void PrintToScreen() = 0;
	virtual int ReadFromFile(FILE* aFr) = 0;
	virtual int PrintToFile(FILE* aFw) = 0;
	
	bool Equals(const ARelation & aOperand) const;
	int ControlExponents();
	int Sort();
	int CombineWithRelation(ARelation* aOperand, int aThetaPolyCount, Polynomial ** aThetaPolys);
	int CombineWithRelation(ARelation* aOperand);

	void ZeroizeNumberOfRelations() { this->iNumberOfRelations = 0; }
	
	static int StartRelationsSerialization(const CBaseParameters & aParam, xmlTextWriterPtr& aWriter);
	static int EndRelationsSerialization(const CBaseParameters & aParam, xmlTextWriterPtr& aWriter);
	static int AppendToRelationsSerialization(const CBaseParameters & aParam, xmlTextWriterPtr& aWriter);
	static int ReadRelationSerialization(const CBaseParameters & aParam, xmlTextReaderPtr& aReader);

	// Getters and Setters
protected:
	void			Set_ClassType(relation_types aType) { this->iClassType = aType; }

public:
	relation_types	Get_ClassType() const { return this->iClassType; }
	int				Get_NumberOfRelations() const { return this->iNumberOfRelations; }
	void			Set_NumberOfRelations(int aNumberOfRelations) { this->iNumberOfRelations = aNumberOfRelations; }
	CRelationPart*	Get_Parts(int aIndex) 
	{
#ifdef DEBUG_RELATION_CONTROL_ACCESS
		if (this->iParts == NULL)
		{
			Utils::ErrorMessage(__LINE__,"ARelation","Get_Parts",0,0,"abstract_relation.h","Field of Parts is NULL");		
			return NULL;
		}

		if (aIndex < 0 || this->iPartsCount <= aIndex)
		{
			Utils::ErrorMessage(__LINE__,"ARelation","Get_Parts",0,0,"abstract_relation.h","Index out of bounds.");				
			return NULL;
		}
#endif
		return this->iParts + aIndex;
	}
	
};

#endif // __ABSTRACT_RELATION_H__
