#include "abstract_sieving_phase.h"

#include <cstring>
#include <math.h>

#include "const_sieving_phase.h"
#include "const_parameter_names.h"
#include "const_element_flags.h"
#include "test_suite.h"
#include "log.h"

ASievingPhase::ASievingPhase()
{
	mpz_init_set_si(this->iM, 1);

	this->iBlockSize = 0;
	this->iModBlockSize = 0;
	this->iLogBlockSize = 0;
	this->iCheckAllMode = false;

	int rc;
	if ((rc = pthread_mutex_init(&(this->iLineSieveSizeMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","ASievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iCheckAllModeMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","ASievingPhase",rc,0,"Cannot initialize mutex.");
	}
}
int ASievingPhase::Reset()
{
	mpz_set_si(this->iM, 1);
	Set_LineSieveSize(ESizeUndefined);

	return APhase::Reset();
}
int ASievingPhase::DisposeGMP()
{
	mpz_clear(this->iM);

	return ConstRC::Ok;
}
int ASievingPhase::DisposeMutexes()
{
	int rc;

	if ((rc = pthread_mutex_destroy(&(this->iLineSieveSizeMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iCheckAllModeMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}

	return APhase::DisposeMutexes();
}
int ASievingPhase::InnerStateFinished() const
{
	return (int)ESPhaseFinished;
}
int ASievingPhase::CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	return APhase::CleanUp(aParameters, aPhaseNBR, aType, aError);
}
int ASievingPhase::InitParameters()
{
	int rc = APhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	AFactorAlgParameters * param = Get_Parameters();

	int auxInt;
	long auxLong;
	param->GetIntegerPhaseParameter(ConstPN::LINE_SIEVE_SIZE, (int)ConstSPhase::LINE_SIEVE_SIZE, auxInt, Get_PhaseNbr());
	Set_LineSieveSize((line_sieve_size)auxInt); 

	param->GetIntegerPhaseParameter(ConstPN::HASHTABLE_SIZE, (int)ConstSPhase::HASHTABLE_SIZE, auxInt, Get_PhaseNbr());
	// control the value and save it - without notification
	Utils::GetClosestPowerOf2(auxLong, auxInt);
	this->iHashtableSize = auxLong;

	param->GetIntegerPhaseParameter(ConstPN::RELATION_RESERVE, (int)ConstSPhase::RELATION_RESERVE, this->iRelationReserve, Get_PhaseNbr());

	param->GetBoolPhaseParameter(ConstPN::CHECK_ALL_MODE, ConstSPhase::CHECK_ALL_MODE, this->iCheckAllMode, Get_PhaseNbr());

	rc = param->GetMpzPhaseParameter(ConstPN::ROOT_M, this->iM, (int)EMain);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","InitParameters",0,rc,"Didn't find m.");
		return rc;
	}

	return ConstRC::Ok;
}
int ASievingPhase::ComputeIFS(int & aResult, Polynomial * aPoly, mpz_t aSievingInt, main_sieving_type aB)
{
	if (aPoly == NULL)
	{
		CLog::LogError(WHEREARG,"ASievingPhase","ComputeIFS",0,0,"Poly is NULL.");
		return ConstRC::GeneralError;
	}

	Polynomial * square = aPoly->AssignAuxPoly();
	Polynomial * f = aPoly->AssignAuxPoly();
	square->Zeroize();
	f->Zeroize();

	int degree = aPoly->GetDegree();
	f->Set(degree, *aPoly->Get2(degree));

	mpz_t integral;
	mpz_t powerB;
	mpz_init(integral);
	mpz_init_set_si(powerB, aB);

	for (int i = degree - 1; i >= 0; i--)
	{
		mpz_mul(integral, *aPoly->Get2(i), powerB);
		f->Set(i, integral);
		mpz_mul_si(powerB, powerB, aB);
	}

	aPoly->Multiply(square, f, f);

	mpz_t intMin, intMax;
	mpz_init(intMin);
	mpz_init_set(intMax, aSievingInt);
	mpz_neg(intMin, intMax);

	// int_{-M}^{M} F^{2}(x, b) dx
	aPoly->DefiniteIntegral(integral, square, intMin, intMax);

	aResult = mpz_sizeinbase(integral, ConstSPhase::LOGARITHM_BASE);
	aResult /= 2;

	mpz_mul_ui(intMax, intMax, 2);

	aResult -= mpz_sizeinbase(intMax, ConstSPhase::LOGARITHM_BASE);

	mpz_clear(integral);
	mpz_clear(powerB);
	mpz_clear(intMin);
	mpz_clear(intMax);

	aPoly->Dispose(f);
	f = NULL;
	aPoly->Dispose(square);
	square = NULL;

	return ConstRC::Ok;
}
int ASievingPhase::TestUpdate(const update_t &aUpdate)
{
	return ConstRC::Ok;
}
void ASievingPhase::PrintNfsElementInfo(nfs_fb_type & aInfo)
{
	printf("\nNFS FB Type: \n");
	printf("     Prime   : %d\n", aInfo.p);
	printf("     Root    : %d\n", aInfo.r);
	printf("     Log     : %d\n", aInfo.log_p);
	printf("     Root1   : %d\n", aInfo.root_1);
	printf("     Next1   : %d\n", aInfo.next_1);
	printf("     Inv     : %d\n", aInfo.auxVal);
}
void ASievingPhase::ControlLatticeElementInfo(lattice_type * aFB, long aMaxIndex)
{
	long max = 0;
	FILE * out = fopen("lattice_fb.txt", FILE_MODE_APPEND);
	fprintf(out, " ---- NEW FB ----\n");
	for (int i = 0; i <= aMaxIndex; i++)
	{
		if (aFB[i].a >= this->iBlockSize ||
			aFB[i].c >= this->iBlockSize)
		{
			fprintf(out, "Lattice Type: \n");
			fprintf(out, "     Prime   : %d\n", aFB[i].p);
			fprintf(out, "     R       : %d\n", aFB[i].r);
			fprintf(out, "     CR      : %d\n", aFB[i].cr);
			fprintf(out, "     B0      : %d\n", aFB[i].b0);
			fprintf(out, "     B1      : %d\n", (aFB[i].b1 + ConstSPhase::LATTICE_FB_B1_CHANGE));
			fprintf(out, "     A       : %d\n", aFB[i].a);
			fprintf(out, "     C       : %d\n", aFB[i].c);
			if (aFB[i].a >= this->iBlockSize &&
				aFB[i].c >= this->iBlockSize)
			{
				fprintf(out, " --- TOO BIG ---\n");
			}

			if (max < aFB[i].a)
			{
				max = aFB[i].a;
			}
			if (max < aFB[i].c)
			{
				max = aFB[i].c;
			}
		}
	}
	fprintf(out, "\n\n --- MAX = %ld ---\n", max);

	fclose(out);
	out = NULL;
}
void ASievingPhase::PrintLatticeElementInfo(lattice_type & aInfo)
{
	printf("\nLattice Type: \n");
	printf("     Prime   : %d\n", aInfo.p);
	printf("     R       : %d\n", aInfo.r);
	printf("     CR      : %d\n", aInfo.cr);
	printf("     B0      : %d\n", aInfo.b0);
	printf("     B1      : %d\n", (aInfo.b1 + ConstSPhase::LATTICE_FB_B1_CHANGE));
	printf("     A       : %d\n", aInfo.a);
	printf("     C       : %d\n", aInfo.c);
}
int ASievingPhase::WriteDumpSieveFile(log_type * aMainBlock)
{
	log_walking_type* cast  = (log_walking_type*)aMainBlock;

	FILE * dumpFILE = fopen("sieve_dump.txt", FILE_MODE_APPEND);
	if (dumpFILE == NULL)
	{
		return ConstRC::CouldNotOpenFile;
	}

	int iters = Get_BlockSize() / sizeof(log_walking_type);
	for (int i = 0; i < iters; i++)
	{
		fprintf(dumpFILE, "%ld", cast[i]);
	}
	fprintf(dumpFILE, "\n");

	fclose(dumpFILE);
	dumpFILE = NULL;

	return ConstRC::Ok;
}
int ASievingPhase::WriteDumpFBFile(void * aField, long aByteLong)
{
	log_type * cast = (log_type *)aField;
	FILE * dumpFILE = fopen("fb_dump.txt", FILE_MODE_APPEND);
	if (dumpFILE == NULL)
	{
		return ConstRC::CouldNotOpenFile;
	}

	int size = sizeof(nfs_fb_type);
	for (long i = 0; i < aByteLong; i++)
	{
		fprintf(dumpFILE, "%d", cast[i]);
		if (i % size == 0)
		{
			fprintf(dumpFILE, "\n");
		}
	}
	fprintf(dumpFILE, "\n\n");

	fclose(dumpFILE);
	dumpFILE = NULL;

	return ConstRC::Ok;
}
line_sieve_size ASievingPhase::Get_LineSieveSize() const
{
	line_sieve_size ret = ESizeUndefined;
	pthread_mutex_lock(&(this->iLineSieveSizeMutex));
	ret = this->iLineSieveSize;
	pthread_mutex_unlock(&(this->iLineSieveSizeMutex));

	return ret;
}
int ASievingPhase::Set_LineSieveSize(line_sieve_size aLineSieveSize)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iLineSieveSizeMutex));
		this->iLineSieveSize = aLineSieveSize;
		this->iBlockSize = (int)aLineSieveSize;
		this->iModBlockSize = this->iBlockSize - 1;
		this->iLogBlockSize = 0;
		if (aLineSieveSize >= ESize1k)
		{
			this->iLogBlockSize = 10;
			if (aLineSieveSize >= ESize2k)
			{
				this->iLogBlockSize++;
				if (aLineSieveSize >= ESize4k)
				{
					this->iLogBlockSize++;
					if (aLineSieveSize >= ESize8k)
					{
						this->iLogBlockSize++;
						if (aLineSieveSize >= ESize16k)
						{
							this->iLogBlockSize++;
							if (aLineSieveSize >= ESize32k)
							{
								this->iLogBlockSize++;
								if (aLineSieveSize >= ESize64k)
								{
									this->iLogBlockSize++;
									if (aLineSieveSize >= ESize128k)
									{
										this->iLogBlockSize++;
										if (aLineSieveSize >= ESize256k)
										{
											this->iLogBlockSize++;
											if (aLineSieveSize >= ESize512k)
											{
												this->iLogBlockSize++;
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		pthread_mutex_unlock(&(this->iLineSieveSizeMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
bool ASievingPhase::Get_CheckAllMode() const
{
	bool ret = false;
	pthread_mutex_lock(&(this->iCheckAllModeMutex));
	ret = this->iCheckAllMode;
	pthread_mutex_unlock(&(this->iCheckAllModeMutex));

	return ret;
}
int ASievingPhase::Set_CheckAllMode(bool aCheckAllMode)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iCheckAllModeMutex));
		this->iCheckAllMode = aCheckAllMode;
		pthread_mutex_unlock(&(this->iCheckAllModeMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
