#ifndef __ABSTRACT_SIEVING_PHASE_H__
#define __ABSTRACT_SIEVING_PHASE_H__

#include "abstract_phase.h"
#include "complex_structures.h"
#include "hashtables.h"
#include "lattice_hashtables.h"

#ifdef DEBUG_PREFETCHING_ON

#if defined(__GNUC__)
#define PREFETCH(x)  __builtin_prefetch(x)
#else
#define PREFETCH(x) /* nothing */
#endif

#else

#define PREFETCH(x) /* nothing */

#endif

///Abstract class for sieving phases
class ASievingPhase : public APhase
{
private:
	// Internal parameters of the line siever. 
	// Cannot be manipulated (set) from the outside, save by special methods as Init()
	///Size of the block as enum
	line_sieve_size iLineSieveSize;
	mutable pthread_mutex_t iLineSieveSizeMutex;
	/*!
		Size of the block as an integer. 
		Should be equal to iLineSieveSize.
		This value should be power of 2.
		Proportionally to the L1 cache size.
	*/
	int	iBlockSize;
	/// Binary length of iBlockSize
	int	iLogBlockSize;
	/// We need modulo iBlockSize so we use bitwise AND with iBlockSize - 1
	int iModBlockSize;
	/*
		Size of all hashtables as an integer.
		This value should be power of 2.
		Proportionally to the L2 cache size.
	*/
	int iHashtableSize;

	///If all relations will be factorized for determine their smoothness
	bool iCheckAllMode;
	mutable pthread_mutex_t iCheckAllModeMutex;

protected:
	///The value of m; f_{i}(m) = 0 mod N.
	mpz_t iM;

	// number of additional relations, which should be collected over the given estimate
	int iRelationReserve;
	
protected:
	// Sieving methods
	//int  SieveBlock(int aBlockIndex, int aBlockMin, int aBlockMax, int aSign, bool aInit);

	// Line Sieving
	//int  SieveWithLargishPrimes(script_t* aScript, log_type* aBlock);
	//int  ShiftLargishPrimes(CHashtables & aHashtables, int aCurrentBlockIndex);
	
	// Lattice Sieving methods
	//int SieveWithLargishPrimes(lattice_script_t* aScript, log_type* aBlock, int aModI);
	//int ShiftLargishPrimes(CLatticeHashtables & aHashtables, int aCurrentBlockIndex, long aModI);

	//virtual int SieveWithMiddlePrimes(void* aStart, void* aEnd, log_type* aBlock, int aSign, bool aInit) = 0;

	//virtual int Swapping(int aIndex) = 0;
	//virtual int Shifting(int aIndex) = 0;
	//virtual int FindingAndFactorizing(int aBlockIndex, int aInnerIndex, int aSign) = 0;
	//virtual int Sieving(int aIndex, int aSign, bool aInit) = 0;

	int ComputeIFS(int & aResult, Polynomial * aPoly, mpz_t aSievingInt, main_sieving_type aB);

	void ReduceBase(main_sieving_type & outA0, main_sieving_type & outB0, 
					main_sieving_type & outA1, main_sieving_type & outB1,
					main_sieving_type inA0, main_sieving_type inB0, 
					main_sieving_type inA1, main_sieving_type inB1, float aSigma);

	// Print methods
	void PrintNfsElementInfo(nfs_fb_type & aInfo);
	void PrintLatticeElementInfo(lattice_type & aInfo);
	void ControlLatticeElementInfo(lattice_type * aFB, long aMaxIndex);
	
	int WriteDumpSieveFile(log_type * aMainBlock);
	int WriteDumpFBFile(void * aField, long aByteLong);

	// Tests methods
	//void SearchForPrimeInHashtable(lattice_hashtable_t* aHashtable, int aMaxIndex, main_sieving_type aPrime, main_sieving_type aCp);
	virtual int TestUpdate(const update_t &aUpdate);

	// Phase methods

	///Init all parameters necessary for calculation - get from iParameters
	int InitParameters();
	///Dispose mutexes in current class - call only from destructor
	int DisposeMutexes();
	///Reset all resources which was used and prepare for new start.
	int Reset();

	///Dispose mpz_t members in current class - call only from destructor
	virtual int DisposeGMP();

	///Return number of finished state
	int InnerStateFinished() const;

public:

	ASievingPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);


	// Getters and Setters
private:
	int				Set_LineSieveSize(line_sieve_size aLineSieveSize);

protected:
	int				Get_BlockSize() const { return this->iBlockSize; }
	int				Get_ModBlockSize() const { return this->iModBlockSize; }
	int				Get_LogBlockSize() const { return this->iLogBlockSize; }
	int				Get_HashtableSize() const { return this->iHashtableSize; }

public:
	line_sieve_size Get_LineSieveSize() const;
	bool			Get_CheckAllMode() const;
	int				Set_CheckAllMode(bool aCheckAllMode);

};

#endif // __ABSTRACT_SIEVING_PHASE_H__
