#include "abstract_square_root_phase.h"

#include "couveignes_root_finder.h"
#include "integral_domain.h"
#include "const_square_root_phase.h"
#include "const_parameter_names.h"
#include "const_element_flags.h"
#include "log.h"

ASquareRootPhase::ASquareRootPhase()
{
	mpz_init_set_si(this->iN, 1);
	mpz_init_set_si(this->iM, 1);
	mpz_init_set_si(this->iFactor1, 1);
	mpz_init_set_si(this->iFactor2, 1);

	this->iMatrixResult = NULL;
	this->iDependencyIndex = 0;

	int rc;
	if ((rc = pthread_mutex_init(&(this->iFactor1Mutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","ASquareRootPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iFactor2Mutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","ASquareRootPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iDependencyIndexMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","ASquareRootPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iCheckAllModeMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","ASquareRootPhase",rc,0,"Cannot initialize mutex.");
	}
}
int ASquareRootPhase::Reset()
{
	if (this->iMatrixResult != NULL)
	{
// already deleted by linear phase ?????
//		delete this->iMatrixResult;
		this->iMatrixResult = NULL;
	}

	mpz_set_si(this->iN, 1);
	mpz_set_si(this->iM, 1);
	pthread_mutex_lock(&(this->iFactor1Mutex));
	mpz_set_si(this->iFactor1, 1);
	pthread_mutex_unlock(&(this->iFactor1Mutex));
	pthread_mutex_lock(&(this->iFactor2Mutex));
	mpz_set_si(this->iFactor2, 1);
	pthread_mutex_unlock(&(this->iFactor2Mutex));
	
	Set_DependencyIndex(0);

	return APhase::Reset();
}
int ASquareRootPhase::DisposeGMP()
{
	mpz_clear(this->iN);
	mpz_clear(this->iM);
	mpz_clear(this->iFactor1);
	mpz_clear(this->iFactor2);

	return ConstRC::Ok;
}
int ASquareRootPhase::DisposeMutexes()
{
	int rc;

	if ((rc = pthread_mutex_destroy(&(this->iFactor1Mutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iFactor2Mutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iDependencyIndexMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iCheckAllModeMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}

	return APhase::DisposeMutexes();
}
int ASquareRootPhase::InnerStateFinished() const
{
	return (int)ESRPhaseFinished;
}
int ASquareRootPhase::CleanUp(AFactorAlgParameters *aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	return APhase::CleanUp(aParameters, aPhaseNBR, aType, aError);
}
int ASquareRootPhase::InitParameters()
{
	int rc = APhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	AFactorAlgParameters * param = Get_Parameters();

	// ASquareRootPhase members
	rc = param->GetMpzPhaseParameter(ConstPN::NUMBER_N, this->iN, (int)EMain);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","InitParameters",0,rc,"Didn't find N.");
		return rc;
	}

	bool auxBool = false;
	param->GetBoolPhaseParameter(ConstPN::CHECK_ALL_MODE, ConstSRPhase::CHECK_ALL_MODE, auxBool, Get_PhaseNbr());
	Set_CheckAllMode(auxBool);

	// result relations file - this file was created by other phase so we should get full name or 
	// we have to assemble it
	if (param->ContainsPhaseParameter(ConstPN::RELATIONS_RESULT_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::RELATIONS_RESULT_FULL_FILENAME, this->iRelationsFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iRelationsFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iRelationsFullFileName, ConstPN::RELATIONS_RESULT_FILENAME, 
			ConstPN::RELATIONS_RESULT_DIRECTORY, ConstPN::RELATIONS_RESULT_COMPRESSION, Get_SerialMainDirectory(),
			ConstSRPhase::RELATIONS_FILENAME, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}
	// Matrix result full filename
	if (param->ContainsPhaseParameter(ConstPN::MATRIX_RESULT_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::MATRIX_RESULT_FULL_FILENAME, this->iMatrixResultFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iMatrixResultFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iMatrixResultFullFileName, ConstPN::MATRIX_RESULT_FILENAME, 
			ConstPN::MATRIX_RESULT_DIRECTORY, ConstPN::MATRIX_RESULT_COMPRESSION, Get_SerialMainDirectory(),
			ConstSRPhase::MATRIX_RESULT_FILENAME, Get_SerialIdentifier(), Get_PhaseNbr());
	}

	int matrixType;
	rc = param->GetIntegerPhaseParameter(ConstPN::MATRIX_RESULT_TYPE, matrixType, Get_PhaseNbr());
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","InitParameters",0,rc,"Didn't find result matrix type.");
		return rc;
	}
	this->iMatrixResultType = (species_of_matrices)matrixType; 

	rc = param->GetMpzPhaseParameter(ConstPN::ROOT_M, this->iM, (int)EMain);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","InitParameters",0,rc,"Didn't find m.");
		return rc;
	}

	return ConstRC::Ok;
}

int ASquareRootPhase::IntegralRoot(CRelationPart *aPart, mpz_t aResult)
{
	if (aPart == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","IntegerRoot",0,0,"Part is NULL.");
		return ConstRC::BadArgument;
	}

	nfs_sieving_element element;
	// multiply integer elements mod N	
	mpz_t powered_integer;

	mpz_init(powered_integer);

	mpz_set_ui(aResult, 1);
	CDivisors * divisors = &aPart->Main;
	long max = divisors->MaxElementIndex;
	long i;
	
	for (i = 0; i <= max; i++)
	{
		element = divisors->Elements[i];
		mpz_set_si(powered_integer, element.prime);
		mpz_powm_ui(powered_integer, powered_integer, element.exponent / 2, this->iN);
		mpz_mul(aResult, aResult, powered_integer);
		mpz_fdiv_r(aResult, aResult, this->iN);
	}

	divisors = &aPart->Large;
	max = divisors->MaxElementIndex;
	for (i = 0; i <= max; i++)
	{
		element = divisors->Elements[i];
		mpz_set_si(powered_integer, element.prime);
		mpz_powm_ui(powered_integer, powered_integer, element.exponent / 2, this->iN);
		mpz_mul(aResult, aResult, powered_integer);
		mpz_fdiv_r(aResult, aResult, this->iN);
	}

	mpz_clear(powered_integer);

	return ConstRC::Ok;
}
int ASquareRootPhase::CheckAlgebraicSquare(Polynomial *aSquareFactors, int aNumFactors, prime_ideal_for_legendre * aQuadChars, long aNumQuadChars, Polynomial *aSievingPoly)
{
	int failed = 0;
	mpz_t temp, temp2;

	mpz_init(temp);
	mpz_init(temp2);

	Polynomial poly, poly2;
	poly2.Zeroize();
	poly2.Set(0, 1);

	for (int i = 0; i < aNumFactors; i++)
	{
		poly.Multiply(&poly, &poly2, &(aSquareFactors[i]));
		poly.PowerMod(&poly2, &poly, 1, aSievingPoly);
	}

	for (long i = 0; i < aNumQuadChars; i++)
	{
		mpz_set_ui(temp, aQuadChars[i].c_p);
		poly2.Evaluate(temp2, temp);
		mpz_set_ui(temp, aQuadChars[i].prime);
		if(mpz_legendre(temp2, temp) == -1)
		{
			gmp_printf("[%i,%i,%i]\n", i, aQuadChars[i].prime, aQuadChars[i].c_p);
			failed++;
		}
	}

	mpz_clear(temp);
	mpz_clear(temp2);

	if (failed > 0) 
	{
		printf("  Check Square Run: failed %d of altogether %ld tries.\n", failed, aNumQuadChars);
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}
int ASquareRootPhase::AlgebraicRoot(CRelationPart *aPart, Polynomial *aThetaPoly, mpz_t aLC, int aNbrOfRel, root_finder_types aType, long aMaxBound, Polynomial **aSquareFactors, long aNumFactors, mpz_t aLCM, mpz_t aResult)
{
	if (aPart == NULL || aThetaPoly == NULL || aSquareFactors == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","AlgebraicRoot",0,0,"Bad arguments, something is missing.");
		return ConstRC::BadArgument;
	}

	Polynomial * derivate = new Polynomial();
	aThetaPoly->Derivate(derivate, aThetaPoly);

	ARootFinder * rootFinder = NULL;

 	switch (aType)
	{
		case ENewton:	
			rootFinder = new CNewtonRootFinder();
			break;
		case ECouveignes:
			rootFinder = new CCouveignesRootFinder();
			break;
		default:
			CLog::LogError(WHEREARG,"ASquareRootPhase","AlgebraicRoot",0,aType,"Unknown square root algorithm required!");
			rootFinder = new CNewtonRootFinder();
			aType = ENewton;
			break;
	}

	rootFinder->SetFactors(aNumFactors, aSquareFactors);
	rootFinder->SetNumberField(aThetaPoly);

	if (aType == ECouveignes)
	{
		IntegralDomain * domain = new IntegralDomain();
		domain->AssignPolynomial(aThetaPoly);
		domain->FindIntegralBase(aMaxBound);

		bool nonmonic = mpz_cmp_ui(aLC, 1) != 0;
		long i;
		long max = aPart->Main.MaxElementIndex + aPart->Large.MaxElementIndex + 3;
		if (nonmonic)
		{
			// nonmonic polynomial - we need to add c_{d} to norm
			max++;
		}
		nfs_sieving_element element;

		mpz_t * bases = new mpz_t[max];
		int * exponents = new int[max];
		CDivisors * divisors = &aPart->Main;
		long countIdeals = divisors->MaxElementIndex + 1;
		for (i = 0; i < countIdeals; i++)
		{
			element = divisors->Elements[i];
			mpz_init_set_si(bases[i], element.prime);
			if (element.flags & ConstEF::INERTIAL_DEGREE_MASK)
			{
				if (element.flags & ConstEF::DIVIDES_L)
				{
					// we get combined valuation of F(a, -b)
					exponents[i] = ((element.flags & ConstEF::F_VALUATION_MASK) >> ConstEF::F_VALUATION_SHIFT);
					if (exponents[i] % 2)
					{
						CLog::LogError(WHEREARG,"ASquareRootPhase","AlgebraicRoot",0,element.flags,"Strange: p-valuation of product F(a,b) is not even.");
					}
					exponents[i] = exponents[i] / 2;
				}
				else
				{
					// this could be special ideal or principle ideal (p)
					exponents[i] = (element.flags & ConstEF::INERTIAL_DEGREE_MASK) * element.exponent / 2;
				}
			}
			else
			{
				// normal prime ideal - inertial degree is 1
				exponents[i] = element.exponent / 2;
			}
		}

		divisors = &aPart->Large;
		for (i = 0; i <= divisors->MaxElementIndex; i++)
		{
			element = divisors->Elements[i];
			mpz_init_set_si(bases[i + countIdeals], element.prime);
			if (element.flags & ConstEF::INERTIAL_DEGREE_MASK)
			{
				// this could be special ideal or principle ideal (p)
				exponents[i + countIdeals] = (element.flags & ConstEF::INERTIAL_DEGREE_MASK) * element.exponent / 2;
			}
			else
			{
				// normal prime ideal - inertial degree is 1
				exponents[i + countIdeals] = element.exponent / 2;
			}
		}

		if (nonmonic)
		{
			// nonmonic polynomial - we need to add c_{d} to norm
			mpz_init_set(bases[max - 2], aLC);
			exponents[max - 2] = aNbrOfRel * (aThetaPoly->GetDegree() - 1) / 2;
		}

		mpz_init(bases[max - 1]);
		domain->NormInZTheta(bases[max - 1], derivate);
		exponents[max - 1] = 1;
	
		((CCouveignesRootFinder*)rootFinder)->SetNorms(bases, exponents, max);
		((CCouveignesRootFinder*)rootFinder)->SetHomomorphism(aLCM, this->iN);
		((CCouveignesRootFinder*)rootFinder)->RootImageFind(aResult);	

		for (i = 0; i < max; i++)
		{
			mpz_clear(bases[i]);
		}

		delete[] bases;
		delete[] exponents;
		delete domain;
	}

	if (aType == ENewton)
	{
		((CNewtonRootFinder*)rootFinder)->FindRoot(derivate);
		derivate->Evaluate(aResult, aLCM);
		mpz_mod(aResult, aResult, this->iN);
	}

	if (rootFinder != NULL)
	{
		delete rootFinder;
		rootFinder = NULL;
	}
	if (derivate != NULL)
	{
		delete derivate;
		derivate = NULL;
	}

	return ConstRC::Ok;
}
int ASquareRootPhase::AllocatePolyField(Polynomial **& aPoly, long aAllocatedSize)
{
	if (aPoly != NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","AllocatePolyField",0,0,"Field is not NULL!");
		return ConstRC::GeneralError;
	}

	aPoly = (Polynomial**)Utils::AllocationMachine((void**)&aPoly, aAllocatedSize * sizeof(Polynomial*)); 
	if (aPoly == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","AllocatePolyField",0,0,"Not enough memory for poly field.");
		return ConstRC::NotEnoughMemory;
	}

	for (long i = 0; i < aAllocatedSize; i++)
	{
		aPoly[i] = new Polynomial();
	}

	return ConstRC::Ok;
}
int ASquareRootPhase::ReallocatePolyField(Polynomial **& aPoly, long &aAllocatedSize)
{
	if (aPoly == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","ReallocatePolyField",0,0,"Field is NULL!");
		return ConstRC::GeneralError;
	}

	long newSize = ConstSRPhase::POLY_FIELD_RESIZE + aAllocatedSize; 
	aPoly = (Polynomial**)Utils::AllocationMachine((void**)&aPoly, newSize * sizeof(Polynomial*)); 
	if (aPoly == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","ReallocatePolyField",0,0,"Not enough memory for poly field.");
		return ConstRC::NotEnoughMemory;
	}

	for (long i = aAllocatedSize; i < newSize; i++)
	{
		aPoly[i] = new Polynomial();
	}

	aAllocatedSize = newSize;

	return ConstRC::Ok;
}
int ASquareRootPhase::DeletePolyField(Polynomial **& aPoly, long & aAllocatedSize)
{
	if (aPoly == NULL)
	{
		CLog::LogError(WHEREARG,"ASquareRootPhase","DeletePolyField",0,0,"Field is NULL!");
		return ConstRC::GeneralError;
	}

	for (long i = 0; i < aAllocatedSize; i++)
	{
		if (aPoly[i] != NULL)
		{
			delete aPoly[i];
			aPoly[i] = NULL;
		}
	}

	free(aPoly);
	aPoly = NULL;
	aAllocatedSize = 0;

	return ConstRC::Ok;
}
int ASquareRootPhase::CheckResultPair(mpz_t aX, mpz_t aY)
{
	int rc = ConstRC::Ok;
	mpz_t aux;
	mpz_init(aux);
	mpz_sub(aux, aX, aY);

	mpz_gcd(aux, this->iN, aux);
	if (mpz_cmp_ui(aux, 1) == 0)
	{
		mpz_add(aux, aX, aY);
		mpz_gcd(aux, this->iN, aux);

		if (Get_InfoMode())
		{
			if (mpz_cmp(aux, this->iN) == 0)
			{
				gmp_printf("\n\t\tTrivial dependency.\n");
			}
			else
			{
				gmp_printf("\n\t\tNot a square in Z[\\alpha]!\n");
			}
		}

		rc = ConstRC::NotFactorized;
	}
	else if(mpz_cmp(aux, this->iN) == 0)
	{
		if (Get_InfoMode())
		{
			gmp_printf("\n\t\tTrivial dependency.\n");
		}
		rc = ConstRC::NotFactorized;
	}
	else 
	{
		Set_Factor1(aux);
		mpz_divexact(aux, this->iN, aux);
		Set_Factor2(aux);
		if (Get_InfoMode())
		{
			gmp_printf("\n\t\tGood dependency.\n");
		}
		rc = ConstRC::Ok;

		if (Get_CheckAllMode())
		{
			rc = ConstRC::NotFactorized;
		}
	}

	mpz_clear(aux);

	return rc;
}
int ASquareRootPhase::Get_Factor1(mpz_t aFactor1) const
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iFactor1Mutex));
		mpz_set(aFactor1, this->iFactor1);
		pthread_mutex_unlock(&(this->iFactor1Mutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int ASquareRootPhase::Set_Factor1(mpz_t aFactor1)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iFactor1Mutex));
		mpz_set(this->iFactor1, aFactor1);
		pthread_mutex_unlock(&(this->iFactor1Mutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int ASquareRootPhase::Get_Factor2(mpz_t aFactor2) const
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iFactor2Mutex));
		mpz_set(aFactor2, this->iFactor2);
		pthread_mutex_unlock(&(this->iFactor2Mutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int ASquareRootPhase::Set_Factor2(mpz_t aFactor2)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iFactor2Mutex));
		mpz_set(this->iFactor2, aFactor2);
		pthread_mutex_unlock(&(this->iFactor2Mutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int ASquareRootPhase::Get_DependencyIndex() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iDependencyIndexMutex));
	ret = this->iDependencyIndex;
	pthread_mutex_unlock(&(this->iDependencyIndexMutex));

	return ret;
}
int ASquareRootPhase::Set_DependencyIndex(int aDependencyIndex)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iDependencyIndexMutex));
		this->iDependencyIndex = aDependencyIndex;
		pthread_mutex_unlock(&(this->iDependencyIndexMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}

bool ASquareRootPhase::Get_CheckAllMode() const
{
	bool ret = 0;
	pthread_mutex_lock(&(this->iCheckAllModeMutex));
	ret = this->iCheckAllMode;
	pthread_mutex_unlock(&(this->iCheckAllModeMutex));

	return ret;
}
int ASquareRootPhase::Set_CheckAllMode(bool aCheckAllMode)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iCheckAllModeMutex));
		this->iCheckAllMode = aCheckAllMode;
		pthread_mutex_unlock(&(this->iCheckAllModeMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
