#ifndef __ABSTRACT_SQUARE_ROOT_PHASE_H__
#define __ABSTRACT_SQUARE_ROOT_PHASE_H__

#include "abstract_phase.h"
#include "relation_part.h"
#include "sparse_matrix_class.h"

///Abstract class for square root phases
class ASquareRootPhase : public APhase
{
private:
	///First factor of N
	mpz_t iFactor1;
	mutable pthread_mutex_t iFactor1Mutex;
	///Second factor of N
	mpz_t iFactor2;
	mutable pthread_mutex_t iFactor2Mutex;
	///Index of now using dependency for factorization
	int iDependencyIndex;
	mutable pthread_mutex_t iDependencyIndexMutex;
	///Check all dependency
	bool iCheckAllMode;
	mutable pthread_mutex_t iCheckAllModeMutex;


protected:
	///Number to factor
	mpz_t iN;
	///The value of m; f(m) = 0 mod N.
	mpz_t iM;
	///Full filename with all relations
	std::string iRelationsFullFileName;
	///Full filename with matrix result
	std::string iMatrixResultFullFileName;
	///LinearPhase results in form of matrix
	AbstractMatrix* iMatrixResult;
	///LinearPhase matrix result type
	species_of_matrices iMatrixResultType;

public:
	ASquareRootPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

protected:
	///Dispose mpz_t members in current class - call only from destructor
	virtual int DisposeGMP();

	///Try all dependencies for factorization
	virtual int TryDependencies() = 0;
	///Try K-th dependency for factorization - index is in iDependencyIndex
	virtual int TryKthDependence() = 0;

	virtual int IntegralRoot(CRelationPart *aPart, mpz_t aResult);
	virtual int AlgebraicRoot(CRelationPart *aPart, Polynomial *aThetaPoly, mpz_t aLC, int aNbrOfRel, root_finder_types aType, long aMaxBound, Polynomial **aSquareFactors, long aNumFactors, mpz_t aLCM, mpz_t aResult);

	int AllocatePolyField(Polynomial **& aPoly, long aAllocatedSize);
	int ReallocatePolyField(Polynomial **& aPoly, long & aAllocatedSize);
	int DeletePolyField(Polynomial **& aPoly, long & aAllocatedSize);

	int CheckResultPair(mpz_t aX, mpz_t aY);
	int CheckAlgebraicSquare(Polynomial *aSquareFactors, int aNumFactors, prime_ideal_for_legendre * aQuadChars, long aNumQuadChars, Polynomial *aSievingPoly);

	int InitParameters();
	int DisposeMutexes();
	int Reset();

	///Return number of finished state
	int InnerStateFinished() const;

	// Getters and Setters
public:
	int Set_CheckAllMode(bool aCheckAllMode);
	bool Get_CheckAllMode() const;

	int Get_Factor1(mpz_t aFactor1) const;
	int Get_Factor2(mpz_t aFactor2) const;
protected:
	int Set_Factor1(mpz_t aFactor1);
	int Set_Factor2(mpz_t aFactor2);
	int Get_DependencyIndex() const;
	int Set_DependencyIndex(int aDependencyIndex);

};


#endif // __ABSTRACT_SQUARE_ROOT_PHASE_H__