#include "algebraic_number_class.h"
#include "const_return_codes.h"

AlgebraicNumber::AlgebraicNumber(unsigned int aRank, HermiteMatrix **aThetaPowers, mpz_t *aDenominator){
    unsigned int i;
    this->Rank = aRank;
    this->vector = NULL;
    this->vector = (mpz_t*) Utils::AllocationMachine((void **) &this->vector, Rank*sizeof(mpz_t));
    this->multiplication = NULL;
    this->mult_allocated = 0;
    for(i=0;i<Rank;i++)
	mpz_init(vector[i]);
    this->ThetaPowers = aThetaPowers;
    this->denominator = aDenominator;
}

AlgebraicNumber::~AlgebraicNumber(){
    unsigned int i;
    for(i=0;i<Rank;i++)
	mpz_clear(vector[i]);
    if (multiplication) delete(multiplication);
    free(vector);
}

int AlgebraicNumber::Set(unsigned int aIndex,long aValue){
    if(aIndex>Rank) return ConstRC::SizeMismatch;
    mpz_set_si(vector[aIndex],aValue);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Set(unsigned int aIndex,mpz_t aValue){
    if(aIndex>Rank) return ConstRC::SizeMismatch;
    mpz_set(vector[aIndex],aValue);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

long AlgebraicNumber::Get(unsigned int aIndex){
    if(aIndex>Rank) return 0;
    return mpz_get_si(vector[aIndex]);
}

mpz_t *AlgebraicNumber::Get2(unsigned int aIndex){
    if(aIndex>Rank) return NULL;
    return vector+aIndex;
}

int AlgebraicNumber::Inc(unsigned int aIndex,mpz_t aValue){
    if(aIndex>Rank) return ConstRC::SizeMismatch;
    mpz_add(vector[aIndex],vector[aIndex],aValue);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Dec(unsigned int aIndex,mpz_t aValue){
    if(aIndex>Rank) return ConstRC::SizeMismatch;
    mpz_sub(vector[aIndex],vector[aIndex],aValue);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Inc(unsigned int aIndex,long aValue){
    if(aIndex>Rank) return ConstRC::SizeMismatch;
    if (aValue>=0)
	mpz_add_ui(vector[aIndex],vector[aIndex],aValue);
    else 
	mpz_sub_ui(vector[aIndex],vector[aIndex],aValue);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Dec(unsigned int aIndex,long aValue){
    if(aIndex>Rank) return ConstRC::SizeMismatch;
    if (aValue<=0)
	mpz_add_ui(vector[aIndex],vector[aIndex],aValue);
    else 
	mpz_sub_ui(vector[aIndex],vector[aIndex],aValue);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Zeroize(){
    unsigned int i;
    for(i=0;i<Rank;i++)
	mpz_set_ui(vector[i],0);
    if (this->mult_allocated==1){
	this->mult_allocated=-1;
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::PrintToScreen(){
    unsigned int i;
    bool plus=FALSE;
    for(i=0;i<Rank;i++){
	if (mpz_cmp_ui(vector[i],0)){
	    if (plus) gmp_printf("%+Zd",vector[i]);
	    else gmp_printf("%Zd",vector[i]);
	    plus=TRUE;
	    switch (i){
		case 0:break;
		case 1:printf("(t)");break;
		default:printf("(t^%d)",i);
	    }
	}
    }
    if (!plus) printf("0");
    return ConstRC::Ok;
}

int AlgebraicNumber::Add(AlgebraicNumber *aTarget, AlgebraicNumber* aOperand,bool aComputeMatrix){
    unsigned int i;
    if((this==aTarget)||(aOperand==aTarget)) aComputeMatrix=FALSE;
    for(i=0;i<Rank;i++)
	mpz_add(aTarget->vector[i],this->vector[i],aOperand->vector[i]);
    if((this->mult_allocated==1)&&(aOperand->mult_allocated==1)&&(aComputeMatrix)){
	if(aTarget->mult_allocated==0){
	    aTarget->multiplication = this->multiplication->Add(aOperand->multiplication);
	} else {
	    this->multiplication->Add(aTarget->multiplication,aOperand->multiplication);
	}
	aTarget->mult_allocated=1;
    } else {
	if (aTarget->mult_allocated==1){
	    aTarget->mult_allocated=-1;
	}
    }
    return ConstRC::Ok;
}

AlgebraicNumber *AlgebraicNumber::Add(AlgebraicNumber *aOperand,bool aComputeMatrix){
    AlgebraicNumber *Res;
    Res = new AlgebraicNumber(Rank,ThetaPowers,denominator);
    this->Add(Res,aOperand,aComputeMatrix);
    return Res;
}

int AlgebraicNumber::Multiply(AlgebraicNumber *aTarget, long aScalar,bool aComputeMatrix){
    unsigned int i;
    for(i=0;i<Rank;i++)
	mpz_mul_si(aTarget->vector[i],this->vector[i],aScalar);
    if((this->mult_allocated==1)&&(aComputeMatrix)){
	if(aTarget->mult_allocated==0){
	    aTarget->multiplication = new HermiteMatrix(Rank,Rank);
	}
	this->multiplication->ScalarMultiply(aTarget->multiplication,aScalar);
	aTarget->mult_allocated=1;
    } else {
	if (aTarget->mult_allocated==1){
	    aTarget->mult_allocated=-1;
	}
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Multiply(AlgebraicNumber *aTarget, mpz_t aScalar,bool aComputeMatrix){
    unsigned int i;
    for(i=0;i<Rank;i++)
	mpz_mul(aTarget->vector[i],this->vector[i],aScalar);
    if((this->mult_allocated==1)&&(aComputeMatrix)){
	if(aTarget->mult_allocated==0){
	    aTarget->multiplication = new HermiteMatrix(Rank,Rank);
	}
	this->multiplication->ScalarMultiply(aTarget->multiplication,aScalar);
	aTarget->mult_allocated=1;
    } else {
	if (aTarget->mult_allocated==1){
	    aTarget->mult_allocated=-1;
	}
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Divide(AlgebraicNumber *aTarget, long aScalar,bool aComputeMatrix){
    unsigned int i;
    for(i=0;i<Rank;i++){
	mpz_fdiv_q_ui(aTarget->vector[i],this->vector[i],abs(aScalar));
	if (aScalar<0) mpz_neg(aTarget->vector[i],aTarget->vector[i]);
    }
    if((this->mult_allocated==1)&&(aComputeMatrix)){
	if(aTarget->mult_allocated==0){
	    aTarget->multiplication = new HermiteMatrix(Rank,Rank);
	}
	this->multiplication->ScalarDivide(aTarget->multiplication,aScalar);
	aTarget->mult_allocated=1;
    } else {
	if (aTarget->mult_allocated==1){
	    aTarget->mult_allocated=-1;
	}
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::Divide(AlgebraicNumber *aTarget, mpz_t aScalar,bool aComputeMatrix){
    unsigned int i;
    if (!mpz_cmp_ui(aScalar,0)) return ConstRC::GeneralError;
    for(i=0;i<Rank;i++)
	mpz_fdiv_q(aTarget->vector[i],this->vector[i],aScalar);
    if((this->mult_allocated==1)&&(aComputeMatrix)){
	if(aTarget->mult_allocated==0){
	    aTarget->multiplication = new HermiteMatrix(Rank,Rank);
	}
	this->multiplication->ScalarDivide(aTarget->multiplication,aScalar);
	aTarget->mult_allocated=1;
    } else {
	if (aTarget->mult_allocated==1){
	    aTarget->mult_allocated=-1;
	}
    }
    return ConstRC::Ok;
}

int AlgebraicNumber::ComputeMultiplicationMatrix(){
    if (mult_allocated==1) return ConstRC::Ok;
    if (!mult_allocated){
	multiplication = new HermiteMatrix(Rank,Rank);
    }
    int ret = multiplication->Zeroize();
    if (ret!=ConstRC::Ok) return ret;
    HermiteMatrix *aux;
    aux = new HermiteMatrix(Rank,Rank);
    ret = aux->Allocate();
    if (ret!=ConstRC::Ok) return ret;
    unsigned int i;
    for(i=0;i<Rank;i++) multiplication->PutNumber(vector[0],i,i);
    for(i=1;i<Rank;i++){
	ThetaPowers[i-1]->ScalarMultiply(aux,this->vector[i]);
	aux->Add(multiplication,multiplication);
    }
    this->mult_allocated=1;
    mpz_set(this->multiplication->denominator,*denominator);
    delete(aux);
    return ConstRC::Ok;
}

int AlgebraicNumber::Multiply(AlgebraicNumber *aTarget,AlgebraicNumber *aOperand,bool aComputeMatrix){
    int ret;
    if((this==aTarget)||(aOperand==aTarget)) return ConstRC::GeneralError;
    if(aOperand->mult_allocated==1){
	ret=aOperand->multiplication->MultiplyVector(aTarget->vector,this->vector);
	if(ret!=ConstRC::Ok) return ret;
    } else {
	if(this->mult_allocated!=1){
	    ret=this->ComputeMultiplicationMatrix();
	    if(ret!=ConstRC::Ok) return ret;
	}
//	this->PrintToScreen();printf(" * ");aOperand->PrintToScreen();printf(" = ");
	ret=this->multiplication->MultiplyVector(aTarget->vector,aOperand->vector);
//	aTarget->PrintToScreen();printf("\n");
	if(ret!=ConstRC::Ok) return ret;
    }
    if((this->mult_allocated==1)&&(aOperand->mult_allocated==1)&&(aComputeMatrix)){
	if(aTarget->mult_allocated==0){
	    aTarget->multiplication = this->multiplication->MultiplyInternal(this->multiplication,aOperand->multiplication);
	} else {
	    this->multiplication->MultiplyInternal(aTarget->multiplication,this->multiplication,aOperand->multiplication);
	}
	aTarget->mult_allocated=1;
    } else {
	if (aTarget->mult_allocated==1){
	    aTarget->mult_allocated=-1;
	}
    }
    aTarget->Divide(aTarget,*denominator,aComputeMatrix);
    return ConstRC::Ok;
}


AlgebraicNumber *AlgebraicNumber::Multiply(AlgebraicNumber *aOperand,bool aComputeMatrix){
    AlgebraicNumber *Res;
    Res = new AlgebraicNumber(Rank,ThetaPowers,denominator);
    this->Multiply(Res,aOperand,aComputeMatrix);
    return Res;
}

int AlgebraicNumber::Norm(mpz_t aNorm, Polynomial *aThetaPolynomial){
    Polynomial *a;
    a = new Polynomial(Rank);
    int ret;
    ret = a->Allocate(Rank);
    if (ret!=ConstRC::Ok) return ret;
    mpz_t aux;
    mpz_init(aux);
    unsigned int i;
    for(i=0;i<Rank;i++){
	a->Set(i,vector[i]);
    }
    a->Resultant(aNorm,aThetaPolynomial,a);
    mpz_pow_ui(aux,*denominator,Rank);
    mpz_divexact(aNorm,aNorm,aux);
    delete a;
    mpz_clear(aux);
    return ConstRC::Ok;
}

int AlgebraicNumber::Power(AlgebraicNumber *aTarget, unsigned int aPower){
    unsigned int i,l;
    AlgebraicNumber *s,*t;
    for(i=1;i<Rank;i++)
	mpz_set_ui(aTarget->vector[i],0);
    mpz_set(aTarget->vector[0],*denominator);
    if (aPower==0){
	return ConstRC::Ok;
    }
    s=new AlgebraicNumber(Rank,ThetaPowers,denominator);
    t=new AlgebraicNumber(Rank,ThetaPowers,denominator);
    this->ComputeMultiplicationMatrix();
    this->Multiply(s,1,TRUE);
    i=aPower;
    for(l=0;i;l++) i=i>>1;
    for(i=0;i<l;i++){
	if (i) {
	    s->Multiply(t,1,TRUE);
	    t->Multiply(s,t,TRUE);
	}
	if (aPower & 1<<i) {
	    s->Multiply(t,aTarget,FALSE);
	    t->Multiply(aTarget,1,FALSE);
	}
    }
    delete(s);
    delete(t);
    return ConstRC::Ok;
}

int AlgebraicNumber::MinimalPolynomial(Polynomial *aPolynomial){
    HermiteMatrix *Poly,*Ker;
    unsigned int i,j;
    Poly = new HermiteMatrix(Rank,Rank+1);
    int ret = Poly->Allocate();
    if (ret!=ConstRC::Ok) return ret;
    Ker = new HermiteMatrix(Rank+1,Rank+1);
    AlgebraicNumber *A,*B;
    A = new AlgebraicNumber(Rank,ThetaPowers,denominator);
    B = new AlgebraicNumber(Rank,ThetaPowers,denominator);
    this->Power(A,0);
    for(j=0;j<=Rank;j++){
	if (j) {
	    A->Multiply(B,1,FALSE);
	    this->Multiply(A,B,FALSE);
	}
	for(i=0;i<Rank;i++)
	    Poly->PutNumber(A->vector[i],i,j);
    }
    long D;
    Poly->PrintToScreen();
    Poly->Kernel(Ker,&D,TRUE);
    Ker->PrintToScreen();
    aPolynomial->Zeroize(0);
    for(j=0;j<=Rank;j++)
	aPolynomial->Set(j,*Ker->GetXY2(j,D-1));
    mpz_t C;
    mpz_init(C);
    aPolynomial->Content(C);
    aPolynomial->Divide(aPolynomial,aPolynomial,C);
    mpz_clear(C);
    delete(Ker);
    delete(Poly);
    return ConstRC::Ok;
}

