#include <cstring>

#include "bit_matrix_class.h"
#include "common_utils.h" // because of allocation machine
#include <stdio.h>
#include "const_return_codes.h"
#include "log.h"
#include "xml_helper.h"

/*!
	Used for bit shift constants: K_SHIFTS[i] = (1 << i), for i=0...(bits_in_m_t - 1). It is
	initialized at construction of first instance of BitMatrix, and is NULL before; but it can be
	forcedly initialized with no instance of BitMatrix in existence by calling static method DefineKShifts().
	
	Before using K_SHIFTS constants (they are public), make sure that they have been initialized, otherwise a null pointer exception occurs.

*/
matrix_type* BitMatrix::K_SHIFTS = NULL;

long BitMatrix::bits_in_m_t = 8*sizeof(matrix_type);

/*!
	The default constructor does not take any parameters and constructs an instance of a "generic" bit matrix,
	with unknown dimensions. The dimensions may be later set by appropriate setter methods.

*/
BitMatrix::BitMatrix() : MSerializable(FILE_NAME_BIT_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->maximal_row_index = -1;
	this->maximal_column_index = -1;
	this->matrix = NULL;
	if (!K_SHIFTS) { DefineKShifts(); }
	this->species = SBitMatrix;
}

/*!
	The second constructor constructs an instance of a bit matrix with known dimensions. Beware that the parameters
	taken mean the actual number of rows and columns, and not their maximal indices; so, if we want to construct
	a matrix of dimensions 17x32, we call
	
		BitMatrix* bm = new BitMatrix(17,32);
		
	Now, we have an instance of a bit matrix; its member variables will be set to:
	
		bm->maximal_row_index = 16;
		
		bm->maximal_column_index = 31;
		
		bm->maximal_allocated_row_index = -1; // no initialization of the data array yet!
		
		bm->maximal_allocated_column_index = -1; // no initialization of the data array yet!
		
	Both of the constructors initialize the data array to NULL pointer, and a real allocation is performed later	- at the time of need. This programming pattern is called lazy initialization, and helps to reduce runtime memory requirements.
	
	Both of the constructors check whether the static member variable K_SHIFTS is equal to zero, and if true, they call DefineKShifts() method in order to initialize the variable.

*/
BitMatrix::BitMatrix(long aRows, long aColumns) : MSerializable(FILE_NAME_BIT_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->maximal_row_index = aRows - 1;
	this->maximal_column_index = aColumns - 1;
	this->matrix = NULL;
	if (!K_SHIFTS) { DefineKShifts(); }
	this->species = SBitMatrix;
}

/*!
	The destructor performs "cleaning up", in this case deallocation of the data array. Its decisions to deallocate
	are based on maximal_allocated_row_index, so it is safe to call it twice. However, I do not see any reason to
	call destructor explicitely; just use the standard C++ pattern
	
		delete bm;
		
	which invConstRC::OKes the destructor implicitely.

*/
BitMatrix::~BitMatrix(){
	if (max_allocated_row_index >= 0){
		long i;
		for (i=0; i<=max_allocated_row_index; i++){
			free(this->matrix[i]);
			this->matrix[i] = NULL;
		}
		max_allocated_row_index = -1;
		free(this->matrix);
		this->matrix = NULL;
	}
}


/*!
	Returns the (protected) value of bits_in_m_t.
	
	Note that unlike other getter methods, this does not have the const modifier. That is because static methods cannot be const (they access no instance-related data, so declaring that they will not change them is meaningless).

*/
int BitMatrix::GetBitsInMT() {
        return bits_in_m_t;
}

/*!
	  This method is used for definition of the K_SHIFTS static array, which is a public member variable of the class.
	  Its length depends on length of matrix_type, and so it cannot be pre-hardcoded into the source code.
	  
	  There is a subtle and not-easily-soluble problem concerning the run of the method. The right place to run this
	  method are the constructors of BitMatrix(), and the constructors must always run O.K. That is why the DefineKShifts()
	  method returns no return code; it would be of no use, since the constructor has no means of telling its caller
	  that it has encountered an error.
	  
	  However, an allocation error MAY theoretically occur in DefineKShifts(), since we really do allocate memory space
	  for the static array. Possible as it is, I still consider this improbable, because of the following facts:
	    - the method is called only once in all program run, at the time of instantiation of the first instance of BitMatrix
	    - the heap space allocated is very small (typically 32*4 or 64*8 bytes, so lesser than 0.5 kB), and, in "normal"
	      systems, there should always be such a small portion of memory available.

*/
void BitMatrix::DefineKShifts() {
	K_SHIFTS = (matrix_type *) Utils::AllocationMachine((void **) &K_SHIFTS,(bits_in_m_t)*sizeof(matrix_type));
	if (K_SHIFTS == NULL) { 
		CLog::LogError(WHEREARG,"BitMatrixClass","error allocating K_SHIFTS",(bits_in_m_t)*sizeof(matrix_type),2001,"There was not enough memory to allocate K_SHIFTS");
	}
	int i;
	for (i=0; i< bits_in_m_t; i++){
		K_SHIFTS[i] = (1 << i);
	}
}

/*!
	  This method takes care of allocation of the internal two-dimensional array of matrix_type, which contains the
	  matrix entries. It tests the need for allocation by evaluating
	  	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1))
	  condition.
	  
	  If there is decision to run the allocation job, two other operations are performed as well. The first one is
	  zeroizing of the result matrix, the other one is computation of max_allocated_column_index by the following
	  formula:
	  	this->max_allocated_column_index = this->maximal_column_index / bits_in_m_t;
	  
	  This is the only place in bit_matrix.cpp, where max_allocated_column_index is determined.
	  
	  Return codes:
	  	
		ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - unsuccessful allocation of the rows and/or columns

*/
int BitMatrix::Allocate(){
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){
		long i;
		this->matrix = (matrix_type **) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(matrix_type*));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"BitMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate bit matrix row");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->matrix,0,(this->maximal_row_index+1)*sizeof(matrix_type*));
		this->max_allocated_row_index = this->maximal_row_index;
		this->max_allocated_column_index = this->maximal_column_index / bits_in_m_t;		
		for (i=0; i<= maximal_row_index; i++){
			this->matrix[i] = (matrix_type *) Utils::AllocationMachine((void **) &(this->matrix[i]),(this->max_allocated_column_index+1)*sizeof(matrix_type));
			if (!this->matrix){
				CLog::LogError(WHEREARG,"BitMatrixClass","error allocating columns",(this->max_allocated_column_index+1),201,"There was not enough memory to allocate bit matrix columns");
				return ConstRC::NotEnoughMemory;
			}
			memset(this->matrix[i],0,(this->max_allocated_column_index+1)*sizeof(matrix_type));
		}
		/*printf("Max alloc row index = %d, max alloc col index = %d, max col index=%d\n\n\n\n\n\n",this->max_allocated_row_index,this->max_allocated_column_index,this->maximal_column_index);
		fflush(stdout);*/
	}
	return ConstRC::Ok;
}

/*!
	  This method at first ensures that the matrix is allocated, by calling Allocate(), and then refills the matrix
	  with random values.
	  The random values are gained from standard randomization in C - calling rand(). The seed is determined
	  at each call of Randomize() by the current clock() value.
	  
	  This method overwrites any previous elements in matrix data array.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.

*/
int BitMatrix::Randomize(){
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok){
		return ret;
	}
	long i, j;
	matrix_type seed = (matrix_type) clock()+time(NULL);
        //printf("BitMatrix: seed is %x.\n",seed);
	srand(seed);

	matrix_type mask = 0;
	/* We must zero out the bit columns whose indices excess maximal_column_index. This is done
	   by masking the random value with a mask of 1's on eligible columns and 0's on invalid columns.
	   This is the construction of the mask.
	   The construction of the mask is a generic operation and need not be repeated at each Randomize()
	   call, but the operation is so fast and so seldom used, that it was left here, instead of changing
	   mask to a static member variable.
	*/
	int diff = bits_in_m_t - ((max_allocated_column_index+1)*bits_in_m_t - (maximal_column_index+1));
	if (diff != bits_in_m_t) {
				   diff %= bits_in_m_t; 
			 	 }
	for (i=0; i< diff; i++){
		mask |= K_SHIFTS[i];
	}
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			/*printf("[%d,%d]\n\n\n\n\n",i,j);
			fflush(stdout);*/
			this->matrix[i][j] = (matrix_type) rand();
			this->matrix[i][j] ^= ((matrix_type) rand()) << 1;// max value of rand() is 0x7fffffff, but we need a full random integer. This is the last bit
			// in case of the highest column+1, we must zero out the excessive bit columns
			if (j == this->max_allocated_column_index){
				this->matrix[i][j] &= mask;
			}
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method at first ensures that the sizes of the calling instance and aSource match and that the calling instance 
	  is allocated, and then copies the entries of aSource into the calling instance, overwriting any previous elements
	  in matrix data array.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions of the calling instance and aSource do not match
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_copy.

*/
int  BitMatrix::Copy(BitMatrix* aSource){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	zero_start = clock();
#endif
	if (this->maximal_row_index != aSource->maximal_row_index)  		return ConstRC::SizeMismatch;
	if (this->maximal_column_index != aSource->maximal_column_index)  	return ConstRC::SizeMismatch;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok){
		return ret;
	}
	long i, j;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			this->matrix[i][j] = aSource->matrix[i][j];
		}
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - zero_start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_copy += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

/*!
	  This method is used to treat the copying of matrices that the compiler things are abstract. The method checks
	  the species of aSource and if it is a BitMatrix, it calls
	  
	  	Copy((BitMatrix*) aSource);
		
	  otherwise it calls the general abstract method.
	  Return codes: the method returns the return code of the launched copy method.

*/
int  BitMatrix::Copy(AbstractMatrix* aSource){
	if (aSource->GetSpecies()==this->GetSpecies()) return this->Copy((BitMatrix*) aSource);
	else return AbstractMatrix::Copy(aSource);
}

/*!
	  This method performs comparison of the calling instance and of aMatrix. It does not call Allocate() to ensure 
	  allocation of anything.
	  The principles for equality are the following:
	  	- if aMatrix is NULL, then the matrices are not equal
		- if the respective dimensions maximal_row_index and maximal_column_index differ, then the matrices are not equal
		- if the respective dimensions max_allocated_row_index and max_allocated_column_index differ, then the matrices are not equal		
		- if any entry on any position differs, then the matrices are not equal.
		
	  The comparison ends as soon as any of the previous conditions is met.
	  
	  This method does not change any data of any operand.
	  
	  Return codes:
	  
	  	TRUE 		   - the matrices are equal
		
		FALSE		   - the matrices are not equal

*/
bool BitMatrix::Equals(BitMatrix* aMatrix){
	if (!aMatrix)		return FALSE; //ConstRC::NullPointerSupplied;
	// performs a "true" comparison of two matrices. They must be equal up to allocation (have the same number of
	// elements allocated
	if (this->max_allocated_row_index != aMatrix->max_allocated_row_index)		return FALSE;
	if (this->max_allocated_column_index != aMatrix->max_allocated_column_index)	return FALSE;
	if (this->maximal_row_index != aMatrix->maximal_row_index)			return FALSE;
	if (this->maximal_column_index != aMatrix->maximal_column_index)		return FALSE;
	long i,j;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){			
			if (this->matrix[i][j] != aMatrix->matrix[i][j])	{	printf("%ld x %ld\n",i,j); return FALSE; }
		}
	}
	return TRUE;
}

/*!
	  This method at first ensures that the matrix is allocated, by calling Allocate(), and then refills the matrix
	  with 0 values, using memset() operation.
	  This method overwrites any previous elements in matrix data array.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.	
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_zeroize.

*/
int  BitMatrix::Zeroize(){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	zero_start = clock();
#endif
	long i = this->max_allocated_row_index;
	long j = this->max_allocated_column_index;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	if ((this->max_allocated_row_index == i) && (this->max_allocated_column_index == j)){
	// zeroizing is needed
		for (i=0; i<= this->max_allocated_row_index; i++){
			memset(this->matrix[i],0,(this->max_allocated_column_index+1)*sizeof(matrix_type));
		}
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - zero_start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_zeroize += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

/*!
	  This method prints the calling instance onto the screen. It does not change any data.
*/
void BitMatrix::PrintToScreen(){
	long i, j, k;
	printf("\n");
	printf("Bit matrix: %ld x %ld\n",this->maximal_row_index+1,this->maximal_column_index+1);
	for (i=0; i<=this->max_allocated_row_index; i++){
		printf("\nRow %4ld:   ",i);
		for (j=0; j<=this->max_allocated_column_index; j++){
			for (k=0; k< this->bits_in_m_t; k++) if(j*bits_in_m_t+k<=maximal_column_index){
				if (this->matrix[i][j] & (1 << k)) printf("1 "); else printf("0 ");
			}
                        printf(" [0x%x] | ",this->matrix[i][j]);
		}
	}
	printf("\n\n");
}

/*!

	Getters and setters for the values in the matrix itself. For historic reasons, the
	setters do not come with Set* prefix, but rather Put* prefix.
		
	All the Put* methods change the value on indices aRow and aColumn to a desired value; 1, 0, or a supplied
	argument aNumber. The least legal index is 0 and the greatest is maximal_row_index (maximal_column_index,
	respectively).
	
	The Put* methods can detect row- and column- overflow and underflow; they
	return error values
	
		ConstRC::ExcessiveRowIndex
		
		ConstRC::ExcessiveColumnIndex
		
		ConstRC::NegativeIndex
		
	defined in definitions.h.
	
	If operation has been done, they return value
	
		ConstRC::Ok
		
	defined in definitions.h
*/
int  BitMatrix::PutOne(long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))		    return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long column_index = aColumn / bits_in_m_t;  // 0 to 31 -> 0, etc.
	long column_bit = aColumn % bits_in_m_t;
	this->matrix[aRow][column_index] |= K_SHIFTS[column_bit];
	return ConstRC::Ok;
}

int  BitMatrix::PutZero(long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))		    return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long column_index = aColumn / bits_in_m_t;  // 0 to 31 -> 0, etc.
	long column_bit = aColumn % bits_in_m_t;
	this->matrix[aColumn][column_index] &= ~K_SHIFTS[column_bit];
	return ConstRC::Ok;
}

/*!

	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  row at given index aRow with zeros (using memset).
	  This method checks validity of aRow index.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveRowIndex-if aRow > maximal_row_index
		
		ConstRC::NegativeIndex	  - if aRow < 0

*/
int  BitMatrix::ZeroizeRow(long aRow){
	int ret;
	if (aRow < 0)				     return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;	
	memset(this->matrix[aRow],0,(this->max_allocated_column_index+1)*sizeof(matrix_type));
	return ConstRC::Ok;
}

/*!
	  This method reads numbers from aRowList array[0 ... aListMaxIndex] and zeroes out rows with those indices.
	  Before this, it checks whether the aRowList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). It checks for validity of all indices and aListMaxIndex.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveRowIndex-if aRow > maximal_row_index
		
		ConstRC::NullPointerSupplied - if aRowList is NULL
		
		ConstRC::NegativeIndex	  - if aListMaxIndex < 0 or any of the indices in list < 0

*/
int  BitMatrix::ZeroizeRow(long* aRowList, long aListMaxIndex){
	if (!aRowList)		return ConstRC::NullPointerSupplied;
	if (aListMaxIndex < 0)				     return ConstRC::NegativeIndex;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long j, index;
	for (j=0; j<= aListMaxIndex; j++){
		if ((index = aRowList[j]) < 0)			     return ConstRC::NegativeIndex;		
		if (index > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		memset(this->matrix[index],0,(this->max_allocated_column_index+1)*sizeof(matrix_type));
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures swap of rows with indices aRow1 and aRow2. If the indices are equal, it does nothing.
	  If they are distinct, it ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.
	  
	  The swap is performed by three XORs.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveRowIndex-if aRow1 or aRow2 > maximal_row_index
		
		ConstRC::NegativeIndex	  - if aRow1 or aRow2 < 0 
		

*/
int  BitMatrix::SwapRows(long aRow1, long aRow2){
	if ((aRow1 < 0) || (aRow2 < 0))				     return ConstRC::NegativeIndex;
	if (aRow1 != aRow2){
		if (aRow1 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		if (aRow2 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		int ret;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		matrix_type *swap;             
		swap = this->matrix[aRow1];
		this->matrix[aRow1]=this->matrix[aRow2];
		this->matrix[aRow2]=swap; 	                         								    
	};
	return ConstRC::Ok;
}

int  BitMatrix::SwapRows(long aRow1, long aRow2, long aMaxIndex){
        if ((aRow1 < 0) || (aRow2 < 0))                              return ConstRC::NegativeIndex;
        if (aRow1 != aRow2){
                if (aRow1 > this->maximal_row_index)                 return ConstRC::ExcessiveRowIndex;
                if (aRow2 > this->maximal_row_index)                 return ConstRC::ExcessiveRowIndex;
                if (this->maximal_column_index < aMaxIndex)          return ConstRC::ExcessiveRowIndex;
                int ret;
                ret = this->Allocate();
                if (ret != ConstRC::Ok)  return ret;
                long k,l;
                l=(aMaxIndex/bits_in_m_t)+((aMaxIndex % bits_in_m_t)!=0)-1;
                for (k=0; k<=l; k++){
                        this->matrix[aRow2][k] ^= this->matrix[aRow1][k];
                        this->matrix[aRow1][k] ^= this->matrix[aRow2][k];
                        this->matrix[aRow2][k] ^= this->matrix[aRow1][k];
                }
        }
        return ConstRC::Ok;
}

/*!
	  This method adds row with index aSource to the row with index aTarget. The indices may be equal, in which
	  case it zeroes out the row by calling ZeroizeRow(aTarget) - multiplication by 2 is equivalent to zero in GF(2).
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveRowIndex-if aTarget or aSource > maximal_row_index
		
		ConstRC::NegativeIndex	  - if aTarget or aSource < 0

*/
int  BitMatrix::AddRows(long aTarget, long aSource){
	int ret;
	if ((aTarget < 0) || (aSource < 0))			     return ConstRC::NegativeIndex;
	if (aTarget != aSource){
		if (aTarget > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		if (aSource > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		for (k=0; k<=this->max_allocated_column_index; k++){
			this->matrix[aTarget][k] ^= this->matrix[aSource][k];
		}

	} else { // adding a row to itself is zeroizing in GF(2)
		ret = ZeroizeRow(aTarget);
		if (ret != ConstRC::Ok) return ret;
	}
	return ConstRC::Ok;
}

/*!
	  This method adds to the row with index aRow the row with index aRow2 of the matrix aOperand.
	  The method ensures the allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveRowIndex-if aTarget or aSource > maximal_row_index
		
		ConstRC::NegativeIndex    - if aTarget or aSource < 0
		
		ConstRC::GeneralError	  - if aOperand2 is not a BitMatrix

*/
int  BitMatrix::AddRows(long aTarget, long aSource, long aMaxIndex){
        int ret;
        if ((aTarget < 0) || (aSource < 0))                          return ConstRC::NegativeIndex;
        if (aTarget != aSource){
                if (aTarget > this->maximal_row_index)               return ConstRC::ExcessiveRowIndex;
                if (aSource > this->maximal_row_index)               return ConstRC::ExcessiveRowIndex;
                if (this->maximal_column_index < aMaxIndex)          return ConstRC::ExcessiveRowIndex;
                ret = this->Allocate();
                if (ret != ConstRC::Ok)  return ret;
                long k,l;
                l=(aMaxIndex/bits_in_m_t)+((aMaxIndex % bits_in_m_t)!=0)-1;
                for (k=0; k<=l; k++){
                        this->matrix[aTarget][k] ^= this->matrix[aSource][k];
                }
        } else { // adding a row to itself is zeroizing in GF(2)
                ret = ZeroizeRow(aTarget);
                if (ret != ConstRC::Ok) return ret;
        }
        return ConstRC::Ok;
}

/*!

	  This method adds to the row with index aRow the row with index aRow2 of the matrix aOperand.
	  The method ensures the allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveRowIndex-if aTarget or aSource > maximal_row_index
		
		ConstRC::NegativeIndex    - if aTarget or aSource < 0
		
		ConstRC::GeneralError	  - if aOperand2 is not a BitMatrix

*/
int  BitMatrix::AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2){
	int ret;
	BitMatrix *aO;
	if ((aRow < 0) || (aRow2 < 0))				     return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)  	  	     return ConstRC::ExcessiveRowIndex;
	if (aRow2 > aOperand->GetMaximalRowIndex())    	     return ConstRC::ExcessiveRowIndex;
	if (aOperand->GetSpecies() != SBitMatrix)	     return ConstRC::GeneralError;
	aO = (BitMatrix*) aOperand;
	ret = aO->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	for (k=0; k<=this->max_allocated_column_index; k++){
		this->matrix[aRow][k] ^= aO->matrix[aRow2][k];
	}
	return ConstRC::Ok;
}

/*!

	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  column at given index aColumn with zeros (by masking the corresponding bit out).
	  The index is checked for validity.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveColumnIndex-if aColumn > maximal_column_index
		
		ConstRC::NegativeIndex	  - if aColumn < 0

*/
int  BitMatrix::ZeroizeColumn(long aColumn){
	int ret;
	if (aColumn < 0)				     return ConstRC::NegativeIndex;
	if (aColumn > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i;
	long col_ind_1, col_bit_1;
	col_ind_1 = aColumn / bits_in_m_t;  // 0 to 31 -> 0, etc.
	col_bit_1 = aColumn % bits_in_m_t;
	matrix_type mask = ~K_SHIFTS[col_bit_1];
	for (i=0; i<=this->max_allocated_row_index; i++){
		this->matrix[i][col_ind_1] &= mask;
	}
	return ConstRC::Ok;
}

/*!
	  This method reads numbers from aColumnList array[0 ... aListMaxIndex] and zeroes out columns with those indices.
	  Before this, it checks whether the aColumnList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). The indices are checked for validity.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveColumnIndex-if aColumn > maximal_column_index
		
		ConstRC::NullPointerSupplied - if aColumnList is NULL
		
		ConstRC::NegativeIndex    - if either aListMaxIndex or any of the indices is < 0

*/
int  BitMatrix::ZeroizeColumn(long* aColumnList, long aListMaxIndex){
	if (!aColumnList)		return ConstRC::NullPointerSupplied;
	if (aListMaxIndex < 0)		return ConstRC::NegativeIndex;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i, j, index;
	long col_ind_1, col_bit_1;

	for (j=0; j<= aListMaxIndex; j++){
		if ((index = aColumnList[j]) < 0)		     return ConstRC::NegativeIndex;
		if (index > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		col_ind_1 = index / bits_in_m_t;  // 0 to 31 -> 0, etc.
		col_bit_1 = index % bits_in_m_t;
		for (i=0; i<=this->max_allocated_row_index; i++){
			this->matrix[i][col_ind_1] &= ~K_SHIFTS[col_bit_1];
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures swap of columns with indices aColumn1 and aColumn2. If the indices are equal, it does nothing.
	  If they are distinct, it ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.
	  
	  The swap is performed iff the corresponding bits differ, by |= operation.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveColumnIndex-if aColumn1 or aColumn2 > maximal_column_index
		
		ConstRC::NegativeIndex    - if aColumn1 or aColumn2 < 0

*/
int  BitMatrix::SwapColumns(long aColumn1, long aColumn2){
	if ((aColumn1 < 0) || (aColumn2 < 0))				     return ConstRC::NegativeIndex;
	if (aColumn1 != aColumn2){
		if (aColumn1 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		if (aColumn2 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		int ret;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		long col_ind_1, col_ind_2, col_bit_1, col_bit_2;
		int bit_1, bit_2;
		col_ind_1 = aColumn1 / bits_in_m_t;  // 0 to 31 -> 0, etc.
		col_bit_1 = aColumn1 % bits_in_m_t;
		col_ind_2 = aColumn2 / bits_in_m_t;  // 0 to 31 -> 0, etc.
		col_bit_2 = aColumn2 % bits_in_m_t;
		for (k=0; k<=this->max_allocated_row_index; k++){
			// do the bits need to be swapped?
			bit_1 = (this->matrix[k][col_ind_1] >> col_bit_1) & 1;
			bit_2 = (this->matrix[k][col_ind_2] >> col_bit_2) & 1;
			if (bit_1 != bit_2){
				this->matrix[k][col_ind_1] &= ~ K_SHIFTS[col_bit_1];
				if (bit_2) this->matrix[k][col_ind_1] |= K_SHIFTS[col_bit_1];
				this->matrix[k][col_ind_2] &= ~K_SHIFTS[col_bit_2];
				if (bit_1) this->matrix[k][col_ind_2] |= K_SHIFTS[col_bit_2];
			}
		}
	}
	return ConstRC::Ok;
}

/*!

	  This method adds column with index aSource to the column with index aTarget. The indices may be equal, in which
	  case it just multiplies the contents of the column by 2.
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  for validity of indices.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveColumnIndex-if aTarget or aSource > maximal_column_index
		
		ConstRC::NegativeIndex    - if aTarget or aSource < 0

*/
int  BitMatrix::AddColumns(long aTarget, long aSource){
	int ret;
	if ((aTarget < 0) || (aSource < 0))				     return ConstRC::NegativeIndex;
	if (aTarget != aSource){
		if (aTarget > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		if (aSource > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		ret = this->Allocate();
		if (ret != ConstRC::Ok)	return ret;
		long k;
		long col_ind_1, col_ind_2, col_bit_1, col_bit_2;
		//matrix_type intermediate;
		col_ind_1 = aTarget / bits_in_m_t;  // 0 to 31 -> 0, etc.
		col_bit_1 = aTarget % bits_in_m_t;
		col_ind_2 = aSource / bits_in_m_t;  // 0 to 31 -> 0, etc.
		col_bit_2 = aSource % bits_in_m_t;
		for (k=0; k<=this->max_allocated_row_index; k++){
			 if (this->matrix[k][col_ind_2] & K_SHIFTS[col_bit_2]){
			 	this->matrix[k][col_ind_1] ^= K_SHIFTS[col_bit_1];  // either 0 or power of 2
			 }
		}
	} else { // adding a column to itself is zeroizing in GF(2)
		ret = ZeroizeColumn(aTarget);
		if (ret != ConstRC::Ok) return ret;
	}
	return ConstRC::Ok;
}

/*!
	  This method adds column with index aColumn2 from aSource matrix to the column with index aColumn1 in aTarget
	  matrix. The method checks validity of indices.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::ExcessiveColumnIndex-if aTarget or aSource > maximal_column_index
		
		ConstRC::NullPointerSupplied if aTarget and/or aSource are NULL.
		
		ConstRC::NegativeIndex    - if aColumn1 or aColumn2 < 0

*/
int  BitMatrix::AddColumnsTwoMatrices(BitMatrix* aTarget, BitMatrix* aSource, long aColumn1, long aColumn2){
	if ((aColumn1 < 0) || (aColumn2 < 0))				     return ConstRC::NegativeIndex;
	if (aColumn1 > aTarget->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aColumn2 > aSource->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aTarget->max_allocated_row_index
		< aSource->max_allocated_row_index)			return ConstRC::ExcessiveRowIndex;
	if (!aTarget->matrix || !aSource->matrix)		     return ConstRC::NullPointerSupplied;

	long k;
	long col_ind_1, col_ind_2, col_bit_1, col_bit_2;	
	col_ind_1 = aColumn1 / bits_in_m_t;  // 0 to 31 -> 0, etc.
	col_bit_1 = aColumn1 % bits_in_m_t;
	col_ind_2 = aColumn2 / bits_in_m_t;  // 0 to 31 -> 0, etc.
	col_bit_2 = aColumn2 % bits_in_m_t;
	for (k=0; k<=aSource->max_allocated_row_index; k++){
		if (aSource->matrix[k][col_ind_2] & K_SHIFTS[col_bit_2]){
			aTarget->matrix[k][col_ind_1] ^= K_SHIFTS[col_bit_1];
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures allocation state, dimension requirements etc., and then transposes the calling instance into
	  the matrix aTarget. The calling instance must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions of the calling instance and aTarget do not match (M x N vs. N x M)
		
		ConstRC::NullPointerSupplied - if aTarget is NULL
		
		ConstRC::BadArgument	  - if aTarget == this.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_transpose.		

*/
int  BitMatrix::Transpose(BitMatrix* aTarget){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	start = clock();
#endif
	if (!aTarget)  							return ConstRC::NullPointerSupplied;
	if (this == aTarget)						return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	long current_column, current_bit, i, j, x, current_segment, current_offset;
	// One by one, we read columns of the original matrix
	for (i=0; i<=this->maximal_column_index; i++)
	{
		current_column = i / bits_in_m_t;
		current_bit    = i % bits_in_m_t;
		x = K_SHIFTS[current_bit];
		for (j=0; j<=this->maximal_row_index; j++){
			current_segment = j / bits_in_m_t;
			current_offset  = j % bits_in_m_t;
			if (this->matrix[j][current_column] & x){
				aTarget->matrix[i][current_segment] |= K_SHIFTS[current_offset];
			}
		}
    	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_transpose += (double) diff;
	}
#endif
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the transposition operation, and then performs
	  the transposition by calling 
	  	int Transpose(NormalMatrix* aTarget)
	  If result has been allocated, but Transpose did not finish well, the result is deleted.
	  
	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)

*/
BitMatrix* BitMatrix::Transpose(){
	BitMatrix* result = new BitMatrix(this->maximal_column_index+1, this->maximal_row_index+1);
	if (!result) return NULL;
	if (this->Transpose(result) != ConstRC::Ok){
		delete result;
		return NULL;
	}
	return result;
}

/*!
	int  Transpose(AbstractMatrix* aTarget);
	  This method is used to treat the transposing of matrices that the compiler things are abstract. The method checks
	  the species of aTarget if is a BitMatrix, it calls
	  	Transpose((BitMatrix*) aTarget);
	  otherwise it calls the general abstract method.
	  
	  Return codes: the method returns the return code of the launched method.

*/
int  BitMatrix::Transpose(AbstractMatrix* aTarget){
	if (aTarget->GetSpecies()==this->GetSpecies()) return this->Transpose((BitMatrix*) aTarget);
	else return AbstractMatrix::Transpose(aTarget);
}

/*!


	  This method ensures allocation state, dimension requirements etc., and then adds aOperand2 to the calling instance
	  and places the result into matrix aTarget. Both the calling instance and aOperand2 must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument      -  if aTarget is equal to any of the operands.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_add.			
*/
int BitMatrix::Add(BitMatrix* aTarget, BitMatrix* aOperand2){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	start = clock();
#endif
	if (!aTarget)  								return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand2) || (aTarget == this))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (!aOperand2)								return ConstRC::NullPointerSupplied;
	if (aOperand2->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aOperand2->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	for (i=0; i<=aTarget->max_allocated_row_index;i++){
		for (j=0; j<= aTarget->max_allocated_column_index;j++){
			aTarget->matrix[i][j] = this->matrix[i][j] ^ aOperand2->matrix[i][j];
		}
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_add += (double) diff;
	}		
#endif
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the addition operation, and then performs
	  the addition by calling 
	  	int Add(BitMatrix* aTarget)
	  If result has been allocated, but Add did not finish well, the result is deleted.
	  
	  Returns:
	  	pointer to result of operation, if everything went all right
		NULL in case of any problem (size mismatches, allocation problems)

*/
BitMatrix* BitMatrix::Add(BitMatrix* aOperand2){
	BitMatrix* result = new BitMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!result) return NULL;
	if (this->Add(result,aOperand2) != ConstRC::Ok){
		delete result;
		return NULL;
	}
	return result;
}

/*!

  	  This method is used to treat the adding of matrices that the compiler things are abstract. The method checks
	  the species of aTarget and aOperand2 and if they are BitMatrices, it calls
	  
	  	Add((BitMatrix*) aTarget, (BitMatrix*) aOperand2);
		
	  otherwise it calls the general abstract method.
	  
	  Return codes: the method returns the return code of the launched method.

*/
int  BitMatrix::Add(AbstractMatrix* aTarget, AbstractMatrix* aOperand2){
	if((aOperand2->GetSpecies()==SBitMatrix)&&(aTarget->GetSpecies()==SBitMatrix))
	  return this->Add((BitMatrix*)aTarget, (BitMatrix*) aOperand2);
	return AbstractMatrix::Add(aTarget, aOperand2);
}

/*!
	  This method is used to treat the multiplying of matrices that the compiler things are abstract. The method checks
	  the species of aTarget, aOperand1 and aOperand2 and if they are BitMatrices, it calls
	  	MultiplyInternal((BitMatrix*) aTarget, (BitMatrix*) aOperand1, (BitMatrix*) aOperand2);
	  otherwise it calls the general abstract method.
	  
	  Return codes: the method returns the return code of the launched method.

*/
int  BitMatrix::MultiplyInternal(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2){
	if((aOperand1->GetSpecies()==SBitMatrix)&&(aOperand2->GetSpecies()==SBitMatrix)&&(aTarget->GetSpecies()==SBitMatrix))
	  return this->MultiplyInternal((BitMatrix*)aTarget, (BitMatrix*) aOperand1, (BitMatrix*) aOperand2);
	return AbstractMatrix::MultiplyInternal(aTarget, aOperand1, aOperand2);
}

/*!
	  This method is used to treat the multiplying of matrices that the compiler things are abstract. The method checks
	  the species of aTarget, aOperand1 and aOperand2 and if they are BitMatrices, it calls
	  
	  	MultiplyInternalTransposed((BitMatrix*) aTarget, (BitMatrix*) aOperand1, (BitMatrix*) aOperand2);
		
	  otherwise it calls the general abstract method.
	  
	  Return codes: the method returns the return code of the launched copy method.

*/
int  BitMatrix::MultiplyInternalTransposed(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2){
	if((aOperand1->GetSpecies()==SBitMatrix)&&(aOperand2->GetSpecies()==SBitMatrix)&&(aTarget->GetSpecies()==SBitMatrix))
	  return this->MultiplyInternalTransposed((BitMatrix*)aTarget, (BitMatrix*) aOperand1, (BitMatrix*) aOperand2);
	return AbstractMatrix::MultiplyInternalTransposed(aTarget, aOperand1, aOperand2);
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_word_multiply.		

*/
int  BitMatrix::MultiplyInternal(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	start = clock();
#endif
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand2) || (aTarget == aOperand1))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_column_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k,l;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;

	//printf("MultiplyInternalWord: aTarget ma rozmery %dx%d\n",aTarget->GetMaxAllocatedRowIndex(),aTarget->GetMaxAllocatedColumnIndex());
	aTarget->Zeroize();
	long max_col = aOperand1->max_allocated_column_index;
	long max_block = aOperand2->GetMaxAllocatedColumnIndex();
	long max_row =  aOperand1->maximal_row_index;
	// for all rows in the first matrix
	for (i = 0; i <= max_row; i++) {			
		// We walk through the whole row ...
		for (j=0; j<= max_col; j++){						
			// .. heeding the values of concrete bits ...
			for (k=0; k< bits_in_m_t; k++){				
				// and if this bit is nonzero, we add the (bits_in_m_t*j+k)-th row to the intermediate sum in target
				if (aOperand1->matrix[i][j] & K_SHIFTS[k]){
					for (l=0; l<=max_block; l++){
						aTarget->matrix[i][l] ^= aOperand2->matrix[bits_in_m_t*j+k][l];						
					}
				}
			}
		}		
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_word_multiply += (double) diff;
		//printf("Pridavame do t.w.m cislo %d, coz je %6.2f sekund, vysledek je %12.4f, coz je %6.2f sekund.\n",diff,diff/(double) CLOCKS_PER_SEC,total_word_multiply,total_word_multiply/(double) CLOCKS_PER_SEC);
	}		
#endif
	return ConstRC::Ok;
}

/*!

	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication (aOperand1)^T x (aOperand2), saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	  
	  	ConstRC::Ok		  - everything all right
		
		ConstRC::NotEnoughMemory - thrown from Allocate() call; there was not enough memory to allocate the requested space.
		
		ConstRC::SizeMismatch     - the dimensions do not match
		
		ConstRC::NullPointerSupplied - if aOperand1, aOperand2 or aTarget is NULL
		
		ConstRC::BadArgument      -  if aTarget is equal to any of the operands.
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_word_transposed_multiply.		
*/
int  BitMatrix::MultiplyInternalTransposed(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2){
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	start = clock();
#endif
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand2) || (aTarget == aOperand1))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_column_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_row_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,l;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;

	long current_column, current_bit;
	//printf("MultiplyInternalWord: aTarget ma rozmery %dx%d\n",aTarget->GetMaxAllocatedRowIndex(),aTarget->GetMaxAllocatedColumnIndex());
	aTarget->Zeroize();	
	long max_real_col = aOperand1->maximal_column_index;
	long max_block = aOperand2->GetMaxAllocatedColumnIndex();
	long max_row =  aOperand1->maximal_row_index;
	// for all columns in the first  matrix ...
	for (i = 0; i <= max_real_col; i++) {
	        current_column = i / bits_in_m_t;
		current_bit    = i % bits_in_m_t;
		for (j=0; j <= max_row; j++){
		// we read the column and if the corresponding bit is 1 ...
			if (aOperand1->matrix[j][current_column] & K_SHIFTS[current_bit]){
				for (l=0; l<= max_block; l++){
					aTarget->matrix[i][l] ^= aOperand2->matrix[j][l];
				}				
			}
		}
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_word_transpose_multiply += (double) diff;
		//printf("Pridavame do t.w.m cislo %d, coz je %6.2f sekund, vysledek je %12.4f, coz je %6.2f sekund.\n",diff,diff/(double) CLOCKS_PER_SEC,total_word_multiply,total_word_multiply/(double) CLOCKS_PER_SEC);
	}		
#endif
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternal(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.
	  
	  Returns:
	  
	  	pointer to result of operation, if everything went all right
		
		NULL in case of any problem (size mismatches, allocation problems)			  

*/
BitMatrix* BitMatrix::MultiplyInternal(BitMatrix* aOperand1, BitMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	BitMatrix* result = new BitMatrix(aOperand1->maximal_row_index+1, aOperand2->maximal_column_index+1);
	if (!result) return NULL;
	if (MultiplyInternal(result,aOperand1,aOperand2) != ConstRC::Ok){
		delete result;
		return NULL;
	}
	return result;
}

BitMatrix* BitMatrix::MultiplyInternalTransposed(BitMatrix* aOperand1, BitMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	BitMatrix* result = new BitMatrix(aOperand1->maximal_column_index+1, aOperand2->maximal_column_index+1);
	if (!result) return NULL;
	if (MultiplyInternalTransposed(result,aOperand1,aOperand2) != ConstRC::Ok){
		delete result;
		return NULL;
	}
	return result;
}

/*!
	  This method instantiates a new NormalMatrix instance of the same dimensions as the caller instance of BitMatrix; then,
	  it allocates the internal matrix of the NormalMatrix instance and converts the matrix represented by caller BitMatrix
	  instance into this freshly allocated NormalMatrix instance.
	  
	  Return values:
	  
	  	pointer to NormalMatrix* instance if everything went all right
		
		NULL if some of the allocations was unsuccessful.

*/
NormalMatrix* BitMatrix::ToNormalMatrix(){
	NormalMatrix* aResult = new NormalMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j,k, index;
	for (i=0; i<=this->max_allocated_row_index; i++){
		for (j=0; j<=this->max_allocated_column_index; j++){
			for (k=0; k< this->bits_in_m_t; k++){
				index = j*bits_in_m_t + k;
				if (index <= this->maximal_column_index){
					if (this->matrix[i][j] & K_SHIFTS[k]){
						aResult->PutOne(i,index);
					}
				}
			}
		}
	}
	return aResult;
}

/*!
	  This method instantiates a new BitMatrix instance of the same dimensions as the parameter aMatrix; then,
	  it allocates the internal matrix of the BitMatrix instance and converts the matrix represented by aMatrix
	  instance into this freshly allocated BitMatrix instance. The rule is that odd entries of aMatrix are turned to bit 1
	  and even entries are turned to bit 0.
	  
	  Return values:
	  
	  	pointer to BitMatrix* instance if everything went all right
		
		NULL if some of the allocations was unsuccessful.

*/
BitMatrix* BitMatrix::ToBitMatrix(NormalMatrix* aMatrix){
	long max_r = aMatrix->GetMaximalRowIndex();
	long max_c = aMatrix->GetMaximalColumnIndex();
	BitMatrix* aResult = new BitMatrix(max_r+1,max_c+1);
	if (!aResult) 		return NULL;
	int ret;
	ret = aResult->Allocate();
	if (ret != ConstRC::Ok)	{
				delete aResult;
				return NULL;
			}
	long i,j;
	normal_matrix_type read;

	for (i=0; i<=max_r; i++){
		for (j=0; j<=max_c; j++){
			read = aMatrix->GetXY(i,j);
			if (read == ConstRC::GeneralError){
				delete aResult;
				return NULL;
			}
			if (read % 2){
				aResult->PutOne(i,j);
			}
		}
	}
	return aResult;
}

/*!
	The getter methods IsOne and IsZero return values
	
	 	TRUE
		
		FALSE
		
	defined in definitions.h according to whether the bit in aRow-row and aColumn-column is nonzero.
	
	If aRow and/or aColumn 
	exceeds maximal_row_index or maximal_column_index, or is below 0, or the matrix has not been allocated, the return value is 
	
		ConstRC::GeneralError
		
	defined in definitions.h. 

*/
int  BitMatrix::IsOne(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))		return ConstRC::GeneralError;
	if (aRow > this->maximal_row_index) 		return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index) 	return ConstRC::GeneralError;
	if (!this->matrix)				return ConstRC::GeneralError;
	long current_column = aColumn / bits_in_m_t;
	long current_bit = aColumn % bits_in_m_t;
	if ((this->matrix[aRow][current_column] & K_SHIFTS[current_bit]))	{
		//printf("%d,%d; %d\n",aRow,aColumn,this->matrix[aRow][current_column] & (1 << current_bit));
		return TRUE;
	}
	else return FALSE;
}

int  BitMatrix::IsZero(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))		return ConstRC::GeneralError;
	if (aRow > this->maximal_row_index) 		return ConstRC::GeneralError;
	if (aColumn > this->maximal_column_index) 	return ConstRC::GeneralError;
	if (!this->matrix)				return ConstRC::GeneralError;
	long current_column = aColumn / bits_in_m_t;
	long current_bit = aColumn % bits_in_m_t;
	if (this->matrix[aRow][current_column] & K_SHIFTS[current_bit])	return FALSE;
	else return TRUE;
}

/*!
	  This method tests whether the calling instance is a zero matrix. It does not call Allocate() to ensure 
	  allocation of anything.
	  The comparison ends as soon as any nonzero matrix_type is met.
	  This method does not change any data of any operand.
	  
	  Return codes:
	  
	  	TRUE 		   - the matrix is a zero matrix
		
		FALSE		   - the matrix is not a zero matrix
		
 	  If the symbolic constant
	  
	        MATRIX_OPERATIONS_TIME_MESSAGE
		
          is defined, the runtime of this method (in processor cycles) is being collected into AbstractMatrix::total_iszero.		

*/
int  BitMatrix::IsZero(){
	long i,j;
	if (!this->matrix)	return ConstRC::GeneralError;
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	zero_start = clock();
#endif
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			if (this->matrix[i][j]) return FALSE;
		}
	}
#ifdef MATRIX_OPERATIONS_TIME_MESSAGE
	diff = clock() - zero_start;
	if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
		total_iszero += (double) diff;
	}
#endif
	return TRUE;
}

int  BitMatrix::IsZeroRow(long aRow){
	long j;
	if (!this->matrix)	return ConstRC::GeneralError;
	for (j=0; j<= this->max_allocated_column_index; j++){
		if (this->matrix[aRow][j]) return FALSE;
	}
	return TRUE;
}

/*!
	  This method counts the total number of nonzero bits in the calling instance matrix. It does not call Allocate() to ensure
	  allocation of anything; the return value for an unallocated matrix is 0, since the counting cycles rely on 
	  max_allocated_row_index and max_allocated_column index, and thus do not run at all. This method does not change any data.	  
	  
	  Return value: the number of nonzero bits in the calling instance matrix. The return value will overflow if the matrix contains
	  more than MAXINT nonzero bits (in normal i386 case, 2^31).

*/
int  BitMatrix::NonzeroElements(){
	long i,j,k;
	int result = 0;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			for (k=0; k< bits_in_m_t; k++){
				if (this->matrix[i][j] & K_SHIFTS[k]){
					result++;
				}
			}				
		}
	}
	return result;
}

/*!
	The last getter/setter methods are designed to influence the whole block of bits (one matrix_type variable).	
	
		matrix_type GetBlock(long aRow, long aBlock) 
				
		void SetBlock(long aRow, long aBlock, matrix_type aValue) 
		
	Both of the methods operate on a whole block in row of index aRow, array column with index aBlock. Remember,
	the aBlock value tells the index in the two-dimensional array, not in the "logical" matrix. That is why
	aBlock must be <= max_allocated_column_index.
	
	The methods check whether the matrix is allocated and legality of aRow and aBlock values; the Get method returns 
	
		ConstRC::GeneralError
		
	if the checks fail, the Set method simply does nothing.

*/
matrix_type BitMatrix::GetBlock(long aRow, long aBlock) {
	if ((aRow < 0) || (aBlock < 0) || (this->matrix == NULL))	return ConstRC::GeneralError;
	if (aRow > this->maximal_row_index) 		return ConstRC::GeneralError;
	if (aBlock > this->max_allocated_column_index)  return ConstRC::GeneralError;
	return this->matrix[aRow][aBlock];
}

void BitMatrix::SetBlock(long aRow, long aBlock, matrix_type aValue) {
	if ((aRow < 0) || (aBlock < 0) || (this->matrix == NULL))	return;
	if (aRow > this->maximal_row_index) 		return;
	if (aBlock > this->max_allocated_column_index)  return;
	this->matrix[aRow][aBlock] = aValue;
}

/*!
	  This method uses linear combinations of rows in order to turn the caller instance into a semidiagonal matrix, with
	  all the nonzero entries above or on the main diagonal. It is designed to run on square matrices only.
	  
	  Return values:
	  
	  	ConstRC::Ok if everything went all right
		
		ConstRC::NullPointerSupplied if the calling matrix has not been allocated yet.

*/
int BitMatrix::ToSemidiagonalShape(){
	long i,j;
	if (!this->matrix)   	return ConstRC::NullPointerSupplied;
	//if (this->maximal_row_index != this->maximal_column_index) return ConstRC::BadArgument;
	// use on square matrices only
	long shift = 0;
	int ret;
	for (i=0; i<=this->maximal_column_index	; i++){
		if ((i<= this->maximal_row_index) && (i+shift <= this->maximal_column_index)){
			for (j=i; j<= this->maximal_row_index; j++){
				if (this->IsZero(i,i+shift) && this->IsOne(j,i+shift)){
					ret = this->SwapRows(i,j);
					if (ret != ConstRC::Ok){
						return ret;
					}
					break;
				}
			}
			if (this->IsOne(i,i+shift)){
				for (j=i+1; j<=this->maximal_row_index; j++){
					if (this->IsOne(j,i+shift)){
						ret = this->AddRows(j,i);
						if (ret != ConstRC::Ok){
							return ret;
						}
					}
				}
			} else {
				shift++;
				i--;
			}

		}
	}
	return ConstRC::Ok;
}


/*!
This method takes as an input matrix aOperand, which has at most bits_in_m_t columns (that is, the value 
	  aOperand->max_allocated_column_index == 0). The aTarget and aOperand matrices must be of the same dimensions.
	  Proper allocation of both aTarget and aOperand is ensured by call of Allocate().
	  The operation is masking of the aOperand entries with aMask, which is equivalent to zeroizing of those columns,
	  whose indices are given by 0 bits in aMask. In turn, this is equivalent to multiplying of aOperand by a 
	  bits_in_m_t x bits_in_m_t matrix, which has zeros outside the main diagonal, and 1s on these indices of main
	  diagonal, where aMask & K_SHIFTS[index] != 0.
	  
	  This is an auxiliary method for the block Lanczos method.
	  
	  Return values:
	  
	  	ConstRC::Ok 		if everything went all right
		
		ConstRC::SizeMismatch   if the requirements on sizes of matrices aOperand and aTarget are not met
		
		ConstRC::NotEnoughMemory if Allocate() fails.
	  
*/
int BitMatrix::PerformColumnMask(BitMatrix* aTarget, BitMatrix* aOperand, matrix_type aMask) {	
	if (!aOperand || !aTarget)   			{
		return ConstRC::NullPointerSupplied;
	}	
	int ret;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok) {
		return ret;
	}
	ret = aOperand->Allocate();
	if (ret != ConstRC::Ok) {
		return ret;
	}
	if (aTarget->maximal_row_index != aOperand->maximal_row_index)		{
		return ConstRC::SizeMismatch;
	}
	if (aTarget->maximal_column_index != aOperand->maximal_column_index){
		return ConstRC::SizeMismatch;
	}
	if (aTarget->max_allocated_column_index) {
		return ConstRC::SizeMismatch;
	}
	int i;
	for (i=0; i<=aTarget->max_allocated_row_index; i++){
		aTarget->matrix[i][0] = aOperand->matrix[i][0] & aMask;
	}
	return ConstRC::Ok;
}

int  BitMatrix::PerformRowMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask)
{
	if(aTarget->GetSpecies()!=SBitMatrix) return ConstRC::GeneralError;
	if(aOperand->GetSpecies()!=SBitMatrix) return ConstRC::GeneralError;
	return this->PerformRowMask((BitMatrix*)aTarget, (BitMatrix*) aOperand, aMask);
}


int  BitMatrix::PerformRowMask(BitMatrix* aTarget, BitMatrix* aOperand, matrix_type aMask)
{
	if (!aOperand || !aTarget)   			{
		return ConstRC::NullPointerSupplied;
	}	
	int ret;
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok) {
		return ret;
	}
	ret = aOperand->Allocate();
	if (ret != ConstRC::Ok) {
		return ret;
	}
	if (aTarget->maximal_row_index != aOperand->maximal_row_index)		{
		return ConstRC::SizeMismatch;
	}
	if (aTarget->maximal_column_index != aOperand->maximal_column_index){
		return ConstRC::SizeMismatch;
	}
	if (aTarget->max_allocated_column_index) {
		return ConstRC::SizeMismatch;
	}
	if ((aOperand->maximal_row_index+1) != bits_in_m_t)
	{
		return ConstRC::SizeMismatch;
	}
	int i;
	aTarget->Copy(aOperand);
	for (i=0; i< bits_in_m_t; i++){
		if (!(aMask & K_SHIFTS[i]))
		{
			aTarget->ZeroizeRow(i);
		}
	}
	return ConstRC::Ok;
}

int  BitMatrix::PerformColumnMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask){
	if(aTarget->GetSpecies()!=SBitMatrix) return ConstRC::GeneralError;
	if(aOperand->GetSpecies()!=SBitMatrix) return ConstRC::GeneralError;
	return this->PerformColumnMask((BitMatrix*)aTarget, (BitMatrix*) aOperand, aMask);
}

/*!
	  This method uses linear combinations of rows in order to turn the caller instance into a semidiagonal matrix, with
	  all the nonzero entries above or on the main diagonal. It is designed to run on square matrices only. 
	  All the linear combinations of rows are in the caller instance are reflected in matrix aMirror (e.g. if one adds
	  row 3 to row 5 in the caller instance, it also adds row 3 to row 5 in aMatrix).
	  This is an auxiliary method for the block Lanczos method.
	  
	  Return values:
	  
	  	ConstRC::Ok if everything went all right
		
		ConstRC::NullPointerSupplied if the calling matrix has not been allocated yet.
		
		ConstRC::SizeMismatch  if the sizes of the caller instance and aMirror differ

*/
int BitMatrix::ToSemidiagonalShapeForLanczos(BitMatrix* aMirror){
	long i,j;
	if (!this->matrix)   		return ConstRC::NullPointerSupplied;
	if (!aMirror->matrix)   	return ConstRC::NullPointerSupplied;
	//if (this->maximal_row_index != this->maximal_column_index) return ConstRC::BadArgument;
	if (this->maximal_row_index != aMirror->maximal_row_index) return ConstRC::SizeMismatch;
	// use on square matrices only
	long shift = 0;
	for (i=0; i<=this->maximal_column_index	; i++){
		if ((i<= this->maximal_row_index) && (i+shift <= this->maximal_column_index)){
			for (j=i; j<= this->maximal_row_index; j++){
				if (this->IsZero(i,i+shift) && this->IsOne(j,i+shift)){
					this->SwapRows(i,j);
					aMirror->SwapRows(i,j);
					break;
				}
			}
			if (this->IsOne(i,i+shift)){
				for (j=i+1; j<=this->maximal_row_index; j++){
					if (this->IsOne(j,i+shift)){
						this->AddRows(j,i);
						aMirror->AddRows(j,i);
					}
				}
				} else {
				shift++;
				i--;
			}
		}
	}
	return ConstRC::Ok;
}

int BitMatrix::ToSemidiagonalShapeForCFRAC(BitMatrix* aMirror,long aRows, long aCols){
        long i,j;
        if (!this->matrix)              return ConstRC::NullPointerSupplied;
        if (!aMirror->matrix)           return ConstRC::NullPointerSupplied;
        if (this->maximal_row_index != aMirror->maximal_row_index) return ConstRC::SizeMismatch;
        if (this->maximal_row_index < aRows-1)  return ConstRC::ExcessiveRowIndex;
        if (this->maximal_row_index < aCols-1)  return ConstRC::ExcessiveColumnIndex;
        long shift = 1;
        for (i=0; i<aCols       ; i++){
                if ((i< aRows) && (i+shift <= aCols)){
                        for (j=i; j< aRows; j++){
                                if (this->IsZero(i,aCols-i-shift) && this->IsOne(j,aCols-i-shift)){
                                        this->SwapRows(i,j,aCols);
                                        aMirror->SwapRows(i,j,aCols);
                                        break;
                                }
                        }
                        if (this->IsOne(i,aCols-i-shift)){
                                for (j=i+1; j<aRows; j++){
                                        if (this->IsOne(j,aCols-i-shift)){
                                                this->AddRows(j,i,aCols);
                                                aMirror->AddRows(j,i,aCols);
                                        }
                                }
                                } else {
                                shift++;
                                i--;
                        }
                }
        }
        return ConstRC::Ok;
}

long BitMatrix::GetMaxAllocatedRowIndex()    const {
	return this->max_allocated_row_index;
}

long BitMatrix::GetMaxAllocatedColumnIndex() const {
	return this->max_allocated_column_index;
}

/*!
	  This method creates a copy of the caller instance, performs ToSemidiagonalShape() on this copy and then counts number
	  of nonzero rows. This number is equal to the rank of the caller instance.
	  Possible errors may result from calls of Copy() and ToSemidiagonalShape().
	  This method does not change the contents of the caller instance.
	  
	  Return values:
	  
	  	nonnegative integer equal to the rank of the caller instance if everything went all right
		
		ConstRC::GeneralError  otherwise

*/
int  BitMatrix::CalculateRank() {
	BitMatrix* result = new BitMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	int res;	
	res = result->Copy(this);
	if (res != ConstRC::Ok){
		delete result;
		return ConstRC::GeneralError;
	}
	res = result->ToSemidiagonalShape();
	if (res != ConstRC::Ok){
		delete result;
		return ConstRC::GeneralError;
	}
	long i,j, rank = 0;
	for (i=0; i<= result->maximal_row_index; i++){
		for (j=0; j<= result->max_allocated_column_index; j++){
			if (result->matrix[i][j]){
				rank++;
				break;
			}
		}
	}
	delete result;
	return rank;
}

/*!
	  This method calls ToSemidiagonalShape() on the caller instance, thereby altering its contents. Then it counts number
	  of nonzero rows. This number is equal to the rank of the caller instance.
	  This approach is usable in situation, when we want to compute rank of an unused matrix, without need to allocate more
	  memory. For example, it is used in optional rank computations in Lanczos method.
	  
	  Return values:
	  
	  	nonnegative integer equal to the rank of the caller instance if everything went all right
		
		ConstRC::GeneralError  otherwise

*/
int  BitMatrix::CalculateRank(bool aOnMyself) {
	int res = this->ToSemidiagonalShape();
	if (res != ConstRC::Ok){
		return ConstRC::GeneralError;
	}
	long i,j, rank = 0;
	for (i=0; i<= this->maximal_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			if (this->matrix[i][j]){
				rank++;
				break;
			}
		}
	}
	return rank;
}

int BitMatrix::Save(char *aName)
{
        FILE *fw;
	long i;
        fw=fopen(aName,"w");
        if (!fw) return ConstRC::GeneralError;
        fwrite(&(this->maximal_row_index),1,sizeof(long),fw);
        fwrite(&(this->maximal_column_index),1,sizeof(long),fw);
	for(i=0;i<=maximal_row_index;i++){
	        fwrite(this->matrix[i],this->max_allocated_column_index+1,sizeof(matrix_type),fw);
	}
        fclose(fw);
	return ConstRC::Ok;
}

int BitMatrix::Load(char *aName)
{
        FILE *fr;
	long i;
        if (this->matrix) return ConstRC::GeneralError;
        fr=fopen(aName,"r");
        if (!fr) return ConstRC::GeneralError;
        fread(&(this->maximal_row_index),1,sizeof(long),fr);
        fread(&(this->maximal_column_index),1,sizeof(long),fr);
        this->Allocate();
	for(i=0;i<=maximal_row_index;i++){
        	fread(this->matrix[i],this->max_allocated_column_index+1,sizeof(matrix_type),fr);
	}
        fclose(fr);
	return ConstRC::Ok;
}

//int BitMatrix::Serialize(const CBaseParameters &aParam) const
//{
//	return MSerializable::Serialize(aParam);
//}
//
//int BitMatrix::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
//{
//	return AbstractMatrix::Serialize(aParam, aWriter);
//}
//
//int BitMatrix::Deserialize(const CBaseParameters &aParam)
//{
//	return MSerializable::Deserialize(aParam);
//}
//
//int BitMatrix::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
//{
//	return AbstractMatrix::Deserialize(aParam, aReader);
//}

int BitMatrix::WriteData(xmlTextWriterPtr aWriter) const
{
	int rc;
	// write all matrix data
	for (long i = 0; i <= maximal_row_index; i++)
	{
		rc = CXMLHelper::SerializeBinaryData(aWriter, XML_SERIAL_MATRIX_ROW_NAME, this->matrix[i], this->max_allocated_column_index + 1,  sizeof(matrix_type));  
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

int BitMatrix::ReadData(xmlTextReaderPtr aReader)
{
	int rc;
	xmlChar * name;

	// we know max_row_index and max_column_index
	Allocate();
	long i;
	// we deserialize every row and get max nonzero columns in row and nonzero indexies
    for (i = 0; i <= this->maximal_row_index; i++)
	{
		name = xmlTextReaderName(aReader);

		if (xmlStrEqual(name,BAD_CAST XML_SERIAL_MATRIX_ROW_NAME) == 0)
		{
			break;
		}

		rc = CXMLHelper::DeserializeBinaryDataIn(aReader, "Deserialize matrix row...", this->matrix[i], this->max_allocated_column_index + 1, sizeof(matrix_type), FALSE);
        if (rc != ConstRC::Ok)
		{
			return rc;
		}

		rc = CXMLHelper::MoveToNextNode(aReader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::EndOfFile)
			{
				return ConstRC::Ok;
			}
			return rc;
		}
    }

	return ConstRC::Ok;
}
