#include "candidate_class.h"

#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_filenames.h"
#include "base_parameters.h"
#include "xml_helper.h"
#include "log.h"


/*!
	Prepares the candidate polynomials arrays.
*/
CandidateClass::CandidateClass(long aAllocation) : MSerializable(ConstFN::CANDIDATE_POLYS)
{
    this->allocated = aAllocation;
	this->Poly = NULL;
	this->Root = NULL;
	this->Skewness = NULL;
	this->Alpha = NULL;
	this->Rating = NULL;

	Allocate();
}

CandidateClass::~CandidateClass()
{
	Deallocate();
}

int CandidateClass::Allocate()
{
	if (this->Poly == NULL)
	{
		if (this->allocated % 2)
		{
			this->allocated++;
		}
		this->Poly = NULL;
		this->Poly = (Polynomial**) Utils::AllocationMachine((void**)(&Poly),allocated*sizeof(Polynomial*));
		this->Root = NULL;
		this->Root = (mpz_t*) Utils::AllocationMachine((void**)(&Root),allocated*sizeof(mpz_t));
		this->Skewness = NULL;
		this->Skewness = (mpf_t*) Utils::AllocationMachine((void**)(&Skewness),allocated*sizeof(mpf_t));
		this->Alpha = NULL;
		this->Alpha = (mpf_t*) Utils::AllocationMachine((void**)(&Alpha),allocated*sizeof(mpf_t));
		this->Rating = NULL;
		this->Rating = (mpf_t*) Utils::AllocationMachine((void**)(&Rating),allocated*sizeof(mpf_t));
		for(long i = 0; i < allocated; i++)
		{
			Poly[i] = new Polynomial();
			mpz_init(this->Root[i]);
			mpf_init_set_si(this->Skewness[i], 1);
			mpf_init(this->Alpha[i]);
			mpf_init(this->Rating[i]);
		}
	}

	return ConstRC::Ok;
}
int CandidateClass::Deallocate()
{
	if (this->Poly != NULL)
	{
		for (long i = 0; i < allocated; i++)
		{
			delete Poly[i];
			mpz_clear(Root[i]);
			mpf_clear(Skewness[i]);
			mpf_clear(Alpha[i]);
			mpf_clear(Rating[i]);
		}
		free(this->Poly);
		this->Poly = NULL;
		free(Root);
		this->Root = NULL;
		free(Skewness);
		this->Skewness = NULL;
		free(Alpha);
		this->Alpha = NULL;
		free(Rating);
		this->Rating = NULL;
	}

	return ConstRC::Ok;
}
int CandidateClass::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CandidateClass::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	std::string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_CANDIDATE_POLYS_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CandidateClass","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_ASSIGNED_ELEMENTS_NAME, this->allocated);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENTS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CandidateClass","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	long max = this->allocated;
	long maxRated = this->allocated;
	aParam.GetLongParameter(ConstPN::MAX_ELEMENT_COUNT, max);
	aParam.GetLongParameter(ConstPN::MAX_RATED_ELEMENT_COUNT, maxRated);
	std::string auxStr;

	for (long i = 0; i < max; i++)
	{
		rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENT_NAME);
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CandidateClass","Serialize",rc,0,"Could not write the start of the element.");
    		return ConstRC::GeneralError;
		}
		rc = CXMLHelper::NewLine(aWriter, FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		auxStr = this->Poly[i]->ToString(ELongFormat);
		rc = CXMLHelper::SerializeString(aWriter, XML_SERIAL_POLYNOMIAL_NAME, auxStr);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_ROOT_NAME, this->Root[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CXMLHelper::SerializeMpf(aWriter, XML_SERIAL_SKEWNESS_NAME, this->Skewness[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (i < maxRated)
		{
			rc = CXMLHelper::SerializeMpf(aWriter, XML_SERIAL_ALPHA_NAME, this->Alpha[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CXMLHelper::SerializeMpf(aWriter, XML_SERIAL_RATING_NAME, this->Rating[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENT_NAME
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CandidateClass","Serialize",rc,0,"Could not end the element.");
			return ConstRC::GeneralError;
		}
		rc = CXMLHelper::NewLine(aWriter,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENTS_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CandidateClass","Serialize",rc,0,"Could not end the element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CandidateClass","Serialize",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);

	return rc;
}
int CandidateClass::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CandidateClass::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	std::string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_CANDIDATE_POLYS_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CandidateClass","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CandidateClass","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);
	long auxLong;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ASSIGNED_ELEMENTS_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize assigned elements...", auxLong, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			
			if (this->allocated < auxLong)
			{
				Deallocate();
				this->allocated = auxLong;
				Allocate();
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENTS_NAME))
		{
			std::string auxStr;
			int index = 0;
			int elementsDepth = xmlTextReaderDepth(aReader);
			int elementDepth = elementsDepth + 1;
			while (CXMLHelper::MoveToNextNode(aReader, elementsDepth) == ConstRC::Ok 
				&& index < this->allocated) // element
			{
				while (CXMLHelper::MoveToNextNode(aReader, elementDepth) == ConstRC::Ok) // inner stuffs
				{
					name = xmlTextReaderName(aReader);
	
					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_POLYNOMIAL_NAME))
					{
						rc = CXMLHelper::DeserializeString(aReader, "Deserialize poly...", auxStr, FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						rc = Polynomial::Parse2(this->Poly[index], auxStr);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}

						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ROOT_NAME))
					{
						rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize root...", this->Root[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ALPHA_NAME))
					{
						rc = CXMLHelper::DeserializeMpf(aReader, "Deserialize alpha...", this->Alpha[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_RATING_NAME))
					{
						rc = CXMLHelper::DeserializeMpf(aReader, "Deserialize rating...", this->Rating[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_SKEWNESS_NAME))
					{
						rc = CXMLHelper::DeserializeMpf(aReader, "Deserialize skewness...", this->Skewness[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}
				} // inner stuffs

				index++;
			} // element
			continue;
		} // elements
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}

