#include "candidate_pairs.h"

#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_filenames.h"
#include "base_parameters.h"
#include "log.h"



/*!
	Prepares the candidate polynomials arrays.
*/
CCandidatePairs::CCandidatePairs(long aAllocation) : MSerializable(ConstFN::CANDIDATE_POLY_PAIRS)
{
	this->allocated = aAllocation;
	this->count = 0;
	this->FirstPoly = NULL;
	this->SecondPoly = NULL;
	this->Root = NULL;
	this->Alpha = NULL;
	this->Rating = NULL;
	this->Skewness = NULL;

	Allocate();
}

CCandidatePairs::~CCandidatePairs()
{
	Deallocate();
}

int CCandidatePairs::AddCandidatePair(Polynomial * aFirstPoly, Polynomial * aSecondPoly, mpz_t aRoot, mpf_t aAlpha, mpf_t aRating, mpf_t aSkewness)
{
	if (aFirstPoly == NULL || aSecondPoly == NULL)
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","AddCandidatePair",0,0,"Polynomials are NULL.");
		return ConstRC::BadArgument;
	}

	if (this->count > 0)
	{
		if (mpf_cmp(this->Rating[this->count - 1], aRating) > 0)
		{
			if (this->count < this->allocated)
			{
				// enought space
				// save result
				aFirstPoly->Copy(this->FirstPoly[this->count]);
				aSecondPoly->Copy(this->SecondPoly[this->count]);
				mpz_set(this->Root[this->count], aRoot);
				mpf_set(this->Alpha[this->count], aAlpha);
				mpf_set(this->Rating[this->count], aRating);
				mpf_set(this->Skewness[this->count], aSkewness);
				this->count++;
			}
			else
			{
				// not good enough and no space
				return ConstRC::Ok;
			}
		}
		else
		{
			int i = this->count - 2;
			
			if (this->count < this->allocated)
			{
				// shift last pair if some space
				this->FirstPoly[this->count - 1]->Copy(this->FirstPoly[this->count]);
				this->SecondPoly[this->count - 1]->Copy(this->SecondPoly[this->count]);
				mpz_set(this->Root[this->count], this->Root[this->count - 1]);
				mpf_set(this->Alpha[this->count], this->Alpha[this->count - 1]);
				mpf_set(this->Rating[this->count], this->Rating[this->count - 1]);
				mpf_set(this->Skewness[this->count], this->Skewness[this->count - 1]);
				this->count++;
			}

			// find how good new pair is
			for (; i >= 0; i--)
			{
				if (mpf_cmp(this->Rating[i], aRating) > 0)
				{
					//new pair is worse than this pair
					//save the new one to the last position
					aFirstPoly->Copy(this->FirstPoly[i + 1]);
					aSecondPoly->Copy(this->SecondPoly[i + 1]);
					mpz_set(this->Root[i + 1], aRoot);
					mpf_set(this->Alpha[i + 1], aAlpha);
					mpf_set(this->Rating[i + 1], aRating);
					mpf_set(this->Skewness[i + 1], aSkewness);
					// no reason to continue
					break;
				}
				else
				{
					//new pair is better - shift this pair
					this->FirstPoly[i]->Copy(this->FirstPoly[i + 1]);
					this->SecondPoly[i]->Copy(this->SecondPoly[i + 1]);
					mpz_set(this->Root[i + 1], this->Root[i]);
					mpf_set(this->Alpha[i + 1], this->Alpha[i]);
					mpf_set(this->Rating[i + 1], this->Rating[i]);
					mpf_set(this->Skewness[i + 1], this->Skewness[i]);
				}
			}

			if (i < 0)
			{
				// first pair is worse than the new one
				//save the new one to the first position
				aFirstPoly->Copy(this->FirstPoly[0]);
				aSecondPoly->Copy(this->SecondPoly[0]);
				mpz_set(this->Root[0], aRoot);
				mpf_set(this->Alpha[0], aAlpha);
				mpf_set(this->Rating[0], aRating);
				mpf_set(this->Skewness[0], aSkewness);
			}
		}
	}
	else if (this->allocated > 0)
	{
		// save result
		aFirstPoly->Copy(this->FirstPoly[this->count]);
		aSecondPoly->Copy(this->SecondPoly[this->count]);
		mpz_set(this->Root[this->count], aRoot);
		mpf_set(this->Alpha[this->count], aAlpha);
		mpf_set(this->Rating[this->count], aRating);
		mpf_set(this->Skewness[this->count], aSkewness);
		this->count++;
	}
	else
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","AddCandidatePair",0,0,"Fields not allocated.");
		return ConstRC::CannotBeDone;
	}

	return ConstRC::Ok;
}
int CCandidatePairs::Copy2(CCandidatePairs * aTarget)
{
	if (aTarget == NULL)
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Copy2",0,0,"Candidate pairs are NULL.");
		return ConstRC::BadArgument;
	}

	if (this->allocated > aTarget->allocated)
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Copy2",0,0,"Not enough allocated space in given candidate pairs.");
		return ConstRC::BadArgument;
	}

	int rc;
	for (int i = 0; i < this->allocated; i++)
	{
		rc = this->FirstPoly[i]->Copy(aTarget->FirstPoly[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		rc = this->SecondPoly[i]->Copy(aTarget->SecondPoly[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		mpz_set(aTarget->Root[i], this->Root[i]);
		mpf_set(aTarget->Alpha[i], this->Alpha[i]);
		mpf_set(aTarget->Rating[i], this->Rating[i]);
		mpf_set(aTarget->Skewness[i], this->Skewness[i]);
	}

	aTarget->count = this->count;

	return ConstRC::Ok;
}

int CCandidatePairs::Allocate()
{
	if (this->FirstPoly == NULL)
	{
		if (this->allocated % 2)
		{
			this->allocated++;
		}
		this->FirstPoly = NULL;
		this->FirstPoly = (Polynomial**) Utils::AllocationMachine((void**)(&FirstPoly),allocated*sizeof(Polynomial*));
		this->SecondPoly = NULL;
		this->SecondPoly = (Polynomial**) Utils::AllocationMachine((void**)(&SecondPoly),allocated*sizeof(Polynomial*));
		this->Root = NULL;
		this->Root = (mpz_t*) Utils::AllocationMachine((void**)(&Root),allocated*sizeof(mpz_t));
		this->Alpha = NULL;
		this->Alpha = (mpf_t*) Utils::AllocationMachine((void**)(&Alpha),allocated*sizeof(mpf_t));
		this->Rating = NULL;
		this->Rating = (mpf_t*) Utils::AllocationMachine((void**)(&Rating),allocated*sizeof(mpf_t));
		this->Skewness = NULL;
		this->Skewness = (mpf_t*) Utils::AllocationMachine((void**)(&Skewness),allocated*sizeof(mpf_t));
		for (long i = 0; i < allocated; i++)
		{
			this->FirstPoly[i] = new Polynomial();
			this->SecondPoly[i] = new Polynomial();
			mpz_init(this->Root[i]);
			mpf_init(this->Alpha[i]);
			mpf_init(this->Rating[i]);
			mpf_init(this->Skewness[i]);
		}
	}

	return ConstRC::Ok;
}
int CCandidatePairs::Deallocate()
{
	if (this->FirstPoly != NULL)
	{
		for (long i = 0; i < allocated; i++)
		{
			delete FirstPoly[i];
			delete SecondPoly[i];
			mpz_clear(Root[i]);
			mpf_clear(Alpha[i]);
			mpf_clear(Rating[i]);
			mpf_clear(Skewness[i]);
		}
		free(this->FirstPoly);
		this->FirstPoly = NULL;
		free(this->SecondPoly);
		this->SecondPoly = NULL;
		free(Root);
		this->Root = NULL;
		free(Alpha);
		this->Alpha = NULL;
		free(Rating);
		this->Rating = NULL;
		free(Skewness);
		this->Skewness = NULL;

		this->allocated = 0;
		this->count = 0;
	}

	return ConstRC::Ok;
}
int CCandidatePairs::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CCandidatePairs::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	std::string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_CANDIDATE_POLY_PAIRS_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Serialize",rc,0,"Could not write the start element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_ASSIGNED_ELEMENTS_NAME, this->allocated);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENTS_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Serialize",rc,0,"Could not write the start element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	long max = this->allocated;
	long maxRated = this->allocated;
	aParam.GetLongParameter(ConstPN::MAX_ELEMENT_COUNT, max);
	aParam.GetLongParameter(ConstPN::MAX_RATED_ELEMENT_COUNT, maxRated);
	std::string auxStr;

	for (long i = 0; i < max; i++)
	{
		rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENT_NAME);
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CCandidatePairs","Serialize",rc,0,"Could not write the start element.");
			return ConstRC::GeneralError;
		}
		rc = CXMLHelper::NewLine(aWriter, FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		auxStr = this->FirstPoly[i]->ToString(ELongFormat);
		rc = CXMLHelper::SerializeString(aWriter, XML_SERIAL_FIRST_POLY_NAME, auxStr);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		auxStr = this->SecondPoly[i]->ToString(ELongFormat);
		rc = CXMLHelper::SerializeString(aWriter, XML_SERIAL_SECOND_POLY_NAME, auxStr);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_ROOT_NAME, this->Root[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (i < maxRated)
		{
			rc = CXMLHelper::SerializeMpf(aWriter, XML_SERIAL_ALPHA_NAME, this->Alpha[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CXMLHelper::SerializeMpf(aWriter, XML_SERIAL_RATING_NAME, this->Rating[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			rc = CXMLHelper::SerializeMpf(aWriter, XML_SERIAL_SKEWNESS_NAME, this->Skewness[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENT_NAME
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CCandidatePairs","Serialize",rc,0,"Could not end the element.");
			return ConstRC::GeneralError;
		}
		rc = CXMLHelper::NewLine(aWriter,FALSE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENTS_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Serialize",rc,0,"Could not end the element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Serialize",rc,0,"Could not end the start element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter,FALSE);

	return rc;
}
int CCandidatePairs::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CCandidatePairs::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	std::string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_CANDIDATE_POLY_PAIRS_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CCandidatePairs","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);
	long auxLong;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ASSIGNED_ELEMENTS_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize assigned elements...", auxLong, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			
			if (this->allocated < auxLong)
			{
				Deallocate();
				this->allocated = auxLong;
				Allocate();
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENTS_NAME))
		{
			std::string auxStr;
			int index = 0;
			int elementsDepth = xmlTextReaderDepth(aReader);
			int elementDepth = elementsDepth + 1;
			while (CXMLHelper::MoveToNextNode(aReader, elementsDepth) == ConstRC::Ok 
				&& index < this->allocated) // element
			{
				while (CXMLHelper::MoveToNextNode(aReader, elementDepth) == ConstRC::Ok) // inner stuffs
				{
					name = xmlTextReaderName(aReader);
	
					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_FIRST_POLY_NAME))
					{
						rc = CXMLHelper::DeserializeString(aReader, "Deserialize first poly...", auxStr, FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						rc = Polynomial::Parse2(this->FirstPoly[index], auxStr);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}

						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_SECOND_POLY_NAME))
					{
						rc = CXMLHelper::DeserializeString(aReader, "Deserialize second poly...", auxStr, FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						rc = Polynomial::Parse2(this->SecondPoly[index], auxStr);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}

						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ROOT_NAME))
					{
						rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize root...", this->Root[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ALPHA_NAME))
					{
						rc = CXMLHelper::DeserializeMpf(aReader, "Deserialize alpha...", this->Alpha[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_RATING_NAME))
					{
						rc = CXMLHelper::DeserializeMpf(aReader, "Deserialize rating...", this->Rating[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

					if (xmlStrEqual(name, BAD_CAST XML_SERIAL_SKEWNESS_NAME))
					{
						rc = CXMLHelper::DeserializeMpf(aReader, "Deserialize skewness...", this->Skewness[index], FALSE);
						if (rc != ConstRC::Ok)
						{
							rc = ConstRC::GeneralError;
							break;
						}
						continue;
					}

				} // inner stuffs

				index++;

			} // element

			continue;

		} // elements
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}

