#include "cc_table.h"

#include "const_return_codes.h"
#include "const_element_flags.h"
#include "const_debug.h"
#include "const_rel_processing_phase.h"
#include "const_hashtables.h"
#include "nfs_utils.h"
#include "log.h"

CCCTable::CCCTable()
{
	this->Table = NULL;
	this->Count = 0;
	this->AllocatedSize = 0;
}
CCCTable::~CCCTable()
{
	DeleteTable();
}
void CCCTable::DeleteTable()
{
	if (this->Table != NULL)
	{
		cycle_construction_entry * aux = this->Table;
		cycle_construction_entry * end = this->Table + this->AllocatedSize;
		do
		{
			if (aux->array != NULL)
			{
				free(aux->array);
				//aux->array = NULL;
				//aux->allocated = 0;
				//aux->current_max_index = -1;
			}
		}
		while (++aux < end);

		free(this->Table);
		this->Table = NULL;
		this->AllocatedSize = 0;
		this->Count = 0;
	}
}
int CCCTable::SetupTable(long aAmount)
{
	if (this->Table != NULL)
	{
		DeleteTable();
		return ConstRC::GeneralError;
	}

	this->Table = (cycle_construction_entry*) Utils::AllocationMachine((void**) &this->Table, aAmount * sizeof(cycle_construction_entry));
	if (this->Table == NULL) 
	{
		CLog::LogError(WHEREARG,"CCCTable","SetupTable",aAmount,aAmount*sizeof(cycle_construction_entry),"Not enough memory to allocate cycle contruction field.");
		return ConstRC::NotEnoughMemory;
	}
	this->AllocatedSize = aAmount;
	this->Count = aAmount;
	cycle_construction_entry * aux = this->Table;
	cycle_construction_entry * end = this->Table + this->AllocatedSize;

	do
	{
		aux->array = NULL;
		aux->array = (int*) Utils::AllocationMachine((void**) &aux->array, DEFAULT_CYCLE_ARRAY_ENTRIES * sizeof(int));
		if (aux->array == NULL) 
		{
			// delete all before
			while (aux-- > this->Table)
			{
				CLog::LogError(WHEREARG,"CCCTable","SetupTable",0,0,"Not enough memory to allocate cycle contruction field.");
				free(aux->array);
			}
			free(this->Table);
			this->Table = NULL;
			return ConstRC::NotEnoughMemory;
		}
		aux->allocated = DEFAULT_CYCLE_ARRAY_ENTRIES;
		aux->current_max_index = -1;
	}
	while (++aux < end);

	return ConstRC::Ok;
}
void CCCTable::ClearTable()
{
	if (this->Table != NULL)
	{
		cycle_construction_entry * aux = this->Table;
		cycle_construction_entry * end = this->Table + this->AllocatedSize;
		do
		{
			if (aux->array != NULL)
			{
				aux->current_max_index = -1;
				aux->running_index = -1;
			}
		}
		while (++aux < end);

		this->Count = 0;
	}
}
/*!
	This auxiliary method will add an index of a relation to a "blueprint" 
	for construction of a full cycle.	

*/
int CCCTable::PutToConstructionEntry(cycle_construction_entry * aEntry, int aRelationIndex) 
{
	aEntry->current_max_index++;
	if (aEntry->current_max_index == aEntry->allocated) 
	{
		// We need to reallocate the array for entries
		aEntry->array = (int*) Utils::AllocationMachine((void**)&aEntry->array, (aEntry->allocated + DEFAULT_CYCLE_ARRAY_ENTRIES) * sizeof(int));
		if (aEntry->array == NULL) 
		{
			CLog::LogError(WHEREARG,"CCCTable","PutToConstructionEntry", aEntry->allocated + DEFAULT_CYCLE_ARRAY_ENTRIES, (aEntry->allocated + DEFAULT_CYCLE_ARRAY_ENTRIES) * sizeof(int),  "There was an error reallocating array.");				
			aEntry->allocated = 0;
			aEntry->current_max_index = -1;
			return ConstRC::NotEnoughMemory;
		} 
		else 
		{
			aEntry->allocated += DEFAULT_CYCLE_ARRAY_ENTRIES;
		}
	}
	// Now, the allocated space suffices
	aEntry->array[aEntry->current_max_index] = aRelationIndex;
	
	return ConstRC::Ok;
}
/*
	This method print number of cycles with given size [0..maxsize].
*/
int CCCTable::PrintCycleSizes() 
{
	Utils::NormalMessage("\t\tCycle sizes:\n");
	if (this->Table == NULL)
	{
		CLog::LogError(WHEREARG,"CCCTable","PrintCycleSizes", 0, 0, "Table is NULL.");
		return ConstRC::BadArgument;
	}

	int max = 0;
	cycle_construction_entry * aux = this->Table;
	cycle_construction_entry * end = this->Table + this->AllocatedSize;

	do
	{
		if (aux->current_max_index > max)
		{
			max = aux->current_max_index;
		}
	}
	while (++aux < end);

	// [-1..maxindex] => maxindex + 2 entries
	max = max + 2;

	int* sizes = NULL;
	sizes = (int*) Utils::AllocationMachine((void**) &sizes, max * sizeof(int));
	if (sizes == NULL)
	{
		CLog::LogError(WHEREARG,"CCCTable","PrintCycleSizes", max, max * sizeof(cycle_construction_entry), "Not enough memory to allocate sizes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(sizes, 0, max * sizeof(int));

	aux = this->Table;
	do
	{
		sizes[aux->current_max_index + 1]++;
	}
	while (++aux < end);

	for (int i = 0; i < max; i++) 
	{
		if (sizes[i] != 0)
		{
			printf("\t\t\t%4d: %7d.\n", i, sizes[i]);
		}
	}

	free(sizes);
	sizes = NULL;

	return ConstRC::Ok;
}
/*!
 * During construction of cycles, it might happen that the two paths constructing
 * the cycle (one going from the root to one of the primes, and the other from
 * the second prime back to the root) have some part in common. Edges which are
 * in common are duplicite and add nothing new to the cycle (their product is 0
 * mod 2 with regard to all exponents); on the other hand, they complicate
 * one-pass cycle construction. Therefore we remove them entirely here. 
 * This method requires the array of edge indices in aEntry to be sorted.
 * From the logic of the construction algorithm, an edge may be at most twice
 * in the array; therefore, we do not care about triple or quadruple occurrences.
 */
int CCCTable::RemoveDuplicitePaths() 
{
	if (this->Table == NULL)
	{
		CLog::LogError(WHEREARG,"CCCTable","RemoveDuplicitePaths",0,0,"Table is NULL.");
		return ConstRC::BadArgument;
	}

	int i, max, j;
	cycle_construction_entry * aux = this->Table;
	cycle_construction_entry * end = this->Table + this->AllocatedSize;
	do
	{
		// Sort cycle definitions
		qsort(aux->array, aux->current_max_index + 1, sizeof(int), &Utils::CompareInts);
		//PrintCycleArray(aux - this->Table, aux->array, aux->current_max_index, true);
		i = 0; 
		max = aux->current_max_index;
		while (true) 
		{
			if ((i + 1) > max) break;
			if (aux->array[i] == aux->array[i + 1]) 
			{
				// We skip both of these entries
				j = i + 2;
				while (j <= max) 
				{
					aux->array[j - 2] = aux->array[j];
					j++;
				}
				max = (aux->current_max_index -= 2);
			} 
			else 
			{
				i++;
			}
		}
		//PrintCycleArray(aux - this->Table, aux->array, aux->current_max_index, false);
		aux->running_index = 0;
	}
	while (++aux < end);

	return ConstRC::Ok;
}
