#include "clique_component_array.h"

#include "nfs_utils.h"
#include "const_return_codes.h"
#include "const_hashtables.h"
#include "log.h"

CCliqueComponentArray::CCliqueComponentArray()
{
	this->AllocatedSize = 0;
	this->Count = 0;
	this->Components = NULL;
}

CCliqueComponentArray::~CCliqueComponentArray()
{
	Reset();
}

void CCliqueComponentArray::Reset()
{
	if (this->Components != NULL)
	{
		free(this->Components);
		this->Components = NULL;
	}

	this->AllocatedSize = 0;
	this->Count = 0;
}

int CCliqueComponentArray::Setup(int aAllocatedSize)
{
	if (this->Components != NULL)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "Setup", 0, 0,  "Components already allocated.");
		return ConstRC::GeneralError;
	}
	this->Components = (clique_component_t *)Utils::AllocationMachine((void**) &this->Components, aAllocatedSize * sizeof(clique_component_t));
	if (this->Components == NULL)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "Setup", aAllocatedSize, aAllocatedSize * sizeof(clique_component_t),  "Not enough memory to allocate data space.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Components, 0, aAllocatedSize * sizeof(clique_component_t));
	this->AllocatedSize = aAllocatedSize;
	this->Count = 0;

	return ConstRC::Ok;
}
int CCliqueComponentArray::GetMainComponent(relation_index_type aComponentIndex, relation_index_type& aMainComponentIndex) const
{
	if (0 > aComponentIndex || aComponentIndex >= this->Count)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "GetMainComponent", 0, 0,  "Index of component out of range.");
		return ConstRC::BadArgument;
	}

	if (this->Components[aComponentIndex].ID == aComponentIndex)
	{
		aMainComponentIndex = aComponentIndex;
		return ConstRC::Ok;
	}
	else
	{
		// recursive call
		return GetMainComponent(this->Components[aComponentIndex].ID, aMainComponentIndex);
	}
}

int CCliqueComponentArray::ConnectComponents(relation_index_type aMainComponentIndex, int aRelationWeight, fhashtable_entry **aPrimeIdeals, int aPrimeIdealsCount)
{
	if (aPrimeIdeals == NULL || aPrimeIdealsCount <= 0)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "ConnectComponents", 0, 0,  "Prime ideals are NULL.");
		return ConstRC::BadArgument;
	}

	if (this->Components == NULL || this->Count <= aMainComponentIndex)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "ConnectComponents", 0, 0,  "Not enough components.");
		return ConstRC::GeneralError;
	}

	relation_index_type * index = NULL;
	clique_component_t * mainComponent = this->Components + aMainComponentIndex;
	// add weight of currently used relation
	mainComponent->weight += aRelationWeight;
	// increase number of relations
	mainComponent->count++;
	clique_component_t * currMainComponent = NULL;
	int rc;
	relation_index_type currIdealMainComponent = 0;
	for (int i = 0; i < aPrimeIdealsCount; i++)
	{
		index = &(aPrimeIdeals[i]->index);
		if (*index != ConstHT::UNDEFINED_RELATION)
		{
			// prime ideal is in some component
			// we need to know the main component in which this prime ideal is
			rc = GetMainComponent(*index, currIdealMainComponent);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}

			// check if this component is different from the target main component
			if (currIdealMainComponent != aMainComponentIndex)
			{
				currMainComponent = this->Components + currIdealMainComponent;
				// set prime ideal into new main component
				*index = aMainComponentIndex;
				// connect main components - the new main component is aMainComponentIndex
				currMainComponent->ID = aMainComponentIndex;
				// just to make things quicker - direct connection
				this->Components[*index].ID = aMainComponentIndex;
				// add weight of main components to the new one
				mainComponent->weight += currMainComponent->weight;
				// add count of prime ideals in the main components to the new one
				mainComponent->count += currMainComponent->count;
				// reset weight of the component
				currMainComponent->weight = 0;
				// reset count of prime ideals in the component
				currMainComponent->count = 0;
			}
			else if (*index != aMainComponentIndex)
			{
				// just set right index of component
				*index = aMainComponentIndex;
			}			
		}
		else
		{
			// this prime ideal is not in any component yet, so we add it to the main component
			*index = aMainComponentIndex;
		}
	}

	return ConstRC::Ok;
}
int CCliqueComponentArray::AddNewComponent(int aRelationWeight, fhashtable_entry **aPrimeIdeals, int aPrimeIdealsCount)
{
	if (aPrimeIdeals == NULL || aPrimeIdealsCount <= 0)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "AddNewComponent", 0, 0,  "Prime ideals are NULL.");
		return ConstRC::BadArgument;
	}

	if (this->Components == NULL || this->Count >= this->AllocatedSize)
	{
		CLog::LogError(WHEREARG, "CCliqueComponentArray", "AddNewComponent", 0, 0,  "Not enough components.");
		return ConstRC::GeneralError;
	}

	clique_component_t * component = this->Components + this->Count;

	// setup
	component->ID = this->Count;
	component->weight = aRelationWeight;
	component->count = 1; // one relation

	// set index to this new component
	for (int i = 0; i < aPrimeIdealsCount; i++)
	{
		aPrimeIdeals[i]->index = this->Count;
	}

	// increase number of components
	this->Count++;

	return ConstRC::Ok;
}
int CCliqueComponentArray::FindLargestComponent(relation_index_type aMaxCountOfRelation, relation_index_type &aComponentSize, relation_index_type &aComponentIndex) const
{
	if (this->Components == NULL)
	{
		Utils::ErrorMessage(WHEREARG, "CCliqueComponentArray", "FindLargestComponent", 0, 0,  "Components are NULL.");
		return ConstRC::GeneralError;
	}

	aComponentIndex = -1;
	INT64 maxWeight = 0;
	clique_component_t * curr = this->Components;
	clique_component_t * end = this->Components + this->Count;
	while (curr < end)
	{
		if (curr->weight > maxWeight && curr->count <= aMaxCountOfRelation)
		{
			aComponentIndex = curr - this->Components;
			aComponentSize = curr->count;
			maxWeight = curr->weight;
		}
		curr++;
	}

	if (aComponentIndex == -1)
	{
		return ConstRC::NotFound;
	}

	return ConstRC::Ok;	
}

int CCliqueComponentArray::ResetWeight(relation_index_type aComponentIndex)
{
	if (0 > aComponentIndex || aComponentIndex >= this->Count)
	{
		Utils::ErrorMessage(WHEREARG, "CCliqueComponentArray", "ResetWeight", 0, 0,  "Index of component out of range.");
		return ConstRC::BadArgument;
	}

	this->Components[aComponentIndex].weight = 0;

	return ConstRC::Ok;
}
