#include "complex_structures.h"

#include "integer_class.h"
#include "abstract_relation.h"
#include "const_return_codes.h"
#include "const_sieving_phase.h"
#include "log.h"

CNumberFieldInfo::CNumberFieldInfo()
{
	this->SievingPoly = NULL;
	this->Domain = NULL;
	this->IdealBase = NULL;
	
	this->ExpectedErrorFactor = 0;
	this->SievingPolyMonic = true;
}
CNumberFieldInfo::~CNumberFieldInfo()
{
	Reset();
}

int CNumberFieldInfo::Reset()
{
	if (this->SievingPoly != NULL)
	{
		delete this->SievingPoly;
		this->SievingPoly = NULL;
		this->SievingPolyMonic = true;
	}

	if (this->Domain != NULL)
	{
		delete this->Domain;
		this->Domain = NULL;
	}

	if (this->IdealBase != NULL)
	{
		delete this->IdealBase;
		this->IdealBase = NULL;
	}

	this->RZeros_Indices.clear();
	int max = this->SpecialIdeals.size();
	for (int i = 0; i < max; i++)
	{
		if (this->SpecialIdeals[i] != NULL)
		{
			delete this->SpecialIdeals[i];
			this->SpecialIdeals[i] = NULL;
		}
	}
	this->SpecialIdeals.clear();
	this->SpecialPrimes.clear();

	return ConstRC::Ok;
}
int CNumberFieldInfo::SetSievingPoly(Polynomial * aPoly)
{
	this->SievingPoly = aPoly;
	this->SievingPolyMonic = mpz_cmp_ui(*SievingPoly->Leading2(), 1) == 0;
	if (this->SievingPoly->GetDegree() > 1)
	{
		this->ExpectedErrorFactor = ConstSPhase::ALGEBRAIC_ERROR_FACTOR;
	}
	else
	{
		this->ExpectedErrorFactor = ConstSPhase::INTEGRAL_ERROR_FACTOR;
	}

	return ConstRC::Ok;
}
int CNumberFieldInfo::InitIntegralDomain(long aUpperBound)
{
	this->Domain = new IntegralDomain();

	if (this->SievingPoly->GetDegree() > 1)
	{
		this->Domain->AssignPolynomial(this->SievingPoly);
		this->Domain->FindIntegralBase(aUpperBound);

		this->IdealBase = new HermiteMatrix(this->SievingPoly->GetDegree(), this->SievingPoly->GetDegree());		
		this->IdealBase->Allocate();
		this->IdealBase->Zeroize();
	}
	else
	{
		// this is just for norm computation
		this->Domain->Rank = 1;

		// special primes - divisors of leading coefficient
		if (mpz_cmp_ui(*(this->SievingPoly)->Leading2(), 1) != 0) {
			long *primes;
			unsigned long primes_count;

			CIntegerClass::GetPrimes(0, aUpperBound, primes_count, primes);

			mpz_t aux_mpz;
			mpz_init(aux_mpz);
			for (unsigned long i = 0; i < primes_count; i++) {
				mpz_mod_ui(aux_mpz,  *(this->SievingPoly)->Leading2(), primes[i]);
				if (mpz_cmp_si(aux_mpz, 0) == 0) this->Domain->AddToListOfSpecialPrimes(primes[i]);
			}
			mpz_clear(aux_mpz);

			free(primes);
		}		
	}

	return ConstRC::Ok;
}
int CNumberFieldInfo::FillSpecialPrimes()
{
	if (this->SievingPoly->GetDegree() > 1)
	{
		for (unsigned i = 0; i < this->Domain->SpecialPrimesCount; i++)
		{
			this->SpecialPrimes.insert(std::make_pair(this->Domain->SpecialPrimes[i], 0));
		}
	}

	return ConstRC::Ok;
}
void CNumberFieldInfo::PrintInfo(int aIndex, fb_element * aFB)
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	int max;
	int i;
	if (this->SievingPoly->GetDegree() > 1)
	{
		max = this->Domain->SpecialPrimesCount;
		if (max == 0)
		{
			CLog::Log(50, "0560000", "\n    No special primes detected for domain #%d.", aIndex + 1);
		} 
		else 
		{
			CLog::Log(50, "0560001", "\n    Special primes detected for domain #%d - %d total: ", aIndex + 1, max);
			for (i = 0; i < max; i++) 
			{
				CLog::Log(50, "", "%3ld ", this->Domain->SpecialPrimes[i]);			
			}
		}
	}

	max = this->RZeros_Indices.size();
	if (max > 0)
	{
		CLog::Log(50, "0560002", "\n    Zero root primes for the field #%d: ", aIndex + 1);
		for (i = 0; i < max; i++)
		{
			CLog::Log(50, "", "%d", aFB[this->RZeros_Indices[i]].prime);
	 		if (i + 1 < max)
			{
				CLog::Log(50, "", ", ");
			} 
			else
			{
				CLog::Log(50, "", ".\n");
			}
		}
	}
#endif
}

CVariationsInfo::CVariationsInfo()
{
	this->FactorAlg = NULL;
	this->Type = EDoNotUseVariations;
	mpz_init(this->MaxPartialRemainder);
	mpz_init(this->MinPartialRemainder);
	this->MaxPartialFactor = 0;	
	this->SizeOfMaxRemainder = 0;
	this->LargePrimeBound = 0;
}
CVariationsInfo::~CVariationsInfo()
{
	if (this->FactorAlg != NULL)
	{
		delete this->FactorAlg;
		this->FactorAlg = NULL;
	}
	mpz_clear(this->MaxPartialRemainder);
	mpz_clear(this->MinPartialRemainder);
	this->MaxPartialFactor = 0;	
	this->SizeOfMaxRemainder = 0;
	this->LargePrimeBound = 0;
}
int CVariationsInfo::Reset()
{
	if (this->FactorAlg != NULL)
	{
		delete this->FactorAlg;
		this->FactorAlg = NULL;
	}
	this->Type = EDoNotUseVariations;
	mpz_set_si(this->MaxPartialRemainder, 0);
	mpz_set_si(this->MinPartialRemainder, 0);
	this->MaxPartialFactor = 0;
	this->SizeOfMaxRemainder = 0;
	this->LargePrimeBound = 0;

	return ConstRC::Ok;
}
int CVariationsInfo::Prepare(long aFBBound, int aPartialFactorMultiplier, int aLogBase)
{
	this->LargePrimeBound = aFBBound;
	if (this->Type != EDoNotUseVariations) 
	{
		mpz_set_ui(this->MaxPartialRemainder, aFBBound);
		mpz_mul_ui(this->MaxPartialRemainder, this->MaxPartialRemainder, aPartialFactorMultiplier);

		if (mpz_fits_slong_p(this->MaxPartialRemainder))
		{
			this->MaxPartialFactor = mpz_get_si(this->MaxPartialRemainder);
		}
		else
		{
			CLog::LogError(WHEREARG,"CVariationsInfo","Prepare",0,0,"Partial factor is too big - it will be reduce.");
			this->MaxPartialFactor = MAXIMAL_LONG_VALUE;
		}

		// B_MAX = B_LP * B_FB^{type - 1}
		// than numbers in [B_MIN, B_MAX] should be composite of desirable size (for Type = 2)
		mpz_ui_pow_ui(this->MinPartialRemainder, aFBBound, ((int)this->Type) - 1);
		mpz_mul(this->MaxPartialRemainder, this->MaxPartialRemainder, this->MinPartialRemainder);

		mpz_ui_pow_ui(this->MinPartialRemainder, aFBBound, 2);

		this->SizeOfMaxRemainder = mpz_sizeinbase(this->MaxPartialRemainder, aLogBase);
	}
	else
	{
		// bigger than 1 
		// 1 is requested remaining
		mpz_set_ui(this->MaxPartialRemainder, 2);
		mpz_set_ui(this->MinPartialRemainder, 0);
		this->MaxPartialFactor = 2;
		this->SizeOfMaxRemainder = 0;
	}
	gmp_printf("\n LargePrimeBound: %d", aFBBound);
	gmp_printf("\n MinPartialRemainder: %Zd", this->MinPartialRemainder);
	gmp_printf("\n MaxPartialRemainder: %Zd", this->MaxPartialRemainder);
	gmp_printf("\n MaxPartialFactor: %d", this->MaxPartialFactor);
	gmp_printf("\n SizeOfMaxRemainder: %d", this->SizeOfMaxRemainder);

	return ConstRC::Ok;
}
bool CVariationsInfo::Control(mpz_t aRemainder)
{
	if (mpz_cmp(aRemainder, this->MaxPartialRemainder) >= 0)
	{
		// too big
		return false;
	}

	if (this->Type > ELargePrimeV &&
		mpz_cmp_si(aRemainder, this->MaxPartialFactor) > 0)
	{
		// we are using double or tripple large prime variation and 
		// the remainder is bigger than max factor so we have to factorize it
		// now we check if the remainder is not just big prime (it is very often)
		// but first we check if is bigger than square of FB bound because we know
		// that all numbers smaller than this square are primes
		if (mpz_cmp(aRemainder, this->MinPartialRemainder) > 0 &&
			mpz_probab_prime_p(aRemainder, ARelation::PROBAB_PRIME_DEGREE) == 0)
		{
			return true;
		}
		else
		{
			// remainder is big prime
			return false;
		}
	}
	else
	{
		// we are not using large primes so it is enough to know that remainder is smaller than max remainder
		return true;
	}
}

CAdvCFBInfo::CAdvCFBInfo()
{
	this->FactorBase = NULL;
	this->FirstMiddlePrime = NULL;
	this->LastMiddlePrime = NULL;
	this->FirstLargishPrime = NULL;
	
	this->Count = 0;
}
CAdvCFBInfo::~CAdvCFBInfo()
{
	Reset();
}

int CAdvCFBInfo::Reset()
{
	if (this->FactorBase != NULL)
	{
		free(this->FactorBase);
		this->FactorBase = NULL;			
	}
	this->FirstMiddlePrime = NULL;
	this->LastMiddlePrime = NULL;
	this->FirstLargishPrime = NULL;
	
	this->Count = 0;

	return ConstRC::Ok;
}
int CAdvCFBInfo::Init(int aBound)
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CAdvCFBInfo","Init",0,0,"Factor base is NULL!");
		return ConstRC::GeneralError;
	}

	int i = 0;
	main_sieving_type p;
	while ((p = this->FactorBase[i].p) != 0)
	{
		if ((!this->FirstMiddlePrime) && (p > ConstSPhase::MIN_PRIME_TO_SIEVE_WITH))
		{
			this->FirstMiddlePrime = this->FactorBase + i;
		}
		// Increments continually to the largest available 'middle-sized prime'
		if (p <= aBound)
		{
			this->LastMiddlePrime = this->FactorBase + i;
		}
		if ((!this->FirstLargishPrime) && (p > aBound))
		{	
			this->FirstLargishPrime = this->FactorBase + i;
			break;
		}
		i++;
	}	
	if (!this->FirstMiddlePrime)
	{
		CLog::LogError(WHEREARG,"CAdvCFBInfo","Init",0,0,"Advance middle prime pointer is NULL - inconsistent state!");
		return ConstRC::InconsistentState;
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (!this->FirstLargishPrime)
	{
		CLog::Log(100, "0560003", "Advance largish prime pointer is NULL - only middle-sized primes possible!\n");
	}
#endif

	return ConstRC::Ok;
}
CAdvLFBInfo::CAdvLFBInfo()
{
	this->FactorBase = NULL;
	this->MiddleFB = NULL;
	this->LastMiddleFB = NULL;
	this->LargishFB = NULL;

	this->MiddlePrimes = NULL;
	this->LastMiddlePrimes = NULL;
	this->LargishPrimes = NULL;
	this->LastLargishPrimes = NULL;
	
	this->Count = 0;
}
CAdvLFBInfo::~CAdvLFBInfo()
{
	Reset();
}

int CAdvLFBInfo::Reset()
{
	if (this->FactorBase != NULL)
	{
		free(this->FactorBase);
		this->FactorBase = NULL;			
	}
	this->MiddleFB = NULL;
	this->LastMiddleFB = NULL;
	this->LargishFB = NULL;

	if (this->MiddlePrimes != NULL)
	{
		free(this->MiddlePrimes);
		this->MiddlePrimes = NULL;
	}
	this->LastMiddlePrimes = NULL;

	if (this->LargishPrimes != NULL)
	{
		free(this->LargishPrimes);
		this->LargishPrimes = NULL;
	}
	this->LastLargishPrimes = NULL;

	this->BadPrimes.clear();

	this->Count = 0;

	return ConstRC::Ok;
}
int CAdvLFBInfo::Init(int aBound)
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CAdvLFBInfo","Init",0,0,"Factor base is NULL!");
		return ConstRC::GeneralError;
	}

	int i = 0;
	lattice_type2* element = this->FactorBase;
	while (element->p != 0 && element->p <= aBound)
	{
		if (element->p <= ConstSPhase::MIN_PRIME_TO_SIEVE_WITH)
		{
			i++;
		}
		element++;
	}

	int sizeofupdate = sizeof(lattice_update_t);
	int countOfMiddlePrimes = (element - this->FactorBase) - i;
	int countOfLargishPrimes = this->Count - (element - this->FactorBase);

	if (countOfMiddlePrimes > 0)
	{
		this->MiddleFB = this->FactorBase + i;
		this->LastMiddleFB = element - 1;
		this->MiddlePrimes = (lattice_update_t*)Utils::AllocationMachine((void**)&this->MiddlePrimes, countOfMiddlePrimes * sizeofupdate);
		if (this->MiddlePrimes == NULL)
		{
			// info
			return ConstRC::NotEnoughMemory;
		}
		memset(this->MiddlePrimes, 0, countOfMiddlePrimes * sizeofupdate);
		// point to last element
		this->LastMiddlePrimes = this->MiddlePrimes + countOfMiddlePrimes - 1;
	}
	
	if (countOfLargishPrimes > 0)
	{
		this->LargishFB = element;
		this->LargishPrimes = (lattice_update_t*)Utils::AllocationMachine((void**)&this->LargishPrimes, countOfLargishPrimes * sizeofupdate);
		if (this->LargishPrimes == NULL)
		{
			// info
			return ConstRC::NotEnoughMemory;
		}
		memset(this->LargishPrimes, 0, countOfLargishPrimes * sizeofupdate);
		// point to last element
		this->LastLargishPrimes = this->LargishPrimes + countOfLargishPrimes - 1;
	}

	if (!this->MiddlePrimes)
	{
		CLog::LogError(WHEREARG,"CAdvLFBInfo","Init",0,0,"Advance middle prime pointer is NULL - inconsistent state!");
		return ConstRC::InconsistentState;
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (!this->LargishPrimes)
	{
		CLog::Log(100, "0560004", "Advance largish prime pointer is NULL - only middle-sized primes possible!\n");
	}
#endif

	return ConstRC::Ok;
}
