#ifndef __COMPLEX_STRUCTURES_H__
#define __COMPLEX_STRUCTURES_H__

#include "integral_domain.h"
#include "structures.h"
#include "abstract_supporting_factor_alg.h"
#include <vector>
#include <map>
#include "enums.h"
#include "abstract_sieve.h"

/*!
 * \brief Information on number field
 */
class CNumberFieldInfo
{
public:

	/*!
		The sieving polynomial.

		This polynomial can be monic or non-monic.
		In case that it is non-monic, the corresponding
		monic polynomial is present in the iDomain as 
		a member variable.
	
		By the corresponding monic polynomial we mean		
		
		f(x/a)*a^{d-1}

		where a is the leading coefficient and d the degree of f.
	*/
	Polynomial*			SievingPoly;
	/// Just quick info if the poly is monic
	bool				SievingPolyMonic;
	/* 
		When we compute expected log value, we have to compute with some error.
		It is different for integral and algebraic part. 
	
		For integral part it can be computed quite exactly.
		For algebraic part it is just estimation.
	*/
	int ExpectedErrorFactor;
	/*
		Integral domain of this number field - O_{K}.
	*/
	IntegralDomain*		Domain;
	/*
		This is auxiliary base of ideal (a + b\theta) used for factorization.
		Used only for prime ideals over special primes.
	*/
	HermiteMatrix* 		IdealBase;
	/*
		This could be sometimes helpful.
		It is field of indexies of prime ideals with 'zero root' - f(x) mod p.
	*/
	std::vector<long>	RZeros_Indices;
	/*
		Field of prime ideals over special primes in O_{K}.
	*/
	std::vector<PrimeIdeal*>			SpecialIdeals;
	/*
		Auxiliary field of special prime used for factorization. 
		Integer element is for p-valuation of the norm of ideal (a + b\theta).

	*/
	std::map<main_sieving_type, int>	SpecialPrimes;

public:
	CNumberFieldInfo();
	~CNumberFieldInfo();

	int Reset();
	int SetSievingPoly(Polynomial * aPoly);
	int InitIntegralDomain(long aUpperBound);
	int FillSpecialPrimes();

	void PrintInfo(int aIndex, fb_element * aFB);

private:
	CNumberFieldInfo(const CNumberFieldInfo & aOperand) {}

};

/*!
 * \brief Advaced factor base information
 */
class CAdvCFBInfo
{
public:
	///A pointer to the advanced factor base. 
	nfs_fb_type*	FactorBase;
	nfs_fb_type*	FirstMiddlePrime;
	/*!
		Must be different from FirstLargishPrime-1, 
		since there are instances where no largish primes exist.
	*/
	nfs_fb_type*	LastMiddlePrime;
	nfs_fb_type*	FirstLargishPrime;
	///Number of entries
	long	Count;

public:
	CAdvCFBInfo();
	~CAdvCFBInfo();

	int Reset();
	int Init(int aBound);

private:
	CAdvCFBInfo(const CAdvCFBInfo& aOperand) {}

};

/*!
 * \brief Advanced factor base for lattice sieving
 */
class CAdvLFBInfo
{
public:
	///A pointer to the advanced factor base. 
	lattice_type2*	FactorBase;
	///Start of the middle primes in FB
	lattice_type2*	MiddleFB;
	///End of the middle primes in FB
	lattice_type2*	LastMiddleFB;
	///Start of the largish primes in FB
	lattice_type2*	LargishFB;

	///Middle primes for sieving
	lattice_update_t* MiddlePrimes;
	///Last middle prime for sieving
	lattice_update_t* LastMiddlePrimes;
	///Largish primes for sieving - for filling hashtables
	lattice_update_t* LargishPrimes;
	///Last largish prime for sieving
	lattice_update_t* LastLargishPrimes;

	///Number of entries
	long	Count;
	///Big primes without sublattice base
	std::vector<lattice_type2> BadPrimes;

public:
	CAdvLFBInfo();
	~CAdvLFBInfo();

	int Reset();
	int Init(int aBound);

private:
	CAdvLFBInfo(const CAdvLFBInfo& aOperand) {}

};


/*!
 * \brief All necessary info for using large primes.
 *
	This class is intended for one number field.
*/
class CVariationsInfo
{
public:
	///Alg. used for factorization of the remainder if necessary
	ASupportingFactorAlg*	FactorAlg;
	///Variation type = EDoNotUseVariations, ELargePrimeV, EDoubleLargePrimeV, ETrippleLargePrimeV
	variations_types	Type;
	mpz_t				MaxPartialRemainder;
	///We know that number smaller than square of the factor base bound is definitely prime.
	mpz_t				MinPartialRemainder;
	///Max value of large prime - we use only "long" for saving primes 
	long				MaxPartialFactor;
	///Factore base size - min value of large prime
	long				LargePrimeBound;
	///Log (used for sieving) of MaxPartialRemainder
	int					SizeOfMaxRemainder;

public:
	CVariationsInfo();
	~CVariationsInfo();
	///Reset all bounds and dispose factoring algorithm
	int Reset();
	///Set all bounds and Leading Coeff - factoring alg. must be setup separately
	int Prepare(long aFBBound, int aPartialFactorMultiplier, int aLogBase);
	/*!
		Control if the remainder is good: 
		- return TRUE if the remainder is 1, < max factor or can be factorized
		- according to variation type
	*/
	bool Control(mpz_t aRemainder);

private:
	CVariationsInfo(const CVariationsInfo& aOperand) {}
};


#endif // __COMPLEX_STRUCTURES_H__
