#include "divisors.h"

#include <cstring>
#include "base_parameters.h"
#include "const_return_codes.h"
#include "const_filenames.h"
#include "const_element_flags.h"
#include "log.h"

#include "nfs_utils.h"
#include "xml_helper.h"

using namespace std;

CDivisors::CDivisors(const int aAssignedElements) : MSerializable(ConstFN::DIVISORS),
	AssignedElements(aAssignedElements)
{
	this->Elements = NULL;
	this->MaxElementIndex = -1;

	if (this->AssignedElements > 0)
	{        
		this->Elements = (nfs_sieving_element*)NFSUtils::AllocationMachine((void **)&(this->Elements), this->AssignedElements * sizeof(nfs_sieving_element));    
        if (this->Elements == NULL) 
		{
			CLog::LogError(WHEREARG,"CDivisors","CDivisors",this->AssignedElements*sizeof(nfs_sieving_element),40,"There was an error in CDivisors, allocating Sieving Elements.");		
			this->AssignedElements = 0;
			return;
        }
		memset(this->Elements, 0, this->AssignedElements * sizeof(nfs_sieving_element));
	}
}

CDivisors::CDivisors(const CDivisors& aOperand) : MSerializable(aOperand)
{
	this->Elements = NULL;

	this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**)&this->Elements, sizeof(nfs_sieving_element)*(aOperand.MaxElementIndex + 3));
	if (this->Elements == NULL) 
	{
		CLog::LogError(WHEREARG,"CDivisors","CDivisors",aOperand.MaxElementIndex+3,40,"There was an error reallocating array elements.");				
		this->AssignedElements = 0;			
		this->MaxElementIndex = -1;
		return;
	} 
	this->AssignedElements = aOperand.MaxElementIndex + 3;
	this->MaxElementIndex = -1;

	memset(this->Elements, 0, sizeof(nfs_sieving_element) * this->AssignedElements);
	if (aOperand.MaxElementIndex > -1)
	{
		memcpy(this->Elements, aOperand.Elements, sizeof(nfs_sieving_element) * (aOperand.MaxElementIndex + 1));
		this->MaxElementIndex = aOperand.MaxElementIndex;
	}
}

CDivisors::~CDivisors()
{
	if (this->Elements != NULL) 
	{
		free(this->Elements);
		this->Elements = NULL;
	    this->MaxElementIndex = -1;            
		this->AssignedElements = 0;
	}
}


int CDivisors::AddDivisor(main_sieving_type aDivisor, unsigned int aExponent, main_sieving_type aRoot, int aFlags, int& aLastIndex, bool aAssertionMode)
{
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (aAssertionMode)
	{
		if ((aLastIndex >= 0) && (!(aFlags & ConstEF::SPECIAL_PRIME)) && (!(aFlags & ConstEF::DIVIDES_L)))
		{
			main_sieving_type primeBefore = this->Elements[aLastIndex].prime;
			if (primeBefore == aDivisor)
			{
				printf("Add Algebraic Divisor: assertion failed. Two distinct nonspecial prime ideals are dividing a single relation. p = %d.\n",aDivisor);
			}
		}
	}
	if ((aExponent > RELATION_PART_2_MAX_EXP) || ((aDivisor > 2) && (aExponent > RELATION_PART_MAX_EXP))) 
	{
		printf("Suspicious: prime = %d, exponent = %d.\n", aDivisor, aExponent);
	}
#endif
	if ((++aLastIndex) >= (signed) this->AssignedElements) 
	{
		// reallocation
		this->AssignedElements += DEFAULT_ELEMENTS_ADD;
		this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**) &this->Elements, this->AssignedElements * sizeof(nfs_sieving_element));
		if (this->Elements == NULL) 
		{
			return ConstRC::NotEnoughMemory;
		}
	}
	this->Elements[aLastIndex].prime = aDivisor;
	this->Elements[aLastIndex].exponent = aExponent;
	this->Elements[aLastIndex].root = aRoot;
	this->Elements[aLastIndex].flags = aFlags;
	this->MaxElementIndex = aLastIndex;
	
	return ConstRC::Ok;
}
int CDivisors::AddDivisor(nfs_sieving_element * aElement, bool aAssertionMode)
{
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (aAssertionMode)
	{
		if ((this->MaxElementIndex >= 0) && (!(aFlags & ConstEF::SPECIAL_PRIME)) && (!(aFlags & ConstEF::DIVIDES_L)))
		{
			main_sieving_type primeBefore = this->Elements[MaxElementIndex].prime;
			if (primeBefore == aElement->prime)
			{
				printf("Add Algebraic Divisor: assertion failed. Two distinct nonspecial prime ideals are dividing a single relation. p = %d.\n",aElement->prime);
			}
		}
	}
	if ((aElement->exponent > RELATION_PART_2_MAX_EXP) || ((aElement->prime > 2) && (aElement->exponent > RELATION_PART_MAX_EXP))) 
	{
		printf("Suspicious: prime = %d, exponent = %d.\n", aElement->prime, aElement->exponent);
	}
#endif
	if ((++this->MaxElementIndex) >= (signed) this->AssignedElements) 
	{
		// reallocation
		this->AssignedElements += DEFAULT_ELEMENTS_ADD;
		this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**) &this->Elements, this->AssignedElements * sizeof(nfs_sieving_element));
		if (this->Elements == NULL) 
		{
			return ConstRC::NotEnoughMemory;
		}
	}
	this->Elements[this->MaxElementIndex].prime = aElement->prime;
	this->Elements[this->MaxElementIndex].exponent = aElement->exponent;
	this->Elements[this->MaxElementIndex].root = aElement->root;
	this->Elements[this->MaxElementIndex].flags = aElement->flags;
	
	return ConstRC::Ok;
}
int CDivisors::Copy2(CDivisors& aOperand)
{
	if (aOperand.Set_AssignedElements(this->AssignedElements) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}

	memcpy(aOperand.Elements, this->Elements, sizeof(nfs_sieving_element) * (this->MaxElementIndex + 1));
	aOperand.MaxElementIndex = this->MaxElementIndex;
	aOperand.Set_DefaultFileName(this->Get_DefaultFileName());

	return ConstRC::Ok;
}
int CDivisors::ControlExponents()
{
	nfs_sieving_element * element = this->Elements;
	for (long i = 0; i <= this->MaxElementIndex; i++, element++) 
	{
		if (element->exponent % 2) 
		{
			CLog::LogError(WHEREARG,"CDivisors","ControlExponents",element->prime,element->exponent,"PrimeIdeal has odd exponent.");
			return ConstRC::GeneralError;
		}
	}

	return ConstRC::Ok;
}
bool CDivisors::Equals(const CDivisors& aOperand) const
{
	if (this->MaxElementIndex != aOperand.MaxElementIndex) return false;
	if (this->Elements == NULL && aOperand.Elements == NULL) return true;
	if (this->Elements != NULL && aOperand.Elements == NULL) return false;
	if (this->Elements == NULL && aOperand.Elements != NULL) return false;

	if (this->MaxElementIndex > 0)
	{
		nfs_sieving_element * e1 = this->Elements;
		nfs_sieving_element * e1End = this->Elements + this->MaxElementIndex;
		nfs_sieving_element * e2 = aOperand.Elements;
		do
		{
			if (e1->prime    != e2->prime) return false;
			if (e1->root     != e2->root) return false;
			if (e1->exponent != e2->exponent) return false;
			if (e1->flags    != e2->flags) return false;

			e2++;
		}
		while (e1++ != e1End);
	}

	return true;
}
int CDivisors::Set_AssignedElements(unsigned long aAssignedElements)
{
	if (this->AssignedElements >= aAssignedElements)
	{
		return ConstRC::Ok;
	}

	nfs_sieving_element * elements = NULL;
	elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**)&elements, sizeof(nfs_sieving_element)*(aAssignedElements));
	if (elements == NULL) 
	{
		// Chyba
		CLog::LogError(WHEREARG,"CDivisors","Set_AssignedElements",aAssignedElements,40,"There was an error reallocating array elements.");				
		return ConstRC::NotEnoughMemory;
	}
	// Nastavime na 0
	memset(elements, 0, sizeof(nfs_sieving_element) * aAssignedElements);

	if (this->MaxElementIndex > -1)
	{
		// Zkopirujeme z puvodni relace
		memcpy(elements, this->Elements, sizeof(nfs_sieving_element) * (this->MaxElementIndex + 1));
	}

	this->AssignedElements = aAssignedElements;
	if (this->Elements != NULL)
	{
		free(this->Elements);
	}
	this->Elements = elements;

	return ConstRC::Ok;
}
int CDivisors::CombineWithDivisors(const CDivisors &aOperand)
{
	CDivisors rel;
	if (NFSUtils::MergeTwoSortedArrays(&rel.Elements,&rel.AssignedElements,&rel.MaxElementIndex,this->Elements,this->MaxElementIndex,aOperand.Elements,aOperand.MaxElementIndex) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}
	
	// Na konci provedeme kopirovani mezi jednotlivymi poli
	// ****** Nejprve pole elements.
	if (this->AssignedElements < (unsigned)(rel.MaxElementIndex + 1)) 
	{
		// Realokovat
		this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**)&this->Elements, sizeof(nfs_sieving_element)*(rel.MaxElementIndex + 3));
		if (this->Elements == NULL) 
		{
			// Chyba
			CLog::LogError(WHEREARG,"CDivisors","CombineTwoRelationParts",rel.MaxElementIndex+3,40,"There was an error reallocating array elements.");				
			this->AssignedElements = 0;			
			this->MaxElementIndex = -1;
			return ConstRC::NotEnoughMemory;
		} 
		this->AssignedElements = rel.MaxElementIndex + 3;
		this->MaxElementIndex = -1;
	}
	// Nastavime target relation na 0
	memset(this->Elements, 0, sizeof(nfs_sieving_element) * this->AssignedElements);
	// Zkopirujeme z rel
	memcpy(this->Elements, rel.Elements, sizeof(nfs_sieving_element) * (rel.MaxElementIndex + 1));
	this->MaxElementIndex = rel.MaxElementIndex;

	return ConstRC::Ok;
}
void CDivisors::PrintToScreen()
{
	if (this->MaxElementIndex > -1) 
	{
		printf("Divisors: ");    
		for (long i = 0; i <= this->MaxElementIndex; i++) 
		{
			printf("(%d [%d])^%d*",this->Elements[i].prime, this->Elements[i].root, this->Elements[i].exponent);
		}
		printf("\n");
	} 
	else 
	{
		printf("No Divisors.\n");
	}
}

/*!
	This method is used to save these divisors into a file.
	File Template:
	==================================================
	Elements count
	Elements as (prime|exponent|c_p|flag)
*/
int CDivisors::PrintToFile(FILE * aFw)
{
	fprintf(aFw, "%ld\n", this->MaxElementIndex + 1);
	for (long i = 0; i <= this->MaxElementIndex; i++) 
	{
		fprintf(aFw, "(%d|%u", this->Elements[i].prime, this->Elements[i].exponent);
		if (this->Elements[i].flags != 0)
		{
			fprintf(aFw, "|%d|%u)", this->Elements[i].root, this->Elements[i].flags);
		}
		else if (this->Elements[i].root != 0)
		{
			fprintf(aFw, "|%d)", this->Elements[i].root);
		}
		else
		{
			fprintf(aFw, ")");
		}
	}
	fprintf(aFw, "\n");

	return ConstRC::Ok;
}
/*!
	This method is used to read these divisors from a file.
	File Template:
	==================================================
	Elements as (prime|exponent|c_p|flag)
*/
int CDivisors::ReadFromFile(long aCount, string aLine)
{
	this->MaxElementIndex = aCount - 1;
	if (aCount > 0) 
	{
		size_t indexStart;
		size_t indexEnd;
		size_t end;

		// At first, we check whether the allocated space for integral elements is big enough.
		if ((signed) this->AssignedElements < aCount)
		{
			// Reallocate 
			this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**) &this->Elements, aCount*sizeof(nfs_sieving_element));
			if (this->Elements == NULL) 
			{
				CLog::LogError(WHEREARG,"CDivisors","ReadFromFile",0,0,"Could not resize elements field to fit all elements");
				return ConstRC::NotEnoughMemory;
			}
			this->AssignedElements = aCount;
		}
		indexStart = 0;

		for (long i = 0; i < aCount; i++)
		{
			indexStart = aLine.find("(", indexStart);
			indexEnd = aLine.find("|", indexStart + 1);
			this->Elements[i].prime = atoi(aLine.substr(indexStart + 1, indexEnd).c_str());
			this->Elements[i].root = 0;
			this->Elements[i].flags = 0;
			
			indexStart = indexEnd;
			end = aLine.find(")", indexStart + 1);

			if ((indexEnd = aLine.find("|", indexStart + 1)) != string::npos && indexEnd < end)
			{
				sscanf(aLine.substr(indexStart + 1, indexEnd).c_str(), "%u", &(this->Elements[i].exponent));
				//this->Elements[i].exponent = atoi(aLine.substr(indexStart + 1, indexEnd).c_str());
				indexStart = indexEnd;
				
				if ((indexEnd = aLine.find("|", indexStart + 1)) != string::npos && indexEnd < end)
				{
					this->Elements[i].root = atoi(aLine.substr(indexStart + 1, indexEnd).c_str());
					indexStart = indexEnd;
					sscanf(aLine.substr(indexStart + 1, end).c_str(), "%u", &(this->Elements[i].flags));
					//this->Elements[i].flags = atoi(aLine.substr(indexStart + 1, end).c_str());
				}
				else
				{
					this->Elements[i].root = atoi(aLine.substr(indexStart + 1, end).c_str());
				}
			}
			else
			{
				sscanf(aLine.substr(indexStart + 1, end).c_str(), "%u", &(this->Elements[i].exponent));
				//this->Elements[i].exponent = atoi(aLine.substr(indexStart + 1, end).c_str());
			}
		}
	}

	return ConstRC::Ok;
}
/*!
	This method is used to read these divisors from a file.
	Return value is ConstRC::NotEnoughMemory in case of error or index of starting large prime ideal.
	File Template:
	==================================================
	Elements as (prime|exponent|c_p|flag)
*/
int CDivisors::ReadFromFile(long aCount, main_sieving_type aLargeBound, string aLine)
{
	this->MaxElementIndex = aCount - 1;
	int largePrimesIndex = 0; // all
	if (aCount > 0) 
	{
		size_t indexStart;
		size_t indexEnd;
		size_t end;

		// At first, we check whether the allocated space for integral elements is big enough.
		if ((signed) this->AssignedElements < aCount)
		{
			// Reallocate 
			this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**) &this->Elements, aCount*sizeof(nfs_sieving_element));
			if (this->Elements == NULL) 
			{
				CLog::LogError(WHEREARG,"CDivisors","ReadFromFile",0,0,"Could not resize elements field to fit all elements");
				return ConstRC::NotEnoughMemory;
			}
			this->AssignedElements = aCount;
		}
		indexStart = 0;

		for (long i = 0; i < aCount; i++)
		{
			indexStart = aLine.find("(", indexStart);
			indexEnd = aLine.find("|", indexStart + 1);
			this->Elements[i].prime = atoi(aLine.substr(indexStart + 1, indexEnd).c_str());
			this->Elements[i].root = 0;
			this->Elements[i].flags = 0;

			if (this->Elements[i].prime < aLargeBound)
			{
				largePrimesIndex = i + 1;
			}
			
			indexStart = indexEnd;
			end = aLine.find(")", indexStart + 1);

			if ((indexEnd = aLine.find("|", indexStart + 1)) != string::npos && indexEnd < end)
			{
				sscanf(aLine.substr(indexStart + 1, indexEnd).c_str(), "%u", &(this->Elements[i].exponent));
				//this->Elements[i].exponent = atoi(aLine.substr(indexStart + 1, indexEnd).c_str());
				indexStart = indexEnd;
				
				if ((indexEnd = aLine.find("|", indexStart + 1)) != string::npos && indexEnd < end)
				{
					this->Elements[i].root = atoi(aLine.substr(indexStart + 1, indexEnd).c_str());
					indexStart = indexEnd;
					sscanf(aLine.substr(indexStart + 1, end).c_str(), "%u", &(this->Elements[i].flags));
					//this->Elements[i].flags = atoi(aLine.substr(indexStart + 1, end).c_str());
				}
				else
				{
					this->Elements[i].root = atoi(aLine.substr(indexStart + 1, end).c_str());
				}
			}
			else
			{
				sscanf(aLine.substr(indexStart + 1, end).c_str(), "%u", &(this->Elements[i].exponent));
				//this->Elements[i].exponent = atoi(aLine.substr(indexStart + 1, end).c_str());
			}
		}
	}

	return largePrimesIndex;
}
int CDivisors::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_DIVISORS_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CDivisors","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_MAX_ELEMENT_INDEX_NAME, this->MaxElementIndex);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeUnsignedLong(aWriter, XML_SERIAL_ASSIGNED_ELEMENTS_NAME, this->AssignedElements);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENTS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CDivisors","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (this->MaxElementIndex > -1)
	{
		for (long i = 0; i <= this->MaxElementIndex; i++)
		{
			rc = CXMLHelper::SerializeSievingElement(aWriter, XML_SERIAL_ELEMENT_NAME, this->Elements[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENTS_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CDivisors","Serialize",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CDivisors","Serialize",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;  
}

int CDivisors::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_DIVISORS_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CDivisors","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CDivisors","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_MAX_ELEMENT_INDEX_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize max element index...", this->MaxElementIndex, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ASSIGNED_ELEMENTS_NAME))
		{
			unsigned long assigned;
			rc = CXMLHelper::DeserializeUnsignedLong(aReader, "Deserialize assigned elements...", assigned, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}

			if (assigned < this->MaxElementIndex + 1)
			{
				assigned = this->MaxElementIndex + 3;
			}

			if (this->AssignedElements < assigned) 
			{
				// Reallocation
				this->Elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**)&this->Elements, sizeof(nfs_sieving_element) * assigned);
				if (this->Elements == NULL) 
				{
					CLog::LogError(WHEREARG,"CDivisors","Deserialize",0,0,"There was an error reallocating array elements.");				
					this->AssignedElements = 0;			
					this->MaxElementIndex = -1;
					rc = ConstRC::NotEnoughMemory;
					break;
				} 
				this->AssignedElements = assigned;
			}
			// Set target relation on 0
			memset(this->Elements, 0, sizeof(nfs_sieving_element) * this->AssignedElements);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENTS_NAME))
		{
			int elementsDepth = xmlTextReaderDepth(aReader);
			nfs_sieving_element * current = this->Elements;
			for (long i = 0; i <= this->MaxElementIndex; i++)
			{
				if (CXMLHelper::MoveToNextNode(aReader, elementsDepth) != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				rc = CXMLHelper::DeserializeSievingElement(aReader, "Deserialize sieving element...", *current, FALSE);
				if (rc != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				current++;
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
