#include "duplicate_hashtable.h"

#include "const_return_codes.h"
#include "const_hashtables.h"
#include "const_debug.h"
#include "log.h"
#include "nfs_utils.h"

CDuplicateHashtable::CDuplicateHashtable()
{
	this->Hashtable = NULL;
	this->AllocatedSize = 0;
	this->Mask = 0;
}
CDuplicateHashtable::~CDuplicateHashtable()
{
	DeleteHashtable();
}
void CDuplicateHashtable::DeleteHashtable()
{
	if (this->Hashtable != NULL)
	{
		free(this->Hashtable);
		this->Hashtable = NULL;
		this->AllocatedSize = 0;
		this->Mask = 0;
	}
}
int CDuplicateHashtable::SetupHashtable(long aAmount)
{
	if (this->Hashtable != NULL)
	{
		CLog::LogError(WHEREARG,"CDuplicateHashtable","SetupHashtableProb",0,0,"Hashtable is not NULL.");
		return ConstRC::BadArgument;
	}

	this->AllocatedSize = Utils::GetSuitableHashtableSize(aAmount);
	this->Hashtable = (relation_hash_type*) NFSUtils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(relation_hash_type));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CDuplicateHashtable","SetupHashtableProb", 0, this->AllocatedSize * sizeof(relation_hash_type),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	ClearHashtable();

	return ConstRC::Ok;
}
int CDuplicateHashtable::ClearHashtable()
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CDuplicateHashtable","ClearHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	memset(this->Hashtable, 0, this->AllocatedSize * sizeof(relation_hash_type));	

	return ConstRC::Ok;
}
/*!
	This method tries to add relation defined by pair (A,B) to hashtable.
	If relation is already in hashtable we indicate this in parameter aPresent.
	We are using special function for relation hash calculation.

	As a hash for relation (A,B) we use:
	PI = pi * 10^17 
	E = e * 10^17
	H(A,B) = PI*A + E*B mod 2^32 (or 2^64) 

	By Cavallar Stefania - Strategies in Filtering in the Number Field Sieve.

*/
int CDuplicateHashtable::AddToTable(main_sieving_type aA, main_sieving_type aB, bool & aPresent)
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CDuplicateHashtable","AddToTable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	relation_index_type index;
	relation_hash_type relationHash = aA * CDuplicateHashtable::PI + aB * CDuplicateHashtable::E;
	
	int rc = FindPrimeInHashtable(relationHash, aPresent, index);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (aPresent == false)
	{
		this->Hashtable[index] = relationHash;
	}

	return ConstRC::Ok;
} 
/*!
	This method tries to find aRelationHash in hashtable, and indicates
	the result of this search in two values: the return value, which is an index
	corresponding to the real or expected position of the relation in the hashtable,
	and value of aPresent, which indicates whether the relation has already been
	in the hashtable or no. In the second case, the return value has meaning
	"if you want to insert relation, this index is the one where it should be inserted".

*/
int CDuplicateHashtable::FindPrimeInHashtable(relation_hash_type aRelationHash, bool & aPresent, relation_index_type & aIndex) const
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CDuplicateHashtable","FindPrimeInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aIndex = 0;
		return ConstRC::BadArgument;
	}

	relation_index_type hash = aRelationHash & this->Mask;			
	if (CollisionResolution(hash, aRelationHash, aIndex) != ConstRC::Ok)
	{
		aPresent = false;
		aIndex = 0;
		return ConstRC::GeneralError;
	}

    if (this->Hashtable[aIndex] == 0) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

	return ConstRC::Ok;
}
/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where relation can be stored.

*/
int CDuplicateHashtable::CollisionResolution(relation_index_type aHash, relation_hash_type aRelationHash, relation_index_type & aIndex) const
{
	aIndex = aHash;
	long iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CDuplicateHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	relation_hash_type * aux = this->Hashtable + aIndex;
	relation_hash_type * end = this->Hashtable + this->AllocatedSize;

    while ((*aux != aRelationHash) && (*aux != 0)) 
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"CDuplicateHashtable","CollisionResolution",0,0,"The hash table has overflown.");
				return ConstRC::GeneralError;
			}
			aux = this->Hashtable;
		}
	}

	aIndex = aux - this->Hashtable;

	return ConstRC::Ok;
}
