#include "factor_base_info.h"

#include <cstring>
#include "integer_class.h"
#include "const_return_codes.h"
#include "const_element_flags.h"
#include "const_sieving_phase.h"
#include "const_parameter_names.h"
#include "const_hashtables.h"
#include <math.h>
#include "log.h"

#include "nfs_utils.h"

using namespace std;

CFBInfo::CFBInfo() : MSerializable(ConstFN::FACTOR_BASE)
{
	this->FactorBase = NULL;
	this->FBMaxAllocatedIndex = -1;
	this->FBMaxIndex = -1;
	this->UpperFBBound = 0;
	this->AssertionMode = false;
	this->InfoMode = false;

	mpz_init(this->iAuxMpz1);
	mpz_init(this->iAuxMpz2);
	mpz_init(this->iAuxMpz3);
}
CFBInfo::~CFBInfo()
{
	Reset();
	mpz_clear(this->iAuxMpz1);
	mpz_clear(this->iAuxMpz2);
	mpz_clear(this->iAuxMpz3);
}

int CFBInfo::Reset()
{
	Dispose();
	this->UpperFBBound = 0;
	this->FBFullFileName = "";
	this->AssertionMode = false;
	this->InfoMode = false;

	return ConstRC::Ok;
}
int CFBInfo::Dispose()
{
	if (this->FactorBase != NULL)
	{
		free(this->FactorBase);
		this->FactorBase = NULL;			
	}
	this->FBMaxAllocatedIndex = -1;
	this->FBMaxIndex = -1;

	return ConstRC::Ok;
}
int CFBInfo::InitFB()
{
	if (this->FactorBase != NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","InitFB",0,0,"FB has to be NULL");
		return ConstRC::BadArgument;
	}

	if (this->UpperFBBound <= 0)
	{
		CLog::LogError(WHEREARG,"CFBInfo","InitFB",this->UpperFBBound,0,"Upper bound is wrong");
		return ConstRC::BadArgument;
	}

	int rc;
	long primeCount;
	// The first problem is to find, how much space is needed for the factor base. To find this, we must 
	// "list" through the file of primes. If the file of primes does not exist, it is regenerated now.	
	rc = CIntegerClass::GetPrimesCount(0, this->UpperFBBound, primeCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	primeCount++; // inclusion of -1 to Integral FB needs 1 more space - algebraic FB will be probably bigger

	// Clear: forces the allocated size to be a multiple of 4.
	int clear = 4 - (primeCount % 4);

	if ((this->FactorBase = (fb_element*) Utils::AllocationMachine((void **) &(this->FactorBase),(primeCount+4+clear)*sizeof(fb_element))) == NULL) 
	{
		// Allocation of the FB has failed!
		CLog::LogError(WHEREARG,"CFBInfo","InitFB",(primeCount+4+clear)*sizeof(fb_element),1,"Problem in allocation of FB");
		return ConstRC::NotEnoughMemory;
	}

	// it is index!!!
	this->FBMaxAllocatedIndex = primeCount + 3 + clear;
	this->FBMaxIndex = -1;

	memset(this->FactorBase,0,(primeCount + 4 + clear) * sizeof(fb_element));

	return ConstRC::Ok;
}

/*!
 * \brief This method fills factor base with prime numbers / prime ideals.
 *
 *  According to degree of the sieving poly it chooses suitable filling method.
 *  Also the field of the zero root indices in the NFInfo is filled.
 *
 */
int CFBInfo::FillFB(CNumberFieldInfo &aNFInfo)
{
	if (aNFInfo.SievingPoly == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillFB",0,0,"Sieving poly is NULL.");
		return ConstRC::BadArgument;
	}

	int rc;
	if (aNFInfo.SievingPoly->GetDegree() > 1)
	{
		rc = FillAlgebraicFB(aNFInfo);
	}
	else
	{
		rc = FillIntegralFB(aNFInfo); 
	}

	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	FillRootZeroPrimes(aNFInfo.RZeros_Indices);

	return ConstRC::Ok;
}
int CFBInfo::FillAuxFields(CNumberFieldInfo &aNFInfo)
{
	if (aNFInfo.SievingPoly->GetDegree() > 1)
	{
		FillSpecialIdeals(aNFInfo.Domain, aNFInfo.SpecialIdeals);
	}

	FillRootZeroPrimes(aNFInfo.RZeros_Indices);

	return ConstRC::Ok;
}
/*!
 * \brief Method which fills the integral factor base with prime numbers.
 *
 * This method fills the integral factor base with prime numbers. For this purpose,
 * it needs read access to a file containing at least as much primes as needed. 
 * The first element (at index 0) is -1.
 *
 */
int CFBInfo::FillIntegralFB(CNumberFieldInfo &aNFInfo)
{
	Polynomial *aPoly = aNFInfo.SievingPoly;

	// We just open file of primes and fill the FB. 
	// Moreover, we calculate logarithms of the primes and put them into the log array
	if (this->FactorBase == NULL || aPoly == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillIntegralFB",0,0,"Integral FB or Poly is NULL");
		return ConstRC::BadArgument;
	}

	this->FBMaxIndex = -1;

	int rc = 0; // Must be comparable to a negative value of EOF
	unsigned long primesCount;
	long * primes = NULL;
	rc = CIntegerClass::GetPrimes(0, this->UpperFBBound, primesCount, primes);
	if (rc != ConstRC::Ok)
	{
		if (primes != NULL)
		{
			free(primes);
			primes = NULL;
		}
		return rc;
	}
	if (primesCount >= this->FBMaxAllocatedIndex)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillIntegralFB",0,0,"Integral FB is not large enough");
		return ConstRC::GeneralError;
	}

	log_type integralLog;
	long prime = 0;
	double realLog     = 0;
	double logConst    = (double) 1/(double) log10((double)ConstSPhase::LOGARITHM_BASE);
	double rounded     = 0;
	bool nonmonic = mpz_cmp_ui(*aPoly->Leading2(), 1) != 0;

	this->FactorBase[0].prime = -1;
	for (unsigned long i = 0, j = 1; i < primesCount; i++, j++)
	{
		prime = primes[i];
		this->FactorBase[j].prime = prime;	
		realLog = log10((double) prime)*logConst;
		rounded = Utils::Round(realLog);
		integralLog = (log_type) rounded;
		this->FactorBase[j].log = integralLog;	

		// divisor of leading coefficient
		if (aNFInfo.Domain->IsASpecialPrime(prime)) {
			this->FactorBase[j].flags |= ConstEF::SPECIAL_PRIME;
			continue;
		}
	
		mpz_set_ui(this->iAuxMpz1, prime);
		mpz_neg(this->iAuxMpz2, *aPoly->Get2(0));
		if (nonmonic)
		{
			mpz_invert(this->iAuxMpz3, *aPoly->Leading2(), this->iAuxMpz1);
			mpz_mul(this->iAuxMpz2, this->iAuxMpz2, this->iAuxMpz3);
		}
		this->FactorBase[j].root = mpz_mod_ui(this->iAuxMpz2, this->iAuxMpz2, prime);
		if (this->FactorBase[j].root > 0)
		{
			// we want to save -root mod p
			this->FactorBase[j].root = prime - this->FactorBase[j].root;
		
			mpz_set_si(this->iAuxMpz2, this->FactorBase[j].root);
			mpz_invert(this->iAuxMpz3, this->iAuxMpz2, this->iAuxMpz1);
			this->FactorBase[j].inversion_of_root = mpz_get_si(this->iAuxMpz3);
		}
	}

	this->FBMaxIndex = primesCount; // primesCount - 1 + 1 (because of -1 in FB)

	std::cout <<"\n FillIntegralFB : "<< FBMaxIndex;
	if (primes != NULL)
	{
		free(primes);
		primes = NULL;
	}

	return ConstRC::Ok;
}
/*!
 * \brief Method which fills the algebraic factor base with prime ideals.
 * 
 * This method fills the algebraic factor base with prime ideals up to a specified
 * norm.
 * 
 * Runtime of this method is not negligible, since (in contrary to filling the
 * algebraic FB) it needs to partially factorize polynomials modulo prime, which is
 * a time-consuming operation.
 *
 */
int CFBInfo::FillAlgebraicFB(CNumberFieldInfo & aNFInfo)
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillAlgebraicFB",0,0,"Algebraic FB is NULL");
		return ConstRC::BadArgument;
	}

	this->FBMaxIndex = -1;

	long * roots = NULL;
	roots = (long*) Utils::AllocationMachine((void**) &roots, aNFInfo.SievingPoly->GetDegree() * sizeof(long));
	if (roots == NULL) 
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillAlgebraicFB",0,0,"Could not allocate roots.");
		return ConstRC::NotEnoughMemory;
	}

	int rc;
	long * primes = NULL;
	long prime;
	unsigned long primesCount;
	rc = CIntegerClass::GetPrimes(0, this->UpperFBBound, primesCount, primes);
	if (rc != ConstRC::Ok)
	{
		if (primes != NULL)
		{
			free(primes);
			primes = NULL;
		}
		free(roots);
		roots = NULL;
		return rc;
	}

	aNFInfo.SpecialIdeals.reserve(CFBInfo::DEFAULT_SPECIAL_IDEALS_COUNT);
	double logConst = (double) 1/(double) log10((double)ConstSPhase::LOGARITHM_BASE);
	log_type log;
	int rootNr;	
	int end;
	long result;
	unsigned int cnt;
	PrimeIdeal** list = NULL;
	int degree = aNFInfo.SievingPoly->GetDegree();
	// We will test whether sieving polynomial has some roots mod p.
	for (unsigned long i = 0; i < primesCount; i++)
	{
		prime = primes[i];
		log = (log_type) Utils::Round(log10((double) prime)*logConst);
		// How to distinguish special primes?		
		if (!aNFInfo.Domain->IsASpecialPrime(prime)) 
		{
			if (aNFInfo.SievingPolyMonic == false)
			{
				// if we have nonmonic polynomial we need to find special kind of primes 
				// which divide leading coeff
				// Only for non-special primes - we treat them differently
				if (mpz_divisible_ui_p(*aNFInfo.SievingPoly->Leading2(), prime))
				{
					// we know that \bar{f} = f{x/c_{d}} * c_{d}^{d - 1}
					// so if p | c_{d} than we have two roots: 0 and c_{d - 1} mod p
					// prime ideal (p, \bar{\theta} - (c_{d - 1} mod p)) will never be used - check yourself !!!
					// therefore we add only the first one
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, 0, ConstEF::DIVIDES_L, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}

					// this is not prime ideal but only principal ideal (p)
					// we use it for case when p | b => (c_{d}a + b\bar{\theta}) = (p)^{k} * ....
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, prime, ConstEF::DIVIDES_L, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}
					continue;
				}
			}

			// We need root of the sieving poly - we will use these roots to identify prime ideals (NOT c_p!!!)
			rootNr = aNFInfo.SievingPoly->RootsModPrimeGeneral(roots, prime);
			if (rootNr > 0) 
			{				
				if (rootNr > degree) 
				{
					CLog::Log(MAXIMAL_INT_VALUE, "0620000", "Degree of polynomial (%d) is lower than number of roots (%d) mod %d!", degree, rootNr, prime);
					CLog::LogError(WHEREARG,"CFBInfo","FillAlgebraicFB",0,0,"Degree of polynomial is lower than number of roots.");
					free(primes);
					primes = NULL;
					free(roots);
					roots = NULL;
					return ConstRC::GeneralError;
				}
				qsort(roots, rootNr, sizeof(long), &Utils::CompareLongs);
				// because we save -root mod p we need to save it as the first element
				if (roots[0] == 0)
				{
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, 0, 0, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}
					end = 0;
				}
				else
				{
					end = -1;
				}
				for (int j = rootNr - 1; j > end; j--) 
				{
					// we want to save -root mod p
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, -roots[j], j, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}
					if (this->AssertionMode) 
					{
						result = aNFInfo.SievingPoly->EvaluateModPrime(roots[j], prime);
						if (result != 0) 
						{
							CLog::Log(MAXIMAL_INT_VALUE, "0620001", "Error determining root of polynomial mod %d. Fake root is %ld, result is %ld. Total roots %d\n",prime,roots[j],result,rootNr);
						} 
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
						else 
						{
							CLog::Log(0, "0620002", "Correct root %d of polynomial mod %d.\n",roots[j],prime);
						}
#endif
					}
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
					CLog::Log(0, "0620003", "\r   Filling algebraic base with prime ideals ... so far %d (%d).       ", aMaxIndex, prime);
#endif
				}
			}
		}
		else // Special primes
		{
			// We find all of the special prime ideals, but we do not use them for sieving; they are usually negli7gible,
			// being only primes of very small norm.
			aNFInfo.Domain->PrimeIdealDecomposition(&list, &cnt, prime);
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
 			CLog::Log(0, "0620004", "Special prime %d has %d prime ideals.\n", prime, cnt);
#endif
			for (unsigned j = 0; j < cnt; j++) 
			{
				rc = CFBInfo::AddPrimeIdealToFB(prime, log, j, cnt-j-1, true);				
				if (rc != ConstRC::Ok) 
				{
					if (list != NULL) 
					{
						free(list);
						list = NULL;
					}
					free(primes);
					primes = NULL;
					free(roots);
					roots = NULL;
					return rc;
				}
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
				CLog::Log(0, "0620005", "  Ideal nr %d has ramification %d and residual %d.\n",j,list[j]->Ramification,list[j]->Residual);
#endif
				list[j]->Index = this->FBMaxIndex;
				list[j]->Cp = j;
				aNFInfo.SpecialIdeals.push_back(list[j]);
			}			
			if (list != NULL) 
			{
				free(list);
				list = NULL;
			}

		} // end if - else special primes

	} // end for cycle - primes
	
	// We add a last '0' value as a stopper flag for the sieving loop
	rc  = CFBInfo::AddPrimeIdealToFB(0, 0, 0, 0, false);
	this->FBMaxIndex--;
	
	free(primes);
	primes = NULL;
	free(roots);
	roots = NULL;

	return rc;
}
int CFBInfo::RefillFB(const long &aNewBound, CNumberFieldInfo & aNFInfo)
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","RefillFB",0,0,"FB is NULL");
		return ConstRC::GeneralError;
	}

	int degree = aNFInfo.SievingPoly->GetDegree();
	long * roots = NULL;
	roots = (long*) Utils::AllocationMachine((void**) &roots, degree * sizeof(long));
	if (roots == NULL) 
	{
		CLog::LogError(WHEREARG,"CFBInfo","RefillFB",0,0,"Could not allocate roots.");
		return ConstRC::NotEnoughMemory;
	}

	int rc;
	long * primes = NULL;
	long prime;
	unsigned long primesCount = 0;
	unsigned int cnt;
	PrimeIdeal** list = NULL;
	rc = CIntegerClass::GetPrimes(this->UpperFBBound, aNewBound, primesCount, primes);
	if (rc != ConstRC::Ok)
	{
		if (primes != NULL)
		{
			free(primes);
			primes = NULL;
		}
		free(roots);
		roots = NULL;
		return rc;
	}

	// reallocation

	// Clear: forces the allocated size to be a multiple of 4.
	int clear = 4 - ((this->FBMaxIndex + 1 + primesCount) % 4);

	// The factor base contains all prime ideals of Z_K which contain prime numbers lesser than aNewBound
	if ((this->FactorBase = (fb_element*) Utils::AllocationMachine((void **) &(this->FactorBase),(this->FBMaxIndex + 1 + primesCount + 4 + clear)*sizeof(fb_element))) == NULL) 
	{
		// Allocation of the algebraic FB has failed!
		CLog::LogError(WHEREARG,"CFBInfo","RefillAlgebraicFB",(this->FBMaxIndex + 1 + primesCount + 4 + clear)*sizeof(fb_element),1,"Problem in allocation of FB");
		return ConstRC::NotEnoughMemory;
	}

	// it is index!!!
	this->FBMaxAllocatedIndex = this->FBMaxIndex + primesCount + 4 + clear;
	memset(&this->FactorBase[this->FBMaxIndex + 1], 0, (this->FBMaxAllocatedIndex - this->FBMaxIndex) * sizeof(fb_element));

	double logConst = (double) 1/(double) log10((double)ConstSPhase::LOGARITHM_BASE);
	log_type log;
	int rootNr;	
	int end;
	long result;
	int j;
	// We will test whether sieving polynomial has some roots mod p.
	for (unsigned long i = 0; i < primesCount; i++)
	{
		prime = primes[i];
		log = (log_type) Utils::Round(log10((double) prime)*logConst);
		// How to distinguish special primes?		
		if (!aNFInfo.Domain->IsASpecialPrime(prime)) 
		{
			if (aNFInfo.SievingPolyMonic == false)
			{
				// if we have nonmonic polynomial we need to find special kind of primes 
				// which divide leading coeff
				// Only for non-special primes - we treat the special primes differently
				if (mpz_divisible_ui_p(*aNFInfo.SievingPoly->Leading2(), prime))
				{
					// we know that \bar{f} = f{x/c_{d}} * c_{d}^{d - 1}
					// so if p | c_{d} than we have two roots: 0 and c_{d - 1} mod p
					// prime ideal (p, \bar{\theta} - (c_{d - 1} mod p)) will never be used - check yourself !!!
					// therefore we add only the first one
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, 0, ConstEF::DIVIDES_L, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}

					// this is not prime ideal but only principal ideal (p)
					// we use it for case when p | b => (c_{d}a + b\bar{\theta}) = (p)^{k} * ....
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, prime, ConstEF::DIVIDES_L, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}
					continue;
				}
			}

			// We need root of the sieving poly - we will use these roots to identify prime ideals (NOT c_p!!!)
			rootNr = aNFInfo.SievingPoly->RootsModPrimeGeneral(roots, prime);
			if (rootNr > 0) 
			{				
				if (rootNr > degree) 
				{
					CLog::Log(MAXIMAL_INT_VALUE, "0620006", "Degree of polynomial (%d) is lower than number of roots (%d) mod %d!", degree, rootNr, prime);
					CLog::LogError(WHEREARG,"CFBInfo","FillAlgebraicFB",0,0,"Degree of polynomial is lower than number of roots.");
					free(primes);
					primes = NULL;
					free(roots);
					roots = NULL;
					return ConstRC::GeneralError;
				}
				qsort(roots, rootNr, sizeof(long), &Utils::CompareLongs);	
				// because we save -root mod p we need to save it as the first element
				if (roots[0] == 0)
				{
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, 0, 0, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}
					end = 0;
				}
				else
				{
					end = -1;
				}
				for (j = rootNr - 1; j > end; j--) 
				{
					rc = CFBInfo::AddPrimeIdealToFB(prime, log, -roots[j], j, false);
					if (rc != ConstRC::Ok) 
					{
						free(primes);
						primes = NULL;
						free(roots);
						roots = NULL;
						return rc;
					}
					if (this->AssertionMode) 
					{
						result = aNFInfo.SievingPoly->EvaluateModPrime(roots[j], prime);
						if (result != 0) 
						{
							CLog::Log(MAXIMAL_INT_VALUE, "0620007", "Error determining root of polynomial mod %d. Fake root is %ld, result is %ld. Total roots %d\n",prime,roots[j],result,rootNr);
						} 
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
						else 
						{
							CLog::Log(0, "0620008", "Correct root %d of polynomial mod %d.\n",roots[j],prime);
						}
#endif
					}
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
					CLog::Log(0, "0620009", "\r   Filling algebraic base with prime ideals ... so far %d (%d).       ", aMaxIndex, prime);
#endif
				}
			}
		}
		else // Special primes
		{
			// We find all of the special prime ideals, but we do not use them for sieving; they are usually negli7gible,
			// being only primes of very small norm.
			aNFInfo.Domain->PrimeIdealDecomposition(&list, &cnt, prime);
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
 			CLog::Log(0, "0620010", "Special prime %d has %d prime ideals.\n", prime, cnt);
#endif
			for (unsigned j = 0; j < cnt; j++) 
			{
				rc = CFBInfo::AddPrimeIdealToFB(prime, log, j, cnt-j-1, true);				
				if (rc != ConstRC::Ok) 
				{
					if (list != NULL) 
					{
						free(list);
						list = NULL;
					}
					free(primes);
					primes = NULL;
					free(roots);
					roots = NULL;
					return rc;
				}
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
				CLog::Log(0, "0620011", "  Ideal nr %d has ramification %d and residual %d.\n",j,list[j]->Ramification,list[j]->Residual);
#endif
				list[j]->Index = this->FBMaxIndex;
				list[j]->Cp = j;
				aNFInfo.SpecialIdeals.push_back(list[j]);
			}			
			if (list != NULL) 
			{
				free(list);
				list = NULL;
			}

		} // end if - else special primes

	} // end for cycle - primes
	
	
	// We add a last '0' value as a stopper flag for the sieving loop
	rc  = CFBInfo::AddPrimeIdealToFB(0, 0, 0, 0, false);
	this->FBMaxIndex--;
	this->UpperFBBound = aNewBound;
	
	free(primes);
	primes = NULL;
	free(roots);
	roots = NULL;

	return rc;
}
int CFBInfo::AddPrimeIdealToFB(main_sieving_type aPrime, log_type aLog, main_sieving_type aRoot, unsigned int aNext, bool aSpecial)
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","AddPrimeIdealToFB",0,0,"Algebraic FB is NULL");
		return ConstRC::BadArgument;
	}

	// At first, we check out whether current allocated space is big enough
	if (this->FBMaxIndex == this->FBMaxAllocatedIndex) 
	{
		this->FactorBase = (fb_element*) Utils::AllocationMachine((void**) &this->FactorBase, 
			(this->FBMaxAllocatedIndex + 1 + CFBInfo::DEFAULT_PRIME_IDEALS_REALLOC)*sizeof(fb_element));
		if (this->FactorBase == NULL) 
		{
			CLog::LogError(WHEREARG,"CFBInfo","AddPrimeIdealToFB",(this->FBMaxAllocatedIndex + 1 + CFBInfo::DEFAULT_PRIME_IDEALS_REALLOC)*sizeof(fb_element),1,"Problem in reallocation of the field");
			return ConstRC::NotEnoughMemory;
		} 
		else 
		{
			memset(&this->FactorBase[this->FBMaxAllocatedIndex + 1], 0, CFBInfo::DEFAULT_PRIME_IDEALS_REALLOC * sizeof(fb_element));
			this->FBMaxAllocatedIndex += CFBInfo::DEFAULT_PRIME_IDEALS_REALLOC;
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
				CLog::Log(0, "0620012", "Realloc to : %d\n",aMaxAllocatedIndex+1);
#endif
		}
	}
	this->FBMaxIndex++;
	this->FactorBase[this->FBMaxIndex].prime = aPrime;
	this->FactorBase[this->FBMaxIndex].log = aLog;
	this->FactorBase[this->FBMaxIndex].flags = 0;
	if (!aSpecial) 
	{		
		if (aRoot < 0)
		{
			// we want to save -root mod p and aRoot = - root
			aRoot = aPrime + aRoot;
		}
		this->FactorBase[this->FBMaxIndex].root = aRoot;
		if (aPrime) 
		{
			if (aRoot) 
			{
				if (aPrime != aRoot)
				{
					// We can, and will, invert root mod p (useful when performing line sieving with respect to the 'a' axis).
					mpz_set_si(this->iAuxMpz1, aRoot);
					mpz_set_ui(this->iAuxMpz2, aPrime);
					mpz_invert(this->iAuxMpz3, this->iAuxMpz1, this->iAuxMpz2);
					this->FactorBase[this->FBMaxIndex].inversion_of_root = mpz_get_si(this->iAuxMpz3);
					if (!this->FactorBase[this->FBMaxIndex].inversion_of_root) 
					{
						printf("AddPrimeIdeal: strange error, inverse of %d mod %d is 0!\n",aRoot,aPrime);
					}	
				}
				//else
				//{
				//	// special kind of ideal - (p,p)
				//}
			}
		} 
		//else 
		//{ 
		//	// prime == 0. end of the sieving base. Nothing to set up.
		//}
		//this->FactorBase[aMaxIndex].flags &= ~ConstEF::SPECIAL_PRIME;
	}
	else 
	{
		// In this case, aRoot is not the real c_p, but rather a flag expressing index of the special prime ideal
		this->FactorBase[this->FBMaxIndex].root = aRoot;
		this->FactorBase[this->FBMaxIndex].flags |= ConstEF::SPECIAL_PRIME;
	}

	this->FactorBase[this->FBMaxIndex].flags |= aNext;
	
	return ConstRC::Ok;
}
int CFBInfo::FillRootZeroPrimes(std::vector<long> &aRZeros_Indices) const
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillRootZeroPrimes",0,0,"FB is NULL");
		return ConstRC::BadArgument;
	}

	aRZeros_Indices.reserve(CFBInfo::DEFAULT_ROOT_ZERO_PRIMES_COUNT);

	for (long i = this->FactorBase->prime > 0 ? 0 : 1; i <= this->FBMaxIndex; i++)
	{
		if (!(this->FactorBase[i].flags & ConstEF::SPECIAL_PRIME))
		{
			if (!this->FactorBase[i].root)
			{
				aRZeros_Indices.push_back(i);
				this->FactorBase[i].flags |= ConstEF::DIVIDES_ROOT;
			}
		}
	}
	
	return ConstRC::Ok;
}
long CFBInfo::GetMaxPrime() const
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","GetMaxPrime",0,0,"FB is NULL");
		return -1;
	}

	return this->FactorBase[this->FBMaxIndex].prime;
}
/**
	This method will find the index of a given fb element (with the required root
	and flags) in the FB, or return -1 if such a element is not found.

	The method uses binary search, so it should be quite fast.

*/
long CFBInfo::FindIndexInFB(nfs_sieving_element * aElement) const 
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FindIndexInFB",0,0,"FB is NULL");
		return -1;
	}

	fb_element pforsieve;
	pforsieve.prime = aElement->prime;
	pforsieve.root  = aElement->root;
	//pforsieve.flags = aElement->flags;
	// Zde je nekdy nejaky problem!!!!
	void* address = bsearch(&pforsieve, this->FactorBase, this->FBMaxIndex + 1, sizeof(fb_element), &NFSUtils::CompareFBElements);
	if (address == NULL) 
	{
		CLog::LogError(WHEREARG,"CFBInfo","FindIndexInFB",aElement->prime,aElement->root,"Could not find indicated element in FB.");				
		return -1;
	} 
	else 
	{
		fb_element* add = (fb_element*) address;
		return (add - this->FactorBase);
	}
}
int CFBInfo::FillSpecialIdeals(IntegralDomain * aDomain, std::vector<PrimeIdeal*> &aSpecialIdeals) const
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","FillSpecialIdeals",0,0,"FB is NULL");
		return ConstRC::BadArgument;
	}

	main_sieving_type last_special_prime = -1;
	main_sieving_type prime = 0;
	unsigned int cnt;
	unsigned int j;
	PrimeIdeal** list = NULL;

	aSpecialIdeals.reserve(CFBInfo::DEFAULT_SPECIAL_IDEALS_COUNT);

	for (long i = 0; i <= this->FBMaxIndex; i++)
	{
		if (this->FactorBase[i].flags & ConstEF::SPECIAL_PRIME)
		{
			prime = this->FactorBase[i].prime;
			if (prime != last_special_prime)
			{
				aDomain->PrimeIdealDecomposition(&list, &cnt, prime);
				for (j = 0; j < cnt; j++) 
				{
					list[j]->Cp = j;
					list[j]->Index = j + i;
					aSpecialIdeals.push_back(list[j]);
				}
				last_special_prime = prime;
				free(list);
				list = NULL;
			}
		}
	}

	return ConstRC::Ok;
}


int CFBInfo::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CFBInfo::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr& aWriter) const
{
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","Serialize",0,0,"FB is NULL");
		return ConstRC::BadArgument;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FACTOR_BASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CFBInfo","Serialize",rc,0,"Could not write the main tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// serialize MaxIndex of FB
	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_MAX_INDEX_NAME, this->FBMaxIndex);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// serialize UpperBound of FB
	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_UPPER_BOUND_NAME, this->UpperFBBound);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENTS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CFBInfo","Serialize",rc,0,"Could not write the elements tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// serialize elements of FB
	fb_element * current = this->FactorBase;
	for (long i = 0; i <= this->FBMaxIndex; i++, current++)
	{
		rc = CXMLHelper::SerializeFBElement(aWriter, XML_SERIAL_ELEMENT_NAME, *current);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENTS_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFBInfo","Serialize",rc,0,"Could not end the elements tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFBInfo","Serialize",rc,0,"Could not end the main tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CFBInfo::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CFBInfo::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr& aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FACTOR_BASE_NAME, mainTagName);

	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CFBInfo","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	// quick reset
	Dispose();

	int depth = xmlTextReaderDepth(aReader);
	int rc;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_MAX_INDEX_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize max index...", this->FBMaxIndex, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_UPPER_BOUND_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize upper bound...", this->UpperFBBound, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENTS_NAME))
		{
			if (this->FBMaxIndex >= 0)
			{
				int clear = 4 - ((this->FBMaxIndex + 1) % 4);
				// it is index!!!
				this->FBMaxAllocatedIndex = this->FBMaxIndex + 1 + 3 + clear;
				if ((this->FactorBase = (fb_element*) Utils::AllocationMachine((void **) &(this->FactorBase),(this->FBMaxAllocatedIndex + 1)*sizeof(fb_element))) == NULL) 
				{
					// Allocation of the integral FB has failed!
					CLog::LogError(WHEREARG,"CFBInfo","DeserializeIntegralFB",(this->FBMaxAllocatedIndex + 1)*sizeof(fb_element),1,"Problem in allocation of the FB");
					rc = ConstRC::NotEnoughMemory;
					break;
				}
				memset(this->FactorBase, 0, (this->FBMaxAllocatedIndex + 1) * sizeof(fb_element));

				int elementsDepth = xmlTextReaderDepth(aReader);
				fb_element * current = this->FactorBase;
				for (long i = 0; i <= this->FBMaxIndex; i++)
				{
					if (CXMLHelper::MoveToNextNode(aReader, elementsDepth) != ConstRC::Ok)
					{
						rc = ConstRC::GeneralError;
						break;
					}
					rc = CXMLHelper::DeserializeFBElement(aReader, "Deserialize fb element...", *current, false);
					if (rc != ConstRC::Ok)
					{
						rc = ConstRC::GeneralError;
						break;
					}
					current++;
				}

				if (rc != ConstRC::Ok)
				{
					break;
				}
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CFBInfo::DeserializeFBInfo(const CBaseParameters &aParam)
{
	// get file name with path
	string fileNameFull = MSerializable::CreateFileNameAuto(aParam, ConstFN::FACTOR_BASE);

	xmlTextReaderPtr reader = xmlNewTextReaderFilename(fileNameFull.c_str());
	if (reader == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","DeserializeFBInfo",0,0,"Could not create xml reader.");
		return ConstRC::GeneralError;
	}
	int rc = CXMLHelper::MoveToNextNode(reader);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextReader(reader);
		if (rc == ConstRC::EndOfFile)
		{
			return ConstRC::Ok;
		}
		return rc;
	}
	
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FACTOR_BASE_NAME, mainTagName);

	xmlChar* name = xmlTextReaderName(reader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","DeserializeFBInfo",0,0,"Deserialization main element name not found.");
		xmlFreeTextReader(reader);
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CFBInfo","DeserializeFBInfo",0,0,"Main element name is wrong.");
		xmlFreeTextReader(reader);
		xmlFree(name);
		return ConstRC::GeneralError; 
	}

	// quick reset
	Dispose();

	int depth = xmlTextReaderDepth(reader);

	while ((rc = CXMLHelper::MoveToNextNode(reader, depth)) == ConstRC::Ok)
	{
		if (NULL != name)  xmlFree(name);
		name = xmlTextReaderName(reader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_MAX_INDEX_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(reader, "Deserialize max index...", this->FBMaxIndex, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			int clear = 4 - ((this->FBMaxIndex + 1) % 4);
			// it is index!!!
			this->FBMaxAllocatedIndex = this->FBMaxIndex + 1 + 3 + clear;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_UPPER_BOUND_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(reader, "Deserialize upper bound...", this->UpperFBBound, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}
	if (NULL != name)  xmlFree(name);
	
	xmlFreeTextReader(reader);
	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CFBInfo::CleanFB(const long *aDetectionField)
{
	std::cerr<<"\nFB - pred:" << FBMaxIndex;
	fb_element* newFB = NULL;
	newFB = (fb_element*) Utils::AllocationMachine((void**) &newFB, (this->FBMaxAllocatedIndex + 1) * sizeof(fb_element));
	if (newFB == NULL) 
	{
		CLog::LogError(WHEREARG,"CFBInfo","CleanFB",0,0,"Not enough memory to allocate data space.");
		return ConstRC::NotEnoughMemory;
	}
	memset(newFB, 0, (this->FBMaxAllocatedIndex + 1) * sizeof(fb_element));
	
	fb_element * current = this->FactorBase;
	long j = -1;
	int removedFinally = 0;
	
	for (long i = 0; i <= this->FBMaxIndex; i++, current++) 
	{
		if (aDetectionField[i] == ConstHT::NOT_YET) 
		{
			removedFinally++;
		} 
		else if (aDetectionField[i] == ConstHT::TWICE) 
		{			
			j++;
			newFB[j].prime = current->prime;
			newFB[j].root = current->root;
			newFB[j].flags = current->flags;
			newFB[j].inversion_of_root = current->inversion_of_root;
			newFB[j].log = current->log;
		} 
		else 
		{
			CLog::Log(MAXIMAL_INT_VALUE, "", "\n");
			CLog::Log(MAXIMAL_INT_VALUE, "0620012", "Error detecting small singletons - prime %d is present exactly once.\n", current->prime);
		}
	}

	free(this->FactorBase);
	this->FactorBase = newFB;
	this->FBMaxIndex = j;

	std::cerr<<"\nFB - po:" << FBMaxIndex;
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
	if (this->InfoMode)
	{
		CLog::Log(0, "0620013", "    Primes finally removed: %d\n", removedFinally);
	}
#endif

	return ConstRC::Ok;
}
int CFBInfo::PrintToScreen() const
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","PrintToScreen",0,0,"FB is NULL");
		return ConstRC::BadArgument;
	}
	CLog::Log(50, "0620014", "\nFactor Base:\n");

	for (long i = 0; i <= this->FBMaxIndex; i++) 
	{
		CLog::Log(50, "0620015", "%8d [%8d]; ", this->FactorBase[i].prime, this->FactorBase[i].root);
	}
	CLog::Log(50, "", "\n");
#endif

	return ConstRC::Ok;
}

int CFBInfo::PrintLCPrimes() const
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	if (this->FactorBase == NULL)
	{
		CLog::LogError(WHEREARG,"CFBInfo","PrintLCPrimes",0,0,"FB is NULL");
		return ConstRC::BadArgument;
	}

	CLog::Log(50, "", "\n");
	CLog::Log(50, "0620016", "    Leading coeff dividing primes: ");
	fb_element *element = this->FactorBase;
	fb_element *end = this->FactorBase + this->FBMaxIndex;

	while (element++ <= end)
	{
		if (element->flags & ConstEF::DIVIDES_L)
		{
			CLog::Log(50, "", "%8d [%8d]; ", element->prime, element->root);
		}
	}
	CLog::Log(50, "", "\n");
#endif

	return ConstRC::Ok;
}
