#include "factorization_info.h"

#include "const_return_codes.h"
#include "const_sieving_phase.h"
#include "log.h"

CFactorInfo::CFactorInfo()
{
	this->Length = 0;
	this->iPrimes = NULL;
	this->iRoots = NULL;
}
CFactorInfo::~CFactorInfo()
{
	Reset();
}
void CFactorInfo::Reset()
{
	if (this->iPrimes != NULL)
	{
		free(this->iPrimes);
		this->iPrimes = NULL;
	}
	if (this->iRoots != NULL)
	{
		free(this->iRoots);
		this->iRoots = NULL;
	}
	this->Length = 0;
}
void CFactorInfo::Clear(int aCount)
{
	int max = aCount * this->Length;
	if (this->iPrimes != NULL)
	{
		memset(this->iPrimes, 0, max * sizeof(int));
	}
	if (this->iRoots != NULL)
	{
		memset(this->iRoots, 0, max * sizeof(int));
	}
}
int CFactorInfo::Init(int aCount, int aLength)
{
	if (this->iPrimes != NULL || this->iRoots != NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","Init",0,0,"Some of the field is not NULL");
		Reset();
	}

	this->Length = aLength;

	int count = aCount * this->Length;

	this->iPrimes = (int*)Utils::AllocationMachine((void**)&this->iPrimes, count * sizeof(int));
	if (this->iPrimes == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","Init",0,0,"Cannot allocate memory for Primes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iPrimes, 0, count * sizeof(int));

	this->iRoots = (int*)Utils::AllocationMachine((void**)&this->iRoots, count * sizeof(int));
	if (this->iRoots == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","Init",0,0,"Cannot allocate memory for Roots.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iRoots, 0, count * sizeof(int));

	return ConstRC::Ok;
}
int CFactorInfo::ResizeCandidates(int aCountOld, int aCountNew)
{
	// delete primes and allocate new field
	if (this->iPrimes != NULL)
	{
		free(this->iPrimes);
		this->iPrimes = NULL;
	}

	int countAll = aCountNew * this->Length;

	this->iPrimes = (int*)Utils::AllocationMachine((void**)&this->iPrimes, countAll * sizeof(int));
	if (this->iPrimes == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","ResizeCandidates",0,0,"Cannot allocate memory for Primes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iPrimes, 0, countAll * sizeof(int));

	// new roots and copy all values (cols and rows)
	int * roots = NULL;
	roots = (int*)Utils::AllocationMachine((void**)&roots, countAll * sizeof(int));
	if (roots == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","ResizeCandidates",0,0,"Cannot allocate memory for roots.");
		return ConstRC::NotEnoughMemory;
	}
	memset(roots, 0, countAll * sizeof(int));

	if (this->iRoots != NULL)
	{
		int max = aCountOld * this->Length;
		for (int i = 0; i < max; i += this->Length)
		{
			roots[i] = this->iRoots[i];
		}

		free(this->iRoots);
		this->iRoots = NULL;
	}

	this->iRoots = roots;

	return ConstRC::Ok;
}
int CFactorInfo::ResizeFactorization(int aCount)
{
	int lengthNew = this->Length + ConstSPhase::FACTOR_RESIZE_COUNT;
	int countAll = lengthNew * aCount;

	int * primes = NULL;
	primes = (int*)Utils::AllocationMachine((void**)&primes, countAll * sizeof(int));
	if (primes == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","ResizeFactorization",0,0,"Cannot allocate memory for primes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(primes, 0, countAll * sizeof(int));

	int * roots = NULL;
	roots = (int*)Utils::AllocationMachine((void**)&roots, countAll * sizeof(int));
	if (roots == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorInfo","ResizeFactorization",0,0,"Cannot allocate memory for roots.");
		return ConstRC::NotEnoughMemory;
	}
	memset(roots, 0, countAll * sizeof(int));

	int indexOld = 0;
	int indexNew = 0;
	if (this->iPrimes != NULL)
	{
		for (int i = 0; i < aCount; i++, indexNew += ConstSPhase::FACTOR_RESIZE_COUNT)
		{
			for (int j = 0; j < this->Length; j++, indexNew++, indexOld++)
			{
				primes[indexNew] = this->iPrimes[indexOld];
			}
		}

		free(this->iPrimes);
		this->iPrimes = NULL;
	}

	indexOld = 0;
	indexNew = 0;
	if (this->iRoots != NULL)
	{
		for (int i = 0; i < aCount; i++, indexNew += ConstSPhase::FACTOR_RESIZE_COUNT)
		{
			for (int j = 0; j < this->Length; j++, indexNew++, indexOld++)
			{
				roots[indexNew] = this->iRoots[indexOld];
			}
		}

		free(this->iRoots);
		this->iRoots = NULL;
	}

	this->iPrimes = primes;
	this->iRoots = roots;
	this->Length = lengthNew;

	return ConstRC::Ok;
}
int CFactorInfo::SaveFactor(int aCount, int aIndex, int aPrime, int aRoot)
{
	int shift = this->Length * aIndex;
	int * count = this->iPrimes + shift;
	(*count)++;
	if (*count < this->Length)
	{
		shift += *count;
	}
	else
	{
		int pcount = *count;
		int rc = ResizeFactorization(aCount);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		shift = this->Length * aIndex + pcount;
	}

	this->iRoots[shift] = aRoot;
	this->iPrimes[shift] = aPrime;

	return ConstRC::Ok;
}
CFactorizationInfo::CFactorizationInfo()
{
	this->Relations = NULL;
	this->Cols = NULL;
	this->Rows = NULL;
	this->Count = 0;
	this->iCurrentCount = 0;
}
CFactorizationInfo::~CFactorizationInfo()
{
	Reset();
}
int * CFactorizationInfo::Get_Primes(int aIndex, int aInnerIndex)
{
	// WARNING without control
	return this->iFactors[aIndex].iPrimes + aInnerIndex * this->iFactors[aIndex].Length;
}
int * CFactorizationInfo::Get_Roots(int aIndex, int aInnerIndex)
{
	// WARNING without control
	return this->iFactors[aIndex].iRoots + aInnerIndex * this->iFactors[aIndex].Length;
}
void CFactorizationInfo::Clear()
{
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iFactors[i].Clear(this->Count);
	}
	if (this->Rows != NULL)
	{
		memset(this->Rows, 0, this->Count * sizeof(int));
	}
	if (this->Cols != NULL)
	{
		memset(this->Cols, 0, this->Count * sizeof(int));
	}
	this->iCurrentCount = 0;
}
void CFactorizationInfo::Reset()
{
	int i;
	if (this->Relations != NULL)
	{
		for (i = 0; i < this->Count; i++)
		{
			if (this->Relations[i] != NULL)
			{
				delete this->Relations[i];
				this->Relations[i] = NULL;
			}
		}
		free(this->Relations);
		this->Relations = NULL;
	}
	if (this->Cols != NULL)
	{
		free(this->Cols);
		this->Cols = NULL;
	}
	if (this->Rows != NULL)
	{
		free(this->Rows);
		this->Rows = NULL;
	}

	this->Count = 0;
	this->iCurrentCount = 0;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iFactors[i].Reset();
	}
}
int CFactorizationInfo::SaveOffset(int aOffset)
{
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iFactors[i].iRoots != NULL)
		{
			if (this->iCurrentCount >= this->Count)
			{
				int rc = ResizeCandidates();
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}

			this->iFactors[i].iRoots[this->iCurrentCount * this->iFactors[i].Length] = aOffset;
		}
		else
		{
			return ConstRC::GeneralError;
		}
	}

	this->iCurrentCount++;

	return ConstRC::Ok;
}
int CFactorizationInfo::FindOffset(int aOffset)
{
	// we used only first factors
	for (int i = 0, j = 0; i < this->iCurrentCount; i++, j += this->iFactors->Length)
	{
		if (this->iFactors->iRoots[j] == aOffset)
		{
			return i;
		}
	}

	return ConstRC::NotFound;
}
int CFactorizationInfo::Init(int aBlockSize, int aRelationMode)
{
	if (this->Rows != NULL || this->Cols != NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","Init",0,0,"Some of the field is not NULL");
		Reset();
	}

	this->iRelationMode = aRelationMode;

	if (aRelationMode > 1) // DoubleLargePrime
	{
		this->Count = aBlockSize / ConstSPhase::CANDIDATE_RATE_DOUBLE_LARGE;
	}
	else
	{
		this->Count = aBlockSize / ConstSPhase::CANDIDATE_RATE;
	}
	if (this->Count < ConstSPhase::MIN_CANDIDATE_COUNT)
	{
		this->Count = ConstSPhase::MIN_CANDIDATE_COUNT;
	}
	int length = ConstSPhase::FACTOR_AVERAGE_COUNT;

	int rc;
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = this->iFactors[i].Init(this->Count, length);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	this->Cols = (int*)Utils::AllocationMachine((void**)&this->Cols, this->Count * sizeof(int));
	if (this->Cols == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","Init",0,0,"Cannot allocate memory for Columns.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Cols, 0, this->Count * sizeof(int));

	this->Rows = (int*)Utils::AllocationMachine((void**)&this->Rows, this->Count * sizeof(int));
	if (this->Rows == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","Init",0,0,"Cannot allocate memory for Rows.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Rows, 0, this->Count * sizeof(int));

	this->Relations = (CNFSRelation**)Utils::AllocationMachine((void**)&this->Relations, this->Count * sizeof(CNFSRelation*));
	if (this->Relations == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","Init",0,0,"Cannot allocate memory for relation pointers.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Relations, 0, this->Count * sizeof(CNFSRelation*));

	for (int i = 0; i < this->Count; i++)
	{
		this->Relations[i] = new CNFSRelation(this->iRelationMode);
	}

	return ConstRC::Ok;
}
int CFactorizationInfo::ResizeCandidates()
{
	int i;
	int countNew = this->Count + ConstSPhase::CANDIDATE_RESIZE_COUNT;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iFactors[i].ResizeCandidates(this->Count, countNew);
	}

	// new rows and copy old values
	int * rows = NULL;
	rows = (int*)Utils::AllocationMachine((void**)&rows, countNew * sizeof(int));
	if (rows == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","ResizeCandidates",0,0,"Cannot allocate memory for rows.");
		return ConstRC::NotEnoughMemory;
	}
	memset(rows, 0, countNew * sizeof(int));

	if (this->Rows != NULL)
	{
		for (i = 0; i < this->Count; i++)
		{
			rows[i] = this->Rows[i];
		}

		free(this->Rows);
		this->Rows = NULL;
	}

	// new cols and copy old values
	int * cols = NULL;
	cols = (int*)Utils::AllocationMachine((void**)&cols, countNew * sizeof(int));
	if (cols == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","ResizeCandidates",0,0,"Cannot allocate memory for cols.");
		return ConstRC::NotEnoughMemory;
	}
	memset(cols, 0, countNew * sizeof(int));

	if (this->Cols != NULL)
	{
		for (i = 0; i < this->Count; i++)
		{
			cols[i] = this->Cols[i];
		}

		free(this->Cols);
		this->Cols = NULL;
	}

	CNFSRelation ** relations = NULL;
	relations = (CNFSRelation**)Utils::AllocationMachine((void**)&relations, countNew * sizeof(CNFSRelation*));
	if (relations == NULL)
	{
		CLog::LogError(WHEREARG,"CFactorizationInfo","ResizeCandidates",0,0,"Cannot allocate memory for relation pointers.");
		return ConstRC::NotEnoughMemory;
	}
	memset(relations, 0, countNew * sizeof(CNFSRelation*));

	if (this->Relations != NULL)
	{
		for (i = 0; i < this->Count; i++)
		{
			relations[i] = this->Relations[i];
		}

		free(this->Relations);
		this->Relations = NULL;

		for (; i < countNew; i++)
		{
			relations[i] = new CNFSRelation(this->iRelationMode);
		}
	}

	this->Cols = cols;
	this->Rows = rows;
	this->Relations = relations;
	this->Count = countNew;

	return ConstRC::Ok;
}
int CFactorizationInfo::ResizeFactorization(int aIndex)
{
	return this->iFactors[aIndex].ResizeFactorization(this->Count);
}


void CFactorizationInfo::ClearRelations()
{
	for (int i = 0; i < this->Count; i++)
	{
		this->Relations[i]->Reset();
	}
}
