#include "filtering_nfs_relation.h"

#include "base_parameters.h"
#include "integer_class.h"
#include "const_return_codes.h"
#include "const_filenames.h"
#include "const_sieving_phase.h"
#include "test_suite.h"
#include "log.h"


using namespace std;

CFNFSRelation::CFNFSRelation(main_sieving_type * aLargePrimeBounds, int aPartCount) : MSerializable(ConstFN::FILTERING_NFS_RELATION)
{
	this->A = 0;
	this->B = 0;
	this->Index = 0;
	this->iNumberOfRelations = 1;

	this->iPartsCount = aPartCount;
	this->iParts = new CFRelationPart[this->iPartsCount];

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].Main.Set_AssignedElements(CFNFSRelation::DEFAULT_DIVISORS_ALLOCATE);
		if (aLargePrimeBounds != NULL)
		{
			this->iParts[i].Set_LargePrimeBound(aLargePrimeBounds[i]);
		}
	}
}

CFNFSRelation::CFNFSRelation(const CFNFSRelation& aOperand) : 
	MSerializable(aOperand), A(aOperand.A), B(aOperand.B)
{
	this->Index = aOperand.Index;
	this->iPartsCount = aOperand.iPartsCount;
	this->iNumberOfRelations = aOperand.iNumberOfRelations;

	this->iParts = new CFRelationPart[this->iPartsCount];
	for (int i = 0; i < aOperand.iPartsCount; i++)
	{
		aOperand.iParts[i].Copy2(this->iParts[i]);
	}
}
CFNFSRelation::~CFNFSRelation()
{
	if (this->iParts != NULL)
	{
		// Delete invokes the destructor.
		delete[] this->iParts;
		this->iParts = NULL;
	}
}
void CFNFSRelation::PrintToScreen()
{	
	printf("================================================== \n");
	printf("#%ld\n", this->Index);
	printf("[a,b] = [%6d,%6d]\n", this->A, this->B);

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].PrintToScreen();
	}

    printf("================================================== \n");
}
int CFNFSRelation::Copy2(CFNFSRelation &aRelation)
{
	aRelation.Index = this->Index;
	aRelation.A = this->A;
	aRelation.B = this->B;
	aRelation.iNumberOfRelations = this->iNumberOfRelations;

	int rc;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].Copy2(aRelation.iParts[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int CFNFSRelation::Copy2(CNFSRelation &aRelation)
{
	aRelation.Index = this->Index;
	aRelation.A = this->A;
	aRelation.B = this->B;
	aRelation.Set_NumberOfRelations(this->iNumberOfRelations);
	aRelation.Type = 0;

	int rc;
	int i;
	CRelationPart * part = NULL;
	for (i = 0; i < this->iPartsCount; i++)
	{
		part = aRelation.Get_Parts(i);
		rc = this->iParts[i].Copy2(*part);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (part->Large.MaxElementIndex == -1)
		{
			aRelation.Type |= (CRelationPart::SMOOTH << (ARelation::PART_TYPE_SHIFT * i));
		}
		else if (part->Large.MaxElementIndex == 0)
		{
			aRelation.Type |= (CRelationPart::LARGE << (ARelation::PART_TYPE_SHIFT * i));
		}
		else if (part->Large.MaxElementIndex == 1)
		{
			aRelation.Type |= (CRelationPart::DOUBLE_LARGE << (ARelation::PART_TYPE_SHIFT * i));
		}
		else
		{
			aRelation.Type |= (CRelationPart::TRIPPLE_LARGE << (ARelation::PART_TYPE_SHIFT * i));
		}
	}

	return ConstRC::Ok;
}
int CFNFSRelation::CopyFrom(CNFSRelation &aRelation)
{
	this->Index = aRelation.Index;
	this->A = aRelation.A;
	this->B = aRelation.B;
	this->iNumberOfRelations = aRelation.Get_NumberOfRelations();

	int rc;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].CopyFrom(*aRelation.Get_Parts(i));
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
bool CFNFSRelation::Equals(const CFNFSRelation &aOperand) const
{
	for (int i = 0; i < this->iPartsCount; i++)
	{
		if (this->iParts[i].Equals(aOperand.iParts[i]) == false) return false;
	}

	if (this->A != aOperand.A) return false;
	if (this->B != aOperand.B) return false;
	if (this->iNumberOfRelations != aOperand.iNumberOfRelations) return false;
	
	return true;
}
int CFNFSRelation::Reset()
{
	this->A = 0;
	this->B = 0;
	this->Index = 0;
	this->iNumberOfRelations = 0;

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].Reset();
	}

	return ConstRC::Ok;
}

CFNFSRelation *CFNFSRelation::CopyItem() {

	return new CFNFSRelation(*this);
}

/*!
Should be called only by objects dynamically allocated by CopyItem().
*/
void CFNFSRelation::FreeItem() {

	delete (CFNFSRelation *)this;
}

int CFNFSRelation::CombineRelations(CFNFSRelation* aOperand1, CFNFSRelation* aOperand2, int aThetaPolyCount, Polynomial ** aThetaPolys, CFNFSRelation *& aResult)
{
	if (aOperand1->iParts != NULL && aOperand2->iParts != NULL && 
		aOperand1->iPartsCount == aOperand2->iPartsCount && 
		aThetaPolys != NULL && aThetaPolyCount == aOperand1->iPartsCount)
	{
		// we do not control if aResult is not NULL because we suppose that it could be but it does not matter... 
		aResult = new CFNFSRelation(*aOperand1);
		int rc;
		CFRelationPart * part1 = aResult->iParts;
		CFRelationPart * part2 = aOperand2->iParts;
		for (int i = 0; i < aThetaPolyCount; i++, part1++, part2++, aThetaPolys++)
		{
			rc = part1->CombineWithPart(*part2, *aThetaPolys);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		aResult->iNumberOfRelations += aOperand2->iNumberOfRelations;

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","CombineRelations",0,0,"Operands inconsistency.");				
		return ConstRC::GeneralError;
	}
}
int CFNFSRelation::Get_HammingWeight()
{
	int i, j;
	int weight = 0;
	int max;
	nfs_sieving_element * primeIdeals = NULL;
	for (i = 0; i < this->iPartsCount; i++)
	{
		max = this->iParts[i].Main.MaxElementIndex;
		primeIdeals = this->iParts[i].Main.Elements;
		for (j = 0; j <= max; j++, primeIdeals++)
		{
			if (primeIdeals->exponent)
			{
				weight++;
			}
		}
	}

	return weight;
}
int CFNFSRelation::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CFNFSRelation::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FILTERING_NFS_RELATION_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CFNFSRelation","Serialize",rc,0,"Could not write the main tag.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_NBR_OF_RELATIONS_NAME, this->iNumberOfRelations);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_A_NAME, this->A);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_B_NAME, this->B);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_INDEX_NAME, this->Index);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// now serialize relation parts
	rc = SerializeParts(aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","Serialize",rc,0,"Could not end the main tag.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CFNFSRelation::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CFNFSRelation::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FILTERING_NFS_RELATION_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);
	string auxStr;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_NBR_OF_RELATIONS_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize number of relations...", this->iNumberOfRelations, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_A_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize a...", this->A, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_B_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize b...", this->B, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INDEX_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize index...", this->Index, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PARTS_NAME))
		{
			rc = DeserializeParts(aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError)
	{
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}
int CFNFSRelation::SerializeParts(xmlTextWriterPtr &aWriter) const
{
	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_PARTS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CFNFSRelation","SerializeParts",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }

	rc = xmlTextWriterWriteFormatAttribute(aWriter, BAD_CAST XML_SERIAL_PARTS_COUNT_ATTR,"%d",this->iPartsCount);	
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","SerializeParts",rc,0,"Could not write the parts count.");
        return ConstRC::GeneralError;
    }

	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters param;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].Serialize(param, aWriter);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","SerializeParts",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;  
}

int CFNFSRelation::DeserializeParts(xmlTextReaderPtr &aReader)
{
	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","DeserializeParts",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST XML_SERIAL_PARTS_NAME) == 0)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","DeserializeParts",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int count;
	xmlChar* attr = xmlTextReaderGetAttribute(aReader, BAD_CAST XML_SERIAL_PARTS_COUNT_ATTR);  
	if (!attr)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","DeserializeParts",0,0,"Parts count not found, aborting");			
		return ConstRC::GeneralError;	
	}
	rc = CXMLHelper::ParseDecimalInt(attr, count);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","DeserializeParts",0,0,"Could not parse int");			
		xmlFree(attr);
		return ConstRC::GeneralError;	
	}
	xmlFree(attr);
	attr = NULL;

	if (count != this->iPartsCount)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","DeserializeParts",0,0,"Parts count incorrect, aborting");			
		return ConstRC::GeneralError;
	}

	if (this->iParts == NULL)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","DeserializeParts",0,0,"Parts don't exist, aborting");			
		return ConstRC::GeneralError;
	}

	CBaseParameters param;
	// int depth = xmlTextReaderDepth(aReader);

	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = CXMLHelper::MoveToNextNode(aReader);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
			{
				return rc;
			}
			return ConstRC::Ok;
		}

		rc = this->iParts[i].Deserialize(param, aReader);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
/*!
	This method is used to save a relation into file. It uses fprintf() for this purpose.
	This method has to be fast.
	File Template:
	==================================================
	Index
	iNumberOfRelations
	A
	B
	-- parts are next - template:
	Poly
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CFNFSRelation::PrintToFile(FILE* aFw)
{
	if (aFw == NULL)
	{
		return ConstRC::CouldNotWriteToNullStream;
	}

	fprintf(aFw, "==================================================\n");
	fprintf(aFw, "%lu\n", this->Index);     
	fprintf(aFw, "%d\n", this->iNumberOfRelations);
	fprintf(aFw, "%d\n", this->A);
	fprintf(aFw, "%d\n", this->B);

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].PrintToFile(aFw);
	}

	fflush(aFw);

	return ConstRC::Ok;
}


/*!
	This method is used to read a relation from file.
	File Template:
	==================================================
	Index
	iNumberOfRelations
	A
	B
	-- parts are next - template:
	Poly
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CFNFSRelation::ReadFromFile(FILE* aFr) 
{
	if (aFr == NULL)
	{
		CLog::LogError(WHEREARG,"CFNFSRelation","ReadFromFile",0,0,"Could not read elements from NULL stream");
		return ConstRC::CouldNotReadFromNullStream;
	}

	char data[CFNFSRelation::DIVISOR_READING_BUFFER_SIZE];

	// ==================================================
	if (fgets(data, CFNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}

	// Index
	if (fgets(data, CFNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	sscanf(data, "%ld", &this->Index);

	// iNumberOfRelations
	if (fgets(data, CFNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->iNumberOfRelations = atoi(data);

	// A
	if (fgets(data, CFNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->A = atoi(data);

	// B
	if (fgets(data, CFNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->B = atoi(data);

	int rc;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].ReadFromFile(aFr, data, CFNFSRelation::DIVISOR_READING_BUFFER_SIZE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
