#include "filtering_phase.h"

#include "matrix_helper.h"
#include "const_filtering_phase.h"
#include "const_parameter_names.h"
#include "const_hashtables.h"
#include "frequency_hashtable.h"
#include "duplicate_hashtable.h"
#include "merge_frequency_hashtable.h"
#include "minimum_spanning_tree.h"
#include "log.h"

#include "nfs_utils.h"

using namespace std;

CFilteringPhase::CFilteringPhase() : MSerializable(ConstFN::REL_PRO_PHASE)
{
	Set_AutoSerializable(ConstFPhase::AUTO_SERIALIZABLE);

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iSievingPoly[i] = NULL;
	}

	this->iFreqHashtable = NULL;
	this->iLargePrimeIdealsCount = NULL;
	this->iMergeLPILowerBounds = NULL;
	this->iPrimeIdealsCount = 0;
	this->iRelationsOverlap = 0;
	this->iRelationsCount = 0;
	this->iRelationsOverlapRatio = 0.0;

	Set_InnerState((int)EFPhaseNotStarted);
}

CFilteringPhase::~CFilteringPhase()
{
	Reset();
	DisposeFunctorField();
	DisposeMutexes();
	DisposeAcceptedReturnCodes();
}
int CFilteringPhase::Reset()
{
	Set_RelationMatrix(NULL);

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iSievingPoly[i] != NULL)
		{
			delete this->iSievingPoly[i];
			this->iSievingPoly[i] = NULL;
		}

		this->iFBs[i].Reset();
	}

	if (this->iFreqHashtable != NULL)
	{
		this->iFreqHashtable->DeleteHashtable();
		this->iFreqHashtable = NULL;
	}
	
	this->iRelationsOverlap = 0;
	this->iRelationsCount = 0;
	this->iPrimeIdealsCount = 0;
	this->iRelationsOverlapRatio = 0.0;
	
	if (this->iLargePrimeIdealsCount != NULL)
	{
		free(this->iLargePrimeIdealsCount);
		this->iLargePrimeIdealsCount = NULL;
	}

	if (this->iMergeLPILowerBounds != NULL)
	{
		free(this->iMergeLPILowerBounds);
		this->iMergeLPILowerBounds = NULL;
	}

	ARelProcessingPhase::Reset();

	Set_InnerState((int)EFPhaseNotStarted);

	return ConstRC::Ok;
}
int CFilteringPhase::FillFunctorField()
{
	if (Get_FunctorsInit() == false)
	{
		Set_FunctorsInit(true);
		this->iPhaseFunctors.reserve(ConstFPhase::INNER_STATE_COUNT);
		this->iPhaseFunctors.clear();
		// EFPhaseNotStarted
		this->iPhaseFunctors.push_back(NULL);
		// EFPhaseInitialization
		this->iPhaseFunctors.push_back(NULL);
		// EFPhaseJoiningFiles
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::JoinRelationFiles));
		// EFPhaseRemovingDuplicates
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::RemoveDuplicates));
		// EFPhaseRemovingSingletons 
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::RemoveSingletons));
		// EFPhaseRemovingRedundancy
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::RemoveRedundantRelations));
		// EFPhaseMergingRelations
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::MergeRelations));
		// EFPhaseCreatingRelMatrix
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::CreateRelationMatrix));
		// EFPhaseSavingResult
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CFilteringPhase>(this, &CFilteringPhase::SaveResult));
		// EFPhaseFinished
		this->iPhaseFunctors.push_back(NULL);
	}

	return ConstRC::Ok;
}

int CFilteringPhase::CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	int phaseNBR;
	if (aParameters->GetIntegerPhaseParameter(ConstPN::REL_PROCESSING_PHASE_NUMBER, phaseNBR, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CleanUp",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}

	string identifier;
	string directory = "";
	string file;
	string auxStr;

	if (aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CleanUp",0,0,"Didn't find serial identifier.");
		return ConstRC::GeneralError;
	}
	aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, directory, (int)EMain);

	switch (aType)
	{
		case ECleanAllIfNotError:
			if (aError)
			{
				break;
			}
		case ECleanAll:
		case ECleanAllExcCF:
		case ECleanAllExcCFRF:
			// All relation full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::ALL_RELATIONS_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::ALL_RELATIONS_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::ALL_RELATIONS_FILENAME, 
					ConstPN::ALL_RELATIONS_DIRECTORY, ConstPN::ALL_RELATIONS_COMPRESSION, directory,
					ConstFN::ALL_RELATIONS, identifier, phaseNBR, TXT_FILE_EXTENSION);
			}
			remove(file.c_str());

			// Temp file full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::TEMP_FILE_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::TEMP_FILE_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::TEMP_FILE_FILENAME, 
					ConstPN::TEMP_FILE_DIRECTORY, ConstPN::TEMP_FILE_COMPRESSION, directory,
					ConstFN::TEMP_FILE, identifier, phaseNBR, TXT_FILE_EXTENSION);
			}
			remove(file.c_str());

			// Quadratic characters full filenames
			for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				auxStr = NFSUtils::ConvertToString(i + 1);
				if (aParameters->ContainsPhaseParameter(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStr, phaseNBR))
				{
					aParameters->GetStringPhaseParameter(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStr, file, phaseNBR);
					MSerializable::AdaptPathSeparator(file);
				}
				else
				{
					aParameters->CreateFullFileName(file, ConstPN::QUAD_CHARS_FILENAME + auxStr, 
						ConstPN::QUAD_CHARS_DIRECTORY + auxStr, ConstPN::QUAD_CHARS_COMPRESSION + auxStr, directory,
						ConstFN::QUAD_CHARS + auxStr, identifier, phaseNBR);
				}
				remove(file.c_str());
			}
			break;
		default:
			break;
	}

	return ARelProcessingPhase::CleanUp(aParameters, phaseNBR, aType, aError);
}
int CFilteringPhase::InitParameters()
{
	Set_InnerState((int)EFPhaseInitialization);

	AFactorAlgParameters * param = Get_Parameters();
	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	// generally we don't know order of the square root phase
	int auxInt;
	if (param->GetIntegerPhaseParameter(ConstPN::REL_PROCESSING_PHASE_NUMBER, auxInt, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","InitParameters",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}
	Set_PhaseNbr(auxInt);

	int rc = ARelProcessingPhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// we support only relations with 2 parts
	if (ConstSPhase::ELEMENT_MASK_COUNT != CNFSRelation::DEFAULT_NFS_PARTS_COUNT)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","InitParameters",0,0,"This phase support only relations with two parts.");
		return ConstRC::BadArgument;
	}

	if (param->ContainsPhaseParameter(ConstPN::RELATIONS_OVERLAP, Get_PhaseNbr()))
	{
		param->GetIntegerPhaseParameter(ConstPN::RELATIONS_OVERLAP, this->iRelationsOverlap, Get_PhaseNbr());
	}

	param->GetDoublePhaseParameter(ConstPN::KEEP_RELATION_PERCENTAGE, ConstFPhase::KEEP_RELATION_PERCENTAGE, this->iRelationsOverlapRatio, Get_PhaseNbr());
		
	int i;
	string auxStr;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = NFSUtils::ConvertToString(i + 1);

		rc = param->GetPolynomialPhaseParameter(ConstPN::SIEVING_POLY + auxStr, this->iSievingPoly[i], (int)EMain);
		if (rc != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","InitParameters",0,0,"Didn't find the sieving polynomial.");
			return rc;
		}

		param->GetIntegerPhaseParameter(ConstPN::PARTIAL_FACTOR_MULTIPLIER + auxStr, ConstSPhase::PARTIAL_FACTOR_MULTIPLIER, auxInt, (int)EMain);
		this->iPartialFactorMultiplier[i] = auxInt;
	}

	if (this->iMergeLPILowerBounds == NULL)
	{
		this->iMergeLPILowerBounds = (main_sieving_type*) NFSUtils::AllocationMachine((void**) &this->iMergeLPILowerBounds, CNFSRelation::DEFAULT_NFS_PARTS_COUNT * sizeof(main_sieving_type));
		if (this->iMergeLPILowerBounds == NULL) 
		{
			CLog::LogError(WHEREARG, "CFilteringPhase", "InitParameters", 0, 0,  "Not enough memory to allocate data space - lower bounds.");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->iMergeLPILowerBounds, 0, CNFSRelation::DEFAULT_NFS_PARTS_COUNT * sizeof(main_sieving_type));
	}

	long bound = 0;
	long auxLong = 0;
	// try to find FB upper bound
	param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND, bound, bound, (int)EMain);
	
	// now we create all filenames which we need - some of then should be in parameters, 
	// others names have to be assembled
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = NFSUtils::ConvertToString(i + 1);
		// Quadratic characters full filename
		if (param->ContainsPhaseParameter(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStr, Get_PhaseNbr()))
		{
			param->GetStringPhaseParameter(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStr, this->iQuadCharsFullFileName[i], Get_PhaseNbr());
			MSerializable::AdaptPathSeparator(this->iQuadCharsFullFileName[i]);
		}
		else
		{
			param->CreateFullFileName(this->iQuadCharsFullFileName[i], ConstPN::QUAD_CHARS_FILENAME + auxStr, 
				ConstPN::QUAD_CHARS_DIRECTORY + auxStr, ConstPN::QUAD_CHARS_COMPRESSION + auxStr, Get_SerialMainDirectory(),
				ConstFN::QUAD_CHARS + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
		}
		// FB full filename
		if (param->ContainsPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, Get_PhaseNbr()))
		{
			param->GetStringPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, this->iFBs[i].FBFullFileName, Get_PhaseNbr());
			MSerializable::AdaptPathSeparator(this->iFBs[i].FBFullFileName);
		}
		else
		{
			param->CreateFullFileName(this->iFBs[i].FBFullFileName, ConstPN::FACTOR_BASE_FILENAME + auxStr, 
				ConstPN::FACTOR_BASE_DIRECTORY + auxStr, ConstPN::FACTOR_BASE_COMPRESSION + auxStr, Get_SerialMainDirectory(),
				ConstFN::FACTOR_BASE + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
		}

		// we try to find individual value for FB upper bound
		param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND + auxStr, bound, bound, (int)EMain);
		// inicialize variations info for later use
		this->iVariationsInfo[i].Prepare(bound, this->iPartialFactorMultiplier[i], ConstSPhase::LOGARITHM_BASE);

		// Merge lower bound for large prime ideals
		if (param->ContainsPhaseParameter(ConstPN::MERGE_LPI_LOWER_BOUND + auxStr, Get_PhaseNbr()))
		{
			param->GetLongPhaseParameter(ConstPN::MERGE_LPI_LOWER_BOUND + auxStr, auxLong, Get_PhaseNbr());
			this->iMergeLPILowerBounds[i] = auxLong;
		}
		else
		{
			// use FB upper bound
			this->iMergeLPILowerBounds[i] = bound;
		}
	}
	// All partial relation full filename
	if (param->ContainsPhaseParameter(ConstPN::ALL_RELATIONS_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::ALL_RELATIONS_FULL_FILENAME, this->iAllRelationsFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iAllRelationsFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iAllRelationsFullFileName, ConstPN::ALL_RELATIONS_FILENAME, 
			ConstPN::ALL_RELATIONS_DIRECTORY, ConstPN::ALL_RELATIONS_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::ALL_RELATIONS, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}
	// Temp file full filename
	if (param->ContainsPhaseParameter(ConstPN::TEMP_FILE_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::TEMP_FILE_FULL_FILENAME, this->iTempFileFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iTempFileFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iTempFileFullFileName, ConstPN::TEMP_FILE_FILENAME, 
			ConstPN::TEMP_FILE_DIRECTORY, ConstPN::TEMP_FILE_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::TEMP_FILE, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}

	if (this->iLargePrimeIdealsCount == NULL)
	{
		this->iLargePrimeIdealsCount = (relation_index_type*) NFSUtils::AllocationMachine((void**) &this->iLargePrimeIdealsCount, CNFSRelation::DEFAULT_NFS_PARTS_COUNT * sizeof(relation_index_type));
		if (this->iLargePrimeIdealsCount == NULL) 
		{
			CLog::LogError(WHEREARG, "CFilteringPhase", "InitParameters", 0, 0,  "Not enough memory to allocate data space - large prime ideals count.");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->iLargePrimeIdealsCount, 0, CNFSRelation::DEFAULT_NFS_PARTS_COUNT * sizeof(relation_index_type));
	}
	
	// should we deserialize or serialize?
	bool deserialize = false;
	param->GetBoolPhaseParameter(ConstPN::DESERIALIZE, false, deserialize, Get_PhaseNbr());
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());


	#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
		if (Get_InfoMode()) 
		{ 
		CLog::Log(10000, "0243211", "Relation Processing Phase is running ...\n\n"); 
		} 
	#endif 

	if (deserialize)
	{
		rc = Deserialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int CFilteringPhase::RunPhase(AFactorAlgParameters *aParameters)
{
	return APhase::RunPhase(aParameters);
}
/*
	This method iterates throught smooth relations and partial relations and add all relations to one file.
	Also the structure of relations is changed.
*/
int CFilteringPhase::JoinRelationFiles()
{
	Set_InnerState((int)EFPhaseJoiningFiles);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243202", "Joining files ...\n\n"); 
	} 
#endif 

	AFactorAlgParameters * param = Get_Parameters();
	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CRelProcessingPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	int rc;
	string relationsFilename;
	// Smooth relation full filename
	if (param->ContainsPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, relationsFilename, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(relationsFilename);
	}
	else
	{
		param->CreateFullFileName(relationsFilename, ConstPN::SMOOTH_RELATIONS_FILENAME, 
			ConstPN::SMOOTH_RELATIONS_DIRECTORY, ConstPN::SMOOTH_RELATIONS_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::SMOOTH_RELATIONS, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}

	FILE * relationsFILE = fopen(relationsFilename.c_str(), FILE_MODE_READ);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveSingletons",0,0,"Could not open file with relations.");
		return ConstRC::CouldNotOpenFile;
	} 

	FILE * allRelationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_WRITE);
	if (allRelationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveSingletons",0,0,"Could not open file with relations.");
		fclose(relationsFILE);
		relationsFILE = NULL;
		return ConstRC::CouldNotOpenFile;
	} 

	this->iRelationsCount = 0;
	CNFSRelation inCandidate(ETrippleLargePrimeV, CNFSRelation::DEFAULT_NFS_PARTS_COUNT);
	CFNFSRelation outCandidate(NULL);

	while (inCandidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
	{
		rc = outCandidate.CopyFrom(inCandidate);
		if (rc != ConstRC::Ok)
		{
			fclose(allRelationsFILE);
			allRelationsFILE = NULL;
			fclose(relationsFILE);
			relationsFILE = NULL;
			return rc;
		}

		outCandidate.Index = this->iRelationsCount;
		this->iRelationsCount++;
		rc = outCandidate.PrintToFile(allRelationsFILE);
		if (rc != ConstRC::Ok)
		{
			fclose(allRelationsFILE);
			allRelationsFILE = NULL;
			fclose(relationsFILE);
			relationsFILE = NULL;
			return rc;
		}
	}

	fclose(relationsFILE);
	relationsFILE = NULL;

	// Partial relation full filename
	if (param->ContainsPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, relationsFilename, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(relationsFilename);
	}
	else
	{
		param->CreateFullFileName(relationsFilename, ConstPN::PARTIAL_REL_FILENAME, 
			ConstPN::PARTIAL_REL_DIRECTORY, ConstPN::PARTIAL_REL_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::PARTIAL_REL, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}

	relationsFILE = fopen(relationsFilename.c_str(), FILE_MODE_READ);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveSingletons",0,0,"Could not open file with relations.");
		fclose(allRelationsFILE);
		allRelationsFILE = NULL;
		return ConstRC::CouldNotOpenFile;
	}

	while (inCandidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
	{
		rc = outCandidate.CopyFrom(inCandidate);
		if (rc != ConstRC::Ok)
		{
			fclose(allRelationsFILE);
			allRelationsFILE = NULL;
			fclose(relationsFILE);
			relationsFILE = NULL;
			return rc;
		}

		outCandidate.Index = this->iRelationsCount;
		this->iRelationsCount++;
		rc = outCandidate.PrintToFile(allRelationsFILE);
		if (rc != ConstRC::Ok)
		{
			fclose(allRelationsFILE);
			allRelationsFILE = NULL;
			fclose(relationsFILE);
			relationsFILE = NULL;
			return rc;
		}
	}

	fclose(allRelationsFILE);
	allRelationsFILE = NULL;
	fclose(relationsFILE);
	relationsFILE = NULL;

	return ConstRC::Ok;
}
/*
	This method remove all duplicate relations. 
	We use hashtable where the main hash is calculating according to Cavallar Stefania (Strategies in Filtering in the Number Field Sieve).
*/
int CFilteringPhase::RemoveDuplicates()
{
	Set_InnerState((int)EFPhaseRemovingDuplicates);

	#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
		if (Get_InfoMode()) 
		{ 
		CLog::Log(10000, "0243203", "Removing duplicates ... \n\n"); 
		} 
	#endif 

	CDuplicateHashtable table;
	int rc = table.SetupHashtable(this->iRelationsCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CFNFSRelation candidate(NULL);
	relation_index_type duplicateCount = 0;
	bool present = false;
	FILE * relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveDuplicates",0,0,"Could not open file with relations.");
		return ConstRC::CouldNotOpenFile;
	}

	FILE * tempFILE = fopen(this->iTempFileFullFileName.c_str(), FILE_MODE_WRITE);
	if (tempFILE == NULL) 
	{
		fclose(relationsFILE);
		relationsFILE = NULL;
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveDuplicates",0,0,"Could not open temp file.");
		return ConstRC::CouldNotOpenFile;
	}

	this->iRelationsCount = 0;
	// read all relations - save them to table and if someof them is duplicate then do not write it back to temp file
	while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
	{
		rc = table.AddToTable(candidate.A, candidate.B, present);
		if (rc != ConstRC::Ok)
		{
			fclose(tempFILE);
			tempFILE = NULL;
			fclose(relationsFILE);
			relationsFILE = NULL;
			return rc;
		}
		if (present)
		{
			// this relation is duplicate so we don't write it to the file
			duplicateCount++;
		}
		else
		{
			// unique relation, we write it to the file with new index
			candidate.Index = this->iRelationsCount;
			this->iRelationsCount++;
			rc = candidate.PrintToFile(tempFILE);
			if (rc != ConstRC::Ok)
			{
				fclose(tempFILE);
				tempFILE = NULL;
				fclose(relationsFILE);
				relationsFILE = NULL;
				return rc;
			}
		}
	}

	fclose(tempFILE);
	tempFILE = NULL;
	fclose(relationsFILE);
	relationsFILE = NULL;

	#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
		if (Get_InfoMode())
		{
			CLog::Log(10000, "0243080", "    Duplicates: %ld\n", duplicateCount);
			CLog::Log(10000, "0243081", "    Relation count: %ld\n\n", this->iRelationsCount);
		}
	#endif 

	return MSerializable::RenameFile(this->iTempFileFullFileName, this->iAllRelationsFullFileName);
}
/*
	This method is inner state method. 
	It just calls method RemoveSingletonsInner which deletes singletons.
*/
int CFilteringPhase::RemoveSingletons()
{
	Set_InnerState((int)EFPhaseRemovingSingletons);

	// serialize phase
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int rc = Serialize(bparam);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

	rc = RemoveSingletonsInner();
	if (rc != ConstRC::Ok) 
	{
		// delete frequency hashtable as soon as possible
		if (this->iFreqHashtable != NULL)
		{
			this->iFreqHashtable->DeleteHashtable();
			this->iFreqHashtable = NULL;
		}
		return rc;
	}

	if (this->iFreqHashtable != NULL)
	{
		// we calculate count of all prime ideals
		
		rc = this->iFreqHashtable->CountPrimeIdeals(this->iLargePrimeIdealsCount, 
			ConstSPhase::ELEMENT_MASK, this->iMergeLPILowerBounds, ConstSPhase::ELEMENT_MASK_COUNT, this->iPrimeIdealsCount);
		if (rc != ConstRC::Ok) 
		{
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			return rc;
		}
	}
	else
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveSingletons",0,0,"Frequency hashtable is NULL.");
		return ConstRC::InconsistentState;
	}

	return ConstRC::Ok;
}
int CFilteringPhase::RemoveSingletonsInner()
{

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243204", "Deleting singletons ...\n\n"); 
	} 
#endif 

	// we allocate a temporary array of bytes, with a single bit per each found relation.	
	// when we found singleton primeideal, we indicate its relation as singleton in this array,
	// so when we iterate throught relations we can quickly decide if it has singleton primeideal.
	// 0 - nonsingleton, 1 - singleton
	// extra care during iteration because of bit values!
	detection_type * detectionField = NULL;
	int detectionFieldSize = (this->iRelationsCount >> ConstHT::BITS_IN_BYTE_SHIFT) + 1;
	detectionField = (detection_type*) NFSUtils::AllocationMachine((void**) &detectionField, detectionFieldSize * sizeof(detection_type));
	if (detectionField == NULL) 
	{
		CLog::LogError(WHEREARG, "CFilteringPhase", "RemoveSingletons", detectionFieldSize, detectionFieldSize * sizeof(detection_type),  "Not enough memory to allocate data space.");
		return ConstRC::NotEnoughMemory;
	}
	memset(detectionField, ConstHT::NONSINGLETON, detectionFieldSize * sizeof(detection_type));
	
	int rc;
	long newSingletonCount = 0;
	CFNFSRelation candidate(NULL);
	if (this->iFreqHashtable == NULL)
	{
		this->iFreqHashtable = new CFrequencyHashtable();
		rc = this->iFreqHashtable->SetupHashtableProb(this->iRelationsCount); // we don't know exact number of prime ideals 
		if (rc != ConstRC::Ok)
		{
			free(detectionField);
			detectionField = NULL;
			return rc;
		}
	}
	else
	{
		// we already have filled hashtable so we will use it
		rc = this->iFreqHashtable->FindSingletonRelations(detectionField, newSingletonCount);
		if (rc != ConstRC::Ok)
		{
			free(detectionField);
			detectionField = NULL;
			return rc;
		}
	}

	FILE * relationsFILE = NULL;
	FILE * tempFILE = NULL;

	int i;
	int iter = 0;
	relation_index_type newRelationIndex = 0;
	// detection array - we use each bit - 0 nonsingleton, 1 singleton
	detection_type * detectionEntry = 0;
	detection_type detectionMask;

	// At first, we will read all relations and according to the detection field we decide 
	// if read relation is singleton. If not then we write it to temp file and its primeideals will be added to the hashtable.
	// The relation index is creating in every iteration newly.
	// Next, we will iterate throught primeideal hashtable and if we find new singletons we will repeate the process. 
	do
	{
		newSingletonCount = 0;
		newRelationIndex = 0;
		detectionEntry = detectionField;
		detectionMask = ConstHT::FIRST_DETECTION_BIT;

		// clear hashtable for new filling
		rc = this->iFreqHashtable->ClearHashtable();
		if (rc != ConstRC::Ok)
		{
			free(detectionField);
			detectionField = NULL;
			return rc;
		}

		// Now, we read the relations one-by-one and put corresponding
		// values into the hashtable if the relation is nonsingleton
		relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
		if (relationsFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","RemoveSingletons",0,0,"Could not open file with relations.");
			free(detectionField);
			detectionField = NULL;
			return ConstRC::CouldNotOpenFile;
		}
		
		tempFILE = fopen(this->iTempFileFullFileName.c_str(), FILE_MODE_WRITE);
		if (tempFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","RemoveSingletons",0,0,"Could not open temp file.");
			free(detectionField);
			detectionField = NULL;
			fclose(relationsFILE);
			relationsFILE = NULL;
			return ConstRC::CouldNotOpenFile;
		}

		while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok) 
		{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000 
			if (Get_InfoMode() && (candidate.Index & 0x40)) 
			{
				CLog::Log(1000, "0243082", "\r    IT: %2d | S: %8d | ReadRelation: %9d | %3.2f %% ", iter, newSingletonCount, candidate.Index, 100 * (float)candidate.Index/ this->iRelationsCount);	
			}
#endif 

			// check if this is not singleton
			if (((*detectionEntry) & detectionMask) == 0)
			{
				// it is not singleton - add primeideals into the hashtable
				for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
				{
					if (candidate.Get_Parts(i)->Main.MaxElementIndex >= 0) 
					{
						// add into the hashtable with new relation index
						rc = this->iFreqHashtable->AddToTable(candidate.Get_Parts(i), ConstSPhase::ELEMENT_MASK[i], newRelationIndex);
						if (rc != ConstRC::Ok)
						{
							free(detectionField);
							detectionField = NULL;
							fclose(relationsFILE);
							relationsFILE = NULL;
							fclose(tempFILE);
							tempFILE = NULL;
							return rc;
						}
					}
				}
				// change index to the new one
				candidate.Index = newRelationIndex;
				// increase index for new nonsingleton relation
				newRelationIndex++;
				// save relation to the temp file - new "nonsingleton" relations
				rc = candidate.PrintToFile(tempFILE);
				if (rc != ConstRC::Ok)
				{
					free(detectionField);
					detectionField = NULL;
					fclose(relationsFILE);
					relationsFILE = NULL;
					fclose(tempFILE);
					tempFILE = NULL;
					return rc;
				}
			}
			else
			{
				i = candidate.Index;
			}

			detectionMask = detectionMask >> 1; // shift to the next bit position
			if (detectionMask == 0)
			{
				// shift to the next detection entry
				detectionMask = ConstHT::FIRST_DETECTION_BIT;
				detectionEntry++;
			}
		}

		fclose(relationsFILE);
		relationsFILE = NULL;
		fclose(tempFILE);
		tempFILE = NULL;

		this->iRelationsCount = newRelationIndex;

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
		if (Get_InfoMode()) 
		{
			CLog::Log(1000, "0243083", "\r    IT: %2d | S: %8d | ReadRelation: %9d | %3.2f %% \n", iter, newSingletonCount, newRelationIndex, 100 * (float)newRelationIndex/ this->iRelationsCount);			
		}
#endif 

		// from temp file create new all relation file
		rc = MSerializable::RenameFile(this->iTempFileFullFileName,this->iAllRelationsFullFileName);
		if (rc != ConstRC::Ok)
		{
			free(detectionField);
			detectionField = NULL;
			return rc;
		}

		// clear detection field for new singleton search
		memset(detectionField, ConstHT::NONSINGLETON, detectionFieldSize * sizeof(detection_type));
		// find singletons
		rc = this->iFreqHashtable->FindSingletonRelations(detectionField, newSingletonCount);
		if (rc != ConstRC::Ok)
		{
			free(detectionField);
			detectionField = NULL;
			return rc;
		}

		iter++;
	}
	while (newSingletonCount > 0);
	
	if (NULL != detectionField)
		free(detectionField);

	return ConstRC::Ok;
}
int CFilteringPhase::RemoveRedundantRelations()
{
	Set_InnerState((int)EFPhaseRemovingRedundancy);

	// serialize phase
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int rc = Serialize(bparam);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

	long i, j;
	FILE * relationsFILE = NULL;
	FILE * tempFILE = NULL;
	CFNFSRelation candidate(NULL);

	if (this->iFreqHashtable == NULL)
	{
		// fill frequency hashtable

		this->iFreqHashtable = new CFrequencyHashtable();
		
		rc = this->iFreqHashtable->SetupHashtableProb(this->iRelationsCount); // we don't know exact number of prime ideals 
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		// Now, we read the relations one-by-one and put corresponding
		// values into the hashtable
		relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
		if (relationsFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","RemoveRedundantRelations",0,0,"Could not open file with relations.");
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			return ConstRC::CouldNotOpenFile;
		}

		while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok) 
		{
			// add primeideals into the hashtable
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				if (candidate.Get_Parts(i)->Main.MaxElementIndex >= 0) 
				{
					// add into the hashtable with real relation index
					rc = this->iFreqHashtable->AddToTable(candidate.Get_Parts(i), ConstSPhase::ELEMENT_MASK[i], candidate.Index);
					if (rc != ConstRC::Ok)
					{
						fclose(relationsFILE);
						relationsFILE = NULL;
						// delete frequency hashtable as soon as possible
						if (this->iFreqHashtable != NULL)
						{
							this->iFreqHashtable->DeleteHashtable();
							this->iFreqHashtable = NULL;
						}
						return rc;
					}
				}
			}			
		}

		fclose(relationsFILE);
		relationsFILE = NULL;
	}

	// create field of components - get max size = #prime ideals with frequency 2
	relation_index_type maxComponentAllocatedSize = 0;
	relation_index_type countPrimeIdeals = 0;
	rc = this->iFreqHashtable->CountPrimeIdealsWithFrequency(ConstFPhase::FREQUENCY_TWO, maxComponentAllocatedSize, countPrimeIdeals);
	if (rc != ConstRC::Ok)
	{
		// delete frequency hashtable as soon as possible
		if (this->iFreqHashtable != NULL)
		{
			this->iFreqHashtable->DeleteHashtable();
			this->iFreqHashtable = NULL;
		}
		return rc;
	}

	/*
		We need fixed min number that #relation + minOverlap > #prime ideals
		So we sum quadCharsCount * #parts and 1
	*/
	int minOverlap = 0;
	if (this->iRelationsOverlap == 0)
	{
		//set overlap according to ratio
		this->iRelationsOverlap = int(countPrimeIdeals * this->iRelationsOverlapRatio);
	}

	// We need to know how many quadratic characters we will use
	int quadCharsCount = 0;
	Get_Parameters()->GetIntegerPhaseParameter(ConstPN::QUAD_CHARS_COUNT, ConstRPPhase::QUAD_CHARS_COUNT, quadCharsCount, Get_PhaseNbr());
	minOverlap = quadCharsCount * CNFSRelation::DEFAULT_NFS_PARTS_COUNT + 1;
	relation_index_type gap = this->iRelationsCount - (countPrimeIdeals + minOverlap + this->iRelationsOverlap);

	if (gap <= 0)
	{
		// we can end now because we have not enough relations

		// delete frequency hashtable as soon as possible
		if (this->iFreqHashtable != NULL)
		{
			this->iFreqHashtable->DeleteHashtable();
			this->iFreqHashtable = NULL;
		}
		return ConstRC::Ok;
	}

	// array for all components 
	CCliqueComponentArray components;
	rc = components.Setup(maxComponentAllocatedSize);
	if (rc != ConstRC::Ok)
	{
		// delete frequency hashtable as soon as possible
		if (this->iFreqHashtable != NULL)
		{
			this->iFreqHashtable->DeleteHashtable();
			this->iFreqHashtable = NULL;
		}
		return rc;
	}

	// array for currenty used prime ideals with frequency FREQUENCY_TWO
	// we need only indexies in hashtable
	int primeIdeals2AllocatedSize = this->iFreqHashtable->MaxPrimeIdeals * CNFSRelation::DEFAULT_NFS_PARTS_COUNT;
	int primeIdeals2MaxSize = 0;
	fhashtable_entry ** primeIdeals2 = NULL;
	primeIdeals2 = (fhashtable_entry **)NFSUtils::AllocationMachine((void**) &primeIdeals2, primeIdeals2AllocatedSize * sizeof(fhashtable_entry*));
	if (primeIdeals2 == NULL)
	{
		CLog::LogError(WHEREARG, "CFilteringPhase", "RemoveRedundantRelations", primeIdeals2AllocatedSize, primeIdeals2AllocatedSize * sizeof(fhashtable_entry*),  "Not enough memory to allocate data space.");
		// delete frequency hashtable as soon as possible
		if (this->iFreqHashtable != NULL)
		{
			this->iFreqHashtable->DeleteHashtable();
			this->iFreqHashtable = NULL;
		}
		return ConstRC::NotEnoughMemory;
	}
	memset(primeIdeals2, 0, primeIdeals2AllocatedSize * sizeof(fhashtable_entry*));

	// weight of current read relation - calculated as
	//		MAX_FREQUENCY_WEIGHT (for relation) or MAX_FREQUENCY_WEIGHT/2 (for free relation)
	//		plus MAX_FREQUENCY_WEIGHT/2^(f-2) (for each big prime ideal), where f is frequency 
	int weightCurrentRelation;
	// index of the component in which lie prime ideals (can be nondefine (UNDEFINE_COMPONENT), one (then index) or more (MORE_COMPONENTS)
	int maxElementIndex;
	relation_index_type oldRelationCount;
	relation_index_type componentSize;
	relation_index_type removingRelationCount;
	relation_index_type gapHalf;
	relation_index_type componentIndex;
	relation_index_type primeidealComponentIndex;
	fhashtable_entry * primeidealEntry = NULL;
	nfs_sieving_element * elements = NULL;
	bool deleteRelation = false;
	prime_ideal_t primeideal;
	set<prime_ideal_t, prime_ideal_comp> removingPrimeIdeals;
	relation_index_type newRelationIndex = 0;
	while (gap > 0)
	{
		oldRelationCount = this->iRelationsCount;
		// clear components array
		components.Clear();
		// clear prime ideals defining which relation should be deleted (these which contain given prime ideals)
		removingPrimeIdeals.clear();
		// Now, we read the relations one-by-one
		relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
		if (relationsFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","RemoveRedundantRelations",0,0,"Could not open file with relations.");
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}
			return ConstRC::CouldNotOpenFile;
		}

		while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok) 
		{
			// prepare weight
			if (candidate.B != 0)
			{
				// it is not free relation
				weightCurrentRelation = ConstFPhase::MAX_FREQUENCY_WEIGHT;
			}
			else
			{
				// it is free relation
				weightCurrentRelation = (ConstFPhase::MAX_FREQUENCY_WEIGHT >> 1);
			}
			primeIdeals2MaxSize = 0;
			componentIndex = ConstFPhase::UNDEFINE_COMPONENT;
			// 1. calculate weight of relation (stepwise)
			// 2. find prime ideals with frequency 2
			//		a) if all prime ideals with f = 2 are not in component, 
			//				then create new component and add all prime ideals with f = 2 to this component
			//		b) if some of prime ideals with f = 2 are in one component, 
			//				then add all others prime ideals with f = 2 to this component
			//		c) if more prime ideals are in different components,
			//				then we have to connect these components
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				maxElementIndex = candidate.Get_Parts(i)->Main.MaxElementIndex;
				elements = candidate.Get_Parts(i)->Main.Elements;
				for (j = 0; j <= maxElementIndex; j++, elements++)
				{
					// take only prime ideals with odd exponent
					if (elements->exponent % 2)
					{
						// get pointer to prime ideal in hashtable
						rc = this->iFreqHashtable->GetPrimeFromHashtable(elements, ConstSPhase::ELEMENT_MASK[i], primeidealEntry);
						if (rc != ConstRC::Ok) // also means that element was not found
						{
							fclose(relationsFILE);
							relationsFILE = NULL;
							// delete frequency hashtable as soon as possible
							if (this->iFreqHashtable != NULL)
							{
								this->iFreqHashtable->DeleteHashtable();
								this->iFreqHashtable = NULL;
							}
							if (primeIdeals2 != NULL)
							{
								free(primeIdeals2);
								primeIdeals2 = NULL;
							}
							return rc;
						}
						
						if (primeidealEntry->frequency == 2)
						{
							// add to array
							primeIdeals2[primeIdeals2MaxSize] = primeidealEntry;
							primeIdeals2MaxSize++;
							// if we still need to decide what to do
							if (primeidealEntry->index != ConstHT::UNDEFINED_RELATION)
							{
								// get main component of current prime ideal
								rc = components.GetMainComponent(primeidealEntry->index, primeidealComponentIndex);
								if (rc != ConstRC::Ok)
								{
									fclose(relationsFILE);
									relationsFILE = NULL;
									// delete frequency hashtable as soon as possible
									if (this->iFreqHashtable != NULL)
									{
										this->iFreqHashtable->DeleteHashtable();
										this->iFreqHashtable = NULL;
									}
									if (primeIdeals2 != NULL)
									{
										free(primeIdeals2);
										primeIdeals2 = NULL;
									}
									return rc;
								}

								// we have prime ideal in some component
								if (componentIndex == ConstFPhase::UNDEFINE_COMPONENT || componentIndex == ConstFPhase::NEW_COMPONENT)
								{
									// use index of component with this prime ideal as main component index (case 2 - for now)
									componentIndex = primeidealComponentIndex;
								}
								else if (componentIndex != primeidealComponentIndex)
								{
									// we already have some index of component and this prime ideal has another (case 3 - forever)
									if (primeidealComponentIndex < componentIndex)
									{
										// set smaller index as the main component
										componentIndex = primeidealComponentIndex;
									}
								}
							}
							else if (componentIndex == ConstFPhase::UNDEFINE_COMPONENT)
							{
								// we don't know which component use so we make a new one (case 1 - for now)
								componentIndex = ConstFPhase::NEW_COMPONENT;
							}
						}
						else if (candidate.Get_Parts(i)->LargePrimeIndex >= j)
						{
							// increase weight only for large prime ideals
							weightCurrentRelation += (ConstFPhase::MAX_FREQUENCY_WEIGHT >> (primeidealEntry->frequency - ConstFPhase::FREQUENCY_TWO));
						}
					}
				}
			}

			// if primeIdeals2MaxSize > 0 then we know that we find some prime ideals with frequency 2
			if (primeIdeals2MaxSize > 0)
			{
				// componentIndex should be ConstFPhase::UNDEFINE_COMPONENT
				if (componentIndex == ConstFPhase::NEW_COMPONENT)
				{
					rc = components.AddNewComponent(weightCurrentRelation, primeIdeals2, primeIdeals2MaxSize);
				}
				else
				{
					// we have one or more prime ideals with frequency 2 and some of them are already in some component
					// so we need to connect these components
					rc = components.ConnectComponents(componentIndex, weightCurrentRelation, primeIdeals2, primeIdeals2MaxSize);
				}
				if (rc != ConstRC::Ok)
				{
					fclose(relationsFILE);
					relationsFILE = NULL;
					// delete frequency hashtable as soon as possible
					if (this->iFreqHashtable != NULL)
					{
						this->iFreqHashtable->DeleteHashtable();
						this->iFreqHashtable = NULL;
					}
					if (primeIdeals2 != NULL)
					{
						free(primeIdeals2);
						primeIdeals2 = NULL;
					}
					return rc;
				}
			}
		}

		fclose(relationsFILE);
		relationsFILE = NULL;

		//remove relations from the largest component
		// NEW - remove so many relations from large components that the gap is cca 50 %
		removingRelationCount = 0;
		componentSize = 0;
		gapHalf = relation_index_type(gap * 0.5);
		rc = components.FindLargestComponent(gap, componentSize, componentIndex);
		if (rc != ConstRC::Ok)
		{
			if (rc == ConstRC::NotFound)
			{
				// we cannot find any component with limited number of relations
				// so we break while cycle
				// new singletons can arise so we need to delete them
				rc = RemoveSingletonsInner();
				if (rc != ConstRC::Ok)
				{
					// delete frequency hashtable as soon as possible
					if (this->iFreqHashtable != NULL)
					{
						this->iFreqHashtable->DeleteHashtable();
						this->iFreqHashtable = NULL;
					}
					if (primeIdeals2 != NULL)
					{
						free(primeIdeals2);
						primeIdeals2 = NULL;
					}

					return rc;
				}
				break;
			}

			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}
			return rc;
		}
		removingRelationCount += componentSize;
		components.ResetWeight(componentIndex);

		while (true)
		{
			// find all prime ideals with frequency two contained in selected component
			// these prime ideals will be used to identify relations which should be deleted
			rc = this->iFreqHashtable->GetAllPrimeIdealsInComponent(componentIndex, &components, &removingPrimeIdeals);
			if (rc != ConstRC::Ok)
			{
				// delete frequency hashtable as soon as possible
				if (this->iFreqHashtable != NULL)
				{
					this->iFreqHashtable->DeleteHashtable();
					this->iFreqHashtable = NULL;
				}
				if (primeIdeals2 != NULL)
				{
					free(primeIdeals2);
					primeIdeals2 = NULL;
				}
			}

			// look for next large component
			rc = components.FindLargestComponent(gap, componentSize, componentIndex);
			if (rc != ConstRC::Ok)
			{
				if (rc == ConstRC::NotFound)
				{
					// we cannot find any component with limited number of relations
					// so we break while cycle
					break;
				}

				// delete frequency hashtable as soon as possible
				if (this->iFreqHashtable != NULL)
				{
					this->iFreqHashtable->DeleteHashtable();
					this->iFreqHashtable = NULL;
				}
				if (primeIdeals2 != NULL)
				{
					free(primeIdeals2);
					primeIdeals2 = NULL;
				}
				return rc;
			}
			removingRelationCount += componentSize;
			components.ResetWeight(componentIndex);

			//check how many relation we want to remove
			if (removingRelationCount > gapHalf)
			{
				break;
			}
		}
		
		// clear hashtable, we will fill it again
		rc = this->iFreqHashtable->ClearHashtable();
		if (rc != ConstRC::Ok)
		{
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}
		}

		// now we have component which can be deleted
		// go throught all relations and delete these which are in this component
		relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
		if (relationsFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","RemoveRedundantRelations",0,0,"Could not open file with relations.");
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}
			return ConstRC::CouldNotOpenFile;
		}

		tempFILE = fopen(this->iTempFileFullFileName.c_str(), FILE_MODE_WRITE);
		if (tempFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","RemoveRedundantRelations",0,0,"Could not open temp file.");
			fclose(relationsFILE);
			relationsFILE = NULL;
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}
			return ConstRC::CouldNotOpenFile;
		}

		// start new relation counting
		newRelationIndex = 0;
		while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok) 
		{
			deleteRelation = false;
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				maxElementIndex = candidate.Get_Parts(i)->Main.MaxElementIndex;
				elements = candidate.Get_Parts(i)->Main.Elements;
				for (j = 0; j <= maxElementIndex; j++, elements++)
				{
					// take only prime ideals with odd exponent
					if (elements->exponent % 2)
					{
						primeideal.prime = elements->prime;
						primeideal.rootmask = elements->root | ConstSPhase::ELEMENT_MASK[i];
						if (removingPrimeIdeals.find(primeideal) != removingPrimeIdeals.end())
						{
							// we find this prime ideal in removing prime ideals so this relation should be deleted
							deleteRelation = true;
							break;
						}
					}
				}

				if (deleteRelation)
				{
					// no need to continue
					break;
				}
			}

			if (deleteRelation == false)
			{
				// this relation is not in the component
				// fill hashtable
				// add into the hashtable with new relation index
				for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
				{
					rc = this->iFreqHashtable->AddToTable(candidate.Get_Parts(i), ConstSPhase::ELEMENT_MASK[i], newRelationIndex);
					if (rc != ConstRC::Ok)
					{
						fclose(relationsFILE);
						relationsFILE = NULL;
						fclose(tempFILE);
						tempFILE = NULL;
						// delete frequency hashtable as soon as possible
						if (this->iFreqHashtable != NULL)
						{
							this->iFreqHashtable->DeleteHashtable();
							this->iFreqHashtable = NULL;
						}
						if (primeIdeals2 != NULL)
						{
							free(primeIdeals2);
							primeIdeals2 = NULL;
						}
						return rc;
					}
				}
				// change index to the new one
				candidate.Index = newRelationIndex;
				// increase index for new relation
				newRelationIndex++;
				// save relation to the temp file
				rc = candidate.PrintToFile(tempFILE);
				if (rc != ConstRC::Ok)
				{
					fclose(relationsFILE);
					relationsFILE = NULL;
					fclose(tempFILE);
					tempFILE = NULL;
					// delete frequency hashtable as soon as possible
					if (this->iFreqHashtable != NULL)
					{
						this->iFreqHashtable->DeleteHashtable();
						this->iFreqHashtable = NULL;
					}
					if (primeIdeals2 != NULL)
					{
						free(primeIdeals2);
						primeIdeals2 = NULL;
					}
					return rc;
				}
			}
		}

		fclose(relationsFILE);
		relationsFILE = NULL;
		fclose(tempFILE);
		tempFILE = NULL;

		// from temp file create new all relation file
		rc = MSerializable::RenameFile(this->iTempFileFullFileName,this->iAllRelationsFullFileName);
		if (rc != ConstRC::Ok)
		{
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}

			return rc;
		}

		// set new relations count
		this->iRelationsCount = newRelationIndex;

		// new singletons can arise so we need to delete them
		rc = RemoveSingletonsInner();
		if (rc != ConstRC::Ok)
		{
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}

			return rc;
		}

		// recalculate gap 
		// firstly - get new count of prime ideals -- maxComponentAllocatedSize is corrupted
		rc = this->iFreqHashtable->CountPrimeIdealsWithFrequency(ConstFPhase::FREQUENCY_TWO, maxComponentAllocatedSize, countPrimeIdeals);
		if (rc != ConstRC::Ok)
		{
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			if (primeIdeals2 != NULL)
			{
				free(primeIdeals2);
				primeIdeals2 = NULL;
			}

			return rc;
		}
		gap = this->iRelationsCount - (countPrimeIdeals + minOverlap + this->iRelationsOverlap);
		if (this->iRelationsCount >= oldRelationCount)
		{
			//no relation was removed - break the cycle
			break;
		}
	}

	// clean up
	if (primeIdeals2 != NULL)
	{
		free(primeIdeals2);
		primeIdeals2 = NULL;
	}

	if (this->iFreqHashtable != NULL)
	{
		// we calculate count of all prime ideals
		rc = this->iFreqHashtable->CountPrimeIdeals(this->iLargePrimeIdealsCount, 
			ConstSPhase::ELEMENT_MASK, this->iMergeLPILowerBounds, ConstSPhase::ELEMENT_MASK_COUNT, this->iPrimeIdealsCount);
		if (rc != ConstRC::Ok) 
		{
			// delete frequency hashtable as soon as possible
			if (this->iFreqHashtable != NULL)
			{
				this->iFreqHashtable->DeleteHashtable();
				this->iFreqHashtable = NULL;
			}
			return rc;
		}
	}
	else
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","RemoveRedundantRelations",0,0,"Frequency hashtable is NULL.");
		return ConstRC::InconsistentState;
	}

	// delete frequency hashtable as soon as possible
	if (this->iFreqHashtable != NULL)
	{
		this->iFreqHashtable->DeleteHashtable();
		this->iFreqHashtable = NULL;
	}
	
	return ConstRC::Ok;
}
int CFilteringPhase::MergeRelations()
{
	Set_InnerState((int)EFPhaseMergingRelations);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243205", "Merging relations ...\n\n"); 
	} 
#endif 

	// serialize phase
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int i, j;
	int rc = Serialize(bparam);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

	// calculate relation size
	int relationSize = 0;
	rc = CalculateRelationSize(relationSize);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

	int memorySize;
	Get_Parameters()->GetIntegerPhaseParameter(ConstPN::MERGE_MEMORY_SIZE, ConstRPPhase::MERGE_MEMORY_SIZE, memorySize, Get_PhaseNbr());
	long relationsInMemory = ((INT64)memorySize)*ConstRPPhase::MEGABIT / relationSize;
	main_sieving_type * upperBound = NULL;
	upperBound = (main_sieving_type*) NFSUtils::AllocationMachine((void**) &upperBound, CNFSRelation::DEFAULT_NFS_PARTS_COUNT * sizeof(main_sieving_type));
	if (upperBound == NULL) 
	{
		CLog::LogError(WHEREARG, "CFilteringPhase", "MergeRelations", 0, 0,  "Not enough memory to allocate data space - upper bound.");
		return ConstRC::NotEnoughMemory;
	}
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		upperBound[i] = this->iVariationsInfo[i].MaxPartialFactor;
	}

	// create relation hashtable
	CRelationHashtable table;
	rc = table.SetupHashtable(this->iLargePrimeIdealsCount, relationsInMemory, this->iMergeLPILowerBounds, upperBound, CNFSRelation::DEFAULT_NFS_PARTS_COUNT);
	free(upperBound);
	upperBound = NULL;
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}
	rc = table.SetDirectoryAndIdentifier(this->Get_SerialMainDirectory(), this->Get_SerialIdentifier());
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

	CFNFSRelation candidate(NULL);
	CFRelationPart * relationPart = NULL;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		// set min bound for large prime ideals
		relationPart = candidate.Get_Parts(i);
		relationPart->Set_LargePrimeBound(this->iMergeLPILowerBounds[i]);
	}
	FILE * relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","MergeRelations",0,0,"Could not open file with relations.");
		return ConstRC::CouldNotOpenFile;
	}

	int maxPrimeIdealsInRelation = 0;

	// read all relations
	while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
	{
		// inicialization of hashtable store files
		rc = table.AddToFile(&candidate);
		if (rc != ConstRC::Ok) 
		{
			fclose(relationsFILE);
			relationsFILE = NULL;
			return rc;
		}

		// we need to know maximum number of prime ideals in relation (in one part)
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			if (candidate.Get_Parts(i)->Main.MaxElementIndex >= maxPrimeIdealsInRelation)
			{
				maxPrimeIdealsInRelation = candidate.Get_Parts(i)->Main.MaxElementIndex + 1;
			}
		}
	}

	fclose(relationsFILE);
	relationsFILE = NULL;

	// get max merge frequency - maximal possible merge
	int maxMergeFrequency;
	Get_Parameters()->GetIntegerPhaseParameter(ConstPN::MERGE_MAX_MERGE_FREQUENCY, ConstRPPhase::MERGE_MAX_MERGE_FREQUENCY, maxMergeFrequency, Get_PhaseNbr());
	int currentMergeFrequency = ConstRPPhase::MERGE_STARTING_FREQUENCY;

	string currentFilename;
	int partIndex;
	int weight;
	int maxPrimeIdeals;
	main_sieving_type minBound;
	main_sieving_type maxBound;
	relation_index_type relationIndex;
	nfs_sieving_element * primeIdeals = NULL;
	merge_fhashtable_entry * currPI = NULL;
	merge_fhashtable_entry * endPI = NULL;
	int frequency;
	bool doMerge = false;	
	CFNFSRelation * relation1;
	CFNFSRelation * relation2;
	// count average Hamming weight of relation
	int averageWeight = this->iPrimeIdealsCount / this->iRelationsCount;

	// variables for merging
	CMinimumSpanningTreeAlg stAlgorithm;
	rc = stAlgorithm.SetupAlg(maxMergeFrequency);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}
	// we need to save results edges - from spanning tree algorithm
	graph_edge_t * results = NULL;
	results = (graph_edge_t*) NFSUtils::AllocationMachine((void**) &results, (maxMergeFrequency - 1) * sizeof(graph_edge_t));
	if (results == NULL) 
	{
		CLog::LogError(WHEREARG, "CFilteringPhase", "MergeRelations", 0, 0,  "Not enough memory to allocate data space - result edges.");
		return ConstRC::NotEnoughMemory;
	}
	memset(results, 0, (maxMergeFrequency - 1) * sizeof(graph_edge_t));
	// we need to save weighted edges
	graph_wedge_t * edges = NULL;
	// number of edges of full graph on n vertices is (n/2) = n * (n-1) / 2
	edges = (graph_wedge_t*) NFSUtils::AllocationMachine((void**) &edges, maxMergeFrequency * (maxMergeFrequency - 1) * sizeof(graph_wedge_t)/2);
	if (edges == NULL) 
	{
		if (results != NULL) 
		{
			free(results);
			results = NULL;
		}
		CLog::LogError(WHEREARG, "CFilteringPhase", "MergeRelations", 0, 0,  "Not enough memory to allocate data space - edges.");
		return ConstRC::NotEnoughMemory;
	}
	memset(edges, 0, maxMergeFrequency * (maxMergeFrequency - 1) * sizeof(graph_wedge_t)/2);
	// we need to save combination of prime ideals from two relations
	nfs_sieving_element * elements = NULL;
	// number of result prime ideals in combination is maximally two times max number of prime ideals in relation
	elements = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**) &elements, maxPrimeIdealsInRelation * 2 * sizeof(nfs_sieving_element));
	if (elements == NULL) 
	{
		if (results != NULL) 
		{
			free(results);
			results = NULL;
		}
		if (edges != NULL) 
		{
			free(edges);
			edges = NULL;
		}
		CLog::LogError(WHEREARG, "CFilteringPhase", "MergeRelations", 0, 0,  "Not enough memory to allocate data space - elements.");
		return ConstRC::NotEnoughMemory;
	}
	memset(elements, 0, maxPrimeIdealsInRelation * 2 * sizeof(nfs_sieving_element));
	// we need to save new merged relations - these relations should not be deleted when destroying this array
	CFNFSRelation ** newRelations = NULL;
	// number of new merged relations is at most maxMergeFrequency - 1
	newRelations = (CFNFSRelation **) NFSUtils::AllocationMachine((void**) &newRelations, (maxMergeFrequency - 1) * sizeof(CFNFSRelation *));
	if (newRelations == NULL) 
	{
		if (results != NULL) 
		{
			free(results);
			results = NULL;
		}
		if (edges != NULL) 
		{
			free(edges);
			edges = NULL;
		}
		if (elements != NULL) 
		{
			free(elements);
			elements = NULL;
		}
		CLog::LogError(WHEREARG, "CFilteringPhase", "MergeRelations", 0, 0,  "Not enough memory to allocate data space - new relations.");
		return ConstRC::NotEnoughMemory;
	}
	memset(newRelations, 0, (maxMergeFrequency - 1) * sizeof(CFNFSRelation *));
	// we need to save indexes of relations to separate array
	relation_index_type * relationIndexes = NULL;
	// number of indexes is max merge frequency
	relationIndexes = (relation_index_type*) NFSUtils::AllocationMachine((void**) &relationIndexes, maxMergeFrequency * sizeof(relation_index_type));
	if (relationIndexes == NULL) 
	{
		if (results != NULL) 
		{
			free(results);
			results = NULL;
		}
		if (edges != NULL) 
		{
			free(edges);
			edges = NULL;
		}
		if (elements != NULL) 
		{
			free(elements);
			elements = NULL;
		}
		if (newRelations != NULL) 
		{
			free(newRelations);
			newRelations = NULL;
		}
		CLog::LogError(WHEREARG, "CFilteringPhase", "MergeRelations", 0, 0,  "Not enough memory to allocate data space - relation indexes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(relationIndexes, 0, maxMergeFrequency * sizeof(relation_index_type));
	// we need theta polynomials of number field sieve to combine relation (for each relation part)
	Polynomial * thetaPoly[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		thetaPoly[i] = NULL;
		if (this->iSievingPoly[i]->GetDegree() > 1)
		{
			rc = IntegralDomain::ComputeThetaPolynomial(this->iSievingPoly[i], thetaPoly[i]);
			if (rc != ConstRC::Ok)
			{
				if (results != NULL) 
				{
					free(results);
					results = NULL;
				}
				if (edges != NULL) 
				{
					free(edges);
					edges = NULL;
				}
				if (elements != NULL) 
				{
					free(elements);
					elements = NULL;
				}
				if (newRelations != NULL) 
				{
					free(newRelations);
					newRelations = NULL;
				}
				if (relationIndexes != NULL) 
				{
					free(relationIndexes);
					relationIndexes = NULL;
				}
				for (j = 0; j <= i; j++)
				{
					if (thetaPoly[j] != NULL)
					{
						delete thetaPoly[j];
						thetaPoly[j] = NULL;
					}
				}
				return rc;
			}
		}
	}

	// prepare frequency hashtable where all indexes of related relations are stored
	CMergeFrequencyHashtable freqTable;
	rc = freqTable.SetupHashtable(this->iRelationsCount, maxMergeFrequency);

	// do individual runs
	// do while any error occurred or we break the cycle
	while (rc == ConstRC::Ok)
	{
		// clear hashtables
		rc = table.ClearHashtable();
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}
		rc = freqTable.ClearHashtable();
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}

		// go to the next file - possibly we have already checked all files
		rc = table.IncreaseCurrentFileIndex(true);
		if (rc == ConstRC::All)
		{
			// set OK
			rc = ConstRC::Ok;
			// we checked all files - reset counter
			table.ResetCurrentFileIndex();
			// if we can do more runs we do
			if (currentMergeFrequency < maxMergeFrequency)
			{
				// increase run max allowed frequency
				currentMergeFrequency++;
				continue;
			}
			else
			{
				// break while cycle - we are finished
				break;
			}
		}
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}

		// get current filename with relations 
		rc = table.GetCurrentFilename(currentFilename);
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}

		// get bounds and relation part index - we need these for finding right prime ideals in relations
		rc = table.GetFromFilePartAndBounds(partIndex, minBound, maxBound);
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}

		// set min bound for large prime ideals
		relationPart = candidate.Get_Parts(partIndex);
		relationPart->Set_LargePrimeBound(minBound);

		// open store hashtable file with relations which have prime ideal of given size (according to current file index)
		relationsFILE = fopen(currentFilename.c_str(), FILE_MODE_READ);
		if (relationsFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","MergeRelations",0,0,"Could not open file with relations.");
			rc = ConstRC::CouldNotOpenFile;
			// break cycle with error
			break;
		}

		// read all relations from store file and put them to relation hashtable, also create frequency hashtable
		while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
		{
			relationIndex = -1;
			if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT)
			{
				// count number of prime ideals with odd exponent (in all relation parts)
				weight = candidate.Get_HammingWeight();
			}

			// get relation divisors (prime ideals) from correct relation part
			primeIdeals = relationPart->Main.Elements;
			maxPrimeIdeals = relationPart->Main.MaxElementIndex + 1;

			for (i = relationPart->LargePrimeIndex, primeIdeals += i; i < maxPrimeIdeals; i++, primeIdeals++)
			{
				if (primeIdeals->prime >= maxBound)
				{
					// too large prime ideals
					break;
				}

				if (primeIdeals->exponent % 2 == 0)
				{
					// we need only prime ideals with odd exponent
					continue;
				}

				// the first prime ideal in given interval 
				if (relationIndex == -1)
				{
					// add relation to relation hashtable (real hashtable in memory)
					// we use the prime of the smallest prime ideal in given interval
					rc = table.AddToTable(primeIdeals->prime, &candidate, relationIndex);
					if (rc != ConstRC::Ok) 
					{
						fclose(relationsFILE);
						relationsFILE = NULL;
						if (results != NULL) 
						{
							free(results);
							results = NULL;
						}
						if (edges != NULL) 
						{
							free(edges);
							edges = NULL;
						}
						if (elements != NULL) 
						{
							free(elements);
							elements = NULL;
						}
						if (newRelations != NULL) 
						{
							free(newRelations);
							newRelations = NULL;
						}
						if (relationIndexes != NULL) 
						{
							free(relationIndexes);
							relationIndexes = NULL;
						}
						for (j = 0; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++)
						{
							if (thetaPoly[j] != NULL)
							{
								delete thetaPoly[j];
								thetaPoly[j] = NULL;
							}
						}
						return rc;
					}
				}

				// calculate relation weight - two possibilities
				if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT == false)
				{
					// use number of relation from which is this relation combine
					// WARNING - we do not recognize free relations
					weight = candidate.Get_NumberOfRelations();
				}
				//else
				//{
				//	// use Hamming distance as a weight
				//	// already set
				//}

				// save prime ideal to the frequency hashtable - root fingerprint is not needed - we use only one relation part each time 
				rc = freqTable.AddToTable(primeIdeals->prime, primeIdeals->root, relationIndex, weight);
				if (rc != ConstRC::Ok) 
				{
					fclose(relationsFILE);
					relationsFILE = NULL;
					if (results != NULL) 
					{
						free(results);
						results = NULL;
					}
					if (edges != NULL) 
					{
						free(edges);
						edges = NULL;
					}
					if (elements != NULL) 
					{
						free(elements);
						elements = NULL;
					}
					if (newRelations != NULL) 
					{
						free(newRelations);
						newRelations = NULL;
					}
					if (relationIndexes != NULL) 
					{
						free(relationIndexes);
						relationIndexes = NULL;
					}
					for (j = 0; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++)
					{
						if (thetaPoly[j] != NULL)
						{
							delete thetaPoly[j];
							thetaPoly[j] = NULL;
						}
					}
					return rc;
				}
			}
		}

		fclose(relationsFILE);
		relationsFILE = NULL;

		// now we will look for set of relations which can be merged
		currPI = freqTable.Hashtable;
		endPI = currPI + freqTable.AllocatedSize;
		while (currPI < endPI)
		{
			frequency = currPI->frequency;
			// check if these prime ideals can be merged
			// TODO - check frequency == 1 => error
			if (1 < frequency && frequency <= currentMergeFrequency)
			{
				// check if this prime ideals have weight small enought
				doMerge = false;
				// relation weight - two possibilities
				if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT)
				{
					// use Hamming distance as a weight
					doMerge = (currPI->weight - frequency) < (currentMergeFrequency * averageWeight);
				}
				else
				{
					// use number of relation from which is this relation combine
					// WARNING - we do not recognize free relations
					// TODO
				}

				if (doMerge)
				{
					// at first, we need to know which relations will be merged
					rc = freqTable.GetRelationIndexes(currPI - freqTable.Hashtable, relationIndexes);
					if (rc != ConstRC::Ok) 
					{
						// break cycle with error
						break;
					}
					// we want to create full graph where vertices are relations
					// now we compute weights of edges between these relations as Hamming distance of union of their prime ideals (from all parts)
					rc = SetupEdges(relationIndexes, frequency, &table, elements, edges);
					if (rc != ConstRC::Ok) 
					{
						// break cycle with error
						break;
					}
					// we find minimal spanning tree of created graph
					rc = stAlgorithm.PrimAlgorithm2(edges, frequency * (frequency - 1) / 2, frequency, results);
					if (rc != ConstRC::Ok) 
					{
						// break cycle with error
						break;
					}

					// merge relations - we merge relations according to minimal spanning tree
					// we create new relations
					// if we merge #"frequency" relations then we get #"frequency" - 1 result relations 
					for (i = 0; i < frequency - 1; i++)
					{
						// get relation for merge
						relation1 = table.Get_Relation(relationIndexes[results[i].vertex1]);
						relation2 = table.Get_Relation(relationIndexes[results[i].vertex2]);
						// do merge
						rc = CFNFSRelation::CombineRelations(relation1, relation2, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, thetaPoly, newRelations[i]);
						if (rc != ConstRC::Ok)
						{
							// break cycle with error
							break;
						}
					}
					// check result of cycle
					if (rc != ConstRC::Ok) 
					{
						// break cycle with error
						break;
					}

					// clean up and update of used hashtables

					// update frequency hashtable - remove old large prime ideals
					for (i = 0; i < frequency; i++)
					{
						// get relation with prime ideals which should be removed from frequency hashtable (update)
						relation1 = table.Get_Relation(relationIndexes[i]);
						relationPart = relation1->Get_Parts(partIndex);
						relationPart->Set_LargePrimeBound(minBound);

						if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT)
						{
							// count number of prime ideals with odd exponent (in all relation parts)
							weight = relation1->Get_HammingWeight();
						}

						// get relation divisors (prime ideals) from correct relation part
						primeIdeals = relationPart->Main.Elements;
						maxPrimeIdeals = relationPart->Main.MaxElementIndex + 1;

						for (j = relationPart->LargePrimeIndex, primeIdeals += j; j < maxPrimeIdeals; j++, primeIdeals++)
						{
							if (primeIdeals->prime >= maxBound)
							{
								// too large prime ideals
								break;
							}

							if (primeIdeals->exponent % 2 == 0)
							{
								// we need only prime ideals with odd exponent
								continue;
							}

							// calculate relation weight - two possibilities
							if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT == false)
							{
								// use number of relation from which is this relation combine
								// WARNING - we do not recognize free relations
								weight = relation1->Get_NumberOfRelations();
							}
							//else
							//{
							//	// use Hamming distance as a weight
							//	// already set
							//}

							// remove prime ideal from the frequency hashtable - root fingerprint is not needed - we use only one relation part each time 
							rc = freqTable.RemoveFromTable(primeIdeals->prime, primeIdeals->root, relationIndexes[i], weight);
							if (rc != ConstRC::Ok)
							{
								// break cycle with error
								break;
							}
						}

						// check result of cycle
						if (rc != ConstRC::Ok)
						{
							// break cycle with error
							break;
						}
					}
					// check result of cycle
					if (rc != ConstRC::Ok) 
					{
						// break cycle with error
						break;
					}

					// save new relations to the relation hashtable
					for (i = 0; i < frequency - 1; i++)
					{
						// delete old relation
						relation1 = table.Get_Relation(relationIndexes[i]);
						delete (relation1);
						// save new relation
						table.Set_Relation(relationIndexes[i], newRelations[i]);
					}
					// last old relation in hashtable just delete
					relation1 = table.Get_Relation(relationIndexes[frequency - 1]);
					delete (relation1);
					// save NULL to the old relation position
					table.Set_Relation(relationIndexes[frequency - 1], NULL);

					// update frequency hashtable - add prime ideals from new relations to frequency hashtable
					for (i = 0; i < frequency - 1; i++)
					{
						// get new relation
						relation1 = newRelations[i];
						relationPart = relation1->Get_Parts(partIndex);
						relationPart->Set_LargePrimeBound(minBound);

						if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT)
						{
							// count number of prime ideals with odd exponent (in all relation parts)
							weight = relation1->Get_HammingWeight();
						}

						// get relation divisors (prime ideals) from correct relation part
						primeIdeals = relationPart->Main.Elements;
						maxPrimeIdeals = relationPart->Main.MaxElementIndex + 1;

						for (j = relationPart->LargePrimeIndex, primeIdeals += j; j < maxPrimeIdeals; j++, primeIdeals++)
						{
							if (primeIdeals->prime >= maxBound)
							{
								// too large prime ideals
								break;
							}

							if (primeIdeals->exponent % 2 == 0)
							{
								// we need only prime ideals with odd exponent
								continue;
							}

							// calculate relation weight - two possibilities
							if (ConstRPPhase::MERGE_USE_RELATION_HAMMING_WEIGHT == false)
							{
								// use number of relation from which is this relation combine
								// WARNING - we do not recognize free relations
								weight = relation1->Get_NumberOfRelations();
							}
							//else
							//{
							//	// use Hamming distance as a weight
							//	// already set
							//}

							// add prime ideal to the frequency hashtable - root fingerprint is not needed - we use only one relation part each time 
							rc = freqTable.AddToTable(primeIdeals->prime, primeIdeals->root, relationIndexes[i], weight);
							if (rc != ConstRC::Ok)
							{
								// break cycle with error
								break;
							}
						}

						// check result of cycle
						if (rc != ConstRC::Ok)
						{
							// break cycle with error
							break;
						}
					}
					// check result of cycle
					if (rc != ConstRC::Ok) 
					{
						// break cycle with error
						break;
					}
				}
			}
			currPI++;
		}

		// check if while cycle ended with error
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}

		// we are done with this store file for this run so we save all relations from memory to hashtable files
		rc = table.StoreCurrentHashtable();
		if (rc != ConstRC::Ok) 
		{
			// break cycle with error
			break;
		}
	}

	if (results != NULL) 
	{
		free(results);
		results = NULL;
	}
	if (edges != NULL) 
	{
		free(edges);
		edges = NULL;
	}
	if (elements != NULL) 
	{
		free(elements);
		elements = NULL;
	}
	if (newRelations != NULL) 
	{
		free(newRelations);
		newRelations = NULL;
	}
	if (relationIndexes != NULL) 
	{
		free(relationIndexes);
		relationIndexes = NULL;
	}
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (thetaPoly[i] != NULL)
		{
			delete thetaPoly[i];
			thetaPoly[i] = NULL;
		}
	}

	if (rc == ConstRC::Ok)
	{
		// join all hastable files to result one file = all relation file
		// recreate index of all relation, new recounting

		this->iRelationsCount = 0;

		FILE * tempFILE = fopen(this->iTempFileFullFileName.c_str(), FILE_MODE_WRITE);
		if (tempFILE == NULL) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","MergeRelations",0,0,"Could not open temp file.");
			return ConstRC::CouldNotOpenFile;
		}

		// at first - use processed relations
		table.Get_ProcessedFileName(currentFilename);
		relationsFILE = fopen(currentFilename.c_str(), FILE_MODE_READ);

		while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
		{
			candidate.Index = this->iRelationsCount;
			this->iRelationsCount++;
			rc = candidate.PrintToFile(tempFILE);
			if (rc != ConstRC::Ok)
			{
				fclose(tempFILE);
				tempFILE = NULL;
				fclose(relationsFILE);
				relationsFILE = NULL;
				return rc;
			}
		}

		fclose(relationsFILE);
		relationsFILE = NULL;

		table.ResetCurrentFileIndex();
		while ((rc = table.IncreaseCurrentFileIndex()) == ConstRC::Ok)
		{
			rc = table.GetCurrentFilename(currentFilename);
			if (rc != ConstRC::Ok)
			{
				fclose(tempFILE);
				tempFILE = NULL;
				fclose(relationsFILE);
				relationsFILE = NULL;
				return rc;
			}
			relationsFILE = fopen(currentFilename.c_str(), FILE_MODE_READ);

			while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
			{
				candidate.Index = this->iRelationsCount;
				this->iRelationsCount++;
				rc = candidate.PrintToFile(tempFILE);
				if (rc != ConstRC::Ok)
				{
					fclose(tempFILE);
					tempFILE = NULL;
					fclose(relationsFILE);
					relationsFILE = NULL;
					return rc;
				}
			}

			fclose(relationsFILE);
			relationsFILE = NULL;
		}

		fclose(tempFILE);
		tempFILE = NULL;

		if (rc == ConstRC::All)
		{
			// rename temp => all relations
			rc = MSerializable::RenameFile(this->iTempFileFullFileName, this->iAllRelationsFullFileName);
		}
	}

	return rc;
}
int CFilteringPhase::CalculateRelationSize(int & aRelationSize)
{
	AFactorAlgParameters * param = Get_Parameters();
	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CalculateRelationSize",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}
	mpz_t factorNumber;
	mpz_init(factorNumber);
	int rc = param->GetMpzPhaseParameter(ConstPN::NUMBER_N, factorNumber, (int)EMain);
	if (rc != ConstRC::Ok)
	{
		mpz_clear(factorNumber);
		CLog::LogError(WHEREARG,"CFilteringPhase","CalculateRelationSize",0,rc,"Didn't find N.");
		return rc;
	}

	int primeIdealsSize = sizeof(nfs_sieving_element) * (this->iPrimeIdealsCount / this->iRelationsCount + 1);
	int polynomialSize = sizeof(Polynomial) + ((*this->iSievingPoly)->GetDegree() + 1) * mpz_sizeinbase(factorNumber, 2);
	aRelationSize = sizeof(CFNFSRelation) + CNFSRelation::DEFAULT_NFS_PARTS_COUNT * (sizeof(CFRelationPart) + polynomialSize) + primeIdealsSize;
	mpz_clear(factorNumber);

	return ConstRC::Ok;
}
int CFilteringPhase::SetupEdges(relation_index_type * aRelationIndexes, int aRelationIndexesCount, 
								CRelationHashtable * aTable, nfs_sieving_element * aElements, graph_wedge_t * aEdges)
{
	int rc;
	int weight;
	long maxIndex;
	CFNFSRelation * relation1;
	CFNFSRelation * relation2;
	int i,j,k,l;
	for (i = 0; i < aRelationIndexesCount; i++)
	{
		for (j = i + 1; j < aRelationIndexesCount; j++, aEdges++)
		{
			weight = 0;
			relation1 = aTable->Get_Relation(aRelationIndexes[i]);
			relation2 = aTable->Get_Relation(aRelationIndexes[j]);
			for (k = 0; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
			{
				rc = NFSUtils::MergeTwoSortedArrays(aElements, maxIndex, 
					relation1->Get_Parts(k)->Main.Elements, relation1->Get_Parts(k)->Main.MaxElementIndex,
					relation2->Get_Parts(k)->Main.Elements, relation2->Get_Parts(k)->Main.MaxElementIndex);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				// count Hamming distance
				for (l = 0; l <= maxIndex; l++)
				{
					if (aElements[l].exponent % 2)
					{
						weight++;
					}
				}
			}
			aEdges->vertex1 = i;
			aEdges->vertex2 = j;
			aEdges->weight = weight;
		}
	}

	return ConstRC::Ok;
}
int CFilteringPhase::CreateRelationMatrix()
{
	Set_InnerState((int)EFPhaseCreatingRelMatrix);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243206", "Converting to bit matrix representation ...\n\n"); 
	} 
#endif 

	// serialize phase
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int rc = Serialize(bparam);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243207", "Deserializing FB Infos ...\n\n"); 
	} 
#endif 

	CBaseParameters paramFB;
	// read factor base infos
	int i, j, k;
	for (i = 0; i < ConstSPhase::ELEMENT_MASK_COUNT; i++)
	{
		paramFB.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iFBs[i].FBFullFileName);
		rc = this->iFBs[i].DeserializeFBInfo(paramFB);
		if (rc != ConstRC::Ok) 
		{
			return rc;
		}
	}

	// *** computing legendre symbols ***
	// we need to be sure that any relation is not divisible by these primeideals
	// we need go throught all relations and create primeideal hashtable
	// than we can find these primeideals
	// we also need to know how many primeideals will be in matrix and their indexes of columns in matrix

	// prepare hashtable - we know that relations count \approx prime ideals count
	CFrequencyHashtable table;
	rc = table.SetupHashtableEx(this->iRelationsCount);
	if (rc != ConstRC::Ok) 
	{
		return rc;
	}

	CFNFSRelation candidate(NULL);	
	CFRelationPart * relationPart = NULL;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		// set min bound for large prime ideals
		relationPart = candidate.Get_Parts(i);
		relationPart->Set_LargePrimeBound(this->iVariationsInfo[i].LargePrimeBound);
	}

	CNFSRelation outputRelation(CNFSRelation::MODE_CYCLE_CONSTRUCTION);
	relation_index_type currCount = 0;
	FILE * relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"Could not open file with relations.");
		return ConstRC::CouldNotOpenFile;
	}

	FILE * relationsResultFILE = fopen(this->iRelationsResultFullFileName.c_str(), FILE_MODE_WRITE);
	if (relationsResultFILE == NULL) 
	{
		fclose(relationsFILE);
		relationsFILE = NULL;
		CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"Could not open file for writting result relations.");
		return ConstRC::CouldNotOpenFile;
	}

	// read all relations - we will know the indexes and count of prime ideals - all of them
	// convert the relations to NFS relation and save them to final file - relation result file
	while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok)
	{
		// convert
		rc = candidate.Copy2(outputRelation);
		if (rc != ConstRC::Ok)
		{
			fclose(relationsFILE);
			relationsFILE = NULL;
			fclose(relationsResultFILE);
			relationsResultFILE = NULL;
			return rc;
		}

		// write to result file
		rc = outputRelation.PrintToFile(relationsResultFILE);
		if (rc != ConstRC::Ok)
		{
			fclose(relationsFILE);
			relationsFILE = NULL;
			fclose(relationsResultFILE);
			relationsResultFILE = NULL;
			return rc;
		}

		for (i = 0; i < ConstSPhase::ELEMENT_MASK_COUNT; i++)
		{
			rc = table.AddToTableCM(candidate.Get_Parts(i), ConstSPhase::ELEMENT_MASK[i], currCount);
			if (rc != ConstRC::Ok)
			{
				fclose(relationsFILE);
				relationsFILE = NULL;
				fclose(relationsResultFILE);
				relationsResultFILE = NULL;
				return rc;
			}
		}
	}

	fclose(relationsFILE);
	relationsFILE = NULL;
	fclose(relationsResultFILE);
	relationsResultFILE = NULL;

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243208", "Filling Quadratic characters ...\n\n"); 
	} 
#endif 

	// We need to know how many quadratic characters we will use
	int quadCharsCount = 0;
	Get_Parameters()->GetIntegerPhaseParameter(ConstPN::QUAD_CHARS_COUNT, ConstRPPhase::QUAD_CHARS_COUNT, quadCharsCount, Get_PhaseNbr());

	// now we can generate prime ideals for legendre symbols
	Polynomial * thetaPoly = NULL;
	prime_ideal_for_legendre * quadChars[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	for (i = 0, j = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iSievingPoly[i]->GetDegree() > 1)
		{
			rc = IntegralDomain::ComputeThetaPolynomial(this->iSievingPoly[i], thetaPoly);
			if (rc != ConstRC::Ok)
			{
				for (k = 0; k < j; k++)
				{
					free(quadChars[k]);
					quadChars[k] = NULL;
				}
				return rc;
			}

			quadChars[j] = NULL;
			rc = PrepareQuadraticCharacters(quadChars[j], quadCharsCount, i, thetaPoly, this->iFBs[i].UpperFBBound, table);
			if (rc != ConstRC::Ok) 
			{
				CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"Problem in filling of the quadratic characters");
				delete thetaPoly;
				thetaPoly = NULL;
				for (k = 0; k <= j; k++)
				{
					if (quadChars[k] != NULL)
					{
						free(quadChars[k]);
						quadChars[k] = NULL;
					}
				}
				return rc;
			}

			delete thetaPoly;
			thetaPoly = NULL;
			j++;
		}
	}

#ifdef DEBUG_REL_PRO_PHASE_SAVE_QUAD_CHARACTERS

	// now we serialize the first quadratic characters - for square root phase (debug)
	CBaseParameters paramChars;
	string auxStrCH;
	for (i = 0, j = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iSievingPoly[i]->GetDegree() > 1)
		{
			auxStrCH = NFSUtils::ConvertToString(i + 1);
			paramChars.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iQuadCharsFullFileName[i]);

			rc = ARelProcessingPhase::SerializeQuadCharacters(paramChars, quadChars[j], quadCharsCount);
			if (rc != ConstRC::Ok) 
			{
				for (k = 0; k < nbrOfAlgParts; k++)
				{
					if (quadChars[k] != NULL)
					{
						free(quadChars[k]);
						quadChars[k] = NULL;
					}
				}
				return rc;
			}
			// we also save this file name to NFSParameters for other phases
			Get_Parameters()->SetParameterForOthers(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStrCH, this->iQuadCharsFullFileName[i], Get_PhaseNbr());
			j++;
		}
	}

#endif

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(10000, "0243209", "Filling Quadratic characters finished\n\n"); 
	} 
#endif 

	long maxRow = this->iRelationsCount - 1;
	// maxColumn ... -1 because of max index, currCount = number of different prime ideals - from all parts
	long maxColumn = currCount -1; 
	int nbrOfAlgParts = 0;
	bool columnOfOnes = false;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iSievingPoly[i]->GetDegree() > 1)
		{
			nbrOfAlgParts++;
		}
		if (mpz_cmp_ui(*this->iSievingPoly[i]->Leading2(), 1) != 0)
		{
			columnOfOnes = true;
		}
	}
	// maxColumn += nbrOfAlgParts * count of characters 
	maxColumn += nbrOfAlgParts * quadCharsCount;
	// maxColumn += 1 as column of 1's 
	if (columnOfOnes)
	{
		maxColumn++;
	}

	// offset setup
	long legendreOffsets[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	long offset = currCount;
	for (i = 0; i < nbrOfAlgParts; i++)
	{
		legendreOffsets[i] = offset;
		offset += quadCharsCount;
	}

	if (maxRow < maxColumn) 
	{ 
		CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"Columns are more than rows.");
		
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
		if (Get_InfoMode()) 
		{
			CLog::Log(1000, "0243084", "    Maximal row is %d, maximal column is %d.\n", maxRow, maxColumn);
		}
#endif

		return ConstRC::NotEnoughSmooths; 
	}
	
	// Now, we will read entries from relation file and
	// put them one-by-one to the (sparse bit) matrix.
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000	
	if (Get_InfoMode())
	{
		CLog::Log(1000, "0243085", "    Matrix: row = %ld, column %ld", maxRow + 1, maxColumn + 1);
	}
#endif

	AbstractMatrix * matrix = CMatrixHelper::CreateMatrix(this->iRelationMatrixType);
	if (matrix == NULL)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"Cannot create relation matrix.");
		for (j = 0; j < nbrOfAlgParts; j++)
		{
			if (quadChars[j] != NULL)
			{
				free(quadChars[j]);
				quadChars[j] = NULL;
			}
		}
		return ConstRC::GeneralError;
	}
	matrix->SetMaximalColumnIndex(maxColumn);
	matrix->SetMaximalRowIndex(maxRow);

	relationsFILE = fopen(this->iAllRelationsFullFileName.c_str(), FILE_MODE_READ);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"Could not open file with smooth relations.");
		for (j = 0; j < nbrOfAlgParts; j++)
		{
			if (quadChars[j] != NULL)
			{
				free(quadChars[j]);
				quadChars[j] = NULL;
			}
		}
		delete matrix;
		matrix = NULL;
		return ConstRC::CouldNotOpenFile;
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "0243086", "    Converting ...");
	}
#endif 

	bool present;
	long count = 0;
	long max;
	long columnIndex;
	nfs_sieving_element * element;
	int iter = 0;
	long row = 0;
	relation_index_type index;
	while (candidate.ReadFromFile(relationsFILE) == ConstRC::Ok) 
	{
		// We will walk the elements one-by-one and put 1's in corresponding locations.
		// At first, the first algebraic divisors are iterated.
		iter++;
		
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
		if (Get_InfoMode() && !(iter % 10))
		{
			CLog::Log(1000, "0243087", "\r    Converting ... %10d       / %10ld      ", iter, this->iRelationsCount);
		}
#endif
		
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			max = candidate.Get_Parts(i)->Main.MaxElementIndex;
			element = candidate.Get_Parts(i)->Main.Elements;
			for (k = 0; k <= max; k++, element++) 
			{	
				if (element->exponent % 2) 
				{
					count++;
					rc = table.FindPrimeInHashtable(element->prime, element->root | ConstSPhase::ELEMENT_MASK[i], present, index);
					if (rc != ConstRC::Ok)
					{
						for (j = 0; j < nbrOfAlgParts; j++)
						{
							if (quadChars[j] != NULL)
							{
								free(quadChars[j]);
								quadChars[j] = NULL;
							}
						}
						delete matrix;
						matrix = NULL;
						fclose(relationsFILE);
						relationsFILE = NULL;
						return ConstRC::GeneralError;
					}

					columnIndex = table.Hashtable[index].index;
					if (columnIndex == ConstHT::UNDEFINED_RELATION) 
					{
						CLog::LogError(WHEREARG,"CFilteringPhase","CreateRelationMatrix",0,0,"PrimeIdeal was not found in the FB.");
						for (j = 0; j < nbrOfAlgParts; j++)
						{
							if (quadChars[j] != NULL)
							{
								free(quadChars[j]);
								quadChars[j] = NULL;
							}
						}
						delete matrix;
						matrix = NULL;
						fclose(relationsFILE);
						relationsFILE = NULL;
						return ConstRC::GeneralError;
					}
					matrix->PutOne(row, columnIndex);
				}
			}
		}

		// Now, generalized Legendre symbols are appended to the matrix.
		for (i = 0, k = 0; i < ConstSPhase::ELEMENT_MASK_COUNT; i++)
		{
			if (this->iSievingPoly[i]->GetDegree() > 1)
			{
				rc = ComputeLegendreSymbols(quadChars[k], quadCharsCount, candidate.Get_Parts(i)->Poly, matrix, row, legendreOffsets[k]);
				if (rc != ConstRC::Ok)
				{
					for (j = 0; j < nbrOfAlgParts; j++)
					{
						if (quadChars[j] != NULL)
						{
							free(quadChars[j]);
							quadChars[j] = NULL;
						}
					}
					delete matrix;
					matrix = NULL;
					fclose(relationsFILE);
					relationsFILE = NULL;
					return rc;
				}
				k++;
			}
		}

		// last column of 1's
		// more precisely: we have to add number of relations mod 2
		// if we use large prime variations then some relations are combined from more relations
		if (columnOfOnes && candidate.Get_NumberOfRelations() % 2)
		{
			matrix->PutOne(row, maxColumn);
		}

		row++;
	}

	// matrix is now ready
	Set_RelationMatrix(matrix);
	matrix = NULL;

	#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
	if (Get_InfoMode())
	{
			CLog::Log(1000, "0243088", "\r    Converting ... %10d       / %10ld      \n", iter, this->iRelationsCount);
			CLog::Log(1000, "0243089", "    Nonzero bits in the matrix: %ld\n", count);
	}
	#endif

	fclose(relationsFILE);
	relationsFILE = NULL;
	
	for (j = 0; j < nbrOfAlgParts; j++)
	{
		if (quadChars[j] != NULL)
		{
			free(quadChars[j]);
			quadChars[j] = NULL;
		}
	}

	return ConstRC::Ok;
}
int CFilteringPhase::SaveResult()
{
	Set_InnerState((int)EFPhaseSavingResult);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000 
	if (Get_InfoMode()) 
	{ 
		CLog::Log(1000, "0243210", "Writing relation matrix ..\n\n"); 
	} 
#endif 

	AbstractMatrix* matrix = Get_RelationMatrix();
	if (matrix == NULL)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","SaveResult",0,0,"Relation matrix doesn't exist.");
		return ConstRC::GeneralError;
	}

	CBaseParameters bparam;
	bparam.AddParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iRelationMatrixFullFileName);

	int rc = matrix->Serialize(bparam);
	if (rc == ConstRC::Ok)
	{
		AFactorAlgParameters * param = Get_Parameters();
		// serialization is ConstRC::Ok so we save file name for other phases
		param->SetParameterForOthers(ConstPN::RELATION_MATRIX_FULL_FILENAME, this->iRelationMatrixFullFileName, Get_PhaseNbr());
		string aux = NFSUtils::ConvertToString((int)this->iRelationMatrixType);
		param->SetParameterForOthers(ConstPN::RELATION_MATRIX_TYPE, aux, Get_PhaseNbr());
		param->SetParameterForOthers(ConstPN::RELATIONS_RESULT_FULL_FILENAME, this->iRelationsResultFullFileName, Get_PhaseNbr());

		Set_InnerState((int)EFPhaseFinished);
	}

	return rc;
}
int CFilteringPhase::Serialize(const CBaseParameters &aParam) const
{
	int rc = MSerializable::Serialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (Running())
	{
		AFactorAlgParameters * param = Get_Parameters();

		// for deserialization at new start
		param->SetPhaseParameter(ConstPN::DESERIALIZE, NFSUtils::ConvertToString(true), Get_PhaseNbr());
		param->SetPhaseParameter(ConstPN::PHASE_FULL_FILENAME, 
			MSerializable::CreateFileNameAuto(aParam, ConstFN::REL_PRO_PHASE), Get_PhaseNbr());

		CBaseParameters paramb;
		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_ParametersFullFileName()); 
		rc = param->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok; 
}
int CFilteringPhase::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_REL_PRO_PHASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","Serialize",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int state = Get_InnerState();
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_INNER_STATE_NAME, state);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_NBR_OF_RELATIONS_NAME, this->iRelationsCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_NUMBER_PRIME_IDEALS_MULTI_NAME, this->iPrimeIdealsCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (this->iLargePrimeIdealsCount != NULL)
	{
		int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_NUMBER_LARGE_PRIME_IDEALS_MULTI_NAME);
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","Serialize",rc,0,"Could not write the large prime ideals count tag.");
			return ConstRC::GeneralError;
		}
		rc = CXMLHelper::NewLine(aWriter, false);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		string auxStr;
		for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			auxStr = XML_SERIAL_NUMBER_LARGE_PRIME_IDEALS_MULTI_NAME + Utils::ConvertToString(i + 1);
			rc = CXMLHelper::SerializeSignedLong(aWriter, auxStr.c_str(), this->iLargePrimeIdealsCount[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		rc = xmlTextWriterEndElement(aWriter); // large prime ideals count
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CFilteringPhase","Serialize",rc,0,"Could not end the parameter tag.");
			return ConstRC::GeneralError;
		}
		rc = CXMLHelper::NewLine(aWriter, false);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","Serialize",rc,0,"Could not end the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok; 
}
int CFilteringPhase::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CFilteringPhase::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != EFPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_REL_PRO_PHASE_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CFilteringPhase","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int auxInt = 0;
	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INNER_STATE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize inner state...", auxInt, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_InnerState(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_NBR_OF_RELATIONS_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize relation count...", this->iRelationsCount, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_NUMBER_PRIME_IDEALS_MULTI_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize prime ideals count...", this->iPrimeIdealsCount, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_NUMBER_LARGE_PRIME_IDEALS_MULTI_NAME))
		{
			if (this->iLargePrimeIdealsCount != NULL)
			{
				for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
				{
					// next node
					rc = CXMLHelper::MoveToNextNode(aReader);
					if (rc != ConstRC::Ok)
					{
						rc = ConstRC::GeneralError;
						break;
					}
					rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize large prime ideals count...", this->iLargePrimeIdealsCount[i], false);
					if (rc != ConstRC::Ok)
					{
						rc = ConstRC::GeneralError;
						break;
					}
				}				
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
