#ifndef __FILTERING_PHASE_H__
#define __FILTERING_PHASE_H__

#include "abstract_rel_processing_phase.h"
#include "filtering_nfs_relation.h"
#include "factor_base_info.h"
#include "relation_hashtable.h"


/*! 
 * \brief Relation processing phase for double large prime variation
 *
 * Relation processing phase for NFS with two parts (can be (integral,
 * algebraic) or (algebraic, algebraic)).  It is recommended for too many
 * relations.  Merge is used for matrix resize, if only max 1,1-partial
 * relations are used, it could be better to use CRelProcessingPhase.
*/
class CFilteringPhase : public ARelProcessingPhase
{

private:

	/*!
		Number of all relations (smooth and partial). We get this number from last phase (or we can compute it, if we join relation files) but 
		we can remove some relations during removing duplicates and singletons. And also we will merge some relations.
		So this number will change and has to be save again.
	*/
	relation_index_type iRelationsCount;
	/*
		Numbers of all large prime ideals with odd power with multiplicity - for each relation part.
		These numbers are for info and also we will need them in merge function for creating hashtable.
		They change quite often and are serialized.
	*/
	relation_index_type * iLargePrimeIdealsCount;

	/*
		Number of all prime ideals with odd power with multiplicity.
	*/
	relation_index_type iPrimeIdealsCount;

	/*
		Lower bounds for large prime ideals. These bounds are used in merge function for relation hashtable.
		If not set in parameters we use FB upper bounds.
	*/
	main_sieving_type * iMergeLPILowerBounds;

	// multiplier of factor base bound which gives upper bound for large primes
	int iPartialFactorMultiplier[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	/*!
		The sieving polynomials.

		This polynomial can be monic or non-monic.
		In case that it is non-monic, the corresponding
		monic polynomial is present in the iDomain as 
		a member variable.
	
		By the corresponding monic polynomial we mean		
		
		f(x/a)*a^{d-1}

		where a is the leading coefficient and d the degree of f.

		In this phase it will be use for some calculations.
	*/
	Polynomial * iSievingPoly[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];


	///Full filenames with the quadratic characters - they will be created in this phase
	std::string iQuadCharsFullFileName[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	///Field of the factor base infos
	CFBInfo		iFBs[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	///Hashtable with frequencies of all prime ideals (at least once with odd exponent) 
	CFrequencyHashtable * iFreqHashtable;

	/*
		Number of relations above number of prime ideals ( = #relations - #primeideals)
		We also use min overlap - see RemoveRedundantRelations
	*/
	int iRelationsOverlap;

	/*
		Ratio of relations above number of prime ideals ( = #relations/#primeideals - 1)
		We also use min overlap - see RemoveRedundantRelations
		Used for calculation of iRelationsOverlap.
	*/
	double iRelationsOverlapRatio;

	std::string iAllRelationsFullFileName;
	std::string iTempFileFullFileName;

private:

	int CalculateRelationSize(int & aRelationSize);
	int SetupEdges(relation_index_type * aRelationIndexes, int aRelationIndexesCount, 
			CRelationHashtable * aTable, nfs_sieving_element * aElements, graph_wedge_t * aEdges);

protected:
	
	// Phase methods
	int JoinRelationFiles();
	int RemoveDuplicates();
	int RemoveSingletons();
	int RemoveRedundantRelations();
	int MergeRelations();
	int CreateRelationMatrix();
	///Save phase's result
	int SaveResult();

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();

	///Init all parameters necessary for calculation - get from iParameters
	int InitParameters();

	int RemoveSingletonsInner();

public:

	CFilteringPhase();
	~CFilteringPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);
	
	///Start method for linear phase
	int RunPhase(AFactorAlgParameters*  aParameters);

	// De/Serialization methods
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

};

#endif // __FILTERING_PHASE_H__
