#include "filtering_relation_part.h"

#include <cstring>
#include "base_parameters.h"
#include "const_debug.h"
#include "const_return_codes.h"
#include "const_filenames.h"
#include "const_element_flags.h"
#include "const_parameter_names.h"
#include "log.h"
#include "xml_helper.h"


using namespace std;

CFRelationPart::CFRelationPart() : MSerializable(ConstFN::FILTERING_RELATION_PART)
{
	this->Poly = new Polynomial(1); 
	this->Poly->Set(0, 1);
	this->iLargePrimeBound = -1; // all primeideal even integral (-1)
	this->LargePrimeIndex = 0; // first large prime - so if no exists it must be bigger than MaxIndex
}

CFRelationPart::CFRelationPart(const CFRelationPart& aOperand) : 
	MSerializable(aOperand), iLargePrimeBound(aOperand.iLargePrimeBound), Main(aOperand.Main),LargePrimeIndex(aOperand.LargePrimeIndex)
{

	this->Poly = new Polynomial();
	aOperand.Poly->Copy(this->Poly);
}

CFRelationPart::~CFRelationPart()
{
	if (this->Poly != NULL)
	{
		delete this->Poly;
		this->Poly = NULL;
	}
}
int CFRelationPart::Copy2(CFRelationPart& aOperand)
{
	aOperand.Set_DefaultFileName(this->Get_DefaultFileName());
	aOperand.Poly->Zeroize();
	this->Poly->Copy(aOperand.Poly);
	
	this->Main.Copy2(aOperand.Main);
	aOperand.iLargePrimeBound = this->iLargePrimeBound;
	aOperand.LargePrimeIndex = this->LargePrimeIndex;

	return ConstRC::Ok;
}
int CFRelationPart::Copy2(CRelationPart& aOperand)
{
	aOperand.Set_DefaultFileName(this->Get_DefaultFileName());
	aOperand.Poly->Zeroize();
	this->Poly->Copy(aOperand.Poly);

	mpz_set_ui(aOperand.Norm, 0);
	mpz_set_ui(aOperand.NormRemaining, 0);
	
	if (this->iLargePrimeBound != -1)
	{
		aOperand.Main.MaxElementIndex = -1;
		aOperand.Large.MaxElementIndex = -1;
		for (int i = 0; i < this->LargePrimeIndex; i++)
		{
			aOperand.Main.AddDivisor(this->Main.Elements + i); 
		}

		for (int i = this->LargePrimeIndex; i <= this->Main.MaxElementIndex; i++)
		{
			aOperand.Large.AddDivisor(this->Main.Elements + i); 
		}
	}
	else
	{
		this->Main.Copy2(aOperand.Main);
		aOperand.Large.MaxElementIndex = -1;
	}

	return ConstRC::Ok;
}
int CFRelationPart::CopyFrom(CRelationPart& aOperand)
{
	this->Set_DefaultFileName(aOperand.Get_DefaultFileName());
	this->Poly->Zeroize();
	aOperand.Poly->Copy(this->Poly);
	
	aOperand.Main.Copy2(this->Main);
	CDivisors * divisors = &aOperand.Large;
	for (int i = 0; i <= divisors->MaxElementIndex; i++)
	{
		this->Main.AddDivisor(divisors->Elements + i); 
	}

	return ConstRC::Ok;
}

bool CFRelationPart::Equals(const CFRelationPart& aOperand) const
{
	if (this->Poly->Equals(*aOperand.Poly) == false)
	{
		return false;
	}
	if (this->Main.Equals(aOperand.Main) == false)
	{
		return false;
	}
	if (this->iLargePrimeBound != aOperand.iLargePrimeBound) 
	{
		return false;
	}
	if (this->LargePrimeIndex != aOperand.LargePrimeIndex) 
	{
		return false;
	}
	return true;
}
int CFRelationPart::Reset()
{
	this->LargePrimeIndex = 0; // first large prime - so if no exists it must be bigger than MaxIndex
	this->Main.MaxElementIndex = -1;
	this->Poly->Zeroize();
	// don't reset iLargePrimeBound !

	return ConstRC::Ok;
}
int CFRelationPart::CombineWithPart(const CFRelationPart &aOperand, Polynomial * aThetaPoly)
{
	// check if it is not just integral part...
	if (aThetaPoly != NULL && aThetaPoly->GetDegree() > 1)
	{
		Polynomial* auxiliary = this->Poly->AssignAuxPoly();
		Polynomial* quotient  = this->Poly->AssignAuxPoly();
		Polynomial* remainder = this->Poly->AssignAuxPoly();	

		auxiliary->Multiply(auxiliary, this->Poly, aOperand.Poly);
		remainder->Divide(quotient, remainder, auxiliary, aThetaPoly);
		remainder->Copy(this->Poly);

		auxiliary->Dispose(auxiliary);
		quotient->Dispose(quotient);
		remainder->Dispose(remainder);
	}

	int rc = this->Main.CombineWithDivisors(aOperand.Main);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// set new LartPrimeIndex
	Set_LargePrimeBound(this->iLargePrimeBound);

	return ConstRC::Ok;
}
int CFRelationPart::CombineWithPart(const CFRelationPart &aOperand)
{
	int rc = this->Main.CombineWithDivisors(aOperand.Main);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// set new LartPrimeIndex
	Set_LargePrimeBound(this->iLargePrimeBound);

	return ConstRC::Ok;
}
void CFRelationPart::PrintToScreen()
{
	printf("PrimeIdeals:\n");
	this->Main.PrintToScreen();
}

/*!
	This method is used to save this relation part into a file.
	File Template:
	==================================================
	Poly
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CFRelationPart::PrintToFile(FILE * aFw)
{
	this->iAuxStr = this->Poly->ToString(ELongFormat);
	fprintf(aFw, "%s\n", this->iAuxStr.c_str());

	int rc = this->Main.PrintToFile(aFw);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
/*!
	This method is used to read this relation part from a file.
	File Template:
	==================================================
	Poly
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CFRelationPart::ReadFromFile(FILE * aFr, char * aBuffer, int aBufferSize)
{
	// Poly
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->iAuxStr.assign(aBuffer);
	int rc = Polynomial::Parse2(this->Poly, this->iAuxStr);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// Get count of PrimeIdeals
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	long count = atol(aBuffer);

	// Read PrimeIdeals as (prime|exponent|c_p|flags)
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	if (count > 0)
	{
		this->iAuxStr.assign(aBuffer);
		this->LargePrimeIndex = this->Main.ReadFromFile(count, this->iLargePrimeBound, this->iAuxStr);
		if (this->LargePrimeIndex < 0)
		{
			return ConstRC::BadArgument;
		}
	}
	else
	{
		this->Main.MaxElementIndex = -1;
		this->LargePrimeIndex = 0; // first large prime - so if no exists it must be bigger than MaxIndex
	}

	return ConstRC::Ok;
}
int CFRelationPart::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FILTERING_RELATION_PART_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CFRelationPart","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// serialize polynomial in string form
	rc = CXMLHelper::SerializeString(aWriter, XML_SERIAL_POLYNOMIAL_NAME, this->Poly->ToString(ELongFormat));
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_PRIMEIDEALS_NAME);
	rc = this->Main.Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CFRelationPart","Serialize",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;  
}

int CFRelationPart::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_FILTERING_RELATION_PART_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CFRelationPart","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CFRelationPart","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);
	CBaseParameters param;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_POLYNOMIAL_NAME))
		{
			rc = CXMLHelper::DeserializeString(aReader, "Deserialize polynomial...", this->iAuxStr, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			rc = Polynomial::Parse2(this->Poly, this->iAuxStr);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIMEIDEALS_NAME))
		{
			param.SetParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_PRIMEIDEALS_NAME);
			rc = this->Main.Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}

void CFRelationPart::Set_LargePrimeBound(main_sieving_type aLargePrimeBound) 
{ 
	this->LargePrimeIndex = this->Main.MaxElementIndex + 1;
	this->iLargePrimeBound = aLargePrimeBound; 
	if (this->Main.Elements != NULL)
	{
		nfs_sieving_element * elements = this->Main.Elements;
		for (int i = 0; i < this->LargePrimeIndex; i++)
		{
			if (elements->prime >= aLargePrimeBound)
			{
				this->LargePrimeIndex = i;
				break;
			}
		}
	}
}
