#include "frequency_hashtable.h"

#include "const_return_codes.h"
#include "const_hashtables.h"
#include "const_debug.h"
#include "const_filtering_phase.h"
#include "const_sieving_phase.h"
#include "nfs_utils.h"
#include "log.h"

using namespace std;


CFrequencyHashtable::CFrequencyHashtable()
{
	this->Hashtable = NULL;
	this->AllocatedSize = 0;
	this->Mask = 0;
	this->MaxPrimeIdeals = 0;
	this->Filled = false;
}
CFrequencyHashtable::~CFrequencyHashtable()
{
	DeleteHashtable();
}
void CFrequencyHashtable::DeleteHashtable()
{
	if (this->Hashtable != NULL)
	{
		free(this->Hashtable);
		this->Hashtable = NULL;
		this->AllocatedSize = 0;
		this->Mask = 0;
		this->MaxPrimeIdeals = 0;
		this->Filled = false;
	}
}
int CFrequencyHashtable::SetupHashtableProb(long aAmount)
{
	if (this->Hashtable != NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","SetupHashtableProb",0,0,"Hashtable is not NULL.");
		return ConstRC::BadArgument;
	}

	this->AllocatedSize = CFrequencyHashtable::GetSuitableHashtableSizeProb(aAmount);
	this->Hashtable = (fhashtable_entry*) Utils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(fhashtable_entry));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","SetupHashtableProb", 0, this->AllocatedSize * sizeof(fhashtable_entry),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	ClearHashtable();

	return ConstRC::Ok;
}
int CFrequencyHashtable::SetupHashtableEx(long aAmount)
{
	if (this->Hashtable != NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","SetupHashtableProb",0,0,"Hashtable is not NULL.");
		return ConstRC::BadArgument;
	}

	this->AllocatedSize = Utils::GetSuitableHashtableSize(aAmount);
	this->Hashtable = (fhashtable_entry*) NFSUtils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(fhashtable_entry));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","SetupHashtableProb", 0, this->AllocatedSize * sizeof(fhashtable_entry),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	ClearHashtable();

	return ConstRC::Ok;
}
int CFrequencyHashtable::ClearHashtable()
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","ClearHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	memset(this->Hashtable, 0, this->AllocatedSize * sizeof(fhashtable_entry));	
	// hashtable will be fill again so clear max number of prime ideals
	this->MaxPrimeIdeals = 0;

	this->Filled = false;

	return ConstRC::Ok;
}
/*
	This method checks whole hashtable for singletons (prime ideals with frequency 1).
	If some singleton is found (prime ideal) then we take corresponding relation index and mark this relation as singleton.
		- only prime ideals with frequency 1 have in field Index index of relation
	For this we use field aDetectionField (lenght is #relations / BITS_IN_BYTE).
	We also count number of unique singletons (return parameter aNewSingletonCount).
*/
int CFrequencyHashtable::FindSingletonRelations(detection_type * aDetectionField, long& aNewSingletonCount)
{
	if (this->Hashtable == NULL || aDetectionField == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","FindSingletonRelations",0,0,"Hashtable or DetectionField is NULL.");
		return ConstRC::BadArgument;
	}

	fhashtable_entry * aux = this->Hashtable;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	relation_index_type index;
	detection_type mask;
	aNewSingletonCount = 0;

#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
	relation_index_type primeIdealsCount = 0;
	relation_index_type primeIdealsHFCount = 0;
	relation_index_type primeIdealsSingletons = 0;
#endif

	do
	{
		// There is a prime on this position ...			
		if (aux->prime != 0) 
		{
#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
			primeIdealsCount++;
#endif
			// is it singleton?
			if (aux->frequency == 1)  
			{
#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
				primeIdealsSingletons++;
#endif
				index = aux->index >> ConstHT::BITS_IN_BYTE_SHIFT;
				mask = ConstHT::FIRST_DETECTION_BIT >> (aux->index & ConstHT::BITS_IN_BYTE_MOD);
				// is already marked?
				if ((aDetectionField[index] & mask) == 0) 
				{
					aNewSingletonCount++;
					aDetectionField[index] = aDetectionField[index] | mask; // mark as singleton
				}
			}
#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
			else
			{
				primeIdealsHFCount++;
			}
#endif
		}
	}
	while (++aux < end);

#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
	CLog::Log(1000, "1143001", "\n\tHF Stats: Prime ideals count: %d \n", primeIdealsCount);
	CLog::Log(1000, "1143002", "\tHF Stats: HF prime ideals count: %d \n", primeIdealsHFCount);
	CLog::Log(1000, "1143003", "\tHF Stats: Singletons prime ideals count: %d \n", primeIdealsSingletons);
	CLog::Log(1000, "1143004", "\tHF Stats: Singletons relation count: %d \n", aNewSingletonCount);
#endif

	return ConstRC::Ok;
}
/*
	This method checks whole hashtable for singletons (prime ideals with frequency 1).
	If some singleton is found (prime ideal) then we take corresponding relation index and mark this relation as singleton.
		- only prime ideals with frequency 1 have in field Index index of relation
	For this we use field aDetectionField (lenght is #relations / BITS_IN_BYTE).
	We also count number of unique singletons (return parameter aNewSingletonCount) and number of prime ideals with high frequency ( >= 2).
*/
int CFrequencyHashtable::FindSingletonRelations(detection_type * aDetectionField, relation_index_type& aNewSingletonCount, relation_index_type& aHFPrimeIdeals)
{
	if (this->Hashtable == NULL || aDetectionField == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","FindSingletonRelations",0,0,"Hashtable or DetectionField is NULL.");
		return ConstRC::BadArgument;
	}

	fhashtable_entry * aux = this->Hashtable;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	relation_index_type index;
	detection_type mask;
	aNewSingletonCount = 0;
	aHFPrimeIdeals = 0;

#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
	relation_index_type primeIdealsCount = 0;
	relation_index_type primeIdealsSingletons = 0;
#endif

	do
	{
		// There is a prime on this position ...			
		if (aux->prime != 0) 
		{
#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
			primeIdealsCount++;
#endif
			// is it singleton?
			if (aux->frequency == 1)  
			{
#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
				primeIdealsSingletons++;
#endif
				index = aux->index >> ConstHT::BITS_IN_BYTE_SHIFT;
				mask = ConstHT::FIRST_DETECTION_BIT >> (aux->index & ConstHT::BITS_IN_BYTE_MOD);
				// is already marked?
				if ((aDetectionField[index] & mask) == 0) 
				{
					aNewSingletonCount++;
					aDetectionField[index] = aDetectionField[index] | mask; // mark as singleton
				}
			}
			else
			{
				aHFPrimeIdeals++;
			}
		}
	}
	while (++aux < end);

#ifdef DEBUG_FREQUENCY_HASHTABLE_PRINT_STATS
	CLog::Log(1000, "1143005", "\n\tHF Stats: Prime ideals count: %d \n", primeIdealsCount);
	CLog::Log(1000, "1143006", "\tHF Stats: HF prime ideals count: %d \n", aHFPrimeIdeals);
	CLog::Log(1000, "1143007", "\tHF Stats: Singletons prime ideals count: %d \n", primeIdealsSingletons);
	CLog::Log(1000, "1143008", "\tHF Stats: Singletons relation count: %d \n", aNewSingletonCount);
#endif

	return ConstRC::Ok;
}
relation_index_type CFrequencyHashtable::GetSuitableHashtableSizeProb(relation_index_type aAmount)
{
	// TODO
	// we can expect that number of relations is bigger than number of prime ideals
	long result;
	// WARNING - convertion between long and relation_index_type !!!
	NFSUtils::GetClosestBiggerPowerOf2(result, aAmount);

	return result;
}
/*
	This method calculates number of prime ideals in hashtable with given frequency (aCountWithFreq).
	It also calculates number of all prime ideals in hashtable.
		- this number is number of all prime ideals (at least once with odd exponent)
*/
int CFrequencyHashtable::CountPrimeIdealsWithFrequency(unsigned int aFrequency, relation_index_type & aCountWithFreq, relation_index_type & aCountAll)
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","CountPrimeIdealsWithFrequency",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	aCountWithFreq = 0;
	aCountAll = 0;
	fhashtable_entry * aux = this->Hashtable;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	do 
	{
		if (aux->frequency > 0)
		{
			aCountAll++;
			if (aux->frequency == aFrequency)
			{
				aCountWithFreq++;
			}
		}
	}
	while (++aux < end);

	return ConstRC::Ok;
}
/*
	This method count number of all prime ideals in hashtable with multiplicity for each relation part separately (large prime ideals).
	Also count number of all prime ideals.
*/
int CFrequencyHashtable::CountPrimeIdeals(relation_index_type * aLargePrimeIdealsCount, const unsigned int * aMasks, main_sieving_type * aLargePrimeBounds, int aRelationPartCount, relation_index_type & aPrimeIdealsCount)
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","CountPrimeIdeals",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	int i;
	aPrimeIdealsCount = 0;
	for (i = 0; i < aRelationPartCount; i++)
	{
		aLargePrimeIdealsCount[i] = 0;
	}
	fhashtable_entry * aux = this->Hashtable;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	do 
	{
		// some prime ideals exist
		if (aux->frequency > 0)
		{
			// find corresponding relation part
			for (i = 0; i < aRelationPartCount; i++)
			{
				if (aMasks[i] == (aux->root_fingerprint & ConstSPhase::ELEMENT_MASK_BIT))
				{
					// check if large prime ideal
					if (aLargePrimeBounds[i] < aux->prime)
					{
						aLargePrimeIdealsCount[i] += aux->frequency;
						break;
					}
				}
			}

			aPrimeIdealsCount += aux->frequency;
		}
	}
	while (++aux < end);

	return ConstRC::Ok;
}
int CFrequencyHashtable::GetAllPrimeIdealsInComponent(int aComponentID, CCliqueComponentArray * aComponents, set<prime_ideal_t, prime_ideal_comp> * aPrimeIdeals)
{
	if (this->Hashtable == NULL || aComponents == NULL || aPrimeIdeals == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","GetAllPrimeIdealsInComponent",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	int rc;
	fhashtable_entry * aux = this->Hashtable;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	relation_index_type primeidealComponentIndex;
	prime_ideal_t primeideal;
	do 
	{
		if (aux->frequency == ConstFPhase::FREQUENCY_TWO)
		{
			// get main component of current prime ideal
			rc = aComponents->GetMainComponent(aux->index, primeidealComponentIndex);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}

			if (primeidealComponentIndex == aComponentID)
			{
				primeideal.prime = aux->prime;
				primeideal.rootmask = aux->root_fingerprint;
				aPrimeIdeals->insert(primeideal);
			}
		}
	}
	while (++aux < end);

	return ConstRC::Ok;
}
int CFrequencyHashtable::AddToTable(CFRelationPart * aPart, unsigned long aRootMask, relation_index_type aIndex)
{
	if (this->Hashtable == NULL || aPart == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","AddToTable",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	int rc;
	bool present = false;
	relation_index_type index;
	int max = aPart->Main.MaxElementIndex;
	// set max number of prime ideals
	if (max >= this->MaxPrimeIdeals)
	{
		this->MaxPrimeIdeals = max + 1;
	}
	nfs_sieving_element * element;
	int i;
	for (i = 0, element = aPart->Main.Elements; i <= max; i++, element++)
	{
		if (element->exponent % 2)
		{
			rc = FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index);
			if (rc == ConstRC::NeedEnlargement)
			{
				rc = EnlargeHashtable();
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				rc = FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index);
			}
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			
			if (!present) 
			{
				// This prime has not been in hashtable yet.					
				this->Hashtable[index].prime = element->prime;
				this->Hashtable[index].root_fingerprint = element->root | aRootMask;
				this->Hashtable[index].frequency = 1;
				this->Hashtable[index].index = aIndex;
			} 
			else 
			{
				// We increase frequency of this primeideal
				this->Hashtable[index].frequency++;
				// reset relation index (this index is used also for removing redundant relations)
				this->Hashtable[index].index = ConstHT::UNDEFINED_RELATION;
			}
		}
	}

	this->Filled = true;

	return ConstRC::Ok;
} 
int CFrequencyHashtable::AddToTable(CRelationPart * aPart, unsigned long aRootMask, relation_index_type aIndex)
{
	if (this->Hashtable == NULL || aPart == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","AddToTable",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	int rc;
	bool present = false;
	relation_index_type index;
	int max = aPart->Main.MaxElementIndex;
	// set max number of prime ideals
	if (aPart->Main.MaxElementIndex + aPart->Large.MaxElementIndex >= this->MaxPrimeIdeals)
	{
		this->MaxPrimeIdeals = aPart->Main.MaxElementIndex + aPart->Large.MaxElementIndex + 1;
	}
	nfs_sieving_element * element;
	int i;

	// add main prime ideals to hashtable
	for (i = 0, element = aPart->Main.Elements; i <= max; i++, element++)
	{
		if (element->exponent % 2)
		{
			rc = FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index);
			if (rc == ConstRC::NeedEnlargement)
			{
				rc = EnlargeHashtable();
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				rc = FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index);
			}
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			
			if (!present) 
			{
				// This prime has not been in hashtable yet.					
				this->Hashtable[index].prime = element->prime;
				this->Hashtable[index].root_fingerprint = element->root | aRootMask;
				this->Hashtable[index].frequency = 1;
				this->Hashtable[index].index = aIndex;
			} 
			else 
			{
				// We increase frequency of this primeideal
				this->Hashtable[index].frequency++;
				// reset relation index (this index is used also for removing redundant relations)
				this->Hashtable[index].index = ConstHT::UNDEFINED_RELATION;
			}
		}
	}

	max = aPart->Large.MaxElementIndex;
	// add large prime ideals to hashtable
	for (i = 0, element = aPart->Large.Elements; i <= max; i++, element++)
	{
		if (element->exponent % 2)
		{
			rc = FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index);
			if (rc == ConstRC::NeedEnlargement)
			{
				rc = EnlargeHashtable();
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				rc = FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index);
			}
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			
			if (!present) 
			{
				// This prime has not been in hashtable yet.					
				this->Hashtable[index].prime = element->prime;
				this->Hashtable[index].root_fingerprint = element->root | aRootMask;
				this->Hashtable[index].frequency = 1;
				this->Hashtable[index].index = aIndex;
			} 
			else 
			{
				// We increase frequency of this primeideal
				this->Hashtable[index].frequency++;
				// reset relation index (this index is used also for removing redundant relations)
				this->Hashtable[index].index = ConstHT::UNDEFINED_RELATION;
			}
		}
	}

	this->Filled = true;

	return ConstRC::Ok;
} 
/*
	This method is used when creating relation matrix.
	The hashtable must be fill with all prime ideal even with prime ideals with even power (but specially marked).
	Parameter aCurrentCount is used for indexing prime ideals.

	If we go throught all relations we get number of all prime ideals which will be in matrix. 
		- in variable aCurrentCount
*/
int CFrequencyHashtable::AddToTableCM(CFRelationPart * aPart, unsigned long aRootMask, relation_index_type& aCurrentCount)
{
	if (this->Hashtable == NULL || aPart == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","AddAllToTable",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	bool present = false;
	relation_index_type index;
	int max = aPart->Main.MaxElementIndex;
	if (max >= this->MaxPrimeIdeals)
	{
		this->MaxPrimeIdeals = max + 1;
	}
	nfs_sieving_element * element;
	int i;
	for (i = 0, element = aPart->Main.Elements; i <= max; i++, element++)
	{
		if (FindPrimeInHashtable(element->prime, element->root | aRootMask, present, index) != ConstRC::Ok)
		{
			return ConstRC::GeneralError;
		}
		
		if (!present) 
		{
			// This prime has not been in hashtable yet.					
			this->Hashtable[index].prime = element->prime;
			this->Hashtable[index].root_fingerprint = element->root | aRootMask;
			this->Hashtable[index].frequency = 1;
			if (element->exponent % 2)
			{
				this->Hashtable[index].index = aCurrentCount++;
			}
			else
			{
				// even exponent = we just mark that this primeideal exists
				this->Hashtable[index].index = ConstHT::UNDEFINED_RELATION;
			}
		} 
		else 
		{
			this->Hashtable[index].frequency++;
			// if the primeideal is in table but was added with even exponent and this exponent is odd, we increase count  
			if (element->exponent % 2 && this->Hashtable[index].index == ConstHT::UNDEFINED_RELATION)
			{
				this->Hashtable[index].index = aCurrentCount++;
			}
		}
	}

	this->Filled = true;

	return ConstRC::Ok;
}
/*!
	This method tries to find prime aPrime in hashtable, and indicates
	the result of this search in two values: the return value, which is an index
	corresponding to the real or expected position of the prime in the hashtable,
	and value of aPresent, which indicates whether the prime has already been
	in the hashtable or no. In the second case, the return value has meaning
	"if you want to insert aPrime, this index is the one where it should be inserted".

*/
int CFrequencyHashtable::FindPrimeInHashtable(main_sieving_type aPrime, main_sieving_type aRootFingerprint, bool & aPresent, relation_index_type & aIndex) const
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","FindPrimeInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aIndex = 0;
		return ConstRC::BadArgument;
	}

	relation_index_type hash = (aPrime >> 1) & this->Mask;		
	int rc = CollisionResolution(hash, aPrime, aRootFingerprint, aIndex);
	if (rc == ConstRC::NeedEnlargement)
	{
		aPresent = false;
		aIndex = 0;
		return rc;
	}
	if (rc != ConstRC::Ok)
	{
		aPresent = false;
		aIndex = 0;
		return ConstRC::GeneralError;
	}

    if (this->Hashtable[aIndex].prime == 0) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

	return ConstRC::Ok;
}
/*!
	This method tries to find prime aPrime in hashtable, and indicates
	the result of this search in two values: the return value, which is an index
	corresponding to the real or expected position of the prime in the hashtable,
	and value of aPresent, which indicates whether the prime has already been
	in the hashtable or no. In the second case, the return value has meaning
	"if you want to insert aPrime, this index is the one where it should be inserted".

	This method looks only for prime - not prime ideal so it does not use root.

*/
int CFrequencyHashtable::FindPrimeInHashtable(main_sieving_type aPrime, bool & aPresent, relation_index_type & aIndex) const
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","FindPrimeInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aIndex = 0;
		return ConstRC::BadArgument;
	}

	relation_index_type hash = (aPrime >> 1) & this->Mask;			
	if (CollisionResolution(hash, aPrime, aIndex) != ConstRC::Ok)
	{
		aPresent = false;
		aIndex = 0;
		return ConstRC::GeneralError;
	}

    if (this->Hashtable[aIndex].prime == 0) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

	return ConstRC::Ok;
}
/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where aPrime can be stored.

*/
int CFrequencyHashtable::CollisionResolution(relation_index_type aHash, main_sieving_type aPrime, main_sieving_type aRootFingerprint, relation_index_type & aIndex) const
{
	aIndex = aHash;
	relation_index_type iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	fhashtable_entry * aux = this->Hashtable + aIndex;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;

    while (((aux->prime != aPrime) || (aux->root_fingerprint != aRootFingerprint)) && (aux->prime != 0)) 
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// The hash table has overflown
				return ConstRC::NeedEnlargement;
			}
			aux = this->Hashtable;
		}
	}

	aIndex = aux - this->Hashtable;

	return ConstRC::Ok;
}
/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where aPrime can be stored.

	This method looks only for prime - not prime ideal so it does not use root.

*/
int CFrequencyHashtable::CollisionResolution(relation_index_type aHash, main_sieving_type aPrime, relation_index_type & aIndex) const
{
	aIndex = aHash;
	relation_index_type iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	fhashtable_entry * aux = this->Hashtable + aIndex;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;

    while ((aux->prime != aPrime) && (aux->prime != 0)) 
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"CFrequencyHashtable","CollisionResolution",0,0,"The hash table has overflown.");
				return ConstRC::GeneralError;
			}
			aux = this->Hashtable;
		}
	}

	aIndex = aux - this->Hashtable;

	return ConstRC::Ok;
}
int CFrequencyHashtable::GetPrimeFromHashtable(nfs_sieving_element *aElement, unsigned int aRootMask, fhashtable_entry *&aEntry) const
{
	if (this->Hashtable == NULL || aElement == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","GetPrimeFromHashtable",0,0,"Hashtable is NULL.");
		aEntry = NULL;
		return ConstRC::BadArgument;
	}

	relation_index_type hash = (aElement->prime >> 1) & this->Mask;	

	if (CollisionResolution(hash, aElement->prime, aElement->root | aRootMask, aEntry) != ConstRC::Ok)
	{
		aEntry = NULL;
		return ConstRC::GeneralError;
	}

    if (aEntry->prime == 0) 
	{
		aEntry = NULL;
		return ConstRC::NotFound;
	} 

	return ConstRC::Ok;
}
int CFrequencyHashtable::CollisionResolution(relation_index_type aHash, main_sieving_type aPrime, main_sieving_type aRootFingerprint, fhashtable_entry *& aEntry) const
{
	aEntry = NULL;
	relation_index_type iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	aEntry = this->Hashtable + aHash;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;

    while (((aEntry->prime != aPrime) || (aEntry->root_fingerprint != aRootFingerprint)) && (aEntry->prime != 0)) 
	{
		iter++;
		aEntry++;
		// Should we go over the bounds of the table, we must reset aEntry to start.
		if (aEntry >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"CFrequencyHashtable","CollisionResolution",0,0,"The hash table has overflown.");
				return ConstRC::GeneralError;
			}
			aEntry = this->Hashtable;
		}
	}

	return ConstRC::Ok;
}
int CFrequencyHashtable::EnlargeHashtable()
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CFrequencyHashtable","EnlargeHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

#ifdef DEBUG_FREQUENCY_HASHTABLE_INFO_MODE_ON
	CLog::Log(1000, "1143009", "\nHashtable enlarged from %d\n", this->AllocatedSize);
#endif

	fhashtable_entry * oldHashtable = this->Hashtable;
	fhashtable_entry * aux = this->Hashtable;
	fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	int oldMaxPrimeIdeals = this->MaxPrimeIdeals;

	//reallocate hashtable
	this->Hashtable = NULL;
	int rc = SetupHashtableEx(this->AllocatedSize);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	this->MaxPrimeIdeals = oldMaxPrimeIdeals;
	this->Filled = true;

	bool present = false;
	relation_index_type index = 0;

	do
	{
		// There is a prime on this position ...			
		if (aux->prime != 0) 
		{
			rc = FindPrimeInHashtable(aux->prime, aux->root_fingerprint, present, index);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			this->Hashtable[index].prime = aux->prime;
			this->Hashtable[index].root_fingerprint = aux->root_fingerprint;
			this->Hashtable[index].frequency = aux->frequency;
			this->Hashtable[index].index = aux->index;
		}
	}
	while (++aux < end);

	free(oldHashtable);
	oldHashtable = NULL;

	return ConstRC::Ok;
}
