#include <gmp.h>
#include "abstract_factor_alg_parameters.h"

/*!
 * \brief This class serves for generating a number C (mpz_t). 
 *
	The number C is a product of primes - small primes can be in higher powers with some bigger primes in the first power (the number of bigger primes is given as a parameter).
	Only primes which are lower then the parameter can be used in the resulting product  .
	The resulting product must be within bounds.
	The main method - NextSequence - get the next product (C).
*/
class CGenerateC : public virtual MSerializable
{	
public:
//constants
	static const int PRIME_COUNT_FOR_C = 50;	///< the number of primes (including repeated primes) for generating C
	static const int MAX_COUNT_BIGGER = 10;		///< the maximum number of bigger primes in the sequence (for define the array)
	static const int FIRST_BIGGER = 9;			///< the index of the first bigger prime in iPrimeForC
	static const int COUNT_BIGGER = 2;			///< the number of bigger primes in the sequence
	static const int MAX_PRIME = 11;			///< the maximal prime that can be used in the sequence
	static const int SEQUENCE_BASE = 60;		///< (2*2*3*5) the product of smalls primes always contained in the resulting product
	static const int LOWER_BOUND = 1000;		///< the lower bound of the resulting product
	static const int UPPER_BOUND = 100000;		///< the upper bound of the resulting product

	static const int GENER_C_LOG_PRIORITY = 10000;	///< the priority for CLog::Log() 

	static const int iPrimesForC[PRIME_COUNT_FOR_C]; ///< the primes for generating C (constants)

private:
//parameters
	int iFirstBigger;   ///< the index of the first bigger prime in iPrimeForC
	int iCountBigger;   ///< the number of bigger primes in the sequence
	int iMaxPrime;		///< the maximal prime that can be used in the sequence
	int iBase;			///< the product of small primes contained in the resulting product
	mpz_t iUpperBound;	///< upper bound of the resulting product 
	mpz_t iLowerBound;	///< lower bound of the resulting product 
	
//auxilliary variables
	int iWorkingSequence[PRIME_COUNT_FOR_C];	///< the last used sequence for generating C
	int iIndex;		///< the number of iWorkingSequence members = the index of iWorkingSequence
	int iNextPrime; ///< index of the next available prime in iPrimesForC
	int iWorkingSequenceBigger[MAX_COUNT_BIGGER];	///< the last used sequence for generating C (biggers primes)
	int iIndexBigger;		///< the number of iWorkingSequenceBigger members = the index of iWorkingSequenceBigger
	int iNextBiggerPrime;   ///< index of the next available bigger prime in iPrimesForC
	
	bool iBegin;		///< true = begining of sequences generating
	mpz_t iResultProduct; ///< the resulting product (the number C)
	bool iSmall;		///< the product is too small - the auxilliary parameter for terminating the next sekvence generation
	

//auxilliary methods
	int CompleteToMax();
	bool ProductInBounds(mpz_t& aProduct, int aValue);
	int Initialisation();
	int NextDifferentPrime(int aIndex);
//	void PrintSequence();
	
public:
//constructors and destructors
	CGenerateC();
	CGenerateC(int aFirstBigger, int aMaxPrime, char* aLowerBound, char* aUpperBound, int aBase, int aCountBigger);
	CGenerateC(int aFirstBigger, int aMaxPrime, mpz_t aLowerBound, mpz_t aUpperBound, int aBase, int aCountBigger);

	~CGenerateC();

//main methods	
	int NextSequence(mpz_t& aResult); //returns the next product (next sequence)
	
//auxilliary methods	
	void PrintAllResults(); //method for testing only - prints all the sequences with the products
	void PrintSequence();

	int ResetAll();

	///This method setup parameters according to factorization algorithm parameters
	int Setup(const AFactorAlgParameters * aParameters, int aPhaseNumber);
	
//getters
	int GetFirstBiggerPrime();
	int GetMaxPrime();
	int GetCountBiggerPrimes();
	int GetBase();
	char* GetLowerBound();
	char* GetUpperBound();
	void GetActualSequence(int aSequence[],int& aSize);

//de/serialization methods
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);


	
};
