#include "hashtables.h"

#include "const_debug.h"
#include "const_return_codes.h"
#include "nfs_utils.h"
#include "log.h"

#if defined(__GNUC__)
#define PREFETCH(x)  __builtin_prefetch(x)
#else
#define PREFETCH(x) /* nothing */
#endif

CHashtables::CHashtables()
{
	this->Hashtables = NULL;
	this->Caches = NULL;
	this->MaxIndex = -1;
	this->AllocatedSize = 0;
}
CHashtables::~CHashtables()
{
	DeleteHashtables();
}
int CHashtables::AddToTable(main_sieving_type aPrime, main_sieving_type aRoot, log_type aLog, main_sieving_type aStart, int aHashtableIndex)
{
	if (aHashtableIndex >= this->AllocatedSize)
	{
		printf("aHashtableIndex too high: %d for prime %d.\n",aHashtableIndex,aPrime);
		return ConstRC::GeneralError;
	}

	int ret;
	cache_t* cache = &(this->Caches[aHashtableIndex]);
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
	if ((aPrime == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) && (aCp))
	{
		printf("Add To Table: prime is %d, root is %d, log is %d, start is %d, index in caches is %d.\n",aPrime,aRoot,aLog,aStart,index);
	}
#endif
	
	if (cache->used == STRUCTURE_CACHE_UPDATES)
	{
		
		hashtable_t* target = &(this->Hashtables[aHashtableIndex]);
		ret = FlushUpdateCache(target,cache);
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
	}
	cache->updates[cache->used].p    = aPrime;
	cache->updates[cache->used].h    = aStart;
	cache->updates[cache->used].logp = aLog;
	cache->updates[cache->used].r	 = aRoot;
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
	if ((aPrime == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) && (aRoot))
	{
		printf("AddToTable: DEBUG_SIEVING_PHASE_CRITICAL_PRIME.\n");
		nfs_fb_type type;
		type.p = aPrime;
		type.root_1 = aStart;
		type.log_p = aLog;
		type.root = aRoot;
		PrintNfsElementInfo(type);
		PrintUpdateInfo(cache->updates[cache->used]);
		Utils::Wait();
	}
#endif
	
	cache->used++;
	return ConstRC::Ok;
}
int CHashtables::FlushUpdateCache(hashtable_t* aHashtable, cache_t* aCache)
{
	int ret;
#ifdef DEBUG_SIEVING_PHASE_CACHE_FLUSHER
	printf("~~~~~~~ Flushing update cache start ~~~~~~~~\n");
#endif
	script_t* script = aHashtable->script;
	script_t* next   = script->next;
	if (!script)
	{
		//printf("Flush update cache: problem, script is NULL.");
		return ConstRC::InconsistentState;
	}

	int previous = 0;
	while (next)
	{
		if (script->num_used < STRUCTURE_MAX_UPDATES)
		{
#ifdef DEBUG_SIEVING_PHASE_CACHE_FLUSHER
			if (next == NULL)
			{
				printf("Flush update cache: the current hashtable is already allocated, but empty.\n");
			}
#endif
			break;
		}
		script = next;
		next = script->next;
		if (next) 
		{
			previous += script->num_used;
		}
	}
	// bool newScript = FALSE;
	// Copy the part of the list to cache
	PREFETCH(&(script->list[script->num_used]));

	// Now script is the last script.
	if ((next == NULL) && ((script->num_used + aCache->used) > STRUCTURE_MAX_UPDATES))
	{
		//printf("Flush update cache: script->num used is %d, used is %d, max_upd is %d, will call next ... \n",script->num_used,aCache->used,MAX_UPDATES);
		ret = AddScript(script);
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
		//newScript = TRUE;
	}
	register int i, max = aCache->used;
	register int j = script->num_used;
	for (i = 0; i < max; i++)
	{
		if (script->num_used == STRUCTURE_MAX_UPDATES)
		{
			//printf("script changed: next is %p\n",script->next);
			previous += script->num_used;
			script = script->next;
			j = 0;			
		}
		j = script->num_used;

		/*printf("Doplnuji z cache: p is %d.\n",aCache->updates[i].p);
		printf("Doplnuji z cache: h is %d.\n",aCache->updates[i].h);*/
		/*if (newScript)
		{
			printf("j = %d, p = %d\n",j,aCache->updates[i].p);
		}*/
		script->list[j].p    = aCache->updates[i].p;
		script->list[j].h    = aCache->updates[i].h;
		script->list[j].logp = aCache->updates[i].logp;
		script->list[j].r	 = aCache->updates[i].r;
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
		/*if (DEBUG_SIEVING_PHASE_CRITICAL_PRIME == script->list[j].p)
		{
			printf("Flush update cache message:\n");
			PrintUpdateInfo(aCache->updates[i]);
			Utils::Wait();
		}*/
#endif

#ifdef DEBUG_SIEVING_PHASE_CACHE_FLUSHER
		printf("Flush update cache, added to index %6d prime %6d with root %6d.\n",j+previous,script->list[j].p,script->list[j].h);
#endif
		script->num_used++;
				
	}
	aCache->used = 0;
#ifdef DEBUG_SIEVING_PHASE_CACHE_FLUSHER
	printf("~~~~~~~ Flushing update cache end ~~~~~~~~\n");
#endif

	return ConstRC::Ok;
}
int CHashtables::FlushAllUpdateCaches()
{
	// Flush all caches
	hashtable_t *h = this->Hashtables;
	cache_t		*c = this->Caches; 
	for (int i = 0; i <= this->MaxIndex; i++, h++, c++) 
	{
		FlushUpdateCache(h, c);
	}

	return ConstRC::Ok;
}
void CHashtables::DeleteHashtables()
{
	if (this->Hashtables != NULL)
	{
		script_t* ptr;
		script_t* next;
		hashtable_t* h1 = this->Hashtables;
		for (int i = 0; i <= this->MaxIndex; i++, h1++) 
		{
			if (h1 != NULL) 
			{
				ptr = h1->script;
				while (ptr)
				{
					next = ptr->next;
					//if (ptr == next)
					//{
					//	printf("LineSiever.deleteHashtable: Strange error: ptr == next");
					//}
					free(ptr);
					ptr = next;
				}
			}
		}

		free(this->Hashtables);
		this->Hashtables = NULL;
		this->AllocatedSize = 0;
		this->MaxIndex = -1;
	}
	if (this->Caches != NULL)
	{
		free(this->Caches);
		this->Caches = NULL;
	}
}
int CHashtables::AllocateHashtables()
{
	if (this->Hashtables != NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","AllocateHashtables",0,0,"Hashtables are not NULL");
		return ConstRC::GeneralError;
	}

	this->Hashtables = (hashtable_t*) Utils::AllocationMachine((void**)&this->Hashtables,this->AllocatedSize*sizeof(hashtable_t));
	if (this->Hashtables == NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","AllocateHashtables",0,this->AllocatedSize*sizeof(hashtable_t),"Could not allocate space for the sieving hashtable");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Hashtables, 0, this->AllocatedSize * sizeof(hashtable_t));	
	
	return ConstRC::Ok;
}
int CHashtables::AllocateCaches()
{
	if (this->Caches != NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","AllocateCaches",0,0,"Caches are not NULL");
		return ConstRC::GeneralError;
	}

	this->Caches = (cache_t*) Utils::AllocationMachine((void**)&this->Caches,this->AllocatedSize*sizeof(cache_t));
	if (this->Caches == NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","AllocateCaches",0,this->AllocatedSize*sizeof(cache_t),"Could not allocate space for the sieving cache");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Caches, 0, this->AllocatedSize * sizeof(cache_t));
	
	return ConstRC::Ok;
}

int CHashtables::SetupHashtables(int aMaxIndex)
{
	if (this->Hashtables != NULL || this->Caches != NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","SetupHashtables",0,0,"Caches or Hashtables are not NULL.");
		return ConstRC::GeneralError;
	}

	// aMaxIndex is just index - we also setup hashtables with the security margin
	this->AllocatedSize = aMaxIndex + 1 + CHashtables::SECURITY_MARGIN;
	this->MaxIndex = aMaxIndex;
	int rc = AllocateHashtables();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = AllocateCaches();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	hashtable_t* t1 = this->Hashtables;
	cache_t * c1 = this->Caches;
	for (int i = 0; i < this->AllocatedSize; i++, t1++, c1++)
	{
		t1->script = (script_t*) malloc(sizeof(script_t));
		//t1->script = (script_t*) Utils::AllocationMachine((void**)&t1->script, sizeof(script_t));
		if (t1->script == NULL)
		{
			CLog::LogError(WHEREARG,"CHashtables","SetupHashtable",i,sizeof(script_t),"Could not allocate space for the script inside a hashtable");
			return ConstRC::NotEnoughMemory;
		}
		t1->script->next = NULL;
		t1->script->num_used = 0;
		c1->used = 0;
	}

	return ConstRC::Ok;
}

void CHashtables::ZeroizeLastHashtable()
{
	hashtable_t* last = this->Hashtables + this->MaxIndex;
	script_t* script = last->script;
	while (script)
	{
		script->num_used = 0;
		script = script->next;
	}
}

int CHashtables::SwapHashtable(int aIndex)
{
	if ((aIndex < 0) || (aIndex > this->MaxIndex))
	{
		return ConstRC::BadArgument;
	}
	script_t* ptr = this->Hashtables[aIndex].script;
	this->Hashtables[aIndex].script = this->Hashtables[this->MaxIndex].script;
	this->Hashtables[this->MaxIndex].script = ptr;
	ZeroizeLastHashtable();

	return ConstRC::Ok;
}
int CHashtables::ClearHashtables()
{
	hashtable_t* target1 = this->Hashtables;
	cache_t*     cache1  = this->Caches;
	script_t*	 script;
	for (int i = 0; i <= this->MaxIndex; i++, target1++, cache1++)
	{
		script = target1->script;
		while (script)
		{
			script->num_used = 0;
			script = script->next;
		}
		cache1->used = 0;
	}

	return ConstRC::Ok;
}
int CHashtables::AddScript(script_t* aLast)
{
	if (aLast == NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","AddScript",0,0,"Add script found that last script is null!");
		return ConstRC::BadArgument;
	}
	// Next script already allocated
	if (aLast->next != NULL)
	{
		return ConstRC::Ok;
	}
	script_t* next = (script_t*) malloc(sizeof(script_t));
	if (next == NULL)
	{
		CLog::LogError(WHEREARG,"CHashtables","AddScript",0,sizeof(script_t),"Not enough memory to allocate new script!");
		return ConstRC::NotEnoughMemory;
	}
	memset(next,0,sizeof(script_t));
	aLast->next = next;
	// We will prefetch a bit of memory, because we add scripts in the 
	PREFETCH(next->list);

	return ConstRC::Ok;
}
void CHashtables::SearchForPrimeInHashtable(main_sieving_type aPrime, main_sieving_type aRoot)
{
	int j, max;
	int total = 0;
	script_t* script;
	update_t *updates;
	hashtable_t* h = this->Hashtables;
	for (int i = 0; i <= this->MaxIndex; i++, h++)
	{
		printf("Searching in hashtable %d ... ", i);
		fflush(stdout);

		script = h->script;
		bool found = false;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (j = 0; j < max; j++, updates++)
			{
				total++;
				if (updates->p == aPrime)
				{
					if (updates->r == aRoot) 
					{
						printf("FOUND!\n");
						found = true;
						PrintUpdateInfo(updates);
					}
				}
			}
			script = script->next;
		}
		if (!found) 
		{
			printf("NOT FOUND (listed total %d).\n", total);
		}
	}
}
void CHashtables::PrintUpdateInfo(update_t * aInfo)
{
	printf("\nUpdate Type: \n");
	printf("     Prime   : %d\n", aInfo->p);
	printf("     Root    : %d\n", aInfo->r);
	printf("     Log     : %d\n", aInfo->logp);
	printf("     H       : %d\n", aInfo->h);
}
void CHashtables::PrintHashtables()
{
	int i, j, max;
	hashtable_t * h = this->Hashtables;
	script_t* script;
	update_t *updates;
	printf("\n\nPrint Hashtables start\n\n");
	for (i = 0; i <= this->MaxIndex; i++, h++)
	{
		printf("\n\n HASHTABLE # %d: \n\n",i);
		script = h->script;
		int counter = 0;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (j=0; j< max; j++, updates++) 
			{
				printf("(%6d:) Prime %6d, h = %6d.\n",counter, updates->p,updates->h);
				counter++;
				
			}
			printf("----- list boundary -----\n");
			script = script->next;
		}
		printf("  This hashtable had %d entries.\n",counter);		
	}
	printf("\n\nPrint Hashtables end\n\n");
}
