#include <cstring>

#include "hermite_matrix_class.h"
#include "nfs_utils.h" // because of allocation machine
#include "polynom_class.h"
#include "const_return_codes.h"
#include "log.h"


/*!
	The default constructor does not take any parameters and constructs an instance of a "generic" normal matrix,
	with unknown dimensions. The dimensions may be later set by appropriate setter methods.

*/
HermiteMatrix::HermiteMatrix() : MSerializable(FILE_NAME_HERMITE_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->matrix = NULL;
	mpz_init_set_ui(this->denominator,1);
}


/*!
	The second constructor constructs an instance of a normal matrix with known dimensions. Beware that the parameters
	taken mean the actual number of rows and columns, and not their maximal indices; so, if we want to construct
	a matrix of dimensions 17x32, we call
	
		HermiteMatrix* nm = new HermiteMatrix(17,32);
		
	Now, we have an instance of a normal matrix; its member variables will be set to:
	
		nm->maximal_row_index = 16;
		
		nm->maximal_column_index = 31;
		
		nm->maximal_allocated_row_index = -1; // no initialization of the data array yet!
		
		nm->maximal_allocated_column_index = -1; // no initialization of the data array yet!
		
	Both of the constructors initialize the data array to NULL pointer, and a real allocation is performed later	- at the time of need. This programming pattern is called lazy initialization, and helps to reduce runtime memory requirements.

*/
HermiteMatrix::HermiteMatrix(long aRows, long aColumns) : MSerializable(FILE_NAME_HERMITE_MATRIX)
{
	this->max_allocated_row_index = -1;
	this->max_allocated_column_index = -1;
	this->maximal_row_index = aRows - 1;
	this->maximal_column_index = aColumns - 1;
	this->matrix = NULL;
	mpz_init_set_ui(this->denominator,1);
}


/*!

	The destructor performs "cleaning up", in this case deallocation of the data array. Its decisions to deallocate
	are based on maximal_allocated_row_index, so it is safe to call it twice. However, I do not see any reason to
	call destructor explicitely; just use the standard C++ pattern
	
		delete nm;
		
	which invConstRC::OKes the destructor implicitely.

*/
HermiteMatrix::~HermiteMatrix(){
	if (max_allocated_row_index >= 0){
		long i,j;
		for (i=0; i<=max_allocated_row_index; i++){
			for(j=0;j<=max_allocated_column_index;j++){
				mpz_clear(matrix[i][j]);
			}
			free(this->matrix[i]);
			this->matrix[i] = NULL;
		}
		max_allocated_row_index = -1;
		free(this->matrix);
		this->matrix = NULL;
		mpz_clear(this->denominator);
	}
}

/*!
	  This method checks whether the calling instance has already been allocated; it uses maximal_allocated_row_index
	  to determine this. If it has not been allocated, but the maximal_row_index is set to 0 or more (that means:
	  if dimensions of this matrix have already been determined), it allocates the rows and columns of the matrix.
	  
	  When allocating, all entries are set to be zero.
	  This function is used to ensure proper allocation of target (and other) matrices in arithmetic functions.
	  Return codes:
	  <table border=0>
	  	<tr><td> ConstRC::Ok 	</td><td>	   everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> there was not enough memory to allocate the requested space.</td></tr>
	  </table>

*/
int  HermiteMatrix::Allocate(){
	if ((this->maximal_row_index >= 0) && (this->max_allocated_row_index == -1)){
		long i,j;
		this->matrix = (mpz_t **) Utils::AllocationMachine((void **) &this->matrix,(this->maximal_row_index+1)*sizeof(mpz_t*));
		if (!this->matrix){
			CLog::LogError(WHEREARG,"HermiteMatrixClass","error allocating rows",(this->maximal_row_index+1),200,"There was not enough memory to allocate normal matrix rows");
			return ConstRC::NotEnoughMemory;
		}
		memset(this->matrix,0,(this->maximal_row_index+1)*sizeof(mpz_t*));
		this->max_allocated_row_index = this->maximal_row_index;
		this->max_allocated_column_index = this->maximal_column_index;
		for (i=0; i<= maximal_row_index; i++){
			this->matrix[i] = (mpz_t *) Utils::AllocationMachine((void **) &(this->matrix[i]),(this->max_allocated_column_index+1)*sizeof(mpz_t));
			if (!this->matrix[i]){
				CLog::LogError(WHEREARG,"HermiteMatrixClass","error allocating columns",(this->max_allocated_column_index+1),201,"There was not enough memory to allocate normal matrix columns");
				return ConstRC::NotEnoughMemory;
			}
			for(j=0;j<=maximal_column_index;j++) mpz_init(this->matrix[i][j]);
		}
	}
	return ConstRC::Ok;
}

HermiteMatrix * HermiteMatrix::clone() {
	HermiteMatrix * new_matrix = new HermiteMatrix();
	return new_matrix;
}

/*!
	The getter method GetXY returns the value on aRow and aColumn indices. It only works if the matrix** array
	has already been allocated. If aRow and/or aColumn exceeds max_allocated_row_index or max_allocated_column_index,
	or if they are below zero, the return value is 
		ConstRC::GeneralError
	If the value does not fit in the long int, it is truncated.
*/

long HermiteMatrix::GetXY(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	return mpz_get_si(this->matrix[aRow][aColumn]);
}

/*!
	The getter method GetXY2 returns the value on aRow and aColumn indices. It only works if the matrix** array
	has already been allocated. If aRow and/or aColumn exceeds max_allocated_row_index or max_allocated_column_index,
	or if they are below zero, the return value is NULL.
	The function actually returns a pointer on the value because the multiprecision integer type cannot be thrown from a function.
*/

mpz_t *HermiteMatrix::GetXY2(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return NULL;
	if (aRow > this->max_allocated_row_index)   return NULL;
	if (aColumn > this->max_allocated_column_index) return NULL;
	return (this->matrix[aRow]+aColumn);
}

/*!
	This function returns TRUE if the rational number on aRow and aColumn is equal to 1.
	Otherwise it returns FALSE.
*/

int HermiteMatrix::IsOne(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	return (mpz_cmp(this->matrix[aRow][aColumn],denominator));
}

/*!
	This function returns TRUE if the rational number on aRow and aColumn is equal to 0.
	Otherwise it returns FALSE.
*/

int HermiteMatrix::IsZero(long aRow, long aColumn){
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::GeneralError;
	if (aRow > this->max_allocated_row_index)   return ConstRC::GeneralError;
	if (aColumn > this->max_allocated_column_index) return ConstRC::GeneralError;
	return (mpz_cmp_ui(this->matrix[aRow][aColumn],0));
}

/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  matrix with pseudorandom values, gained from standard randomization in C - calling rand(). The seed is determined
	  at each call of Randomize() by the current clock() value.
	  
	  This method overwrites any previous elements in matrix data array.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>

*/
int  HermiteMatrix::Randomize(){
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok){
		return ret;
	}
	long i, j;
	long seed = (long) clock();
	srand(seed);

	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			mpz_set_si(this->matrix[i][j], (long)rand());
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method at first ensures allocation of the calling instance (by calling Allocate()). Then it switches its behaviour
	  according to the action taken by Allocate().
	     - If Allocate() really allocated the matrix, it is zeroized already; no need to zeroize it again.
	     - Otherwise the matrix is filled with zeros using memset().
	     
	  This method overwrites any previous elements in matrix data array and set the denominator to be 1.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td> ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>

*/
int  HermiteMatrix::Zeroize(){
	long i = this->max_allocated_row_index;
	long j = this->max_allocated_column_index;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	// This strangely loConstRC::OKing condition in fact checks whether Allocate() was needed or no. In case that
	// Allocate() run the allocation job, it changed values of this->max_allocated_column_index and
	// this->max_allocated_row_index. Zeroizing is a part of allocation job; we do not want to perform
	// it twice. So we run this zeroizing if and only if it has not been already run by Allocate()
	if ((this->max_allocated_row_index == i) && (this->max_allocated_column_index == j)){
		// zeroizing is needed
		for (i=0; i<= this->max_allocated_row_index; i++){
		    for(j=0; j<=this->max_allocated_column_index; j++) mpz_set_ui(matrix[i][j],0);
		}
	}
	mpz_set_ui(this->denominator,1);
	return ConstRC::Ok;
}

/*!
	  This method performs comparison of the calling instance and of aMatrix. It does not call Allocate() to ensure 
	  allocation of anything.
	  
	  The principles for equality are the following:
	  	- if aMatrix is NULL then the matrices are not equal
		- if the respective dimensions maximal_row_index and maximal_column_index differ then the matrices are not equal
		- if the allocated arrays are not of the same size then the matrices are not equal (discutable)
		- if any entry on any position differs then the matrices are not equal.
		- if the denominators are not equal then the matrices are not equal.
		
	  The comparison ends as soon as any of the previous conditions is met.
	  
	  This method does not change any data of any operand.
	  
	  Return codes:
	  <table border=0>
	  	<tr><td>TRUE 		</td><td>    the matrices are equal</td></tr>
		<tr><td>FALSE		</td><td>    the matrices are not equal</td></tr>
	  </table>

*/
bool HermiteMatrix::Equals(HermiteMatrix* aMatrix){
	if (!aMatrix)		return FALSE; // ConstRC::NullPointerSupplied
	// performs a "true" comparison of two matrices. They must be equal up to allocation (have the same number of
	// elements allocated
	if (this->max_allocated_row_index != aMatrix->max_allocated_row_index)		return FALSE;
	if (this->max_allocated_column_index != aMatrix->max_allocated_column_index)	return FALSE;
	if (this->maximal_row_index != aMatrix->maximal_row_index)			return ConstRC::GeneralError;
	if (this->maximal_column_index != aMatrix->maximal_column_index)		return ConstRC::GeneralError;
	long i,j;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			if (mpz_cmp(this->matrix[i][j] , aMatrix->matrix[i][j]))	return ConstRC::GeneralError;
		}
	}
	return (mpz_cmp(this->denominator,aMatrix->denominator) == 0);
}

/*!
	  This method prints the calling instance onto the screen. It is suitable for printing matrices with entries smaller
	  than 8 decimal digits.
	  
	  This method does not change any data.

*/
void HermiteMatrix::PrintToScreen(){
	long i, j;
	printf("\n");
	gmp_printf("Hermite matrix: %d x %d divided by the number %Zd\n",this->maximal_row_index+1,this->maximal_column_index+1,this->denominator);
	for (i=0; i<=this->max_allocated_row_index; i++){
		printf("\nRow %4ld:  ",i);
		for (j=0; j<=this->max_allocated_column_index; j++){
			gmp_printf("%8Zd ",this->matrix[i][j]);
		}
	}
	printf("\n\n");
}

/*!	This method sets the rational number on aRow and aColumn to be 1
	(more precisely denominator/denominator).
	The method can detect row- and column- overflow and underflow; it
	returns error values
		ConstRC::ExcessiveRowIndex,
		ConstRC::ExcessiveColumnIndex,
		ConstRC::NegativeIndex,
	defined in definitions.h.
	
	If operation has been done, they return value
		ConstRC::OK
	defined in definitions.h
*/

int  HermiteMatrix::PutOne(long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	mpz_set(this->matrix[aRow][aColumn], denominator);
	return ConstRC::Ok;
}

/*!	This method sets the rational number on aRow and aColumn to be 0.
	The method can detect row- and column- overflow and underflow; it
	returns error values
		ConstRC::ExcessiveRowIndex,
		ConstRC::ExcessiveColumnIndex,
		ConstRC::NegativeIndex,
	defined in definitions.h.
	
	If operation has been done, they return value
		ConstRC::Ok
	defined in definitions.h
*/

int  HermiteMatrix::PutZero(long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	mpz_set_ui(this->matrix[aRow][aColumn], 0);
	return ConstRC::Ok;
}

/*!	This method puts the number aNumber on aRow and aColumn, regardlessly on the denominator.
	The method can detect row- and column- overflow and underflow; it
	returns error values
		ConstRC::ExcessiveRowIndex,
		ConstRC::ExcessiveColumnIndex,
		ConstRC::NegativeIndex,
	defined in definitions.h.
	
	If operation has been done, they return value
		ConstRC::Ok
	defined in definitions.h
*/

int  HermiteMatrix::PutNumber(mpz_t aNumber, long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	mpz_set(this->matrix[aRow][aColumn], aNumber);
	return ConstRC::Ok;
}

/*!	This method puts the number aNumber on aRow and aColumn, regardlessly on the denominator.
	The method can detect row- and column- overflow and underflow; it
	returns error values
		ConstRC::ExcessiveRowIndex,
		ConstRC::ExcessiveColumnIndex,
		ConstRC::NegativeIndex,
	defined in definitions.h.
	
	If operation has been done, they return value
		ConstRC::Ok
	defined in definitions.h
*/

int  HermiteMatrix::PutNumber(long aNumber, long aRow, long aColumn){
	int ret;
	if ((aRow < 0) || (aColumn < 0))	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aColumn > this->maximal_column_index)   return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	mpz_set_si(this->matrix[aRow][aColumn], aNumber);
	return ConstRC::Ok;
}

/*!
	  This method recalculates the current data of the calling matrix modulo argument. In an instantiated & unallocated
	  matrix it does simply nothing.
	  
	  So far, this method has no return codes. If aModulus < 2, nothing is done and the method returns immediately.

*/
void HermiteMatrix::Modulo(long aModulus){
	long i,j;
	if (aModulus < 2) return;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			mpz_fdiv_r_ui(this->matrix[i][j],this->matrix[i][j], aModulus);
		}
	}
}

/*!
	  This method recalculates the current data of the calling matrix modulo argument. In an instantiated & unallocated
	  matrix it does simply nothing.
	  
	  So far, this method has no return codes. If aModulus < 2, nothing is done and the method returns immediately.

*/

void HermiteMatrix::Modulo(mpz_t aModulus){
	long i,j;
	if (mpz_cmp_ui(aModulus,2)<0) return;
	for (i=0; i<= this->max_allocated_row_index; i++){
		for (j=0; j<= this->max_allocated_column_index; j++){
			mpz_fdiv_r(this->matrix[i][j],this->matrix[i][j], aModulus);
		}
	}
}

/*!
	This methods performs copying of the matrix aSource into the called matrix.

	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::SizeMismatch </td><td>if some sizes are not equal</td></tr>
	  </table>
*/

int  HermiteMatrix::Copy(HermiteMatrix* aSource){
        if (this->maximal_row_index != aSource->maximal_row_index)              return ConstRC::SizeMismatch;
        if (this->maximal_column_index != aSource->maximal_column_index)        return ConstRC::SizeMismatch;
        int ret;
        ret = this->Allocate();
        if (ret != ConstRC::Ok){
                return ret;
        }
        long i, j;
        for (i=0; i<= this->max_allocated_row_index; i++){
                for (j=0; j<= this->max_allocated_column_index; j++){
                        mpz_set(this->matrix[i][j], aSource->matrix[i][j]);
                }
        }
	mpz_set(this->denominator,aSource->denominator);
        return ConstRC::Ok;
}

/*!	 
	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  row at given index aRow with zeros (using memset).
	  
	  Checking whether 0 <= aRow <= maximal_row_index is enabled.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if aRow > maximal_row_index</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td> if aRow < 0</td></tr>
	  </table>

*/

int  HermiteMatrix::ZeroizeRow(long aRow){
	int ret;
	if (aRow < 0)	return ConstRC::NegativeIndex;
	if (aRow > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i;
	for(i=0; i<=maximal_column_index; i++)
	    mpz_set_ui(matrix[aRow][i],0);
	return ConstRC::Ok;
}

/*!
	  This method reads numbers from aRowList array[0 ... aListMaxIndex] and zeroes out rows with those indices.
	  
	  Before this, it checks whether the aRowList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). If aListMaxIndex < 0, an error is returned.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		</td><td>  everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::ExcessiveRowIndex</td><td>if any of the indices in aRowList > maximal_row_index</td></tr>
		<tr><td>ConstRC::NullPointerSupplied</td><td> if aRowList is NULL</td></tr>
		<tr><td>ConstRC::NegativeIndex    </td><td>  if aListMaxIndex < 0 or any of the indices om aRowList is negative</td></tr>
	  </table>

*/
int  HermiteMatrix::ZeroizeRow(long* aRowList, long aListMaxIndex){
	if (!aRowList)		return ConstRC::NullPointerSupplied;
	if (aListMaxIndex < 0)  return ConstRC::NegativeIndex;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i,j;
	long index;
	for (j=0; j<= aListMaxIndex; j++){
		index = aRowList[j];
		if (index < 0)					     return ConstRC::NegativeIndex;
		if (index > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
		for(i=0; i<=maximal_column_index; i++)
		    mpz_set_ui(matrix[index][i],0);
	}
	return ConstRC::Ok;
}

/*!
	  This method adds row with index aSource to the row with index aTarget. The indices may be equal, in which
	  case it just multiplies the contents of the row by 2.
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are between 0 and maximal_row_index.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveRowIndex</td><td>if aTarget or aSource > maximal_row_index</td></tr>
		<tr><td> ConstRC::NegativeIndex    </td><td> if aTarget or aSource < 0</td></tr>
	  </table>

*/
int  HermiteMatrix::AddRows(long aTarget, long aSource){
	int ret;
	if ((aTarget < 0) || (aSource < 0))		     return ConstRC::NegativeIndex;
	if (aTarget > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aSource > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	for (k=0; k<=this->max_allocated_column_index; k++){
		mpz_add(this->matrix[aTarget][k],this->matrix[aTarget][k],this->matrix[aSource][k]);
	}
	return ConstRC::Ok;
}

/*!
	  This method swaps the row with index aRow1 with the row with index aRow2. The indices may be equal, in which
	  case nothing happens. The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are between 0 and maximal_row_index.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveRowIndex</td><td>if aRow1 or aRow2 > maximal_row_index</td></tr>
		<tr><td> ConstRC::NegativeIndex    </td><td> if aRow1 or aRow2 < 0</td></tr>
	  </table>

*/

int  HermiteMatrix::SwapRows(long aRow1, long aRow2){
	int ret;
	if ((aRow1 < 0) || (aRow2 < 0))		     return ConstRC::NegativeIndex;
	if (aRow1 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aRow2 > this->maximal_row_index)    	     return ConstRC::ExcessiveRowIndex;
	if (aRow1==aRow2)	return ConstRC::Ok;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	mpz_t aux;
	mpz_init(aux);
	for (k=0; k<=this->max_allocated_column_index; k++){
		mpz_set(aux,this->matrix[aRow1][k]);
		mpz_set(this->matrix[aRow1][k],this->matrix[aRow2][k]);
		mpz_set(this->matrix[aRow2][k],aux);
	}
	mpz_clear(aux);
	return ConstRC::Ok;
}

/*!
	  This method swaps the column with index acolumn1 with the column with index aColumn2. The indices may be equal, in which
	  case nothing happens. The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are between 0 and maximal_row_index.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveColumnIndex</td><td>if aColumn1 or aColumn2 > maximal_column_index</td></tr>
		<tr><td> ConstRC::NegativeIndex    </td><td> if aColumn1 or aColumn2 < 0</td></tr>
	  </table>

*/

int  HermiteMatrix::SwapColumns(long aColumn1, long aColumn2){
	int ret;
	if ((aColumn2 < 0) || (aColumn1 < 0))		     return ConstRC::NegativeIndex;
	if (aColumn2 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aColumn1 > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aColumn1==aColumn2)	return ConstRC::Ok;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	mpz_t aux;
	mpz_init(aux);
	for (k=0; k<=this->max_allocated_row_index; k++){
		mpz_set(aux,this->matrix[k][aColumn1]);
		mpz_set(this->matrix[k][aColumn1],this->matrix[k][aColumn2]);
		mpz_set(this->matrix[k][aColumn2],aux);
	}
	mpz_clear(aux);
	return ConstRC::Ok;
}

/*!
	  This method adds to the row with index aRow the row with index aRow2 of the matrix aOperand.
	  The method ensures the allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		 </td><td>  everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveRowIndex</td><td>if aTarget or aSource > maximal_row_index</td></tr>
		<tr><td> ConstRC::NegativeIndex    </td><td> if aTarget or aSource < 0</td></tr>
		<tr><td> ConstRC::GeneralError	  </td><td> if aOperand2 is not a HermiteMatrix</td></tr>
	  </table>

*/
int  HermiteMatrix::AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2){
        int ret;
        HermiteMatrix *aO;
        if ((aRow < 0) || (aRow2 < 0))                               return ConstRC::NegativeIndex;
        if (aRow > this->maximal_row_index)                  return ConstRC::ExcessiveRowIndex;
        if (aRow2 > aOperand->GetMaximalRowIndex())                 return ConstRC::ExcessiveRowIndex;
        aO = (HermiteMatrix*) aOperand;
        ret = aO->Allocate();
        if (ret != ConstRC::Ok)  return ret;
        long k;
        for (k=0; k<=this->maximal_column_index; k++){
                mpz_add(this->matrix[aRow][k],this->matrix[aRow][k],aO->matrix[aRow2][k]);
        }
        return ConstRC::Ok;
}

/*!

	  This method at first ensures allocation of the calling instance (by calling Allocate()), and then fills the
	  column at given index aColumn with zeros.
	  The method checks whether aColumn is between 0 and maximal_column_index.	  
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveColumnIndex</td><td>if aColumn > maximal_column_index</td></tr>
		<tr><td> ConstRC::NegativeIndex    </td><td> if aColumn < 0</td></tr>
	  </table>

*/
int  HermiteMatrix::ZeroizeColumn(long aColumn){
	int ret;
	if (aColumn < 0)    	     		return ConstRC::NegativeIndex;
	if (aColumn > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i;
	for (i=0; i<=this->max_allocated_row_index; i++){
		mpz_set_ui(this->matrix[i][aColumn],0);
	}
	return ConstRC::Ok;
}

/*!
This method reads numbers from aColumnList array[0 ... aListMaxIndex] and zeroes out columns with those indices.
	  Before this, it checks whether the aColumnList is not NULL, and ensures allocation of the calling instance 
	  (by calling Allocate()). 
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveColumnIndex</td><td>if any column index in list > maximal_column_index</td></tr>
		<tr><td> ConstRC::NullPointerSupplied </td><td> if aColumnList is NULL</td></tr>
		<tr><td> ConstRC::NegativeIndex	</td><td> if any column index in list < 0 or aListMaxIndex < 0.</td></tr>
	  </table>
	
*/
int  HermiteMatrix::ZeroizeColumn(long* aColumnList, long aListMaxIndex){
	if (!aColumnList)		return ConstRC::NullPointerSupplied;
	if (aListMaxIndex < 0)		return ConstRC::NegativeIndex;
	int ret;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long i, j, index;
	for (j=0; j<= aListMaxIndex; j++){
		index = aColumnList[j];
		if (index < 0)					     return ConstRC::NegativeIndex;
		if (index > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
		for (i=0; i<=this->max_allocated_row_index; i++){
			mpz_set_ui(this->matrix[i][index],0);
		}
	}
	return ConstRC::Ok;
}

/*!
	  This method adds column with index aSource to the column with index aTarget.	  
	  The method ensures allocation of the calling instance (by calling Allocate()). It checks
	  whether the indices are legal.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		</td><td>   everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::ExcessiveColumnIndex</td><td>if aTarget or aSource > maximal_column_index</td></tr>
		<tr><td> ConstRC::NegativeIndex	</td><td> if aTarget or aSource < 0</td></tr>
	  </table>

*/
int  HermiteMatrix::AddColumns(long aTarget, long aSource){
	int ret;
	if ((aTarget < 0) || (aSource < 0))		     return ConstRC::NegativeIndex;
	if (aTarget > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	if (aSource > this->maximal_column_index)    	     return ConstRC::ExcessiveColumnIndex;
	ret = this->Allocate();
	if (ret != ConstRC::Ok)	return ret;
	long k;
	for (k=0; k<=this->max_allocated_row_index; k++){
		mpz_add(this->matrix[k][aTarget],this->matrix[k][aTarget],this->matrix[k][aSource]);
	}
	return ConstRC::Ok;
}

/*!
	  This method ensures allocation state, dimension requirements etc., and then transposes the calling instance into
	  the matrix aTarget. The calling instance must NOT be equal to aTarget.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td> ConstRC::Ok		 </td><td>  everything all right</td></tr>
		<tr><td> ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td> ConstRC::SizeMismatch     </td><td> the dimensions of the calling instance and aTarget do not match (M x N vs. N x M)</td></tr>
		<tr><td> ConstRC::NullPointerSupplied </td><td> if aTarget is NULL</td></tr>
		<tr><td> ConstRC::BadArgument	  </td><td> if the calling instance is equal to aTarget</td></tr>
	  </table>
*/

int  HermiteMatrix::Transpose(HermiteMatrix* aTarget){
	if (!aTarget)  							return ConstRC::NullPointerSupplied;
	if (this == aTarget)						return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	long i, j;
	for (i=0; i<=this->max_allocated_row_index; i++)
	{
		for (j=0; j<=this->max_allocated_column_index; j++){
			mpz_set(aTarget->matrix[j][i],this->matrix[i][j]);
		}
	}
	mpz_set(aTarget->denominator,this->denominator);
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the transposition operation, and then performs
	  the transposition by calling 
	  	int Transpose(HermiteMatrix* aTarget)
	  If result has been allocated, but Transpose did not finish well, the result is deleted again.
	  
	  Returns:
	   <table border=0>
	  	<tr><td>pointer to result of operation if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>

*/
HermiteMatrix* HermiteMatrix::Transpose(){
	HermiteMatrix* result = new HermiteMatrix(this->maximal_column_index+1, this->maximal_row_index+1);
	if (!result) return NULL;
	if (this->Transpose(result) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method ensures allocation state, dimension requirements etc., and then adds aOperand2 to the calling instance
	  and places the result into matrix aTarget. Both the calling instance and aOperand2 must NOT be equal to aTarget.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand2 or aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if aOperand2 == this or aTarget == this.</td></tr>
	  </table>

*/
int  HermiteMatrix::Add(HermiteMatrix* aTarget, HermiteMatrix* aOperand2){
	if (!aTarget)  								return ConstRC::NullPointerSupplied;
	if (aTarget->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	if (!aOperand2)								return ConstRC::NullPointerSupplied;
	if (aOperand2->maximal_row_index != this->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aOperand2->maximal_column_index != this->maximal_column_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j;
	mpz_t a,b,aux;
	mpz_init(a);
	mpz_init(b);
	mpz_init(aux);
	ret = aTarget->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	mpz_lcm(aTarget->denominator,this->denominator,aOperand2->denominator);
	mpz_divexact(a,aTarget->denominator,this->denominator);
	mpz_divexact(b,aTarget->denominator,aOperand2->denominator);
	for (i=0; i<=aTarget->max_allocated_row_index;i++){
		for (j=0; j<= aTarget->max_allocated_column_index;j++){
			mpz_mul(aux,b,aOperand2->matrix[i][j]);
			mpz_mul(aTarget->matrix[i][j],a,this->matrix[i][j]);
			mpz_add(aTarget->matrix[i][j],aTarget->matrix[i][j],aux);
		}
	}
	mpz_clear(a);
	mpz_clear(b);
	mpz_clear(aux);
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the addition operation, and then performs
	  the addition by calling 
	  	int Add(HermiteMatrix* aTarget)
	  If result has been allocated, but Add did not finish well, the result is deleted again.
	  
	  Returns:
	   <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>

*/
HermiteMatrix* HermiteMatrix::Add(HermiteMatrix* aOperand2){
	HermiteMatrix* result = new HermiteMatrix(this->maximal_row_index+1, this->maximal_column_index+1);
	if (!result) return NULL;
	if (this->Add(result,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand1, aOperand2 or aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if aOperand1 or aOperand2 == aTarget.</td></tr>
	  </table>

*/
int  HermiteMatrix::MultiplyInternal(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2){
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand1) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_column_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k;
	mpz_t intermediate;
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	mpz_init(intermediate);
	for (i=0; i<= aTarget->maximal_row_index; i++){
		for (j=0; j<= aTarget->maximal_column_index; j++){
			for (k=0; k<= aOperand1->maximal_column_index; k++){
				mpz_mul(intermediate,aOperand1->matrix[i][k],aOperand2->matrix[k][j]);
				mpz_add(aTarget->matrix[i][j],aTarget->matrix[i][j],intermediate);
			}
		}
	}
	mpz_clear(intermediate);
	mpz_mul(aTarget->denominator,aOperand1->denominator,aOperand2->denominator);
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1^T x aOperand2, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand1, aOperand2 or aTarget is NULL</td></tr>
		<tr><td>ConstRC::BadArgument	  </td><td> if aOperand1 or aOperand2 == aTarget.</td></tr>
	  </table>

*/
int  HermiteMatrix::MultiplyInternalTransposed(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2){
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand1) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_column_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_row_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k;
	mpz_t intermediate;
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	mpz_init(intermediate);
	for (i=0; i<= aTarget->maximal_row_index; i++){
		for (j=0; j<= aTarget->maximal_column_index; j++){
			for (k=0; k<= aOperand1->maximal_column_index; k++){
				mpz_mul(intermediate,aOperand1->matrix[k][i],aOperand2->matrix[k][j]);
				mpz_add(aTarget->matrix[i][j],aTarget->matrix[i][j],intermediate);
			}
		}
	}
	mpz_clear(intermediate);
	mpz_mul(aTarget->denominator,aOperand1->denominator,aOperand2->denominator);
	return ConstRC::Ok;
}

/*!
	  This method ensures that all input parameters are allocated, checks their dimensions, and then performs
	  multiplication aOperand1 x (aOperand2)^T, saving the result into aTarget. Both aOperand1 and aOperand2 
	  must NOT be equal to aTarget.
	  Return codes:
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> if aOperand1, aOperand2 or aTarget is NULL</td></tr>
	  </table>
*/

int  HermiteMatrix::MultiplyInternalWithTransposition(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2){
	if (!aOperand1 || !aOperand2 || !aTarget)   				return ConstRC::NullPointerSupplied;
	if ((aTarget == aOperand1) || (aTarget == aOperand2))			return ConstRC::BadArgument;
	if (aTarget->maximal_row_index != aOperand1->maximal_row_index)		return ConstRC::SizeMismatch;
	if (aTarget->maximal_column_index != aOperand2->maximal_row_index)	return ConstRC::SizeMismatch;
	if (aOperand1->maximal_column_index != aOperand2->maximal_column_index)	return ConstRC::SizeMismatch;
	int ret;
	long i,j,k;
	mpz_t intermediate;
	ret = aTarget->Zeroize();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand1->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	ret = aOperand2->Allocate();
	if (ret != ConstRC::Ok)		return ret;
	mpz_init(intermediate);
	for (i=0; i<= aTarget->maximal_row_index; i++){
		for (j=0; j<= aTarget->maximal_column_index; j++){
			for (k=0; k<= aOperand1->maximal_column_index; k++){
				mpz_mul(intermediate,aOperand1->matrix[i][k],aOperand2->matrix[j][k]);
				mpz_add(aTarget->matrix[i][j],aTarget->matrix[i][j],intermediate);
			}
		}
	}
	mpz_clear(intermediate);
	mpz_mul(aTarget->denominator,aOperand1->denominator,aOperand2->denominator);
	return ConstRC::Ok;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternal(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.
	  
	  Returns:
	   <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>

*/
HermiteMatrix* HermiteMatrix::MultiplyInternal(HermiteMatrix* aOperand1, HermiteMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	HermiteMatrix* result = new HermiteMatrix(aOperand1->maximal_row_index+1, aOperand2->maximal_column_index+1);
	if (!result) return NULL;
	if (MultiplyInternal(result,aOperand1,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method allocates a new matrix for placement of the result of the multiplication operation, and then performs
	  the multiplication by calling 
	  	int MultiplyInternalTransposed(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2)
	  If result has been allocated, but MultiplyInternal did not finish well, the result is deleted.
	  
	  Returns:
	   <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>

*/
HermiteMatrix* HermiteMatrix::MultiplyInternalTransposed(HermiteMatrix* aOperand1, HermiteMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	HermiteMatrix* result = new HermiteMatrix(aOperand1->maximal_column_index+1, aOperand2->maximal_column_index+1);
	if (!result) return NULL;
	if (MultiplyInternalTransposed(result,aOperand1,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;
}

/*!
	  This method allocates a new matrix for placement of the result of the operation, and then performs the transposed
	  multiplication by calling 
	  
	  	int MultiplyInternalWithTransposition(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2)
		
	  If result has been allocated, but MultiplyInternalWT did not finish well, the result is deleted.
	  
	  Returns:
	   <table border=0>
	  	<tr><td>pointer to result of operation, if everything went all right</td></tr>
		<tr><td>NULL in case of any problem (size mismatches, allocation problems)</td></tr>
	  </table>

*/
HermiteMatrix* HermiteMatrix::MultiplyInternalWithTransposition(HermiteMatrix* aOperand1, HermiteMatrix* aOperand2){
	if (!aOperand1 || !aOperand2)   return NULL;
	HermiteMatrix* result = new HermiteMatrix(aOperand1->maximal_row_index+1, aOperand2->maximal_row_index+1);
	if (!result) return NULL;
	if (MultiplyInternalWithTransposition(result,aOperand1,aOperand2) != ConstRC::Ok) {
	 	delete result;
	 	return NULL;
	} 
	return result;	
}

/*!
	Returns the inverse matrix of the matrix which is supposed to be upper triangular.
	The denominator of the source is ignored. The function checks sizes of matrices

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch     </td><td> the dimensions do not match</td></tr>
	  </table>
*/

int HermiteMatrix::InverseFromTriangular(HermiteMatrix *aTarget){
    if (this->maximal_column_index!=this->maximal_row_index) return ConstRC::SizeMismatch;
    if (aTarget->maximal_row_index!=this->maximal_row_index) return ConstRC::SizeMismatch;
    if (aTarget->maximal_column_index!=this->maximal_column_index) return ConstRC::SizeMismatch;
    long i,j,k;
    int ret;
    ret = aTarget->Zeroize();
    if (ret!=ConstRC::Ok) return ret;
    mpz_t aux;
    mpz_init(aux);
    for(i=0;i<=maximal_column_index;i++) mpz_mul(aTarget->denominator,aTarget->denominator,matrix[i][i]);
    for(j=maximal_column_index;j>=0;j--){
	mpz_divexact(aTarget->matrix[j][j],aTarget->denominator,matrix[j][j]);
	for(i=j-1;i>=0;i--){
	    for(k=j;k<=maximal_column_index;k++){
		mpz_mul(aux,aTarget->matrix[j][k],matrix[i][j]);
		mpz_divexact(aux,aux,matrix[i][i]);
		mpz_sub(aTarget->matrix[i][k],aTarget->matrix[i][k],aux);
	    }
	}
    }
    mpz_clear(aux);
    return ConstRC::Ok;
}

/*!
	Converts the matrix into the Hermite normal form.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>
*/

int HermiteMatrix::ToHermiteNormalForm(){
    int ret = Allocate();
    if (ret!= ConstRC::Ok) return ret;
    long i,j,k,c=maximal_column_index;
    mpz_t a,b,d,e,f,aux,aux2;
    mpz_init(a);
    mpz_init(b);
    mpz_init(d);
    mpz_init(e);
    mpz_init(f);
    mpz_init(aux);
    mpz_init(aux2);
    for(i=maximal_row_index;i>=0;i--){
	for(j=0;j<c;j++) if (mpz_cmp_ui(matrix[i][j],0)) {
	    if(!mpz_cmp_ui(matrix[i][j+1],0)) SwapColumns(j,j+1);
	    else {
		mpz_gcdext(d,a,b,matrix[i][j],matrix[i][j+1]);
		mpz_divexact(e,matrix[i][j+1],d);
		mpz_divexact(f,matrix[i][j],d);
		for(k=0;k<=i;k++){
		    mpz_set(aux,matrix[k][j]);
		    mpz_mul(aux2,matrix[k][j+1],f);
		    mpz_mul(matrix[k][j],matrix[k][j],e);
		    mpz_sub(matrix[k][j],matrix[k][j],aux2);
		    mpz_mul(aux2,matrix[k][j+1],b);
		    mpz_mul(matrix[k][j+1],aux,a);
		    mpz_add(matrix[k][j+1],matrix[k][j+1],aux2);
		}
	    }
	}
	if (mpz_cmp_ui(matrix[i][c],0)){
	    if(mpz_cmp_ui(matrix[i][c],0)<0) {
		for(k=0;k<=i;k++) mpz_neg(matrix[k][c],matrix[k][c]);
	    }
	    for(j=c+1;j<=maximal_column_index;j++){
		mpz_fdiv_q(d,matrix[i][j],matrix[i][c]);
		for(k=0;k<=i;k++){
		    mpz_mul(aux,matrix[k][c],d);
		    mpz_sub(matrix[k][j],matrix[k][j],aux); 
		}
	    }
	    c--;
	}
    }
    mpz_clear(a);
    mpz_clear(b);
    mpz_clear(d);
    mpz_clear(e);
    mpz_clear(f);
    mpz_clear(aux);
    mpz_clear(aux2);
    return ConstRC::Ok;
}

/*!
	Converts the matrix into the Hermite normal form when a multiple of the determinant is known.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
	  </table>
*/

int HermiteMatrix::ToHermiteNormalForm(mpz_t aModulus){
    int ret = Allocate();
    if (ret!= ConstRC::Ok) return ret;
    long i,j,k,c=maximal_column_index;
    mpz_t a,b,d,e,f,aux,aux2,R;
    mpz_init(a);
    mpz_init(b);
    mpz_init(d);
    mpz_init(e);
    mpz_init(f);
    mpz_init(aux);
    mpz_init(aux2);
    if(mpz_cmp_ui(aModulus,0)<0){
	mpz_init(R);
	mpz_neg(R,aModulus);
    } else
	mpz_init_set(R,aModulus);
    for(i=maximal_row_index;i>=0;i--){
	if (!mpz_cmp_ui(matrix[i][c],0)) mpz_set(matrix[i][c],R);
	for(j=0;j<c;j++) if (mpz_cmp_ui(matrix[i][j],0)) {
	    if(!mpz_cmp_ui(matrix[i][j+1],0)) SwapColumns(j,j+1);
	    else {
		mpz_gcdext(d,a,b,matrix[i][j],matrix[i][j+1]);
		mpz_divexact(e,matrix[i][j+1],d);
		mpz_divexact(f,matrix[i][j],d);
		for(k=0;k<=i;k++){
		    mpz_set(aux,matrix[k][j]);
		    mpz_mul(aux2,matrix[k][j+1],f);
		    mpz_mul(matrix[k][j],matrix[k][j],e);
		    mpz_sub(matrix[k][j],matrix[k][j],aux2);
		    mpz_fdiv_r(matrix[k][j],matrix[k][j],R);
		    mpz_mul(aux2,matrix[k][j+1],b);
		    mpz_mul(matrix[k][j+1],aux,a);
		    mpz_add(matrix[k][j+1],matrix[k][j+1],aux2);
		    mpz_fdiv_r(matrix[k][j+1],matrix[k][j+1],R);
		}
	    }
//	    PrintToScreen();
	}
	mpz_gcdext(d,a,b,matrix[i][c],R);
	if (!mpz_cmp_ui(a,0)) mpz_set_ui(a,1);
        for(k=0;k<=i;k++){
//	    gmp_printf("(%Zd * %Zd)mod %Zd=",a,matrix[i][k],R);
	    mpz_mul(matrix[k][c],a,matrix[k][c]);
	    mpz_fdiv_r(matrix[k][c],matrix[k][c],R);
//	    gmp_printf("%Zd\n",matrix[i][k]);
	}
//	    PrintToScreen();
	if (!mpz_cmp_ui(matrix[i][c],0)) mpz_set(matrix[i][c],R);
        for(j=c+1;j<=maximal_column_index;j++){
	    mpz_fdiv_q(e,matrix[i][j],matrix[i][c]);
	    for(k=0;k<=i;k++){
		mpz_mul(aux,matrix[k][c],e);
		mpz_sub(matrix[k][j],matrix[k][j],aux);
		mpz_fdiv_r(matrix[k][j],matrix[k][j],R);
	   }
        }
        c--;
        mpz_divexact(R,R,d);
    }
    mpz_clear(a);
    mpz_clear(b);
    mpz_clear(d);
    mpz_clear(e);
    mpz_clear(f);
    mpz_clear(aux);
    mpz_clear(aux2);
    mpz_clear(R);
    return ConstRC::Ok;
}

/*!
	Divides all the coefficients and the denominator by a common factor.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::GeneralError </td><td> division by zero</td></tr>
	  </table>
*/

int HermiteMatrix::Reduce(mpz_t aFactor){
    if (!mpz_cmp_ui(aFactor,0)) return ConstRC::GeneralError;
    long i,j;
    for(i=0;i<=maximal_row_index;i++)
	for(j=0;j<=maximal_column_index;j++)
	    mpz_divexact(matrix[i][j],matrix[i][j],aFactor);
    mpz_divexact(denominator,denominator,aFactor);
    return ConstRC::Ok;
}

/*!
	Divides all the coefficients and the denominator by a common factor.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::GeneralError </td><td> division by zero</td></tr>
	  </table>
*/

int HermiteMatrix::Reduce(long aFactor){
    if (!aFactor) return ConstRC::GeneralError;
    long i,j;
    for(i=0;i<=maximal_row_index;i++)
	for(j=0;j<=maximal_column_index;j++)
	    mpz_divexact_ui(matrix[i][j],matrix[i][j],aFactor);
    mpz_divexact_ui(denominator,denominator,aFactor);
    return ConstRC::Ok;
}

/*!
	Multiplies the matrix by a column vector from the right. The vectors are supposed to be
	sufficiently allocated.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> a vector is NULL</td></tr>
	  </table>
*/

int HermiteMatrix::MultiplyVector(mpz_t *aTarget, mpz_t *aSource){
    long i,j;
    if ((!aTarget)||(!aSource)) return ConstRC::NullPointerSupplied;
    mpz_t c;
    mpz_init(c);
    for(i=0;i<=maximal_row_index;i++){
	mpz_set_ui(aTarget[i],0);
	for(j=0;j<=maximal_column_index;j++){
	    if(mpz_cmp_ui(aSource[j],0)){
		mpz_mul(c,matrix[i][j],aSource[j]);
		mpz_add(aTarget[i],aTarget[i],c);
	    }
	}
    }
    mpz_clear(c);
    return ConstRC::Ok;
}

/*!
	Multiplies the matrix by a vector from the left. The vectors are supposed to be
	sufficiently allocated.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NullPointerSupplied </td><td> a vector is NULL</td></tr>
	  </table>
*/

int HermiteMatrix::VectorMultiply(mpz_t *aTarget, mpz_t *aSource){
    long i,j;
    if ((!aTarget)||(!aSource)) return ConstRC::NullPointerSupplied;
    mpz_t c;
    mpz_init(c);
    for(i=0;i<=maximal_column_index;i++){
	mpz_set_ui(aTarget[i],0);
	for(j=0;j<=maximal_row_index;j++){
	    if(mpz_cmp_ui(aSource[j],0)){
		mpz_mul(c,matrix[j][i],aSource[j]);
		mpz_add(aTarget[i],aTarget[i],c);
	    }
	}
    }
    mpz_clear(c);
    return ConstRC::Ok;
}

/*!
	Multiplies the matrix by a scalar.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int HermiteMatrix::ScalarMultiply(HermiteMatrix *aTarget, mpz_t aScalar){
    if (this->maximal_row_index != aTarget->maximal_row_index)              return ConstRC::SizeMismatch;
    if (this->maximal_column_index != aTarget->maximal_column_index)        return ConstRC::SizeMismatch;
    int ret;
    ret = aTarget->Allocate();
    if (ret != ConstRC::Ok){
        return ret;
    }
    long i,j;
    for(i=0; i<=maximal_row_index; i++){
	for(j=0; j<=maximal_column_index; j++){
	    mpz_mul(aTarget->matrix[i][j],this->matrix[i][j],aScalar);
	}
    }
    return ConstRC::Ok;
}

/*!
	Multiplies the matrix by a scalar.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int HermiteMatrix::ScalarMultiply(HermiteMatrix *aTarget, long aScalar){
    if (this->maximal_row_index != aTarget->maximal_row_index)              return ConstRC::SizeMismatch;
    if (this->maximal_column_index != aTarget->maximal_column_index)        return ConstRC::SizeMismatch;
    int ret;
    ret = aTarget->Allocate();
    if (ret != ConstRC::Ok){
        return ret;
    }
    long i,j;
    for(i=0; i<=maximal_row_index; i++){
	for(j=0; j<=maximal_column_index; j++){
	    mpz_mul_si(aTarget->matrix[i][j],this->matrix[i][j],aScalar);
	}
    }
    return ConstRC::Ok;
}

/*!
	Divides the matrix by a scalar.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int HermiteMatrix::ScalarDivide(HermiteMatrix *aTarget, mpz_t aScalar){
    if (this->maximal_row_index != aTarget->maximal_row_index)              return ConstRC::SizeMismatch;
    if (this->maximal_column_index != aTarget->maximal_column_index)        return ConstRC::SizeMismatch;
    int ret;
    ret = aTarget->Allocate();
    if (ret != ConstRC::Ok){
        return ret;
    }
    long i,j;
    for(i=0; i<=maximal_row_index; i++){
	for(j=0; j<=maximal_column_index; j++){
	    mpz_divexact(aTarget->matrix[i][j],this->matrix[i][j],aScalar);
	}
    }
    return ConstRC::Ok;
}

/*!
	Divides the matrix by a scalar.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int HermiteMatrix::ScalarDivide(HermiteMatrix *aTarget, long aScalar){
    if (this->maximal_row_index != aTarget->maximal_row_index)              return ConstRC::SizeMismatch;
    if (this->maximal_column_index != aTarget->maximal_column_index)        return ConstRC::SizeMismatch;
    int ret;
    ret = aTarget->Allocate();
    if (ret != ConstRC::Ok){
        return ret;
    }
    long i,j;
    for(i=0; i<=maximal_row_index; i++){
	for(j=0; j<=maximal_column_index; j++){
	    mpz_fdiv_q_ui(aTarget->matrix[i][j],this->matrix[i][j],abs(aScalar));
	    if (aScalar<0) mpz_neg(aTarget->matrix[i][j],aTarget->matrix[i][j]);
	}
    }
    return ConstRC::Ok;
}

/*!
	Given a Hermite Basis aSource of a domain R and vectors aNewVectors generating the ring
	S/pR for a prime p=aPrime, the function outputs the Hermite Base of S.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int HermiteMatrix::FormHermiteBase(HermiteMatrix *aTarget, HermiteMatrix *aSource, IntegerMatrix *aNewVectors, unsigned int aDimension, long aPrime, mpz_t aDelta){
    if (aTarget->maximal_column_index!=aSource->maximal_column_index) return ConstRC::SizeMismatch;
    if (aTarget->maximal_row_index!=aSource->maximal_row_index) return ConstRC::SizeMismatch;
    HermiteMatrix *B;
    B = new HermiteMatrix(aSource->maximal_row_index+1,aSource->maximal_row_index+aDimension+1);
    int ret=B->Zeroize();
    if (ret!=ConstRC::Ok) return ret;
    ret=aTarget->Allocate();
    if (ret!=ConstRC::Ok) return ret;
    long i,j,k;
    mpz_t aux;
    mpz_init(aux);
    for(i=0;i<=B->maximal_row_index;i++)
        for(j=i;j<=aSource->maximal_column_index;j++)
            mpz_mul_ui(B->matrix[i][j],aSource->matrix[i][j],aPrime);
    for(i=0;i<=B->maximal_row_index;i++)
        for(j=0;(unsigned long)j<aDimension;j++)
            for(k=0;k<=aSource->maximal_row_index;k++){
                mpz_mul_si(aux,aSource->matrix[k][i],aNewVectors->GetXY(i,j));
                mpz_add(B->matrix[k][j+aSource->maximal_column_index+1],B->matrix[k][j+aSource->maximal_column_index+1],aux);
            }
    B->ToHermiteNormalForm(aDelta);
    mpz_mul_ui(aTarget->denominator,aSource->denominator,aPrime);
    for(i=0;i<=B->maximal_row_index;i++)
        for(j=0;j<=aTarget->maximal_column_index;j++)
            mpz_set(aTarget->matrix[i][j],B->matrix[i][j+aDimension]);
    delete(B);
    mpz_clear(aux);
    return ConstRC::Ok;
}

/*!
	Returns the kernel of given homomorphism having the dimension aDimension.

	Return codes
	   <table border=0>
	  	<tr><td>ConstRC::Ok		  </td><td> everything all right</td></tr>
		<tr><td>ConstRC::NotEnoughMemory </td><td> thrown from Allocate() call; there was not enough memory to allocate the requested space.</td></tr>
		<tr><td>ConstRC::SizeMismatch </td><td> the matrices have different sizes</td></tr>
	  </table>
*/

int HermiteMatrix::Kernel(HermiteMatrix *aTarget, long *aDimension, bool aOnMyself){
    if (aTarget->maximal_column_index!=aTarget->maximal_row_index) return ConstRC::SizeMismatch;
    if (aTarget->maximal_row_index!=this->maximal_column_index) return ConstRC::SizeMismatch;
    HermiteMatrix *M;
    int ret;
    if (aOnMyself){
	M=this;
    } else {
	M = new HermiteMatrix(this->maximal_row_index+1,this->maximal_column_index+1);
	ret=M->Copy(this);
	if (ret!=ConstRC::Ok) return ret;
    }
    ret=aTarget->Zeroize();
    if (ret!=ConstRC::Ok) return ret;
    mpz_t aux,divisor;
    mpz_init(aux);
    mpz_init_set_ui(divisor,1);
    long i,j,k,p,r=0;
    for(k=0;k<=maximal_column_index;k++) if (r<=maximal_row_index) {
	for (p=r;(p<=maximal_row_index)&&(!mpz_cmp_ui(M->matrix[p][k],0));p++);
	if (p<=maximal_row_index) {
	    if(r!=p) {
		for (j=k;j<=maximal_column_index;j++){
		    mpz_xor(M->matrix[p][j],M->matrix[p][j],M->matrix[r][j]);
		    mpz_xor(M->matrix[r][j],M->matrix[r][j],M->matrix[p][j]);
		    mpz_xor(M->matrix[p][j],M->matrix[p][j],M->matrix[r][j]);
		}
	    }
	    for(i=r+1;i<=maximal_row_index;i++) {
		for(j=k+1;j<=maximal_column_index;j++) {
		    mpz_mul(M->matrix[i][j],M->matrix[r][k],M->matrix[i][j]);
		    mpz_mul(aux,M->matrix[r][j],M->matrix[i][k]);
		    mpz_sub(M->matrix[i][j],M->matrix[i][j],aux);
		    mpz_divexact(M->matrix[i][j],M->matrix[i][j],divisor);
		}
		mpz_set_ui(M->matrix[i][k],0);
	    }
	    mpz_set(divisor,M->matrix[r][k]);
	    r++;
	}
    }
    *aDimension=0;
    r=maximal_column_index;
    for(k=maximal_row_index;k>=0;k--) {
	for(p=0;(p<=r)&&(mpz_cmp_ui(M->matrix[k][p],0)==0);p++);
	if(p<=r){
	    while (r>p) {
		mpz_set_ui(aTarget->matrix[r--][(*aDimension)++],1);
	    }
	    aTarget->ScalarMultiply(aTarget,M->matrix[k][r]);
	    for(i=0;i<*aDimension;i++){
		for(j=r+1;j<=maximal_column_index;j++){
		    mpz_mul(aux,M->matrix[k][j],aTarget->matrix[j][i]);
		    mpz_add(aTarget->matrix[r][i],aTarget->matrix[r][i],aux);
		}
		mpz_divexact(aTarget->matrix[r][i],aTarget->matrix[r][i],M->matrix[k][r]);
		mpz_neg(aTarget->matrix[r][i],aTarget->matrix[r][i]);
	    }
	    r--;
	}
    }
    mpz_clear(aux);
    mpz_clear(divisor);
    if (!aOnMyself) delete(M);
    return ConstRC::Ok;
}

int HermiteMatrix::Save(char *aName){
    return ConstRC::NotSupported;
}
int HermiteMatrix::Load(char *aName){
    return ConstRC::NotSupported;
}


//int HermiteMatrix::Serialize(const CBaseParameters &aParam) const
//{
//	return MSerializable::Serialize(aParam);
//}
//
//int HermiteMatrix::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
//{
//	return ConstRC::NotSupported;
//}
//
//int HermiteMatrix::Deserialize(const CBaseParameters &aParam)
//{
//	return MSerializable::Deserialize(aParam);
//}
//
//int HermiteMatrix::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
//{
//	return ConstRC::NotSupported;
//}

int HermiteMatrix::WriteData(xmlTextWriterPtr aWriter) const
{
	return ConstRC::NotSupported;
}

int HermiteMatrix::ReadData(xmlTextReaderPtr aReader)
{
	return ConstRC::NotSupported;
}
