#ifndef __HERMITE_MATRIX_CLASS__
#define __HERMITE_MATRIX_CLASS__

#include "integer_matrix_class.h"
#include <stdlib.h>
#include <gmp.h>

/*!
\brief	A class of matrices with arbitrary rational coefficients

	The Hermite matrix is a pair (<I>M</I>,<i>f</i>), where <i>M</i> is an integer
	matrix and <i>f</i> is an integer denominator. Each coefficient here is of type
	mpz_t - multiprecision integer.

	The purpose of this matrix is twofold: it represents integral bases of integral domains
	in Hermite normal form - here the name, and it serves for representing the inverse matrices
	of integer matrices.
*/

class HermiteMatrix : public AbstractMatrix {
public:
	HermiteMatrix();
	HermiteMatrix(long aRows, long aColumns);
	~HermiteMatrix();
	int  Allocate();
	HermiteMatrix * clone();

	int  Randomize();
	int  Zeroize();
	bool Equals(HermiteMatrix* aMatrix);
	void PrintToScreen();
	int  PutOne(long aRow, long aColumn);
	int  PutZero(long aRow, long aColumn);
	int  IsOne(long aRow, long aColumn);
	int  IsZero(long aRow, long aColumn);
	int  PutNumber(mpz_t aNumber, long aRow, long aColumn);
	int  PutNumber(long aNumber, long aRow, long aColumn);
	void Modulo(long aModulus);
	void Modulo(mpz_t aModulus);
	long GetXY(long aRow, long aColumn);
	mpz_t *GetXY2(long aRow, long aColumn);
	int  Copy(HermiteMatrix *aSource);
	// Swaps etc.
	int  ZeroizeRow(long aRow);
	int  ZeroizeRow(long* aRowList, long aListMaxIndex);
	int  SwapRows(long aRow1, long aRow2);
	int  AddRows(long aTarget, long aSource);
	int  AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2);
	int  ZeroizeColumn(long aColumn);
	int  ZeroizeColumn(long* aColumnList, long aListMaxIndex);
	int  SwapColumns(long aColumn1, long aColumn2);
	int  AddColumns(long aTarget, long aSource);
	// Mathematical operations
	int  	   Transpose(HermiteMatrix* aTarget);
	HermiteMatrix* Transpose();
	int  	   Add(HermiteMatrix* aTarget, HermiteMatrix* aOperand2);
	HermiteMatrix* Add(HermiteMatrix* aOperand2);
	int	MultiplyInternal(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2);
	int	MultiplyInternalTransposed(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2);
	int	MultiplyInternalWithTransposition(HermiteMatrix* aTarget, HermiteMatrix* aOperand1, HermiteMatrix* aOperand2);
	HermiteMatrix*   MultiplyInternalWithTransposition(HermiteMatrix* aOperand1, HermiteMatrix* aOperand2);
	HermiteMatrix*	MultiplyInternal(HermiteMatrix* aOperand1, HermiteMatrix* aOperand2);
	HermiteMatrix*	MultiplyInternalTransposed(HermiteMatrix* aOperand1, HermiteMatrix* aOperand2);
	int  MultiplyVector(mpz_t *aTarget, mpz_t *aSource);
	int  VectorMultiply(mpz_t *aTarget, mpz_t *aSource);
	int  ScalarMultiply(HermiteMatrix *aTarget, mpz_t aScalar);
	int  ScalarMultiply(HermiteMatrix *aTarget, long aScalar);
	int  ScalarDivide(HermiteMatrix *aTarget, mpz_t aScalar);
	int  ScalarDivide(HermiteMatrix *aTarget, long aScalar);
	int  InverseFromTriangular(HermiteMatrix *aTarget);
	int  ToHermiteNormalForm(void);
	int  ToHermiteNormalForm(mpz_t aModulus);
	int  FormHermiteBase(HermiteMatrix *aTarget, HermiteMatrix *aSource, IntegerMatrix *aNewVectors, unsigned int aDimension, long aPrime, mpz_t aDelta);
	int  Reduce(mpz_t aFactor);
	int  Reduce(long aFactor);
	int  Kernel(HermiteMatrix *aTarget, long *aDimension, bool aOnMyself);
	int  Save(char *aName);
	int  Load(char *aName);

	mpz_t denominator;///<	The denominator
	
	// De/Serialization methods
	//int Serialize(const CBaseParameters& aParam) const;
	//int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	//int Deserialize(const CBaseParameters& aParam);
	//int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

protected:
	int WriteData(xmlTextWriterPtr aWriter) const;
	int ReadData(xmlTextReaderPtr aReader);

	
protected:
	long max_allocated_row_index;///<	Maximal allocated column index
	long max_allocated_column_index;///<	Maximal allocated row index
	mpz_t** matrix; ///<	The matrix array
};

#endif

