#include <cstring>

#include "integral_domain.h"
#include "nfs_utils.h"
#include "integer_class.h"
#include "const_return_codes.h"
#include "log.h"


IntegralDomain::IntegralDomain()
{
    Rank=0;
    Base=NULL;
    BaseInverse=NULL;
    omega=NULL;

    mpz_init_set_ui(this->Index,1);
	mpz_init(this->Discriminant);

    SpecialPrimesCount=0;
    SpecialPrimesAllocated=0;

    ThetaPowers = NULL;
    ThetaPolynomial = NULL;
    AllocatedThetaPowers = 0;

	mpz_init(iOp1Aux);
	mpz_init(iOp2Aux);
}

IntegralDomain::~IntegralDomain()
{
    if (Base)
	{
		delete(Base);
	}
    if (BaseInverse)
	{
		delete(BaseInverse);
	}

    unsigned int i,j;
    if (omega) 
	{	
		for(i=0;i<Rank;i++)
		{
		    delete(omega[i]);
		    for(j=0;j<Rank;j++) 
			{
				delete(OmegaMultiplication[i][j]);
				OmegaMultiplication[i][j] = NULL;
		    }
		  free(OmegaMultiplication[i]);	    
		    /*if (i < (Rank-1)) 
			{
			printf("Mazu ThetaPowers %d\n",i);
			delete(ThetaPowers[i]);
			ThetaPowers[i] = NULL;
			 }*/
		}
	
		free(omega);
		free(OmegaMultiplication);
    }

    if (SpecialPrimesAllocated)
	{
		free(SpecialPrimes);
	}

    mpz_clear(Index);
    mpz_clear(Discriminant);
    delete (ThetaPolynomial); 

    if (ThetaPowers != NULL) 
	{
		for (j=0; (signed)j< AllocatedThetaPowers; j++)
		{		
			delete(ThetaPowers[j]);
			ThetaPowers[j] = NULL;
		}
		free (ThetaPowers);
    }    
	mpz_clear(iOp1Aux);
	mpz_clear(iOp2Aux);
}

int IntegralDomain::ComputeThetaPolynomial(Polynomial *aPolynomial, Polynomial *&aThetaPolynomial)
{
	if (aPolynomial == NULL)
	{
		return ConstRC::BadArgument;
	}

	if (aPolynomial->GetDegree() < 2) 
	{
		return ConstRC::GeneralError;
	}

	if (mpz_cmp_ui(*aPolynomial->Leading2(), 1) != 0)
	{
		mpz_t aux, pow;

		mpz_init(aux);
		mpz_init_set_ui(pow, 1);

		int degree = aPolynomial->GetDegree();
		aThetaPolynomial = new Polynomial(degree);
		aThetaPolynomial->Set(degree, 1);
		for (int i = degree - 1; i >= 0; i--)
		{
			mpz_mul(aux, pow, *aPolynomial->Get2(i));
			aThetaPolynomial->Set(i, aux);
			mpz_mul(pow, pow, *aPolynomial->Get2(degree));
		}

		mpz_clear(aux);
		mpz_clear(pow);
	}
	else
	{
		aThetaPolynomial = new Polynomial(*aPolynomial);
	}

	return ConstRC::Ok;
}
/*!
	Assigns an irreducible polynomial <I>t</I>. The domain is now equal to the
	domain <B>Z</B>\htmlonly[&theta;\endhtmlonly\latexonly$[\theta$\endlatexonly].
	The only computation made here is the computation of disc(<i>t</i>).
*/


int IntegralDomain::AssignPolynomial(Polynomial *aPolynomial)
{
	if (ComputeThetaPolynomial(aPolynomial, this->ThetaPolynomial) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}
    if (ThetaPowers)
	{
		return ConstRC::GeneralError;
	}

    Rank = (unsigned)aPolynomial->GetDegree();
    ThetaPowers = (HermiteMatrix**) Utils::AllocationMachine((void **) &ThetaPowers, (Rank-1)*sizeof(HermiteMatrix*));
    if (!ThetaPowers)
	{
		return ConstRC::NotEnoughMemory;
	}

    AllocatedThetaPowers = Rank-1;
    int ret;
    signed int i,d;
    mpz_t aux;
    mpz_init(aux);
    d=aPolynomial->GetDegree();
    ThetaPolynomial->PrintToScreen();puts("");fflush(stdout);
    ThetaPowers[0] = new HermiteMatrix(Rank,Rank);
    ret = ThetaPowers[0]->Zeroize();

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    // Here, a signed value (i) gets compared to an unsigned value Rank.
    // That is why we use a cast instead.   
    for(i=0;i<(signed) Rank-1;i++)
	{
	ThetaPowers[0]->PutNumber(1,i+1,i);
	}

    for(i=0;i<(signed) Rank;i++)
	{
        mpz_neg(aux,*ThetaPolynomial->Get2(i));
        ThetaPowers[0]->PutNumber(aux,i,(signed) Rank-1);
    }
    for(i=1;i<(signed) Rank-1;i++)
	{
	ThetaPowers[i]=ThetaPowers[0]->MultiplyInternal(ThetaPowers[0],ThetaPowers[i-1]);
        if(!ThetaPowers[i]) return ConstRC::NotEnoughMemory;
    }
    //mpz_init(this->Discriminant);
    this->Rank=(unsigned)d;
    aPolynomial->Discriminant(Discriminant);

    //gmp_printf("%Zd\n",Discriminant);
    mpz_pow_ui(aux,*aPolynomial->Get2(d),d*d-3*d+2);
    mpz_mul(Discriminant,Discriminant,aux);

    //gmp_printf("%Zd\n",Discriminant);
    Base = new HermiteMatrix(Rank,Rank);
    ret = Base->Zeroize();

	if (ret!=ConstRC::Ok)
	{
		return ret;
	}
    for(i=0;i<(signed) Rank;i++)
	{
		Base->PutOne(i,i);
	}
	mpz_clear(aux);
    return ConstRC::Ok;
}

/*!
	Computes an extension of the domain in the direction of a prime aPrime
	using the Dedekind Criterion. The dimension of the extension over the former
	domain is aDimension.
*/


int IntegralDomain::DedekindCriterion(unsigned int *aDimension, long aPrime)
{
    PolynomialSplitting *PS;
    int ret,i;
    unsigned j;
    PS = new PolynomialSplitting();
    PS -> Characteristic = aPrime;
    ret = PS -> Split(ThetaPolynomial);

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    Polynomial *gh,*ghdiv,*f,*aux;
    gh = new Polynomial(ThetaPolynomial->GetDegree());
    ghdiv = new Polynomial(ThetaPolynomial->GetDegree());
    f = new Polynomial(ThetaPolynomial->GetDegree());
    aux = new Polynomial(ThetaPolynomial->GetDegree());
    gh->Set(0,1);
    ghdiv->Set(0,1);

    for(i=0;(unsigned)i<PS->Length;i++)
	{
		if(PS->Exponents[i]>1)
		{
		    ghdiv->Multiply(aux,ghdiv,PS->Factors[i]);
		    ghdiv->Multiply(ghdiv,aux,1);
		}
		gh->Power(aux,PS->Factors[i],PS->Exponents[i]);
		gh->Multiply(f,gh,aux);
		gh->Multiply(gh,f,1);
    }
/*  printf("gh= ");
    gh->PrintToScreen();
    printf("\n(g,h)= ");
    ghdiv->PrintToScreen();
*/
    f->Substract(f,gh,ThetaPolynomial);
    f->Divide(f,f,aPrime);
/*
    printf("\nf= ");
    f->PrintToScreen();
*/
    aux->GCD_p(aux,ghdiv,f,aPrime);
    *aDimension=(unsigned)(aux->GetDegree());
    f->Divide(f,ghdiv,gh,aux);
/*  printf("\nResult= ");
    f->PrintToScreen();
    puts("");
*/
    if(*aDimension)
	{
		mpz_t aux2,one;
		mpz_init(aux2);
		mpz_init_set_ui(one,1);
		HermiteMatrix *B;
		AlgebraicNumber *A,*V,*S;
		A = new AlgebraicNumber(Rank,ThetaPowers,&one);
		V = new AlgebraicNumber(Rank,ThetaPowers,&one);
		S = new AlgebraicNumber(Rank,ThetaPowers,&one);
		B = new HermiteMatrix(Rank,Rank+(*aDimension));
		ret=B->Allocate();

		if (ret!=ConstRC::Ok)
		{
			return ret;
		}

		for(i=0;i<(signed)Rank;i++)
		{
		    for(j=i;j<Rank;j++)
			{
				mpz_mul_ui(aux2,*Base->GetXY2(i,j),aPrime);
				B->PutNumber(aux2,i,j);
		    }
		}
		for(i=0;i<(signed)Rank;i++)
		{
		   A->Set(i,*f->Get2(i));
		}

		A->ComputeMultiplicationMatrix();
		for(i=0;(unsigned)i<*aDimension;i++)
		{
		    for(j=0;j<Rank;j++)
			{
				V->Set(j,*Base->GetXY2(j,i));
			}
		  A->Multiply(S,V,FALSE);
		  for(j=0;j<Rank;j++)
		  {
			B->PutNumber(*S->Get2(j),j,Rank+i);
		  }
		}
		mpz_set_ui(aux2,1);

		for(j=0;j<Rank;j++)
		{
		   mpz_mul(aux2,aux2,*B->GetXY2(j,j));
		}
		B->ToHermiteNormalForm(aux2);
		mpz_mul_ui(Base->denominator,Base->denominator,aPrime);

		for(i=0;i<(int)Rank;i++)
		{
		   for(j=0;j<Rank;j++)
		   {
				Base->PutNumber(*B->GetXY2(i,j+*aDimension),i,j);
		   }
		}
		delete(B);

		mpz_set_ui(aux2,aPrime);
		mpz_pow_ui(aux2,aux2,*aDimension);
		mpz_mul(Index,Index,aux2);
		mpz_mul(aux2,aux2,aux2);
		mpz_divexact(this->Discriminant,this->Discriminant,aux2);
		mpz_clear(aux2);

		delete(A);
		delete(S);
		delete(V);
		mpz_clear(one);
    }

    delete(gh);
    delete(ghdiv);
    delete(f);
    delete(aux);
    delete(PS);

    return ConstRC::Ok;
}

/*!
	Computes an extension of the domain in the direction of a prime aPrime
	using the Pohst-Zassenhaus theorem. The dimension of the extension over the former
	domain is aDimension.
*/

int IntegralDomain::PohstZassenhaus(unsigned int *aDimension, long aPrime)
{
    int ret;

    HermiteMatrix *Baze,*Inverz,*aux,*aux2;
    IntegerMatrix *Frob,*Kernel,*Mapping;
    AlgebraicNumber **Bazove=NULL,*AN;

    unsigned int i;
    int j,k;
    long p=aPrime;

    mpz_t aux3,aux4,aux5;
    mpz_init(aux3);
    mpz_init(aux4);
    mpz_init(aux5);

    while ((unsigned)p<Rank) 
	{
		p*=aPrime;
	}

    AN = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    Bazove = (AlgebraicNumber**) Utils::AllocationMachine((void **) &Bazove,(this->Rank)*sizeof(AlgebraicNumber*));
    Kernel = new IntegerMatrix(Rank,Rank);
    Kernel->Allocate();
    Frob = new IntegerMatrix(Rank,Rank);
    ret = Frob->Zeroize();

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

//    Base->PrintToScreen();
    for(i=0;i<Rank;i++)
	{
		Bazove[i] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
		for(j=0;(unsigned)j<Rank;j++)
		{
		    Bazove[i]->Set(j,*Base->GetXY2(j,i));
		}
		Bazove[i]->Power(AN,p);
//		printf("\n%d:",i);fflush(stdout);
//		AN->PrintToScreen();fflush(stdout);
		for(j=Rank-1;j>=0;j--)
		{
		    mpz_divexact(aux3,*AN->Get2(j),*Base->GetXY2(j,j));
		    Frob->PutNumber(mpz_fdiv_ui(aux3,aPrime),j,i);
//			 gmp_printf(" %Zd %d %d",aux3,j,i);
			for(k=j;k>=0;k--)
			{
				mpz_mul(aux4,aux3,*Base->GetXY2(k,j));
				AN->Dec(k,aux4);
			}
		}
    }

//    Frob->PrintToScreen();
    Frob->KernelModPrime(Kernel,&p,aPrime,TRUE);
//    Kernel->PrintToScreen();

    Baze = new HermiteMatrix(Rank,Rank);
    Inverz = new HermiteMatrix(Rank,Rank);
    aux = new HermiteMatrix(Rank,Rank);    
    aux2 = new HermiteMatrix(Rank,Rank);    
    Mapping = new IntegerMatrix(Rank*Rank,Rank);
    mpz_mul_ui(aux4,Base->denominator,aPrime);
    mpz_pow_ui(aux4,aux4,Rank);
    ret = Base->FormHermiteBase(Baze,Base,Kernel,p,aPrime,aux4);

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

//    Baze->PrintToScreen();
    ret = Baze->InverseFromTriangular(Inverz);
//    Inverz->PrintToScreen();

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    for(i=0;i<Rank;i++)
	{
		Baze->MultiplyInternal(aux,Inverz,Bazove[i]->multiplication);
		Baze->MultiplyInternal(aux2,aux,Baze);
		if (i==0)
		{
			mpz_set(aux3,*aux2->GetXY2(0,0));
		}
		aux2->Reduce(aux3);
//		aux2->PrintToScreen();
		for(j=0;(unsigned)j<Rank;j++)
		{
			for(k=0;(unsigned)k<Rank;k++)
			{
			Mapping->PutNumber(mpz_mod_ui(aux5,*aux2->GetXY2(j,k),aPrime),j+k*Rank,i);
		    }
		}
    }

//    Mapping->PrintToScreen();
    ret = Mapping->KernelModPrime(Kernel,&p,aPrime,TRUE);
//    Kernel->PrintToScreen();

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    *aDimension=p;
    if(*aDimension)
	{
//		Base->PrintToScreen();
		ret = Base->FormHermiteBase(Base,Base,Kernel,*aDimension,aPrime,aux4);
//		Base->PrintToScreen();
		
		if (ret!=ConstRC::Ok)
		{
			return ret;
		}

		mpz_set_ui(aux3,aPrime);
		mpz_pow_ui(aux3,aux3,*aDimension);
		mpz_mul(Index,Index,aux3);
		mpz_mul(aux3,aux3,aux3);
		mpz_divexact(this->Discriminant,this->Discriminant,aux3);
    }

    mpz_clear(aux3);
    mpz_clear(aux4);
    mpz_clear(aux5);

    delete(Kernel);
    delete(Mapping);
    delete(Baze);
    delete(Inverz);
    delete(aux);
    delete(aux2);
    delete(Frob);
    delete(AN);

    for(i=0;i<Rank;i++)
	{
		delete Bazove[i];
	}

    free(Bazove);

    return ConstRC::Ok;
}

/*!
	Finds the integrally closed domain, e.g. the integral base of the extension field.
	Actually, the domain is almost integrally closed in the sense that its index in its
	integral closure is not divisible by any prime smaller than aBound.
*/


int IntegralDomain::FindIntegralBase(long aBound)
{
	// printf("Integral Domain: Find Integral Base!\n");
    if (!Base)
	{
		return ConstRC::GeneralError;
	}
    unsigned long primes;
    long *small_primes = NULL;

	int ret = CIntegerClass::GetPrimes(0, aBound, primes, small_primes);
	if (ret != ConstRC::Ok)
	{
		return ret;
	}
	//primes = CIntegerClass::FindSizesByReadingPrimesFile(aBound);
	//if (primes < 0) 
	//{
	//	return primes;
	//}
	//small_primes = (long*) Utils::AllocationMachine((void **) &small_primes, primes*sizeof(long));
	//if (!small_primes) 
	//{
	//	return ConstRC::NotEnoughMemory;
	//}
	//ret = CIntegerClass::ReadPrimesFromFile(0,small_primes,primes);

    mpz_t N,aux;
    unsigned long i;
    unsigned int d;
    mpz_init(N);
    mpz_init(aux);
    int deli;

    ThetaPolynomial->Discriminant(N);
    i=mpz_cmp_ui(N,0);

    if (i==0)
	{
		return ConstRC::GeneralError;
	}
    if (i<0)
	{
		mpz_neg(N,N);
	}

    for(i=0;(i<primes)&&(mpz_cmp_ui(N,small_primes[i]>0));i++)
	{
// 		printf(" FIB: %ld ",small_primes[i]);
		deli=0;
		while(mpz_divisible_ui_p(N,small_primes[i]))
		{
// 			gmp_printf("Divisible %d %Zd\n",small_primes[i],N);
		    mpz_divexact_ui(N,N,small_primes[i]);
		    if (deli & 1) 
			{
				if (deli==1)
				{
					 ret=DedekindCriterion(&d,small_primes[i]);
					 if (ret!=ConstRC::Ok)
					 {
						 return ret;
					 }

					 if (d)
					 {
						mpz_ui_pow_ui(aux,small_primes[i],(d<<1)-2);
						mpz_divexact(N,N,aux);
						AddToListOfSpecialPrimes(small_primes[i]);
//						printf("%d ",small_primes[i]);
					 }
					 else 
					 {
						 deli=-11;
					 }

				}
				else
				{
				    ret=PohstZassenhaus(&d,small_primes[i]);
				    if (ret!=ConstRC::Ok)
					{
						return ret;
					}
					if (d)
					{
						mpz_ui_pow_ui(aux,small_primes[i],(d<<1)-2);
						mpz_divexact(N,N,aux);
					}
					else
					{
						deli=-11;
					}
				}
//				Base->PrintToScreen();
			}
			else
			{
				if (deli==-10)
				{
					deli=-11;
				}
			}
			deli++;
		}
    }
    mpz_set(aux,*Base->GetXY2(Rank-1,Rank-1));
    Base->Reduce(aux);
    mpz_clear(N);
    mpz_clear(aux);
    free(small_primes);

    return ConstRC::Ok;
}

/*!
	Computes the multiplication table of integral base vertices.
*/


int IntegralDomain::ComputeOmegas()
{
    int ret;

    BaseInverse = new HermiteMatrix(Rank,Rank);
    ret=Base->InverseFromTriangular(BaseInverse);
    mpz_divexact(BaseInverse->denominator,BaseInverse->denominator,Base->denominator);

    if (ret!=ConstRC::Ok)
	{
		return ret;
	}

    this->OmegaMultiplication = NULL;
    this->OmegaMultiplication = (AlgebraicNumber***) Utils::AllocationMachine((void **) &this->OmegaMultiplication, Rank*sizeof(AlgebraicNumber**));
    
	if (!this->OmegaMultiplication)
	{
		return ConstRC::NotEnoughMemory;
	}

    this->omega = (AlgebraicNumber**) Utils::AllocationMachine((void **) &this->omega, Rank*sizeof(AlgebraicNumber*));
    
	if (!this->omega)
	{
		return ConstRC::NotEnoughMemory;
	}

    unsigned int i,j;
    int k;
    AlgebraicNumber *A,*B;
    A = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    B = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    mpz_t aux;
    mpz_init(aux);

    for(i=0;i<Rank;i++)
	{
		OmegaMultiplication[i] = NULL;
		OmegaMultiplication[i] = (AlgebraicNumber**) Utils::AllocationMachine((void **) this->OmegaMultiplication+i, Rank*sizeof(AlgebraicNumber*));
		if (!this->OmegaMultiplication[i])
		{
			return ConstRC::NotEnoughMemory;
		}
		omega[i] = NULL;
		omega[i] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
		for(j=0;j<Rank;j++)
		{
		    omega[i]->Set(j,*Base->GetXY2(j,i));
		}
		ret=omega[i]->ComputeMultiplicationMatrix();
		if (ret!=ConstRC::Ok) 
		{
			return ret;
		}
    }

    for(i=0;i<Rank;i++)
	{
		for(j=0;j<Rank;j++)
		{
		    OmegaMultiplication[i][j] = NULL;
		    OmegaMultiplication[i][j] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
		    omega[i]->Multiply(A,omega[j],FALSE);

		    for(k=(signed)Rank-1;k>=0;k--)
			{
				mpz_divexact(aux,*A->Get2(k),*omega[k]->Get2(k));
				OmegaMultiplication[i][j]->Set(k,aux);
				mpz_neg(aux,aux);
				omega[k]->Multiply(B,aux,FALSE);
				A->Add(A,B,FALSE);
			}
		}
	}
    delete(A);
    delete(B);
    mpz_clear(aux);
    return ConstRC::Ok;
}

/*!
	Computes the valuation of the prime ideal generated by aGenCount elements in aGenerators
	in the ideal with Hermite base aIdeal.
*/


int IntegralDomain::ValuationOfPrime(long *aValuation, AlgebraicNumber **aGenerators, unsigned int aGenCount, long aPrime, HermiteMatrix *aIdeal)
{
    unsigned int i,j;
    int k;
    AlgebraicNumber *Alfa,*Beta,*Gama;
    int ret;

    if (!BaseInverse)
	{
		ret=ComputeOmegas();
		if (ret!=ConstRC::Ok)
		{
			return ret;
		}
    }

    IntegerMatrix *H,*K;
    long d;
    H = new IntegerMatrix((long)Rank*(long)aGenCount,Rank);
    ret = H->Zeroize();

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    Alfa = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    Beta = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);

    mpz_t aux;
    mpz_init(aux);

    for(i=0;i<Rank;i++)
	{
		for(j=0;j<aGenCount;j++)
		{
		    omega[i]->Multiply(Alfa,aGenerators[j],false);
		    for(k=(signed)Rank-1;k>=0;k--)
			{
				mpz_divexact(aux,*Alfa->Get2(k),*omega[k]->Get2(k));
				H->PutNumber(mpz_fdiv_ui(aux,aPrime),k+(long)j*(long)Rank,i);
				mpz_neg(aux,aux);
				omega[k]->Multiply(Beta,aux,false);
				Alfa->Add(Alfa,Beta,false);
			}
		}
    }

    K = new IntegerMatrix(Rank,Rank);
    ret = H->KernelModPrime(K,&d,aPrime,true);

    if (ret!=ConstRC::Ok) 
	{	
		return ret;
    }

    if (!d) 
	{
		return ConstRC::GeneralError;
	}
/*
    Beta->Zeroize();
    for(i=0;i<Rank;i++){
	omega[i]->Multiply(Alfa,K->GetXY(i,0),FALSE);
	Beta->Add(Beta,Alfa,FALSE);
    }
    Beta->ComputeMultiplicationMatrix();
    Beta->PrintToScreen();
    Beta->multiplication->PrintToScreen();
*/
    HermiteMatrix *A;
    A = new HermiteMatrix(Rank,Rank);
    ret = aIdeal -> MultiplyInternal(A,BaseInverse,aIdeal);

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    mpz_divexact(aux,BaseInverse->denominator,Base->denominator);
    mpz_mul(aux,aIdeal->denominator,aux);
    A->Reduce(aux);
    mpz_set_ui(aux,1);

    for(i=0;i<Rank;i++)
	{
		mpz_mul(aux,aux,*A->GetXY2(i,i));
	}

    *aValuation=0;
    bool divisible = (mpz_divisible_ui_p(aux,aPrime) != 0);
    Gama = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    
	while (divisible)
	{
		for(j=0;j<Rank;j++)
		{
		    Alfa->Zeroize();
			for(i=0;i<Rank;i++)
			{
				for(k=0;k<(signed)Rank;k++)
				{
					mpz_mul_si(aux,*A->GetXY2(k,j),K->GetXY(i,0));
					OmegaMultiplication[i][k]->Multiply(Gama,aux,FALSE);
					Alfa->Add(Alfa,Gama,FALSE);
				}
			}
			for(i=0;i<Rank;i++)
			{
				A->PutNumber(*Alfa->Get2(i),i,j);
			}
		}
		if (mpz_divisible_ui_p(*A->GetXY2(Rank-1,Rank-1),aPrime))
		{
			if(mpz_divisible_ui_p(Index,aPrime))
			{
				for(i=0;(divisible)&&(i<Rank);i++)
				{
					for(j=0;(divisible)&&(j<Rank);j++)
					{
						divisible = mpz_divisible_ui_p(*A->GetXY2(i,j),aPrime) != 0;
					}
				}
			} 
			else
			{
				divisible = mpz_divisible_ui_p(*A->GetXY2(0,0),aPrime) != 0;
			}

		} 
		else
		{
			divisible = false;
		}

		if (divisible) 
		{
		    A->ScalarDivide(A,aPrime);
		    (*aValuation)++;
		}
    }

    delete(Gama);
    delete(Beta);
    delete(Alfa);

    mpz_clear(aux);

    delete(H);
    delete(K);
    delete(A);

    return ConstRC::Ok;
}

//#define MEMLEAK_DEBUG_1

/*!
	Computes the valuation of the prime ideal aPrime
	in the ideal with a base aIdeal.
*/

int IntegralDomain::ValuationOfPrime(long *aValuation, PrimeIdeal *aPrime, HermiteMatrix *aIdeal)
{
#ifdef MEMLEAK_DEBUG_1
    printf("\n======================================\n");
    printf("Valuation of prime: prime %p, ideal %p.\n",aPrime,aIdeal);
#endif

    unsigned int i,j;
    int k;
    AlgebraicNumber *Alfa,*Gama;
    int ret;

    if (!this->BaseInverse)
	{
		ret=ComputeOmegas();
		if (ret!=ConstRC::Ok) 
		{
			CLog::LogError(WHEREARG,"Integral domain","valuation of a prime with input base",1,ret,"Error in compute omegas.");
			return ret;
		}
    }

    Alfa = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    Gama = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    mpz_t aux;
    mpz_init(aux);	

    if (!aPrime->pGenerator) 
	{
#ifdef MEMLEAK_DEBUG_1
	printf("pGenerator needs to be allocated, Rank= %d\n",Rank);
#endif
		IntegerMatrix *H,*K;
		long d;
		H = new IntegerMatrix(Rank,Rank);
		ret = H->Zeroize();
		if (ret!=ConstRC::Ok) 
		{
			CLog::LogError(WHEREARG,"Integral domain","valuation of a prime with input base",1,ret,"Zeroize failed.");
			return ret;
		}
		for(i=0;i<Rank;i++)
		{
		   omega[i]->Multiply(Alfa,aPrime->Generator,FALSE);
		   for(k=(signed)Rank-1;k>=0;k--)
		   {
				mpz_divexact(aux,*Alfa->Get2(k),*omega[k]->Get2(k));
				H->PutNumber(mpz_fdiv_ui(aux,aPrime->Prime),k,i);
				mpz_neg(aux,aux);
				omega[k]->Multiply(Gama,aux,FALSE);
				Alfa->Add(Alfa,Gama,FALSE);
			}
		}
//		H->PrintToScreen();
		K = new IntegerMatrix(Rank,Rank);
		ret = H->KernelModPrime(K,&d,aPrime->Prime,TRUE);
//		K->PrintToScreen();
		
		if (ret!=ConstRC::Ok) 
		{
			CLog::LogError(WHEREARG,"Integral domain","valuation of a prime with input base",1,ret,"Kernel mod prime error.");
			return ret;
		}
		if (!d) 
		{
			CLog::LogError(WHEREARG,"Integral domain","valuation of a prime with input base",1,ret,"d equal to zero error.");
			return ConstRC::GeneralError;
		}
		aPrime->pGenerator = (long*) Utils::AllocationMachine((void **) &aPrime->pGenerator, Rank*sizeof(long));
		
		if (!aPrime->pGenerator) 
		{
			CLog::LogError(WHEREARG,"Integral domain","valuation of a prime with input base",1,ret,"Not enough memory error.");
			return ConstRC::NotEnoughMemory;
		}
		memset(aPrime->pGenerator,0,Rank*sizeof(long));
#ifdef MEMLEAK_DEBUG_1
	printf("..... Array pGenerator allocated. Its address is %p.\n",aPrime->pGenerator);
#endif	
	
		for(i=0;i<Rank;i++)
		{
			aPrime->pGenerator[i]=K->GetXY(i,0);
		}

		delete(H);
		delete(K);
    }

    HermiteMatrix *A;
    A = new HermiteMatrix(Rank,Rank);
   //printf("\nBaseInverse: %d x %d, aIdeal: %d x %d, A = %d x %d.\n",BaseInverse->GetMaximalRowIndex(),BaseInverse->GetMaximalColumnIndex(),aIdeal->GetMaximalRowIndex(),aIdeal->GetMaximalColumnIndex(),A->GetMaximalRowIndex(),A->GetMaximalColumnIndex());
    ret = aIdeal -> MultiplyInternal(A,BaseInverse,aIdeal);

    if (ret!=ConstRC::Ok) 
	{
		CLog::LogError(WHEREARG,"Integral domain","valuation of a prime with input base",1,ret,"Multiply internal failed.");
		return ret;
    }

//    A->PrintToScreen();
    A->Reduce(A->denominator);
//    A->PrintToScreen();

    *aValuation=0;
    bool divisible=true;

    /*	    Opsano z Cohena, ale evidentne nefunguje
    mpz_set_ui(aux,1);
    for(i=0;i<Rank;i++)
	mpz_mul(aux,aux,*A->GetXY2(i,i));
    divisible = (mpz_divisible_ui_p(aux,aPrime->Prime));
    */
	// Pouze za ucelem zjisteni problemu pomoci valgrindu
//    mpz_t docasny; mpz_init(docasny);
//    int docasny_int;

#ifdef MEMLEAK_DEBUG_1
    printf("Current Rank: %d.\n",Rank);
#endif

    while (divisible)
	{
		for(j=0;j<Rank;j++)
		{
#ifdef MEMLEAK_DEBUG_1
            printf("~ j=%d\n",j);
	    printf(";;; Reading from pGenerator array. Its address is %p, rank %d.",aPrime->pGenerator,Rank);
#endif	
			Alfa->Zeroize();
			for(i=0;i<Rank;i++)
			{
//				docasny_int = aPrime->pGenerator[i];
	//			printf("%3d ",docasny_int);
				for(k=0;k<(signed)Rank;k++)
				{		    
		    
//					mpz_set(docasny,*A->GetXY2(k,j));
//					mpz_set(docasny,aux);
#ifdef MEMLEAK_DEBUG_1
	            printf("%d (%d): \n",aPrime->pGenerator[i],i);
		    gmp_printf("%Zd (%d, %d): \n",*A->GetXY2(k,j),k,j);
		    gmp_printf("%Zd \n",aux);
		    fflush(NULL);
#endif
					mpz_mul_si(aux,*A->GetXY2(k,j),aPrime->pGenerator[i]);
					OmegaMultiplication[i][k]->Multiply(Gama,aux,FALSE);
					Alfa->Add(Alfa,Gama,FALSE);
				}
			}

	  //	printf("\n");
			for(i=0;i<Rank;i++)
			{
				A->PutNumber(*Alfa->Get2(i),i,j);
			}
		}

//		A->PrintToScreen();
		if (mpz_divisible_ui_p(*A->GetXY2(Rank-1,Rank-1),aPrime->Prime))
		{
			for(i=0;(divisible)&&(i<Rank);i++)
			{
				for(j=0;(divisible)&&(j<Rank);j++)
				{
					divisible = mpz_divisible_ui_p(*A->GetXY2(i,j),aPrime->Prime) != 0;
		
				}
			}
			
		}
		else
		{
			divisible = false;
		}

		if (divisible)
		{
			A->ScalarDivide(A,aPrime->Prime);
			(*aValuation)++;
		}
    }

#ifdef MEMLEAK_DEBUG_1
    printf("Konec zpracovani.\n");
#endif

    delete(Gama);
    delete(Alfa);
    mpz_clear(aux);
//    mpz_clear(docasny);
    delete(A);

    return ConstRC::Ok;
}

/*!
	Finds a two element generating set of the prime ideal over aPrime with Hermite normal base HNF.
*/
int IntegralDomain::TwoElementRepresentation(PrimeIdeal **aTarget, long aPrime, HermiteMatrix *HNF)
{
    if (*aTarget==NULL) 
	{
		*aTarget = new PrimeIdeal();
    }
	else 
	{
		if ((*aTarget)->Generator)
		{
		    delete((*aTarget)->Generator);
		    (*aTarget)->Generator = NULL;
		}
		(*aTarget)->Ramification = 0;
		if ((*aTarget)->pGenerator)
		{
			free((*aTarget)->pGenerator);
			(*aTarget)->pGenerator = NULL;
		}
    }

    long i,j;
    mpz_t aux,aux2;
    mpz_init(aux);
    mpz_init(aux2);

    if ((*aTarget)->Residual==0)
	{
		mpz_set(aux,this->Index);

		for(i=0;i<(long)Rank;i++)
		{
		    mpz_mul(aux,aux,*HNF->GetXY2(i,i));
		}

		mpz_pow_ui(aux2,HNF->denominator,Rank);
		mpz_divexact(aux,aux,aux2);

		for(i=0;mpz_cmp_ui(aux,1)>0;i++)
		{
			mpz_divexact_ui(aux,aux,aPrime);
		}

		(*aTarget)->Residual = i;
    }

    if (!this->BaseInverse)
	{
		int ret=ComputeOmegas();
		if (ret!=ConstRC::Ok) 
		{
			return ret;
		}
    }

    AlgebraicNumber **Beta,*Alfa,*Gama;
    short int *Lambda,R;
    Beta=NULL;

    Beta = (AlgebraicNumber**) Utils::AllocationMachine((void **) &Beta, (2*Rank-1)*sizeof(AlgebraicNumber*));
	
	if (!Beta) 
	{
		return ConstRC::NotEnoughMemory;
	}

    Lambda=NULL;
    Lambda = (short*) Utils::AllocationMachine((void **) &Lambda, (2*Rank-1)*sizeof(short));
    
	if (!Lambda) 
	{
		return ConstRC::NotEnoughMemory;
	}

    for(i=1;i<(long)Rank;i++)
	{
		Beta[i-1] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
		omega[i]->Multiply(Beta[i-1],aPrime,FALSE);
    }

    mpz_divexact(aux2,this->Base->denominator,HNF->denominator);

    for(i=0;i<(long)Rank;i++)
	{
		Beta[i+Rank-1] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
		for(j=0;j<(long)Rank;j++)
		{
			mpz_mul(aux,aux2,*HNF->GetXY2(j,i));
			Beta[i+Rank-1]->Set(j,aux);
		}
    }

    Alfa = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    Gama = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    bool finished=FALSE;
    mpz_ui_pow_ui(aux2,aPrime,(*aTarget)->Residual);

    for(R=1;!finished;R++)
	{
		for(i=0;i<(long)Rank*2-1;i++)
		{
			Lambda[i]=R;
		}
		j=1;

		while((j<(long)Rank*2-1)&&(!finished)) 
		{
			Alfa->Zeroize();
			for(i=0;i<(long)Rank*2-1;i++)
			{
				Beta[i]->Multiply(Gama,Lambda[i],FALSE);
				Alfa->Add(Alfa,Gama,FALSE);
//				printf(" %2d",Lambda[i]);
			}
//			puts("");
			Alfa->Norm(aux,ThetaPolynomial);
			mpz_divexact(aux,aux,aux2);

			if(!mpz_divisible_ui_p(aux,aPrime)) 
			{
				finished=TRUE;
			}
			else 
			{
				for(j=Rank*2-2;(j>=0)&&(!(Lambda[j]+R));j--);
				{
					if (j>=0)
					{
						Lambda[j]--;
						for(j++;j<(long)Rank*2-1;j++)
						{
							Lambda[j]=R;
						}
						for(j=0;(j<(long)Rank*2-1)&&(!Lambda[j]);j++);
						{
						//ISN'T HERE MISSING SOMETHING??
						}
					} else j=Rank<<1;
				}
			}
		}

		(*aTarget)->Prime=aPrime;
		(*aTarget)->Generator=Alfa;
		delete(Gama);

		for(i=0;i<(long)Rank*2-1;i++)
		{
			delete(Beta[i]);
		}
    }

    free(Beta);
    free(Lambda);
    mpz_clear(aux);
    mpz_clear(aux2);

    return ConstRC::Ok;
}

/*!
	Decomposes the principle ideal of aPrime as a product of prime ideals.
	The pointer aPrimeList should be a valid pointer to a NULL pointer.
	The method fills the pointer with an array of prime ideals of length aCount.
*/


int IntegralDomain::PrimeIdealDecomposition(PrimeIdeal ***aPrimeList, unsigned int *aCount, long aPrime)
{
    if (!Base) 
	{
		return ConstRC::GeneralError;
	}

    if (!aPrimeList) 
	{
		return ConstRC::GeneralError;
	}

    if (*aPrimeList)
	{
		return ConstRC::GeneralError;
	}

    //   The pointer aPrimeList has to be a valid pointer to a NULL pointer
    if (IsASpecialPrime(aPrime))
	{
		int ret;
		IntegerMatrix *Frob,*Kernel;
	    AlgebraicNumber **Bazove=NULL,*AN;
		unsigned int i;
		int j,k;
		long p=aPrime;
        mpz_t aux3,aux4;
        mpz_init(aux3);
        mpz_init(aux4);

        while ((unsigned)p<Rank)
		{
			p*=aPrime;
		}

        AN = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
	
		Bazove = (AlgebraicNumber**) Utils::AllocationMachine((void **) &Bazove,(this->Rank)*sizeof(AlgebraicNumber*));
        Kernel = new IntegerMatrix(Rank,Rank);
        Kernel->Allocate();
        Frob = new IntegerMatrix(Rank,Rank);
        ret = Frob->Zeroize();

        if (ret!=ConstRC::Ok)
		{
			return ret;
		}

    //    Base->PrintToScreen();

        for(i=0;i<Rank;i++)
		{
			Bazove[i] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
			for(j=0;(unsigned)j<Rank;j++)
			{
				Bazove[i]->Set(j,*Base->GetXY2(j,i));
			}

			Bazove[i]->Power(AN,p);

			for(j=Rank-1;j>=0;j--)
			{
				mpz_divexact(aux3,*AN->Get2(j),*Base->GetXY2(j,j));
				Frob->PutNumber(mpz_fdiv_ui(aux3,aPrime),j,i);
				for(k=j;k>=0;k--)
				{
					mpz_mul(aux4,aux3,*Base->GetXY2(k,j));
					AN->Dec(k,aux4);
				}
			}
		}

//		Frob->PrintToScreen();
		Frob->KernelModPrime(Kernel,&p,aPrime,TRUE);
//		Kernel->PrintToScreen();

		if (p>=(signed)Rank) 
		{
			return ConstRC::GeneralError;
		}

		(*aCount)=0;
        (*aPrimeList) = (PrimeIdeal**) Utils::AllocationMachine((void **) aPrimeList, (Rank)*sizeof(PrimeIdeal*));
		ret=BuchmannLenstra(Kernel,p,aPrimeList,aCount,aPrime);

		if (ret!=ConstRC::Ok) 
		{
			return ret;
		}

		if((*aCount)<Rank)
		{
			PrimeIdeal **PL= NULL;
			PL = (PrimeIdeal**) Utils::AllocationMachine((void **) &PL, (*aCount)*sizeof(PrimeIdeal*));
			if (!PL) 
			{
				return ConstRC::NotEnoughMemory;
			}

			for(i=0;i<*aCount;i++)
			{
				PL[i]=(*aPrimeList)[i];
			}

			if (*aPrimeList)
			{
				free(*aPrimeList);
			}
			(*aPrimeList) = PL;
		}

		delete(Kernel);
		delete(Frob);
		delete(AN);

		mpz_clear(aux3);
		mpz_clear(aux4);

		for(i=0;i<Rank;i++) 
		{
			delete(Bazove[i]);
		}

		free(Bazove);
	}
	else
	{
		mpz_t aux;
		mpz_init(aux);
		PolynomialSplitting *S;
		S = new PolynomialSplitting();
		S->Characteristic = aPrime;
		S->Split(ThetaPolynomial);
			(*aPrimeList) = (PrimeIdeal**) Utils::AllocationMachine((void **) aPrimeList, (S->Length)*sizeof(PrimeIdeal*));
			*aCount = S->Length;
			unsigned int i;
			int j;

        for(i=0;i<(*aCount);i++)
		{
			(*aPrimeList)[i] = new PrimeIdeal();
				(*aPrimeList)[i]->Prime = aPrime;
				(*aPrimeList)[i]->Generator = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
				(*aPrimeList)[i]->Residual = S->Factors[i]->GetDegree();
				
				if ((*aPrimeList)[i]->Residual<(signed)Rank)
				{
					for(j=(*aPrimeList)[i]->Residual;j>=0;j--)
					{
						 mpz_mul(aux,*S->Factors[i]->Get2(j),Base->denominator);
						(*aPrimeList)[i]->Generator->Set(j,aux);
					}
				}
				(*aPrimeList)[i]->Ramification = S->Exponents[i];
        }

		delete(S);
		mpz_clear(aux);
    }

    return ConstRC::Ok;
}

/*!
	Performs the Buchmann's-Lenstra's method of decomposing an algebra as a product of prime ideals.
*/
int IntegralDomain::BuchmannLenstra(IntegerMatrix *aIdeal, long aDimension, PrimeIdeal ***aPrimeList, unsigned int *aCount, long aPrime)
{
    unsigned i;
    int p,j,k,ret;
    long l,d;

    p=Rank-aDimension;
    IntegerMatrix *Gammas=NULL,*Map,*Ker,*I=NULL;
    AlgebraicNumber *AN,**Bazove=NULL;

    long *A=NULL,*S=NULL,*B=NULL;

    mpz_t aux3,aux4;
    mpz_init(aux3);
    mpz_init(aux4);

//    aIdeal->PrintToScreen();

    AN = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    Bazove = (AlgebraicNumber**) Utils::AllocationMachine((void **) &Bazove,(p)*sizeof(AlgebraicNumber*));
    A = (long*) Utils::AllocationMachine((void **) &A,(p)*sizeof(long));
    B = (long*) Utils::AllocationMachine((void **) &B,(p)*sizeof(long));
    S = (long*) Utils::AllocationMachine((void **) &S,(p)*sizeof(long));

    Bazove[0]=NULL;

    if (aDimension+1<(long)Rank) 
	{
		aIdeal->PutOne(0,aDimension);
		I = new IntegerMatrix(Rank,Rank);
		aIdeal->SupplementModPrime(I,aDimension+1,aPrime);
//			I->PrintToScreen();fflush(stdout);
		for(j=0;j<p;j++)
		{
			Bazove[j] = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
			Bazove[j]->Zeroize();
			for(i=0;i<Rank;i++)
			{
				for(k=0;k<(signed)Rank;k++)
				{
					mpz_mul_si(aux3,*Base->GetXY2(k,i),I->GetXY(i,j+aDimension));
					Bazove[j]->Inc(k,aux3);
				}
			}
		}
    
		Gammas = new IntegerMatrix(Rank,p*p);
		
		for(j=0;j<p;j++)
		{
			for(i=j;i<(unsigned)p;i++)
			{
				Bazove[i]->Multiply(AN,Bazove[j],FALSE);
/*				Bazove[i]->PrintToScreen();
				printf(" * ");
				Bazove[j]->PrintToScreen();
				printf(" = ");
				AN->PrintToScreen();
				puts("");
*/				for(k=Rank-1;k>=0;k--)
				{
					mpz_divexact(aux3,*AN->Get2(k),*Base->GetXY2(k,k));
					Gammas->PutNumber(mpz_fdiv_ui(aux3,aPrime),k,j*p+i);
		
					if (i!=(unsigned)j)
					{
						Gammas->PutNumber(mpz_fdiv_ui(aux3,aPrime),k,i*p+j);
					}

					for(l=k;l>=0;l--)
					{
						mpz_mul(aux4,aux3,*Base->GetXY2(l,k));
						AN->Dec(l,aux4);
					}
				}
			}
		}

//			Gammas->PrintToScreen();
		ret = I->InverseImageOfMatrixModPrime(Gammas,Gammas,aPrime,FALSE);
//			Gammas->PrintToScreen();

		if (ret!=ConstRC::Ok) 
		{
			return ret;
		}

		Map = new IntegerMatrix(p,p);
		Ker = new IntegerMatrix(p,p);

		for(j=0;j<p;j++)
		{
			for(k=1;k<p;k++) A[k]=0;
			A[0]=1;
			for(l=aPrime;l;l=l>>1)
			{
				if (l==aPrime) 
				{
					for(k=0;k<p;k++) 
					{
						S[k]=(k==j);
					}

				}
				else 
				{
					this->MultiplyInAlgebra(B,S,S,Gammas,p,aPrime);
					for(k=0;k<p;k++)
					{
						S[k]=B[k];
					}
				}
			
				if (l & 1) 
				{
					this->MultiplyInAlgebra(B,A,S,Gammas,p,aPrime);
					for(k=0;k<p;k++)
					{
						A[k]=B[k];
					}
				}
			}
	
			A[j]--;
	
			for(k=0;k<p;k++) 
			{
				Map->PutNumber(A[k],k,j);
			}
		}

//		Map->PrintToScreen();
		Map->KernelModPrime(Ker,&d,aPrime,TRUE);
//		Ker->PrintToScreen();
    
		for(j=0;j<p;j++)
		{
			A[j]=Ker->GetXY(j,1);
			B[j]=0;
		}

		delete(Map);
		delete(Ker);
	}
    else
	{
		d=1;
	}

    if (d==1) 
	{
		(*aPrimeList)[*aCount] = new PrimeIdeal();
		(*aPrimeList)[*aCount]->Residual = p;
		HermiteMatrix *Baze;
		Baze = new HermiteMatrix(Rank,Rank);
	
		mpz_mul_ui(aux4,Base->denominator,aPrime);
		mpz_pow_ui(aux3,aux4,Rank);
	
		Base->FormHermiteBase(Baze,Base,aIdeal,aDimension,aPrime,aux3);
		mpz_set(Baze->denominator,Base->denominator);
//		Baze->PrintToScreen();
	
		TwoElementRepresentation((*aPrimeList)+(*aCount),aPrime,Baze);
		Baze->Copy(Base);
		mpz_divexact_ui(Baze->denominator,Baze->denominator,aPrime);
	
		ValuationOfPrime(&d,(*aPrimeList)[*aCount],Baze);
		(*aPrimeList)[*aCount]->Ramification = d;
		(*aCount)++;
	
		delete(Baze);
    }
	else
	{
		Map = new IntegerMatrix(p,p+1);
		Ker = new IntegerMatrix(p+1,p+1);
		B[0]=1;
		Map->Zeroize();
		Map->PutOne(0,0);
	
		for(i=0;i<(unsigned)p;i++)
		{
			MultiplyInAlgebra(S,A,B,Gammas,p,aPrime);
			for(j=0;j<p;j++)
			{
				B[j]=S[j];
				Map->PutNumber(B[j],j,i+1);
			}
		}
	
//		Map->PrintToScreen();
		Map->KernelModPrime(Ker,&d,aPrime,TRUE);
//		Ker->PrintToScreen();
		Polynomial *P;
		P = new Polynomial(p-d+1);
		P->Zeroize(0);
	
		for(j=p-d+1;j>=0;j--)
		{
			P->Set(j,Ker->GetXY(j,0));
		}
	
		P->RootsModPrimeSpecial(B,aPrime);
//		P->PrintToScreen();
		delete(Map);
		Map = new IntegerMatrix(Rank,Rank+aDimension);
	
		for(i=0;i<Rank;i++)
		{
			for(j=0;j<aDimension;j++)
			{
				Map->PutNumber(aIdeal->GetXY(i,j),i,j+Rank);
			}
		}

		AN->Zeroize();
	
		for(j=0;j<p;j++)
		{
			for(k=0;k<(signed)Rank;k++)
			{
				for(i=0;i<Rank;i++)
				{
					mpz_mul_si(aux3,*Base->GetXY2(i,k),A[j]*I->GetXY(k,j+aDimension));
					AN->Inc(i,aux3);
				}
			}
		}
	
		if (!this->BaseInverse)
		{
			ComputeOmegas();
		}

		for(j=P->GetDegree()-1;j>=0;j--)
		{
			mpz_mul_si(aux4,Base->denominator,B[j]);
			AN->Dec(0,aux4);
//			AN->PrintToScreen();
	    
			for(i=0;i<Rank;i++)
			{
				omega[i]->Multiply(Bazove[0],AN,FALSE);
		
				for(l=Rank-1;l>=0;l--)
				{
					mpz_divexact(aux3,*Bazove[0]->Get2(l),*Base->GetXY2(l,l));
					Map->PutNumber(mpz_fdiv_ui(aux3,aPrime),l,i);
					
					for(k=l;k>=0;k--)
					{
					mpz_mul(aux4,aux3,*Base->GetXY2(k,l));
					Bazove[0]->Dec(k,aux4);
					}
				}
			}
	    
			mpz_mul_si(aux4,Base->denominator,B[j]);
			AN->Inc(0,aux4);
//			Map->PrintToScreen();
			Map->ImageModPrime(I,&d,aPrime);
			BuchmannLenstra(I,d,aPrimeList,aCount,aPrime);
		}
	
		delete(Ker);
		delete(Map);
		delete(P);
	}

    delete(AN);

    if (I) 
	{
		delete(I);
	}

    if (Gammas)
	{
		delete(Gammas);
	}

    if (Bazove[0]) 
	{
		for(k=0;k<p;k++)
		{
			delete(Bazove[k]);
		}
    }

    free(Bazove);
    free(A);
    free(B);
    free(S);

    mpz_clear(aux3);
    mpz_clear(aux4);

    return ConstRC::Ok;
}

/*!
	Performs the multiplication of two vectors in an algebra when the multiplication in the algebra
	is given by lower rows of the matrix aMultTable.
*/  

int IntegralDomain::MultiplyInAlgebra(long *aTarget, long *aOperand1, long *aOperand2, IntegerMatrix *aMultTable, unsigned int aSize, long aPrime)
{
    unsigned int i,j,k;
    INT64 a;

    for(k=0;k<aSize;k++) 
	{
		aTarget[k]=0;
	}

    for(i=0;i<aSize;i++)
	{
		for(j=0;j<aSize;j++)
		{
			for(k=0;k<aSize;k++)
			{
				a=aMultTable->GetXY(k+(Rank-aSize),i*aSize+j);
				a*=aOperand1[i];
				a%=aPrime;
				a*=aOperand2[j];
				a%=aPrime;
				a+=aTarget[k];
				aTarget[k]=a%aPrime;
			}
		}
    }
    return ConstRC::Ok;
}

/*!
	Adds a prime into the list of special primes.
*/


int IntegralDomain::AddToListOfSpecialPrimes(long aPrime)
{
    unsigned int i;

    if (SpecialPrimesAllocated==SpecialPrimesCount) 
	{
		SpecialPrimesAllocated+=DEFAULT_SPEC_PRIMES_REALLOC;
		long *S=NULL;
		S = (long*) Utils::AllocationMachine((void **) &S, (SpecialPrimesAllocated)*sizeof(long));
	
		if (!S) 
		{
			return ConstRC::NotEnoughMemory;
		}

		if (SpecialPrimesCount) 
		{
			for(i=0;i<SpecialPrimesCount;i++)
			{
				S[i]=SpecialPrimes[i];
			}
			free(SpecialPrimes);
		}

		SpecialPrimes=S;
    }
    SpecialPrimes[SpecialPrimesCount++]=aPrime;
    return ConstRC::Ok;
}

/*!
	Returns TRUE if aPrime is a special prime and FALSE otherwise.
*/

bool IntegralDomain::IsASpecialPrime(long aPrime)
{
    unsigned int i;

    for(i=0;i<SpecialPrimesCount;i++)
	{
		if (aPrime==SpecialPrimes[i]) 
		{
			return TRUE;
		}
	}
    return FALSE;
}

long IntegralDomain::Norm(Polynomial* aPoly,mpz_t aResult, mpz_t aOperand1, mpz_t aOperand2)
{
    if (!mpz_cmp_ui(aOperand1, 0))
	{
		mpz_pow_ui(aResult, aOperand2, Rank);
		mpz_mul(aResult, aResult, *aPoly->Get2(0));

		return mpz_get_si(aResult);
    }

    mpz_set_ui(aResult ,0);
	 

    mpz_pow_ui(iOp1Aux, aOperand1, Rank);

    int i, j = 0;
    for (i = Rank; i >= 0; i--)
	{	
		mpz_mul(iOp2Aux, iOp1Aux, *aPoly->Get2(i));
        if (!((j++) % 2)) 
		{
			mpz_add(aResult, aResult, iOp2Aux);
		} 
		else 
		{
			mpz_sub(aResult, aResult, iOp2Aux);
		}
		if (i) 
		{
			mpz_divexact(iOp1Aux, iOp1Aux, aOperand1);
			mpz_mul(iOp1Aux, iOp1Aux, aOperand2);
		}
    }

    
	if (mpz_fits_slong_p(aResult)) 
	{
		return mpz_get_si(aResult);
    } 
	else
	{
		return RESULT_OVERFLOW;
	}
}

/*!
	Computes the norm of <I>a</I>+<I>b</I>&theta;\endhtmlonly\latexonly$\theta$\endlatexonly.
*/

long IntegralDomain::Norm(mpz_t aResult, mpz_t aOperand1, mpz_t aOperand2){
	return Norm(ThetaPolynomial,aResult,aOperand1,aOperand2);
}

/*!
	Computes the norm of <I>a</I>+<I>b</I>&theta;\endhtmlonly\latexonly$\theta$\endlatexonly.
*/

long IntegralDomain::Norm(mpz_t aResult, long aOperand1, long aOperand2, Polynomial* aPoly){
	if (!aPoly)
	{
		aPoly = ThetaPolynomial;
	}
	mpz_set_si(iOp1Aux, aOperand1);
	mpz_set_si(iOp2Aux, aOperand2);
	long result = Norm(aPoly, aResult, iOp1Aux, iOp2Aux);
	return result;
}

/*!
	Computes the norm of <I>a</I>+<I>b</I>&theta;\endhtmlonly\latexonly$\theta$\endlatexonly
	modulo aPrime.
*/

long IntegralDomain::NormModPrime(long aOperand1, long aOperand2, long aPrime)
{
	INT64 aux;
	if (aOperand2 % aPrime)
	{
		aux = aOperand1 * CIntegerClass::InverseModN(-aOperand2,aPrime);
		aux = ThetaPolynomial->EvaluateModPrime(aux % aPrime,aPrime);
		aux *= CIntegerClass::PowerModN(aOperand2,Rank,aPrime);
		return aux % aPrime;
	} 
	else 
	{
		aux = CIntegerClass::PowerModN(aOperand1,Rank,aPrime) * mpz_fdiv_ui(*ThetaPolynomial->Get2(Rank),aPrime);
		return aux % aPrime;
	}
}

/*!
	Returns a base of the principal ideal (<I>a</I>+<I>b</I>&theta;\endhtmlonly\latexonly$\theta$\endlatexonly)<B>Z</B><sub>K</sub>.
*/
int IntegralDomain::ABaseOfPrincipal(HermiteMatrix *aTarget, long aOperand1, long aOperand2)
{
    mpz_t(aux);
    mpz_init(aux);
    AlgebraicNumber *A;
    A = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    mpz_mul_si(aux,this->Base->denominator,aOperand1);
    int ret = A->Set(0,aux);

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    mpz_mul_si(aux,this->Base->denominator,aOperand2);
    A->Set(1,aux);
    mpz_clear(aux);
    ret = ABaseOfPrincipal(aTarget, A);

    delete A;
    return ret;
}
/*!
	Returns a base of the principal ideal (<I>a</I>+<I>b</I>&theta;\endhtmlonly\latexonly$\theta$\endlatexonly)<B>Z</B><sub>K</sub>.
*/
int IntegralDomain::ABaseOfPrincipal(HermiteMatrix *aTarget, mpz_t aOperand1, long aOperand2)
{
    mpz_t(aux);
    mpz_init(aux);
    AlgebraicNumber *A;
    A = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);
    mpz_mul(aux,this->Base->denominator,aOperand1);
    int ret = A->Set(0,aux);

    if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    mpz_mul_si(aux,this->Base->denominator,aOperand2);
    A->Set(1,aux);
    mpz_clear(aux);
    ret = ABaseOfPrincipal(aTarget, A);

    delete A;
    return ret;
}

/*!
	Returns a base of the principal ideal generated by aGenerator.
*/

int IntegralDomain::ABaseOfPrincipal(HermiteMatrix *aTarget, AlgebraicNumber *aGenerator)
{
    if (aTarget->GetMaximalColumnIndex()!=(signed)Rank-1) 
	{
		printf("A base of principal col: %ld vs. %d.\n",aTarget->GetMaximalColumnIndex(),(signed)Rank-1);
		return ConstRC::SizeMismatch;
    }

    if (aTarget->GetMaximalRowIndex()!=(signed)Rank-1) 
	{
		printf("A base of principal row: %ld vs. %d.\n",aTarget->GetMaximalRowIndex(),(signed)Rank-1);
		return ConstRC::SizeMismatch;
    }

    int ret = aTarget->Zeroize();
    
	if (ret!=ConstRC::Ok) 
	{
		return ret;
	}

    if (!this->BaseInverse)
	{
		ret=ComputeOmegas();
		if (ret!=ConstRC::Ok) 
		{
			return ret;
		}
    }

    unsigned i,j;
    AlgebraicNumber *B;
    B = new AlgebraicNumber(Rank,ThetaPowers,&Base->denominator);

    for(i=0;i<Rank;i++)
	{
		omega[i]->Multiply(B,aGenerator,FALSE);
		for(j=0;j<Rank;j++)
		{
			aTarget->PutNumber(*B->Get2(j),j,i);
		}
    }

    mpz_set(aTarget->denominator,this->Base->denominator);
    delete B;
    return ConstRC::Ok;
}

int IntegralDomain::NormInZTheta(mpz_t aNorm, Polynomial *aPolynomial){
    aPolynomial->Resultant(aNorm,ThetaPolynomial,aPolynomial);
    return ConstRC::Ok;
}
