#ifndef __INTEGRAL_DOMAIN__
#define __INTEGRAL_DOMAIN__
#include "prime_ideal.h"

/*!
\brief	An Integrally-closed domain

The number field sieve works in a Dedekind domain. In the begining we have the domain
<B>Z</B>\htmlonly[&theta;\endhtmlonly\latexonly$[\theta$\endlatexonly] where \htmlonly&theta;\endhtmlonly\latexonly$\theta$\endlatexonly is a root of an irreducible polynomial.
Then we construct the integral closure and all the consecutive computations are made in a Dedekind domain
(integrally closed domain with all prime ideals maximal).
*/


class IntegralDomain {
public:
	IntegralDomain();
	~IntegralDomain();
	int AssignPolynomial(Polynomial *aPolynomial);
	int FindIntegralBase(long aBound);
	int AddToListOfSpecialPrimes(long aPrime);
	int TwoElementRepresentation(PrimeIdeal **aTarget, long aPrime, HermiteMatrix *HNF);
	int ValuationOfPrime(long *aValuation, AlgebraicNumber **aGenerators, unsigned int aGenCount, long aPrime, HermiteMatrix *aIdeal);
	int ValuationOfPrime(long *aValuation, PrimeIdeal *aPrime, HermiteMatrix *aIdeal);
	int PrimeIdealDecomposition(PrimeIdeal ***aPrimeList, unsigned int *aCount, long aPrime);
	bool IsASpecialPrime(long aPrime);
	long Norm(mpz_t aResult, mpz_t aOperand1, mpz_t aOperand2);
	long Norm(mpz_t aResult, long aOperand1, long aOperand2, Polynomial* aPoly = NULL);
	long Norm(Polynomial* aPoly,mpz_t aResult, mpz_t aOperand1, mpz_t aOperand2);
	long NormModPrime(long aOperand1, long aOperand2, long aPrime);
	int ABaseOfPrincipal(HermiteMatrix *aTarget, long aOperand1, long aOperand2);
	int ABaseOfPrincipal(HermiteMatrix *aTarget, mpz_t aOperand1, long aOperand2);
	int ABaseOfPrincipal(HermiteMatrix *aTarget, AlgebraicNumber *aGenerator);
	int NormInZTheta(mpz_t aNorm, Polynomial *aPolynomial);

	static int ComputeThetaPolynomial(Polynomial *aPolynomial, Polynomial *& aThetaPolynomial);

protected:
	int DedekindCriterion(unsigned int *aDimension, long aPrime);
	int PohstZassenhaus(unsigned int *aDimension, long aPrime);
	int ComputeOmegas();
	int BuchmannLenstra(IntegerMatrix *aIdeal, long aDimension, PrimeIdeal ***aPrimeList, unsigned int *aCount, long aPrime);
	int MultiplyInAlgebra(long *aTarget, long *aOperand1, long *aOperand2, IntegerMatrix *aMultTable, unsigned int aSize, long aPrime);

public:
	unsigned int Rank;///<	The rank of the ideal
	HermiteMatrix *Base;///<	The integral base in Hermite normal form
	mpz_t Discriminant;///<		The discriminant of the domain
	mpz_t Index;///<	The index of the starting domain in its integral closure
	long *SpecialPrimes;///<    The special primes list
	unsigned int SpecialPrimesCount;///<	The number of special primes
	Polynomial *ThetaPolynomial;	

protected:
	HermiteMatrix *BaseInverse;///<		The inverse matrix to the Base matrix
	AlgebraicNumber **omega;///<	The integral base
	AlgebraicNumber ***OmegaMultiplication;///<	The multiplication table of the integral base
	unsigned int SpecialPrimesAllocated;///<    The allocated length of SpecialPrimes array
// they used to be in algebraic_number_class.h
	HermiteMatrix **ThetaPowers;
	

private:
	int AllocatedThetaPowers;
	mpz_t iOp1Aux;
	mpz_t iOp2Aux;
};

#endif

