#ifndef __KLEINJUNG2_POLY_SELECTION_PHASE_H__
#define __KLEINJUNG2_POLY_SELECTION_PHASE_H__

#include "abstract_poly_selection_phase.h"
#include "candidate_pairs.h"
#include "const_poly_selection_phase.h"
#include "polynomial_improvement.h"
#include "structures.h"

/*!
 * \brief	Kleinjung's polynomial selection phase for NFS. Type (d,1).
 *

	Class is written according to paper "On Polynomial Selection for the General Number Field Sieve" by Thorsten Kleinjung.

	This algorithm generates two nonmonical polynomials. The first polynomial is of degree d > 3 and the second one has degree one.
	
	This is the best algorithm for generating polynomials for NFS for number N with more than 140 cipher.
*/
class CKleinjung2PolySelPhase : public APolyPhase
{

protected:

	//Logarithm of factor bases bounds
	mpf_t		iFBBoundLog[ConstPSPhase::GENERATED_POLYS_COUNT];
	///The inverse value of chi
	int			iChiInverse;

	mpz_t				iSievingIntervalLength;
	CCandidatePairs *	iCandidatePolyPairs; ///<results polynomial pairs - we save more for eventual sofisticated selection of the best poly pair, count is iCandidatesCount
	CPolynomialImprovement *	iPolyImprovement; ///<class for improving current polynomial pair if it is worth

	Polynomial *		iResultFirstPoly;
	Polynomial *		iResultSecondPoly;

	long				iRSerialFoundCandPolyCount; ///<used for serialization - another thread
	long				iRSerialRound; ///<used for serialization - another thread
	mpz_t				iRSerialAd; ///<used for serialization - another thread
	CCandidatePairs *	iRSerialCandidatePolyPairs; ///<used for serialization - another thread

	long				iRSerialRunAdChangeCount;
	long				iRSerialRunPolyCount;

	std::string	iCandPolyPairsFullFileName;
	std::string iRunningSerialFullFileName;

	int			iPolyDegree; ///{K2}<degree of the first polynomial, second polynomial has degree one
	int			iRegSize; ///Size of the sieving region, needed for rating MurphyE
	mpz_t		iAuxBound; ///<bound M, auxiliary bound used for calculation
	mpz_t		iAuxBound2; ///<bound M^{2}, auxiliary bound used for calculation
	mpz_t		iAuxBoundD; ///<bound M^{(2d - 6)/(d - 2)}, auxiliary bound used for calculation, d is poly degree
	mpz_t		iAdMax; ///{K2}<upper bound for the leading coeff of the first polynomial - a_{d}
	int *		iSetP; ///{K2}<set P of prime r fulfilling r = 1 mod 4
	int			iSetPSize; ///{K2}<size of the set P
	int *		iSetQ; ///{K2}<set Q of prime r fulfilling r != 1 mod 4
	int			iSetQSize; ///{K2}<size of the set Q
	int			iNumberOfPairsPr; ///{K2}number of identified pair of solutions (p, rp)
	int			iNumberOfPairsQr; ///{K2}number of identified pair of solutions (q, rq)
	int			iNumberOfPairsPi; ///{K2}number of identified pair of solutions (p, i)
	int			iPrimeBound; ///{K2}<Bound for primes in factorization of p, denoted as p_{b}, default value is 1000
	int *		iPrimitiveElementsP; ///{K2}<Array of primitive elements modulo primes r from P, size of the array is same as P = iSetPSize
	int *		iPrimitiveElementsQ; ///{K2}<Array of primitive elements modulo primes r from Q, size of the array is same as Q = iSetQSize
	int *		iNModSetPQ; ///{K2}<Array of N modulo prime r from P
	int			iNModSetPQSize; ///{K2}<Size of the array of N modulo prime r from P, should be same as P = iSetPSize
	int **		iMatrixPowersP; ///{K2}< iSetPSize x iPrimeBound matrix, for element a_{i,j} holds e_{i}^a_{i,j} = j mod r_{i} where r_{i} from P and e_{i} is primitive element modulo r_{i}
	int **		iMatrixPowersQ; ///{K2}< iSetQSize x iPrimeBound matrix, for element a_{i,j} holds e_{i}^a_{i,j} = j mod r_{i} where r_{i} from Q and e_{i} is primitive element modulo r_{i}
	mpz_t		iAd; ///{K2}<current leading coeff of the first polynomial
	mpz_t		idAd; ///{K2}<current leading coeff of the first polynomial multiplied by the degree of the polynomial
	mpz_t		iAdMin; ///<lower bound for the leading coeff of the first polynomial - a_{d}
	long		iRound; ///{K2}<current round, round = new one a_{d}
	long		iFoundCandidatePolyCount; ///<number of found polynomial - total number, not all are saved
	unsigned long	iAdBase; ///{K2}<leading coeff of the first polynomial has form a_{d} = base * ...

	mpz_t		iNtilde; ///{K2}auxilary value N tilde = d^d * a_{d}^(d-1) * N
	mpz_t		iMRoot0; ///{K2}rename to iMRoot0 - auxilarity value to count roots m_1 and m_2
	mpz_t		iMRoot1; ///{K2}root m_1 for base-(m_1, m_2) method
	mpz_t		iMRoot2; ///{K2}root m_2 for base-(m_1, m_2) method
	mpz_t		iAuxAdm1; ///{K2}temporary value for counting m_1 value
	volatile int		iPreM2; ///{K2}composition of m_2 from p1 p2 and q
	int			iCollision; ///{K2}collision of solutions of congruences modulo P1, P2 and Q1
	multimap<int,int>	iPairSolutionPr;///{K2}a solution of congruence, a pair (p, r), where p is a prime from the set P
	multimap<int,int>	iPairSolutionQr;///{K2}a solution of congruence, a pair (q, r), where q is a prime from the set Q
	multimap<int,int>	iPairSolutionPi;///{K2}a solution of congruence, a pair (p, i), where p is a prime from the set P and i is from rq + i * q^2 = rp (mod p^2)

	long *		iRoots; ///<roots modulo prime of selected polynomails - for rating calculation

	mpf_t		iRating; ///<final rating of current poly pair
	mpf_t		iAlpha; ///<aplha of current poly pair
	mpf_t		iSkewness; ///<skewness of current poly pair

	mutable mpz_t	iAuxMpz1; ///{K2}
	mutable mpz_t	iAuxMpz2;
	mutable mpz_t	iAuxMpz3;
	mutable mpz_t	iAuxMpz4;
	mutable mpz_t	iAuxMpz5;

	mutable mpf_t	iAuxMpfSum;
	mutable mpf_t	iAuxMpf;
	mutable mpf_t	iAUxMpfS1;
	mutable mpf_t	iAUxMpfS2;
	mutable mpf_t	iAUxMpfA1;
	mutable mpf_t	iAUxMpfA2;

protected:


	// Phase methods

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();
	int DisposeGMP();
	int DisposeMutexes();

	int InitParameters();
	int GenerateCandidates();
	int ChooseBest();
	int SaveResult();

	// auxiliary methods

	///Preparation of auxiliary fields for polynomial pairs generating
	int PrepareForGenerating();

	///Partial reset of supported values for selected a_d
	int PartialResetAd();

	double MurphyE(double Bf, double Bg, double area, int K);
	double DickmanRho(double x);

	double GetSpecialValuation (Polynomial * f, unsigned long p, mpz_t disc);
	int GetAlpha(Polynomial * thePolynomial, unsigned long B);

	//returns invers modulo: if((n*x) % p == 1) return x
	int modInverse(int n, int p);
	///Allocate and fill set P and Q
	int GenerateSetP();
	///Allocate and fill matrix Powers and array of primitive elements
	int GeneratePowers();
	///Allocate and fill array of N mod r from set P
	int GenerateNModSetPQ(int * SetPQ, int SetSize);
	///{K2}Solve r from a congruence N tilde == (m0 + r)^d mod p^2, where p is from set P or Q
	int SolveCongrModSqare( multimap<int,int>& PairSolution, int& NumberOfPairs, int * SetPQ, int SetSize, int * iPrimitiveElements, int ** iMatrixPowers );
	///Calculate i for: rq + i*q^2 == rp (mod p^2) for each p from iSetP
	int CalculateRpRqI();
	///Function finding a collision in pairs from solving congruency d-th root
	int FindCollision( multimap<int,int> PairSolution, int q );
	/// Find polynomials from prepared values, iAd, m_1, m_2, etc.
	int FindPolynomials();
	//Prepare values m1 and m2 from known p1, p2, q1 and ri collision
	int PrepareM1M2(int p1p2q1, int ri);
	///Base-(m1, m2) method, where N, m1 and m2 are the input with a_d already known(?)
	int BaseMPExpansion(mpz_t aM1, mpz_t aM2);
	///From indexes generates and saves polynomial pair
	int SaveCandidate();
	int GenerateESearchArray(e_search_element * aResults, double aSum, int aI, int aMaxI, int aIndex, int aDPower);
	///Computes the rating of the polynomial pair. 
	int Rating(mpf_t aResult, mpf_t aAlpha, mpf_t aSkewness);
	///Main executing method - searching algorithm
	int Generate();

	// Printing methods
/*	void PrintHeader();
	void PrintValues();
	void PrintEffectiveness();
*/
	// For running serialization
	int PrepareRunningSerialization();

public:

	CKleinjung2PolySelPhase();
	~CKleinjung2PolySelPhase();

	
	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;
	int Serialize(const CBaseParameters& aParam) const;

	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
	int Deserialize(const CBaseParameters& aParam);


};

#endif // __KLEINJUNG2_POLY_SELECTION_PHASE_H__
