#ifndef __KLEINJUNG_POLY_SELECTION_PHASE_H__
#define __KLEINJUNG_POLY_SELECTION_PHASE_H__

#include "abstract_poly_selection_phase.h"
#include "candidate_pairs.h"
#include "const_poly_selection_phase.h"
#include "polynomial_improvement.h"
#include "structures.h"

/*!
 * \brief Kleinjung's polynomial selection phase for NFS. Type (d,1).
 *
	Class is written according to paper "On Polynomial Selection for the General Number Field Sieve" by Thorsten Kleinjung.

	This algorithm generates two nonmonical polynomials. The first polynomial is of degree d > 3 and the second one has degree one.
	
	This is the best algorithm for generating polynomials for NFS for number N with more than 140 cipher.
*/
class CKleinjungPolySelPhase : public APolyPhase
{

protected:

	//Logarithm of factor bases bounds
	mpf_t		iFBBoundLog[ConstPSPhase::GENERATED_POLYS_COUNT];
	///The inverse value of chi
	int			iChiInverse;

	mpz_t				iSievingIntervalLength;
	CCandidatePairs *	iCandidatePolyPairs; ///<results polynomial pairs - we save more for eventual sofisticated selection of the best poly pair, count is iCandidatesCount
	CPolynomialImprovement *	iPolyImprovement; ///<class for improving current polynomial pair if it is worth

	Polynomial *		iResultFirstPoly;
	Polynomial *		iResultSecondPoly;

	long				iRSerialFoundCandPolyCount; ///<used for serialization - another thread
	long				iRSerialRound; ///<used for serialization - another thread
	mpz_t				iRSerialAd; ///<used for serialization - another thread
	CCandidatePairs *	iRSerialCandidatePolyPairs; ///<used for serialization - another thread

	long				iRSerialRunAdChangeCount;
	long				iRSerialRunPolyCount;

	std::string	iCandPolyPairsFullFileName;
	std::string iRunningSerialFullFileName;

	int			iPolyDegree; ///<degree of the first polynomial, second polynomial has degree one
	mpz_t		iAuxBound; ///<bound M, auxiliary bound used for calculation
	mpz_t		iAuxBound2; ///<bound M^{2}, auxiliary bound used for calculation
	mpz_t		iAuxBoundD; ///<bound M^{(2d - 6)/(d - 2)}, auxiliary bound used for calculation, d is poly degree
	mpz_t		iAdMax; ///<upper bound for the leading coeff of the first polynomial - a_{d}
	mpz_t		iAd1Max; ///<upper bound for the second coeff of the first polynomial - a_{d - 1}
	mpz_t		iAd2Max; ///<upper bound for the third coeff of the first polynomial  - a_{d - 2}
	mpz_t		iAuxMRoot; ///<auxiliary root m of f_{i} mod N, used for calculation
	mpz_t		iMRoot0; ///<the smallest integer bigger then auxiliary root m and divisible by p
	int *		iSetP; ///<set P of prime r fulfilling r = 1 mod d, r -| N, r < p_{b}
	int			iSetPSize; ///<size of the set P
	int *		iSetQad; ///<set Q_{a_{d}} of primes r from P fulfilling a_{d}/N != 0 mod r and it is d-th power modulo r, allocated size of the array is same as P = iSetPSize
	int			iSetQadSize; ///<actual size of the set Q_{a_{d}} - number of elements
	int *		iPP; ///<subset of Q_{a_{d}} of minimal size l (allocated as l_{max})
	int			iL; ///<Minimal number of primes in factorization of p, denoted as l, default value is 7
	int			iLMax; ///<Maximal number of primes in factorization of p, denoted as l_{max}, default value is l * 2
	int *		iLSizes; ///<Array of actual sizes of subsets of Q_{a_{d}} (allocated as iMatrixPPRowsAllocated)
	int			iPrimeBound; ///<Bound for primes in factorization of p, denoted as p_{b}, default value is 1000
	int			iMatrixPPRowsAllocated; ///<Number of allocated rows of matrix iMatrixPP
	mpz_t		iP; ///<Number p = \prod r from PP subset of Q_{a_{d}} - the leading coefficient of the second polynomial - f_{2}(x) = px - m 
	int *		iPrimitiveElements; ///<Array of primitive elements modulo primes r from P, size of the array is same as P = iSetPSize
	int *		iNModSetP; ///<Array of N modulo prime r from P, size of the array is same as P = iSetPSize
	int **		iMatrixPP; ///<matrix of subsets of Q_{a_{d}} of size minimal size l (allocated as iMatrixPPRowsAllocated x l_{max})
	int **		iMatrixPowers; ///< iSetPSize x iPrimeBound matrix, for element a_{i,j} holds e_{i}^a_{i,j} = j mod r_{i} where r_{i} from P and e_{i} is primitive element modulo r_{i} 
	int **		iMatrixXMod; ///< iSetPSize x d matrix of solutions of N = a_{d} * x^{d} mod r from Q_{a_{d}}, number of used rows is actual size of Q_{a_{d}}
	mpz_t **	iMatrixX; ///< l_{max} x d matrix of x_{i,j} - d^{l} solutions of N = a_{d} * x^{d} mod p
	mpz_t **	iMatrixM; ///< l_{max} x d matrix of m_{i,j} 
	mpz_t **	iMatrixE; ///< l_{max} x d matrix of e_{i,j}
	double **	iMatrixF; ///< l_{max} x d matrix of f_{i,j}
	double		iF0; ///<value f_{0}
	mpz_t		iAd; ///<current leading coeff of the first polynomial
	mpz_t		iAdMin; ///<lower bound for the leading coeff of the first polynomial - a_{d}
	long		iRound; ///<current round, round = new one a_{d}
	long		iFoundCandidatePolyCount; ///<number of found polynomial - total number, not all are saved
	unsigned long	iAdBase; ///<leading coeff of the first polynomial has form a_{d} = base * ...
	e_search_element *	iESearch1; ///<final search for values in epsilon neighbourhood - positive part
	e_search_element *	iESearch2; ///<final search for values in epsilon neighbourhood - negative part

	long *		iRoots; ///<roots modulo prime of selected polynomails - for rating calculation

	mpf_t		iRating; ///<final rating of current poly pair
	mpf_t		iAlpha; ///<aplha of current poly pair
	mpf_t		iSkewness; ///<skewness of current poly pair

	mutable mpz_t	iAuxMpz1;
	mutable mpz_t	iAuxMpz2;
	mutable mpz_t	iAuxMpz3;
	mutable mpz_t	iAuxMpz4;
	mutable mpz_t	iAuxMpz5;

	mutable mpf_t	iAuxMpfSum;
	mutable mpf_t	iAuxMpf;
	mutable mpf_t	iAUxMpfS1;
	mutable mpf_t	iAUxMpfS2;
	mutable mpf_t	iAUxMpfA1;
	mutable mpf_t	iAUxMpfA2;

protected:


	// Phase methods

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();
	int DisposeGMP();
	int DisposeMutexes();

	int InitParameters();
	int GenerateCandidates();
	int ChooseBest();
	int SaveResult();

	// auxiliary methods

	///Preparation of auxiliary fields for polynomial pairs generating
	int PrepareForGenerating();

	///Allocate and fill set P
	int GenerateSetP();
	///Allocate and fill matrix Powers and array of primitive elements
	int GeneratePowers();
	///Allocate and fill array of N mod r from set P
	int GenerateNModSetP();
	///Fill set Q_{a_{d}} and auxiliary matrix iMatrixXMod
	int FillSetQAd();
	///Add subset of Q_{a_{d}} to matrix PP
	int AddSubsetMatrixPP(int aCount, int & aRowsCount);
	///Recursion method for generation of subsets Q_{a_{d}} 
	int GenerateSubsetPP(int aPostion, int aCount, mpz_t aProd, int & aRowsCount);
	///Fill matrix PP of subsets Q_{a_{d}} 
	int FillMatrixPP(int & aRowsCount);
	///Fill matrix X with solutions of N = a_{d} * x^{d} mod p, where p is product of subset PP at position aSubsetPPIndex in matrix PP, also calculate p
	int GenerateMatrixXandP(int aSubsetPPIndex);
	///Fill matrix M - partial roots m
	int GenerateMatrixM(int aSubsetPPIndex);
	///Fill matrix E - partial values for calculation of a_{d - 1}
	int GenerateMatrixE(int aSubsetPPIndex);
	///Calculate coeff a_{d - 1} for current N, a_{d} and given root m modulo p
	int CalculateAd1ModP(mpz_t aM, mpz_t &Ad1);
	///Fill matrix F and f_{0} and bound epsilon = a_{d-2, max} / m_{0}
	int GenerateMatrixF(int aSubsetPPIndex, double & aEpsilon);
	int FindPolynomials(int aSubsetPPIndex, double aEpsilon);
	/// Base-(m,p) expansion of N (see Kleinjung proof of Lemma 1.2)
	int BaseMPExpansion(mpz_t aM, mpz_t aP);
	///From indexes generates and saves polynomial pair
	int SaveCandidate(int aIndex1, int aIndex2, int aDPower1, int aDPower2, int aL, int aLHalf);
	int GenerateESearchArray(e_search_element * aResults, double aSum, int aI, int aMaxI, int aIndex, int aDPower);
	///Computes the rating of the polynomial pair. 
	int Rating(mpf_t aResult, mpf_t aAlpha, mpf_t aSkewness);
	///Main executing method - searching algorithm
	int Generate();


	// Printing methods
	void PrintHeader();
	void PrintValues();
//	void PrintEffectiveness();

	// For running serialization
	int PrepareRunningSerialization();

public:

	CKleinjungPolySelPhase();
	~CKleinjungPolySelPhase();

	
	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;
	int Serialize(const CBaseParameters& aParam) const;

	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
	int Deserialize(const CBaseParameters& aParam);


};

#endif // __KLEINJUNG_POLY_SELECTION_PHASE_H__
