#include "lattice_hashtables.h"

#include "const_return_codes.h"
#include "const_element_flags.h"
#include "const_debug.h"
#include "const_sieving_phase.h"
#include "nfs_utils.h"
#include "log.h"

#ifdef DEBUG_PREFETCHING_ON

#if defined(__GNUC__)
#define PREFETCH(x)  __builtin_prefetch(x)
#else
#define PREFETCH(x) /* nothing */
#endif

#else

#define PREFETCH(x) /* nothing */

#endif

CLatticeHashtables::CLatticeHashtables()
{
	this->Hashtables = NULL;
	this->MaxIndex = -1;
	this->AllocatedSize = 0;

	this->iBlockSize = 0;
	this->iLogBlockSize = 0;
	this->iModBlockSize = 0;
}
CLatticeHashtables::~CLatticeHashtables()
{
	DeleteHashtables();
}
int CLatticeHashtables::SetBlockSize(int aBlockSize, int aLog, int aMod)
{
	this->iBlockSize = aBlockSize;
	this->iLogBlockSize = aLog;
	this->iModBlockSize = aMod;

	return ConstRC::Ok;
}
void CLatticeHashtables::DeleteHashtables()
{
	if (this->Hashtables != NULL)
	{
		lattice_script_t* ptr;
		lattice_script_t* next;
		lattice_hashtable_t* h1 = this->Hashtables;
		for (int i = 0; i <= this->MaxIndex; i++, h1++) 
		{
			if (h1 != NULL) 
			{
				ptr = h1->script;
				while (ptr)
				{
					next = ptr->next;
					free(ptr);
					ptr = next;
				}
			}
		}

		free(this->Hashtables);
		this->Hashtables = NULL;
		this->AllocatedSize = 0;
		this->MaxIndex = -1;
	}
}

int CLatticeHashtables::AllocateHashtables()
{
	if (this->Hashtables != NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeHashtables","AllocateHashtables",0,0,"Hashtables are not NULL");
		return ConstRC::GeneralError;
	}

	this->Hashtables = (lattice_hashtable_t*) Utils::AllocationMachine((void**) &this->Hashtables, this->AllocatedSize * sizeof(lattice_hashtable_t));
	if (this->Hashtables == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeHashtables","AllocateHashtables", 0, this->AllocatedSize * sizeof(lattice_hashtable_t),"Could not allocate space for the sieving hashtable");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->Hashtables, 0, this->AllocatedSize * sizeof(lattice_hashtable_t));	
	
	return ConstRC::Ok;
}
int CLatticeHashtables::SetupHashtables(int aCount, int aProbPrimeCount)
{
	if (this->Hashtables != NULL)// || this->Caches != NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeHashtables","SetupHashtables",0,0,"Caches or Hashtables is not NULL.");
		return ConstRC::BadArgument;
	}

	// we also setup hashtables with the security margin
	this->AllocatedSize = aCount;// + CLatticeHashtables::SECURITY_MARGIN;
	this->MaxIndex = aCount - 1;
	int rc = AllocateHashtables();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int j, max;
	max = aProbPrimeCount / STRUCTURE_MAX_UPDATES + 1;

	lattice_hashtable_t* t1 = this->Hashtables;
	lattice_script_t * next = NULL;
	for (int i = 0; i < this->AllocatedSize; i++, t1++)
	{
		t1->script = NULL;
		t1->script = (lattice_script_t*) Utils::AllocationMachine((void**)&t1->script, sizeof(lattice_script_t));
		if (t1->script == NULL)
		{
			CLog::LogError(WHEREARG,"CLatticeHashtables","SetupHashtables",i,sizeof(lattice_script_t),"Could not allocate space for the script inside a hashtable");
			return ConstRC::NotEnoughMemory;
		}
		t1->script->next = NULL;
		t1->script->num_used = 0;
		next = t1->script;
		for (j = 0; j < max; j++)
		{
			rc = AddScript(next);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			next = next->next;
		}
		t1->last_free_script = t1->script;
	}

	return ConstRC::Ok;
}
void CLatticeHashtables::ZeroizeLastHashtable()
{
	lattice_hashtable_t* last = this->Hashtables + this->MaxIndex;
	lattice_script_t* script = last->script;
	while (script)
	{
		script->num_used = 0;
		script = script->next;
	}
	last->last_free_script = last->script;
}

int CLatticeHashtables::SwapHashtable(int aIndex)
{
	if ((aIndex < 0) || (aIndex > this->MaxIndex))
	{
		return ConstRC::BadArgument;
	}
	lattice_script_t * ptrLast = this->Hashtables[aIndex].last_free_script;
	lattice_script_t * ptr = this->Hashtables[aIndex].script;
	this->Hashtables[aIndex].script = this->Hashtables[this->MaxIndex].script;
	this->Hashtables[aIndex].last_free_script = this->Hashtables[this->MaxIndex].last_free_script;
	this->Hashtables[this->MaxIndex].script = ptr;
	this->Hashtables[this->MaxIndex].last_free_script = ptrLast;
	ZeroizeLastHashtable();

	return ConstRC::Ok;
}

int CLatticeHashtables::ClearHashtables()
{
	lattice_hashtable_t* t1 = this->Hashtables;
	lattice_script_t* script;
	for (int i = 0; i <= this->MaxIndex; i++, t1++)
	{
		script = t1->script;
		while (script)
		{
			script->num_used = 0;
			script = script->next;
		}
		t1->last_free_script = t1->script;
	}

	return ConstRC::Ok;
}

int CLatticeHashtables::AddScript(lattice_script_t* aLast)
{
	if (aLast == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeHashtables","AddScript",0,0,"Add script found that last script is null!");
		return ConstRC::BadArgument;
	}
	// Next script already allocated
	if (aLast->next != NULL)
	{
		return ConstRC::Ok;
	}
	lattice_script_t* next = (lattice_script_t*) malloc(sizeof(lattice_script_t));
	if (next == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeHashtables","AddScript",0,sizeof(lattice_script_t),"Not enough memory to allocate new script!");
		return ConstRC::NotEnoughMemory;
	}
	memset(next,0,sizeof(lattice_script_t));
	aLast->next = next;
	// We will prefetch a bit of memory, because we add scripts in the 
	PREFETCH(next->list);

	return ConstRC::Ok;
}

void CLatticeHashtables::PrintHashtables()
{
	int i, j, max, counter;
	printf("\n\nPrint Lattice Hashtables start\n\n");
	lattice_script_t * script;
	lattice_update2_t *updates;
	for (i = 0; i <= this->MaxIndex; i++)
	{
		printf("\n\n HASHTABLE # %d: \n\n",i);
		script = this->Hashtables[i].script;
		counter = 0;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (j=0; j< max; j++) 
			{
				printf("(%6d:) Prime %6d, Root %6d, h = %6d.\n",counter, updates[j].element->p, updates[j].element->r, updates[j].h);
				counter++;
				
			}
			printf("----- list boundary -----\n");
			script = script->next;
		}
		printf("  This hashtable had %d entries.\n",counter);		
	}
	printf("\n\nPrint Lattice Hashtables end\n\n");
}

void CLatticeHashtables::PrintUpdateInfo(lattice_update2_t * aInfo)
{
	printf("\nLattice Update Type: \n");
	printf("     Log     : %d\n", aInfo->logp);
	printf("     H       : %d\n", aInfo->h);
	printf("     Prime   : %d\n", aInfo->element->p);
	printf("     R       : %d\n", aInfo->element->r);
	printf("     CR      : %d\n", aInfo->element->cr);
}
void CLatticeHashtables::SearchForPrimeInHashtable(main_sieving_type aPrime, main_sieving_type aRoot)
{
	int j, max;
	int total = 0;
	lattice_hashtable_t * h = this->Hashtables;
	lattice_script_t * script;
	lattice_update2_t * updates;
	for (int i = 0; i <= this->MaxIndex; i++, h++)
	{
		printf("Searching in hashtable %d ... ", i);
		fflush(stdout);

		script = h->script;
		bool found = false;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (j = 0; j < max; j++, updates++)
			{
				total++;
				if (updates->element->p == aPrime)
				{
					if (updates->element->r == aRoot) 
					{
						printf("FOUND!\n");
						found = true;
						PrintUpdateInfo(updates);
					}
				}
			}
			script = script->next;
		}
		if (!found) 
		{
			printf("NOT FOUND (listed total %d).\n", total);
		}
	}
}
