#ifndef __LATTICE_HASHTABLE_H__
#define __LATTICE_HASHTABLE_H__

#include "structures.h"
#include "const_sieving_phase.h"

/*!
 * \brief Hashtables for lattice sieving
 */
class CLatticeHashtables
{
private:
	///Security margin for the field of hashtables
	static const int SECURITY_MARGIN = 1;

public:
	///Internal hashtable for sieving with largish prime ideals.
	lattice_hashtable_t *	Hashtables;
	///Number of the allocated hashtables
	int				AllocatedSize;
	///Max index of elements
	int				MaxIndex;

private:

	// these values are for quick computations with blocks

	/*!
		Size of the block as an integer. 
		Should be equal to iLineSieveSize.
		This value should be power of 2.
	*/
	int	iBlockSize;
	/// Binary length of iBlockSize
	int	iLogBlockSize;
	/// We need modulo iBlockSize so we use bitwise AND with iBlockSize - 1
	int iModBlockSize;

public:

	CLatticeHashtables();
	~CLatticeHashtables();

#ifdef WIN32
	__forceinline int AddToTable(lattice_update_t * aUpdate, int aOffset, int aHashtableIndex)
#else
	int AddToTable(lattice_update_t * aUpdate, int aOffset, int aHashtableIndex) __attribute__((always_inline))
#endif
{

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (aHashtableIndex >= this->AllocatedSize)
	{
		printf("aHashtableIndex too high: %d for prime %d.\n", aHashtableIndex, aUpdate->element->p);
		return ConstRC::Ok;
	}
#endif

	int ret;

	lattice_script_t * script = this->Hashtables[aHashtableIndex].last_free_script;

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
	if ((aPrime == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) && (aCp))
	{
		printf("Add To Table: prime is %d, offset is %d, log is %d, c_p is %d, index in caches is %d.\n",aElement->p,aOffset,aLog,aElement->c_p,index);
	}
#endif
	
	if (script->num_used == STRUCTURE_MAX_UPDATES)
	{
		if (script->next == NULL)
		{
			ret = AddScript(script);
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
		}
		script = script->next;
		this->Hashtables[aHashtableIndex].last_free_script = script;
	}

	lattice_update2_t * update = script->list + script->num_used;
	script->num_used++;

	update->h		= aOffset;
	update->logp	= aUpdate->logp;
	update->element	= aUpdate->element;

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
	if ((aUpdate.element->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) && (aUpdate.element->c_p))
	{
		printf("AddToTable: DEBUG_SIEVING_PHASE_CRITICAL_PRIME.\n");
		PrintUpdateInfo(update);
		Utils::Wait();
	}
#endif
	
	return ConstRC::Ok;
}

	void DeleteHashtables();
	int  SetupHashtables(int aCount, int aProbPrimeCount);
	void ZeroizeLastHashtable();
	int  SwapHashtable(int aIndex);
	int  ClearHashtables();
	int  AddScript(lattice_script_t* aLast);

	int SetBlockSize(int aBlockSize, int aLog, int aMod);

	// Print methods
	void PrintHashtables();
	void PrintUpdateInfo(lattice_update2_t * aInfo);

	// Test methods
	void SearchForPrimeInHashtable(main_sieving_type aPrime, main_sieving_type aRoot);

private:
	int  AllocateHashtables();

	CLatticeHashtables(const CLatticeHashtables & aOperand) {}

};

#endif // __LATTICE_HASHTABLE_H__
