#ifndef __LATTICE_SIEVING_PHASE_H__
#define __LATTICE_SIEVING_PHASE_H__

#include "abstract_sieving_phase.h"
#include "nfs_relation.h"
#include "complex_structures.h"
#include "factor_base_info.h"
#include "nfs_threshold_optimizer.h"
#include "factorization_info.h"
#include "lp_hashtable_type1.h"
#include "frequency_hashtable.h"

///Class for lattice sieving phases - poly type (d,1)
class CLatticeSievingPhase : public ASievingPhase
{
#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
private:
	int iTimeSievingMiddle[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeSievingMiddleStart[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeSievingLargish[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeSievingLargishStart[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeFactorMiddle[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeFactorMiddleStart[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeFactorLargish[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeFactorLargishStart[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeFactorDivideSpecial[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	int iTimeFactorDivideSpecialStart[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
#endif

private:
	///Field of the factor bases
	CFBInfo iFBs[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	/*
		Array with number field info:
			- sieving poly -> generating polynomial Theta
			- integral domain
			- info about special primes - for factorization
			- other helpful stuff
	*/
	CNumberFieldInfo iNumberFields[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	/*!
		Info about variations used. 
		We use two "number fields" - (we can think about integral part as a number field)
		VariationInfo contains these things:
		 - variations mode = EDoNotUseVariations, ELargePrimeV, EDoubleLargePrimeV, ETrippleLargePrimeV
		 - factorization algorithm for factorizing remainder after FB factorization
		 - bounds - max remainder, min remainder, max factor
		 - leading coeff of the sieving polynomial - just for c_p calculation
	*/
	CVariationsInfo	iVariationsInfo[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	/*
		If we should use cycle construction method for creating smooth relations from partial relations.
		We can do that only if we use max 1,1-partial relations.
	*/
	bool			iUseCyclesForRelConstruction;
	// true iff we use n,m-partial relations, where n > 1 or m > 1 
	bool iUseHighVariations;

	int						iIteration;
	mutable pthread_mutex_t iIterationMutex;
	int						iMaxIteration;
	mutable pthread_mutex_t iMaxIterationMutex;

	int						iTimeMessage;
	mutable pthread_mutex_t iTimeMessageMutex;

	/*
		Number of smooth relations needed for factorization.
		BUT:
		If we use more than 1,1-relations than it is number of all relations - smooth + partial relations.
		This number is calculated according to some formula.
	*/
	long			iRequestedRelations;

	// For cache optimilized sieving

	///Number of block in the set of block - also number of preallocated and filled hashtables 
	int				iBlockSetSize;
	///Total number of sieving block in sieving area
	int				iBlockCount;

	///Number of lines in the sieving block
	int				iLinesInBlock;
	///Number of lines in the block set
	int				iLinesInBlockSet;

	int				iFactorBlockShift;
	int				iFactorBlockMask;
	int				iFactorBlockSize;
	log_type *		iFactorBlock;

	CFactorizationInfo	iFactorsInfo;
	/// Bound between middle primes and largish primes
	int			iLargishPrimesThreshold[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	// multiplier of factor base bound which gives bound for large primes
	int			iPartialFactorMultiplier[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	///Hashtables for sieving with largish primes
	CLatticeHashtables	iHashtables[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	

	///The sieving block itself
	log_type *		iSievingBlocks[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	///Advance factor bases - these bases are used for sieving - middle primes - and for filling hashtables
	CAdvLFBInfo		iAdvFB[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];


	/*!
		The value that will be used to fill the sieving block
		at the beginning of a new sieving cycle.

		It is basically 0x80 - (threshold log), so that aggregation of threshold_log
		would exceed 0x80.
	*/
	log_type		iInitBlockValues[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	/*!
		Change of the InitBlockValue - this value is optimized by CThresholdOptimizer
	*/
	int				iInitBlockValueChange[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	mpz_t			iAuxMpz1;
	mpz_t			iAuxMpz2;

	/*!
		This structure contains informations about main lattice - reducted base and sieving region.
		
		Sieving interval must be power of 2. If it is not then will be reset to closest power of 2.
		This value is full interval length but we expect to get I/2 from parameters.
		Some info about used interval length should be printed.
	*/
	main_lattice_info iLatticeInfo;
	int				iModI; // I - 1 for modulus
	int				iLogI; // Binary size of I

	long			iSpecialQ;
	int			iFirstSpecialQIndex;  // for distributed case
	int			iSpecialQCount;  // for distributed case
	long			iSpecialR;
	int				iRootIndex;

	double			iSkewness;

	// if given, sieving should stop after collecting this number of smooth relations (including cycles)
	// overrides usual check enough
	int iAssignedNumberOfSmooths;

	CNFSRelation*	iRelation;

	CLPType1Hashtable *			iLPHashtable;
	std::vector<long>			iIntermediateRoots;

	// hashtable to remember primes in higher variations
	CFrequencyHashtable *iHighVarHashtable;

	// serialization stuff
	std::vector<CNFSRelation*>* iRelationBuffer;
	mutable pthread_mutex_t		iRelationBufferMutex;
	int							iRelationBufferMaxCount;
	std::vector<CNFSRelation*>* iRelationWorkingBuffer;
	int							iRelationWorkingBufferMaxCount;
	int							iWorkingIteration;
	mutable pthread_mutex_t		iWorkingIterationMutex;
	long						iRSerialSpecialQ;
	int							iRSerialRootIndex;

	///For optimizing sieving threshold 
	CThresholdOptimizer		iOptimizer[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	int iThresholdOptInterval;

	std::string		iSmoothRelFullFileName;
	std::string		iPartialRelFullFileName;

	// for running serialization
	bool			iRSerialAccordingIters;
	int				iRSerialIterInterval;
	bool			iRSerialAccordingTime;
	long			iRSerialTimeInterval;
	bool			iRSerialAccordingRelations;
	int				iRSerialRelationsCount;
	std::string		iRunningSerialFullFileName;
	int				iRSerialInitBlockValueChange[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	// for distributed case

	// cumulative count of walk steps made with special Q since start of sieving (used by center through WalkWithSpecialQ method)
	long iSpecialQWalkSteps;

protected:

	// Sieving methods
	int SieveRegion();
	int SieveBlock(int aBlockSetIndex, int aBlockMin, int aBlockMax);

	int Sieving(int aIndex);

	int SieveWithMiddlePrimes(lattice_update_t* aStart, lattice_update_t* aEnd, log_type* aBlock);
	int SieveWithLargishPrimes(lattice_script_t* aScript, log_type* aBlock);

	int FindingAndFactorizing(int aBlockIndex, int aInnerIndex);

	int AddToCountStructures(CNFSRelation *aRelation);

	int Factorize2(int aBlockSetIndex, int aInnerIndex);
	int FactorizePrepareRelations();
	int FactorizeWithMiddlePrimes(int aIndex);
	int FactorizeWithLargishPrimes(int aIndex, int aHashtableIndex);
	int FactorizeDivide(int aIndex, int aInnerIndex, int aBlockSetIndex);
	int FactorizeProcessRelations();

	int Factorize(main_sieving_type aOffset, int aBlockSetIndex, int aInnerIndex);
	int FactorizeIntegral(int aI, int aJ, main_sieving_type aOffset, lattice_hashtable_t * aHashtable, int aIndex);
	int FactorizeAlgebraic(int aI, int aJ, main_sieving_type aOffset, lattice_hashtable_t *aHashtable, bool aDivisible, int aIndex);

	int FillHashtable(int aIndex, CLatticeHashtables & aHashtables, CAdvLFBInfo & aLFB);

	int ComputeExpectedLog(CNumberFieldInfo & aNF, CVariationsInfo & aVarInfo);

	int PrepareForSieving();

	void ReduceBase(main_sieving_type & outA0, main_sieving_type & outB0, 
					main_sieving_type & outA1, main_sieving_type & outB1,
					main_sieving_type inA0, main_sieving_type inB0, 
					main_sieving_type inA1, main_sieving_type inB1, float aSigma);

	// Debug methods
	void ControlBlock(int aBlockIndex, int aInnerIndex);

	// Large Prime Variation methods
	int PrepareCycleTable();
	int AddToCycleCountStructures(CNFSRelation* aRelation);
	int IndependentCyclesCount();
	int ReloadHashtable();
	int ReloadFreqHashtable();

	int FindFreeRelations();

	// Phase methods

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();
	int DisposeGMP();
	int DisposeMutexes();

	int RunSieve();

	int InitFactorBases();
	int InitAdvFactorBases();

	int InitParameters();

	// For running serialization
	int PrepareRunningSerialization();

public:

	CLatticeSievingPhase();
	~CLatticeSievingPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;
	int Serialize(const CBaseParameters& aParam) const;
	int SerializePart(int aIndex, xmlTextWriterPtr &aWriter) const;

	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
	int Deserialize(const CBaseParameters& aParam);
	int DeserializePart(int aIndex, xmlTextReaderPtr &aReader);

	// made public to allow call from distributed center

	int DetermineRequestedRelations();
	int CheckEnough();
	int SaveResult();

	// Printing methods
	void PrintHeader();
	void PrintValues();
	void PrintEffectiveness();

	// methods for distributed case

        // initializes sieving phase and generates factor bases (called by center)
        int InitDistributedSieving(AFactorAlgParameters* aParameters);

        // reads relations from given file, adds relations to count structures and appends to smooths or partials files (called by center)
        int SupplyRelations(const char *aPath, bool aSmooth, int &aRelsCount);

        // confirm all statistics before saving result (called by center)
        void ConfirmAllStats();

	// increment special Q by aSteps-iSpecialQWalkSteps
	void WalkWithSpecialQ(long aSteps);

	// Getters and Setters
public:
	int Get_Iteration() const;
	int	Set_Iteration(int aIteration);
	int Get_MaxIteration() const;
	int	Set_MaxIteration(int aMaxIteration);
	int Get_TimeMessage() const;
	int Set_TimeMessage(int aTimeMessage);
protected:
	int Get_WorkingIteration() const;
	int	Set_WorkingIteration(int aWorkingIteration);
	std::vector<CNFSRelation*>* Get_RelationBuffer() const;
	int							Set_RelationBuffer(std::vector<CNFSRelation*>* aRelationBuffer, int aCount);

};

#endif // __LATTICE_SIEVING_PHASE_H__
