#include "linear_phase.h"
#include "lanczos_class.h"
#include "matrix_helper.h"
#include "const_linear_phase.h"
#include "const_parameter_names.h"
#include "log.h"


using namespace std;

CLinearPhase::CLinearPhase() : MSerializable(ConstFN::LINEAR_PHASE), ALinearPhase()
{
	Set_AutoSerializable(ConstLPhase::AUTO_SERIALIZABLE);
	Set_InnerState((int)ELPhaseNotStarted);
	this->iMatrixSolver = NULL;
}

CLinearPhase::~CLinearPhase()
{
	Reset();
	DisposeFunctorField();
	DisposeMutexes();
	DisposeAcceptedReturnCodes();
}
int CLinearPhase::Reset()
{
	Set_MatrixResult(NULL);

	if (this->iMatrixSolver != NULL)
	{
		delete this->iMatrixSolver;
		this->iMatrixSolver = NULL;
	}

	APhase::Reset();

	Set_InnerState((int)ELPhaseNotStarted);
	return ConstRC::Ok;
}
int CLinearPhase::FillFunctorField()
{
	if (Get_FunctorsInit() == false)
	{
		Set_FunctorsInit(true);
		this->iPhaseFunctors.reserve(ConstLPhase::INNER_STATE_COUNT);
		this->iPhaseFunctors.clear();
		this->iPhaseFunctors.push_back(NULL); // ELPhaseNotStarted
		this->iPhaseFunctors.push_back(NULL); // ELPhaseInitialization
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CLinearPhase>(this, &CLinearPhase::SolveSystem)); // ELPhaseSolvingSystem
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CLinearPhase>(this, &CLinearPhase::SaveResult)); // ELPhaseSavingResult
		this->iPhaseFunctors.push_back(NULL); // ELPhaseFinished
	}

	return ConstRC::Ok;
}
int CLinearPhase::CleanUp(AFactorAlgParameters *aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	int auxInt;
	if (aParameters->GetIntegerPhaseParameter(ConstPN::LINEAR_PHASE_NUMBER, auxInt, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","CleanUp",0,0,"Didn't find phase number.");
    	return ConstRC::GeneralError;
	}

	return ALinearPhase::CleanUp(aParameters, auxInt, aType, aError);
}
int CLinearPhase::InitParameters()
{
	Set_InnerState((int)ELPhaseInitialization);

	AFactorAlgParameters * param = Get_Parameters();
	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	// generally we don't know order of the linear phase
	int auxInt;
	if (param->GetIntegerPhaseParameter(ConstPN::LINEAR_PHASE_NUMBER, auxInt, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","InitParameters",0,0,"Didn't find phase number.");
    	return ConstRC::GeneralError;
	}
	Set_PhaseNbr(auxInt);

	int rc = ALinearPhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// should we deserialize or serialize?
	bool deserialize = false;
	param->GetBoolPhaseParameter(ConstPN::DESERIALIZE, false, deserialize, Get_PhaseNbr());
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

	if (Get_InfoMode())
	{
		Utils::NormalMessage("Linear Phase is running ...");
    }

	if (deserialize)
	{
		rc = Deserialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

int CLinearPhase::SaveResult()
{
	Set_InnerState((int)ELPhaseSavingResult);

	AbstractMatrix * matrix = Get_MatrixResult();
	if (matrix == NULL)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","SaveResult",0,0,"Matrix result doesn't exist.");
		return ConstRC::GeneralError;
	}

	if (Get_InfoMode())
	{
		Utils::NormalMessage("   Writing results ...", NO_NUMBER);
    }

	// create parameters for serialization - assemble file name for result matrix
	CBaseParameters bparam;
	bparam.AddParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_MatrixResultFullFileName());
	
	// serialize matrix
	int rc = matrix->Serialize(bparam);
	if (rc == ConstRC::Ok)
	{
		// serialization is OK so we save file name for other phases

		AFactorAlgParameters * param = Get_Parameters();
		param->SetParameterForOthers(ConstPN::MATRIX_RESULT_FULL_FILENAME, 
			Get_MatrixResultFullFileName(), Get_PhaseNbr());
		string aux = Utils::ConvertToString((int)matrix->GetSpecies());		
		param->SetParameterForOthers(ConstPN::MATRIX_RESULT_TYPE, aux, Get_PhaseNbr());

		bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		Set_InnerState((int)ELPhaseFinished);
	}

	return rc;
}

int CLinearPhase::SolveSystem()
{
	Set_InnerState((int)ELPhaseSolvingSystem);

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tSolving linear equation system ...");
    }

	// read matrix to solve
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iRelationMatrixFullFileName);
	
	// create matrix 
	AbstractMatrix * relationMatrix = CMatrixHelper::CreateMatrix(this->iMatrixType);
	if (relationMatrix == NULL)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","SolveSystem",0,0,"Cannot create relation matrix.");
		return ConstRC::GeneralError;
	}

	// deserialize
	int rc = relationMatrix->Deserialize(bparam);
	if (rc != ConstRC::Ok)
	{
		if (relationMatrix != NULL)
		{
			delete relationMatrix;
			relationMatrix = NULL;
		}
		return rc;
	}

	// solve linear equation system
	AbstractMatrix * matrixResult = this->iMatrixSolver->Compute(relationMatrix);
	// delete used matrix
	if (relationMatrix != NULL)
	{
		delete relationMatrix;
		relationMatrix = NULL;
	}
	if (matrixResult == NULL)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","SolveSystem",0,0,"Cannot compute result.");
		return ConstRC::GeneralError;
	}

	Set_MatrixResult(matrixResult);

	return ConstRC::Ok;
}

int CLinearPhase::RunPhase(AFactorAlgParameters *aParameters)
{
	return APhase::RunPhase(aParameters);
}

int CLinearPhase::Serialize(const CBaseParameters &aParam) const
{
	int rc = MSerializable::Serialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (Running())
	{
		AFactorAlgParameters * param = Get_Parameters();

		// for deserialization at new start
		param->SetPhaseParameter(ConstPN::DESERIALIZE, Utils::ConvertToString(true), Get_PhaseNbr());
		param->SetPhaseParameter(ConstPN::PHASE_FULL_FILENAME, 
			MSerializable::CreateFileNameAuto(aParam, ConstFN::LINEAR_PHASE), Get_PhaseNbr());

		CBaseParameters paramb;
		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_ParametersFullFileName()); 
		rc = param->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok; 
}

int CLinearPhase::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_LINEAR_PHASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CLinearPhase","Serialize",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_INNER_STATE_NAME, Get_InnerState());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLinearPhase","Serialize",rc,0,"Could not end the parameter tag.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok; 
}

int CLinearPhase::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}

int CLinearPhase::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != ELPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_LINEAR_PHASE_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CLinearPhase","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int auxInt = 0;
	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INNER_STATE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize inner state...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_InnerState(auxInt);
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}

