#include "merge_frequency_hashtable.h"

#include "const_return_codes.h"
#include "const_hashtables.h"
#include "const_debug.h"
#include "const_filtering_phase.h"
#include "nfs_utils.h"
#include "log.h"

using namespace std;


CMergeFrequencyHashtable::CMergeFrequencyHashtable()
{
	this->Hashtable = NULL;
	this->AllocatedSize = 0;
	this->Mask = 0;
	this->RelationIndexes = NULL;
	this->iMaxFrequency = 0;
}
CMergeFrequencyHashtable::~CMergeFrequencyHashtable()
{
	DeleteHashtable();
}
void CMergeFrequencyHashtable::DeleteHashtable()
{
	if (this->Hashtable != NULL)
	{
		free(this->Hashtable);
		this->Hashtable = NULL;
		this->AllocatedSize = 0;
		this->Mask = 0;
	}

	if (this->RelationIndexes != NULL)
	{
		free(this->RelationIndexes);
		this->RelationIndexes = NULL;
	}

	this->iMaxFrequency = 0;
}
int CMergeFrequencyHashtable::SetupHashtable(relation_index_type aRelationCount, int aMaxFrequency)
{
	if (this->Hashtable != NULL || this->RelationIndexes != NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","SetupHashtable",0,0,"Hashtable or Indexes are not NULL.");
		return ConstRC::BadArgument;
	}

	this->iMaxFrequency = aMaxFrequency;

	long auxLong = 0;
	Utils::GetClosestBiggerPowerOf2(auxLong, aRelationCount << 1);
	this->AllocatedSize = auxLong;
	this->Hashtable = (merge_fhashtable_entry*) Utils::AllocationMachine((void**) &this->Hashtable, this->AllocatedSize * sizeof(merge_fhashtable_entry));
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","SetupHashtable", 0, this->AllocatedSize * sizeof(merge_fhashtable_entry),"Could not allocate space for the hashtable");
		return ConstRC::NotEnoughMemory;
	}
	this->Mask = this->AllocatedSize - 1;
	this->RelationIndexes = (relation_index_type*)Utils::AllocationMachine((void**) &this->RelationIndexes, this->iMaxFrequency * this->AllocatedSize * sizeof(relation_index_type));
	if (this->RelationIndexes == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","SetupHashtable", 0, 0,"Could not allocate space for the indexes");
		return ConstRC::NotEnoughMemory;
	}
	
	ClearHashtable();

	return ConstRC::Ok;
}
int CMergeFrequencyHashtable::ClearHashtable()
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","ClearHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	memset(this->Hashtable, 0, this->AllocatedSize * sizeof(merge_fhashtable_entry));	

	if (this->RelationIndexes == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","ClearHashtable",0,0,"Indexes are NULL.");
		return ConstRC::BadArgument;
	}

	memset(this->RelationIndexes, 0, this->iMaxFrequency * this->AllocatedSize * sizeof(relation_index_type));	

	return ConstRC::Ok;
}
/*
	This method calculates number of prime ideals in hashtable with given frequency (aCountWithFreq).
	It also calculates number of all prime ideals in hashtable.
		- this number is number of all prime ideals (at least once with odd exponent)
*/
int CMergeFrequencyHashtable::CountPrimeIdealsWithFrequency(unsigned int aFrequency, relation_index_type & aCountWithFreq, relation_index_type & aCountAll)
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","CountPrimeIdealsWithFrequency",0,0,"Hashtable or Part is NULL.");
		return ConstRC::BadArgument;
	}

	aCountWithFreq = 0;
	aCountAll = 0;
	merge_fhashtable_entry * aux = this->Hashtable;
	merge_fhashtable_entry * end = this->Hashtable + this->AllocatedSize;
	do 
	{
		if (aux->frequency > 0)
		{
			aCountAll++;
			if (aux->frequency == aFrequency)
			{
				aCountWithFreq++;
			}
		}
	}
	while (++aux < end);

	return ConstRC::Ok;
}
int CMergeFrequencyHashtable::AddToTable(main_sieving_type aPrime, main_sieving_type aRootFingerprint, relation_index_type aRelationIndex, int aWeight)
{
	if (this->Hashtable == NULL || this->RelationIndexes == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","AddToTable",0,0,"Hashtable or Indexes are NULL.");
		return ConstRC::InconsistentState;
	}

	bool present = false;
	relation_index_type index;
	
	if (FindPrimeInHashtable(aPrime, aRootFingerprint, present, index) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}
	
	if (!present) 
	{
		// This prime has not been in hashtable yet.					
		this->Hashtable[index].prime = aPrime;
		this->Hashtable[index].root_fingerprint = aRootFingerprint;
		this->Hashtable[index].frequency = 1;
		this->Hashtable[index].weight = aWeight;
		// we save index of the relation to the auxiliary field on the first position
		this->RelationIndexes[index * this->iMaxFrequency] = aRelationIndex;
	} 
	else 
	{
		// check if we have enough place for saving relation index
		if (this->Hashtable[index].frequency < this->iMaxFrequency)
		{
			// we save index of the relation to the auxiliary field on the next position
			this->RelationIndexes[index * this->iMaxFrequency + this->Hashtable[index].frequency] = aRelationIndex;
		}
		// We increase frequency of this primeideal
		this->Hashtable[index].frequency++;
		// increase weight
		this->Hashtable[index].weight += aWeight;
	}

	return ConstRC::Ok;
} 
int CMergeFrequencyHashtable::RemoveFromTable(main_sieving_type aPrime, main_sieving_type aRootFingerprint, relation_index_type aRelationIndex, int aWeight)
{
	if (this->Hashtable == NULL || this->RelationIndexes == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","RemoveFromTable",0,0,"Hashtable or Indexes are NULL.");
		return ConstRC::InconsistentState;
	}

	bool present = false;
	relation_index_type index;
	
	if (FindPrimeInHashtable(aPrime, aRootFingerprint, present, index) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}
	
	if (present) 
	{
		// we found prime ideal in hashtable so we decrease statistics
		if (this->Hashtable[index].frequency == 0)
		{
			// strange - this prime ideal should be in hashtable with positive frequency
			CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","RemoveFromTable",0,0,"Prime ideal has frequency zero.");
			return ConstRC::Ok;
		}
		int i;
		int frequency = this->Hashtable[index].frequency;
		relation_index_type * indexes = this->RelationIndexes + (index * this->iMaxFrequency);
		// remove index to the relation
		// first find index in array
		for (i = 0; i < frequency; i++)
		{
			if (indexes[i] == aRelationIndex)
			{
				break;
			}
		}
		// replace found index with the last one
		if (i < (frequency - 1))
		{
			indexes[i] = indexes[frequency - 1];
		}
		// descrease frequency - these also remove last index to the relation
		this->Hashtable[index].frequency--;
		// decrease weight
		this->Hashtable[index].weight -= aWeight;
	} 
	else 
	{
		// strange - this prime ideal should be in hashtable
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","RemoveFromTable",0,0,"Prime ideal not found in hashtable.");
	}

	return ConstRC::Ok;
}
/*!
	This method tries to find prime aPrime in hashtable, and indicates
	the result of this search in two values: the return value, which is an index
	corresponding to the real or expected position of the prime in the hashtable,
	and value of aPresent, which indicates whether the prime has already been
	in the hashtable or no. In the second case, the return value has meaning
	"if you want to insert aPrime, this index is the one where it should be inserted".

*/
int CMergeFrequencyHashtable::FindPrimeInHashtable(main_sieving_type aPrime, main_sieving_type aRootFingerprint, bool & aPresent, relation_index_type & aIndex) const
{
	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","FindPrimeInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aIndex = 0;
		return ConstRC::BadArgument;
	}

	relation_index_type hash = (aPrime >> 1) & this->Mask;			
	if (CollisionResolution(hash, aPrime, aRootFingerprint, aIndex) != ConstRC::Ok)
	{
		aPresent = false;
		aIndex = 0;
		return ConstRC::GeneralError;
	}

    if (this->Hashtable[aIndex].prime == 0) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

	return ConstRC::Ok;
}
/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where aPrime can be stored.

*/
int CMergeFrequencyHashtable::CollisionResolution(relation_index_type aHash, main_sieving_type aPrime, main_sieving_type aRootFingerprint, relation_index_type & aIndex) const
{
	aIndex = aHash;
	relation_index_type iter = 0;

	if (this->Hashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	merge_fhashtable_entry * aux = this->Hashtable + aIndex;
	merge_fhashtable_entry * end = this->Hashtable + this->AllocatedSize;

    while (((aux->prime != aPrime) || (aux->root_fingerprint != aRootFingerprint)) && (aux->prime != 0)) 
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->AllocatedSize) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","CollisionResolution",0,0,"The hash table has overflown.");
				return ConstRC::GeneralError;
			}
			aux = this->Hashtable;
		}
	}

	aIndex = aux - this->Hashtable;

	return ConstRC::Ok;
}

/*
	Copy all array of relation indexes which contain prime ideal wih given index.
	aRelationIndexes must be array of size at least iMaxFrequency.
*/
int CMergeFrequencyHashtable::GetRelationIndexes(relation_index_type aIndex, relation_index_type *& aRelationIndexes)
{
	if (this->RelationIndexes == NULL)
	{
		CLog::LogError(WHEREARG,"CMergeFrequencyHashtable","GetRelationIndexes",0,0,"Relation indexes are NULL.");
		return ConstRC::BadArgument;
	}

	// copy array of indexes to result array
	memcpy(aRelationIndexes, this->RelationIndexes + aIndex * this->iMaxFrequency, this->iMaxFrequency * sizeof(relation_index_type));

	return ConstRC::Ok;
}
