#include "minimum_spanning_tree.h"

#include "nfs_utils.h"
#include "const_return_codes.h"
#include "log.h"


CMinimumSpanningTreeAlg::CMinimumSpanningTreeAlg()
{
	this->iVerticesCount = 0;
	this->iAllocatedSize = 0;
	this->iUsedVertices = NULL;

	this->iCurrentUsed = 0;
	this->iCurrentRest = 0;
	this->i = 0;
	this->j = 0;
	this->iProcessedUsed = 0;
	this->iProcessedRest = 0;
	this->iValue = 0;
	this->iMinEdge.vertex1 = 0;
	this->iMinEdge.vertex2 = 0;
	this->iMinWeight = 0;
}

CMinimumSpanningTreeAlg::~CMinimumSpanningTreeAlg()
{
	DeleteAlg();
}

void CMinimumSpanningTreeAlg::DeleteAlg()
{
	this->iVerticesCount = 0;
	this->iAllocatedSize = 0;
	if (this->iUsedVertices != NULL)
	{
		free(this->iUsedVertices );
		this->iUsedVertices = NULL;
	}
	this->iCurrentUsed = 0;
	this->iCurrentRest = 0;
}

void CMinimumSpanningTreeAlg::ClearAlg()
{
	if (this->iUsedVertices != NULL)
	{
		for (this->i = 0; i < this->iAllocatedSize; i++)
		{
			this->iUsedVertices[i] = false;
		}
	}

	this->iVerticesCount = 0;
	this->iCurrentUsed = 0;
	this->iCurrentRest = 0;
}

int CMinimumSpanningTreeAlg::SetupAlg(int aVerticesCount)
{
	if (aVerticesCount <= 0)
	{
		CLog::LogError(WHEREARG,"CMinimumSpanningTreeAlg","SetupAlg ",0,0,"Vertices count cannot be nonpositive.");
		return ConstRC::BadArgument;
	}

	// setup field
	if (aVerticesCount > this->iAllocatedSize)
	{
		DeleteAlg();
		this->iAllocatedSize = aVerticesCount;
		this->iUsedVertices = (bool*)NFSUtils::AllocationMachine((void**)&this->iUsedVertices, this->iAllocatedSize * sizeof(bool));
		if (this->iUsedVertices == NULL)
		{
			CLog::LogError(WHEREARG,"CMinimumSpanningTreeAlg","SetupAlg ",0,0,"Could not allocate used vertices.");
			return ConstRC::NotEnoughMemory;
		}
	}

	ClearAlg();
	// setup values
	this->iVerticesCount = aVerticesCount;
	this->iCurrentUsed = 0;
	this->iCurrentRest = aVerticesCount;

	return ConstRC::Ok;
}
/*
	Solve minimum spanning tree problem with Prim's algorithm.
	We use adjecency matrix for saving graph and weights.
		- aVerticesCount is number of vertices
		- aAdjacencyMatrix can be greater than just aVerticesCount
		- aTree must be allocated with the size at least aVerticesCount-1

*/
int CMinimumSpanningTreeAlg::PrimAlgorithm(IntegerMatrix *aAdjacencyMatrix, int aVerticesCount, graph_edge_t *aTree)
{
	if (aAdjacencyMatrix == NULL || aTree == NULL)
	{
		CLog::LogError(WHEREARG,"CMinimumSpanningTreeAlg","PrimAlgorithm",0,0,"Matrix or Tree is NULL.");
		return ConstRC::BadArgument;
	}

	if (SetupAlg(aVerticesCount) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}


	this->iUsedVertices[0] = true;
	this->iCurrentUsed++;
	this->iCurrentRest--;


	while (this->iCurrentRest)
	{
		// reset counters for this run
		this->iProcessedRest = 0;
		this->iProcessedUsed = 0;
		this->iMinWeight = MAXIMAL_INT_VALUE;
		// go throught all vertices
		// find the shortest edge from used vertices to the unused vertices
		for (this->i = 0; this->i < this->iVerticesCount; this->i++)
		{
			if (this->iUsedVertices[i])
			{
				// i is now index of vertex which is used (it is in calculated component)
				// increase used vertices for this run
				this->iProcessedUsed++;
				// go throught all vertices
				for (this->j = i + 1; this->j < this->iVerticesCount; this->j++)
				{
					if (this->iUsedVertices[i] == false)
					{
						// j is now index of vertex which is not used (we can connect it with calculated component)
						// increase used vertices for this run
						this->iProcessedRest++;

						// get edge weight
						this->iValue = aAdjacencyMatrix->GetXY(i, j);
						// check if edge exists
						if (this->iValue > 0)
						{
							// check if value is the smallest
							if (this->iValue < this->iMinWeight)
							{
								this->iMinWeight = this->iValue;
								this->iMinEdge.vertex1 = i;
								this->iMinEdge.vertex2 = j;
							}
						}

						// if we processed all unused vertices then we are finished for this run
						if (this->iProcessedRest == this->iCurrentRest)
						{
							break;
						}
					}
				}

				// if we processed all used vertices then we are finished for this run
				if (this->iProcessedUsed == this->iCurrentUsed)
				{
					break;
				}
			}
		}

		if (this->iMinWeight < MAXIMAL_INT_VALUE)
		{
			// save result for this run
			aTree[this->iCurrentUsed - 1].vertex1 = this->iMinEdge.vertex1;
			aTree[this->iCurrentUsed - 1].vertex2 = this->iMinEdge.vertex2;
			// flag end vertex as used
			this->iUsedVertices[this->iMinEdge.vertex2] = true;
		}
		else
		{
			CLog::LogError(WHEREARG,"CMinimumSpanningTreeAlg","PrimAlgorithm",0,0,"We did not find the shortest edge.");
			return ConstRC::GeneralError;
		}

		this->iCurrentUsed++;
		this->iCurrentRest--;
	}

	return ConstRC::Ok;
}

/*
	Solve minimum spanning tree problem with Prim's algorithm.
	We use array of edges for saving graph and weights.
		- aEdgedCount is number of edges
		- aVerticesCount is number of vertices
		- aEdges is array of all edges with weights
		- aTree must be allocated with the size at least aVerticesCount-1

*/
int CMinimumSpanningTreeAlg::PrimAlgorithm2(graph_wedge_t *aEdges, int aEdgedCount, int aVerticesCount, graph_edge_t *aTree)
{
	if (aEdges == NULL || aTree == NULL)
	{
		CLog::LogError(WHEREARG,"CMinimumSpanningTreeAlg","PrimAlgorithm2",0,0,"Edges or Tree is NULL.");
		return ConstRC::BadArgument;
	}

	if (SetupAlg(aVerticesCount) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}

	// sort array of edges - from the shortest to the longest
	qsort(aEdges, aEdgedCount, sizeof(graph_wedge_t), &NFSUtils::CompareEdges);
	// use the first edge
	this->iUsedVertices[aEdges[0].vertex1] = true;
	this->iUsedVertices[aEdges[0].vertex2] = true;
	// save result for first edge
	aTree[0].vertex1 = aEdges[0].vertex1;
	aTree[0].vertex2 = aEdges[0].vertex2;
	// set counters
	this->iCurrentUsed += 2;
	this->iCurrentRest -= 2;


	while (this->iCurrentRest)
	{
		for (this->i = 1; this->i < aEdgedCount; this->i++)
		{
			// check if start vertex is in calculated component and end vertex is not
			if (this->iUsedVertices[aEdges[i].vertex1] == true && 
				this->iUsedVertices[aEdges[i].vertex2] == false || 
				this->iUsedVertices[aEdges[i].vertex1] == false && 
				this->iUsedVertices[aEdges[i].vertex2] == true)
			{
				// save result for this run
				aTree[this->iCurrentUsed - 1].vertex1 = aEdges[i].vertex1;
				aTree[this->iCurrentUsed - 1].vertex2 = aEdges[i].vertex2;
				// flag end vertex as used
				this->iUsedVertices[aEdges[i].vertex1] = true;
				this->iUsedVertices[aEdges[i].vertex2] = true;
				// set counters
				this->iCurrentUsed++;
				this->iCurrentRest--;
				break;
			}
		}
	}

	return ConstRC::Ok;
}
