#include "montgomery_poly_sel_phase.h"

#include <math.h>
#include "test_suite.h"
#include "const_parameter_names.h"
#include "log.h"

using namespace std;

CMontgomeryPolySelPhase::CMontgomeryPolySelPhase() : MSerializable(ConstFN::POLY_PHASE)
{
	this->iFoundCandidatePolyCount = 0;
	this->iRatedCandidatePolyCount = 0;
	this->iRSerialFoundCandPolyCount = 0;
	this->iRSerialRatedCandPolyCount = 0;
	this->iRSerialRunPrimeChangeCount = 0;
	this->iRSerialRunPolyCount = 0;
	this->iRound = 0;
	this->iRSerialRound = 0;

	for (int i = 0; i < ConstPSPhase::GENERATED_POLYS_COUNT; i++)
	{
		this->iFBBound[i] = 0;
	}

	this->iCandidatePolyPairs = NULL;
	this->iResultFirstPoly = NULL;
	this->iResultSecondPoly = NULL;

	mpz_init(this->iSievingIntervalLength);
	mpz_init(this->iRSerialRunningPrime);
	mpz_init(this->iRunningPrime);

	Set_AutoSerializable(ConstPSPhase::AUTO_SERIALIZABLE);
	Set_InnerState((int)EPSPhaseNotStarted);

	this->iStats = new CStatContainer(ConstPSPhase::INFO_STAT_COUNT);
	this->iTimes = new CTimeContainer(ConstPSPhase::INFO_TIME_COUNT);
}

CMontgomeryPolySelPhase::~CMontgomeryPolySelPhase()
{
	Reset();

	if (this->iStats != NULL)
	{
		delete this->iStats;
		this->iStats = NULL;
	}
	if (this->iTimes != NULL)
	{
		delete this->iTimes;
		this->iTimes = NULL;
	}

	DisposeFunctorField();
	DisposeGMP();
	DisposeMutexes();
	DisposeAcceptedReturnCodes();
}
int CMontgomeryPolySelPhase::Reset()
{
	this->iFoundCandidatePolyCount = 0;
	this->iRatedCandidatePolyCount = 0;
	this->iRSerialFoundCandPolyCount = 0;
	this->iRSerialRatedCandPolyCount = 0;
	this->iRound = 0;
	this->iRSerialRound = 0;

	for (int i = 0; i < ConstPSPhase::GENERATED_POLYS_COUNT; i++)
	{
		this->iFBBound[i] = 0;
	}

	if (this->iCandidatePolyPairs != NULL)
	{
		delete this->iCandidatePolyPairs;
		this->iCandidatePolyPairs = NULL;
	}

	if (this->iResultFirstPoly != NULL)
	{
		delete this->iResultFirstPoly;
		this->iResultFirstPoly = NULL;
	}

	if (this->iResultSecondPoly != NULL)
	{
		delete this->iResultSecondPoly;
		this->iResultSecondPoly = NULL;
	}

	if (this->iStats != NULL)
	{
		this->iStats->ResetAll();
	}
	if (this->iTimes != NULL)
	{
		this->iTimes->ResetAll();
	}

	APolyPhase::Reset();

	Set_InnerState((int)EPSPhaseNotStarted);
	return ConstRC::Ok;
}
int CMontgomeryPolySelPhase::DisposeGMP()
{
	mpz_clear(this->iSievingIntervalLength);
	mpz_clear(this->iRSerialRunningPrime);
	mpz_clear(this->iRunningPrime);

	return APolyPhase::DisposeGMP();
}
int CMontgomeryPolySelPhase::DisposeMutexes()
{
	return APolyPhase::DisposeMutexes();
}
int CMontgomeryPolySelPhase::FillFunctorField()
{
	if (Get_FunctorsInit() == false)
	{
		Set_FunctorsInit(true);
		this->iPhaseFunctors.reserve(ConstPSPhase::INNER_STATE_COUNT);
		this->iPhaseFunctors.clear();
		this->iPhaseFunctors.push_back(NULL); // EPSPhaseNotStarted
		this->iPhaseFunctors.push_back(NULL); // EPSPhaseInitialization
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CMontgomeryPolySelPhase>(this, &CMontgomeryPolySelPhase::GenerateCandidates)); // EPSPhaseSearchingForCandidates
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CMontgomeryPolySelPhase>(this, &CMontgomeryPolySelPhase::ChooseBest)); // EPSPhaseChoosingFromCandidates
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CMontgomeryPolySelPhase>(this, &CMontgomeryPolySelPhase::SaveResult)); // EPSPhaseSavingResult
		this->iPhaseFunctors.push_back(NULL); // EPSPhaseFinished
	}

	return ConstRC::Ok;
}
int CMontgomeryPolySelPhase::CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	string identifier;
	string directory = "";
	string file;

	int phaseNBR;
	if (aParameters->GetIntegerPhaseParameter(ConstPN::POLY_PHASE_NUMBER, phaseNBR, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","CleanUp",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}

	if (aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","CleanUp",0,0,"Didn't find serial identifier.");
		return ConstRC::GeneralError;
	}
	aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, directory, (int)EMain);

	switch (aType)
	{
		case ECleanAllIfNotError:
			if (aError)
			{
				break;
			}
		case ECleanAll:
		case ECleanAllExcCF:
		case ECleanAllExcCFRF:
			// Running Serialization full filename - default is same as iSerialPhaseFullFileName
			if (aParameters->ContainsPhaseParameter(ConstPN::RSERIAL_POLY_SEL_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::RSERIAL_POLY_SEL_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::RSERIAL_POLY_SEL_FILENAME, 
					ConstPN::RSERIAL_POLY_SEL_DIRECTORY, ConstPN::RSERIAL_POLY_SEL_COMPRESSION, directory,
					ConstFN::POLY_PHASE, identifier, phaseNBR);
			}
			remove(file.c_str());

			// Candidate polynomials full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::CAND_POLY_PAIR_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::CAND_POLY_PAIR_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::CAND_POLY_PAIR_FILENAME, 
					ConstPN::CAND_POLY_PAIR_DIRECTORY, ConstPN::CAND_POLY_PAIR_COMPRESSION, directory,
					ConstFN::CANDIDATE_POLY_PAIRS, identifier, phaseNBR);
			}
			remove(file.c_str());
			break;
		default:
			break;
	}

	return APolyPhase::CleanUp(aParameters, phaseNBR, aType, aError);
}
int CMontgomeryPolySelPhase::InitParameters()
{
	Set_InnerState((int)ESPhaseInitialization);

	AFactorAlgParameters * param = Get_Parameters();

	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	// generally we don't know order of the poly selection phase ;)
	int phaseNBR;
	if (param->GetIntegerPhaseParameter(ConstPN::POLY_PHASE_NUMBER, phaseNBR, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","InitParameters",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}
	Set_PhaseNbr(phaseNBR);

	// now we initialize members from parents's classes
	int rc = APolyPhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// some parameters must be resetup
	if (this->iAlphaAbort == ConstPSPhase::ALPHA_ABORT)
	{
		this->iAlphaAbort = ConstPSPhase::MONTGOMERY_ALPHA_ABORT;
	}
	if (this->iAlphaBound == ConstPSPhase::ALPHA_BOUND)
	{
		this->iAlphaBound = ConstPSPhase::MONTGOMERY_ALPHA_BOUND;
	}

	if (param->GetMpzPhaseParameter(ConstPN::SIEVING_REGION_WIDTH, this->iSievingIntervalLength, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","InitParameters",0,0,"Didn't find sieving interval length.");
		return ConstRC::GeneralError;
	}

	string auxStr;
	long bound = 0;
	// we try to find value for all 
	param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND, bound, bound, (int)EMain);
	for (int i = 0; i < ConstPSPhase::GENERATED_POLYS_COUNT; i++)
	{
		auxStr = Utils::ConvertToString(i + 1);
		// we try to find individual value
		param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND + auxStr, bound, bound, (int)EMain);
		this->iFBBound[i] = bound;
	}

	param->GetIntegerPhaseParameter(ConstPN::CHI_INVERSE, ConstPSPhase::MONTGOMERY_CHI_INVERSE, this->iChiInverse, phaseNBR);
	param->GetLongPhaseParameter(ConstPN::RSERIALIZATION_PRIME_CHANGE_COUNT, ConstPSPhase::PRIME_CHANGE_COUNT, this->iRSerialRunPrimeChangeCount, phaseNBR);
	param->GetLongPhaseParameter(ConstPN::RSERIALIZATION_POLY_COUNT, this->iCandidatesCount, this->iRSerialRunPolyCount, phaseNBR);

	// Running Serialization full filename - default is same as iSerialPhaseFullFileName
	if (param->ContainsPhaseParameter(ConstPN::RSERIAL_POLY_SEL_FULL_FILENAME, phaseNBR))
	{
		param->GetStringPhaseParameter(ConstPN::RSERIAL_POLY_SEL_FULL_FILENAME, this->iRunningSerialFullFileName, phaseNBR);
		MSerializable::AdaptPathSeparator(this->iRunningSerialFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iRunningSerialFullFileName, ConstPN::RSERIAL_POLY_SEL_FILENAME, 
			ConstPN::RSERIAL_POLY_SEL_DIRECTORY, ConstPN::RSERIAL_POLY_SEL_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::POLY_PHASE, Get_SerialIdentifier(), phaseNBR);
	}

	// Candidate polynomials full filename
	if (param->ContainsPhaseParameter(ConstPN::CAND_POLY_PAIR_FULL_FILENAME, phaseNBR))
	{
		param->GetStringPhaseParameter(ConstPN::CAND_POLY_PAIR_FULL_FILENAME, this->iCandPolyPairsFullFileName, phaseNBR);
		MSerializable::AdaptPathSeparator(this->iCandPolyPairsFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iCandPolyPairsFullFileName, ConstPN::CAND_POLY_PAIR_FILENAME, 
			ConstPN::CAND_POLY_PAIR_DIRECTORY, ConstPN::CAND_POLY_PAIR_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::CANDIDATE_POLY_PAIRS, Get_SerialIdentifier(), phaseNBR);
	}

	// should we deserialize or serialize?
	bool deserialize = false;
	param->GetBoolPhaseParameter(ConstPN::DESERIALIZE, false, deserialize, Get_PhaseNbr());
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

	if (Get_InfoMode())
	{
		Utils::NormalMessage("Montgomery Polynomial Selection Phase is running ...");
	}

	if (deserialize)
	{
		rc = Deserialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int CMontgomeryPolySelPhase::GenerateCandidates()
{
	Set_InnerState(EPSPhaseSearchingForCandidates);

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tGenerating polynomial ...");
	}

	this->iTimes->Start(EPSTGeneratePoly);

	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iRunningSerialFullFileName);
	int rc = StartRunningSerialization(param);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = Generate(false, false);
	if (rc != ConstRC::Ok)
	{
		EndRunningSerialization();
		return rc;
	}

	rc = EndRunningSerialization();	

	this->iTimes->End(EPSTGeneratePoly);

	return rc;
}
int CMontgomeryPolySelPhase::ChooseBest()
{
	Set_InnerState(EPSPhaseChoosingFromCandidates);

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tChoosing the best polynomial ...\n");
	}

	if (this->iFoundCandidatePolyCount <= 0)
	{
		return ConstRC::NoCandidates;
	}

	this->iTimes->Start(EPSTChooseBest);

	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iRunningSerialFullFileName);
	int rc = StartRunningSerialization(param);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	long * roots = NULL;
	roots = (long *)Utils::AllocationMachine((void **)&roots, ConstPSPhase::QUADRATIC_POLY_DEGREE * sizeof(long));
	if (roots == NULL)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","ChooseBest",0,0,"Could not allocate memory for roots.");
		return ConstRC::NotEnoughMemory;
	}

	mpf_t base1;
	mpf_t base2;
	mpf_t mx;

	mpf_init_set_d(base1, log((double)this->iFBBound[0]));
	mpf_init_set_d(base2, log((double)this->iFBBound[1]));
	mpf_init_set_ui(mx,0);

	long index;

	for (long i = 0; i < this->iRatedCandidatePolyCount; i++)
	{
		if (mpf_cmp(mx, this->iCandidatePolyPairs->Rating[i]) < 0) 
		{
			index = i;
			mpf_set(mx, this->iCandidatePolyPairs->Rating[i]);
		}
	}
	
	for (; this->iRatedCandidatePolyCount < this->iFoundCandidatePolyCount; this->iRatedCandidatePolyCount++)
	{
		if (Get_InfoMode())
		{
			printf("\r\t\tCandidate polynomials rated: %ld", this->iRatedCandidatePolyCount);
			fflush(stdout);
		}

		Rating(this->iCandidatePolyPairs->Rating[this->iRatedCandidatePolyCount], 
			this->iCandidatePolyPairs->Alpha[this->iRatedCandidatePolyCount], roots,
			this->iCandidatePolyPairs->FirstPoly[this->iRatedCandidatePolyCount],
			this->iCandidatePolyPairs->SecondPoly[this->iRatedCandidatePolyCount], base1, base2);

		if (mpf_cmp(mx, this->iCandidatePolyPairs->Rating[this->iRatedCandidatePolyCount]) < 0) 
		{
			index = this->iRatedCandidatePolyCount;
			mpf_set(mx, this->iCandidatePolyPairs->Rating[this->iRatedCandidatePolyCount]);
		}
		if ((this->iRatedCandidatePolyCount + 1) % this->iRSerialRunPolyCount == 0)
		{
			rc = RunRunningSerialization();
			if (rc != ConstRC::Ok)
			{
				mpf_clear(base1);
				mpf_clear(base2);
				mpf_clear(mx);

				free(roots);
				roots = NULL;

				EndRunningSerialization();

				return rc;
			}
		}
	}

	free(roots);
	roots = NULL;

	this->iResultFirstPoly = new Polynomial();
	this->iCandidatePolyPairs->FirstPoly[index]->Copy(this->iResultFirstPoly);
	this->iResultSecondPoly = new Polynomial();
	this->iCandidatePolyPairs->SecondPoly[index]->Copy(this->iResultSecondPoly);
	mpz_set(this->iRoot, this->iCandidatePolyPairs->Root[index]);

	if (Get_InfoMode())
	{
		printf("\n\tThe best polynomial pair found is\n");
		this->iResultFirstPoly->PrintToScreen();
		this->iResultSecondPoly->PrintToScreen();
		gmp_printf("\n\t\twith alpha = %.*Ff and rating = %.*Ff\n\n", 10, this->iCandidatePolyPairs->Alpha[index], 10, mx);
	}

	mpf_clear(base1);
	mpf_clear(base2);
	mpf_clear(mx);

	rc = EndRunningSerialization();

	this->iTimes->End(EPSTChooseBest);

	return rc;
}
int CMontgomeryPolySelPhase::SaveResult()
{
	if (this->iResultFirstPoly == NULL || this->iResultSecondPoly == NULL)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","SaveResult",0,0,"Result polynomial is not determined.");
		return ConstRC::GeneralError;
	}

	Set_InnerState(EPSPhaseSavingResult);

	AFactorAlgParameters * param = Get_Parameters();
	string auxStr = Utils::ConvertToString(this->iRoot);
	param->SetPhaseParameter(ConstPN::ROOT_M, auxStr, (int)EMain);
	param->SetPhaseParameter(ConstPN::SIEVING_POLY1, this->iResultFirstPoly->ToString(ELongFormat), (int)EMain);
	param->SetPhaseParameter(ConstPN::SIEVING_POLY2, this->iResultSecondPoly->ToString(ELongFormat), (int)EMain);

	Set_InnerState(EPSPhaseFinished);

	return ConstRC::Ok;
}
/*!
	Computes the rating of the polynomial pair. 
	Returns also the average alpha of the polynomials.
*/
int CMontgomeryPolySelPhase::Rating(mpf_t aRating, mpf_t aAlpha, long * aRoots, Polynomial *aPoly1, Polynomial *aPoly2, mpf_t aLogBase1, mpf_t aLogBase2)
{
	mpf_t aux, sum, alpha1, alpha2;
	mpf_init(aux);
	mpf_init(sum);
	mpf_init(alpha1);
	mpf_init(alpha2);
	AlphaPoly(alpha1, aRoots, aPoly1);
	AlphaPoly(alpha2, aRoots, aPoly2);

	mpf_set_ui(aRating ,0);
	for (long i = 0; i < this->iSubintervalsCount; i++)
	{
		EvaluateAngle(sum, aPoly1, i);
		mpf_set_d(aux, Utils::Ln(sum));
		mpf_add(sum, alpha1, aux);
		mpf_div(aux, sum, aLogBase1);
		Rho(aAlpha, aux); // we use aAlpha as aux variable

		EvaluateAngle(sum, aPoly2, i);
		mpf_set_d(aux, Utils::Ln(sum));
		mpf_add(sum, alpha2, aux);
		mpf_div(aux, sum, aLogBase2);
		Rho(sum, aux);

		mpf_mul(sum, sum, aAlpha);
		mpf_add(aRating, aRating, sum);
	}

	mpf_add(sum, alpha1, alpha2);
	mpf_div_ui(aAlpha, sum, 2);

	mpf_clear(aux);
	mpf_clear(sum);
	mpf_clear(alpha1);
	mpf_clear(alpha2);

	return ConstRC::Ok;
}
/*!
	Gauss base reduction (like LLL for 2-dim lattice) 
	more Cohen - A Course in Computational Algebraic Number Theory - algorithm 1.3.14
	same notation, using class Polynomial as mpz_t vector
	Polynomial a is the smallest one as the result.
*/
int CMontgomeryPolySelPhase::GaussReduction(Polynomial *& a, Polynomial *& b)
{
	if (a != NULL && b != NULL)
	{
		mpz_t A,B,T,TEMP,N,R;
		Polynomial * temp;
		
		mpz_init(A);
		mpz_init(B);
		mpz_init(T);
		mpz_init(TEMP);
		mpz_init(N);
		mpz_init(R);
		
		Polynomial::EuclideNormSqr(A, a);
		Polynomial::EuclideNormSqr(B, b);

		if (mpz_cmp(B, A) > 0)
		{
			temp = a;
			a = b;
			b = temp;
			mpz_swap(A, B);
		}
		
		// N = a * b
		Polynomial::InnerProduct(N, a, b);
		
		// R = [N/B]
		mpz_cdiv_q_ui(TEMP, B, 2);
		mpz_add(TEMP, TEMP, N);
		mpz_fdiv_q(R, TEMP, B);
		
		// T = A - 2RN + RRB
		mpz_set(T, A);
		mpz_pow_ui(TEMP, R, 2);
		mpz_mul(TEMP, TEMP, B);
		mpz_add(T,T,TEMP);
		mpz_mul(TEMP, R, N);
		mpz_mul_si(TEMP, TEMP, -2);
		mpz_add(T, T, TEMP);		
		
		while (mpz_cmp_ui(R, 0) != 0)
		{
			// a = a - Rb
			mpz_neg(R, R);
			a->AddMultiply(a, b, R);
			
			temp = b;
			b = a;
			a = temp;
			
			mpz_set(A, B);
			mpz_set(B, T);
			
			// N = a * b
			Polynomial::InnerProduct(N, a, b);
			
			// R = [N/B]
			mpz_cdiv_q_ui(TEMP, B, 2);
			mpz_add(TEMP, TEMP, N);
			mpz_fdiv_q(R, TEMP, B);
			
			// T = A - 2RN + RRB
			mpz_set(T, A);
			mpz_pow_ui(TEMP, R, 2);
			mpz_mul(TEMP, TEMP, B);
			mpz_add(T, T, TEMP);
			mpz_mul(TEMP, R, N);
			mpz_mul_si(TEMP, TEMP, -2);
			mpz_add(T, T, TEMP);			
		}
		
		
		mpz_clear(A);
		mpz_clear(B);
		mpz_clear(T);
		mpz_clear(TEMP);
		mpz_clear(N);
		mpz_clear(R);
		
		return ConstRC::Ok;
	}
	else
	{
		return ConstRC::BadArgument;		
	}
}

int CMontgomeryPolySelPhase::Generate(bool aAdvanced, bool aEnforceRealRoots)
{
	// init
	// ----------------------------------------------------
	long * roots = NULL;
	roots = (long *)Utils::AllocationMachine((void **)&roots, ConstPSPhase::QUADRATIC_POLY_DEGREE * sizeof(long));
	if (roots == NULL)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","Generate",0,0,"Could not allocate memory for roots.");
		return ConstRC::NotEnoughMemory;
	}
	int rc;
	long delay = 0;
	Polynomial * a = new Polynomial(ConstPSPhase::QUADRATIC_POLY_DEGREE); // quadratic polynomial
	Polynomial * b = new Polynomial(ConstPSPhase::QUADRATIC_POLY_DEGREE); // quadratic polynomial
	mpz_t squareRootOfN, c1, c2, s, aux1, aux2, auxM, squareM, coeffBound, zero;
	mpf_t alpha;
	mpz_init(squareRootOfN);
	mpz_init(c1);
	mpz_init(c2);
	mpz_init(s);
	mpz_init(aux1);
	mpz_init(aux2);
	mpz_init(auxM);
	mpz_init(squareM);
	mpf_init(alpha);
	mpz_init(coeffBound);
	mpz_init_set_ui(zero, 0);

	// expected size of coeff
	mpz_root(aux1, this->iN, (unsigned long)(2 * ConstPSPhase::QUADRATIC_POLY_DEGREE));
	// max size of poly coeff - according to chi
	mpz_fdiv_q_ui(coeffBound, aux1, (unsigned long)this->iChiInverse);
	
	mpz_sqrt(squareRootOfN, this->iN);
	mpz_pow_ui(squareM, this->iSievingIntervalLength, 2);
	
	if (aAdvanced)
	{
		mpz_fdiv_q(squareRootOfN, squareRootOfN, this->iSievingIntervalLength);
		mpz_mul(coeffBound, coeffBound, this->iSievingIntervalLength);
	}
	
	if (this->iCandidatePolyPairs == NULL) // can be deserialize
	{
		// first run init
		this->iCandidatePolyPairs = new CCandidatePairs(this->iCandidatesCount);
		mpz_set(this->iRunningPrime, squareRootOfN);
		this->iRound = 0;
	}
	// ---------------------------------------------------

#ifndef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
	if (Get_InfoMode())
	{
		PrintHeader();
	}
#endif

	// generating candidates
	while (this->iFoundCandidatePolyCount < this->iCandidatesCount)
	{
		this->iRound++;
		if (this->iRound % this->iRSerialRunPrimeChangeCount == 0)
		{
			rc = RunRunningSerialization();
			if (rc != ConstRC::Ok)
			{
				goto montgomery_generate_clear_gmp;
			}
		}

		Utils::PreviousPrime(this->iRunningPrime, this->iRunningPrime);

#ifndef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
		if (Get_InfoMode())
		{
			PrintValues();
		}
#endif
		
		if (mpz_jacobi(this->iN, this->iRunningPrime) != 1)
		{
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			//gmp_printf("\t\tprime is %Zd; (this->iN/p) = -1 \n", this->iRunningPrime);
#endif			
			// this->iN has to be quadratic residue mod p
			continue;
		}
		
		// compute c1
		Utils::TonelliShanks(c1, this->iN, this->iRunningPrime);
		
		mpz_sub(aux1, squareRootOfN, c1);
		mpz_fdiv_q_ui(aux2, this->iRunningPrime, 2);
				
		if (mpz_cmp(aux1, aux2) > 0)
		{
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			gmp_printf("\t\tprime is %Zd; c1 is %Zd; abs(c1-sqN) is %Zd \n", this->iRunningPrime, c1, aux1);
#endif			
			// c1 is too big
			continue;
		}
		
		// compute c2
		mpz_invert(s, c1, this->iRunningPrime);
		mpz_pow_ui(aux1, c1, 2);
		mpz_sub(aux1, aux1, this->iN);
		mpz_divexact(c2, aux1, this->iRunningPrime);
		
		if (aAdvanced)
		{
			// set a = (c1, -p * this->iSievingIntervalLength, 0 * this->iSievingIntervalLength^2)
			a->Set(0, c1);   // first 
			mpz_neg(aux1, this->iRunningPrime);
			mpz_mul(aux1, aux1, this->iSievingIntervalLength);
			a->Set(1, aux1); // second
			a->Set(2, zero); // third
			
			// set b = ((c1(c2*s mod p) - c2)/p, -(c2*s mod p) * this->iSievingIntervalLength, 1 * this->iSievingIntervalLength^2)
			mpz_mul(aux1, c2, s);
			mpz_mod(aux1, aux1, this->iRunningPrime);
			mpz_mul(aux2, c1, aux1);
			mpz_sub(aux2, aux2, c2);
			mpz_divexact(aux2, aux2, this->iRunningPrime);
			b->Set(0, aux2);     // first
			mpz_neg(aux1, aux1);
			mpz_mul(aux1, aux1, this->iSievingIntervalLength);
			b->Set(1, aux1);     // second
			b->Set(2, squareM);	 // third
		}
		else
		{
			// set a = (c1, -p, 0)
			a->Set(0, c1);   // first 
			mpz_neg(aux1, this->iRunningPrime);
			a->Set(1, aux1); // second
			a->Set(2, zero); // third
			
			// set b = ((c1(c2*s mod p) - c2)/p, -(c2*s mod p), 1)
			mpz_mul(aux1, c2, s);
			mpz_mod(aux1, aux1, this->iRunningPrime);
			mpz_mul(aux2, c1, aux1);
			mpz_sub(aux2, aux2, c2);
			mpz_divexact(aux2, aux2, this->iRunningPrime);
			b->Set(0, aux2);  // first
			mpz_neg(aux1, aux1);
			b->Set(1, aux1);  // second
			b->Set(2, 1);	  // third
		}

		GaussReduction(a, b);	

#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
		Polynomial::EuclideNormSqr(aux1, a);
		gmp_printf("\t\tNorm a after reduction is %Zd \n", aux1);

		Polynomial::EuclideNormSqr(aux1, b);
		gmp_printf("\t\tNorm b after reduction is %Zd \n", aux1);
#endif

		// compute common root this->iRoot = c1/p
		mpz_invert(aux1, this->iRunningPrime, this->iN);
		mpz_mul(auxM, aux1, c1);
		mpz_mod(auxM, auxM, this->iN);
		
		if (aAdvanced)
		{
			mpz_set(aux1, *a->Get2(2));
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			if (mpz_divisible_p(aux1, squareM) == 0)
			{
				printf("Error a2 is not divisible by this->iSievingIntervalLength^2 \n");
			}
			else
			{
#endif
				mpz_divexact(aux1, aux1, squareM);
				a->Set(2, aux1);
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			}
#endif
			
			mpz_set(aux1, *a->Get2(1));
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			if (mpz_divisible_p(aux1, this->iSievingIntervalLength) == 0)
			{
				printf("Error a1 is not divisible by this->iSievingIntervalLength \n");
			}
			else
			{
#endif
				mpz_divexact(aux1, aux1, this->iSievingIntervalLength);
				a->Set(1, aux1);
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			}	
#endif

			mpz_set(aux1, *b->Get2(2));
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			if (mpz_divisible_p(aux1, squareM) == 0)
			{
				printf("Error b2 is not divisible by this->iSievingIntervalLength^2 \n");
			}
			else
			{
#endif
				mpz_divexact(aux1, aux1, squareM);
				b->Set(2, aux1);
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			}
#endif
			
			mpz_set(aux1, *b->Get2(1));
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			if (mpz_divisible_p(aux1, this->iSievingIntervalLength) == 0)
			{
				printf("Error b1 is not divisible by this->iSievingIntervalLength \n");
			}
			else
			{
#endif
				mpz_divexact(aux1, aux1, this->iSievingIntervalLength);
				b->Set(1, aux1);
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
			}	
#endif
			
		}
				
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
		printf("\t\tcontrol: \n");
		
		mpz_set(aux1, *a->Get2(0));
		mpz_mul(aux1, aux1, auxM);
		mpz_mul(aux1, aux1, auxM);
		mpz_set(aux2, *a->Get2(1));
		mpz_mul(aux2, aux2, auxM);
		mpz_add(aux1, aux1, aux2);
		mpz_set(aux2, *a->Get2(2));
		mpz_add(aux1, aux1, aux2);
		mpz_mod(aux1, aux1, this->iN);
		if (mpz_cmp_ui(aux1, 0) == 0)
		{
			printf("\t\t\ta is OK \n");
		}
		else
		{
			gmp_printf("\t\t\ta is WRONG, result is %Zd \n", aux1);
		}
		
		mpz_set(aux1, *b->Get2(0));
		mpz_mul(aux1, aux1, auxM);
		mpz_mul(aux1, aux1, auxM);
		mpz_set(aux2, *b->Get2(1));
		mpz_mul(aux2, aux2, auxM);
		mpz_add(aux1, aux1, aux2);
		mpz_set(aux2, *b->Get2(2));
		mpz_add(aux1, aux1, aux2);
		mpz_mod(aux1, aux1, this->iN);
		if (mpz_cmp_ui(aux1, 0) == 0)
		{
			printf("\t\t\tb is OK \n");
		}
		else
		{
			gmp_printf("\t\t\tb is WRONG, result is %Zd \n", aux1);
		}
#endif
		if (aEnforceRealRoots)
		{
			// has a real roots?
			mpz_set(aux1, *a->Get2(2));
			mpz_set(aux2, *a->Get2(0));
			mpz_mul(aux1, aux2, aux1);
			mpz_mul_ui(aux1, aux1, 4);
			mpz_set(aux2, *a->Get2(1));
			mpz_pow_ui(aux2, aux2, 2);
			if (mpz_cmp(aux2,aux1) <= 0)
			{
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
				//printf("a hasn't real roots \n");
#endif
				// a hasn't real roots
				continue;
			}
#ifdef DEBUG_POLY_SEL_PHASE_WRITE_ROOTS
			else
			{
				mpz_sub(aux2, aux2, aux1);
				mpz_sqrt(aux2, aux2);
				mpz_set(s, *a->Get2(1));
				mpz_neg(s, s);
				mpz_add(c1, s, aux2);
				mpz_set(aux1, *a->Get2(0));
				mpz_mul_ui(aux1, aux1, 2);
				mpz_fdiv_q(c1, c1, aux1);
				gmp_printf("\t\tRoot of a is %Zd", c1);
				mpz_sub(c1, s, aux2);
				mpz_fdiv_q(c1, c1, aux1);
				gmp_printf("    Root of a is %Zd \n", c1);
			}
#endif
			
			// has b real roots?
			mpz_set(aux1, *b->Get2(2));
			mpz_set(aux2, *b->Get2(0));
			mpz_mul(aux1, aux2, aux1);
			mpz_mul_ui(aux1, aux1, 4);
			mpz_set(aux2, *b->Get2(1));
			mpz_pow_ui(aux2, aux2, 2);
			if (mpz_cmp(aux2,aux1) <= 0)
			{
#ifdef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
				//printf("b hasn't real roots \n");
#endif
				// b hasn't real roots
				continue;
			}
#ifdef DEBUG_POLY_SEL_PHASE_WRITE_ROOTS
			else
			{
				mpz_sub(aux2, aux2, aux1);
				mpz_sqrt(aux2, aux2);
				mpz_set(s, *b->Get2(1));
				mpz_neg(s, s);
				mpz_add(c1, s, aux2);
				mpz_set(aux1, *b->Get2(0));
				mpz_mul_ui(aux1, aux1, 2);
				mpz_fdiv_q(c1, c1, aux1);
				gmp_printf("\t\tRoot of b is %Zd", c1);
				mpz_sub(c1, s, aux2);
				mpz_fdiv_q(c1, c1, aux1);
				gmp_printf("    Root of b is %Zd \n", c1);
			}
#endif
		}

#ifdef DEBUG_POLY_SEL_PHASE_INFO_STATS_ON
		this->iStats->Increase(EPSSTested);
#endif
			
		if (a->ControlCoeffSize(coeffBound) && b->ControlCoeffSize(coeffBound))
		{
#ifdef DEBUG_POLY_SEL_PHASE_TIME_MESSAGE_ON
			this->iTimes->Start(EPSTCompAlpha);
#endif

			if (aAdvanced)
			{
				AlphaEstimatePoly(alpha, roots, a);
			}
			else
			{
				AlphaEstimateHomPoly(alpha, roots, a);
			}

#ifdef DEBUG_POLY_SEL_PHASE_TIME_MESSAGE_ON
			this->iTimes->End(EPSTCompAlpha);
#endif

			if (mpf_cmp_d(alpha, this->iAlphaBound) <= 0)
			{
#ifdef DEBUG_POLY_SEL_PHASE_TIME_MESSAGE_ON
				this->iTimes->Start(EPSTCompAlpha);
#endif

				if (aAdvanced)
				{
					AlphaEstimatePoly(alpha, roots, b);
				}
				else
				{
					AlphaEstimateHomPoly(alpha, roots, b);
				}

#ifdef DEBUG_POLY_SEL_PHASE_TIME_MESSAGE_ON
				this->iTimes->End(EPSTCompAlpha);
#endif
				if (mpf_cmp_d(alpha, this->iAlphaBound) <= 0)
				{
					if (this->iSelfInitializingDelay)
					{
						this->iAlphaBound -= ConstPSPhase::ALPHA_STEP_BACK;
					}
					delay = 0;

					// save result
					a->Copy(this->iCandidatePolyPairs->FirstPoly[this->iFoundCandidatePolyCount]);
					b->Copy(this->iCandidatePolyPairs->SecondPoly[this->iFoundCandidatePolyCount]);
					mpz_set(this->iCandidatePolyPairs->Root[this->iFoundCandidatePolyCount], auxM);
					this->iFoundCandidatePolyCount++;
				}
				else
				{
#ifdef DEBUG_POLY_SEL_PHASE_INFO_STATS_ON
					this->iStats->Increase(EPSSRejectedAlpha);
#endif
#ifdef DEBUG_POLY_SEL_PHASE_WRITE_ALPHA
					gmp_printf("\t\tToo big alpha of b. Alpha = %.*Ff\n", 10, alpha);
#endif

					delay++; // change alpha bound for finding something?
					if ((this->iSelfInitializingDelay) && (delay > this->iSelfInitializingDelay))
					{
						delay = 0;
						this->iAlphaBound += ConstPSPhase::ALPHA_STEP;
					}
				}
			}
			else
			{
#ifdef DEBUG_POLY_SEL_PHASE_INFO_STATS_ON
				this->iStats->Increase(EPSSRejectedAlpha);
#endif
#ifdef DEBUG_POLY_SEL_PHASE_WRITE_ALPHA
				gmp_printf("\t\tToo big alpha of a. Alpha = %.*Ff\n", 10, alpha);
#endif

				delay++; // change alpha bound for finding something?
				if ((this->iSelfInitializingDelay) && (delay > this->iSelfInitializingDelay))
				{
					delay = 0;
					this->iAlphaBound += ConstPSPhase::ALPHA_STEP;
				}
			}
		} // end if - coeff size
#ifdef DEBUG_POLY_SEL_PHASE_INFO_STATS_ON
		else 
		{
			this->iStats->Increase(EPSSRejectedCoeff);
		}
#endif

	}

#ifndef DEBUG_POLY_SEL_PHASE_GENERATE_POLY
	if (Get_InfoMode())
	{
		PrintValues();
		printf("\n");
	}
#endif
	

	// cleanup
	// -------------------------
montgomery_generate_clear_gmp:

	delete a;
	a = NULL;
	delete b;
	b = NULL;
	
	mpz_clear(squareRootOfN);
	mpz_clear(c1);
	mpz_clear(c2);
	mpz_clear(s);
	mpz_clear(aux1);
	mpz_clear(aux2);
	mpz_clear(auxM);
	mpz_clear(squareM);
	mpz_clear(coeffBound);
	mpz_clear(zero);
	//mpz_clear(sizeBound);
	mpf_clear(alpha);
	// -------------------------
	
	return ConstRC::Ok;
}
void CMontgomeryPolySelPhase::PrintHeader()
{
	printf("\n\n");
	printf(" __________________________________________________________________________");
	printf("\n");
	printf("|  round  |              prime             |   alpha |  found  |  percent  |");
	printf("\n");
}
void CMontgomeryPolySelPhase::PrintValues()
{
	printf("\r|");
	printf(" %7ld |", this->iRound);
	gmp_printf(" %30Zd |", this->iRunningPrime);	
	printf(" %7.3f |", this->iAlphaBound);				
	printf(" %7ld |", this->iFoundCandidatePolyCount);
	printf(" %7.3f %% |", 100 * (float) this->iFoundCandidatePolyCount / (float) this->iCandidatesCount);
	fflush(stdout);	
}
void CMontgomeryPolySelPhase::PrintEffectiveness()
{

}
int CMontgomeryPolySelPhase::PrepareRunningSerialization()
{
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();

	this->iRSerialFoundCandPolyCount = this->iFoundCandidatePolyCount;
	this->iRSerialRatedCandPolyCount = this->iRatedCandidatePolyCount;
	this->iRSerialRound = this->iRound;

	mpz_set(this->iRSerialRunningPrime, this->iRunningPrime);

	return ConstRC::Ok;
}
/*!
	TODO: If call from more threads then could be a problem with writing candidates - same files...
*/
int CMontgomeryPolySelPhase::Serialize(const CBaseParameters &aParam) const
{
	int rc = MSerializable::Serialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters paramb;
	if (Running())
	{
		AFactorAlgParameters * param = Get_Parameters();

		// for deserialization at new start
		param->SetPhaseParameter(ConstPN::DESERIALIZE, Utils::ConvertToString(true), Get_PhaseNbr());
		param->SetPhaseParameter(ConstPN::PHASE_FULL_FILENAME, 
			MSerializable::CreateFileNameAuto(aParam, ConstFN::POLY_PHASE), Get_PhaseNbr());

		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_ParametersFullFileName()); 
		rc = param->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	if (this->iCandidatePolyPairs != NULL)
	{
		// we want serialize only assigned polynomials
		string auxStr = Utils::ConvertToString(this->iRSerialFoundCandPolyCount);
		paramb.SetParameter(ConstPN::MAX_ELEMENT_COUNT, auxStr);
		// we want serialize rating only for rated polynomials
		auxStr = Utils::ConvertToString(this->iRSerialRatedCandPolyCount);
		paramb.SetParameter(ConstPN::MAX_RATED_ELEMENT_COUNT, auxStr);
		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iCandPolyPairsFullFileName);
		rc = this->iCandidatePolyPairs->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok; 
}
int CMontgomeryPolySelPhase::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_MONTGOMERY_POLY_SEL_PHASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","Serialize",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_INNER_STATE_NAME, Get_InnerState());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_FOUND_CAND_POLY_NAME, this->iRSerialFoundCandPolyCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_RATED_CAND_POLY_NAME, this->iRSerialRatedCandPolyCount);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_ROUND_NAME, this->iRSerialRound);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_PRIME_NAME, this->iRSerialRunningPrime);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters param;
	rc = this->iStats->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = this->iTimes->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","Serialize",rc,0,"Could not end the parameter tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok; 
}
/*!
	Shouldn't be call from outside!!!
*/
int CMontgomeryPolySelPhase::Deserialize(const CBaseParameters &aParam)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != EPSPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	if (Get_InfoMode())
	{
		Utils::NormalMessage(" Deserializing Montgomery Polynomial Selection Phase...");
	}

	int rc = MSerializable::Deserialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (MSerializable::ExistFile(this->iCandPolyPairsFullFileName))
	{
		if (Get_InfoMode())
		{
			Utils::NormalMessage("\tDeserializing found polynomial pairs...");
		}

		CBaseParameters param;
		param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iCandPolyPairsFullFileName);
		this->iCandidatePolyPairs = new CCandidatePairs(this->iCandidatesCount);
		rc = this->iCandidatePolyPairs->Deserialize(param);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
/*!
	Shouldn't be call from outside!!!
*/
int CMontgomeryPolySelPhase::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != EPSPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_MONTGOMERY_POLY_SEL_PHASE_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CMontgomeryPolySelPhase","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int auxInt = 0;
	int depth = xmlTextReaderDepth(aReader);
	CBaseParameters param;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INNER_STATE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize inner state...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_InnerState(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIME_NAME))
		{
			rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize running prime...", this->iRunningPrime, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			mpz_set(this->iRSerialRunningPrime, this->iRunningPrime);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ROUND_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize round...", this->iRound, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iRSerialRound = this->iRound;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_FOUND_CAND_POLY_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize found candidate poly pairs count...", this->iFoundCandidatePolyCount, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iRSerialFoundCandPolyCount = this->iFoundCandidatePolyCount;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_RATED_CAND_POLY_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize rated candidate poly pairs count...", this->iRatedCandidatePolyCount, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iRSerialRatedCandPolyCount = this->iRatedCandidatePolyCount;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_STATISTIC_CONTAINER_NAME))
		{
			rc = this->iStats->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_TIME_CONTAINER_NAME))
		{
			rc = this->iTimes->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
