#include "multivariable_polynomial.h"
#include "const_return_codes.h"

/*!
	Constructs a multivariable polynomial that can hold at most aAllocation terms. All this terms can be
	x-polynomials with degree at most aPolynomialDegree. If aPolynomialDegree is at least 1, the polynomial
	has integer coefficients, otherwise it has real coefficients.
*/


MultVarPolynomial::MultVarPolynomial(unsigned aAllocation, unsigned aPolynomialDegree){
	unsigned i;
	degree = aPolynomialDegree;
	length = 0;
	allocated = aAllocation;
	Term = NULL;
	Term = (MultVarTerm*) malloc(aAllocation*sizeof(MultVarTerm));
	for(i=0;i<allocated;i++){
	if (degree) {
		Term[i].XPolynomial = new Polynomial(degree);
	} else {
		mpf_init2(Term[i].Coeff, ConstPSPhase::MPF_T_HIGH_PRECISION);
	}
	}

	mpf_init2(this->iAuxMPF, ConstPSPhase::MPF_T_HIGH_PRECISION);
}

MultVarPolynomial::~MultVarPolynomial(){
	unsigned i;
	for(i=0;i<allocated;i++) 
	{
	if (degree)
		delete Term[i].XPolynomial;
	else mpf_clear(Term[i].Coeff);
	}
	free(Term);

	mpf_clear(this->iAuxMPF);
}

/*!
	Returns true if all exponents in aVariables1 and aVariables2 are same.
*/
	
bool MultVarPolynomial::EqualPowers(signed int *aVariables1,signed int *aVariables2){
	unsigned int i;
	for(i=0;i<MULT_VAR_VARIABLES;i++)
	if (aVariables1[i]!=aVariables2[i]) return false;
	return true;
}

/*!
	Prints the multivariable polynomial to the screen.
*/

int MultVarPolynomial::PrintToScreen(){
	unsigned i,k;
	for(i=0;i<length;i++){
		if(degree){
			if(i) printf("+");
			printf("(");
			Term[i].XPolynomial->PrintToScreen();
			printf(")");
		}
		else
			gmp_printf("%+Ff",Term[i].Coeff);
		for(k=0;k<MULT_VAR_VARIABLES;k++)
			if (Term[i].Variables[k])
				printf("(v%d)^%d",k,Term[i].Variables[k]);
	}
	puts("");
	return  ConstRC::Ok;
}

/*!
	Makes a copy of the polynomial into aResult.
*/

int MultVarPolynomial::Copy(MultVarPolynomial *aResult){
	if (aResult->allocated<this->length) return  ConstRC::SizeMismatch;
	unsigned int j,k;
	for(j=0;j<length;j++){
	if (degree)
		this->Term[j].XPolynomial->Copy(aResult->Term[j].XPolynomial);
	else
		mpf_set(aResult->Term[j].Coeff,this->Term[j].Coeff);
	for(k=0;k<MULT_VAR_VARIABLES;k++)
		aResult->Term[j].Variables[k]=this->Term[j].Variables[k];
	}
	aResult->length=this->length;
	return ConstRC::Ok;
}

/*!
	Computes a sum this + aOperand = aResult.
*/

int MultVarPolynomial::Add(MultVarPolynomial *aResult, MultVarPolynomial *aOperand){
	unsigned i,j,k;
	if (this != aResult) {
	if (this->Copy(aResult)!=ConstRC::Ok) return ConstRC::SizeMismatch;
	}
	for(i=0;i<aOperand->length;i++){
	for(j=0;(j<aResult->length)&&(!EqualPowers(aResult->Term[j].Variables,aOperand->Term[i].Variables));j++);
	if (j>=aResult->allocated) return ConstRC::SizeMismatch;
	if (j>=aResult->length){
		if (degree)
		aResult->Term[j].XPolynomial->SetDegree(-1);
		else
		mpf_set_ui(aResult->Term[j].Coeff,0);
		for(k=0;k<MULT_VAR_VARIABLES;k++)
		aResult->Term[j].Variables[k]=aOperand->Term[i].Variables[k];
		aResult->length++;
	}
	if (degree)
		aResult->Term[j].XPolynomial->Add(aResult->Term[j].XPolynomial,aResult->Term[j].XPolynomial,aOperand->Term[i].XPolynomial);
	else
		mpf_add(aResult->Term[j].Coeff,aResult->Term[j].Coeff,aOperand->Term[i].Coeff);
	}
	return ConstRC::Ok;
}

/*!
	Computes a product this * aOperand = aResult.
*/

int MultVarPolynomial::Multiply(MultVarPolynomial *aResult, MultVarPolynomial *aOperand){
	if ((this==aResult)||(aResult==aOperand)) return ConstRC::GeneralError;
	unsigned i,j,k,l;
	Polynomial *auxP = NULL;
	if (degree)
	{
		auxP = new Polynomial(degree);
	}
	int v[MULT_VAR_VARIABLES];
	aResult->length=0;
	for(i=0;i<this->length;i++)
	for(j=0;j<aOperand->length;j++){
		for(k=0;k<MULT_VAR_VARIABLES;k++)
		v[k]=this->Term[i].Variables[k]+aOperand->Term[j].Variables[k];
		for(l=0;(l<aResult->length)&&(!EqualPowers(aResult->Term[l].Variables,v));l++);
		if (l>=aResult->allocated) return ConstRC::SizeMismatch;
		if (l>=aResult->length){
		if (degree)
			aResult->Term[l].XPolynomial->SetDegree(-1);
		else
			mpf_set_ui(aResult->Term[l].Coeff,0);
		for(k=0;k<MULT_VAR_VARIABLES;k++)
			aResult->Term[l].Variables[k]=v[k];
		aResult->length++;
		}
		if (degree)
		{
			auxP->Multiply(auxP,this->Term[i].XPolynomial,aOperand->Term[j].XPolynomial);
			aResult->Term[l].XPolynomial->Add(aResult->Term[l].XPolynomial,aResult->Term[l].XPolynomial,auxP);
		}
		else
		{
			mpf_mul(this->iAuxMPF,this->Term[i].Coeff,aOperand->Term[j].Coeff);
			mpf_add(aResult->Term[l].Coeff,aResult->Term[l].Coeff,this->iAuxMPF);
		}
	}
	if (degree)
	{
		delete auxP;
	}

	return ConstRC::Ok;
}

/*!
	Evaluates the polynomial in the given values. The s variable is given in SPowerTable by its powers from
	-aDegree to aDegree, the other variables are given in PowerTable by its powers from 0 to aDegree.
	The polynomial must have real coefficients otherwise the program can crash.
*/

int MultVarPolynomial::Evaluate(mpf_t aResult, mpf_t **PowerTable, mpf_t *SPowerTable, unsigned aDegree){
	unsigned i,k;

	mpf_set_ui(aResult,0);
	for(i=0;i<length;i++){
	mpf_set(this->iAuxMPF,Term[i].Coeff);
	for(k=0;k<MULT_VAR_VARIABLES-1;k++)
		mpf_mul(this->iAuxMPF,this->iAuxMPF,PowerTable[k][Term[i].Variables[k]]);
	mpf_mul(this->iAuxMPF,this->iAuxMPF,SPowerTable[Term[i].Variables[S_LOC]+aDegree+1]);
	mpf_add(aResult,aResult,this->iAuxMPF);
	}

	return ConstRC::Ok;
}

/*!
	Forms a function in variables c<sub>0</sub>, c<sub>1</sub> and c<sub>2</sub> and inserts its derivatives
	into a matrix so that the minimum can be computed by the Gauss elimination.
	The polynomial must have real coefficients otherwise the program can crash.
*/

int MultVarPolynomial::ThreeVariableFunction(mpf_t aResult[T_LOC][T_LOC + 1], mpf_t **PowerTable, mpf_t *SPowerTable, unsigned aDegree){
	unsigned i,k,a,b;

	for(a=0;a<T_LOC;a++) {
		for(b=0;b<=T_LOC;b++)
			mpf_set_ui(aResult[a][b],0);
	}
	for(i=0;i<length;i++){
		for(a=0;(a<T_LOC)&&(!Term[i].Variables[a]);a++);
		if (a<T_LOC){
			if (Term[i].Variables[a]==2)
				b=a;
			else
				for(b=a+1;(b<T_LOC)&&(!Term[i].Variables[b]);b++);
			mpf_set(this->iAuxMPF,Term[i].Coeff);
			for(k=T_LOC;k<MULT_VAR_VARIABLES-1;k++){
				mpf_mul(this->iAuxMPF,this->iAuxMPF,PowerTable[k][Term[i].Variables[k]]);
			}
			mpf_mul(this->iAuxMPF,this->iAuxMPF,SPowerTable[Term[i].Variables[S_LOC]+aDegree+1]);
			mpf_add(aResult[a][b],aResult[a][b],this->iAuxMPF);
		}
	}
	for(a=0;a<T_LOC;a++){
		mpf_mul_2exp(aResult[a][a],aResult[a][a],1);
		mpf_neg(aResult[a][T_LOC],aResult[a][T_LOC]);
		for(b=0;b<a;b++){
			mpf_set(aResult[a][b],aResult[b][a]);
		}
	}

	return ConstRC::Ok;
}

/*!
	Returns the partial derivative with respect to the variable aVariable.
	The polynomial must have real coefficients otherwise the program can crash.
*/

int MultVarPolynomial::PartialDerivative(MultVarPolynomial *aResult, unsigned aVariable){
	unsigned i,k,l=0;
	for(i=0;i<length;i++) if (Term[i].Variables[aVariable]) {
	for(k=0;k<MULT_VAR_VARIABLES;k++)
		aResult->Term[l].Variables[k] = Term[i].Variables[k];
	aResult->Term[l].Variables[aVariable]--;
	mpf_mul_ui(aResult->Term[l].Coeff,Term[i].Coeff,abs(Term[i].Variables[aVariable]));
	if (Term[i].Variables[aVariable]<0) mpf_neg(aResult->Term[l].Coeff,aResult->Term[l].Coeff);
	l++;
	}
	aResult->length=l;
	return ConstRC::Ok;
}

/*
	Reset all values to 0.
*/
int MultVarPolynomial::Reset()
{
	int i, j;
	for(i = 0; i < this->allocated; i++)
	{
		if (degree) 
		{
			this->Term[i].XPolynomial->Zeroize();
			for (j = 0; j < MULT_VAR_VARIABLES; j++)
			{
				this->Term[i].Variables[j] = 0;
			}
		} 
		else 
		{
			mpf_set_d(this->Term[i].Coeff, 0);
			for (j = 0; j < MULT_VAR_VARIABLES; j++)
			{
				this->Term[i].Variables[j] = 0;
			}
		}
	}

	return ConstRC::Ok;
}
