#include "nfs_classical_sieving_phase.h"

#include "const_parameter_names.h"
#include "const_sieving_phase.h"
#include "const_element_flags.h"
#include "adv_factor_base_helper.h"
#include "abstract_rel_processing_phase.h"
#include "supporting_factor_alg_creator.h"
#include "integer_class.h"
#include <cstring>
#include <math.h>
#include "log.h"
#include <iostream>
#include <iomanip>


#include "test_suite.h"

using namespace std;

typedef map<main_sieving_type, int>::const_iterator SpecialPrimeIt;

CNFSClassicalSievingPhase::CNFSClassicalSievingPhase() : MSerializable(ConstFN::SIEVING_PHASE), ASievingPhase()
{
	Set_AutoSerializable(ConstSPhase::AUTO_SERIALIZABLE);
	Set_InnerState((int)ESPhaseNotStarted);

	this->iTimes = new CTimeContainer(ConstSPhase::INFO_TIME_COUNT);
	this->iStats = new CStatContainer(ConstSPhase::INFO_STAT_COUNT);

	this->iRelationBuffer = NULL;
	this->iRelationBufferMaxCount = 0;
	this->iRelationWorkingBuffer = NULL;
	this->iRelationWorkingBufferMaxCount = 0;

	this->iRelation = NULL;

	this->iLPHashtable = NULL;

	mpz_init(this->iAuxMpz1);
	mpz_init(this->iAuxMpz2);

	this->iIteration = 0;
	this->iWorkingIteration = 0;

	this->iRSerialAccordingLines = false;
	this->iRSerialAccordingTime = false;
	this->iRSerialAccordingRelations = false;

	this->iNfsCompressEvenIntervals = false;
	this->iNfsOddToEven = false;
	this->iNfsEvenToOdd = false;
	this->iIsCompressedNow = false;

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iSievingBlocks[i] = NULL;
		this->iInitBlockValueChange[i] = 0;
		this->iInitBlockValues[i] = 0;
		this->iRSerialInitBlockValueChange[i] = 0;
		this->iInitBlockValueReference[i] = 0;
	}

	int rc;
	if ((rc = pthread_mutex_init(&(this->iBMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iAMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iIterationMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iMaxIterationMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iTimeMessageMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iWorkingIterationMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iRelationBufferMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CNFSClassicalSievingPhase",rc,0,"Cannot initialize mutex.");
	}
}
CNFSClassicalSievingPhase::~CNFSClassicalSievingPhase()
{
	Reset();
	if (this->iTimes != NULL)
	{
		delete this->iTimes;
		this->iTimes = NULL;
	}
	if (this->iStats != NULL)
	{
		delete this->iStats;
		this->iStats = NULL;
	}
	
	if (this->iRelationBuffer != NULL)
	{
		int max = this->iRelationBuffer->size();
		for (int i = 0; i < max; i++)
		{
			if (this->iRelationBuffer->at(i) != NULL)
			{
				delete this->iRelationBuffer->at(i);
				this->iRelationBuffer->at(i) = NULL;
			}
		}
		this->iRelationBufferMaxCount = 0;
		this->iRelationBuffer->clear();

		delete this->iRelationBuffer;
		this->iRelationBuffer = NULL;
	}

	if (this->iRelationWorkingBuffer != NULL)
	{
		int max = this->iRelationWorkingBuffer->size();
		for (int i = 0; i < max; i++)
		{
			if (this->iRelationWorkingBuffer->at(i) != NULL)
			{
				delete this->iRelationWorkingBuffer->at(i);
				this->iRelationWorkingBuffer->at(i) = NULL;
			}
		}
		this->iRelationWorkingBufferMaxCount = 0;
		this->iRelationWorkingBuffer->clear();

		delete this->iRelationWorkingBuffer;
		this->iRelationWorkingBuffer = NULL;
	}
	DisposeFunctorField();
	DisposeGMP();
	DisposeMutexes();
	DisposeAcceptedReturnCodes();
}
int CNFSClassicalSievingPhase::Reset()
{
	if (this->iTimes != NULL)
	{
		this->iTimes->ResetAll();
	}
	if (this->iStats != NULL)
	{
		this->iStats->ResetAll();
	}
	if (this->iRelation != NULL)
	{
		delete this->iRelation;
		this->iRelation = NULL;
	}

	// we use just one instance of the same factoring algorithm - TODO better
	if (this->iVariationsInfo[0].FactorAlg == this->iVariationsInfo[1].FactorAlg)
	{
		this->iVariationsInfo[1].FactorAlg = NULL;
	}

	int i;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iOptimizer[i].Reset();
		this->iInitBlockValueReference[i] = 0;
		this->iRSerialInitBlockValueChange[i] = 0;
		this->iInitBlockValueChange[i] = 0;
		this->iInitBlockValues[i] = 0;
		
		this->iAdvFB[i].Reset();
		
		this->iFBs[i].Reset();
		
		this->iNumberFields[i].Reset();

		this->iHashtables[i].DeleteHashtables();

		if (this->iSievingBlocks[i] != NULL)
		{
			free(this->iSievingBlocks[i]);
			this->iSievingBlocks[i] = NULL;
		}

		this->iVariationsInfo[i].Reset();
	}

	Set_Iteration(0);
	Set_MaxIteration(0);
	Set_WorkingIteration(0);
	Set_TimeMessage(0);

	this->iRSerialAccordingLines = false;
	this->iRSerialAccordingTime = false;
	this->iRSerialAccordingRelations = false;

	this->iNfsCompressEvenIntervals = false;
	this->iNfsOddToEven = false;
	this->iNfsEvenToOdd = false;
	this->iIsCompressedNow = false;

	this->iIntermediateRoots.clear();

	if (this->iLPHashtable != NULL)
	{
		delete this->iLPHashtable;
		this->iLPHashtable = NULL;
	}

	int max;
	if (this->iRelationBuffer != NULL)
	{
		max = this->iRelationBuffer->size();
		for (i = 0; i < max; i++)
		{
			if (this->iRelationBuffer->at(i) != NULL)
			{
				this->iRelationBuffer->at(i)->Reset();
			}
		}
		this->iRelationBufferMaxCount = 0;
	}

	if (this->iRelationWorkingBuffer != NULL)
	{
		max = this->iRelationWorkingBuffer->size();
		for (i = 0; i < max; i++)
		{
			if (this->iRelationWorkingBuffer->at(i) != NULL)
			{
				this->iRelationWorkingBuffer->at(i)->Reset();
			}
		}
		this->iRelationWorkingBufferMaxCount = 0;
	}

	ASievingPhase::Reset();

	Set_InnerState((int)ESPhaseNotStarted);
	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::DisposeGMP()
{
	mpz_clear(this->iAuxMpz1);
	mpz_clear(this->iAuxMpz2);

	return ASievingPhase::DisposeGMP();
}
int CNFSClassicalSievingPhase::DisposeMutexes()
{
	int rc;

	if ((rc = pthread_mutex_destroy(&(this->iBMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iAMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iIterationMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iMaxIterationMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iTimeMessageMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iWorkingIterationMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iRelationBufferMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}

	return ASievingPhase::DisposeMutexes();
}
int CNFSClassicalSievingPhase::FillFunctorField()
{
	if (Get_FunctorsInit() == false)
	{
		Set_FunctorsInit(true);
		this->iPhaseFunctors.reserve(ConstSPhase::INNER_STATE_COUNT);
		this->iPhaseFunctors.clear();
		this->iPhaseFunctors.push_back(NULL); // ESPhaseNotStarted
		this->iPhaseFunctors.push_back(NULL); // ESPhaseInitialization
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CNFSClassicalSievingPhase>(this, &CNFSClassicalSievingPhase::InitFactorBases)); // ESPhaseInitFB
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CNFSClassicalSievingPhase>(this, &CNFSClassicalSievingPhase::InitAdvancedFactorBases)); // ESPhaseInitAdvFB
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CNFSClassicalSievingPhase>(this, &CNFSClassicalSievingPhase::RunSieve)); // ESPhaseSieving
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CNFSClassicalSievingPhase>(this, &CNFSClassicalSievingPhase::SaveResult)); // ESPhaseSavingResult
		this->iPhaseFunctors.push_back(NULL); // ESPhaseFinished
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	string identifier;
	string directory = "";
	string file;
	string auxStr;

	int phaseNBR;
	if (aParameters->GetIntegerPhaseParameter(ConstPN::SIEVING_PHASE_NUMBER, phaseNBR, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CleanUp",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}

	if (aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","CleanUp",0,0,"Didn't find serial identifier.");
		return ConstRC::GeneralError;
	}
	aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, directory, (int)EMain);

	switch (aType)
	{
		case ECleanAllIfNotError:
			if (aError)
			{
				break;
			}
		case ECleanAll:
		case ECleanAllExcCF:
			// Smooth relation full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::SMOOTH_RELATIONS_FILENAME, 
					ConstPN::SMOOTH_RELATIONS_DIRECTORY, ConstPN::SMOOTH_RELATIONS_COMPRESSION, directory,
					ConstFN::SMOOTH_RELATIONS, identifier, phaseNBR, TXT_FILE_EXTENSION);
			}
			remove(file.c_str());

			// Partial relation full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::PARTIAL_REL_FILENAME, 
					ConstPN::PARTIAL_REL_DIRECTORY, ConstPN::PARTIAL_REL_COMPRESSION, directory,
					ConstFN::PARTIAL_REL, identifier, phaseNBR, TXT_FILE_EXTENSION);
			}
			remove(file.c_str());

		case ECleanAllExcCFRF:
			// FB full filenames
			for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				auxStr = Utils::ConvertToString(i + 1);
				if (aParameters->ContainsPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, Get_PhaseNbr()))
				{
					aParameters->GetStringPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, file, Get_PhaseNbr());
					MSerializable::AdaptPathSeparator(file);
				}
				else
				{
					aParameters->CreateFullFileName(file, ConstPN::FACTOR_BASE_FILENAME + auxStr, 
						ConstPN::FACTOR_BASE_DIRECTORY + auxStr, ConstPN::FACTOR_BASE_COMPRESSION + auxStr, Get_SerialMainDirectory(),
						ConstFN::FACTOR_BASE + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
				}
				remove(file.c_str());
			}
			
			// Running Serialization full filename - default is same as iSerialPhaseFullFileName
			if (aParameters->ContainsPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::RSERIAL_SIEVING_FILENAME, 
					ConstPN::RSERIAL_SIEVING_DIRECTORY, ConstPN::RSERIAL_SIEVING_COMPRESSION, directory,
					ConstFN::SIEVING_PHASE, identifier, phaseNBR);
			}
			remove(file.c_str());

			break;
		default:
			break;
	}

	return ASievingPhase::CleanUp(aParameters, phaseNBR, aType, aError);
}
int CNFSClassicalSievingPhase::InitParameters()
{
	Set_InnerState((int)ESPhaseInitialization);

	AFactorAlgParameters * param = Get_Parameters();

	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	// generally we don't know order of the sieving phase
	int auxInt;
	if (param->GetIntegerPhaseParameter(ConstPN::SIEVING_PHASE_NUMBER, auxInt, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","InitParameters",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}
	Set_PhaseNbr(auxInt);

	// now we initialize members from parents's classes
	int rc = ASievingPhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int i;
	int varType = (int)ConstSPhase::VARIATIONS_TYPE;
	int algType = (int)ConstSPhase::SUPPORTING_FACTOR_ALG;
	ASupportingFactorAlg * alg = NULL;
	bool deleteAlg = true;
	string auxStr;
	long bound = 0;
	double ratio = 1.0;
	int policy = ConstSPhase::THRESHOLD_OPTIMIZER_POLICY;
	Polynomial * poly = NULL; // do not delete - it is saved in NumberField
	
	// we try to find value for all 
	param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND, bound, bound, (int)EMain);
	param->GetIntegerPhaseParameter(ConstPN::THRESHOLD_OPT_POLICY, policy, policy, Get_PhaseNbr());
	param->GetDoublePhaseParameter(ConstPN::THRESHOLD_OPT_RATIO, ratio, ratio, Get_PhaseNbr());
	param->GetIntegerPhaseParameter(ConstPN::VARIATIONS_TYPE, varType, varType, (int)EMain);
	param->GetIntegerPhaseParameter(ConstPN::SUPPORTING_FACTOR_ALG, algType, algType, Get_PhaseNbr());
	rc = CFactorAlgCreator::CreateFactorAlg((supporting_factor_alg_types)algType, alg);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = Utils::ConvertToString(i + 1);
		rc = param->GetPolynomialPhaseParameter(ConstPN::SIEVING_POLY + auxStr, poly, (int)EMain);
		if (rc != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","InitParameters",0,0,"Didn't find the sieving polynomial.");
			return rc;
		}
		this->iNumberFields[i].SetSievingPoly(poly);
		poly = NULL;

		// we try to find individual value
		param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND + auxStr, bound, bound, (int)EMain);
		this->iFBs[i].UpperFBBound = bound;

		param->GetDoublePhaseParameter(ConstPN::THRESHOLD_OPT_RATIO + auxStr, ratio, ratio, Get_PhaseNbr());
		this->iOptimizer[i].Set_DesiredRatio(ratio);

		param->GetIntegerPhaseParameter(ConstPN::THRESHOLD_OPT_POLICY + auxStr, policy, policy, Get_PhaseNbr());
		this->iOptimizer[i].Set_OptimizationPolicy(policy);

		param->GetIntegerPhaseParameter(ConstPN::PARTIAL_FACTOR_MULTIPLIER + auxStr, ConstSPhase::PARTIAL_FACTOR_MULTIPLIER, auxInt, (int)EMain);
		this->iPartialFactorMultiplier[i] = auxInt;

		param->GetIntegerPhaseParameter(ConstPN::VARIATIONS_TYPE + auxStr, varType, auxInt, (int)EMain);
		this->iVariationsInfo[i].Type = (variations_types)auxInt;

		if (this->iVariationsInfo[i].Type > ELargePrimeV)
		{
			// we want only one instance of the same factorization algorithm
			param->GetIntegerPhaseParameter(ConstPN::SUPPORTING_FACTOR_ALG + auxStr, algType, auxInt, Get_PhaseNbr());
			if (auxInt != algType)
			{
				rc = CFactorAlgCreator::CreateFactorAlg((supporting_factor_alg_types)auxInt, this->iVariationsInfo[i].FactorAlg);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			else
			{
				this->iVariationsInfo[i].FactorAlg = alg;
				deleteAlg = false;
			}	
		}
	}

	if (deleteAlg)
	{
		delete alg;
		alg = NULL;
	}
	else
	{
		// factorization algorithm will be deleted in Reset method
		alg = NULL;
	}

	if (bound == 0)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","InitParameters",0,0,"Didn't find the factor base upper bound.");
		return ConstRC::GeneralError;
	}

	// sieving region
	rc = param->GetLongPhaseParameter(ConstPN::SIEVING_REGION_WIDTH, this->iSievingRegion.width, (int)EMain);
	if (rc != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","InitParameters",0,0,"Didn't find the sieving region width.");
		return rc;
	}
	rc = param->GetLongPhaseParameter(ConstPN::SIEVING_REGION_HEIGHT, this->iSievingRegion.height, (int)EMain);
	if (rc != ConstRC::Ok) {
		iSievingRegion.height = SIEVING_REGION_NO_HEIGHT_BOUND;
		CLog::Log(50, "", "Sieving region height was not given. We assume there is no bound for the height.");
	}
	rc = param->GetLongPhaseParameter(ConstPN::SIEVING_REGION_START, this->iSievingRegion.start, (int)EMain);
	if (rc != ConstRC::Ok) {
		iSievingRegion.start = SIEVING_REGION_DEFAULT_START;
		CLog::Log(50, "", "Sieving region start was not given. We set the start to the default value %d.", SIEVING_REGION_DEFAULT_START);
	}

	param->GetIntegerPhaseParameter(ConstPN::THRESHOLD_OPT_INT, (int)ConstSPhase::THRESHOLD_OPTIMIZER_INT, this->iThresholdOptInterval, Get_PhaseNbr());

	param->GetIntegerPhaseParameter(ConstPN::MAX_ITERATIONS, (int)ConstSPhase::MAX_ITERATIONS, this->iMaxIteration, Get_PhaseNbr());

	param->GetIntegerPhaseParameter(ConstPN::SIEVING_DIRECTION, (int)ConstSPhase::SIEVING_DIRECTION, auxInt, Get_PhaseNbr());
	this->iSievingDirection = (nfs_sieving_direction)auxInt;

	if (param->ContainsPhaseParameter(ConstPN::SIEVING_NUMBER_A, Get_PhaseNbr()) || 
		param->ContainsPhaseParameter(ConstPN::SIEVING_NUMBER_B, Get_PhaseNbr()))
	{
		param->GetIntegerPhaseParameter(ConstPN::SIEVING_NUMBER_A, 0, this->iA, Get_PhaseNbr());
		param->GetIntegerPhaseParameter(ConstPN::SIEVING_NUMBER_B, 0, this->iB, Get_PhaseNbr());
	}
	else
	{
		switch (this->iSievingDirection)
		{
				case EIncrementA:
					this->iA = iSievingRegion.start;
					this->iB = 0;
					break;
				case EDecrementA:
					this->iA = -iSievingRegion.start;
					this->iB = 0;
					break;
				case EIncrementB:
					this->iB = iSievingRegion.start;
					this->iA = 0;
					break;
				case EDecrementB:
					this->iB = -iSievingRegion.start;
					this->iA = 0;
					break;		
		}
	}

	param->GetIntegerPhaseParameter(ConstPN::TIME_MESSAGE_INTERVAL, ConstSPhase::TIME_MESSAGE_INTERVAL, auxInt, Get_PhaseNbr());
	Set_TimeMessage(auxInt);

	// Running serialization setup

	if (param->ContainsPhaseParameter(ConstPN::RSERIALIZATION_ITER_INTERVAL, Get_PhaseNbr()))
	{
		this->iRSerialAccordingLines = true;
		param->GetIntegerPhaseParameter(ConstPN::RSERIALIZATION_ITER_INTERVAL, this->iRSerialLineInterval, Get_PhaseNbr());
	}
	if (param->ContainsPhaseParameter(ConstPN::RSERIALIZATION_TIME_INTERVAL, Get_PhaseNbr()))
	{
		this->iRSerialAccordingTime = true;
		param->GetLongPhaseParameter(ConstPN::RSERIALIZATION_TIME_INTERVAL, this->iRSerialTimeInterval, Get_PhaseNbr());
	}
	if (param->ContainsPhaseParameter(ConstPN::RSERIALIZATION_REL_COUNT, Get_PhaseNbr()))
	{
		this->iRSerialAccordingRelations = true;
		param->GetIntegerPhaseParameter(ConstPN::RSERIALIZATION_REL_COUNT, this->iRSerialRelationsCount, Get_PhaseNbr());
	}
	// Running Serialization full filename - default is same as iSerialPhaseFullFileName
	if (param->ContainsPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, this->iRunningSerialFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iRunningSerialFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iRunningSerialFullFileName, ConstPN::RSERIAL_SIEVING_FILENAME, 
			ConstPN::RSERIAL_SIEVING_DIRECTORY, ConstPN::RSERIAL_SIEVING_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::SIEVING_PHASE, Get_SerialIdentifier(), Get_PhaseNbr());
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = Utils::ConvertToString(i + 1);
		// FB full filename
		if (param->ContainsPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, Get_PhaseNbr()))
		{
			param->GetStringPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, this->iFBs[i].FBFullFileName, Get_PhaseNbr());
			MSerializable::AdaptPathSeparator(this->iFBs[i].FBFullFileName);
		}
		else
		{
			param->CreateFullFileName(this->iFBs[i].FBFullFileName, ConstPN::FACTOR_BASE_FILENAME + auxStr, 
				ConstPN::FACTOR_BASE_DIRECTORY + auxStr, ConstPN::FACTOR_BASE_COMPRESSION + auxStr, Get_SerialMainDirectory(),
				ConstFN::FACTOR_BASE + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
		}
	}

	// Smooth relation full filename
	if (param->ContainsPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, this->iSmoothRelFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iSmoothRelFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iSmoothRelFullFileName, ConstPN::SMOOTH_RELATIONS_FILENAME, 
			ConstPN::SMOOTH_RELATIONS_DIRECTORY, ConstPN::SMOOTH_RELATIONS_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::SMOOTH_RELATIONS, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}
	// Partial relation full filename
	if (param->ContainsPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, this->iPartialRelFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iPartialRelFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iPartialRelFullFileName, ConstPN::PARTIAL_REL_FILENAME, 
			ConstPN::PARTIAL_REL_DIRECTORY, ConstPN::PARTIAL_REL_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::PARTIAL_REL, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}

	// should we deserialize or serialize?
	bool deserialize = false;
	param->GetBoolPhaseParameter(ConstPN::DESERIALIZE, false, deserialize, Get_PhaseNbr());
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

	if (Get_InfoMode())
	{
		Utils::NormalMessage("Classical Sieving Phase is running ...");
    }

	if (Get_InfoMode()) fprintf(stdout, "\n\tSieving region: width %ld, height %ld, start %ld.\n", iSievingRegion.width, iSievingRegion.height, iSievingRegion.start);

	if (deserialize)
	{
		rc = Deserialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
inline void CNFSClassicalSievingPhase::UpdateNfsRoot(nfs_fb_type* aEntry)
{
	if (aEntry->root_1 < 0)
	{
		return;
	}

	register main_sieving_type p;
	p = aEntry->p;
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
	bool res = p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME;
	if (res) 
	{
		printf("Update Nfs Root: old information ");
		PrintNfsElementInfo(*aEntry);		
	}
#endif
	if (this->iNfsCompressEvenIntervals)
	{
		// Return from a compressed interval to a normal interval.
		if (this->iNfsEvenToOdd)
		{
			if (p != 2) 
			{
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
				if (res) 
				{
					printf("Update Nfs Root: (de)compression from even to odd. Old value of a root for %d is %d \n",p,aEntry->root_1);
				}	
#endif
				aEntry->root_1 <<= 1;
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
				if (res) 
				{
					printf("Update Nfs Root: (de)compression from even to odd. After multiplication by 2 is the root for %d equal to %d \n",p,aEntry->root_1);
				}	
#endif
				if (aEntry->root_1 >= p)
				{
					aEntry->root_1 -= p;
				}
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
				if (res) 
				{
					printf("Update Nfs Root: (de)compression from even to odd. After possible reduction by p is the root for %d equal to %d \n",p,aEntry->root_1);
				}	
#endif
			}
		}
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
		if (res) 
		{
			printf("Update Nfs Root: old value of a root is %d \n",aEntry->root_1);
		}	
#endif
	}
	aEntry->root_1 += aEntry->auxVal;
	// auxVal can be negative
	if (aEntry->root_1 < 0)
	{
		aEntry->root_1 += p;
	}
	else if (aEntry->root_1 >= p)	
	{
		aEntry->root_1 -= p;
	}
	if (this->iNfsCompressEvenIntervals)
	{
		if (this->iNfsOddToEven)
		{
			if (p != 2)
			{
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
				if (res) 
				{
					printf("Update Nfs Root: compression from odd to even. Old value of a root for %d is %d \n",p,aEntry->root_1);
				}	
#endif
				// We will check whether the result is even. If no, we will add p. 
				// Then, we will divide the result by 2.
				// ? What about algebraic prime 2?
				if (aEntry->root_1 % 2)
				{
					aEntry->root_1 += p;

#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
					if (res) 
					{
						printf("Update Nfs Root: compression from odd to even. Root must be expanded by p, result is %d \n",aEntry->root_1);
					}	
#endif
				}
				aEntry->root_1 >>= 1; // division by 2
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
				if (res) 
				{
					printf("Update Nfs Root: compression from odd to even. After division by 2, the root for %d is %d \n",p,aEntry->root_1);
				}	
#endif	
			}
		}
	}
#ifdef DEBUG_SIEVING_PHASE_UPDATE_NFS_ROOT
	if (res)
	{
		printf("Update Nfs Root: new information ");
		PrintNfsElementInfo(*aEntry);
	}
#endif
	
}
/*!
	The hashtable principle:
	
	We have at least (iBlockSetSize+1) hashtables available.

	Of this, at most aBlockSetSize hashtables are meaningful
	at the moment. The last hashtable is used as an auxiliary
	variable.

	The greatest possible jump from current update to next update
	is by iBlockSetSize hashtables. If it lands in the same hashtable,
	we will rather update the last hashtable, and then switch the contents.
*/
int CNFSClassicalSievingPhase::SieveBlock(int aBlockIndex, int aBlockMin, int aBlockMax, int aSign, bool aInit)
{
	int rc;

	for (int i = aBlockMin; i < aBlockMax; i++) 
	{			
		rc = Sieving(i, aSign, aInit);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

		// Now the factorizable candidates are found and factorized
		rc = FindingAndFactorizing(aBlockIndex, i, aSign);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

		// Now, we must shift the sieving values for the largish primes in the hashtable for the next block. 
		// This has already been done for the middle-sized primes.
		// An open question: is it better to do this apart from the sieving, or within the
		// largish sieving cycle?
		rc = Shifting(i);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

		rc = Swapping(i);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}
	}
	
	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::SieveWithMiddlePrimes(void* aStart, void* aEnd, log_type* aBlock, int aSign, bool aInit) 
{
	register main_sieving_type p;
	register main_sieving_type offset;
	register log_type          log;
	int shiftConst = 2;
	if (this->iNfsCompressEvenIntervals && this->iIsCompressedNow)
	{
		shiftConst = 1;
	}

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	bool assertionMode = Get_AssertionMode();
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	long sum = 0;
	long hits = 0;
#endif
	
	nfs_fb_type* end = (nfs_fb_type*)aEnd;
	nfs_fb_type* running = (nfs_fb_type*)aStart;
	do
	{
		p = running->p;		
		log = running->log_p;

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		if (!log)
		{
			printf("Log is zero for p = %d, strange.\n",p);
		}
#endif

		if (running->root_1 >= 0) 
		{				
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
			printf("SievingWithMiddlePrimes: p = %d, root = %d, curr = %p, end = %p.\n",p,running->root_1,running, aEnd);
#endif
			if (aInit)
			{
				if (aSign == NEGATIVE)
				{
					// shiftConst is to adjust values and offsets for the negative sieve.
					offset = (running->next_1 = p - running->root_1 - shiftConst); 
					if (offset < 0) 
					{
						offset = (running->next_1 += p);
					}
				} 
				else 
				{
					offset = (running->next_1 = running->root_1);
				}
			} 
			else 
			{
				offset = running->next_1;
			}

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
			if (assertionMode)
			{
				if (offset >= 0)
				{								
					//TestNfsAlgebraic(*running,offset);
				}			
			}
#endif

			do 
			{
				*(aBlock + offset) += log;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
				sum += log;
				hits++;
#endif
				offset += p;
			} 
			while (offset < Get_BlockSize());

			// The new index for the next block
			running->next_1 = offset - Get_BlockSize();
			//printf(" The next mod %d would be %d, diff is %d ... test ...\n",p,offset, running->next_1);
			
		} // else the root is invalid

		// Finish sieving with middle primes.
		//if (running == end)
		//{
		//	break;
		//}
		//running++;
	} 
	while (running++ != end); // Will be broken by internal check.

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	if (hits != 0)
	{
		printf("Sieve mid primes: added total %d logs in %d hits (%d on avg).\n",sum,hits,sum/hits);
	}
#endif

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::SieveWithLargishPrimes(script_t* aScript, log_type* aBlock)
{
	script_t* script = aScript;
	update_t *updates = NULL;
	int j, max;
	int blocksize = Get_BlockSize();
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	int total = 0;
	int listIndex = 0;
	int runningIndex = 0;

	long sum = 0;
	int hits = 0;
#endif
	while (script)
	{	
		max = script->num_used;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		total += max;
		listIndex++;
#endif
		updates = script->list;
		for (j = 0; j < max; j++)
		{
	        if ((j & 7) == 0) //if (j % 8 == 0) 
			{
          		PREFETCH(updates + j + 16);
			}

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
			if (Get_AssertionMode()) 
			{
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
				if (updates[j].p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME)
				{
					PrintUpdateInfo(updates[j]);
				}
#endif
				TestUpdate(updates[j]);
			}
#endif
			if (updates[j].h >= blocksize)
			{
				printf("\n..... Line siever, sieve with largish primes: index too big ... %5d[%d]",updates[j].p,updates[j].h);
				return ConstRC::InconsistentState;
			} 
			else 
			{
	        	aBlock[updates[j].h] += updates[j].logp;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
				sum += updates[j].logp;
				hits++;
#endif
			}
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
			runningIndex++;	
#endif
		}
		script = script->next;
	}
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	if (hits) 
	{
		printf("Sieve largish primes: added total %d logs in %d hits (avg %d).\n",sum,hits,sum/hits);
	} 
	else 
	{
		printf("Sieve largish primes: No largish prime availables\n");
	}
#endif

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::ShiftLargishPrimes(CHashtables & aHashtables, int aCurrentBlockIndex)
{
	main_sieving_type newHashtableIndex;
	main_sieving_type newUpdateH;
	int blocksize = Get_BlockSize();
	//int total = 0;
	//int listIndex = 0;
	int j, max, result;
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	int errors  = 0;
#endif
	script_t* script = aHashtables.Hashtables[aCurrentBlockIndex].script;
	update_t *updates = NULL;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
 	printf("Shift largish primes: current block index is %d.\n",aCurrentBlockIndex);
#endif
	while (script)
	{
		max = script->num_used;
		//printf("List Index %d: num used is %d.\n",listIndex,max);
		//total += max;
		//listIndex++;
		updates = script->list;
		for (j = 0; j < max; j++, updates++)
		{
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
			if (errors >= 10)
			{
				return ConstRC::InconsistentState;
			}
#endif
	       	if ((j & 7) == 0) //if (j % 8 == 0) 
			{
          		PREFETCH(updates + 16);
			}
			
			newHashtableIndex = aCurrentBlockIndex + (newUpdateH = (updates->h + updates->p)) / blocksize;
			//if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME)
			//{
			//	printf("Shift largish primes: current h is %d, new h is %d.\n",updates->h,newUpdateH);	
			//	PrintUpdateInfo(*updates);		
			//}
			if (newHashtableIndex > aHashtables.MaxIndex) //newHashtableIndex %= aHashtables.MaxIndex;
			{
				newHashtableIndex -= aHashtables.MaxIndex;
			}
			//if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME)
			//{
			//	printf("Mod block is %d.\n",newHashtableIndex);			
			//}
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			if (updates[j].p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME)
			{
				printf("------------------------------------\n");
				printf("Shift largish primes: p = %d, alg = %d, current h is %d, new h is %d.\n",updates->p, iSievingAlgebraic, updates->h,newUpdateH);	
				PrintUpdateInfo(*updates);		
				printf("Mod block is %d.\n",newHashtableIndex);	
			}
#endif
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
			if (Get_AssertionMode())
			{
				if (TestUpdate(*updates) != ConstRC::Ok)
				{
					errors++;
				}
			}
#endif
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			//if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) 
			//{
			//	printf("New update index    je %6d.\n",newUpdateH);
			//	printf("New hashtable index je %6d.\n",newHashtableIndex);
			//	printf("Block set size je rove %6d.\n",aHashtables.MaxIndex);
			//}
#endif
			newUpdateH %= blocksize;
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			//if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) {
			//	printf("~~After mod:.\n");
			//	printf("New hashtable index je %6d.\n",newHashtableIndex);
			//	printf("New update h je roven  %6d.\n",newUpdateH);
			//}
#endif
			if (newHashtableIndex == aCurrentBlockIndex)
			{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
				if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) 
				{
					printf("\n   We will go to k+1 with prime %d at index %d.\n",updates->p, newUpdateH);
				}
#endif

				if (newUpdateH >= updates->p) 
				{
					printf("II. Root1 %d > p %d.\n",newUpdateH,updates->p);
				}
				
				result = aHashtables.AddToTable(updates->p, updates->r, updates->logp, newUpdateH, aHashtables.MaxIndex);
				if (result != ConstRC::Ok)
				{
					return result;
				}
			} 
			else 
			{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
				if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) 
				{
					printf("We will go to %d with prime %d at index %d. Block set size is %d. \n\n",newHashtableIndex,updates->p, newUpdateH,iBlockSetSize);
				}
#endif

				result = aHashtables.AddToTable(updates->p, updates->r, updates->logp, newUpdateH, newHashtableIndex);
				if (result != ConstRC::Ok)
				{
					return result;
				}
			}
		}
		script = script->next;
	}

	// Flush all caches for all hashtables.
	aHashtables.FlushAllUpdateCaches();

	return ConstRC::Ok;
}


int CNFSClassicalSievingPhase::FillHashtable(CHashtables& aHashtables, nfs_fb_type * aStart, int aSign)
{
	main_sieving_type p;
	main_sieving_type root_1;	
	main_sieving_type r;
	log_type log_p;
	int ret;
	int shiftConst = 2;
	int blocksize = Get_BlockSize();
	if (this->iNfsCompressEvenIntervals && this->iIsCompressedNow)
	{
		shiftConst = 1;
	}
	
	aHashtables.ClearHashtables();

	if (aStart == NULL)
	{
		// In case that all primes are middle (no largish exist).
		return ConstRC::Ok;
	}

	//printf(" Fill hashtable running. ...");
	// The upper bound of the FB is denoted by a zero block;
	nfs_fb_type* running = aStart;
	while ((p = running->p) != 0)
	{		
 		root_1 = running->root_1;
		log_p  = running->log_p;
		// if root is valid ...
		if (root_1 >= 0)
		{
			//printf("Running.root_1 = %d.\n",root_1);
			if (aSign == NEGATIVE)	
			{
				// Recalculate the root for the negative sieving
				root_1 = p - root_1 - shiftConst; // shiftConst ... offset move caused by offset policy
				if (root_1 < 0)
				{
					root_1 += p;
				}
			}

			r = running->r;

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			if (p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) 
			{
				printf("\nFill Hashtable Algebraic: Adding to table: p = %d, r = %d, root_1 = %d.\n",p,r,root_1);
				Utils::Wait();
			}
#endif
			
			if (root_1 >= p) 
			{
				printf("I. Root1 %d > p %d.\n", root_1, p);
			}

			// Add to the update table
			ret = aHashtables.AddToTable(p, r, log_p, root_1 % blocksize, root_1 / blocksize);			
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
		} 

		// Recalculation of roots must be done for each largish entry.
		if (aSign == NEGATIVE)
		{
			UpdateNfsRoot(running);	
		}
		running++;
	}

	// Flush all caches
	aHashtables.FlushAllUpdateCaches();

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::Swapping(int aIndex)
{
	int rc;
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = this->iHashtables[i].SwapHashtable(aIndex);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::Shifting(int aIndex)
{
	int rc;
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
		printf("Calling Shift Largish Primes FB nbr %d.\n", i);
#endif

		rc = ShiftLargishPrimes(this->iHashtables[i], aIndex);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
		printf("Shift Largish Primes Finished.\n");
#endif
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::FindingAndFactorizing(int aBlockIndex, int aInnerIndex, int aSign)
{
#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
	//for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	//{
	//	WriteDumpSieveFile(this->iSievingBlocks[i]);
	//}
#endif

	// FOR DEBUG
	//if (aInnerIndex > 64)
	//{
	//	ControlBlock(aBlockIndex, aInnerIndex, aSign);
	//}
	//return ConstRC::Ok;

	// we start with last block
	log_type * block = this->iSievingBlocks[1]; // TODO - without magic numbers
	log_walking_type* cast  = (log_walking_type*)block;
	log_walking_type  mask  = SIEVING_PHASE_CHECK_MASK;
	int size = sizeof(log_walking_type);
	if (size == SIEVING_PHASE_64BITS_SIZE)
	{
		mask <<= ((size / 2) * 8);
		mask += SIEVING_PHASE_CHECK_MASK;
	}
	int rc;
	bool checkAll = Get_CheckAllMode();
	int blockSize = Get_BlockSize();
	int iters = blockSize / size;
	int offset = 0;
	int i, j;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	int mask_hits  = 0;
	int candidates = 0;
	int sizes[SIEVING_PHASE_SIZES_COUNT];
	memset(sizes, 0, SIEVING_PHASE_SIZES_COUNT * sizeof(int));
#endif

	for (i = 0; i < iters; i++)
	{
		if ((cast[i] & mask)  || checkAll) 
		{ 
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
			mask_hits++;
#endif
			offset = size * i;	
			for (j = 0; j < size; j++)
			{
				
				if ((block[offset + j] & SIEVING_PHASE_MASK) || checkAll)
				{					
					if ((offset + j) >= blockSize)
					{
						printf("Size: %d, i = %d, offset = %d, j = %d, offset+j = %d, pointer = 0x%p, iters = %d.\n",size,i,offset,j,offset+j,&(block[offset+j]),iters);
						Utils::Wait();
					}

					if ((this->iSievingBlocks[0][offset + j] & SIEVING_PHASE_MASK) || checkAll)
					{
						rc = Factorize(offset + j, aBlockIndex, aInnerIndex, aSign);	
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
						candidates++;	
#endif
					}
				} 
				else 
				{
					if ((offset + j) >= blockSize)
					{
						printf("Size: %d, i = %d, offset = %d, j = %d, offset+j = %d, pointer = 0x%p, iters = %d.\n",size,i,offset,j,offset+j,&(block[offset+j]),iters);
						Utils::Wait();
					}
				}							
			}
		} 
	}

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
 	printf("Had %d mask hits and %d candidates for factorization. Size statistics:\n",mask_hits,candidates);
#endif

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::Sieving(int aIndex, int aSign, bool aInit)
{
	int rc;
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_COMPUTE_EXP_ALG_LOG
		if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
		{
			this->iInitBlockValues[i] = SIEVING_PHASE_MASK - ComputeQuickExpectedAlgLog(this->iBlockIndex, aIndex, aSign, 
				this->iInitBlockValues[i], this->iNumberFields[i].SievingPoly, this->iVariationsInfo[i]);
		}
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE		
		printf("Sieving: index = %d. Memset the block to 0x%x, size %d.\n",aIndex,this->iInitBlockValues[i] - this->iInitBlockValueChange[i],Get_BlockSize()*sizeof(log_type));
#endif

		// iInitAlgebraic1BlockValue = iNfsInstance->ComputeExpectedAlgebraicLog(0); was called in SieveLine		
		if (memset(this->iSievingBlocks[i], this->iInitBlockValues[i] - this->iInitBlockValueChange[i], Get_BlockSize() * sizeof(log_type)) == NULL)
		{
			CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Sieving",0,Get_BlockSize() * sizeof(log_type),"Memset failed!");
			return ConstRC::NotEnoughMemory;
		}

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
		printf("Calling Sieve With Middle Primes For FB nbr. %d.\n",i);
#endif

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstSieving + i);
#endif
		// Then, we sieve with the middle-sized primes.
		rc = SieveWithMiddlePrimes(this->iAdvFB[i].FirstMiddlePrime, this->iAdvFB[i].LastMiddlePrime, 
			this->iSievingBlocks[i], aSign, aInit && (aIndex == 0));
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
		WriteDumpSieveFile(this->iSievingBlocks[i]);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
		printf("Sieve With Middle Primes For FB nbr. %d Finished. Calling Sieve With Largish Primes.\n",i);
#endif

		// Then, we will sieve with the largish primes	
		rc = SieveWithLargishPrimes(this->iHashtables[i].Hashtables[aIndex].script, this->iSievingBlocks[i]);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
		WriteDumpSieveFile(this->iSievingBlocks[i]);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
		printf("Sieve With Largish Primes For FB nbr. %d Finished.\n", i);
#endif

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstSieving + i);
#endif

	}

	return ConstRC::Ok;
}

int CNFSClassicalSievingPhase::AddToCountStructures(CNFSRelation *aRelation) {
	int rc;

	if ((aRelation->Type & RELATION_FULL_ALGEBRAIC1) && (aRelation->Type & RELATION_FULL_ALGEBRAIC2))
	{
		aRelation->Index = this->iStats->Get_Total(ESSFF);
		this->iStats->Increase(ESSFF);
	}
	else
	{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		if ((aRelation->Type & RELATION_PARTIAL_ALGEBRAIC1) && (aRelation->Type & RELATION_FULL_ALGEBRAIC2))
		{
			this->iStats->Increase(ESSPF);
		}
		else if ((aRelation->Type & RELATION_FULL_ALGEBRAIC1) && (aRelation->Type & RELATION_PARTIAL_ALGEBRAIC2))
		{
			this->iStats->Increase(ESSFP);
		}
		else
		{
			this->iStats->Increase(ESSPP);
		}
#endif
		aRelation->Index = this->iStats->Get_Total(ESSPartialR);
		this->iStats->Increase(ESSPartialR);
		if (this->iVariationsInfo[0].Type == ELargePrimeV &&
			this->iVariationsInfo[1].Type == ELargePrimeV &&
			IndependentCyclesCount() < ConstSPhase::MAX_INDEPENDENT_CYCLES_COUNT)  
		{
			if ((rc = AddToCycleCountStructures(aRelation)) != ConstRC::Ok) return rc;
		}
	}

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
	this->iStats->Increase(ESSAccepted);
#endif

	return ConstRC::Ok;
}

int CNFSClassicalSievingPhase::Factorize(main_sieving_type aOffset, int aBlockIndex, int aInnerIndex, int aSign)
{
	int shiftConst = 2;
	if (this->iNfsCompressEvenIntervals && this->iIsCompressedNow)
	{
		shiftConst = 1;
	}

	bool directionA = ((this->iSievingDirection == EIncrementA) || (this->iSievingDirection == EDecrementA));	

	int blockSize = Get_BlockSize();
	// This will be comparable (via mod) to the root ...
	main_sieving_type auxiliaryOffset = aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize + aOffset;
	// At first, we will calculate the relation-related values
	int x = aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize + aOffset;
	
	if (this->iNfsCompressEvenIntervals && this->iIsCompressedNow)
	{
		x *= 2;
	}

	x += 1;
	if (aSign == -1)
	{
		x *= -1;				
#ifdef FACTOR_DEBUG
		printf("\n");
		printf("======= NEGATIVE ======= ");
#endif
	} 
	else 
	{
#ifdef FACTOR_DEBUG
		printf("\n");
		printf("======= POSITIVE ======= ");
#endif
	}

#ifdef FACTOR_DEBUG
	printf("Factor: x is equal to %d.\n",x);
#endif

	this->iRelation->Reset();

	main_sieving_type a;
	main_sieving_type b;
	switch (this->iSievingDirection) 
	{
		case EIncrementA:
		case EDecrementA:							
			a = this->iWorkingA;
			b = x;
			break;
		case EIncrementB:
		case EDecrementB:
			a = x;
			b = this->iWorkingB;
			break;			
	}

	this->iRelation->Setup(a, b, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iNumberFields);

#ifdef FACTOR_DEBUG
	printf("Factor: [a,b] = [%d,%d].\n", this->iRelation->A, this->iRelation->B);
#endif
	
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
	this->iStats->Increase(ESSTested);
#endif
	if (this->iRelation->TestCoprimality() == false)
	{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		this->iStats->Increase(ESSRejected);
		this->iStats->Increase(ESSNotCoprime);
#endif
		return ConstRC::Ok;
	}

	int i;
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iStats->Increase(ESSAverageSizeIP_log2 + i, mpz_sizeinbase(this->iRelation->Get_Parts(i)->Norm, 2));
		this->iStats->Increase(ESSAverageSizeIP_log10 + i, mpz_sizeinbase(this->iRelation->Get_Parts(i)->Norm, 10));
	}
#endif

	int rc;

	// Now, we have factored the relation with all factor base primes. But we
	// need to find potential large primes too.
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstFactorization + i);
#endif
		if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
		{
			rc = FactorizeAlgebraic(shiftConst, directionA, aSign, aOffset, auxiliaryOffset, &this->iHashtables[i].Hashtables[aInnerIndex], i);
		}
		else
		{
			rc = FactorizeIntegral(shiftConst, directionA, aSign, aOffset, auxiliaryOffset, &this->iHashtables[i].Hashtables[aInnerIndex], i);
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstFactorization + i);
#endif
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (this->iVariationsInfo[i].Control(this->iRelation->Get_Parts(i)->NormRemaining) == false)
		{
			this->iOptimizer[i].IncrementRejected();

			// remaining factor is too big, we don't have to continue. 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
			this->iStats->Increase(ESSRejected);
#endif
			return ConstRC::Ok;
		}
		// this part is good
		this->iOptimizer[i].IncrementAccepted();	
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTPartialFactorization);
#endif
	rc = this->iRelation->DetermineType(CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iVariationsInfo);
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTPartialFactorization);
#endif
	if (rc == ConstRC::Ok)
	{
		if ((rc = AddToCountStructures(iRelation)) != ConstRC::Ok) return rc;

		if (this->iRelationWorkingBufferMaxCount < this->iRelationWorkingBuffer->size())
		{
			rc = this->iRelation->Copy2(*this->iRelationWorkingBuffer->at(this->iRelationWorkingBufferMaxCount));
		}
		else
		{
			this->iRelationWorkingBuffer->push_back(new CNFSRelation(*this->iRelation));
		}
		this->iRelationWorkingBufferMaxCount++;

		return rc;
	}
	else if (rc == ConstRC::Impropriate)
	{

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		this->iStats->Increase(ESSRejected);
#endif
		return ConstRC::Ok;
	}
	else
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::FactorizeIntegral(int aShiftConst, bool aDirectionA, int aSign, 
		main_sieving_type aOffset, main_sieving_type aAuxOffset, hashtable_t *aHashtable, int aIndex)
{
	bool assertionMode = false;

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	assertionMode = Get_AssertionMode();
#endif

	int lastWrittenIndex = -1;
	int rc;
	int exponent;
	main_sieving_type prime;
	main_sieving_type rem    = 0;	
	main_sieving_type root_1 = 0;

	CDivisors * divisors = &this->iRelation->Get_Parts(aIndex)->Main;
	mpz_t * remaining = &this->iRelation->Get_Parts(aIndex)->NormRemaining;
	nfs_fb_type* running = this->iAdvFB[aIndex].FactorBase + 1; // skip -1
	nfs_fb_type* moddivision_end = this->iAdvFB[aIndex].LastMiddlePrime;	

	// The integral part
#ifdef FACTOR_DEBUG
	gmp_printf("Factor: remaining is %Zd.\n",*remaining);
#endif

	// At first, we will take care of -1.
	if (mpz_cmp_si(*remaining, 0) < 0) 
	{
		prime = -1;
		exponent = 1;
		rc = divisors->AddDivisor(prime, exponent, 0, 0, lastWrittenIndex, assertionMode);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}
		mpz_neg(*remaining, *remaining);

#ifdef FACTOR_DEBUG
		gmp_printf("Factor: was negative, added -1 to list.\n");
#endif
	}

	// now we check prime = 2
	exponent = mpz_scan1(*remaining, 0);
	if (exponent)
	{
		prime = 2;
		rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}       

#ifdef FACTOR_DEBUG
		gmp_printf("Factor: Reduced %Zd by 2^%d.\n", *remaining, exponent);
		//Utils::Wait();
#endif

		mpz_fdiv_q_2exp(*remaining, *remaining, exponent);				
	}
	running++; // next prime after 2
	

	/* Now, the tiny primes and middle primes*/
	while (running <= moddivision_end)
	{
		// Compare the required 
		prime = running->p;

		// We must specially solve the p's for which p | m.			
		if (aDirectionA)
		{
			if (running->flags & ConstEF::DIVIDES_ROOT_MINI_FLAG)
			{
				//printf("Factor: prime %d divides M.\n",prime);
				if (!(this->iRelation->A % prime))
				{
					exponent = 0;			
					while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
					
					if (exponent) 	
					{
						rc = divisors->AddDivisor(prime, exponent, running->r, ConstEF::DIVIDES_ROOT, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok) 
						{ 
							return rc; 
						}      
					} 
					else 
					{
						gmp_printf("\nINT FactorError: reduction of %Zd by %d | m impossible: exponent %d, aOffset = %d, auxiliaryOffset = %d, rem = %d, root_1 = %d, running_root1 = %d, condition1 = %d\n",*remaining,prime,exponent,aOffset,aAuxOffset,rem,root_1,running->root_1,(rem == root_1));
					}
				}
				running++;
				continue;
			}
		}

		if (aSign == -1) 
		{
			rem = aAuxOffset % prime;
			// if (valid root 1)
			if (running->root_1 >= 0) 
			{
				root_1 = prime - running->root_1 - aShiftConst;
				if (root_1 < 0) 
				{
					root_1 += prime;
				}
			} 
			else 
			{
				root_1 = -1;
			}
		} 
		else 
		{
			rem = aAuxOffset % prime;
			root_1 = running->root_1;
		}
		if (rem < 0) 
		{ 
			rem += prime; 
		}

		//if ((rem == root_1) || (running->flags & DIVIDES_L_MINI_FLAG))
		if (rem == root_1)
		{
			exponent = 0;
		
			while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
			{
				exponent++;
				mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
			this->iStats->Increase(ESSDivrem);
#endif

			if (exponent) 	
			{
#ifdef FACTOR_DEBUG
				gmp_printf("Factor: reducing %Zd by %d^%d.\n", *remaining, prime, exponent);
#endif
				rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok) 
				{ 
					return rc; 
				}            
			} 
			else
			{
				gmp_printf("\nINT FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, auxiliaryOffset = %d, rem = %d, root_1 = %d, running_root1 = %d, condition1 = %d\n",*remaining,prime,exponent,aOffset,aAuxOffset,rem,root_1,running->root_1,(rem == root_1));
			}		
		}
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
		// Not divisible by p
		else if (assertionMode)
		{
			rem = mpz_mod_ui(this->iAuxMpz1, *remaining, prime);
			if (!rem)
			{
				gmp_printf("FactorError: the value %Zd should be divisible by %d, but offset indicates otherwise. aOffset = %d. Auxiliary offset is %d, rem is %d, root1 is %d, r is %d \n",*remaining,prime,aOffset,aAuxOffset,rem,running->root_1,running->r);
			}

		} 
#endif
		running++;
	}

	// Now, the largish primes. We will walk through the current update table ...
	int k, max;
	update_t *updates;

	if (aHashtable)
	{
		script_t* script = aHashtable->script;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (k = 0; k < max; k++, updates++)
			{
				if (updates->h == aOffset)
				{
					prime = updates->p;
					//gmp_printf("Factor: %Zd should be divisible by %d.\n",*remaining,updates->p);
					exponent = 0;			
					while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
						gmp_printf("Factor: reducing %Zd by %d^%d.\n", *remaining, prime, exponent);
#endif
						rc = divisors->AddDivisor(prime, exponent, updates->r, 0, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok) 
						{ 
							return rc; 
						}            
					} 
					else 
					{
						gmp_printf("\nINT FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, auxiliaryOffset = %d, h = %d.\n",*remaining,prime,exponent,aOffset,aAuxOffset,updates->h);
					}			
				} 
			}
			script = script->next;
		}
	}


#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (assertionMode)
	{
#ifdef FACTOR_DEBUG
		gmp_printf("Factor: assertion for large primes. Remaining %Zd, ass start %p\n", *remaining, this->iAdvFB[aIndex].FirstLargishPrime);
#endif
		if (mpz_cmp_ui(*remaining, 1) != 0 && this->iAdvFB[aIndex].FirstLargishPrime != NULL)
		{
			
			// We will try all remaining large primes
			running = this->iAdvFB[aIndex].FirstLargishPrime;
			while (running != NULL)
			{
				prime = running->p;
				//printf("%7d",prime);	fflush(stdout);
				if (!prime) 
				{
					break;
				}
				if (mpz_mod_ui(this->iAuxMpz1, *remaining, prime) == 0)
				{
					gmp_printf("\nFactorIntegralError: remaining %Zd is divisible by largish prime %d, but this prime hasn't been foung during the process.\n", *remaining, prime);
					printf("Now searching for position of the prime in hashtables ....\n");
					// Search
					this->iHashtables[aIndex].SearchForPrimeInHashtable(prime, running->r);
					Utils::Wait();
				}
				running++;
			}
		}
	}
#endif

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::FactorizeAlgebraic(int aShiftConst, bool aDirectionA, int aSign, 
		main_sieving_type aOffset, main_sieving_type aAuxOffset, hashtable_t *aHashtable, int aIndex)
{
	int rc;
	int lastWrittenIndex = -1;

	bool assertionMode = false;
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	assertionMode = Get_AssertionMode();
#endif

	CNumberFieldInfo * nf = this->iNumberFields + aIndex;
	CDivisors * divisors = &this->iRelation->Get_Parts(aIndex)->Main;
	mpz_t * remaining = &this->iRelation->Get_Parts(aIndex)->NormRemaining;
	nfs_fb_type* running = this->iAdvFB[aIndex].FactorBase;
	nfs_fb_type* end = this->iAdvFB[aIndex].LastMiddlePrime;
	
	main_sieving_type exponent = 0, prime, rem, root_1;
	unsigned int flag;
	long valuation;

	/* Now, the tiny primes and middle primes*/
	while (running <= end)
	{
		prime = running->p;
		exponent = 0;
		//printf("Prime %d.\n",prime);
		if (!(running->flags & ConstEF::SPECIAL_PRIME_MINI_FLAG) && !(running->flags & ConstEF::DIVIDES_L_MINI_FLAG)) 
		{
			//printf("Not a special prime. \n");
			if (aDirectionA && (running->flags & ConstEF::DIVIDES_ROOT_MINI_FLAG))
			{
				//printf("Prime %d divides c_p. \n",prime);
				// In this case, prime divides norm iff prime divides current 'a'.
				main_sieving_type a = this->iRelation->A;
				if (!(a % prime))
				{									
					while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					}
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					// Some serious error happened in the algorithm 
					if (!exponent)
					{
						// TODO: elegantneji
						if ((aShiftConst != 1) || (prime !=  2))
						{
							gmp_printf("\nALG FactorError: reduction of %Zd by %d impossible: exponent %d, aAuxiliaryOffset = %d. c_p divides case.\n", *remaining, prime, exponent, aAuxOffset);
						}
					} 
					else 
					{
						//printf("Added exponent %d for %d.\n",exponent,prime);
					}
				}
			} 
			else // does not divide c_p. Normal test
			{
				//if (prime == 3) printf("3 does not divide c_p,\n");
				if (aSign == -1) 
				{
					rem = aAuxOffset % prime;
					// if (valid root 1)
					if (running->root_1 >= 0) 
					{
						root_1 = prime - running->root_1 - aShiftConst;
						if (root_1 < 0) 
						{
							root_1 += prime;
						}
					} 
					else 
					{
						root_1 = -1;
					}
				} 
				else 
				{ 
					// sieving positive
					rem = aAuxOffset % prime;
					root_1 = running->root_1;
	 			}		
				if (rem < 0) 
				{ 
					rem += prime; 
				}
				if (rem == root_1)
				{
					//printf(" rem = %d = root_1.\n",rem);
								
					while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
						gmp_printf("FactorAlgebraic: reducing %Zd by %d^%d.\n",*remaining,prime,exponent);
#endif
					} 
					else  // some problem has occurred
					{
						// TODO: elegantneji
						if ((aShiftConst != 1) || (prime != 2))
						{
							gmp_printf("\nALG FactorError: reduction of %Zd by %d impossible: exponent %d, aAuxiliaryOffset = %d. c_p divides case.\n", *remaining, prime, exponent, aAuxOffset);
						}

					}	
				} // if rem == root_1
			}
			// Save the divisor if appropriate
			if (exponent)
			{
				rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
				exponent = 0;
			}
		} // end if - normal prime ideal
		else if ((running->flags & ConstEF::SPECIAL_PRIME_MINI_FLAG) == 0 &&
			(running->flags & ConstEF::DIVIDES_L_MINI_FLAG) && prime != running->r)
		{
			// we do two ideals at once

			// ideal (p) valuation of (c_{d}a + \bar{\theta}b)
			if (this->iRelation->B % prime == 0)
			{
				root_1 = this->iRelation->B;
				do
				{
					// TODO better
					exponent++;
					root_1 /= prime;
				}
				while (root_1 % prime == 0);
			}

			// p - valuation of F(a, -b)
			valuation = 0;
			if (mpz_divisible_ui_p(*remaining, prime))
			{
				while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
				{
					valuation++;
					mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
				} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			}

			// we have prime ideal (p, \bar{\theta} - 0) and ideal (p,p)
			root_1 = nf->Domain->ThetaPolynomial->GetDegree();
			rem = valuation + (running->flags & ConstEF::L_VALUATION_MINI_MASK) * (root_1 - 1) - exponent * root_1;
			
			flag = ConstEF::DIVIDES_L | root_1;
			flag |= ((running->flags & ConstEF::L_VALUATION_MINI_MASK) << ConstEF::L_VALUATION_SHIFT);
			if (exponent == 0)
			{
				flag |= (valuation << ConstEF::F_VALUATION_SHIFT);
			}

			if (rem > 0)
			{
				// save (p, \bar{\theta} - 0)
				rc = divisors->AddDivisor(prime, rem, running->r, flag, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}

			if (exponent > 0)
			{
				// save (p, p) with p-valuation of F(a, -b)
				flag |= (valuation << ConstEF::F_VALUATION_SHIFT);
				rc = divisors->AddDivisor(prime, exponent, prime, flag, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
		}

		running++;
	} // while (running <= moddivison_end

	// Now, the largish primes. We will walk through the current update table ...
	if (aHashtable)
	{
		int k, max;
		update_t *updates;
		script_t* script = aHashtable->script;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (k = 0; k < max; k++, updates++)
			{
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
				if (updates->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME)
				{
					printf("Factor algebraic script message:\n");						
					PrintUpdateInfo(*updates);
					Utils::Wait();
				}
#endif
				if (updates->h == aOffset)
				{
					prime = updates->p;
					//gmp_printf("FactorAlgebraic: %Zd should be divisible by %d.\n",*remaining,updates->p);
					exponent = 0;			
					while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
						gmp_printf("FactorAlgebraic: reducing %Zd by %d^%d.\n", *remaining, prime, exponent);
#endif
						rc = divisors->AddDivisor(prime, exponent, updates->r, 0, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok) 
						{ 
							return rc; 
						}            		
					} 
					else 
					{							
						gmp_printf("\nALG FactorError: reduction of %Zd by %d impossible: exponent %d, auxiliaryOffset = %d, h = %d.\n", *remaining, prime, exponent, aAuxOffset, updates->h);
						this->iRelation->PrintToScreen();
						//Utils::Wait();
					}			
				} 
			}
			script = script->next;
		}
	}

		// factorize with special ideals
	int size = nf->Domain->SpecialPrimesCount;
	if (size > 0)
	{
		// firstly - we find out if the remaining is low enough
		// and save exponents for other computing...
		int i;
		bool good = false;
		for (i = 0; i < size; i++)
		{
			exponent = 0;
			prime = nf->Domain->SpecialPrimes[i];
			while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
			{
				good = true;
				exponent++;
				mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			}
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
			this->iStats->Increase(ESSDivrem);
#endif
			nf->SpecialPrimes[prime] = exponent;
		}

		// is it worth to continue?
		if (good && this->iVariationsInfo[aIndex].Control(*remaining))
		{
			// find base of ideal (c_{d}a + b\bar{\theta})
			if (nf->SievingPolyMonic)
			{
				nf->Domain->ABaseOfPrincipal(nf->IdealBase, this->iRelation->A, this->iRelation->B);
			}
			else
			{
				// non-monic poly
				mpz_mul_si(this->iAuxMpz2, *nf->SievingPoly->Leading2(), this->iRelation->A);
				nf->Domain->ABaseOfPrincipal(nf->IdealBase, this->iAuxMpz2, this->iRelation->B);
			}

			log_type flags; 
			SpecialPrimeIt it;
			PrimeIdeal * ideal = NULL;
			// we find out exact factorization
			size = nf->SpecialIdeals.size();
			for (i = 0; i < size; i++)
			{
				ideal = nf->SpecialIdeals[i];
				prime = ideal->Prime;
				it = nf->SpecialPrimes.find(prime);
				flags = this->iAdvFB[aIndex].FactorBase[ideal->Index].flags;
				if ((*it).second > 0 || flags & ConstEF::DIVIDES_L_MINI_FLAG)
				{
					rc = nf->Domain->ValuationOfPrime(&valuation, ideal, nf->IdealBase);
					if (rc != ConstRC::Ok)
					{
						return rc;
					}

					//printf("Valuation of ideal with prime %d is %ld.\n",prime,valuation);
					exponent = valuation;
					if (exponent > 0) 
					{
						// we note that this is special prime and we save its residual
						flag = ConstEF::SPECIAL_PRIME | ideal->Residual;
						if (flags & ConstEF::DIVIDES_L_MINI_FLAG)
						{
							valuation = (*it).second;
							nf->SpecialPrimes[prime] = 0;
							// if this prime divides the leading coeff than we save exact valuation of F(a,-b)
							flag |= ConstEF::DIVIDES_L | ((unsigned int)valuation << ConstEF::F_VALUATION_SHIFT);
							// we also save this prime valuation of the leading coeff
							flag |= ((flags & ConstEF::L_VALUATION_MINI_MASK) << ConstEF::L_VALUATION_SHIFT);
						}
						else
						{
							// norm of prime ideal is p^{residual} so if valuation > 0 than F(a,-b) is divisible by p^{valuation*residual}
							valuation *= ideal->Residual;

							nf->SpecialPrimes[prime] = (*it).second - valuation;
						}

						rc = divisors->AddDivisor(prime, exponent, ideal->Cp, flag, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
					
					} // end if - exponent

				} // end if - used this special idesl

			} // end of for cycle - special ideals
		
		} // end if - continue?

	} // end if - special primes

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (assertionMode)
	{
		//gmp_printf("Factor: assertion for large prime ideals. Remaining %Zd, ass start %p\n",*remaining,this->iAdvFB[aIndex].FirstLargishPrime);
		if (mpz_cmp_ui(*remaining, 1) && this->iAdvFB[aIndex].FirstLargishPrime != NULL)
		{
			// We will try all remaining large primes
			running = this->iAdvFB[aIndex].FirstLargishPrime;
			while (running != NULL)
			{
				prime = running->p;
				//printf("%7d",prime);				
				if (!prime) 
				{
					break;
				}
				if (mpz_mod_ui(this->iAuxMpz1, *remaining, prime) == 0)
				{
					gmp_printf("\nFactorAlgebraicError: remaining %Zd is divisible by largish prime %d, but this prime hasn't been found during the process.\nOffset for this element is %d.\n",*remaining,prime,aOffset);
					printf("Now searching for position of the prime in hashtables ....\n");
					// Search
					this->iRelation->PrintToScreen();
					this->iHashtables[aIndex].SearchForPrimeInHashtable(prime, running->r);
					Utils::Wait();
				}
				running++;
			}			
		}
	}
#endif

	return ConstRC::Ok;
}

void CNFSClassicalSievingPhase::DetermineRequestedRelations() {
	int i;

	this->iRequestedRelations = ConstSPhase::RELATION_RESERVE;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iRequestedRelations += this->iFBs[i].FBMaxIndex + 1;
		if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
		{
			this->iRequestedRelations += ConstSPhase::QUAD_CHARS_COUNT;
		}
	}

	return;
}

int CNFSClassicalSievingPhase::RunSieve()
{
	Set_InnerState(ESPhaseSieving);

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTRunSieve);
#endif

	int rc = PrepareForSieving();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	DetermineRequestedRelations();

	bool enough = false;
	//printf("\nLine Siever: Run Sieve active. Now calling run sieve...");

	bool region_end = false;

	int iter = Get_Iteration() - 1;
	main_sieving_type now;

	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iRunningSerialFullFileName);
	param.SetParameter(ConstPN::RUNNING_SERIALIZATION, Utils::ConvertToString(true));
	rc = StartRunningSerialization(param);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	if (this->iRSerialAccordingTime)
	{
		this->iTimes->Start(ESTRSerial);
	}

	if (Get_InfoMode())
	{
		PrintHeader();
	}

	do
	{
		iter++;

		Set_WorkingIteration(iter);

		if (Get_InfoMode())
		{
			PrintValues();
		}

		if (Get_InfoMode() && Get_TimeMessage() > 0)
		{
			if ((iter % Get_TimeMessage()) == 0)
			{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
				this->iTimes->End(ESTRunSieve);
				this->iTimes->Start(ESTRunSieve);
#endif
				PrintEffectiveness();
				PrintHeader();
				
				this->iStats->ResetAllIntermediate();
				this->iTimes->ResetAllIntermediate();
			}
		}

		// running serialization
		if (this->iRSerialAccordingTime)
		{
			this->iTimes->End(ESTRSerial);
		}
		if (this->iRelationWorkingBufferMaxCount > ConstSPhase::MAX_RELATION_BUFFER_SIZE || 
			(this->iRSerialAccordingRelations && this->iRelationWorkingBufferMaxCount > this->iRSerialRelationsCount) ||
			(this->iRSerialAccordingLines && ((iter % this->iRSerialLineInterval) == 0)) ||
			(this->iRSerialAccordingTime && this->iTimes->Get_TotalTime(ESTRSerial) > this->iRSerialTimeInterval))
		{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
			this->iTimes->End(ESTRunSieve);
			this->iTimes->Start(ESTRunSieve);
#endif
			rc = RunRunningSerialization();
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			if (this->iRSerialAccordingTime)
			{
				this->iTimes->Reset(ESTRSerial);
			}
		}
		if (this->iRSerialAccordingTime)
		{
			this->iTimes->Start(ESTRSerial);
		}

		now = (main_sieving_type)0;
		switch (this->iSievingDirection)
		{
			case EIncrementA:
			case EDecrementA:
				now = this->iWorkingA;
				break;
			case EIncrementB:
			case EDecrementB:
				now = this->iWorkingB;
				break;
			default:
				CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","RunSieve",0,0,"Error in sieving direction.");
				break;
		}

		if (!(now % 2))
		{
			// The next value will be odd, current is even
			this->iNfsEvenToOdd = true;
			this->iNfsOddToEven = false;
			this->iIsCompressedNow = true;
		} 
		else 
		{
			// The next value will be even, current is odd
			this->iNfsEvenToOdd = false;
			this->iNfsOddToEven = true;
			this->iIsCompressedNow = false;
		}

 		//printf("LineSiever: Calling SieveLine()\n");
		rc = SieveLine();
		if (rc != ConstRC::Ok)
		{
 			//printf("LineSiever: The result of SieveLine() was not OK = %d.\n", rc);
			return rc;
		}	

		if (Get_MaxIteration() != ConstSPhase::UNLIMITED_ITERATIONS)
		{
			if (iter >= Get_MaxIteration()) 
			{
				printf("\n      Total time spent by sieving itself is %12.4f sec.\n", this->iTimes->Get_TotalTime(ESTLineSieving));
			 	break; 
			}
		}

		enough = (CheckEnough() == ConstRC::EnoughSmooths);		

		switch (this->iSievingDirection)
		{
			case EIncrementA:
				this->iWorkingA = ++now;
				break;
			case EDecrementA:
				this->iWorkingA = --now;
				break;
			case EIncrementB:
				this->iWorkingB = ++now;
				break;
			case EDecrementB:
				this->iWorkingB = --now;
				break;
		}

		// check whether sieving region height was reached
		if (iSievingRegion.height != SIEVING_REGION_NO_HEIGHT_BOUND) {
			long l;
			switch (this->iSievingDirection) {
				case EIncrementA:
					l = this->iWorkingA;
					break;
				case EDecrementA:
					l = -(this->iWorkingA);
					break;
				case EIncrementB:
					l = this->iWorkingB;
					break;
				case EDecrementB:
					l = -(this->iWorkingB);
					break;
			}
			if (l >= iSievingRegion.start+iSievingRegion.height) region_end = true;
		}
	
	} 
	while ((!enough) && (!region_end));

	if (this->iRSerialAccordingTime)
	{
		this->iTimes->End(ESTRSerial);
	}

	rc = EndRunningSerialization();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTRunSieve);
#endif
	if (Get_InfoMode())
	{
		PrintValues();
		if (Get_TimeMessage() > 0)
		{
			PrintEffectiveness();
		}
	}

	// The sieving ended before enough relations have been found.
	if (!enough) {
		if (region_end) {
			CLog::Log(50, "", "Sieving region was exhausted.");
			fprintf(stdout, "\n\tSieving region was exhausted.\n");
			return ConstRC::SievingRegionExhausted;
		}
		return ConstRC::NotEnoughSmooths;
	}

	return ConstRC::Ok;
}
/*!
	\short This method will perform sieving in a single line and readjust the data for the next line.

	This method takes care of the line sieving. It expects the factor base(s) to be already
	divided into middle and largish primes.

	The method will at first sieve the positive part of the sieving interval: [1,M], and
	then, separately, the negative part [-M,-1]. The value of 0 is ignored.

	The positive and negative parts will be sieved with use of "block sets". The "block set" 
	is a set of blocks big enough to cover the whole factor base (in case of NFS, to cover
	the bigger of the two factor bases). The sieving of each of the block sets is performed
	using the SieveBlock method.

	Before both the positive and the negative part of the sieving, a hashtable with largish
	primes is filled in. During the negative hashtable fill, the roots are also adjusted for
	the next sieving line.

*/
int CNFSClassicalSievingPhase::SieveLine()
{
	int rc;
	int k, i;
	int block_max;

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTLineSieving);
#endif

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstFillHT + i);
#endif
		// Fill hashtable
		rc = FillHashtable(this->iHashtables[i], this->iAdvFB[i].FirstLargishPrime, POSITIVE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstFillHT + i);
#endif
	}

	if (this->iWorkingIteration % this->iThresholdOptInterval == 0)
	{
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			this->iOptimizer[i].AnalyzeAndOptimize(this->iInitBlockValueChange[i]);
		}
	}

	// Positive sieving (from 1 to M)
	int bStart = 0;
	int bEnd = this->iPositiveBlocks;

#ifdef DEBUG_SIEVING_PHASE_COMPUTE_EXP_ALG_LOG
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iInitBlockValues[i] = 0;
	}
#endif

	for (k = bStart, i = 0; k < bEnd; k+= this->iBlockSetSize, i++) 
	{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		printf("\n Block set sieve: positive k = %d start.\n", k);
#endif
		block_max = this->iBlockSetSize;		
		if ((this->iPositiveBlocks - k) < block_max)
		{
			block_max = this->iPositiveBlocks - k;
			//printf("Block max lesser than set: %d.\n", block_max);
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTBlockSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_COMPUTE_EXP_ALG_LOG
		this->iBlockIndex = i;
#endif
		// The sieving itself
		rc = SieveBlock(i, 0, block_max, POSITIVE, (k == 0)); 
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTBlockSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		printf("\n Block set sieve: positive k = %d end.\n", k);
#endif
	} // end of for cycle
	
	// Negative sieving, from -M to -1.

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstFillHT + i);
#endif
		// Fill hashtable
		rc = FillHashtable(this->iHashtables[i], this->iAdvFB[i].FirstLargishPrime, NEGATIVE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstFillHT + i);
#endif
	}

	bStart = 0;
	bEnd = this->iNegativeBlocks;

	for (k = bStart, i = 0; k < bEnd; k += this->iBlockSetSize, i++) 
	{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		printf("\n Block set sieve: negative k = %d start.\n", k);
#endif
		block_max = this->iBlockSetSize;		
		if ((this->iNegativeBlocks - k) < block_max)
		{
			block_max = this->iNegativeBlocks - k;
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTBlockSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_COMPUTE_EXP_ALG_LOG
		this->iBlockIndex = i;
#endif
		rc = SieveBlock(i, 0, block_max, NEGATIVE, (k == 0)); // init negative iff k == 0
		if (rc != ConstRC::Ok)
		{
			return rc;
		}	

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTBlockSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		printf("\n Block set sieve: negative k = %d end.\n", k);
#endif
	} // end of for cycle


#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTRootUpdate);
#endif

	nfs_fb_type* running;
	nfs_fb_type* end; 

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		running = this->iAdvFB[i].FactorBase;
		end = this->iAdvFB[i].LastMiddlePrime;

		// Shift the middle-sized first algebraic ideals
		// Identical code is present in FillHashtable for largish prime ideals.
		do 
		{
			UpdateNfsRoot(running);	
			//if (running == end)
			//{
			//	break;
			//}
			//running++;
		} 
		while (running++ != end); // Will be broken by internal check
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTRootUpdate);
#endif

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTLineSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	Utils::Wait();
#endif

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::ComputeExpectedIntegralLog(int aIndex)
{
	mpz_mul_si(this->iAuxMpz1, this->iM, this->iWorkingB);
	mpz_add_ui(this->iAuxMpz1, this->iAuxMpz1, this->iWorkingA);
	int rough = mpz_sizeinbase(this->iAuxMpz1, ConstSPhase::LOGARITHM_BASE);
	rough -= this->iNumberFields[aIndex].ExpectedErrorFactor;
	//printf("NFS: ComputeExpectedIntegralLog: rough = %d.\n",rough);
	if (this->iVariationsInfo[aIndex].Type != EDoNotUseVariations) 
	{
		// Space for maximal partial factor (integral)
		rough -= this->iVariationsInfo[aIndex].SizeOfMaxRemainder;
	} 
	else 
	{
		//printf("Not using variations");
	}
	if ((this->iSievingDirection == EIncrementA) || (this->iSievingDirection == EDecrementA)) 
	{
		// We will add the logarithms of primes p which divide current a
		int max = this->iNumberFields[aIndex].RZeros_Indices.size();
		long index;
		main_sieving_type prime;
		main_sieving_type a = this->iWorkingA;
		for (int i = 0; i < max; i++) 
		{
			index = this->iNumberFields[aIndex].RZeros_Indices[i];
			prime = this->iFBs[aIndex].FactorBase[index].prime;
			if (!(a % prime)) 
			{
				// A+MB is divisible by p for all B
				rough += this->iFBs[aIndex].FactorBase[index].log;
				//printf("All integers will be divisible by %d, adding log %d to the sieve...\n",prime,this->iFBs[aIndex].FactorBase[index].log);
			}
			// else A+MB is never divisible by B
		}
	}

	//printf("Min integral log is %d. Prefill memset to 0x%x.\n", rough, 0x80 - rough);
	return rough;
}
int CNFSClassicalSievingPhase::ComputeExpectedAlgebraicLog(int aIndex)
{
#ifdef DEBUG_SIEVING_PHASE_COMPUTE_EXP_ALG_LOG
	return SIEVING_PHASE_MASK;
#endif

	int i, j;
 	int rough = 0;
	int step = this->iSievingRegion.width >> 6;
	long index  = 0;
	Polynomial * poly = this->iNumberFields[aIndex].SievingPoly;
	int deg =  poly->GetDegree();
	mpz_t a, result;
	mpz_init(a);
	mpz_init(result);

	for (mpz_set_si(a, -this->iSievingRegion.width); 
		mpz_cmp_si(a, this->iSievingRegion.width) < 0; 
		mpz_add_ui(a, a, step))
	{
		mpz_set_ui(result, 0);
		index++;

		if (mpz_cmp_ui(a, 0) == 0)
		{
			mpz_ui_pow_ui(result, this->iWorkingB, deg);
			mpz_mul(result, result, *poly->Get2(0));
		}
		else
		{
			mpz_pow_ui(this->iAuxMpz1, a, deg);

			j = 0;
			for (i = deg; i >= 0; i--)
			{	
				mpz_mul(this->iAuxMpz2, this->iAuxMpz1, *poly->Get2(i));
				if (!((j++) % 2)) 
				{
					mpz_add(result, result, this->iAuxMpz2);
				} 
				else 
				{
					mpz_sub(result, result, this->iAuxMpz2);
				}

				if (i) 
				{
					mpz_divexact(this->iAuxMpz1, this->iAuxMpz1, a);
					mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, this->iWorkingB);
				}
			}
		}

		rough += mpz_sizeinbase(result, ConstSPhase::LOGARITHM_BASE);
	}

	rough = rough / index;
	mpz_clear(a);
	mpz_clear(result);

	if (this->iVariationsInfo[aIndex].Type != EDoNotUseVariations) 
	{
		// Space for maximal partial factor (algebraic)
		rough -= this->iVariationsInfo[aIndex].SizeOfMaxRemainder;
	}

	if ((this->iSievingDirection == EIncrementA) || (this->iSievingDirection == EDecrementA)) 
	{
		// Pricteme hodnotu tech logaritmu p, ktere deli soucasne this->a
		int max = this->iNumberFields[aIndex].RZeros_Indices.size();
		main_sieving_type prime;
		main_sieving_type a = this->iWorkingA;
		for (i = 0; i < max; i++) 
		{
			index = this->iNumberFields[aIndex].RZeros_Indices[i];
			prime = this->iFBs[aIndex].FactorBase[index].prime;
			//printf("Refill Sieving: i = %d, cpZerosMax = %d, index = %d, prime = %d.\n",i,max+1,index,prime);			
			if (!(a % prime)) 
			{
				// A+C_PB is divisible by p for all B
				rough += this->iFBs[aIndex].FactorBase[index].log;
				//printf("All prime ideals will be divisible by %d, adding log %d to the sieve...\n",prime,this->iFBs[aIndex].FactorBase[index].log);
			}
			// else A+C_PB is never divisible by B
		}
	}

	rough -= this->iNumberFields[aIndex].ExpectedErrorFactor;

	//printf("Min algebraic log is %d. Prefill memset to 0x%x.\n", rough, 0x80-rough);
 	return rough;
}

int CNFSClassicalSievingPhase::ComputeQuickExpectedAlgLog(int aBlockIndex, int aInnerIndex, int aSign, 
			int aInitValue, Polynomial * aSievingPoly, CVariationsInfo & aVarInfo)
{
	int blockSize = Get_BlockSize();
	long middle = aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize + (blockSize >> 1);

	int i, j;
	int rough;
	int deg = aSievingPoly->GetDegree();
	mpz_t a, result;
	mpz_init_set_si(a, aSign * middle);
	mpz_init(result);

	if (mpz_cmp_ui(a, 0) == 0)
	{
		mpz_ui_pow_ui(result, this->iWorkingB, deg);
		mpz_mul(result, result, *aSievingPoly->Get2(0));
	}
	else
	{
		mpz_pow_ui(this->iAuxMpz1, a, deg);

		j = 0;
		for (i = deg; i >= 0; i--)
		{	
			mpz_mul(this->iAuxMpz2, this->iAuxMpz1, *aSievingPoly->Get2(i));
			if (!((j++) % 2)) 
			{
				mpz_add(result, result, this->iAuxMpz2);
			} 
			else 
			{
				mpz_sub(result, result, this->iAuxMpz2);
			}

			if (i) 
			{
				mpz_divexact(this->iAuxMpz1, this->iAuxMpz1, a);
				mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, this->iWorkingB);
			}
		}
	}

	rough = mpz_sizeinbase(result, ConstSPhase::LOGARITHM_BASE);

	mpz_clear(a);
	mpz_clear(result);

	if (aVarInfo.Type != EDoNotUseVariations) 
	{
		// Space for maximal partial factor (algebraic)
		rough -= aVarInfo.SizeOfMaxRemainder;
	}

	rough -= ConstSPhase::ALGEBRAIC_ERROR_FACTOR;

	if (aInitValue != 0 && aBlockIndex > 1)
	{
		// not too big change
		if (SIEVING_PHASE_MASK - rough - aInitValue > 2)
		{
			rough =  SIEVING_PHASE_MASK - aInitValue - 2;
		}
		else if (SIEVING_PHASE_MASK - rough - aInitValue < -2)
		{
			rough = SIEVING_PHASE_MASK - aInitValue + 2;
		}
	}

 	return rough;
}
int CNFSClassicalSievingPhase::CheckEnough() 
{
	return this->iRequestedRelations > (this->iStats->Get_Total(ESSFF) + this->iStats->Get_Total(ESSEdges) 
		+ this->iStats->Get_Total(ESSComponents) - this->iStats->Get_Total(ESSVertices)) 
		? ConstRC::NotEnoughSmooths : ConstRC::EnoughSmooths;
}
/*!
 * \brief One-time method for allocation of the factor bases.
 *
 * There are two algebraic factor bases used during the NFS or one integral and one algebraic.
 * We must allocate some space for them.
 * 
 * In both factor bases, we allocate some extra space. It is caused by the fact that
 * we want to sieve with use of a zero-byte block at the end, which will serve as a 
 * limit value during sieving ("read until you meet a zero"); we want the total 
 * allocated block to be multiple of 4 long.
 */
int CNFSClassicalSievingPhase::InitFactorBases() 
{
	Set_InnerState(ESPhaseInitFB);

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tInit integral domain ... ");
	}

	int i;
	int rc;

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = this->iNumberFields[i].InitIntegralDomain(this->iFBs[i].UpperFBBound);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tInit factor bases ... ");
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTInitFB);
#endif

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = this->iFBs[i].InitFB();
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = this->iFBs[i].FillFB(this->iNumberFields[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iNumberFields[i].FillSpecialPrimes();
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTInitFB);
#endif

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\t\tSerializing factor bases ... ");
	}

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// serialize factor bases 
	// and set FB filenames for other phases
	string auxStr;
	CBaseParameters paramFB;
	AFactorAlgParameters * param = Get_Parameters();
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		paramFB.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iFBs[i].FBFullFileName);
		rc = this->iFBs[i].Serialize(paramFB);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		auxStr = Utils::ConvertToString(i + 1);
		param->SetParameterForOthers(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, this->iFBs[i].FBFullFileName, Get_PhaseNbr());
	}
#endif

	// write some info
	if (Get_InfoMode())
	{
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			this->iNumberFields[i].PrintInfo(i, this->iFBs[i].FactorBase);
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		printf("\n		Init FB took: %12.4f sec", this->iTimes->Get_TotalTime(ESTInitFB));
#endif
	}

	bool useFreeRelations = ConstSPhase::USE_FREE_RELATIONS;
	Get_Parameters()->GetBoolPhaseParameter(ConstPN::USE_FREE_RELATIONS, ConstSPhase::USE_FREE_RELATIONS, useFreeRelations, Get_PhaseNbr());

	if (useFreeRelations)
	{
		rc = FindFreeRelations();
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	if (this->iVariationsInfo[0].Type == ELargePrimeV &&
		this->iVariationsInfo[1].Type == ELargePrimeV)
	{
		// Now, we prepare structures for counting cycles in the LPV	
		rc = PrepareCycleTable();
	}

	return rc;
}
/*!
	This method find all free relations and save them to the smooth relation file.
	We suppose that this file doesn't exist - it shouldn't, so we delete it if there is some one.

	The free relations is a factorization of the ideal (p), where p is prime, for which 
	f_{1} mod p and f_{2} mod p fully factorize to linear factors.

	For finding we use FBs, where we are looking for prime ideals with same prime. 
	If we find d (degree of sieving poly) of them, than we know that f mod p is fully factorized (in both bases).
*/
int CNFSClassicalSievingPhase::FindFreeRelations()
{
	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tFinding free relations ... ");
	}

	int k;
	int degrees[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	fb_element * fb[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	fb_element * end[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	for (k = 0; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
	{
		degrees[k] = this->iNumberFields[k].SievingPoly->GetDegree();
		fb[k] = this->iFBs[k].FactorBase;
		end[k] = this->iFBs[k].FactorBase + this->iFBs[k].FBMaxIndex;
	}

	CNFSRelation relation(EDoNotUseVariations);
	
	int rc;
	int i;
	int count = 0;
	main_sieving_type prime = 0;
	bool good = false;
	int lastWrittenIndex = -1;
	unsigned int flag = 0;

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// prepare file - file will be deleted if there is some one
	FILE * smoothsFILE = fopen(this->iSmoothRelFullFileName.c_str(), FILE_MODE_WRITE);
	if (smoothsFILE == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","FindFreeRelations",0,0,"Could not open file with smooth relations.");
		return ConstRC::CouldNotOpenFile;
	}
#endif

	while (fb[0] < end[0])
	{
		prime = fb[0]->prime;
		count = 1;
		fb[0]++;
		while (fb[0]->prime == prime && prime != fb[0]->root)
		{
			fb[0]++;
			count++;
		}

		// we find factorization of ideal (p) in FB1
		if (count == degrees[0])
		{
			good = true;
			for (k = 1; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
			{
				// find position before prime ideals above p
				while (fb[k] < end[k] && fb[k]->prime < prime)
				{
					fb[k]++;
				}
				count = 0;
				// count number of prime ideals above p in FBk
				while (fb[k]->prime == prime && prime != fb[k]->root)
				{
					fb[k]++;
					count++;
				}
				if (count != degrees[k])
				{
					good = false;
					break;
				}
			}

			if (good)
			{
				// setup relation - we want just ideal (p)
				relation.Reset();
				// we need to set 0 relation count - we never use leading coeff
				relation.ZeroizeNumberOfRelations();
				relation.A = prime;
				relation.B = 0;

				for (k = 0; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
				{
					fb[k]--; // go back
					relation.Get_Parts(k)->Poly->Set(0, prime);
					mpz_ui_pow_ui(relation.Get_Parts(k)->Norm, (unsigned long)prime, degrees[k]);
					mpz_set_ui(relation.Get_Parts(k)->NormRemaining, 1);
					if (fb[k]->flags & ConstEF::SPECIAL_PRIME)
					{
						// we note that this is special prime and we save its residual which has to be 1
						flag = ConstEF::SPECIAL_PRIME | (unsigned int)1;
						// TODO check
						//if (fb[k]->flags & ConstEF::DIVIDES_L)
						//{
						//	// if this prime divides the leading coeff than we save exact valuation which is still 1
						//	flag |= ConstEF::DIVIDES_L | ((unsigned int)1 << ConstEF::F_VALUATION_SHIFT);
						//}
					}
					else
					{
						flag = 0;
					}

					// set prime ideals divisors - we go through found prime ideals with same prime
					lastWrittenIndex = -1;
					for (i = - degrees[k] + 1; i <= 0; i++)
					{
						rc = relation.Get_Parts(k)->Main.AddDivisor(prime, 1, fb[k][i].root, flag, lastWrittenIndex);
						if (rc != ConstRC::Ok)
						{
#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
							fclose(smoothsFILE);
							smoothsFILE = NULL;
#endif
							return rc;
						}
					}

					relation.Type |= (CRelationPart::SMOOTH << (ARelation::PART_TYPE_SHIFT * k));
				}

				relation.Type |= RELATION_FREE;
				relation.Index = this->iStats->Get_Total(ESSFF);
				this->iStats->Increase(ESSFF);

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
				this->iStats->Increase(ESSTested);
				this->iStats->Increase(ESSAccepted);
#endif

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
				rc = relation.PrintToFile(smoothsFILE);
				if (rc != ConstRC::Ok)
				{
					fclose(smoothsFILE);
					smoothsFILE = NULL;
					return rc;
				}
#endif
			}
		}
	}

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	fclose(smoothsFILE);
	smoothsFILE = NULL;
#endif

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::PrepareCycleTable()
{
	if (this->iVariationsInfo[0].Type == ELargePrimeV &&
		this->iVariationsInfo[1].Type == ELargePrimeV) 
	{	
		if (CNFSRelation::DEFAULT_NFS_PARTS_COUNT != ConstSPhase::ELEMENT_MASK_COUNT)
		{
			CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","PrepareCycleTable",0,0,"Relation size mismatch.");
			return ConstRC::GeneralError;
		}

		if (Get_InfoMode())
		{
			Utils::NormalMessage("   Preparing cycle table ... ");
		}			
		
		this->iLPHashtable = new CLPType1Hashtable();
		this->iLPHashtable->SetupHashtableFB(max(this->iFBs[0].FBMaxIndex, this->iFBs[1].FBMaxIndex));

		// Special case: PRIME=1, root of its own component.
		this->iStats->Increase(ESSVertices);
		this->iStats->Increase(ESSComponents);
		//this->vertices = 1;
		//this->components = 1;
		//this->edges = 0;

		this->iIntermediateRoots.reserve(ConstSPhase::INTERMEDIATE_ROOTS_COUNT);
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::AddToCycleCountStructures(CNFSRelation* aRelation) 
{
	int oldCycles = IndependentCyclesCount();
	int rc;
	long roots[ConstSPhase::ELEMENT_MASK_COUNT];
	long index[ConstSPhase::ELEMENT_MASK_COUNT];
	bool present = false;
	int  searched;
	int  root_fingerprint;

	bool enlarged = false;
	bool recalculate = false;

	for (int j = 0; j < ConstSPhase::ELEMENT_MASK_COUNT*2; j++)
	{
		int i = j % ConstSPhase::ELEMENT_MASK_COUNT;
		if (aRelation->Type & (CRelationPart::SMOOTH << (ARelation::PART_TYPE_SHIFT * i)))
		{
			searched = 1;	
			root_fingerprint = 0;
		} 
		else 
		{
			searched = aRelation->Get_Parts(i)->Large.Elements[0].prime;
			root_fingerprint = aRelation->Get_Parts(i)->Large.Elements[0].root;
			root_fingerprint |= ConstSPhase::ELEMENT_MASK[i];
		}

		//printf("calling FindPrimeInHashtable: searched = %d, present = %d.\n", searched, present);
		rc = this->iLPHashtable->FindPrimeInHashtable(searched, root_fingerprint, present, index[i], enlarged);
		if (enlarged) recalculate = true; 
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		if (!present) 
		{
			// This prime has not been in hashtable yet.					
			//printf("   Pridavame do hashtable prvocislo %d na index %d \n", searched, index[i]);
			this->iLPHashtable->Hashtable[index[i]].SetValue(searched);
			this->iLPHashtable->Hashtable[index[i]].SetRootFingerprint(root_fingerprint);
			this->iLPHashtable->Hashtable[index[i]].SetAncestor(index[i]);
			this->iLPHashtable->Added();
			this->iStats->Increase(ESSVertices);
			this->iStats->Increase(ESSComponents);
			//this->vertices++;
			//this->components++;
		} 
		else 
		{
	   		//printf("   Prvocislo %d jiz bylo v tabulce na indexu %d \n", searched, d_1);
			if (searched != 1) 
			{
	  			//printf("   Opakovane jsme zaznamenali netrivialni prvocislo %d.\n", searched, d_1);
			}
		}

		if (ConstSPhase::ELEMENT_MASK_COUNT-1==j && !recalculate) break;
	}

	for (int i = 0; i < ConstSPhase::ELEMENT_MASK_COUNT; i++)
	{
		// Finding roots of components
		roots[i] = this->iLPHashtable->FindAndResetRoot(index[i], this->iIntermediateRoots);
	};
	this->iStats->Increase(ESSEdges);
	//this->edges++;

	//printf("   Koreny komponent: %d a %d.\n------------\n", roots[0], roots[1]);
	if (roots[0] != roots[1]) 
	{			
		// The component with the larger prime at root is joined to the other one
		if (this->iLPHashtable->Hashtable[roots[0]].GetValue() > this->iLPHashtable->Hashtable[roots[1]].GetValue()) 
		{
			this->iLPHashtable->Hashtable[roots[0]].SetAncestor(roots[1]);	
		} 
		else 
		{
			this->iLPHashtable->Hashtable[roots[1]].SetAncestor(roots[0]);	
		}
		this->iStats->Decrease(ESSComponents);
		//this->components--;
		//printf("Koreny komponent %d, %d, zmerge-ovany.\n", roots[0], roots[1]);
	}

	if (IndependentCyclesCount() != oldCycles) 
	{
		//printf("Pridani teto relace zpusobilo vznik cyklu: \n");
		aRelation->Type |= RELATION_FINISHING_CYCLE;
	}

	return ConstRC::Ok;
}


int CNFSClassicalSievingPhase::IndependentCyclesCount()
{
	return this->iStats->Get_Total(ESSEdges) + this->iStats->Get_Total(ESSComponents) - this->iStats->Get_Total(ESSVertices);
}
/*!
	This method will set up the factor bases
	for the optimized line siever. They are very sensitive to special
	values of root, and such special cases must be treated separately.
*/
int CNFSClassicalSievingPhase::InitAdvancedFactorBases()
{
	Set_InnerState(ESPhaseInitAdvFB);

	if (Get_InfoMode())
	{
		Utils::NormalMessage("\tInitializing the advanced FB ...");
	}

	// we serialize this phase
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();
	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int rc = Serialize(param);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int blockSize = Get_BlockSize();
	int i;
	// init a fill advance factor base - this field is used for sieving
	// we have to reflect updated value of A or B
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = CAdvFBHelper::InitAndFillAdvFB(this->iAdvFB[i].FactorBase, this->iFBs[i], this->iSievingDirection,
				Get_A(), Get_B(), *this->iNumberFields[i].SievingPoly->Leading2(), this->iM);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iAdvFB[i].Count = this->iFBs[i].FBMaxIndex + 1;
		rc = this->iAdvFB[i].Init(blockSize);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		WriteDumpFBFile(this->iAdvFB[i].FactorBase, (this->iAdvFB[i].Count) * sizeof(nfs_fb_type));
	}
#endif

	return ConstRC::Ok;
}

int CNFSClassicalSievingPhase::PrepareForSieving()
{
	// We calculate the maximal partial factor (in case of Single and Double LPV).
	int maxx = 0;
	int i;

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iVariationsInfo[i].Prepare(this->iFBs[i].UpperFBBound, 
			this->iPartialFactorMultiplier[i], ConstSPhase::LOGARITHM_BASE);
		maxx = max(maxx, (int)this->iVariationsInfo[i].Type);
	}

	int rc;

	this->iWorkingA = Get_A();
	this->iWorkingB = Get_B();

	this->iRelation = new CNFSRelation(maxx);

	this->iRelationBuffer = new vector<CNFSRelation*>;
	this->iRelationBuffer->reserve(ConstSPhase::RELATION_BUFFER_RESERVE);
	this->iRelationWorkingBuffer = new vector<CNFSRelation*>;
	this->iRelationWorkingBuffer->reserve(ConstSPhase::RELATION_BUFFER_RESERVE);

	int blockSize = Get_BlockSize();
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iSievingBlocks[i] = (log_type*) Utils::AllocationMachine((void**) &this->iSievingBlocks[i],(blockSize)*sizeof(log_type));
		if (this->iSievingBlocks[i] == NULL)
		{
			CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","PrepareForSieving",0,blockSize*sizeof(log_type),"Could not allocate space for internal sieving block");
			return ConstRC::NotEnoughMemory;
		} 
		else 
		{
			if (Get_InfoMode())
			{
				cout << "\n\tSieving block nbr " << i << " is allocated to " << (blockSize) * sizeof(log_type) << " bytes.";
			}
		}
		memset(this->iSievingBlocks[i], 0, blockSize * sizeof(log_type));

		if (this->iInitBlockValueReference[i] == 0)
		{
			// first setup of the init block value
			if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
			{
				this->iInitBlockValueReference[i] = ComputeExpectedAlgebraicLog(i);
			}
			else
			{
				this->iInitBlockValueReference[i] = ComputeExpectedIntegralLog(i);
			}
			
			this->iInitBlockValueChange[i] = 0;
			this->iInitBlockValues[i] = SIEVING_PHASE_MASK - this->iInitBlockValueReference[i];
		}
		else
		{
			// we deserialize some values
			this->iInitBlockValues[i] = SIEVING_PHASE_MASK - this->iInitBlockValueReference[i];
		}
#ifdef DEBUG_SIEVING_PHASE_COMPUTE_EXP_ALG_LOG
		if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
		{
			// max limits for algebraic part
			this->iOptimizer[i].SetChangeLimits(SIEVING_PHASE_MASK >> 1);
		}
		else
		{
			// integral part is not computed for each block so we can setup limits
			this->iOptimizer[i].SetChangeLimits(this->iInitBlockValueReference[i]);
		}
#else
		this->iOptimizer[i].SetChangeLimits(this->iInitBlockValueReference[i]);
#endif
	}

	if (this->iSievingRegion.width <= 0) 
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","PrepareForSieving",this->iSievingRegion.width,0,"Value of line upper bound must be positive!");
		return ConstRC::InconsistentState;
	}
	if (this->iSievingRegion.width > ConstSPhase::MAX_LINE_SIEVE_HALFLINE)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","PrepareForSieving",this->iSievingRegion.width,ConstSPhase::MAX_LINE_SIEVE_HALFLINE,"Value of line upper bound too big - must not exceed a particular value!");
		return ConstRC::InconsistentState;
	}
	this->iLineUpperBound = this->iSievingRegion.width;
	this->iLineLowerBound = - this->iSievingRegion.width;

	long upper = 0;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		upper = CIntegerClass::Max(upper, this->iFBs[i].UpperFBBound);
	}
	if (upper <= 0)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","PrepareForSieving",upper,0,"Value of factor base upper bound must be positive.");
		return ConstRC::InconsistentState;
	}
	if (upper > ConstSPhase::MAX_FACTOR_BASE_UPPER_BOUND)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","PrepareForSieving",upper,ConstSPhase::MAX_FACTOR_BASE_UPPER_BOUND,"Value of factor base upper bound too big - must not exceed a particular value!");
		return ConstRC::InconsistentState;
	}

	this->iPositiveBlocks = this->iLineUpperBound / blockSize; 
	if (this->iLineUpperBound % blockSize)
	{
		this->iPositiveBlocks++;
	}
	this->iNegativeBlocks = (-1) * this->iLineLowerBound / blockSize;
	if (((-1) * this->iLineLowerBound) % blockSize) 
	{
		this->iNegativeBlocks++;
	}
	this->iBlockSetSize = upper / blockSize;
	if (upper % blockSize)
	{
		this->iBlockSetSize++;
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		// iBlockSetSize is number of blocks in block set - it is also the count of hashtables
		rc = this->iHashtables[i].SetupHashtables(this->iBlockSetSize);
		if (rc != ConstRC::Ok)	
		{
			return rc;
		}
	}

	return rc;
}
int CNFSClassicalSievingPhase::SaveResult()
{
	Set_InnerState(ESPhaseSavingResult);

	AFactorAlgParameters * param = Get_Parameters();

	string auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSFF));
	param->SetParameterForOthers(ConstPN::SMOOTH_REL_COUNT, auxStr, Get_PhaseNbr());

	auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSEdges));
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_GRAPH_EDGES, auxStr, Get_PhaseNbr());

	auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSComponents));
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_GRAPH_COMPONENTS, auxStr, Get_PhaseNbr());

	auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSVertices));
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_GRAPH_VERTICES, auxStr, Get_PhaseNbr());

	param->SetParameterForOthers(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, this->iSmoothRelFullFileName, Get_PhaseNbr());
	
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_FULL_FILENAME, this->iPartialRelFullFileName, Get_PhaseNbr());

	// we serialize this phase
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();
	CBaseParameters paramSerial;
	paramSerial.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int rc = Serialize(paramSerial);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	Set_InnerState(ESPhaseFinished);

	return ConstRC::Ok;
}
/*!
	This is an auxiliary method for printing out the intermediate results
	of the sieving step.

*/
void CNFSClassicalSievingPhase::PrintHeader() 
{
	printf("\n\n");
	printf(" __________________________________________________________________________________");
	printf("\n");
	printf("|   LINE  |  SMOOTH |  PARTS  |  CYCLES |   EDGES |  COMPON |  VERTEX |  PERCENT   |");
	printf("\n");
}
/*!
	This is an auxiliary method for printing out the intermediate results
	of the sieving step.

*/
void CNFSClassicalSievingPhase::PrintValues() 
{
	cout << "\r|";
	cout << setw(7) <<  Get_WorkingIteration() << " | ";
	cout << setw(7) <<  this->iStats->Get_Total(ESSFF) << " | ";
	cout << setw(7) <<  this->iStats->Get_Total(ESSFP) + this->iStats->Get_Total(ESSPF) + this->iStats->Get_Total(ESSPP) << " | ";	
	int independentCycles = this->iStats->Get_Total(ESSEdges) + 
		this->iStats->Get_Total(ESSComponents) - this->iStats->Get_Total(ESSVertices);			
	cout << setw(7) <<  independentCycles << " | ";					
	cout << setw(7) <<  this->iStats->Get_Total(ESSEdges) << " | ";				
	cout << setw(7) <<  this->iStats->Get_Total(ESSComponents) << " | ";				
	cout << setw(7) <<  this->iStats->Get_Total(ESSVertices) << " | ";					
	int collected = this->iStats->Get_Total(ESSFF) + independentCycles;
	float pert = (float) collected * (float) 100;
	float ratio = (float) pert / (float) this->iRequestedRelations;
	printf("   %6.2f %% |",ratio);
	fflush(stdout);	
}
void CNFSClassicalSievingPhase::PrintEffectiveness()
{
	printf("\nSo far, %d iterations have been run.\n", Get_WorkingIteration());
	float seconds = this->iTimes->Get_TotalTime(ESTLineSieving);
	float abs;
	int i;
	float secondsBlock;
	abs = this->iTimes->Get_TotalTime(ESTRunSieve); 
		printf("\tOverall time spent in         RunSieve is %12.4f sec \n", abs);
	if (abs > 0)
	{
		printf("\t\tFrom this in             SieveLine is %12.4f sec (%6.4f %%)\n", seconds, 100 * (seconds / abs));
	}		
	if (seconds > 0)
	{
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			abs = this->iTimes->Get_TotalTime(ESTFirstFillHT + i); 
			printf("\t\t\tFrom this in       Fill HT #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / seconds));
		}
		secondsBlock = this->iTimes->Get_TotalTime(ESTBlockSieving);
		printf("\t\t\tFrom this in        SieveBlock is %12.4f sec (%6.4f %%)\n", secondsBlock, 100 * (secondsBlock / seconds));
		if (secondsBlock > 0)
		{
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				abs = this->iTimes->Get_TotalTime(ESTFirstSieving + i); 
				printf("\t\t\t\tFrom this in Sieving #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
			}
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				abs = this->iTimes->Get_TotalTime(ESTFirstFactorization + i); 
				printf("\t\t\t\tFrom this in Factor. #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
			}
			abs = this->iTimes->Get_TotalTime(ESTPartialFactorization); 
			printf("\t\t\t\tFrom this in Part. Factor. is %12.4f sec (%6.4f %%)\n", abs, 100 * (abs / secondsBlock));
		}
		abs = this->iTimes->Get_TotalTime(ESTRootUpdate); 
		printf("\t\t\tFrom this in        RootUpdate is %12.4f sec (%6.4f %%)\n", abs, 100 * (abs / seconds));
	}	 

	printf("\nEFFECTIVENESS: \n");
	printf("\n\tGlobal: \n");
	int succ = 0;
	int tested = this->iStats->Get_Total(ESSTested);
	printf("     Total tested      :  %10d.\n", tested);
	int value = this->iStats->Get_Total(ESSFF);
	printf("     Total full-full   :  %10d.\n", value);
	succ += value;
	value = this->iStats->Get_Total(ESSFP);
	printf("     Total full-part   :  %10d.\n", value);
	succ += value;
	value = this->iStats->Get_Total(ESSPF);
	printf("     Total part-full   :  %10d.\n", value);
	succ += value;
	value = this->iStats->Get_Total(ESSPP);
	printf("     Total part-part   :  %10d.\n", value);
	succ += value;
	if (tested > 0)
	{
		printf("     Total successes   :  %10d (%6.4f %%).\n", succ, (((double) succ)*(double) 100)/(double) tested);
	}
	else
	{
		printf("     Total successes could not be determined - 0 relations tested.\n");
	}
	int rejected = this->iStats->Get_Total(ESSRejected);
	int notcoprime = this->iStats->Get_Total(ESSNotCoprime);
	if (tested > 0)
	{
		printf("     Total rejected    :  %10d (%6.4f %%).\n", rejected, (((double) rejected)*(double) 100)/(double) tested);
		printf("                GCD > 1:  %10d (%6.4f %%).\n", notcoprime, (((double) notcoprime)*(double) 100)/(double) tested);
		printf("                Expensi:  %10d (%6.4f %%).\n", (rejected - notcoprime), (((double) rejected - notcoprime)*(double) 100)/(double) tested);
	}
	printf("\n");
	seconds = this->iTimes->Get_IntermediateTime(ESTLineSieving);
	tested = this->iStats->Get_Intermediate(ESSTested);
	printf("\n  Local (%6.2f sec): \n", seconds);
	printf("     Local tested      :  %10d (%6.2f per sec).\n", tested, (seconds > 0) ? (double) tested / seconds : 0);
	succ = 0;
	value = this->iStats->Get_Intermediate(ESSFF);
	printf("     Local full-full   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
	succ += value;
	value = this->iStats->Get_Intermediate(ESSFP);
	printf("     Local full-part   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
	succ += value;
	value = this->iStats->Get_Intermediate(ESSPF);
	printf("     Local part-full   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
	succ += value;
	value = this->iStats->Get_Intermediate(ESSPP);
	printf("     Local part-part   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
	succ += value;
	if (tested > 0 && seconds > 0)
	{
		printf("     Local successes   :  %10d (%6.4f %%).\n", succ, (((double) succ)*(double) 100)/(double) tested);
	}
	else
	{
		printf("     Local successes could not be determined - 0 relations tested or 0 seconds.\n");
	}
	rejected = this->iStats->Get_Intermediate(ESSRejected);
	notcoprime = this->iStats->Get_Intermediate(ESSNotCoprime);
	if (tested > 0 && seconds > 0)
	{
		printf("     Local rejected    :  %10d (%6.4f %%).\n", rejected, (((double) rejected) * (double) 100)/(double) tested);
		printf("                GCD > 1:  %10d (%6.4f %%).\n", notcoprime, (((double) notcoprime)*(double) 100)/(double) tested);
		printf("                Expensi:  %10d (%6.4f %%).\n", (rejected - notcoprime), (((double) rejected - notcoprime) * (double) 100)/(double) tested);
	}
	printf("\n");

	if (tested - notcoprime > 0) 
	{
		tested -= notcoprime;
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			printf("     Avg size - Part #%d:  %4d (bits), %4d (dec)\n", i + 1, this->iStats->Get_Intermediate(ESSAverageSizeIP_log2 + i) / tested, this->iStats->Get_Intermediate(ESSAverageSizeIP_log10 + i) / tested);
		}
	} 
	else 
	{
		printf("     Avgs sizes could not be determined - 0 relations tested\n");
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		printf("     Part #%d: Expected value:  %4d; Change: %4d (base: %d)\n", i + 1, SIEVING_PHASE_MASK - this->iInitBlockValues[i], this->iInitBlockValueChange[i], ConstSPhase::LOGARITHM_BASE);
	}
	
	double totalEntries = 2 * (double)this->iSievingRegion.width * Get_TimeMessage();	
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		seconds = this->iTimes->Get_IntermediateTime(ESTFirstSieving + i) * 1000;
		printf("     Sieving #%d speed :  %8.2f * 10^3 per sec\n", i + 1, totalEntries / seconds);
	}

	fflush(stdout);	
}
int CNFSClassicalSievingPhase::TestUpdate(const update_t& aUpdate)
{
	nfs_fb_type running;					
	running.p = aUpdate.p;
	running.r = aUpdate.r;
	running.root_1 = aUpdate.h;
	running.log_p = aUpdate.logp;
	running.flags = 0; // Largish primes do not usually have flags
	
	// TODO jak dalsi parametry??
	//TestNfsAlgebraic(running, aUpdate.h);
	
	return ConstRC::Ok;
}
bool CNFSClassicalSievingPhase::TestNfsAlgebraic(nfs_fb_type aInfo, main_sieving_type aUpdateIndex, int aBlockIndex, int aInnerIndex, int aSign, mpz_t aLeadingCoeff)
{
	bool result = FALSE;
	main_sieving_type plus = 0;
	int blockSize = Get_BlockSize();
	main_sieving_type currentArgument = aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize;	

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
	if (aInfo.p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME) 
	{
		printf("\nTest NFS algebraic: Printing for control ...\n");
		PrintNfsElementInfo(aInfo);
		printf("\nEnd of printing for control ...\n");
	}
#endif

	if (aUpdateIndex != NO_UPDATE_INDEX) 
	{
		currentArgument += aUpdateIndex;
	} 
	else 
	{
		currentArgument += aInfo.root_1;
	}
	if (this->iNfsCompressEvenIntervals && this->iIsCompressedNow)
	{
		currentArgument <<= 1;
	}
	currentArgument += 1;
	currentArgument *= aSign;

	if ((this->iSievingDirection == EIncrementB) || (this->iSievingDirection == EDecrementB)) 
	{
		mpz_set_si(this->iAuxMpz1, this->iWorkingB);
#ifdef DEBUG_SIEVING_PHASE_NFS_ALGEBRAIC_TEST
		printf("LineSiever: Test Algebraic. B = %d.\n",this->iWorkingB);
#endif
		mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, aInfo.r);
#ifdef DEBUG_SIEVING_PHASE_NFS_ALGEBRAIC_TEST
		printf("LineSiever: Test Algebraic. c_p = %d.\n",aInfo.r);		
#endif
		plus = currentArgument;
		mpz_set_si(this->iAuxMpz2,plus);
		mpz_mul(this->iAuxMpz2,this->iAuxMpz2,aLeadingCoeff);
		mpz_mod_ui(this->iAuxMpz2,this->iAuxMpz2,aInfo.p);
		plus = mpz_get_si(this->iAuxMpz2);
		if (plus < 0) plus += aInfo.p;
	}
	if ((this->iSievingDirection == EIncrementA) || (this->iSievingDirection == EDecrementA)) 
	{
		mpz_set_si(this->iAuxMpz1, currentArgument);
		mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, aInfo.r);
#ifdef DEBUG_SIEVING_PHASE_NFS_ALGEBRAIC_TEST
		printf("LineSiever: Test Algebraic. c_p = %d.\n",aInfo.r);
#endif
		plus = this->iWorkingA;
		mpz_set_si(this->iAuxMpz2,plus);
		mpz_mul(this->iAuxMpz2,this->iAuxMpz2,aLeadingCoeff);
		mpz_mod_ui(this->iAuxMpz2,this->iAuxMpz2,aInfo.p);
		plus = mpz_get_si(this->iAuxMpz2);
		if (plus < 0) plus += aInfo.p;
#ifdef DEBUG_SIEVING_PHASE_NFS_ALGEBRAIC_TEST		
		printf("LineSiever: Test Algebraic. LA mod p = %d.\n",plus);
#endif
	}
	if ((!(aInfo.flags & ConstEF::SPECIAL_PRIME_MINI_FLAG)) && (!(aInfo.flags & ConstEF::DIVIDES_ROOT_MINI_FLAG))) 
	{
		int res1 = mpz_mod_ui(this->iAuxMpz1,this->iAuxMpz1,aInfo.p);
		int res2 = res1 + plus;					 
		res2 %= aInfo.p;
		if (res2) 
		{
			PrintNfsElementInfo(aInfo);
			gmp_printf("Line Siever: Test Algebraic. For prime %d: current root is %d, which means argument %d, mod is %d. .\n",aInfo.p,aInfo.root_1,currentArgument,res2);
			Utils::Wait();
			result = false;
		} 
		else 
		{
			result = true;
		}
		
	} 
	else 
	{
		result = true;
	}

	return result;
}
void CNFSClassicalSievingPhase::ControlBlock(int aBlockIndex, int aInnerIndex, int aSign)
{
	int blockSize = Get_BlockSize();

	// aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize

	bool directionA = ((this->iSievingDirection == EIncrementA) || (this->iSievingDirection == EDecrementA));	
	int shiftConst = 2;
	int start = aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize;
	int max = start + blockSize;

	int base;

	string aux = "e:\\Data\\data_m" + Utils::ConvertToString(this->iWorkingB) + "block_" + Utils::ConvertToString(aSign * (aInnerIndex + 1)) + ".txt";

	FILE * data = fopen(aux.c_str(), FILE_MODE_WRITE);
	// col, row, Int Block, Alg Block, Int Value, Alg Value, Int Rem, Alg Rem, gcd

	int i;
	int index = 0;
	for (int a = start; a < max; a++)
	{
		this->iRelation->Reset();
		this->iRelation->Setup(aSign * (a + 1), this->iWorkingB, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iNumberFields);

		fprintf(data, "%d\t%d\t", a, this->iWorkingB);
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			fprintf(data, "%d\t", (- (int)this->iInitBlockValues[i] + this->iInitBlockValueChange[i] + (int)this->iSievingBlocks[i][index]));
		}
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			base = mpz_sizeinbase(this->iRelation->Get_Parts(i)->Norm, ConstSPhase::LOGARITHM_BASE);
			fprintf(data, "%d\t", base);
		}
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{		
			if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
			{
				FactorizeAlgebraic(shiftConst, directionA, aSign, index, a, &this->iHashtables[i].Hashtables[aInnerIndex], i); 
			}
			else
			{
				FactorizeIntegral(shiftConst, directionA, aSign, index, a, &this->iHashtables[i].Hashtables[aInnerIndex], i);
			}
		}
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			base = mpz_sizeinbase(this->iRelation->Get_Parts(i)->NormRemaining, ConstSPhase::LOGARITHM_BASE);
			fprintf(data, "%d\t", base);
		}
		Utils::GCD(abs(a + 1), abs(this->iWorkingB), base);
		fprintf(data, "%d\t", base);
		fprintf(data, "\n");
		fflush(data);

		index++;
	}

	fclose(data);
	data = NULL;
}
int CNFSClassicalSievingPhase::PrepareRunningSerialization()
{
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();
	Set_A(this->iWorkingA);
	Set_B(this->iWorkingB);
	Set_Iteration(Get_WorkingIteration());

	int i;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iRSerialInitBlockValueChange[i] = this->iInitBlockValueChange[i];
	}

	// buffer switch
	vector<CNFSRelation*>* aux = Get_RelationBuffer();
	int size = this->iRelationBufferMaxCount;
	Set_RelationBuffer(this->iRelationWorkingBuffer, this->iRelationWorkingBufferMaxCount);
	this->iRelationWorkingBuffer = aux;
	this->iRelationWorkingBufferMaxCount = 0;
	for (i = 0; i < size; i++)
	{
		this->iRelationWorkingBuffer->at(i)->Reset();
	}

	return ConstRC::Ok;
}

/*!
	Throughout the 1-partials processing, a hashtable containing large primes
	is used to track the state of the "partial relation graph". It serves us
      to determine the current number of independent cycles found.

	As such, this hashtable can be (re)constructed from the list of partial relations
    collected so far. That is why we do not save it into any XML form; instead,
 	we recreate it during deserialization from the partials file.

*/
int CNFSClassicalSievingPhase::ReloadHashtable()
{
	FILE *partialsFILE;
	CNFSRelation relation(CNFSRelation::MODE_CYCLE_CONSTRUCTION);
	int rc, rel_count;
	bool infoMode;
	
	if (Get_InfoMode())
	{
		Utils::NormalMessage("\t\tReconstructing hashtable for partial relations ... ");
	}

	// all new
	this->iStats->Reset(ESSEdges);
	this->iStats->Reset(ESSVertices);
	this->iStats->Reset(ESSComponents);

	// create hashtable and add ONE to hashtable
	rc = PrepareCycleTable();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (MSerializable::ExistFile(this->iPartialRelFullFileName) == false)
	{
		return ConstRC::Ok;
	}

	if ((partialsFILE = fopen(this->iPartialRelFullFileName.c_str(), FILE_MODE_READ)) == NULL) {
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","ReloadHashtable",0,0,"Could not open file with partial relations.");
		return ConstRC::CouldNotOpenFile;
	}

	if (Get_InfoMode()) {
		printf("\t\tPartial relations file name: %s.\n", this->iPartialRelFullFileName.c_str());
	}

	rel_count = 0;
	infoMode = Get_InfoMode();
	while (relation.ReadFromFile(partialsFILE) == ConstRC::Ok)
	{
		if (infoMode)
		{
			if (!(rel_count % 500))
			{
				printf("\r\t\tAdding %8d. relation...", rel_count);
				fflush(stdout);
			}
			++rel_count;
		}
		rc = AddToCycleCountStructures(&relation);
		if (rc != ConstRC::Ok)
		{
			fclose(partialsFILE);
			return rc;
		}
	}

	fclose(partialsFILE);

	if (Get_InfoMode())
	{
		printf("\r\t\tHashtable reconstructed. Altogether %d partials read, forming %d independent cycles.", rel_count, IndependentCyclesCount());
	}

	return ConstRC::Ok;
}

/*!
	TODO: If call from more threads then there could be a problem with writing relations - same files...
*/
int CNFSClassicalSievingPhase::Serialize(const CBaseParameters &aParam) const
{
#ifdef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// quick end
	return ConstRC::Ok;
#endif
	int rc = MSerializable::Serialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (Running())
	{
		AFactorAlgParameters * param = Get_Parameters();

		// for deserialization at new start
		param->SetPhaseParameter(ConstPN::DESERIALIZE, Utils::ConvertToString(true), Get_PhaseNbr());
		param->SetPhaseParameter(ConstPN::PHASE_FULL_FILENAME, 
			MSerializable::CreateFileNameAuto(aParam, ConstFN::SIEVING_PHASE), Get_PhaseNbr());

		CBaseParameters paramb;
		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_ParametersFullFileName()); 
		rc = param->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	// if it is running serialization, we serialize also found relations
	bool rserial = false;
	aParam.GetBoolParameter(ConstPN::RUNNING_SERIALIZATION, rserial);
	if (rserial)
	{
		vector<CNFSRelation*>* relations = Get_RelationBuffer();
		if (this->iRelationBufferMaxCount > 0 && relations != NULL)
		{
			FILE * partialsFILE = NULL;
			FILE * smoothsFILE = fopen(this->iSmoothRelFullFileName.c_str(), FILE_MODE_APPEND);
			if (smoothsFILE == NULL)
			{
				CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Serialize",0,0,"Could not open file with smooth relations.");
				return ConstRC::CouldNotOpenFile;
			}

			if (this->iVariationsInfo[0].Type != EDoNotUseVariations ||
				this->iVariationsInfo[1].Type != EDoNotUseVariations)
			{
				partialsFILE = fopen(this->iPartialRelFullFileName.c_str(), FILE_MODE_APPEND);
				if (partialsFILE == NULL)
				{
					fclose(smoothsFILE);
					smoothsFILE = NULL;
					CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Serialize",0,0,"Could not open file with partial relations.");
					return ConstRC::CouldNotOpenFile;
				}
			}

			CNFSRelation * relation = NULL;
			for (int i = 0; i < this->iRelationBufferMaxCount; i++)
			{
				relation = relations->at(i);
				if ((relation->Type & RELATION_FULL_ALGEBRAIC1) && (relation->Type & RELATION_FULL_ALGEBRAIC2))
				{
					// smooth relation
					rc = relation->PrintToFile(smoothsFILE);
					if (rc != ConstRC::Ok)
					{
						fclose(smoothsFILE);
						smoothsFILE = NULL;
						if (partialsFILE != NULL)
						{
							fclose(partialsFILE);
							partialsFILE = NULL;
						}
						return rc;
					}
				}
				else
				{
					// partial relation
					rc = relation->PrintToFile(partialsFILE);
					if (rc != ConstRC::Ok)
					{
						fclose(smoothsFILE);
						smoothsFILE = NULL;
						fclose(partialsFILE);
						partialsFILE = NULL;
						return rc;
					}
				}
			}

			fclose(smoothsFILE);
			smoothsFILE = NULL;

			if (partialsFILE != NULL)
			{
				fclose(partialsFILE);
				partialsFILE = NULL;
			}
		}
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
#ifdef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// quick end
	return ConstRC::Ok;
#endif
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_SIEVING_PHASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_INNER_STATE_NAME, Get_InnerState());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters param;
	rc = this->iStats->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = this->iTimes->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_ITERATION_NAME, Get_Iteration());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_A_NAME, (int)Get_A());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_B_NAME, (int)Get_B());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// write info about individual parts - generaly we have only two
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_PARTS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Serialize",rc,0,"Could not write the elements tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = SerializePart(i, aWriter);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_PARTS_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Serialize",rc,0,"Could not end the elements tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Serialize",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);

	return rc;
}
/*!
	Serialize info about individual part of sieving phase which corresponded to relation part.
*/
int CNFSClassicalSievingPhase::SerializePart(int aIndex, xmlTextWriterPtr &aWriter) const
{
#ifdef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// quick end
	return ConstRC::Ok;
#endif
	if (aIndex < 0 || aIndex >= CNFSRelation::DEFAULT_NFS_PARTS_COUNT)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","SerializePart",aIndex,0,"Index out of range.");
		return ConstRC::BadArgument;
	}

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENT_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","SerializePart",rc,0,"Could not write the elements tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_THRESHOLD_REFERENCE_NAME, (int)this->iInitBlockValueReference[aIndex]);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_THRESHOLD_CHANGE_NAME, this->iRSerialInitBlockValueChange[aIndex]);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENT_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","SerializePart",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, false);

	return rc;
}
/*!
	Shouldn't be call from outside!!!
*/
int CNFSClassicalSievingPhase::Deserialize(const CBaseParameters &aParam)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != ESPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	if (Get_InfoMode())
	{
		Utils::NormalMessage(" Deserializing Algebraic Sieving Phase...");
	}

	int rc = MSerializable::Deserialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int state;
	if ((state = Get_InnerState()) >= ESPhaseInitFB)
	{
		// for setup
		if (state == ESPhaseSieving)
		{
			Set_InnerState(ESPhaseInitAdvFB);
		}
		
		if (Get_InfoMode())
		{
			Utils::NormalMessage("\tInit integral domains ... ");
		}

		int i;
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			rc = this->iNumberFields[i].InitIntegralDomain(this->iFBs[i].UpperFBBound);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		if (Get_InfoMode())
		{
			Utils::NormalMessage("\tDeserializing FBs ... ");
		}

		CBaseParameters param;
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iFBs[i].FBFullFileName);
			rc = this->iFBs[i].Deserialize(param);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		if (Get_InfoMode())
		{
			Utils::NormalMessage("\tFilling auxiliary fields ... ");
		}

		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			this->iFBs[i].FillAuxFields(this->iNumberFields[i]);
			this->iNumberFields[i].FillSpecialPrimes();
		}

		if (this->iVariationsInfo[0].Type == ELargePrimeV &&
			this->iVariationsInfo[1].Type == ELargePrimeV)
		{
			//partial relations hashtable
			rc = ReloadHashtable();
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}

	return ConstRC::Ok;
}
/*!
	Shouldn't be call from outside!!!
*/
int CNFSClassicalSievingPhase::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != ESPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_SIEVING_PHASE_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	CBaseParameters param;
	int auxInt = 0;
	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INNER_STATE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize inner state...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_InnerState(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_STATISTIC_CONTAINER_NAME))
		{
			rc = this->iStats->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_TIME_CONTAINER_NAME))
		{
			rc = this->iTimes->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_A_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize A...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_A((main_sieving_type)auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_B_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize B...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_B((main_sieving_type)auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ITERATION_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize iteration...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_Iteration(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PARTS_NAME))
		{
			int depthParts = xmlTextReaderDepth(aReader);
			for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				if (CXMLHelper::MoveToNextNode(aReader, depthParts) != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				rc = DeserializePart(i, aReader);
				if (rc != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
			}

			if (rc != ConstRC::Ok)
			{
				break;
			}

			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CNFSClassicalSievingPhase::DeserializePart(int aIndex, xmlTextReaderPtr &aReader)
{
	if (aIndex < 0 || aIndex >= CNFSRelation::DEFAULT_NFS_PARTS_COUNT)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DeserializePart",aIndex,0,"Index out of range.");
		return ConstRC::BadArgument;
	}

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DeserializePart",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENT_NAME) == 0)
	{
		CLog::LogError(WHEREARG,"CNFSClassicalSievingPhase","DeserializePart",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int auxInt = 0;
	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_THRESHOLD_REFERENCE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize threshold reference...", auxInt, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iInitBlockValueReference[aIndex] = (log_type)auxInt;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_THRESHOLD_CHANGE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize threshold change...", this->iRSerialInitBlockValueChange[aIndex], false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iInitBlockValueChange[aIndex] = this->iRSerialInitBlockValueChange[aIndex];
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
main_sieving_type CNFSClassicalSievingPhase::Get_A() const
{
	main_sieving_type ret = 0;
	pthread_mutex_lock(&(this->iAMutex));
	ret = this->iA;
	pthread_mutex_unlock(&(this->iAMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_A(main_sieving_type aA)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iAMutex));
		this->iA = aA;
		pthread_mutex_unlock(&(this->iAMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
main_sieving_type CNFSClassicalSievingPhase::Get_B() const
{
	main_sieving_type ret = 0;
	pthread_mutex_lock(&(this->iBMutex));
	ret = this->iB;
	pthread_mutex_unlock(&(this->iBMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_B(main_sieving_type aB)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iBMutex));
		this->iB = aB;
		pthread_mutex_unlock(&(this->iBMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CNFSClassicalSievingPhase::Get_Iteration() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iIterationMutex));
	ret = this->iIteration;
	pthread_mutex_unlock(&(this->iIterationMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_Iteration(int aIteration)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iIterationMutex));
		this->iIteration = aIteration;
		pthread_mutex_unlock(&(this->iIterationMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CNFSClassicalSievingPhase::Get_WorkingIteration() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iWorkingIterationMutex));
	ret = this->iWorkingIteration;
	pthread_mutex_unlock(&(this->iWorkingIterationMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_WorkingIteration(int aWorkingIteration)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iWorkingIterationMutex));
		this->iWorkingIteration = aWorkingIteration;
		pthread_mutex_unlock(&(this->iWorkingIterationMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CNFSClassicalSievingPhase::Get_MaxIteration() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iMaxIterationMutex));
	ret = this->iMaxIteration;
	pthread_mutex_unlock(&(this->iMaxIterationMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_MaxIteration(int aMaxIteration)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iMaxIterationMutex));
		this->iMaxIteration = aMaxIteration;
		pthread_mutex_unlock(&(this->iMaxIterationMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CNFSClassicalSievingPhase::Get_TimeMessage() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iTimeMessageMutex));
	ret = this->iTimeMessage;
	pthread_mutex_unlock(&(this->iTimeMessageMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_TimeMessage(int aTimeMessage)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iTimeMessageMutex));
		this->iTimeMessage = aTimeMessage;
		pthread_mutex_unlock(&(this->iTimeMessageMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
vector<CNFSRelation*>* CNFSClassicalSievingPhase::Get_RelationBuffer() const
{
	vector<CNFSRelation*>* ret = NULL;
	pthread_mutex_lock(&(this->iRelationBufferMutex));
	ret = this->iRelationBuffer;
	pthread_mutex_unlock(&(this->iRelationBufferMutex));

	return ret;
}
int CNFSClassicalSievingPhase::Set_RelationBuffer(vector<CNFSRelation*>* aRelationBuffer, int aCount)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iRelationBufferMutex));
		this->iRelationBuffer = aRelationBuffer;
		this->iRelationBufferMaxCount = aCount;
		pthread_mutex_unlock(&(this->iRelationBufferMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}

/*
	Methods for distributed case.
*/

// initializes sieving phase and generates factor bases (called by center)
int CNFSClassicalSievingPhase::InitDistributedSieving(AFactorAlgParameters *aParameters) {

	Set_Parameters(aParameters);

	if (InitParameters() != ConstRC::Ok) {
		CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "InitDistributedSieving", 0, 0, "Parameters initialization failed.");
		return ConstRC::GeneralError;
	}

	if (InitFactorBases() != ConstRC::Ok) {
		CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "InitDistributedSieving", 0, 0, "Factor bases initialization failed.");
		return ConstRC::GeneralError;
	}

	if (InitAdvancedFactorBases() != ConstRC::Ok) {
		CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "InitDistributedSieving", 0, 0, "Advanced factor bases initialization failed.");
		return ConstRC::GeneralError;
	}

	DetermineRequestedRelations();

	return ConstRC::Ok;
}

// reads relations from given file, adds relations to count structures and appends to smooths or partials files
int CNFSClassicalSievingPhase::SupplyRelations(const char *aPath, bool aSmooth, int &aRelsCount) {
	FILE *f_rel, *f_append;
	string path_append;
	CNFSRelation relation(CNFSRelation::MODE_CYCLE_CONSTRUCTION);

	if (aSmooth) path_append = iSmoothRelFullFileName;
	else path_append = iPartialRelFullFileName;

	if ((f_rel = fopen(aPath, "r")) == NULL) {
		CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "SupplyRelations", 0, 0, "Could not open relations file.");
		return ConstRC::CouldNotOpenFile;
	}
	if ((f_append = fopen(path_append.c_str(), "a")) == NULL) {
		CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "SupplyRelations", 0, 0, "Could not open append file.");
		fclose(f_rel);
		return ConstRC::CouldNotOpenFile;
	}

	aRelsCount = 0;
	while (relation.ReadFromFile(f_rel) == ConstRC::Ok) {
//		if (CheckEnough() == ConstRC::EnoughSmooths) break;
		if (AddToCountStructures(&relation) != ConstRC::Ok) {
			CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "SupplyRelations", 0, 0, "Add to count structures failed.");
			fclose(f_rel);
			fclose(f_append);
			return ConstRC::GeneralError;
		}
		if (relation.PrintToFile(f_append) != ConstRC::Ok) {
			CLog::LogError(WHEREARG, "CNFSClassicalSievingPhase", "SupplyRelations", 0, 0, "Print relation to file failed.");
			fclose(f_rel);
			fclose(f_append);
			return ConstRC::GeneralError;
		}
		++aRelsCount;
	}

	fclose(f_rel);
	fclose(f_append);

	return ConstRC::Ok;
}

void CNFSClassicalSievingPhase::ConfirmAllStats() {

	iStats->ConfirmAll();

	return;
}

