#ifndef __NFS_CLASSICAL_SIEVING_PHASE_H__
#define __NFS_CLASSICAL_SIEVING_PHASE_H__

#include "abstract_sieving_phase.h"
#include "nfs_relation.h"
#include "complex_structures.h"
#include "factor_base_info.h"
#include "nfs_threshold_optimizer.h"
#include "lp_hashtable_type1.h"

///Abstract class for sieving phases
class CNFSClassicalSievingPhase : public ASievingPhase
{
private:
	///Field of the factor bases
	CFBInfo iFBs[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	/*
		Array with number field info:
			- sieving poly -> generating polynomial Theta
			- integral domain
			- info about special primes - for factorization
			- other helpful stuff
	*/
	CNumberFieldInfo iNumberFields[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	// multiplier of factor base bound which gives bound for large primes 
	int iPartialFactorMultiplier[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	/*!
		Info about variations used. 
		We use two "number fields" - (we can think about integral part as a number field)
		VariationInfo contains these things:
		 - variations mode = EDoNotUseVariations, ELargePrimeV, EDoubleLargePrimeV, ETrippleLargePrimeV
		 - factorization algorithm for factorizing remainder after FB factorization
		 - bounds - max remainder, min remainder, max factor
		 - leading coeff of the sieving polynomial - just for c_p calculation
	*/
	CVariationsInfo	iVariationsInfo[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	main_sieving_type		iA;
	mutable pthread_mutex_t iAMutex;
	main_sieving_type		iB;
	mutable pthread_mutex_t iBMutex;

	int						iIteration;
	mutable pthread_mutex_t iIterationMutex;
	int						iMaxIteration;
	mutable pthread_mutex_t iMaxIterationMutex;

	int						iTimeMessage;
	mutable pthread_mutex_t iTimeMessageMutex;


	long			iRequestedRelations;


	// Current sieving properties.
	int				iLineLowerBound; ///< must be < 0; determines M_0 such that we sieve through [-M_0, -1]. Is set directly by Init().
	int				iLineUpperBound; ///< must be > 0; determines M_1 such that we sieve through [1, M_1]. Is set directly by Init().
	int				iPositiveBlocks; ///< Is equal to number of positive blocks (basically, upper integral part of M_1/ iBlockSize)
	int				iNegativeBlocks; ///< Is equal to number of negative blocks (basically, upper integral part of -M_0/ iBlockSize)
	/*!
		Size of the set of blocks determined by 
		upper integral part of iFactorBaseUpperBound/ iBlockSize. 
		Denoted by K in the accompanying material.
	*/
	int				iBlockSetSize; 

	///Hashtables for sieving with largish primes
	CHashtables		iHashtables[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	

	///The sieving block itself
	log_type *		iSievingBlocks[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	///Advance factor bases - these bases are used for sieving - middle primes - and for filling hashtables
	CAdvCFBInfo		iAdvFB[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	/*!
		The value that will be used to fill the sieving block
		at the beginning of a new sieving cycle.

		It is basically 0x80 - (threshold log), so that aggregation of threshold_log
		would exceed 0x80.
	*/
	log_type		iInitBlockValues[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	/*!
		Change of the InitBlockValue - this value is optimized by CThresholdOptimizer
	*/
	int				iInitBlockValueChange[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	///Reference value used for computation of the init block values
	log_type		iInitBlockValueReference[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	
	///For optimizing sieving threshold 
	CThresholdOptimizer		iOptimizer[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	///just for now
	int				iBlockIndex;

	mpz_t			iAuxMpz1;
	mpz_t			iAuxMpz2;

	/*!
		If true than we use fact that e.g if B is even 
		(sieving direction is EIncrementB or EDecrementB) 
		than all even A are not coprime so we can omit them in the sieving.
	*/
	bool			iNfsCompressEvenIntervals;
	bool			iNfsOddToEven;
	bool			iNfsEvenToOdd;
	///True if current line (value of A or B) is even - it is used with iNfsCompressEvenIntervals
	bool			iIsCompressedNow;

	// sieving region
	sieving_region iSievingRegion;

	CNFSRelation*	iRelation;

	nfs_sieving_direction		iSievingDirection;

	CLPType1Hashtable *			iLPHashtable;
	std::vector<long>			iIntermediateRoots;

	// serialization stuff
	std::vector<CNFSRelation*>* iRelationBuffer;
	mutable pthread_mutex_t		iRelationBufferMutex;
	int							iRelationBufferMaxCount;
	std::vector<CNFSRelation*>* iRelationWorkingBuffer;
	int							iRelationWorkingBufferMaxCount;
	main_sieving_type			iWorkingA;
	main_sieving_type			iWorkingB;
	int							iWorkingIteration;
	mutable pthread_mutex_t		iWorkingIterationMutex;

	///How often we should try to find better threshold
	int iThresholdOptInterval;

	std::string		iSmoothRelFullFileName;
	std::string		iPartialRelFullFileName;

	// for running serialization
	bool			iRSerialAccordingLines;
	int				iRSerialLineInterval;
	bool			iRSerialAccordingTime;
	long			iRSerialTimeInterval;
	bool			iRSerialAccordingRelations;
	int				iRSerialRelationsCount;
	std::string		iRunningSerialFullFileName;
	int				iRSerialInitBlockValueChange[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];


protected:
	// Sieving methods
	int SieveLine();
	int SieveBlock(int aBlockIndex, int aBlockMin, int aBlockMax, int aSign, bool aInit);

	int Sieving(int aIndex, int aSign, bool aInit);

	int SieveWithMiddlePrimes(void* aStart, void* aEnd, log_type* aBlock, int aSign, bool aInit);
	int SieveWithLargishPrimes(script_t* aScript, log_type* aBlock);
	int ShiftLargishPrimes(CHashtables & aHashtables, int aCurrentBlockIndex);

	int FindingAndFactorizing(int aBlockIndex, int aInnerIndex, int aSign);

	int AddToCountStructures(CNFSRelation *aRelation);

	int Factorize(main_sieving_type aOffset, int aBlockIndex, int aInnerIndex, int aSign);
	int FactorizeIntegral(int aShiftConst, bool aDirectionA, int aSign, main_sieving_type aOffset,
			main_sieving_type aAuxOffset, hashtable_t *aHashtable, int aIndex);
	int FactorizeAlgebraic(int aShiftConst, bool aDirectionA, int aSign, main_sieving_type aOffset,
			main_sieving_type aAuxOffset, hashtable_t *aHashtable, int aIndex);
	void UpdateNfsRoot(nfs_fb_type* aEntry);

	int Shifting(int aIndex);
	int Swapping(int aIndex);

	int FillHashtable(CHashtables& aHashtables, nfs_fb_type * aStart, int aSign);

	int FindFreeRelations();
	int ComputeExpectedIntegralLog(int aIndex);
	int ComputeExpectedAlgebraicLog(int aIndex);
	int ComputeQuickExpectedAlgLog(int aBlockIndex, int aInnerIndex, int aSign, int aInitValue,
			Polynomial * aSievingPoly, CVariationsInfo & aVarInfo);

	int PrepareForSieving();

	// Tests methods - TODO
	int TestUpdate(const update_t &aUpdate);
	bool TestNfsAlgebraic(nfs_fb_type aInfo, main_sieving_type aUpdateIndex, int aBlockIndex, int aInnerIndex, int aSign, mpz_t aLeadingCoeff);
	void ControlBlock(int aBlockIndex, int aInnerIndex, int aSign);

	// Large Prime Variation methods
	int PrepareCycleTable();
	int AddToCycleCountStructures(CNFSRelation* aRelation);
	int IndependentCyclesCount();
	int ReloadHashtable();

	// Phase methods

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();
	int DisposeGMP();
	int DisposeMutexes();

	void DetermineRequestedRelations();

	int RunSieve();

	int InitAdvancedFactorBases();

	int InitFactorBases();

	int InitParameters();

	// For running serialization
	int PrepareRunningSerialization();

	// For serialization
	int SerializePart(int aIndex, xmlTextWriterPtr& aWriter) const;
	int DeserializePart(int aIndex, xmlTextReaderPtr& aReader);

public:

	CNFSClassicalSievingPhase();
	~CNFSClassicalSievingPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;
	int Serialize(const CBaseParameters& aParam) const;

	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
	int Deserialize(const CBaseParameters& aParam);

	// made public to allow call from distributed center

	int CheckEnough();

	int SaveResult();

	// Printing methods
	void PrintHeader();
	void PrintValues();
	void PrintEffectiveness();

	// methods for distributed case

	// initializes sieving phase and generates factor bases (called by center)
        int InitDistributedSieving(AFactorAlgParameters* aParameters);

	// reads relations from given file, adds relations to count structures and appends to smooths or partials files (called by center)
	int SupplyRelations(const char *aPath, bool aSmooth, int &aRelsCount);

	// confirm all statistics before saving result (called by center)
	void ConfirmAllStats();

	// Getters and Setters
public:
	main_sieving_type Get_A() const;
	int				  Set_A(main_sieving_type aA);
	main_sieving_type Get_B() const;
	int				  Set_B(main_sieving_type aB);
	int Get_Iteration() const;
	int	Set_Iteration(int aIteration);
	int Get_MaxIteration() const;
	int	Set_MaxIteration(int aMaxIteration);
	int Get_TimeMessage() const;
	int Set_TimeMessage(int aTimeMessage);
protected:
	int Get_WorkingIteration() const;
	int	Set_WorkingIteration(int aWorkingIteration);
	std::vector<CNFSRelation*>* Get_RelationBuffer() const;
	int							Set_RelationBuffer(std::vector<CNFSRelation*>* aRelationBuffer, int aCount);

};

#endif // __NFS_CLASSICAL_SIEVING_PHASE_H__
