#include "nfs_parameters.h"
#include "nfs_utils.h"
#include "enums.h"
#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_filenames.h"
#include "nfs_definitions.h"

using namespace std;

CNFSParameters::CNFSParameters() : AFactorAlgParameters(), MSerializable(ConstFN::NFS_PARAMETERS)
{
	this->iPhasesCount = NFS_PARAMETERS_PHASES_COUNT;
	this->iPhasesParameters = new CBaseParameters[this->iPhasesCount];

	this->iPhasesNames = new string[this->iPhasesCount];
	this->iPhasesNames[(int)EMain] = XML_SERIAL_NFS_PARAM_MAIN_NAME;
	this->iPhasesNames[(int)EPolySelection] = XML_SERIAL_NFS_PARAM_POLY_NAME;
	this->iPhasesNames[(int)ESieving] = XML_SERIAL_NFS_PARAM_SIEVE_NAME;
	this->iPhasesNames[(int)ERelProcessing] = XML_SERIAL_NFS_PARAM_REL_NAME;
	this->iPhasesNames[(int)ELinear] = XML_SERIAL_NFS_PARAM_LINEAR_NAME;
	this->iPhasesNames[(int)ESquareRoot] = XML_SERIAL_NFS_PARAM_SQUARE_NAME;

	// set nonserializable parameters
	AddPhaseNonSerializableParam(ConstPN::WORKING_DIRECTORY);
	AddPhaseNonSerializableParam(ConstPN::SERIALIZATION_ELEMENT_NAME);
	AddPhaseNonSerializableParam(ConstPN::POLY_PHASE_NUMBER);
	AddPhaseNonSerializableParam(ConstPN::SIEVING_PHASE_NUMBER);
	AddPhaseNonSerializableParam(ConstPN::REL_PROCESSING_PHASE_NUMBER);
	AddPhaseNonSerializableParam(ConstPN::LINEAR_PHASE_NUMBER);
	AddPhaseNonSerializableParam(ConstPN::SQUARE_ROOT_PHASE_NUMBER);
}

CNFSParameters::CNFSParameters(const CNFSParameters& aOperand) :
	AFactorAlgParameters(aOperand)
{
}
CNFSParameters::~CNFSParameters()
{
	if (this->iPhasesParameters != NULL)
	{
		delete[] this->iPhasesParameters;
		this->iPhasesParameters = NULL;
	}
	if (this->iPhasesNames != NULL)
	{
		delete[] this->iPhasesNames;
		this->iPhasesNames = NULL;
	}
}

/*
	This method parse all input parameters from command line. 
	These parameters are prefered (before deserialized parameters, but only these which are after deserialization). 
	Form of parameters is:
		"-phaseNbr:parameterName:parameterValue"
	or
		"-parameterName:parameterValue"
*/
int CNFSParameters::ParseInputParameters(int argc, char *argv[])
{
	// first parameter is path to this program
	string dir = Utils::GetDirectoryFromPath(argv[0]);
	AddPhaseParameter(ConstPN::WORKING_DIRECTORY, dir);
	
	for (int i = 1; i < argc; i++)
	{
		ParseInputParameter(argv[i]);
	}

	// we try to add serialization directory same as working if there isn't any
	//AddPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, dir);
	// we try to add serialization identifier if there isn't any
	AddPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, MSerializable::CreateNewIdentifier());
	// add phase numbers
	AddPhaseParameter(ConstPN::POLY_PHASE_NUMBER, Utils::ConvertToString((int)EPolySelection));
	AddPhaseParameter(ConstPN::SIEVING_PHASE_NUMBER, Utils::ConvertToString((int)ESieving));
	AddPhaseParameter(ConstPN::REL_PROCESSING_PHASE_NUMBER, Utils::ConvertToString((int)ERelProcessing));
	AddPhaseParameter(ConstPN::LINEAR_PHASE_NUMBER, Utils::ConvertToString((int)ELinear));
	AddPhaseParameter(ConstPN::SQUARE_ROOT_PHASE_NUMBER, Utils::ConvertToString((int)ESquareRoot));

	return ConstRC::Ok;
}

// ignores escaped separators
size_t CNFSParameters::FindInputParamSeparator(string aStr, size_t aPos) {
	string str(aStr);
	size_t pos = aPos;

	while ((pos = str.find(INPUT_PARAM_SEPARATOR, pos)) != string::npos) {
		if (pos == 0) return pos;
		if (str[pos-1] != '\\') return pos;
		++pos;
	}

	return string::npos;
}

// unescapes all escaped separator chars
// should be called on parsed parameter tokens
string CNFSParameters::UnescapeInputParameter(string aStr) {
	string str(aStr);
	size_t pos;

	while ((pos = str.find(string("\\")+INPUT_PARAM_SEPARATOR)) != string::npos) str.erase(pos, 1);

	return str;
}

/*
	This method parse one input parameter.
	Form of parameter is:
		"-phaseNbr:parameterName:parameterValue"
	or
		"-parameterName:parameterValue"
*/
int CNFSParameters::ParseInputParameter(const char *aParam)
{
	string str(aParam);

	if (str.length() == 0 || str[0] != '-')
	{
		return ConstRC::BadArgument;
	}

	// we erase '-' in the start position
	str.erase(0, 1);

	int i = FindInputParamSeparator(str, 0);
	int count = 0;
	while(i != string::npos) 
	{
		++count;
		i = FindInputParamSeparator(str, ++i);
	}

	int phase = (int)EMain;
	string name;
	string value;

	if (count == INPUT_PARAM_SEPARATOR_COUNT)
	{
		// input parameter form: nbrOfPhase:name:value
		int firstSeparator = FindInputParamSeparator(str, 0);
		int secondSeparator = FindInputParamSeparator(str, firstSeparator + 1);

		phase = atoi(UnescapeInputParameter(str.substr(0, firstSeparator)).c_str());
		name = UnescapeInputParameter(str.substr(firstSeparator + 1, secondSeparator - firstSeparator - 1));
		value = UnescapeInputParameter(str.substr(secondSeparator + 1, str.length() - secondSeparator));
	}
	else if (count == INPUT_PARAM_SEPARATOR_COUNT_DEFAULT)
	{
		// input parameter form: name:value
		int firstSeparator = FindInputParamSeparator(str, 0);
		name = UnescapeInputParameter(str.substr(0, firstSeparator));
		value = UnescapeInputParameter(str.substr(firstSeparator + 1, str.length() - firstSeparator));
	}
	else
	{
		return ConstRC::BadArgument;
	}

	// if we have to deserialize NFSParameters we will do it
	if (name.compare(DESERIALIZE_NFS_PARAMETERS) == 0 &&
		value.length() > 0)
	{
		CBaseParameters param;
		param.AddParameter(PARAM_SERIALIZATION_FULL_FILENAME, value);

		return MSerializable::Deserialize(param);
	}

	return SetPhaseParameter(name, value, phase);
}



int CNFSParameters::Serialize(const CBaseParameters& aParam, xmlTextWriterPtr &aWriter) const
{
	CBaseParameters param(aParam);
	// we just change main element if there isn't any
	param.AddParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_NFS_PARAM_ELEMENT_NAME);

	return AFactorAlgParameters::Serialize(param, aWriter);  
}
int CNFSParameters::Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr &aReader)
{
	CBaseParameters param(aParam);
	// we just change main element if there isn't any
	param.AddParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_NFS_PARAM_ELEMENT_NAME);

	return AFactorAlgParameters::Deserialize(param, aReader);
}

/*
        Methods of JobParameters.
*/

const string CNFSParameters::GetStringParameter(const char *aName) const {
        string r;

        if (GetStringPhaseParameter(string(aName), r, (int)EMain) != ConstRC::Ok) return KStringParameterUndefined;

        return r;
}

int CNFSParameters::GetIntegerParameter(const char *aName) const {
        int r;

        if (GetIntegerPhaseParameter(string(aName), r, (int)EMain) != ConstRC::Ok) return KIntegerParameterUndefined;

        return r;
}

double CNFSParameters::GetDoubleParameter(const char *aName) const {
        double r;

        if (GetDoublePhaseParameter(string(aName), r, (int)EMain) != ConstRC::Ok) return KDoubleParameterUndefined;

        return r;
}

bool CNFSParameters::GetBoolParameter(const char *aName) const {
        bool r;

        if (GetBoolPhaseParameter(string(aName), r, (int)EMain) != ConstRC::Ok) return false;

        return r;
}

int CNFSParameters::GetMpzParameter(const char *aName, mpz_t aResult) const {

        return GetMpzPhaseParameter(string(aName), aResult, (int)EMain);
}

bool CNFSParameters::Complete() const {
// TODO
        return true;
}

string CNFSParameters::GetFailureReason() const {
// TODO
        return "";
}

JobParameters *CNFSParameters::CreateCopy() const {

        return new CNFSParameters(*this);
}

void CNFSParameters::Print() const {

	AFactorAlgParameters::Print();

	return;
}

void CNFSParameters::PrintInBatch(const char *aPrefix) const {
	sieving_region sr;

	if (GetLongPhaseParameter(ConstPN::SIEVING_REGION_WIDTH, sr.width, (int)EMain) != ConstRC::Ok) return;
        if (GetLongPhaseParameter(ConstPN::SIEVING_REGION_HEIGHT, sr.height, (int)EMain) != ConstRC::Ok) return;
        if (GetLongPhaseParameter(ConstPN::SIEVING_REGION_START, sr.start, (int)EMain) != ConstRC::Ok) return;

	fprintf(stdout, "%ssieving region: width = %ld, height = %ld, start = %ld\n", aPrefix, sr.width, sr.height, sr.start);

	return;
}

