#include "nfs_relation.h"

#include "base_parameters.h"
#include "integer_class.h"
#include "const_return_codes.h"
#include "const_filenames.h"
#include "const_sieving_phase.h"
#include "test_suite.h"
#include "log.h"


using namespace std;

/*!
	Parameter aMode:
		Mode of parts initializations - how the number of elements should be determine
		- ARelation::MODE_CYCLE_CONSTRUCTION	- this relation will be used for cycle construction 
												- other relation will be combined with this relation so 
												starting number of relations is 0
												- probably many different large primes will be used
		- Other way is used only number in interval [0, 3] - it means no large prime, 1 large prime ...
*/
CNFSRelation::CNFSRelation(int aMode, int aPartCount) : MSerializable(ConstFN::NFS_RELATION)
{
	mpz_init(this->iFactor1);
	mpz_init(this->iFactor2);

	Set_ClassType(ERelNfs);

	this->A = 0;
	this->B = 0;
	this->Index = 0;
	this->Type = 0;

	int maxDivisors = ARelation::DEFAULT_DIVISORS_ALLOCATE;
	int maxLargeDivisors = 0;
	if (aMode & CNFSRelation::MODE_CYCLE_CONSTRUCTION) 
	{
		this->iNumberOfRelations = 0;
		maxLargeDivisors = maxDivisors;
	}
	else if (aMode & CNFSRelation::MODE_SQUARE_ROOT)
	{
		this->iNumberOfRelations = 0;
		maxDivisors = CNFSRelation::DEFAULT_SR_DIVISORS_ALLOCATE;
		maxLargeDivisors = maxDivisors >> 1;
	}
	else
	{
		this->iNumberOfRelations = 1;
		// just for sure
		int maxLR = ARelation::MAX_LARGE_PRIMES;
		maxLargeDivisors = min(aMode, maxLR);
	}

	this->iPartsCount = aPartCount;
	this->iParts = new CRelationPart[this->iPartsCount];

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].Main.Set_AssignedElements(ARelation::DEFAULT_DIVISORS_ALLOCATE);
		this->iParts[i].Large.Set_AssignedElements(maxLargeDivisors);
	}
}

CNFSRelation::CNFSRelation(const CNFSRelation& aOperand) : 
	MSerializable(aOperand), A(aOperand.A), B(aOperand.B)
{
	this->Index = aOperand.Index;
	this->Type = aOperand.Type;
	this->iPartsCount = aOperand.iPartsCount;
	Set_ClassType(ERelNfs);
	this->iNumberOfRelations = aOperand.iNumberOfRelations;

	mpz_init(this->iFactor1);
	mpz_init(this->iFactor2);

	this->iParts = new CRelationPart[this->iPartsCount];
	for (int i = 0; i < aOperand.iPartsCount; i++)
	{
		aOperand.iParts[i].Copy2(this->iParts[i]);
	}
}
CNFSRelation::~CNFSRelation()
{
	DisposeMpz();

	if (this->iParts != NULL)
	{
		// Delete invokes the destructor.
		delete[] this->iParts;
		this->iParts = NULL;
	}
}
void CNFSRelation::DisposeMpz()
{
	mpz_clear(this->iFactor1);
	mpz_clear(this->iFactor2);

	ARelation::DisposeMpz();
}

void CNFSRelation::PrintToScreen()
{	
	printf("================================================== \n");
	printf("#%lu\n", this->Index);
	printf("[a,b] = [%6ld,%6ld]\n", this->A, this->B);

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].PrintToScreen();
	}

    printf("================================================== \n");
}
bool CNFSRelation::TestCoprimality()
{
	long result = 0;
	if (Utils::LGCD(labs(this->A), labs(this->B), result) != ConstRC::Ok)
	{
		return false;
	}
	
	return (result == 1);
}
int CNFSRelation::Copy2(CNFSRelation &aRelation)
{
	aRelation.Index = this->Index;
	aRelation.Type = this->Type;
	aRelation.A = this->A;
	aRelation.B = this->B;
	aRelation.iNumberOfRelations = Get_NumberOfRelations();

	int rc;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].Copy2(aRelation.iParts[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
bool CNFSRelation::Equals(const CNFSRelation &aOperand) const
{
	if (ARelation::Equals(aOperand) == false) return false;

	if (this->A != aOperand.A) return false;
	if (this->B != aOperand.B) return false;
	
	return true;
}
int CNFSRelation::Reset()
{
	this->A = 0;
	this->B = 0;

	return ARelation::Reset();
}
int CNFSRelation::Setup(algebraic_coeff_type aA, algebraic_coeff_type aB, int aNFInfoCount, CNumberFieldInfo * aNFInfos)
{
	this->A = aA;
	this->B = aB;

	if (aNFInfos == NULL || this->iPartsCount != aNFInfoCount)
	{
		CLog::LogError(WHEREARG,"CNFSRelation","Setup",0,0,"Parameters mismatch.");
		return ConstRC::GeneralError; 
	}

	CRelationPart * part = this->iParts;

	for (int i = 0; i < this->iPartsCount; i++, part++, aNFInfos++)
	{
		if (aNFInfos->SievingPoly->GetDegree() > 1)
		{
			part->Poly->Set(0, aA);
			part->Poly->Set(1, aB);
			// if nonmonic than we represent c_{d}a + b \bar{\vartheta}
			if (aNFInfos->SievingPolyMonic == false)
			{
				mpz_mul(*part->Poly->Get2(0), *part->Poly->Get2(0), *aNFInfos->SievingPoly->Leading2());
			}
		}

		mpz_set_si(this->iAuxMpz1, this->A);
		mpz_set_si(this->iAuxMpz2, this->B);

		aNFInfos->Domain->Norm(aNFInfos->SievingPoly, part->Norm, this->iAuxMpz1, this->iAuxMpz2);	
		if (aNFInfos->SievingPoly->GetDegree() > 1)
		{
			// only for prime ideals
			mpz_abs(part->Norm, part->Norm);
		}
		mpz_set(part->NormRemaining, part->Norm);
	}

	return ConstRC::Ok;
}
int CNFSRelation::DetermineType(int aInfoCount, CVariationsInfo * aInfos)
{
	if (aInfos == NULL || this->iPartsCount != aInfoCount)
	{
		CLog::LogError(WHEREARG,"CNFSRelation","DetermineType",0,0,"Parameters mismatch.");
		return ConstRC::GeneralError; 
	}

	int rc;
	main_sieving_type prime;
	main_sieving_type root;
	int index = -1;
	// we want to go backwards
	CRelationPart * part = this->iParts + this->iPartsCount - 1;
	aInfos = aInfos + aInfoCount - 1;

	for (int i = this->iPartsCount - 1; i >= 0; i--, part--, aInfos--)
	{
		if (!mpz_cmp_ui(part->NormRemaining, 1)) 
		{
			this->Type |= (CRelationPart::SMOOTH << (ARelation::PART_TYPE_SHIFT * i));
		} 
		else  
		{
			if (aInfos->Type != EDoNotUseVariations)
			{
				// gmp_printf("NORM >%Zd<",part->NormRemaining);
				// we use variation - we need to know - if remainder is small enough - else it is wrong relation
				if (mpz_cmp(part->NormRemaining, aInfos->MaxPartialRemainder) < 0) 
				{
#ifdef DEBUG_RELATION_DETERMINE_TYPE
					if (mpz_cmp_ui(part->NormRemaining, DEBUG_RELATION_MIN_PARTIAL_FACTOR) < 0) 
					{
						gmp_printf("\t\tStrange!!! Relation second algebraic norm remaining factor is too small: %Zd\n", part->NormRemaining);
					}
#endif
					// we know that remainder is small enough
					// now we need to know if we have to factorize or the remainder fits into max large prime
					if (aInfos->Type > ELargePrimeV &&
						mpz_cmp_si(part->NormRemaining, aInfos->MaxPartialFactor) > 0)
					{
						// the remainder is bigger than max large prime so we need to factorize
						// but we factorize only if the remainder is not prime
						// if it is then this is wrong relation
						// we know that if the remainder is smaller than square of FB bound then it is definitely prime
						if (mpz_cmp(part->NormRemaining, aInfos->MinPartialRemainder) > 0 &&
							mpz_probab_prime_p(part->NormRemaining, ARelation::PROBAB_PRIME_DEGREE) == 0)
						{
							rc = aInfos->FactorAlg->FactorWithStatistics(part->NormRemaining, this->iFactor1, this->iFactor2);
							// we control if the factors are small enough but it is not necessary...
							if (rc == ConstRC::Ok &&
								mpz_cmp_si(this->iFactor1, aInfos->MaxPartialFactor) < 0 &&
								mpz_cmp_si(this->iFactor2, aInfos->MaxPartialFactor) < 0)
							{
								this->Type |= (CRelationPart::DOUBLE_LARGE << (ARelation::PART_TYPE_SHIFT * i));
								prime = mpz_get_si(this->iFactor1);
								root = GetRoot(this->A, this->B, prime);

								index = -1;
								rc = part->Large.AddDivisor(prime, 1, root, 0, index);
								if (rc != ConstRC::Ok)
								{
									return rc;
								}

								prime = mpz_get_si(this->iFactor2);
								root = GetRoot(this->A, this->B, prime);

								rc = part->Large.AddDivisor(prime, 1, root, 0, index);
								if (rc != ConstRC::Ok)
								{
									return rc;
								}

								mpz_set_ui(part->NormRemaining, 1);
							}
							else
							{
								this->Type |= (CRelationPart::TOO_LARGE << (ARelation::PART_TYPE_SHIFT * i));
								return ConstRC::FactorTooLarge;
							}
						}
						else
						{
							this->Type |= (CRelationPart::TOO_LARGE << (ARelation::PART_TYPE_SHIFT * i));
							return ConstRC::PrimeTooLarge;
						}
					}
					else
					{
						this->Type |= (CRelationPart::LARGE << (ARelation::PART_TYPE_SHIFT * i));
						
						prime = mpz_get_ui(part->NormRemaining);
						root = GetRoot(this->A, this->B, prime); 

						index = -1;
						rc = part->Large.AddDivisor(prime, 1, root, 0, index);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}

						mpz_set_ui(part->NormRemaining, 1);
					}
				} 
				else 
				{
					this->Type |= (CRelationPart::TOO_LARGE << (ARelation::PART_TYPE_SHIFT * i));
					return ConstRC::Impropriate;
				}
			}
			else
			{
				return ConstRC::Impropriate;
			}
		}
	}	

	return ConstRC::Ok;
}

CNFSRelation *CNFSRelation::CopyItem() {

	return new CNFSRelation(*this);
}

/*!
Should be called only by objects dynamically allocated by CopyItem().
*/
void CNFSRelation::FreeItem() {

	delete (CNFSRelation *)this;
}

int CNFSRelation::Serialize(const CBaseParameters &aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CNFSRelation::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_NFS_RELATION_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CNFSRelation","Serialize",rc,0,"Could not write the main tag.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_NBR_OF_RELATIONS_NAME, this->iNumberOfRelations);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_A_NAME, this->A);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_B_NAME, this->B);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeUnsignedInt(aWriter, XML_SERIAL_TYPE_NAME, this->Type);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeUnsignedLong(aWriter, XML_SERIAL_INDEX_NAME, this->Index);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// now serialize relation parts
	rc = ARelation::SerializeParts(aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CNFSRelation","Serialize",rc,0,"Could not end the main tag.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CNFSRelation::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CNFSRelation::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_NFS_RELATION_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSRelation","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CNFSRelation","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);
	string auxStr;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_NBR_OF_RELATIONS_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize number of relations...", this->iNumberOfRelations, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_A_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize a...", this->A, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_B_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize b...", this->B, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_TYPE_NAME))
		{
			rc = CXMLHelper::DeserializeUnsignedInt(aReader, "Deserialize type...", this->Type, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INDEX_NAME))
		{
			rc = CXMLHelper::DeserializeUnsignedLong(aReader, "Deserialize index...", this->Index, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PARTS_NAME))
		{
			rc = ARelation::DeserializeParts(aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError)
	{
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}
/*!
	This method is used to save a relation into file. It uses fprintf() for this purpose.
	This method has to be fast.
	File Template:
	==================================================
	iClassType
	Index
	Type
	iNumberOfRelations
	A
	B
	-- parts are next - template:
	Poly
	Norm
	NormRemaining
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
	Large PrimeIdeals count
	Large PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CNFSRelation::PrintToFile(FILE* aFw)
{
	if (aFw == NULL)
	{
		return ConstRC::CouldNotWriteToNullStream;
	}
	// At first, we sort the divisors in relation
	Sort();
	fprintf(aFw, "==================================================\n");
	fprintf(aFw, "%d\n", (int)Get_ClassType());
	fprintf(aFw, "%lu\n", this->Index);     
	fprintf(aFw, "%u\n", this->Type);
	fprintf(aFw, "%d\n", this->iNumberOfRelations);
	fprintf(aFw, "%ld\n", this->A);
	fprintf(aFw, "%ld\n", this->B);

	for (int i = 0; i < this->iPartsCount; i++)
	{
		this->iParts[i].PrintToFile(aFw);
	}

	fflush(aFw);

	return ConstRC::Ok;
}


/*!
	This method is used to read a relation from file.
	File Template:
	==================================================
	iClassType
	Index
	Type
	iNumberOfRelations
	A
	B
	-- parts are next - template:
	Poly
	Norm
	NormRemaining
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
	Large PrimeIdeals count
	Large PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CNFSRelation::ReadFromFile(FILE* aFr) 
{
	if (aFr == NULL)
	{
		CLog::LogError(WHEREARG,"CNFSRelation","ReadFromFile",0,0,"Could not read elements from NULL stream");
		return ConstRC::CouldNotReadFromNullStream;
	}

	char data[CNFSRelation::DIVISOR_READING_BUFFER_SIZE];

	// ==================================================
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}

	// iClassType
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	if (atoi(data) != (int)Get_ClassType())
	{
		CLog::LogError(WHEREARG,"CNFSRelation","ReadFromFile",atoi(data),(int)Get_ClassType(),"Relation class type mismatch.");
		return ConstRC::GeneralError;
	}

	// Index
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	sscanf(data, "%lu", &this->Index);

	// Type
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	sscanf(data, "%u", &this->Type);

	// iNumberOfRelations
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->iNumberOfRelations = atoi(data);

	// A
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->A = atol(data);

	// B
	if (fgets(data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->B = atol(data);

	int rc;
	for (int i = 0; i < this->iPartsCount; i++)
	{
		rc = this->iParts[i].ReadFromFile(aFr, data, CNFSRelation::DIVISOR_READING_BUFFER_SIZE);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
