#include "nfs_threshold_optimizer.h"
#include "nfs_definitions.h"
#include <stdio.h>
#include <stdlib.h>
#include "log.h"

#define DEBUG_THRESHOLD_OPTIMIZER 0

CThresholdOptimizer::CThresholdOptimizer(int aOptimizationPolicy, float aDesiredRatio)
	: iAccepted(0), iRejected(0), iOptimizationPolicy(aOptimizationPolicy), iDesiredRatio(aDesiredRatio)
{
	if ((iOptimizationPolicy > 5) || (iOptimizationPolicy < 0))
	{
		CLog::LogError(WHEREARG, "CThresholdOptimizer", "CThresholdOptimizer", iOptimizationPolicy, 0,  "Bad optimization policy.");
	}		

	this->iMinChange = - CThresholdOptimizer::DEFAULT_CHANGE;
	this->iMaxChange = CThresholdOptimizer::DEFAULT_CHANGE;
}

CThresholdOptimizer::~CThresholdOptimizer()
{
}

void CThresholdOptimizer::SetChangeLimits(int aExpectedValue)
{
	// minimal result value can be 1 ==> threshold is then SIEVING_PHASE_MASK - 1
	this->iMinChange = -aExpectedValue + 1;
	// maximal result value can be SIEVING_PHASE_MASK - 1 ==> threshold is then 1 
	this->iMaxChange = SIEVING_PHASE_MASK - aExpectedValue - 1;
}
/*!
	This method change the expected value of the F(x,y) over sieving interval.
	result = expected + change

	Change is computed according to sieving success.
*/
void	CThresholdOptimizer::AnalyzeAndOptimize(int& aBlockChange)
{
#if DEBUG_THRESHOLD_OPTIMIZER
	PrintStatus();
	CLog::Log(100, "", "\n"); 
	CLog::Log(100, "1060000", "AnalyzeAndOptimize\n");
	CLog::Log(100, "1060001", "    Block change: %d\n", aBlockChange);
#endif

	// check info
	if (!HasEnoughInformationToAnalyze())
	{
		// it's possible that we cannot find anything
		aBlockChange -= 1;
		// threshold can be too high
#if DEBUG_THRESHOLD_OPTIMIZER
		CLog::Log(100, "1060002", "  We do not have enough information to analyze and optimize the threshold.\n");
		CLog::Log(100, "1060003", "    Block change: %d\n", aBlockChange);
#endif
		return;
	}

	// Has enough information 
	if (!iRejected)
	{
		// We have at least KMinAccepted, but no rejected.
		aBlockChange -= 1;
		// threshold can be too high so we find only smooth numbers

#if DEBUG_THRESHOLD_OPTIMIZER
		CLog::Log(100, "1060004", "  Special case: rejected == 0");
		CLog::Log(100, "1060005", "    Block change: %d\n", aBlockChange);
#endif

		return;
	}

	iRatio = (float) iAccepted / (float) iRejected;
	iDiff = iRatio / iDesiredRatio;
#if DEBUG_THRESHOLD_OPTIMIZER
	CLog::Log(100, "1060006", "  Ratio now: %4.2f\n", iRatio);
	CLog::Log(100, "1060007", "  Diff now: %4.2f\n", iDiff);
#endif
	signed int move = 0;
	if (this->iDiff < 0.05) //(iDiff < 0.05)
	{
		// A lot of rejected, just a few of accepted
		move = -3;
	} 
	else if ((this->iDiff >= 0.05) && (this->iDiff < 0.25)) //((iDiff >= 0.05) && (iDiff < 0.10))
	{
		move = -2;
	} 
	else if ((this->iDiff >= 0.25) && (this->iDiff < 0.80)) //((iDiff >= 0.10) && (iDiff < 0.33))
	{
		move = -1;
	} 
	else if ((this->iDiff >= 0.80) && (this->iDiff <= 1.25)) //((iDiff >= 0.33) && (iDiff < 1))
	{
		move = 0;
	} 
	else if ((this->iDiff > 1.25) && (this->iDiff <= 4.00)) //((iDiff >= 1) && (iDiff < 2))
	{
		move = 1;
	} 
	else if ((this->iDiff > 4.00) && (this->iDiff <= 20.00)) //((iDiff >= 2) && (iDiff < 5))
	{
		move = 2;
	} 
	else if (this->iDiff > 20) //(iDiff >= 5)
	{
		move = 3;
	}
	bool optimize = true;
	switch (iOptimizationPolicy)
	{
		case 0:
			optimize = false;
			break;
		case 1:
			move -= 2;
			break;	
		case 2:
			move -= 1;
			break;	
		case 3:
			// Nothing		
			break;	
		case 4:
			move += 1;
			break;	
		case 5:
			move += 2;
			break;		
		default:
			CLog::LogError(WHEREARG,"CThresholdOptimizer","AnalyzeAndOptimize",iOptimizationPolicy,0,"Bad optimization policy.");
			return;
	}
	
	if (optimize)
	{
		aBlockChange -= move;
#if DEBUG_THRESHOLD_OPTIMIZER
	CLog::Log(100, "1060008", "  Move calculated as: %d\n", move);
	CLog::Log(100, "1060009", "    Block change: %d\n", aBlockChange);	
#endif
	}
	if (aBlockChange < this->iMinChange)
	{
		aBlockChange = this->iMinChange;
#if DEBUG_THRESHOLD_OPTIMIZER
	CLog::Log(100, "1060010", "  Min change!!!\n");
	CLog::Log(100, "1060011", "    Block change: %d\n", aBlockChange);	
#endif
	}
	else if (aBlockChange > this->iMaxChange)
	{
		aBlockChange = this->iMaxChange;
#if DEBUG_THRESHOLD_OPTIMIZER
	CLog::Log(100, "1060012", "  Max change!!!\n");
	CLog::Log(100, "1060013", "    Block change: %d\n", aBlockChange);	
#endif
	}

	ClearStatistics();
}

void	CThresholdOptimizer::AnalyzeAndOptimize(log_type& aBlockValue)
{
#if DEBUG_THRESHOLD_OPTIMIZER
	PrintStatus();
	CLog::Log(100, "", "\n"); 
	CLog::Log(100, "1060014", "AnalyzeAndOptimize\n");
	CLog::Log(100, "1060015", "    Block value: %d\n", aBlockValue);
#endif

	// check info
	if (!HasEnoughInformationToAnalyze())
	{
		// it's possible that we cannot find anything
		aBlockValue += 1;
		// threshold can be too high
#if DEBUG_THRESHOLD_OPTIMIZER
		CLog::Log(100, "1060016", "  We do not have enough information to analyze and optimize the threshold.\n");
		CLog::Log(100, "1060017", "    Block value: %d\n", aBlockValue);
#endif
		return;
	}

	// Has enough information 
	if (!iRejected)
	{
		// We have at least KMinAccepted, but no rejected.
		aBlockValue += 1;
		// threshold can be too high so we find only smooth numbers

#if DEBUG_THRESHOLD_OPTIMIZER
		CLog::Log(100, "1060018", "  Special case: rejected == 0");
		CLog::Log(100, "1060019", "    Block value: %d\n", aBlockValue);
#endif

		return;
	}

	iRatio = (float) iAccepted / (float) iRejected;
	iDiff = iRatio / iDesiredRatio;
#if DEBUG_THRESHOLD_OPTIMIZER
	CLog::Log(100, "1060020", "  Ratio now: %4.2f\n", iRatio);
	CLog::Log(100, "1060021", "  Diff now: %4.2f\n", iDiff);
#endif
	signed int move = 0;
	if (iDiff < 0.05)
	{
		// A lot of rejected, just a few of accepted
		move = -3;
	} 
	else if ((iDiff >= 0.05) && (iDiff < 0.10))
	{
		move = -2;
	} 
	else if ((iDiff >= 0.10) && (iDiff < 0.33))
	{
		move = -1;
	} 
	else if ((iDiff >= 0.33) && (iDiff < 1))
	{
		move = 0;
	} 
	else if ((iDiff >= 1) && (iDiff < 2))
	{
		move = 1;
	} 
	else if ((iDiff >= 2) && (iDiff < 5))
	{
		move = 2;
	} 
	else if (iDiff >= 5)
	{
		move = 3;
	}
	bool optimize = true;
	switch (iOptimizationPolicy)
	{
		case 0:
			optimize = false;
			break;
		case 1:
			move -= 2;
			break;	
		case 2:
			move -= 1;
			break;	
		case 3:
			// Nothing		
			break;	
		case 4:
			move += 1;
			break;	
		case 5:
			move += 2;
			break;		
		default:
			CLog::LogError(WHEREARG,"CThresholdOptimizer","AnalyzeAndOptimize",iOptimizationPolicy,0,"Bad optimization policy.");
			return;
	}

	log_type newBlockValue = aBlockValue + move;

	if (newBlockValue >= SIEVING_PHASE_MASK)
	{
		if (move > 0)
		{
#if (!DEBUG_THRESHOLD_OPTIMIZER)
			// If not in debug mode, print the status now.
			// Otherwise, it has already been printed.
			PrintStatus();
#endif
			newBlockValue = SIEVING_PHASE_MASK - 1;
			CLog::Log(MAXIMAL_INT_VALUE, "1060022", "Threshold optimizer: Warning. BlockValue already too high. Set to %d.\n", newBlockValue);
		}
		else
		{
#if (!DEBUG_THRESHOLD_OPTIMIZER)
			// If not in debug mode, print the status now.
			// Otherwise, it has already been printed.
			PrintStatus();
#endif
			newBlockValue = 0;
			CLog::Log(MAXIMAL_INT_VALUE, "1060023", " ^^^^^ Threshold optimizer: Warning. BlockValue already too low. Set to %d.\n", newBlockValue);
		}
	}
	
#if DEBUG_THRESHOLD_OPTIMIZER
	CLog::Log(100, "1060024", "  Move calculated as: %d\n", move);
	CLog::Log(100, "1060025", "    Block value: %d\n", newBlockValue);	
#endif
	if (optimize)
	{
		aBlockValue = newBlockValue;	
	}
	ClearStatistics();
}

void	CThresholdOptimizer::PrintStatus() const
{
	CLog::Log(100, "", "\n"); 
	CLog::Log(100, "1060026", "Threshold optimizer status: policy %d ^^^^^\n", iOptimizationPolicy);
	CLog::Log(100, "1060027", "  Accepted: %6d\n", (int)iAccepted);
	CLog::Log(100, "1060028", "  Rejected: %6d\n", (int)iRejected);
	if (iRejected)
	{
		CLog::Log(100, "1060029", "  Ratio: %4.2f\n", (float) iAccepted / (float) iRejected);
	}
}

void	CThresholdOptimizer::Reset()	
{
	ClearStatistics();
	this->iMinChange = - CThresholdOptimizer::DEFAULT_CHANGE;
	this->iMaxChange = CThresholdOptimizer::DEFAULT_CHANGE;
	this->iOptimizationPolicy = 3;
	this->iDesiredRatio = (float) 1;
}

void	CThresholdOptimizer::ClearStatistics()	
{
	this->iAccepted = 0;
	this->iRejected = 0;
	this->iRatio = (float) 0;
	this->iDiff = (float) 0;
}

void	CThresholdOptimizer::IncrementAccepted(unsigned int aBy)
{
	iAccepted += aBy;
}

void	CThresholdOptimizer::IncrementRejected(unsigned int aBy)
{
	iRejected += aBy;
}


bool	CThresholdOptimizer::HasEnoughInformationToAnalyze() const
{
	return ((iAccepted >= CThresholdOptimizer::MIN_ACCEPTED) || (iRejected >= CThresholdOptimizer::MIN_REJECTED));
}
