#ifndef __THRESHOLD_OPTIMIZER_H_
#define __THRESHOLD_OPTIMIZER_H_

#include "types.h" // For log_type

/*!
	\class CThresholdOptimizer
	\brief Optimizes threshold values during sieving. 

	This class analyzes performance of current sieving run, mainly the
	ratio of accepted to rejected relations. If the ratio indicates suboptimal
	choice of the sieving threshold, the class computes a new value.

*/
class CThresholdOptimizer
{
public:
	// Construction and destruction
	/**
	\brief Constructs the optimizer.
	@param aOptimizationPolicy Either 0, 1, 2, 3, 4, 5. 
	0 = off. In this case, the optimized does not optimize. 
	Otherwise, the higher is the number, the higher the block value will be (less restrictive). 
	Value other than 5 results in exit of the program.

	*/
	CThresholdOptimizer(int aOptimizationPolicy = 3, float aDesiredRatio = 1.0);
	~CThresholdOptimizer();

	// Enter new information to analyze
	void	IncrementAccepted(unsigned int aBy = 1);
	void	IncrementRejected(unsigned int aBy = 1);

	/**
	\brief Determines new threshold value.
	@param aCurrentThreshold The current value of threshold. Will be rewritten with the new value.

	For the analysis logic, see the document sieving-threshold-optimization.odt

	*/
	void	AnalyzeAndOptimize(log_type& aCurrentThreshold);
	void	AnalyzeAndOptimize(int& aBlockChange);		

	void	SetChangeLimits(int aExpectedValue);


	void	Reset();

// Getters and Setters
	int   Get_OptimizationPolicy() const { return this->iOptimizationPolicy; }
	void  Set_OptimizationPolicy(int aOptimizationPolicy) {this->iOptimizationPolicy = aOptimizationPolicy; }
	float Get_DesiredRatio() const { return this->iDesiredRatio; }
	void  Set_DesiredRatio(float aDesiredRatio) {this->iDesiredRatio = aDesiredRatio; }

	void  Set_MinChange(int aMinChange) { this->iMinChange = aMinChange; }
	void  Set_MaxChange(int aMaxChange) { this->iMaxChange = aMaxChange; }
	int	  Get_MinChange() { return this->iMinChange; }
	int	  Get_MaxChange() { return this->iMaxChange; }



private:
	/**
          Clears the accepted / rejected statistics.
	*/
	void	ClearStatistics();

	/**
	  Returns TRUE if iAccepted >= KMinAccepted or iRejected >= KMinRejected.
	*/
	bool	HasEnoughInformationToAnalyze() const;

	void	PrintStatus() const;

private:
	// Data members
	unsigned int	iAccepted;
	unsigned int	iRejected;
	float		iRatio;
	float		iDesiredRatio;
	float		iDiff;

	int iMinChange;
	int iMaxChange;

// Constants
	static const unsigned int MIN_ACCEPTED = 5;
	static const unsigned int MIN_REJECTED = 50;
	static const int DEFAULT_CHANGE = 3;

	int 		iOptimizationPolicy;
};

#endif //  __THRESHOLD_OPTIMIZER_H_
