#include "nfs_utils.h"
#include "const_return_codes.h"
#include "const_hashtables.h"
#include "const_element_flags.h"
#include "structures.h"
#include "log.h"

/*!
 * \brief Compare two "nfs_sieving_element" types.
 *
 * This function is used in qsort calls operating
 * on arrays of nfs_sieving_element types.
 *
 */
int NFSUtils::CompareSievingElements(const void* aArg1, const void* aArg2) 
{ 
	nfs_sieving_element* element1 = (nfs_sieving_element*) aArg1;
	nfs_sieving_element* element2 = (nfs_sieving_element*) aArg2;
	int result = 0;
	if  ((result = (element1->prime - element2->prime)) != 0) {
		return result;
	}	
	return (element1->root - element2->root);
}

/*
	The method's name and signature explains all. The merged fields should be
	different, and the target should be also different from all of them.

*/
int NFSUtils::MergeTwoSortedArrays(nfs_sieving_element** aTarget, unsigned long* aTargetAllocated, signed long* aTargetMaxIndex, nfs_sieving_element* aOperand1, int aOperand1MaxIndex, nfs_sieving_element* aOperand2, int aOperand2MaxIndex) 
{
	int	maximal_number_of_elements = aOperand1MaxIndex + aOperand2MaxIndex + 3; // to be sure.
	// If the allocated target is smaller than max.number of elements, we reallocate it to accomodate more elements	
	if (((signed) (*aTargetAllocated)) < maximal_number_of_elements) 
	{
		//printf("MTSA: reallocing target to %d.\n",maximal_number_of_elements); 
		*aTarget = (nfs_sieving_element*) NFSUtils::AllocationMachine((void**) aTarget,maximal_number_of_elements*sizeof(nfs_sieving_element));
		if (*aTarget == NULL) 
		{
			CLog::LogError(WHEREARG,"NFSUtils","MergeTwoSortedArrays",maximal_number_of_elements*sizeof(nfs_sieving_element),40,"There was an error reallocating array.");			
			return ConstRC::NotEnoughMemory;	
		}
		*aTargetAllocated = maximal_number_of_elements;
	}
	// simple copy iff some of the operands is empty
	nfs_sieving_element* source = NULL;
	int max = 0;
	if (aOperand1MaxIndex == -1) 
	{
		source = aOperand2;
		max = aOperand2MaxIndex;
	}
	if (aOperand2MaxIndex == -1) 
	{
		source = aOperand1;
		max = aOperand1MaxIndex;
	}
	int i;
	if (source != NULL) 
	{
		//printf("MTSA: simple copy %d, %d, %d.\n",max,aOperand1MaxIndex,aOperand2MaxIndex);
		for (i=0; i<= max; i++) 
		{
			(*aTarget)[i] = source[i];
			//printf("Copied %d.\n",source[i]);
		}
		*aTargetMaxIndex = max;
		return ConstRC::Ok;
	}
	// If both operands are nonempty ...
	int last_written = -1;
	int last_read_1 = 0;
	int last_read_2 = 0;
	int reading_stream_index = 1;
	int compare;
	bool finished = false;
	bool finished_1 = false;
	bool finished_2 = false;
	while (!finished) 
	{		
		// Both operands are unfinished yet. We will write down to target array the lesser entry
		compare = NFSUtils::CompareSievingElements((void*) &aOperand1[last_read_1],(void*) &aOperand2[last_read_2]);		
		last_written++;
		if (compare < 0) 
		{
			// The current element in the first stream was lesser than the current
			// element in the second stream.
			if (reading_stream_index == 2) 
			{
				reading_stream_index = 1;
			}
			// We will copy the element from the first stream/
			(*aTarget)[last_written] = aOperand1[last_read_1];
			//printf("Zapsano z prvniho: %d, %d.\n",(*aTarget)[last_written],last_written,last_read_1);
		} 
		else if (compare == 0) 
		{
			// The current elements in stream 1 and stream 2 are identical.
			// We will read them both.
			reading_stream_index = 4;
			(*aTarget)[last_written] = aOperand1[last_read_1];
			(*aTarget)[last_written].exponent += aOperand2[last_read_2].exponent;
			if ((*aTarget)[last_written].flags & ConstEF::DIVIDES_L)
			{
				// we need to count p-valuation of F(a,b) of these elements
				(*aTarget)[last_written].flags += (aOperand2[last_read_2].flags & ConstEF::F_VALUATION_MASK);
			}
			//printf("Zapsano z obou: %d, %d.\n",(*aTarget)[last_written],last_written,last_read_1);
		} 
		else 
		{
			// The current element in the first stream was greater than the current
			// element in the second stream.
			if (reading_stream_index == 1) 
			{
				reading_stream_index = 2;
			}
			// Zkopirujeme nizsi prvek do targetu
			(*aTarget)[last_written] = aOperand2[last_read_2];
			//printf("Zapsano ze druheho: %d, %d.\n",(*aTarget)[last_written],last_written,last_read_2);
		}		
		// We will make a read in the currently used stream (1 or 2 or both)
		if ((reading_stream_index == 1) || (reading_stream_index == 4)) 
		{
			last_read_1++;
			if (last_read_1 > aOperand1MaxIndex) 
			{
				finished_1 = true;
			}
		} 	
		if ((reading_stream_index == 2) || (reading_stream_index == 4)) 
		{
			last_read_2++;
			if (last_read_2 > aOperand2MaxIndex) 
			{
				finished_2 = true;
			}
		}		
		if (finished_1 || finished_2) 
		{
			// We will finish the copying process by writing down remaining entries from the unfinished operand
			if (!finished_1) 
			{
				while (last_read_1 <= aOperand1MaxIndex) 
				{
					last_written++;
					(*aTarget)[last_written] = aOperand1[last_read_1];
					last_read_1++;
				}				
			}
			if (!finished_2) 
			{
				while (last_read_2 <= aOperand2MaxIndex) 
				{
					last_written++;
					(*aTarget)[last_written] = aOperand2[last_read_2];
					last_read_2++;
				}				
			}
			finished = true;
		} 
		else 
		{
			if (reading_stream_index == 4) 
			{
				// None of the both streams is finished.
				if (CompareSievingElements( &aOperand1[last_read_1], &aOperand2[last_read_2]) <= 0) 
				{
					reading_stream_index = 1;
				} 
				else 
				{
					reading_stream_index = 2;
				}
			}
		}
	}
	*aTargetMaxIndex = last_written;
	return ConstRC::Ok;
}

/*
	The method's name and signature explains all. The merged fields should be
	different, and the target should be also different from all of them and big enough.

*/
int NFSUtils::MergeTwoSortedArrays(nfs_sieving_element* aTarget, long & aTargetMaxIndex, nfs_sieving_element* aOperand1, int aOperand1MaxIndex, nfs_sieving_element* aOperand2, int aOperand2MaxIndex) 
{
	// simple copy iff some of the operands is empty
	nfs_sieving_element* source = NULL;
	int max = 0;
	if (aOperand1MaxIndex == -1) 
	{
		source = aOperand2;
		max = aOperand2MaxIndex;
	}
	if (aOperand2MaxIndex == -1) 
	{
		source = aOperand1;
		max = aOperand1MaxIndex;
	}
	int i;
	if (source != NULL) 
	{
		for (i=0; i<= max; i++) 
		{
			aTarget[i] = source[i];
		}
		aTargetMaxIndex = max;
		return ConstRC::Ok;
	}
	// If both operands are nonempty ...
	int last_written = -1;
	int last_read_1 = 0;
	int last_read_2 = 0;
	int reading_stream_index = 1;
	int compare;
	bool finished = false;
	bool finished_1 = false;
	bool finished_2 = false;
	while (!finished) 
	{		
		// Both operands are unfinished yet. We will write down to target array the lesser entry
		compare = CompareSievingElements((void*) &aOperand1[last_read_1],(void*) &aOperand2[last_read_2]);		
		last_written++;
		if (compare < 0) 
		{
			// The current element in the first stream was lesser than the current
			// element in the second stream.
			if (reading_stream_index == 2) 
			{
				reading_stream_index = 1;
			}
			// We will copy the element from the first stream/
			aTarget[last_written] = aOperand1[last_read_1];
		} 
		else if (compare == 0) 
		{
			// The current elements in stream 1 and stream 2 are identical.
			// We will read them both.
			reading_stream_index = 4;
			aTarget[last_written] = aOperand1[last_read_1];
			aTarget[last_written].exponent += aOperand2[last_read_2].exponent;
			if (aTarget[last_written].flags & ConstEF::DIVIDES_L)
			{
				// we need to count p-valuation of F(a,b) of these elements
				aTarget[last_written].flags += (aOperand2[last_read_2].flags & ConstEF::F_VALUATION_MASK);
			}
		} 
		else 
		{
			// The current element in the first stream was greater than the current
			// element in the second stream.
			if (reading_stream_index == 1) 
			{
				reading_stream_index = 2;
			}
			// Zkopirujeme nizsi prvek do targetu
			aTarget[last_written] = aOperand2[last_read_2];
		}		
		// We will make a read in the currently used stream (1 or 2 or both)
		if ((reading_stream_index == 1) || (reading_stream_index == 4)) 
		{
			last_read_1++;
			if (last_read_1 > aOperand1MaxIndex) 
			{
				finished_1 = true;
			}
		} 	
		if ((reading_stream_index == 2) || (reading_stream_index == 4)) 
		{
			last_read_2++;
			if (last_read_2 > aOperand2MaxIndex) 
			{
				finished_2 = true;
			}
		}		
		if (finished_1 || finished_2) 
		{
			// We will finish the copying process by writing down remaining entries from the unfinished operand
			if (!finished_1) 
			{
				while (last_read_1 <= aOperand1MaxIndex) 
				{
					last_written++;
					aTarget[last_written] = aOperand1[last_read_1];
					last_read_1++;
				}				
			}
			if (!finished_2) 
			{
				while (last_read_2 <= aOperand2MaxIndex) 
				{
					last_written++;
					aTarget[last_written] = aOperand2[last_read_2];
					last_read_2++;
				}				
			}
			finished = true;
		} 
		else 
		{
			if (reading_stream_index == 4) 
			{
				// None of the both streams is finished.
				if (CompareSievingElements( &aOperand1[last_read_1], &aOperand2[last_read_2]) <= 0) 
				{
					reading_stream_index = 1;
				} 
				else 
				{
					reading_stream_index = 2;
				}
			}
		}
	}
	aTargetMaxIndex = last_written;
	return ConstRC::Ok;
}

/*!
 * \brief Compare two "wedge" types.
 *
 * This function is used in qsort calls operating
 * on arrays of graph_wedge_t types.
 *
 */	
int NFSUtils::CompareEdges(const void *aArg1, const void *aArg2)
{
	return (((graph_wedge_t*)aArg1)->weight - ((graph_wedge_t*)aArg2)->weight);
}

/*!
 * \brief Compare two "fb_element" types.
 *
 * This function is used in qsort calls operating
 * on arrays of fb_element types.
 *
 */	
int NFSUtils::CompareFBElements(const void* aArg1, const void* aArg2) 
{
	fb_element* element1 = (fb_element*) aArg1;
	fb_element* element2 = (fb_element*) aArg2;

	int result = 0;
	if ((result = (element1->prime - element2->prime)) != 0) 
	{
		return result;
	}
	else
	{
		return (element1->root - element2->root);
	}

	//else if ((result = (element1->root - element2->root)) != 0)
	//{
	//	return result;
	//}
	//else
	//{
	//	return (element1->flags - element2->flags);
	//}
}


/*!
 * \brief Compare two "e_search_element" types.
 *
 * This function is used in qsort calls operating
 * on arrays of e_search_element types.
 *
 */	
int NFSUtils::CompareESearchElements(const void* aArg1, const void* aArg2) 
{
	//return (int)(((e_search_element*)aArg1)->value - ((e_search_element*)aArg2)->value); //takhle to nefunguje - porovnavam double z int [0,1] => pretypovani na int vzdy vrati 0
	if (((e_search_element*)aArg1)->value > ((e_search_element*)aArg2)->value){
		return 1;
	} else if (((e_search_element*)aArg1)->value > ((e_search_element*)aArg2)->value) {
		return -1;
	} else {
		return 0;
	}
}


