#include "number_field_sieve.h"

#include "phase_creator.h"
#include "const_parameter_names.h"
#include "const_nfs.h"
#include "const_filenames.h"
#include "const_sieving_phase.h"
#include "const_debug.h"
#include "log.h"
# include "xml_nfs_definitions.h"
# include "nfs_classical_sieving_phase.h"
# include "lattice_sieving_phase.h"

# include <typeinfo>  // for std::bad_cast

#ifndef GIT_VERSION
	#define GIT_VERSION ""
#endif
#ifndef VERSION
	#define VERSION "without autotools - 0.8"
#endif

using namespace std;


CNumberFieldSieve::CNumberFieldSieve() : MSerializable(FILE_NAME_NFS), AbstractSerializator()
{
	this->iThreadPool = NULL;
	this->iParameters = NULL;
	this->iPhases = NULL;
	iNodeParameters = NULL;
	iNextSievingRegion = NULL;
	iCenterSievingPhase = NULL;

	if (pthread_mutex_init(&iComputationRunningMutex, NULL) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","CNumberFieldSieve",0,0,"Cannot initialize mutex.");
	pthread_mutex_lock(&iComputationRunningMutex);
	iComputationRunning = false;
	pthread_mutex_unlock(&iComputationRunningMutex);

	if (pthread_mutex_init(&iNodeInstanceRunningMutex, NULL) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","CNumberFieldSieve",0,0,"Cannot initialize mutex.");
	pthread_mutex_lock(&iNodeInstanceRunningMutex);
	iNodeInstanceRunning = false;
	pthread_mutex_unlock(&iNodeInstanceRunningMutex);

	return;
}

void CNumberFieldSieve::ClearPhases() {
	int max, i;

	if (this->iPhases != NULL)
	{
		max = (int)this->iPhases->size();
		for (i = 0; i < max; i++)
		{
			if (this->iPhases->at(i) != NULL)
			{
				delete this->iPhases->at(i);
				this->iPhases->at(i) = NULL;
			}
		}

		this->iPhases->clear();
		delete this->iPhases;
		this->iPhases = NULL;
	}

	return;
}

CNumberFieldSieve::~CNumberFieldSieve()
{
	if (this->iThreadPool != NULL)
	{
		if (!this->iThreadPool->Get_Destroyed())
		{
			// thread pool is still runnig, we have to destroy it
			// we don't want to wait for finishing of the working threads
			this->iThreadPool->Destroy(false);
		}
		delete this->iThreadPool;
		this->iThreadPool = NULL;
	}

	ClearPhases();

	if (iNodeParameters != NULL) delete iNodeParameters;

	if (iNextSievingRegion != NULL) free(iNextSievingRegion);

	if (pthread_mutex_destroy(&iComputationRunningMutex) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","~CNumberFieldSieve",0,0,"Cannot destroy mutex.");

	if (pthread_mutex_destroy(&iNodeInstanceRunningMutex) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","~CNumberFieldSieve",0,0,"Cannot destroy mutex.");

	return;
}

int CNumberFieldSieve::CleanUp(AFactorAlgParameters *aParameters, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CNumberFieldSieve","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	int rc;
	int auxInt;
	aParameters->GetIntegerPhaseParameter(ConstPN::CLEANING_TYPE, (int)ECleanNothing, auxInt, (int)EMain);
	cleaning_types type = (cleaning_types)auxInt;

	if (type != ECleanNothing)
	{
		if (this->iPhases != NULL && this->iPhases->size() > 0)
		{
			// last phase is SquareRootPhase
			if (this->iPhases->at(this->iPhases->size() - 1)->Finished() == false)
			{
				// we are not finished yet - no cleaning
				return ConstRC::Ok;
			}
		}

		APhase * phase = NULL;
		aParameters->GetIntegerPhaseParameter(ConstPN::NFS_POLY_PHASE_TYPE, ConstNFS::DEFAULT_POLY_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		phase->CleanUp(aParameters, (int)EPolySelection, type, aError);
		delete phase;
		phase = NULL;

		aParameters->GetIntegerPhaseParameter(ConstPN::NFS_SIEVING_PHASE_TYPE, ConstNFS::DEFAULT_SIEVING_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		phase->CleanUp(aParameters, (int)ESieving, type, aError);
		delete phase;
		phase = NULL;

		aParameters->GetIntegerPhaseParameter(ConstPN::NFS_REL_PRO_PHASE_TYPE, ConstNFS::DEFAULT_REL_PRO_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		phase->CleanUp(aParameters, (int)ERelProcessing, type, aError);
		delete phase;
		phase = NULL;

		aParameters->GetIntegerPhaseParameter(ConstPN::NFS_LINEAR_PHASE_TYPE, ConstNFS::DEFAULT_LINEAR_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		phase->CleanUp(aParameters, (int)ELinear, type, aError);
		delete phase;
		phase = NULL;

		aParameters->GetIntegerPhaseParameter(ConstPN::NFS_SQUARE_ROOT_PHASE_TYPE, ConstNFS::DEFAULT_SQUARE_ROOT_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		phase->CleanUp(aParameters, (int)ESquareRoot, type, aError);
		delete phase;
		phase = NULL;

		string identifier;
		string directory = "";
		string file;

		if (aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CNumberFieldSieve","CleanUp",0,0,"Didn't find serial identifier.");
			return ConstRC::GeneralError;
		}
		aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, directory, (int)EMain);


		switch (type)
		{
			case ECleanAllIfNotError:
				if (aError)
				{
					break;
				}
			case ECleanAll:
				// Parameters full filename
				if (aParameters->ContainsPhaseParameter(ConstPN::CONTROL_FILE_FULL_FILENAME, (int)EMain))
				{
					aParameters->GetStringPhaseParameter(ConstPN::CONTROL_FILE_FULL_FILENAME, file, (int)EMain);
					MSerializable::AdaptPathSeparator(file);
				}
				else
				{
					aParameters->CreateFullFileName(file, ConstPN::CONTROL_FILE_FILENAME, 
						ConstPN::CONTROL_FILE_DIRECTORY, ConstPN::CONTROL_FILE_COMPRESSION, directory,
						ConstFN::CONTROL_FILE, identifier, (int)EMain);
				}
				remove(file.c_str());
			case ECleanAllExcCF:
			case ECleanAllExcCFRF:
				break;
			default:
				break;
		}
	}

	return ConstRC::Ok;
}
int CNumberFieldSieve::Initialization()
{
	int rc;
	int currentPhase;
	bool auxBool;
	int auxInt;

	this->iParameters->GetIntegerPhaseParameter(ConstPN::CURRENT_PHASE, ConstNFS::START_PHASE_NBR, currentPhase, (int)EMain);

	this->iParameters->GetBoolPhaseParameter(ConstPN::INFO_MODE, false, this->iInfoMode, (int)EMain);

	// log setting
	this->iParameters->GetBoolPhaseParameter(ConstPN::INFO_MODE_TECHNICAL, false, auxBool, (int)EMain);
	CLog::Set_TechnicalMode(auxBool);
	
	// log setting
	this->iParameters->GetBoolPhaseParameter(ConstPN::INFO_MODE_USE_FILE, false, auxBool, (int)EMain);
	if (auxBool)
	{
		string strDirectory = "";
		string strIndentifier = "";
		this->iParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, strDirectory, (int)EMain);
		this->iParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, strIndentifier, (int)EMain);
		CLog::UseFile(strIndentifier, strDirectory);
	}
	else
	{
		CLog::UseConsole();
	}

	// log setting
	this->iParameters->GetIntegerPhaseParameter(ConstPN::INFO_MODE_PRIORITY, CLog::DEFAULT_MESSAGE_PRIORITY, auxInt, (int)EMain);
	CLog::Set_Priority(auxInt);

	bool onePhaseRun; // if only one phase will be run - for debug
	this->iParameters->GetBoolPhaseParameter(ConstPN::ONE_PHASE_RUN, false, onePhaseRun, (int)EMain);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (this->iInfoMode)
	{
		CLog::Log(10000, "0042000", "NFS Factorization algorithm\n");
		CLog::Log(10000, "0042001", "  version: " VERSION " " GIT_VERSION "\n\n");
		CLog::Log(10000, "0042002", "  Running info:\n");
		if (onePhaseRun)
		{
			CLog::Log(10000, "0042003", "    --- ONE PHASE RUN ---\n");
			CLog::Log(10000, "0042004", "    - phase: %d\n", currentPhase);
		}
		else
		{
			CLog::Log(10000, "0042005", "    - starting phase: %d\n", currentPhase);
		}
	}
#endif

	mpz_t numberN;
	mpz_init(numberN);

	// generate random number for factorization?
	bool generate = false;
	this->iParameters->GetBoolPhaseParameter(ConstPN::GENERATOR_GENERATE_N, generate, generate, (int)EMain);
	if (generate)
	{
		int factor1size;
		int factor2size;
		this->iParameters->GetIntegerPhaseParameter(ConstPN::GENERATOR_FACTOR1_SIZE, 
			ConstNFS::FACTOR_SIZE, factor1size, (int)EMain);
		this->iParameters->GetIntegerPhaseParameter(ConstPN::GENERATOR_FACTOR2_SIZE, 
			ConstNFS::FACTOR_SIZE, factor2size, (int)EMain);

		GenerateModulus(numberN, factor1size, factor2size);

		char * auxChar = mpz_get_str(NULL, 10, numberN);
		if (auxChar == NULL)
		{
			mpz_clear(numberN);
			CLog::LogError(WHEREARG,"CNumberFieldSieve","Initialization",0,0,"Didn't convert number N.");
			return ConstRC::GeneralError;
		}

		string auxStr(auxChar);

		free(auxChar);
		auxChar = NULL;

		this->iParameters->SetPhaseParameter(ConstPN::NUMBER_N, auxStr, (int)EMain);
	}
	else
	{
		rc = this->iParameters->GetMpzPhaseParameter(ConstPN::NUMBER_N, numberN, (int)EMain);
		if (rc != ConstRC::Ok)
		{
			mpz_clear(numberN);
			CLog::LogError(WHEREARG,"CNumberFieldSieve","Initialization",0,0,"Didn't find number N.");
			return ConstRC::GeneralError;
		}
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (this->iInfoMode)
	{
		CLog::Log(10000, "0042006", "    - number to factorize: %Zd\n", numberN);
		CLog::Log(10000, "0042007", "      - bit length: %d, dec length: %d\n", (int)mpz_sizeinbase(numberN, 2), (int)mpz_sizeinbase(numberN, 10));
		CLog::Log(10000, "", "\n\n");
	}
#endif

	mpz_clear(numberN);

	this->iPhases = new vector<APhase*>;
	this->iPhases->reserve(ConstNFS::PHASE_COUNT);

	this->iThreadPool = new CThreadPool(ConstNFS::WORKING_THREADS_COUNTS, ConstNFS::PHASE_COUNT, false);

	APhase * phase = NULL;
	if (currentPhase == (int)EPolySelection || (onePhaseRun == false && currentPhase < (int)EPolySelection))
	{
		this->iParameters->GetIntegerPhaseParameter(ConstPN::NFS_POLY_PHASE_TYPE, ConstNFS::DEFAULT_POLY_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iPhases->push_back(phase);
		this->iThreadPool->AddWork(phase, this->iParameters);
		phase = NULL;
	}
	if (currentPhase == (int)ESieving || (onePhaseRun == false && currentPhase < (int)ESieving))
	{
		this->iParameters->GetIntegerPhaseParameter(ConstPN::NFS_SIEVING_PHASE_TYPE, ConstNFS::DEFAULT_SIEVING_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		// in case of one phase run exhausting of sieving region is not an error
		if (onePhaseRun) phase->AcceptReturnCode(ConstRC::SievingRegionExhausted);
		this->iPhases->push_back(phase);
		this->iThreadPool->AddWork(phase, this->iParameters);
		phase = NULL;
	}
	if (currentPhase == (int)ERelProcessing || (onePhaseRun == false && currentPhase < (int)ERelProcessing))
	{
		this->iParameters->GetIntegerPhaseParameter(ConstPN::NFS_REL_PRO_PHASE_TYPE, ConstNFS::DEFAULT_REL_PRO_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iPhases->push_back(phase);
		this->iThreadPool->AddWork(phase, this->iParameters);
		phase = NULL;
	}
	if (currentPhase == (int)ELinear || (onePhaseRun == false && currentPhase < (int)ELinear))
	{
		this->iParameters->GetIntegerPhaseParameter(ConstPN::NFS_LINEAR_PHASE_TYPE, ConstNFS::DEFAULT_LINEAR_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iPhases->push_back(phase);
		this->iThreadPool->AddWork(phase, this->iParameters);
		phase = NULL;
	}
	if (currentPhase == (int)ESquareRoot || (onePhaseRun == false && currentPhase < (int)ESquareRoot))
	{
		this->iParameters->GetIntegerPhaseParameter(ConstPN::NFS_SQUARE_ROOT_PHASE_TYPE, ConstNFS::DEFAULT_SQUARE_ROOT_PHASE, auxInt, (int)EMain);
		rc = CPhaseCreator::CreatePhase((phase_types)auxInt, phase);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iPhases->push_back(phase);
		this->iThreadPool->AddWork(phase, this->iParameters);
		phase = NULL;
	}

	return ConstRC::Ok;
}

int CNumberFieldSieve::RunFactorization(AFactorAlgParameters *aParameters)
{
	this->iParameters = aParameters;

	int rc = Initialization();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// this thread will wait for the end
	rc = this->iThreadPool->Destroy(true);
	CleanUp(this->iParameters, this->iThreadPool->Get_InnerError());

#ifndef MEASUREMENT_ON
	// we do not check errors if we just do measurement
	if (this->iThreadPool->Get_InnerError())
	{
		CLog::LogError(WHEREARG,"CNumberFieldSieve","RunFactorization",0,0,"Some error in the phase.");
		return ConstRC::GeneralError;
	}
#endif

	return rc;
}

// method assumes that iParameters were already initialized by SetupParameters
int CNumberFieldSieve::RunFactorization() {

	return RunFactorization(iParameters);
}

int CNumberFieldSieve::Serialize(const CBaseParameters& aParam) const
{
	return MSerializable::Serialize(aParam);
}
int CNumberFieldSieve::Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const
{
	return ConstRC::NotSupported;
}
int CNumberFieldSieve::Deserialize(const CBaseParameters& aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CNumberFieldSieve::Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader)
{
	return ConstRC::NotSupported;
}

/*
	Algorithm methods.
*/

const char* CNumberFieldSieve::AlgorithmName() const {
	return "NFS";
}

JobParameters *CNumberFieldSieve::CreateNewParameters() const {

	return new CNFSParameters();
}

int CNumberFieldSieve::SetupParameters(const JobParameters &aParameters) {

	if (iParameters != NULL) {
		delete iParameters;
		iParameters = NULL;
	}

        try {
                const CNFSParameters &np = dynamic_cast<const CNFSParameters &>(aParameters);
                iParameters = (AFactorAlgParameters *) new CNFSParameters(np);
        }
        catch (const std::bad_cast & e) {
                CLog::LogError(WHEREARG,"NumberFieldSieve","SetupParameters",0,0,"Unknown subtype of JobParameters, doing nothing.");
                return ConstRC::GeneralError;
        }

	return ConstRC::Ok;
}

int CNumberFieldSieve::DeleteParameters(const JobParameters *aParameters) {

	try {
                const CNFSParameters *np = dynamic_cast<const CNFSParameters *>(aParameters);
                delete np;
        }
        catch (const std::bad_cast & e) {
                CLog::LogError(WHEREARG,"NumberFieldSieve","DeleteParameters",0,0,"Unknown subtype of JobParameters, doing nothing.");
                return ConstRC::GeneralError;
        }

	return ConstRC::Ok;
}

/*
	DistributedAlgorithm methods.
*/

Batch CNumberFieldSieve::CreateNewJobs(unsigned int aBatchSize) {
	Batch batch;
	CNFSParameters *np;
	char buff[100];
	int i;

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","CreateNewJobs",0,0,"Parameters were not initialized.");
		return batch;
	}

	// the first batch
	if (iNextSievingRegion == NULL) {
		// initialize next sieving region
		iNextSievingRegion = new sieving_region;
		if (iParameters->GetLongPhaseParameter(ConstPN::SIEVING_REGION_WIDTH, iNextSievingRegion->width, (int)EMain) != ConstRC::Ok) {
        	        CLog::LogError(WHEREARG,"CNumberFieldSieve","CreateNewJobs",0,0,"Didn't find the sieving region width.");
                	return batch;
	        }
	        if (iParameters->GetLongPhaseParameter(ConstPN::SIEVING_REGION_HEIGHT, iNextSievingRegion->height, (int)EMain) != ConstRC::Ok) {
	                CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateNewJobs", 0, 0, "Didn't find the sieving region height.");
			return batch;
        	}
		if (iNextSievingRegion->height == SIEVING_REGION_NO_HEIGHT_BOUND) {
	                CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateNewJobs", 0, 0, "Sieving region has no height bound.");
			return batch;
		}
	        if (iParameters->GetLongPhaseParameter(ConstPN::SIEVING_REGION_START, iNextSievingRegion->start, (int)EMain) != ConstRC::Ok) iNextSievingRegion->start = SIEVING_REGION_DEFAULT_START;

		// clear phases
		ClearPhases();

		// set phase numbers
		SetPhaseNumbers();

		// set serialization identifier
		iParameters->SetPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, MSerializable::CreateNewIdentifier(), (int)EMain);

		// set cleaning type
		sprintf(buff, "%d", (int)ECleanNothing);
		iParameters->SetPhaseParameter(ConstPN::CLEANING_TYPE, buff, (int)EMain);

		// select polynomials
		if (SelectPolynomials() != ConstRC::Ok) {
	                CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateNewJobs", 0, 0, "Select polynomials failed.");
			return batch;		
		}

		// set common node parameters
		if (iNodeParameters != NULL) {
			delete iNodeParameters;
			iNodeParameters = NULL;
		}
		try {
			np = dynamic_cast<CNFSParameters *>(iParameters);
		}
		catch (const std::bad_cast & e) {
			CLog::LogError(WHEREARG,"CNumberFieldSieve","CreateNewJobs",0,0,"Unknown subtype of AFactorAlgParameters.");
			return batch;
		}
		iNodeParameters = new CNFSParameters(*np);

		// create sieving phase
		if (CreateSievingPhase() != ConstRC::Ok) {
	                CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateNewJobs", 0, 0, "Create sieving phase failed.");
			return batch;		
		}

		// set computation running
		pthread_mutex_lock(&iComputationRunningMutex);
		iComputationRunning = true;
		pthread_mutex_unlock(&iComputationRunningMutex);
	}

	if (iNodeParameters == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","CreateNewJobs",0,0,"Node parameters were not initialized.");
		return batch;
	}

	if (IsDistributedPhaseFinished()) return batch;

	for (i = 0; i < aBatchSize; i++) {
		np = new CNFSParameters(*iNodeParameters);
		// set sieving region
		sprintf(buff, "%ld", iNextSievingRegion->width);
		np->SetPhaseParameter(ConstPN::SIEVING_REGION_WIDTH, buff, (int)EMain);
		sprintf(buff, "%ld", iNextSievingRegion->height);
		np->SetPhaseParameter(ConstPN::SIEVING_REGION_HEIGHT, buff, (int)EMain);
		sprintf(buff, "%ld", iNextSievingRegion->start);
		np->SetPhaseParameter(ConstPN::SIEVING_REGION_START, buff, (int)EMain);
		batch.push_back(np);
		iNextSievingRegion->start += iNextSievingRegion->height;
	}

	return batch;
}

void CNumberFieldSieve::RunNodeInstance() {
	char buff[10];

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNodeInstance", 0, 0, "Parameters were not initialized.");
		return;
	}

	ClearPhases();

	// set phase numbers
	SetPhaseNumbers();

	// set serialization identifier
	iParameters->SetPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, MSerializable::CreateNewIdentifier(), (int)EMain);

	// order sieving phase
	sprintf(buff, "%d", (int)ESieving);
	iParameters->SetPhaseParameter(ConstPN::CURRENT_PHASE, buff, (int)EMain);
	iParameters->SetPhaseParameter(ConstPN::ONE_PHASE_RUN, "true", (int)EMain);

	// set node instance running
	pthread_mutex_lock(&iNodeInstanceRunningMutex);
        iNodeInstanceRunning = true;
        pthread_mutex_unlock(&iNodeInstanceRunningMutex);

	// run sieving phase
	if (RunFactorization() != ConstRC::Ok) CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNodeInstance", 0, 0, "Run factorization failed.");

	delete iThreadPool;
	iThreadPool = NULL;

	// set node instance not running
	pthread_mutex_lock(&iNodeInstanceRunningMutex);
        iNodeInstanceRunning = false;
        pthread_mutex_unlock(&iNodeInstanceRunningMutex);

	return;
}

bool CNumberFieldSieve::NodeInstanceRunning() const {
	bool running;

	pthread_mutex_lock(&iNodeInstanceRunningMutex);
        running = iNodeInstanceRunning;
        pthread_mutex_unlock(&iNodeInstanceRunningMutex);

	return running;
}

bool CNumberFieldSieve::HasDataToSend() const {

	return DataToSend();
}

unsigned int CNumberFieldSieve::DataToSend() const {
	int smooths_count, partials_count;

	if (iParameters == NULL) return 0;

	if (NodeInstanceRunning()) return 0;

	smooths_count = 0;
	if (iParameters->GetIntegerPhaseParameter(ConstPN::SMOOTH_REL_COUNT, smooths_count, (int)ERelProcessing) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","DataToSend",0,0,"Missing smooth relations count.");

	partials_count = 0;
	if (iParameters->GetIntegerPhaseParameter(ConstPN::PARTIAL_REL_GRAPH_EDGES, partials_count, (int)ERelProcessing) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","DataToSend",0,0,"Missing partial relations count.");

	return smooths_count+partials_count;
}

const char* CNumberFieldSieve::DataUnit() const {
	return "relations";
}

void CNumberFieldSieve::LockDataMutex() {

	while (NodeInstanceRunning()) SLEEP(3);

	return;
}

void CNumberFieldSieve::UnlockDataMutex() {
	return;
}

void CNumberFieldSieve::ClearFreshData() {

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","ClearFreshData",0,0,"Parameters were not initialized.");
		return;
	}

	iParameters->SetPhaseParameter(ConstPN::SMOOTH_REL_COUNT, "0", (int)ERelProcessing);
	iParameters->SetPhaseParameter(ConstPN::PARTIAL_REL_GRAPH_EDGES, "0", (int)ERelProcessing);

	return;
}

bool CNumberFieldSieve::IsDistributedPhaseFinished() const {
	int sieving_phase_type;

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "IsDistributedPhaseFinished", 0, 0, "Parameters were not initialized.");
		return true;
	}

	if (iCenterSievingPhase == NULL) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "IsDistributedPhaseFinished", 0, 0, "Center sieving phase was not initialized.");
		return true;
	}	

	if (iParameters->GetIntegerPhaseParameter(ConstPN::NFS_SIEVING_PHASE_TYPE, ConstNFS::DEFAULT_SIEVING_PHASE, sieving_phase_type, (int)EMain) != ConstRC::Ok) sieving_phase_type = ConstNFS::DEFAULT_SIEVING_PHASE;

	switch (sieving_phase_type) {
		case ENFSClassicalSieving:
			return (((CNFSClassicalSievingPhase *)iCenterSievingPhase)->CheckEnough() == ConstRC::EnoughSmooths);
		case ENFSLatticeSieving:
			((CLatticeSievingPhase *)iCenterSievingPhase)->WalkWithSpecialQ(iNextSievingRegion->start);
			((CLatticeSievingPhase *)iCenterSievingPhase)->DetermineRequestedRelations();
			return (((CLatticeSievingPhase *)iCenterSievingPhase)->CheckEnough() == ConstRC::EnoughSmooths);
		default:
			CLog::LogError(WHEREARG, "CNumberFieldSieve", "IsDistributedPhaseFinished", 0, 0, "Sieving phase type is not implemented in distributed mode.");
			return true;
	}
}

int CNumberFieldSieve::RunNondistributedPhase() {
	int sieving_phase_type, rc;
	char buff[10];

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNondistributedPhase", 0, 0, "Parameters were not initialized.");
		return ConstRC::GeneralError;
	}

	if (iCenterSievingPhase == NULL) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNondistributedPhase", 0, 0, "Center sieving phase was not initialized.");
		return ConstRC::GeneralError;
	}

	if (iParameters->GetIntegerPhaseParameter(ConstPN::NFS_SIEVING_PHASE_TYPE, ConstNFS::DEFAULT_SIEVING_PHASE, sieving_phase_type, (int)EMain) != ConstRC::Ok) sieving_phase_type = ConstNFS::DEFAULT_SIEVING_PHASE;

	switch (sieving_phase_type) {
		case ENFSClassicalSieving:
			((CNFSClassicalSievingPhase *)iCenterSievingPhase)->ConfirmAllStats();
			if (((CNFSClassicalSievingPhase *)iCenterSievingPhase)->SaveResult() != ConstRC::Ok) {
				CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNondistributedPhase", 0, 0, "Save center sieving phase result failed.");
				return ConstRC::GeneralError;
			}
			break;
		case ENFSLatticeSieving:
			((CLatticeSievingPhase *)iCenterSievingPhase)->ConfirmAllStats();
			((CLatticeSievingPhase *)iCenterSievingPhase)->WalkWithSpecialQ(iNextSievingRegion->start);
			if (((CLatticeSievingPhase *)iCenterSievingPhase)->SaveResult() != ConstRC::Ok) {
				CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNondistributedPhase", 0, 0, "Save center sieving phase result failed.");
				return ConstRC::GeneralError;
			}
			break;
		default:
			CLog::LogError(WHEREARG, "CNumberFieldSieve", "RunNondistributedPhase", 0, 0, "Sieving phase type is not implemented in distributed mode.");
			return ConstRC::GeneralError;
	}

	// set remaining phases
	sprintf(buff, "%d", (int)ERelProcessing);
	iParameters->SetPhaseParameter(ConstPN::CURRENT_PHASE, buff, (int)EMain);
	iParameters->SetPhaseParameter(ConstPN::ONE_PHASE_RUN, "false", (int)EMain);

	if ((rc = RunFactorization()) != ConstRC::Ok) CLog::LogError(WHEREARG,"CNumberFieldSieve","RunNondistributedPhase",0,0,"Run factorization failed.");

	delete iThreadPool;
	iThreadPool = NULL;

	pthread_mutex_lock(&iComputationRunningMutex);
	iComputationRunning = false;
	pthread_mutex_unlock(&iComputationRunningMutex);

	return rc;
}

void CNumberFieldSieve::InterruptNodeNow() {

	if (iThreadPool == NULL) return;

	if (!(iThreadPool->Get_Destroyed())) iThreadPool->Destroy(false);

	return;
}

bool CNumberFieldSieve::IsComputationFinished() {
	bool val;

	pthread_mutex_lock(&iComputationRunningMutex);
	val = iComputationRunning;
	pthread_mutex_unlock(&iComputationRunningMutex);

	return !val;
}

void CNumberFieldSieve::SetComputationFinished(bool aValue) {
	return;
}

void CNumberFieldSieve::ResetCenter() {

	ClearPhases();

	if (iParameters != NULL) {
		delete iParameters;
		iParameters = NULL;
	}


	if (iNodeParameters != NULL) {
		delete iNodeParameters;
		iNodeParameters = NULL;
	}

	if (iNextSievingRegion != NULL) {
		free(iNextSievingRegion);
		iNextSievingRegion = NULL;
	}

	if (iCenterSievingPhase != NULL) {
		delete iCenterSievingPhase;
		iCenterSievingPhase = NULL;
	}

	pthread_mutex_lock(&iComputationRunningMutex);
	iComputationRunning = false;
	pthread_mutex_unlock(&iComputationRunningMutex);

	return;
}

/*
	AbstractSerializator methods.
*/

int CNumberFieldSieve::Serialize(int aAction) {
	return 0;
}

int CNumberFieldSieve::Deserialize(int aAction) {
	return 0;
}

int CNumberFieldSieve::SerializeResult() {
	return 0;
}

int CNumberFieldSieve::SerializeJob(xmlTextWriterPtr &aWriter, const JobParameters &aParameters) {

	// parent element
        if (xmlTextWriterStartElement(aWriter, BAD_CAST JOB_DATA) < 0) {
                CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeJob",0,0,"Could not write the serializator tag.");
                return ConstRC::GeneralError;
        }
	// newline
        if (CommonXMLHelper::NewLine(aWriter,FALSE) != ConstRC::Ok) return ConstRC::GeneralError;
	// serializator info
        if (CommonXMLHelper::WriteSerializatorData(aWriter,SERIALIZATOR_VAL, SERIALIZATOR_VERSION) != ConstRC::Ok) return ConstRC::GeneralError;
	// time
        if (CommonXMLHelper::SerializeTime(aWriter) != ConstRC::Ok) return ConstRC::GeneralError;

	try {
		const CNFSParameters &np = dynamic_cast<const CNFSParameters &>(aParameters);
		CBaseParameters bp;
		if (np.Serialize(bp, aWriter) != ConstRC::Ok) return ConstRC::GeneralError;
        }
        catch (const std::bad_cast & e) {
                CLog::LogError(WHEREARG,"NumberFieldSieve","SerializeJob",0,0,"Unknown subtype of JobParameters, doing nothing.");
                return ConstRC::GeneralError;
        }

	// close parent element
	if (xmlTextWriterEndElement(aWriter) < 0) {
                CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeJob",0,0,"Could not close parent element.");
                return ConstRC::GeneralError;
        }
	// newline
        if (CommonXMLHelper::NewLine(aWriter,FALSE) != ConstRC::Ok) return ConstRC::GeneralError;

	return ConstRC::Ok;
}

int CNumberFieldSieve::DeserializeJob(xmlTextReaderPtr &aReader, JobParameters &aParameters) {
	xmlChar *name, *value;
	int depth;
	bool name_found, param_found;

	CommonXMLHelper::MoveToNextNode(aReader);

	if ((name = xmlTextReaderName(aReader)) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Deserialization main element name not found.");
                return ConstRC::GeneralError;
	}
	if (!xmlStrEqual(name, BAD_CAST JOB_DATA)) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Wrong deserialization main element name.");
		xmlFree(name);
                return ConstRC::GeneralError;
	}
	xmlFree(name);

	depth = xmlTextReaderDepth(aReader);

	name_found = false;
	param_found = false;
	while (CommonXMLHelper::MoveToNextNode(aReader, depth) == ConstRC::Ok) {
		if ((name = xmlTextReaderName(aReader)) == NULL) continue;

		if (xmlStrEqual(name, BAD_CAST SERIALIZATOR_NAME)) {
			name_found = true;
			if ((value = xmlTextReaderGetAttribute(aReader, BAD_CAST SERIALIZATOR_ATTR_NAME)) == NULL) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Missing serializator name.");
				xmlFree(name);
				continue;
			}
			if (!xmlStrEqual(value, BAD_CAST SERIALIZATOR_VAL)) CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Unknown serializator name.");
			xmlFree(name);
			xmlFree(value);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_NFS_PARAM_ELEMENT_NAME)) {
			param_found = true;
			try {
				CNFSParameters &np = dynamic_cast<CNFSParameters &>(aParameters);
				CBaseParameters bp;
				if (np.Deserialize(bp, aReader) != ConstRC::Ok) {
					xmlFree(name);
					return ConstRC::GeneralError;
				}
		        }
		        catch (const std::bad_cast & e) {
                		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Unknown subtype of JobParameters, doing nothing.");
				xmlFree(name);
		                return ConstRC::GeneralError;
		        }
			xmlFree(name);
			continue;
		}
	}

	if (!name_found) CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Missing serializator tag.");

	if (!param_found) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Missing parameters tag.");
		return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}

int CNumberFieldSieve::DeserializeJob(const char *aPath, JobParameters &aParameters) {
	xmlTextReaderPtr reader;
	int rc;

	if ((reader = xmlNewTextReaderFilename(aPath)) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeJob",0,0,"Could not create xml reader.");
                return ConstRC::GeneralError;
	}

	rc = DeserializeJob(reader, aParameters);

	xmlFreeTextReader(reader);

	return rc;
}

int CNumberFieldSieve::DeserializeJob(JobParameters &aParameters) {
	return 0;
}

int CNumberFieldSieve::SerializeData(xmlTextWriterPtr &aWriter) {
	FILE *f;
	string rel_file, identifier;
	int i;
	char buff[IO_BUFF_SIZE];

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Parameters were not initialized.");
                return ConstRC::GeneralError;
	}

	if (iParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Missing serialization identifier.");
                return ConstRC::GeneralError;
	}

	// main element
	if (xmlTextWriterStartElement(aWriter, (xmlChar *)"relations") < 0) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not write element.");
                return ConstRC::GeneralError;
	}
	if (CommonXMLHelper::NewLine(aWriter,FALSE) != ConstRC::Ok) return ConstRC::GeneralError;

	// smooth
	if (xmlTextWriterStartElement(aWriter, (xmlChar *)"smooth") < 0) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not write element.");
                return ConstRC::GeneralError;
	}

	iParameters->CreateFullFileName(rel_file, ConstPN::SMOOTH_RELATIONS_FILENAME, ConstPN::SMOOTH_RELATIONS_DIRECTORY, ConstPN::SMOOTH_RELATIONS_COMPRESSION, "", ConstFN::SMOOTH_RELATIONS, identifier, (int)ERelProcessing, TXT_FILE_EXTENSION);
	MSerializable::AdaptPathSeparator(rel_file);

	if ((f = fopen(rel_file.c_str(), "rb")) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not open file.");
                return ConstRC::GeneralError;
	}
	do {
		i = fread(buff, sizeof(char), IO_BUFF_SIZE, f);
		xmlTextWriterWriteRawLen(aWriter, (xmlChar *)buff, i);
	}
	while (i == IO_BUFF_SIZE);
	if (!feof(f)) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"File read error.");
		fclose(f);
                return ConstRC::GeneralError;
	}
	fclose(f);

	if (xmlTextWriterEndElement(aWriter) < 0) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not write element.");
                return ConstRC::GeneralError;
	}
	if (CommonXMLHelper::NewLine(aWriter,FALSE) != ConstRC::Ok) return ConstRC::GeneralError;

	//partial
	if (xmlTextWriterStartElement(aWriter, (xmlChar *)"partial") < 0) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not write element.");
                return ConstRC::GeneralError;
	}

	iParameters->CreateFullFileName(rel_file, ConstPN::PARTIAL_REL_FILENAME, ConstPN::PARTIAL_REL_DIRECTORY, ConstPN::PARTIAL_REL_COMPRESSION, "", ConstFN::PARTIAL_REL, identifier, (int)ERelProcessing, TXT_FILE_EXTENSION);
	MSerializable::AdaptPathSeparator(rel_file);

	if ((f = fopen(rel_file.c_str(), "rb")) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not open file.");
                return ConstRC::GeneralError;
	}
	do {
		i = fread(buff, sizeof(char), IO_BUFF_SIZE, f);
		xmlTextWriterWriteRawLen(aWriter, (xmlChar *)buff, i);
	}
	while (i == IO_BUFF_SIZE);
	if (!feof(f)) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"File read error.");
		fclose(f);
                return ConstRC::GeneralError;
	}
	fclose(f);

	if (xmlTextWriterEndElement(aWriter) < 0) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not write element.");
                return ConstRC::GeneralError;
	}
	if (CommonXMLHelper::NewLine(aWriter,FALSE) != ConstRC::Ok) return ConstRC::GeneralError;

	// end main element
	if (xmlTextWriterEndElement(aWriter) < 0) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SerializeData",0,0,"Could not write element.");
                return ConstRC::GeneralError;
	}
	if (CommonXMLHelper::NewLine(aWriter,FALSE) != ConstRC::Ok) return ConstRC::GeneralError;

	return ConstRC::Ok;
}

int CNumberFieldSieve::DeserializeData(const char *aPath) {
	int sieving_phase_type, len, rels_count;
	xmlTextReaderPtr reader;
	xmlNodePtr node;
	xmlChar *name, *content;
	FILE *f;
	string smooths_file, partials_file;

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Parameters were not initialized.");
                return ConstRC::GeneralError;
	}

	if (iCenterSievingPhase == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Center sieving phase was not initialized.");
                return ConstRC::GeneralError;	
	}

	if (IsDistributedPhaseFinished()) return ConstRC::Ok;

	// first create empty files
	smooths_file = string(aPath)+".smooth";
	if ((f = fopen(smooths_file.c_str(), "wb")) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not open file.");
                return ConstRC::GeneralError;
	}
	fclose(f);
	partials_file = string(aPath)+".partial";
	if ((f = fopen(partials_file.c_str(), "wb")) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not open file.");
                return ConstRC::GeneralError;
	}
	fclose(f);

	// fill files with relations
	if ((reader = xmlNewTextReaderFilename(aPath)) == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not create xml reader.");
                return ConstRC::GeneralError;
	}
	while (CommonXMLHelper::MoveToNextNode(reader) == ConstRC::Ok) {
		if ((name = xmlTextReaderName(reader)) == NULL) continue;

		if (xmlStrEqual(name, BAD_CAST "smooth")) {
			xmlFree(name);
			if ((node = xmlTextReaderExpand(reader)) == NULL) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not expand node.");
				xmlFreeTextReader(reader);
                		return ConstRC::GeneralError;
			}
			if ((content = xmlNodeGetContent(node)) == NULL) continue;
			if ((f = fopen(smooths_file.c_str(), "wb")) == NULL) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not open file.");
				xmlFree(content);
				xmlFreeTextReader(reader);
                		return ConstRC::GeneralError;
			}
			len = xmlStrlen(content);
			if (fwrite(content, sizeof(xmlChar), len, f) != len) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"File write failed.");
				fclose(f);
				xmlFree(content);
				xmlFreeTextReader(reader);
                		return ConstRC::GeneralError;
			}
			fclose(f);
			xmlFree(content);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST "partial")) {
			xmlFree(name);
			if ((node = xmlTextReaderExpand(reader)) == NULL) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not expand node.");
				xmlFreeTextReader(reader);
                		return ConstRC::GeneralError;
			}
			if ((content = xmlNodeGetContent(node)) == NULL) continue;
			if ((f = fopen(partials_file.c_str(), "wb")) == NULL) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Could not open file.");
				xmlFree(content);
				xmlFreeTextReader(reader);
                		return ConstRC::GeneralError;
			}
			len = xmlStrlen(content);
			if (fwrite(content, sizeof(xmlChar), len, f) != len) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"File write failed.");
				fclose(f);
				xmlFree(content);
				xmlFreeTextReader(reader);
                		return ConstRC::GeneralError;
			}
			fclose(f);
			xmlFree(content);
			continue;
		}

		xmlFree(name);
	}
	xmlFreeTextReader(reader);

	// read relations from file
	if (iParameters->GetIntegerPhaseParameter(ConstPN::NFS_SIEVING_PHASE_TYPE, ConstNFS::DEFAULT_SIEVING_PHASE, sieving_phase_type, (int)EMain) != ConstRC::Ok) sieving_phase_type = ConstNFS::DEFAULT_SIEVING_PHASE;

	switch (sieving_phase_type) {
		case ENFSClassicalSieving:
			if (((CNFSClassicalSievingPhase *)iCenterSievingPhase)->SupplyRelations(smooths_file.c_str(), true, rels_count) != ConstRC::Ok) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Supply smooth relations failed.");
        	        	return ConstRC::GeneralError;
			}
			fprintf(stdout, "\nSupply relations from file \"%s\": %d smooth relations\n", aPath, rels_count);
			if (((CNFSClassicalSievingPhase *)iCenterSievingPhase)->SupplyRelations(partials_file.c_str(), false, rels_count) != ConstRC::Ok) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Supply partial relations failed.");
        	        	return ConstRC::GeneralError;
			}
			fprintf(stdout, "Supply relations from file \"%s\": %d partial relations", aPath, rels_count);
			((CNFSClassicalSievingPhase *)iCenterSievingPhase)->PrintHeader();
			((CNFSClassicalSievingPhase *)iCenterSievingPhase)->PrintValues();
			fprintf(stdout, "\n\n");
			break;
		case ENFSLatticeSieving:
			if (((CLatticeSievingPhase *)iCenterSievingPhase)->SupplyRelations(smooths_file.c_str(), true, rels_count) != ConstRC::Ok) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Supply smooth relations failed.");
        	        	return ConstRC::GeneralError;
			}
			fprintf(stdout, "\nSupply relations from file \"%s\": %d smooth relations\n", aPath, rels_count);
			if (((CLatticeSievingPhase *)iCenterSievingPhase)->SupplyRelations(partials_file.c_str(), false, rels_count) != ConstRC::Ok) {
				CLog::LogError(WHEREARG,"CNumberFieldSieve","DeserializeData",0,0,"Supply partial relations failed.");
        	        	return ConstRC::GeneralError;
			}
			fprintf(stdout, "Supply relations from file \"%s\": %d partial relations", aPath, rels_count);
			((CLatticeSievingPhase *)iCenterSievingPhase)->PrintHeader();
			((CLatticeSievingPhase *)iCenterSievingPhase)->PrintValues();
			fprintf(stdout, "\n\n");
			break;
		default:
			CLog::LogError(WHEREARG, "CNumberFieldSieve", "DeserializeData", 0, 0, "Sieving phase type is not implemented in distributed mode.");
			return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}

Algorithm *CNumberFieldSieve::CreateInstance(const JobParameters &aParameters) const {
	CNumberFieldSieve *nfs;

	nfs = new CNumberFieldSieve();
	if (nfs->SetupParameters(aParameters) != ConstRC::Ok) return NULL;

	return nfs;
}

void CNumberFieldSieve::RegisterInstance(Algorithm *aAlgorithm) {
	iAlgorithmInstance = aAlgorithm;

	return;
}

/*
	Other methods for distributed case.
*/

// sets phase numbers from predefined values
void CNumberFieldSieve::SetPhaseNumbers() {
	char buff[10];

	sprintf(buff, "%d", (int)EPolySelection);
	iParameters->SetPhaseParameter(ConstPN::POLY_PHASE_NUMBER, buff, (int)EMain);
	sprintf(buff, "%d", (int)ESieving);
	iParameters->SetPhaseParameter(ConstPN::SIEVING_PHASE_NUMBER, buff, (int)EMain);
	sprintf(buff, "%d", (int)ERelProcessing);
	iParameters->SetPhaseParameter(ConstPN::REL_PROCESSING_PHASE_NUMBER, buff, (int)EMain);
	sprintf(buff, "%d", (int)ELinear);
	iParameters->SetPhaseParameter(ConstPN::LINEAR_PHASE_NUMBER, buff, (int)EMain);
	sprintf(buff, "%d", (int)ESquareRoot);
	iParameters->SetPhaseParameter(ConstPN::SQUARE_ROOT_PHASE_NUMBER, buff, (int)EMain);

	return;
}

// find polynomials for sieving phase (called by center)
int CNumberFieldSieve::SelectPolynomials() {
	char buff[10];
	int rc;

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG,"CNumberFieldSieve","SelectPolynomials",0,0,"Parameters were not initialized.");
		return ConstRC::GeneralError;
	}

	sprintf(buff, "%d", (int)EPolySelection);
	iParameters->SetPhaseParameter(ConstPN::CURRENT_PHASE, buff, (int)EMain);
	iParameters->SetPhaseParameter(ConstPN::ONE_PHASE_RUN, "true", (int)EMain);

	rc = ConstRC::Ok;

	if ((rc = RunFactorization()) != ConstRC::Ok) CLog::LogError(WHEREARG, "CNumberFieldSieve", "SelectPolynomials", 0, 0, "Run factorization failed.");

	delete iThreadPool;
	iThreadPool = NULL;

	return rc;
}

// create sieving phase and generate factor bases (called by center)
int CNumberFieldSieve::CreateSievingPhase() {
	int sieving_phase_type;
	char buff[10];

	if (iParameters == NULL) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateSievingPhase", 0, 0, "Parameters were not initialized.");
		return ConstRC::GeneralError;
	}

	// order sieving phase
	sprintf(buff, "%d", (int)ESieving);
	iParameters->SetPhaseParameter(ConstPN::CURRENT_PHASE, buff, (int)EMain);
	iParameters->SetPhaseParameter(ConstPN::ONE_PHASE_RUN, "true", (int)EMain);

	// create phase
	if (iCenterSievingPhase != NULL) {
		delete iCenterSievingPhase;
		iCenterSievingPhase = NULL;
	}
	if (iParameters->GetIntegerPhaseParameter(ConstPN::NFS_SIEVING_PHASE_TYPE, ConstNFS::DEFAULT_SIEVING_PHASE, sieving_phase_type, (int)EMain) != ConstRC::Ok) sieving_phase_type = ConstNFS::DEFAULT_SIEVING_PHASE;
	if (CPhaseCreator::CreatePhase((phase_types)sieving_phase_type, iCenterSievingPhase) != ConstRC::Ok) {
		CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateSievingPhase", 0, 0, "Could not create phase.");
		return ConstRC::GeneralError;
	}

	// initialize phase and generate factor bases
	switch (sieving_phase_type) {
		case ENFSClassicalSieving:
			((CNFSClassicalSievingPhase *)iCenterSievingPhase)->InitDistributedSieving(iParameters);
			break;
		case ENFSLatticeSieving:
			((CLatticeSievingPhase *)iCenterSievingPhase)->InitDistributedSieving(iParameters);
			break;
		default:
			
			delete iCenterSievingPhase;
			iCenterSievingPhase = NULL;
			CLog::LogError(WHEREARG, "CNumberFieldSieve", "CreateSievingPhase", 0, 0, "Sieving phase type is not implemented in distributed mode.");
			return ConstRC::GeneralError;
	}

	return ConstRC::Ok;
}

