#ifndef __NUMBER_FIELD_SIEVE_H__
#define __NUMBER_FIELD_SIEVE_H__

#include "abstract_factor_alg.h"
# include "distributed_algorithm.h"
# include "abstract_serializator.h"
# include "nfs_parameters.h"

/*!
 * \brief Number field sieve algorithm
 */
class CNumberFieldSieve : public AFactorAlg, public virtual DistributedAlgorithm, public AbstractSerializator
{
private:
	// common parameters for nodes
	CNFSParameters *iNodeParameters;
	// for distributed case (center is assigning disjoint sieving regions)
	sieving_region *iNextSievingRegion;
	// center sieving phase
	APhase *iCenterSievingPhase;
	// set by center in CreateNewJobs() method
	bool iComputationRunning;
	mutable pthread_mutex_t iComputationRunningMutex;
	// set by node in RunNodeInstance() method
	bool iNodeInstanceRunning;
	mutable pthread_mutex_t iNodeInstanceRunningMutex;

protected:
	///Field with all phases
	std::vector<APhase*>* iPhases;


protected:
	
	///Init all parameters necessary for calculation - get from iParameters
	int Initialization();

	///Clena up all used files
	int CleanUp(AFactorAlgParameters * aParameters, bool aError);

	// release iPhases
	void ClearPhases();

public:
	CNumberFieldSieve();
	~CNumberFieldSieve();

	int RunFactorization(AFactorAlgParameters * aParameters);
	int RunFactorization();  // method assumes that iParameters were already initialized by SetupParameters

	// De/Serialization methods
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);

	// Algorithm methods
	const char* AlgorithmName() const;
	JobParameters* CreateNewParameters() const;
	int SetupParameters(const JobParameters &aParameters);

	// DistributedAlgorithm methods
	Batch CreateNewJobs(unsigned int aBatchSize);
	void RunNodeInstance();
	bool NodeInstanceRunning() const;
	bool HasDataToSend() const;
	unsigned int DataToSend() const;
	const char* DataUnit() const;
	void LockDataMutex();
	void UnlockDataMutex();
	void ClearFreshData();
	bool IsDistributedPhaseFinished() const;
	int RunNondistributedPhase();
	void InterruptNodeNow();
	bool IsComputationFinished();
	void SetComputationFinished(bool);
	void ResetCenter();
	int DeleteParameters(const JobParameters *aParameters);

	// AbstractSerializator methods
	virtual int Serialize(int aAction);
	virtual int Deserialize(int aAction);
	virtual int SerializeResult();
	virtual int SerializeJob(xmlTextWriterPtr &aWriter, const JobParameters &aParameters);
	virtual int DeserializeJob(xmlTextReaderPtr &aReader, JobParameters &aParameters);
	virtual int DeserializeJob(const char *aPath, JobParameters &aParameters);
	virtual int DeserializeJob(JobParameters &aParameters);
	virtual int SerializeData(xmlTextWriterPtr &aWriter);
	virtual int DeserializeData(const char *aPath);
	virtual Algorithm* CreateInstance(const JobParameters &aParameters) const;
	virtual void RegisterInstance(Algorithm *aAlgorithm);

private:

	// other methods for distributed case

	// sets phase numbers from predefined values
	void SetPhaseNumbers();
	// find polynomials for sieving phase (called by center)
	int SelectPolynomials();
	// create sieving phase and generate factor bases (called by center)
	int CreateSievingPhase();

};

#endif // __NUMBER_FIELD_SIEVE_H__

