#ifndef __POLY_SELECTION_PHASE_H__
#define __POLY_SELECTION_PHASE_H__

#include "abstract_poly_selection_phase.h"
#include "multivariable_polynomial.h"
#include "candidate_class.h"
#include "const_poly_selection_phase.h"


///Classical polynomial selection phase for NFS. Type (n,1)
class CPolySelectionPhase : public APolyPhase
{	
private:
	///The small primes
	static const int iCoefficientPrimes[ConstPSPhase::SMALL_PRIME_COUNT];
	static int iCPrimes[ConstPSPhase::C_PRIME_COUNT]; 

protected:
	///An array used when sieving for the best alpha
	unsigned char jValuations[ConstPSPhase::MAX_J_PRIME][ConstPSPhase::MAX_J_PRIME];
	///An array used when sieving for the best alpha
	double jAlphaP[ConstPSPhase::MAX_J_PRIME];
	///An array used when sieving for the best alpha
	double jAlpha[2 * ConstPSPhase::J0 + 1];

	long		iFBBound;
	poly_gen	iPolyGenType;

	int			iMinAdPerC;				///<The minimal number of a_{d} per c constant
	int			iChiMinLog;				///<The logarithm of value of chi_{min} - used for bottom bound
	int			iChiMaxLog;				///<The logarithm of value of chi_{max} - used for upper bound
	int			iChiInverse;			///<The inverse value of chi
	long		iPrimesDividingLeading;	///<The number of primes used for leading coefficients construction
	long		iMaxPerCoefficient;		///<The maximal number of candidate polynomials per leading coefficient
	long		iSteepestDescentLevel;	///<The steepest descent minimum accuracy - TODO inicializovat
	double		iThirdCoefficientBound;	///<The bound for the third polynomial coefficient during a skewed polynomial selection
	double		iIFSBound;				///<The bound for I(F,S)
	double		iIFSImprovement;		///<The value of improvement of I(F,S) by alpha

	long		iFoundCandidatePolyCount;
	long		iRatedCandidatePolyCount;
	CandidateClass *	iCandidatePolynomials;
	Polynomial *		iResultPoly;
	double				iResultSkewness;

	mpz_t		iMainAd;
	mpz_t		iC;

	mpz_t		iWorkingMonicM;
	int *		iWorkingCPrimes;

	mpz_t		iRSerialMonicM;
	int *		iRSerialCPrimes;
	int			iRSerialCPrimesCount;
	long		iRSerialFoundCandPolyCount;
	long		iRSerialRatedCandPolyCount;

	long		iRSerialRunMChangeCount;
	long		iRSerialRunCChangeCount;
	long		iRSerialRunPolyCount;

	std::string iCandidatePolyFullFileName;
	std::string iRunningSerialFullFileName;

protected:
	//self-initializing procedures
	int FindOptimalChiValue(unsigned int aDegree, mpz_t aValue, mpz_t aLowLeading, mpz_t aHighLeading, int aStrictness);
	int FindOptimalBounds(unsigned int aDegree, mpz_t aValue, mpz_t aLowAd, mpz_t aHighAd);

	int SieveForBestAlpha(int *aJ0, int *aJ1, double *aAlpha, Polynomial *aPolynomial, mpz_t aM);
	int Skewness(mpf_t aResult, Polynomial *aPolynomial, mpf_t aStartingSkewness);
	int SkewPolynomial(Polynomial *aResult, mpf_t aSkewness, mpz_t aNewRoot, double *aIntegral, Polynomial *aPolynomial, mpz_t aOldRoot);
	int DoubleIntegralAcrossSkewedRegion(mpf_t *aResult, Polynomial *aOperand);
	int DoubleIntegralAcrossSkewedRegion(MultVarPolynomial *aResult, MultVarPolynomial *aOperand, unsigned aDegree);
	int CMinimum(mpf_t *aResult, MultVarPolynomial *aOperand, mpf_t **aPowerTable, mpf_t *aSPowerTable, unsigned aDegree);
	
	int Rating(mpf_t aResult, mpf_t aAlpha, long * aRoots, Polynomial *aPolynomial, mpf_t aLogBase);
	int Rating(mpf_t aResult, mpf_t aAlpha, long * aRoots, Polynomial *aPolynomial, mpf_t aSkewness, mpf_t aLogBase);
	
	int ChoosePolynomial(long& aResult, mpf_t aLogBase);
	int ChooseLeadingCoefficient(mpz_t aResult, mpz_t aHighCoeff);

	/// Generate only monic polynomials as candidates
	int GenerateMonic();
	/// Generate only nonmonic unskewed polynomials as candidates
	int GenerateNonmonic();
	/// Generate only skewed polynomials as candidates
	int GenerateSkewed();

	bool ComputeRemainingCoeff(Polynomial * aPoly, int aDegree, mpz_t aM, mpz_t aPowerM, mpz_t aRemainder, mpz_t aHalfM, mpz_t aCoeffBound, mpz_t aAux);

	// Phase methods

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();
	int DisposeGMP();
	int DisposeMutexes();

	int InitParameters();
	int GenerateCandidates();
	int ChooseBest();
	int SaveResult();

	// Printing methods
	void PrintHeader();
	void PrintValues();
	void PrintEffectiveness();

	// For running serialization
	int PrepareRunningSerialization();

public:

	CPolySelectionPhase();
	~CPolySelectionPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);

	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;
	int Serialize(const CBaseParameters& aParam) const;

	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
	int Deserialize(const CBaseParameters& aParam);

};

#endif // __POLY_SELECTION_PHASE_H__