#include "polynomial_improvement.h"

#include "integer_class.h"
#include <math.h>
#include "log.h"
#include "const_return_codes.h"


const int CPolynomialImprovement::iCoefficientPrimes[ConstPSPhase::SMALL_PRIME_COUNT] = 
{ 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 
59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131,
137,139,149,151,157,163,167,173,179,181,191,193,197,199,
211,223,227,229,233,239,241,251,257,263,269,271,277,281,283,
293,307,311,313,317,331,337,347,349,353,359,367,373,379,383,389,
397,401,409,419,421,431,433,439,443,449,457,461,463,467,479,487,491,499};


CPolynomialImprovement::CPolynomialImprovement(int aBufferSize)
{
	this->iDegree = 0;
	this->iSteepestDescentLevel = ConstPSPhase::STEEPEST_DESCENT_LEVEL;

	this->iIntegral = NULL;
	this->iTr = NULL;
	this->iRot = NULL;
	this->iF = NULL;
	this->iAuxF = NULL;
	this->iInteg = NULL;
	this->iSPowers = NULL;
	this->iEvaluatedPowers = NULL;
	this->iChars = NULL;

	mpf_set_default_prec(640);
	mpz_init(this->iAuxMPZ);
	mpz_init(this->iAuxMPZ2);
	mpz_init(this->iSumMPZ2);

	mpf_init(this->iDeriv);
	mpf_init(this->iAux);
	mpf_init(this->iVal);
	mpf_init(this->iPow);
	mpf_init(this->iSq);
	mpf_init2(this->iAuxMPF, ConstPSPhase::MPF_T_HIGH_PRECISION);
	mpf_init2(this->iValP, ConstPSPhase::MPF_T_HIGH_PRECISION);
	mpf_init2(this->iValP1, ConstPSPhase::MPF_T_HIGH_PRECISION);
	mpf_init2(this->iAux1, ConstPSPhase::MPF_T_HIGH_PRECISION);
	mpf_init2(this->iAux2, ConstPSPhase::MPF_T_HIGH_PRECISION);

	int i, j;
	for (i = 0; i < T_LOC; i++)
	{
		for (j = 0; j <= T_LOC; j++)
		{
			mpf_init(this->iAuxMPFMatrix[i][j]);
		}
	}
	for (i = 0; i < MULT_VAR_VARIABLES; i++)
	{
		this->iDerivMVP[i] = NULL;

		mpf_init2(this->iGrad[i], ConstPSPhase::MPF_T_HIGH_PRECISION);
		mpf_init(this->iP[i]);
		mpf_init(this->iP1[i]);
	} 

	int aux = MULT_VAR_VARIABLES - 1;
	for (i = 0; i < aux; i++)
	{
		mpz_init(this->iFC[i]);
	}

	this->iBufferSize = aBufferSize;
	this->iAlphaBoundBuffer = NULL;
	this->iIFSBoundBuffer = NULL;
	this->iAlphaBound = ConstPSPhase::ALPHA_BOUND;
	this->iIFSBound = 1;
}

CPolynomialImprovement::~CPolynomialImprovement()
{
	Reset();
	DisposeGMP();
}

int CPolynomialImprovement::Init(int aDegree)
{
	this->iDegree = aDegree;
	if (this->iIntegral != NULL)
	{
		CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",0,0,"Array already allocated!");
		return ConstRC::GeneralError;
	}
	this->iIntegral = (mpf_t*) Utils::AllocationMachine((void**)&this->iIntegral, (this->iDegree + 1) * sizeof(mpf_t));
	if (this->iIntegral == NULL)
	{
		CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",this->iDegree,0,"There was not enough memory to allocate integral array.");
		return ConstRC::NotEnoughMemory;
	}

	int i, aux;
	for (i = 0; i <= this->iDegree; i++)
	{
		mpf_init(this->iIntegral[i]);
	}

	aux = 3 * this->iDegree * (this->iDegree + 1)/2 + (this->iDegree + 1) * 12;
	this->iTr = new MultVarPolynomial(2, 1);
	this->iRot = new MultVarPolynomial(4, 2);
	this->iF = new MultVarPolynomial(this->iDegree + 7, this->iDegree);
	this->iAuxF = new MultVarPolynomial(5 * this->iDegree + 25, 2 * this->iDegree);
	this->iInteg = new MultVarPolynomial(aux, 0);

	for (i = 0; i < MULT_VAR_VARIABLES; i++)
	{
		this->iDerivMVP[i] = new MultVarPolynomial(aux, 0);
	}

	aux = 2 * this->iDegree + 1;
	this->iEvaluatedPowers = (mpf_t**) Utils::AllocationMachine((void**)&this->iEvaluatedPowers, (MULT_VAR_VARIABLES - 1) * sizeof(mpf_t*));
	if (this->iEvaluatedPowers == NULL)
	{
		CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",MULT_VAR_VARIABLES,0,"There was not enough memory to allocate evaluated powers matrix.");
		return ConstRC::NotEnoughMemory;
	}
	for (int k = 0; k < MULT_VAR_VARIABLES - 1; k++)
	{
		this->iEvaluatedPowers[k] = NULL;
		this->iEvaluatedPowers[k] = (mpf_t*) Utils::AllocationMachine((void**)&this->iEvaluatedPowers[k], aux * sizeof(mpf_t));
		if (this->iEvaluatedPowers[k] == NULL)
		{
			CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",aux,0,"There was not enough memory to allocate evaluated powers matrix.");
			return ConstRC::NotEnoughMemory;
		}
		for (i = 0; i < aux; i++)
		{
			mpf_init(this->iEvaluatedPowers[k][i]);
		}
	}

	aux = 2 * this->iDegree + 2;
	this->iSPowers = (mpf_t*) Utils::AllocationMachine((void**)&this->iSPowers, aux * sizeof(mpf_t));
	if (this->iSPowers == NULL)
	{
		CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",aux,0,"There was not enough memory to allocate SPower array.");
		return ConstRC::NotEnoughMemory;
	}
	for (i = 0; i < aux; i++)
	{
		mpf_init(this->iSPowers[i]);
	}

	if (this->iChars == NULL)
	{
		this->iChars = (char*) Utils::AllocationMachine((void**)&this->iChars, 258 * sizeof(char));
		if (this->iChars == NULL)
		{
			CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",258,0,"There was not enough memory to allocate char array.");
			return ConstRC::NotEnoughMemory;
		}
	}

	if (this->iIFSBoundBuffer == NULL)
	{
		this->iIFSBoundBuffer = (double*) Utils::AllocationMachine((void**)&this->iIFSBoundBuffer, this->iBufferSize * sizeof(double));
		if (this->iIFSBoundBuffer == NULL)
		{
			CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",this->iBufferSize,0,"There was not enough memory to allocate IFS buffer array.");
			return ConstRC::NotEnoughMemory;
		}
		for (i = 0; i < this->iBufferSize; i++)
		{
			this->iIFSBoundBuffer[i] = this->iIFSBound;
		}
	}

	if (this->iAlphaBoundBuffer == NULL)
	{
		this->iAlphaBoundBuffer = (double*) Utils::AllocationMachine((void**)&this->iAlphaBoundBuffer, this->iBufferSize * sizeof(double));
		if (this->iAlphaBoundBuffer == NULL)
		{
			CLog::LogError(WHEREARG, "CPolynomialImprovement","Init",this->iBufferSize,0,"There was not enough memory to allocate alpha buffer array.");
			return ConstRC::NotEnoughMemory;
		}
		for (i = 0; i < this->iBufferSize; i++)
		{
			this->iAlphaBoundBuffer[i] = this->iAlphaBound;
		}
	}

	return ConstRC::Ok;
}

int CPolynomialImprovement::Reset()
{
	int i, auxInt;

	mpz_set_si(this->iAuxMPZ, 0);
	mpz_set_si(this->iAuxMPZ2, 0);
	mpz_set_si(this->iSumMPZ2, 0);

	mpf_set_str(this->iAuxMPF, "0", 10);
	mpf_set_str(this->iDeriv, "0", 10);
	mpf_set_str(this->iAux, "0", 10);
	mpf_set_str(this->iVal, "0", 10);
	mpf_set_str(this->iPow, "0", 10);
	mpf_set_str(this->iSq, "0", 10);
	mpf_set_str(this->iValP, "0", 10);
	mpf_set_str(this->iValP1, "0", 10);
	mpf_set_str(this->iAux1, "0", 10);
	mpf_set_str(this->iAux2, "0", 10);

	if (this->iIntegral != NULL)
	{
		for (i = 0; i <= this->iDegree; i++)
		{
			mpf_clear(this->iIntegral[i]);
		}
		free(this->iIntegral);
		this->iIntegral = NULL;
	}

	if (this->iTr != NULL)
	{
		delete this->iTr;
		this->iTr = NULL;
	}
	if (this->iRot != NULL)
	{
		delete this->iRot;
		this->iRot = NULL;
	}
	if (this->iF != NULL)
	{
		delete this->iF;
		this->iF = NULL;
	}
	if (this->iAuxF != NULL)
	{
		delete this->iAuxF;
		this->iAuxF = NULL;
	}
	if (this->iInteg != NULL)
	{
		delete this->iInteg;
		this->iInteg = NULL;
	}

	for (i = 0; i < MULT_VAR_VARIABLES; i++)
	{
		if (this->iDerivMVP[i] != NULL)
		{
			delete this->iDerivMVP[i];
			this->iDerivMVP[i] = NULL;
		}
	}

	if (this->iChars != NULL)
	{
		free(this->iChars);
		this->iChars = NULL;
	}

	if (this->iIFSBoundBuffer != NULL)
	{
		free(this->iIFSBoundBuffer);
		this->iIFSBoundBuffer = NULL;
	}

	if (this->iAlphaBoundBuffer != NULL)
	{
		free(this->iAlphaBoundBuffer);
		this->iAlphaBoundBuffer = NULL;
	}

	if (this->iIntegral != NULL)
	{
		for (i = 0; i <= this->iDegree; i++)
		{
			mpf_clear(this->iIntegral[i]);
		}
		free(this->iIntegral);
		this->iIntegral = NULL;
	}

	if (this->iSPowers != NULL)
	{
		auxInt = 2 * this->iDegree + 2;
		for (i = 0; i < auxInt; i++)
		{
			mpf_clear(this->iSPowers[i]);
		}
		free(this->iSPowers);
		this->iSPowers = NULL;
	}

	if (this->iEvaluatedPowers != NULL)
	{
		auxInt = 2 * this->iDegree + 1;
		for (int j = 0; j < MULT_VAR_VARIABLES - 1; j++)
		{
			for (i = 0; i < auxInt; i++)
			{
				mpf_clear(this->iEvaluatedPowers[j][i]);
			}
			free(this->iEvaluatedPowers[j]);
			this->iEvaluatedPowers[j] = NULL;
		}
		free(this->iEvaluatedPowers);
		this->iEvaluatedPowers = NULL;
	}

	return ConstRC::Ok;
}

int CPolynomialImprovement::DisposeGMP()
{
	mpz_clear(this->iAuxMPZ);
	mpz_clear(this->iAuxMPZ2);
	mpz_clear(this->iSumMPZ2);

	mpf_clear(this->iAuxMPF);
	mpf_clear(this->iDeriv);
	mpf_clear(this->iAux);
	mpf_clear(this->iVal);
	mpf_clear(this->iPow);
	mpf_clear(this->iSq);
	mpf_clear(this->iValP);
	mpf_clear(this->iValP1);
	mpf_clear(this->iAux1);
	mpf_clear(this->iAux2);

	int i, j, auxInt;
	for (i = 0; i < T_LOC; i++)
	{
		for (j = 0; j <= T_LOC; j++)
		{
			mpf_clear(this->iAuxMPFMatrix[i][j]);
		}
	}

	for (i = 0; i < MULT_VAR_VARIABLES; i++)
	{
		mpf_clear(this->iGrad[i]);
		mpf_clear(this->iP[i]);
		mpf_clear(this->iP1[i]);
	} 

	auxInt = MULT_VAR_VARIABLES - 1;
	for (i = 0; i < auxInt; i++)
	{
		mpz_clear(this->iFC[i]);
	}

	return ConstRC::Ok;
}


/*!
	Performs a sieve over different values of j<sub>0</sub> and j<sub>1</sub> to find
	such values that the polynomial aPolynomial with a root aM has the best alpha estimate aAlpha.
*/
int CPolynomialImprovement::SieveForBestAlpha(int *aJ0, int *aJ1, double *aAlpha, Polynomial *aPolynomial, mpz_t aM, mpz_t aP)
{
	int p,k,l,j1,j0,pi,mpi,pk,m,lm,lmp,flp,pkl,j0b, pp;
	long fl;
	double cont, Malfa = 10.0;
	for (mpi = 0; this->iCoefficientPrimes[mpi] < ConstPSPhase::MAX_J_PRIME; mpi++);

	for (j1 = - ConstPSPhase::J1; j1 <= ConstPSPhase::J1; j1++)
	{
		for (j0 = - ConstPSPhase::J0; j0 <= ConstPSPhase::J0; j0++)
		{
			this->jAlpha[j0 + ConstPSPhase::J0] = 0.0;
		}

		for (pi = 0; pi < mpi; pi++)
		{
			p = this->iCoefficientPrimes[pi];
			pk = p;
			for (k = 1; p * pk < ConstPSPhase::MAX_J_PRIME; k++)
			{
				pk *= p;
			}

			m = mpz_fdiv_ui(aM, pk);
			pp = mpz_fdiv_ui(aP, pk);
			for (l = 0; l < pk; l++)
			{
				for (j0 = 0; j0 < pk; j0++)
				{
					this->jValuations[j0][l] = 0;
				}

				fl = aPolynomial->EvaluateModPrime(l, pk);
				fl += j1 * l * l * pp;
				fl -= j1 * l * m;
				fl %= pk;
				lm = pp * l - m;
				lmp = 0;
				pkl = pk;

				while ((lmp < k) && (!(lm % p))) 
				{
					lm /= p;
					pkl /= p;
					lmp++;
				}

				flp = 0;
				while ((flp < lmp) && (!(fl % p))) 
				{
					fl /= p;
					flp++;
				}

				if ((flp < lmp)||(lmp == k)) 
				{
					if (flp) 
					{
						for (j0 = 0; j0 < pk; j0++)
						{
							this->jValuations[j0][l] = flp;
						}
					}
				} 
				else 
				{
					j0b = (CIntegerClass::InverseModN(lm, pkl) * fl) % pkl;
					if (j0b < 0)
					{
						j0b += pkl;
					}

					while (pkl > 1) 
					{
						for (j0 = j0b; j0 < pk; j0 += pkl)
						{
							this->jValuations[j0][l]++;
						}

						pkl /= p;
						j0b %= pkl;
					}
					if (lmp) 
					{
						for (j0 = 0; j0 < pk; j0++)
						{
							this->jValuations[j0][l] += lmp;
						}
					}
				}
			}

			for (j0 = 0; j0 < pk; j0++)
			{
				j0b = 0;
				for (l = 0; l < pk; l++)
				{
					j0b += this->jValuations[j0][l];
				}

				cont = (double)j0b / (double)pk;

				if (mpz_divisible_ui_p(*aPolynomial->Leading2(), p))
				{
					cont += (double)(p) / (double)(2 * p * p - 1);
				}

				this->jAlphaP[j0] = (1.0/(double)(p - 1) - cont) * log((double)p);
			}

			for (j0 = - ConstPSPhase::J0; j0 <= ConstPSPhase::J0; j0++)
			{
				k = j0 % pk;
				if (k < 0) 
				{
					k += pk;
				}
				this->jAlpha[j0 + ConstPSPhase::J0] += this->jAlphaP[k];
			}
		}

		for (j0 = - ConstPSPhase::J0; j0 <= ConstPSPhase::J0; j0++)
		{
			if (this->jAlpha[j0 + ConstPSPhase::J0] < Malfa) 
			{
				(*aJ0) = j0;
				(*aJ1) = j1;
				Malfa = this->jAlpha[j0 + ConstPSPhase::J0];
			}
		}
	}

	(*aAlpha) = Malfa;

	return ConstRC::Ok;
}

/*!
	Evaluates the double integral in given skewness.
*/
int CPolynomialImprovement::EvalIntInSkewness(mpf_t aResult, mpf_t * aInt, mpf_t aSkewness, int d)
{
       int i;

       mpf_pow_ui(this->iSq, aSkewness, 2);
       mpf_pow_ui(this->iPow, aSkewness, 2);
       mpf_set(this->iVal, aInt[0]);
       mpf_set_str(this->iAux, "0", 10);
       for (i = 1; i <= d; i++){
               mpf_mul(this->iAux, aInt[i], this->iPow);
               mpf_add(this->iVal, this->iVal, this->iAux);
               mpf_mul(this->iPow, this->iPow, this->iSq);
       }

       mpf_set(aResult, this->iVal);

       return ConstRC::Ok;
}


/*!
	Returns the skewness of the polynomial aPolynomial. The iterating process starts at aStartingSkewness.
*/
int CPolynomialImprovement::SkewnessKleinjung(mpf_t aResult, Polynomial *aPolynomial, mpf_t aStartingSkewness)
{
	int d = aPolynomial->GetDegree();
	if (d < 0)
	{
		return ConstRC::GeneralError;
	}

	int i;
	bool found = false, jumped = false;

	DoubleIntegralAcrossSkewedRegionKleinjung(this->iIntegral, aPolynomial);

	mpf_set(aResult, aStartingSkewness);

	EvalIntInSkewness(this->iAux, this->iIntegral, aResult, d);

	while (mpf_cmp_d(this->iAux, 0.1) > 0) {
		mpf_div_ui(aResult, aResult, 2);
		EvalIntInSkewness(this->iAux, this->iIntegral, aResult, d);
	}
	do {
		mpf_mul_ui(aResult, aResult, 2);
		EvalIntInSkewness(this->iAux, this->iIntegral, aResult, d);
	} while (mpf_cmp_d(this->iAux, -0.1) < 0);

	double ub, lb, mp;
	ub = mpf_get_d(aResult);
	lb = ub/2;
	mp = (ub + lb)/2;
	for (int ii = 100; ii > 0; ii--) {      //this precision should be enough
		mp = (ub + lb)/2;
		mpf_set_d(aResult,mp);
		EvalIntInSkewness(this->iAux, this->iIntegral, aResult, d);
		if (mpf_cmp_d(this->iAux, 0.1) > 0)
			ub = mp;
		else
			lb = mp;
	}
	mp = (ub + lb)/2;
	mpf_set_d(aResult,mp);

	return ConstRC::Ok;
}
/*!
	Returns the minimum of the function in variables c<sub>0</sub>, c<sub>1</sub> and c<sub>2</sub>.
*/
int CPolynomialImprovement::CMinimumKleinjung(mpf_t *aResult, MultVarPolynomial *aOperand, mpf_t **aPowerTable, mpf_t *aSPowerTable, unsigned aDegree)
{
	int i, j, k;

	aOperand->ThreeVariableFunction(this->iAuxMPFMatrix, aPowerTable, aSPowerTable, aDegree);
	for (i = 0; i < T_LOC; i++) 
	{
		if (mpf_cmp_ui(this->iAuxMPFMatrix[i][i],0))
		{
			for (j = T_LOC; j >= i; j--)
			{
				mpf_div(this->iAuxMPFMatrix[i][j], this->iAuxMPFMatrix[i][j], this->iAuxMPFMatrix[i][i]);
			}
			for (k = 0; k < T_LOC; k++) 
			{
				if (k != i)
				{
					for (j = T_LOC; j > i; j--)
					{
						mpf_mul(this->iAuxMPF, this->iAuxMPFMatrix[i][j], this->iAuxMPFMatrix[k][i]);
						mpf_sub(this->iAuxMPFMatrix[k][j], this->iAuxMPFMatrix[k][j], this->iAuxMPF);
					}
					mpf_set_ui(this->iAuxMPFMatrix[k][i], 0);
				}
			}
		}
	}

	for (i = 0; i < T_LOC; i++)
	{
		mpf_set(aResult[i], this->iAuxMPFMatrix[i][T_LOC]);
	}
	
	return ConstRC::Ok;
}

/*!
	Returns the s-coefficients of the partial derivative d/ds of the definite integral of (aOperand)<sup>2</sup>
	across the sieving region (-s,s).
*/
int CPolynomialImprovement::DoubleIntegralAcrossSkewedRegionKleinjung(mpf_t *aResult, Polynomial *aOperand)
{
	unsigned i, j, d = aOperand->GetDegree();
	signed int k;


	for (i = 0; i <= d; i++)
	{
		mpf_set_str(aResult[i], "0", 10);
	}

	for (i = 0; i <= d; i++)
	{
		for (j = 0; j <= i; j++)
		{
			if (!((i + j) & 1))
			{
				mpz_mul(this->iAuxMPZ, *aOperand->Get2(i), *aOperand->Get2(j));
				mpf_set_z(this->iAuxMPF, this->iAuxMPZ);
				if (i != j)
				{
					mpf_mul_2exp(this->iAuxMPF, this->iAuxMPF, 1);
				}
				k = -4*d + 4*(i+j);
				if (k < 0) {    //we multiply unsigned int
				k = -k;
				mpf_neg(this->iAuxMPF, this->iAuxMPF);
				}
				mpf_mul_ui(this->iAuxMPF, this->iAuxMPF, k);
				mpf_div_ui(this->iAuxMPF, this->iAuxMPF, (i + j)*(2*d - i - j) + 2*d + 1);
				mpf_add(aResult[(i + j)/2], aResult[(i + j)/2], this->iAuxMPF);
			}
		}
	}

	return ConstRC::Ok;
}

/*!
	Returns the multivariate polynomial being the definite integral of aOperand across the sieving region.
*/
int CPolynomialImprovement::DoubleIntegralAcrossSkewedRegionKleinjung(MultVarPolynomial *aResult, MultVarPolynomial *aOperand, unsigned int aDegree)
{
	unsigned int i,j,k;
	unsigned int d = aDegree << 1;
	unsigned int l = 0;

	for (i = 0; i < aOperand->length; i++)
	{
		for (j = 0; j <= d; j++) 
		{
			if ((!(j & 1)) && (mpz_cmp_ui(*aOperand->Term[i].XPolynomial->Get2(j), 0)))
			{
				mpf_set_z(this->iAuxMPF, *aOperand->Term[i].XPolynomial->Get2(j));
				mpf_mul_2exp(this->iAuxMPF, this->iAuxMPF, 2);
				mpf_div_ui(this->iAuxMPF, this->iAuxMPF, j + 1);
				mpf_div_ui(this->iAuxMPF, this->iAuxMPF, d - j + 1);
				for (k = 0; k < MULT_VAR_VARIABLES - 1; k++)
				{
					aResult->Term[l].Variables[k] = aOperand->Term[i].Variables[k];
				}
				aResult->Term[l].Variables[k] = (signed)j-(signed)aDegree;
				mpf_set(aResult->Term[l].Coeff, this->iAuxMPF);
				l++;
			}
		}
	}

	aResult->length = l;

	return ConstRC::Ok;
}


/*!
	Skews the polynomial using the steepest descent method. Input is a polynomial aPolynomial with a root aOldRoot and leading coeff aP (of second polynomial)
	and output is a polynomial aResult with a root aNewRoot, skewness aSkewness and its I(F,S) aIntegral.
*/
int CPolynomialImprovement::SkewPolynomial(Polynomial *aResult, mpf_t aSkewness, mpz_t aNewRoot, double *aIntegral, Polynomial *aPolynomial, mpz_t aOldRoot, mpz_t aP)
{
	// translation by p*t => f_{t}(x) = f(x - p*t) 
	// rotation by R(x) => R(x) = c_{2}x^{2} + c_{1}x + c_{0} = c_{2}(x - p*t)^{2} + c_{1}(x - p*t) + c_{0}
	// f(x) = f_{1,m}(x_pt) + R(x_pt) * f_{2,m}(x_pt) = f_{m}(x - p*t) + (c_{2}(x - p*t)^{2} + c_{1}(x - p*t) + c_{0}) * (p*x - p^{2}*t - m)
	unsigned i,k,d;
	int j;
	d = aPolynomial->GetDegree();
	
	this->iTr->Reset();
	this->iF->Reset();
	this->iRot->Reset();
	this->iInteg->Reset();
	this->iAuxF->Reset();
	for (i = 0; i < MULT_VAR_VARIABLES; i++)
	{
		this->iDerivMVP[i]->Reset();
	}

	// Tr = x + 0
	this->iTr->Term[0].XPolynomial->Set(1, 1);
	// Tr = x - p
	mpz_neg(this->iAuxMPZ2, aP);
	this->iTr->Term[1].XPolynomial->Set(0, this->iAuxMPZ2);
	// Tr = x - p*t
	this->iTr->Term[1].Variables[T_LOC] = 1;
	this->iTr->length = 2;
	this->iF->length = 1;
	// F = a_{d}
	this->iF->Term[0].XPolynomial->Set(0, *aPolynomial->Get2(d));

	// Horner scheme - F = f_{m}(x_{pt}) = f_{m}(x - p*t)
	for (i = 0; i < d; i++)
	{
		// AuxF = F * Tr
		this->iF->Multiply(this->iAuxF, this->iTr);
		// F = AuxF
		this->iAuxF->Copy(this->iF);
		// F = F + a_{d - i - 1}
		this->iF->Term[0].XPolynomial->Inc(0, *aPolynomial->Get2(d - i - 1));
	}

	this->iRot->Term[0].XPolynomial->Set(0, 1);
	// Rot = c_{0}
	this->iRot->Term[0].Variables[C0_LOC] = 1;
	if (d > 4) 
	{
		this->iRot->Term[1].XPolynomial->Set(1, 1);
		// Rot = c_{1}x + c_{0}
		this->iRot->Term[1].Variables[C1_LOC] = 1;
		if (d > 5) 
		{
			// Rot = c_{2}x^{2} + c_{1}x + c_{0}
			this->iRot->Term[2].XPolynomial->Set(2, 1);
			this->iRot->Term[2].Variables[C2_LOC] = 1;
			this->iRot->length = 3;
		} 
		else
		{
			this->iRot->length = 2;
		}
	} 
	else
	{
		this->iRot->length = 1;
	}

	// Tr = x - p*t - m
	this->iTr->Term[0].XPolynomial->Dec(0, aOldRoot);
	// Tr = p*x - p*t - m
	this->iTr->Term[0].XPolynomial->Set(1, aP);
	// Tr = p*x - p^2*t - m
	mpz_mul(this->iAuxMPZ2, aP, aP);
	mpz_neg(this->iAuxMPZ2, this->iAuxMPZ2);
	this->iTr->Term[1].XPolynomial->Set(0, this->iAuxMPZ2);
	// AuxF = Tr * Rot = (p*x - p^{2}*t - m) * (c_{2}x^{2} + c_{1}x + c_{0})
	this->iTr->Multiply(this->iAuxF, this->iRot);
	// F = F + AuxF = f_{m}(x - t) + (p*x - p^{2}*t - m) * (c_{2}x^{2} + c_{1}x + c_{0})
	this->iF->Add(this->iF, this->iAuxF);
	// AuxF = F^2
	this->iF->Multiply(this->iAuxF, this->iF);

	DoubleIntegralAcrossSkewedRegionKleinjung(this->iInteg, this->iAuxF, d);

	// Steepest descent minimum algorithm
	
	for (i = 0; i < MULT_VAR_VARIABLES; i++)
	{
		this->iInteg->PartialDerivative(this->iDerivMVP[i],i);
	}

	double step = ConstPSPhase::INITIAL_STEEPEST_DESCENT_STEP, r, ir;
	unsigned cont = 0, good = 0, smallImprove = 0;
	int auxInt;

	auxInt = 2 * d + 1;
	for (k = 0; k < MULT_VAR_VARIABLES - 1; k++)
	{
		for (i = 0; i < auxInt; i++)
		{
			mpf_set_ui(this->iEvaluatedPowers[k][i], 0);
		}
		mpf_set_ui(this->iEvaluatedPowers[k][0], 1);
	}
   
	auxInt = 2 * d + 2;
	for (i = 0; i < auxInt; i++)
	{
		mpf_set_ui(this->iSPowers[i], 0);
	}

	for (i = 0; i < MULT_VAR_VARIABLES - 1; i++)
	{
		mpf_set_ui(this->iP[i], 0);
	}

	mpf_set_ui(this->iP[T_LOC], 1);
	mpf_set_ui(this->iP[S_LOC], ConstPSPhase::SKEWNESS);
	mpf_set_ui(this->iSPowers[d + 1], 1);

	auxInt = 2 * d + 1;
	for (i = 1; i < auxInt; i++)
	{
		mpf_mul(this->iEvaluatedPowers[T_LOC][i], this->iEvaluatedPowers[T_LOC][i-1], this->iP[T_LOC]);
	}
	for (i = 1; i <= d; i++)
	{   
		mpf_mul(this->iSPowers[d + i + 1], this->iSPowers[d + i], this->iP[S_LOC]);
		mpf_ui_div(this->iSPowers[d - i + 1], 1, this->iSPowers[d + i +1 ]);
	}
	mpf_div(this->iSPowers[0], this->iSPowers[1], this->iP[S_LOC]);
	CMinimumKleinjung(this->iP, this->iInteg, this->iEvaluatedPowers, this->iSPowers, d);

	do 
	{
		if (!cont)
		{
			for (k = 0; k <= C2_LOC; k++)
			{
				for (i = 1; i <= 2; i++)
				{
					mpf_mul(this->iEvaluatedPowers[k][i], this->iEvaluatedPowers[k][i - 1], this->iP[k]);
				}
			}

			auxInt = 2 * d + 1;
			for (i = 1; i < auxInt; i++)
			{
				mpf_mul(this->iEvaluatedPowers[T_LOC][i], this->iEvaluatedPowers[T_LOC][i - 1], this->iP[T_LOC]);
			}
			for (i = 1; i <= d; i++)
			{   
				mpf_mul(this->iSPowers[d + i + 1], this->iSPowers[d + i], this->iP[S_LOC]);
				mpf_ui_div(this->iSPowers[d - i + 1], 1, this->iSPowers[d + i + 1]);
			}

			mpf_div(this->iSPowers[0], this->iSPowers[1], this->iP[S_LOC]);
			for (k = T_LOC; k < MULT_VAR_VARIABLES; k++)
			{
				this->iDerivMVP[k]->Evaluate(this->iGrad[k], this->iEvaluatedPowers, this->iSPowers, d);
			}
		}
		mpf_set_ui(this->iAux2, 0);
		for (k = T_LOC; k < MULT_VAR_VARIABLES; k++)
		{
			mpf_pow_ui(this->iAux1, this->iGrad[k], 2);
			mpf_add(this->iAux2, this->iAux2, this->iAux1);
		}
		mpf_sqrt(this->iAux1, this->iAux2);

		for (k = T_LOC; k < MULT_VAR_VARIABLES; k++)
		{
			mpf_div(this->iGrad[k], this->iGrad[k], this->iAux1);
		}

		this->iInteg->Evaluate(this->iValP, this->iEvaluatedPowers, this->iSPowers, d);

		do 
		{
			cont = 0;
			for (k = T_LOC; k < MULT_VAR_VARIABLES; k++)
			{
				mpf_set_d(this->iP1[k], -step);
				if (mpf_sgn(this->iGrad[k]) < 0)
					mpf_neg(this->iP1[k], this->iP1[k]);
				mpf_add(this->iP1[k], this->iP[k], this->iP1[k]);
			}
			
			mpf_abs(this->iP1[S_LOC], this->iP1[S_LOC]);

			auxInt = 2 * d + 1;
			for (i = 1; i < auxInt; i++)
			{
				mpf_mul(this->iEvaluatedPowers[T_LOC][i], this->iEvaluatedPowers[T_LOC][i - 1], this->iP1[T_LOC]);
			}
			for (i = 1; i <= d; i++)
			{   
				mpf_mul(this->iSPowers[d + i + 1], this->iSPowers[d + i], this->iP1[S_LOC]);
				mpf_ui_div(this->iSPowers[d - i + 1], 1, this->iSPowers[d + i + 1]);
			}
			mpf_div(this->iSPowers[0], this->iSPowers[1], this->iP1[S_LOC]);
			this->iInteg->Evaluate(this->iValP1, this->iEvaluatedPowers, this->iSPowers, d);
			if (mpf_cmp(this->iValP, this->iValP1) < 0)
			{
				step = step / 2;
				cont = 1;
				good = 0;
			} 
			else 
			{
				good++;
				r = log(sqrt(mpf_get_d(this->iValP))) - log(sqrt(mpf_get_d(this->iValP1)));
				ir = fabs(r / (0.5 * log(mpf_get_d(this->iValP))));
				if (ir < 0.0002) 
				{
					smallImprove++;
					if (ir < 0.000001)
					smallImprove += 10;
				} 
				else 
				{
					smallImprove = 0;
				}
			}
			
			if (good >= 4) 
			{
				step *= 2;
				good = 0;
			}
		} 
		while (cont);

		auxInt = 2 * d + 1;
		for (i = 1; i < auxInt; i++)
		{
			mpf_mul(this->iEvaluatedPowers[T_LOC][i], this->iEvaluatedPowers[T_LOC][i - 1], this->iP1[T_LOC]);
		}
		for (i = 1; i <= d; i++)
		{   
			mpf_mul(this->iSPowers[d + i + 1], this->iSPowers[d + i], this->iP1[S_LOC]);
			mpf_ui_div(this->iSPowers[d - i + 1], 1, this->iSPowers[d + i + 1]);
		}
		mpf_div(this->iSPowers[0], this->iSPowers[1], this->iP1[S_LOC]);
		CMinimumKleinjung(this->iP1, this->iInteg, this->iEvaluatedPowers, this->iSPowers, d);
		
		for (k = 0; k < MULT_VAR_VARIABLES; k++)
		{
			mpf_set(this->iP[k], this->iP1[k]);
		}
	} 
	while ((step > (0.1 / this->iSteepestDescentLevel) && (smallImprove < 100 * this->iSteepestDescentLevel)));

	mpf_set_d(this->iAux1, 0.5);

	auxInt = MULT_VAR_VARIABLES - 1;
	for (k = 0; k < auxInt; k++)
	{
		mpf_add(this->iP[k], this->iP[k], this->iAux1);
		mpf_floor(this->iP[k], this->iP[k]);
		mpf_get_str(this->iChars, &this->iT, 36, 256, this->iP[k]);
		mpz_set_str(this->iFC[k], this->iChars, 36);
	}

	aResult->SetDegree(-1);

	auxInt = MULT_VAR_VARIABLES - 1;
	for (i = 0; i < this->iF->length; i++)
	{
		mpz_set_ui(this->iSumMPZ2, 1);
		for (k = 0; k < auxInt; k++)
		{
			mpz_pow_ui(this->iAuxMPZ2, this->iFC[k], this->iF->Term[i].Variables[k]);
			mpz_mul(this->iSumMPZ2, this->iSumMPZ2, this->iAuxMPZ2);
		}
		for (j = this->iF->Term[i].XPolynomial->GetDegree(); j >= 0; j--)
		{
			mpz_mul(this->iAuxMPZ2, this->iSumMPZ2, *this->iF->Term[i].XPolynomial->Get2(j));
			aResult->Inc(j, this->iAuxMPZ2);
		}
	}

	SkewnessKleinjung(aSkewness, aResult, this->iP[S_LOC]);

	mpz_add(aNewRoot, aOldRoot, this->iFC[T_LOC]);
	for (k = 0; k <= C2_LOC; k++)
	{
		for (i = 1; i <= 2; i++)
		{
			mpf_mul(this->iEvaluatedPowers[k][i], this->iEvaluatedPowers[k][i - 1], this->iP[k]);
		}
	}
	auxInt = 2 * d + 1;
	for (i = 1; i < auxInt; i++)
	{
		mpf_mul(this->iEvaluatedPowers[T_LOC][i], this->iEvaluatedPowers[T_LOC][i - 1], this->iP[k]);
	}
	for (i = 1; i <= d; i++)
	{   
		mpf_mul(this->iSPowers[d + i + 1], this->iSPowers[d + i], aSkewness);
		mpf_ui_div(this->iSPowers[d - i + 1], 1, this->iSPowers[d + i + 1]);
	}
	this->iInteg->Evaluate(this->iValP, this->iEvaluatedPowers, this->iSPowers, d);
	(*aIntegral) = Utils::Ln(this->iValP);
	(*aIntegral) /= 2;
	
	return ConstRC::Ok;
}

/*
	Main method. Perform improvement of given polynomial. Use translation and rotation, adjust by sieving for better alpha.
*/
int CPolynomialImprovement::Improve(Polynomial * aResultPoly, mpz_t aResultM, mpf_t aSkewness, mpz_t aM, mpz_t aP)
{
	int logM = mpz_sizeinbase(aM, 2);
	double aux = logM * sqrt(sqrt((double)this->iDegree));
	double gIFS = aux * this->iIFSBound;
	double integral = 0;
	double alpha = 0;
	int j0 = 0;
	int j1 = 0;
	int rc;
	rc = SkewPolynomial(aResultPoly, aSkewness, aResultM, &integral, aResultPoly, aM, aP);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (integral < gIFS)
	{ 
		// f_{j0,j1}(x) = f(x) + (j1*x - j0)(p*x - m)
		rc = SieveForBestAlpha(&j0, &j1, &alpha, aResultPoly, aResultM, aP);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (alpha + integral < gIFS + this->iAlphaBound)
		{
			// selfinitialization of IFSBound and AlphaBound
			this->iIFSBoundBuffer[this->iBufferSize - 1] = integral / aux;
			qsort(this->iIFSBoundBuffer, this->iBufferSize, sizeof(double), &Utils::CompareDoubles);
			this->iIFSBound = this->iIFSBoundBuffer[this->iBufferSize - 1];

			if (alpha < this->iAlphaBound)
			{
				this->iAlphaBoundBuffer[this->iBufferSize - 1] = alpha;
				qsort(this->iAlphaBoundBuffer, this->iBufferSize, sizeof(double), &Utils::CompareDoubles);
				this->iAlphaBound = this->iAlphaBoundBuffer[this->iBufferSize - 1];				
			}

			// j1 * p * x^{2}
			mpz_mul_si(this->iAuxMPZ, aP, j1);
			aResultPoly->Inc(2, this->iAuxMPZ);

			// (- j0 * p - j1 * m ) * x
			mpz_mul_si(this->iAuxMPZ, aM, j1);
			aResultPoly->Dec(1, this->iAuxMPZ);
			mpz_mul_si(this->iAuxMPZ, aP, j0);
			aResultPoly->Dec(1, this->iAuxMPZ);

			// j0 * m
			mpz_mul_si(this->iAuxMPZ, aM, j0);
			aResultPoly->Inc(0, this->iAuxMPZ);

			rc = SkewnessKleinjung(aSkewness, aResultPoly, aSkewness);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}	

			return ConstRC::GoodCandidate;
		} 
	}

	return ConstRC::Ok;
}
