#ifndef __REL_PROCESSING_PHASE_H__
#define __REL_PROCESSING_PHASE_H__

#include "abstract_rel_processing_phase.h"
#include "nfs_relation.h"
#include "factor_base_info.h"
#include "lp_hashtable_type1.h"


/*!
 *\brief Relation processing phase for NFS with two parts (can be (integral, algebraic) or (algebraic, algebraic)). 

	Use only for max 1,1-partial relations. Otherwise use CFilteringPhase.
	It is not also recommended for too many relations.
*/
class CRelProcessingPhase : public ARelProcessingPhase
{

protected:
	///Number of vertices in partial relations graph - we get this number from parameters
	int iVertices;
	///Number of endes in partial relations graph - we get this number from parameters
	int iEdges;
	///Number of components in partial relations graph - we get this number from parameters
	int iComponents;

	///Number of expected vertices - aux member
	int iExpectedVertices;
	///Number of independent cycles - aux member
	int iIndependentCycles;

	/*!
		Number of all smooth relations. We get this number from last phase but 
		we can construct new smooth relations from partial relations and also 
		we can remove some relations during removing singletons.
		So this number will change and has to be save again.
	*/
	long iSmoothRelationsCount;

	/*!	List of roots:
		The graph of remaining partial relations can have more than 1 component of connectivity.
		We must save roots of all components.
	*/
	std::vector<CNFS_HashEntry1> iRootsList;

	/*!
		The sieving polynomials.

		This polynomial can be monic or non-monic.
		In case that it is non-monic, the corresponding
		monic polynomial is present in the iDomain as 
		a member variable.
	
		By the corresponding monic polynomial we mean		
		
		f(x/a)*a^{d-1}

		where a is the leading coefficient and d the degree of f.

		In this phase it will be use for some calculations.
	*/
	Polynomial * iSievingPoly[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	std::string iSmoothRelFullFileName;
	std::string iPartialRelFullFileName;

	std::string iAllPartialRelFullFileName;
	std::string iTempFileFullFileName;
	std::string iUsablePartialRelFullFileName;
	std::string iAllSmoothRelFullFileName;
	///Full filenames with the quadratic characters - they will be created in this phase
	std::string iQuadCharsFullFileName[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	///Field of the factor base infos
	CFBInfo		iFBs[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

protected:
	
	// Phase methods
	int RemoveLargeSingletons();
	int RemoveSmallSingletons();
	int ConstructCycles();
	int CreateRelationMatrix();

	/*!
		Fill the iPhaseFunctors with correct function pointers.
		This method is called from InitParameters in APhase.
	*/
	int FillFunctorField();

	///Dispose all resources which was used and prepare for new start. Also set inner state.
	int Reset();

	///Save phase's result
	int SaveResult();

	///Init all parameters necessary for calculation - get from iParameters
	int InitParameters();

public:

	CRelProcessingPhase();
	~CRelProcessingPhase();

	///Delete all used files according to enum cleaning_types
	int CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError);
	
	///Start method for linear phase
	int RunPhase(AFactorAlgParameters*  aParameters);

	// De/Serialization methods
	int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);
};

#endif // __REL_PROCESSING_PHASE_H__
