#include "relation_hashtable.h"

#include "nfs_utils.h"
//#include <stdio.h>
#include "mserializable.h"

#include "const_filenames.h"
#include "const_hashtables.h"
#include "const_return_codes.h"
#include "log.h"

using namespace std;

CRelationHashtable::CRelationHashtable()
{
	this->iBaseFileName = ConstFN::RELATION_HASHTABLE_STORE;
	this->iProcessedFileName = ConstFN::PROCESSED_RELATIONS;
	this->iTempFileName = ConstFN::RELATION_HASHTABLE_TEMP;

	this->iTotalFileCount = 0;
	this->iRelationPartCount = 0;

	this->iMask = 0;
	this->iMaxNbrRelationInFile = 0;
	// no file selected - will be used the first prime ideal in the first relation part
	this->iCurrentFileIndex = -1; 

	this->iFileCounts = NULL;
	this->iMinPrimes = NULL;
	this->iMaxPrimes = NULL;
	this->iBlockLenghts = NULL;

	this->iHashtable = NULL;
}

CRelationHashtable::~CRelationHashtable()
{
	DeleteHashtable();
}

void CRelationHashtable::DeleteHashtable()
{
	if (this->iTotalFileCount != 0)
	{
		ClearHashtableFiles();
		this->iTotalFileCount = 0;
	}

	this->iBaseFileName = "";
	this->iProcessedFileName = "";
	this->iTempFileName = "";
	this->iRelationPartCount = 0;

	this->iMask = 0;
	this->iMaxNbrRelationInFile = 0;
	// no file selected - will be used the first prime ideal in the first relation part
	this->iCurrentFileIndex = -1; 

	if (this->iFileCounts != NULL)
	{
		free(this->iFileCounts);
		this->iFileCounts = NULL;
	}
	if (this->iMinPrimes != NULL)
	{
		free(this->iMinPrimes);
		this->iMinPrimes = NULL;
	}
	if (this->iMaxPrimes != NULL)
	{
		free(this->iMaxPrimes);
		this->iMaxPrimes = NULL;
	}
	if (this->iBlockLenghts != NULL)
	{
		free(this->iBlockLenghts);
		this->iBlockLenghts = NULL;
	}

	if (this->iHashtable != NULL)
	{
		CFNFSRelation ** current = this->iHashtable;
		for (relation_index_type i = 0; i < this->iMaxNbrRelationInFile; i++, current++)
		{
			if (*current != NULL)
			{
				delete (*current);
				*current = NULL;
			}
		}

		free(this->iHashtable);
		this->iHashtable = NULL;
	}
}

/*
	This method clear hashtable store files.
	When we use files for storing hashtable entries then it means that we just remove all used files.
*/
int CRelationHashtable::ClearHashtableFiles()
{
	if (this->iTotalFileCount != 0)
	{
		// remove all hashtable store files
		string strFilename;
		int rc;
		for (int i = 0; i < this->iTotalFileCount; i++)
		{
			rc = GetFilename(i, strFilename);
			if (rc != ConstRC::Ok) 
			{
				return rc;
			}

			//we don't control if file exists (in that case the method remove just return nonzero value)
			rc = remove(strFilename.c_str());
		}

		// remove processed file
		rc = remove(this->iProcessedFileName.c_str());

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","ClearHashtableFiles",0,0,"Hashtable is not setup.");
		return ConstRC::GeneralError;
	}
}

/*
	This method creates all hashtable store files.
*/
int CRelationHashtable::CreateHashtableFiles()
{
	if (this->iTotalFileCount != 0)
	{
		// get all hashtable store files
		string strFilename;
		int rc;
		FILE * file;
		for (int i = 0; i < this->iTotalFileCount; i++)
		{
			rc = GetFilename(i, strFilename);
			if (rc != ConstRC::Ok) 
			{
				return rc;
			}
			//create new store file
			file = fopen(strFilename.c_str(), FILE_MODE_WRITE);
			if (file == NULL) 
			{
				CLog::LogError(WHEREARG,"CRelationHashtable","CreateHashtableFiles",0,0,"Hashtable store file cannot be created.");
				return ConstRC::GeneralError;
			}
			else
			{
				fclose(file);
				file = NULL;
			}
		}

		// create processed file
		file = fopen(this->iProcessedFileName.c_str(), FILE_MODE_WRITE);
		if (file == NULL) 
		{
			CLog::LogError(WHEREARG,"CRelationHashtable","CreateHashtableFiles",0,0,"Hashtable store file cannot be created.");
			return ConstRC::GeneralError;
		}
		else
		{
			fclose(file);
			file = NULL;
		}

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","CreateHashtableFiles",0,0,"Hashtable is not setup.");
		return ConstRC::GeneralError;
	}	
}


/*
	This method clear hashtable.
	All relations are deleted. But hashtable field remains allocated.
*/
int CRelationHashtable::ClearHashtable()
{
	if (this->iHashtable != NULL)
	{
		CFNFSRelation ** current = this->iHashtable;
		for (relation_index_type i = 0; i < this->iMaxNbrRelationInFile; i++, current++)
		{
			if (*current != NULL)
			{
				delete (*current);
				*current = NULL;
			}
		}

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable",0,0,"Hashtable is not setup.");
		return ConstRC::GeneralError;
	}
}

int CRelationHashtable::SetupHashtable(relation_index_type * aMultiRelationCounts, long aRelationsInMemory,
									   main_sieving_type * aMinPrimes, main_sieving_type * aMaxPrimes, int aRelationPartCount)
{
	if (this->iTotalFileCount != 0 || this->iHashtable != NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable",0,0,"Hashtable is already setup.");
    	return ConstRC::GeneralError;
	}

	this->iRelationPartCount = aRelationPartCount;

	this->iMinPrimes = (main_sieving_type*)Utils::AllocationMachine((void**)&this->iMinPrimes, this->iRelationPartCount * sizeof(main_sieving_type));
	if (this->iMinPrimes == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable ",0,0,"Could not allocate min primes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iMinPrimes, 0, this->iRelationPartCount * sizeof(main_sieving_type));

	this->iMaxPrimes = (main_sieving_type*)Utils::AllocationMachine((void**)&this->iMaxPrimes, this->iRelationPartCount * sizeof(main_sieving_type));
	if (this->iMaxPrimes == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable ",0,0,"Could not allocate max primes.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iMaxPrimes, 0, this->iRelationPartCount * sizeof(main_sieving_type));

	this->iFileCounts = (int*)Utils::AllocationMachine((void**)&this->iFileCounts, this->iRelationPartCount * sizeof(int));
	if (this->iFileCounts == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable ",0,0,"Could not allocate file counts.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iFileCounts, 0, this->iRelationPartCount * sizeof(int));

	this->iBlockLenghts = (main_sieving_type*)Utils::AllocationMachine((void**)&this->iBlockLenghts, this->iRelationPartCount * sizeof(main_sieving_type));
	if (this->iBlockLenghts == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable ",0,0,"Could not allocate block lenghts.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iBlockLenghts, 0, this->iRelationPartCount * sizeof(main_sieving_type));

	long relationsInMemory = 0;
	Utils::GetClosestBiggerPowerOf2(relationsInMemory, aRelationsInMemory);
	relationsInMemory = relationsInMemory >> 1; //we need smaller number than memory size
	// for hash calculation
	this->iMask = relationsInMemory - 1;
	this->iMaxNbrRelationInFile = relationsInMemory;

	// allocate hashtable -- size + SECURITY_MARGIN
	this->iHashtable = (CFNFSRelation**)Utils::AllocationMachine((void**)&this->iHashtable, (this->iMaxNbrRelationInFile + CRelationHashtable::SECURITY_MARGIN) * sizeof(CFNFSRelation*));
	if (this->iHashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","SetupHashtable ",0,0,"Could not allocate hashtable.");
		return ConstRC::NotEnoughMemory;
	}
	memset(this->iHashtable, 0, (this->iMaxNbrRelationInFile + CRelationHashtable::SECURITY_MARGIN) * sizeof(CFNFSRelation*));

	// for file count calculation we need to consider fullness of hashtable
	relationsInMemory = (long)(relationsInMemory * ConstHT::MAX_FULLNESS);

	this->iTotalFileCount = 0;
	for (int i = 0; i < this->iRelationPartCount; i++)
	{
		// copy sizes
		this->iMinPrimes[i] = aMinPrimes[i];
		this->iMaxPrimes[i] = aMaxPrimes[i];
		// get number of files for each part
		this->iFileCounts[i] = aMultiRelationCounts[i] / relationsInMemory;
		if (this->iFileCounts[i] == 0)
		{
			this->iFileCounts[i] = 1;
		}
		// get length of block of interval [min prime, max prime]
		this->iBlockLenghts[i] = ((aMaxPrimes[i] - aMinPrimes[i]) / this->iFileCounts[i]) + 1;
		// count total number of files
		this->iTotalFileCount += this->iFileCounts[i];
	}

	// no file selected - will be used the first prime ideal in the first relation part
	this->iCurrentFileIndex = -1; 
		
	return ConstRC::Ok;
}

int CRelationHashtable::SetDirectoryAndIdentifier(string aWorkingDirectoryPath, string aIdentifier)
{
	this->iBaseFileName = MSerializable::CreateFileNamePlain(aWorkingDirectoryPath, ConstFN::RELATION_HASHTABLE_STORE, aIdentifier, "");
	this->iProcessedFileName = MSerializable::CreateFileNamePlain(aWorkingDirectoryPath, ConstFN::PROCESSED_RELATIONS, aIdentifier, TXT_FILE_EXTENSION);
	this->iTempFileName = MSerializable::CreateFileNamePlain(aWorkingDirectoryPath, ConstFN::RELATION_HASHTABLE_TEMP, aIdentifier, TXT_FILE_EXTENSION);
	
	//create all store files
	return CreateHashtableFiles();
}

/*
	Same as AddToFile. If file index is not specified we use current file index.
*/
int CRelationHashtable::AddToFile(CFNFSRelation * aRelation)
{
	return AddToFile(aRelation, this->iCurrentFileIndex);
}
/*
	This method save relation to the suitable file.

	Relations are put to the files according to large prime ideals. We have file groups for each relation parts.
	Prime ideal are sorted by size and according to relation part (first part sorted by size, second part sorted by size etc.).
	Relations are save in file according to choosen prime ideal (we want relations with same prime ideal to be close to each other).
		- Relations are save in the right file subsequently. They are just appended to the file.
*/
int CRelationHashtable::AddToFile(CFNFSRelation * aRelation, int aFileIndex)
{
	if (aRelation == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","AddToFile ",0,0,"Relation is NULL.");
		return ConstRC::BadArgument;
	}

	if (this->iTotalFileCount == 0)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","AddToFile ",0,0,"Hashtable is not setup.");
		return ConstRC::GeneralError;
	}

	int index;
	int rc;
	int i, j;
	int max;
	CFRelationPart * part = NULL;
	int partIndex = -1;
	main_sieving_type prime = -1;
	bool sameblock = false;
	if (aFileIndex == -1)
	{
		// We will use the first large prime ideal.
		// We try all relation parts.
		// We suppose that prime ideals are sorted by size.
		// Everything what we need is stored in variables "partIndex" and "prime"
		for (i = 0; i < this->iRelationPartCount; i++)
		{
			part = aRelation->Get_Parts(i);
			max = part->Main.MaxElementIndex;
			for (j = part->LargePrimeIndex; j <= max; j++)
			{
				if (part->Main.Elements[j].exponent % 2)
				{
					prime = part->Main.Elements[j].prime;
					partIndex = i;
					break;
				}
			}

			if (prime != -1)
			{
				// we already find the first prime ideal
				break;
			}
		}
	}
	else
	{
		// we need to do more complex decision
		// firstly, we need to know in which part of relation we are
		// and we need to calculate current block bounds in current relation part
		int currentPartIndex = 0;
		main_sieving_type blockStart;
		main_sieving_type blockEnd;
		rc = GetFromFilePartAndBounds(aFileIndex, currentPartIndex, blockStart, blockEnd);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		// we are looking for the first smallest large prime ideal (if exists)
		// then for the first large prime ideal in the same block (if exists)
		// then for the first larger prime ideal than the given block (if exists)
		main_sieving_type p = -1;
		for (i = 0; i < this->iRelationPartCount; i++)
		{
			part = aRelation->Get_Parts(i);
			max = part->Main.MaxElementIndex;
			for (j = part->LargePrimeIndex; j <= max; j++)
			{
				if (part->Main.Elements[j].exponent % 2)
				{
					p = part->Main.Elements[j].prime;
					// check if this prime is the first smallest large prime ideal
					if (prime == -1 && 
						((i == currentPartIndex && p < blockStart) ||
						 (i < currentPartIndex)))
					{
						// we save this relation according to the first smallest large prime ideal 
						// because this relation must be processed in the next round
						prime = p;
						partIndex = i;
						// can go to the next prime ideal and may be find something else
						continue;
					}

					// check if this prime is in the same block
					if (i == currentPartIndex && blockStart <= p && p < blockEnd)
					{
						sameblock = true;
						prime = p;
						partIndex = i;
						// can go to the next prime ideal and may be find something else
						continue;
					}

					// check if this prime is first larger large prime ideal
					if (i == currentPartIndex && blockEnd <= p || currentPartIndex < i)
					{
						// we save this relation according to the first larger large prime ideal 
						// because this relation must be processed in this round one more					
						prime = p;
						partIndex = i;
						// we have everything so we can end
						i = this->iRelationPartCount;
						break;
					}
				}
			}
		}
	}

	rc = ConstRC::Ok;
	string strFilename = "";
	if (prime == -1)
	{
		// we didn't find any large prime ideal so this relation is processed...
		// we save this relation to the special file
		strFilename = this->iProcessedFileName;

	}
	else
	{
		// we found the large prime ideal so we save this relation to the file according to this prime ideal
		rc = FindFileIndex(prime, partIndex, index);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (sameblock)
		{
			// we need to use temp file because current file is in used
			strFilename = this->iTempFileName;
		}
		else
		{
			// put together the file name
			strFilename = this->iBaseFileName + "_" + Utils::ConvertToString(partIndex) + "_" + Utils::ConvertToString(index) + TXT_FILE_EXTENSION;
		}
	}

	// we append this relation to the given file
	FILE * relationsFILE = fopen(strFilename.c_str(), FILE_MODE_APPEND);
	if (relationsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","AddToFile ",0,0,"Could not append to the file with relations.");
		return ConstRC::CouldNotOpenFile;
	}
	rc = aRelation->PrintToFile(relationsFILE);
	fclose(relationsFILE);
	relationsFILE = NULL;

	return rc;
}

int CRelationHashtable::FindFileIndex(main_sieving_type aPrime, int aPartIndex, int & aIndex) const
{
	if (this->iTotalFileCount == 0)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","FindFileIndex ",0,0,"Hashtable is not setup.");
		return ConstRC::GeneralError;
	}

	if (aPartIndex < 0 || aPartIndex >= this->iRelationPartCount)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","FindFileIndex ",0,0,"Relation part out of range.");
		return ConstRC::BadArgument;
	}

	aIndex = 0;

	main_sieving_type current = this->iMinPrimes[aPartIndex];
	while (current <= aPrime)
	{
		current += this->iBlockLenghts[aPartIndex];
		aIndex++;
	}

	aIndex--;

	return ConstRC::Ok;
}

int CRelationHashtable::ConvertTotalIndexToFileIndex(int aTotalIndex, int & aPartIndex, int & aIndex) const
{
	aIndex = -1;
	aPartIndex = -1;
	for (int i = 0; i < this->iRelationPartCount; i++)
	{
		if (aTotalIndex < this->iFileCounts[i])
		{
			aIndex = aTotalIndex;
			aPartIndex = i;
			return ConstRC::Ok;
		}
		else
		{
			aTotalIndex -= this->iFileCounts[i];
		}
	}

	CLog::LogError(WHEREARG,"CRelationHashtable","ConvertTotalIndexToFileIndex ",0,0,"Total index file out of range.");
	return ConstRC::BadArgument;
}

int CRelationHashtable::AddToTable(main_sieving_type aPrime, CFNFSRelation *aRelation, relation_index_type & aIndex)
{
	if (this->iHashtable == NULL || aRelation == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","AddToTable",0,0,"Hashtable or Relation is NULL.");
		return ConstRC::BadArgument;
	}

	bool present = false;

	if (FindRelationInHashtable(aPrime, aRelation, present, aIndex) != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}
	
	if (present == false) 
	{
		// This relation has not been in hashtable yet.
		// We create copy of given relation and save it to the right position in hashtable
		this->iHashtable[aIndex] = new CFNFSRelation(*aRelation);
	} 

	return ConstRC::Ok;
}

/*!
	This method tries to find relation according to aPrime in hashtable, and indicates
	the result of this search in two values: the return value, which is an index
	corresponding to the real or expected position of the relation in the hashtable,
	and value of aPresent, which indicates whether the relation has already been
	in the hashtable or no. In the second case, the return value has meaning
	"if you want to insert relation, this index is the one where it should be inserted".

*/
int CRelationHashtable::FindRelationInHashtable(main_sieving_type aPrime, CFNFSRelation * aRelation, bool & aPresent, relation_index_type & aHashtableIndex) const
{
	if (this->iHashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","FindRelationInHashtable",0,0,"Hashtable is NULL.");
		aPresent = false;
		aHashtableIndex = 0;
		return ConstRC::BadArgument;
	}

	relation_index_type hash = (aPrime >> 1) & this->iMask;			
	if (CollisionResolution(hash, aRelation, aHashtableIndex) != ConstRC::Ok)
	{
		aPresent = false;
		aHashtableIndex = 0;
		return ConstRC::GeneralError;
	}

    if (this->iHashtable[aHashtableIndex] == NULL) 
	{
		aPresent = false;
	} 
	else 
	{
		aPresent = true;
	}

	return ConstRC::Ok;
}

/*!
	
	A collision resolution algorithm for insertion into hashtable. Returns in aIndex
	the index of first suitable free position, where relation can be stored.

*/
int CRelationHashtable::CollisionResolution(relation_index_type aHash, CFNFSRelation * aRelation, relation_index_type & aRelationIndex) const
{
	aRelationIndex = aHash;
	relation_index_type iter = 0;

	if (this->iHashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","CollisionResolution",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	CFNFSRelation ** aux = this->iHashtable + aRelationIndex;
	CFNFSRelation ** end = this->iHashtable + this->iMaxNbrRelationInFile;

	while (*aux != NULL && ((*aux)->A != aRelation->A || (*aux)->B != aRelation->B))
	{
		iter++;
		aux++;
		// Should we go over the bounds of the table, we must reset aux to start.
		if (aux >= end) 
		{
			if (iter > this->iMaxNbrRelationInFile) 
			{
				// this shouldn't happend so we use just stupid control
				CLog::LogError(WHEREARG,"CRelationHashtable","CollisionResolution",0,0,"The hash table has overflown.");
				return ConstRC::GeneralError;
			}
			aux = this->iHashtable;
		}
	}

	aRelationIndex = aux - this->iHashtable;

	return ConstRC::Ok;
}

void CRelationHashtable::ResetCurrentFileIndex()
{
	remove(this->iTempFileName.c_str());
	this->iCurrentFileIndex = -1;
}

int CRelationHashtable::IncreaseCurrentFileIndex()
{
	return IncreaseCurrentFileIndex(false);
}

int CRelationHashtable::IncreaseCurrentFileIndex(bool aDeleteTemp)
{
	if (this->iCurrentFileIndex >= -1 && this->iCurrentFileIndex < this->iTotalFileCount)
	{
		if (this->iCurrentFileIndex > -1 && aDeleteTemp)
		{
			// we need to know filename of current file
			string strFilename = "";
			int rc = GetCurrentFilename(strFilename);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
			// we overwrite store file with temp file where are relations which should be processed in next round
			if (MSerializable::ExistFile(this->iTempFileName))
			{
				rc = MSerializable::RenameFile(this->iTempFileName, strFilename);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
		}

		this->iCurrentFileIndex++;
		if (this->iCurrentFileIndex >= this->iTotalFileCount)
		{
			return ConstRC::All;
		}

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","IncreaseCurrentFileIndex",0,0,"Current file index is out of range.");
		return ConstRC::InconsistentState;
	}
}

int CRelationHashtable::GetFilename(int aIndex, string & aFileName) const
{
	if (aIndex > -1 && aIndex < this->iTotalFileCount)
	{
		int index = -1;
		int partIndex = -1;
		ConvertTotalIndexToFileIndex(aIndex, partIndex, index);
		aFileName = this->iBaseFileName + "_" + Utils::ConvertToString(partIndex) + "_" + Utils::ConvertToString(index) + TXT_FILE_EXTENSION;

		return ConstRC::Ok;
	}
	else
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","GetFilename ",0,0,"Current file index is out of range.");
		return ConstRC::InconsistentState;
	}
}

int CRelationHashtable::GetCurrentFilename(string & aFileName) const
{
	return GetFilename(this->iCurrentFileIndex, aFileName);
}
/*
	This method saves current loaded hashtable with relations to the store files.
	It uses AddToFile method for saving relations.
*/
int CRelationHashtable::StoreCurrentHashtable()
{
	if (this->iHashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","StoreCurrentHashtable",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	int rc;
	CFNFSRelation ** aux = this->iHashtable;
	CFNFSRelation ** end = this->iHashtable + this->iMaxNbrRelationInFile;

	while (aux < end)
	{
		if (*aux != NULL)
		{
			rc = AddToFile(*aux);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		aux++;
	}

	return ConstRC::Ok;
}
/*
	This method just call GetFromFilePartAndBounds with aFileIndex set to current file index.
	For more see GetFromFilePartAndBounds.
*/
int CRelationHashtable::GetFromFilePartAndBounds(int &aPartIndex, main_sieving_type &aMinBound, main_sieving_type &aMaxBound)
{
	return GetFromFilePartAndBounds(this->iCurrentFileIndex, aPartIndex, aMinBound, aMaxBound);
}

/*
	This method calculate according to file index (total file index) part of relation and bounds for large prime ideals 
	which are used as a criterion for adding relation to the file with given file index.

	Relation can be add to the file with given index if has some large prime ideal on odd power which is in the result relation part and between the result bounds.
*/
int CRelationHashtable::GetFromFilePartAndBounds(int aFileIndex, int &aPartIndex, main_sieving_type &aMinBound, main_sieving_type &aMaxBound)
{
	if (this->iFileCounts == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","GetFromFilePartAndBounds ",0,0,"Hashtable is not setup.");
		return ConstRC::GeneralError;
	}

	if (aFileIndex < 0 || aFileIndex >= this->iTotalFileCount)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","GetFromFilePartAndBounds ",0,0,"File index is out of range.");
		return ConstRC::InconsistentState;
	}

	// firstly, we need to know in which part of relation we are
	aPartIndex = 0;
	int auxInt = aFileIndex;
	for (int i = 0; i < this->iRelationPartCount; i++)
	{
		if (auxInt < this->iFileCounts[i])
		{
			aPartIndex = i;
			break;
		}
		else
		{
			auxInt -= this->iFileCounts[i];
		}
	}
	// now we can calculate current block bounds in current relation part
	aMinBound = this->iMinPrimes[aPartIndex] + auxInt * this->iBlockLenghts[aPartIndex];
	aMaxBound = aMinBound + this->iBlockLenghts[aPartIndex];

	return ConstRC::Ok;
}
int CRelationHashtable::DeleteRelation(relation_index_type aIndex)
{
	if (this->iHashtable == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","DeleteRelation",0,0,"Hashtable is NULL.");
		return ConstRC::BadArgument;
	}

	if (this->iHashtable[aIndex] != NULL)
	{
		delete (this->iHashtable[aIndex]);
		this->iHashtable[aIndex] = NULL;
	}
	else
	{
		CLog::LogError(WHEREARG,"CRelationHashtable","DeleteRelation",0,0,"Relation is already deleted.");
		return ConstRC::BadArgument;
	}

	return ConstRC::Ok;
}

