#ifndef __RELATION_HASHTABLE_H__
#define __RELATION_HASHTABLE_H__

#include "filtering_nfs_relation.h"


/*!
 * \brief used in Filtering phase for merging.
 *
	Hashtable contains invidual relations which are saved in files bacause of their count.
	For processing all relations from one file are loaded to the memory.

	Relations are put to the files according to large prime ideals. We have file groups for each relation parts.
	Prime ideal are sorted by size and according to relation part (first part sorted by size, second part sorted by size etc.).
	Relations are save in file according to choosen prime ideal (we want relations with same prime ideal to be close to each other).
		- Relations are save in the right file subsequently. They are just appended to the file.
		
	Relations loaded to memory are saved as in hashtable.
		- Hash is calculated from given prime (from large prime ideal) and hash mask.
		- Collisions are treated by finding the first free position (Each relation is unique - is only once in hashtable).

	Number of relations in file is defined by memory size and relation size.
	Number of files is defined by count of relations with multiplicity per relation part and number of relations in file.
	When we have choosen large prime ideal of given relation then we save relation to the file according to size of this large prime ideal.
		- We have divided the interval (min large prime ideal, max large prime ideal) according to the number of files.
		- We choose the large prime ideal of given relation as follows:
			- If parameter fileindex is not given (fileindex = -1), we choose the first large prime ideal
			- If we have parameter fileindex:
				- If the large prime ideal, which is larger than min size for given interval, does not exist then ...
*/
class CRelationHashtable
{
private:
	///Security margin for the field of hashtables
	static const int SECURITY_MARGIN = 1;

private:
	///Base name of all files used as hashtable store files (also with directory path and identifier if necessary)
	std::string iBaseFileName;
	///Name of the file with relations which are processed (also with directory path and identifier if necessary)
	std::string iProcessedFileName;
	///Name of the temp file with processed relations from current file (also with directory path and identifier if necessary)
	std::string iTempFileName;
	///Total number of hashtable store files
	int iTotalFileCount;
	///Number of hashtable store files for each relation part
	int * iFileCounts;
	///Count of relation parts
	int iRelationPartCount;
	///Hash mask - hash is defined by file nbr. and in file by large prime AND mask (same for each relation part)
	relation_index_type	iMask;
	/*
		Maximum number of relation in hashtable file (same for each relation part).
		But we want less relations in hashtable (cca 80% fullness).
		It is also size of hashtable.
		Allocated size is bigger (+ SECURITY_MARGIN) - security margin.
			- for better iteration
	*/
	relation_index_type iMaxNbrRelationInFile;
	///Current file index which is used for processing relations
	int iCurrentFileIndex;
	///Own field of minimum primes for each relation part
	main_sieving_type * iMinPrimes;
	///Own field of maximum primes for each relation part
	main_sieving_type * iMaxPrimes;
	///Field of block length for each relation part (division of the interval [min prime, max prime])
	main_sieving_type * iBlockLenghts;
	///Hashtable with loaded relations
	CFNFSRelation ** iHashtable;
	

public:

	CRelationHashtable();
	~CRelationHashtable();

	void DeleteHashtable();
	int ClearHashtableFiles();
	int ClearHashtable();
	int CreateHashtableFiles();
	int SetupHashtable(relation_index_type * aMultiRelationCounts, long aRelationsInMemory, 
						main_sieving_type * aMinPrimes, main_sieving_type * aMaxPrimes, int aRelationPartCount);
	int SetDirectoryAndIdentifier(std::string aWorkingDirectoryPath, std::string aIdentifier);

	int AddToFile(CFNFSRelation * aRelation, int aFileIndex);
	int AddToFile(CFNFSRelation * aRelation);

	int AddToTable(main_sieving_type aPrime, CFNFSRelation * aRelation, relation_index_type & aIndex);

	int FindRelationInHashtable(main_sieving_type aPrime, CFNFSRelation * aRelation, bool & aPresent, relation_index_type & aHashtableIndex) const;

	// methods for processing relations in Filtering phase
	int GetFilename(int aIndex, string & aFileName) const;
	int GetCurrentFilename(std::string & aFileName) const;
	void ResetCurrentFileIndex();
	int IncreaseCurrentFileIndex();
	int IncreaseCurrentFileIndex(bool aDeleteTemp);

	int StoreCurrentHashtable();

	int GetFromFilePartAndBounds(int & aPartIndex, main_sieving_type& aMinBound, main_sieving_type& aMaxBound);
	int GetFromFilePartAndBounds(int aFileIndex, int & aPartIndex, main_sieving_type& aMinBound, main_sieving_type& aMaxBound);

	int DeleteRelation(relation_index_type aIndex);

public:

	relation_index_type Get_MaxNbrRelationInFile() const { return this->iMaxNbrRelationInFile; }

	void Get_ProcessedFileName(std::string & aFileName) const { aFileName = this->iProcessedFileName; } 

	// Directly get or set relations in hashtable - WARNING no control!!!
	CFNFSRelation * Get_Relation(relation_index_type aIndex) const { return this->iHashtable[aIndex]; }
	void Set_Relation(relation_index_type aIndex, CFNFSRelation * aRelation) { this->iHashtable[aIndex] = aRelation; }

private:

	CRelationHashtable(const CRelationHashtable& aOperand) {}

	int CollisionResolution(relation_index_type aHash, CFNFSRelation * aRelation, relation_index_type & aRelationIndex) const;

	int FindFileIndex(main_sieving_type aPrime, int aPartIndex, int & aIndex) const;
	int ConvertTotalIndexToFileIndex(int aTotalIndex, int & aPartIndex, int & aIndex) const;

};

#endif // __RELATION_HASHTABLE_H__
