#include "relation_part.h"

#include <cstring>
#include "base_parameters.h"
#include "const_debug.h"
#include "const_return_codes.h"
#include "const_filenames.h"
#include "const_element_flags.h"
#include "const_parameter_names.h"
#include "log.h"
#include "xml_helper.h"


using namespace std;

CRelationPart::CRelationPart(const string aDescription) : MSerializable(ConstFN::RELATION_PART),
	iDescription(aDescription)
{
	mpz_init(this->Norm);
	mpz_init(this->NormRemaining);

	this->Poly = new Polynomial(1);
	this->Poly->Set(0, 1);
}

CRelationPart::CRelationPart(const CRelationPart& aOperand) : 
	MSerializable(aOperand), iDescription(aOperand.iDescription), Large(aOperand.Large), Main(aOperand.Main)
{
	mpz_init_set(this->Norm, aOperand.Norm);
	mpz_init_set(this->NormRemaining, aOperand.NormRemaining);

	this->Poly = new Polynomial();
	aOperand.Poly->Copy(this->Poly);
}

CRelationPart::~CRelationPart()
{
	if (this->Poly != NULL)
	{
		delete this->Poly;
		this->Poly = NULL;
	}

	mpz_clear(this->Norm);
	mpz_clear(this->NormRemaining);
}
int CRelationPart::Copy2(CRelationPart& aOperand)
{
	aOperand.Set_Description(this->iDescription);
	aOperand.Set_DefaultFileName(this->Get_DefaultFileName());
	aOperand.Poly->Zeroize();
	this->Poly->Copy(aOperand.Poly);
	
	mpz_set(aOperand.Norm, this->Norm);
	mpz_set(aOperand.NormRemaining, this->NormRemaining);

	this->Main.Copy2(aOperand.Main);
	this->Large.Copy2(aOperand.Large);

	return ConstRC::Ok;
}
bool CRelationPart::Equals(const CRelationPart& aOperand) const
{
	if (mpz_cmp(this->Norm, aOperand.Norm) != 0)
	{
		return false;
	}
	if (mpz_cmp(this->NormRemaining, aOperand.NormRemaining) != 0)
	{
		return false;
	}
	if (this->Poly->Equals(*aOperand.Poly) == false)
	{
		return false;
	}
	if (this->Main.Equals(aOperand.Main) == false)
	{
		return false;
	}
	if (this->Large.Equals(aOperand.Large) == false)
	{
		return false;
	}
	return true;
}
int CRelationPart::Reset()
{
	this->Main.MaxElementIndex = -1;
	this->Large.MaxElementIndex = -1;
	this->Poly->Zeroize();
	mpz_set_si(this->Norm, 0);
	mpz_set_si(this->NormRemaining, 0);

	return ConstRC::Ok;
}
int CRelationPart::CombineWithPart(const CRelationPart &aOperand, Polynomial * aThetaPoly)
{
	// check if it is not just integral part...
	if (aThetaPoly != NULL && aThetaPoly->GetDegree() > 1)
	{
		Polynomial* auxiliary = this->Poly->AssignAuxPoly();
		Polynomial* quotient  = this->Poly->AssignAuxPoly();
		Polynomial* remainder = this->Poly->AssignAuxPoly();	

		auxiliary->Multiply(auxiliary, this->Poly, aOperand.Poly);
		remainder->Divide(quotient, remainder, auxiliary, aThetaPoly);
		remainder->Copy(this->Poly);

		auxiliary->Dispose(auxiliary);
		quotient->Dispose(quotient);
		remainder->Dispose(remainder);
	}

#ifdef DEBUG_RELATION_PART_COMBINE_NORMS
	mpz_mul(this->Norm, this->Norm, aOperand.Norm);
#endif

	int rc = this->Main.CombineWithDivisors(aOperand.Main);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = this->Large.CombineWithDivisors(aOperand.Large);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CRelationPart::CombineWithPart(const CRelationPart &aOperand)
{
#ifdef DEBUG_RELATION_PART_COMBINE_NORMS
	mpz_mul(this->Norm, this->Norm, aOperand.Norm);
#endif

	int rc = this->Main.CombineWithDivisors(aOperand.Main);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = this->Large.CombineWithDivisors(aOperand.Large);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
void CRelationPart::PrintToScreen()
{
	gmp_printf("N(a,b) = %Zd.\n", this->Norm);	
	gmp_printf("Remaining = %Zd\n",this->NormRemaining);
	printf("PrimeIdeals:\n");
	this->Main.PrintToScreen();
	printf("Large PrimeIdeals:\n");
	this->Large.PrintToScreen();
}

/*!
	This method is used to save this relation part into a file.
	File Template:
	==================================================
	Poly
	Norm
	NormRemaining
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
	Large PrimeIdeals count
	Large PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CRelationPart::PrintToFile(FILE * aFw)
{
	this->iAuxStr = this->Poly->ToString(ELongFormat);
	fprintf(aFw, "%s\n", this->iAuxStr.c_str());
	Utils::PrintMpzToFile(aFw, 10, this->Norm);
	fprintf(aFw,"\n");
	Utils::PrintMpzToFile(aFw, 10, this->NormRemaining);
	fprintf(aFw,"\n");

	int rc = this->Main.PrintToFile(aFw);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = this->Large.PrintToFile(aFw);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
/*!
	This method is used to read this relation part from a file.
	File Template:
	==================================================
	Poly
	Norm
	NormRemaining
	Main PrimeIdeals count
	Main PrimeIdeals as (prime|exponent|c_p|flags)
	Large PrimeIdeals count
	Large PrimeIdeals as (prime|exponent|c_p|flags)
*/
int CRelationPart::ReadFromFile(FILE * aFr, char * aBuffer, int aBufferSize)
{
	// Poly
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	this->iAuxStr.assign(aBuffer);
	int rc = Polynomial::Parse2(this->Poly, this->iAuxStr);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// Norm
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	mpz_set_str(this->Norm, aBuffer, 10);

	// NormRemaining
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	mpz_set_str(this->NormRemaining, aBuffer, 10);

	// Get count of PrimeIdeals
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	long count = atol(aBuffer);

	// Read PrimeIdeals as (prime|exponent|c_p|flags)
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	if (count > 0)
	{
		this->iAuxStr.assign(aBuffer);
		rc = this->Main.ReadFromFile(count, this->iAuxStr);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		this->Main.MaxElementIndex = -1;
	}

	// Get count of Large PrimeIdeals
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	count = atol(aBuffer);

	// Read Large PrimeIdeals as (prime|exponent|c_p|flags)
	if (fgets(aBuffer, aBufferSize, aFr) == NULL)
	{
		return ConstRC::EndOfFile;
	}
	if (count > 0)
	{
		this->iAuxStr.assign(aBuffer);
		rc = this->Large.ReadFromFile(count, this->iAuxStr);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		this->Large.MaxElementIndex = -1;
	}

	return ConstRC::Ok;
}
int CRelationPart::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_RELATION_PART_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CRelationPart","Serialize",rc,0,"Could not write the start element.");
    	return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeString(aWriter, XML_SERIAL_DESCRIPTION_NAME, this->iDescription);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	// serialize polynomial in string form
	rc = CXMLHelper::SerializeString(aWriter, XML_SERIAL_POLYNOMIAL_NAME, this->Poly->ToString(ELongFormat));
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_ALGEBRAIC_NORM_NAME, this->Norm);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_ALGEBRAIC_NORM_REMAINING_NAME, this->NormRemaining);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}	

	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_PRIMEIDEALS_NAME);
	rc = this->Main.Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	param.SetParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_LARGE_PRIMEIDEALS_NAME);
	rc = this->Large.Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CRelationPart","Serialize",rc,0,"Could not end the start element.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;  
}

int CRelationPart::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(PARAM_SERIALIZATION_ELEMENT_NAME, XML_SERIAL_RELATION_PART_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationPart","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CRelationPart","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);
	CBaseParameters param;

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_DESCRIPTION_NAME))
		{
			rc = CXMLHelper::DeserializeString(aReader, "Deserialize description...", this->iDescription, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_POLYNOMIAL_NAME))
		{
			rc = CXMLHelper::DeserializeString(aReader, "Deserialize polynomial...", this->iAuxStr, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			rc = Polynomial::Parse2(this->Poly, this->iAuxStr);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ALGEBRAIC_NORM_NAME))
		{
			rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize algebraic norm...", this->Norm, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ALGEBRAIC_NORM_REMAINING_NAME))
		{
			rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize algebraic norm remaining...", this->NormRemaining, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PRIMEIDEALS_NAME))
		{
			param.SetParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_PRIMEIDEALS_NAME);
			rc = this->Main.Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_LARGE_PRIMEIDEALS_NAME))
		{
			param.SetParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_LARGE_PRIMEIDEALS_NAME);
			rc = this->Large.Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
