#include "relation_reader.h"

#include "nfs_relation.h"
#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "base_parameters.h"
#include "log.h"


using namespace std;

CRelationReader::CRelationReader(const CBaseParameters* aParameters)
{
	this->iParameters = aParameters;
	this->iReader = NULL;
	this->iCurrentRelation = NULL;
	this->iCount = RELATION_UNDEFINED_COUNT;
	this->iCurrentIndex = -1;
	this->iType = ERelAbstract;
}
CRelationReader::~CRelationReader()
{
	if (this->iReader != NULL)
	{
		xmlFreeTextReader(this->iReader);
		this->iReader = NULL;
	}
	this->iParameters = NULL;
	this->iCurrentRelation = NULL;
}
int CRelationReader::Reset()
{
	if (this->iReader != NULL)
	{
		xmlFreeTextReader(this->iReader);
		this->iReader = NULL;
	}
	this->iCurrentRelation = NULL;

	this->iCurrentIndex = -1;

	return ConstRC::Ok;
}
int CRelationReader::Close()
{
	return FinishReading();
}
int CRelationReader::FinishReading()
{
	if (this->iReader != NULL)
	{
		xmlFreeTextReader(this->iReader);
		this->iReader = NULL;
	}
	this->iParameters = NULL;
	this->iCurrentRelation = NULL;

	this->iCount = RELATION_UNDEFINED_COUNT;
	this->iCurrentIndex = -1;
	this->iType = ERelAbstract;

	return ConstRC::Ok;
}
int CRelationReader::PrepareForReading()
{
	if (Reset() != ConstRC::Ok)
	{
		return ConstRC::GeneralError;
	}
	if (this->iParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationReader","PrepareForReading",0,0,"Parameters are NULL.");
		return ConstRC::GeneralError;
	}

	string mainTagName;	
	this->iParameters->GetStringParameter(ConstPN::SERIALIZATION_MAIN_ELEMENT_NAME, XML_SERIAL_ELEMENTS_NAME, mainTagName);

	// may be main element is 'elements'
	int rc = MSerializable::PrepareXMLReader(*this->iParameters, mainTagName, this->iReader);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	this->iCount = RELATION_UNDEFINED_COUNT;
	this->iType = ERelAbstract;

	// Type + Count
	if (!xmlTextReaderHasAttributes(this->iReader))
	{
		// this could be problem...
		CLog::LogError(WHEREARG,"CRelationReader","PrepareForReading",0,0,"No attributes of main element.");
		// next element is already first relation
		return CXMLHelper::MoveToNextNode(this->iReader);
	}

	xmlChar* attr = xmlTextReaderGetAttribute(this->iReader, BAD_CAST XML_SERIAL_COUNT_ATTR);  
	if (attr != NULL)
	{
		rc = CXMLHelper::ParseDecimalInt(attr, this->iCount);
		xmlFree(attr);
		attr = NULL;
		if (rc != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CRelationReader","PrepareForReading",0,0,"Could not parse int");			
			return ConstRC::GeneralError;	
		}
	}

	attr = xmlTextReaderGetAttribute(this->iReader, BAD_CAST XML_SERIAL_TYPE_ATTR);  
	if (attr != NULL)
	{
		int auxInt;
		rc = CXMLHelper::ParseDecimalInt(attr, auxInt);
		xmlFree(attr);
		attr = NULL;
		if (rc != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CRelationReader","PrepareForReading",0,0,"Could not parse int");			
			return ConstRC::GeneralError;	
		}
		this->iType = (relation_types)auxInt;
	}

	return ConstRC::Ok;
}
int CRelationReader::ReadRelation(ARelation * aRelation)
{
	if (this->iParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationReader","ReadRelation",0,0,"Parameters are NULL.");
		return ConstRC::GeneralError;
	}
	if (aRelation == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationReader","ReadRelation",0,0,"Relation is NULL.");
		return ConstRC::BadArgument;
	}
	int rc;
	if (this->iReader == NULL)
	{
		// first reading - we get basic info and move iReader to the first relation
		rc = PrepareForReading(); 
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	//switch (this->iType)
	//{
	//	case ERelNfs:
	//		// values are not important because they will change during deserialization
	//		*aRelation = new CNFSRelation(RELATION_DEFAULT_POLY_DEGREE, RELATION_DEFAULT_MODE);
	//		break;
	//	case ERelMpqs:
	//		CLog::LogError(WHEREARG,"CRelationReader","ReadRelation",0,0,"MPQS relations are not supported.");
	//		return ConstRC::NotSupported;
	//	case ERelAbstract:
	//		CLog::LogError(WHEREARG,"CRelationReader","ReadRelation",0,0,"Abstract relation is not possible.");
	//		return ConstRC::CannotBeDone;
	//	default:
	//		CLog::LogError(WHEREARG,"CRelationReader","ReadRelation",0,0,"Relation is not specified!");
	//		return ConstRC::NotSpecified;
	//}

	if (aRelation->Get_ClassType() != this->iType)
	{
		CLog::LogError(WHEREARG,"CRelationReader","ReadRelation",0,0,"Incorrect type of relation.");
		return ConstRC::BadArgument;
	}

	rc = CXMLHelper::MoveToNextNode(this->iReader);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = aRelation->Deserialize(*this->iParameters, this->iReader);

	this->iCurrentIndex++;
	this->iCurrentRelation = aRelation;

	// TODO some control for end of file
	if (rc != ConstRC::Ok)
	{
		this->iCount = this->iCurrentIndex + 1;
		return ConstRC::All;
	}

	return ConstRC::Ok;
}
