#include "relation_writer.h"

#include "nfs_relation.h"
#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_filenames.h"
#include "base_parameters.h"
#include "relation_reader.h"
#include "log.h"


using namespace std;

CRelationWriter::CRelationWriter(const CBaseParameters* aParameters, int aCount)
{
	this->iParameters = aParameters;
	this->iWriter = NULL;
	this->iCount = aCount;
	this->iCurrentIndex = -1;
	this->iType = ERelAbstract;
}
CRelationWriter::~CRelationWriter()
{
	if (this->iWriter != NULL)
	{
		FinishWriting();
	}
	this->iParameters = NULL;
}
int CRelationWriter::Close()
{
	if (this->iWriter != NULL)
	{
		return FinishWriting();
	}

	return ConstRC::Ok;
}
int CRelationWriter::FinishWriting()
{
	if (this->iWriter != NULL)
	{
		int rc = xmlTextWriterEndElement(this->iWriter); // mainTagName
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CRelationWriter","FinishWriting",rc,0,"Could not end the main tag.");
			xmlFreeTextWriter(this->iWriter);
			this->iWriter = NULL;
			this->iParameters = NULL;
			return ConstRC::GeneralError;
		}

		rc = CXMLHelper::NewLine(this->iWriter,FALSE);
		if (rc != ConstRC::Ok)
		{
			xmlFreeTextWriter(this->iWriter);
			this->iWriter = NULL;
			this->iParameters = NULL;
			return rc;
		}

		// end the document
		rc = xmlTextWriterEndDocument(this->iWriter);

		xmlFreeTextWriter(this->iWriter);
		this->iWriter = NULL;
		if (rc < 0) 
		{
			CLog::LogError(WHEREARG,"CRelationWriter","FinishWriting",rc,0,"Could not close the document.");
			this->iParameters = NULL;
			return ConstRC::GeneralError;
		}
	}

	this->iParameters = NULL;

	this->iCount = RELATION_UNDEFINED_COUNT;
	this->iCurrentIndex = -1;
	this->iType = ERelAbstract;

	return ConstRC::Ok;
}
int CRelationWriter::PrepareForWriting()
{
	if (this->iWriter != NULL)
	{
		return ConstRC::GeneralError;
	}

	int compressionLevel;

	// get parameters for serialization - with default values
	this->iParameters->GetIntegerParameter(ConstPN::COMPRESSION_LEVEL, DEFAULT_COMPRESSION_LEVEL, compressionLevel);
	
	// get file name with path
	string fileNameFull = MSerializable::CreateFileNameAuto(*this->iParameters, ConstFN::RELATIONS);

	this->iWriter = xmlNewTextWriterFilename(fileNameFull.c_str(), compressionLevel);

	// we write only start of document now
	int rc = xmlTextWriterStartDocument(this->iWriter, NULL, XML_ENCODING, NULL);
    if (rc < 0) 
	{
	    CLog::LogError(WHEREARG,"CRelationWriter","PrepareForWriting",rc,0,"Could not write the document start.");
		xmlFreeTextWriter(this->iWriter);
		this->iWriter = NULL;
		return ConstRC::GeneralError;        	
    }

	string mainTagName;	
	this->iParameters->GetStringParameter(ConstPN::SERIALIZATION_MAIN_ELEMENT_NAME, XML_SERIAL_ELEMENTS_NAME, mainTagName);

	rc = xmlTextWriterStartElement(this->iWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CRelationWriter","PrepareForWriting",rc,0,"Could not write the main tag.");
		xmlFreeTextWriter(this->iWriter);
		this->iWriter = NULL;
		return ConstRC::GeneralError;
    }

	rc = xmlTextWriterWriteFormatAttribute(this->iWriter, BAD_CAST XML_SERIAL_COUNT_ATTR,"%d",this->iCount);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CRelationWriter","PrepareForWriting",rc,0,"Could not write the count.");
        xmlFreeTextWriter(this->iWriter);
		this->iWriter = NULL;
		return ConstRC::GeneralError;
    }

	rc = xmlTextWriterWriteFormatAttribute(this->iWriter, BAD_CAST XML_SERIAL_TYPE_ATTR,"%d",(int)this->iType);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CRelationWriter","PrepareForWriting",rc,0,"Could not write the type.");
        xmlFreeTextWriter(this->iWriter);
		this->iWriter = NULL;
		return ConstRC::GeneralError;
    }

	rc = CXMLHelper::NewLine(this->iWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		xmlFreeTextWriter(this->iWriter);
		this->iWriter = NULL;
		return rc;
	}

	return ConstRC::Ok;
}
int CRelationWriter::WriteRelation(ARelation *aRelation)
{
	if (this->iParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationWriter","WriteRelation",0,0,"Parameters are NULL.");
		return ConstRC::GeneralError;
	}
	if (aRelation == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationWriter","WriteRelation",0,0,"Relation is NULL.");
		return ConstRC::BadArgument;
	}
	if (this->iWriter == NULL)
	{
		this->iType = aRelation->Get_ClassType();
		// first writing - prepare file - main tag with attributes
		int rc = PrepareForWriting(); 
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	this->iCurrentIndex++;
	return aRelation->Serialize(*this->iParameters, this->iWriter);
}
int CRelationWriter::CopyRelationsFrom(const CBaseParameters &aParam)
{
	if (this->iParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Parameters are NULL.");
		return ConstRC::GeneralError;
	}
	int rc;

	xmlTextReaderPtr reader = NULL;
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_MAIN_ELEMENT_NAME, XML_SERIAL_ELEMENTS_NAME, mainTagName);

	// may be main element is 'elements'
	rc = MSerializable::PrepareXMLReader(aParam, mainTagName, reader);
	if (rc != ConstRC::Ok)
	{
		if (reader != NULL)
		{
			xmlFreeTextReader(reader);
		}
		return rc;
	}

	// Type + Count
	if (xmlTextReaderHasAttributes(reader))
	{
		int auxInt;
		xmlChar* attr = xmlTextReaderGetAttribute(reader, BAD_CAST XML_SERIAL_COUNT_ATTR);  
		if (attr != NULL)
		{
			rc = CXMLHelper::ParseDecimalInt(attr, auxInt);
			xmlFree(attr);
			attr = NULL;
			if (rc != ConstRC::Ok)
			{
				CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Could not parse int");			
				xmlFreeTextReader(reader);
				return ConstRC::GeneralError;	
			}
			if (this->iCount != RELATION_UNDEFINED_COUNT && auxInt != RELATION_UNDEFINED_COUNT)
			{
				this->iCount += auxInt;
			}
		}

		attr = xmlTextReaderGetAttribute(reader, BAD_CAST XML_SERIAL_TYPE_ATTR);  
		if (attr != NULL)
		{
			rc = CXMLHelper::ParseDecimalInt(attr, auxInt);
			xmlFree(attr);
			attr = NULL;
			if (rc != ConstRC::Ok)
			{
				CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Could not parse int");			
				xmlFreeTextReader(reader);
				return ConstRC::GeneralError;	
			}
			if (this->iType != ERelAbstract && this->iType != (relation_types)auxInt)
			{
				CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Relation type mismatch.");			
				xmlFreeTextReader(reader);
				return ConstRC::GeneralError;
			}
			this->iType = (relation_types)auxInt;
		}
	}

	if (this->iWriter == NULL)
	{
		if (this->iType == ERelAbstract)
		{
			CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Cannot determine relation type.");			
			xmlFreeTextReader(reader);
			return ConstRC::GeneralError;
		}
		// first writing - prepare file - main tag with attributes
		rc = PrepareForWriting(); 
		if (rc != ConstRC::Ok)
		{
			xmlFreeTextReader(reader);
			return rc;
		}
	}

	xmlChar * outer = NULL;
	int depth = xmlTextReaderDepth(reader);
	int maxDepth = depth + 2;

	// now we read all relations - they should be same...
	// the reader is on the start element of current relation
	while ((rc = CXMLHelper::MoveToNextNode(reader, depth, maxDepth)) == ConstRC::Ok)
	{
		outer = xmlTextReaderReadOuterXml(reader);
		if (outer == NULL)
		{
			CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Readed relation is empty.");
			rc = ConstRC::GeneralError;
			break;
		}
		rc = xmlTextWriterWriteRaw(this->iWriter, outer);
		xmlFree(outer);
		if (rc == -1)
		{
			CLog::LogError(WHEREARG,"CRelationWriter","CopyRelationsFrom",0,0,"Cannot write xml text.");
			rc = ConstRC::GeneralError;
			break;
		}
		rc = CXMLHelper::NewLine(this->iWriter, FALSE);
		if (rc != ConstRC::Ok)
		{
			break;
		}
	}

	xmlFreeTextReader(reader);

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
