#include "square_root_phase.h"

#include "abstract_rel_processing_phase.h"
#include "matrix_helper.h"
#include "const_square_root_phase.h"
#include "const_parameter_names.h"
#include "const_element_flags.h"
#include "test_suite.h"
#include "integral_domain.h"
#include "integer_class.h"
#include "log.h"


using namespace std;

CSquareRootPhase::CSquareRootPhase() : MSerializable(ConstFN::SQUARE_ROOT_PHASE)
{
	Set_AutoSerializable(ConstSRPhase::AUTO_SERIALIZABLE);
	Set_InnerState((int)ESRPhaseNotStarted);

	this->iAlgElementsCount = 0;
	this->iStartOfAlg = 0;

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iSievingPoly[i] = NULL;
		this->iThetaPoly[i] = NULL;
		this->iAlgElements[i] = NULL;
		this->iAlgElementsAllocated[i] = 0;
	}

	this->iRelation = NULL;

	this->iTimes = new CTimeContainer(ConstSRPhase::INFO_TIME_COUNT);
	this->iStats = new CStatContainer(ConstSRPhase::INFO_STAT_COUNT);
	
	this->iRootFinderType = ENewton;
}

CSquareRootPhase::~CSquareRootPhase()
{
	Reset();
	if (this->iTimes != NULL)
	{
		delete this->iTimes;
		this->iTimes = NULL;
	}
	if (this->iStats != NULL)
	{
		delete this->iStats;
		this->iStats = NULL;
	}
	if (this->iMatrixResult != NULL) 
	{
		delete this->iMatrixResult;
		this->iMatrixResult = NULL;
	};

	DisposeFunctorField();
	DisposeGMP();
	DisposeMutexes();
	DisposeAcceptedReturnCodes();
}
int CSquareRootPhase::Reset()
{
	this->iStartOfAlg = 0;
	this->iAlgElementsCount = 0;
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iSievingPoly[i] != NULL)
		{
			delete this->iSievingPoly[i];
			this->iSievingPoly[i] = NULL;
		}
		if (this->iThetaPoly[i] != NULL)
		{
			delete this->iThetaPoly[i];
			this->iThetaPoly[i] = NULL;
		}
		if (this->iAlgElements[i] != NULL)
		{
			DeletePolyField(this->iAlgElements[i], this->iAlgElementsAllocated[i]);
		}
	}

	if (this->iRelation != NULL)
	{
		delete this->iRelation;
		this->iRelation = NULL;
	}

	ASquareRootPhase::Reset();

	Set_InnerState((int)ESRPhaseNotStarted);
	return ConstRC::Ok;
}
int CSquareRootPhase::FillFunctorField()
{
	if (Get_FunctorsInit() == false)
	{
		Set_FunctorsInit(true);
		this->iPhaseFunctors.reserve(ConstSRPhase::INNER_STATE_COUNT);
		this->iPhaseFunctors.clear();
		this->iPhaseFunctors.push_back(NULL); // ESRPhaseNotStarted
		this->iPhaseFunctors.push_back(NULL); // ESRPhaseInitialization
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CSquareRootPhase>(this, &CSquareRootPhase::TryDependencies)); // ESRPhaseTryingDependencies
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CSquareRootPhase>(this, &CSquareRootPhase::SaveResult)); // ESRPhaseSavingResult
		this->iPhaseFunctors.push_back(NULL); // ESRPhaseFinished
	}

	return ConstRC::Ok;
}
int CSquareRootPhase::CleanUp(AFactorAlgParameters *aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	int phaseNBR;
	if (aParameters->GetIntegerPhaseParameter(ConstPN::SQUARE_ROOT_PHASE_NUMBER, phaseNBR, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","CleanUp",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}

	return ASquareRootPhase::CleanUp(aParameters, phaseNBR, aType, aError);
}
int CSquareRootPhase::InitParameters()
{
	Set_InnerState((int)ESRPhaseInitialization);

	AFactorAlgParameters * param = Get_Parameters();

	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	// generally we don't know order of the square root phase
	int auxInt;
	if (param->GetIntegerPhaseParameter(ConstPN::SQUARE_ROOT_PHASE_NUMBER, auxInt, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","InitParameters",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}
	Set_PhaseNbr(auxInt);

	// now we initialize members from parents's classes
	int rc = ASquareRootPhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	
	long maxBound = 0;
	// we try to find value for all 
	param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND, maxBound, maxBound, (int)EMain);
	this->iMaxFBBound = 0;
	string auxStr;
	// we go backwards because we want to find first index of algebraic part 
	// this means that the degree of the sieving poly is bigger than 1
	// we suppose that integral part should be first...
	for (int i = CNFSRelation::DEFAULT_NFS_PARTS_COUNT - 1; i >= 0; i--)
	{
		auxStr = Utils::ConvertToString(i + 1);
		rc = param->GetPolynomialPhaseParameter(ConstPN::SIEVING_POLY + auxStr, this->iSievingPoly[i], (int)EMain);
		if (rc != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CSquareRootPhase","InitParameters",0,0,"Didn't find the sieving polynomial.");
			return rc;
		}
		if (this->iSievingPoly[i]->GetDegree() > 1)
		{
			IntegralDomain::ComputeThetaPolynomial(this->iSievingPoly[i], this->iThetaPoly[i]);
			this->iStartOfAlg = i;
		}

		param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND + auxStr, maxBound, maxBound, (int)EMain);

		// find max
		this->iMaxFBBound = CIntegerClass::Max(this->iMaxFBBound, maxBound);

		// quadratic characters file - this file was created by other phase so we should get full name or 
		// we have to assemble it
		if (param->ContainsPhaseParameter(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStr, Get_PhaseNbr()))
		{
			param->GetStringPhaseParameter(ConstPN::QUAD_CHARS_FULL_FILENAME + auxStr, this->iQuadCharsFullFileName[i], Get_PhaseNbr());
			MSerializable::AdaptPathSeparator(this->iQuadCharsFullFileName[i]);
		}
		else
		{
			param->CreateFullFileName(this->iQuadCharsFullFileName[i], ConstPN::QUAD_CHARS_FILENAME + auxStr, 
				ConstPN::QUAD_CHARS_DIRECTORY + auxStr, ConstPN::QUAD_CHARS_COMPRESSION + auxStr, Get_SerialMainDirectory(),
				ConstFN::QUAD_CHARS + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
		}
	}

	if (this->iMaxFBBound == 0)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","InitParameters",0,0,"Didn't find the factor base upper bound.");
		return ConstRC::GeneralError;
	}

	int rootFinderType;
	param->GetIntegerPhaseParameter(ConstPN::ROOT_FINDER_TYPE, (int)ConstSRPhase::ROOT_FINDER, rootFinderType, Get_PhaseNbr());
	this->iRootFinderType = (root_finder_types)rootFinderType;

	// should we deserialize or serialize?
	bool deserialize = false;
	param->GetBoolPhaseParameter(ConstPN::DESERIALIZE, false, deserialize, Get_PhaseNbr());
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "0325000", "Square Root Phase is running ...");
    }
#endif

	if (deserialize)
	{
		rc = Deserialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	
	return ConstRC::Ok;
}
int CSquareRootPhase::SaveResult()
{
	Set_InnerState((int)ESRPhaseSavingResult);

	mpz_t f1;
	mpz_t f2;
	mpz_init(f1);
	mpz_init(f2);
	Get_Factor1(f1);
	Get_Factor2(f2);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50000
	if (Get_InfoMode())
	{
		CLog::Log(50000, "", "\n\n");
		CLog::Log(50000, "0325001", "Result of number field sieve factorization\n");
		CLog::Log(50000, "0325002", "==========================================\n");
		CLog::Log(50000, "0325003", "Factors %Zd = %Zd * %Zd.", this->iN, f1, f2);
		CLog::Log(50000, "", "\n\n");
    }
#endif

	mpz_clear(f1);
	mpz_clear(f2);

	Set_InnerState((int)ESRPhaseFinished);
	
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

	return Serialize(bparam);
}
int CSquareRootPhase::TryKthDependence()
{
	int depIndex = Get_DependencyIndex();

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "", "\n");
		CLog::Log(10000, "0325004", "    Trying dependency: %d\n", depIndex);
    }
#endif

#ifdef DEBUG_SQUARE_ROOT_PHASE_INFO_STATS_ON
	this->iStats->ResetAllIntermediate();
#endif
#ifdef DEBUG_SQUARE_ROOT_PHASE_TIME_MESSAGE_ON
	this->iTimes->ResetAllIntermediate();
#endif

	int rc;

	this->iRelation->Reset();
	this->iRelation->ZeroizeNumberOfRelations();

	CNFSRelation read(CNFSRelation::MODE_SQUARE_ROOT);
	long max =  this->iMatrixResult->GetMaximalColumnIndex() + 1;
	// We will read relations one-by-one from the relations result file
	FILE * relationFILE = fopen(this->iRelationsFullFileName.c_str(), FILE_MODE_READ);
	if (relationFILE == NULL)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","TryKthDependence",0,0,"Could not open file with smooth relations.");
		return ConstRC::CouldNotOpenFile;
	}

	long i;
	this->iAlgElementsCount = 0;
	long index = 0;
	while (read.ReadFromFile(relationFILE) == ConstRC::Ok && index < max)
	{
		if (this->iMatrixResult->IsOne(depIndex, index))
		{
			if (this->iAlgElementsCount >= this->iAlgElementsAllocated[this->iStartOfAlg])
			{
				for (i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
				{
					rc = ReallocatePolyField(this->iAlgElements[i], this->iAlgElementsAllocated[i]);
					if (rc != ConstRC::Ok)
					{
						fclose(relationFILE);
						relationFILE = NULL;
						return rc;
					}
				}
			}
			for (i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				read.Get_Parts(i)->Poly->Copy(this->iAlgElements[i][this->iAlgElementsCount]);
			}
			this->iAlgElementsCount++;

#ifdef DEBUG_SQUARE_ROOT_PHASE_COMBINE_INFO_ON
			#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
				if (Get_InfoMode())
				{
					CLog::Log(1000, "0325020", "Try %d-th dependence. Relations combined, iter %d.\n", depIndex, numAlgFact);
			   }
			#endif
			for (i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				if (this->iRelation->Get_Parts(i)->Poly == NULL) 
				{
					#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
						if (Get_InfoMode())
						{
							CLog::Log(1000, "0325021", "Target polynomial is null.\n");
					   }
					#endif
				} 
				else 
				{
					#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
						if (Get_InfoMode())
						{
							CLog::Log(1000, "0325022", "Target polynomial %p.\n", this->iRelation->Get_Parts(i)->Poly);
					   }
					#endif
					this->iRelation->Get_Parts(i)->Poly->PrintToScreen();
					#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
						if (Get_InfoMode())
						{
							CLog::Log(10000, "\n");
					   }
					#endif
				}
				if (read.Get_Parts(i)->Poly == NULL) 
				{
					#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
						if (Get_InfoMode())
						{
							CLog::Log(1000, "0325023", "Read polynomial is null.\n");
					   }
					#endif
				} 
				else 
				{
					#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
						if (Get_InfoMode())
						{
							CLog::Log(1000, "0325024", "Read polynomial %p.\n", read.Get_Parts(i)->Poly);
					   }
					#endif
					read.Get_Parts(i)->Poly->PrintToScreen();
					#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
						if (Get_InfoMode())
						{
							CLog::Log(10000, "\n");
					   }
					#endif
				}
			}
#endif

#ifdef DEBUG_SQUARE_ROOT_PHASE_INFO_STATS_ON
			if (read.Type & RELATION_FREE)
			{
				this->iStats->Increase(ESRSFree);
			}
			else if ((read.Type & CNFSRelation::P1_SMOOTH) && (read.Type & CNFSRelation::P2_SMOOTH))
			{
				this->iStats->Increase(ESRSSmooth);
			}
			else
			{
				this->iStats->Increase(ESRSPartial);
			}					
#endif

#ifdef DEBUG_SQUARE_ROOT_PHASE_ALG_REPRESENTATION_ON
			rc = this->iRelation->CombineWithRelation(&read, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iThetaPoly);
#else
			// combine only divisors - not algebraic representation
			rc = this->iRelation->CombineWithRelation(&read);
#endif
			if (rc != ConstRC::Ok)
			{
				fclose(relationFILE);
				relationFILE = NULL;
				return rc;
			}
		}
		index++;
	}

	fclose(relationFILE);
	relationFILE = NULL;

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "0325005", "      - altogether %ld relations were used\n", this->iAlgElementsCount);
		CLog::Log(10000, "0325006", "      - independent relations were %ld\n", this->iRelation->Get_NumberOfRelations());
    }
#endif

	rc = this->iRelation->ControlExponents();
	if (rc != ConstRC::Ok)
	{
		this->iRelation->PrintToScreen();
		return ConstRC::NotFactorized;
	}

	long j;
#ifdef DEBUG_SQUARE_ROOT_PHASE_INFO_STATS_ON
	
	CRelationPart * part = NULL;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		part = this->iRelation->Get_Parts(i);
		for (j = part->Main.MaxElementIndex; j > -1; j--)
		{
			if (part->Main.Elements[j].flags & ConstEF::SPECIAL_PRIME)
			{
				this->iStats->Increase(ESRSSpecialPrime);
			}
			else
			{
				this->iStats->Increase(ESRSNormalPrime);
			}
		}
	}


#endif

	// We need to multiply also by square of derivation of f(x)
	if (this->iAlgElementsCount + 2 >= this->iAlgElementsAllocated[this->iStartOfAlg])
	{
		for (i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			rc = ReallocatePolyField(this->iAlgElements[i], this->iAlgElementsAllocated[i]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}

	Polynomial * derivates = new Polynomial[CNFSRelation::DEFAULT_NFS_PARTS_COUNT]();
	for (i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iThetaPoly[i]->Derivate(&derivates[i], this->iThetaPoly[i]);
		derivates[i].Copy(this->iAlgElements[i][this->iAlgElementsCount]);
		derivates[i].Copy(this->iAlgElements[i][this->iAlgElementsCount + 1]);
	}
	this->iAlgElementsCount += 2;

#ifdef DEBUG_SQUARE_ROOT_PHASE_CONTROL_WITH_CHARACTERS_ON

	prime_ideal_for_legendre * quadChars = NULL;
	long numQuadChars = 0;
	CBaseParameters quadCharsParam;
	for (i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		quadCharsParam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iQuadCharsFullFileName[i]);
		rc = ARelProcessingPhase::DeserializeQuadCharacters(quadCharsParam, quadChars, numQuadChars);
		if (rc != ConstRC::Ok)
		{
			if (quadChars != NULL)
			{
				delete quadChars;
			}
			delete[] derivates;
			derivates = NULL;
			return rc;
		}

		rc = CheckAlgebraicSquare(*(this->iAlgElements[i]), this->iAlgElementsCount, quadChars, numQuadChars, this->iThetaPoly[i]);
		delete quadChars;
		quadChars = NULL;
		if (rc != ConstRC::Ok)
		{
			delete[] derivates;
			derivates = NULL;
			CLog::LogError(WHEREARG,"CSquareRootPhase","TryKthDependence",0,0,"WARNING: Algebraic part is not a Square! Error in LA phase or in Legendre symbols!");
			return ConstRC::NotFactorized;
		}
	}

#endif

	mpz_t * phis = new mpz_t[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	mpz_t * roots = new mpz_t[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		mpz_init(phis[i]);
		mpz_mul(phis[i], this->iM, *this->iSievingPoly[i]->Leading2());
		mpz_init(roots[i]);
	}

	mpz_t auxMpz;
	mpz_init(auxMpz);
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SQUARE_ROOT_PHASE_TIME_MESSAGE_ON
		// ESRTFirstRoot = 0, ESRTSecondRoot, ESRTThirdRoot, ESRTFourthRoot = 3 !!!
		this->iTimes->Start(i);
#endif
		if (this->iSievingPoly[i] != NULL && this->iSievingPoly[i]->GetDegree() > 1)
		{
			// algebraic root
			// phi : \theta -> c_{1, d}m
			rc = AlgebraicRoot(this->iRelation->Get_Parts(i), this->iThetaPoly[i], *this->iSievingPoly[i]->Leading2(),
				this->iRelation->Get_NumberOfRelations(), this->iRootFinderType, this->iMaxFBBound, this->iAlgElements[i], 
				this->iAlgElementsCount, phis[i], roots[i]);
			if (rc != ConstRC::Ok)
			{
				break;
			}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 1000
	if (Get_InfoMode())
	{
		CLog::Log(1000, "\n");
    }
#endif

		}
		else
		{
			// integral root
			rc = IntegralRoot(this->iRelation->Get_Parts(i), roots[i]);
			if (rc != ConstRC::Ok)
			{
				break;
			}
		}

		// c_{2, d}^{|M|/2} * \bar{f}_{2}'(c_{2, d}*m) * alg_root_1 mod N
		// if we have integral part - we don't want to use it
		for (j = this->iStartOfAlg; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++)
		{
			if (j != i)
			{
				derivates[j].Evaluate(auxMpz, phis[j]);
				mpz_mod(auxMpz, auxMpz, this->iN);
				mpz_mul(roots[i], roots[i], auxMpz);
				mpz_mod(roots[i], roots[i], this->iN);
				if (mpz_cmp_ui(*this->iSievingPoly[j]->Leading2(), 1) != 0)
				{
					// nonmonic sieving poly 2 - we have to multiply by c_{2, d}^{|M|/2} (coeff from poly 2)
					mpz_powm_ui(auxMpz, *this->iSievingPoly[j]->Leading2(), this->iRelation->Get_NumberOfRelations()/2, this->iN);
					mpz_mul(roots[i], roots[i], auxMpz);
					mpz_mod(roots[i], roots[i], this->iN);
				}
			}
		}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "", "\n");
		CLog::Log(10000, "0325007", "    The root #%d is %Zd\n", i + 1, roots[i]);
    }
#endif

#ifdef DEBUG_SQUARE_ROOT_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(i);
#endif
	}

	delete[] derivates;
	derivates = NULL;
	mpz_clear(auxMpz);

	if (rc != ConstRC::Ok)
	{
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			mpz_clear(phis[i]);
			mpz_clear(roots[i]);
		}
		delete [] phis;
		delete [] roots;
		return ConstRC::NotFactorized;
	}

	for (i = CNFSRelation::DEFAULT_NFS_PARTS_COUNT - 2; i > -1; i--)
	{
		rc = CheckResultPair(roots[i + 1], roots[i]);
		if (rc == ConstRC::Ok)
		{
			break;
		}
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		mpz_clear(phis[i]);
		mpz_clear(roots[i]);
	}

	delete [] phis;
	delete [] roots;

	if (Get_InfoMode())
	{
		PrintDetails();
	}
	
	return rc;
}
int CSquareRootPhase::TryDependencies()
{
	Set_InnerState(ESRPhaseTryingDependencies);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "0325008", "Trying dependencies ...");
    }
#endif

	// now read matrix result
	// first - prepare parameters for deserialization of result matrix
	CBaseParameters bparam;
	bparam.AddParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iMatrixResultFullFileName);

	this->iMatrixResult = CMatrixHelper::CreateMatrix(this->iMatrixResultType);
	if (this->iMatrixResult == NULL)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","RunPhase",0,0,"Cannot create matrix results.");
		return ConstRC::GeneralError;
	}

	int rc = this->iMatrixResult->Deserialize(bparam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = ConstRC::NotFactorized;
	int max = this->iMatrixResult->GetMaximalRowIndex() + 1;
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "0325009", "    Nonzero dependencies: ", max);
    }
#endif	
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

	this->iAlgElementsCount = 0;
	long alloc = this->iMatrixResult->GetMaximalColumnIndex() / 2;

	for (int i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iAlgElementsAllocated[i] = alloc;
		rc = AllocatePolyField(this->iAlgElements[i], this->iAlgElementsAllocated[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	this->iRelation = new CNFSRelation(CNFSRelation::MODE_SQUARE_ROOT);

	for (int k = Get_DependencyIndex(); k < max; k++)
	{
		Set_DependencyIndex(k);
#ifdef DEBUG_SQUARE_ROOT_PHASE_INFO_STATS_ON
		this->iStats->ConfirmAll();
#endif
#ifdef DEBUG_SQUARE_ROOT_PHASE_TIME_MESSAGE_ON
		this->iTimes->ConfirmAll();
#endif
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			break;
		}
		rc = TryKthDependence();
		if (rc != ConstRC::NotFactorized)
		{
			// case of finding a factor, or of a memory error etc.
			break;
		}
	}

	//for (int i = this->iStartOfAlg; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	//{
	//	DeletePolyField(this->iAlgElements[i], this->iAlgElementsAllocated[i]);
	//}
	
	if (this->iMatrixResult != NULL)
	{
		delete this->iMatrixResult;
		this->iMatrixResult = NULL;
	}


	return rc;
}

void CSquareRootPhase::PrintDetails()
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 99
	if (Get_InfoMode())
	{
		CLog::Log(99, "\n");
		CLog::Log(99, "0325010", "Details:\n");

		CLog::Log(99, "0325011", "  Used relations:\n");
		CLog::Log(99, "0325012", "    Smooth relation: %d\n", this->iStats->Get_Intermediate(ESRSSmooth));
		CLog::Log(99, "0325013", "    Partial relation: %d\n", this->iStats->Get_Intermediate(ESRSPartial));
		CLog::Log(99, "0325014", "    Free relation: %d\n", this->iStats->Get_Intermediate(ESRSFree));

		CLog::Log(99, "0325015", "  PrimeIdeals:\n");
		CLog::Log(99, "0325016", "    Normal: %d\n", this->iStats->Get_Intermediate(ESRSNormalPrime));
		CLog::Log(99, "0325017", "    Special: %d\n", this->iStats->Get_Intermediate(ESRSSpecialPrime));

		CLog::Log(99, "0325018", "  Time:\n");

		for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			CLog::Log(99, "0325019", "    The root #%d: %12.4f sec\n", i + 1, this->iTimes->Get_IntermediateTime(ESRTFirstRoot + i));
		}
		CLog::Log(99, "\n\n");
    }
#endif

}
int CSquareRootPhase::RunPhase(AFactorAlgParameters * aParameters)
{
	return APhase::RunPhase(aParameters);
}
int CSquareRootPhase::Serialize(const CBaseParameters &aParam) const
{
	int rc = MSerializable::Serialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (Running())
	{
		AFactorAlgParameters * param = Get_Parameters();

		// for deserialization at new start
		param->SetPhaseParameter(ConstPN::DESERIALIZE, Utils::ConvertToString(true), Get_PhaseNbr());
		param->SetPhaseParameter(ConstPN::PHASE_FULL_FILENAME, 
			MSerializable::CreateFileNameAuto(aParam, ConstFN::SQUARE_ROOT_PHASE), Get_PhaseNbr());

		CBaseParameters paramb;
		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_ParametersFullFileName()); 
		rc = param->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int CSquareRootPhase::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_SQUARE_ROOT_PHASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CSquareRootPhase","Serialize",rc,0,"Could not write the parameter tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int state = Get_InnerState();
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_INNER_STATE_NAME, state);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (state >= ESRPhaseTryingDependencies)
	{
		rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_DEPENDENCY_INDEX_NAME, Get_DependencyIndex());
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (state >= ESRPhaseSavingResult)
		{
			mpz_t auxMpz;
			mpz_init(auxMpz);

			Get_Factor1(auxMpz);
			rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_FACTOR_1_NAME, auxMpz);
			if (rc != ConstRC::Ok)
			{
				mpz_clear(auxMpz);
				return rc;
			}

			Get_Factor2(auxMpz);
			rc = CXMLHelper::SerializeMpz(aWriter, XML_SERIAL_FACTOR_2_NAME, auxMpz);
			if (rc != ConstRC::Ok)
			{
				mpz_clear(auxMpz);
				return rc;
			}
			mpz_clear(auxMpz);
		}
	}

	CBaseParameters param;
	rc = this->iStats->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = this->iTimes->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","Serialize",rc,0,"Could not end the parameter tag.");
        return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter,FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok; 
}
int CSquareRootPhase::Deserialize(const CBaseParameters &aParam)
{
	return MSerializable::Deserialize(aParam);
}
int CSquareRootPhase::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != ESRPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_SQUARE_ROOT_PHASE_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CSquareRootPhase","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	CBaseParameters param;
	int auxInt = 0;
	mpz_t auxMpz;
	mpz_init(auxMpz);
	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INNER_STATE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize inner state...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_InnerState(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_DEPENDENCY_INDEX_NAME))
		{

			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize dependency index...", auxInt, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_DependencyIndex(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_FACTOR_1_NAME))
		{

			rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize factor 1...", auxMpz, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_Factor1(auxMpz);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_FACTOR_2_NAME))
		{

			rc = CXMLHelper::DeserializeMpz(aReader, "Deserialize factor 2...", auxMpz, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_Factor2(auxMpz);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_STATISTIC_CONTAINER_NAME))
		{
			rc = this->iStats->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_TIME_CONTAINER_NAME))
		{
			rc = this->iTimes->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}
	}

	mpz_clear(auxMpz);

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
