#include "statistic_container.h"

#include "base_parameters.h"
#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_filenames.h"
#include "log.h"


using namespace std;

CStatContainer::CStatContainer(int aMaxSize) : MSerializable(ConstFN::STATISTIC_CONTAINER),
	iTotals(aMaxSize, 0), iIntermediates(aMaxSize, 0), iConfirmTotals(aMaxSize, 0), iMutexes(aMaxSize)
{
	this->iMaxSize = aMaxSize;
	int rc;
	for (int i = 0; i < aMaxSize; i++)
	{
		if ((rc = pthread_mutex_init(&(this->iMutexes[i]), NULL)) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CStatContainer","CStatContainer",rc,0,"Cannot initialize mutex.");
		}
	}
}
CStatContainer::~CStatContainer()
{
	int rc;
	int max = this->iMutexes.size();
	for (int i = 0; i < max; i++)
	{
		if ((rc = pthread_mutex_destroy(&(this->iMutexes[i]))) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CStatContainer","~CStatContainer",rc,0,"Cannot destroy mutex.");
		}
	}
}
int CStatContainer::Increase(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iTotals[aIndex]++;
		this->iIntermediates[aIndex]++;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::Increase(int aIndex, int aValue)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iTotals[aIndex] += aValue;
		this->iIntermediates[aIndex] += aValue;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::Decrease(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iTotals[aIndex]--;
		this->iIntermediates[aIndex]--;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::Decrease(int aIndex, int aValue)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iTotals[aIndex] -= aValue;
		this->iIntermediates[aIndex] -= aValue;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::Reset(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iIntermediates[aIndex] = 0;
		this->iTotals[aIndex] = 0;
		this->iConfirmTotals[aIndex] = 0;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::ResetAll()
{
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iIntermediates[i] = 0;
		this->iTotals[i] = 0;
		this->iConfirmTotals[i] = 0;
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CStatContainer::Confirm(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iConfirmTotals[aIndex] = this->iTotals[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::ConfirmAll()
{
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iConfirmTotals[i] = this->iTotals[i];
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CStatContainer::ResetIntermediate(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iIntermediates[aIndex] = 0;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CStatContainer::ResetAllIntermediate()
{
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iIntermediates[i] = 0;
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CStatContainer::Get_ConfirmTotal(int aIndex) const
{
	if (Control(aIndex))
	{
		int ret;
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		ret = this->iConfirmTotals[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ret;
	}

	return 0;
}
int CStatContainer::Get_Total(int aIndex) const
{
	if (Control(aIndex))
	{
		int ret;
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		ret = this->iTotals[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ret;
	}

	return 0;
}
int CStatContainer::Get_Intermediate(int aIndex) const
{
	if (Control(aIndex))
	{
		int ret;
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		ret = this->iIntermediates[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ret;
	}

	return 0;
}
int CStatContainer::Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_STATISTIC_CONTAINER_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CStatContainer","Serialize",rc,0,"Could not write the main tag.");
		return ConstRC::GeneralError;
    }
	
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_MAX_BOUND_NAME, this->iMaxSize);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_CONFIRM_TOTALS_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CStatContainer","Serialize",rc,0,"Could not write the tag.");
		return ConstRC::GeneralError;
    }
	
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int value;
	for (int i = 0; i < this->iMaxSize; i++)
	{
		value = Get_ConfirmTotal(i);
		rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_ELEMENT_NAME, value);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_CONFIRM_TOTALS_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CStatContainer","Serialize",rc,0,"Could not end the tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CStatContainer","Serialize",rc,0,"Could not end the main tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	
	return rc;
}
int CStatContainer::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_STATISTIC_CONTAINER_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CStatContainer","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CStatContainer","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	// delete all data
	ResetAll();

	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_MAX_BOUND_NAME))
		{
			int max;
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize max size...", max, FALSE);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}

			if (this->iMaxSize < max)
			{
				// add new elements
				while (this->iTotals.size() < max)
				{
					this->iConfirmTotals.push_back(0);
					this->iTotals.push_back(0);
					this->iIntermediates.push_back(0);
					pthread_mutex_t mutex;
					this->iMutexes.push_back(mutex);
					if ((rc = pthread_mutex_init(&(mutex), NULL)) != ConstRC::Ok)
					{
						CLog::LogError(WHEREARG,"CStatContainer","Deserialize",rc,0,"Cannot initialize mutex.");
					}
				}
				this->iMaxSize = max;
			}

			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_CONFIRM_TOTALS_NAME))
		{
			int elementsDepth = xmlTextReaderDepth(aReader);
			int value;
			for (int i = 0; i < this->iMaxSize; i++)
			{
				if (CXMLHelper::MoveToNextNode(aReader, elementsDepth) != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize confirm total...", value, FALSE);
				if (rc != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				pthread_mutex_lock(&this->iMutexes[i]);
				this->iConfirmTotals[i] = value;
				this->iTotals[i] = value;
				pthread_mutex_unlock(&this->iMutexes[i]);
			}
			continue;
		}

		//if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INTERMEDIATES_NAME))
		//{
		//	int elementsDepth = xmlTextReaderDepth(aReader);
		//	int value;
		//	for (int i = 0; i < this->iMaxSize; i++)
		//	{
		//		if (CXMLHelper::MoveToNextNode(aReader, elementsDepth) != ConstRC::Ok)
		//		{
		//			rc = ConstRC::GeneralError;
		//			break;
		//		}
		//		rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize intermediates...", value, FALSE);
		//		if (rc != ConstRC::Ok)
		//		{
		//			rc = ConstRC::GeneralError;
		//			break;
		//		}
		//		pthread_mutex_lock(&this->iMutexes[i]);
		//		this->iIntermediates[i] = value;
		//		pthread_mutex_unlock(&this->iMutexes[i]);
		//	}
		//	continue;
		//}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
