#ifndef __STRUCTURES_H__
#define __STRUCTURES_H__

#include "types.h"
#include "nfs_definitions.h"

/*! 
 * \brief Divisor in a relation
 *
	This short structure is intended to host a divisor in a relation. Each relation has an array of such divisors. One can easily see
	that only two variables are used: column index, which means the index of the prime number in the factor base, and exponent, which
	means to which power this prime divides the relation.

	Flags: B1.B2.B3.B4

	Last 4 bits in B1: flag for special prime, divides M, divides leading, divides c_p
	-- other bits only for special primes or divide L:
	First 4 bits in B1, all in B2 and all in B3: prime-valuation of the F(a,b)
	Last 4 bits in B4: prime-valuation of the leading coefficient
	First 4 bits in B4: inertial degree of a special prime ideal or an ideal (p)

 */
typedef struct {
	main_sieving_type prime; // May include -1
	unsigned int      exponent;
	main_sieving_type root;
	unsigned int      flags;		
} nfs_sieving_element;

/*!
 * \brief New general factor base element
 *
	This structure is for elements in new general factor base.
	It combines two old structures: integer_for_sieving and prime_ideal_for_sieving
		- they had different names for same things...
		- this was big problem for sieving phase where we want one class for (int, alg) and (alg, alg) sieving.
*/
typedef struct {
	main_sieving_type prime;
	main_sieving_type root;
	main_sieving_type inversion_of_root;
	unsigned int      flags;
	log_type	      log;	
} fb_element;


/*!
 * \brief Prime ideal for quadratic characters
 */
typedef struct {
	main_sieving_type prime;	
	main_sieving_type c_p;
} prime_ideal_for_legendre;

/*!
 * \brief First hashtable entry
 *
	This structure is the base type for the "first" hashtable used in both large prime variations. Intent of this hashtable is to give exact information
	about number of collected fundamental cycles throughout the runtime. As this hashtable needs to be large (to hold many large primes), its 
	elements must be as small as possible.

*/
typedef struct {
	large_prime_type value;	
	large_prime_type root_fingerprint;
	unsigned int ancestor;
} nfs_hashtable_entry_type_1;

/*!
 * \brief Seconf hashtable entry
 *
	This structure is the base type for the "second" hashtable used in both large prime variations during the cycle construction phase. As this
	hashtable needs only to contain information about large primes taking part in at least one of the fundamental cycles, its elements can be
	larger; however, only one extra field is needed. 
	The most significant bit of the relation field is used to denote the iteration parity (see Lenstra, Manasse: Factoring with Two Large Primes).

*/
typedef struct {
	large_prime_type value;	
	large_prime_type root_fingerprint;
	unsigned int ancestor;
	unsigned int relation;
} nfs_hashtable_entry_type_2;

/*! 
 * \brief Update entry for bucket sieving
 *
   Type taken and slightly redefined from Chris Papadopoulos' line siever.

   As the size of the sieving block can well exceed 64K, the 'h' value must be 32-bit.	

*/
typedef struct {
  main_sieving_type		p;		/* prime */
  main_sieving_type		h;		/* offset within block */
  main_sieving_type 	r;		/* root of the poly mod p */
  log_type 				logp;	/* log of prime */
} update_t;

/*!
 * \brief type for cache optimalization
 */
typedef struct {
	update_t	updates[STRUCTURE_CACHE_UPDATES];
	int		used;	
} cache_t;

/*!
 * \brief Linked list of lists of updates for bucket sieving
 *
	Type taken from Chris Papadopoulos' line siever.

	This is a type suitable for construction of a linked list of lists, where each member
	list is of reasonable size (to fit into cache).

	Since the size of update_t has grown, STRUCTURE_MAX_UPDATES should be lower than original 1000.
	Currently, we use 500 (see nfs_definitions.h)
*/
typedef struct script_t {
	struct script_t *next;
	unsigned int	 num_used;
	//double align_me;
	update_t list[STRUCTURE_MAX_UPDATES];
} script_t;

/*! 
 * \brief Outer structure for updates for bucket sieving
 *
	The blocks are indexed via a hashtable. Each block
	also maintains a cache of update structures that are
	packed together to conserve TLB resources 
*/
typedef struct {
	script_t *script;	/* linked list for this block */
	//int cache_off;    /* offset into the cache array where block cache starts */
	//int cache_used;   /* number of entries currently used in block's cache */
} hashtable_t;

/*!
 * \brief Typedef for the new approach to line sieving
 */
typedef struct {
	main_sieving_type   p; 	    ///<The prime value
	main_sieving_type   r;		///<Recomputed c_p or m mod p for integers - root of poly mod p
	main_sieving_type   root_1; ///<The offset s.t. Q(x) = 0. Note that root_1 != x.
	main_sieving_type   next_1;
	main_sieving_type   auxVal; ///<Necessary value for shift X if special prime: p valuation of leading coeff
	log_type	    log_p;
	log_type	    flags; ///<Some usefull flags (special prime, divide L prime ... and valuation of L)
} nfs_fb_type;

/*!
 * \brief General sieving region
 *
	This structure defines general sieving region. It is used both for classical and lattice sieving.
	It is a "stripe" in full sieving region (intended for job distribution). Width of the region is
	sieving interval length.
*/
typedef struct {
	long width;  // sieving interval
	long height;
	long start;  // starting value in height dimension
} sieving_region;

// FOR LATTICE SIEVING

/*!
 * \brief Lattice sieving advanced factor base
 *
	Typedef for the lattice sieving.
	Some values are according to paper by Franke Jens, Kleinjung Thorsten - Continued Fraction and Lattice Sieving.
*/
typedef struct 
{
	main_sieving_type   p; 	    ///<The prime value
	/*!
		First shift value (if p < I then special value for shift on next row).
		If it is negative than this element cannot be used for sieving.
	*/
	main_sieving_type	a;
	main_sieving_type	z;		///<Start value for block
	///First bound (if p < I then I - p) -- 16bit
	unsigned short		b0;
	///Second bound (if p < I then I - a) -- 16bit -- for real value of b1 we must add 1 to first half
	unsigned short		b1;
	main_sieving_type	c;		///<Second shift value (if p < I then just p)
	main_sieving_type	r;		///<Recomputed c_p or m mod p for integers
	main_sieving_type	cr;		///<New r for sublattice of main lattice
	///Some usefull flags (special prime, divide L prime ... and valuation of L)
	log_type			flags;
	///Log of prime
	log_type			logp;
} lattice_type;

/*!
 * \brief Lean factor base
 */
typedef struct 
{
	main_sieving_type   p; 	    ///<The prime value
	main_sieving_type	r;		///<Recomputed c_p or m mod p for integers
	main_sieving_type	cr;		///<New r for sublattice of main lattice
	///Some usefull flags (special prime, divide L prime ... and valuation of L)
	log_type			flags;
	///Log of prime
	log_type			logp;
} lattice_type2;

// similar hashtable structure as for the new approach to line sieving

/*!
 * \brief Advanced working factor base  for lattice sieving
 */
typedef struct 
{
	main_sieving_type	h;			/* offset within block */
	main_sieving_type	start;		/* offset within block - for factorization sieving */
	///Some usefull flags (special prime, divide L prime ... and valuation of L)
	log_type			flags;
	///Log of prime
	log_type			logp;
	///First bound (if p < I then I - p) -- 16bit
	unsigned short		b0;
	///Second bound (if p < I then I - a) -- 16bit -- for real value of b1 we must add 1 to first half
	unsigned short		b1;
	main_sieving_type	a;			/* shift a */
	main_sieving_type	c;			/* shift c */
	lattice_type2 *		element;	/* more info about used prime ideal or prime */
} lattice_update_t;

/*!
 * \brief Bucket sieving update type
 */

typedef struct 
{
	main_sieving_type	h;			/* offset within block */
	///Log of prime
	log_type			logp;
	lattice_type2 *		element;	/* more info about used prime ideal or prime */
} lattice_update2_t;

/*!
 * \brief Linked list of list for lattice sieving
 */
typedef struct lattice_script_t
{
	struct lattice_script_t *next;
	unsigned int			num_used;
	lattice_update2_t		list[STRUCTURE_MAX_UPDATES];
} lattice_script_t;

/*!
 * \brief Outer structure for lattice bucket sieving
 */
typedef struct 
{
	lattice_script_t *script;	/* linked list for this block */
	lattice_script_t *last_free_script;
} lattice_hashtable_t;

/*!
 * \brief Base lattice information
 */
typedef struct 
{
	main_sieving_type	a0; // the first coeff of the first base vector
	main_sieving_type	b0; // the second coeff of the first base vector
	main_sieving_type	a1; // the first coeff of the second base vector
	main_sieving_type	b1; // the second coeff of the second base vector
	long				I;  // sieving interval - [-I/2, I/2]
	long				J;  // [1, J] - J < I/2
} main_lattice_info;

/*!
 * \brief Used for counting prime ideal frequencies.
*/
typedef struct {
	main_sieving_type	prime;	
	main_sieving_type	root_fingerprint;
	unsigned int		frequency;
	relation_index_type	index; // if the prime ideal is singleton, then there is index of the single relation
} fhashtable_entry;

/*!
 * \brief Used in Filtering phase  for discarding prime ideals in clique algorithm
 *
	When we have more relations than prime ideals we want to discard few of them.
	This comes from Clique algorithm.
*/
typedef struct {
	/*
		ID of main component - it is index in the field with all components
		If index of this component is different from ID then this is not main component
	*/
	int ID; 
	/// Weight of the component - calculated from relations inside component
	INT64 weight;
	/// count of relations in this component
	int count;

} clique_component_t;

/*!
 * \brief Simple type for saving prime ideal. We can also distinguish which part of relation this prime ideal belongs to.
*/
typedef struct {
	main_sieving_type prime;
	/// c_p AND mask (flag of relation part)
	main_sieving_type rootmask;
} prime_ideal_t;

/*!
 * \brief  Comparing prime ideals
 */
struct prime_ideal_comp {
  bool operator() (const prime_ideal_t& lhs, const prime_ideal_t& rhs) const
  {
	  if (lhs.prime == rhs.prime) return lhs.rootmask < rhs.rootmask;
	  return lhs.prime < rhs.prime;
  }
};

/*!
 * \brief Simple type for saving edge of a graph.
*/
typedef struct {
	int vertex1;
	int vertex2;
} graph_edge_t;

/*!
*\brief 	Simple type for saving edge of a graph with weight.
*/
typedef struct {
	int vertex1;
	int vertex2;
	int weight;
} graph_wedge_t;

/*!
 * \brief Counting prime ideal frequencies in filtering phase
 *
	This structure is used for counting prime ideal frequencies in filtering phase for merge part.
	So we need to know weight of this set of relations (relations which share same prime ideal).
*/
typedef struct {
	main_sieving_type	prime;	
	main_sieving_type	root_fingerprint;
	unsigned int		frequency;
	/// Weight of relations which share given prime ideal.
	int	weight;
} merge_fhashtable_entry;


/*!
 * \brief Searching useful values in Kleinjung algorithm
 *
	This structure is used for searching of values mod Z in [-e, e] - finding suitable polynomial in Kleinjung algorithm.	
*/
typedef struct {
	double	value; // calculated value
	int		index; // ll-tuples (u_{1}, ..., u_{ll}) saved as d-based value
} e_search_element;

#endif // __STRUCTURES_H__
