#include "test_suite.h"

#include "nfs_utils.h"
#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_element_flags.h"
#include "nfs_relation.h"
#include "const_sieving_phase.h"
#include "lp_hashtable_type1.h"
#include "const_hashtables.h"
#include "sparse_matrix_class.h"
#include "normal_matrix_class.h"
#include "integer_matrix_class.h"
#include "integer_class.h"
#include "nfs_classical_sieving_phase.h"
#include "relation_writer.h"
#include "relation_reader.h"
#include "sparse_matrix_class.h"
#include "nfs_parameters.h"
#include "candidate_pairs.h"
#include "log.h"

using namespace std;

int CTestSuite::ControlSmoothNFSRelationsCount(string aSmoothRelFullFileName, long aSmooths)
{
	if (aSmooths == DEBUG_NO_TEST)
	{
		return ConstRC::Ok;
	}

	Utils::NormalMessage("\t\tControlling number of smooth relations ... ");
	
	CNFSRelation relation(EDoubleLargePrimeV);
	FILE * smoothsFILE = fopen(aSmoothRelFullFileName.c_str(), FILE_MODE_READ);
	if (smoothsFILE == NULL) 
	{
		CLog::LogError(WHEREARG,"CTestSuite","ControlSmoothNFSRelationsCount",0,0,"Could not open file with smooth relations.");
		return ConstRC::CouldNotOpenFile;
	}

	long count = 0;
	while (relation.ReadFromFile(smoothsFILE) == ConstRC::Ok)
	{
		count++;
	}

	fclose(smoothsFILE);
	smoothsFILE = NULL;

	if (aSmooths == count)
	{
		printf("\n\t\tSmooth relations count OK. Count: %ld\n\n", aSmooths);
	}
	else
	{
		printf("\n\t\tSmooth relations count WRONG. SaveCount: %ld =!= RealCount: %ld\n\n", aSmooths, count);
	}

	return ConstRC::Ok;
}
int CTestSuite::ControlPartialNFSRelationsCount(string aPartialRelFullFileName, string aIntFBFullFileName, string aAlgFBFullFileName, long aPartials, long aVertices, long aEdges, long aComponents)
{
	Utils::NormalMessage("\t\tControlling number of partial relations ... ");

	// all new
	long edges = 0;
	long vertices = 0;
	long components = 0;

	long integralFBMaxIndex = -1;
	long integralFBMaxAllocIndex = 0;
	long algebraicFBMaxIndex = -1;
	long algebraicFBMaxAllocIndex = 0;

	CFBInfo info;
	CBaseParameters paramFB;
	paramFB.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, aIntFBFullFileName);
	int rc = info.DeserializeFBInfo(paramFB);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	integralFBMaxIndex = info.FBMaxIndex;
	integralFBMaxAllocIndex = info.FBMaxAllocatedIndex;

	paramFB.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, aAlgFBFullFileName);
	rc = info.DeserializeFBInfo(paramFB);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	algebraicFBMaxIndex = info.FBMaxIndex;
	algebraicFBMaxAllocIndex = info.FBMaxAllocatedIndex;

	// At first, we find out whichever of the actual factor bases is larger
	int fbsize = max(integralFBMaxIndex, algebraicFBMaxIndex);

	// The main hashtable for detection of cycles. Each entry corresponds to a large prime found in a x-partial.
	CLPType1Hashtable lpTable;
	rc = lpTable.SetupHashtableFB(fbsize);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// Special case: PRIME=1, root of its own component.
	vertices++;
	components++;

	vector<long> intermediateRoots;
	intermediateRoots.reserve(ConstSPhase::INTERMEDIATE_ROOTS_COUNT);

	CNFSRelation relation(EDoubleLargePrimeV);

	FILE * partialsFILE = fopen(aPartialRelFullFileName.c_str(), FILE_MODE_READ);
	if (partialsFILE == NULL)
	{
		CLog::LogError(WHEREARG,"CTestSuite","ControlPartialNFSRelationsCount",0,0,"Could not open file with partial relations.");
		return ConstRC::CouldNotOpenFile;
	}

	int oldCycles = 0;
	long d_1  = 0;
	long d_2  = 0;
	int r_1  = 0;
	int r_2  = 0;
	bool present = false;
	int  searched;
	int  root_fingerprint;
	int index = 0;
	while (relation.ReadFromFile(partialsFILE) == ConstRC::Ok)
	{
		if (!(index % 10))
		{
			printf("\r\t\t\tAdding %8d. relation...", index);
			fflush(stdout);
		}
		index++;

		oldCycles = edges + components - vertices;
		bool enlarged = true;
		while (enlarged) {
			if (relation.Type & ARelation::P1_SMOOTH) 
			{
				searched = 1;
				root_fingerprint	= 0;
			} 
			else 
			{
				searched = relation.Get_Parts(0)->Large.Elements[0].prime;
				root_fingerprint = ConstSPhase::ELEMENT_MASK[0];
			}
			rc = lpTable.FindPrimeInHashtable(searched, root_fingerprint, present, d_1, enlarged);
			if (rc != ConstRC::Ok)
			{
				fclose(partialsFILE);
				partialsFILE = NULL;
				return rc;
			}
			if (!present) 
			{
				// This prime has not been in hashtable yet.					
				lpTable.Hashtable[d_1].SetValue(searched);
				lpTable.Hashtable[d_1].SetRootFingerprint(root_fingerprint);
				lpTable.Hashtable[d_1].SetAncestor(d_1);
				lpTable.Added();
				vertices++;
				components++;
			} 

			if (relation.Type & ARelation::P2_SMOOTH) 
			{
				searched = 1;
				root_fingerprint = 0;
			} 
			else 
			{
				searched = relation.Get_Parts(1)->Large.Elements[0].prime;
				root_fingerprint = relation.Get_Parts(1)->Large.Elements[0].root;
				root_fingerprint |= ConstSPhase::ELEMENT_MASK[1];
			}
			rc = lpTable.FindPrimeInHashtable(searched, root_fingerprint, present, d_2, enlarged);
			if (rc != ConstRC::Ok)
			{
				fclose(partialsFILE);
				partialsFILE = NULL;
				return rc;
			}
			if (!present) 
			{
				//This prime has not been in hashtable yet.					
				lpTable.Hashtable[d_2].SetValue(searched);
				lpTable.Hashtable[d_2].SetRootFingerprint(root_fingerprint);
				lpTable.Hashtable[d_2].SetAncestor(d_2);
				lpTable.Added();
				vertices++;
				components++;
			} 
		}
			
		edges++;

		// Finding roots of components
		r_1 = lpTable.FindAndResetRoot(d_1, intermediateRoots);
		r_2 = lpTable.FindAndResetRoot(d_2, intermediateRoots);
		if (r_1 != r_2) 
		{			
			// The component with the larger prime at root is joined to the other one
			if (lpTable.Hashtable[r_1].GetValue() > lpTable.Hashtable[r_2].GetValue()) 
			{
				lpTable.Hashtable[r_1].SetAncestor(r_2);	
			} 
			else 
			{
				lpTable.Hashtable[r_2].SetAncestor(r_1);	
			}
			components--;
		}

		if (edges + components - vertices != oldCycles) 
		{
			relation.Type |= RELATION_FINISHING_CYCLE;
		}
	}

	fclose(partialsFILE);
	partialsFILE = NULL;

	printf("\n");

	if (aPartials != DEBUG_NO_TEST)
	{
		if (index != aPartials)
		{
			printf("\n\t\t\tPartial relations count WRONG. SaveCount: %ld =!= RealCount: %ld\n", aPartials, index);
		}
		else
		{
			printf("\n\t\t\tPartial relations count OK. Count: %ld\n", aPartials);
		}
	}

	if (aVertices != DEBUG_NO_TEST)
	{
		if (aVertices != vertices)
		{
			printf("\n\t\t\tVertices count WRONG. SaveCount: %ld =!= RealCount: %ld\n", aVertices, vertices);
		}
		else
		{
			printf("\n\t\t\tVertices count OK. Count: %ld\n", aVertices);
		}
	}

	if (aComponents != DEBUG_NO_TEST)
	{
		if (aComponents != components)
		{
			printf("\n\t\t\tComponents count WRONG. SaveCount: %ld =!= RealCount: %ld\n", aComponents, components);
		}
		else
		{
			printf("\n\t\t\tComponents count OK. Count: %ld\n", aComponents);
		}
	}

	if (aEdges != DEBUG_NO_TEST)
	{
		if (aEdges != edges)
		{
			printf("\n\t\t\tEdges count WRONG. SaveCount: %ld =!= RealCount: %ld\n", aEdges, edges);
		}
		else
		{
			printf("\n\t\t\tEdges count OK. Count: %ld\n", aEdges);
		}
	}

	printf("\n");

	return ConstRC::Ok;
}
int CTestSuite::ControlFactorization(mpz_t aNumber, mpz_t aRemaining, nfs_fb_type * aFB, nfs_fb_type * aFBEnd)
{
	mpz_t aux;
	mpz_init_set(aux, aNumber);

	do
	{
		while (mpz_divisible_ui_p(aux, (*aFB).p))
		{
			mpz_divexact_ui(aux, aux, (*aFB).p);
		}
	}
	while (++aFB != aFBEnd);

	if (mpz_cmp(aux, aRemaining) == 0)
	{
		mpz_clear(aux);
		return ConstRC::Ok;
	}
	else
	{
		gmp_printf("\n\t\t\tRemaining is different. Expected: %Zd, Computed: %Zd\n", aRemaining, aux);
		mpz_clear(aux);
		return ConstRC::Impropriate;
	}
}
int CTestSuite::ControlFactorization(mpz_t aNumber, mpz_t aRemaining, lattice_type * aFB, lattice_type * aFBEnd, long aSpecialQ, bool aDivisible)
{
	mpz_t aux;
	mpz_init_set(aux, aNumber);

	if (aDivisible)
	{
		if (mpz_divisible_ui_p(aux, aSpecialQ) == 0)
		{
			gmp_printf("\n\t\t\tNumber should be divisible by special Q but it is not! N: %Zd, SpecialQ: %ld\n", aux, aSpecialQ);
		}
		else
		{
			mpz_divexact_ui(aux, aux, aSpecialQ);
			while (mpz_divisible_ui_p(aux, aSpecialQ))
			{
				mpz_divexact_ui(aux, aux, aSpecialQ);
			}
		}
	}

	do
	{
		while (mpz_divisible_ui_p(aux, (*aFB).p))
		{
			mpz_divexact_ui(aux, aux, (*aFB).p);
		}
	}
	while (++aFB != aFBEnd);

	if (mpz_cmp(aux, aRemaining) == 0)
	{
		mpz_clear(aux);	
		return ConstRC::Ok;
	}
	else
	{
		gmp_printf("\n\t\t\tRemaining is different. Expected: %Zd, Computed: %Zd\n", aRemaining, aux);
		mpz_clear(aux);
		return ConstRC::Impropriate;
	}
}

void CTestSuite::TestPoly()
{
	Polynomial * poly = new Polynomial(10);
	poly->Set(10, 435236);
	poly->Set(7, 1);
	poly->Set(5, 423464375);
	poly->Set(4, 1111);
	poly->Set(3, 457588);
	poly->Set(2, 10979875);
	poly->Set(0, 8589);

	std::string str = poly->ToString(EShortFormat);

	Polynomial * poly2 = Polynomial::Parse(str);
}

void CTestSuite::TestPolyMod()
{
	Polynomial * poly1 = new Polynomial(2); // (x + 1)(x + 2)  = x^{2} + 2
	Polynomial * poly2 = new Polynomial(3); // (x + 2)(x + 2)x = x^{3} + x^{2} + x
	Polynomial * poly3 = new Polynomial(1); // (x + 2)
	
	poly1->Set(0, 2);
	poly1->Set(2, 1);

	poly2->Set(1, 1);
	poly2->Set(2, 1);
	poly2->Set(3, 1);

	poly3->Set(0, 5);
	poly3->Set(1, -2);

	// modulo test - result is x + 2
	poly3->Modulo(3);	
	if (poly3->Get(0) != 2)
	{
		printf("TestPolyMod: Error modulo - coeff 0 is %ld, should be 2\n", poly3->Get(0));
	}
	if (poly3->Get(1) != 1)
	{
		printf("TestPolyMod: Error modulo - coeff 1 is %ld, should be 1\n", poly3->Get(1));
	}

	// substract test - result is 2x^{3} + 2x^{2} + 2
	Polynomial * aux = new Polynomial();
	poly3->SubstractModPrime(aux, poly3, poly2, 3);
	if (aux->Get(0) != 2 || aux->Get(1) != 0 || aux->Get(2) != 2 || aux->Get(3) != 2)
	{
		printf("TestPolyMod: Error substract\n");
		aux->PrintToScreen();
		printf("\n");
	}

	// divide test - result is x + 1
	Polynomial * auxx = new Polynomial();
	mpz_t prime;
	mpz_init_set_ui(prime, 3);
	poly1->DivideModPrime(aux, auxx, poly1, poly3, prime);
	if (aux->Get(0) != 1 || aux->Get(1) != 1)
	{
		printf("TestPolyMod: Error divide\n");
		aux->PrintToScreen();
		printf("\n");
	}

	mpz_clear(prime);

	delete poly1;
	poly1 = NULL;
	delete poly2;
	poly2 = NULL;
	delete poly3;
	poly3 = NULL;
	delete aux;
	aux = NULL;
	delete auxx;
	auxx = NULL;
}

void CTestSuite::TestRelations()
{
	CNFSRelation rel(ETrippleLargePrimeV);
	rel.A = 12;
	rel.B = 15;
	mpz_set_str(rel.Get_Parts(1)->Norm, "24523642", 10);
	mpz_set_str(rel.Get_Parts(1)->NormRemaining, "9476943", 10);
	mpz_set_str(rel.Get_Parts(0)->Norm, "1346777777", 10);
	mpz_set_str(rel.Get_Parts(0)->NormRemaining, "72048205", 10);
	rel.Get_Parts(1)->Poly->Set(0, 12445);
	rel.Get_Parts(1)->Poly->Set(1, 8346);
	rel.Get_Parts(1)->Poly->Set(2, 5377224);
	rel.Get_Parts(1)->Poly->Set(3, 1);
	rel.Index = 4;

	int index = -1;
	rel.Get_Parts(0)->Main.AddDivisor(2, 6, 0, 0, index);
	rel.Get_Parts(0)->Main.AddDivisor(3, 2, 0, 0, index);
	rel.Get_Parts(0)->Main.AddDivisor(11, 1, 0, 0, index);
	rel.Get_Parts(0)->Main.AddDivisor(53, 2, ConstEF::DIVIDES_L, 0, index);
	rel.Get_Parts(0)->Main.AddDivisor(224521, 1, 0, 0, index);
	rel.Get_Parts(0)->Main.AddDivisor(225585633, 1, 0, 0, index);

	index = -1;
	rel.Get_Parts(0)->Large.AddDivisor(452166347, 1, 0, 0, index);
	rel.Get_Parts(0)->Large.AddDivisor(962145551, 2, 0, 0, index);

	index = -1;
	rel.Get_Parts(1)->Main.AddDivisor(2, 10, 1, ConstEF::SPECIAL_PRIME, index);
	rel.Get_Parts(1)->Main.AddDivisor(5, 1, 4, 0, index);
	rel.Get_Parts(1)->Main.AddDivisor(97, 2, 45, 0, index);
	rel.Get_Parts(1)->Main.AddDivisor(25632251, 1, 2425, 0, index);
	rel.Get_Parts(1)->Main.AddDivisor(34959157, 1, 124555, 0, index);

	index = -1;
	rel.Get_Parts(1)->Large.AddDivisor(959142241, 1, 12445555, 0, index);
	rel.Get_Parts(1)->Large.AddDivisor(134974393, 1, 32466, ConstEF::DIVIDES_L, index);

	int rc;
	//rc = rel.PrintToScreen();

	FILE * relFILE = fopen("relations.txt", "w");
	if (relFILE == NULL)
	{
		printf("Error: TestRelations - Cannot create a file with relations.\n");
		return;
	}

	rc = rel.PrintToFile(relFILE);
	if (rc != ConstRC::Ok)
	{
		printf("Error: TestRelations - Cannot write relation.\n");
		return;
	}

	fclose(relFILE);
	relFILE = NULL;

	CNFSRelation read(ETrippleLargePrimeV);

	relFILE = fopen("relations.txt", "r");
	if (relFILE == NULL)
	{
		printf("Error: TestRelations - Cannot open a file with relations.\n");
		return;
	}

	rc = read.ReadFromFile(relFILE);
	if (rc != ConstRC::Ok)
	{
		printf("Error: TestRelations - Cannot read relation.\n");
		return;
	}

	if (rel.Equals(read) == false)
	{
		printf("Error: TestRelations - Relations are different!!!\n");
	}
	else
	{
		printf("TestRelations - OK.\n");
	}
}

void CTestSuite::TestMatrix(int argc, char *argv[])
{
	//CNFSParameters param;
	//param.ParseInputParameters(argc, argv);


	//param.AddParameter(PARAM_COMPRESSION_LEVEL, "1");

	//param.Serialize(param);

	CBaseParameters base;
	base.AddParameter(PARAM_SERIALIZATION_DIRECTORY, Utils::GetDirectoryFromPath(argv[0]));

	//IntegerMatrix matrix(10, 5);
	//NormalMatrix matrix(20, 30);
	//SparseMatrix matrix(20, 30);
	//matrix.PutOne(2, 6);
	//matrix.PutOne(2, 7);
	//matrix.PutOne(2, 9);
	//matrix.PutOne(2, 15);
	//matrix.PutOne(4, 20);
	//matrix.PutOne(10, 9);
	//matrix.PutOne(10, 20);
	//matrix.PutOne(19, 29);
	//matrix.PutNumber(26, 2, 1);
	//matrix.PutNumber(116, 2, 2);
	//matrix.PutNumber(336, 2, 4);
	//matrix.PutNumber(-76, 3, 0);
	//matrix.PutNumber(466, 5, 4);
	//matrix.PutNumber(-316, 6, 4);
	//matrix.PutNumber(292386, 7, 1);
	//matrix.PutNumber(142246, 7, 3);
	//matrix.PutNumber(22356, 9, 0);
	//matrix.PutNumber(-226576, 9, 4);

	//matrix.Serialize(base);

	IntegerMatrix matrix;
	//SparseMatrix matrix;
	matrix.Deserialize(base);

	//int t = matrix.IsOne(2,15);
	matrix.PrintToScreen();



	//Polynomial poly;
	//int rc = poly.Deserialize(base);

	//param.Deserialize(base);
}
void CTestSuite::TestPrimes()
{
	unsigned long count;
	long * primes = NULL;
	CIntegerClass::GetPrimes(0, 355000, count, primes);

	if (count > 0)
	{

	}

	free(primes);
}
void CTestSuite::TestTimes()
{
	int max = 1000;
	int delay = 100;

	long start = clock();
	long total;
	long k = 0;

	CTimeContainer * time = new CTimeContainer(12);


	for (int i = 0; i < max; i++)
	{
		time->Start(2);
		for (int j = 0; j < delay; j++)
		{
			k += k;
		}
		time->End(2);
		time->Start(1);
		time->Start(3);
		time->Start(4);
		time->Start(10);
		time->Start(11);
		time->End(1);
		for (int j = 0; j < delay; j++)
		{
			time->End(3);
			k += k;
			time->Start(3);
		}

		time->EndAll();
		time->ResetAllIntermediate();
		time->ConfirmAll();
		time->StartAll();
		for (int j = 0; j < delay; j++)
		{
			k += k;
			time->EndAll();
			time->ConfirmAll();
		}

		time->ResetAll();
	}

	total = clock() - start;
	printf("\nTimeContainer: %d", total);


	long timeS;
	start = clock();
	for (int i = 0; i < max; i++)
	{
		timeS = clock();
		for (int j = 0; j < delay; j++)
		{
			k += k;
		}
		timeS = clock();
		timeS = clock();
		timeS = clock();
		timeS = clock();
		timeS = clock();
		timeS = clock();
		timeS = clock();
		for (int j = 0; j < delay; j++)
		{
			timeS = clock();
			k += k;
			timeS = clock();
		}

		for (int j = 0; j < 12; j++)
		{
			timeS = clock() - timeS;
			timeS = 0;
			timeS = 1;
			timeS = clock();
		}
		for (int j = 0; j < delay; j++)
		{
			k += k;
			for (int l = 0; l < 12; l++)
			{
				timeS = clock() - timeS;
				timeS = 1;
			}
		}

		
		for (int j = 0; j < 12; j++)
		{
			timeS = 0;
		}
	}

	total = clock() - start;
	printf("\nTime: %d", total);

}
void CTestSuite::TestSieving(int argc, char *argv[])
{
	// for using libxml2 in more threads
	xmlInitParser();

	// parameters specialized for NFS
	CNFSParameters param;
	param.ParseInputParameters(argc, argv);


	CNFSClassicalSievingPhase phase;
	int rc = phase.RunPhase(&param);
	if (rc != ConstRC::Ok)
	{
		Utils::Wait();
	}
}
void CTestSuite::TestWriter()
{
	long start = clock();

	// read new relations
	CBaseParameters neww;
	neww.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, "relations_new.xml");

	CNFSRelation relation(ETrippleLargePrimeV);
	std::vector<ARelation*> relations;
	relations.reserve(256);

	CRelationReader reader(&neww);

	int rc;

	while ((rc = reader.ReadRelation(&relation)) == ConstRC::Ok)
	{
		relations.push_back(new CNFSRelation(relation));
	}

	float total = (clock() - start)/ (float)CLOCKS_PER_SEC;
	int nbrRel = reader.Get_CurrentIndex() + 1;
	reader.Close();

	printf("Read relations: %d, Total time: %12.4f sec \n", nbrRel, total);


	// write relations to new file temp, also copy all old relations
	start = clock();

	CBaseParameters old;
	old.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, "relations_old.xml");
	neww.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, "relations_temp.xml");

	CRelationWriter writer(&neww);

	// USE CopyRelationsFrom

	//rc = writer.CopyRelationsFrom(old);
	//if (rc != ConstRC::Ok)
	//{
	//	printf("Error %d", rc);
	//}

	//total = (clock() - start)/ (float)CLOCKS_PER_SEC;
	//printf("Copy relations... Total time: %12.4f sec \n", total);

	// USE reader and writer

	reader.Set_Parameters(&old);
	while ((rc = reader.ReadRelation(&relation)) == ConstRC::Ok)
	{
		rc = writer.WriteRelation(&relation);
		if (rc != ConstRC::Ok)
		{
			printf("Error %d", rc);
		}
	}

	total = (clock() - start)/ (float)CLOCKS_PER_SEC;
	printf("Copy relations with reader... Total time: %12.4f sec \n", total);


	start = clock();

	int max = relations.size();
	for (int i = 0; i < max; i++)
	{
		rc = writer.WriteRelation(relations[i]);
		if (rc != ConstRC::Ok)
		{
			printf("Error %d", rc);
		}
	}

	writer.Close();

	total = (clock() - start)/ (float)CLOCKS_PER_SEC;
	printf("Writed relations: %d, Total time: %12.4f sec \n", max, total);
	start = clock();

	for (int i = 0; i < max; i++)
	{
		delete relations[i];
	}

	total = (clock() - start)/ (float)CLOCKS_PER_SEC;
	printf("Delete relations... Total time: %12.4f sec \n", total);

	Utils::Wait();

	rc = remove("relations_neww.xml");
	rc = rename("relations_temp.xml", "relations_neww.xml");

	Utils::Wait();
}
void CTestSuite::TestRewrite()
{
	CBaseParameters neww;
	neww.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, "temp.xml");
	CBaseParameters old;
	old.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, "partial_relations_20081105150525.xml");

	//CRelationWriter writer(&neww);
	CRelationReader reader(&old);


	int rc;
	CNFSRelation relation(ETrippleLargePrimeV);

	long start = clock();

	FILE * file = fopen("temp.xml", "w");

	while ((rc = reader.ReadRelation(&relation)) == ConstRC::Ok)
	{
		relation.Index = reader.Get_CurrentIndex();
		//rc = writer.WriteRelation(&relation);
		rc = relation.PrintToFile(file);
		if (rc != ConstRC::Ok)
		{
			printf("Error %d", rc);
		}
	}

	//writer.Close();
	fclose(file);
	float total = (clock() - start)/ (float)CLOCKS_PER_SEC;
	printf("Copy %d relations with reader... Total time: %12.4f sec \n", (reader.Get_CurrentIndex() + 1), total);

	reader.Close();

	//rc = remove("partial_relations_20081105150525.xml");
	//rc = rename("temp.xml", "partial_relations_20081105150525.xml");

	Utils::Wait();
}


void CTestSuite::TestMPF()
{
	mpf_t n;
	mpf_init(n);

	mpf_set_str(n, "123456@1", 10);

	gmp_printf("n = %.*Ff", 10, n);


	Utils::Wait();
}
void CTestSuite::TestCandidates()
{
	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, "candidate_polynomial_pairs_20090115112852.xml");
	CCandidatePairs * cand = new CCandidatePairs(100);
	int rc = cand->Deserialize(param);
	if (rc != ConstRC::Ok)
	{
		printf("Error deserialize...\n");
		return;
	}

	mpz_t result;
	mpz_t n;
	mpz_init(result);
	mpz_init_set_str(n, "630341520131416462199663050694481763", 10);

	for (int i = 0; i < 100; i++)
	{
		cand->FirstPoly[i]->Evaluate(result, cand->Root[i]);
		mpz_mod(result, result, n);

		if (mpz_cmp_ui(result, 0) != 0)
		{
			printf("Error 1 - no root!\n");
			printf("\t- index =  %d\n", i);
			gmp_printf("\t- m = %Zd\n", cand->Root[i]);
			cand->FirstPoly[i]->PrintToScreen();
			printf("\n");
		}

		cand->SecondPoly[i]->Evaluate(result, cand->Root[i]);
		mpz_mod(result, result, n);

		if (mpz_cmp_ui(result, 0) != 0)
		{
			printf("Error 2 - no root!\n");
			printf("\t- index =  %d\n", i);
			gmp_printf("\t- m = %Zd\n", cand->Root[i]);
			cand->SecondPoly[i]->PrintToScreen();
			printf("\n");
		}
	}

	mpz_clear(result);
	mpz_clear(n);
}
