#include "thread_pool.h"
#include "nfs_utils.h"
#include "constants.h"
#include "const_return_codes.h"
#include "log.h"

CThreadPool::CThreadPool(const int &aThreadsCount, const int &aMaxQueueSize, const bool &aDoNotBlockWhenFull) :
	iThreadsCount(aThreadsCount), iMaxQueueSize(aMaxQueueSize), iDoNotBlockWhenFull(aDoNotBlockWhenFull)
{
	this->iThreads = NULL;
	this->iThreads = (pthread_t *)Utils::AllocationMachine((void**)&this->iThreads, this->iThreadsCount * sizeof(pthread_t));
	if (this->iThreads == NULL)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot allocate memory for threads.");
		return;
	}
	this->iCurQueueSize = 0;	
	this->iQueueHead = NULL;	
	this->iQueueTail = NULL;	
	this->iQueueClosed = false;	
	this->iShutdown = false;
	InitDestroyed(false);
	this->iInnerError = false;

	int rtn;
	if ((rtn = pthread_mutex_init(&(this->iQueueMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",rtn,0,"Cannot initialize mutex.");
	}
	if ((rtn = pthread_cond_init(&(this->iQueueNotEmptyCond), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",rtn,0,"Cannot initialize condition.");
	}	
	if ((rtn = pthread_cond_init(&(this->iQueueNotFullCond), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",rtn,0,"Cannot initialize condition.");
	}
	if ((rtn = pthread_cond_init(&(this->iQueueEmptyCond), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",rtn,0,"Cannot initialize condition.");
	}

	/* create threads */	
	for (int i = 0; i < this->iThreadsCount; i++) 
	{	
		if ((rtn = pthread_create(&(this->iThreads[i]),	
						NULL,	
						&CThreadPool::WorkerThreadStart,	
						(void*)this)) != 0)
		{
			CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot create thread.");
			return;
		}	
	}
}

CThreadPool::CThreadPool(const CThreadPool &aOperand) :
	iThreadsCount(aOperand.iThreadsCount), iMaxQueueSize(aOperand.iMaxQueueSize), 
	iDoNotBlockWhenFull(aOperand.iDoNotBlockWhenFull), iCurQueueSize(aOperand.iCurQueueSize),
	iQueueClosed(aOperand.iQueueClosed), iShutdown(aOperand.iShutdown), 
	iInnerError(aOperand.iInnerError)
{
	this->iThreads = NULL;
	this->iThreads = (pthread_t *)Utils::AllocationMachine((void**)&this->iThreads, this->iThreadsCount * sizeof(pthread_t));
	if (this->iThreads == NULL)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot allocate memory for threads.");
		return;
	}
	this->iQueueHead = NULL;	
	this->iQueueTail = NULL;	
	InitDestroyed(aOperand.iDestroyed);

	int rtn;
	if ((rtn = pthread_mutex_init(&(this->iQueueMutex), NULL)) != 0)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot inititilize mutex.");
	}
	if ((rtn = pthread_cond_init(&(this->iQueueNotEmptyCond), NULL)) != 0)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot initilize condition.");
	}	
	if ((rtn = pthread_cond_init(&(this->iQueueNotFullCond), NULL)) != 0)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot initilize condition.");
	}
	if ((rtn = pthread_cond_init(&(this->iQueueEmptyCond), NULL)) != 0)
	{
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot initilize condition.");
	}

	/* create threads */	
	for (int i = 0; i < this->iThreadsCount; i++) 
	{	
		if ((rtn = pthread_create(&(this->iThreads[i]),	
						NULL,	
						&CThreadPool::WorkerThreadStart,	
						NULL)) != 0)
		{
			CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot create thread.");
			return;
		}	
	}
}

void CThreadPool::InitDestroyed(bool aValue) {

	if (pthread_mutex_init(&iDestroyedMutex, NULL) != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"CThreadPool","CThreadPool",0,0,"Cannot initialize mutex.");
		return;
	}
	pthread_mutex_lock(&iDestroyedMutex);
	this->iDestroyed = aValue;
	pthread_mutex_unlock(&iDestroyedMutex);

	return;
}

CThreadPool::~CThreadPool()
{
	bool destroyed;

	pthread_mutex_lock(&iDestroyedMutex);
	destroyed = iDestroyed;
	pthread_mutex_unlock(&iDestroyedMutex);

	if (!destroyed) {
		Destroy(false);
	}

	if (pthread_mutex_destroy(&iDestroyedMutex) != ConstRC::Ok) CLog::LogError(WHEREARG,"CThreadPool","~CThreadPool",0,0,"Cannot destroy mutex.");

	int rtn;
	if ((rtn = pthread_mutex_destroy(&(this->iQueueMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","~CThreadPool",rtn,0,"Cannot destroy mutex.");
	}
	if ((rtn = pthread_cond_destroy(&(this->iQueueNotEmptyCond))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","~CThreadPool",rtn,0,"Cannot destroy condition.");
	}	
	if ((rtn = pthread_cond_destroy(&(this->iQueueNotFullCond))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","~CThreadPool",rtn,0,"Cannot destroy condition.");
	}
	if ((rtn = pthread_cond_destroy(&(this->iQueueEmptyCond))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","~CThreadPool",rtn,0,"Cannot destroy condition.");
	}

	/* Now free pool structures */	
	free(this->iThreads);	
	thread_work * curNode;
	while (this->iQueueHead != NULL) 
	{	
		curNode = this->iQueueHead->next;
		free(this->iQueueHead);
		this->iQueueHead = curNode;	
	}	
}

bool CThreadPool::Get_Destroyed() const {
	bool destroyed;

	pthread_mutex_lock(&iDestroyedMutex);
	destroyed = iDestroyed;
	pthread_mutex_unlock(&iDestroyedMutex);

	return destroyed;
}

int CThreadPool::Destroy(bool aWaitForFinish)
{
	int rtn;
	if ((rtn = pthread_mutex_lock(&(this->iQueueMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot lock mutex.");	
		return ConstRC::GeneralError;
	}

	/* Is a shutdown already in progress? */	
	if (this->iQueueClosed || this->iShutdown) 
	{	
		if ((rtn = pthread_mutex_unlock(&(this->iQueueMutex))) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot unlock mutex.");
			return ConstRC::GeneralError;
		}
		return ConstRC::Ok;	
	}	

	this->iQueueClosed = true;	

	/* If the finish flag is set, wait for workers to drain queue */	
	if (aWaitForFinish) 
	{	
		while (this->iCurQueueSize > 0 && this->iInnerError == false) 
		{	
			if ((rtn = pthread_cond_wait(&(this->iQueueEmptyCond), &(this->iQueueMutex))) != ConstRC::Ok)
			{
				CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot wait for condition.");
				return ConstRC::GeneralError;
			}
		}	
	}	

	this->iShutdown = true;	

	if ((rtn = pthread_mutex_unlock(&(this->iQueueMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot unlock mutex.");
		return ConstRC::GeneralError;
	}

	/* Wake up any workers so they recheck shutdown flag */	
	if ((rtn = pthread_cond_broadcast(&(this->iQueueNotEmptyCond))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot broadcast condition.");
		return ConstRC::GeneralError;
	}
	if ((rtn = pthread_cond_broadcast(&(this->iQueueNotFullCond))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot broadcast condition.");
		return ConstRC::GeneralError;
	}

	/* Wait for workers to exit */	
	for(int i = 0; i < this->iThreadsCount; i++) 
	{	
		if ((rtn = pthread_join(this->iThreads[i], NULL)) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CThreadPool","Destroy",rtn,0,"Cannot join thread.");
			return ConstRC::GeneralError;
		}
	}

	pthread_mutex_lock(&iDestroyedMutex);
	this->iDestroyed = true;
	pthread_mutex_unlock(&iDestroyedMutex);

	return ConstRC::Ok;
}


int CThreadPool::AddWork(APhase * aPhase, AFactorAlgParameters * aParam)
{
	pthread_mutex_lock(&this->iQueueMutex);	

	if ((this->iCurQueueSize == this->iMaxQueueSize) &&	this->iDoNotBlockWhenFull) 
	{	
		pthread_mutex_unlock(&this->iQueueMutex);	
		return ConstRC::GeneralError;	
	}	
	while ((this->iCurQueueSize == this->iMaxQueueSize) &&	
		(!(this->iShutdown || this->iQueueClosed))) 
	{	
		pthread_cond_wait(&this->iQueueNotFullCond, &this->iQueueMutex);	
	}	

	if (this->iShutdown || this->iQueueClosed) 
	{	
		pthread_mutex_unlock(&this->iQueueMutex);	
		return ConstRC::GeneralError;	
	}	

	/* allocate work structure */	
	thread_work * work = NULL;
	work = (thread_work *)Utils::AllocationMachine((void**)&work, sizeof(thread_work));
	if (work == NULL)
	{
		CLog::LogError(WHEREARG,"CThreadPool","AddWork",0,0,"Cannot allocate memory for work.");
		return ConstRC::NotEnoughMemory;
	}
	work->param = aParam;	
	work->phase = aPhase;	
	work->next = NULL;	
	if (this->iQueueTail == NULL) 
	{	
		this->iQueueTail = this->iQueueHead = work;	
		pthread_cond_broadcast(&this->iQueueNotEmptyCond);	
	} 
	else 
	{	
		(this->iQueueTail)->next = work;	
		this->iQueueTail = work;	
	}	
	
	this->iCurQueueSize++;	
	pthread_mutex_unlock(&this->iQueueMutex);

	return ConstRC::Ok;
}

void * CThreadPool::WorkerThreadStart(void * aArg)
{
	if (aArg == NULL)
	{
		CLog::LogError(WHEREARG,"CThreadPool","WorkerThreadStart",0,0,"Arg is NULL.");
		pthread_exit(NULL);	
		return NULL;
	}

	thread_work * myWork = NULL;
	CThreadPool * threadPool = (CThreadPool *)aArg;
	int rc;

	while (true)
	{
		pthread_mutex_lock(&(threadPool->iQueueMutex));	

		while ((threadPool->iCurQueueSize == 0) && (!threadPool->iShutdown)) 
		{	
			pthread_cond_wait(&(threadPool->iQueueNotEmptyCond), &(threadPool->iQueueMutex));	
		}	

		if (threadPool->iShutdown)
		{	
			pthread_mutex_unlock(&(threadPool->iQueueMutex));	
			pthread_exit(NULL);	
			break;
		}

		myWork = threadPool->iQueueHead;
		threadPool->iCurQueueSize--;	
		if (threadPool->iCurQueueSize == 0)
		{
			threadPool->iQueueHead = threadPool->iQueueTail = NULL;
		}
		else
		{
			threadPool->iQueueHead = myWork->next;	
		}

		if ((!threadPool->iDoNotBlockWhenFull) && 
			(threadPool->iCurQueueSize == (threadPool->iMaxQueueSize - 1)))
		{
			pthread_cond_broadcast(&(threadPool->iQueueNotFullCond));
		}

		if (threadPool->iCurQueueSize == 0)
		{
			pthread_cond_signal(&(threadPool->iQueueEmptyCond));
		}

		pthread_mutex_unlock(&(threadPool->iQueueMutex));

		// Do work
		rc = myWork->phase->RunPhase(myWork->param);
		// TODO some inner error if return value is not ConstRC::Ok...

		if (myWork != NULL)
		{
			free(myWork);
			myWork = NULL;
		}
		// error - should we react or not
		if (rc != ConstRC::Ok && threadPool->Get_ThreadsCount() == 1)
		{
			CLog::LogError(WHEREARG,"CThreadPool","WorkerThreadStart",rc,0,"Computation unsuccessfull.");
			// pipe line model - we cannot continue
			pthread_mutex_lock(&(threadPool->iQueueMutex));	
			threadPool->iInnerError = true;
			pthread_cond_signal(&(threadPool->iQueueEmptyCond));
			pthread_mutex_unlock(&(threadPool->iQueueMutex));
			pthread_exit(NULL);	
			break;
		}
	}

	return NULL;
}

