#include "time_container.h"

#include "base_parameters.h"
#include "const_return_codes.h"
#include "const_parameter_names.h"
#include "const_filenames.h"
#include "log.h"


using namespace std;

CTimeContainer::CTimeContainer(int aMaxSize) : MSerializable(ConstFN::TIME_CONTAINER),
	iTotalTimes(aMaxSize, 0), iStartTimes(aMaxSize, 0), iConfirmTotalTimes(aMaxSize, 0), iIntermediateTimes(aMaxSize, 0), iMutexes(aMaxSize)
{
	this->iMaxSize = aMaxSize;
	this->iStorageRatio = 0;
	if (CLOCKS_PER_SEC != CTimeContainer::STANDARD_CLOCKS_PER_SEC)
	{
		this->iStorageRatio = CLOCKS_PER_SEC / CTimeContainer::STANDARD_CLOCKS_PER_SEC; 
	}
	int rc;
	for (int i = 0; i < aMaxSize; i++)
	{
		if ((rc = pthread_mutex_init(&(this->iMutexes[i]), NULL)) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CTimeContainer","CTimeContainer",rc,0,"Cannot initialize mutex.");
		}
	}
}
CTimeContainer::~CTimeContainer()
{
	int rc;
	int max = (int)this->iMutexes.size();
	for (int i = 0; i < max; i++)
	{
		if ((rc = pthread_mutex_destroy(&(this->iMutexes[i]))) != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CTimeContainer","~CTimeContainer",rc,0,"Cannot destroy mutex.");
		}
	}
}
int CTimeContainer::Start(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iStartTimes[aIndex] = clock();
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CTimeContainer::StartAll()
{
	long start = clock();
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iStartTimes[i] = start;
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CTimeContainer::End(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		long diff = clock() - this->iStartTimes[aIndex];
		if (diff > 0)
		{
			diff = GetStoreValue(diff);
			this->iTotalTimes[aIndex] += diff;
			this->iIntermediateTimes[aIndex] += diff;
		}
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CTimeContainer::EndAll()
{
	long end = clock();
	long diff;
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		diff = end - this->iStartTimes[i];
		if (diff > 0)
		{
			diff = GetStoreValue(diff);
			this->iTotalTimes[i] += diff;
			this->iIntermediateTimes[i] += diff;
		}
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CTimeContainer::Reset(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iStartTimes[aIndex] = 0;
		this->iTotalTimes[aIndex] = 0;
		this->iConfirmTotalTimes[aIndex] = 0;
		this->iIntermediateTimes[aIndex] = 0;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CTimeContainer::ResetAll()
{
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iStartTimes[i] = 0;
		this->iTotalTimes[i] = 0;
		this->iConfirmTotalTimes[i] = 0;
		this->iIntermediateTimes[i] = 0;
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CTimeContainer::ResetIntermediate(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iIntermediateTimes[aIndex] = 0;
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CTimeContainer::ResetAllIntermediate()
{
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iIntermediateTimes[i] = 0;
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
int CTimeContainer::Confirm(int aIndex)
{
	if (Control(aIndex))
	{
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		this->iConfirmTotalTimes[aIndex] = this->iTotalTimes[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return ConstRC::Ok;
	}

	return ConstRC::BadArgument;
}
int CTimeContainer::ConfirmAll()
{
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		this->iConfirmTotalTimes[i] = this->iTotalTimes[i];
		pthread_mutex_unlock(&this->iMutexes[i]);
	}

	return ConstRC::Ok;
}
float CTimeContainer::Get_TotalTime(int aIndex) const
{
	if (Control(aIndex))
	{
		long ret;
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		ret = this->iTotalTimes[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return GetSeconds(ret);
	}

	return 0;
}
float CTimeContainer::Get_ConfirmTotalTime(int aIndex) const
{
	if (Control(aIndex))
	{
		long ret;
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		ret = this->iConfirmTotalTimes[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return GetSeconds(ret);
	}

	return 0;
}
float CTimeContainer::Get_IntermediateTime(int aIndex) const
{
	if (Control(aIndex))
	{
		long ret;
		pthread_mutex_lock(&this->iMutexes[aIndex]);
		ret = this->iIntermediateTimes[aIndex];
		pthread_mutex_unlock(&this->iMutexes[aIndex]);

		return GetSeconds(ret);
	}

	return 0;
}
int CTimeContainer::Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_TIME_CONTAINER_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CTimeContainer","Serialize",rc,0,"Could not write the main tag.");
		return ConstRC::GeneralError;
    }
	
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_MAX_BOUND_NAME, this->iMaxSize);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_CONFIRM_TOTAL_TIMES_NAME);
    if (rc < 0) 
	{
        CLog::LogError(WHEREARG,"CTimeContainer","Serialize",rc,0,"Could not write the tag.");
		return ConstRC::GeneralError;
    }
	
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	long time;
	for (int i = 0; i < this->iMaxSize; i++)
	{
		pthread_mutex_lock(&this->iMutexes[i]);
		time = this->iConfirmTotalTimes[i];
		pthread_mutex_unlock(&this->iMutexes[i]);
		rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_ELEMENT_NAME, time);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_CONFIRM_TOTAL_TIMES_NAME
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CTimeContainer","Serialize",rc,0,"Could not end the tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
    if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CTimeContainer","Serialize",rc,0,"Could not end the main tag.");
		return ConstRC::GeneralError;
    }
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	
	return rc;
}
int CTimeContainer::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_TIME_CONTAINER_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CTimeContainer","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CTimeContainer","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	// delete all data
	ResetAll();

	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_MAX_BOUND_NAME))
		{
			int max;
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize max size...", max, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}

			if (this->iMaxSize < max)
			{
				// add new elements
				while ((int)this->iTotalTimes.size() < max)
				{
					this->iTotalTimes.push_back(0);
					this->iStartTimes.push_back(0);
					this->iConfirmTotalTimes.push_back(0);
					pthread_mutex_t mutex;
					this->iMutexes.push_back(mutex);
					if ((rc = pthread_mutex_init(&(mutex), NULL)) != ConstRC::Ok)
					{
						CLog::LogError(WHEREARG,"CTimeContainer","Deserialize",rc,0,"Cannot initialize mutex.");
					}
				}
				this->iMaxSize = max;
			}

			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_CONFIRM_TOTAL_TIMES_NAME))
		{
			int elementsDepth = xmlTextReaderDepth(aReader);
			long time;
			for (int i = 0; i < this->iMaxSize; i++)
			{
				if (CXMLHelper::MoveToNextNode(aReader, elementsDepth) != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize confirm total time...", time, false);
				if (rc != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				pthread_mutex_lock(&this->iMutexes[i]);
				this->iTotalTimes[i] = time;
				this->iConfirmTotalTimes[i] = time;
				pthread_mutex_unlock(&this->iMutexes[i]);
			}
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
