#ifndef __BIT_MATRIX_CLASS__
#define __BIT_MATRIX_CLASS__

#include "normal_matrix_class.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

typedef unsigned int matrix_type;

/*!
	\brief A class for memory-savvy representation of not-too-sparse matrices over GF(2).
	
This class is used for memory-savvy representation of matrices over GF(2). For this representation, it uses
	a two-dimensional array of custom type matrix_type, which is in i386 version defined as unsigned int. 
	The aim of this class is to provide suitable representation of not very sparse matrices over GF(2), and to
	exploit fast assembler operations (addition, xor, modulo) in implementation of matrix arithmetics.
	This type is suitable for representation of not very sparse bit matrices, say having at least 2 per cent of
	nonzero bits. For matrices which are sparser, there is another type defined in this program - see the next
	section.
	As for details of the representation: each bit in the two-dimensional array is treated separately, using
	bitwise operations >>, &, | etc. Let us say that the custom type matrix_type has B bits (usual value is 32).
	Then, to represent a matrix with R rows and C columns, a two-dimensional array of matrix_type is used, which
	has R rows and /C/B\ columns (where operator /x\  denotes "the upper integer part of x"). So, to represent
	a matrix 64x32, a field having 64 rows and 1 column in each row is used in i386 version.
	Let us imagine that I is index of a row, J is index of a column to seek (so, I>=0 and I lesser than C). The bit on [I,J]-th 
	position is physically located in the following way:
		- at first, we determine which column of the two-dimensional array is to be chosen. Index of this
		  column is given by value of \J/B/ (where operator \x/ denotes "the lower integer part of x"). So,
		  in case of B=32, bits with indices 0 to 31 are stored in the 0-th column of the two-dimensional
		  array, bits with indices 32 to 63 in the 1-st column etc. Let us denote the "right" column by COL.
		- now, we must determine which  bit in this column is the right one. To determine this, we compute
		  K = J % B, which is a number from 0 to 31; now, we perform the operation (COL & (1 << K)) and test
		  the result. If it is nonzero, the bit was nonzero, and vice versa.
		  
	This means that the bit number zero is located in the LSB of matrix_type in the first column, and that the bit
	number 31 is located in the MSB of matrix_type.	
	"Bit matrices" implemented in this class have been designed for use in the block Lanczos algorithm, which often
	uses matrices with width of a computer word. Their arithmetics exploits the fact that they can be added or 
	multiplied very fast, with use of XOR operator. However, they have weak points. The most problematic of them is 
	transposition, which is not too fast. If you need to perform transpositions very frequently, use rather another
	type. Several (say 3) transpositions per second are, on the other hand, tolerable even with this type.
	
	

*/
class BitMatrix : public AbstractMatrix {
public:
	BitMatrix();
	BitMatrix(long aRows, long aColumns);
	~BitMatrix();
        static void DefineKShifts();
	long GetMaxAllocatedRowIndex()    const;
	long GetMaxAllocatedColumnIndex() const;
	int  Allocate();
	int  Randomize();
	int  Zeroize();
	int  Copy(BitMatrix* aSource);
	int  Copy(AbstractMatrix* aSource);
	bool Equals(BitMatrix* aMatrix);
	void PrintToScreen();
	int  PutOne(long aRow, long aColumn);
	int  PutZero(long aRow, long aColumn);
	int  IsOne(long aRow, long aColumn);
	int  IsZero(long aRow, long aColumn);
	int  IsZero();
	int  IsZeroRow(long aRow);
	int  NonzeroElements();
	matrix_type GetBlock(long aRow, long aBlock);
	void SetBlock(long aRow, long aBlock, matrix_type aValue);
        static int GetBitsInMT();
	// Swaps etc.
	int  ZeroizeRow(long aRow);
	int  ZeroizeRow(long* aRowList, long aListMaxIndex);
	int  SwapRows(long aRow1, long aRow2);
	int  SwapRows(long aRow1, long aRow2, long aMaxIndex);
	int  AddRows(long aTarget, long aSource);
	int  AddRows(long aTarget, long aSource, long aMaxIndex);
	int  AddToRow(long aRow, AbstractMatrix *aOperand, long aRow2);
	int  ZeroizeColumn(long aColumn);
	int  ZeroizeColumn(long* aColumnList, long aListMaxIndex);
	int  SwapColumns(long aColumn1, long aColumn2);
	int  AddColumns(long aTarget, long aSource);
	// Mathematical operations
	int  	   Transpose(BitMatrix* aTarget);
	int  	   Transpose(AbstractMatrix* aTarget);
	BitMatrix* Transpose();
	int  	   Add(BitMatrix* aTarget, BitMatrix* aOperand2);
	int  	   Add(AbstractMatrix* aTarget, AbstractMatrix* aOperand2);
	BitMatrix* Add(BitMatrix* aOperand2);
	int	MultiplyInternal(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2);
	int     MultiplyInternalTransposed(BitMatrix* aTarget, BitMatrix* aOperand1, BitMatrix* aOperand2);
	int	MultiplyInternal(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	int     MultiplyInternalTransposed(AbstractMatrix* aTarget, AbstractMatrix* aOperand1, AbstractMatrix* aOperand2);
	BitMatrix*	MultiplyInternal(BitMatrix* aOperand1, BitMatrix* aOperand2);
	BitMatrix*	MultiplyInternalTransposed(BitMatrix* aOperand1, BitMatrix* aOperand2);

	// Type conversions
	NormalMatrix*     ToNormalMatrix();
	static BitMatrix* ToBitMatrix(NormalMatrix* aMatrix);
	// Others
	int CalculateRank();
	int CalculateRank(bool aOnMyself);
	int ToSemidiagonalShape();
	int ToSemidiagonalShapeForLanczos(BitMatrix* aMirror);
	int ToSemidiagonalShapeForCFRAC(BitMatrix* aMirror,long aRows, long aCols);
	int PerformColumnMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask);
	int PerformColumnMask(BitMatrix* aTarget, BitMatrix* aOperand, matrix_type aMask);
	int PerformRowMask(AbstractMatrix* aTarget, AbstractMatrix* aOperand, matrix_type aMask);
	int PerformRowMask(BitMatrix* aTarget, BitMatrix* aOperand, matrix_type aMask);
	int AddColumnsTwoMatrices(BitMatrix* aTarget, BitMatrix* aSource, long aColumn1, long aColumn2);
	BitMatrix* BlockLanczosMethod(BitMatrix* aSymmetric, BitMatrix* aMatrixB, BitMatrix* aTransposed); // finds solutions of a symmetric matrix.
    int Save(char *aName);
    int Load(char *aName);

	// De/Serialization methods
	/*int Serialize(const CBaseParameters& aParam) const;
	int Serialize(const CBaseParameters& aParam, xmlTextWriterPtr& aWriter) const;

	int Deserialize(const CBaseParameters& aParam);
	int Deserialize(const CBaseParameters& aParam, xmlTextReaderPtr& aReader);*/
protected:
	int WriteData(xmlTextWriterPtr aWriter) const;
	int ReadData(xmlTextReaderPtr aReader);

protected:
	long max_allocated_row_index;
	long max_allocated_column_index;
	matrix_type** matrix;
	static long bits_in_m_t; ///< 	Common to all instances, reflects the number of bits in matrix_type. This value	is needed for type-independent implementation of the BitMatrix class.

public:
	static matrix_type* K_SHIFTS;
};

#endif

