// This definition will force QS to save relations with primes instead of prime indices in FB.
#define USE_PRIMES

#include <string.h>
#include <assert.h>
#include <typeinfo>   // For std::bad_cast
#include <boost/version.hpp>
#include <iostream>
#include "quadratic_sieve.h"
#include "cfrac_class.h"
#include "pollard_p_m_1.h"
#include "ecm_old.h"
#include "filesystem_connector.h"
#include "filesystem_message.h"
#include "qs_parameters.h"
#include "mulm.h"
#include "enums_common.h"
#include "common_definitions.h"
#include "const_hashtables.h"

extern "C" {
#include "gmp_tests.h"
}

// The definition of this constant is in definitions.h
#ifdef WITH_XML_SERIALIZATOR
#include "xml_qs_serializator.h"
#endif

//#define SIQS_DEBUG

#include "algorithm_exception.h"
#include "log.h"


int start;
int local_size_statistics[9];
int global_size_statistics[9];
double total_logs_1 = 0, total_logs_2 = 0, iters_1 = 0, iters_2 = 0;
  int*    prime_iters_1;
  int*    prime_iters_2;
  int*    prime_logs_1;
  int*    prime_logs_2;
//int unusables = 0;  

int  CompareLong(const void* aArg1, const void* aArg2){	 // a non-member function used for qsort
	return (*((long *) aArg1) - *((long *) aArg2));
}

int  CompareCCE(const void* aArg1, const void* aArg2) {	// a non-member function used for qsort
	return (*((int*) aArg1) -  *((int*) aArg2));
}

int  CompareDivisors(const void* aArg1, const void* aArg2) { // a non-member function used for qsort
	return ((*((sieving_element*) aArg1)).p - (*((sieving_element*) aArg2)).p);
}

int  CompareMainSievingTypes(const void* aArg1, const void* aArg2)
{
	return ((*((main_sieving_type*) aArg1)) - (*((main_sieving_type*) aArg2)));
}



/*int AdHoc(int aIndex){
 return ((aIndex == 793) || (aIndex == 1531) || (aIndex == 2144) || (aIndex == 2224) || (aIndex == 2441) || (aIndex == 2940)  || (aIndex == 2968) || (aIndex == 3910) || (aIndex == 3999) || (aIndex == 4200) || (aIndex == 4375));
}*/

/*
	This class implements the Multiple Polynomial Quadratic Sieve. It is designed to
	be easily maintainable and expandable. As the number of potential set-ups of QS
	rises with every new feature implemented, multilevel approach has been chosen.
	That means that the top-level methods are oblivious about real internal state of
	the QS, and any program branches are implemented only on lower levels (methods called
	from methods, or perhaps a level lower).

*/
// CONSTRUCTORS AND DESTRUCTORS
QuadraticSieve::QuadraticSieve(){
	// at first, default constructor of AbstractSieve should run and initiate
	// mpz_t N and kN, writing corresponding bits to inited; anyway, we'll check it
	if (!(this->inited & N_INITED)){
		mpz_init_set_si(this->N,INIT_VALUE_N);
		this->inited |= N_INITED;
	}
	if (!(this->inited & kN_INITED)){
		mpz_init(this->kN);
		this->inited |= kN_INITED;
	}
	if (!(this->inited & F1_INITED)){
		mpz_init_set_ui(this->factor_1,INIT_VALUE_FACTOR);
		this->inited |= F1_INITED;
	}
	if (!(this->inited & F2_INITED)){
		mpz_init_set_ui(this->factor_2,INIT_VALUE_FACTOR);
		this->inited |= F2_INITED;
	}
	if (!(this->inited & M_INITED)){
		mpz_init_set_ui(this->M,DEFAULT_VALUE_M);
		this->inited |= M_INITED;
	}
	if (!(this->inited & ROOT_INITED)){
		mpz_init(this->root);
		this->inited |= ROOT_INITED;
	}
	if (!(this->inited & MPF_INITED)){
		mpz_init(this->maximal_partial_factor);
		this->inited |= MPF_INITED;
	}
	if (!(this->inited & MIN_C_VALUE_INITED)) {
		mpz_init(this->minimal_candidate_value);
		this->inited |= MIN_C_VALUE_INITED;
	}
	if (!(this->inited & MAX_C_VALUE_INITED)) {
		mpz_init(this->maximal_candidate_value);
		this->inited |= MAX_C_VALUE_INITED;
	}
	if (!(this->inited & IDEAL_A_VALUE_INITED)){
		mpz_init(this->ideal_a_value);
		this->inited |= IDEAL_A_VALUE_INITED;
	}
	this->extK		  = FALSE;
	this->div_allocate        = DEFAULT_DIV_ALLOCATE;
	this->k 		   = INIT_VALUE_k;
	this->checklevel           = 0;
	this->info_mode 	   = TRUE;
	this->assertion_mode 	   = FALSE;
	this->density_info_mode    = FALSE;
	this->rank_calculation_mode= FALSE;
	this->value_distribution_mode = FALSE;
	this->contini_threshold    = 0;	
	this->error_factor         = SIEVE_ROUNDOFF_ERROR;
	this->integer_M          = mpz_get_ui(this->M);
	this->using_variations          = EDoNotUseVariations;
	this->QSMode		        = EMpqs;
	this->power_sieving_mode        = EDoNotSieveWithPrimePowers;
	this->factoring_algorithm       = EPollardRho;
	this->sieving_interval_mode     = EFixed;
	this->prime_powers		= NULL;
	this->allocated_prime_powers    = 0;
	this->is_this_prime_divisor_of_a= NULL;
	this->divisors_of_a		= NULL;
	this->indices_of_divisors_of_a	= NULL;
	this->s				= DEFAULT_S;
	this->B2			= NULL;
	this->B2_inited			= NULL;
	this->Ba_inv			= NULL;
	this->nu			= 0;
	this->subtract			= 0;
	this->rough_factorbase_index = DEFAULT_ROUGH_INDEX;
	this->margin                 = DEFAULT_MARGIN;
	this->upper_bound_on_factor_base = DEFAULT_FB_BOUND;
	this->so_far_smooths_found= 0;
	this->so_far_1_partials_found = 0;
	this->so_far_2_partials_found = 0;
	this->so_far_x_partials_found = 0;
	this->number_of_polynomials   = DEFAULT_NR_POLYNOMIALS;
	this->allocated_polynomials   = 0;
	this->current_polynomial      = 0;	
	this->index_in_polynomial = -1; // there is NO polynomial yet
	this->time_message_interval   = DEFAULT_TIME_MESSAGE_INTERVAL;
	this->maximal_factor_base_index = -1;
	this->original_maximal_fb_index = -1;
	this->enriched_maximal_factor_base_index = -1;
	this->factor_base_allocated   = 0;
	this->intermediate_array      = NULL;
	this->advanced_factor_base    = NULL;
	this->ia_allocated            = 0;
	this->lanczos_time_message_mode = FALSE;
	this->without_primes_listing = TRUE;
	this->checking_mode        = ECheckNormal;
	this->printing_sort        = EDoNotPrintSort;
	this->SGTMode		   = EDeleteSingletons;
	this->open                  = 0;
	this->polynomials	     = NULL;
	this->factor_base            = NULL;	
	this->binary_logarithms_of_factor_base_times_log_factor = NULL;
	this->exact_binary_logarithms_of_factor_base = NULL;
	this->roots_of_kN_mod_p      = NULL;
	this->hash_root                 = NULL;
	this->partials_file              = NULL;
	this->smooth_values_file        = NULL;
	this->solve		= 1;
	this->auxCharForMpz = (char*) malloc(1000*sizeof(char));
	AllocateTexts();
	this->nx_env		= NULL;
	small_primes 		= NULL;
	this->main_matrix 	= NULL;
	this->smooth_values 	= NULL;
	this->class_result_bit    = NULL;
	this->class_transpose_bit = NULL;
	this->class_result_sparse = NULL;
	this->class_transpose_sparse = NULL;
	this->class_dependencies  = NULL;
	this->lanczos          = NULL;
	this->fact_alg         = NULL;
	this->a_inverses       = NULL;
	this->triples	       = NULL;
	this->aux_inited       = 0;
	this->cycle_table_1	= NULL;
	this->cycle_table_2	= NULL;
	this->components        = 0;
        this->edges             = 0;
	this->vertices          = 0;
	this->cycles            = 0;
	this->intermediate_roots = NULL;
	this->max_ir_index   = -1;
	this->root_list	      = NULL;
	this->rl_allocated    = 0;
	this->rl_max_index    = -1;
	this->matrix_type      = ESparseBitMatrix;
	if (!(this->aux_inited & AUX_1_INITED)){
		mpz_init(this->aux_1);
		this->aux_inited |= AUX_1_INITED;
	}
	if (!(this->aux_inited & AUX_2_INITED)){
		mpz_init(this->aux_2);
		this->aux_inited |= AUX_2_INITED;
	}
	if (!(this->aux_inited & AUX_3_INITED)){
		mpz_init(this->aux_3);
		this->aux_inited |= AUX_3_INITED;
	}
	if (!(this->aux_inited & AUX_4_INITED)){
		mpz_init(this->aux_4);
		this->aux_inited |= AUX_4_INITED;
	}
	if (!(this->aux_inited & AUX_5_INITED)){
		mpz_init(this->aux_5);
		this->aux_inited |= AUX_5_INITED;
	}
	if (!(this->aux_inited & AUX_A_INITED)){
		mpz_init(this->aux_a);
		this->aux_inited |= AUX_A_INITED;
	}
	if (!(this->aux_inited & AUX_B_INITED)){
		mpz_init(this->aux_b);
		this->aux_inited |= AUX_B_INITED;
	}
	if (!(this->aux_inited & AUX_C_INITED)){
		mpz_init(this->aux_c);
		this->aux_inited |= AUX_C_INITED;
	}
	if (!(this->aux_inited & AUX_D_INITED)){
		mpz_init(this->aux_d);
		this->aux_inited |= AUX_D_INITED;
	}
	if (!(this->aux_inited & AUX_P_INITED)){
		mpz_init(this->aux_p);
		this->aux_inited |= AUX_P_INITED;
	}
	if (!(this->aux_inited & AUX_P_EXP_INITED)){
		mpz_init(this->aux_p_exp);
		this->aux_inited |= AUX_P_EXP_INITED;
	}
	if (!(this->aux_inited & AUX_A_INV_INITED)){
		mpz_init(this->aux_a_inv);
		this->aux_inited |= AUX_A_INV_INITED;
	}
	if (!(this->aux_inited & AUX_K_INITED)){
		mpz_init(this->aux_k);
		this->aux_inited |= AUX_K_INITED;
	}
	if (!(this->aux_inited & AUX_L_INITED)){
		mpz_init(this->aux_l);
		this->aux_inited |= AUX_L_INITED;
	}
	this->st_aux_inited = 0;
	if (!(this->st_aux_inited & ST_AUX_A_INITED)){
		mpz_init(this->st_aux_a);
		this->st_aux_inited |= ST_AUX_A_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_A_INVERSE_INITED)){
		mpz_init(this->st_aux_a_inverse);
		this->st_aux_inited |= ST_AUX_A_INVERSE_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_B_INITED)){
		mpz_init(this->st_aux_b);
		this->st_aux_inited |= ST_AUX_B_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_R_INITED)){
		mpz_init(this->st_aux_r);
		this->st_aux_inited |= ST_AUX_R_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_N_INITED)){
		mpz_init(this->st_aux_n);
		this->st_aux_inited |= ST_AUX_N_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_P_INITED)){
		mpz_init(this->st_aux_p);
		this->st_aux_inited |= ST_AUX_P_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_1_INITED)){
		mpz_init(this->st_aux_1);
		this->st_aux_inited |= ST_AUX_1_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_ROOT_INITED)){
		mpz_init(this->st_aux_root);
		this->st_aux_inited |= ST_AUX_ROOT_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_P_MINUS_1_INITED)){
		mpz_init(this->st_aux_p_minus_1);
		this->st_aux_inited |= ST_AUX_P_MINUS_1_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_B_EXP_2I_INITED)){
		mpz_init(this->st_aux_b_exp_2i);
		this->st_aux_inited |= ST_AUX_B_EXP_2I_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_NON_INITED)){
		mpz_init(this->st_aux_non);
		this->st_aux_inited |= ST_AUX_NON_INITED;
	}
	this->wib_inited = 0;
	if (!(this->wib_inited & WIB_DIFF_INITED)){
		mpz_init(this->wib_diff);
		this->wib_inited |= WIB_DIFF_INITED;
	}
	if (!(this->wib_inited & WIB_RATIO_INITED)){
		mpz_init(this->wib_ratio);
		this->wib_inited |= WIB_RATIO_INITED;
	}
	if (!(this->wib_inited & WIB_MUL_P_INITED)){
		mpz_init(this->wib_mul_p);
		this->wib_inited |= WIB_MUL_P_INITED;
	}
	if (!(this->wib_inited & WIB_ARG_INITED)){
		mpz_init(this->wib_arg);
		this->wib_inited |= WIB_ARG_INITED;
	}
	if (!(this->wib_inited & WIB_ARG_2_INITED)){
		mpz_init(this->wib_arg_2);
		this->wib_inited |= WIB_ARG_2_INITED;
	}
	if (!(this->wib_inited & WIB_ARG_PLUS_INITED)){
		mpz_init(this->wib_arg_plus);
		this->wib_inited |= WIB_ARG_PLUS_INITED;
	}
	// Added in version 1.04
	this->int_sols_of_cur_poly.allocated = 0;
	this->int_sols_of_cur_poly.sol_1     = NULL;
	this->int_sols_of_cur_poly.sol_2     = NULL;
	this->int_sols_of_cur_poly.usable    = NULL;

	this->class_normal_result_bit   	= NULL;
	this->class_normal_transpose_bit	= NULL;
	this->class_normal_result_sparse	= NULL;
	this->class_normal_transpose_sparse	= NULL;

	this->total_sieve = 0;
	this->total_refill = 0;
	this->total_inner = 0;
	this->total_solv = 0;
	this->total_sort = 0;
	this->total_divisors_of_a = 0;
	this->total_double_factors = 0;
	this->total_cyclecount = 0;
	this->total_poly = 0;
	this->total_psbtp = 0;
	this->total_perform = 0;
	this->total_short_divisions = 0;
	this->total_hits = 0;
	this->total_divrem = 0;
	this->total_tested = 0;
	this->accepted_nr = 0;
	this->rejected_nr = 0;	
	this->iOptimizationPolicy=3;
	this->iOptimizer = NULL;
	int i;
	for (i=0; i<= MAX_DISTANCE; i++) {	
    		this->smooths_distribution[i] = 0;
		this->part_1_distribution[i] = 0;
		this->part_2_distribution[i] = 0;
    	}	
	this->iLineSiever = NULL;
	this->iLineSieveSize = ESize64k;
	this->iFirstLineSieverPolynomial = FALSE;
	this->iUseRootShort = FALSE;
	this->iDirectoryName = NULL;	
// The definition of this constant is in definitions.h	
#ifdef WITH_XML_SERIALIZATOR
	this->iSerializator = new XmlQsSerializator(this,".",0);
#else
	this->iSerializator = NULL;
#endif
	this->iLineNr = 0;
	this->iLoadFlags = 0;
	this->iIsInterruptable = FALSE;
	this->iInterruptNow = FALSE;
	this->iRelationProcessingPhase = EBeforeRelationCollection;
	this->iRelationProcessingFlags = 0;
	this->lastSaved = 0;

	this->iDistributionMachineType = ENone;
	
	
	this->iCommitInterval = IMPLICIT_COMMIT_INTERVAL;
	this->iMSDSpecifiedExternally = FALSE;
	this->iLowerMachineBound = 0;	
	this->iUpperMachineBound = 0;

	this->iLowerMachineBoundIndex = DEFAULT_MACHINE_BOUND_START;	
	this->iUpperMachineBoundIndex = DEFAULT_MACHINE_BOUND_END;
	
	this->machine_specific_divisors_number = 0;
	this->iMachineSpecificGeneration = MSGTwoDiffer;
	this->iNodeInstanceRunning = FALSE;

	iRelationsSinceLastDataTransmission = 0;
	iRelationsSent = 0;

	iDistributedMutexInited = FALSE;

	factoring_structures_inited = FALSE;

	iInitEnvironmentAlreadyRun = FALSE;
}

QuadraticSieve::QuadraticSieve(AbstractSieve* aSieve){
	if (!(this->inited & N_INITED)){
		mpz_init(this->N);
		this->inited |= N_INITED;
	}
	aSieve->Get_N(this->N);
	if (!(this->inited & kN_INITED)){
		mpz_init(this->kN);
		this->inited |= kN_INITED;
	}
	this->k 		  = aSieve->Get_k();
	mpz_mul_ui(this->kN,this->N,this->k);
	if (!(this->inited & F1_INITED)){
		mpz_init_set_ui(this->factor_1,INIT_VALUE_FACTOR);
		this->inited |= F1_INITED;
	}
	if (!(this->inited & F2_INITED)){
		mpz_init_set_ui(this->factor_2,INIT_VALUE_FACTOR);
		this->inited |= F2_INITED;
	}
	if (!(this->inited & M_INITED)){
		mpz_init_set_ui(this->M,DEFAULT_VALUE_M);
		this->inited |= M_INITED;
	}
	if (!(this->inited & ROOT_INITED)){
		mpz_init(this->root);
		this->inited |= ROOT_INITED;
	}
	if (!(this->inited & MPF_INITED)){
		mpz_init(this->maximal_partial_factor);
		this->inited |= MPF_INITED;
	}	
	if (!(this->inited & MIN_C_VALUE_INITED)) {
		mpz_init(this->minimal_candidate_value);
		this->inited |= MIN_C_VALUE_INITED;
	}
	if (!(this->inited & MAX_C_VALUE_INITED)) {
		mpz_init(this->maximal_candidate_value);
		this->inited |= MAX_C_VALUE_INITED;
	}
	if (!(this->inited & IDEAL_A_VALUE_INITED)){
		mpz_init(this->ideal_a_value);
		this->inited |= IDEAL_A_VALUE_INITED;
	}
	this->extK		  = FALSE;
	this->checklevel           = 0;
	this->div_allocate        = DEFAULT_DIV_ALLOCATE;
	this->info_mode 	   = aSieve->GetInfoMode();
	this->assertion_mode 	   = aSieve->GetAssertionMode();
	this->contini_threshold    = 0; // Log_2 (accepted).
	this->error_factor         = SIEVE_ROUNDOFF_ERROR;
	this->integer_M          = mpz_get_ui(this->M);	
	this->using_variations          = EDoNotUseVariations;
	this->QSMode		        = EMpqs;
	this->power_sieving_mode        = EDoNotSieveWithPrimePowers;
	this->factoring_algorithm       = EPollardRho;
	this->sieving_interval_mode     = EFixed;
	this->prime_powers		= NULL;
	this->allocated_prime_powers    = 0;
	this->is_this_prime_divisor_of_a= NULL;
	this->divisors_of_a		= NULL;
	this->indices_of_divisors_of_a	= NULL;
	this->s				= DEFAULT_S;
	this->B2			= NULL;
	this->B2_inited			= NULL;
	this->Ba_inv			= NULL;
	this->cycle_table_1		= NULL;
	this->cycle_table_2		= NULL;
	this->components        = 0;
        this->edges             = 0;
	this->vertices          = 0;
	this->cycles            = 0;
	this->nu			= 0;
	this->subtract			= 0;	
	this->rough_factorbase_index = DEFAULT_ROUGH_INDEX;
	this->margin                 = DEFAULT_MARGIN;
	this->upper_bound_on_factor_base = DEFAULT_FB_BOUND;
	this->div_allocate        = DEFAULT_DIV_ALLOCATE;
	this->so_far_smooths_found= 0;
	this->so_far_1_partials_found = 0;
	this->so_far_2_partials_found = 0;
	this->so_far_x_partials_found = 0;
	this->number_of_polynomials   = DEFAULT_NR_POLYNOMIALS;
	this->allocated_polynomials   = 0;
	this->index_in_polynomial = -1; // there is NO polynomial yet
	this->current_polynomial      = 0;
	this->time_message_interval   = DEFAULT_TIME_MESSAGE_INTERVAL;
	this->maximal_factor_base_index = -1;
	this->original_maximal_fb_index = -1;
	this->enriched_maximal_factor_base_index = -1;
	this->factor_base_allocated   = 0;
	this->intermediate_array      = NULL;
	this->ia_allocated            = 0;
	this->lanczos_time_message_mode = FALSE;
	this->without_primes_listing= TRUE;	
	this->density_info_mode    = FALSE;
	this->rank_calculation_mode= FALSE;
	this->value_distribution_mode = FALSE;
	this->checking_mode         = ECheckNormal;
	this->printing_sort        = EDoNotPrintSort;	
	this->SGTMode		   = EDeleteSingletons;
	this->open                  = 0;
	this->polynomials	     = NULL;
	this->factor_base            = NULL;	
	this->binary_logarithms_of_factor_base_times_log_factor = NULL;
	this->exact_binary_logarithms_of_factor_base = NULL;	
	this->roots_of_kN_mod_p      = NULL;
	this->hash_root                 = NULL;
	this->partials_file              = NULL;
	this->smooth_values_file        = NULL;
	this->a_inverses       = NULL;
	this->triples	= NULL;
	this->root_list	      = NULL;
	this->rl_allocated    = 0;
	this->rl_max_index    = -1;
	this->auxCharForMpz = (char*) malloc(1000*sizeof(char));
	AllocateTexts();
	this->nx_env		= NULL;
	small_primes = NULL;
	this->main_matrix 	= NULL;
	this->smooth_values 	= NULL;
	this->class_result_bit    = NULL;
	this->class_transpose_bit = NULL;
	this->class_result_sparse = NULL;
	this->class_transpose_sparse = NULL;
	this->class_dependencies     = NULL;
	this->lanczos          = NULL;
	this->fact_alg         = NULL;
	this->aux_inited        = 0;
	this->matrix_type      = ESparseBitMatrix;
	this->advanced_factor_base    = NULL;
	if (!(this->aux_inited & AUX_1_INITED)){
		mpz_init(this->aux_1);
		this->aux_inited |= AUX_1_INITED;
	}
	if (!(this->aux_inited & AUX_2_INITED)){
		mpz_init(this->aux_2);
		this->aux_inited |= AUX_2_INITED;
	}
	if (!(this->aux_inited & AUX_3_INITED)){
		mpz_init(this->aux_3);
		this->aux_inited |= AUX_3_INITED;
	}
	if (!(this->aux_inited & AUX_4_INITED)){
		mpz_init(this->aux_4);
		this->aux_inited |= AUX_4_INITED;
	}
	if (!(this->aux_inited & AUX_5_INITED)){
		mpz_init(this->aux_5);
		this->aux_inited |= AUX_5_INITED;
	}
	if (!(this->aux_inited & AUX_A_INITED)){
		mpz_init(this->aux_a);
		this->aux_inited |= AUX_A_INITED;
	}
	if (!(this->aux_inited & AUX_B_INITED)){
		mpz_init(this->aux_b);
		this->aux_inited |= AUX_B_INITED;
	}
	if (!(this->aux_inited & AUX_C_INITED)){
		mpz_init(this->aux_c);
		this->aux_inited |= AUX_C_INITED;
	}
	if (!(this->aux_inited & AUX_D_INITED)){
		mpz_init(this->aux_d);
		this->aux_inited |= AUX_D_INITED;
	}
	if (!(this->aux_inited & AUX_P_INITED)){
		mpz_init(this->aux_p);
		this->aux_inited |= AUX_P_INITED;
	}
	if (!(this->aux_inited & AUX_P_EXP_INITED)){
		mpz_init(this->aux_p_exp);
		this->aux_inited |= AUX_P_EXP_INITED;
	}
	if (!(this->aux_inited & AUX_A_INV_INITED)){
		mpz_init(this->aux_a_inv);
		this->aux_inited |= AUX_A_INV_INITED;
	}
	if (!(this->aux_inited & AUX_K_INITED)){
		mpz_init(this->aux_k);
		this->aux_inited |= AUX_K_INITED;
	}
	if (!(this->aux_inited & AUX_L_INITED)){
		mpz_init(this->aux_l);
		this->aux_inited |= AUX_L_INITED;
	}
	this->st_aux_inited = 0;
	if (!(this->st_aux_inited & ST_AUX_A_INITED)){
		mpz_init(this->st_aux_a);
		this->st_aux_inited |= ST_AUX_A_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_A_INVERSE_INITED)){
		mpz_init(this->st_aux_a_inverse);
		this->st_aux_inited |= ST_AUX_A_INVERSE_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_B_INITED)){
		mpz_init(this->st_aux_b);
		this->st_aux_inited |= ST_AUX_B_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_R_INITED)){
		mpz_init(this->st_aux_r);
		this->st_aux_inited |= ST_AUX_R_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_N_INITED)){
		mpz_init(this->st_aux_n);
		this->st_aux_inited |= ST_AUX_N_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_P_INITED)){
		mpz_init(this->st_aux_p);
		this->st_aux_inited |= ST_AUX_P_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_1_INITED)){
		mpz_init(this->st_aux_1);
		this->st_aux_inited |= ST_AUX_1_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_ROOT_INITED)){
		mpz_init(this->st_aux_root);
		this->st_aux_inited |= ST_AUX_ROOT_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_P_MINUS_1_INITED)){
		mpz_init(this->st_aux_p_minus_1);
		this->st_aux_inited |= ST_AUX_P_MINUS_1_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_B_EXP_2I_INITED)){
		mpz_init(this->st_aux_b_exp_2i);
		this->st_aux_inited |= ST_AUX_B_EXP_2I_INITED;
	}
	if (!(this->st_aux_inited & ST_AUX_NON_INITED)){
		mpz_init(this->st_aux_non);
		this->st_aux_inited |= ST_AUX_NON_INITED;
	}
	this->wib_inited = 0;
	if (!(this->wib_inited & WIB_DIFF_INITED)){
		mpz_init(this->wib_diff);
		this->wib_inited |= WIB_DIFF_INITED;
	}
	if (!(this->wib_inited & WIB_RATIO_INITED)){
		mpz_init(this->wib_ratio);
		this->wib_inited |= WIB_RATIO_INITED;
	}
	if (!(this->wib_inited & WIB_MUL_P_INITED)){
		mpz_init(this->wib_mul_p);
		this->wib_inited |= WIB_MUL_P_INITED;
	}
	if (!(this->wib_inited & WIB_ARG_INITED)){
		mpz_init(this->wib_arg);
		this->wib_inited |= WIB_ARG_INITED;
	}
	if (!(this->wib_inited & WIB_ARG_2_INITED)){
		mpz_init(this->wib_arg_2);
		this->wib_inited |= WIB_ARG_2_INITED;
	}
	if (!(this->wib_inited & WIB_ARG_PLUS_INITED)){
		mpz_init(this->wib_arg_plus);
		this->wib_inited |= WIB_ARG_PLUS_INITED;
	}
	// Added in version 1.04
	this->int_sols_of_cur_poly.allocated = 0;
	this->int_sols_of_cur_poly.sol_1     = NULL;
	this->int_sols_of_cur_poly.sol_2     = NULL;
	this->int_sols_of_cur_poly.usable    = NULL;

	this->class_normal_result_bit   	= NULL;
	this->class_normal_transpose_bit	= NULL;
	this->class_normal_result_sparse	= NULL;
	this->class_normal_transpose_sparse	= NULL;
	this->solve		= 1;
	this->total_sieve = 0;
	this->total_refill = 0;
	this->total_inner = 0;
	this->total_solv = 0;
	this->total_sort = 0;
	this->total_divisors_of_a = 0;
	this->total_double_factors = 0;
	this->total_cyclecount = 0;
	this->total_poly = 0;
	this->total_psbtp = 0;
	this->total_perform = 0;
	this->total_short_divisions = 0;
	this->total_hits = 0;
	this->total_divrem = 0;
	this->total_tested = 0;
	this->accepted_nr = 0;
	this->rejected_nr = 0;	
	int i;
	for (i=0; i<= MAX_DISTANCE; i++) {	
    		this->smooths_distribution[i] = 0;
		this->part_1_distribution[i] = 0;
		this->part_2_distribution[i] = 0;
    	}	
	this->iLineSiever = NULL;
	this->iLineSieveSize = ESize64k;
	this->iFirstLineSieverPolynomial = FALSE;
	this->iUseRootShort = FALSE;
	this->iDirectoryName = NULL;
// The definition of this constant is in definitions.h
#ifdef WITH_XML_SERIALIZATOR
	this->iSerializator = new XmlQsSerializator(this,".",0);
#else
	this->iSerializator = NULL;
#endif
	this->iLineNr = 0;
	this->iLoadFlags = 0;
	this->iIsInterruptable = FALSE;
	this->iInterruptNow = FALSE;
	this->iRelationProcessingPhase = EBeforeRelationCollection;
	this->iRelationProcessingFlags = 0;
	this->lastSaved = 0;
	this->iLowerMachineBound = 0;	
	this->iUpperMachineBound = 0;

	this->iLowerMachineBoundIndex = DEFAULT_MACHINE_BOUND_START;	
	this->iUpperMachineBoundIndex = DEFAULT_MACHINE_BOUND_END;
	this->machine_specific_divisors_number = 0;
	this->iNodeInstanceRunning = FALSE;

	iRelationsSinceLastDataTransmission = 0;
	iRelationsSent = 0;

	iDistributedMutexInited = FALSE;

	factoring_structures_inited = FALSE;

	iInitEnvironmentAlreadyRun = FALSE;
}

int QuadraticSieve::InitFileNames()
{
	this->temporary_file_name = iSerializator->CreateFileNamePlain(DEFAULT_TEMPORARY_FILE_NAME);
	

	this->partial_values_file_name = iSerializator->CreateFileNamePlain(DEFAULT_PARTIAL_FILE_NAME);
	

	this->smooth_values_file_name = iSerializator->CreateFileNamePlain(DEFAULT_SMOOTH_FILE_NAME);
	

	this->prime_numbers_file_name.assign(DEFAULT_PRIMEFILE_NAME);

	this->all_partial_values_file_name = iSerializator->CreateFileNamePlain(DEFAULT_ALL_PARTIALS_FILE_NAME);	

	this->usable_partial_values_file_name = iSerializator->CreateFileNamePlain(DEFAULT_USABLE_PARTIALS_FILE_NAME);
	
	return ConstRC::Ok;
}

void QuadraticSieve::AllocateTexts()
{
	iGenerateNewPolyText.assign(     GEN_NEW_POLY_TEXT);
	iQuadraticMachineText.assign(    QUAD_MACHINE_TEXT);
	iPerformSievingByPolyText.assign(PERF_SIEVING_TEXT);
	iSievingLoopText.assign(         SIEVING_LOOP_TEXT);
	iRefillMainMatrixText.assign(    REFILL_MAINM_TEXT);
	iSortSmoothText.assign(          SORT_SMOOTHS_TEXT);
	iDivisorsOfAText.assign(         DIVISORS_OFA_TEXT);
	iSmallFactorText.assign(         SMALL_FACTOR_TEXT);
	iCountCyclesText.assign(	 COUNT_CYCLES_TEXT);
	iBlockSieveText.assign(          BLOCK_SIEVES_TEXT);
	iFillHashtableText.assign(       FILL_HSTABLE_TEXT);
	iFillHashtableNegText.assign(    FILL_HST_NEG_TEXT);
	iRootsUpdateText.assign(         ROOTS_UPDATE_TEXT);
	iFindAndFactorText.assign(       FIND_FACTORZ_TEXT);
	iUnknownPartText.assign(         UNKNOWN_PART_TEXT);
}

void QuadraticSieve::AllocPolynomial(quad_polynomial& poly)
{
	mpz_init(poly.a);
	mpz_init(poly.b);
	mpz_init(poly.c);
	mpz_init(poly.d);
	mpz_init(poly.d2);
	poly.inited = (A_INITED | B_INITED | C_INITED | D_INITED | D2_INITED);
}

void QuadraticSieve::ClearPolynomial(quad_polynomial& poly)
{
	if (poly.inited & A_INITED){
		mpz_clear(poly.a);
		poly.inited &= ~(A_INITED);
	}
	if (poly.inited & B_INITED){
		mpz_clear(poly.b);
		poly.inited &= ~(B_INITED);
	}
	if (poly.inited & C_INITED){
		mpz_clear(poly.c);
		poly.inited &= ~(C_INITED);
	}
	if (poly.inited & D_INITED){
		mpz_clear(poly.d);
		poly.inited &= ~(D_INITED);
	}
	if (poly.inited & D2_INITED){
		mpz_clear(poly.d2);
		poly.inited &= ~(D2_INITED);
	}
}

QuadraticSieve::~QuadraticSieve(){
	long i,j;
//	printf("QS1\n\n");
	if (this->open & PARTIALS_OPEN){
		if ((i=fclose(this->partials_file)) != ConstRC::Ok){
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file partials (in destructor)",i,1,"Could not close file partials.");
		} else {
			this->open &= ~(PARTIALS_OPEN);
		}
	}
//	printf("QS2\n\n");
	if (this->open & SMOOTHS_OPEN){
		if ((j=fclose(this->smooth_values_file)) != ConstRC::Ok){
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file smooths (in destructor)",i,2,"Could not close file smooths.");
		} else {
			this->open &= ~(SMOOTHS_OPEN);
		}
	}
//	printf("QS3\n\n");
	if (this->allocated_polynomials){ // there were polynomials allocated
		quad_polynomial poly;
		for (i=0; i<allocated_polynomials; i++){			
			ClearPolynomial(this->polynomials[i]);
		} // end of for
		free(this->polynomials);
		this->polynomials = NULL;
		this->allocated_polynomials = 0;
		this->index_in_polynomial = -1;
	} // end of if
	if (this->hash_root){
		DeleteHashes(this->hash_root);
	}
	// now, we clear all mpz_t values
	if (this->inited & N_INITED){
		mpz_clear(this->N);
		this->inited &= (~N_INITED);
	}
	if (this->inited & kN_INITED){
		mpz_clear(this->kN);
		this->inited &= (~kN_INITED);
	}
	if (this->inited & F1_INITED){
		mpz_clear(this->factor_1);
		this->inited &= (~F1_INITED);
	}
	if (this->inited & F2_INITED){
		mpz_clear(this->factor_2);
		this->inited &= (~F2_INITED);
	}
	if (this->inited & M_INITED){
		mpz_clear(this->M);
		this->inited &= (~M_INITED);
	}
	if (this->inited & ROOT_INITED){
		mpz_clear(this->root);
		this->inited &= (~ROOT_INITED);
	}
	if (this->inited & MPF_INITED){
		mpz_clear(this->maximal_partial_factor);
		this->inited &= (~MPF_INITED);
	}
	if (this->inited & MIN_C_VALUE_INITED) {
		mpz_clear(this->minimal_candidate_value);
		this->inited &= (~MIN_C_VALUE_INITED);
	}
	if (this->inited & MAX_C_VALUE_INITED) {
		mpz_clear(this->maximal_candidate_value);
		this->inited &= (~MAX_C_VALUE_INITED);
	}
	if (this->inited & IDEAL_A_VALUE_INITED){
		mpz_clear(this->ideal_a_value);
		this->inited &= (~IDEAL_A_VALUE_INITED);
	}
	if (this->factor_base != NULL) {
		free(this->factor_base);
		this->factor_base = NULL;
	}
	
	if (this->roots_of_kN_mod_p != NULL) {
		free(this->roots_of_kN_mod_p);
		this->roots_of_kN_mod_p = NULL;
	}

	if (this->binary_logarithms_of_factor_base_times_log_factor != NULL){
		free(this->binary_logarithms_of_factor_base_times_log_factor);
		this->binary_logarithms_of_factor_base_times_log_factor = NULL;
	}
	
//	printf("QS4a\n\n");
	if (small_primes){
		free(small_primes);
	}
	if (polynomials){
		for (i=0; i< this->number_of_polynomials; i++){
			ClearPolynomial(this->polynomials[i]);
		}
		// cyclic removal of polys
		free(polynomials);
	}
	if (main_matrix){
		if (this->main_matrix->	accumulated_logarithm){
			free(this->main_matrix->	accumulated_logarithm);
			this->main_matrix->	accumulated_logarithm = NULL;
		}
		if (main_matrix->assigned) {;
			main_matrix->assigned = 0;
		}
		free(main_matrix);
		main_matrix = NULL;
	}
//	printf("QS4b\n\n");
	if (this->a_inverses){
		free(this->a_inverses);
		a_inverses = NULL;
	}
	if (this->is_this_prime_divisor_of_a){
		free(this->is_this_prime_divisor_of_a);
		this->is_this_prime_divisor_of_a = NULL;
	}
	if (this->divisors_of_a){
		free(this->divisors_of_a);
		this->divisors_of_a = NULL;
	}
	if (this->indices_of_divisors_of_a){
		free(this->indices_of_divisors_of_a);
		this->indices_of_divisors_of_a	= NULL;
	}
//	printf("QS4c\n\n");
	if (smooth_values){
		for (i=0; i< this->smooth_values->assigned_rows; i++){
//			printf("Clearing: %d.\n\n",i);
			ClearRelation(&(this->smooth_values->matrix[i]));			
		} // end of for (row deletion)
		if (smooth_values->assigned_rows) {
//			printf("Clearing smooth values matrix.\n\n");
			free(smooth_values->matrix);
			smooth_values->matrix = NULL;
			smooth_values->assigned_rows = 0;
		}
//		printf("Clearing smooth values.\n\n");
		free(smooth_values);
		smooth_values = NULL;
	}
//	printf("QS4c1\n\n");

//	printf("QS4d\n\n");
	if (this->class_result_bit){
		delete this->class_result_bit;
		this->class_result_bit = NULL;
	}
	if (this->class_transpose_bit){
		delete this->class_transpose_bit;
		this->class_transpose_bit = NULL;
	}
	if (this->class_result_sparse){
		delete this->class_result_sparse;
		this->class_result_sparse = NULL;
	}
	if (this->class_transpose_sparse){
		delete this->class_transpose_sparse;
		this->class_transpose_sparse = NULL;
	}
	if (this->class_dependencies){
		delete this->class_dependencies;
		this->class_dependencies = NULL;
	}
//	printf("QS4e\n\n");
	if (this->lanczos){
		delete this->lanczos;
		this->lanczos = NULL;
	}
	if (this->aux_inited & AUX_1_INITED){
		mpz_clear(this->aux_1);
		this->aux_inited &= ~(AUX_1_INITED);
	}
	if (this->aux_inited & AUX_2_INITED){
		mpz_clear(this->aux_2);
		this->aux_inited &= ~(AUX_2_INITED);
	}
	if (this->aux_inited & AUX_3_INITED){
		mpz_clear(this->aux_3);
		this->aux_inited &= ~(AUX_3_INITED);
	}
	if (this->aux_inited & AUX_4_INITED){
		mpz_clear(this->aux_4);
		this->aux_inited &= ~(AUX_4_INITED);
	}
	if (this->aux_inited & AUX_5_INITED){
		mpz_clear(this->aux_5);
		this->aux_inited &= ~(AUX_5_INITED);
	}
	if (this->aux_inited & AUX_A_INITED){
		mpz_clear(this->aux_a);
		this->aux_inited &= ~(AUX_A_INITED);
	}
	if (this->aux_inited & AUX_B_INITED){
		mpz_clear(this->aux_b);
		this->aux_inited &= ~(AUX_B_INITED);
	}
	if (this->aux_inited & AUX_C_INITED){
		mpz_clear(this->aux_c);
		this->aux_inited &= ~(AUX_C_INITED);
	}
	if (this->aux_inited & AUX_D_INITED){
		mpz_clear(this->aux_d);
		this->aux_inited &= ~(AUX_D_INITED);
	}
	if (this->aux_inited & AUX_P_INITED){
		mpz_clear(this->aux_p);
		this->aux_inited &= ~(AUX_P_INITED);
	}
	if (this->aux_inited & AUX_P_EXP_INITED){
		mpz_clear(this->aux_p_exp);
		this->aux_inited &= ~(AUX_P_EXP_INITED);
	}
	if (this->aux_inited & AUX_A_INV_INITED){
		mpz_clear(this->aux_a_inv);
		this->aux_inited &= ~(AUX_A_INV_INITED);
	}
	if (this->aux_inited & AUX_K_INITED){
		mpz_clear(this->aux_k);
		this->aux_inited &= ~(AUX_K_INITED);
	}
	if (this->aux_inited & AUX_L_INITED){
		mpz_clear(this->aux_l);
		this->aux_inited &= ~(AUX_L_INITED);
	}
	if (this->aux_inited) {
		CLog::LogError(WHEREARG,"Quadratic sieve destructor","something aux was not deleted",25,this->aux_inited,"Check destructors and table of symbolic constants for this value.");
	}
	if (this->st_aux_inited & ST_AUX_A_INITED){
		mpz_clear(this->st_aux_a);
		this->st_aux_inited &= ~ST_AUX_A_INITED;
	}
	if (this->st_aux_inited & ST_AUX_A_INVERSE_INITED){
		mpz_clear(this->st_aux_a_inverse);
		this->st_aux_inited &= ~ST_AUX_A_INVERSE_INITED;
	}
	if (this->st_aux_inited & ST_AUX_B_INITED){
		mpz_clear(this->st_aux_b);
		this->st_aux_inited &= ~ST_AUX_B_INITED;
	}
	if (this->st_aux_inited & ST_AUX_R_INITED){
		mpz_clear(this->st_aux_r);
		this->st_aux_inited &= ~ST_AUX_R_INITED;
	}
	if (this->st_aux_inited & ST_AUX_N_INITED){
		mpz_clear(this->st_aux_n);
		this->st_aux_inited &= ~ST_AUX_N_INITED;
	}
	if (this->st_aux_inited & ST_AUX_P_INITED){
		mpz_clear(this->st_aux_p);
		this->st_aux_inited &= ~ST_AUX_P_INITED;
	}
	if (this->st_aux_inited & ST_AUX_1_INITED){
		mpz_clear(this->st_aux_1);
		this->st_aux_inited &= ~ST_AUX_1_INITED;
	}
	if (this->st_aux_inited & ST_AUX_ROOT_INITED){
		mpz_clear(this->st_aux_root);
		this->st_aux_inited &= ~ST_AUX_ROOT_INITED;
	}
	if (this->st_aux_inited & ST_AUX_P_MINUS_1_INITED){
		mpz_clear(this->st_aux_p_minus_1);
		this->st_aux_inited &= ~ST_AUX_P_MINUS_1_INITED;
	}
	if (this->st_aux_inited & ST_AUX_B_EXP_2I_INITED){
		mpz_clear(this->st_aux_b_exp_2i);
		this->st_aux_inited &= ~ST_AUX_B_EXP_2I_INITED;
	}
	if (this->st_aux_inited & ST_AUX_NON_INITED){
		mpz_clear(this->st_aux_non);
		this->st_aux_inited &= ~ST_AUX_NON_INITED;
	}
	if (this->st_aux_inited) {
		CLog::LogError(WHEREARG,"Quadratic sieve destructor","something st_aux was not deleted",26,this->st_aux_inited,"Check destructors and table of symbolic constants for this value.");
	}
	if (this->wib_inited & WIB_DIFF_INITED){
		mpz_clear(this->wib_diff);
		this->wib_inited &= ~WIB_DIFF_INITED;
	}
	if (this->wib_inited & WIB_RATIO_INITED){
		mpz_clear(this->wib_ratio);
		this->wib_inited &= ~WIB_RATIO_INITED;
	}
	if (this->wib_inited & WIB_MUL_P_INITED){
		mpz_clear(this->wib_mul_p);
		this->wib_inited &= ~WIB_MUL_P_INITED;
	}
	if (this->wib_inited & WIB_ARG_INITED){
		mpz_clear(this->wib_arg);
		this->wib_inited &= ~WIB_ARG_INITED;
	}
	if (this->wib_inited & WIB_ARG_2_INITED){
		mpz_clear(this->wib_arg_2);
		this->wib_inited &= ~WIB_ARG_2_INITED;
	}
	if (this->wib_inited & WIB_ARG_PLUS_INITED){
		mpz_clear(this->wib_arg_plus);
		this->wib_inited &= ~WIB_ARG_PLUS_INITED;
	}
//	printf("QS5\n\n");
	// new in version 1.04
	Clear_Solutions();
	// prime powers structures (since 1.15)
	if (this->prime_powers){
		for (i=0; i< this->allocated_prime_powers; i++){
			if ((this->prime_powers[i]).powers){
				free((this->prime_powers[i]).powers);
				this->prime_powers[i].powers = NULL;
			}
			this->prime_powers[i].allocated_powers = -1;
			if ((this->prime_powers[i]).roots_of_kN_mod_powers){
				free((this->prime_powers[i]).roots_of_kN_mod_powers);
				this->prime_powers[i].roots_of_kN_mod_powers = NULL;
			}
			if ((this->prime_powers[i]).solutions_1){
				free((this->prime_powers[i]).solutions_1);
				this->prime_powers[i].solutions_1 = NULL;
			}
			if ((this->prime_powers[i]).a_inverses){
				free((this->prime_powers[i]).a_inverses);
				this->prime_powers[i].a_inverses = NULL;
			}
			if ((this->prime_powers[i]).Ba_inv){
				free((this->prime_powers[i]).Ba_inv);
				this->prime_powers[i].Ba_inv = NULL;
			}
			if ((this->prime_powers[i]).solutions_2){
				free((this->prime_powers[i]).solutions_2);
				this->prime_powers[i].solutions_2 = NULL;
			}
			if ((this->prime_powers[i]).usable){
				free((this->prime_powers[i]).usable);
				this->prime_powers[i].usable = NULL;
			}	
		}
		this->prime_powers = NULL;
		this->allocated_prime_powers = 0;
	}
	// siqs structures (since 1.09)
	if (this->B2_inited != NULL){
		for (i=0; i<s; i++){
			if (this->B2_inited[i]){
				mpz_clear(B2[i]);
				this->B2_inited[i] = 0;
			}
		}
		free(this->B2);		
		free(this->B2_inited);
		this->B2 = NULL;
		this->B2_inited = NULL;
	}
	FreeBa_inv();
	if (this->nx_env){
		if (this->nx_env->subset){
			free(this->nx_env->subset);
			this->nx_env->subset = NULL;
		}
		free(this->nx_env);
		this->nx_env = NULL;
	}
	if (this->exact_binary_logarithms_of_factor_base){
		free(this->exact_binary_logarithms_of_factor_base);
		this->exact_binary_logarithms_of_factor_base = NULL;
	}
//	printf("QS6\n\n");
	if (this->triples){
		for (i=0; i< this->triple_count; i++){
			if (this->triples[i]){
				free(this->triples[i]);
				this->triples[i] = NULL;
			}
		}
		free(this->triples);
		this->triples = NULL;
	}
	if (this->cycle_table_1) {
		delete this->cycle_table_1;
		this->cycle_table_1 = NULL;
	}	
	if (this->cycle_table_2) {
		delete this->cycle_table_2;
		this->cycle_table_2 = NULL;
	}	
	if (this->intermediate_roots) {
		free(this->intermediate_roots);
		this->intermediate_roots = NULL;
		this->max_ir_index = -1;
	}	
	if (this->root_list) {
		free(this->root_list);
		this->root_list = NULL;
		this->rl_allocated    = 0;
		this->rl_max_index    = -1;
	}
	if (this->intermediate_array) {
		free(this->intermediate_array);
		this->intermediate_array = NULL;
		this->ia_allocated    = 0;
	}
	if (this->auxCharForMpz) {
		free(this->auxCharForMpz);
	}
	delete this->class_normal_result_bit;
	delete this->class_normal_transpose_bit;
	delete this->class_normal_result_sparse;
	delete this->class_normal_transpose_sparse;	
	delete this->iLineSiever;
	if (this->advanced_factor_base)
	{
		if (this->QSMode == ESiqs)
		{
			qs_fb_type* running = this->advanced_factor_base;
			while (running->p)
			{
				if (running->ba_inv != NULL)
				{
					free(running->ba_inv);
					running->ba_inv = NULL;
				}
				running++;
			}
		}
		free(this->advanced_factor_base);
		this->advanced_factor_base = NULL;
	}

	if (this->fact_alg)
	{
		delete this->fact_alg;
		this->fact_alg = NULL;
	}

	if (this->iDirectoryName)
	{
		free(this->iDirectoryName);
		iDirectoryName = NULL;
	}	

	if (iDistributedMutexInited) {
		pthread_mutex_destroy(&iDistributedMutex);
		iDistributedMutexInited = FALSE;
	};

	factoring_structures_inited = FALSE;

	if (iOptimizer)
	{
		delete iOptimizer;
		iOptimizer = NULL;
	}
}

void QuadraticSieve::DestroyPolynomial(quad_polynomial* aPoly)
{
	if (aPoly->inited & A_INITED){
		mpz_clear(aPoly->a);
		aPoly->inited &= ~(A_INITED);
	}
	if (aPoly->inited & B_INITED){
		mpz_clear(aPoly->b);
		aPoly->inited &= ~(B_INITED);
	}
	if (aPoly->inited & C_INITED){
		mpz_clear(aPoly->c);
		aPoly->inited &= ~(C_INITED);
	}
	if (aPoly->inited & D_INITED){
		mpz_clear(aPoly->d);
		aPoly->inited &= ~(D_INITED);
	}
	if (aPoly->inited & D2_INITED){
		mpz_clear(aPoly->d2);
		aPoly->inited &= ~(D2_INITED);
	}
}

void QuadraticSieve::FreeBa_inv()
{
	int i;
	if (this->Ba_inv != NULL){			
		for (i=0; i<=this->original_maximal_fb_index; i++){			
			if (this->Ba_inv[i] != NULL){			
				free(this->Ba_inv[i]);
				this->Ba_inv[i] = NULL;
			}
		}
		free(this->Ba_inv);
		this->Ba_inv = NULL;
	} 
}

int QuadraticSieve::AllocateBa_inv()
{
	int i;
	this->Ba_inv = (main_sieving_type**) AbstractSieve::AllocationMachine((void**) &(this->Ba_inv),(this->maximal_factor_base_index+1)*sizeof(main_sieving_type*));
	if (this->Ba_inv == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating Ba_inv",s,72,"There was an error allocating array of Ba_inv inited's");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->Ba_inv,0,(this->maximal_factor_base_index+1)*sizeof(main_sieving_type*));
	for (i=0; i<= this->maximal_factor_base_index; i++){
		this->Ba_inv[i] = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->Ba_inv[i],(this->s)*sizeof(main_sieving_type));
		if (this->Ba_inv[i] == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating Ba_inv",s,i,"There was an error allocating array of Ba_inv inited's");
			return NOT_ENOUGH_MEMORY;
		}
		memset(this->Ba_inv[i],0,(this->s)*sizeof(main_sieving_type));
	}
	return ConstRC::Ok;
}


// GETTERS AND SETTERS
void QuadraticSieve::Set_UsingVariations(variations_types aValue) {
	
	this->using_variations = aValue;
}

void QuadraticSieve::Set_Nu(int aValue)
{
	this->nu = aValue;
}

void QuadraticSieve::Set_Subtract(int aValue)
{
	this->subtract = aValue;
}

void QuadraticSieve::Set_DirectoryName(const char* aValue)
{
	int len = strlen(aValue)+2*sizeof(char);
	if (!iDirectoryName)
	{			
		iDirectoryName = (char*) malloc(len);
	} else if ((strlen(iDirectoryName)+2) < strlen(aValue))
	{
		iDirectoryName = (char*) realloc((void*) iDirectoryName,len);
	}	
	strcpy(iDirectoryName,aValue);
}

void QuadraticSieve::Set_ExtK(bool aValue) {
	this->extK = aValue;
}

void QuadraticSieve::Set_PrintingSort(print_sort aValue){
	this->printing_sort = aValue;
}

void QuadraticSieve::Set_FactoringAlgorithm(factor_alg aValue){
	this->factoring_algorithm = aValue;
}

void QuadraticSieve::Set_SievingIntervalMode(siev_i_mode aValue) {
	this->sieving_interval_mode = aValue;
}

void QuadraticSieve::Set_QSMode(qs_mode aValue){
	this->QSMode = aValue;
}

void QuadraticSieve::Set_S(int aValue){
	this->s = aValue;
}

void QuadraticSieve::Set_UseRootShort(bool aValue)
{
	this->iUseRootShort = aValue;
}

void QuadraticSieve::Set_M(mpz_t aValue){
	if (!(this->inited & M_INITED)){
		mpz_init(this->M);
		this->inited |= M_INITED;
	}
	mpz_set(this->M,aValue);
	this->integer_M = mpz_get_ui(aValue);
};

void QuadraticSieve::Set_Root(mpz_t aValue){
	if (!(this->inited & ROOT_INITED)){
		mpz_init(this->root);
		this->inited |= ROOT_INITED;
	}
	mpz_set(this->root,aValue);
};

void QuadraticSieve::Set_MPF(mpz_t aValue){
	if (!(this->inited & MPF_INITED)){
		mpz_init(this->maximal_partial_factor);
		this->inited |= MPF_INITED;
	}
	mpz_set(this->maximal_partial_factor,aValue);
};

void QuadraticSieve::Set_LTMode(bool aValue) {
	this->lanczos_time_message_mode = aValue;
}

void QuadraticSieve::Set_WPList(bool aValue){
	this->without_primes_listing = aValue;
}


void QuadraticSieve::Set_NrPolynomials(long aValue){
	this->number_of_polynomials = aValue;
}

void QuadraticSieve::Set_UpperFBBound(long aValue){
	this->upper_bound_on_factor_base = aValue;
}

void QuadraticSieve::Set_DIMode(bool aValue){
	this->density_info_mode = aValue;
}

void QuadraticSieve::Set_CheckMode(check_mode aMode){
	this->checking_mode = aMode;
}

void QuadraticSieve::Set_CheckLevel(int aValue){
	this->error_factor += CHECK_LEVEL_STEP*aValue;
	this->checklevel           = aValue;
}

void QuadraticSieve::Set_TimeMessageInterval(int aValue){
	this->time_message_interval = aValue;
}

void QuadraticSieve::Set_SingletonMode(sgt_mode aValue){
	this->SGTMode = aValue;
}

void QuadraticSieve::Set_PowerMode(power_mode aValue){
	this->power_sieving_mode = aValue;
}

void QuadraticSieve::Set_ValDistMode(bool aValue) {
	this->value_distribution_mode = aValue;
}

void QuadraticSieve::Set_MemblockSize(int aValue)
{
	switch (aValue)
	{
		case 16: 
			this->iLineSieveSize = ESize16k;
			break;
		case 32:
			this->iLineSieveSize = ESize32k;
			break;
		case 48:
			this->iLineSieveSize = ESize48k;
			break;
		case 64:
			this->iLineSieveSize = ESize64k;
			break;
		case 96:
			this->iLineSieveSize = ESize96k;
			break;
		case 128:
			this->iLineSieveSize = ESize128k;
			break;
		case 160:
			this->iLineSieveSize = ESize160k;
			break;
		case 192:
			this->iLineSieveSize = ESize192k;
			break;
		case 256:
			this->iLineSieveSize = ESize256k;
			break;
		case 384:
			this->iLineSieveSize = ESize384k;
			break;
		case 512:
			this->iLineSieveSize = ESize512k;
			break;
		default: 
			this->iLineSieveSize = ESize64k;
			break;
	}
}

void QuadraticSieve::Set_MemblockSizeLSS(line_sieve_size aValue)
{
	this->iLineSieveSize = aValue;
}

void	QuadraticSieve::Set_RelationProcessingFlag(int aFlag)
{
	this->iRelationProcessingFlags |= aFlag;
}

void	QuadraticSieve::Clear_RelationProcessingFlag(int aFlag)
{
	this->iRelationProcessingFlags &= ~(aFlag);
}


line_sieve_size QuadraticSieve::Get_MemblockSize() const
{
	return this->iLineSieveSize;
}

int QuadraticSieve::Get_CheckLevel() const
{
	return this->checklevel;
}

quad_polynomial* QuadraticSieve::Get_Polynomials(long& aNumberOfPolynomials, long& aAllocatedPolynomials, long& aIndexInPolynomial, long& aCurrentPolynomial) const
{
	aNumberOfPolynomials = this->number_of_polynomials;
	aAllocatedPolynomials = this->allocated_polynomials;
	aIndexInPolynomial = this->index_in_polynomial;
	aCurrentPolynomial = this->current_polynomial;
	return this->polynomials;
}

void QuadraticSieve::Set_Polynomials(quad_polynomial* aPolynomials, long& aNumberOfPolynomials, long& aAllocatedPolynomials, long& aIndexInPolynomial, long& aCurrentPolynomial)
{
	
	if (this->polynomials)
	{
		int i;
		for (i=0; i< allocated_polynomials; i++)
		{
			ClearPolynomial(polynomials[i]);
		}
		free(this->polynomials);
		this->polynomials = NULL;
	}
	this->number_of_polynomials = aNumberOfPolynomials;
	this->allocated_polynomials = aAllocatedPolynomials;
	this->index_in_polynomial   = aIndexInPolynomial;
	this->current_polynomial    = aCurrentPolynomial;
	this->polynomials = aPolynomials;
}

void QuadraticSieve::Set_FactorBase(main_sieving_type* aFactorBase, long aMaxFBIndex, long aFbAllocated)
{
	if (this->factor_base)
	{
		free(this->factor_base);
	}
	this->factor_base = aFactorBase;
	this->maximal_factor_base_index = aMaxFBIndex;
	this->factor_base_allocated = aFbAllocated;	
}

void    QuadraticSieve::Set_ContiniThreshold(log_type aValue)
{	
	this->contini_threshold = aValue;
}

void    QuadraticSieve::Set_ErrorFactor(log_type aValue)
{
	this->error_factor = aValue;
}

void QuadraticSieve::Set_SmoothsFound(long aValue) 
{
	this->so_far_smooths_found = aValue;
}

void QuadraticSieve::Set_1PartialsFound(long aValue) 
{
	this->so_far_1_partials_found = aValue;
}

void QuadraticSieve::Set_2PartialsFound(long aValue) 
{
	this->so_far_2_partials_found = aValue;
}

void QuadraticSieve::Set_XPartialsFound(long aValue) 
{
	this->so_far_x_partials_found = aValue;
}


long QuadraticSieve::Get_SmoothsFound() const
{
	return this->so_far_smooths_found;
}

long QuadraticSieve::Get_1PartialsFound() const
{
	return this->so_far_1_partials_found;
}

long QuadraticSieve::Get_2PartialsFound() const
{
	return this->so_far_2_partials_found;
}

long QuadraticSieve::Get_XPartialsFound() const
{
	return this->so_far_x_partials_found;
}

log_type QuadraticSieve::Get_ContiniThreshold() const
{
	return this->contini_threshold;
}

log_type QuadraticSieve::Get_ErrorFactor() const
{
	return this->error_factor;
}

int   QuadraticSieve::Get_RelationProcessingFlags() const
{
	return this->iRelationProcessingFlags;
}

void    QuadraticSieve::Get_kN(mpz_t aRop) const
{
	mpz_set(aRop,this->kN);
}

main_sieving_type* QuadraticSieve::Get_FactorBase(long& aMaxFBIndex, long& aFbAllocated)
{
	aMaxFBIndex = this->maximal_factor_base_index;
	aFbAllocated = this->factor_base_allocated;
	return this->factor_base;
}

variations_types QuadraticSieve::Get_UsingVariations() const {
	return this->using_variations;
}

factor_alg QuadraticSieve::Get_FactoringAlgorithm() const {
	return this->factoring_algorithm;
}

check_mode QuadraticSieve::Get_CheckMode() const {
	return this->checking_mode;
}

siev_i_mode QuadraticSieve::Get_SievingIntervalMode() const {
	return this->sieving_interval_mode;
}

void QuadraticSieve::Get_M(mpz_t aRop) const {
	if (this->inited & M_INITED){
		mpz_set(aRop,this->M);
	}
}
int  QuadraticSieve::Get_Integer_M() const
{
	if (this->inited & M_INITED){
		return mpz_get_ui(this->M);
	}
	return 0;
}

void QuadraticSieve::Get_Root(mpz_t aRop) const {
	if (this->inited & ROOT_INITED){
		mpz_set(aRop,this->root);
	}
}

void QuadraticSieve::Get_MPF(mpz_t aRop) const {
	if (this->inited & MPF_INITED){
		mpz_set(aRop,this->maximal_partial_factor);
	}
}

void QuadraticSieve::Get_IdealAValue(mpz_t aValue) const
{
	if (this->inited & IDEAL_A_VALUE_INITED)
	{
		mpz_set(aValue,this->ideal_a_value);
	}
}

LineSiever*	QuadraticSieve::Get_AdvancedSieverInstance() const
{
	return this->iLineSiever;
}


void QuadraticSieve::Set_IdealAValue(mpz_t aValue)
{
	if (this->inited & IDEAL_A_VALUE_INITED)
	{
		mpz_set(this->ideal_a_value,aValue);
	}
}


bool QuadraticSieve::Get_LTMode() const {
	return (this->lanczos_time_message_mode);
}

bool QuadraticSieve::Get_WPMode() const {
	return (this->without_primes_listing);
}


long QuadraticSieve::Get_NrPolynomials() const {
	return (this->number_of_polynomials);
}

long QuadraticSieve::Get_UpperFBBound() const {
	return (this->upper_bound_on_factor_base);
}

bool QuadraticSieve::Get_DIMode() const {
	return (this->density_info_mode);
}

print_sort QuadraticSieve::Get_PrintingSort() const {
	return this->printing_sort;
}

qs_mode QuadraticSieve::Get_QSMode() const {
	return this->QSMode;
}

int QuadraticSieve::Get_S() const {
	return this->s;
}

char* QuadraticSieve::Get_DirectoryName() const
{
	return this->iDirectoryName;
}

int QuadraticSieve::Get_TimeMessageInterval() const {
	return this->time_message_interval;
}

sgt_mode QuadraticSieve::Get_SingletonMode() const {
	return this->SGTMode;
}

power_mode QuadraticSieve::Get_PowerMode() const {
	return this->power_sieving_mode;
}

bool QuadraticSieve::Get_ValDistMode() const {
	return this->value_distribution_mode;
}

bool QuadraticSieve::Get_ExtK() const {
	return this->extK;
}

bool QuadraticSieve::Get_UseRootShort() const {
	return this->iUseRootShort;
}


int QuadraticSieve::ReadDouble(double* aTarget, int aNameLength, FILE* aFile) {
	char  data[1000];
     	char* return_code;	
	return_code = fgets(data,aNameLength,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	return_code = fgets(data,1000,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	*aTarget = atof(return_code);
	return ConstRC::Ok;
}

/*!
	This method tries to read an integral value from a text file aFile. It expects that the value
	is given by a single line entry, started by aNameLength characters (including '=' and the NULL
	value at the presumed end of the name), and then the value itself follows.
	
	The result is put into variable *aTarget. The return value does not specify the result; it
	only gives the information whether the reading has been successful or no.
	
	Return values: - INCONSISTENT_STATE if reading of the line failed (distinguished by NULL result
	               of the fgets() call)
		       - ConstRC::Ok if everything went well
*/
int QuadraticSieve::ReadInt(int* aTarget, int aNameLength, FILE* aFile) {
	char  data[1000];
     	char* return_code;	
	return_code = fgets(data,aNameLength,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	return_code = fgets(data,1000,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	*aTarget = atoi(return_code);
	return ConstRC::Ok;
}

/*!
	These three methods are very similar to ReadInt. The only difference is that ReadMpz needs
	the argument aTarget to be init'ed before the call.
	
	Return values: - INCONSISTENT_STATE if reading of the line failed (distinguished by NULL result
	               of the fgets() call)
		       - ConstRC::Ok if everything went well
*/
int QuadraticSieve::ReadLong(long* aTarget, int aNameLength, FILE* aFile) {
	char  data[1000];
     	char* return_code;	
	return_code = fgets(data,aNameLength,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	return_code = fgets(data,1000,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	*aTarget = atol(return_code);
	return ConstRC::Ok;
}


int QuadraticSieve::ReadMpz(mpz_t aTarget, int aNameLength, FILE* aFile) {
	char  data[1000];
     	char* return_code;	
	return_code = fgets(data,aNameLength,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	return_code = fgets(data,1000,aFile);
	if (!return_code) {
		return INCONSISTENT_STATE;
	}
	mpz_set_str(aTarget,return_code,10);
	return ConstRC::Ok;
}

// GENERIC FUNCTIONS, UNAWARE OF INTERNAL QS STATE
/*!
	\brief The primary method for MPQS run (top level).
	
	This method calls all four top-level business methods in fixed order and returns one of these integer values,
	depending on their results:
	
		- PRIME_NUMBER, 			if the input number has been prime
		- ERROR_INIT_ENVIRONMENT,		if there has been any other error in environment initialization
		- ERROR_SIEVING			if there has been any error during the sieving phase
		- NOT_ENOUGH_SMOOTHS		if the sieving phase has not generated enough smooths to yield factorization
		- ConstRC::Factorized                      if the number has been factorized successfully
		- NOT_FACTORIZED			if the number has not been factorized, despite all steps have succeeded		
		- NO_DEPENDENCIES			if the method for solution of the sieving matrix returned no results
		- ERROR_LINALG			if there has been any error during the linear algebra phase

*/
int QuadraticSieve::RunMPQS(){
	int ret = 0;
	mpz_set_ui(this->start_time_main_process,time(NULL));
	//gmp_printf("Start time main process je rovno %Zd\n",this->start_time_main_process);
	ret = InitEnvironment();
		if (ret == PRIME_NUMBER) { return PRIME_NUMBER; }
		if (ret != ConstRC::Ok) { return ERROR_INIT_ENVIRONMENT; }
	ret = PerformSieving();
		if (ret != ConstRC::Ok) { 
			if (ret == TERMINATED)
			{
				return NOT_FACTORIZED;
			}
			return ERROR_SIEVING; 
		}
	if (ENode ==iDistributionMachineType)
	{
		return NOT_FACTORIZED;
	}
        ret = ProcessRelations();
		if (ret != ConstRC::Ok) { return ERROR_PROCESSING_RELATIONS; }
	ret = LinearAlgebraStep();
		if (ret == NOT_ENOUGH_SMOOTHS) { return NOT_ENOUGH_SMOOTHS; }
		if (ret == ConstRC::Factorized) 	    { return ConstRC::Factorized; }
		if (ret == NO_DEPENDENCIES) { return NO_DEPENDENCIES; }
		if (ret != ConstRC::Ok) { return ERROR_LINALG; } // all other errors (memory etc.)
	ret = Factor();
		if (ret == ConstRC::Factorized) { return ConstRC::Factorized; }
		if (ret == NOT_FACTORIZED) { return NOT_FACTORIZED; }
	return NOT_FACTORIZED;
}

int QuadraticSieve::DemonstrateBatches()
{
	int i, ret = InitEnvironment();
	if (ret != ConstRC::Ok) { return ERROR_INIT_ENVIRONMENT; }
	int batchCount = 1 + (time(NULL) % 3);
	printf("Now, demonstrating creation of batches. Batches: %d\n",batchCount);
	for (i=0; i < batchCount; i++)
	{
		int batchSize = 8 + (time(NULL) % 8);
		printf("\t\tThe batch #%d is requested to have %d jobs...\n",i,batchSize);
		Batch b1 = CreateNewJobs(batchSize);
		printf("\t\tThe batch returned has %d jobs in it.\n",b1.size());
	}
	return ConstRC::Ok;
}

int QuadraticSieve::RunTTR(testing_mode aMode){
        TTR* ttr_instance = new TTR(aMode);
	ttr_instance->InitialMessage();
	ttr_instance->TestBitMatrix();
	ttr_instance->TestSparseMatrix();
	ttr_instance->FinalMessage();
	delete ttr_instance;
	return TTR_OK;
}

// 1ST LEVEL METHODS, CALLED FROM GENERIC METHODS
/*!
	This method performs operations needed for setting up the algorithm environment. This includes creation of
	the factor base, data structures for intermediate results of operation, etc. The vast majority of memory
	allocations takes place here, and failing due to lack of memory could happen here.
	This method calls many second-level methods, and should they fail (by returning something different from ConstRC::Ok
	constant), it aborts the computation and propagates the result upwards. Due to this property, it is virtually
	impossible to count all possible return codes of this method.
	For the list of second-level methods called from this method, see section e2).

*/
int QuadraticSieve::InitEnvironment()
{
	if (iInitEnvironmentAlreadyRun)
	{
		return ConstRC::Ok;
	}
	int ret;
	ret = CheckInitConsistency();
		if (ret != ConstRC::Ok) 		 { return ret; }
	ret = InitFileNames();
		if (ret != ConstRC::Ok) 		 { return ret; }
	ret = CheckPrimality(this->kN);
		if (ret == PRIME_NUMBER) { return ret; }
	ret = ReadPrimesFromFile();
		if (ret != ConstRC::Ok) 		 { return ERROR_INIT_ENVIRONMENT; }
	ret = CheckDivisibility();
		if (ret != ConstRC::Ok)		 { return ret; }
	ret = CalculateKRoot();
		if (ret != ConstRC::Ok)		 { return ret; }
	ret = AllocVariationsAndHashes();
		if (ret != ConstRC::Ok)		 { return ret; }
	ret = DefineFactorBase();
		if (ret != ConstRC::Ok)		 { return ret; }
	if (this->power_sieving_mode == ESieveWithPrimePowers){
		ret = DefinePrimePowers();
		if (ret != ConstRC::Ok)		 { return ret; }
		ret = CalculateRootsOfkNForPrimePowers();
		if (ret != ConstRC::Ok)		 { return ret; }
	}
	if (this->using_variations != EDoNotUseVariations) {
		ret = SetFactoringStructures();
		if (ret != ConstRC::Ok)		 { return ret; }
	} 
	ret = AllocRootsOfkN();
		if (ret != ConstRC::Ok)		 { return ret; }
	ret = CalculateRootsOfkN();
		if (ret != ConstRC::Ok)		 { return ret; }
	ret = AllocSolutions();
		if (ret != ConstRC::Ok)		 { return ret; }
	ret = AllocMatrices();
		if (ret != ConstRC::Ok)		 { return ret; }	

	// In case of the new sieving methods, set-up the line siever.
	if ((this->QSMode == EMpqs) || (this->QSMode == ESiqs))
	{
		if (this->info_mode){
			NormalMessage("   Setting up the advanced line siever ...",NO_NUMBER);
		}
		this->iLineSiever = new LineSiever();
		int localM = mpz_get_ui(this->M);
		ret = this->iLineSiever->Init(this->iLineSieveSize,(-1)*localM,localM,this->upper_bound_on_factor_base);
		if (ret != ConstRC::Ok) 
		{
			return ret;
		}
		if (this->assertion_mode)
		{
			this->iLineSiever->SetAssertionMode(TRUE);
		}
		this->iLineSiever->SetInitBlockValue(0x80 - this->contini_threshold+16);
		this->iLineSiever->SetMaxLines(number_of_polynomials);
		this->iLineSiever->SetQsUseRootShort(this->iUseRootShort);
	}
	if (!iOptimizer)
	{
		iOptimizer = new ThresholdOptimizer(iOptimizationPolicy);
	}
	if ((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs)) {
		ret = SetSValue();
		if (ret != ConstRC::Ok)		 { return ret; }
		ret = InitiateSiqsStructures();
		if (ret != ConstRC::Ok)		 { return ret; }
	} else {
		this->s = 0;   // There are no divisors of A in the factor base		
	}
	ret = AllocPolynomials();	// In SIQS, this method must be called after determining of 's' value.
		if (ret != ConstRC::Ok)		 { return ret; }
	
	this->CheckEnough(); // pouze ku stanoveni "threshold-u"
	SetComputationFinished(false);
	if (this->info_mode){
	      PrintInformation();
	}
	if (!(this->iLoadFlags & SIEVING_LOAD_FLAG))
	{
		// Destructive opening of the files (all previous content is removed).
		printf("Init Environment: destructive opening of the files. Load flags = %x\n",this->iLoadFlags);
		ret = OpenFiles(WRITE);
		if (ret != ConstRC::Ok)		 
		{ 
			return ret; 
		}
	} 
	iInitEnvironmentAlreadyRun = TRUE;
	return ConstRC::Ok;
}

int  QuadraticSieve::RootCopy(main_sieving_type& aTarget1, main_sieving_type& aTarget2)
{
	//int result = 0;
	aTarget1  = this->root_1;	
	aTarget2  = this->root_2;
	if (this->assertion_mode) 
	{
		printf("\nQS.RootCopy: roots are %d, %d.\n",aTarget1,aTarget2); 
	}
	return FIRST_ROOT_VALID | SECOND_ROOT_VALID;
}

/*!
	This method is run during InitEnvironment(), if -reload command line option has been specified.
	It tries to read back all of the relations gathered during previous partial sieving, from both
	"smooths" file and "partials" file. Note that the implementation does not check exactly whether the
	relations really conform to the given input parameters; some relation consistency check is done,
	but, for exaple, no one checks whether the sieving interval has been set accordingly.
	If info mode is specified, some output is printed to ensure the user that things really go
	well. Especially reading of the partial relations may be very lengthy, and so regular
	progress info is very useful. By "reloading" we mean reading of the relations and putting
	the appropriate values into the variation hashtables etc.
	Having read the relations, the method determines which will be the next polynomial to 
	sieve with; it is NR+1, where NR is the largest polynomial index from union of the relations
	from "smooths" and "partials" file. After that, a "Polynomial fast forward" is performed:
	which means that polynomials from 0 to NR are cycled through as fast as possible (e.g. by
	use of NexKSB, without real calculations of 'A' and roots), to get the sieving process into
	a state as if the process of sieving had never been interrupted. 
	
	Return values:  - COULD_NOT_OPEN_FILE if opening of one of the files used failed
			- anything thrown by submethods called, e.g. GenerateNewPolynomial, ReadRelationFromFile
			etc.			
			- ConstRC::Ok if everything went well	
*/
int  QuadraticSieve::ReloadRelations() {
	int ret;
	// otevreme soubor smooths a zjistime, ktery je posledni zaznamenany polynom
	OpenFile(this->smooth_values_file_name,READ);
	if (this->smooth_values_file == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","reload relations",0,1,"Problem opening smooths file");
		return COULD_NOT_OPEN_FILE;
	}
	qs_relation* candidate = InitRelation();	
	int max_smooth_poly = -1;
	int max_smooth_index = -1;
	int max_partial_poly = -1;
	int index = 0;
	if (this->info_mode) { printf("\n"); }
	while(ReadRelationFromFile(this->smooth_values_file,candidate) == ConstRC::Ok) {
		if (candidate->index > max_smooth_index) {
			max_smooth_index = candidate->index;
		}
		if (candidate->index_of_used_polynomial > max_smooth_poly) {
			max_smooth_poly = candidate->index_of_used_polynomial;
		}
		if ((this->info_mode) && (!((index++)%10000))) {
			printf("\r   Reloading smooth values, so far %d ...",index);			
			fflush(stdout);
		}
	}	
	ret = CloseFile(this->smooth_values_file_name);
	this->so_far_smooths_found = index;
	if (this->info_mode) {
		printf("\r   Reloading smooth values finished (%d read).        \n",index);			
		fflush(stdout);
	}
	if (ret != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","reload relations",0,1,"Problem closing smooths file");
		return ret;
	}
	index = 0;	
	int nr_1_partials = 0;
	int nr_2_partials = 0;
	// pokud mame pouzivat parcialni relace, otevreme soubor s nimi a zjistime, ktery je posledni zaznamenany polynom
	if (this->using_variations != EDoNotUseVariations) {
		OpenFile(this->partial_values_file_name,READ);
		if (this->partials_file == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","reload relations",0,1,"Problem opening partials file");
			return COULD_NOT_OPEN_FILE;
		}
		while(ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
			if (candidate->index_of_used_polynomial > max_partial_poly) {
				max_partial_poly = candidate->index_of_used_polynomial;
			}
			if ((candidate->lprime1 != 1) && (candidate->lprime2 !=1)) {
				nr_2_partials++;
			} else {
				nr_1_partials++;
			}
			if ((this->info_mode) && (!((index++)%10000))) {
				printf("\r   Reloading partial values, so far %d ...",index);			
				fflush(stdout);
			}
			AddToCycleCountStructures(candidate);		
		}
		if (this->info_mode) {
			printf("\r   Reloading partial values finished (%d read).        \n",index);			
			fflush(stdout);
		}
		ret = CloseFile(this->partial_values_file_name);
		if (ret != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","reload relations",0,1,"Problem closing partials file");
			return ret;
		}
	}	
	this->so_far_x_partials_found = index;
	this->so_far_1_partials_found = nr_1_partials;
	this->so_far_2_partials_found = nr_2_partials;
	int max_poly_used = max_smooth_poly;
	if (max_partial_poly > max_poly_used) {
		max_poly_used = max_partial_poly;
	}
	// Provedeme "fast forward"
	if ((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs)) {
		int length_of_polys = 1 << (this->s - 1);
		int nexksb_shifts = max_poly_used / length_of_polys;
		for (index = 0; index < nexksb_shifts; index++) {
			NEXKSB_Next();
		}
		this->current_polynomial = nexksb_shifts*length_of_polys;
		this->index_in_polynomial = 0;
		// Nyni mame posunuto o "dolni cast" indexu.
		while (this->current_polynomial < max_poly_used) {
			ret = GenerateNewPolynomial();
			if (ret != ConstRC::Ok)   { return ret; }
			ret = QuadraticMachine();
			if (ret != ConstRC::Ok)   { return ret; }		
			this->index_in_polynomial++;
			this->current_polynomial++;
			if (this->index_in_polynomial == this->allocated_polynomials) {
				this->index_in_polynomial = 0;
			}			
		}
		printf("   SIQS Fast Forward: length of polys %d, max poly %d, nx sh %d, cur poly %ld, index of poly %ld.\n",length_of_polys,max_poly_used,nexksb_shifts,current_polynomial,index_in_polynomial);
	} else if (this->QSMode == EMpqs) {
	}
	return ConstRC::Ok;
}

/*double accumulated_sizes_rough;
double accumulated_sizes_refined;
double accumulated_sizes_accepted;
double accumulated_sizes_smooth;
int diff;
*/

void PrintStatistics(int* statistics) {
	int i;
	printf("      Statistika:\n");
	int max = 0;
	for (i=0; i<9; i++){
	    max += statistics[i];	
	}
	double max_double = (double) max;
	double current;
	double ratio;
	for (i=0; i<8; i++){
		current = (double) statistics[i];
		ratio = current/max_double;
		ratio *=100;
		printf("0x%d0 - 0x%d0:  %14d [%2.2lf %%]\n",i,i+1,statistics[i],ratio);
	}
	current = (double) statistics[8];
	ratio = current/max_double;
	ratio *=100;
		printf("0x80 a vice:  %14d [%2.2lf %%]\n",statistics[8],ratio);
}

void QuadraticSieve::PrintTimeInfo(int aWhich, double& aInProcess, double& aTotal)
{
	string	text = iUnknownPartText;
	switch  (aWhich)
	{
		case GEN_NEW_POLY:
			text = iGenerateNewPolyText;
			break;
		case QUAD_MACHINE:
			text = iQuadraticMachineText;
			break;
		case PERF_SIEVING:
			text = iPerformSievingByPolyText;
			break;
		case SIEVING_LOOP:
			text = iSievingLoopText;
			break;
		case REFILL_MAINM:
			text = iRefillMainMatrixText;
			break;
		case SORT_SMOOTHS:
			text = iSortSmoothText;
			break;
		case DIVISORS_OFA:
			text = iDivisorsOfAText;
			break;
		case SMALL_FACTOR:
			text = iSmallFactorText;
			break;
		case COUNT_CYCLES:	
			text = iCountCyclesText;
			break;
		case BLOCK_SIEVES:
			text = iBlockSieveText;
			break;
		case FILL_HSTABLE:
			text = iFillHashtableText;
			break;
		case FILL_HST_NEG:
			text = iFillHashtableNegText;
			break;
		case ROOTS_UPDATE:
			text = iRootsUpdateText;
			break;
		case FIND_FACTORZ:
			text = iFindAndFactorText;
			break;
		default:		
			break;
	}
	float abs = aInProcess/(double) CLOCKS_PER_SEC;
	float pct = 0;
	if (aTotal)
	{
		pct = 100*(aInProcess/(double) aTotal);
	}	 
	printf("%s %12.4f sec (%6.4f %%)\n",text.c_str(),abs,pct);
}

void QuadraticSieve::PrintEfficiencyInfo(int aBigDiv, int aAccepted, int aRejected, int aLines, double aTotalTime)
{
	printf("\n");			
	printf("      -----------------------------------------------\n");
	printf("      Overall # of big int divisions is          %6d.\n",aBigDiv);
	//printf("      IntDivA : %5d\n",total_short_divisions);
	//printf("      Hits    : %5d\n",total_hits);
	printf("      Tested  : %5d\n",aAccepted+aRejected);
	printf("      Accepted: %5d\n",aAccepted);
	printf("      Smooth  : %5ld\n",this->so_far_smooths_found);
	printf("      Rejected: %5d\n",aRejected);
	
	if (aAccepted) {
	
	printf("      Acc:rej : 1:%3.2f\n",(double) aRejected / (double) aAccepted);
	}
	if (aTotalTime > 0){
		printf("      Acce/sec: %8.4f\n",aAccepted*(double) CLOCKS_PER_SEC/(double) aTotalTime);
		printf("      Smoo/sec: %8.4f\n",this->so_far_smooths_found*(double) CLOCKS_PER_SEC/(double) aTotalTime);			
	}
	if (aLines) {
		printf("      Acce/pol: %8.4f\n",aAccepted/(double) aLines);
		printf("      Smoo/pol: %8.4f\n",this->so_far_smooths_found/(double) aLines);	
	}
}

bool QuadraticSieve::Verify_Solutions()
{
	int i, max = this->maximal_factor_base_index, res1;
	isol_type usable, sol_1, sol_2;
	main_sieving_type prime;
	bool result = TRUE;
	for (i=2; i<= max; i++)
	{
		usable = int_sols_of_cur_poly.usable[i];
		sol_1  = int_sols_of_cur_poly.sol_1[i];
		sol_2  = int_sols_of_cur_poly.sol_2[i];
		prime  = factor_base[i];
		if (usable & VALID_1) {
			res1 = Evaluate(sol_1,prime);
			if (res1)
			{
				printf("Verify_Solutions 1: x=%4d, p=%4d not ok.\n",sol_1,prime);			
				result = FALSE;	
			}
		}
		if (usable & VALID_2) {
			res1 = Evaluate(sol_2,prime);
			if (res1)
			{
				printf("Verify_Solutions 2: x=%4d, p=%4d not ok.\n",sol_2,prime);			
				result = FALSE;	
			}
		}
	}
	return result;
}

/*!
	This method takes care of the sieving process. This is the computationally most intensive part of the whole
	calculation. Memory allocations happen here, though not as often as in InitEnvironment; their frequency is
	greater if no intermediate files are used to store smooth values.
	This method calls many second-level methods, and should they fail (by returning something different from ConstRC::Ok
	constant), it aborts the computation and propagates the result upwards. Due to this property, it is virtually
	impossible to count all possible return codes of this method.

*/
int QuadraticSieve::PerformSieving() {
	int diff;
	int ret, startingPoly;	
	long i;
	int pbstp_start = 0;
	int ps_start = 0;
 /* prime_iters_1 = (int*) malloc((size_FB+3)*sizeof(int));
  prime_iters_2 = (int*) malloc((size_FB+3)*sizeof(int));
  prime_logs_1  = (int*) malloc((size_FB+3)*sizeof(int));
  prime_logs_2  = (int*) malloc((size_FB+3)*sizeof(int));
  memset(prime_iters_1,0,(size_FB+3)*sizeof(int));
  memset(prime_iters_2,0,(size_FB+3)*sizeof(int));
  memset(prime_logs_1,0,(size_FB+3)*sizeof(int));
  memset(prime_logs_2,0,(size_FB+3)*sizeof(int));*/
	
	memset(global_size_statistics,0,9*sizeof(int));
	memset(local_size_statistics,0,9*sizeof(int));
	if (this->info_mode){
		mpz_set_ui(this->start_time_subprocess_level_1,time(NULL));
		//gmp_printf("Subprocess 1 set as %Zd.\n",this->start_time_subprocess_level_1);
	}
  	/*accumulated_sizes_rough = 0;
  	accumulated_sizes_refined = 0;
  	accumulated_sizes_accepted = 0;
  	accumulated_sizes_smooth = 0;*/
	start = 0;
	//total_perform_sieving = time(NULL);		
	// Header for sieving information (runtime)
	
	//printf("\n Print Header and values ...");
	iIsInterruptable = TRUE;
	SetSievingPhase(ESievingRunning);	
	this->iRelationProcessingPhase = ECollectingRelations;
	bool loadCollectionFinished = FALSE;
	bool loadSievingRunning = FALSE;
	bool reloadedInstance = FALSE;
	if (this->iLoadFlags & SIEVING_LOAD_FLAG)
	{
		if (iSerializator)
		{
			ret = iSerializator->Deserialize(EQsSieving);
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
		}
		loadCollectionFinished = (this->iRelationProcessingFlags & RELATION_COLLECTION_FINISHED_FLAG);
		printf("Load Collection Finished: %d (%x, %x).\n",loadCollectionFinished,this->iRelationProcessingFlags, RELATION_COLLECTION_FINISHED_FLAG);
		loadSievingRunning = (this->iRelationProcessingFlags & SIEVING_RUNNING_FLAG);
		if (loadCollectionFinished && loadSievingRunning)
		{	
			CLog::LogError(WHEREARG,"Quadratic sieve","Perform sieving",iLoadFlags,iRelationProcessingFlags,"Inconsistent flags");
			return GENERAL_ERROR;
		}
		if (loadCollectionFinished)
		{
			// Nothing to be done now.
			NormalMessage("   Perform Sieving: no need to continue with sieving, enough smooths have been collected before.",NO_NUMBER);
			return ConstRC::Ok;
		} 
		else if (loadSievingRunning)
		{
			NormalMessage("   Perform Sieving: Will load sieving. ",NO_NUMBER);
			return ConstRC::Ok;
			ret = ReloadSieverState();
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
			Check_LastSolutions();			
			reloadedInstance = TRUE;
		}
		startingPoly = iLineNr;
	} 
	else {
		ret = iSerializator->Serialize(EQsSieving);
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
		startingPoly = 0;
	}	
	if (info_mode) {
		PrintHeader();
		PrintValues();
	}
	if ((this->QSMode == ESiqs) || (this->QSMode == EMpqs))
	{	
		//printf("\n New MPQS and SIQS ... - fresh instance");
		if (!loadSievingRunning)
		{
			// At first, we will generate the first sieving polynomial 
			this->iFirstLineSieverPolynomial = TRUE;
			ret = GenerateNewPolynomial();
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
			this->iFirstLineSieverPolynomial = FALSE;
			//printf("\n New polynomial generated ...");
			// ... and determine the roots. After that, the poly generation and
		// shift of the sieving values will be a matter for the iLineSiever instance.
			ret = QuadraticMachine();
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
			//printf("\n Quad machine run ...");
			ret = InitAdvancedFactorBase();
			//printf("\n Advanced FB Inited ...");
			//fflush(stdout);
			if (ret != ConstRC::Ok) 
			{
				return ret;
			}
		} else 
		{
			// Reloaded sieving instance
			//printf("\n New MPQS and SIQS ... - reloaded instance");
		}
		iLineSiever->SetQsFactorBase(advanced_factor_base);		
		iLineSiever->SetQsInstance(this);
		Set_RelationProcessingFlag(SIEVING_RUNNING_FLAG);
		iStartSievingClock = clock();
		ret = iLineSiever->RunSieve();		
		iFinishSievingClock = clock();
		if (ENode == this->iDistributionMachineType)
		{
			while ((this->HasDataToSend()) && (!(this->Get_InterruptNow()))) { 
				fprintf(stderr, "\n Finishing communication \n");
				SLEEP(3);
			};
		};
		Clear_RelationProcessingFlag(SIEVING_RUNNING_FLAG);
		if (ret != ConstRC::Ok) { return ret; }
		ret = CloseFiles();  // we close files that we wrote into (next methods will need them open for reading)
		if (ret != ConstRC::Ok) { return ret; }
		ret = DeleteMainMatrix(); // we do not need main matrix anymore
		if (ret != ConstRC::Ok) { return ret; }
		if (this->info_mode)
		{
			printf("\n"); // to get to new line after "sieving step" messages
			TimeMessage("Perform Sieving",LEVEL_1);
		}
		if (CheckEnough() == NOT_ENOUGH)
		{
			CLog::LogError(WHEREARG,"Quadratic sieve","Perform sieving",0,1,"Not enough relations gathered");
			return NOT_ENOUGH;
		}
		this->iRelationProcessingFlags |= RELATION_COLLECTION_FINISHED_FLAG;
		if (iSerializator)
		{
			ret = iSerializator->Serialize(EQsSieving);
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
		}
		return ConstRC::Ok;
	}
	
	/*printf("Total perform sieving je %d.\n",total_perform_sieving);
	The sieving: * generate a new polynomial
			* compute its roots
	             * refill the main matrix 
	             * sieve through the interval
	             * sort out the candidates for full trial factorization * and perform the factorization*/
	Set_RelationProcessingFlag(SIEVING_RUNNING_FLAG);
	if (iOptimizer)
	{
		delete iOptimizer;
	}
	iOptimizer = new ThresholdOptimizer(iOptimizationPolicy);
	for (i= startingPoly; i< this->number_of_polynomials; i++){
		iLineNr = i;
// Testovani
		/*if (i == 25)
		{
			printf("\nQSLineSiever: do you wish to interrupt sieving?");
			char c;
			scanf("%c",&c);
			if ((c=='y') || (c=='Y'))
			{
				this->iInterruptNow = TRUE;
			}
		}*/
		if (this->iInterruptNow)
		{
			printf("(Old) QS line siever received request for serialization.\n");
			if (iSerializator)
			{
				ret = iSerializator->Serialize(EQsSieving);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
				ret = iSerializator->Serialize(EQsSieverState);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}				
				if (this->assertion_mode) {
					Check_LastSolutions();
				}
			}
			Clear_RelationProcessingFlag(SIEVING_RUNNING_FLAG);
			return TERMINATED;
		} else {
			// We will dump the information regarding sieving anyway
			// in regular intervals;
			time_t now = time(NULL);
			if (difftime(now,lastSaved) > AUTOSAVE_INTERVAL_IN_SECONDS)
			{
				lastSaved = now;
				printf("\rSaving sieving status, please wait ...\n");
				if (iSerializator)
				{
					ret = iSerializator->Serialize(EQsSieving);
					if (ret != ConstRC::Ok)
					{
						return ret;
					}
					ret = iSerializator->Serialize(EQsSieverState);
					if (ret != ConstRC::Ok)
					{
						return ret;
					}				
					if (this->assertion_mode) {
						Check_LastSolutions();
					}
				}			
			}
		}
#ifdef TIME_MESSAGE		
		ps_start = clock();				
		start = clock();
#else
		// Even in case when TIME_MESSAGE is undef'd, we need to collect data about runtime for CheckAuto feature
		if (this->checking_mode == ECheckAuto) {
			ps_start = clock();
		}
#endif
		ret = GenerateNewPolynomial();
		if (ret != ConstRC::Ok)   { return ret; }
#ifdef TIME_MESSAGE		
		AddTime(total_poly,start);
		start = clock();
#endif		
		ret = QuadraticMachine();

#ifdef TIME_MESSAGE	
		AddTime(total_solv,start);	
		/*diff = clock() - start;
		if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
			total_solv += (double) diff;
		}*/		
#endif		
		if (ret != ConstRC::Ok)   { return ret; }		
#ifdef SIQS_DEBUG	
	quad_polynomial* poly;
	poly = &(this->polynomials[this->index_in_polynomial]);		
	gmp_printf("\n\n\n\nSoucasny polynom (#%d) je (%Zd,%Zd).\n\n\n\n",i,poly->a, poly->b);
#endif			
#ifdef TIME_MESSAGE		
		pbstp_start = clock();
#endif		
		ret = PerformSievingByThisPolynomial();		
#ifdef TIME_MESSAGE		
		AddTime(total_psbtp,pbstp_start);		
#endif				
		if (ret == ENOUGH) {
			SetSievingPhase(ESievingFinished);
			if (this->info_mode){
				printf("\n"); // to get to new line after "sieving step" messages
				TimeMessage("Perform Sieving",LEVEL_1);	
			}
			if (this->frontend) {
				fprintf(this->frontend_pipe,"\nPhaseSieve> (%4.2f%%) \n",
						100.0);
				fflush(this->frontend_pipe);
			}
			ret = CloseFiles();
			if (ret != ConstRC::Ok) { return ret; }
			ret = DeleteMainMatrix();
			if (ret != ConstRC::Ok) { return ret; }
			Clear_RelationProcessingFlag(SIEVING_RUNNING_FLAG);
			this->iRelationProcessingFlags |= RELATION_COLLECTION_FINISHED_FLAG;
			if (iSerializator)
			{
				printf("!!!! Serializing Sieving, rpflags are 0x%x.\n",iRelationProcessingFlags);
				ret = iSerializator->Serialize(EQsSieving);
				if (ret != ConstRC::Ok)
				{
					return ret;
				}
			}
			return ConstRC::Ok;
		} // we have found enough values
		if (ret != ConstRC::Ok)     { return ret;}	
		if ((i % THRESHOLD_OPTIMIZER_INTERVAL) == 0)
			{
			iOptimizer->AnalyzeAndOptimize(this->contini_threshold);
			}
		if ((this->QSMode == EOldSiqs) || (this->QSMode == EOldMpqs))	
		{
			IncrementPolynomialCounter();
		}
#ifdef TIME_MESSAGE		
		AddTime(total_perform,ps_start);
		if ((this->time_message_interval) && (!(i%this->time_message_interval)) && 1 && i) {						
			if (total_perform > 0){
				//printf("time(NULL) je %d, total perform sieving je %d.\n",time(NULL),total_perform_sieving);
				printf("\nSo far, %ld iterations have been run.\n",i);
				PrintTimeInfo(GEN_NEW_POLY,total_poly,total_perform);
				PrintTimeInfo(QUAD_MACHINE,total_solv,total_perform);
				PrintTimeInfo(PERF_SIEVING,total_psbtp,total_perform);
				PrintTimeInfo(SIEVING_LOOP,total_sieve,total_perform);
				PrintTimeInfo(REFILL_MAINM,total_refill,total_perform);
				PrintTimeInfo(SORT_SMOOTHS,total_sort,total_perform);	PrintTimeInfo(DIVISORS_OFA,total_divisors_of_a,total_perform);			
				if (this->using_variations == EDoubleLargePrimeV) {
					PrintTimeInfo(SMALL_FACTOR,total_double_factors,total_perform);
					PrintTimeInfo(COUNT_CYCLES,total_cyclecount,total_perform);
				}
			}
			printf("\n      TOTAL TIME SPENT IN PERFORM SIEVING  IS   "); TimeFormat((int) (total_perform/(double) CLOCKS_PER_SEC)); 	PrintEfficiencyInfo(total_divrem, accepted_nr, rejected_nr,i, total_perform);
			/*for (int j=0; j<=maximal_factor_base_index; j++){
				printf("Prvocislo %5d: logs_1 = %9d, logs_2 = %9d, iters_1 = %9d, iters_2 = %9d.\n",factor_base[j],prime_logs_1[j],prime_logs_2[j],prime_iters_1[j],prime_iters_2[j]);
				//global_size_statistics[j] += local_size_statistics[j];
			}*/
			// We'll prepare header for runtime information
			if (info_mode) {
				PrintHeader();
				PrintValues();		
			}
		}
	
#endif 
		if (this->checking_mode == ECheckAuto) {			
			if (this->autocheck.state != EFinished)/* && (((this->flexicheck.state == EFinished) || (this->sieving_interval_mode != EFlexible)))) */{			
				diff = clock() - ps_start;			
				if (diff > 0){
					this->autocheck.collected_cycles += diff;
				}
				if (this->autocheck.collected_cycles > this->autocheck.next_check_moment) {
					//printf("Autocheck: next check moment\n");
					switch (this->autocheck.state) {
						case EBeforeStart:      this->autocheck.state = EStartUp;
									break;
						case EStartUp:		if (AddToScore() == ConstHT::FINISHED) {
										this->autocheck.score_start = CountNewScore();
										this->contini_threshold +=1;
										this->autocheck.state = EStartDown;				
										//printf("CheckAuto: Menime state na StartDown. Score start je %6.2f\n",this->autocheck.score_start);		
									}						
									break;
						case EStartDown:        if (AddToScore() == ConstHT::FINISHED) {
										this->autocheck.score_up = CountNewScore();
										this->contini_threshold -= 2;
										this->autocheck.state = EDecide;
										//printf("CheckAuto: Menime state na Decide. Score up je %6.2f\n",this->autocheck.score_up);		
									}
									break;
						case EDecide:           if (AddToScore() == ConstHT::FINISHED) {
										this->autocheck.score_down = CountNewScore();
										//printf("CheckAuto: Score down je %6.2f\n",this->autocheck.score_down);		
										if ((this->autocheck.score_start > this->autocheck.score_up) && (this->autocheck.score_start > this->autocheck.score_down)) {
											// Zafixujeme hodnotu na puvodni a ukoncime
										 	this->contini_threshold += 1;
											this->autocheck.state = EFinished;
											break;
										}
										if (this->autocheck.score_up > this->autocheck.score_down) {
											this->autocheck.state = EGoUp;
											this->autocheck.score_now = this->autocheck.score_up;
											this->contini_threshold += 3;
											//printf("CheckAuto: Budeme pokracovat nahoru\n");
										} else {
											this->autocheck.state = EGoDown;
											this->autocheck.score_now = this->autocheck.score_down;
											this->contini_threshold -= 1;
											//printf("CheckAuto: Budeme pokracovat dolu\n");
										}
									}
									break;
						case EGoUp:		if (AddToScore() == ConstHT::FINISHED) {
										this->autocheck.score_last = this->autocheck.score_now;
										this->autocheck.score_now = CountNewScore();
										if (this->autocheck.score_last > this->autocheck.score_now) {
											// zastavit
											//printf("CheckAuto: Koncime s rustem\n");
											this->contini_threshold -= 1;
											this->autocheck.state = EFinished;
										} else {
											threshold += 1;
											}
									}
									break;
						case EGoDown:           if (AddToScore() == ConstHT::FINISHED) {
										this->autocheck.score_last = this->autocheck.score_now;
										this->autocheck.score_now = CountNewScore();
										if (this->autocheck.score_last > this->autocheck.score_now) {
											// zastavit
											//printf("CheckAuto: Koncime s klesanim\n");
											this->contini_threshold += 1;
											this->autocheck.state = EFinished;										
										} else {
											threshold -= 1;
											}
									}
									break;									
						case EFinished:
									break;
					}
					this->autocheck.collected_cycles = 0;
				}
			}

		}
		// We start shifting the sieving interval if and only if either checkmode is not auto, or it is auto, but it has already determined the optimal value of the contini threshold
		if (this->sieving_interval_mode == EFlexible) {
			//if ((this->autocheck.state == EFinished) || (this->checking_mode != ECheckAuto)) {
				if (this->flexicheck.state != EFinished) {			
					diff = clock() - ps_start;			
					if (diff > 0){
						this->flexicheck.collected_cycles += diff;
					}
					if (this->flexicheck.collected_cycles > 	this->flexicheck.next_check_moment) {
						//printf("Flexicheck: next check moment (%d clocks)\n",this->flexicheck.next_check_moment);
						switch (this->flexicheck.state) {
							case EBeforeStart:      this->flexicheck.state = EStartUp;
										break;
							case EStartUp:		if (AddToScore() == ConstHT::FINISHED) {
											this->flexicheck.score_start = CountNewScore();
											this->integer_M +=1*DEFAULT_INTERVAL_STEP;
											ResetMainMatrix();
											this->flexicheck.state = EStartDown;				
											//printf("CheckFlexi: Menime state na StartDown. Score start je %6.2f\n",this->flexicheck.score_start);		
										}						
										break;
							case EStartDown:        if (AddToScore() == ConstHT::FINISHED) {
											this->flexicheck.score_up = CountNewScore();
											this->integer_M -= 2*DEFAULT_INTERVAL_STEP;
											ResetMainMatrix();
											this->flexicheck.state = EDecide;
											//printf("CheckFlexi: Menime state na Decide. Score up je %6.2f\n",this->flexicheck.score_up);		
										}
										break;
							case EDecide:           if (AddToScore() == ConstHT::FINISHED) {
											this->flexicheck.score_down = CountNewScore();
											//printf("CheckFlexi: Score down je %6.2f\n",this->flexicheck.score_down);
											if ((this->flexicheck.score_start > this->flexicheck.score_up) && (this->flexicheck.score_start > this->flexicheck.score_down)) {
												this->integer_M +=1*DEFAULT_INTERVAL_STEP;
												ResetMainMatrix();
												this->flexicheck.state = EFinished;
												break;
											}		
											if (this->flexicheck.score_up > this->flexicheck.score_down) {
												this->flexicheck.state = EGoUp;
												this->flexicheck.score_now = this->flexicheck.score_up;
												this->integer_M += 3*DEFAULT_INTERVAL_STEP;
												ResetMainMatrix();
												//printf("CheckFlexi: Budeme pokracovat nahoru\n");
											} else {
												this->flexicheck.state = EGoDown;
												this->flexicheck.score_now = this->flexicheck.score_down;
												this->integer_M -=1*DEFAULT_INTERVAL_STEP;
												ResetMainMatrix();
												//printf("CheckFlexi: Budeme pokracovat dolu\n");
											}
										}
										break;
							case EGoUp:		if (AddToScore() == ConstHT::FINISHED) {
											this->flexicheck.score_last = this->flexicheck.score_now;
											this->flexicheck.score_now = CountNewScore();
											if (this->flexicheck.score_last > this->flexicheck.score_now) {
												// zastavit
												this->flexicheck.state = EFinished;
												integer_M -= 1*DEFAULT_INTERVAL_STEP;
												ResetMainMatrix();
										printf("CheckFlexi: Koncime s rustem. Optimalni interval M stanoven jako %d.\n",this->integer_M);												
											} else {
												integer_M += 1*DEFAULT_INTERVAL_STEP;
												ResetMainMatrix();
											}
										}
										break;
							case EGoDown:           if (AddToScore() == ConstHT::FINISHED) {
											this->flexicheck.score_last = this->flexicheck.score_now;
											this->flexicheck.score_now = CountNewScore();
											if (this->flexicheck.score_last > this->flexicheck.score_now) {
												// zastavit												
												this->flexicheck.state = EFinished;										
												integer_M += 1*DEFAULT_INTERVAL_STEP;
												//printf("CheckFlexi: Koncime s klesanim. Optimalni interval M stanoven jako %d.\n",this->integer_M);												
												ResetMainMatrix();
											} else {
												integer_M -= 1*DEFAULT_INTERVAL_STEP;
												ResetMainMatrix();
											}
										}
										break;									
							case EFinished:
										break;
						}
						this->flexicheck.collected_cycles = 0;
				}
			}
			
		}
		
	}
	
	SetSievingPhase(ESievingFinished);
	ret = CloseFiles();  // we close files that we wrote into (next methods will need them open for reading)
		if (ret != ConstRC::Ok) { return ret; }
	ret = DeleteMainMatrix(); // we do not need main matrix anymore
		if (ret != ConstRC::Ok) { return ret; }
	if (this->info_mode){
		printf("\n"); // to get to new line after "sieving step" messages
		TimeMessage("Perform Sieving",LEVEL_1);
	}
	Clear_RelationProcessingFlag(SIEVING_RUNNING_FLAG);
	this->iRelationProcessingFlags |= RELATION_COLLECTION_FINISHED_FLAG;
	if (iSerializator)
	{
		ret = iSerializator->Serialize(EQsSieving);
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
	}
	return ConstRC::Ok;
}

bool  QuadraticSieve::Get_Interruptable() const
{
	return this->iIsInterruptable;
}

void  QuadraticSieve::Set_InterruptNow(bool aValue)
{
	this->iInterruptNow = aValue;
	this->iCenterRunInterrupted = aValue;
}

void   QuadraticSieve::Set_RelationProcessingFlags(int aValue) 
{
	this->iRelationProcessingFlags = aValue;
}


bool  QuadraticSieve::Get_InterruptNow() const
{
	return this->iInterruptNow;
}

relproc_phase QuadraticSieve::Get_RelationProcessingPhase() const
{
	return this->iRelationProcessingPhase;
}

void QuadraticSieve::Set_RelationProcessingPhase(relproc_phase aValue)
{
	this->iRelationProcessingPhase = aValue;
}

void QuadraticSieve::IncrementPolynomialCounter()
{
	this->index_in_polynomial++;
	this->current_polynomial++;
	if (this->index_in_polynomial == this->allocated_polynomials) {
		this->index_in_polynomial = 0;
	}
}
/*!
	This method processes the smooth, 1-partial and 2-partial variables found during the sieving phase. The
	processing includes:
		- removal of both large and normal singletons,
		- construction of fundamental cycles from x-partials,
		- checking duplicates among the collected rows.
	This method calls many second-level methods, and should they fail (by returning something different from ConstRC::Ok
	constant), it aborts the computation and propagates the result upwards. Due to this property, it is virtually
	impossible to count all possible return codes of this method.		

*/
int QuadraticSieve::ProcessRelations() {
	int ret;
	if (this->iLoadFlags & SIEVING_LOAD_FLAG)
	{
		if (this->iRelationProcessingFlags & RELATIONS_READY_FLAG)
		{
			NormalMessage("   Relation processing will be skipped.",NO_NUMBER);
			// We will read the smooths into smooth_values 
			this->smooth_values_file = OpenFile(this->smooth_values_file_name,READ);
			ret = AddSmoothValuesToRelationMatrix();
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
			CloseFile(this->smooth_values_file_name);
			return ConstRC::Ok;
		}
	}  
	// Processing of found relations has not been finished in the last run
	{
		if (this->info_mode){
			mpz_set_ui(this->start_time_subprocess_level_1,time(NULL));
		}
		if (this->using_variations != EDoNotUseVariations) {		
			
			bool needsHashtableReload = !(this->iRelationProcessingFlags & (LARGE_SINGLETONS_REMOVED_FLAG | CYCLES_CONSTRUCTED_FLAG));
			if (needsHashtableReload)
			{
				printf("\nReconstruction of hashtable running...");
				ret = ReconstructHashtable();
				if (ret != ConstRC::Ok)
				{
				}
				printf("\nReconstruction of hashtable done. Res = %d.\n",ret);
			}
			this->independent_cycles = this->edges + this->components - this->vertices;
			if (!(this->iRelationProcessingFlags & LARGE_SINGLETONS_REMOVED_FLAG))
			{					
				
				this->independent_cycles = this->edges + this->components - this->vertices;
				ret = RemoveLargeSingletons();
				if (ret != ConstRC::Ok) {
		        		CLog::LogError(WHEREARG,"Quadratic sieve","relation processing step",this->collected,this->threshold,"There has been some error in removing large singletons");
					return ret;	
				}
			}
			if (!(this->iRelationProcessingFlags & CYCLES_CONSTRUCTED_FLAG))
			{
				ret = ConstructCycles();
				if (ret != ConstRC::Ok) {
		        		CLog::LogError(WHEREARG,"Quadratic sieve","relation processing step",this->collected,ret,"There has been an error in cycle construction.");			
					return ret;
				}
			}
        	}
		ret = CheckEnough();
		if (ret == ENOUGH)
		{
		// Everything ConstRC::Ok
		}
		if (ret == ENOUGH_BUT_WARNING)
		{
		// Print out warning.
			if (info_mode)
			{
			printf("Warning: barely enough smooth values!\n");
			}
		}
		if (ret == NOT_ENOUGH) {
				if (this->value_distribution_mode) {
					PrintDistributionInfo();
				}
				// Check whether the collected values are still more numerous than the factor base.
				
				CLog::LogError(WHEREARG,"Quadratic sieve","relation processing step",this->collected,this->threshold,"There is not enough smooth and partial values to perform linear algebraic operations");
				return NOT_ENOUGH_SMOOTHS;
			}
	}
	ret = OpenFiles(READ);
		if (ret != ConstRC::Ok)	{ return ret; }
	ret = AddSmoothValuesToRelationMatrix();
		if (ret != ConstRC::Ok)  { return ret; }
	ret = CheckAndDeleteDuplicates();
		if (ret != ConstRC::Ok)  { return ret; }
	ret = CloseFiles();
		if (ret != ConstRC::Ok)	{ return ret; }
	if (this->SGTMode == EDeleteSingletons){
		ret = CheckAndDeleteSingletons();
			if (ret != ConstRC::Ok)  { return ret; }
	}
	AssertSmooths();
	if (this->info_mode){
		printf("\n");
		TimeMessage("Relation processing step",LEVEL_1);
	}
	this->iRelationProcessingPhase = ERelationsReady;
	this->iRelationProcessingFlags |= RELATIONS_READY_FLAG;
	if (this->iSerializator)
	{
		ret = iSerializator->Serialize(EQsSieving);
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
	}
	return ConstRC::Ok;
}

int  QuadraticSieve::InitAdvancedFactorBase()
{
	/*if (this->info_mode){
		NormalMessage("   Initializing the advanced FB ...",NO_NUMBER);
	}*/
	if (this->advanced_factor_base)
	{
		CLog::LogError(WHEREARG,"Quadratic sieve","init advanced fb",0,0,"Advanced FB already inited?");
		return INCONSISTENT_STATE;
	}
	this->advanced_factor_base = (qs_fb_type*) AllocationMachine((void**) &this->advanced_factor_base, (this->factor_base_allocated)*sizeof(qs_fb_type));
	if (this->advanced_factor_base == NULL)
	{
		CLog::LogError(WHEREARG,"Quadratic sieve","init advanced fb",this->factor_base_allocated,(this->factor_base_allocated)*sizeof(qs_fb_type),"Could not allocate advanced.");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->advanced_factor_base,0,(this->factor_base_allocated)*sizeof(qs_fb_type));
	int i, max = maximal_factor_base_index;	
	main_sieving_type res1, res2;
	qs_fb_type* curr;
	isol_type*     usable =  this->int_sols_of_cur_poly.usable;
    	isol_type*     sol_1  =  this->int_sols_of_cur_poly.sol_1;
    	isol_type*     sol_2  =  this->int_sols_of_cur_poly.sol_2;

	//printf(" Max = %d.\n",max);
	for (i=0; i<= max; i++)
	{
		//printf("\r Now running = %6d.",i);
		fflush(stdout);
		curr = & (this->advanced_factor_base[i]);
		curr->p = factor_base[i];
		// The positive sieving block in new sieving always starts with 1. Therefore, offset 0 is equal to x = 1; that is why we subtract 1. Value of -1 means that this root is unusable for sieving
		if (usable[i] & VALID_1) {
			if (this->assertion_mode)
			{
				res1 = Evaluate(sol_1[i],curr->p);
				if (res1)
				{
					printf("InitAdvFB: x=%4d, p=%4d not ok.\n",sol_1[i],curr->p);
				}
			}
			curr->root_1 = sol_1[i] - 1;
			if (curr->root_1 < 0) {
				curr->root_1 += curr->p;
			}
		} else {
			curr->root_1 = -1;
		}
		if (usable[i] & VALID_2) {
			if (this->assertion_mode)
			{
				res2 = Evaluate(sol_2[i],curr->p);
				if (res2)
				{
					printf("InitAdvFB: x=%4d, p=%4d not ok.\n",sol_2[i],curr->p);
				}
			}
			curr->root_2 = sol_2[i] - 1;
			if (curr->root_2 < 0) {
				curr->root_2 += curr->p;
			}
		} else {
			curr->root_2 = -1;
		}
		curr->log_p = binary_logarithms_of_factor_base_times_log_factor[i];
		curr->sqrt_kn = roots_of_kN_mod_p[i];
		if (this->QSMode == ESiqs)
		{
			curr->ba_inv = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &curr->ba_inv,(this->s)*sizeof(main_sieving_type));
			int j;
			for (j=0; j< this->s; j++)
			{
				curr->ba_inv[j] = Ba_inv[i][j];
			}
		}
	}
	//printf("Returning O.K. from init Adv FB...\n");
	return ConstRC::Ok;
}

/*!
	This method takes care about the linear algebraic operations, which means: converting the resulting set of 
	relations to a bit matrix and solving the resulting sparse GF(2)-matrix. 
	Since operation with large matrices takes place in this method, another allocation failure may happen here.
	However, I have never experienced such behavior.
	This method calls many second-level methods, and should they fail (by returning something different from ConstRC::Ok
	constant), it aborts the computation and propagates the result upwards. Due to this property, it is virtually
	impossible to count all possible return codes of this method.

*/
int QuadraticSieve::LinearAlgebraStep(){
	int ret;
	if (this->info_mode){
		mpz_set_ui(this->start_time_subprocess_level_1,time(NULL));
	}
	ret = ToBitMatrix();
		if (ret == NOT_ENOUGH){
			CLog::LogError(WHEREARG,"Quadratic sieve","linear algebra step",125,this->threshold,"There is not enough smooth and partial values to perform linear algebraic operations");
			return ret;
		}
		if (ret != ConstRC::Ok) { return ret; }
	ret = SolveSystem();
		if (ret == NO_DEPENDENCIES) {
			NormalMessage("   No proper dependencies have been found.",NO_NUMBER);
			TimeMessage("Linear Algebra Step",LEVEL_1);
			PrintApology();
		}
		if (ret != ConstRC::Ok) { return ret; }
	if (this->info_mode){
		if (this->using_variations) {	printf("\n"); // to get a new line after messages "reading from files"
				 	}
		TimeMessage("Linear Algebra Step",LEVEL_1);
	}

	return ConstRC::Ok;
}

/*!
	This is the shortest of the first-level methods. It takes the dependecies generated by the previous method
	and tries to find a proper factor of the input number. It returns one of the following values:
	
	    ConstRC::Factorized,       if a proper factor has been found
	    
	    NOT_FACTORIZED,   if no proper factor has been found

*/
int QuadraticSieve::Factor(){
	int ret;
	int dep = this->class_dependencies->GetMaximalRowIndex()+1;

	if (this->info_mode){
		mpz_set_ui(this->start_time_subprocess_level_1,time(NULL));
		NormalMessage("   Trying to factor the number ...",NO_NUMBER);
		printf("(%d nonzero dependencies)",dep);
	}
	if (this->frontend){
		fprintf(this->frontend_pipe,"\nPhaseFactorize> \n");
		fflush(this->frontend_pipe);
	}
	ret = FindFactors();
	if (ret == ConstRC::Factorized) {
		SetComputationFinished(true);
		PrintFactors();
	}
	if (ret == NOT_FACTORIZED) PrintApology();
	if (this->info_mode){
		printf("\n");
		TimeMessage("Factor",LEVEL_1);
		printf("\n"); // the last running executive process
		TimeMessage("Quadratic Sieve",MAIN_LEVEL);
	}
	return ret;
}

// 2ND LEVEL METHODS
/*!
	This method checks whether N and M have been defined. If no, it returns CALLED_TOO_EARLY, otherwise ConstRC::Ok.

*/
int QuadraticSieve::CheckInitConsistency(){
	if (this->info_mode){
		NormalMessage("   Checking init consistency ...",NO_NUMBER);
	}
	if (!(this->inited & N_INITED)){
		CLog::LogError(WHEREARG,"Quadratic sieve","check init consistency",CALLED_TOO_EARLY,3,"Calling InitEnvironment before defining N.");
		return CALLED_TOO_EARLY;
	}
	if (!(mpz_cmp_si(this->N,INIT_VALUE_N))){
		CLog::LogError(WHEREARG,"Quadratic sieve","check init consistency",CALLED_TOO_EARLY,4,"Calling InitEnvironment before defining N.");
		return CALLED_TOO_EARLY;
	}
	if (!(this->inited & M_INITED)){
		CLog::LogError(WHEREARG,"Quadratic sieve","check init consistency",CALLED_TOO_EARLY,5,"Calling InitEnvironment before defining M - warning. Got default value for M.");
		mpz_init_set_ui(this->M,DEFAULT_VALUE_M);
	}
	return ConstRC::Ok;
}

/*!
	This method checks whether N is a composite number or no, by calling  MILLER_RABIN_ROUNDS_STRICT
	rounds of mpz_probab_prime_p test. It returns either ConstRC::Ok, when N is composite, or PRIME_NUMBER, if it is prime.

*/
int QuadraticSieve::CheckPrimality(mpz_t aArg){
	if (mpz_probab_prime_p(aArg,MILLER_RABIN_ROUNDS_STRICT)) {
		CLog::LogError(WHEREARG,"Quadratic sieve","checking primality of input",0,35,"The number to be factored is prime");
		return PRIME_NUMBER;
	}
	else    { return ConstRC::Ok; }
}

/*!
	This method tries to read primes from file primefil.fil to memory. If it does not find the file in the
	current working directory, it tries to generate it from scratch.
	It returns ERROR_INIT_ENVIRONMENT in case of any problem, or ConstRC::Ok when everything went smoothly.	

*/
int QuadraticSieve::ReadPrimesFromFile(){
	int ret;
	if (this->iLoadFlags & FB_LOAD_FLAG)
	{
		NormalMessage("   FB will be loaded from file, no need to read primes ...",NO_NUMBER);
		return ConstRC::Ok;
	}
	if (this->info_mode){
		NormalMessage("   Reading primes from file ...",NO_NUMBER);
	}
	//generate_file_of_small_primes(prime_numbers_file_name,2,10000000,0,1);
	ret=copy_file_of_small_primes_into_memory(prime_numbers_file_name.c_str(),this->upper_bound_on_factor_base);
	if (ret == COULD_NOT_OPEN_FILE) {
		NormalMessage("   Reading from file failed. Trying to regenerate the file of primes, please wait ...",NO_NUMBER);
		ret = generate_file_of_small_primes(prime_numbers_file_name.c_str(),2,DEFAULT_PRIMEFILE_BOUND,0,1,this->frontend_pipe);
		if (ret != ConstRC::Ok){
			CLog::LogError(WHEREARG,"Quadratic sieve","generating prime file",ret,20,"There was an error generating prime file");
			return ERROR_INIT_ENVIRONMENT;
		}
		ret=copy_file_of_small_primes_into_memory(prime_numbers_file_name.c_str(),this->upper_bound_on_factor_base);
	}
	if (ret != length_of_small_primes) {
		CLog::LogError(WHEREARG,"Quadratic sieve","reading primes from file",ret,20,"There was an error reading primes from designed file");
		return ERROR_INIT_ENVIRONMENT;
	}
	else { return ConstRC::Ok; }
}

/*!
	This method tries to find a nontrivial divisor of N among small primes read into memory by the previous
	method by trial division. If it finds one, it returns DIVISIBLE_BY_SMALL_PRIME, otherwise it returns ConstRC::Ok.

*/
int QuadraticSieve::CheckDivisibility(){
	int i;
	if (this->info_mode){
		NormalMessage("   Checking divisibility ...",NO_NUMBER);
	}
	for (i=0;i< length_of_small_primes; i++){
		if (mpz_divisible_ui_p(this->N,small_primes[i])){
			CLog::LogError(WHEREARG,"Quadratic sieve","checking divisibility",small_primes[i],21,"The number to be factored is divisible by a small prime");
			mpz_set_ui(this->factor_1,small_primes[i]);
			mpz_set(this->factor_2,this->N);
			mpz_divexact_ui(this->factor_2,this->factor_2,small_primes[i]);
			return DIVISIBLE_BY_SMALL_PRIME;
		}
	}
	return ConstRC::Ok;
}

void QuadraticSieve::PrintPolynomial(quad_polynomial* aPoly)
{
	gmp_printf("\nNew polynomial: %Zd x^2 + %Zd x + %Zd.\n",aPoly->a,aPoly->b,aPoly->c);
}

void QuadraticSieve::PrintPolynomialMod(quad_polynomial* aPoly, main_sieving_type aP)
{
	mpz_t aux_1;
	mpz_init(aux_1);
	main_sieving_type a = mpz_mod_ui(aux_1,aPoly->a,aP);
	main_sieving_type b = mpz_mod_ui(aux_1,aPoly->b,aP);
	main_sieving_type c = mpz_mod_ui(aux_1,aPoly->c,aP);
	printf("\nNew polynomial: %d x^2 + %d x + %d mod %d.\n",a,b,c,aP);
	mpz_clear(aux_1);
}

/*!
	This method calculates three useful values:
	- small multiplier k, which is calculated as N mod 8, and used to alter the factorized value to allow
	inclusion of 2 into the FB.
	- value of lower integer part of square root of kN.
	It returns only ConstRC::Ok value.	

*/
int QuadraticSieve::CalculateKRoot(){
	mpz_t pom;
	int flag=0, flag1=0, flag2=0;
	if (this->info_mode){
		NormalMessage("   Calculating KDRoot ...",NO_NUMBER);
	}
	mpz_init(pom);
	/*
	 * We will walk all values of potential 'k', until we meet some which multiplied by N is a quadratic residue mod 3,5 and 7 at the same time, while still equal to N mod 8.
	 *
	 */
	if (this->extK) {
		int p1; 
		int pom_2 = mpz_mod_ui(pom,this->N,(unsigned long int) 8);
		int pom_3 = mpz_mod_ui(pom,this->N,(unsigned long int) 3);
		int pom_5 = mpz_mod_ui(pom,this->N,(unsigned long int) 5);
		int pom_7 = mpz_mod_ui(pom,this->N,(unsigned long int) 7);
		printf("N mod 3 = %d, N mod 5 = %d, N mod 7 = %d\n",pom_3,pom_5,pom_7);
		int trailingK = pom_2 - 8;	
		int ok = 0;
		do {
			trailingK += 8;
			if ((pom_3*trailingK) % 3 == 2) {
				continue;
			}
			p1 = pom_5 * trailingK;
			if ((p1 == 2) || (p1 == 3)) continue;
			p1 = pom_7 * trailingK;
			if ((p1 == 3) || (p1 == 5) || (p1 == 6)) continue;
			ok = 1;
		} while (!ok);
		this->k	= trailingK; // mpz_mod_ui(pom,this->N,(unsigned long int) 8);
	} else {
		this->k = mpz_mod_ui(pom,this->N,(unsigned long int) 8);
	}
	mpz_mul_ui(this->kN,this->N,this->k);
	/*mpz_set(pom,this->kN);
	mpz_fdiv_q_ui(pom,pom,2);
	mpz_sqrt(pom,pom);
	mpz_fdiv_q(pom,pom,M);
	mpz_sqrt(pom,pom);
	while (!flag)
	{
		mpz_nextprime(pom,pom);
		if ((mpz_fdiv_ui(pom,(unsigned long int) 4))==3) flag1=1; else flag1=0;
		if ((mpz_kronecker(pom,kN))==1) flag2=1; else flag2=0;
		flag=(flag1 & flag2);
	}*/
	mpz_root(this->root,this->kN,2);
	mpz_clear(pom);
	return ConstRC::Ok;
}

/*!

	In this method, structures for holding sieving polynomials are created. Their size is determined differently 
	in MPQS and in SIQS mode: in MPQS, space for DEFAULT_POLYNOMIALS_ALLOCATED is allocated, while in SIQS, the
	number of polynomials kept in memory is (1 << (s-1)). Therefore, this method must be called after determination
	of number 's'.
	
	During allocation of polynomial structures, all involved mpz_t's are inited as well.
	
	Return codes:
	- NOT_ENOUGH_MEMORY if any allocation fails
	- ConstRC::Ok otherwise.

*/
int QuadraticSieve::AllocPolynomials(){
	this->allocated_polynomials = DEFAULT_POLYNOMIALS_ALLOCATED;	
	if ((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs)){
		this->allocated_polynomials = (1 << (this->s-1));
	}
	if (this->info_mode){
		NormalMessage("   Allocating polynomials ... ",NO_NUMBER);
	}
	this->polynomials = (quad_polynomial*) AbstractSieve::AllocationMachine((void**) &this->polynomials,this->allocated_polynomials*sizeof(quad_polynomial));
	if (this->polynomials == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating polynomials",0,20,"There was an error allocating array of sieving polynomials");
		return NOT_ENOUGH_MEMORY;
	}	
	long i;
	for (i=0; i< this->allocated_polynomials; i++){
		AllocPolynomial(this->polynomials[i]);
	}	
	this->index_in_polynomial = 0;
	this->current_polynomial  = 0;
	if (this->info_mode) {
		printf("memory for %ld polynomials allocated.",this->allocated_polynomials);
	}
	return ConstRC::Ok;
}

/*!
This method is a historic relict from earlier versions of MPQS/SIQS. It only performs the following operations:
	- sets maximal_partial_factor to 128*maximal element of factor base
	- sets this->tree and this->hash_root to NULL.
	It returns only ConstRC::Ok value.
	
*/
int QuadraticSieve::AllocVariationsAndHashes(){
	if (this->info_mode){
		NormalMessage("   Allocating variations and hashes ...",NO_NUMBER);
	}
	if (this->using_variations != EDoNotUseVariations){
		mpz_set_ui(this->maximal_partial_factor,upper_bound_on_factor_base);
		if (this->using_variations == EDoubleLargePrimeV) {
			mpz_mul(this->minimal_candidate_value,this->maximal_partial_factor,this->maximal_partial_factor);
		}
		mpz_mul_ui(this->maximal_partial_factor,this->maximal_partial_factor,VARIATION_FACTOR);		
		if (!mpz_fits_uint_p(this->maximal_partial_factor)) {
			CLog::LogError(WHEREARG,"Quadratic sieve","checking max fb size",0,20,"Maximal value of a large prime does not fit into unsigned int. Aborting.");
			return EXCESSIVE_FACTOR_BASE;
		}
		if (this->using_variations == EDoubleLargePrimeV) {
			mpz_mul(this->maximal_candidate_value,this->maximal_partial_factor,this->maximal_partial_factor);
		}
	}
	this->hash_root = NULL;
	return ConstRC::Ok;
}

/*!
	This method tries to open files for writing intermediate results. The aMode parameter may have the following values: READ, WRITE, APPEND.
	
	Possible return codes:
	- BAD_ARGUMENT, if the aMode argument was different from READ, WRITE, APPEND.
	- COULD_NOT_OPEN_FILE, if the file handle returned by OS is NULL.
	- ConstRC::Ok in case of success.

*/
int QuadraticSieve::OpenFiles(int aMode){	
	if (this->using_variations != EDoNotUseVariations){
		this->partials_file = OpenFile(partial_values_file_name,aMode);
		if (this->partials_file == NULL){
			return COULD_NOT_OPEN_FILE;
		}
	}
	this->smooth_values_file = OpenFile(smooth_values_file_name,aMode);
	if (this->smooth_values_file == NULL){
		return COULD_NOT_OPEN_FILE;
	}	
	return ConstRC::Ok;
}

FILE* QuadraticSieve::OpenFile(const string& aName, int aMode) {
	FILE* file = NULL;	
	switch (aMode){
		case READ:	file = fopen(aName.c_str(),"r");						
				break;
		case WRITE:	file = fopen(aName.c_str(),"w");
				//printf("\nOpened file %s for writing.\n",aName);
				break;
		case APPEND:	file = fopen(aName.c_str(),"a");
				break;
		default:	CLog::LogError(WHEREARG,"Quadratic sieve","opening file",aMode,121,"Could not open file in unknown mode");
				break;
	}
	if (file != NULL) {
		if (aName == this->smooth_values_file_name) {
			this->smooth_values_file = file;
			this-> open |= SMOOTHS_OPEN;
		}
		if (aName == this->partial_values_file_name) {
			this->partials_file = file;
			this->open |= PARTIALS_OPEN;
		}
		if (aName == this->all_partial_values_file_name) {
			this->all_partials_file = file;
			this->open |= ALL_PARTIALS_OPEN;
		}
		if (aName == this->usable_partial_values_file_name) {
			this->usable_partials_file = file;
			this->open |= USABLE_PARTIALS_OPEN;
		}
		if (aName == this->temporary_file_name) {
			this->temporary_file = file;
			this->open |= TEMPORARY_OPEN;
		}
	}
	return file;
}

/*!
	This method is responsible for closing open files. 
	
	Return codes:
	- COULD_NOT_CLOSE_FILE if there has been an error closing any of the files
	- ConstRC::Ok if everything went well.

*/
int QuadraticSieve::CloseFiles(){
    int a, b, c, d, e;
    if (this->info_mode){
    	printf("\n");
    	printf("   Closing files ... ");
	fflush(stdout);
    }
	
    a = CloseFile(this->usable_partial_values_file_name);
    if (a) CLog::LogError(WHEREARG,"Quadratic sieve","closing file usable partials (in close files)",a,11,"Could not close file usable partials.");

    b = CloseFile(this->all_partial_values_file_name);
    if (b) CLog::LogError(WHEREARG,"Quadratic sieve","closing file all partials (in close files)",b,11,"Could not close file all partials.");

    c = CloseFile(this->partial_values_file_name);
    if (c) CLog::LogError(WHEREARG,"Quadratic sieve","closing file partials (in close files)",c,11,"Could not close file partials.");

    d = CloseFile(this->smooth_values_file_name);
    if (d) CLog::LogError(WHEREARG,"Quadratic sieve","closing file smooths (in close files)",d,11,"Could not close file smooths.");

    e = CloseFile(this->temporary_file_name);
    if (e) CLog::LogError(WHEREARG,"Quadratic sieve","closing file temporary (in close files)",e,11,"Could not close file temporary.");
   if (a || b || c || d || e){
	    if (this->info_mode){
	    	printf("problem!");
		fflush(stdout);
    	    }
	    return COULD_NOT_CLOSE_FILE;
	} else {
	    if (this->info_mode){
	    	printf("Ok");
		fflush(stdout);
    	    }
	    return ConstRC::Ok;
	  }
}

	// Useful for XmlQsSerializator and distribution
int	QuadraticSieve::CloseDataFile(data_file_type aType)
{
	if (aType == ESmooth)
	{
		return CloseFile(this->smooth_values_file_name);
	}
	if (aType == EPartial)
	{
		return CloseFile(this->partial_values_file_name);
	}
	return GENERAL_ERROR;
}

FILE*	QuadraticSieve::OpenDataFile(data_file_type aType, int aMode)
{
	if (aType == ESmooth)
	{
		return OpenFile(this->smooth_values_file_name, aMode);
	}
	if (aType == EPartial)
	{
		return OpenFile(this->partial_values_file_name, aMode);
	}
	return NULL;
}

int QuadraticSieve::CloseFile(const string& aName) {
	int ret = ConstRC::Ok;
	if ((aName == this->partial_values_file_name) && (this->open & PARTIALS_OPEN)) {
		fflush(this->partials_file);
		ret = fclose(this->partials_file);
		if (ret != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file partials (in close file)",ret,11,"Could not close file partials.");
		}
		this->partials_file = NULL;
		this->open &= ~(PARTIALS_OPEN);
	} 
        else
	if ((aName == this->smooth_values_file_name) && (this->open & SMOOTHS_OPEN)) {
		fflush(this->smooth_values_file);
		ret = fclose(this->smooth_values_file);
		if (ret != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file smooths (in close file)",ret,11,"Could not close file smooths.");
		}
		this->smooth_values_file = NULL;
		this->open &= ~(SMOOTHS_OPEN);
	}
	else
	if ((aName == this->all_partial_values_file_name) && (this->open & ALL_PARTIALS_OPEN)) {
		fflush(this->all_partials_file);
		ret = fclose(this->all_partials_file);
		if (ret != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file all partials (in close file)",ret,11,"Could not close file all partials.");
		}
		this->all_partials_file = NULL;
		this->open &= ~(ALL_PARTIALS_OPEN);
	}
	else
	if ((aName == this->usable_partial_values_file_name) && (this->open & USABLE_PARTIALS_OPEN)) {
		fflush(this->usable_partials_file);
		ret = fclose(this->usable_partials_file);
		if (ret != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file usable partials (in close file)",ret,11,"Could not close file usable partials.");
		}
		this->usable_partials_file = NULL;
		this->open &= ~(USABLE_PARTIALS_OPEN);
	}
	else
	if ((aName == this->temporary_file_name) && (this->open & TEMPORARY_OPEN)) {
		fflush(this->temporary_file);
		ret = fclose(this->temporary_file);
		if (ret != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","closing file temporary (in close file)",ret,11,"Could not close file temporary.");
		}
		this->temporary_file = NULL;
		this->open &= ~(TEMPORARY_OPEN);
	} 
	return ret;
}


int QuadraticSieve::SetFactoringStructures() {
	if (this->info_mode){
		NormalMessage("   Setting factorization structures ...",NO_NUMBER);
	}
	if (this->using_variations == EDoubleLargePrimeV) {		
		switch (this->factoring_algorithm) {
			case EPollardRho:	fact_alg = new PollardRho();
						break;
			case ECFRAC:		fact_alg = new CFRAC(DEFAULT_CFRAC_BASE_SIZE);
						break;
			case EPMinus1:		fact_alg = new PollardPMinus1(DEFAULT_FB_PRIME_BOUND);
						break;
			case EEllipticCurve:	fact_alg = new oldECM(100);					
						break;
			default:		return BAD_ARGUMENT;					
		}
	}	
	int table_size = DEFAULT_CYCLE_TABLE_SIZE;	// 512 K. Must be power of 2, in order for the following to work
	if (this->maximal_factor_base_index > 10000) { 
		table_size <<= 1;			// 1 M
	}
	if (this->maximal_factor_base_index > 20000) { 
		table_size <<= 1;			// 2 M
	}
	if (this->maximal_factor_base_index > 40000) { 
		table_size <<= 1;			// 4 M
	}	
	if (this->maximal_factor_base_index > 80000) { 
		table_size <<= 1;			// 8 M
	}	
	if (this->using_variations == ELargePrimeV) {
	// In case of single LPV, we do not need so much space
		table_size >>= 3; 
	}

	// The main hashtable for detection of cycles. Each entry corresponds to a large prime found in a x-partial.
	cycle_table_1 = new lp_hashtable_common<CCommonHashEntry>();
	cycle_table_1->SetupHashtable(table_size);
	if (this->cycle_table_1 == NULL) {
		return NOT_ENOUGH_MEMORY;
	}
	ResetHashtable();
	this->independent_cycles = 0;
	this->intermediate_roots = (unsigned int*) AllocationMachine((void**) &this->intermediate_roots,DEFAULT_INTERMEDIATE_ROOTS*sizeof(unsigned int));
	if (this->intermediate_roots == NULL) {
		return NOT_ENOUGH_MEMORY;
	}
	this->max_ir_index = DEFAULT_INTERMEDIATE_ROOTS - 1;
	// Structure for clearance of large singletons (large primes occuring in one x-partial only)
	this->root_list	      = (hashtable_entry_type_1*) AllocationMachine((void**) &this->root_list,DEFAULT_ROOT_LIST*sizeof(hashtable_entry_type_1));
	if (this->root_list == NULL) {
		return NOT_ENOUGH_MEMORY;
	}
	this->rl_allocated    = DEFAULT_ROOT_LIST-1;
	this->rl_max_index    = -1;

	factoring_structures_inited = TRUE;
	return ConstRC::Ok;
}

/*!
	This method clears the hashtable used for representation of the partials graph during the
	sieving phase.
	
*/
void QuadraticSieve::ResetHashtable() {
 	int i;
	// Special case of i=0: PRIME=1, root of its own component.
	cycle_table_1->ClearHashtable();
	this->vertices = 1;
	this->components = 1;
	this->edges = 0;
	this->cycles = 0;
}

/*!
	This method clears the hashtable used for representation of the partials graph during the
	partial processing phase (and construction of cycles).
*/
void QuadraticSieve::ResetHashtable_2() {
	int i;
	// Special case of i=0: PRIME=1, root of its own component.
	this->cycle_table_2->ClearHashtable();
	for (i=1; i< cycle_table_2->AllocatedSize; i++) {
		this->cycle_table_2->Hashtable[i].SetRelation(ConstHT::FULL_MASK_NFS);
	}	
}

/*!
	This method is used in autocheck processes (both finding the best threshold and finding the best
	sieving interval).
	
	This method tries to determine the check interval in cycles, depending on the size of kN.
	The reasoning is as follows: for smaller kN's, the sieving is a really fast process, determining
	many smooths and 1-partials per sec, so a shorter interval must be used. For larger kN's, the
	process of sieving is slower, and having a short check interval would result in many statistical
	errors. For example, if in 10 seconds, 5 partials are found, and then 6, the difference may be
	caused by random processes and need not be significant. Therefore, a longer interval for collection
	of data is needed.
	
	Return value: the interval in cycles.	

*/
int QuadraticSieve::FindOptimalCheckInterval() {
	int result = 0;
	if (this->kN_decimal_length < 60) {
			
			result = CLOCKS_PER_SEC / 4;
		} else if (this->kN_decimal_length < 63) {
			result = CLOCKS_PER_SEC / 2;
		} else if (this->kN_decimal_length < 66) {
			result = CLOCKS_PER_SEC;
		} else if (this->kN_decimal_length < 70) {
			result = CLOCKS_PER_SEC*2;
		} else if (this->kN_decimal_length < 73) {
			result = CLOCKS_PER_SEC*4;
		} else if (this->kN_decimal_length < 76) {
			result = CLOCKS_PER_SEC*8;
		} else if (this->kN_decimal_length < 80) {
			result = CLOCKS_PER_SEC*12;
		} else if (this->kN_decimal_length < 83) {
			result = CLOCKS_PER_SEC*20;
		} else {
			result = CLOCKS_PER_SEC*30;
		}
	return result;
}

/*!
This method defines (and allocates) the factor base array. 
	At first, one cannot determine the size of FB exactly before actually defining it. So, the method allocates 
	rich space for the factor base by malloc-ing a relatively large space (large enough to contain ALL primes smaller than B). Then, it puts -1 and 2 into the FB on the first two indices. After that, the method
	searches the array of primes defined by ReadPrimesFromFile(); and if kN is quadratic residue modulo a prime,
	that prime is inserted into the FB.
	
	Once the FB is generated, the method generates other auxiliary data structures whose size depends on size
	of the factor base, which are arrays:
	- a_inverses
	- exact_binary_logarithms_of_factor_base
	- binary_logarithms_of_factor_base_times_log_factor
	- is_this_prime_divisor_of_a
	
	Each of these allocations may fail, in which case the method returns NOT_ENOUGH_MEMORY.
	
	In rough logarithmic mode, not all primes are sieved with; the sieving begins at some position in the FB,
	say 100 or 200. If the factor base generated is smaller than this threshold, an error TOO_SMALL_FACTORBASE
	is returned.
	
	Moreover, several integer values are updated:	    
	
	this->factor_base_allocated
	
	this->maximal_factor_base_index
	
	this->original_maximal_fb_index (this value is needed at deallocation time in destructor, since maximal_factor_base_index changes when singletons are removed)
	
	this->contini_threshold
	
	this->accepted
	
	this->margin	
	
	Aside from the already mentioned return values, ConstRC::Ok is returned when everything went well.	
*/
int QuadraticSieve::DefineFactorBase(){
	if (this->info_mode){
		NormalMessage("   Defining factor base ...",NO_NUMBER);
	}
	int i=2,j=1;
	bool loadedSuccessfully = FALSE;
	mpz_t pom;
	mpz_init(pom);

	if (this->iLoadFlags & FB_LOAD_FLAG)
	{
		if (iSerializator)
		{
			int res = iSerializator->Deserialize(EQsFactorBase);
			if (res == ConstRC::Ok)
			{
				loadedSuccessfully = TRUE;
				i = this->maximal_factor_base_index+1;
			}
		}
	}
	if (!loadedSuccessfully)
	{
		// ALLOC FB
		this->factor_base = (main_sieving_type *) AbstractSieve::AllocationMachine((void**) &this->factor_base,(length_of_small_primes+4)*sizeof(main_sieving_type));
		if (this->factor_base == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating factor base",0,40,"There was an error allocating factor base");
			return NOT_ENOUGH_MEMORY;
		}
		this->factor_base_allocated = length_of_small_primes+2;
		this->factor_base[0] = -1;
		this->factor_base[1] =  2;
		for (j=1;j<=length_of_small_primes; j++){ // should not be here '<' ?
			mpz_set_ui(pom,small_primes[j]);
			if (mpz_jacobi(this->kN,pom)+1)
			{
				this->factor_base[i]= (main_sieving_type) small_primes[j];
				i++;
			}
		}
		this->maximal_factor_base_index=i-1;
		this->original_maximal_fb_index=i-1;	
	}
	// We will reduce the factor base memory requirements.
	this->factor_base_allocated = this->maximal_factor_base_index + 5;
	this->factor_base = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->factor_base,(this->factor_base_allocated)*sizeof(main_sieving_type));
	// ALLOC A_INVERSES (are of the same length)	
	i = RefillInverses_Of_A(FALSE /* do not calculate values*/);
	if (i != ConstRC::Ok)
	{
		return i;
	}
	this->factor_base[this->maximal_factor_base_index+1] = 0; // This is "terminal sign" - will show that the factor base is finished	
	for (i=1;i<=this->maximal_factor_base_index;i++){
		if (this->factor_base[i] >= ROUGH_MODE_LOWER_BOUND){
			break;
		}
	}
	if (this->rough_factorbase_index < i) this->rough_factorbase_index = i;	
	if (this->info_mode) {
		printf(" rough factorbase index set to %ld (prime  = %d).",this->rough_factorbase_index,this->factor_base[this->rough_factorbase_index]);
	}
	/* Determine "Contini Threshold"  - 0x5*log_2(kN) + log_2(M) - log_2(maximal_partial_factor) - ROUNDOFF_ERROR */
	this->contini_threshold = mpz_sizeinbase(this->kN,2)/2 + mpz_sizeinbase(this->M,2) - this->error_factor;
	
	if (this->using_variations == ELargePrimeV) {
		this->contini_threshold -= mpz_sizeinbase(this->maximal_partial_factor,2);
	}
	if (this->using_variations == EDoubleLargePrimeV) {
		this->contini_threshold -= mpz_sizeinbase(this->maximal_candidate_value,2);
	}
	this->kN_decimal_length = mpz_sizeinbase(this->kN,10);
	// In case of autocheckmode, we init the corresponding structure
	for (i=0; i<= CHECKAUTO_MAXINDEX; i++) {
		this->checkinfo.partials_1[i] = 0;
		this->checkinfo.partials_2[i] = 0;
	}
	this->checkinfo.index       = -1;
        if (this->checking_mode == ECheckAuto) {
	
		this->autocheck.score_start = 0;
		this->autocheck.score_up    = 0;
		this->autocheck.score_down  = 0;
		this->autocheck.score_last  = 0;
		this->autocheck.score_now   = 0;
		this->autocheck.state       = EBeforeStart;		
		this->autocheck.collected_cycles = 0;
		this->autocheck.next_check_moment = FindOptimalCheckInterval();
		//printf("kN length: %d, next check moment = %d.\n",this->kN_decimal_length,this->autocheck.next_check_moment);
	}
    	//if (this->assertion_mode){
	//	AbstractSieve::NormalMessage("Logarithm_accepted was calculated as",this->logarithm_accepted);
	//}
	if (this->sieving_interval_mode == EFlexible) {		
		this->flexicheck.score_start = 0;
		this->flexicheck.score_up    = 0;
		this->flexicheck.score_down  = 0;
		this->flexicheck.score_last  = 0;
		this->flexicheck.score_now   = 0;
		this->flexicheck.state       = EBeforeStart;		
		this->flexicheck.collected_cycles = 0; 
		this->flexicheck.next_check_moment = FindOptimalCheckInterval();		
	}
	if (this->rough_factorbase_index > this->maximal_factor_base_index){
		CLog::LogError(WHEREARG,"Quadratic sieve","rough mode index determination",0,50,"too small factor base for rough mode");
		mpz_clear(pom);
		return TOO_SMALL_FACTORBASE;
	}
	double log_2 = (double) 1/(double) log10(2);
	double actual_log;
	int help;
	this->exact_binary_logarithms_of_factor_base = (float*) AbstractSieve::AllocationMachine((void**) &this->exact_binary_logarithms_of_factor_base,(this->maximal_factor_base_index+1)*sizeof(float));
	if (this->exact_binary_logarithms_of_factor_base == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating logarithms of divisors of A",this->maximal_factor_base_index+1,74,"There was an error allocating array of logarithms of divisors of A");
		return NOT_ENOUGH_MEMORY;
	}
	this->binary_logarithms_of_factor_base_times_log_factor = (log_type*) AbstractSieve::AllocationMachine((void **) &(this->binary_logarithms_of_factor_base_times_log_factor),(this->maximal_factor_base_index+1)*sizeof(log_type));
	if (this->binary_logarithms_of_factor_base_times_log_factor == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating binary logarithms",0,40,"There was an error allocating binary logarithms");
		return NOT_ENOUGH_MEMORY;
	}
	this->binary_logarithms_of_factor_base_times_log_factor[0] = 0;
	this->binary_logarithms_of_factor_base_times_log_factor[1] = 0;
	for (i=2;i<=this->maximal_factor_base_index;i++){
		actual_log = log10((double) this->factor_base[i]);
		actual_log *= log_2;
		this->exact_binary_logarithms_of_factor_base[i] = actual_log;
		actual_log *= _LOG_FACTOR;		
		help = (log_type) actual_log+1;		
		this->binary_logarithms_of_factor_base_times_log_factor[i] = help;
	}
	this->is_this_prime_divisor_of_a = (main_sieving_type*) AbstractSieve::AllocationMachine((void **) &(this->is_this_prime_divisor_of_a),(this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	if (this->is_this_prime_divisor_of_a == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating is this prime divisor of a",0,40,"There was an error allocating is this prime divisor of a");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->is_this_prime_divisor_of_a,0,(this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	mpz_clear(pom);	
	this->margin *= (maximal_factor_base_index)/1000 + 1;
	if ((this->using_variations == ELargePrimeV))// && (this->processing_partials == EAddingToFB)){
        {		this->margin *= 2;
	}
	if (iSerializator)
	{
		int res = iSerializator->Serialize(EQsFactorBase);
		if (res != ConstRC::Ok)
		{
			return res;
		}
	}
	return ConstRC::Ok;
}

/*!
	This method reallocates and possibly refills the array of inverses of 'a'
	
	Return value: the score.
*/

int  QuadraticSieve::RefillInverses_Of_A(bool aCalculate) {
	if (this->a_inverses)
	{
		free(this->a_inverses);
		this->a_inverses = NULL;
	}
	this->a_inverses = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->a_inverses, (this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	if (this->a_inverses == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating a_inverses",0,40,"There was an error allocating a_inverses");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->a_inverses,0,(this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	if (aCalculate)
	{
		ComputeAInverses();
	}
	return ConstRC::Ok;
}


/*!
	This method is used by autocheck processes to save current data about number of relations found
	into the score_info structures.
	
	Return codes: - RUNNING if the number of values collected is not big enough yet
		      - ConstHT::FINISHED if the number of values collected is already big enough and the final score can be calculated	
*/
int  QuadraticSieve::AddToScore() {
	this->checkinfo.index++;
	//printf("\nAdd to score: pridavame na index %d, nalezene hodnoty jsou (%d 1-partials) a (%d 2-partials).\n",this->checkinfo.index,this->so_far_1_partials_found,this->so_far_2_partials_found);
	this->checkinfo.partials_1[this->checkinfo.index] = this->so_far_1_partials_found;
	this->checkinfo.partials_2[this->checkinfo.index] = this->so_far_2_partials_found;
	if (this->checkinfo.index == CHECKAUTO_MAXINDEX) {
		this->checkinfo.index = -1;
		return ConstHT::FINISHED;
	} else return RUNNING;	
}

/*!
	This method counts a score from the score_info structures. This score measures effect of
	current parameter values onto the efficiency of the whole sieving algorithm. For details,
	see description in documentation of types.h file.
	
	Return value: the score.
*/
double QuadraticSieve::CountNewScore() {
	int i, total = 0;
	for (i=0; i < CHECKAUTO_MAXINDEX; i++) {
		total += PARTIAL_1_WEIGHT*(this->checkinfo.partials_1[i+1] - this->checkinfo.partials_1[i]) + PARTIAL_2_WEIGHT*(this->checkinfo.partials_2[i+1] - this->checkinfo.partials_2[i]);
	}
	double score = ((double) total / ((double) CHECKAUTO_MAXINDEX));
	//printf("\nNew score: %6.2f.\n",score);
	return score;
}

/*!
	This method is called if and only if sieving with prime powers is enabled. It allocates data structures 
	needed for sieving with prime powers and sets several member variables. This means that allocation failure
	may happen.
	
	The following arrays are allocated:
	- this->prime_powers
	- for each element of this->prime_powers, the arrays 
	  - this->prime_powers[i].powers
	  - this->prime_powers[i].roots_of_kN_mod_powers
	  - this->prime_powers[i].a_inverses
	  - this->prime_powers[i].solutions_1
	  - this->prime_powers[i].solutions_2
	  - this->prime_powers[i].usable	
	  	    
	  In case of SIQS, also	  
	  
	  - this->prime_powers[i].Ba_inv
	  
	If any of the allocation fails, NOT_ENOUGH_MEMORY is returned, else ConstRC::Ok.

*/
int  QuadraticSieve::DefinePrimePowers() {
	if (this->info_mode){
		NormalMessage("   Defining structures for prime powers ...",NO_NUMBER);
	}
	// At first, we determine the prime power limit - that is, the maximal acceptable value of a prime power sieved with
	this->prime_power_limit = DEFAULT_LIMIT_MULTIPLIER*this->integer_M;
	
	// Now, we allocate the structure for prime powers
	this->allocated_prime_powers = DEFAULT_PRIME_POWERS_ALLOCATED;
	this->prime_powers = (prime_power_structure*) AbstractSieve::AllocationMachine((void **) &(this->prime_powers),(this->allocated_prime_powers)*sizeof(prime_power_structure));
	if (this->prime_powers == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating prime powers",0,this->allocated_prime_powers,"There was an error allocating prime powers");
		this->allocated_prime_powers = 0;
		return NOT_ENOUGH_MEMORY;
	}
	
	// We zeroize the whole structure array.
	memset(this->prime_powers,0,(this->allocated_prime_powers)*sizeof(prime_power_structure));	
	int i,j;
	int max_exp;
	double power;
	int p;
	int min_exp = this->factor_base[this->rough_factorbase_index];
	int index_in_fb;
	// Definition of the prime powers themselves.
	// We begin with prime on index 2 (first odd prime geq than 0)
	this->primes_with_powers = 0;
	this->powers_together = 0;
	for (i=0; i< this->allocated_prime_powers; i++){		
		index_in_fb = i+2;
		if ((index_in_fb) > this->maximal_factor_base_index) break; // we do not want to run out of the factor base
		p					   = this->factor_base[i+2];
		this->prime_powers[i].prime 		   = p;
		this->prime_powers[i].index_in_factor_base = index_in_fb;
		this->prime_powers[i].powers		   = NULL;
		// We do not want to sieve with prime powers which exceed 10*M (remember that 2*M+1 is the size of the sieving interval)
		max_exp = 0;
		power = 1; ;
		while ((power *= p) <= this->prime_power_limit) {		  	
			max_exp++;
		}				
		if (max_exp == 1) {	//	already the square of this prime is too big to be sieved with
			// We zeroize the prime to notify that this is the real end of the relevant data
			this->prime_powers[i].prime 		   = 0;	
			this->prime_powers[i].index_in_factor_base = 0;
			// ... and we break the cycle
			break;
		}
		// Number of really used primes with powers increases ...
		this->primes_with_powers++;		
		this->prime_powers[i].allocated_powers = max_exp - 1;				
		// We allocate the corresponding structures
		this->prime_powers[i].powers = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].powers,(this->prime_powers[i].allocated_powers)*sizeof(main_sieving_type));
		if (this->prime_powers[i].powers == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating prime powers II",i,this->prime_powers[i].allocated_powers,"There was an error allocating prime powers II");
			this->prime_powers[i].allocated_powers = 0;
			return NOT_ENOUGH_MEMORY;
		}
		this->prime_powers[i].roots_of_kN_mod_powers = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].roots_of_kN_mod_powers,(this->prime_powers[i].allocated_powers)*sizeof(main_sieving_type));
		if (this->prime_powers[i].roots_of_kN_mod_powers == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating roots of kN mod prime powers",i,this->prime_powers[i].allocated_powers,"There was an error allocating roots of kN mod prime powers");			
			return NOT_ENOUGH_MEMORY;
		}
		this->prime_powers[i].a_inverses = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].a_inverses,(this->prime_powers[i].allocated_powers)*sizeof(main_sieving_type));
		if (this->prime_powers[i].a_inverses == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating 'a' inverses mod prime powers",i,this->prime_powers[i].allocated_powers,"There was an error allocating 'a' inverses mod prime powers");			
			return NOT_ENOUGH_MEMORY;
		}
		// Beware: this is a 2D array whose second dimension is given by 's'. Value of 's' is not known at this moment;
		// allocation of rows must be done afterwards.
		if ((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs)) {
			this->prime_powers[i].Ba_inv = (main_sieving_type**) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].Ba_inv,(this->prime_powers[i].allocated_powers)*sizeof(main_sieving_type*));
			if (this->prime_powers[i].Ba_inv == NULL){
				CLog::LogError(WHEREARG,"Quadratic sieve","allocating Ba_inv mod prime powers",i,this->prime_powers[i].allocated_powers,"There was an error allocating Ba_inv mod prime powers");			
				return NOT_ENOUGH_MEMORY;
			}
			memset(this->prime_powers[i].Ba_inv,0,(this->prime_powers[i].allocated_powers)*sizeof(main_sieving_type*));
		}
				
		this->prime_powers[i].solutions_1 = (isol_type*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].solutions_1,(this->prime_powers[i].allocated_powers)*sizeof(isol_type));
		if (this->prime_powers[i].solutions_1 == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating solutions 1 mod prime powers",i,this->prime_powers[i].allocated_powers,"There was an error allocating solutions 1 mod prime powers");			
			return NOT_ENOUGH_MEMORY;
		}
		this->prime_powers[i].solutions_2 = (isol_type*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].solutions_2,(this->prime_powers[i].allocated_powers)*sizeof(isol_type));
		if (this->prime_powers[i].solutions_2 == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating solutions 2 mod prime powers",i,this->prime_powers[i].allocated_powers,"There was an error allocating solutions 1 mod prime powers");			
			return NOT_ENOUGH_MEMORY;
		}
		this->prime_powers[i].usable = (int*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].usable,(this->prime_powers[i].allocated_powers)*sizeof(int));
		if (this->prime_powers[i].usable == NULL){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating usable mod prime powers",i,this->prime_powers[i].allocated_powers,"There was an error allocating usable mod prime powers");			
			return NOT_ENOUGH_MEMORY;
		}
		// We initialize the powers array
		this->prime_powers[i].powers[0] = p*p;		
		for (j=1; j< this->prime_powers[i].allocated_powers; j++){			
			this->prime_powers[i].powers[j] = this->prime_powers[i].powers[j-1]*p;				
		}
		// We find out which of the powers is the first to be large enough to be sieved with
		this->prime_powers[i].least_index_to_sieve_with = 0;
		for (j = 0; j < this->prime_powers[i].allocated_powers; j++){
			if (this->prime_powers[i].powers[j] >= min_exp) break;
			this->prime_powers[i].least_index_to_sieve_with++;	
		}
		this->powers_together += (this->prime_powers[i].allocated_powers - this->prime_powers[i].least_index_to_sieve_with);
		/*printf("\nPrime Powers: prime %d. We have generated %d powers: ",p,this->prime_powers[i].allocated_powers);
		for (j=0;  j < this->prime_powers[i].allocated_powers; j++){
			printf(" %d",this->prime_powers[i].powers[j]);
		}
		printf(". Least index is %d.",this->prime_powers[i].least_index_to_sieve_with);*/
		// These are two forms of the logarithm, which will be added to corresponding location in sieving with almost
		// all the powers. Of course, this is equal to the logarithm of the prime itself.
		this->prime_powers[i].exact_standard_logarithm_to_use = this->exact_binary_logarithms_of_factor_base[index_in_fb];
		this->prime_powers[i].binary_standard_logarithm_to_use = this->binary_logarithms_of_factor_base_times_log_factor[index_in_fb];
		// The only exception can happen in case when 
		//		 either the prime itself, or its square is too small to be sieved with (lesser than rough factorbase index prime). 
		// In this situation, we must, in the first prime-power sieving iteration, add not only log p, but k*log p, to 
		// represent also the lower powers of p.
		this->prime_powers[i].exact_least_logarithm_to_use = this->exact_binary_logarithms_of_factor_base[index_in_fb];
		this->prime_powers[i].binary_least_logarithm_to_use = this->binary_logarithms_of_factor_base_times_log_factor[index_in_fb];
		// This is the first case; even p^2 is too small to be sieved with.
		if (this->prime_powers[i].least_index_to_sieve_with){
			this->prime_powers[i].exact_least_logarithm_to_use *= (this->prime_powers[i].least_index_to_sieve_with+2);	
			this->prime_powers[i].binary_least_logarithm_to_use *= (this->prime_powers[i].least_index_to_sieve_with+2);	
		} else {
			// p is too small to be sieved with, but p^2 is already big enough
			if (p < min_exp){
				this->prime_powers[i].exact_least_logarithm_to_use *= 2;	
				this->prime_powers[i].binary_least_logarithm_to_use *= 2;	
			}
		}		
	}	
	return ConstRC::Ok;
}

/*!
	This method calculates square root of kN mod all prime powers used for sieving. It is invoked
	if and only if sieving with prime powers is enabled.
	
	For this purpose, Shanks-Tonelli algorithm is used. If the Shanks-Tonelli fails at some prime power,
	it is considered a fatal failure and the computation ends.
	
	Returns: either error code of Shanks-Tonelli, or ConstRC::Ok.

*/
int  QuadraticSieve::CalculateRootsOfkNForPrimePowers() {
	if (this->info_mode){
		NormalMessage("   Calculating roots of kN mod prime powers ...",NO_NUMBER);
	}
	int i, j, ret;	
	prime_power_structure pps;
	mpz_t sol; mpz_t aux_p;
	mpz_init(sol);
	mpz_init(aux_p);
	for (i=0; i< this->primes_with_powers; i++){
		pps = this->prime_powers[i];
		mpz_set_ui(aux_p,pps.prime);		
		for (j=0; j < pps.allocated_powers; j++){			
			if (pps.prime != this->k) { // n = kN mod p
				ret = ShanksTonelliMultiexpMpz(sol, this->kN, aux_p, j+2);
				if (ret != ConstRC::Ok){
					printf("    Shanks-Tonelli: error. Return code %d.\n",ret);
					return ret;
				}			
				pps.roots_of_kN_mod_powers[j] = mpz_get_si(sol);
			} else { // p = k, the multiplier
				pps.roots_of_kN_mod_powers[j] = 0;
			}			
			//printf("Root of kN mod %d is %d.\n",pps.powers[j],pps.roots_of_kN_mod_powers[j]);
		}
	}
	mpz_clear(aux_p);
	mpz_clear(sol);
	return ConstRC::Ok;
}

/*!
	This method performs sieving with large prime powers (if sieving with large prime powers is
	enabled).
	
	Return codes: ConstRC::Ok only.
*/
int  QuadraticSieve::PerformSievingWithPrimePowersLogarithmic() {
    register log_type log;
    int i, j;    
    register main_sieving_type    p;    
    isol_type*     usable;
    isol_type*     sol_1;
    isol_type*     sol_2;
    log_type* 	   acc    =  this->main_matrix->accumulated_logarithm;
    log_type*      zero   =  acc - this->main_matrix->lower_arg;
    register log_type*      curr ;
    register log_type*      acc_max=  acc + this->main_matrix->maximal_index;
    //mpz_t logs_mpz;
    //mpz_init(logs_mpz);
    //unusables = 0;
    prime_power_structure pps;
    for (i=0; i< this->primes_with_powers; i++){
    	  //local_iters = 0;
      	  pps = this->prime_powers[i];
	  usable = pps.usable;
	  sol_1  = pps.solutions_1;
	  sol_2  = pps.solutions_2;
	  //printf("Sieving with prime powers: i = %d, usable = %p, sol_1 = %p, sol_2 = %p.\n",i,usable, sol_1,sol_2);
	  for (j=pps.least_index_to_sieve_with; j < pps.allocated_powers; j++){
	  	p = pps.powers[j];
	  	if (j != pps.least_index_to_sieve_with) log = pps.binary_standard_logarithm_to_use;
		else   log = pps.binary_least_logarithm_to_use;
		if (usable[j] & VALID_1){ 
			curr = zero+sol_1[j] - p;			
			while (1) {		
		  		if ((curr+=p) > acc_max) break;
				*(curr) += log;			 	           				
		 	}		 
		 	curr = zero+sol_1[j];		
		 	while (1) {		
		  		if ((curr-=p) < acc) break;
				*(curr) += log;			 	           				
		 	}
		}	
		if (usable[j] & VALID_2){ 
			curr = zero+sol_2[j] - p;
			
			while (1) {		
		  		if ((curr+=p) > acc_max) break;
				*(curr) += log;			 	           				
		 	}		 
		 	curr = zero+sol_2[j];		
		 	while (1) {		
		  		if ((curr-=p) < acc) break;
				*(curr) += log;			 	           				
		 	}
		}	
	  }
     } // end of while ((p=*fb++) != 0     
     return ConstRC::Ok;
}

/*!
	This method allocates space for square roots of kN modulo primes of the factor base. If this allocation fails,
	NOT_ENOUGH_MEMORY is returned; otherwise ConstRC::Ok.

*/
int QuadraticSieve::AllocRootsOfkN(){
	if (this->info_mode){
		NormalMessage("   Allocating roots of kN mod p ...",NO_NUMBER);
	}
	this->roots_of_kN_mod_p     = (main_sieving_type*) AllocationMachine((void**) &(this->roots_of_kN_mod_p),(this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	if (!this->roots_of_kN_mod_p){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating roots of kN mod p",this->maximal_factor_base_index+1,40,"There was an error allocating roots of kN mod p");
		return NOT_ENOUGH_MEMORY;	
	}
	return ConstRC::Ok;
}

integer_solutions& QuadraticSieve::Get_Solutions()
{
	return this->int_sols_of_cur_poly;
}

void    QuadraticSieve::Set_Solutions(integer_solutions aSolutions)
{
	Clear_Solutions();
	this->int_sols_of_cur_poly = aSolutions;
}

void    QuadraticSieve::Clear_Solutions()
{
	if (this->int_sols_of_cur_poly.allocated & ALLOCATED_1_SOLS){
		free(this->int_sols_of_cur_poly.sol_1);
		this->int_sols_of_cur_poly.sol_1       = NULL;
		this->int_sols_of_cur_poly.allocated &= ~ALLOCATED_1_SOLS;
	}
	if (this->int_sols_of_cur_poly.allocated & ALLOCATED_2_SOLS){
		free(this->int_sols_of_cur_poly.sol_2);
		this->int_sols_of_cur_poly.sol_2       = NULL;
		this->int_sols_of_cur_poly.allocated &= ~ALLOCATED_1_SOLS;
	}
	if (this->int_sols_of_cur_poly.allocated & ALLOCATED_USABLE_SOLS){
		free(this->int_sols_of_cur_poly.usable);
		this->int_sols_of_cur_poly.usable       = NULL;
		this->int_sols_of_cur_poly.allocated &= ~ALLOCATED_USABLE_SOLS;
	}
}

/*!
	For sieving, solutions of current polynomial modulo all primes of the factor base are needed. These solutions
	occupy some memory space; this memory space is allocated in this procedure.
	The solutions are contained in member variable
	
	this->int_sols_of_cur_poly
	
	respectively 3 arrays in this struct-variable:
	
	this->int_sols_of_cur_poly.sol_1
	
	this->int_sols_of_cur_poly.sol_2 (this expresses the fact that Q(x) mod p may have, and usually has, 2 solutions)
	
	this->int_sols_of_cur_poly.usable, which expresses whether solutions at given indices are valid or no (sometimes, Q(x) reduces to linear equation and then it has at most 1 solution mod p; therefore, the second
	solution is invalid).
	
	If any of the allocations fails, the method returns NOT_ENOUGH_MEMORY, else ConstRC::Ok.

*/
int QuadraticSieve::AllocSolutions(){
	if (this->info_mode){
		NormalMessage("   Allocating polynomial solutions ...",NO_NUMBER);
	}

	this->int_sols_of_cur_poly.allocated = 0;

	this->int_sols_of_cur_poly.sol_1     = (isol_type*) AllocationMachine((void**) &(this->int_sols_of_cur_poly.sol_1),(this->maximal_factor_base_index+1)*sizeof(isol_type));
	if (!this->int_sols_of_cur_poly.sol_1){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating integer solutions sol_1",this->maximal_factor_base_index+1,40,"There was an error allocating int solutions 1");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->int_sols_of_cur_poly.sol_1,0,(this->maximal_factor_base_index+1)*sizeof(isol_type));
	this->int_sols_of_cur_poly.allocated |= ALLOCATED_1_SOLS;

	this->int_sols_of_cur_poly.sol_2     = (isol_type*) AllocationMachine((void**)
&(this->int_sols_of_cur_poly.sol_2),(this->maximal_factor_base_index+1)*sizeof(isol_type));
	if (!this->int_sols_of_cur_poly.sol_2){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating integer solutions sol_2",this->maximal_factor_base_index+1,40,"There was an error allocating int solutions 2");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->int_sols_of_cur_poly.sol_2,0,(this->maximal_factor_base_index+1)*sizeof(isol_type));
	this->int_sols_of_cur_poly.allocated |= ALLOCATED_2_SOLS;
	
	this->int_sols_of_cur_poly.usable    = (isol_type*) AllocationMachine((void**) &(this->int_sols_of_cur_poly.usable),(this->maximal_factor_base_index+1)*sizeof(isol_type));
	if (!this->int_sols_of_cur_poly.usable){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating integer solutions usable",this->maximal_factor_base_index+1,40,"There was an error allocating int solutions usable");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->int_sols_of_cur_poly.usable,0,(this->maximal_factor_base_index+1)*sizeof(isol_type));
	this->int_sols_of_cur_poly.allocated |= ALLOCATED_USABLE_SOLS;	
	return ConstRC::Ok;
}

/*!
	In this method, three matrices needed for sieving process are allocated. The first one is the sieving matrix,
	which must contain 1 entry per each integer in the sieving interval. The other ones are "relation matrices",
	which contain so far found relations; they are always used for keeping the smooth and partial values in memory
	during the linear algebra phase. Moreover, if no intermediate files are used for keeping so far found relations,
	the relations found during sieving are kept in those matrices.
	
	Only the struct-s which make the matrices themselves are allocated, but no content; this is just a few bytes
	of memory, so a NOT_ENOUGH_MEMORY should not happen.  However, the possibility is still checked for.

*/
int QuadraticSieve::AllocMatrices(){
	if (this->info_mode){
		NormalMessage("   Allocating matrices ...",NO_NUMBER);
	}
	this->main_matrix = (sieve_matrix*) AbstractSieve::AllocationMachine((void**) &(this->main_matrix),sizeof(sieve_matrix));
	if (this->main_matrix == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating main matrix",0,70,"There was an error allocating main matrix");
		return NOT_ENOUGH_MEMORY;
	}
	this->main_matrix->accumulated_logarithm = NULL;
	this->main_matrix->assigned = 0;
	this->main_matrix->maximal_index = -1;
	this->smooth_values = (relation_matrix*) AbstractSieve::AllocationMachine((void**) &(this->smooth_values),sizeof(relation_matrix));
	if (this->smooth_values == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating smooth matrix",0,70,"There was an error allocating matrix for smooth values");
		return NOT_ENOUGH_MEMORY;
	}
	this->smooth_values->matrix = NULL;
	this->smooth_values->assigned_rows = 0;
	this->smooth_values->maximal_row_index = -1;
	return ConstRC::Ok;
}

/*!
	This method sets machine-specific divisors of 'a' in Carrier-Wagstaff SIQS mode. So far, it
	has not been implemented to work really; the generated values are constant and do not depend
	on machine number. Must be rewritten for Carrier-Wagstaff to work in parallel sieving mode (!!!).
	
*/
void QuadraticSieve::SetMachineSpecificDivisors(){
	int i;
	int decimals = mpz_sizeinbase(this->kN,10);
	this->machine_specific_divisors_number = 3;
	if (decimals < 50) 
        {
		this->machine_specific_divisors_number = 0;
	} 
        else if (decimals < 60) 
        {
		this->machine_specific_divisors_number = 1;
	} 
        else if (decimals < 70) 
        {
		this->machine_specific_divisors_number = 2;
	}
	//printf("Set Machine Specific Divisors: their # set to %d (decimals = %d).\n",this->machine_specific_divisors_number,decimals);
	this->mspecdivisors_log = 0;
	for (i=0; i< machine_specific_divisors_number; i++){
		machine_specific_divisors[i] = this->factor_base[DEFAULT_MACHINE_BOUND_END - 2+i];
		machine_specific_divisor_indices[i] = DEFAULT_MACHINE_BOUND_END - 2 + i;
		this->is_this_prime_divisor_of_a[DEFAULT_MACHINE_BOUND_END-2+i]     = 1;
		//printf("Machine specific divisor [%d]: %d\n",i,machine_specific_divisors[i]);
		this->mspecdivisors_log += this->exact_binary_logarithms_of_factor_base[DEFAULT_MACHINE_BOUND_END - 2+i];
	}
	this->iLowerMachineBound = factor_base[iLowerMachineBoundIndex];
	this->iUpperMachineBound = factor_base[DEFAULT_MACHINE_BOUND_END];
	this->iUpperMachineBoundIndex = DEFAULT_MACHINE_BOUND_END;	
}

/*!
	This method is only called in SIQS sieving mode.
	
	In order to determine 's', it divides the set of all 'a' divisors into "machine specific divisors", "even-indexed
	divisors" and "odd-indexed divisors", as described by Carrier-Wagstaff (only the procedure is more complicate, to
	avoid some practical traps). If the number of "odd-indexed divisors" is lower than 3, the Carrier-Wagstaff method
	would not work well; the calculation is therefore aborted and return code UNSUITABLE_FOR_SIQS is returned.
	
	Otherwise, ConstRC::Ok is returned.

*/
int  QuadraticSieve::SetSValue(){
	if (this->info_mode){
		NormalMessage("   Setting s value ...",NO_NUMBER);
	}
	this->machine_divisors_log_times_triples_log_factor = 0;
	// At first, we determine value of 0-3 small, "machine specific" divisors of 'a'. So far there is no
	// 'genuine' algorithm to do this, only a simple assignment (which is not "machine specific" at all).	
	if (!iMSDSpecifiedExternally)
	{
		SetMachineSpecificDivisors();				
	} else {
		// Here, we must set iUpper[Lower]MachineBoundIndex
		// and machine_specific_divisor_indices[i], because
		// iUpper[Lower]MachineBound and MSDs were specified
		// externally before construction of the factor base
		// took place
		iUpperMachineBoundIndex = GetTrueIndexInFB(&iUpperMachineBound);
		if (iUpperMachineBoundIndex == -1)
		{
			printf("Upper Machine Bound %d not found in the factor base!\n",iUpperMachineBound);
			return GENERAL_ERROR;
		}
		iLowerMachineBoundIndex = GetTrueIndexInFB(&iLowerMachineBound);
		if (iLowerMachineBoundIndex == -1)
		{
			printf("Lower Machine Bound %d not found in the factor base!\n",iLowerMachineBound);
			return GENERAL_ERROR;
		}
		int i;
		this->mspecdivisors_log = 0;
		int index;
		for (i=0; i< machine_specific_divisors_number; i++)
		{

			machine_specific_divisor_indices[i] = GetTrueIndexInFB(&(machine_specific_divisors[i]));
			index = machine_specific_divisor_indices[i];
			if (index == -1)
			{
				printf("Machine specific divisor %d not found in the factor base!\n",index);
				return GENERAL_ERROR;
			}
			this->is_this_prime_divisor_of_a[index] = 1;
			this->mspecdivisors_log += this->exact_binary_logarithms_of_factor_base[index];
		}
	}
	this->machine_divisors_log_times_triples_log_factor = (int) (this->mspecdivisors_log*((float) TRIPLES_LOG_FACTOR));
// 	printf("SetSValue: mspdlog %d\n",this->mspecdivisors_log);
	//printf("Machine specific divisors log, altered by a constant, is equal to: %d.\n",machine_divisors_log_times_triples_log_factor);
	// we want 'a' to be of roughly the same size as sqrt(2N)/M. 
	mpz_fdiv_q_ui(aux_1,this->kN,2);
	mpz_sqrt(aux_1,aux_1);
	mpz_fdiv_q(aux_1,aux_1,this->M);	
	mpz_set(this->ideal_a_value,aux_1);		
	mpz_pow_ui(aux_1,aux_1,TRIPLES_LOG_FACTOR);
	this->ideal_log_of_a_times_triples_log_factor = mpz_sizeinbase(aux_1,2); // This is a very rough calculation. Will be (once) replaced by a more exact method.
	
	//gmp_printf("Ideal value of 'a' is set to %Zd, its altered log to %d.\n",this->ideal_a_value,this->ideal_log_of_a_times_triples_log_factor);
	
	this->ideal_divisor_of_a = this->factor_base[CW_EVEN_INDEX_START + CW_EVEN_INDEX_SIZE / 2];
	float ideal_div_of_a_log = (double) log10(this->ideal_divisor_of_a) / (double) log10(2);
// 	printf("SetSValue: ideal log times %d\n",this->ideal_log_of_a_times_triples_log_factor);
// 	printf("SetSValue: mdlog     times %d\n",this->machine_divisors_log_times_triples_log_factor);
// 	printf("SetSValue: ideal dv of log %d\n",ideal_div_of_a_log);
	float first_s_assessment = (this->ideal_log_of_a_times_triples_log_factor - this->machine_divisors_log_times_triples_log_factor) / ((float) TRIPLES_LOG_FACTOR*ideal_div_of_a_log);
	//printf("First 's' assessment is %d / %6.2f = % 6.2f.\n",this->ideal_log_of_a_times_triples_log_factor-this->machine_divisors_log_times_triples_log_factor,((float) TRIPLES_LOG_FACTOR*ideal_div_of_a_log),first_s_assessment);
	int   first_s_assessment_int = (int) first_s_assessment;
	if ((first_s_assessment - (double) first_s_assessment_int) >= 0.5){
		first_s_assessment_int++;
	}
	this->s = first_s_assessment_int + this->machine_specific_divisors_number;
	//printf("Definite value of 's' was determined as %d.\n",this->s);	
	this->number_of_odd_index_divisors = this->s - this->machine_specific_divisors_number - 3;
	if (this->number_of_odd_index_divisors < 3){
		CLog::LogError(WHEREARG,"Quadratic sieve","determining 's' in SIQS",this->s,this->number_of_odd_index_divisors,"This number is too small for Carrier-Wagstaff SIQS, use plain mpqs (-q)");
		return UNSUITABLE_FOR_SIQS;
	}
	if (this->iDistributionMachineType == ENode)
	{
		// Machine specific divisors should be saved to job.xml.
		int ret = iSerializator->Serialize(EQsJob);
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
	}

	return ConstRC::Ok;
	/*
	int index_of_average_divisor = MACHINE_BOUND + A_BOUND/2;
	if (index_of_average_divisor > this->maximal_factor_base_index){
		// error, factor base too small!
	}
	
	float divisor_log = this->exact_binary_logarithms_of_factor_base[index_of_average_divisor];
	//printf("Ideal log of 'a' je %d; prumerny divisor je %d; jeho logaritmus je %6.2f; podil je %d;\n",this->ideal_log_of_a_times_triples_log_factor,this->factor_base[index_of_average_divisor],divisor_log,(this->ideal_log_of_a_times_triples_log_factor - machine_divisors_log_times_triples_log_factor) / divisor_log);
	
	// it is better to have smaller 'a'
	this->s = (int) ((this->ideal_log_of_a_times_triples_log_factor - machine_divisors_log_times_triples_log_factor) / (divisor_log*(float)TRIPLES_LOG_FACTOR)) + machine_specific_divisors_number;	
	*/
}

/*!
	This is an auxiliary method determining the true FB index of an even-indexed prime dividing 'a' from its index in 
	NexKSB-structure.
	It is needed because NexKSB uses symbolic numbers 0...n-1, and we need to know which prime in the FB corresponds	
	to this symbolic number.
	
	Return value: index of nexksb-indexed prime in the factor base.

*/
int  QuadraticSieve::TrueIndexEven(int nexksb_index){
	return (nexksb_index*2 + even_index_start - 2);
}

/*!
	This is an auxiliary method determining the true FB index of an odd-indexed prime dividing 'a' from its index in 
	NexKSB-structure.
	It is needed because NexKSB uses symbolic numbers 0...n-1, and we need to know which prime in the FB corresponds
	to this symbolic number.
	
	Return value: index of nexksb-indexed prime in the factor base.

*/
int  QuadraticSieve::TrueIndexOdd(int nexksb_index){
	return (nexksb_index*2 + odd_index_start - 2);
}

/*!
	This is an auxiliary method used for transforming NexKSB abstract representation of one ordered triple into the 
	representation used during sieving.

*/
void QuadraticSieve::CopyNexKSBToEvens(){
	int   index   = this->nx_env->subset[0];
	float acc_log = 0;
	int i;
	//printf("===================\nCopy nexKSB to evens, dalsi cast.");
	for (i=1; i<=3; i++){
		this->triples[index][i] = this->nx_env->subset[i];
		acc_log += this->exact_binary_logarithms_of_factor_base[TrueIndexEven(this->nx_env->subset[i])];
		//printf("Delitel: %5d, logaritmus: %6.2f, celkovy logaritmus: %6.2f.\n",this->factor_base[TrueIndexEven(this->nx_env->subset[i])],this->exact_binary_logarithms_of_factor_base[TrueIndexEven(this->nx_env->subset[i])],acc_log);
	}
	this->triples[index][0] = (int) (acc_log*(float) TRIPLES_LOG_FACTOR);
	/*if (!(index % 1000)) {
		printf("[%6d]: Log = %10d: factors %6d %6d %6d.\n",index,this->triples[index][0],this->triples[index][1],this->triples[index][2],this->triples[index][3]);
 	}*/
}

/*!
	This method calculates the logarithm of product of the current nexksb-n-tuple of primes. Used in Carrier-Wagstaff
	method of SIQS.
	
	Return value: the logarithm.

*/
int  QuadraticSieve::CalculateLogOfOdd(){
	int i;
	float acc_log = 0;
	for (i=1; i<= number_of_odd_index_divisors; i++){
		acc_log += this->exact_binary_logarithms_of_factor_base[TrueIndexOdd(this->nx_env->subset[i])];
		//printf("CalculateLogOfOdd: vynasobeno prvkem %d.\n",this->factor_base[TrueIndexOdd(this->nx_env->subset[i])]);
	}	
	int result = ((int) (acc_log*(float) TRIPLES_LOG_FACTOR));
	//printf("   Calculate Log of Odd skoncilo jako %d.\n",result);
	return result;
}

/*!
	This method calculates the logarithm of product of a sequence of primes in the factor base. This sequence can be
	"spaced" by "pause": e.g., taking every other prime. Used in environment settings for Carrier-Wagstaff.
	
	Return value: the logarithm.

*/
int  QuadraticSieve::CalculateLogOfSequence(int start_index, int length, int pause){
	int i;
	float acc_log = 0;
	for (i=0; i< length; i++){
		acc_log += this->exact_binary_logarithms_of_factor_base[start_index + i*pause];
	}
	int result = ((int) (acc_log*(float) TRIPLES_LOG_FACTOR));
	//printf("   Calculate Log of Odd skoncilo jako %d.\n",result);
	return result;
}

/*!
	This method tries to swap two pointers in an array of undefined type. It is extensively used in the quicksort algorithm.

*/
void QuadraticSieve::Swap(void** array, int index_1, int index_2){
	void* temp = array[index_1];
	array[index_1] = array[index_2];
	array[index_2] = temp;
}

/*!
	This recursive method quicksorts the triples of even-indexed primes dividing 'a'.
*/
void QuadraticSieve::Quicksort_Triples(int** array, int left, int right){
	int i, last;
	if (left >= right) return;
	Swap((void**) array,left,(left+right)/2);
	last = left;
	for (i=left+1; i<= right; i++){
		if (array[i][0] < array[left][0]){
			Swap((void**) array,++last,i);
		}
	}
	Swap((void**) array,left,last);
	Quicksort_Triples(array,left,last-1);
	Quicksort_Triples(array,last+1,right);
}

/*!
	This recursive method searches for the value in array[] between indices left and right, which is closest to 'what'.
	
	It is used in SIQS, when a new 'a' value is generated.

*/
int  QuadraticSieve::Binary_Search_Triples(int** array, int left, int right, int what){
	if (right == left) {		
		 return left;
	}
	int middle_index = (left+right)/2;
	int middle_value = array[middle_index][0];
	if (middle_value < what){
		return Binary_Search_Triples(array,middle_index+1,right,what);
	}
	else {
		return Binary_Search_Triples(array,left,middle_index,what);
	}
}

/*!
	When running Carrier-Wagstaff method of SIQS coefficient generation, it is desirable to have the odd-indexed and
	even-indexed intervals chosen suitably: so that the product of the smallest odd-indexed n-tuple and the machine-
	specific divisors, and the product of the greatest even-indexed 3-tuple are big enough to get an almost ideal value
	of 'a'; and symmetrically, this should work for the smallest even-indexed 3-tuple and largest odd-indexed n-tuple.
	Otherwise, many generated 'a' values would differ only in 1 divisor and frequency of duplicate relations would
	be great (proved experimentally).
	
	This method shifts the odd-indexed and even-indexed intervals in this sense.
	
	See Carrier-Wagstaff, page 6

*/

void QuadraticSieve::AdjustLowerAndUpperOddBound(){	
	/*// at first, we determine odd_index_center; it should be as close to the ideal 'a' divisor value, as possible
	if (this->ideal_divisor_of_a_index%2){
		this->odd_index_center = this->ideal_divisor_of_a_index;
	} else {
		// index of the closest value of ideal 'a' divisor is even; we pick the one of the neighbors, which is closer 
		if (AbsInt(this->ideal_divisor_of_a - this->factor_base[this->ideal_divisor_of_a_index-1]) <= AbsInt(this->ideal_divisor_of_a - this->factor_base[this->ideal_divisor_of_a_index+1])){
			this->odd_index_center = this->ideal_divisor_of_a_index-1;
		} else {
			this->odd_index_center = this->ideal_divisor_of_a_index+1;
		}
	}
	this->odd_index_start = this->odd_index_center - CW_ODD_INDEX_SIZE/2;
	this->odd_index_end   = this->odd_index_start + CW_ODD_INDEX_SIZE;
	printf("Adjust Lower and Upper Bound: we have at first determined odd_index_start as %d (p=%d), odd_index_end as %d (p=%d), odd_index_center as %d (p=%d).\n",odd_index_start,this->factor_base[odd_index_start],odd_index_end,this->factor_base[odd_index_end],odd_index_center,this->factor_base[odd_index_center]);
	// now, we determine the shift of the interval. Realize that the 'odd indexed part' must be, from the very first iteration, large enough to
	// generate at least the ideal value of a. This means that log of the smallest combination, with added logarithms of machine divisors and
	// the maximal possible logarithm of triples, must exceed this->ideal_log_of_a_times_triples_log_factor.
	int accepted_least_log = this->ideal_log_of_a_times_triples_log_factor - this->maximal_even_log - this->machine_divisors_log_times_triples_log_factor;
	printf("Accepted least log is %d, and we want to excess it.\n",accepted_least_log);
	int accepted_max_log = this->ideal_log_of_a_times_triples_log_factor - this->minimal_even_log - this->machine_divisors_log_times_triples_log_factor;
	int this_least_log = CalculateLogOfOdd();
	while (this_least_log < accepted_least_log) {
		this->odd_index_start += 2;
		this->odd_index_end   += 2;
		this->odd_index_center+= 2;
		this_least_log = CalculateLogOfOdd();
		printf("   New this least log: %d.\n",this_least_log);
	}
	// On the other hand, in the very first iteration, the 'odd indexed part' may not exceed limit given by the minimal logarithm of triples
	printf("Accepted max log is %d, and we want to get under it.\n",accepted_max_log);
	int this_max_log = CalculateLogOfOdd();	
	while (this_max_log > accepted_max_log) {
		this->odd_index_start -= 6;
		this->odd_index_end   -= 6;
		this->odd_index_center-= 6;
		this_max_log = CalculateLogOfOdd();
		printf("   New this max log: %d.\n",this_max_log);
	}
	printf("Adjust Lower and Upper Bound: we have finally determined odd_index_start as %d (p=%d), odd_index_end as %d (p=%d), odd_index_center as %d (p=%d).\n",odd_index_start,this->factor_base[odd_index_start],odd_index_end,this->factor_base[odd_index_end],odd_index_center,this->factor_base[odd_index_center]);
	//printf("Puvodni odd lower bound byla %d, posunuli jsme ji na %d. Soucasny this_least_log je %d, maximalni triple je %d, idealni log je %d",MACHINE_BOUND,this->odd_lower_bound, this_least_log,this->triples[this->triple_count-1][0],this->ideal_log_of_a_times_triples_log_factor);*/
	
	int start_index = this->iUpperMachineBoundIndex+1;
	int least_log;
	if (!(start_index%2)) start_index++;
	// At first, we determnine the lower bound; that is, the place where the product of odd-indexed primes is, for the first time, big enough to
	// exceed accepted_least_log;
	int accepted_least_log = this->ideal_log_of_a_times_triples_log_factor - this->maximal_even_log - this->machine_divisors_log_times_triples_log_factor;
	do {
		if ((start_index + 2*this->number_of_odd_index_divisors) >= this->maximal_factor_base_index) break;		
		if ((least_log = CalculateLogOfSequence(start_index,this->number_of_odd_index_divisors,2)) > accepted_least_log) break;
		start_index += 2;
		//printf("Shift of start index = %d: least log is %d, accepted least log is %d.\n",start_index,least_log,accepted_least_log);
	} while (1);
	//printf("We have finally determined start index as %d. Least log is %d. Accepted least log is %d\n",start_index,least_log,accepted_least_log);
	int end_index = start_index;
	int accepted_max_log = this->ideal_log_of_a_times_triples_log_factor - this->minimal_even_log - this->machine_divisors_log_times_triples_log_factor;
	do{	
		int max_log;
		if ((end_index + 2*(this->number_of_odd_index_divisors-1)) >= (this->maximal_factor_base_index)) { break; }
		if ((max_log = CalculateLogOfSequence(end_index,this->number_of_odd_index_divisors,2)) > accepted_max_log) { break; }
		//printf("Shift of end index = %d: max log is %d, accepted max log is %d.\n",end_index,max_log,accepted_max_log);
		end_index += 2;
	} while (1);
	//printf("We have finally determined end index as %d. Max log is %d. Accepted max log is %d.\n",end_index,max_log,accepted_max_log);
	this->odd_index_start = start_index;
	this->odd_index_end   = end_index + 2*(this->number_of_odd_index_divisors-1);
	this->odd_index_center = (this->odd_index_end-this->odd_index_start)/2;
	//printf("The span of odd-indexed primes is from %d (p=%d) to %d (p=%d).\n",this->odd_index_start,this->factor_base[this->odd_index_start],this->odd_index_end,this->factor_base[this->odd_index_end]);
}

/*!
	This method tries to determine the ideal span of even indices. We consider the centre of the even-indexed interval
	to be ideal, if the closest neighboring value to the left is smaller than ideal divisor of 'a', and the closest
	neighboring value to the right is greater than ideal divisor of 'a'.

*/
void QuadraticSieve::DetermineTriplesRange() {
	int i = this->iUpperMachineBoundIndex;
	if (i%2) i++; //set 'i' to even index
	do {
		if ((this->factor_base[i] <= this->ideal_divisor_of_a) && (this->factor_base[i+2] >= this->ideal_divisor_of_a)) {
			break;
		}
	} while ((i+=2) <= this->maximal_factor_base_index);
	// the loop was broken in the moment when ideal a divisor had been between FB[i] and FB[i+2]
	// determination of the closest FB-prime value on even index
	if (AbsInt(this->ideal_divisor_of_a - this->factor_base[i]) <= AbsInt(this->factor_base[i+2] - this->ideal_divisor_of_a)){
		this->even_index_center = i;
	} else {
		this->even_index_center = i+2;
	}
	this->even_index_start = this->even_index_center - CW_EVEN_INDEX_SIZE/2;
	this->even_index_end   = this->even_index_start  + CW_EVEN_INDEX_SIZE;
	//printf("Determine Triples Range: start index %d (p=%d), end index %d (p=%d).\n",this->even_index_start,this->factor_base[this->even_index_start],this->even_index_end,this->factor_base[this->even_index_end]);
	//printf("Centre of the interval is on index %d (p=%d). Neighboring primes are %d and %d.\n",this->even_index_center,this->factor_base[this->even_index_center],this->factor_base[this->even_index_center-1],this->factor_base[this->even_index_center+1]);
	//printf("Ideal 'a' divisor was chosen to be %d.\n",this->ideal_divisor_of_a);
	// determination of the closest FB-prime value on any index
	this->ideal_divisor_of_a_index = this->even_index_center;
	if (AbsInt(this->ideal_divisor_of_a - this->factor_base[this->even_index_center]) >= AbsInt(this->factor_base[this->even_index_center+1] - this->ideal_divisor_of_a)){
		this->ideal_divisor_of_a_index += 1;
	} else
	  if (AbsInt(this->ideal_divisor_of_a - this->factor_base[this->even_index_center]) >= AbsInt(this->factor_base[this->even_index_center-1] - this->ideal_divisor_of_a)){
	  	this->ideal_divisor_of_a_index -= 1;
	  }
	this->ideal_divisor_of_a_closest_value = this->factor_base[this->ideal_divisor_of_a_index];
	//printf("Closest value to the ideal divisor: %d, on index %d.\n",this->ideal_divisor_of_a_closest_value,this->ideal_divisor_of_a_index);	
}

/*!
	This method is only called in SIQS sieving mode.
	
	It takes care of allocation of data structures used in SIQS, namely arrays
	 - B2, plus initialization of all its mpz_t members
	 - B2_inited
	 - Ba_inv
	 - divisors_of_a
	 - indices_of_divisors_of_a
	 
	If any of the allocations fail, the method returns NOT_ENOUGH_MEMORY.
	
	Furthermore, this method takes care of the NexKSB-algorithm structures, as used by the Carrier-Wagstaff SIQS
	algorithm. It creates an auxiliary instance of NexKSB environment variable, and uses this instance to generate
	all triples of even-indexed primes (for details on even- and odd- indexed primes see the Carrier-Wagstaff article
	'Implementing the Hypercube Quadratic Sieve ....'). Then it sorts the outcoming set of triples (it should be 
	almost sorted already) by binary logarithms of their products. After that, the set of even-indexed triples is
	determined; now, the bounds of the odd-indexed interval must be determined so that the method would not generate
	many similar sets of 'a' divisors (see thesis of M.Kechlibar). This determination is performed by calling a third-
	level method; then, the NexKSB-structures are reset to serve the main purpose of generation of the odd-indexed
	n-tuples.
	
	Finally, if sieving with prime powers is enabled, corresponding structures Ba_inv are allocated as well.
	
	Return codes:
	- anything thrown by NexKSB_Init();
	- NOT_ENOUGH_MEMORY if any allocation fails.
	- ConstRC::Ok otherwise.

*/
int  QuadraticSieve::InitiateSiqsStructures() {
#ifdef SIQS_DEBUG	
	printf("this->s = %d\n",this->s);
#endif
	int i;
	if (this->info_mode){
		NormalMessage("   Initiating SIQS structures ...",NO_NUMBER);
	}
	// at first, we allocate B2
	this->B2 = (mpz_t*) AbstractSieve::AllocationMachine((void**) &(this->B2),this->s*sizeof(mpz_t));
	if (this->B2 == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating B2",s,71,"There was an error allocating array of B2's");
		return NOT_ENOUGH_MEMORY;
	}
	this->B2_inited = (int*) AbstractSieve::AllocationMachine((void**) &(this->B2_inited),this->s*sizeof(int));
	if (this->B2_inited == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating B2_inited",s,72,"There was an error allocating array of B2 inited's");
		return NOT_ENOUGH_MEMORY;
	}
	// no B2's are inited now
	memset(this->B2_inited,0,s*sizeof(int));	
	for (i=0; i< s; i++){
		mpz_init(this->B2[i]);
		this->B2_inited[i] = 1;
	}	
	int ret = AllocateBa_inv();
	if (ret != ConstRC::Ok)
	{
		return ret;
	}
	// Now, we allocate array for 'a' divisors (we must know 's' in order to do that)
	this->divisors_of_a = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &(this->divisors_of_a),(this->s)*sizeof(main_sieving_type));
	if (this->divisors_of_a == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating divisors of A",s,72,"There was an error allocating array of divisors of A");
		return NOT_ENOUGH_MEMORY;
	}	
	memset(this->divisors_of_a,0,(this->s)*sizeof(main_sieving_type));
	this->indices_of_divisors_of_a = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &(this->indices_of_divisors_of_a),(this->s)*sizeof(main_sieving_type));
	if (this->indices_of_divisors_of_a == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating indices_of_divisors of A",s,72,"There was an error allocating array of indices_of_divisors of A");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->indices_of_divisors_of_a,0,(this->s)*sizeof(main_sieving_type));	
	for (i=0; i< machine_specific_divisors_number; i++){
		this->divisors_of_a[i] = machine_specific_divisors[i];
		this->indices_of_divisors_of_a[i] = machine_specific_divisor_indices[i];
	}
	
	// now, we generate all triples of potential "end divisors" of 'a'. This needs knowledge of (binary) logarithms of corresponding primes (in
	// interval <even_index_start, even_index_end>
	
	// The next step is to determine how many triples will be there. 
	// The triples consist of primes with even indices in interval <even_index_start, even_index_end>.
	// There are precisely CW_EVEN_INDEX_SIZE such primes, starting with even_index_start.
	// For details see Carrier-Wagstaff: Implementing the Hypercube Quadratic Sieve ....
	mpz_t comb; mpz_init(comb);
	mpz_bin_uiui(comb,CW_EVEN_INDEX_SIZE,3);
	this->triple_count = mpz_get_ui(comb);
	
	this->triples = (int**) AbstractSieve::AllocationMachine((void**) &this->triples,(this->triple_count)*sizeof(int*));
	if (this->triples == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","allocating main triples array",this->triple_count,74,"There was an error allocating main triples array of A");
		return NOT_ENOUGH_MEMORY;
	}
	memset(this->triples,0,(this->triple_count)*sizeof(int*));
	for (i=0; i< this->triple_count; i++) {	
		this->triples[i] = (int*) AbstractSieve::AllocationMachine((void**) &this->triples[i],TRIPLE_SIZE*sizeof(int));		
		if (this->triples[i] == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating triples row",i,74,"There was an error allocating triples row.");
			return NOT_ENOUGH_MEMORY;
		}
	}
	ret = NEXKSB_Init(CW_EVEN_INDEX_SIZE,3);
	if (ret != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","initiating NEXKSB in triples",i,74,"There was an error in NEXKSB row.");
		return ret;
	}
	DetermineTriplesRange();
	CopyNexKSBToEvens();
	// save the first generated k-set	
	while (!this->nx_env->finished){		
		NEXKSB_Next();
		// save all other generated k-sets
		CopyNexKSBToEvens();
	}
	NEXKSB_Delete();
	// We sort the triples by their logarithm.
	Quicksort_Triples(this->triples,0,this->triple_count-1);		
	this->minimal_even_log = this->triples[0][0];
	this->maximal_even_log = this->triples[this->triple_count-1][0];
	//printf("The triples have been sorted. Now, their least logarithm is %d, and their maximal logarithm is %d.\n",this->minimal_even_log,this->maximal_even_log);
	// Calculation of the odd-indexed primes
	AdjustLowerAndUpperOddBound();
	// Now, we reset the NEXKSB structures in order to create the "odd"-tuples.	
	this->odd_index_size = (this->odd_index_end - this->odd_index_start)/2 + 1;
	ret = NEXKSB_Init(this->odd_index_size,this->number_of_odd_index_divisors);
	if (ret != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","initiating NEXKSB in triples",this->number_of_odd_index_divisors,74,"There was an error in NEXKSB row.");
		return ret;
	}
	mpz_bin_uiui(comb,this->odd_index_size,this->number_of_odd_index_divisors);
	this->odds_count = mpz_get_ui(comb);
	//printf("There will be altogether %d odd combinations\n",this->odds_count);
	// Average number of "odd prime combinations" exceeds 10^9. As lexicographic ordering quite well corresponds to absolute value of the
	// product of all array members, we must "adjust" the generated arrays a little. Otherwise it could happen that products of first 100 000 000
	// (or similar number of) combinations are too small to be used efficiently.	
	
	mpz_clear(comb);	
	if (this->info_mode) {
		printf(" %d triples generated.",this->triple_count);
	}
	// For prime power sieving
	if (this->power_sieving_mode == ESieveWithPrimePowers){
		int j, max;
		for (i=0; i< this->primes_with_powers; i++){	
			max = this->prime_powers[i].allocated_powers;
			for (j=0; j< max; j++){
				//printf("Chystame se k alokaci. i=%d, j=%d, ukazatel = %p.\n",i,j,this->prime_powers[i].Ba_inv[j]);
				//fflush(stdout);
				this->prime_powers[i].Ba_inv[j] = (main_sieving_type*) AbstractSieve::AllocationMachine((void**) &this->prime_powers[i].Ba_inv[j],(this->s)*sizeof(main_sieving_type));
				//printf("Po alokaci je ukazatel = %p.\n",i,j,this->prime_powers[i].Ba_inv[j]);
				//fflush(stdout);
				if (this->prime_powers[i].Ba_inv[j] == NULL){
					CLog::LogError(WHEREARG,"Quadratic sieve","allocating Ba_inv[j] mod prime powers",i,j,"There was an error allocating Ba_inv[j] mod prime powers");			
					return NOT_ENOUGH_MEMORY;
				}
			}
		}
	}
	return ConstRC::Ok;
}

/*!
	This method prints out the results of environment initialization. If listing of FB primes is enabled, it
	also prints out their list.

*/
void QuadraticSieve::PrintInformation(){
	if (this->frontend){
		fprintf(this->frontend_pipe, "Phase2>");
		fflush(this->frontend_pipe);
	}
	gmp_printf("\n\n");
	gmp_printf("The environment has been initialized as follows:\n");
	gmp_printf("------------------------------------------------\n");
	gmp_printf("The factored number is  :  %Zd\n",this->N);	 
	gmp_printf("The small coefficent    :  %d\n",this->k);
	gmp_printf("The input kN is then    :  %Zd\n",this->kN);
	gmp_printf("Decimal length of kN is :  %d\n",mpz_sizeinbase(this->kN,10));
	gmp_printf("Binary length of kN is  :  %d\n",mpz_sizeinbase(this->kN,2));
	gmp_printf("The sieving interval    :  %Zd\n",this->M);
	gmp_printf("We use # of polynomials :  %d\n", this->number_of_polynomials);
	gmp_printf("We use factorbase up to :  %d\n", this->upper_bound_on_factor_base);
	/*if (!this->without_primes_listing){
		int i;
		gmp_printf("The following primes are in factor base: ");
		for (i=0;i<=this->maximal_factor_base_index;i++){
			if (!(i%8)) gmp_printf("\n");
			printf("%10d",this->factor_base[i]);
		}
		gmp_printf("\n");
	}*/

	gmp_printf("Factorbase # elements   :  %d\n",this->maximal_factor_base_index+1);
	gmp_printf("Threshold elements      :  %d\n",this->threshold);	
	if (this->power_sieving_mode == ESieveWithPrimePowers){
	gmp_printf("Primes with high powers :  %d\n",this->primes_with_powers);
	gmp_printf("Higher powers used      :  %d\n",this->powers_together);
	}
	gmp_printf("Using Large Primes      :  ");
	if (this->using_variations == EDoNotUseVariations) {
		printf("None.\n");
	}
	int size = (cycle_table_1->AllocatedSize)*sizeof(hashtable_entry_type_1);
	int mb = size/1048576;
	int kb = (size - mb*1048576)/1024;
	if (this->using_variations == ELargePrimeV) {
		printf("Single.\n");		
		gmp_printf("   Max Single           :  %Zd\n",this->maximal_partial_factor);
		if (mb) {
			gmp_printf("   Hashtable Size       :  %d MB\n",mb);
		} else {
			gmp_printf("   Hashtable Size       :  %d kB\n",kb);
		}
	}
	if (this->using_variations == EDoubleLargePrimeV) {
		printf("Double.\n");
		gmp_printf("   Max Single           :  %Zd\n",this->maximal_partial_factor);
		gmp_printf("   Min Double           :  %Zd\n",this->minimal_candidate_value);
		gmp_printf("   Max Double           :  %Zd\n",this->maximal_candidate_value);
		if (mb) {
			gmp_printf("   Hashtable Size       :  %d MB\n",mb);
		} else {
			gmp_printf("   Hashtable Size       :  %d kB\n",kb);
		}
	}
	if ((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs)) {
	gmp_printf("Ideal value of 'a'      :  %Zd\n",this->ideal_a_value);
	gmp_printf("# 's' of divisors of 'a':  %d\n",this->s);	
	gmp_printf("# of machine-spec. div's:  %d\n",this->machine_specific_divisors_number);
	gmp_printf("Carrier-Wagstaff triples:  %d\n",this->triple_count);
	gmp_printf("Carrier-Wagstaff odds   :  %d\n",this->odds_count);
#if defined(__GNUC__)
	gmp_printf("GNU C compiler detected :  True\n");
#else
	gmp_printf("GNU C compiler detected :  False\n");
#endif
	    printf("\n");
	    printf("Libraries\n");
	    printf("LibXML2		    :  %s\n",LIBXML_DOTTED_VERSION);
	    printf("LibGMP		    :  %s\n",gmp_version);	   
	    // We must transform the original BOOST_LIB_VERSION
            // which uses underscores to the normal dotted version
	    char boostV[25];
	    memset(boostV,0,25);
	    strcat(boostV, BOOST_LIB_VERSION);
	    int i, max = strlen(boostV);
            for (i=0; i< max; i++)
	    {
		if (boostV[i] == '_') { boostV[i] = '.'; }
	    }
	    printf("LibBoost    	    :  %s\n",boostV);

	if (this->assertion_mode) {
            printf("Assertion mode          :  Active.\n");
	}
	if (ENone != this->iDistributionMachineType)
	{
	    printf("Distributed process     : ");
	    if (this->iDistributionMachineType == ECenter)
	    {
		printf(" Center.\n");
	    } 
	    if (this->iDistributionMachineType == ENode)
	    {
		printf(" Node.\n");
		printf("Commit interval         :  %d\n",iCommitInterval);
            }
	    
	}
	printf("\n");
	}
	
	TimeMessage("Init Environment",MAIN_LEVEL);
}

int     QuadraticSieve::Get_Solve() const
{
	return this->solve;
}

void    QuadraticSieve::Set_Solve(int aValue)
{
	this->solve = aValue;
}

// This method solves equation of shape (ax+b)^2 - kN = 0 mod p, p being a prime number from factor base.
// Values of 'a' and 'b' are saved in polynomials. 
// In case of SIQS, value 'a' does not change often, only 'b'; that means that only some shifting of
// previous results needs to be done.
/*!
	This is front-end method for solving polynomial equations (ax+b)^2 - kN = 0 mod p, p being a prime number from factor base.
	In case of MPQS, SolveQuadraticEquation() is run anytime, while in case of SIQS, it is run only after changing of 'a'
	coefficient; otherwise, ShiftSolutionsForSiqs() is called.
	
	Return values:
	- anything thrown by SolveQuadraticEquation()
	- anything thrown by ShiftSolutionsForSiqs()
	- ConstRC::Ok if everything went well.

*/
int QuadraticSieve::QuadraticMachine(){
	/*if (this->info_mode){
		NormalMessage("   Running quadratic machine ...",NO_NUMBER);
	}*/
	int ret;
	switch (this->QSMode){
		case EOldMpqs:
		case EMpqs:	ret = SolveQuadraticEquation();
				break;
		case EOldSiqs:
		case ESiqs:     if (this->solve) {		                
			        	//printf("Calling solve siqs()");		
					ret = SolveQuadraticEquation();
					this->solve = 0;
				} 
				else {
#ifdef SIQS_DEBUG			
					//printf("Calling ShiftSolutionsForSiqs()");		
#endif										
					// 2^{s-1} polynomials generated with use of SIQS
					ret = ShiftSolutionsForSiqs();					

				}
				break;
	}
	return ret;
}

/*!
	This method calculates modular inverses of \a (2a) for all odd primes in the factor base, and, if sieving with
	large prime powers is enabled, also for the prime powers. If \a (2a) is not invertible \a mod \a p, it puts 0 into
	the corresponding \a a_inverses entry.

*/
void QuadraticSieve::ComputeAInverses(){
	quad_polynomial* poly = &(this->polynomials[index_in_polynomial]);	
	main_sieving_type a_inv, p;	
	int  i, start = 2; // first odd number in FB		
	mpz_set(aux_a,poly->a);
	//gmp_printf("ComputeAInverses: index = %d, A = %Zd.\n",index_in_polynomial,poly->a);	
	mpz_mul_ui(aux_a,aux_a,2);
	for (i=start; i<=this->maximal_factor_base_index; i++){		
		if (!this->is_this_prime_divisor_of_a[i]){	
			p = this->factor_base[i];
			a_inv = 0;			
			mpz_set_ui(aux_p,p);			
			if (mpz_invert(aux_a_inv,aux_a,aux_p)){
				
				a_inv = mpz_get_ui(aux_a_inv);	
			} else { 
				a_inv = 0; 
			}		
			this->a_inverses[i] = a_inv;			
// 			printf("A_inverses = %d for %d.\n",a_inv,p);
			
			if (this->assertion_mode){
				int vysl = Mulm(a_inv,mpz_mod_ui(aux_b,aux_a,p),p);
				//printf("Pri vynasobeni %d x %d mod %d dostavame %d\n",a_inv,mpz_mod_ui(aux_b,aux_a,p),p,vysl);
				if ((vysl != 1) && (vysl != 0)){
					printf("Chyba! Pri vypoctu inverze z 2A jsme dostali spatny vysledek. 2A mod p: %lu, spocitany inverz: %d, vysledek nasobeni mod %d: %d.\n",mpz_mod_ui(aux_b,aux_a,p),this->a_inverses[i],p,vysl);
				
				}
			}
		} else { // delitel
			this->a_inverses[i] = 0;
// 			printf("A_inverses delitel.\n",a_inv);
		}
	}
// 	printf("Compute A inverses finished.\n");
// 	wait();
	// Large prime powers
	if (this->power_sieving_mode == ESieveWithPrimePowers){
		int j;
		prime_power_structure pps;
		for (i=0; i< primes_with_powers; i++){
			pps = this->prime_powers[i];
			for (j=0; j< pps.allocated_powers; j++){
				p = pps.powers[j];
				mpz_set_ui(aux_p,p);
				if (mpz_invert(aux_a_inv,aux_a,aux_p)){
					a_inv = mpz_get_ui(aux_a_inv);
				} else  a_inv = 0;
				pps.a_inverses[j] = a_inv;
			}
		}
	}
}

void CalculateAverages(mpz_t res, int x, int y, mpz_t a, mpz_t b, mpz_t kN){
	mpz_t x1, x2, x3, out;
	mpz_init(x1);
	mpz_init(x2);
	mpz_init(x3);
	mpz_init(out);
	mpz_set_si(x1,y);
	mpz_set_si(x2,x);
	mpz_sub(x1,x1,x2);
	mpz_add_ui(x1,x1,1); // x1 = y-x+1
	mpz_mul_ui(x2,x1,2); // x2 = 2y - 2x +2
	mpz_sub_ui(x2,x2,1); // x2 = 2y - 2x +1
	mpz_mul(out,x1,x2);
	mpz_fdiv_q_ui(out,out,6);
	mpz_mul_si(x3,x1,x);
	mpz_add(out,out,x3);
	mpz_set_si(x3,x);
	mpz_mul(x3,x3,x3);
	mpz_add(out,out,x3);
	mpz_mul_ui(out,out,4);
	mpz_mul(out,out,a);
	mpz_mul(out,out,a); //4a^2(x^2 + x(y-x+1) + (y-x+1)(2y-2x+1)/6)
	mpz_fdiv_q_ui(x1,x1,2);
	if (x > 0) mpz_add_ui(x1,x1,x); else mpz_sub_ui(x1,x1,-x);
	mpz_mul_ui(x1,x1,4);
	mpz_mul(x1,x1,a);
	mpz_mul(x1,x1,b);
	mpz_add(out,out,x1);
	mpz_sub(out,out,kN);
	mpz_mul(x1,b,b);
	mpz_add(out,out,x1);
	mpz_fdiv_q(x1,out,a);
	mpz_fdiv_q_ui(x1,x1,4);
	gmp_printf("Na intervalu [%d,%d] ma polynom Q(x) prumernou hodnotu o delce %d, po odecteni vlivu 4a pak %d.\nJeji presne vyjadreni je %Zd\n",x,y,mpz_sizeinbase(out,2),mpz_sizeinbase(out,2)-mpz_sizeinbase(a,2)-2,x1);
 //+ 4ab(x+(y-x+1)/2) + b^2 - kN
 	mpz_set(res,x1);
	mpz_clear(x1);
	mpz_clear(x2);
	mpz_clear(x3);
	mpz_clear(out);
}

void QuadraticSieve::PolyCopy(quad_polynomial* aTarget)
{
	if (!(aTarget->inited & A_INITED)){
		mpz_init(aTarget->a);
		aTarget->inited |= (A_INITED);
	}
	if (!(aTarget->inited & B_INITED)){
		mpz_init(aTarget->b);
		aTarget->inited |= (B_INITED);
	}
	if (!(aTarget->inited & C_INITED)){
		mpz_init(aTarget->c);
		aTarget->inited |= (C_INITED);
	}
	if (!(aTarget->inited & D_INITED)){
		mpz_init(aTarget->d);
		aTarget->inited |= (D_INITED);
	}
	if (!(aTarget->inited & D2_INITED)){
		mpz_init(aTarget->d2);
		aTarget->inited |= (D2_INITED);
	}
	quad_polynomial* poly = &(this->polynomials[this->index_in_polynomial]);
	mpz_set(aTarget->a,poly->a);
	mpz_set(aTarget->b,poly->b);
	mpz_set(aTarget->c,poly->c);
	mpz_set(aTarget->d,poly->d);
	mpz_set(aTarget->d2,poly->d2);
}

int  QuadraticSieve::Get_Subtract() const
{
	return this->subtract;
}

int  QuadraticSieve::Get_Nu() const
{
	//printf("Quadratic Sieve: called Get_Nu, returning %d.\n",this->nu);
	return this->nu;
}

/*!
	This is front-end method for determining a new sieving polynomial. It separates the second-level method PerformSieving
	from need to know whether we run MPQS or SIQS sieving.
	
	In case of MPQS, it simply calls CreateNewMpqsPolynomial(). In case of SIQS, it behaves differently depending
	on the index of the next polynomial. If there is need to create new 'a' coefficient, it runs CreateNewSiqsPolynomial(),
	and otherwise it runs RecalculateBForSiqs().
	
	Return values: ConstRC::Ok only.

*/
int QuadraticSieve::GenerateNewPolynomial(){
	/*if (this->info_mode){
		NormalMessage("   Generating new polynomial ...",NO_NUMBER);
	}*/	
	long j = this->current_polynomial;
	long i = this->index_in_polynomial;
	long last = i-1;
	if (last ==-1) last = this->allocated_polynomials-1;
	quad_polynomial* poly;
	poly = &(this->polynomials[i]);
	switch (this->QSMode){		
		case EOldMpqs:
		case EMpqs:
				if (j!= 0) mpz_set(poly->d,this->polynomials[last].d);				
				CreateNewMpqsPolynomial();				
				break;
		case EOldSiqs:
		case ESiqs:	if (((this->index_in_polynomial) % (1 << (s-1))) == 0){
#ifdef SIQS_DEBUG
				printf("Calling CreateNewAForSiqs:\n");
#endif		
				
					CreateNewSiqsPolynomial();
					this->solve = 1;
				} else {
#ifdef SIQS_DEBUG
				printf("Calling RecalculateBForSiqs:\n");
#endif						
					RecalculateBForSiqs();				
				}				
				if (this->iUseRootShort)
				{
					// calculate and print roots
					mpz_t root_1;
					mpz_init(root_1);
					mpz_sub(root_1,this->root,poly->b);
					mpz_fdiv_q(root_1,root_1,poly->a);
					mpz_fdiv_q_ui(root_1,root_1,2);
					this->root_2 = mpz_get_si(root_1); // > 0		
#ifdef SIQS_DEBUG				
					gmp_printf("Koren 1 je roven %Zd.\n",root_1);
#endif
					mpz_set(root_1,this->root);
					mpz_mul_si(root_1,root_1,-1);
					mpz_sub(root_1,root_1,poly->b);
					mpz_fdiv_q(root_1,root_1,poly->a);
					mpz_fdiv_q_ui(root_1,root_1,2);				
					this->root_1 = mpz_get_si(root_1); // < 0
				/*printf("\n");
				mpz_t res1,res2,res3; mpz_init(res1); mpz_init(res2); mpz_init(res3);
				CalculateAverages(res1,-this->integer_M,this->root_1, poly->a, poly->b, this->kN);
				CalculateAverages(res2,this->root_1,this->root_2, poly->a, poly->b, this->kN);
				CalculateAverages(res3,this->root_2,this->integer_M, poly->a, poly->b, this->kN);		
				//gmp_printf("Soucin %Zd a %d dava ... ",res1,this->root_1+this->integer_M);
				mpz_mul_ui(res1,res1,this->root_1+this->integer_M);
				//gmp_printf("%Zd\n",res1);
				//gmp_printf("Soucin %Zd a %d dava ... ",res2,this->root_1 - this->root_2);
				mpz_mul_si(res2,res2,this->root_1 - this->root_2); // change of sign
				//gmp_printf("%Zd\n",res2);
				//gmp_printf("Soucin %Zd a %d dava ... ",res3,this->integer_M - this->root_2);
				mpz_mul_si(res3,res3,this->integer_M - this->root_2);
				//gmp_printf("%Zd\n",res3);				
				mpz_add(res1,res1,res2);
				//gmp_printf("Soucet prvnich dvou hodnot je %Zd.\n",res1);
				mpz_add(res1,res1,res3);
				//gmp_printf("Soucet vsech tri hodnot je %Zd.\n",res1);
				mpz_fdiv_q_ui(res1,res1,2*this->integer_M);
				gmp_printf("Na celem intervalu je prumerna absolutni hodnota polynomu rovna %Zd.\n",res1);
				mpz_clear(res1); mpz_clear(res2); mpz_clear(res3);		*/
#ifdef SIQS_DEBUG												
					gmp_printf("Koren 2 je roven %Zd.\n",root_1);
#endif						
					mpz_clear(root_1);
				} // iUseRootShort
				// at first,
				break;
	}
	// Count the roots
	if (this->value_distribution_mode) {
		mpz_sub(aux_1,this->root,poly->b);
		mpz_fdiv_q(aux_1,aux_1,poly->a);
		mpz_fdiv_q_ui(aux_1,aux_1,2);
		// Now we have the positive root
		root_1 = mpz_get_si(aux_1);			
		mpz_set(aux_1,this->root);
		mpz_mul_si(aux_1,aux_1,-1);
		mpz_sub(aux_1,aux_1,poly->b);
		mpz_fdiv_q(aux_1,aux_1,poly->a);
		mpz_fdiv_q_ui(aux_1,aux_1,2);
		// Now we have the negative root
		root_2 = mpz_get_si(aux_1);		
	}	
	/*mpz_mul(aux_1,poly->b,poly->b);
	mpz_mul(aux_2,poly->a,poly->c);
	mpz_mul_si(aux_2,aux_2,-4);
	mpz_add(aux_1,aux_1,aux_2); // now aux_1 = kN
	mpz_sqrt(aux_1,aux_1);
	mpz_sub(aux_1,aux_1,poly->b);
	mpz_fdiv_q_2exp(aux_1,aux_1,1);
  	mpz_fdiv_q(aux_1,aux_1,poly->a);*/
  	
	return ConstRC::Ok;
}

bool QuadraticSieve::Check_LastSolutions()
{	
	isol_type* sol_1       = this->int_sols_of_cur_poly.sol_1;
	isol_type* sol_2       = this->int_sols_of_cur_poly.sol_2;
	main_sieving_type*  fb = this->factor_base;
	int max = this->maximal_factor_base_index;
	int correct = 0;
	bool result = TRUE;
	if (this->assertion_mode)
	{
		int last_index = index_in_polynomial - 1;
		if (last_index == -1)
		{
			printf("Check_LastSolutions ... last index is -1?");
		} else {
			index_in_polynomial--;
			isol_type* usable       = this->int_sols_of_cur_poly.usable;
	
			for (register int i=2; i<= max; i++){
				register main_sieving_type p = fb[i];					
				if (usable[i] & VALID_1) {
					/*if (i == 5)
					{
						printf("i=5,p=%d,sol_1=%d.\n",p,sol_1[i]);
						printf("Index in poly is %d.\n",index_in_polynomial);
						printf("Evaluate is %d.\n",Evaluate(sol_1[i],p));
						wait();
					}*/
					if (Evaluate(sol_1[i],p))	
					{
						printf("Check_LastSolutions: not ok sol_1 mod %d. Index in poly is %ld.\n",p,index_in_polynomial);
						result = FALSE;
					} else correct++;
				}
				if (usable[i] & VALID_2) {
					if (Evaluate(sol_2[i],p))	
					{
						printf("Check_LastSolutions: not ok sol_2 mod %d.\n",p);
						result = FALSE;
					} else correct++;
				}
			}
// 			printf("Check_LastSolutions finished\n");
			index_in_polynomial++;
		}
	}
// 	printf("Check_LastSolutions: correct %d.\n",correct);
	return result;
}

//#define SHIFT_DEBUG

/*!
	In SIQS, "heavy" polynomial change involving \a 'a' inversion happens only scarcely; most of the polynomial
	changes are "light", where only \a 'b' is being changed. In this situation, solutions of Q(x) are only shifted
	with use of normal integer addition or subtraction.
	This method performs recalculations of Q(x) solutions during such "light" polynomial change. It is heavily
	optimized for performance, since this is one of the time-critical operations in SIQS sieving.
	
	Return value: ConstRC::Ok.

*/
int  QuadraticSieve::ShiftSolutionsForSiqs(){
	// This method is called for all SIQS polynomials save those, whose A is 'freshly generated' (that means, used for the very first time in Ax^2+Bx+C)
	quad_polynomial* new_poly;
	isol_type* sol_1       = this->int_sols_of_cur_poly.sol_1;
	isol_type* sol_2       = this->int_sols_of_cur_poly.sol_2;
	main_sieving_type** Ba = this->Ba_inv;
	main_sieving_type*  fb = this->factor_base;
	new_poly = &(this->polynomials[this->index_in_polynomial]);
	quad_polynomial* old_poly;
	old_poly = &(this->polynomials[this->index_in_polynomial-1]);
	register int i = 1;
	register int n = this->nu;
	int max = this->maximal_factor_base_index;
	register main_sieving_type p;
	main_sieving_type minus_c;
//#define SIQS_RECALC_DEBUG
	/* This is heavily optimized method for recalculation of polynomial roots. The main optimizations are:
	  -  the condition test "subtract vs. add" is placed outside the 'for' loop to be performed only once;
	     this results in need to have 2 almost identical loops
	  -  the loop is done for all primes in FB, without testing whether a given prime divides 'a' coefficient
	     or no; the 'a' divisors are then recalculated afterwards. This saves a condition test whether p divides 'a'.
	     inside the loop.
	  -  usage of local variables sol_1, sol_2, Ba and max removes the need to access member variables of the class.
	     As this happens 500 000 times per second, this is an important operation
	  -  usage of operation % (modulo) is deprecated. Instead, we expect that solutions from last iteration are in range
	     0 to p-1 and perform only a single addition or subtraction to get them back into this range after being altered.
	     This means that the method responsible for creating the initial solutions (create new 'a' for siqs) must initialize
	     them well.*/
	Check_LastSolutions();	
	if (!subtract){
		for (i=2; i<= max; i++){
			p = fb[i];		
#ifdef SIQS_RECALC_DEBUG	
			printf("SIQS Subtract, p = %7d, root_1 = %7d, root_2 = %7d, Ba_inv = %d, nu = %d\n",p,sol_1[i],sol_2[i],Ba[i][n],n);
#endif			
			if ((sol_1[i] -= Ba[i][n]) < 0){
				sol_1[i] += p;
			}			
			if ((sol_2[i] -= Ba[i][n]) < 0){
				sol_2[i] += p;
			}			
#ifdef SIQS_RECALC_DEBUG	
			printf("   After: p = %7d, root_1 = %7d, root_2 = %7d.\n\n",p,sol_1[i],sol_2[i]);
#endif			

		}
	}
	else { // case of adding Bainv_i to solutions
		for (i=2; i<= max; i++){		
			p = fb[i];     			 			
#ifdef SIQS_RECALC_DEBUG	
			printf("SIQS Add, p = %7d, root_1 = %7d, root_2 = %7d, Ba_inv = %d, nu = %d.\n",p,sol_1[i],sol_2[i],Ba[i][n],n);
#endif			

			if ((sol_1[i] += Ba[i][n]) >= p){
				sol_1[i] -= p;
			}			
			if ((sol_2[i] += Ba[i][n]) >= p){
				sol_2[i] -= p;
			}			
#ifdef SIQS_RECALC_DEBUG	
			printf("   After: p = %7d, root_1 = %7d, root_2 = %7d.\n",p,sol_1[i],sol_2[i]);
#endif			

		}	
	}	
	int j;
	for (i=0; i< this->s; i++){
		p = this->divisors_of_a[i];
		j = this->indices_of_divisors_of_a[i];
		minus_c = mpz_fdiv_ui(new_poly->c,p);
		minus_c = p-minus_c; // -c;
		CalculateSolutionForADivisor(j, p,new_poly->b,minus_c);
	}
	if (this->assertion_mode)
	{
		int correct = 0;
		isol_type* usable       = this->int_sols_of_cur_poly.usable;
		main_sieving_type res;
		for (i=2; i<= max; i++)
		{	
			p = fb[i];
			if (usable[i] & VALID_1)
			{
				res = Evaluate(sol_1[i],p);
				if (res)
				{
					printf("Shift solutions for SIQS failed: mod %d, %d, ba_inv[%d,%d]: %d\n",p,res,i,n,Ba[i][n]);
					wait_for_Enter();
				} else correct++;
			}
			if (usable[i] & VALID_2)
			{
				res = Evaluate(sol_2[i],p);
				if (res)
				{
					printf("Shift solutions for SIQS failed: mod %d, %d,ba_inv: %d\n",p,res,Ba[i][n]);
					wait_for_Enter();
				} else correct++;
			}
// 			if (p == 7)
// // 			{
// 				printf("Sol_1: %d, sol_2: %d for p == 7.\n",sol_1[i],sol_2[i]);
// 			}
		}
// 		printf("Shift SIQS Solutions: Correct: %d.\n",correct);
	}		
	// sieving with prime powers (since 1.15)
	if (this->power_sieving_mode == ESieveWithPrimePowers) {
		prime_power_structure pps;
		int* usable;
		if (!subtract){
			for (i=0; i< this->primes_with_powers; i++){
				pps = this->prime_powers[i];			
				Ba = pps.Ba_inv;
				sol_1  = pps.solutions_1;
				sol_2  = pps.solutions_2;
				usable = pps.usable;
				for (j=0; j< pps.allocated_powers; j++){
					p = pps.powers[j];
					if (usable[j] & VALID_1) {
						if ((sol_1[j] -= Ba[j][n]) < 0){
							sol_1[j] += p;
						}
					}
					if (usable[j] & VALID_2) {
						if ((sol_2[j] -= Ba[j][n]) < 0){
							sol_2[j] += p;
						}
					}
				}
			}
		} else { // case of adding Bainv_i to solutions
			for (i=0; i< this->primes_with_powers; i++){
				pps = this->prime_powers[i];			
				Ba = pps.Ba_inv;
				sol_1  = pps.solutions_1;
				sol_2  = pps.solutions_2;
				usable = pps.usable;
				for (j=0; j< pps.allocated_powers; j++){
					p = pps.powers[j];
					if (usable[j] & VALID_1) {
						if ((sol_1[j] += Ba[j][n]) >= p){
							sol_1[j] -= p;
						}
					}
					if (usable[j] & VALID_2) {
						if ((sol_2[j] += Ba[j][n]) >= p){
							sol_2[j] -= p;
						}
					}
				}
			}
		}
	}
	return ConstRC::Ok;
}

mpz_t*	QuadraticSieve::Get_B2Block()
{
	return B2;
}

void  	QuadraticSieve::Clear_B2Block()
{
	if (B2)
	{
		int i;
		for (i=0; i< this->s; i++)
		{
			if (B2_inited[i])
			{
				mpz_clear(B2[i]);
				B2_inited[i] = 0;
			}
		}
	}
	free(B2);	
	free(B2_inited);
	B2 = NULL;
	B2_inited = NULL;
}

void 	QuadraticSieve::Set_B2Block(mpz_t* aB2Block, int* aB2_Inited)
{
	if (B2)
	{
		Clear_B2Block();
	}
	B2 = aB2Block;	
	B2_inited = aB2_Inited;
}

main_sieving_type* QuadraticSieve::Get_DivisorsOfA()
{
	return divisors_of_a;
}

void	QuadraticSieve::Set_DivisorsOfA(main_sieving_type* aDivisorsOfA)
{
	if (divisors_of_a)
	{
// 		printf("QS: cleaning previous divisors of 'a'.\n");
		free(divisors_of_a);
		divisors_of_a = NULL;
	}
// 	printf("QS: setting new divisors of a of 'a'.\n");
	divisors_of_a = aDivisorsOfA;
// 	printf("QS: before memset.\n");
	if (indices_of_divisors_of_a)
	{
		free(indices_of_divisors_of_a);
		indices_of_divisors_of_a = NULL;
	}
	indices_of_divisors_of_a = (main_sieving_type*) AllocationMachine((void**) &indices_of_divisors_of_a,(this->s)*sizeof(main_sieving_type));

	if (is_this_prime_divisor_of_a)
	{
		free(is_this_prime_divisor_of_a);
		is_this_prime_divisor_of_a = NULL;
	}
	is_this_prime_divisor_of_a = (main_sieving_type*) AllocationMachine((void**) &is_this_prime_divisor_of_a,(this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	
	memset(indices_of_divisors_of_a,0,(this->s)*sizeof(main_sieving_type));
	memset(this->is_this_prime_divisor_of_a,0,(this->maximal_factor_base_index+1)*sizeof(main_sieving_type));
	int i;
	for (i=0; i< this->s; i++)
	{
		// Set the indices_of_divisors+of+a correctly
		int  index = GetTrueIndexInFB(&divisors_of_a[i]);
		this->indices_of_divisors_of_a[i]        = index;
		this->is_this_prime_divisor_of_a[index]  = 1;
// 		printf("The index of %d in the FB is %d.\n",divisors_of_a[i],index);
	}
}


int  QuadraticSieve::ReloadSieverState()
{
	if (!(iLoadFlags & SIEVING_LOAD_FLAG))
	{
		return ConstRC::Ok;
	}
	if (!iSerializator)
	{
		CLog::LogError(WHEREARG,"Quadratic sieve","Reload Siever State ",0,0,"No Serializator instance available");
		return GENERAL_ERROR;
	};
	int ret = iSerializator->Deserialize(EQsSieverState);
	if (ret != ConstRC::Ok)
	{
		printf("Reload Siever State aborted.\n");
		return ret;
	}
	if ((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs))
	{
		// Recalculate indices of divisors of 'a' in the FB.
		// Refill is_this_prime_divisor_of_a
		// This has been done through Set_DivisorsOfA
		//printf("Recalc inverses of a.\n");
		// Recalculate a_inverses
		// We must shift the index in polynomial back for a moment
		this->index_in_polynomial--;
		if (this->index_in_polynomial >= 0)
		{
			ret = RefillInverses_Of_A(TRUE /* recalculate the inverses */);
			if (ret != ConstRC::Ok)
			{
				return ret;
			}	
		}			
		this->index_in_polynomial++;
		//printf("Recalc auxiliary siqs values.\n");
		// Recalculate Ba_inv.
		ret = RecalculateAuxiliarySiqsValues();
		if (ret != ConstRC::Ok)
		{
			return ret;
		}			

		// Recalculate Carrier-Wagstaff triples.
		
	}
	// Reconstruct the hashtable.
	if ((this->using_variations == ELargePrimeV) || (this->using_variations == EDoubleLargePrimeV))
	{
		ret = SetFactoringStructures();
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
		ret = ReconstructHashtable();
		if (ret != ConstRC::Ok)
		{
			return ret;
		}
	}
	// Nondestructive opening of the files
	ret = OpenFiles(APPEND);
	if (ret != ConstRC::Ok)		 
	{ 
		return ret; 
	}
	return ConstRC::Ok;
}

int  QuadraticSieve::ReconstructHashtable()
{	
	qs_relation* candidate = InitRelation();
	if (candidate == NULL)
	{
		return NOT_ENOUGH_MEMORY;
	}
	int x = 0;
	ResetHashtable();
	this->partials_file 	= OpenFile(this->partial_values_file_name,READ);	
	while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
		x++;
		printf("\rReconstructing hashtable %8d ...",x);
		fflush(stdout);
		AddToCycleCountStructures(candidate);
	}
	this->independent_cycles = this->edges + this->components - this->vertices;
	printf("\rHashtable reconstructed. Altogether %d partials read, forming %d independent cycles.",x,this->independent_cycles);
	CloseFile(this->partial_values_file_name);
	ClearRelation(candidate);
	return ConstRC::Ok;
}

long  QuadraticSieve::Get_MaxFBIndex() const
{
	return this->maximal_factor_base_index;
}

int QuadraticSieve::RecalculateAuxiliarySiqsValues()
{
	int i;
	FreeBa_inv();
	AllocateBa_inv();
	
	main_sieving_type inv_a;
	main_sieving_type B2_mod_p;
// Now, for each B_i and each p_j from FB that does not divide A, we must calculate 2B_j (2a)^{-1} mod p and store into a corresponding location.
	main_sieving_type* poi = this->factor_base;
	while (*(++poi)){ // The end of factor base is signalled by zero value. The first factor base element we want to examine is probably 2.
		if (!this->is_this_prime_divisor_of_a[poi - this->factor_base]){ //only for primes not dividing A ...
		//if ((poi < this->start_of_qi) || (poi >= this->end_of_qi)){ 
			inv_a = a_inverses[poi-this->factor_base];
			//printf("inv_a %d, %d.\n",inv_a,*poi);
			main_sieving_type* ba_inv = NULL;
			if (this->QSMode == ESiqs) 
			{
				if (this->advanced_factor_base != NULL)
				{
					ba_inv = this->advanced_factor_base[poi-this->factor_base].ba_inv;
					
				}
			}
			for (i=0; i<this->s; i++){				
				B2_mod_p = mpz_fdiv_ui(B2[i],*poi);
				this->Ba_inv[poi-this->factor_base][i] = Mulm(B2_mod_p,inv_a,*poi);
//  				gmp_printf("*Poi: %d, inv_a: %d, B2modp: %d\n",*poi,inv_a,B2_mod_p);
//  				printf("Recalculate Aux Siqs: Ba_inv[%d,%d] = %d\n",poi-this->factor_base,i,this->Ba_inv[poi-this->factor_base][i]);
		
				if (ba_inv) {
					ba_inv[i] = this->Ba_inv[poi-this->factor_base][i];
				}
#ifdef SIQS_DEBUG
				
				mpz_t B,A,p_mpz;
				mpz_init_set(A,poly->a);
				mpz_mul_ui(A,A,2);
				mpz_init(B);
				mpz_init_set_ui(p_mpz,*poi);
				mpz_invert(A,A,p_mpz);
				mpz_mul(B,B2[i],A);
				int check = mpz_fdiv_ui(B,*poi);
				if ((check-this->Ba_inv[poi-this->factor_base][i])%(*poi)){
					printf("Error in 2Ba_inv. i=%d, p = %d.\n",i,*poi);					
				} //else printf("*");
				mpz_clear(A);
				mpz_clear(B);
				mpz_clear(p_mpz);
#endif				
			}
#ifdef SIQS_DEBUG
// 			//printf("\n");
#endif			
		}
	}	
	return ConstRC::Ok;
}

/*!
	This method generates a new polynomial (with a new 'a') for SIQS. The recalculation of 'a' is done by calling
	CreateNewAForSiqs(); then, contents of Ba_inv array are updated.

*/
void QuadraticSieve::CreateNewSiqsPolynomial(){
	quad_polynomial* poly;
	poly = &(this->polynomials[this->index_in_polynomial]);
	// at first, we calculate the value of 'a'
	CreateNewAForSiqs();
	// then we calculate B_i's. This includes two steps:
	int i;
	main_sieving_type inv_a;
	main_sieving_type gamma;
	main_sieving_type B2_mod_p;
	for (i=0; i<this->s; i++){
		// we calculate sqrt(kN) mod q_i and multiply it with (A/q_i)^{-1} mod q_l. If result gamma exceeds q_i/2, we substitute it by q_i-gamma
		mpz_divexact_ui(aux_1,poly->a, this->divisors_of_a[i]);
		mpz_set_ui(aux_2,this->divisors_of_a[i]);
		mpz_invert(aux_3,aux_1,aux_2);
		inv_a = mpz_get_ui(aux_3);
		gamma = Mulm(roots_of_kN_mod_p[this->indices_of_divisors_of_a[i]],inv_a,this->divisors_of_a[i]);
		if (gamma > (this->divisors_of_a[i] >> 1)){
			gamma = this->divisors_of_a[i] - gamma;
		}
		// Now, B_i = A/q_l x gamma. We want to store 2B_i.
		mpz_mul_ui(B2[i],aux_1,2*gamma);		
		
	}	
#ifdef SIQS_DEBUG
	printf("Checking whether B_i's are correct ...\n");
	mpz_t Bsquare;
	mpz_t cmp;
	int result_1;
	mpz_init(Bsquare);	
	mpz_init(cmp);
	for (i=0; i<this->s; i++){
		mpz_set(Bsquare,B2[i]);
		mpz_fdiv_q_2exp(Bsquare,Bsquare,1);
		mpz_mul(Bsquare,Bsquare,Bsquare);
		for (int j=0; j<this->s; j++){			
			if (i==j){
				mpz_sub(cmp,Bsquare,this->kN);
				result_1 = mpz_fdiv_ui(cmp,this->divisors_of_a[j]);
				if (result_1) printf("B[%d]^2 is not equal to kN mod %d",i,this->divisors_of_a[j]);
				else printf("*");
			}
			else {
				result_1 = mpz_fdiv_ui(Bsquare,this->divisors_of_a[j]);
				if (result_1) printf("B[%d]^2 is not equal to 0 mod %d",i,this->divisors_of_a[j]);
				else printf("*");
			}
		}
		printf("\n");		
	}
	gmp_printf("CreateNewAForSiqs: new a = %Zd\n",poly->a);
	printf("Verifying 2Ba_inv: \n");
	mpz_clear(Bsquare);
	mpz_clear(cmp);
#endif	
	RecalculateAuxiliarySiqsValues();	
	// higher prime powers
	if (this->power_sieving_mode == ESieveWithPrimePowers){
		int j, index, l;
		main_sieving_type p;
		prime_power_structure pps;
		for (i=0; i< primes_with_powers; i++){
			pps = this->prime_powers[i];
			index = pps.index_in_factor_base;
			if (!this->is_this_prime_divisor_of_a[index]){ //only for primes not dividing A ...
				for (j=0; j< pps.allocated_powers; j++){					
					inv_a = pps.a_inverses[j];
					p     = pps.powers[j];					
					for (l=0; l<this->s; l++){
						B2_mod_p = mpz_fdiv_ui(B2[l],p);
						pps.Ba_inv[j][l] = Mulm(B2_mod_p,inv_a,p);
					}					
				}
			}
		}
	}
	// now, we determine the initial value of b
	mpz_set_ui(poly->b,0);
	for (i=0; i<this->s; i++){
		mpz_add(poly->b,poly->b,B2[i]);
	}
	mpz_divexact_ui(poly->b,poly->b,2); // b_0 = B_1 + ... + B_s
	mpz_t pom_1; mpz_init(pom_1);
	//mpz_mod(pom_1,poly->b,poly->a);
	// in case that B % A is even, we change B_s to A - B_s. This should turn B % A to odd value.
	// DULEZITA ZMENA!!
	int changed = 0;
	if (mpz_divisible_2exp_p(poly->b,1)){		
		/*gmp_printf("\nB mod A = %Zd je sude. Zmenime ho na liche.\n",pom_1);
		changed = 1;
		int t = this->s - 1;
		mpz_divexact_ui(aux_1,B2[t],2);
		mpz_sub(poly->b,poly->b,aux_1); // B - B_s
		mpz_sub(B2[t],poly->a,aux_1); // A - B_s
		mpz_add(poly->b,poly->b,B2[t]); // B + B_s
		mpz_mul_ui(B2[t],B2[t],2);
		poi = this->factor_base;
		while (*(++poi)){ // The end of factor base is signalled by zero value. The first factor base element we want to examine is probably 2.
			if (!this->is_this_prime_divisor_of_a[poi - this->factor_base]){ //only for primes not dividing A ...
				inv_a = a_inverses[poi-this->factor_base];					
				B2_mod_p = mpz_fdiv_ui(B2[t],*poi);
				this->Ba_inv[poi-this->factor_base][t] = Mulm(B2_mod_p,inv_a,*poi);
			}
		}*/
		mpz_add(poly->b,poly->b,poly->a);
	}
	//mpz_mod(poly->b,poly->b,poly->a);
	mpz_clear(pom_1);
	if (changed) {
		printf("\nCreateNewSiqsPolynomial: Generovani noveho B: ");
		gmp_printf("     %Zd\n",poly->b);  	
	}
#ifdef SIQS_DEBUG	
	printf("\nCreateNewSiqsPolynomial: Generovani noveho B: ");
	gmp_printf("     %Zd\n",poly->b);  	
	/*if (mpz_divisible_2exp_p(poly->b,1)){
		printf("CreateNewSiqsPolynomial: B je sude. Dosadime na jeho misto A-B.\n");
		mpz_sub(poly->b,poly->a,poly->b);
	}*/
#endif			
	/*
		c = (b^2-N)/4a. If this is not true, there was some serious computational problem
	*/	
	mpz_mul(aux_1,poly->b,poly->b);
	mpz_sub(aux_1,aux_1,this->kN);
	mpz_fdiv_q(poly->c,aux_1,poly->a);
	//mpz_fdiv_q_ui(poly->c,poly->c,4);
	// we will calculate b^2-4ac and compare it to kN.	
	if (changed){
		mpz_mod(aux_1,this->kN,poly->a);
		    printf("kN  mod 4 = %lu\n",mpz_fdiv_ui(this->kN,4));
		gmp_printf("kN  mod a = %Zd\n",aux_1);
	            
		mpz_mul(aux_3,poly->b,poly->b);
		    printf("b^2  mod 4 = %lu\n",mpz_fdiv_ui(aux_3,4));
		mpz_mod(aux_3,aux_3,poly->a);
		gmp_printf("b^2 mod a = %Zd\n",aux_3);
		    
	}
	ComputeC(poly);
#ifdef SIQS_DEBUG
	
	mpz_mod(aux_1,this->kN,poly->a);
	gmp_printf("kN  mod a = %Zd\n",aux_1);
	mpz_mul(aux_3,poly->b,poly->b);
	mpz_mod(aux_3,aux_3,poly->a);
	gmp_printf("b^2 mod a = %Zd\n",aux_3);
#endif	
	this->solve = 1;
}

/*!
	This method generates new 'a' coefficient for SIQS sieving. It uses Carrier-Wagstaff algorithm for this
	purpose, generating an odd-indexed n-tuple of 'a' divisors at first and then choosing the most suitable
	even-indexed triple of 'a' divisors to get the whole of 'a'. An even-indexed triple is considered most
	suitable, if the overall product of 'a' divisors is the closest possible to ideal value of 'a'.
	
	This method does not, however, check for an important condition: that the set of newly generated 'a'
	divisors differs from the previous set of 'a' divisors in at least 2 entries. This might not be 
	satisfied if the limits for odd-indexed and even-indexed divisor set have been chosen unsuitably.

*/
void QuadraticSieve::CreateNewAForSiqs(){
	quad_polynomial* poly;
	//printf("Pocitame pro polynom cislo %d.\n",this->index_in_polynomial);
	poly = &(this->polynomials[this->index_in_polynomial]);
	// Preparation: we clear structures holding information about divisors of 'a', except for those determined by machine number
	memset(this->is_this_prime_divisor_of_a+(this->iUpperMachineBoundIndex+1),0,(this->maximal_factor_base_index+1-(this->iUpperMachineBoundIndex+1))*sizeof(main_sieving_type));
	memset(this->divisors_of_a+machine_specific_divisors_number,0,(this->s-3)*sizeof(main_sieving_type));
	memset(this->indices_of_divisors_of_a+machine_specific_divisors_number,0,(this->s-3)*sizeof(main_sieving_type));	
	
	// We get the value of current NEXKSB Combination
	int i, max = this->s - 3 - machine_specific_divisors_number;
	int index;
	
	mpz_set_ui(poly->a,1);
	float acc_log = 0;
	double log_2 = (double) 1/(double) log10(2);
	for (i=0; i<machine_specific_divisors_number; i++){
		mpz_mul_ui(poly->a,poly->a,this->divisors_of_a[i]);		
		//gmp_printf("Nasobime machine specific divisorem %d, vysledek a je %Zd.\n",this->divisors_of_a[i],poly->a);
	}
	acc_log += log10((double) mpz_get_ui(poly->a))*log_2;		
	
	for (i=1; i<=max; i++){
		index = TrueIndexOdd(this->nx_env->subset[i]);
		//printf("CreateNewAForSiqs:  index je %5d.\n",index);
		//printf("Prvocislo na tomto indexu je %5d.\n",this->factor_base[index]);
		this->is_this_prime_divisor_of_a[index] = 1;
		this->divisors_of_a[machine_specific_divisors_number+i-1] = this->factor_base[index];
		this->indices_of_divisors_of_a[machine_specific_divisors_number+i-1] = index;
		acc_log += this->exact_binary_logarithms_of_factor_base[index];
		mpz_mul_ui(poly->a,poly->a,this->factor_base[index]);
	    	//gmp_printf("Soucasna hodnota 'a' je %Zd, jeji delka v bitech je %d. Dosavadni nasbirany log je %6.4f, po podeleni %d je %d\n",poly->a,mpz_sizeinbase(poly->a,2),acc_log,TRIPLES_LOG_FACTOR,(int) (acc_log/(double)TRIPLES_LOG_FACTOR));
	}
	//printf("Pokusime se doplnit tyto prvky na cely faktor 'a'.\n");
	int acc_log_2 = (int) (acc_log*(double) TRIPLES_LOG_FACTOR);
	int ideal_log = this->ideal_log_of_a_times_triples_log_factor;
	//printf("Nasbirany log je %d, idealni log je %d, budeme hledat log blizky %d.\n",acc_log_2,ideal_log,ideal_log-acc_log_2);
	// index in polynomial has already been shifted in GenerateNewPolynomial
	int triple = Binary_Search_Triples(this->triples, 0, this->triple_count-1,ideal_log-acc_log_2);
	int triple_minus_1 = triple-1;
	int a1, a2;
	// We want the smallest possible difference in absolute value
	if (triple_minus_1 >= 0){
		a1 = this->triples[triple][0]-ideal_log+acc_log_2;
		if (a1 < 0) {	a1 *= -1;  }
		a2 = this->triples[triple_minus_1][0]-ideal_log+acc_log_2;
		if (a2 < 0) {	a2 *= -1;  }
		if (a2 < a1) {
			triple = triple_minus_1;
		}
	}
	//printf("Idealni pry bude tato triple: %d, s logaritmem %d. Celkova delka triples je %d.\n",triple,this->triples[triple][0],this->triple_count);
	for (i=1; i<4; i++){
		index = TrueIndexEven(this->triples[triple][i]);
		//printf("Dodatecne prvocislo je %d.\n",this->factor_base[index]);
		this->is_this_prime_divisor_of_a[index] = 1;
		this->divisors_of_a[this->s-4+i] = this->factor_base[index];
		this->indices_of_divisors_of_a[this->s-4+i] = index;
		acc_log += this->exact_binary_logarithms_of_factor_base[index];
		mpz_mul_ui(poly->a,poly->a,this->factor_base[index]);
	    	//gmp_printf("Soucasna hodnota 'a' je %Zd, jeji delka v bitech je %d. Dosavadni nasbirany log je %6.4f.\n",poly->a,mpz_sizeinbase(poly->a,2),acc_log);
	}
	// As the final step, we will sort the arrays "divisors_of_a" and "indices_of_divisors_if_a"
        qsort(divisors_of_a,this->s,sizeof(main_sieving_type), &CompareMainSievingTypes);
        qsort(indices_of_divisors_of_a,this->s,sizeof(main_sieving_type), &CompareMainSievingTypes);

	//gmp_printf("Idealni hodnota 'a' je  %Zd, jeji delka v bitech je %d. Logaritmus prenasobeny je %d, logaritmus skutecny je %6.4f.\n",this->ideal_a_value,mpz_sizeinbase(this->ideal_a_value,2),ideal_log_of_a_times_triples_log_factor,(ideal_log_of_a_times_triples_log_factor/(float) TRIPLES_LOG_FACTOR));
	//gmp_printf("CreateNewAForSiqs: new a = %Zd\n",poly->a);
	//printf("Its binary length is %d",mpz_sizeinbase(poly->a,2));
	ComputeAInverses(); // stores into a_inverses (2a)^{-1} mod p		
	NEXKSB_Next(); // we generate another 'a' value	
}

/* This method is called only when index_in_polynomial > 1. Only under this assumption it performs valid operations; otherwise mem overflows may occur! */
/*!
	This method calculates new 'b' in SIQS sieving. For this, it uses the Gray code formula. For detailed
	description of this algorithm, see theoretical documentation.
	
	This method is called only when index_in_polynomial > 1 (which is natural in normal SIQS). Only under this assumption it performs valid operations; otherwise mem overflows may occur!

*/
void QuadraticSieve::RecalculateBForSiqs(){
	// at first, we determine nu. This is the maximal exponent such that 2^nu | 2i, where i is the index of previous polynomial mod 2^{s-1} plus 1.
	// It is clear that nu is at least 1. Value of i may not be zero!
	int  i  = ((this->index_in_polynomial) % (1 << (s-1)));
	int  i2 = 2*i;
	if (i2 == 0) {
		CLog::LogError(WHEREARG,"Quadratic sieve","recalculating B",s,72,"There was a logical error in SIQS!");		
	}
	this->nu = 0;
	// The determination of nu itself.
	while (!(i2 & 1)){
		i2 >>=1;
		this->nu++;
	}	
	
	quad_polynomial* new_poly;
	new_poly = &(this->polynomials[this->index_in_polynomial]);
	quad_polynomial* old_poly;
	old_poly = &(this->polynomials[this->index_in_polynomial-1]);
	// Values of 'a' and 'c' remain the same
	mpz_set(new_poly->a,old_poly->a);
	mpz_set(new_poly->c,old_poly->c);
	// Now, we iterate b by use of Gray code formula. At first we determine whether we will subtract or add the value of 2*B_nu from/to current B.
	// This depends on parity of 'ceiling division' i/2^nu. At first we determine i/2^nu in a normal way. Definitely 2^nu does not divide i, as
	// then 2^{nu+1} would divide 2i, a contradiction to the fact that 2^nu is maximal power of 2 with this quality. So we must divide i by 2^nu,
	// then add 1, and only then mod it by 2.
	this->subtract = ((i / (1 << this->nu))+1) % 2; // i/2^nu must be rounded UP! How to manage this?
	// Then we either add or subtract 2*B_nu
	if (this->subtract){
		mpz_sub(new_poly->b,old_poly->b,this->B2[this->nu]);
	} else {
		mpz_add(new_poly->b,old_poly->b,this->B2[this->nu]);	
	}	
#ifdef SIQS_DEBUG	
	printf("\nGenerovani noveho B: ");
	gmp_printf("     %Zd\n",new_poly->b);  	
	gmp_printf("RecalculateBForSiqs: new_poly->b = %Zd. Subtract = %d. Nu=%d.\n",new_poly->b,subtract,this->nu);
#endif	
	mpz_mul(aux_1,new_poly->b,new_poly->b);
	mpz_sub(aux_1,aux_1,this->kN);
	mpz_fdiv_q(new_poly->c,aux_1,new_poly->a);
	//mpz_fdiv_q_ui(new_poly->c,new_poly->c,4);
	// we will calculate b^2-4ac and compare it to kN.	
#ifdef SIQS_DEBUG	
	mpz_mod(aux_1,this->kN,new_poly->a);
	gmp_printf("kN  mod a = %Zd\n",aux_1);
	mpz_mul(aux_3,new_poly->b,new_poly->b);
	mpz_sub(aux_5,aux_3,this->kN);
	mpz_mod(aux_3,aux_3,new_poly->a);	
	
	gmp_printf("b^2 mod a = %Zd\n",aux_3);
	
	mpz_fdiv_qr(aux_1,aux_2,aux_5,new_poly->a);	
	if (mpz_cmp_ui(aux_2,0)){
		gmp_printf("Recalculate B for SIQS: Problem: b^2-kN / a = %Zd * %Zd + %Zd!\n",aux_1,new_poly->a,aux_2);
	} else {
	        gmp_printf("Recalculate B for SIQS: b^2-kN is divisible by a, ConstRC::Ok. The ratio is %Zd.\n",aux_1);
		
	}
#endif	
	ComputeC(new_poly);
}

/*!
	This method is called to create a new sieving polynomial (e.g., calculate its coefficients) in MPQS sieving.
	It generates pseudoprime numbers 'd' of size sqrt(ideal value of 'a') until it meets 'd' being = 3 mod 4 and
	being a quadratic residue mod kN. This 'd' is taken as square root of 'a', and 'b' is calculated as value
	satisfying b^2 = 1 mod 4, b^2 = 1 mod kN. This calculation uses Shanks-Tonelli multiexp algorithm.
	Also, (2a) inverses mod p in FB are calculated.

*/
void QuadraticSieve::CreateNewMpqsPolynomial(){
	/*
		Generation of a new quad polynomial for MPQS. We must get an 'a' such that a is roughly sqrt(2kN)/M, a = d^2, d prime
		and kN is a quadratic residue mod d.
	*/
	quad_polynomial* poly;
	poly = &(this->polynomials[this->index_in_polynomial]);
	mpz_t pom, pom2, h_0, h_1, h_2;
	int flag=0,
	flag1=0, flag2=0;
	mpz_init(pom);
//#define CREATE_DEBUG	
	// in GenerateNewPolynomial we got poly->d := poly[index-1]->d
	if (!this->current_polynomial){
		// case of the first polynomial, with a = d = 1. We must calculate pom as sqrt(sqrt(kN/2)/M))
		mpz_set(pom,this->kN);	
		mpz_fdiv_q_ui(pom,pom,2);
		mpz_sqrt(pom,pom);
		mpz_fdiv_q(pom,pom,this->M);		
		
		mpz_sqrt(pom,pom);
#ifdef CREATE_DEBUG		
		printf("\n");
		gmp_printf("CreateNewMpqsPolynomial #1: we have generated first d candidate as = %Zd, its binary length being %d.\n",pom,mpz_sizeinbase(pom,2));
#endif		
	} else {
		mpz_set(pom,poly->d); 
	}
	mpz_init(pom2); mpz_init(h_0); mpz_init(h_1); mpz_init(h_2);
	do {
		mpz_nextprime(pom,pom);
		if ((mpz_fdiv_ui(pom,4))==3) {
			 flag1 = 1; 
		} else {
			 flag1 = 0;
		}
		if (mpz_kronecker(this->kN,pom) == 1) {
			 flag2 = 1; 
		} else {
			 flag2 = 0;
		}
		flag = (flag1 & flag2);			
	} while (!flag);
#ifdef CREATE_DEBUG			
	gmp_printf("CreateNewMpqsPolynomial: we have generated d as = %Zd.\n",pom);
#endif
	mpz_set(poly->d,pom);
	mpz_set(poly->a,poly->d);
	mpz_mul(poly->a,poly->a,poly->a); 
#ifdef CREATE_DEBUG			
	gmp_printf("CreateNewMpqsPolynomial: we have generated a as = %Zd.\n",poly->a);
#endif		
	if ((this->QSMode == EOldMpqs) || (this->QSMode == EOldSiqs) || (iFirstLineSieverPolynomial))
	{
		ComputeAInverses(); // stores into a_inverses (2a)^{-1} mod p	
	}
	/* Now, we must generate b (in expression (ax+b)^2 - kN) so that b^2-kN = 0 mod a. This is modular equation with respect to a square
	   of a prime. This can be solved with Shanks-Tonelli algorithm 'mod q' and then lifted to q^2 = a. 
	*/
	ShanksTonelliMultiexpMpz(poly->b, this->kN, poly->d, 2);
	// if b is even, we substitute it with a-b mod a, to get odd b. If b is odd, then b^2 = 1 mod 4 and, assuming that
	// multiplier k was chosen so that kN = 1 mod 4, we get 4 | (b^2 - kN), and thus 4a | (b^2 - kN), b^2 - 4ac = 0 mod kN.
	if (mpz_divisible_2exp_p(poly->b,1)){
		mpz_sub(poly->b,poly->a,poly->b);
	}
//#define CREATE_DEBUG			

	ComputeC(poly);	

	mpz_clear(pom); mpz_clear(pom2); mpz_clear(h_0); mpz_clear(h_1); mpz_clear(h_2);
}

void QuadraticSieve::ComputeC(quad_polynomial* aPoly)
{
	mpz_mul(aux_4,aPoly->b,aPoly->b);
	mpz_sub(aux_4,aux_4,this->kN);
#ifdef CREATE_DEBUG			
	gmp_printf("b^2 - kN = %Zd.\n",aux_4);
#endif	
	mpz_fdiv_qr(aux_5,aux_1,aux_4,aPoly->a);
	if (mpz_cmp_ui(aux_1,0)){

		gmp_printf("ComputeC: Problem: b^2-kN / a = %Zd * %Zd + %Zd!\n",aux_5,aPoly->a,aux_1);

	} else {
		//gmp_printf("CreateNewMpqsPolynomial: b^2-kN is divisible by a, ConstRC::Ok. The ratio is %Zd.\n",pom2);
		mpz_fdiv_qr_ui(aux_5,aux_1,aux_5,4);
		if (mpz_cmp_ui(aux_1,0)){
			gmp_printf("ComputeC: Problem: b^2-kN / a is not divisible by 4!");	

		} else {
			mpz_set(aPoly->c,aux_5);
#ifdef CREATE_DEBUG			
	        	gmp_printf("ComputeC: b^2-kN is divisible by 4a, ConstRC::Ok. The ratio c is %Zd.\n",aux_4);
#endif		
		}
	}

}

/*!
	This is an auxiliary method for printing out the intermediate results
	of the sieving step.

*/
void QuadraticSieve::PrintHeader() {
	printf("\n");
	printf(" ________________________________________________________________________________________________________");
	printf("\n");
	printf("|  POLYNOM |  SMOOTH |  1-PART |  2-PART |  CYCLES |   EDGES |  COMPON |  VERTEX |  RELATIONS |  PERCENT |");
	printf("\n");
}

/*!
	This is an auxiliary method for printing out the intermediate results
	of the sieving step.

*/
void QuadraticSieve::PrintValues() {
	printf("\r|");
	printf(" %8ld |",this->current_polynomial);
	printf(" %7ld |",this->so_far_smooths_found);
	printf(" %7ld |",this->so_far_1_partials_found);
	printf(" %7ld |",this->so_far_2_partials_found);				
	this->cycles = this->edges+this->components-this->vertices;			
	printf(" %7d |",this->cycles);					
	printf(" %7d |",this->edges);				
	printf(" %7d |",this->components);				
	printf(" %7d |",this->vertices);				
	this->collected = this->so_far_smooths_found + (this->cycles);
	printf(" %10ld |",this->collected);				
	printf(" %7.2f%% |",(100*this->collected)/((double) this->threshold));	
	fflush(stdout);	
	if (this->frontend)
	{
		fprintf(this->frontend_pipe,"\nPhaseSieve> (%4.2f%%) (%ld) (%ld) (%ld) (%ld) (%ld)",
				(100*this->collected)/((double) this->threshold),this->collected,
				this->so_far_smooths_found,this->so_far_1_partials_found, this->so_far_2_partials_found, time(NULL));
		fflush(this->frontend_pipe);
	}
}

bool	QuadraticSieve::IsDistributedPhaseFinished() const
{
	return (this->collected >= this->threshold);
}

int	QuadraticSieve::RunNondistributedPhase()
{
	int ret;
	ret = ProcessRelations();
	if (ret != ConstRC::Ok) { return ERROR_PROCESSING_RELATIONS; }
	ret = LinearAlgebraStep();
	if (ret == NOT_ENOUGH_SMOOTHS) { return NOT_ENOUGH_SMOOTHS; }
	if (ret == ConstRC::Factorized) 	    { return ConstRC::Factorized; }
	if (ret == NO_DEPENDENCIES) { return NO_DEPENDENCIES; }
	if (ret != ConstRC::Ok) { return ERROR_LINALG; } // all other errors (memory etc.)
	ret = Factor();
	if (ret == ConstRC::Factorized) { return ConstRC::Factorized; }
	if (ret == NOT_FACTORIZED) { return NOT_FACTORIZED; }
	return NOT_FACTORIZED;
}

bool QuadraticSieve::IsComputationFinished() {
	return iComputationFinished;
}

void QuadraticSieve::SetComputationFinished(bool aFlag) {
	iComputationFinished = aFlag;
}

void    QuadraticSieve::ResetCenter()
{
	printf("QuadraticSieve::ResetCenter(); incomplete yet!\n");
	SetInfoMode(TRUE);
}

void	QuadraticSieve::InterruptNodeNow()
{
	Set_InterruptNow(TRUE);
}

/*!
	This method is responsible for calling the innermost sieving routines in the right order. If TIME_MESSAGE is defined
	and time messages requested (from the command line), it also collects timing data about runtime of these routines.
	
	Return codes:
	- ConstRC::Ok if everything went well
	- ENOUGH if the number of collected smooths and partial exceeded threshold value (linear algebra step can now be run).

*/
int QuadraticSieve::PerformSievingByThisPolynomial(){
	int  diff;
	int ret;
		if (this->info_mode){						
			if (!(this->current_polynomial%10)) {
				PrintValues();
			}
		}
		this->main_matrix->lower_arg = (-1)*this->integer_M;
		this->main_matrix->upper_arg = this->integer_M - 1;
//#define PERFORM_DEBUG		
#ifdef PERFORM_DEBUG		
		printf("\nLower arg = %d, upper arg = %d\n",this->main_matrix->lower_arg,this->main_matrix->upper_arg);
#endif	
/*//#define	NEW_METHOD_OF_ACCEPTED
#ifdef NEW_METHOD_OF_ACCEPTED
// We will calculate the greatest value of Q(x) in current interval; then, we will subtract 20*_LOG_FACTOR and maximal possible variation.
	        mpz_t val_1, val_2;
		mpz_init(val_1); mpz_init(val_2); 
	    	if (extent_i <= max_extent_cycle/2) mpz_set(val_2,this->main_matrix->lower_arg); else mpz_set(val_2,this->main_matrix->upper_arg);
		mpz_set(val_1,val_2);
	    	mpz_mul(val_1,val_1,this->polynomials[index_in_polynomial].a);
	    	mpz_add(val_1,val_1,this->polynomials[index_in_polynomial].b);
		mpz_mul(val_1,val_1,val_2);
		mpz_add(val_1,val_1,this->polynomials[index_in_polynomial].c);
		int size = mpz_sizeinbase(val_1,2);
		//size -= 20;
		if (this->using_variations == ELargePrimeV) size -= mpz_sizeinbase(this->maximal_partial_factor,2);
		size *= _LOG_FACTOR;
		this->logarithm_accepted = size;
		mpz_clear(val_1); mpz_clear(val_2);
#endif
//#define  SILVERMAN_BOUND		
#ifdef SILVERMAN_BOUND
		mpz_t val_1, val_2;
		mpz_init(val_1); mpz_init(val_2); 
		mpz_fdiv_q_ui(val_1,this->kN,2);
		mpz_sqrt(val_1,val_1);
		mpz_mul(val_1,val_1,this->M);
		mpz_set_ui(val_2,this->factor_base[this->maximal_factor_base_index]);
		mpz_sqrt(val_2,val_2);
		mpz_mul_ui(val_2,val_2,this->factor_base[this->maximal_factor_base_index]);
		mpz_mul_ui(val_2,val_2,this->factor_base[this->maximal_factor_base_index]); // pmax^2.5
		mpz_fdiv_q(val_1,val_1,val_2);
		this->logarithm_accepted = mpz_sizeinbase(val_1,2) * _LOG_FACTOR;
		mpz_clear(val_1); mpz_clear(val_2);
#endif*/
#ifdef TIME_MESSAGE
		start = clock();
#endif		
		ret = RefillMainMatrix();
		if (ret != ConstRC::Ok){
			return ret;
		}
#ifdef TIME_MESSAGE		
		diff = clock() - start;
		if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
			total_refill += (double) diff;
		}		
		start = clock();		
#endif		
		// The sieving step itself		
		ret = SievingLoopLogarithmic();				
		if (ret != ConstRC::Ok) 		{
			CLog::LogError(WHEREARG,"Quadratic sieve","sieving loop",ret,0,"check error code - method not supported?");
			return ret;
		}		
		if (this->power_sieving_mode == ESieveWithPrimePowers) {
			ret = PerformSievingWithPrimePowersLogarithmic();
			if (ret != ConstRC::Ok) {
				return ret;
			}
		}
#ifdef TIME_MESSAGE		
		diff = clock() - start;
		if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
			total_sieve += (double) diff;
		}
#endif				

#ifdef TIME_MESSAGE			
		start = clock();
#endif				
		ret = SortSmoothValuesLogarithmic();
#ifdef TIME_MESSAGE							
		diff = clock() - start;
		if (diff >=0){ // in this way we avoid problems with integer overflow, using addition of negative values 
			total_sort += (double) diff;
		}	
#endif									
		
		if (ret != ConstRC::Ok) 		{
			CLog::LogError(WHEREARG,"Quadratic sieve","sorting smooth values",ret,0,"check error code - method not supported?");
			return ret;
		}
		if (this->using_variations != EDoNotUseVariations) {
			this->independent_cycles = this->edges + this->components - this->vertices;
		}
		ret = CheckEnough();
		if (ret == ENOUGH)  { return ret; }

	return ConstRC::Ok;
}

/*!
	This method calculates square roots of kN modulo primes of the factor base (excluding -1 and 2).  
	    
	For this purpose, a faster variant of Shanks-Tonelli, using only integer arithmetics, is used.
	
	The only return value is ConstRC::Ok.

*/
int QuadraticSieve::CalculateRootsOfkN(){
    	int  i,starting_i = 2;   // from first odd prime number
    	main_sieving_type p, n;    	
    	for (i=starting_i;i<=this->maximal_factor_base_index;i++)
    	{
    		  p = this->factor_base[i];	
		  n = mpz_fdiv_ui(this->kN,p); // n = kN mod p
		  if (n){
		  	this->roots_of_kN_mod_p[i] = ShanksTonelliIntegral(n,p);
		  } else this->roots_of_kN_mod_p[i] = 0;
	}
	return ConstRC::Ok;	  
}

/*!
	This method is used to calculate arg1*arg2 mod mod, where arg1 and arg2 are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.

*/
/*inline int QuadraticSieve::Mulm(INT64 arg1, INT64 arg2, int mod){
	INT64 inter =  arg1* arg2;
	//printf("Mulm soucin argumentu: %lld\n",inter);
	return (inter % mod);	
}*/

/*!
	This method is used to calculate arg1*arg2 mod mod, where arg1 and arg2 are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.

*/
/*inline int QuadraticSieve::Mulm(INT64 arg1, int arg2, int mod){
	INT64 inter =  arg1* (INT64) arg2;
	//printf("Mulm soucin argumentu: %lld\n",inter);
	return (inter % mod);	
}*/

/*!
	This method is used to calculate arg1*arg2 mod mod, where arg1 and arg2 are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.

*/
/*inline int QuadraticSieve::Mulm(int arg1, int arg2, int mod){
	INT64 inter = (INT64) arg1* (INT64) arg2;	
	//printf("Mulm soucin argumentu: %lld, mod %d: %lld.\n",inter, mod,(inter%mod));
	return (inter % mod);
}*/

/*!
	This method is used to calculate arg1^exp mod mod, where arg1 is are assumed to be 32-bit
	values. It uses an intermediate 64-bit variable, dependent on computer platform.
	A standard binary algorithm is used for calculation of powers.

*/
/*inline int QuadraticSieve::Powm(int arg, int exp, int mod){
	if (!exp) return 0;
	if (exp == 1) return (arg%mod);	
	INT64 inter = (INT64) arg;	
        // nejprve zjistime maximalni jednickovy bit v cisle exp
	int i = -1;	
	while (exp >> (++i)){
	} 
	int max = i-1;  // nejvyssi signifikantni bit.
	int MASK = (1 << (max-1)); // zacneme od druheho nejvyssiho signifikatniho bitu
	while (MASK > 0){
		inter = Mulm(inter,inter,mod);
		if (exp & MASK) {
			inter = Mulm(inter,arg,mod);
		}
		MASK = MASK >> 1;
	}
	return (inter%mod);
}*/

// This method solves equations (2ax+b)^2 = kN mod p for all odd factorbase primes p. It does not expect
// any precomputed values, and so it calls 'invert a-s' method first. If a is not invertible mod p (as
// often happens in SIQS), the method just pronounces the whole equation invalid (and therefore it will
// not be sieved with, though it usually happens that this prime divides ANY Q(x)).
//#define QUADR_DEBUG
/*!
	This method solves equations (2ax+b)^2 = kN mod p for all odd factorbase primes p. It does not expect
	any precomputed values, and so it calls 'invert a-s' method first. If a is not invertible mod p (as
	often happens in SIQS), the method just pronounces the whole equation invalid (and therefore it will
	not be sieved with).
	
	Return value: ConstRC::Ok.

*/
int QuadraticSieve::SolveQuadraticEquation(){		
	int i, l,          start = 2;	// first odd positive number in FB
	quad_polynomial*   poly  = &(this->polynomials[index_in_polynomial]);
	integer_solutions* sols  = &this->int_sols_of_cur_poly;
	main_sieving_type  a, a_inv, b, p, res_1, res_2,  minus_c;
	
	for (i=start; i<=this->maximal_factor_base_index; i++){		
		p = this->factor_base[i];
		// now, we must take care of p|a case
		// GCD(a,p) > 1
		if (!(a_inv = this->a_inverses[i])){ // equation has no 'true' solutions; in SIQS, this happens for q_i | a and then q_i|Q(x) for all x
			minus_c = mpz_fdiv_ui(poly->c,p);
			minus_c = p-minus_c; // -c;
//#define A_DIVISOR_DEBUG
#ifdef  A_DIVISOR_DEBUG
			printf("Solve quadratic equation: equation is linear for p=%d.\n",p);
#endif
			CalculateSolutionForADivisor(i, p, poly->b, minus_c);			

//#define QUADR_DEBUG			
    			continue; // We continue iterating to next prime in factorbase
		} // Case of GCD(a,p) > 1 finished
		// True quadratic equation. There will be two results.		
		this->int_sols_of_cur_poly.usable[i] = VALID_1 | VALID_2;				
		// It is obvious that:  (ax+b)^2 = kN mod p
		//			 ax+b    = sqrt(kN)mod p	 remember that sqrt is done modularly, mod p
		//			 ax      = sqrt(kN) - b mod p
		//			  x      = (sqrt(kN) - b)*a_inv mod p
		a  = mpz_fdiv_ui(poly->a,p);
		b  = mpz_fdiv_ui(poly->b,p);
#ifdef QUADR_DEBUG				
		kn = mpz_fdiv_ui(this->kN,p);
		printf("\nCase of quadratic equation (2*%dx+%d)^2 = %d mod %d.\n",a,b,kn,p);	
		printf("   Root of kN mod p was calculated as %d.\n",this->roots_of_kN_mod_p[i]);
		printf("   Its square mod p is equal to       %d.\n",Mulm(this->roots_of_kN_mod_p[i],this->roots_of_kN_mod_p[i],p));
#endif		
		res_1  = (this->roots_of_kN_mod_p[i] - b) % p;
#ifdef QUADR_DEBUG		
		printf("   Difference of root and b mod p is  %d.\n",res_1);
		printf("   Corresponding a_inv            is  %d.\n",a_inv);		
#endif		
		res_1  = Mulm((int) res_1,(int) a_inv,p); // This MUST be done with Mulm. Sizes of res_1 and a_inv may well be over 16 bits, and so their product would overflow a 32-bit integer.
		if (res_1 < 0) { // We want our solutions to be nonnegative. As this is not always guaranteed by built-in machine operator %, we check it ands possibly correct it manually.
			res_1 += p; 
		}
#ifdef QUADR_DEBUG		
		printf("   Their product mod p            is  %d.\n",res_1);
#endif		
		res_2  = ((-1)*this->roots_of_kN_mod_p[i] - b) % p;
#ifdef QUADR_DEBUG		
		printf("   Difference of -root and b mod p is  %d.\n",res_2);
#endif		
		res_2  = Mulm((int) res_2,(int) a_inv,p); // This MUST be done with Mulm. Sizes of res_1 and a_inv may well be over 16 bits, and so their
#ifdef QUADR_DEBUG		
		printf("   Corresponding a_inv            is  %d.\n",a_inv);
#endif		
		if (res_2 < 0) {
			res_2 += p; 
		}
#ifdef QUADR_DEBUG
		printf("   Their product mod p            is  %d.\n",res_1);
		printf("   Solutions: %d and %d.\n",res_1,res_2);				
		mpz_set(aux_1,poly->a);
		mpz_mul_ui(aux_1,aux_1,res_1);
		mpz_mul_ui(aux_1,aux_1,2);
		mpz_mul_ui(aux_2,poly->a,res_2);
		mpz_mul_ui(aux_2,aux_2,2);
		mpz_add(aux_1,aux_1,poly->b);
		mpz_add(aux_2,aux_2,poly->b);
		mpz_mul(aux_1,aux_1,aux_1);
		mpz_mul(aux_2,aux_2,aux_2);
		mpz_sub(aux_1,aux_1,this->kN);
		mpz_sub(aux_2,aux_2,this->kN);
		if (mpz_congruent_ui_p(aux_1,0,p) && mpz_congruent_ui_p(aux_2,0,p)){
		        printf("Quadr debug: everything ConstRC::Ok for p=%d\n",p);
		}
		else {
			printf("Quadr debug: some problem in quadratic equation solution! First = %d, second = %d.\n",mpz_fdiv_ui(aux_1,p),mpz_fdiv_ui(aux_2,p));
		}
#endif		
		(res_1 < 0)  ? sols->sol_1[i] = p+res_1 : sols->sol_1[i] = res_1 ;
		(res_2 < 0)  ? sols->sol_2[i] = p+res_2 : sols->sol_2[i] = res_2 ;
		// In case of identical solutions (very rare, only if p|k) we mark the second one as invalid, to avoid "double sieving"
		if (sols->sol_1[i] == sols->sol_2[i]) this->int_sols_of_cur_poly.usable[i] &= ~VALID_2;
	}
	// Solution for prime powers. New from 1.15. For simplicity, we ignore the case of p^k such that p|a; we mark such solutions
	// as invalid.
	if (this->power_sieving_mode == ESieveWithPrimePowers){
		prime_power_structure pps;
		int 		      index;
		main_sieving_type     p;
		for (i=0; i< this->primes_with_powers; i++){
			pps   = this->prime_powers[i];
			index = pps.index_in_factor_base;			
			for (l=0; l< pps.allocated_powers; l++){							
				// We ignore powers of primes dividing 'a'.				
				if (!(a_inv = pps.a_inverses[l])){			
					pps.usable[l] = 0;
				} else {
					p     = pps.powers[l];
					a  = mpz_fdiv_ui(poly->a,p);
					b  = mpz_fdiv_ui(poly->b,p);
					res_1  = (pps.roots_of_kN_mod_powers[l] - b) % p;			
					res_1  = Mulm((int) res_1,(int) a_inv,p); // This MUST be done with Mulm. Sizes of res_1 and a_inv may well be over 16 bits, and so their product would overflow a 32-bit integer.
					if (res_1 < 0) { // We want our solutions to be nonnegative. As this is not always guaranteed by built-in machine operator %, we check it ands possibly correct it manually.
						res_1 += p; 
					}
					res_2  = ((-1)*pps.roots_of_kN_mod_powers[l] - b) % p;
					res_2  = Mulm((int) res_2,(int) a_inv,p); // This MUST be done with Mulm. Sizes of res_1 and a_inv may well be over 16
					if (res_2 < 0) {
						res_2 += p; 
					}
//#define	POWERS_QUADR_DEBUG					
#ifdef POWERS_QUADR_DEBUG					
					mpz_set(aux_1,poly->a);
					mpz_mul_ui(aux_1,aux_1,res_1);
					mpz_mul_ui(aux_1,aux_1,2);
					mpz_mul_ui(aux_2,poly->a,res_2);
					mpz_mul_ui(aux_2,aux_2,2);
					mpz_add(aux_1,aux_1,poly->b);
					mpz_add(aux_2,aux_2,poly->b);
					mpz_mul(aux_1,aux_1,aux_1);
					mpz_mul(aux_2,aux_2,aux_2);
					mpz_sub(aux_1,aux_1,this->kN);
					mpz_sub(aux_2,aux_2,this->kN);
					if (mpz_congruent_ui_p(aux_1,0,p) && mpz_congruent_ui_p(aux_2,0,p)){
		        			printf("Quadr debug: everything ConstRC::Ok for p=%d\n",p);
					}
					else {
						printf("Quadr debug: some problem in quadratic equation solution! P = %d, first = %d, second = %d.\n",p,mpz_fdiv_ui(aux_1,p),mpz_fdiv_ui(aux_2,p));
					}
#endif
					(res_1 < 0)  ? pps.solutions_1[l] = p+res_1 : pps.solutions_1[l] = res_1 ;
					(res_2 < 0)  ? pps.solutions_2[l] = p+res_2 : pps.solutions_2[l] = res_2 ;
					pps.usable[l] |= (VALID_1 | VALID_2);
					// In case of identical solutions (very rare, only if p|k) we mark the second one as invalid, to avoid "double sieving"
					if (pps.solutions_1[l] == pps.solutions_2[l]) pps.usable[l] &= ~VALID_2;
					if (pps.prime == this->k) { // case of the multiplier (p=k). We do not sieve with it.										        
						pps.usable[l] = 0;
					}
				}	
			}
		}
	}
	return ConstRC::Ok;
}

main_sieving_type  QuadraticSieve::Evaluate(main_sieving_type aX, main_sieving_type aP)
{
	quad_polynomial*   poly  = &(this->polynomials[index_in_polynomial]);	
	return Evaluate(poly,aX,aP);
}

main_sieving_type QuadraticSieve::Evaluate(quad_polynomial* aPoly,main_sieving_type aX, main_sieving_type aP)
{
	mpz_set(aux_1,aPoly->a);
	mpz_mul_si(aux_1,aux_1,aX);
	mpz_mul_ui(aux_1,aux_1,2);
	mpz_add(aux_1,aux_1,aPoly->b);
	mpz_mul(aux_1,aux_1,aux_1);
	mpz_sub(aux_1,aux_1,this->kN);
	
	mpz_set(aux_2,aPoly->a);
	mpz_mul_si(aux_2,aux_2,aX); // ax
	mpz_add(aux_2,aux_2,aPoly->b); //ax + b
	mpz_mul_si(aux_2,aux_2,aX); // ax^2 + bx
	mpz_add(aux_2,aux_2,aPoly->c); // ax^2 + bx +c. Should be equal to ((2ax+b)^2 - kN)

	mpz_cdiv_q(aux_3,aux_1,aPoly->a);
	mpz_cdiv_q_ui(aux_3,aux_3,4);

	if (mpz_cmp(aux_3,aux_2)) {
		gmp_printf("Method 1: %Zd\n",aux_3);
		gmp_printf("Method 2: %Zd\n",aux_2);
	}

	main_sieving_type result = mpz_mod_ui(aux_2,aux_1,aP);
	return result;
}

main_sieving_type QuadraticSieve::EvaluateWithPrint(quad_polynomial* aPoly,main_sieving_type aX, main_sieving_type aP)
{
	mpz_set(aux_1,aPoly->a);
	mpz_mul_si(aux_1,aux_1,aX);
	mpz_mul_ui(aux_1,aux_1,2);
	mpz_add(aux_1,aux_1,aPoly->b);
	mpz_mul(aux_1,aux_1,aux_1);
	mpz_sub(aux_1,aux_1,this->kN);
	
	mpz_set(aux_2,aPoly->a);
	mpz_mul_si(aux_2,aux_2,aX); // ax
	mpz_add(aux_2,aux_2,aPoly->b); //ax + b
	mpz_mul_si(aux_2,aux_2,aX); // ax^2 + bx
	mpz_add(aux_2,aux_2,aPoly->c); // ax^2 + bx +c. Should be equal to ((2ax+b)^2 - kN)

	mpz_cdiv_q(aux_3,aux_1,aPoly->a);
	mpz_cdiv_q_ui(aux_3,aux_3,4);

	if (mpz_cmp(aux_3,aux_2)) {
		gmp_printf("Method 1: %Zd\n",aux_3);
		gmp_printf("Method 2: %Zd\n",aux_2);
	}
	mpz_t mod;
	mpz_init(mod);
	gmp_printf("Evaluate with print: x = %d, p = %d, Ax^2 + Bx + C    = %Zd, mod = %d\n",aX, aP, aux_2, mpz_mod_ui(mod,aux_2,aP));
	gmp_printf("Evaluate with print: x = %d, p = %d, (2Ax + B)^2 - kN = %Zd, mod = %d\n",aX, aP, aux_1, mpz_mod_ui(mod,aux_1,aP));
	gmp_printf("Evaluate with print: x = %d, p = %d, ((2Ax + B)^2 - kN) / 4A = %Zd = %d\n",aX, aP, aux_3, mpz_mod_ui(mod,aux_3,aP));
	mpz_clear(mod);

	main_sieving_type result = mpz_mod_ui(aux_2,aux_3,aP);
	return result;
}

//#define A_DIVISOR_DEBUG

main_sieving_type  QuadraticSieve::CalculateSolutionForADivisor(quad_polynomial* aPoly, main_sieving_type aPrimeP)
{
	mpz_set_ui(aux_2,aPrimeP); // aux_2 == p
	mpz_invert(aux_1,aPoly->b,aux_2); // aux_1 = b^{-1} mod p
	main_sieving_type bInverse = mpz_get_ui(aux_1);
	main_sieving_type minusC   = mpz_mod_ui(aux_1,aPoly->c,aPrimeP);
	main_sieving_type res;
	minusC = aPrimeP - minusC;
	if (minusC >= aPrimeP)
	{
		minusC %= aPrimeP;
	}
	if (minusC < 0)
	{
		minusC += aPrimeP;
	}
	if (bInverse)
	{
		res = Mulm(minusC,bInverse,aPrimeP);
#ifdef  A_DIVISOR_DEBUG
		printf("  Proposed solution mod %d: x=%d.\n",aPrimeP, res);
		main_sieving_type c = (-1)*minusC;
		main_sieving_type coeffB = mpz_mod_ui(aux_1,aPoly->b,aPrimeP);
		mpz_mul_ui(aux_1,aPoly->b,res);
		mpz_add_si(aux_1,aux_1,c);
		c = mpz_mod_ui(aux_1,aux_1,aPrimeP);
		printf("  Modulo: %d.\n",c);
		// Spocteme hodnotu Q(x) v bode x a zjistime, zda je delitelna cislem p^2.
		c = Evaluate(aPoly,res,aPrimeP*aPrimeP);	
		printf("  Evaluated: %d.\n",c);
#endif	
		return res;
	} 
	else {
		return -1;
	}
}

/*!
	This straightforward method solves the linear equation bx+c mod p. This is needed for primes dividing
	'a' in SIQS sieving.

*/
void QuadraticSieve::CalculateSolutionForADivisor(main_sieving_type aIndex, main_sieving_type aPrimeP, mpz_t aCoeffB, main_sieving_type aMinusC){
	quad_polynomial* new_poly = &(this->polynomials[this->index_in_polynomial]);
	mpz_t x1, x2;
	mpz_init_set_ui(x1,aPrimeP);
	mpz_init(x2);	
	mpz_invert(x2,aCoeffB,x1);
	main_sieving_type inv_b = mpz_get_ui(x2);
	main_sieving_type res;
	if (inv_b) {
#ifdef  A_DIVISOR_DEBUG
		res = mpz_mod_ui(aux_1,aCoeffB,aPrimeP);
		
		printf("Calc solution for a divisor: eq %d x + %d = 0 mod p=%d.\n",res,(-1)*aMinusC,aPrimeP);
		PrintPolynomial(new_poly);
		PrintPolynomialMod(new_poly,aPrimeP);
#endif

		res = Mulm(aMinusC,inv_b,aPrimeP);
		if (res < 0) res+= aPrimeP;
#ifdef  A_DIVISOR_DEBUG
		printf("  Proposed solution: x=%d.\n",res);
		main_sieving_type c = (-1)*aMinusC;
		mpz_mul_ui(aux_1,aCoeffB,res);
		mpz_add_si(aux_1,aux_1,c);
		c = mpz_mod_ui(aux_1,aux_1,aPrimeP);
		printf("  Modulo: %d.\n",c);
		// Spocteme hodnotu Q(x) v bode x a zjistime, zda je delitelna cislem p^2.
		c = Evaluate(new_poly,res,aPrimeP*aPrimeP);	
		printf("  Evaluated: %d.\n",c);
#endif
		this->int_sols_of_cur_poly.usable[aIndex] &= ~VALID_2;
		this->int_sols_of_cur_poly.usable[aIndex] |= VALID_1;
		this->int_sols_of_cur_poly.sol_1[aIndex] = res;
	}
	else {
	// pripad p | b
		printf("Patologicky pripad, prvocislo p deli jak a, tak b; p = %d.\n",aPrimeP);
		this->int_sols_of_cur_poly.usable[aIndex] &= (~VALID_2 & ~VALID_1);
	}	
	mpz_clear(x1); mpz_clear(x2);
}
/*void QuadraticSieve::AssertQuadraticEquation(solutions* sols){
	if (this->info_mode){
		NormalMessage("   Running quadratic assertion ...",NO_NUMBER);
	}
	long i, k, k_max;
	for (i=2; i<=this->maximal_factor_base_index; i++){
       		mpz_set_ui(aux_p,this->factor_base[i]);
		k_max = mpz_get_ui((sols[i]).sol_1[0]);
      		for (k=1;k<=k_max;k++){
        		mpz_pow_ui(aux_p,aux_p,k);
        		mpz_set(aux_1,(sols[i]).sol_1[k]);
        		mpz_set(aux_2,this->polynomials[this->index_in_polynomial].c);
        		mpz_mul(aux_3,this->polynomials[this->index_in_polynomial].b,aux_1);
        		mpz_add(aux_2,aux_2,aux_3);
        		mpz_mul(aux_1,aux_1,aux_1);
        		mpz_mul(aux_3,this->polynomials[this->index_in_polynomial].a,aux_1);
        		mpz_add(aux_3,aux_3,aux_2);
        		// nyni by v aux_3 melo byt ax^2+bx+c;
        		mpz_mod(aux_3,aux_3,aux_p);
        		if (mpz_cmp_ui(aux_3,0)){
				gmp_printf("\nAssertion failed (quadratic machine, root 1): %Zdx^2+%Zdx+%Zd mod %Zd (=%d^%d) = %Zd != 0.\n",this->polynomials[this->index_in_polynomial].a,this->polynomials[this->index_in_polynomial].b,this->polynomials[this->index_in_polynomial].c,aux_p,this->factor_base[i],k,aux_3);
	           	}
        	}
		k_max = mpz_get_ui((sols[i]).sol_2[0]);
		for (k=1;k<=k_max;k++){
        		mpz_pow_ui(aux_p,aux_p,k);
        		mpz_set(aux_1,(sols[i]).sol_2[k]);
        		mpz_set(aux_2,this->polynomials[this->index_in_polynomial].c);
        		mpz_mul(aux_3,this->polynomials[this->index_in_polynomial].b,aux_1);
        		mpz_add(aux_2,aux_2,aux_3);
        		mpz_mul(aux_1,aux_1,aux_1);
        		mpz_mul(aux_3,this->polynomials[this->index_in_polynomial].a,aux_1);
        		mpz_add(aux_3,aux_3,aux_2);
        		// nyni by v aux_3 melo byt ax^2+bx+c;
        		mpz_mod(aux_3,aux_3,aux_p);
        		if (mpz_cmp_ui(aux_3,0)){
				gmp_printf("\nAssertion failed (quadratic machine, root2): %Zdx^2+%Zdx+%Zd mod %Zd (=%d^%d) = %Zd != 0.\n",this->polynomials[this->index_in_polynomial].a,this->polynomials[this->index_in_polynomial].b,this->polynomials[this->index_in_polynomial].c,aux_p,this->factor_base[i],k,aux_3);
	           	}
		}
	}
}*/

// this solves x^2 = n mod p only

/*!
	This method runs the Shanks-Tonelli algorithm to find x such that x^2 = n mod p and returns 
	either x (>= 0) or -1.

*/
int QuadraticSieve::ShanksTonelliIntegral(long n, long p){
	int j;	
	mpz_set_ui(st_aux_n,n);
   	// Case of n being a quadratic st_aux_nonresidue - this should not happen in QSieve!
	if (mpz_kronecker_ui(st_aux_n,(unsigned int) p)-1)
	{
//		printf("   Insoluble equation.\n");
		return -1;
	}
	unsigned long max_power_of_two;
	unsigned long odd_part;
	//max_power_of_two = maximalni_mocnina_dvou(st_aux_p);
	max_power_of_two = MaxPowerOfTwo(p);
	odd_part=((p-1) >> max_power_of_two);
	// p-1 = t2^s	
	mpz_set_ui(st_aux_non,(unsigned int) 2);
	mpz_set_ui(st_aux_p,p);
	// We look for a quadratic st_aux_nonresidue mod p by simple linear search
	while(mpz_kronecker_ui(st_aux_non,p)+1)
	{
		mpz_add_ui(st_aux_non,st_aux_non,1);
	}
	// nereziduum 
	int non = mpz_get_ui(st_aux_non);
//	printf("      Nereziduum bylo vygenerovano jako %d\n",non);
	int root;
	int b, r, inv, b_exp_2j, aux_1;
	/*
	   Usage of Shanks-Tonelli algorithm, see Koblitz, A Course in Nr Theory and 		            Cryptography, II.2, For a prime congruent to 3 mod 4 there is an easier
	   algorithm.
	*/
	if (p%4 == 3){
		root = Powm(n,(p+1)/4,p);
	}
	else
	{
		b = Powm(non,odd_part,p);
		//printf("      b = %d\n",b);
		r = Powm(n,((odd_part+1)>>1),p);
		//printf("      r = %d\n",r);
		mpz_invert(st_aux_a_inverse,st_aux_n,st_aux_p);
		//gmp_printf("      st_aux_a_inverse = %Zd\n",st_aux_a_inverse);
		inv = mpz_get_ui(st_aux_a_inverse);
		b_exp_2j = b;		
		aux_1 = Mulm(r,r,p);
		aux_1 = Mulm(aux_1,inv,p);
		
		aux_1 = Powm(aux_1, (1 << (max_power_of_two-2)),p);
		if (aux_1 == 1){
			root = r;
		}
		else {
			root = Mulm(b,r,p);
		}
//		printf("     root je %d\n",root);
		for (j=1;j<=(max_power_of_two-2);j++)
		{
			b_exp_2j = Mulm(b_exp_2j,b_exp_2j,p);
			aux_1 = Mulm(root,inv,p);
			aux_1 = Mulm(aux_1,root,p);
			aux_1 = Powm(aux_1, 1 << (max_power_of_two-2-j), p);
			if (aux_1 != 1){
				root = Mulm(root,b_exp_2j,p);
			}
		}

	} // end of else - quadratic equation is solved
	if (this->assertion_mode){
		b = Mulm(root,root,p);
		if ( b!= n){
			gmp_printf("\nAssertion failed in Shanks-Tonelli Integral: %d is not a square st_aux_root of %d mod %d.",root,n,p);
		} //else printf("Assertion ok in Shanks-Tonelli for p=%d.\n",p);
	}
	//printf("\nWe have calculated that (%d)^2 = %d mod %d",root,n,p);	
	if (root < 0) root += p;
	return root;	
}

// solves (x+k)^2 = n mod p
// At first, it calculates 2nd root of n mod p, then it subtracts k mod p from the result.
// i is index of p in factorbase
/*!
	This method runs the Shanks-Tonelli algorithm to find x such that x^2 = n mod p^l, for l=1..k,
	and saves the results into this->int_sols_of_cur_poly on index i.

*/
int QuadraticSieve::ShanksTonelliIntegral(long n, long p, long i, long k){		
	int j;	
	integer_solutions* sols = &this->int_sols_of_cur_poly;
	// Foolproof assertions
	if (!sols || !(sols->sol_1) || !(sols->sol_2) || (sols->allocated < 2)){
		CLog::LogError(WHEREARG,"Quadratic sieve","Shanks-Tonelli called on uninitialised solution array",p,1000,"This Shanks-Tonelli works only with initialized solution array");
		printf("%p, %p, %p, %d.\n",sols,sols->sol_1,sols->sol_2,sols->allocated);
		return CALLED_TOO_EARLY;

	}
  	if (p<3){
		CLog::LogError(WHEREARG,"Quadratic sieve","Shanks-Tonelli called on too small p",p,1000,"This Shanks-Tonelli works only for p>2");
		return NOT_SUITABLE_MODULUS;
	}
	if (!(n%p)){ // dual root
		sols->usable[i] = VALID_1;
		sols->sol_1[i] = 0;
		return ConstRC::Ok;
	}
	mpz_set_ui(st_aux_n,n);
   	// Case of n being a quadratic st_aux_nonresidue - this should not happen in QSieve!
	if (mpz_kronecker_ui(st_aux_n,(unsigned int) p)-1)
	{
//		printf("   Insoluble equation.\n");
		return -1;
	}
	unsigned long max_power_of_two;
	unsigned long odd_part;
	//max_power_of_two = maximalni_mocnina_dvou(st_aux_p);
	max_power_of_two = MaxPowerOfTwo(p);
	odd_part=((p-1) >> max_power_of_two);
	// p-1 = t2^s	
	mpz_set_ui(st_aux_non,(unsigned int) 2);
	mpz_set_ui(st_aux_p,p);
	// We look for a quadratic st_aux_nonresidue mod p by simple linear search
	while(mpz_kronecker_ui(st_aux_non,p)+1)
	{
		mpz_add_ui(st_aux_non,st_aux_non,1);
	}
	// nereziduum 
	int non = mpz_get_ui(st_aux_non);
//	printf("      Nereziduum bylo vygenerovano jako %d\n",non);
	int root;
	int b, r, inv, b_exp_2j, aux_1;
	/*
	   Usage of Shanks-Tonelli algorithm, see Koblitz, A Course in Nr Theory and 		            Cryptography, II.2, For a prime congruent to 3 mod 4 there is an easier
	   algorithm.
	*/
	if (p%4 == 3){
		root = Powm(n,(p+1)/4,p);
	}
	else
	{
		b = Powm(non,odd_part,p);
		//printf("      b = %d\n",b);
		r = Powm(n,((odd_part+1)>>1),p);
		//printf("      r = %d\n",r);
		mpz_invert(st_aux_a_inverse,st_aux_n,st_aux_p);
		//gmp_printf("      st_aux_a_inverse = %Zd\n",st_aux_a_inverse);
		inv = mpz_get_ui(st_aux_a_inverse);
		b_exp_2j = b;		
		aux_1 = Mulm(r,r,p);
		aux_1 = Mulm(aux_1,inv,p);
		
		aux_1 = Powm(aux_1, (1 << (max_power_of_two-2)),p);
		if (aux_1 == 1){
			root = r;
		}
		else {
			root = Mulm(b,r,p);
		}
//		printf("     root je %d\n",root);
		for (j=1;j<=(max_power_of_two-2);j++)
		{
			b_exp_2j = Mulm(b_exp_2j,b_exp_2j,p);
			aux_1 = Mulm(root,inv,p);
			aux_1 = Mulm(aux_1,root,p);
			aux_1 = Powm(aux_1, 1 << (max_power_of_two-2-j), p);
			if (aux_1 != 1){
				root = Mulm(root,b_exp_2j,p);
			}
		}

	} // end of else - quadratic equation is solved
	if (this->assertion_mode){
		b = Mulm(root,root,p);
		if ( b!= n){
			gmp_printf("\nAssertion failed in Shanks-Tonelli Integral: %d is not a square st_aux_root of %d mod %d.",root,n,p);
		} //else printf("Assertion ok in Shanks-Tonelli for p=%d.\n",p);
	}
	sols->usable[i] = VALID_1 | VALID_2;
	long ret;
	// this ugly construction is necessary, as % may return negative results, and that would cause mistakes in following procedures
	(ret=(root-k)%p) < 0 ? sols->sol_1[i] = p+ret : sols->sol_1[i] = ret;
	(ret=(p-root-k)%p) < 0 ? sols->sol_2[i] = p+ret : sols->sol_2[i] = ret;
	return ConstRC::Ok;
}
/*
	This solves x^2 = n_in mod p^exp, including large moduli p. This is needed in generating new polynomials in MPQS - especially value of 'b'.
	We expect that GCD(n_in,p) = 1;
*/
/*!
	This method runs the Shanks-Tonelli algorithm to find x such that x^2 = n mod p^k and saves
	the result into solution. We expect that GCD(n_in,p) = 1.
	
	This method is used for example in calculation of sqrt(kN) mod elements of FB.

*/
int QuadraticSieve::ShanksTonelliMultiexpMpz(mpz_t solution, mpz_t n_in, mpz_t p, unsigned int exp)
{
	mpz_t n;
	int i;
	
	mpz_init_set(n,n_in);		
	if (!mpz_cmp_ui(n,0))
	{	
		/* reseni rovnice x^2=0, vsechny vysledky uz de facto mame */
		mpz_set_ui(solution,0);
		return ConstRC::Ok;
	}
	mpz_t p_minus_1_mpz;
	mpz_init_set(p_minus_1_mpz,p);
	mpz_sub_ui(p_minus_1_mpz,p_minus_1_mpz,1);	
//#define RESIDUE_DEBUG	
	if (mpz_kronecker(n,p)-1)
	{
#ifdef RESIDUE_DEBUG
		gmp_printf("Shanks-Tonelli multiexp mpz: %Zd is not a quadratic residue modulo p=%Zd",n,p);
#endif		
		mpz_clear(n);
		return INSOLUBLE_EQUATION;
	} 	

	int   max_even; // maximal power of 2 such that 2^max_even divides p_mpz-1
	mpz_t odd_part; // remainder of division (p_mpz-1)/2^max_even
	mpz_init_set(odd_part,p_minus_1_mpz);
	if (max_even = mpz_scan1(p_minus_1_mpz,0)){ // we calculate max_even and perform a test on its zeroity at the same time
		mpz_fdiv_q_2exp(odd_part,p_minus_1_mpz,max_even);						
	}
#ifdef RESIDUE_DEBUG	
	gmp_printf("Shanks-Tonelli multiexp mpz: Number p=%Zd is equal to 2^%d * %Zd.\n",p,max_even,odd_part);
#endif	
	mpz_t a,b,r,a_inverse,root,ratio,b_exp_2i,x,p_exp_alpha,pom;
	mpz_init_set(a,n);
	mpz_init_set_ui(b,0);
	mpz_init_set_ui(r,0);
	mpz_init_set_ui(a_inverse,0);
	mpz_init_set_ui(b_exp_2i,0);
	mpz_init_set_ui(p_exp_alpha,1);
	mpz_init_set_ui(root,0);
	mpz_init_set_ui(ratio,0);
	mpz_init_set_ui(pom,0);
	mpz_t non;
	mpz_init_set_ui(non,(unsigned int) 2);
	/* hledame kvadraticke nereziduum modulo p, a to prostym prohledavanim od dvojky vyse */
	while(mpz_kronecker(non,p)+1)
	{
		mpz_add_ui(non,non,1);
	}
#ifdef RESIDUE_DEBUG	
	gmp_printf("Shanks-Tonelli multiexp mpz: We have found a first nonresidue mod %Zd. It is equal to %Zd.\n",p,non);
#endif
	if (mpz_congruent_ui_p(p,3,4)) /* pripad p=3 mod 4 */
	{
#ifdef RESIDUE_DEBUG			
		gmp_printf("Shanks-Tonelli multiexp mpz: the easier branch.\n");
#endif
		mpz_add_ui(pom,p,1);
		mpz_divexact_ui(pom,pom,4);
		mpz_powm(root,a,pom,p);	
	}
	else
	{
		int j;
		mpz_powm(b,non,odd_part,p);	
		mpz_add_ui(aux_1,odd_part,1);
		mpz_fdiv_q_2exp(aux_1,aux_1,1);
		mpz_powm(r,a,aux_1,p);
		mpz_invert(a_inverse,a,p);
		mpz_set(root,b);
		mpz_set(b_exp_2i,b);		
		mpz_mul(pom,r,r);
		mpz_mul(pom,pom,a_inverse);
		mpz_mod(ratio,pom,p);
		mpz_set(pom,ratio);
		for (i=1;i<=(max_even-2);i++)
		{
			mpz_mul(pom,pom,pom);
			mpz_mod(pom,pom,p);
		}	
		if (!(mpz_cmp_ui(pom,1)))
		{
		/*		printf("j0 will be 0\n");*/
			mpz_set(root,r);
		}
		else
		{
			mpz_mul(root,b,r);
			mpz_mod(root,root,p);
		}
		for (i=1;i<=(max_even-2);i++)
		{
			mpz_mul(b_exp_2i,b_exp_2i,b_exp_2i);
			mpz_mod(b_exp_2i,b_exp_2i,p);
			mpz_mul(pom,root,a_inverse);
			mpz_mod(ratio,pom,p);
			mpz_mul(ratio,ratio,root);
			mpz_mod(ratio,ratio,p);
			for(j=1;j<=(max_even-2-i);j++)
			{
				mpz_powm_ui(ratio,ratio,2,p);
			}
			if (!mpz_cmp_ui(ratio,1))
			{
			/*	printf("j%d bude 0\n",i);*/
			}
			else
			{
				/*printf("j%d bude 1\n",i);*/
				mpz_mul(root,root,b_exp_2i);
				mpz_mod(root,root,p);
			}
		}

	} /* end of else */

	mpz_init(x);
	mpz_set(x,root);
	if (this->assertion_mode){ // ASSERTION: We check whether computed root is indeed sqrt of n_in mod p.
		mpz_mul(aux_1,x,x);		
		mpz_mod(aux_1,aux_1,p);
		mpz_mod(aux_2,n_in,p);
		if (mpz_cmp(aux_1,aux_2)){
			gmp_printf("\nAssertion failed (ShanksTonelliMultiexpMpz): %Zd^2 = %Zd != %Zd mod %Zd.\n",x,aux_1,aux_2,p);
		}
	}
#ifdef RESIDUE_DEBUG	
		gmp_printf("We have calculated root of %Zd mod %Zd as %Zd.\n",n_in,p,x);
		mpz_mul(aux_1,x,x);		
		mpz_mod(aux_1,aux_1,p);
		gmp_printf("Its square mod %Zd is %Zd.\n",p,aux_1);
		mpz_mod(aux_1,n_in,p);
		gmp_printf("%Zd mod %Zd is %Zd.\n",n_in,p,aux_1);
#endif					
	mpz_t inv;
	if (exp>1)
	{
		mpz_init(inv);
		mpz_mul_si(pom,root,-2);
		mpz_invert(inv,pom,p);
		int alpha;
		for (alpha=1;alpha<exp;alpha++) /* cast pro vyssi mocninu */
		{		
			mpz_mul(p_exp_alpha,p_exp_alpha,p);
			mpz_mul(pom,x,x);
			mpz_sub(pom,pom,a);
			mpz_divexact(b,pom,p_exp_alpha);
			mpz_mul(pom,b,inv);
			mpz_mod(pom,pom,p);	
			mpz_mul(pom,pom,p_exp_alpha);
			mpz_add(x,x,pom);		
			if (this->assertion_mode){ // ASSERTION: We check whether computed root is indeed sqrt of n_in mod p^(alpha+1).
				mpz_mul(aux_1,x,x);		
				mpz_pow_ui(aux_2,p,alpha+1);
				mpz_mod(aux_1,aux_1,aux_2);
				mpz_mod(aux_3,n_in,aux_2);
				if (mpz_cmp(aux_1,aux_3)){
					gmp_printf("\nAssertion failed (ShanksTonelliMultiexpMpz): %Zd^2 = %Zd != %Zd mod %Zd^%d.\n",x,aux_1,aux_3,p,alpha+1);
				}
			}

#ifdef RESIDUE_DEBUG	
			gmp_printf("We have calculated root of %Zd mod %Zd^%d as %Zd.\n",n_in,p,alpha+1,x);
			mpz_mul(aux_1,x,x);
			mpz_pow_ui(aux_2,p,alpha+1);
			mpz_mod(aux_1,aux_1,aux_2);
			gmp_printf("Its square mod %Zd^%d is %Zd.\n",p,alpha+1,aux_1);
			mpz_mod(aux_1,n_in,aux_2);
			gmp_printf("%Zd mod %Zd^%d is %Zd.\n",n_in,p,alpha+1,aux_1);
			
#endif				
		} // end of for(alpha ...)
		
	}
	mpz_set(solution,x);
	mpz_clear(a);
	mpz_clear(a_inverse);
	mpz_clear(b);
	mpz_clear(p_exp_alpha);
	mpz_clear(b_exp_2i);
	if (exp>1) mpz_clear(inv);
	mpz_clear(p_minus_1_mpz);
	mpz_clear(pom);
	mpz_clear(ratio);
	mpz_clear(root);
	mpz_clear(r);
	mpz_clear(x);
	mpz_clear(n);
	mpz_clear(non);
	return ConstRC::Ok;
}

/*!
	A trivial method returning in 'rop' the least absolute value of 'op' mod 'modulus'.
*/
void QuadraticSieve::ToLeastAbsoluteValue(mpz_t rop, mpz_t op, mpz_t modulus)
{
	mpz_t threshold;
	mpz_init(threshold);
	mpz_sub_ui(threshold,modulus,1);
	mpz_divexact_ui(threshold,threshold,2);
	if (mpz_cmpabs(op,threshold)>0)
	{
		if (mpz_cmp_ui(op,0)>0)  /* vstupni cislo bylo kladne */
		{
			mpz_sub(rop,op,modulus);
		}
		else
		{
			mpz_add(rop,op,modulus);
		}
	}
	mpz_clear(threshold);
}
// Calculates s such that 2^s t = p-1, where t is odd.
/*!
This method returns 's' such that 2^s t = p-1, where 't' is odd.
*/
int QuadraticSieve::MaxPowerOfTwo(long p){
	int res = 0; int p_m = p-1;
	while (!(p_m%2)){
		p_m >>= 1;
		res++;
	}
	return res;
}

/*!
	This method is used in search for optimal size of the sieving interval. This requires some
	reallocations of the sieving matrix, and this method makes the first step by freeing the matrix
	memory.
*/
void QuadraticSieve::ResetMainMatrix() {
	sieve_matrix* target = this->main_matrix;	
	if (target->assigned){
		free(target->accumulated_logarithm);
		target->accumulated_logarithm = NULL;
		target->assigned = 0;
		target->maximal_index = -1;
	}	
}

/*!
	This method refills the sieving matrix by bytes 0x80-T, where T is the threshold, whose exceeding makes
	a value of Q(x) suitable for detailed factorization. This is a kludge invented by Scott Contini; using
	this, we distinguish suitable Q(x) values by the fact that their entries in the sieving matrix exceed
	0x80. This can be tested effectively by taking the bytes in quadruples and testing whether (4 bytes) &
	0x80808080 == 0 or no.
	
	If the sieving matrix has not been allocated yet, it is allocated at first by this method.
	
	There is a small complication using this approach on very large numbers. There, the threshold T can well
	exceed 0x80, which makes this approach unusable. A small modification is needed then: instead of calculating
	all the intermediate values and T as logarithms in base 2, we may choose the base as 4 or 8. This, however,
	needs subtle modifications in other QS methods, and will be implemented only if really needed.
	
	Since this method is called at every sieving loop, it must be as fast as possible. However, there are no
	real problems with speed of this method, as it only serves as a caller for low-level memset() method,
	performing a few other operations.
	
	Return value: 
	- ConstRC::Ok if everything went smoothly
	- NOT_ENOUGH_MEMORY if the allocation was unsuccessful.

*/
int QuadraticSieve::RefillMainMatrix(){
	long current_minimal = 0;
	long current_maximal = 2*this->integer_M-1;
	long to_allocate = current_maximal - current_minimal+1;
	sieve_matrix* target = this->main_matrix;	
	if (!target->assigned){
		target->accumulated_logarithm = (log_type*) AllocationMachine((void**) &(target->accumulated_logarithm),to_allocate*sizeof(log_type));
		if (target->accumulated_logarithm == NULL) {
			return NOT_ENOUGH_MEMORY;
		}
		target->assigned = to_allocate;
		target->maximal_index = to_allocate-1;
	}
	// A clever trick due to Scott Contini. We will initialize the acc_log structure by 0x80 - T, then testing whether accumulated
	// logarithms have exceeded 0x80.
	if (this->contini_threshold & 0x80){
		this->contini_threshold = 0x80;
	}
	memset(target->accumulated_logarithm,0x80-this->contini_threshold,to_allocate*sizeof(log_type));
	return ConstRC::Ok;
}

/*!
	Both these methods perform the same operation and differ only in types of supplied arguments. They take
	the current sieving polynomial, given by index_in_polynomial, calculates its value in aArg and modulo
	aModulus. If the resul`t is nonzero, it prints out a warning saying "Assertion failed" and giving details
	of this failure.
	
	This method is used in assertion mode to test reliability of polynomial root calculation.

*/
void QuadraticSieve::AssertQuadPoly(mpz_t aArg, long aModulus){
	int poly_index = this->index_in_polynomial;
	quad_polynomial* poly = &(this->polynomials[poly_index]);
   	mpz_set(aux_a,aArg);
   	mpz_set(aux_1,poly->a);
	mpz_mul(aux_1,aux_1,aux_a); // ax
	mpz_add(aux_1,aux_1,poly->b); // ax+b
	mpz_mul(aux_1,aux_1,aux_a); // ax^2+bx
	mpz_add(aux_1,aux_1,poly->c); // ax^2+bx+c
	mpz_abs(aux_1,aux_1);
	mpz_set_ui(aux_p,aModulus);
	mpz_mod(aux_1,aux_1,aux_p);
	if (mpz_cmp_ui(aux_1,0)){
		gmp_printf("\nAssertion failed (assert quad poly #%d): %Zdx^2+%Zdx+%Zd mod %Zd = %Zd != 0.\n",poly_index,poly->a,poly->b,poly->c,aux_p,aux_1);
	}
}

void QuadraticSieve::AssertQuadPoly(long aArg, long aModulus){
	int poly_index = this->index_in_polynomial;
	quad_polynomial* poly = &(this->polynomials[poly_index]);
   	mpz_set_si(aux_a,aArg);
   	mpz_set(aux_1,poly->a);
	mpz_mul(aux_1,aux_1,aux_a); // ax
	mpz_add(aux_1,aux_1,poly->b); // ax+b
	mpz_mul(aux_1,aux_1,aux_a); // ax^2+bx
	mpz_add(aux_1,aux_1,poly->c); // ax^2+bx+c
	mpz_abs(aux_1,aux_1);
	mpz_set_ui(aux_p,aModulus);
	mpz_mod(aux_1,aux_1,aux_p);
	if (mpz_cmp_ui(aux_1,0)){
		gmp_printf("\nAssertion failed (assert quad poly #%d): %Zdx^2+%Zdx+%Zd mod %Zd = %Zd != 0.\n",poly_index,poly->a,poly->b,poly->c,aux_p,aux_1);
	}
}

/*!
	This method is similar to AssertQuadPoly methods, but it uses special qualities of Q(x) to determine
	the result faster; instead of aX^2 + bX + c, it calculates (aX + b)^2 - kN, where X = aArg, and modules
	the result of this calculation.

*/
void QuadraticSieve::AssertSpecialQuadPoly(long aArg, long aModulus){
	int poly_index = this->index_in_polynomial;
	quad_polynomial* poly = &(this->polynomials[poly_index]);
  	// We calculate (Ax+B)^2 - kN
   	mpz_set_si(aux_a,aArg);
   	mpz_set(aux_1,poly->a);
	mpz_mul(aux_1,aux_1,aux_a); // ax
	mpz_mul_ui(aux_1,aux_1,2);  // 2ax
	mpz_add(aux_1,aux_1,poly->b); // ax+b
	mpz_mul(aux_1,aux_1,aux_1); // (ax+b)^2
	mpz_sub(aux_1,aux_1,this->kN); 
	mpz_set_ui(aux_p,aModulus);
	mpz_mod(aux_1,aux_1,aux_p);
	if (mpz_cmp_ui(aux_1,0)){
		gmp_printf("\nAssertion failed (assert quad poly #%d): (2*%Zdx + %Zd)^2 - %Zd mod %Zd = %Zd != 0.\n",poly_index,poly->a,poly->b,this->kN,aux_p,aux_1);
	}
}

/*!
	This method is planned to perform the sieving in exact division mode, however, now, it only returns
	NOT_SUPPORTED value.

*/
int QuadraticSieve::SievingLoopExact(){
	return NOT_SUPPORTED;
}

/*!
	This method is planned to perform the sieving in rough division mode, however, now, it only returns
	NOT_SUPPORTED value.

*/
int QuadraticSieve::SievingLoopRough(){
	return NOT_SUPPORTED;
}

//#define LINE_SIEVER_DEBUG
#define BOUND 16384
/*!
	This method performs sieving in rough logarithmic mode - now, the only mode supported by MPQS/SIQS.
	This is a time-critical routine, which takes up as much as 50% of the overall MPQS/SIQS runtime.
	
	To achieve the high performance, the following techniques are used (working on x86 PC; other platforms
	may vary).
	   - variables used in inner loops are declared as register
	   - pointer arithmetic is used extensively
	   - tests of cycle end conditions are constructed as trivially as possible: for example
	     - end of FB is marked by zero integer
	     - end of sieving loops is tested at the inside of the loops and performed by break call
	     
	At the end of sieving, sieving for prime powers is called, if allowed.
	
	Return values:	
	- ConstRC::Ok normally
	- anything thrown by PerformSievingWithPrimePowersLogarithmic().     

*/
int QuadraticSieve::SievingLoopLogarithmic(){    
    register log_type log;
    int i = this->rough_factorbase_index - 1;    
    register main_sieving_type    p;    
    main_sieving_type*   fb 	  =  this->factor_base;    
    log_type* logs   =  this->binary_logarithms_of_factor_base_times_log_factor;
    isol_type* 	   usable =  this->int_sols_of_cur_poly.usable;
    isol_type*     sol_1  =  this->int_sols_of_cur_poly.sol_1;
    isol_type*     sol_2  =  this->int_sols_of_cur_poly.sol_2;
    log_type* 	   acc    =  this->main_matrix->accumulated_logarithm;
    log_type*      zero   =  acc - this->main_matrix->lower_arg;
#ifdef LINE_SIEVER_DEBUG 
    log_type*	   bound  =  zero + BOUND;
    int sum_minimal = 0;    
    int sum_middle  = 0;
    int sum_largish = 0;
    int hits_minimal = 0;
    int hits_middle  = 0;
    int hits_largish = 0;
#endif
    register log_type*      curr ;
    register log_type*      st1;
    register log_type*      acc_max=  acc + this->main_matrix->maximal_index;
#ifdef TIME_MESSAGE		    
    start = clock();
#endif    
    int local_logs_1 = 0;
    int local_iters_1 = 0;
    int local_logs_2 = 0;
    int local_iters_2 = 0;
    int hits = 0;
    //mpz_t logs_mpz;
    //mpz_init(logs_mpz);
    //unusables = 0;
//#define OLD_VARIANT
#ifdef OLD_VARIANT
    while ((p=fb[++i]) != 0) // The end of FB is marked by 0. This nice little trick was inspired by LiDIA
#else
    while (1)
#endif
    {   
#ifdef OLD_VARIANT
#else  
	  i++;
	  p = fb[i];
	  if (!p) break;
#endif
    	  //local_iters = 0;
      	  log = logs[i];	  	  
	  //	printf("\nSieving Loop Logarithmic: p = %d, log = %4x: ",p,log);	      	  	  

	  if ((usable[i]) & VALID_1){   // aserce:	  	
#ifdef OLD_VARIANT
		curr = zero+sol_1[i]-p;
#else
		st1 = zero+sol_1[i];
		curr = st1;
#endif
		while (1) {
#ifdef OLD_VARIANT
	  		if ((curr+=p) > acc_max) break;
			*(curr) += log;					 
#else					
		  	if (curr > acc_max) break;
			*(curr) += log;		
			curr += p;	 
			//hits++;
#endif
#ifdef LINE_SIEVER_DEBUG 
			if (curr <= bound)
			{
				if (p < 256) 
				{
					sum_minimal += log;
					hits_minimal++;
				} else if (p < BOUND)
				{
					sum_middle += log;
					hits_middle ++;
				} else {
					sum_largish += log;
					hits_largish++;
				}
			}
#endif	           	
			//printf(" %x | ",*curr);
			/*local_iters_1++;
			local_logs_1+= log;
			prime_logs_1[i] += log;
		        prime_iters_1[i]++;*/
			//mpz_add_ui(logs_mpz,logs_mpz,log);
		 }
#ifdef OLD_VARIANT
		
		curr = zero+sol_1[i];		
#else
		curr = st1;
#endif		 
		 
		while (1) {		
		  	if ((curr-= p) < acc) break;
			*(curr) += log;			 	           	
			//hits++;
			//printf(" %x | ",*curr);
			/*local_iters_1++;
			local_logs_1+= log;
			//mpz_add_ui(logs_mpz,logs_mpz,log);
			 prime_logs_1[i] += log;
	  		prime_iters_1[i]++;*/
		 }
	  } // end of if (first_valid)	  
	  if ((usable[i]) & VALID_2){	  	
#ifdef OLD_VARIANT
		curr = zero+sol_2[i]-p;
#else
		st1 =  zero+sol_2[i];
		curr = st1;
#endif
		while (1) {
#ifdef OLD_VARIANT
	  		if ((curr+=p) > acc_max) break;
			*(curr) += log;					 
#else					
		  	if (curr > acc_max) break;
			*(curr) += log;		
			curr += p;	 
			//hits++;
#endif
#ifdef LINE_SIEVER_DEBUG 
			if (curr <= bound)
			{
				if (p < 256) 
				{
					sum_minimal += log;
					hits_minimal++;
				} else if (p < BOUND)
				{
					sum_middle += log;
					hits_middle ++;
				} else {
					sum_largish += log;
					hits_largish++;
				}
			}
#endif	           	
	
			//printf(" %x | ",*curr);
			/*-local_iters_2++;
			local_logs_2+= log;
			//mpz_add_ui(logs_mpz,logs_mpz,log);
			prime_logs_2[i] += log;
		        prime_iters_2[i]++;*/

		} 
#ifdef OLD_VARIANT
		curr = zero + sol_2[i];		
#else
		curr = st1;
#endif
		
		while (1) {		
		  	if ((curr-=p) < acc) break;
			*(curr) += log;			 	           	
			//hits++;
			//printf(" %x | ",*curr);
			/*local_iters_2++;
			local_logs_2+= log;
			//mpz_add_ui(logs_mpz,logs_mpz,log);
			prime_logs_2[i] += log;
		        prime_iters_2[i]++;*/
		 }
	} // end of if (second_valid)	
	//else unusables++;
	//iters += local_iters;
	//printf("Prime %5d: local iters %5d, total iters %7d.\n",p,local_iters,iters);
     } // end of while ((p=*fb++) != 0     
     //gmp_printf("\nSieving Loop Logarithmic: iters %d, logs %Zd.\n",iters,logs_mpz);
     //printf("Sieving Loop Logarithmic: unusables %d.\n",unusables);
     //mpz_clear(logs_mpz);     
     if (this->power_sieving_mode == ESieveWithPrimePowers){
     	int ret = PerformSievingWithPrimePowersLogarithmic();
	if (ret != ConstRC::Ok) return ret;
     }
     //printf("\nSieving: hits %d.\n", hits);
#ifdef LINE_SIEVER_DEBUG 
	printf("\nSieving, interval 0 to %5d:\n",BOUND);
 	printf("Minimal primes (< %6d): sum %7d in %5d hits (avg %5d).\n",256,sum_minimal,hits_minimal,sum_minimal/hits_minimal);
	printf("Middle  primes (< %6d): sum %7d in %5d hits (avg %5d).\n",BOUND,sum_middle,hits_middle,sum_middle/hits_middle);
	printf("Largish primes (> %6d): sum %7d in %5d hits (avg %5d).\n",BOUND,sum_largish,hits_largish,sum_largish/hits_largish);
	wait();
#endif	           	

     total_logs_1 += local_logs_1;
     iters_1 += local_iters_1;
     total_logs_2 += local_logs_2;
     iters_2 += local_iters_2;     
#ifdef TIME_MESSAGE		     
     int diff = clock() - start;
     if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
	total_inner += (double) diff;
     }		     
#endif     
     return ConstRC::Ok;
}

/*!
	This method is planned to perform the factorizations of 'candidate values' in exact division mode, 
	however, now, it only returns NOT_SUPPORTED value.

*/
int QuadraticSieve::SortSmoothValuesExact(){
     return NOT_SUPPORTED;
}

/*!
	This method is planned to perform the factorizations of 'candidate values' in exact division mode, 
	however, now, it only returns NOT_SUPPORTED value.

*/
int QuadraticSieve::SortSmoothValuesRough(){
     return NOT_SUPPORTED;
}

/*!
	
	This method searches through the main sieving matrix and detects entries suitable for detailed
	factorization. The search is performed by technique of Scott Contini: re-casting the array to
	integer type and testing the entries by AND 0x80808080 != 0. If a suitable entry is found, the
	corresponding Q(x) is calculated, factorized into primes, and tested for smoothness or partial
	smoothness.
	
	Since this method is called at every sieving loop, it must be highly optimized to yield optimal
	performance. There are two main problems involving speed of this method:
	  - determination of optimal "test threshold" T, which is beyond scope of this documentation
	    (see theory). This threshold determines amount of values tested.
	  - programmatic solution of the factorization, so that it would be fast enough.
	To achieve the second goal on x86 PCs, variety of techniques is used:
	  - Q(x) is calculated as (2ax+b)^2 - kN.
	  - maximal power of 2 dividing Q(x) is determined by scanning zero bits from the right. Since
	    4 | Q(x) for each x, this is always effective.
	  - pointer arithmetics is used extensively
	  - testing for end of FB is done by comparation of p to 0.
	  - testing whether p | Q(x) is done with use of relatively sophisticated code:
	    Aim of this code is to replace as much mods (%) by less expensive comparations (>, >=),
	    shifts (<<) and additions (+,-) as possible, with regard to effectivity. This code speeds 
	    up the calculation by 30%.      
	  - If SIQS method of sieving is used, a specific situation is solved: that is, for p | 'a',
	    Q(x) degenerates to linear equation mod p, but Q(x) may still have 2 solutions mod p^2.
	    There is an attempt to determine such divisors if the remnant after earlier division
	    fits into signed int, and it can be found in loop headed by
	    	if (!one && mpz_fits_sint_p (candidate.remaining))	
			
	If a value of Q(x)is found to be smooth or *-partial, it is put into corresponding structure
	(table, file).
	
	Return values:
	- ConstRC::Ok if everything went smoothly	
	- NOT_ENOUGH_MEMORY if there were not enough memory to allocate or reallocate list of divisors
	(very improbable, since it is only a few integers).

*/
int QuadraticSieve::SortSmoothValuesLogarithmic(){
    // Pretty long declaration of intermediate variables
    int ret;
    int accepted = 0;
    int rejected = 0;
    int max =  this->maximal_factor_base_index+1; 
    int one = 0;
    register int current_argument;      
    int collected_a_divisors = 0;
        
    qs_relation* candidate = InitRelation();
    if (candidate == NULL) {
	return NOT_ENOUGH_MEMORY;
    }	

    main_sieving_type last_written_factor;
    main_sieving_type* fb_j;    
    register main_sieving_type p, minus_p;
    int max_matrix = this->main_matrix->assigned;
    int  checkall  = (this->checking_mode == ECheckAll);
    log_type* acc_logs = this->main_matrix->accumulated_logarithm;    
    int local_start = 0;
    register int l;
    int i, j, mm, exponent;
    
    isol_type* usable = this->int_sols_of_cur_poly.usable;
    isol_type* sol_1  = this->int_sols_of_cur_poly.sol_1;
    isol_type* sol_2  = this->int_sols_of_cur_poly.sol_2;
    isol_type  this_usable;    
    
    mpz_t a;
    mpz_t b;
    mpz_t d;
    
    mpz_init_set(a, this->polynomials[index_in_polynomial].a);
    mpz_init_set(b, this->polynomials[index_in_polynomial].b);
    mpz_init_set(d,this->polynomials[index_in_polynomial].d);  
      
    int type_ratio = sizeof(int) / sizeof(log_type);
    int mask = 0;
    for (i=0; i< type_ratio; i++){
    	mask |= (TEST_VALUE << 8*i);
    }
    
    int max_cycle = max_matrix / type_ratio;    
    bool write;

    /*
        A clever trick by Scott Contini. Instead of testing each entry of acc_logs, we only test acc_logs_as_ints for nonzeroity.
	This is a scarce situation - with at least one of the entries being a candidate for trial division.
    */
    int* acc_logs_as_ints = (int*) acc_logs;
    int min_1;
    int max_1;
    bool divA;
#ifdef LINE_SIEVER_DEBUG
    int logsizes[10];
    for (mm=0; mm< 10; mm++)
	{
	 logsizes[mm] = 0;
	}
#endif
    for (mm = 0; mm < max_cycle; mm++){
    	if ((acc_logs_as_ints[mm] & mask) || checkall){
		min_1 = mm*type_ratio;
		max_1 = (mm+1)*type_ratio - 1;
		for (i = min_1; i< max_1; i++){
			exponent = 0;   
        		local_size_statistics[acc_logs[i]/0x10]++;
			if ((acc_logs[i] & 0x80) || checkall){

			
			// We calculate all parameters.
			// We calculate left and right side value - that is (Ax+b) on left side, d on the right side.
			// Also, we must calculate factorized_value = remaining. It is equal to (Ax+b)^2 - kN, and if we use MPQS, it is divided by a. 						
	
			current_argument = this->main_matrix->lower_arg+i; // Current value of x. Will be used frequently in divisor search.
#ifdef LINE_SIEVER_DEBUG
			if ((current_argument > 0) && (current_argument <= BOUND))
			{
				if ((acc_logs[i] / 16) < 10)
				{
					logsizes[acc_logs[i] / 16]++;
				}
			}
#endif
			mpz_set_si(candidate->argument,current_argument); 
			candidate->lprime1 = 1;	// Setting all x-partial-specific values to 1			
			candidate->lprime2 = 1;					
			mpz_mul(candidate->remaining,candidate->argument,a);	// ax
			mpz_mul_2exp (candidate->remaining,candidate->remaining,1); // 2ax
			mpz_add(candidate->remaining,candidate->remaining,b); // 2ax+b
			mpz_set(candidate->left_side_value,candidate->remaining); // square root on the left side
			mpz_mul(candidate->remaining,candidate->remaining,candidate->remaining); // (2ax+b)^2
			mpz_sub(candidate->remaining,candidate->remaining,this->kN); // (2ax+b)^2 - kN; factorized value on the right side
			if (this->assertion_mode){
				if (!mpz_divisible_p(candidate->remaining,a)) {
					printf("Assertion failed:  a does not divide Q(x)!\n");
				}
			}
			/*if (index_in_polynomial == 25)
			{
				gmp_printf("a = %Zd, b = %Zd, d = %Zd, kN = %Zd\n",a,b,d,kN);
				gmp_printf("The value at %d is %Zd.\n",current_argument,candidate->remaining);
				wait();
			}*/
			// The factorized value (2ax+b)^2 - kN is always divisible by 'a'. We factor only the quotient.
			mpz_fdiv_q(candidate->remaining,candidate->remaining,a);
			/*if (index_in_polynomial == 25)
			{
				gmp_printf("The value after division by 'a' is %Zd.\n",candidate->remaining);
				wait();
			} */ //  ((2ax+b)^2 mod kN)/a		
			if ((this->QSMode == EMpqs) || (this->QSMode == EOldMpqs)) { 						
				mpz_set(candidate->right_side_value,d); // MPQS: a square factor on the right side.		
			} else {
				mpz_set_ui(candidate->right_side_value,1); // Siqs: no square factor on the right side. Factors of a will be among small divisors of the right side.
			}
			mpz_set(candidate->factorized_value,candidate->remaining);
			total_tested++;
			// Now, the initialization of the Q(x) value is over and search for its small divisors
			// begins.
			// This will be the index of the last written small factor on the right side
			last_written_factor = -1;        	        
			// We anticipate max_factor >= 2 (that is, room for at least 2 divisors is allocated) 		
	            	// In case of negative value, its first divisor is -1
        	    	if (mpz_cmp_si(candidate->remaining,0) < 0) {              	
#ifndef USE_PRIMES	
              			(candidate->elements[++last_written_factor]).p = 0;
#else
				(candidate->elements[++last_written_factor]).p = -1;
#endif				
              			(candidate->elements[last_written_factor]).exponent = 1;
              			mpz_mul_si(candidate->remaining,candidate->remaining,-1);
            		}			
			// Case of p=2. This should always divide Q(x) in at least second power.
			exponent = mpz_scan1(candidate->remaining,0);
			if (exponent){
				mpz_fdiv_q_2exp(candidate->remaining,candidate->remaining,exponent);					
#ifndef USE_PRIMES	
              			(candidate->elements[++last_written_factor]).p = 1;
#else
				(candidate->elements[++last_written_factor]).p = 2;
#endif				
              			(candidate->elements[last_written_factor]).exponent = exponent;              		
			} 
			if (exponent < 2) {
				printf("Assertion failed: Q(x) not divisible by 4!\n");		
			}
			// This variable takes values 0 or 1 and shows whether remaining == 1 without need to call mpz_cmp
			one = 0;
			// Index of the last small prime in factorization of Q(x). Mostly 0.
			exponent = 0;	
			// In case of SIQS, number of 'a' divisors already met on the way.
			collected_a_divisors = 0;			    
		 	// Now, we divide the remaining value by all primes
            		fb_j = factor_base+1;
			j = 1;
			while ((p = *(++fb_j))){				
				j++;
	    			this_usable = *(usable+j);              						
				/* if this line is divisible by p ...
				This is a sophisticated optimization. We want to perform single-precision operation mod (%)
				only as often as really needed. Expecting that an 'if' operation is much less time-expensive
				than a single precision mod (%), we test at first the size of the current argument.
				If  arg in [-p,0), it is enough to add a single value of p to l. If it is in interval [0,p],
				nothing must be done to get the real value of 'l'. All these extra operations (if, if, += are
				together less expensive than a single %. This is especially true when large primes (over M)
				are used.
				This unelegant code speeds up the run of the whole SortSmoothValues method by 20% - on Athlon XP 1600
				(61dec: 24.3 sec vs. 19.3 dec.).*/
				l = current_argument;
				if (l < 0) {					
					if (l >= (minus_p = (-1)*p)) { l += p; }					
					else if (l >= (minus_p <<= 1)) { l -= minus_p; }
					else if	((l %= p) < 0) { l+= p; }
				}
				else if (l >= p) {
					if (l < (p << 1)) {
						l -= p;
					} 
					else {
						l %= p;
					}
				}								
				// in case of SIQS, divisors of 'a' must be put into the list of all divisors in at least first power. We do not check for higher powers of those yet and leave it for later processing.
				divA = FALSE;
				if ((((this->QSMode) == ESiqs) || (this->QSMode == EOldSiqs)) && (is_this_prime_divisor_of_a[j]))	{				
					collected_a_divisors++;					
					exponent = 1;										
				}				
				if (( (this_usable & VALID_1) && ( l == sol_1[j])) || (((this_usable & VALID_2) && (l == sol_2[j])))) {					
					if (exponent == 1)
					{
					        divA = TRUE;
					}
					while( ! mpz_fdiv_qr_ui(aux_1,aux_2,candidate->remaining,p) ){
#ifdef TIME_MESSAGE				
	
						total_divrem++;
#endif					
						
						exponent++;
						mpz_swap(candidate->remaining,aux_1);
					} 
					if (this->assertion_mode)
					{
						if (!exponent)
						{
							printf("\nSortSmoothValuesLogarithmic: Divisibility problem on main matrix row [%d]: No factor %d at argument %d!\n",i,p,current_argument);
							printf("Sol1: %d. sol2: %d, l=%d.",sol_1[j],sol_2[j],l);
						}
					}
#ifdef TIME_MESSAGE					
					total_divrem++; // this is for the last operation
#endif					
				}
				if (divA)
				{
					if (exponent == 1) 
					{
						CLog::LogError(WHEREARG,"Quadratic sieve","sorting smooth",p,sol_1[j],"Problem with divisibility by 'a' divisor.");
					} /*else {
						CLog::LogError(WHEREARG,"Quadratic sieve","sorting smooth",p,exponent,"Divisibility by 'a' divisor was good");
					}*/
				}
				if (exponent){									
					// If the space for divisors is full, it must be enlarged by realloc.
                       			if ((++last_written_factor) > (candidate->currently_assigned_sieving_elements-1)){
						candidate->currently_assigned_sieving_elements += div_allocate;
						candidate->elements = (sieving_element*) AllocationMachine((void**) &candidate->elements,candidate->currently_assigned_sieving_elements*sizeof(sieving_element));
						if (!candidate->elements)
						{
							CLog::LogError(WHEREARG,"Quadratic sieve","sorting smooth",0,candidate->currently_assigned_sieving_elements*sizeof(sieving_element),"Not enough memory for reallocing the element field.");
						 	// give a warning and exit
							return NOT_ENOUGH_MEMORY;
						}
                        		}
					// Now, after potential realloc, it is safe to write to index last_written_factor
#ifndef USE_PRIMES	
              				(candidate->elements[last_written_factor]).p = j;
#else
					(candidate->elements[last_written_factor]).p = p;
#endif				

					candidate->elements[last_written_factor].exponent = exponent;
					if (!mpz_cmp_ui(candidate->remaining,1)) {
						// We have found all divisors of Q(x), and therefore we can end our search
						// for more divisors. The only possibly remaining divisors are divisors of 'a' in SIQS. Could be optimized a little.
						one = 1;
						// we should not end the divisor-finding loop before all divisors of a are included.
						if ((this->QSMode == EMpqs) || (((this->QSMode == ESiqs) || (this->QSMode == EOldSiqs)) && (collected_a_divisors == this->s))){
   						 	break;
						} 
					}
					// Reset exponent to 0 again for the next iteration(s).
					exponent = 0;
                		} // end of if(exponent)
                	} // end of while (*(++fb_j ....
			candidate->maximal_sieving_element_index = last_written_factor;			

			/* Now, almost all factors of this row have been determined. 
			As far ar the divisors are concerned, there is an important exception in case of SIQS. In this case,
			we have solution of the linear equation Q(x) mod p for all p dividing 'a'; however, Q(x) is a quadratic
			equation mod p^2, and might have extra solutions as such. These solutions cannot be caught in the previous
			process; on the other hand, trial division of all the tested values by all the primes dividing 'a' is
			not worth the effort (it might raise the number of multiple precision operations by more than 25%).
			A short recipe against this is the following: if the remaining part of Q(x) can be exported losslessly
			into a normal main_sieving_type (usually signed int) variable, we export it and try to perform the operations in single precision. This does not catch all instances of the previous problem, but most of them.
			Single precision operations are not so heavy, and do not involve runtime more than
			by 1/1000. */
			
#ifdef TIME_MESSAGE		     
		        local_start = clock();
#endif     								
			
			TrialDivideByADivisors(candidate,&one);			
			
#ifdef TIME_MESSAGE		     
		        int diff = clock() - local_start;
 			if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
				total_divisors_of_a += (double) diff;
     			}		     
#endif     				
			
			if (this->assertion_mode){ // We will check whether sorting was correct - e.g., whether we did not miss some of the legitimate divisors
				for (j=2; j < max; j++){
					if (mpz_divisible_ui_p(candidate->remaining,factor_base[j])){
						printf("\nSortSmoothValuesLogarithmic: Assertion failed on main matrix row [%d]: Forgotten factor %d!",i,factor_base[j]);
						PrintRelation(candidate);
						printf("\n");
						printf("Usable: %x, sol_1 = %d, sol_2 = %d.\n",*(usable+j),sol_1[j],sol_2[j]);              							
					}
				}
			}
			
			if (this->assertion_mode){			
		    		AssertRelationConsistency(candidate, "Problem", 100);
			}
			ret = SaveCandidate(candidate,accepted, rejected);
			if (ret != ConstRC::Ok)
			{
				CLog::LogError(WHEREARG,"Quadratic sieve","sorting smooth",0,ret,"Some problem in saving a candidate.");
				return ret;
			}
			/*if (one){
				// At first we check whether this row is nonzero. We do not want to include zero rows in smooth matrix.
				if (CheckNulity(candidate) == NONZERO){		
					// Must be defined before writing to file/memory:
					candidate->index_of_used_polynomial = this->current_polynomial;
					//accumulated_sizes_smooth += (mpz_sizeinbase(candidate->factorized_value,10) - mpz_sizeinbase(a,10));
					//accumulated_sizes_accepted += (mpz_sizeinbase(candidate->factorized_value,10) - mpz_sizeinbase(a,10));			
					accepted++;
					accepted_nr++;					
					PrintRelationToFile(smooth_values_file,candidate,this->so_far_smooths_found++);
					if (this->value_distribution_mode) {
						AddDistributionInfo(candidate,SMOOTH_VALUE);
					}
				}
			}
			// Case of a partially smooth value.
			else if (this->using_variations != EDoNotUseVariations) {
				write = false;
				// 1-partial.
			 	if (mpz_cmp(candidate->remaining,this->maximal_partial_factor) <= 0) { 			candidate->index_of_used_polynomial = this->current_polynomial;	
					candidate->lprime1 = mpz_get_ui(candidate->remaining);
					candidate->lprime2 = 1;
					mpz_set_ui(candidate->remaining,1);					
					accepted++;
					accepted_nr++;
					so_far_1_partials_found++;
					write = true;
				} else
				// 2-partial
				if ((this->using_variations == EDoubleLargePrimeV) && (mpz_cmp(candidate->remaining,this->maximal_candidate_value) <= 0) &&
				(mpz_cmp(candidate->remaining,this->minimal_candidate_value) >= 0)) {
					
					// Then, we test the remaining part for primality
					if (!mpz_probab_prime_p(candidate->remaining,25)) {
						ret = FactorizeCandidate(candidate);		
						if (ret == ConstRC::Factorized) {											
							mpz_set_ui(candidate->remaining,1);
							accepted++;
							accepted_nr++;
							if (candidate->lprime1 == candidate->lprime2) {
							// We have incidentally found a smooth value!
							// This happens surprisingly often in case of small FB.
							  	mpz_mul_ui(candidate->right_side_value,candidate->right_side_value,candidate->lprime1);
								candidate->lprime1 = 1;
								candidate->lprime2 = 1;
								PrintRelationToFile(smooth_values_file,candidate,this->so_far_smooths_found++);
								if (this->value_distribution_mode) {
									AddDistributionInfo(candidate,SMOOTH_VALUE);
								}
								continue; // we continue to the next candidate
							}
							so_far_2_partials_found++;
							
							write = true;
						}
						else if (ret == NOT_FACTORIZED) {
							// Unable to factorize a 2-partial. Sorry.
							rejected++;
							rejected_nr++;
						} else if (ret == FACTOR_TOO_LARGE) {
							// one of the large primes is too large
							rejected++;
							rejected_nr++;
						} else {
							// Any other serious problem.
							ClearRelation(candidate);							
							return ret;
						}
					} else {
						rejected++;
						rejected_nr++;
					}
				} else {
					rejected++;
					rejected_nr++;
				}
				// An acceptable x-partial; will be saved into a file and added into graph structures.
				if (write) {	
#ifdef TIME_MESSAGE		     
					int local_start = clock();
#endif  					
					AddToCycleCountStructures(candidate);		
#ifdef TIME_MESSAGE		     
					int diff = clock() - local_start;
    					if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
						total_cyclecount += (double) diff;
    					}		     
#endif     		    					
					PrintRelationToFile(partials_file,candidate,this->so_far_x_partials_found++);
					if (this->value_distribution_mode) {
						if (candidate->lprime2 == 1) {
							AddDistributionInfo(candidate,PARTIAL_1_VALUE);
						} else {
							AddDistributionInfo(candidate,PARTIAL_2_VALUE);
						}
					}									
				} 
				
			   }*/ // end of else if (partial) 
			} // end of if ((acc_logs[i] >= max_acce) || checkall) 
		
		} // end of for (i = min_1; i< max_1; i++)
	}  // end of if ((acc_logs_as_ints & mask) || checkall)
    }
#ifdef LINE_SIEVER_DEBUG
	for (mm=0; mm< 10; mm++)
	{
		printf("      to 0x%d0: %d.\n",mm,logsizes[mm]);
	}
#endif

    ClearRelation(candidate);
    free(candidate);
    candidate = NULL;
    mpz_clear(a); mpz_clear(b); mpz_clear(d);    
    if (this->printing_sort == EPrintSort) printf(" %d/%d accepted/rejected\n",accepted,rejected);

    if (this->open & SMOOTHS_OPEN) fflush(smooth_values_file);
    if (this->open & PARTIALS_OPEN) fflush(partials_file);

    return ConstRC::Ok;
}

int  QuadraticSieve::SaveCandidate(qs_relation* candidate, TCandidateType type)
{
  //   	AssertRelationConsistency(candidate,"SaveCandidate - type", type);    
	int ret = ConstRC::Ok;
// Smooth value
	if (type == ESmoothCandidate)
	{		
		//gmp_printf("Save candidate, Smooth value\n");
		if (CheckNulity(candidate) == NONZERO){		
			// Must be defined before writing to file/memory:
			candidate->index_of_used_polynomial = this->current_polynomial;
			accepted_nr++;					
			PrintRelationToFile(smooth_values_file,candidate,this->so_far_smooths_found++);
			if (this->value_distribution_mode) {
				AddDistributionInfo(candidate,SMOOTH_VALUE);
			}
		}
	}
	if (type == EPartialCandidate)
	{
		//gmp_printf("Save candidate, Smooth value\n");
		if (CheckNulity(candidate) == NONZERO){		
			// Must be defined before writing to file/memory:
			candidate->index_of_used_polynomial = this->current_polynomial;
			accepted_nr++;					
			if (candidate->lprime2 == 1)
			{
				so_far_1_partials_found++;
			} else
			{
				so_far_2_partials_found++;
			}				
			accepted_nr++;			
			PrintRelationToFile(partials_file,candidate,this->so_far_x_partials_found++);
			
			AddToCycleCountStructures(candidate);		
		}
	}
}

int  QuadraticSieve::SaveCandidate(qs_relation* candidate, int& accepted, int& rejected)
{
     //	AssertRelationConsistency(candidate,"SaveCandidate", accepted);    
	int ret = ConstRC::Ok;
	bool write = FALSE;
	//gmp_printf("Save candidate, remaining = %Zd.\n",candidate->remaining);
	bool isSmooth = !mpz_cmp_ui(candidate->remaining,1);
	bool isPartial = !isSmooth;

	// Smooth value
	if (isSmooth)
	{		
		//gmp_printf("Save candidate, Smooth value\n");
		if (CheckNulity(candidate) == NONZERO){		
			// Must be defined before writing to file/memory:
			candidate->index_of_used_polynomial = this->current_polynomial;
			//accumulated_sizes_smooth += (mpz_sizeinbase(candidate->factorized_value,10) - mpz_sizeinbase(a,10));
			//accumulated_sizes_accepted += (mpz_sizeinbase(candidate->factorized_value,10) - mpz_sizeinbase(a,10));			
			accepted++;
			accepted_nr++;					
			PrintRelationToFile(smooth_values_file,candidate,this->so_far_smooths_found++);
			if (this->value_distribution_mode) {
				AddDistributionInfo(candidate,SMOOTH_VALUE);
			}
			if (iOptimizer)
			{
				iOptimizer->IncrementAccepted();
			}
		}
	}
	else if ((this->using_variations != EDoNotUseVariations) && (isPartial)) {
		//gmp_printf("Save candidate, May be partial value\n");
		write = false;
		// 1-partial.
		if (mpz_cmp(candidate->remaining,this->maximal_partial_factor) <= 0) {
 			candidate->index_of_used_polynomial = this->current_polynomial;	
			candidate->lprime1 = mpz_get_ui(candidate->remaining);
			candidate->lprime2 = 1;
			mpz_set_ui(candidate->remaining,1);					
			accepted++;
			accepted_nr++;
			so_far_1_partials_found++;
			if (iOptimizer)
			{
				iOptimizer->IncrementAccepted();
			}
			//gmp_printf("Save candidate, is a 1-partial value, LP = %d\n",candidate->lprime1);
			write = true;
		} else
		// 2-partial
		if ((this->using_variations == EDoubleLargePrimeV) && (mpz_cmp(candidate->remaining,this->maximal_candidate_value) <= 0) &&
		(mpz_cmp(candidate->remaining,this->minimal_candidate_value) >= 0)) {
			// Then, we test the remaining part for primality
			if (!mpz_probab_prime_p(candidate->remaining,25)) {
				candidate->index = -1;
				ret = FactorizeCandidate(candidate);		
				if (ret == ConstRC::Factorized) 
				{	
					mpz_set_ui(candidate->remaining,1);
					accepted++;
					accepted_nr++;
					if (iOptimizer)
					{
						iOptimizer->IncrementAccepted();
					}
					if (candidate->lprime1 == candidate->lprime2) {
						// We have incidentally found a smooth value!
						// This happens surprisingly often in case of small FB.
					  	mpz_mul_ui(candidate->right_side_value,candidate->right_side_value,candidate->lprime1);
						candidate->lprime1 = 1;
						candidate->lprime2 = 1;
						PrintRelationToFile(smooth_values_file,candidate,this->so_far_smooths_found++);
						if (this->value_distribution_mode) {
							AddDistributionInfo(candidate,SMOOTH_VALUE);
						}
						return ConstRC::Ok; // we continue to the next candidate
					}
					so_far_2_partials_found++;			
					//gmp_printf("Save candidate, is a 2-partial value, LP = %d\n",candidate->lprime1,candidate->lprime2);
					if ((candidate->lprime1 <= 10000)|| (candidate->lprime2 <= 10000))
					{
						printf("Error in 2-partial variation, probably SIQS-related.\n");
						PrintRelation(candidate);
						ABNORMAL_EXIT(1);
					}
					write = true;
				}
				else if (ret == ConstRC::NotFactorized) {
					// Unable to factorize a 2-partial. Sorry.
					//gmp_printf("Save candidate, not factorized, sorry.\n");
					rejected++;
					rejected_nr++;
					if (iOptimizer)
					{
						iOptimizer->IncrementRejected();
					}
				} else if (ret == ConstRC::FactorTooLarge) {
					// one of the large primes is too large
					//gmp_printf("Save candidate, factor too large.\n",this->maximal_candidate_value);
					rejected++;
					rejected_nr++;
					if (iOptimizer)
					{
						iOptimizer->IncrementRejected();
					}
				} else {
					// Any other serious problem.
					CLog::LogError(WHEREARG,"Quadratic sieve","SortSmoothValuesLogarithmic",ret,0,"other serious problem");				
					ClearRelation(candidate);				return ret;
				}
			} else {
				//gmp_printf("Save candidate, is prime, sorry.\n");
				rejected++;
				rejected_nr++;
				if (iOptimizer)
				{
					iOptimizer->IncrementRejected();
				}

			}
		} else {
			//gmp_printf("Save candidate, Probably too big (> %Zd).\n",this->maximal_candidate_value);
			rejected++;
			rejected_nr++;
			if (iOptimizer)
			{
				iOptimizer->IncrementRejected();
			}

		}
		// An acceptable x-partial; will be saved into a file and added into graph structures.
		if (write) {	
#ifdef TIME_MESSAGE		     
			int local_start = clock();
#endif  				
			AddToCycleCountStructures(candidate);		
#ifdef TIME_MESSAGE		     
			int diff = clock() - local_start;
    			if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
				total_cyclecount += (double) diff;
    			}		     
#endif     		    					
			PrintRelationToFile(partials_file,candidate,this->so_far_x_partials_found++);
			if (this->value_distribution_mode) {
				if (candidate->lprime2 == 1) {
					AddDistributionInfo(candidate,PARTIAL_1_VALUE);
				} else {
					AddDistributionInfo(candidate,PARTIAL_2_VALUE);
				}
			}									
		} // if (write)				
	} // end of variation
	return ConstRC::Ok;
}


/*!
	This method is used in value distribution update. It calculates the distance between argument 'x'
	in aCandidate and the closer of the two polynomial roots, and updates the distribution statistics
	accordingly. The aType variable serves for distinguishing smooth, 1-partial and 2-partial relations,
	though, possibly, the method could distinguish them itself in the future.

*/
void QuadraticSieve::AddDistributionInfo(qs_relation* aCandidate, int aType) {
    int arg = mpz_get_si(aCandidate->argument);
    //printf("\nAdd distr.info: argument = %d, root1 = %d, root2 = %d.\n",arg,root_1,root_2);
    // spocitame min abs (arg-root_1), arg-root_2
    int d1 = arg - root_1;    
    if (d1 < 0) {d1 *= -1; }    
    int d2 = arg - root_2;
    if (d2 < 0) {d2 *= -1; }
    int dist = ( d1 < d2 ? d1 : d2);
    //printf("                 d1 = %d, d2 = %d. Vybrano %d.\n",d1,d2,dist);
    int step = dist / DISTRIBUTION_STEP;
    if (step > MAX_DISTANCE) {
    	step = MAX_DISTANCE;
    }
    switch (aType) {
    	case SMOOTH_VALUE:	this->smooths_distribution[step]++;
				break;
	case PARTIAL_1_VALUE:	this->part_1_distribution[step]++;
				break;
	case PARTIAL_2_VALUE:	this->part_2_distribution[step]++;				
    				break;
    }
}

/*!
	This method prints out the distribution info - in steps determined by DISTRIBUTION_STEP variable,
	which is currently defined as 10000. It always prints out distribution info about smooth values;
	the corresponding info about 1-partials and 2-partials is printed out if and only if such
	relations have been collected (e.g. -v or -v2 has been used).

*/
void QuadraticSieve::PrintDistributionInfo() {
   // print out the most interesting distribution details: first 20 K
   int i, max = 19;
   double percent = 0;
   double total = 0;
   printf("\n  Value Distribution: smooth values.\n");
   printf(" From - To: amount - perc (cumul).\n");   
   for (i=0; i<= max; i++) {
   	percent = this->smooths_distribution[i]*100 / (double) this->so_far_smooths_found;
	total += percent;
   	printf("%7d - %7d: %10d - %2.2f %%  (%5.2f %%)\n",i*DISTRIBUTION_STEP,(i+1)*DISTRIBUTION_STEP-1,this->smooths_distribution[i],percent,total);
   }
   if (this->using_variations != EDoNotUseVariations) {
   	total = 0;
	printf("\n  Value Distribution: 1-partial values.\n");
   	printf(" From - To: amount - perc (cumul).\n");   
   	for (i=0; i<= max; i++) {
   		percent = this->part_1_distribution[i]*100 / (double) this->so_far_1_partials_found;
		total += percent;
   		printf("%7d - %7d: %10d - %2.2f %%  (%5.2f %%)\n",i*DISTRIBUTION_STEP,(i+1)*DISTRIBUTION_STEP-1,this->part_1_distribution[i],percent,total);
   	}	
	if (this->using_variations == EDoubleLargePrimeV) {
		total = 0;
		printf("\n  Value Distribution: 2-partial values.\n");
   		printf(" From - To: amount - perc (cumul).\n");   
   		for (i=0; i<= max; i++) {
   			percent = this->part_2_distribution[i]*100 / (double) this->so_far_2_partials_found;
			total += percent;
			printf("%7d - %7d: %10d - %2.2f %%  (%5.2f %%)\n",i*DISTRIBUTION_STEP,(i+1)*DISTRIBUTION_STEP-1,this->part_2_distribution[i],percent,total);
   		}	
	}
   }
   
}

/*!
	This method tries to factor "remaining" value of aCandidate into large prime factors, using
	the factorization algorithm indicated by this->factoring_algorithm. As such, it is used
	only during the Double LPV sieving process.
	
	Return values: - ConstRC::Factorized, NOT_FACTORIZED, FACTOR_TOO_LARGE according to the result of 
				     the requested factorization.
		       - NOT_SUPPORTED if this->factoring algorithm contains name of an unsupported
		                     algorithm
		       - BAD_ARGUMENT  if this->factoring algorithm is outside the acceptable range.

*/
int  QuadraticSieve::FactorizeCandidate(qs_relation* aCandidate) {
#ifdef TIME_MESSAGE		     
    int local_start = clock();
#endif  
    int result = NOT_FACTORIZED;
    result = fact_alg->Factor(aCandidate->remaining, aCandidate->lprime1, aCandidate->lprime2);
    if ((aCandidate->lprime1 <= 0) || (aCandidate->lprime2 <= 0))
    {
		CLog::LogError(WHEREARG,"Quadratic sieve","Factorizing 2-partial candidate smooth",aCandidate->lprime1,aCandidate->lprime2,"One of the factors is negative, possible overflow?");				
    }
    assert((result != ConstRC::Factorized) || ((aCandidate->lprime1 != 1) && (aCandidate->lprime2 != 1)));

#ifdef TIME_MESSAGE		     
    int diff = clock() - local_start;
    if (diff >=0){ // in this way we avoid problems with integer overflow, causing addition of negative values 
	total_double_factors += (double) diff;
    }		     
#endif     		    
    return result;
}

/*!
	A trivial method to print contents of the main sieving matrix (logarithms) on screen.
*/
void QuadraticSieve::PrintMainMatrix(){
     int i;
     if (this->main_matrix != NULL){
	printf("\n Main matrix contents: ");
	for (i=0; i< this->main_matrix->assigned; i++){
		printf("\nRow %d: log %x.",i,this->main_matrix->accumulated_logarithm[i]);
	}
	printf("\n");
     } else printf("\n Matrix is empty.");
}

#define PRINT_A_DIVISION 0

/*!
	This method tries to divide the indicated relation by all divisors of 'a' (has only meaning
	in SIQS sieving). If the relation has been reduced to 1 (has no more divisors), the value
	of aOne is changed.

*/
void QuadraticSieve::TrialDivideByADivisors(qs_relation* aCandidate, int* aOne) {	
	unsigned int rem = 0;
	bool toInt = false;		
	if (mpz_fits_uint_p(aCandidate->remaining)) {
		rem = mpz_get_ui(aCandidate->remaining);
		toInt = true;
	}
	int x1, y1;
	int res;
	int exponent;
	register int p;
	for (x1=0; x1< this->s; x1++){
		exponent = 0;
		p = divisors_of_a[x1];
#if PRINT_A_DIVISION
		printf("TrialDivideByADivisors: dividing by %d.\n",p);
#endif		
		// Pokud je to male cislo (vejde se do Integeru)
		if (toInt) {
			while (! (res = rem % p)){
				exponent++;
				rem /= p;
			}
			total_short_divisions += exponent+1;
			total_hits += exponent;
		} else {
			while( ! mpz_fdiv_qr_ui(aux_1,aux_2,aCandidate->remaining,p) ){
#ifdef TIME_MESSAGE			
				total_divrem++;
#endif						
				exponent++;
				mpz_swap(aCandidate->remaining,aux_1);
			} 
#ifdef TIME_MESSAGE					
			total_divrem++; // this is for the last operation
#endif					
		}
#if PRINT_A_DIVISION
		printf("TrialDivideByADivisors: %d^%d found.\n", p, exponent);
#endif				
		if (exponent){
			//printf("\nNalezena zvlastni hladka hodnota; delitel je %d s indexem .\n",p,indices_of_divisors_of_a[x1]);
			y1 = 0;
#ifndef USE_PRIMES	
			while (aCandidate->elements[y1].p != indices_of_divisors_of_a[x1]){
				y1++;
			}
#else 
			while (aCandidate->elements[y1].p != divisors_of_a[x1]){
				y1++;
			}
#endif				
			//printf("\nZapsano na %d-tou pozici.\n",y1);
			aCandidate->elements[y1].exponent += exponent;
			if (toInt) {
				mpz_set_ui(aCandidate->remaining,rem);
			}
			if (!mpz_cmp_ui(aCandidate->remaining,1)) {
				if (aOne != NULL) (*aOne) = 1;
				break;
			} //else printf("A hele, partial se stejnou vlastnosti. Po vydeleni zbyva %d\n",rem);
		}
	}
}

/*!
	A trivial method to print contents of a matrix containing relations (factorizations of Q(x) 
	values) on screen. It calls PrintRelation() for each entry.

*/
void QuadraticSieve::PrintRelationMatrix(relation_matrix* aSource){
    int i;
    if (aSource){
       printf("\n Matrix contents (%ld): ",aSource->maximal_row_index);
       for (i=0; i<= aSource->maximal_row_index; i++){
       		printf("\nRow %d: ",i);
		PrintRelation(&(aSource->matrix[i]));
       } // end of row printout
       printf("\n");
    } else printf("\n Matrix is empty");
}

/*!
	A trivial method to print a relation (factorization of Q(x) value) on screen.
*/
void QuadraticSieve::PrintRelation(qs_relation* aElement){
    printf("\n");
    printf("Relation index %lu. ",aElement->index);
    if (aElement->index_of_used_polynomial == NO_INDEX){
    	printf("No poly used; ");
    } else printf("Poly #%ld; ",aElement->index_of_used_polynomial);
    if (aElement->inited & ARG_INITED){
    	gmp_printf("Arg=%Zd; ",aElement->argument);
    } else printf("Arg not inited; ");
    if (aElement->inited & LSV_INITED){
    	gmp_printf("LSV=%Zd; ",aElement->left_side_value);
    } else printf("LSV not inited; ");
    if (aElement->inited & FAC_INITED){
    	gmp_printf("Q(x)=%Zd; ",aElement->factorized_value);
    } else printf("Q(x) not inited; ");
    int i;
    if (aElement->maximal_sieving_element_index > -1){    	
	printf("Divisors: ");
	for (i=0; i<=aElement->maximal_sieving_element_index; i++){
#ifndef USE_PRIMES
		printf("(%d^%d)*",this->factor_base[aElement->elements[i].p],aElement->elements[i].exponent);
#else
		printf("(%d^%d)*",aElement->elements[i].p,aElement->elements[i].exponent);
#endif
	}
    } else printf("No divisors");
    if (aElement->inited & REM_INITED){
    	gmp_printf("%Zd; ",aElement->remaining);
    } else printf("Rem not inited; ");
    if (aElement->inited & RSV_INITED){
    	gmp_printf("RSV=%Zd; ",aElement->right_side_value);
    } else printf("RSV not inited; ");
    
    gmp_printf("LPrime1=%d; ",aElement->lprime1);
    gmp_printf("LPrime2=%d; ",aElement->lprime2);    
    printf("Allocated elements %ld, used elements %ld.", aElement->currently_assigned_sieving_elements, aElement->maximal_sieving_element_index+1);
    if (aElement->max_lp_index > -1) {
    	printf("\nThis relation is a partial cycle and has some large primes, altogether %ld: ",aElement->max_lp_index+1);
	for (i=0; i<=aElement->max_lp_index; i++) {
		std:cout << "(" << aElement->lps[i].prime << "^" << aElement->lps[i].exponent <<")*";
	}
	printf("1.");
    }
   printf("\nComponents: %d\n",aElement->components);
}

//#define TRUE_INDEX_DEBUG

int  QuadraticSieve::GetTrueIndexInFB(main_sieving_type* aP) const
{	
#ifdef TRUE_INDEX_DEBUG
	printf("Searching for true FB index of %p: ",aP);
#endif
	void* result = bsearch(aP,factor_base,maximal_factor_base_index+1,sizeof(main_sieving_type),&CompareMainSievingTypes);
	if (result != NULL)
	{
		main_sieving_type* end = (main_sieving_type*) result;
		int diff =  end-factor_base;
#ifdef TRUE_INDEX_DEBUG
		printf(" ... it is %d.\n",diff);
#endif
		return diff;
	} else {
#ifdef TRUE_INDEX_DEBUG
		printf(" ... not found.\n");
#endif
		return -1;
	}
}

/*!
This method is used to save a relation into file. It uses fprintf() for this purpose. 

The reason why the FILE pointer is supplied as a reference to a pointer is 
as follows:
The pointer is intended to be a member variable of class QuadraticSieve.
In distributed environment, another thread may close and reopen the pointer
at the same time. The mutex solves problems of concurrent access. However, the
pointer aFw could get invalidated by the close / reopen operation from the other
thread. That is why we use a reference to a member variable instead; the 
referenced member variable will be valid again after reopen operation.
*/
void QuadraticSieve::PrintRelationToFile(FILE*& aFw, qs_relation* aElement, long aIndex){
     if (aFw == NULL){
     	return;
     }	
     if (ENone != iDistributionMachineType)
     {
		if (!iDistributedMutexInited)
		{
			pthread_mutex_init(&iDistributedMutex,NULL);
			iDistributedMutexInited = TRUE;
		}
		//printf("Distributed / MUTEX\n");
// 		printf("PrintRelation to file: locking mutex.\n");
		pthread_mutex_lock(&iDistributedMutex);
// 		printf("PrintRelation to file: lock mutex successful.\n");
     }
     // At first, we sort the divisors in relation
     SortRelation(aElement);
     fprintf(aFw,"==================================================\n");
     fprintf(aFw,"#%ld\n",aIndex);
     fprintf(aFw,"poly=%ld\n",aElement->index_of_used_polynomial);
     fprintf(aFw,"inited=%d\n",aElement->inited);
     fprintf(aFw,"arg=");
     if (aElement->inited & ARG_INITED) PrintMpzToFile(aFw,10,aElement->argument); else fprintf(aFw,"NONE");
     fprintf(aFw,"\n");
     fprintf(aFw,"fac=");
     if (aElement->inited & FAC_INITED) PrintMpzToFile(aFw,10,aElement->factorized_value); else fprintf(aFw,"NONE");
     fprintf(aFw,"\n");
     fprintf(aFw,"rem=");
     if (aElement->inited & REM_INITED) PrintMpzToFile(aFw,10,aElement->remaining); else fprintf(aFw,"NONE");
     fprintf(aFw,"\n");
     fprintf(aFw,"lsv=");
     if (aElement->inited & LSV_INITED) PrintMpzToFile(aFw,10,aElement->left_side_value); else fprintf(aFw,"NONE");
     fprintf(aFw,"\n");
     fprintf(aFw,"rsv=");
     if (aElement->inited & RSV_INITED) PrintMpzToFile(aFw,10,aElement->right_side_value); else fprintf(aFw,"NONE");
     fprintf(aFw,"\n");
     if ((aElement->lprime1 < 0) || (aElement->lprime2 < 0))
     {
	CLog::LogError(WHEREARG,"Quadratic sieve","PrintRelationTofile",aElement->lprime1,aElement->lprime2,"One (or both) of the partials is negative");     	
     }
#ifdef LONG_LARGE_PRIME
     fprintf(aFw,"lp1=%lu\n",aElement->lprime1);     
     fprintf(aFw,"lp2=%lu\n",aElement->lprime2);     
#else
     fprintf(aFw,"lp1=%u\n",aElement->lprime1);     
     fprintf(aFw,"lp2=%u\n",aElement->lprime2);     
#endif
     fprintf(aFw,"msi=%ld\n",aElement->maximal_sieving_element_index);
     fprintf(aFw,"cas=%lu\n",aElement->currently_assigned_sieving_elements);
     int i;
     for (i=0;i<=aElement->maximal_sieving_element_index;i++){
     fprintf(aFw,"(%d^%d)",aElement->elements[i].p,aElement->elements[i].exponent);
     }
     fprintf(aFw,"\n");
     if (ENone != iDistributionMachineType)
     {
		//printf("Distributed / MUTEX end\n");
// 		printf("PrintRelation to file: unlocking mutex.\n");
		pthread_mutex_unlock(&iDistributedMutex);
// 		printf("PrintRelation to file: unlock mutex successful.\n");
		iRelationsSinceLastDataTransmission++;
     }
     /*if (AdHoc(aIndex)) {		
	AssertRelationConsistency(aElement,"Ad hoc kontrola",100);
	//PrintRelation(aElement);
     }*/
}

/*!
	This method tries to read a single relation from the indicated file and save it into aElement. It only
	accepts file formed in a specific way (see WriteRelationToFile), and must start reading from the beginning
	of a relation. It uses fgets() to read from stream.
	At the end of reading, the obtained relation is checked for consistency (e.g. whether the left side and
	the right side are equal mod kN).
	
	Return codes:
	- COULD_NOT_READ_FROM_NULL_STREAM if the supplied file stream is NULL.
	- BAD_ARGUMENT in case of bad position in stream (the first character on second line is not #)
	- END_OF_FILE if no more data can be read from file stream
	- NOT_ENOUGH_MEMORY if memory space for relation cannot be allocated
	- ConstRC::Ok if everything went well.	

*/
int QuadraticSieve::ReadRelationFromFile(FILE* aFr, qs_relation* aElement){
     if (aFr == NULL){
        CLog::LogError(WHEREARG,"Quadratic sieve","reading smooth elements from file",30,30,"Could not read elements from NULL stream");
     	return COULD_NOT_READ_FROM_NULL_STREAM;
     }
  //   printf("A\n");
     char  data[50000];
     char* data_2;
     char* return_code;
     int   ret;
     int   item;
     long  i;

     return_code = fgets(data,500,aFr); // leading ====='s
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,30,aFr); // 'index of element'
     if (data[0]!= '#'){
	CLog::LogError(WHEREARG,"Quadratic sieve","reading relations from file",30,30,"The data pointer does not begin with #");
	printf("   %s\n",data);
     	return BAD_ARGUMENT;
     } else data[0] = '0';
     item = atol(data);
     aElement->index = item;
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,6,aFr); // 'poly='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,30,aFr); // actual number of a polynomial
     if (!return_code) return END_OF_FILE;
     aElement->index_of_used_polynomial=atol(data);
     return_code = fgets(data,8,aFr); // 'inited='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,30,aFr); // actual number of a polynomial
     if (!return_code) return END_OF_FILE;
     ret=atol(data);
     int newly_inited = ret;
     int p, exponent;
  //   printf("B\n");
     if (aElement->inited != newly_inited){
		// == argument
		if (newly_inited & ARG_INITED){
			if (!(aElement->inited & ARG_INITED)){
				mpz_init(aElement->argument);
				aElement->inited |= ARG_INITED;
			}
		} else {
			if (aElement->inited & ARG_INITED){
				mpz_clear(aElement->argument);
				aElement->inited &= ~(ARG_INITED);
			}
		}
		// == factorized value
		if (newly_inited & FAC_INITED){
			if (!(aElement->inited & FAC_INITED)){
				mpz_init(aElement->factorized_value);
				aElement->inited |= FAC_INITED;
			}
		} else {
			if (aElement->inited & FAC_INITED){
				mpz_clear(aElement->factorized_value);
				aElement->inited &= ~(FAC_INITED);
			}
		}
		// == remaining value
		if (newly_inited & REM_INITED){
			if (!(aElement->inited & REM_INITED)){
				mpz_init(aElement->remaining);
				aElement->inited |= REM_INITED;
			}
		} else {
			if (aElement->inited & REM_INITED){
				mpz_clear(aElement->remaining);
				aElement->inited &= ~(REM_INITED);
			}
		}
		// == left side value
		if (newly_inited & LSV_INITED){
			if (!(aElement->inited & LSV_INITED)){
				mpz_init(aElement->left_side_value);
				aElement->inited |= LSV_INITED;
			}
		} else {
			if (aElement->inited & LSV_INITED){
				mpz_clear(aElement->left_side_value);
				aElement->inited &= ~(LSV_INITED);
			}
		}
		// == right side value
		if (newly_inited & RSV_INITED){
			if (!(aElement->inited & RSV_INITED)){
				mpz_init(aElement->right_side_value);
				aElement->inited |= RSV_INITED;
			}
		} else {
			if (aElement->inited & RSV_INITED){
				mpz_clear(aElement->right_side_value);
				aElement->inited &= ~(RSV_INITED);
			}
		}		
		if (newly_inited != aElement->inited){		
			CLog::LogError(WHEREARG,"Quadratic sieve","reading from file",newly_inited,aElement->inited,"there was a (nonfatal) mismatch between inited fields");
		}
     }
//     printf("C\n");
     return_code = fgets(data,5,aFr); // 'arg='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr); // the value itself
     mpz_set_str(aElement->argument,data,10); // does this really work?
     if (!return_code) return END_OF_FILE;

     return_code = fgets(data,5,aFr); // 'fac='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr);
     mpz_set_str(aElement->factorized_value,data,10); // does this really work?
     if (!return_code) return END_OF_FILE;

     return_code = fgets(data,5,aFr); // 'rem='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr);
     mpz_set_str(aElement->remaining,data,10); // does this really work?
     if (!return_code) return END_OF_FILE;

     return_code = fgets(data,5,aFr); // 'lsv='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr);
     mpz_set_str(aElement->left_side_value,data,10); // does this really work?
     if (!return_code) return END_OF_FILE;

     return_code = fgets(data,5,aFr); // 'rsv='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr);
     mpz_set_str(aElement->right_side_value,data,10); // does this really work?
     if (!return_code) return END_OF_FILE;
     
     return_code = fgets(data,5,aFr); // 'lp1='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr);
     if (!return_code) return END_OF_FILE;
#ifdef LONG_LARGE_PRIME
     sscanf(data,"%lu",&(aElement->lprime1));
#else
     sscanf(data,"%u",&(aElement->lprime1));
#endif
     //aElement->lprime1=atol(data);   
    // printf("D\n");     
     
     return_code = fgets(data,5,aFr); // 'lp2='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,500,aFr);
     if (!return_code) return END_OF_FILE;
#ifdef LONG_LARGE_PRIME
     sscanf(data,"%lu",&(aElement->lprime2));
#else
     sscanf(data,"%u",&(aElement->lprime2));
#endif
     //aElement->lprime2=atol(data);

     return_code = fgets(data,5,aFr); // 'msi='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,30,aFr); //
     if (!return_code) return END_OF_FILE;

     return_code = fgets(data,5,aFr); // 'cas='
     if (!return_code) return END_OF_FILE;
     return_code = fgets(data,30,aFr); //
     if (!return_code) return END_OF_FILE;
     int currently_assigned_sieving_elements=atol(data);
    // printf("E\n");
     if (aElement->currently_assigned_sieving_elements < currently_assigned_sieving_elements){
	aElement->elements = (sieving_element*) AllocationMachine((void **) &(aElement->elements),currently_assigned_sieving_elements*sizeof(sieving_element));
	if (!aElement->currently_assigned_sieving_elements){
 		CLog::LogError(WHEREARG,"Quadratic sieve","reallocating elements in reading from file",item,100,"There was not enough memory to reallocate space needed for read elements");
		return NOT_ENOUGH_MEMORY;
	}
	aElement->currently_assigned_sieving_elements = currently_assigned_sieving_elements;
	for (i=0; i< aElement->currently_assigned_sieving_elements; i++){
		aElement->elements[i].p = 0;
		aElement->elements[i].exponent     = 0;
	}
     }
     return_code = fgets(data,50000,aFr); // actual divisors
     if (!return_code) return END_OF_FILE;
     data_2 = data;
     i = 0;
     do {
     	return_code = strtok(data_2,"(^)");
	if (return_code) {
		p = atol(return_code);
	}
	data_2 = NULL;
	return_code = strtok(data_2,"(^)");
	if (return_code) { exponent=atol(return_code);
			   aElement->elements[i].p = p;
			   aElement->elements[i].exponent = exponent;
#ifndef USE_PRIMES
			   if ((this->QSMode == EMpqs) || (this->QSMode == ESiqs))
				{
					// Prepocitat column indexy.
					aElement->elements[i].p = GetTrueIndexInFB(&p);
				}
#endif
			   i++;
			  }
	data_2 = NULL;
     } while(return_code);
     aElement->maximal_sieving_element_index = i-1;
    // printf("F\n");
    if (aElement->lprime2 == 2147483647)
	{
     		AssertRelationConsistency(aElement,"lprime2=2147483647", item);    
	}
     if (this->assertion_mode)
	{
     		AssertRelationConsistency(aElement,"reading smooths", item);    
		 		
   	}
  //   printf("G\n");
  //   PrintRelation(aElement);
     return ConstRC::Ok;
}


/*!
	This method is used to save a relation into matrix aTarget. Unlike PrintRelationToFile, it has
	a return value indicating status of operation; reallocation is sometimes needed, and it can fail.
	
	Return codes:
	- NOT_ENOUGH_MEMORY if any reallocation failed
	- ConstRC::Ok otherwise.

*/
int QuadraticSieve::PrintRelationToMatrix(relation_matrix* aTarget, qs_relation* aElement, long aIndex){
     long i;
     // at first, we check whether there is need to reallocate the target matrix
     if (aIndex == aTarget->assigned_rows){ // situation, where we would write beyond boundaries
     		aTarget->assigned_rows += SMOOTH_REALLOCATION_STEP;
		aTarget->matrix = (qs_relation*) AllocationMachine((void**) &(aTarget->matrix), aTarget->assigned_rows*sizeof(qs_relation));
		if (!aTarget->matrix){
			CLog::LogError(WHEREARG,"Quadratic sieve","reallocating one of smooth matrices",mpz_cmp_ui(aElement->remaining,0),aIndex,"New memory could not be allocated");
			return NOT_ENOUGH_MEMORY;
		}
		for (i=aIndex; i<aTarget->assigned_rows; i++){
			aTarget->matrix[i].inited = 0;
			aTarget->matrix[i].maximal_sieving_element_index = -1;
			aTarget->matrix[i].currently_assigned_sieving_elements = 0;
			aTarget->matrix[i].elements = NULL;
			aTarget->matrix[i].lps = NULL;
		        aTarget->matrix[i].max_lp_index = -1;	
		        aTarget->matrix[i].assigned_lps = 0;	
		}
     } // end of reallocation check
     // mpz initialization check
     if (!(aTarget->matrix[aIndex].inited & ARG_INITED)){
     		mpz_init_set_ui(aTarget->matrix[aIndex].argument,1);
		aTarget->matrix[aIndex].inited |= ARG_INITED;
     }
     if (!(aTarget->matrix[aIndex].inited & REM_INITED)){
     		mpz_init_set_ui(aTarget->matrix[aIndex].remaining,1);
		aTarget->matrix[aIndex].inited |= REM_INITED;
     }
     if (!(aTarget->matrix[aIndex].inited & FAC_INITED)){
     		mpz_init_set_ui(aTarget->matrix[aIndex].factorized_value,1);
		aTarget->matrix[aIndex].inited |= FAC_INITED;
     }
     if (!(aTarget->matrix[aIndex].inited & LSV_INITED)){
     		mpz_init_set_ui(aTarget->matrix[aIndex].left_side_value,1);
		aTarget->matrix[aIndex].inited |= LSV_INITED;
     }
     if (!(aTarget->matrix[aIndex].inited & RSV_INITED)){
     		mpz_init_set_ui(aTarget->matrix[aIndex].right_side_value,1);
		aTarget->matrix[aIndex].inited |= RSV_INITED;
     }
     // We do not copy the large primes. This method is not intended to work like this.
     // the copy process itself
     if (aElement->inited & ARG_INITED) mpz_set(aTarget->matrix[aIndex].argument, aElement->argument);
     if (aElement->inited & REM_INITED) mpz_set(aTarget->matrix[aIndex].remaining,aElement->remaining);
     if (aElement->inited & FAC_INITED) mpz_set(aTarget->matrix[aIndex].factorized_value,aElement->factorized_value);
     if (aElement->inited & LSV_INITED) mpz_set(aTarget->matrix[aIndex].left_side_value,aElement->left_side_value);
     if (aElement->inited & RSV_INITED) mpz_set(aTarget->matrix[aIndex].right_side_value,aElement->right_side_value);
     aTarget->matrix[aIndex].lprime1 = aElement->lprime1;
     aTarget->matrix[aIndex].lprime2 = aElement->lprime2;
     aTarget->matrix[aIndex].index_of_used_polynomial = aElement->index_of_used_polynomial;
     aTarget->matrix[aIndex].currently_assigned_sieving_elements = 0;
     aTarget->matrix[aIndex].maximal_sieving_element_index = -1;
     if (aElement->currently_assigned_sieving_elements){
     		aTarget->matrix[aIndex].elements = (sieving_element*) AllocationMachine((void**) &(aTarget->matrix[aIndex].elements),aElement->currently_assigned_sieving_elements*sizeof(sieving_element));
		if (!aTarget->matrix[aIndex].elements){
			CLog::LogError(WHEREARG,"Quadratic sieve","allocating elements in smooth matrix",mpz_cmp_ui(aElement->remaining,0),aIndex,"New memory could not be allocated");
			return NOT_ENOUGH_MEMORY;
		}
		aTarget->matrix[aIndex].currently_assigned_sieving_elements = aElement->currently_assigned_sieving_elements;
		for (i=0; i< aTarget->matrix[aIndex].currently_assigned_sieving_elements; i++){
			if (i<= aElement->maximal_sieving_element_index){
				aTarget->matrix[aIndex].maximal_sieving_element_index++;
				aTarget->matrix[aIndex].elements[i].p = aElement->elements[i].p;
				aTarget->matrix[aIndex].elements[i].exponent = aElement->elements[i].exponent;
			}
			else {
				aTarget->matrix[aIndex].elements[i].p = 0;
				aTarget->matrix[aIndex].elements[i].exponent = 0;
			}
		}
     }
     aTarget->maximal_row_index = aIndex; // maximal meaningful entry in matrix
     return ConstRC::Ok;
}

/*!
	This method checks whether the input relation has some odd exponent of a small prime. If no, such
	relation appears as zero row in the linear algebra step, and so should be omitted.
	Zero values are found very scarcely, less than 1 in 10000, but if they do, they could cause
	serious havoc in the linear algebra step.
	
	Return values:
	- ZERO 
	- NONZERO.

*/
int QuadraticSieve::CheckNulity(qs_relation* aElement){
     int ret = ZERO;
     long i;
     for (i=0; i<=aElement->maximal_sieving_element_index; i++){
     		// Checking whether this row is nonzero modulo 2
     		if ((aElement->elements[i].exponent)%2){
			ret = NONZERO;
			break;
		}
     }
     return ret;
}

/*!
	This method serves for determination whether there has been enough smooth and partial values collected. 
	It returns ENOUGH if # of collected values exceeds the FB count + margin.
        It returns ENOUGH_BUT_WARNING if # of collected values exceeds the FB count, but not the FB count + margin.
        It returns NOT_ENOUGH otherwise.
	Besides this, it also determines the value of threshold, which is the amount of smooth and partial values
	suitable for running the linear algebra phase.

*/
int QuadraticSieve::CheckEnough(){
    this->threshold = this->maximal_factor_base_index + this->margin;
    this->collected = this->so_far_smooths_found;
    
	this->independent_cycles = this->edges + this->components - this->vertices;
	this->collected += (this->independent_cycles);
    
    if (this->collected >= this->threshold)
    {
    	return ENOUGH;
    }
   if (this->collected > (this->maximal_factor_base_index+1))
    {
	return ENOUGH_BUT_WARNING;
    }
    return NOT_ENOUGH;
}

/*!
	This method detects those x-partials, which cannot be used in construction of any
	fundamental cycle, and removes them from the set of usable partials. This is done
	in the following way: the partial relations are read one-by-one and their primes are
	written into a table. At the end of relation reading, we have a table of frequency
	of occurrence of all collected large primes. If a prime happens to occur in
	one x-partial relation only, the corresponding relation cannot take part in any
	fundamental cycle (since it is a leaf in graph of all relations). Therefore, it
	must be removed from the collection of x-partials; and this precisely is done.
	
	Of course, such removal could, and in fact often does, create new leaves in the
	graph of x-partial relations; therefore, the process must be iterated again and
	again, until the graph consists of cycles only.
	
	The process requires repetitive reading from a large file (in first iteration the
	size can well exceed 100 M) and writing to another large file (though usually much 
	smaller). Therefore, it is lengthy and dependent on actual speed of sequential disk
	access and fragmentation of the underlying file system. It can take several minutes
	to remove all large singletons.
	
	Return values: - NOT_ENOUGH_MEMORY if allocation of some table fails (this may
		       happen, since their sizes often exceed several MB). 
		       - anything thrown by RereadRelations, including ConstRC::Ok as sign that everything
		       went well.

*/
int QuadraticSieve::RemoveLargeSingletons() {
	// No partial values available? A legitimate situation in *very* short sieving.	
	if (!this->edges)
	{
		printf("Remove large singletons: no edges.\n");
		return ConstRC::Ok;
	}
	if (this->iLoadFlags & SIEVING_LOAD_FLAG)
	{
		
	}
	if (this->info_mode) {
		printf("\n");
		PrintGraphInformation();
	}
	// at first, we remove the large array cycle_table, to ensure memory space
	int i, max;
	if (this->cycle_table_1) {
		delete cycle_table_1;
		this->cycle_table_1 = NULL;
	}
	// Then, we allocate a temporary array of bytes, with a single entry per each found x-partial relation.	
	detection_type* detection_field = NULL;
	int max_field = this->edges-1;
	detection_field = (detection_type *) AllocationMachine((void**) &detection_field,(max_field+1)*sizeof(detection_type));
	if (detection_field == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","could not allocate detection field",this->edges,(max_field+1)*sizeof(detection_type),"Not enough memory to allocate data space.");
		return NOT_ENOUGH_MEMORY;
	}
	// At first, none of the x-partial relations is marked as "bearing a singleton"
	memset(detection_field,ConstHT::NONSINGLETON,(max_field+1)*sizeof(detection_type));
	int current_vertices = this->vertices;
	int current_edges    = this->edges;
	int currently_removed_vertices = 0;
	int currently_removed_edges    = 0;
	int iter = 0;
	int total = 0;
	int total_edges = 0;
	int size = 0;
	bool present;
	long d_1;
	long d_2;
	int edge;
	if (this->info_mode) {
		printf("   Deleting singletons from x-partials ... ");
		fflush(stdout);
	}
	if (this->debug_mode)
	{
		printf("\n      Allocated detection field of size %d\n",max_field+1);
	}
	// At first, we copy all partial relations to file "all partials".
	CloseFile(this->partial_values_file_name);	
	qs_relation* candidate = InitRelation();
	this->partials_file 	= OpenFile(this->partial_values_file_name,READ);
	this->all_partials_file = OpenFile(this->all_partial_values_file_name,WRITE);
	int countRead = 0;
	while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
		countRead++;
		if ((info_mode || debug_mode) && (!(countRead % 10000)))
		{
			printf("\r\t\tCopying partial values to all_partials file, so far %d .	...",countRead);			
			fflush(stdout);
		}
		PrintRelationToFile(this->all_partials_file,candidate,candidate->index);
		/*fflush(this->all_partials_file);
		printf("Printed relation to file ...");
		wait();*/
	}
	CloseFile(this->partial_values_file_name);
	CloseFile(this->all_partial_values_file_name);
	while (1) {
		currently_removed_vertices = 0;
		currently_removed_edges    = 0;
		// At first, let us allocate hashtable with space suitable for the remaining large primes
		this->cycle_table_1 = new lp_hashtable_common<CCommonHashEntry>();
		this->cycle_table_1->SetupHashtable(current_vertices);
		if (this->cycle_table_1 == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","could not allocate hashtable",this->edges,size*sizeof(detection_type),"Not enough memory to allocate hashtable space.");
			return NOT_ENOUGH_MEMORY;
		}
		ResetHashtable();
		if (this->debug_mode)
		{
			printf("\n\tLarge singleton removal, iteration %02d: allocated hashtable of size %d\n",iter,size);
		}
		// Now, we read the relations from file "partials" one-by-one and put corresponding
		// values into the hashtable  
		this->partials_file 	= OpenFile(this->partial_values_file_name,READ);
		int countRead = 0;
		while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
			countRead++;		
			if ((info_mode || debug_mode) && (!(countRead % 10000)))
			{
				printf("\r\t\tReading partial values to hashtable, so far %d .	...",countRead);	
				fflush(stdout);		
			}
			/*if (iter)
			{
				PrintRelation(candidate);
				wait();
			}*/
			bool enlarged = false;
			// Zkontrolujeme pritomnost prvocisel lprime1 a lprime2 v hashovaci tabulce
			this->cycle_table_1->FindPrimeInHashtable(candidate->lprime1,0,present,d_1, enlarged);
			//printf("%d ",d_1); fflush(stdout);
			if (!present) {
			// This prime has not been in hashtableyet.					
				cycle_table_1->Hashtable[d_1].SetValue(candidate->lprime1);
				cycle_table_1->Hashtable[d_1].SetAncestor(candidate->index);
				cycle_table_1->Added();
			} else {
			// We mark this prime as nonsingleton by bit-masking the "ancestor" field.
				//printf("We have found second occurrence of prime %d\n",candidate->lprime1);
				cycle_table_1->Hashtable[d_1].SetAncestor(cycle_table_1->Hashtable[d_1].GetAncestor() | ConstHT::NONSINGLETON_MASK_NFS);
				//printf("Current value of ancestor: 0x%x\n",cycle_table_1[d_2].ancestor);
			}
			this->cycle_table_1->FindPrimeInHashtable(candidate->lprime2,0,present,d_2, enlarged);
			if (enlarged) this->cycle_table_1->FindPrimeInHashtable(candidate->lprime1,0,present,d_1, enlarged);
			//printf("%d ",d_2); fflush(stdout);
			if (!present) {
			// This prime has not been in hashtable yet.					
				cycle_table_1->Hashtable[d_2].SetValue(candidate->lprime2);
				cycle_table_1->Hashtable[d_2].SetAncestor(candidate->index);
				cycle_table_1->Added();
			} else {
			// We mark this prime as nonsingleton by bit-masking the "ancestor" field.
				//printf("We have found second occurrence of prime %d\n",candidate->lprime2);
				cycle_table_1->Hashtable[d_2].SetAncestor(cycle_table_1->Hashtable[d_1].GetAncestor() | ConstHT::NONSINGLETON_MASK_NFS);
				//printf("Current value of ancestor: 0x%x\n",cycle_table_1[d_2].ancestor);
			}
		}
		CloseFile(this->partial_values_file_name);
		if (this->debug_mode)
		{
			printf("\n\t\tIter %02d: loaded %d partial relations from file.\n",iter,countRead);
			fflush(stdout);
		}
		// Now, we count the number of removed primes and relations
		max = cycle_table_1->AllocatedSize;
		int freshSingletons = 0;
		for (i=0; i< max; i++) {
			// There is a prime on this position ...			
			if (cycle_table_1->Hashtable[i].GetValue() != ConstHT::FULL_MASK_NFS) {
				// This is a singleton prime
                                if (!(cycle_table_1->Hashtable[i].GetAncestor() & ConstHT::NONSINGLETON_MASK_NFS)) {
					edge = cycle_table_1->Hashtable[i].GetAncestor() & ~ConstHT::NONSINGLETON_MASK_NFS;
					currently_removed_vertices++;
					if (edge > max_field)
					{
						CLog::LogError(WHEREARG,"Quadratic sieve","edge index too big",edge,max_field,"Edge overran size of the detection field.");
						return GENERAL_ERROR;
					}
					if (detection_field[edge] == ConstHT::NONSINGLETON) {
						detection_field[edge] = ConstHT::FRESH_SINGLETON;
						freshSingletons++;
					}
				} else {
					//printf("Index %d contains nonsingleton.\n",i);
				}
			}
		}
		if (this->debug_mode)
		{
			printf("\t\tIter %02d: counting fresh singletons finished: %d.\n",iter, freshSingletons);
		}

		for (i=0; i<= max_field; i++) {
			if (detection_field[i] == ConstHT::FRESH_SINGLETON) {
				currently_removed_edges++;
				detection_field[i] = ConstHT::SINGLETON;
			}
		}
		// Now, we open the file "relations" for reading and copy only those relations
		// which are not marked as singletons
		this->partials_file = OpenFile(this->partial_values_file_name, READ);
		this->temporary_file = OpenFile(this->temporary_file_name, WRITE);
		int countWritten = 0;
		while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
			if (detection_field[candidate->index] == ConstHT::NONSINGLETON) {
				PrintRelationToFile(this->temporary_file,candidate,candidate->index);
				countWritten++;
				if ((info_mode || debug_mode) && (!(countWritten % 10000)))
				{
					printf("\r\t\tChoosing usable partial values from file, so far %d .	...",countWritten);	
					fflush(stdout);		
				}
			}
		}
		CloseFile(this->partial_values_file_name);
		CloseFile(this->temporary_file_name);
		remove(this->partial_values_file_name.c_str());
		rename(this->temporary_file_name.c_str(),this->partial_values_file_name.c_str());
		// We clear the hashtable
		delete this->cycle_table_1;
		this->cycle_table_1 = NULL;
		if (this->debug_mode)
		{
			printf("\n\t\tIter %02d: removed %d relations (%d primes) from partial relations.\n",iter,currently_removed_edges,currently_removed_vertices);
			printf("\t\tIter %02d: now %d partial relations remaining\n",iter,countWritten);
		} else if (info_mode) {
			printf(" %d V, %d E |",currently_removed_vertices,currently_removed_edges);			
		}
		total += currently_removed_vertices;
		total_edges += currently_removed_edges;
		current_vertices -= currently_removed_vertices;
		current_edges    -= currently_removed_edges;
		iter++;
		if (!(currently_removed_vertices && currently_removed_edges)) break; 
	}
	printf(" (%d V and %d E removed in %d iterations)\n",total,total_edges,iter);
	fflush(stdout);
	// Now, we reconstruct the cycles again, in a way similar to the sieving process.
	// At first: we currently know the number of all vertices, so we create a hashtable
	// of suitable size.
	cycle_table_1 = new lp_hashtable_common<CCommonHashEntry>();
	cycle_table_1->SetupHashtable(current_vertices);
	if (this->cycle_table_1 == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","could not allocate hashtable",this->edges,size*sizeof(detection_type),"Not enough memory to allocate hashtable space.");
		return NOT_ENOUGH_MEMORY;
	}
	ResetHashtable();	
	int result =	RereadRelations(detection_field,max_field);	
	// We find all roots of all components. This will be needed in ConstructCycles
	CountRoots();
	this->cycles = this->edges + this->components - this->vertices;
	if (this->info_mode) {
		PrintGraphInformation();
	}
	// Now, we find out all roots of all components and put them into the list "root_list"
	// Remove the temporary array of bytes
	ClearRelation(candidate);
	free(candidate);
	candidate = NULL;
	free(detection_field);
	detection_field = NULL;
	iRelationProcessingFlags |= LARGE_SINGLETONS_REMOVED_FLAG;	
	return result;
}

/*!
	This method reads relations from partials_file and puts appropriate value into the
 	hashtable 1. This is performed just before construction of cycles, when we look 
	for roots of the graph components.	
	
	Return values: - ConstRC::Ok if everything went well.
	               - COULD_NOT_OPEN_FILE if opening of partials_file failed.

*/
int  QuadraticSieve::RereadRelations(detection_type* aField, int aMax) {
	int result = ConstRC::Ok;
	this->partials_file= OpenFile(this->partial_values_file_name,READ);
	if (this->partials_file == NULL) {
		return COULD_NOT_OPEN_FILE;
	}
	// Declaration of a new "relation" variable.
	qs_relation* candidate = InitRelation();	
	if  (this->partials_file != NULL) {
		while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
			AddToCycleCountStructures(candidate);
		}
		result = CloseFile(this->partial_values_file_name);
		if (result != ConstRC::Ok) {
			CloseFile(this->partial_values_file_name);			
		} 
	} else {
		result = COULD_NOT_OPEN_FILE;
	}
	ClearRelation(candidate);
	free(candidate); candidate = NULL;
	return result;
}

/*!
	If the intermediate smooth and partial values have been collected into a file, this method reads the smooth values and
	saves them into a special matrix in memory for instant access. This approach is suitable for smaller factorizations,
	since it is quite memory-expensive. Must be rewritten for larger factorizations, where number of possible smooths
	reaches millions.
	
	Return codes:
	- anything thrown by ReadRelationFromFile() and PrintRelationToMatrix()
	- ConstRC::Ok if everything went well.

*/
int QuadraticSieve::AddSmoothValuesToRelationMatrix(){	
	    if (this->info_mode){
	        NormalMessage("   Reading smooth values from file ...",NO_NUMBER);
	    }
	    qs_relation* candidate = InitRelation();
	    int ret;
	    int ret_2;
	    int i = 0;
	    int countRead = 0;
	    // Read values from file smooth_values.
   	    while ((ret = ReadRelationFromFile(smooth_values_file,candidate)) == ConstRC::Ok){
		countRead++;
		if ((info_mode || debug_mode) && (!(countRead % 10000)))
		{
			printf("\r\t\tReading smooth values from file, so far %d .	...",countRead);			
			fflush(stdout);
		}
                if ((ret_2=PrintRelationToMatrix(smooth_values,candidate,i++)) != ConstRC::Ok){
		    CLog::LogError(WHEREARG,"Quadratic sieve","AddSmoothValuesToRelationMatrix",countRead,ret_2,"Print Relation to Matrix failed");		
		    ClearRelation(candidate);
		    free(candidate); candidate = NULL;
		    return ret_2;
		 }
	    } // while
	    ClearRelation(candidate);
	    free(candidate); candidate = NULL;
	    if (ret != END_OF_FILE){
		CLog::LogError(WHEREARG,"Quadratic sieve","AddSmoothValuesToRelationMatrix",countRead,ret,"ret was not end of file");
		return ret;
	    }
	    if (this->info_mode){
	    	printf(" (%lu read)",this->smooth_values->maximal_row_index+1);
		fflush(stdout);
	    }
	return ConstRC::Ok;
}

/*!
	This method calculates hashes of all values in the smooth matrix and if two of them match, it removes one of the
	"duplicate values". This is needed because duplicate rows in the sieving matrix reduce the number of possible
	nonzero solutions.
	Return codes:
	- NOT_ENOUGH_MEMORY if allocation of auxiliary array fails,
	- ConstRC::Ok otherwise.

*/
int QuadraticSieve::CheckAndDeleteDuplicates(){
    if (this->info_mode){
    	NormalMessage("   Checking and deleting duplicates ...",NO_NUMBER);
    }
    long i;
    long* list_of_rows_to_delete;    
    long max_index_in_list = -1;
    list_of_rows_to_delete = (long*) malloc(sizeof(long)*(this->smooth_values->maximal_row_index+1));
    if (!list_of_rows_to_delete){
        CLog::LogError(WHEREARG,"Quadratic sieve","Checking and deleting duplicates",this->smooth_values->maximal_row_index,10,"Could not allocate list of rows to delete");
    	return NOT_ENOUGH_MEMORY;
    }
    // at first, we find out which rows to delete and add their numbers in deletion list
    for (i=0; i<= this->smooth_values->maximal_row_index; i++){
	CalculateHash(aux_1,&(this->smooth_values->matrix[i]));
	if (AddHashToTree(aux_1,i) == FOUND){
		max_index_in_list++;
		list_of_rows_to_delete[max_index_in_list] = i;
		/*printf("\n%d,%d",first_row,i);
		PrintRelation(&(this->smooth_values->matrix[first_row]));
		PrintRelation(&(this->smooth_values->matrix[i]));*/
	}
    }
    // now, we delete the rows 
    DeleteRowsFromRelationMatrix(this->smooth_values,list_of_rows_to_delete,max_index_in_list);    
    // we delete the occupied memory
    free(list_of_rows_to_delete);
    list_of_rows_to_delete = NULL;
    if (this->info_mode){
    	printf(" (%ld deleted)",max_index_in_list+1);
	fflush(stdout);
    }
    return ConstRC::Ok;
}

/*!
	This method is given a list of indices (by the second and third argument) and a matrix, and tries to 
	remove them from the matrix.
	The removal is done in the following way: we find all the rows mentioned in the list of indices, and
	move them to the bottom of the matrix by swapping them with some "innocent" rows from the bottom; this
	is done using methods IsInList() and SwapTwoRows(). Some precautions must be taken not to swap two
	rows which are both intended for removal. After this process, the upper bound of the matrix is determined
	so that it does not include any of the "removed rows".
	Note that this is "logical", not "physical" deletion. The memory taken by "removed rows" is not freed,
	and it is available to write some new rows. The real memory freeing is performed by the destructor.

*/
void QuadraticSieve::DeleteRowsFromRelationMatrix(relation_matrix *aTarget, long* aListOfRowsToDelete, long aMaxIndexInList){
    // now, we delete the rows as follows: we move them to the end of the smooth matrix and then just decrease
    // its max_row_index attribute
    long i;
    long top_index;
    long bottom_index = aTarget->maximal_row_index;
    /*for (i=0; i<=aMaxIndexInList; i++){
    	printf("\nBudeme mazat radu cislo %d.",aListOfRowsToDelete[i]);
    }*/
    if (aMaxIndexInList > -1){
    	while (IsInList(aListOfRowsToDelete,aMaxIndexInList,bottom_index) != NOT_FOUND){
   		bottom_index--;
	    }
    	for (i=0; i<=aMaxIndexInList;i++){
		top_index = aListOfRowsToDelete[i];
		if (top_index >= bottom_index)
			break;
		SwapTwoRows(aTarget, top_index, bottom_index);
		//printf("\nProhodili jsme rady cislo %d a %d.",top_index,bottom_index);		
		bottom_index--;
		while (IsInList(aListOfRowsToDelete,aMaxIndexInList,bottom_index) != NOT_FOUND){
   			bottom_index--;
   		}
		if (top_index >= bottom_index)
		        break;
    	} // end of for
    	// we truncate the matrix at the end, thus deleting the extra rows
    	aTarget->maximal_row_index -= (aMaxIndexInList+1);
    }
}


/*!
	This method removes all singletons from the set of collected smooth values. It repeatedly (since the process is
	iterative) walks through the list of the smooth values and notes occurrences of primes. If a prime is detected
	only once or never in the whole list, it is unsuitable and the corresponding row will be removed. The process
	ends when there are no unsuitable primes in the FB; then, those primes which have been marked as unsuitable, are
	removed from the FB.
	
	Return codes:
	- NOT_ENOUGH_MEMORY if any allocation of auxiliary arrays fails
	- ConstRC::Ok otherwise.

*/
int  QuadraticSieve::CheckAndDeleteSingletons(){
    if (this->info_mode){
    	NormalMessage("   Checking and deleting singletons ...",NO_NUMBER);
    }
    long max_column = this->maximal_factor_base_index;
    long i = 0, j = 0;
    // This is a field, originally initialized by -1's. Each time we meet an occurrence of a prime factor in a smooth value,
    // we store index of the particular row here - but only in case of first and second occurrence. At the end of this checking
    // loop, we remove all rows with singletons and mark unusable primes (those generating singletons or not occurring at all)
    // for next iteration.
    long** checking_field = NULL;
    checking_field = (long**) AllocationMachine((void**) &checking_field,3*sizeof(long*));
    if (!checking_field) {
	CLog::LogError(WHEREARG,"Quadratic sieve","checking and deleting singletons",2,-1,"There was not enough memory to allocate row field");
	return NOT_ENOUGH_MEMORY;
    }
    for (i=0; i<3; i++){
    	checking_field[i] = NULL;
	checking_field[i] = (long*) AllocationMachine((void**) &checking_field[i],(max_column+1)*sizeof(long));
	if (!checking_field[i]){
		CLog::LogError(WHEREARG,"Quadratic sieve","checking and deleting singletons",2,i,"There was not enough memory to allocate this row.");
		return NOT_ENOUGH_MEMORY;
	}
    } 
    // Fields for storage of unusable primes and unusable row indices. The first field grows through iterations, the second
    // is always reinitialized at each iteration.
    long* unusable_primes = NULL;
    long allocated_unusable_primes = 0;
    signed long current_index_in_unusable_primes = -1;
    unusable_primes = (long*) AllocationMachine((void**) &unusable_primes,USELESS_PRIME_REALLOCATION_STEP*sizeof(long));
    if (!unusable_primes) {
	CLog::LogError(WHEREARG,"Quadratic sieve","checking and deleting singletons",3,100,"There was not enough memory to allocate useless prime field");
	return NOT_ENOUGH_MEMORY;    
    }
    allocated_unusable_primes = USELESS_PRIME_REALLOCATION_STEP;
    long* unusable_rows = NULL;
    long allocated_unusable_rows;
    signed long current_index_in_unusable_rows;    
    unusable_rows = (long*) AllocationMachine((void**) &unusable_rows,USELESS_ROW_REALLOCATION_STEP*sizeof(long));
    if (!unusable_rows) {
	CLog::LogError(WHEREARG,"Quadratic sieve","checking and deleting singletons",4,100,"There was not enough memory to allocate useless row field");
	return NOT_ENOUGH_MEMORY;    
    }
    allocated_unusable_rows = USELESS_ROW_REALLOCATION_STEP;
    // end of allocations.
    // initialization. This one is done just once.
    RefillByConstant(unusable_primes,allocated_unusable_primes-1,-1);
    // Now we iterate the following search. We walk through the whole matrix and find singletons. We put indices of corresponding rows
    // into unusable_rows and indices of corresponding primes into unusable_primes. At end of each iteration, we delete unusable rows
    // off the smooth matrix. This may lead to occurrence of other singletons, and so we must iterate this process as long as any
    // singletons occur.
    qs_relation* current = NULL;
    int index = 0;
    long aux_zerooccur  = 0;
    long aux_singletons = 0;
    long aux_iterations = 0;
    long aux_lone_squares = 0;
    do {
    aux_lone_squares = 0;
    RefillByConstant(unusable_rows,allocated_unusable_rows-1,-1);
	RefillByConstant(checking_field[0],max_column,-1);
	RefillByConstant(checking_field[1],max_column,-1);
	RefillByConstant(checking_field[2],max_column,-1);
	current_index_in_unusable_rows = -1;
	// matrix checking
	for (i=0; i<= smooth_values->maximal_row_index; i++){
		current = &smooth_values->matrix[i];
		for (j=0; j<= current->maximal_sieving_element_index; j++){
			if (current->elements[j].exponent) {
#ifndef USE_PRIMES
				index = current->elements[j].p;
#else
				index = GetTrueIndexInFB(&(current->elements[j].p));
#endif
				if (current->elements[j].exponent % 2){
					if (checking_field[0][index] == -1){
						checking_field[0][index] = i;
					}
					else if (checking_field[1][index] == -1){
						checking_field[1][index] = i;
					}
				} else {
					if (checking_field[2][index] == -1) {
						checking_field[2][index] = i;
					}
				}
			}
		}
	}
	for (i=0; i<= max_column; i++){
		if ((checking_field[0][i] == -1) || (checking_field[1][i] == -1)) { 
			if (checking_field[2][i] == -1) {
			// we insert this prime into general list of unusable primes. we do not check yet for duplicities.
			current_index_in_unusable_primes++;
			if (current_index_in_unusable_primes == allocated_unusable_primes){
    				unusable_primes = (long*) AllocationMachine((void**) &unusable_primes,(allocated_unusable_primes+USELESS_PRIME_REALLOCATION_STEP)*sizeof(long));
    				if (!unusable_primes) {
					CLog::LogError(WHEREARG,"Quadratic sieve","checking and deleting singletons",allocated_unusable_primes+USELESS_PRIME_REALLOCATION_STEP,100,"There was not enough memory to reallocate useless prime field");
					return NOT_ENOUGH_MEMORY;    
    				}
    				allocated_unusable_primes += USELESS_PRIME_REALLOCATION_STEP;				
			}
			unusable_primes[current_index_in_unusable_primes] = i;			
			//printf("\nUkladame do seznamu nepouzitelnych prvocisel cislo s indexem %d",i);
			} else {
				aux_lone_squares++;
			}
		}
		if ((aux_iterations == 0) && (checking_field[0][i] == -1) && (checking_field[2][i] == -1)) { // counter of zero-incidence primes (in 0th iteration only)
			aux_zerooccur++;
		}
		if ((checking_field[0][i] != -1) && (checking_field[1][i] == -1)) {
			current_index_in_unusable_rows++;			
			if (current_index_in_unusable_rows == allocated_unusable_rows){
    				unusable_rows = (long*) AllocationMachine((void**) &unusable_rows,(allocated_unusable_rows+USELESS_ROW_REALLOCATION_STEP)*sizeof(long));
    				if (!unusable_rows) {
					CLog::LogError(WHEREARG,"Quadratic sieve","checking and deleting singletons",allocated_unusable_rows+USELESS_PRIME_REALLOCATION_STEP,100,"There was not enough memory to reallocate useless prime field");
					return NOT_ENOUGH_MEMORY;    
    				}
    				allocated_unusable_rows += USELESS_ROW_REALLOCATION_STEP;				
			}
			unusable_rows[current_index_in_unusable_rows] = checking_field[0][i];			
			//printf("\nUkladame do seznamu nepouzitelnych rad radu s indexem %d",checking_field[0][i]);
		}
	}	
	// end of smooth matrix check. Now processing the results
	qsort(unusable_rows,(current_index_in_unusable_rows+1),sizeof(long), &CompareLong);
	current_index_in_unusable_rows = DeleteDuplicatesFromSortedArray(unusable_rows,current_index_in_unusable_rows);
	// a paranoid check: we really do not wish to have some row in list twice
	if (this->assertion_mode){
		for (i=1; i<= current_index_in_unusable_rows; i++){						
			if (unusable_rows[i] == unusable_rows[i-1]){
				printf("Assertion failed in redundant row list check: rows[%ld] = rows[%ld] = %ld.\n",i,i-1,unusable_rows[i]);
			}
			if (unusable_rows[i] < unusable_rows[i-1]){
				printf("Assertion failed in redundant row list check: rows[%ld] = %ld < rows[%ld] = %ld.\n",i,unusable_rows[i-1],i-1,unusable_rows[i-1]);
			}				
		}
	}
	DeleteRowsFromRelationMatrix(this->smooth_values,unusable_rows,current_index_in_unusable_rows);
	printf(" %ld |",current_index_in_unusable_rows+1);
	aux_singletons += current_index_in_unusable_rows+1;
	aux_iterations++;
    } while ((current_index_in_unusable_rows > -1));    
    /*printf("Redundant primes: ");
    for (i=0; i<= current_index_in_unusable_primes; i++){
    	printf("%d ",unusable_primes[i]);
    }
    printf("\n");
    printf("Redundant rows A: ");
    for (i=0; i<= current_index_in_unusable_rows; i++){
    	printf("%d ",unusable_rows[i]);
    }    
    printf("\n");
    
    printf("Redundant rows B: ");
    for (i=0; i<= current_index_in_unusable_rows; i++){
    	printf("%d ",unusable_rows[i]);
    }    
    printf("\n");*/
    printf(" (%ld deleted in %ld iterations, %ld with zero occurrence, %ld lone squares)",aux_singletons,aux_iterations,aux_zerooccur, aux_lone_squares);
    qsort(unusable_primes,(current_index_in_unusable_primes+1),sizeof(long), &CompareLong);
    current_index_in_unusable_primes = DeleteDuplicatesFromSortedArray(unusable_primes,current_index_in_unusable_primes);
    /*printf("Redundant primes: ");
    for (i=0; i<= current_index_in_unusable_primes; i++){
    	printf("%d ",unusable_primes[i]);
    }
    printf("\n"); */
    int ret = ConstRC::Ok;
    ret = RemoveUnusablePrimesFromFB(unusable_primes, current_index_in_unusable_primes);        
    free((void*) unusable_rows);
    free((void*) unusable_primes);
    free((void*) checking_field[0]);
    free((void*) checking_field[1]);
    free((void*) checking_field);
    return ret;
}

/*!
	
	This is a complicated method taking part in singleton removal. It removes all singleton primes from the
	factor base, shrinking the FB; it also recalculates the prime indices in the FB, and then rewrites the relations
	gained using the new index set.
	
	Return values: 
	- ConstRC::Ok if everything went well
	- NOT_ENOUGH_MEMORY if new factor base could not be established.

*/
int  QuadraticSieve::RemoveUnusablePrimesFromFB(long* aListOfPrimesToDelete, long aMaxIndexInList){
    if (this->info_mode){
    	NormalMessage("   Removing unusable primes from factor base ...",NO_NUMBER);
	/*printf("\n");
	for (int a = 0; a<= aMaxIndexInList; a++){
		printf("%7d",aListOfPrimesToDelete[a]);
	}
	printf("\n");*/
    }    
    int max_column = this->maximal_factor_base_index;
    main_sieving_type* new_factor_base = NULL;
    /*
    	At first we allocate a new factorization base. Then we copy all primes except the unusable ones into this new FB.
	Finally, we must recalculate all smooth values in smooth matrix, because indices of primes in FB have changed.
	At the end of the method the old factor base is deleted and new put into its place.
    */
    new_factor_base = (main_sieving_type*) AllocationMachine((void**) &new_factor_base,(this->factor_base_allocated)*sizeof(main_sieving_type));
    if (!new_factor_base) {
	CLog::LogError(WHEREARG,"Quadratic sieve","removing unusable primes",2,-1,"There was not enough memory to allocate new factor base");
	return NOT_ENOUGH_MEMORY;
    }
    main_sieving_type* recalculation_indices = NULL;
    recalculation_indices = (main_sieving_type*) AllocationMachine((void**) &recalculation_indices,(this->factor_base_allocated)*sizeof(main_sieving_type));
    if (!recalculation_indices) {
	CLog::LogError(WHEREARG,"Quadratic sieve","removing unusable primes",2,-1,"There was not enough memory to allocate recalculation indices");
	return NOT_ENOUGH_MEMORY;
    }
    RefillByConstantMST(recalculation_indices,this->factor_base_allocated-1,-1);
    main_sieving_type i = 0, j = 0, k = 0, l = 0;
    /*
           Index i runs in old factor base, index j runs in redundant primes. Index k is the position written to new factor base.
	   A general rule is that i <= j. At the beginning this is sure.
	   Index l checks the difference between factor base and enriched factor base in case of EAddingToFB
    */
	while (i <= max_column) {
		if (i == aListOfPrimesToDelete[j]) { /* In case that we meet a 'forbidden' value, we do not add it into new factor base */
			i++;
			if (j < aMaxIndexInList) {
				j++;
			}
			continue; /* skip of adding into new FB*/
		}
		new_factor_base[k] = factor_base[i];
		recalculation_indices[i] = k;
		i++;
		k++;
		if (i <= maximal_factor_base_index)
			l++;
	}
    main_sieving_type mfbi = --k; 
    qs_relation* current = NULL;
    int index = 0;    
#ifndef USE_PRIMES
    for (i=0; i<= smooth_values->maximal_row_index; i++){
	current = &smooth_values->matrix[i];
	for (j=0; j<= current->maximal_sieving_element_index; j++){		
		index = current->elements[j].p;
		if ((current->elements[j].p = recalculation_indices[index]) == -1){
			// This may happen. Even if a prime number was marked as a singleton, it COULD be present in some other
			// row in EVEN power. If this happens, we must remove it from elements array (by setting exponent to zero)
			// and put the corresponding root into RSV.
			if (current->elements[j].exponent % 2){
			    printf("\nStrange situation. Singleton should not come to recalculation! Row number %d, index=%d, exponent = %d",i,index,current->elements[j].exponent);	
			}
			else {
			    //printf("In row %d we have %d^%d. Setting exponent to zero and moving value to RSV ... ",i,factor_base[index],current->elements[j].exponent);			    
			    for (int m=0; m < (current->elements[j].exponent)/2; m++){
			    	mpz_mul_si(current->right_side_value,current->right_side_value,factor_base[index]);
			    }
			    current->elements[j].exponent = 0;			    
			}			
		}
	}
    }    
#endif
    free(factor_base);
    factor_base = new_factor_base;
    maximal_factor_base_index = mfbi;    
    k = mfbi;
    printf(" now it has %d elements.",k+1);
    free(recalculation_indices);  
    return ConstRC::Ok;
}

/*!
	These two methods refill the given array from 0 to aMaxIndex by aConstant, not using the memset(), but rather
	walking through the whole array and replacing the entries by aConstant manually. This was done because
	memset(), though it accepts int as an entry, can only take the LSbyte from the int and fill the array\
	as array of bytes.

*/
void QuadraticSieve::RefillByConstant(long* aArray, long aMaxIndex, long aConstant){
    long i;
    for (i=0; i<= aMaxIndex; i++){
    	aArray[i] = aConstant;
    }
}

void QuadraticSieve::RefillByConstantMST(main_sieving_type* aArray, main_sieving_type aMaxIndex, main_sieving_type aConstant){
    int i;
    for (i=0; i<= aMaxIndex; i++){
    	aArray[i] = aConstant;
    }
}

/*!
	This method walks through the supplied array, assuming that it is sorted (up), and removes all duplicates
	from it. It returns the value of new maximal index.

*/
int QuadraticSieve::DeleteDuplicatesFromSortedArray(long* aArray, long aMaxIndex){
    long i = 0, j = 0;
    while (j <= aMaxIndex){
    	aArray[i] = aArray[j];
	i++;
	j++;
	while ((j <= aMaxIndex) && (aArray[j-1] == aArray[j])) j++;
    } 
    return (--i);
}

/*!
	This method is used in removal of duplicite rows from the set of collected relations. The "hash" is a primitive
	algorithm: the product of all divisors having odd exponent is calculated and saved into aResult.

*/
void QuadraticSieve::CalculateHash(mpz_t aResult, qs_relation* aRow){
   mpz_set_ui(aResult,1);
   int i;
   for (i=0; i<=aRow->maximal_sieving_element_index; i++){
   	if ((aRow->elements[i].exponent)%2){
#ifndef USE_PRIMES
		mpz_mul_ui(aResult,aResult,this->factor_base[aRow->elements[i].p]);
#else
		mpz_mul_si(aResult,aResult,aRow->elements[i].p);	
#endif
	}
   }
}

// the 'aType' variable makes difference between combination of smooth values (which is easier) and
// between combination of partial values (where we define a partial value as having lprime1 or lprime2 != 1 
/*
 We want to be able to input aTarget identical to aRow1 or aRow2. That is why we postpone modifications of
 aTarget inner structures to the end of the method, or, at least, to place where they are no more needed
 (like in code combining small divisors)
*/
/*!
	This method combines two relations into a single relation. It is used in Large Prime Variations.
	
	Return codes:
	- NOT_ENOUGH_MEMORY if allocation of space for divisors of combined relations fails.
	- ConstRC::Ok otherwise.

*/
int QuadraticSieve::CombineTwoRelations(qs_relation* aTarget, qs_relation* aRow1, qs_relation* aRow2, int aType){
   long i;
   // printf("================================================ \nCombineTwoRelations:\n\n");
   // PrintRelation(aRow1);
   // printf("\n\nand\n\n");
   // PrintRelation(aRow2);	
   // printf("\n\n");
   // This combines two given rows into aTarget, including combination of their left- and right side values.
   // The LSV's and RSV's of inputs MUST be allocated. This is guaranteed by this version of program.
   if (!(aTarget->inited & LSV_INITED)){
   	mpz_init(aTarget->left_side_value);
	aTarget->inited |= LSV_INITED;
   }
   if (!(aTarget->inited & RSV_INITED)){
   	mpz_init(aTarget->right_side_value);
	aTarget->inited |= RSV_INITED;
   }
   if (!(aTarget->inited & REM_INITED)){
   	mpz_init(aTarget->remaining);
	aTarget->inited |= REM_INITED;
   }
   //printf("CombineTwoRelations 2.\n");
   // this makes no sense in combined row
   aTarget->index_of_used_polynomial = NO_INDEX; 
   mpz_mul(aux_1, aRow1->left_side_value, aRow2->left_side_value);
   mpz_mod(aux_1,aux_1,this->kN);
   mpz_mul(aux_2, aRow1->right_side_value,aRow2->right_side_value);
   mpz_mod(aux_2,aux_2,this->kN);
   
   // Now we combine the divisors. For this purpose, we use an intermediate array.
   // We expect that the divisors are sorted in ascending order according to their index. This is guaranteed
   // by this version of this program.
   if (!this->intermediate_array) {
   	this->intermediate_array = (intermediate_array_element*) AllocationMachine((void**) &this->intermediate_array,DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE*sizeof(intermediate_array_element));
	if (this->intermediate_array == NULL) {
 		CLog::LogError(WHEREARG,"Quadratic sieve","combining two rows",0,DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE*sizeof(intermediate_array_element),"Could not allocate memory for intermediate array");
		ClearRelation(aTarget);
		return NOT_ENOUGH_MEMORY;
	}
	this->ia_allocated = DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE;
   }	
   int ret;
   int current_ia_max_index = -1;
   bool unfinished_1, unfinished_2;
   sieving_element* elements_1 = aRow1->elements;
   sieving_element* elements_2 = aRow2->elements;
   int min_1 = 0;
   int min_2 = 0;   
   qs_relation* last;
   int index;
   int index_1 = 0;
   int index_2 = 0;
   main_sieving_type prime_1;
   main_sieving_type prime_2;
   int exponent_1;
   int exponent_2;
   int max;
   int max_1 = aRow1->maximal_sieving_element_index;
   int max_2 = aRow2->maximal_sieving_element_index;
   // Is the first relation negative?
#ifndef USE_PRIMES
   if ((max_1 >= 0) && (elements_1[0].p == 0) && ((elements_1[0].exponent)%2 == 1)) { min_1 = 1; }
   // Is the second relation negative?
   if ((max_2 >= 0) && (elements_2[0].p == 0) && ((elements_2[0].exponent)%2 == 1)) { min_2 = 1; }
#else
	if ((max_1 >= 0) && (elements_1[0].p == -1) && ((elements_1[0].exponent)%2 == 1)) { min_1 = 1; }
   // Is the second relation negative?
   if ((max_2 >= 0) && (elements_2[0].p == -1) && ((elements_2[0].exponent)%2 == 1)) { min_2 = 1; }
#endif
   // Is their product negative?
   if (min_1 ^ min_2) {
   	// We know that there is at least DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE elements in intermediate array.
	// Whoever changes this constant to 0, will face dire consequences :-)
	// So, no checking of array limits here.
#ifndef USE_PRIMES
	this->intermediate_array[0].p = 0;  // setting -1 as the first divisor
#else
	this->intermediate_array[0].p = -1; // setting -1 as the first divisor
#endif
	this->intermediate_array[0].exponent     = 1;
	current_ia_max_index = 0;
   }
   index_1 = min_1;
   index_2 = min_2;
   // Now, we walk through the sequence of divisors and try to form their combination, as long as there are any left.
   // Beware of the OR condition !!! Here only one | is placed instead of ||, in order to force evaluation of BOTH
   // OR-ed expressions. Should there be two ||, only the first expression would be evaluated at each pass, and the
   // second one only sometimes. This would cause havoc in value of unfinished_2, which could be wrong, inherited
   // from previous iterations (and it indeed used to be, before this small change).
   while ((unfinished_1 = (index_1 <= max_1)) | (unfinished_2 = (index_2 <= max_2))) {
	//printf("Index_1 = %d, max_1 = %d, index_2 = %d, max_2 = %d, unf_1 = %d, unf_2 = %d, XOR = %d\n",index_1,max_1,index_2,max_2,unfinished_1,unfinished_2,unfinished_1 ^ unfinished_2);
   	// At the beginning of each iteration, we distinguish between two cases
	// 1. The divisors of exactly one of the relations have been already spent	
	if (unfinished_1 ^ unfinished_2) { 
		if (unfinished_1) {
			//printf("Unfinished 1\n");
			index = index_1;
			max   = max_1;
			last = aRow1;
		}	
		if (unfinished_2) {
			//printf("Unfinished 2\n");
			index = index_2;
			max   = max_2;
			last = aRow2;
		}
		while (index <= max) {
			//printf("Taking %d (index %d).\n",this->factor_base[last->elements[index].column_index],index);
			ret = AddDivisor(&current_ia_max_index,&(last->elements[index].p),NULL,last->elements[index].exponent);
			if (ret != ConstRC::Ok) {
				return ret;
			}
			index++;			
		}
		break; // Finishing the upper while cycle
	} else {
	// 2. We may still take divisors from each relation. So, reading from both arrays at index_1 and index_2 is safe.
	// At first, we find out which one of the divisors is smaller.
		prime_1 = elements_1[index_1].p;			
		prime_2 = elements_2[index_2].p;
		exponent_1=elements_1[index_1].exponent;			
		exponent_2=elements_2[index_2].exponent;			
		// if they are equal:
		if (prime_1 == prime_2) {
			//printf("Taking %d from both relations. (r1[%d]=%d, r2[%d]=%d])\n",this->factor_base[column_1],index_1,this->factor_base[column_1],index_2,this->factor_base[column_2]);
			ret = AddDivisor(&current_ia_max_index,&prime_1,NULL,exponent_1+exponent_2);
			if (ret != ConstRC::Ok) {
				return ret;
			}
			index_1++;
			index_2++;
			continue; // Next round of "while" cycle
		}
		if (prime_1 < prime_2) {
			//printf("Taking %d from relation 1 (r1[%d]=%d, r2[%d]=%d]).\n",this->factor_base[column_1],index_1,this->factor_base[column_1],index_2,this->factor_base[column_2]);
			ret = AddDivisor(&current_ia_max_index,&prime_1,NULL,exponent_1);
			if (ret != ConstRC::Ok) {
				return ret;
			}
			index_1++;
			continue; // Next round of "while" cycle
		}
		if (prime_1 > prime_2) {
			//printf("Taking %d from relation 2 (r1[%d]=%d, r2[%d]=%d]).\n",this->factor_base[column_2],index_1,this->factor_base[column_1],index_2,this->factor_base[column_2]);
			ret = AddDivisor(&current_ia_max_index,&prime_2,NULL,exponent_2);
			if (ret != ConstRC::Ok) {
				return ret;
			}
			index_2++;
			continue; // Next round of "while" cycle
		}
		// Here should be unreachable territory.
	}
   }
   // Now, the small divisors are present in intermediate_array, sorted ascendingly by their index in FB.
   // We will copy them into the aTarget array.
   // At first, ensure that there is enough space
   if (aTarget->currently_assigned_sieving_elements < (current_ia_max_index+1)) {
   	// realloc
	aTarget->elements = (sieving_element*) AllocationMachine((void **) &(aTarget->elements),(current_ia_max_index+3)*sizeof(sieving_element));					
	if (aTarget->elements == NULL){
		CLog::LogError(WHEREARG,"Quadratic sieve","combine two relations",0,(current_ia_max_index+3)*sizeof(sieving_element),"There was not enough space for divisors.");
		aTarget->currently_assigned_sieving_elements = 0;
		aTarget->maximal_sieving_element_index = -1;
		return NOT_ENOUGH_MEMORY;
	}
	aTarget->currently_assigned_sieving_elements = (current_ia_max_index+3);	
   }
   for (i=0; i<= current_ia_max_index; i++) {
   	aTarget->elements[i].p = intermediate_array[i].p;
	aTarget->elements[i].exponent     = intermediate_array[i].exponent;
   }
   aTarget->maximal_sieving_element_index = current_ia_max_index;
   // In case of smooth values, our job is done
   if (aType == SMOOTH_REL) {
   	
   } else
   if (aType == CYCLE) {
   	// In case of cycle combination, at least one (if not both) relations aRow1 and aRow2 has quality
	// "lprime1 = 1 = lprime2". We must distinguish between those two cases.
	bool is_partial_cycle_1 = false;
	bool is_partial_cycle_2 = false;
	if ((aRow1->lprime1 == 1) && (aRow1->lprime2 == 1)) {
		is_partial_cycle_1 = true;
	}
	if ((aRow2->lprime1 == 1) && (aRow2->lprime2 == 1)) {
		is_partial_cycle_2 = true;
	}
	if (is_partial_cycle_1 ^ is_partial_cycle_2) {
		qs_relation* partial_cycle = NULL;
		large_prime_type prime_1;
		large_prime_type prime_2;
		large_prime_type inter;
		if (is_partial_cycle_1) {
			partial_cycle = aRow1;
			prime_1 = aRow2->lprime1;
			prime_2 = aRow2->lprime2;
		}
		if (is_partial_cycle_2) {
			partial_cycle = aRow2;
			prime_1 = aRow1->lprime1;
			prime_2 = aRow1->lprime2;
		}		
		if (prime_1 > prime_2) {
			inter = prime_1;
			prime_1 = prime_2;
			prime_2 = inter;
		}
		// Now, we have 1 <= prime_1 <= prime_2, and prime_2 is not 1
		// We start copying large primes from partial_cycle to intermediate array
		index = 0;
		max   = partial_cycle->max_lp_index;
		bool  written_1 = false;
		if (prime_1 == 1) written_1 = true;
		bool  written_2 = false;
		current_ia_max_index = -1;
		while (index <= max) {	
// 			printf("Cycle[%d] = %d.\n",index,partial_cycle->lps[index].prime);
			if (prime_1 != 1) {
				if (partial_cycle->lps[index].prime < prime_1) {
					AddDivisor(&current_ia_max_index,NULL,&(partial_cycle->lps[index].prime),partial_cycle->lps[index].exponent);
// 					printf("(a)Added %d from cycle, on position %d.\n",partial_cycle->lps[index].prime,current_ia_max_index);
					index++;
					continue;
				} else
				if (partial_cycle->lps[index].prime == prime_1) {
					if (prime_1 == prime_2) {
						AddDivisor(&current_ia_max_index,NULL,&(partial_cycle->lps[index].prime),partial_cycle->lps[index].exponent+2);
// 						printf("(b)Added %d doubly from cycle+rel, on position %d.\n",prime_1,current_ia_max_index);
						written_1 = true;
						written_2 = true;						
					} else {
						AddDivisor(&current_ia_max_index,NULL,&(partial_cycle->lps[index].prime),partial_cycle->lps[index].exponent+1);
// 						printf("(c)Added %d from cycle+rel, on position %d.\n",prime_1,current_ia_max_index);
						written_1 = true;						
					}	
					index++;		
					continue;		
				} else 
				if ((partial_cycle->lps[index].prime > prime_1) && (!written_1)) {
					if (prime_1 == prime_2) {
						AddDivisor(&current_ia_max_index,NULL,&prime_1,2);
// 						printf("(d)Added %d doubly from rel, on position %d.\n",prime_1,current_ia_max_index);

						written_1 = true;
						written_2 = true;
					} else {
						AddDivisor(&current_ia_max_index,NULL,&prime_1,1);
// 						printf("(e)Added %d from rel, on position %d.\n",prime_1,current_ia_max_index);
						written_1 = true;
					}
					// We do not raise index this time
					continue;
				}
			}
			if (!written_2) {
				if (partial_cycle->lps[index].prime < prime_2) {
					AddDivisor(&current_ia_max_index,NULL,&(partial_cycle->lps[index].prime),partial_cycle->lps[index].exponent);
// 					printf("(f)Added %d from cycle, on position %d.\n",partial_cycle->lps[index].prime,current_ia_max_index);

					index++;
					continue;
				} else
				if (partial_cycle->lps[index].prime == prime_2) {					

					AddDivisor(&current_ia_max_index,NULL,&(partial_cycle->lps[index].prime),partial_cycle->lps[index].exponent+1);
// 					printf("(g)Added %d from cycle+rel, on position %d.\n",prime_2,current_ia_max_index);

					written_2 = true;						
					index++;
					continue;				
				} else 
				if (partial_cycle->lps[index].prime > prime_2) {	
				
					AddDivisor(&current_ia_max_index,NULL,&prime_2,1);
// 					printf("(h)Added %d from rel, on position %d.\n",prime_2,current_ia_max_index);

					written_2 = true;					
					// We do not raise index this time
					continue;
				}
			} else {
				// written_2 = true -> written_1 = true; therefore, both of the primes have been written
				AddDivisor(&current_ia_max_index,NULL,&(partial_cycle->lps[index].prime),partial_cycle->lps[index].exponent);
// 					printf("(i)Added %d from cycle, on position %d.\n",partial_cycle->lps[index].prime,current_ia_max_index);

				index++;
				continue;
			}
		}
		// Finally, if the primes are larger than any primes contained in current partial cycle ...
		if (!written_1) {
			if (prime_1 == prime_2) {
				AddDivisor(&current_ia_max_index,NULL,&prime_1,2);
// 				printf("(j)Added %d from rel doubly, on position %d.\n",prime_1,current_ia_max_index);

				written_1 = true;
				written_2 = true;
			} else {
				AddDivisor(&current_ia_max_index,NULL,&prime_1,1);
// 				printf("(k)Added %d from rel, on position %d.\n",prime_1,current_ia_max_index);
				written_1 = true;
			}
		} 
		if (!written_2) {
			AddDivisor(&current_ia_max_index,NULL,&prime_2,1);
// 			printf("(l)Added %d from rel, on position %d.\n",prime_2,current_ia_max_index);
			written_2 = true;
		}
		// Now, the list of large primes should be finished. We will copy it into "target" location.
		if (aTarget->assigned_lps < (current_ia_max_index+1)) {
			aTarget->lps = (lp_element*) AllocationMachine((void**) &aTarget->lps,(current_ia_max_index+3)*sizeof(lp_element));
			if (aTarget->lps == NULL) {
				CLog::LogError(WHEREARG,"Quadratic sieve","combine two relations",0,(current_ia_max_index+3)*sizeof(lp_element),"There was not enough space for large primes.");
				aTarget->assigned_lps = 0;
				aTarget->max_lp_index = -1;
				return NOT_ENOUGH_MEMORY;
			}
		}
		for (i=0; i<= current_ia_max_index; i++) {
			aTarget->lps[i].prime = intermediate_array[i].large_prime;
			aTarget->lps[i].exponent = intermediate_array[i].exponent;
		}
		aTarget->max_lp_index = current_ia_max_index;
	} else {
		CLog::LogError(WHEREARG,"Quadratic sieve","combine two relations",0,0,"We have tried to combine two relations, neither of which is a partial cycle, or both of them are partial cycles.");		
		return NOT_SUPPORTED;
	}
	
   }
   mpz_set(aTarget->left_side_value,aux_1);
   mpz_set(aTarget->right_side_value,aux_2);
   mpz_set_ui(aTarget->remaining,1);
   aTarget->lprime1 = 1;
   aTarget->lprime2 = 1;
   // these two variables make no sense in combined row
   if (aTarget->inited & FAC_INITED) { mpz_set_ui(aTarget->factorized_value,0); }
   if (aTarget->inited & ARG_INITED) { mpz_set_ui(aTarget->argument,0); }

//    printf("CombineTwoRelations: result\n");
//    PrintRelation(aTarget);
//    printf("\n");	 
   int res = AssertRelationConsistency(aTarget,"Problem",__LINE__);   	
   if (res != ConstRC::Ok)
	{
	return res;
	}
   if (this->assertion_mode) {
   	// the "< max_index" is deliberate; we operate with [i+1]
   	for (i=0; i< aTarget->maximal_sieving_element_index; i++) {
		if (aTarget->elements[i].p >= aTarget->elements[i+1].p) {
			NormalMessage("Assertion failed: unsorted divisors. ",NO_NUMBER);
			PrintRelation(aTarget);	
		}
	}
	// same here
	for (i=0; i< aTarget->max_lp_index; i++) {
		if (aTarget->lps[i].prime >= aTarget->lps[i+1].prime) {
			NormalMessage("Assertion failed: unsorted large primes. ",NO_NUMBER);
			PrintRelation(aTarget);	
		}		
	}
   }
   return ConstRC::Ok;
}


/*!
	An auxiliary method used during combination of relations. Returns NOT_ENOUGH_MEMORY 
	or ConstRC::Ok, depending on whether the inner allocation failed.
	
	A multi-purpose function. Precisely one of aSmallPrime and aLargePrime should be NULL.
*/	
int  QuadraticSieve::AddDivisor(int* aMaxIndex,main_sieving_type* aSmallPrime,large_prime_type* aLargePrime, int aExponent) {
	(*aMaxIndex)++;
	if ((*aMaxIndex) == this->ia_allocated) {
		// Need to reallocate the intermediate array
		this->intermediate_array = (intermediate_array_element*) AllocationMachine((void**) &this->intermediate_array,(this->ia_allocated+DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE)*sizeof(intermediate_array_element));
		if (this->intermediate_array == NULL) {
 			CLog::LogError(WHEREARG,"Quadratic sieve","combining two rows",0,(this->ia_allocated+DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE)*sizeof(intermediate_array_element),"Could not allocate memory for intermediate array");
			return NOT_ENOUGH_MEMORY;
		}
		this->ia_allocated += DEFAULT_INTERMEDIATE_ARRAY_ALLOCATE;
	}
	if (aSmallPrime != NULL) {
		this->intermediate_array[(*aMaxIndex)].p = *aSmallPrime;
		this->intermediate_array[(*aMaxIndex)].exponent = aExponent;
	}
	if (aLargePrime != NULL) {
		this->intermediate_array[(*aMaxIndex)].large_prime = *aLargePrime;
		this->intermediate_array[(*aMaxIndex)].exponent = aExponent;
	}
	return ConstRC::Ok;
}

// Limited! Only for use with smooth relations (having RSV's) or 1-partials with the same lprime1
int QuadraticSieve::CombineTwoRelations(qs_relation* aTarget, qs_relation* aRow1, qs_relation* aRow2){
   long i, j;
   // Now, the combination of the divisors takes place.
   int	max_1 = aRow1->maximal_sieving_element_index;
   int	max_2 = aRow2->maximal_sieving_element_index;
   //printf("CombineTwoRelations 5.\n");
   int	maximal_number_of_elements = max_1 + max_2 + 3; // to be sure.
   // (-1)^2 is, of course, 1. it would be silly to write this factor to the product. So we check the fac's sign
   int write_minus_1 = (mpz_cmp_si(aRow1->factorized_value,0)*mpz_cmp_si(aRow2->factorized_value,0) < 0);
   //printf("CombineTwoRelations 6.\n");
   // in case of need we realloc the corresponding element array
   if (aTarget->currently_assigned_sieving_elements < maximal_number_of_elements) {
   	//printf("\nWe will realloc the corresponding element array %p (%d vs %d).\n",aTarget->elements,aTarget->currently_assigned_sieving_elements,maximal_number_of_elements);
	aTarget->elements = (sieving_element*) AllocationMachine((void**) &(aTarget->elements),maximal_number_of_elements*sizeof(sieving_element));
	//printf("We have reallocated the corresponding element array.\n");
	if (!aTarget->elements){
		CLog::LogError(WHEREARG,"Quadratic sieve","combining two rows",max_1,max_2,"Could not allocate memory for divisors of a combined row");
		return NOT_ENOUGH_MEMORY;
	}
	aTarget->currently_assigned_sieving_elements = maximal_number_of_elements;
   }
   aTarget->maximal_sieving_element_index = -1;
   //printf("CombineTwoRelations 7.\n");
   if (write_minus_1){
	aTarget->maximal_sieving_element_index++;
#ifndef USE_PRIMES
	aTarget->elements[aTarget->maximal_sieving_element_index].p = 0;
#else
	aTarget->elements[aTarget->maximal_sieving_element_index].p = -1;
#endif
	aTarget->elements[aTarget->maximal_sieving_element_index].exponent = 1;
   }
   // elements from the first row
   //printf("CombineTwoRelations 8.\n");
  for (i=0; i<=max_1; i++){
		// in case of the element not being -1, we write it down
		if (aRow1->elements[i].p){
			aTarget->maximal_sieving_element_index++;
			aTarget->elements[aTarget->maximal_sieving_element_index] = aRow1->elements[i];
		}
  }
  // with regard to the already copied factors, we add factors of the second summand
#ifndef USE_PRIMES
  long index;
#else
  main_sieving_type index;
#endif
  int not_found;
  //printf("CombineTwoRelations 9.\n");
  for (i=0; i<=max_2; i++){
        // in case of the factor not being -1 (that is, its index be nonzero ...)
#ifndef USE_PRIMES
	if (index=aRow2->elements[i].p){
#else
	if ((index=aRow2->elements[i].p) != -1){
#endif
		// the factors are not sorted, so we must use a dirty linear search
		not_found = 1;
		for (j=0; j<=aTarget->maximal_sieving_element_index; j++){
			if (aTarget->elements[j].p == index){		  							aTarget->elements[j].exponent += aRow2->elements[i].exponent;
				not_found = 0;
				break;
			}
		} // end of for (j=0
		if (not_found){
			aTarget->maximal_sieving_element_index++;
			aTarget->elements[aTarget->maximal_sieving_element_index] = aRow2->elements[i];
	        } // end of if(not_found)
	} // end of if(index)
   } // end of for (i=0
   //printf("CombineTwoRelations 10.\n");
   return ConstRC::Ok;
}

/*!
	When collected relations are checked for duplicities, their hashes are saved into a binary tree. This method
	saves the hash to the binary tree, altogether with index of this relation in matrix of relations.
	
	Return codes: NOT_FOUND or FOUND.

*/
int QuadraticSieve::AddHashToTree(mpz_t aValue, long aIndex){
  ROW_HASH* hash  = NULL;
  ROW_HASH* pred_poi;
  ROW_HASH** pred = &pred_poi;
  int last = NONE;
  if (!FindHash(&last,pred,aValue)){
	    hash = (ROW_HASH*) malloc(sizeof(ROW_HASH));
	    mpz_init(hash->hash_value);
	    hash->hash_inited = HASH_INITED;
	    mpz_set(hash->hash_value,aValue);
            hash->first_row_index = aIndex;
            hash->left = NULL;
            hash->right = NULL;
	    //gmp_printf("\nVari %Zd has not been found in variations. Pointer to var is %p, do *pred is %p, last is %d (NONE=%d, LEFT=%d, RIGHT=%d).\n",aValue,var,*pred,last,NONE,LEFT,RIGHT);
	    //fflush(stdout);

	    switch (last) {
            	case LEFT:  hash->left = (*pred)->left;
                	     (*pred)->left = hash;
                             break;
            	case RIGHT:  hash->right = (*pred)->right;
                             (*pred)->right = hash;
                             break;
            	case NONE:   this->hash_root = hash;
			     break;
            }
	    return NOT_FOUND;
  }
  return FOUND;
}

/*!
This method recursively releases memory taken by the binary tree of hashes. It is called by the 
	destructor.
	
*/
void QuadraticSieve::DeleteHashes(ROW_HASH* aHash){
     ROW_HASH* right;
     if (aHash->left) {
     	DeleteHashes(aHash->left);
	aHash->left = NULL;
     }
     if (aHash->hash_inited & HASH_INITED){
     	mpz_clear(aHash->hash_value);
	aHash->hash_inited &= (~HASH_INITED);
     }
     right = aHash->right;
     free(aHash);
     aHash = NULL;
     if (right) {
     	DeleteHashes(right);
     }
}

/*!
	
	This method recursively searches the binary tree of hashes for occurence of aValue. The pointer variable
	aLast serves for messagging between various recursive levels of FindHash calls and carries values LEFT
	and RIGHT.
	
	Return codes: TRUE, FALSE.
	
*/
bool QuadraticSieve::FindHash(int* aLast,ROW_HASH** aPred, mpz_t aValue){
	*aPred = NULL;
	*aLast = NONE;
	if (this->hash_root){
	   ROW_HASH* hash = this->hash_root;
	   bool found = FALSE;
	   while (hash && (!found)){
	   	if (!mpz_cmp(hash->hash_value,aValue)){			
			found = TRUE;
		} else { // We were not lucky this round
           		*aPred = hash;
           		if (mpz_cmpabs(hash->hash_value,aValue)>0){ // we go left from predecessor

			     if (hash->left == NULL){
			     	*aLast = LEFT; // adding leaf to the tree
               			break;
               		     }
              		     else (hash = hash->left);
             		} else { // we go right from predecessor
			     if (hash->right == NULL){
				*aLast = RIGHT;
               			break;
               		     }
             		     else (hash=hash->right);
             		} // end of else(right)
		} // end of else(we were not lucky this round)
	   } // end of while
	   return found;
	} // end of if
	else { // nothing in tree
		return FALSE;
	}
}

/*!
This method swaps two rows in relation matrix aTarget.
*/
void QuadraticSieve::SwapTwoRows(relation_matrix* aTarget, long aInd_1, long aInd_2){
        qs_relation* r1 = &aTarget->matrix[aInd_1];
	qs_relation* r2 = &aTarget->matrix[aInd_2];
	mpz_swap(r1->argument,r2->argument);
	mpz_swap(r1->factorized_value,r2->factorized_value);
	mpz_swap(r1->remaining,r2->remaining);
	mpz_swap(r1->left_side_value,r2->left_side_value);
	mpz_swap(r1->right_side_value,r2->right_side_value);

	long pom;
	void* pointer;

	pointer = r2->elements;
	r2->elements = r1->elements;
	r1->elements = (sieving_element*) pointer;

	pom = r2->maximal_sieving_element_index;
	r2->maximal_sieving_element_index = r1->maximal_sieving_element_index;
	r1->maximal_sieving_element_index = pom;

	pom = r2->currently_assigned_sieving_elements;
	r2->currently_assigned_sieving_elements = r1->currently_assigned_sieving_elements;
	r1->currently_assigned_sieving_elements = pom;

	pom = r2->index_of_used_polynomial;
	r2->index_of_used_polynomial = r1->index_of_used_polynomial;
	r1->index_of_used_polynomial = pom;

	pom = r2->inited;
	r2->inited = r1->inited;
	r1->inited = pom;
}

// A binary searcher.
/*!
	This method uses binary search to find index of aValue in sorted list aList, whose greatest index
	is given by aMaxIndex. It returns either this index, or NOT_FOUND.

*/
long QuadraticSieve::IsInList(long* aList, long aMaxIndex, long aValue){
	long min = 0;
	long max = aMaxIndex;
	long s;
	while ((max-min)>1){
	    	s = min+(max-min)/2;
		    if (aList[s] <= aValue){
  		  	min = s;
  	  	}
  		  else{
    			max = s;
  	  	}
        } // end of while
        if (aList[min] == aValue) { s = min; }
        else if (aList[max] == aValue) { s = max; }
        else s = NOT_FOUND;
        return s;
}

void QuadraticSieve::DetailedAssertion(qs_relation* aElement) {
		main_sieving_type i, j;
     		// We assert whether the read value does really satisfy the rules of quadratic sieve.
		// Namely, we must have (2Ax+b)^2 = product_of_primes*remaining*lpv1*lpv2*(RSV)^2
		printf("\nDetailed assertion of relation.\n");
     		mpz_mul(aux_1,aElement->left_side_value,aElement->left_side_value);
		mpz_mod(aux_1,aux_1,this->kN);
#ifndef USE_PRIMES		
		if (this->factor_base[aElement->elements[0].p] == -1) {
#else
		if (aElement->elements[0].p == -1) {
#endif
			mpz_sub(aux_1,aux_1,this->kN);
		}
		gmp_printf("LSV is %Zd.\n",aux_1);
		for (i=0; i<= aElement->maximal_sieving_element_index; i++){
#ifndef USE_PRIMES		
			if (this->factor_base[aElement->elements[i].p] == -1) {
#else
			if (aElement->elements[i].p == -1) {
#endif
				printf("Division by    -1 ...");
				mpz_mul_si(aux_1,aux_1,-1);
				gmp_printf("Ok. Remaining %Zd.\n",aux_1);
			} else
			for (j=0; j< aElement->elements[i].exponent; j++){
#ifndef USE_PRIMES		

				printf("Division by %4d ...",this->factor_base[aElement->elements[i].p]);
				mpz_fdiv_qr_ui(aux_3,aux_2,aux_1,this->factor_base[aElement->elements[i].p]);	
#else
				printf("Division by %4d ...",aElement->elements[i].p);
				mpz_fdiv_qr_ui(aux_3,aux_2,aux_1,aElement->elements[i].p);	
#endif
				if (mpz_cmp_ui(aux_2,0)){
					printf("FAILED!\n");
				} else {
					mpz_swap(aux_1,aux_3);
					gmp_printf("Ok. Remaining %Zd.\n",aux_1);
				}
			}
		}	
		if (aElement->lprime1 != 1) {
			std::cout << "Division by lprime1 ("<< aElement->lprime1 <<") ...";
			mpz_fdiv_qr_ui(aux_3,aux_2,aux_1,aElement->lprime1);	
			if (mpz_cmp_ui(aux_2,0)){
					printf("FAILED!\n");
				} else {
					mpz_swap(aux_1,aux_3);
					gmp_printf("Ok. Remaining %Zd.\n",aux_1);
				}
		}
		if (aElement->lprime2 != 1) {
			std::cout << "Division by lprime2 ("<< aElement->lprime2 <<") ...";
			mpz_fdiv_qr_ui(aux_3,aux_2,aux_1,aElement->lprime2);	
			if (mpz_cmp_ui(aux_2,0)){
					printf("FAILED!\n");
				} else {
					mpz_swap(aux_1,aux_3);
					gmp_printf("Ok. Remaining %Zd.\n",aux_1);
				}
		}
}

/*!

	This method checks the given relation. It asserts whether the read value does really satisfy 
	the rules of quadratic sieve: whether (2Ax+b)^2 = product_of_primes*remaining*(RSV)^2
	
	In case of failure, it prints out "Assertion failed in assert row consistency".

*/
int  QuadraticSieve::AssertRelationConsistency(qs_relation* aElement, const char* aMessage, int aValue){
		main_sieving_type i, j;
     		// We assert whether the read value does really satisfy the rules of quadratic sieve.
		// Namely, we must have (2Ax+b)^2 = product_of_primes*remaining*lpv1*lpv2*(RSV)^2
     		mpz_mul(aux_1,aElement->left_side_value,aElement->left_side_value);
		mpz_set(aux_2,aElement->remaining);
		mpz_mul_ui(aux_2,aux_2,aElement->lprime1);
		mpz_mul_ui(aux_2,aux_2,aElement->lprime2);
		mpz_mul(aux_2,aux_2,aElement->right_side_value);
		mpz_mul(aux_2,aux_2,aElement->right_side_value);
		for (i=0; i<= aElement->maximal_sieving_element_index; i++){
			for (j=0; j< aElement->elements[i].exponent; j++){
				// The "si" here is important - often, you have a -1 among factors.
#ifndef USE_PRIMES
				mpz_mul_si(aux_2,aux_2,this->factor_base[aElement->elements[i].p]);
#else
				mpz_mul_si(aux_2,aux_2,aElement->elements[i].p);
#endif
			}
		}
		if (aElement->lps != NULL) {
			for (i=0; i<= aElement->max_lp_index; i++) {
				for (j=0; j<aElement->lps[i].exponent; j++) {
					mpz_mul_ui(aux_2,aux_2,aElement->lps[i].prime);
				}
			}
		}
		mpz_sub(aux_1,aux_1,aux_2);
		mpz_mod(aux_1,aux_1,this->kN);
		if (mpz_cmp_ui(aux_1,0)){
			NormalMessage("Assertion failed in assert row consistency, ",NO_NUMBER);
			PrintRelation(aElement);
			mpz_mul(aux_1,aElement->left_side_value,aElement->left_side_value);
			gmp_printf("Left side squared is %Zd, right side (prime powers product) is %Zd.\n",aux_1,aux_2);
			if (aMessage) { printf("%s, %d",aMessage,aValue); }
			DetailedAssertion(aElement);
			ABNORMAL_EXIT(1);
			return GENERAL_ERROR;
		} 
		return ConstRC::Ok;
}

int QuadraticSieve::InitForList(const char* aDirectoryName)
{
	if (aDirectoryName)
	{
		Set_DirectoryName(aDirectoryName);
	}
	int ret = InitFileNames();
	if (ret != ConstRC::Ok)
	{
		return ret;
	}
	ret = CalculateKRoot();
	if (ret != ConstRC::Ok)
	{ 
		return ret; 
	}
	return ConstRC::Ok;
}

/*!
	This method deallocates indicated relation_matrix object.
	Return codes: ConstRC::Ok only.	

*/
int  QuadraticSieve::DeleteRelationMatrix(relation_matrix* aTarget){
	long j;
	qs_relation* element;
	if (this->info_mode){
		NormalMessage("   Deleting one of smooth matrices ...",NO_NUMBER);
        }
	for (j=0; j< aTarget->assigned_rows; j++){
		element = &aTarget->matrix[j];
		ClearRelation(element);
	}
	free(aTarget->matrix);
	aTarget->maximal_row_index = -1;
	aTarget->assigned_rows = 0;
	return ConstRC::Ok;
}

/*!
	This method frees the array main_matrix. 
	
	Return codes: ConstRC::Ok

*/
int  QuadraticSieve::DeleteMainMatrix(){
	if (this->main_matrix->accumulated_logarithm){
		free(this->main_matrix->accumulated_logarithm);
		this->main_matrix->accumulated_logarithm = NULL;
	}
	this->main_matrix->maximal_index = -1;
	this->main_matrix->assigned = 0;
	return ConstRC::Ok;
}

/*!

	There is a fundamental problem in coherency of the saved siever status,
	given by the fact that serialization of the siever status takes place
	only from time to time (say, every 2 minutes), while the files "smooths"
	and "partials" grow continuously. 
	
	Therefore, if the sieving is killed externally (not by SIGTERM, which is
	caught by the program, but by SIGKILL or SIGSEGV or whatever), the last
 	saved siever status may (and usually will) not reflect the last state
	of the "smooths" and "partials"; and if we then restart the sieving
	from the last saved siever status, all the smooth and partial relations
	that have been written down after the last save of the siever status,
	would be found twice - once in the previous sieving, once in the restarted
	sieving.

	Therefore, we will remove all the smooths and partials that were found
	after the last save of the siever status.

*/
int     QuadraticSieve::Revert(long aSmooths, long aPartials)
{
	qs_relation* candidate = InitRelation();
	if (candidate == NULL)
	{
		return NOT_ENOUGH_MEMORY;
	}
	long counter = 0;
	
	if (aSmooths > 0)	
	{
		this->smooth_values_file      = OpenFile(this->smooth_values_file_name,READ);	
		this->temporary_file	= OpenFile(this->temporary_file_name,WRITE);	
		while (ReadRelationFromFile(this->smooth_values_file,candidate) == ConstRC::Ok) {
			counter++;
			PrintRelationToFile(this->temporary_file,candidate,candidate->index);				
			if (counter >= aSmooths)
			{
				break;
			}
		}
		CloseFile(this->smooth_values_file_name);
		CloseFile(this->temporary_file_name);
		remove(this->smooth_values_file_name.c_str());
		rename(this->temporary_file_name.c_str(),this->smooth_values_file_name.c_str());
		if (this->info_mode){
    			printf("   Reverted %ld smooths\n",aSmooths);
    		}
	}
	counter = 0;
	if (aPartials > 0)
	{
		this->partials_file 	= OpenFile(this->partial_values_file_name,READ);	
		this->temporary_file	= OpenFile(this->temporary_file_name,WRITE);	
		while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
			counter++;
			PrintRelationToFile(this->temporary_file,candidate,candidate->index);				
			if (counter >= aPartials)
			{
				break;
			}
		}
		CloseFile(this->partial_values_file_name);
		CloseFile(this->temporary_file_name);
		remove(this->partial_values_file_name.c_str());
		rename(this->temporary_file_name.c_str(),this->partial_values_file_name.c_str());
		if (this->info_mode){
    			printf("   Reverted %ld partials\n",aPartials);
    		}
	}
	ClearRelation(candidate);
	return ConstRC::Ok;
}

/*!
	This method has effect only in assertion mode. It asserts consistency of the collected rows: that is, if the
	saved values really form a smooth value. This is useful in development, especailly after the singletons have
	been deleted.	

*/
void QuadraticSieve::AssertSmooths(){
	if (this->assertion_mode){
		if (this->info_mode){
			NormalMessage("   Asserting consistency of smooth values ...",NO_NUMBER);
	        }
		long i;
		for (i=0; i<=this->smooth_values->maximal_row_index; i++){
			int res = AssertRelationConsistency(&this->smooth_values->matrix[i],"Asserting all smooths",i);
			if (res != ConstRC::Ok)
			{
				ABNORMAL_EXIT(1);
			}
		}
	}
}

/*!
	This method creates either a SparseMatrix or a BitMatrix with number of rows equal to number of collected smooth
	values. Then it walks through the matrix containing smooth values and if it meets a prime with odd index, it 
	adds 1 to the SparseMatrix or BitMatrix at corresponding place.
	
	Return values:
	- NOT_ENOUGH if the matrix has lesser rows than columns
	- BAD_ARGUMENT if this->matrix_type is neither ESparseBitMatrix not EBitMatrix
	- ConstRC::Ok if everything went well

*/
int QuadraticSieve::ToBitMatrix(){
	int max_row = this->smooth_values->maximal_row_index;
	int max_column = this->maximal_factor_base_index;
	if (max_row < max_column) { 
		printf("ToBitMatrix: maximal row is %d, maximal column is %d.\n",max_row,max_column);
		return NOT_ENOUGH; 
	}
	sieving_element* current = NULL;
	int i = 0;
	int k = 0;
	int how_much = 0;
	int total = 0;
	if (this->info_mode){
		NormalMessage("   Converting to bit matrix representation ... ",NO_NUMBER);
		fflush(stdout);
        }
	switch (this->matrix_type){
		case ESparseBitMatrix: this->class_result_sparse = new SparseMatrix(max_row+1,max_column+1);
				        for (i=0; i<=this->smooth_values->maximal_row_index; i++){
						how_much = 0;
						for (k=0; k<=this->smooth_values->matrix[i].maximal_sieving_element_index; k++){
							current = &(this->smooth_values->matrix[i].elements[k]);
							if ((current->exponent)%2){
#ifndef USE_PRIMES
								this->class_result_sparse->PutOne(i,current->p);
#else
								this->class_result_sparse->PutOne(i,GetTrueIndexInFB(&current->p));
#endif
								how_much++;
							}
						}
						//this->result_sparse->matrix[i][1] = how_much;
						total += how_much;
					}
 					this->class_result_sparse->Reduce();
#ifdef CHECK_SPARSE_MATRICES
					this->class_result_bit = new BitMatrix(max_row+1,max_column+1);
					for (i=0; i<=this->smooth_values->maximal_row_index; i++){
						for (k=0; k<=this->smooth_values->matrix[i].maximal_sieving_element_index; k++){
							current = &(this->smooth_values->matrix[i].elements[k]);
							if ((current->exponent)%2){							this->class_result_bit->PutOne(i,current->column_index);
								total++;								}
						}
					}
					this->class_normal_result_sparse = this->class_result_sparse->ToNormalMatrix();
					this->class_normal_result_bit = this->class_result_bit->ToNormalMatrix();
					printf("\n   Comparing bit and sparse matrix via normal ones ");
					if (!(this->class_normal_result_sparse->Equals(this->class_normal_result_bit))){
						printf(" Do not match!  ");
						return GENERAL_ERROR;
					} else printf(" Match!   ");
#endif
					break;
		case EBitMatrix:	this->class_result_bit = new BitMatrix(max_row+1,max_column+1);
					this->class_result_bit->Allocate();
					total = 0;
					for (i=0; i<=this->smooth_values->maximal_row_index; i++){
						for (k=0; k<=this->smooth_values->matrix[i].maximal_sieving_element_index; k++){
							current = &(this->smooth_values->matrix[i].elements[k]);
							if ((current->exponent)%2){	
#ifndef USE_PRIMES	
								this->class_result_bit->PutOne(i,current->p);
#else
								this->class_result_bit->PutOne(i,GetTrueIndexInFB(&current->p));
#endif
								total++;								}
						}
					}
					break;
		default:		CLog::LogError(WHEREARG,"Quadratic Sieve","conversion to bit matrix",(int) this->matrix_type,105,"Unknown type of matrix");
					return BAD_ARGUMENT;
	}
	if (this->info_mode){
		printf("(%d nonzero bits in matrix)",total);
	}
	return ConstRC::Ok;
}

/*!
	This method transposes the created (sparse) bit matrix and calls the Lanczos method on it.
	
	Return values:
	- NOT_SUPPORTED if block Wiedemann method is requested
	- BAD_ARGUMENT if anything else than block Wiedemann or block Lanczos method is requested. 
	- ConstRC::Ok if everything went well

*/
int QuadraticSieve::SolveSystem(){
	if (this->frontend){
		fprintf(this->frontend_pipe,"\nPhaseLin> \n");
		fflush(this->frontend_pipe);
	}
	if (this->info_mode){
		NormalMessage("   Solving linear equation system ...",NO_NUMBER);
        }
	switch (this->algorithm_type){
		case EBlockLanczos:     if (this->matrix_type == ESparseBitMatrix){
						/*this->class_transpose_sparse = new SparseMatrix(this->class_result_sparse->GetMaximalColumnIndex()+1, this->class_result_sparse->GetMaximalRowIndex()+1);
						ret = this->class_result_sparse->Transpose(this->class_transpose_sparse);
						if (ret != ConstRC::Ok){
							CLog::LogError(WHEREARG,"Quadratic sieve","solve linear equation system",115,120,"Some error transposing result sparse matrix");
							return ret;
						}
#ifdef CHECK_SPARSE_MATRICES
						this->class_transpose_bit = new BitMatrix(this->class_result_bit->GetMaximalColumnIndex()+1,this->class_result_bit->GetMaximalRowIndex()+1);
						ret = this->class_result_bit->Transpose(this->class_transpose_bit);
						this->class_normal_transpose_sparse = this->class_transpose_sparse->ToNormalMatrix();
						this->class_normal_transpose_bit = this->class_transpose_bit->ToNormalMatrix();
						printf("\n   Comparing bit and sparse transpose matrix via normal ones ");
						if (!(this->class_normal_transpose_sparse->Equals(this->class_normal_transpose_bit))){
							printf(" Do not match!  ");
							return GENERAL_ERROR;
						} else printf(" Match!   ");
#endif
						*/
						this->lanczos = new Lanczos(SSparseMatrix);
						this->lanczos->SetAuxiliaryMatricesType(SBCMatrix);
						this->lanczos->Set_InfoMode(this->info_mode);
						this->lanczos->Set_AssertionMode(this->assertion_mode);
						this->lanczos->Set_RankCalculationMode(this->rank_calculation_mode);
						this->lanczos->Set_DensityMode(this->density_info_mode);
						this->lanczos->Set_TimeMessageMode(this->lanczos_time_message_mode);
						//this->lanczos->SetFrontend(this->frontend_pipe);
						printf("Here, the frontend pipe should be set.\n");
						this->class_dependencies = this->lanczos->Compute(this->class_result_sparse/*,this->class_transpose_sparse*/);
						if (!this->class_dependencies) { return NO_DEPENDENCIES; }
					 }
					 if (this->matrix_type == EBitMatrix){
						/*this->class_transpose_bit = new BitMatrix(this->class_result_bit->GetMaximalColumnIndex()+1,this->class_result_bit->GetMaximalRowIndex()+1);
						ret = this->class_result_bit->Transpose(this->class_transpose_bit);
						if (ret != ConstRC::Ok){
							CLog::LogError(WHEREARG,"Quadratic sieve","solve linear equation system",215,220,"Some error transposing result bit matrix");
							return ret;
						}*/
						this->lanczos = new Lanczos(SBitMatrix);
						this->lanczos->SetAuxiliaryMatricesType(SBCMatrix);
						this->lanczos->Set_InfoMode(this->info_mode);
						this->lanczos->Set_AssertionMode(this->assertion_mode);
						this->lanczos->Set_RankCalculationMode(this->rank_calculation_mode);
						this->lanczos->Set_DensityMode(this->density_info_mode);
						this->lanczos->Set_TimeMessageMode(this->lanczos_time_message_mode);
						//this->lanczos->SetFrontend(this->frontend_pipe);
						printf("Here, the frontend pipe should be set.\n");
						this->class_dependencies = this->lanczos->Compute(this->class_result_bit/*,this->class_transpose_bit*/);
						if (!this->class_dependencies) { return NO_DEPENDENCIES; }
					 }
					 break;
		case EBlockWiedemann:   CLog::LogError(WHEREARG,"Quadratic sieve","solving system of linear equations",EBlockWiedemann,NOT_SUPPORTED,"This algorithm is not supported");
					 return NOT_SUPPORTED;
		default:		 CLog::LogError(WHEREARG,"Quadratic sieve","solving system of linear equations",this->algorithm_type,NOT_SUPPORTED,"This algorithm is not supported, even not defined");
					 return BAD_ARGUMENT;
	}
	return ConstRC::Ok;
}

/*!
	This method tries to find nontrivial factors of N by taking all of the found dependencies one by one and calling
	TryKthDependence() on it. It returns either
	- ConstRC::Factorized
	- NOT_FACTORIZED
	- or anything thrown by TryKthDependence (in case of mem allocation failures etc.)

*/
int  QuadraticSieve::FindFactors(){
	long k;
	long max = this->class_dependencies->GetMaximalRowIndex();
	int ret;
	for (k=0; k<=max; k++){
		ret = TryKthDependence(k);
		if (ret == ConstRC::NotFactorized) continue;
		else return ret; // case of finding a factor, or of a memory error etc.
	}
	return ConstRC::NotFactorized;
}

/*!
	This method takes the aK-th dependence from list of dependencies generated by linear algebra
	step and combines the indicated relations into X,Y such that X^2 = Y^2 mod kN. If the
	combination does not satisfy those rules, an error message is printed.
	
	After this combination, it tries to factor the number kN by Fermat factorization.
	
	Return codes:
	- BAD_ARGUMENT if X^2 != Y^2 mod kN
	- ConstRC::Factorized if this dependence has led to nontrivial factorization of kN
	- NOT_FACTORIZED otherwise.

*/
int  QuadraticSieve::TryKthDependence(int aK){
	// aux_1 will be left_root and aux_2 left_square
	// aux_3 will be right_root and aux_4 right_square
	// aux_5 will serve as "pom"
	sieving_element *right_side = NULL;
	int size_of_right_side = this->maximal_factor_base_index+1;
	int i, j;
#ifndef USE_PRIMES
	long prime_index;
#else
	long prime_index;
	main_sieving_type prime;
#endif
	int exponent;
	int half;
	int max = this->class_dependencies->GetMaximalColumnIndex();
	bool flag;

	right_side = (sieving_element*) AllocationMachine((void**) &right_side,size_of_right_side*sizeof(sieving_element));
	if (!right_side) {
		CLog::LogError(WHEREARG,"Quadratic sieve","trying k-th dependence",k,size_of_right_side,"There was not enough memory to allocate right side");
		return NOT_ENOUGH_MEMORY;
	}
//#define CHECK_RELATIONS_AFTER_LINEAR_PHASE
#ifdef CHECK_RELATIONS_AFTER_LINEAR_PHASE 
	for (i=0; i<size_of_right_side; i++){
		right_side[i].p = i;
		right_side[i].exponent = 0;
	}
	for (i=0; i<=max; i++){
		mpz_set_ui(aux_1,1);
		mpz_set_ui(aux_2,1);
		mpz_set_ui(aux_3,1);
		mpz_set_ui(aux_4,1);	
		flag = this->class_dependencies->IsOne(aK,i);
		if (flag){
			mpz_mul(aux_2,aux_1,this->smooth_values->matrix[i].left_side_value);
			mpz_mul(aux_2,aux_2,this->smooth_values->matrix[i].left_side_value);
			mpz_mul(aux_4,aux_4,this->smooth_values->matrix[i].right_side_value);
			mpz_mul(aux_4,aux_4,this->smooth_values->matrix[i].right_side_value);
			gmp_fprintf(stderr, " [ %Zd ] ",aux_4);
			for (j=0; j<=this->smooth_values->matrix[i].maximal_sieving_element_index; j++){
				prime = this->smooth_values->matrix[i].elements[j].p;
				exponent    = this->smooth_values->matrix[i].elements[j].exponent;
				mpz_set_ui(aux_5,1);
				for (int k=0; k< exponent; k++){ // this looks stupid, but remember that -1 is in factorbase and we
				                         // cannot easily use mpz_ui_pow_ui
					mpz_mul_si(aux_5,aux_5,prime);
				}
				mpz_mul(aux_4,aux_4,aux_5);  // square
				gmp_fprintf(stderr, " [ %Zd  %Zd ] ",aux_4, aux_5);
			}
			gmp_fprintf(stderr, "....%Zd....%Zd....",aux_2,aux_4);
			 mpz_sub(aux_5, aux_4, aux_2);
			gmp_fprintf(stderr, "...%Zd...",aux_5);
			mpz_mod(aux_5, aux_5, this->kN);
			if (mpz_cmp_ui(aux_5,0)){
				CLog::LogError(WHEREARG,"Quadratic sieve","trying dependence",aK,100,"Unexpected error: squares are not the same modulo kN");
				ABNORMAL_EXIT(1);
				free(right_side);
				return BAD_ARGUMENT;
			}

		}
	}
#endif
	for (i=0; i<size_of_right_side; i++){
		right_side[i].p = i;
		right_side[i].exponent = 0;
	}

	mpz_set_ui(aux_1,1);
	mpz_set_ui(aux_2,1);
	mpz_set_ui(aux_3,1);
	mpz_set_ui(aux_4,1);
	for (i=0; i<=max; i++){

		flag = this->class_dependencies->IsOne(aK,i);

		if (flag){
			// left side
			mpz_mul(aux_1,aux_1,this->smooth_values->matrix[i].left_side_value); // root
			mpz_mul(aux_2,aux_2,this->smooth_values->matrix[i].left_side_value);
			mpz_mul(aux_2,aux_2,this->smooth_values->matrix[i].left_side_value); // square
			mpz_mod(aux_1,aux_1,this->kN);
			mpz_mod(aux_2,aux_2,this->kN);
			// right side
			mpz_mul(aux_3,aux_3,this->smooth_values->matrix[i].right_side_value); // root
			mpz_mul(aux_4,aux_4,this->smooth_values->matrix[i].right_side_value);
			mpz_mul(aux_4,aux_4,this->smooth_values->matrix[i].right_side_value); // square
			mpz_mod(aux_3,aux_3,this->kN);
			mpz_mod(aux_4,aux_4,this->kN);
			for (j=0; j<=this->smooth_values->matrix[i].maximal_sieving_element_index; j++){
#ifndef USE_PRIMES
				prime_index = this->smooth_values->matrix[i].elements[j].p;
#else 
				prime = this->smooth_values->matrix[i].elements[j].p;
#endif
				exponent    = this->smooth_values->matrix[i].elements[j].exponent;
#ifndef USE_PRIMES
				right_side[prime_index].exponent += exponent;
#else 
				prime_index = GetTrueIndexInFB(&prime);
				if (prime_index == -1)
				{
					CLog::LogError(WHEREARG,"Quadratic sieve","trying dependence",aK,prime,"Unexpected error: Could not find a prime in the factor base.");
				}
				right_side[prime_index].exponent += exponent;
				//return GENERAL_ERROR;
#endif
			}
		}
	}
	for (i=0; i< size_of_right_side; i++){
		if (right_side[i].exponent){
			if (right_side[i].exponent % 2){
				CLog::LogError(WHEREARG,"Quadratic sieve","trying dependence",aK,i,"Unexpected error: odd exponent. Probably serious bug in linalg step");
			//	free(right_side);
			//	return BAD_ARGUMENT;
			}
			half = right_side[i].exponent / 2;
			mpz_set_ui(aux_5,1);
			for (j=0; j< half; j++){ // this looks stupid, but remember that -1 is in factorbase and we
			                         // cannot easily use mpz_ui_pow_ui
				mpz_mul_si(aux_5,aux_5,this->factor_base[i]);
			}
			mpz_mul(aux_3,aux_3,aux_5);  // root
			mpz_mul(aux_4,aux_4,aux_5);
			mpz_mul(aux_4,aux_4,aux_5);  // square
			mpz_mod(aux_3,aux_3,this->kN);
			mpz_mod(aux_4,aux_4,this->kN);
		}
	}
	// at the end, squares modulo kN should be the same
	mpz_sub(aux_5, aux_4, aux_2);
	mpz_mod(aux_5, aux_5, this->kN);
	if (mpz_cmp_ui(aux_5,0)){
		CLog::LogError(WHEREARG,"Quadratic sieve","trying dependence",aK,100,"Unexpected error: squares are not the same modulo kN");
		printf("\n");
		mpz_mod(aux_1,aux_1,this->kN);
		gmp_printf("Left side root    mod kN: %Zd.\n",aux_1);
		mpz_mod(aux_2,aux_2,this->kN);
		gmp_printf("Left side square  mod kN: %Zd.\n",aux_2);
		mpz_mul(aux_5,aux_1,aux_1);
		mpz_mod(aux_5,aux_5,this->kN);
		gmp_printf("Left side root sq.mod kN: %Zd.\n",aux_5);
		mpz_mod(aux_3,aux_3,this->kN);
		gmp_printf("Right side root    mod kN: %Zd.\n",aux_3);
		mpz_mod(aux_4,aux_4,this->kN);
		gmp_printf("Right side square mod kN: %Zd.\n",aux_4);
		mpz_mul(aux_5,aux_3,aux_3);
		mpz_mod(aux_5,aux_5,this->kN);
		gmp_printf("Right side root sqmod kN: %Zd.\n",aux_5);
		ABNORMAL_EXIT(1);
		free(right_side);
		return BAD_ARGUMENT;
	}
//#define DEP_DEBUG	
#ifdef DEP_DEBUG
		printf("\n");
		mpz_mod(aux_1,aux_1,this->kN);
		gmp_printf("Left side root    mod kN: %Zd.\n",aux_1);
		mpz_mod(aux_2,aux_2,this->kN);
		gmp_printf("Left side square  mod kN: %Zd.\n",aux_2);
		mpz_mul(aux_5,aux_1,aux_1);
		mpz_mod(aux_5,aux_5,this->kN);
		gmp_printf("Left side root sq.mod kN: %Zd.\n",aux_5);
		mpz_mod(aux_3,aux_3,this->kN);
		gmp_printf("Right side root    mod kN: %Zd.\n",aux_3);
		mpz_mod(aux_4,aux_4,this->kN);
		gmp_printf("Right side square mod kN: %Zd.\n",aux_4);
		mpz_mul(aux_5,aux_3,aux_3);
		mpz_mod(aux_5,aux_5,this->kN);
		gmp_printf("Right side root sqmod kN: %Zd.\n",aux_5);
#endif	
	mpz_add(aux_5, aux_1, aux_3);
	mpz_gcd(this->factor_1, aux_5, this->kN);

#ifdef DEP_DEBUG
	gmp_printf("GCD(X+Y = %Zd,%Zd) = %Zd\n",aux_5,this->kN,this->factor_1);	
	printf("Small coefficient k = %d\n",this->k);
#endif
	// Now, we will try to remove all small factors of factor_1, especially those common with the multiplier this->k	
	mpz_set_ui(aux_5,this->k);
	mpz_gcd(aux_2,this->factor_1,aux_5);
#ifdef DEP_DEBUG
	gmp_printf("GCD(%Zd,%Zd) = %Zd.\n",this->factor_1,aux_5,aux_2);
#endif	
	if (mpz_cmp_ui(aux_2,1)){
		mpz_fdiv_q(this->factor_1,this->factor_1,aux_2);
	}
#ifdef DEP_DEBUG
	gmp_printf("After elimination of common factors with k: factor_1 is %Zd.\n",this->factor_1);
#endif	
	if (mpz_cmp(this->factor_1,this->N) > 0){
		gmp_printf("Strange. Factor_1 =%Zd is strictly greater than N = %Zd.\n",this->factor_1,this->N);
	}
	if (mpz_cmpabs_ui(this->factor_1,this->k) <= 0){
		mpz_set_ui(this->factor_1,0); // a trivial divisor (1 or k)
	}
	if (!mpz_cmpabs(this->factor_1,this->N)){
		mpz_set_ui(this->factor_1,0); // a trivial divisor (N)
	}
	if (!mpz_cmpabs(this->factor_1,this->kN)){
		mpz_set_ui(this->factor_1,0); // a trivial divisor (kN)
	}
	if (mpz_cmp_ui(this->factor_1,0)){	// we have found a nontrivial factor
		mpz_gcd(this->factor_1,this->factor_1,this->N);
		mpz_div(this->factor_2,this->N,this->factor_1);
		free(right_side);
		return ConstRC::Factorized;
	}
	mpz_sub(aux_5, aux_1, aux_3);
	mpz_abs(aux_5, aux_5);
	mpz_gcd(this->factor_1, aux_5, this->kN);
	if (mpz_cmpabs_ui(this->factor_1,this->k) <= 0){
		mpz_set_ui(this->factor_1,0); // a trivial divisor (1 or k)
	}
	if (!mpz_cmpabs(this->factor_1,this->N)){
		mpz_set_ui(this->factor_1,0); // a trivial divisor (N)
	}
	if (!mpz_cmpabs(this->factor_1,this->kN)){
		mpz_set_ui(this->factor_1,0); // a trivial divisor (kN)
	}
	if (mpz_cmp_ui(this->factor_1,0)){	// we have found a nontrivial factor
		mpz_gcd(this->factor_1,this->factor_1,this->N);
		mpz_div(this->factor_2,this->N,this->factor_1);
		free(right_side);
		return ConstRC::Factorized;
	}
	free(right_side);
	return ConstRC::NotFactorized;
}

/*!
	This method is called iff the factor search has been successful. It prints out the factors and exits.

*/
void QuadraticSieve::PrintFactors(){
	if (this->value_distribution_mode) {
		PrintDistributionInfo();
	}
	gmp_printf("\n\n");
	gmp_printf("Result of quadratic sieve factorization\n");
	gmp_printf("=======================================\n");
	gmp_printf("%Zd = %Zd * %Zd\n",this->N, this->factor_1, this->factor_2);
	gmp_printf("\n\n");
	if (this->frontend){
		gmp_fprintf(this->frontend_pipe,"\nPhaseFactors> 	%Zd = %Zd * %Zd\n",this->N,this->factor_1,this->factor_2);
		fflush(this->frontend_pipe);
	}
	if (iSerializator) {
		iSerializator->Serialize(EQsResult);
	}
}

/*!
	This method is called only if there has been no dependence giving proper factor of N. It prints out the corresponding
	message and exits.

*/
void QuadraticSieve::PrintApology(){
	gmp_printf("\n\n");
	gmp_printf("Result of quadratic sieve factorization\n");
	gmp_printf("=======================================\n");
	gmp_printf("We were not able to find any proper nontrivial factor.");
	gmp_printf("\n\n");
}

/*!
	An auxiliary method - substitution of missing method from GMP. "si" stands for "signed int".
*/
void QuadraticSieve::mpz_add_si(mpz_t aRop, mpz_t aOp1, long aOp2){
	if (aOp2 >=0){
		mpz_add_ui(aRop,aOp1,aOp2);
	} else {
		mpz_sub_ui(aRop,aOp1,abs((int) aOp2));
	}
}

/*!
An auxiliary method - substitution of missing method from GMP. "si" stands for "signed int".
*/
void QuadraticSieve::mpz_sub_si(mpz_t aRop, mpz_t aOp1, long aOp2){
	if (aOp2 >=0){
		mpz_sub_ui(aRop,aOp1,aOp2);
	} else {
		mpz_add_ui(aRop,aOp1,abs((int) aOp2));
	}
}
/*
    Functionality of built-in NEXKSB algorithm.
*/
/*!
	This method allocates memory and structures for NEXKSB algorithm. For details (including
	meaning of parameters k and n) see theoretic description of the algorithm.
	
	Return codes:
	- NOT_ENOUGH_MEMORY if any allocation has failed
	- ConstRC::Ok otherwise

*/
int  QuadraticSieve::NEXKSB_Init(nexksb_type n, nexksb_type k){
	if (this->nx_env == NULL){
		this->nx_env = (nexksb_environment*) AllocationMachine((void**) &this->nx_env,sizeof(nexksb_environment));	
		if (this->nx_env == NULL){
			return NOT_ENOUGH_MEMORY;
		}	
		this->nx_env->subset = NULL;
	} 
	this->nx_env->k = k;
	this->nx_env->n  = n;
	this->nx_env->subset = (nexksb_type*) AllocationMachine((void**) &this->nx_env->subset,(this->nx_env->k+1)*sizeof(nexksb_type));
	if (nx_env->subset == NULL) 
		return NOT_ENOUGH_MEMORY;
	memset(this->nx_env->subset,0, (this->nx_env->k+1)*sizeof(nexksb_type));
	this->nx_env->h = this->nx_env->k;
	this->nx_env->m = 0;
	this->nx_env->finished = 0;
	NEXKSB_Shift();
	return ConstRC::Ok;	
}

int   QuadraticSieve::NEXKSB_Set(nexksb_type n, nexksb_type k, nexksb_type h, nexksb_type m, nexksb_type* subset, nexksb_type finished)
{
	if (this->nx_env == NULL)
	{
		NEXKSB_Init(n,k);
	}
	this->nx_env->n = n;
	this->nx_env->k = k;
	this->nx_env->h = h;
	this->nx_env->m = m;
	if (this->nx_env->subset)
	{
		free(this->nx_env->subset);
		this->nx_env->subset = NULL;
	}
	this->nx_env->subset = subset;
	this->nx_env->finished = finished;
	return ConstRC::Ok;
}

nexksb_environment* QuadraticSieve::NEXKSB_Get() const
{
	return this->nx_env;
}

/*!
	By invocation of this method, one iteration of NEXKSB algorithm is performed. It is used
	- to initiate the NEXKSB algorithm structures in Init
	- to yield a new k-subset from within NEXKSB_Next()

*/
void QuadraticSieve::NEXKSB_Shift(){
	nexksb_type j, max = this->nx_env->h;
	for (j = 1; j <= max; j++){
		this->nx_env->subset[this->nx_env->k + j - this->nx_env->h] = this->nx_env->m + j;
	}
	if (this->nx_env->subset[1] == (this->nx_env->n - this->nx_env->k + 1)) {
		this->nx_env->finished = 1; // end of algorithm	
	}
}

/*!
	This method is a wrapper for NEXKSB_Shift(), performing nothing if the algorithm has finished.

*/
void QuadraticSieve::NEXKSB_Next(){
	if (!this->nx_env->finished){
		this->nx_env->subset[0]++; // we raise the counter by 1
		if (this->nx_env->m < (this->nx_env->n - this->nx_env->h)){
			this->nx_env->h = 0;	
		} 
		this->nx_env->h++;
		this->nx_env->m = this->nx_env->subset[this->nx_env->k + 1 - this->nx_env->h];
		NEXKSB_Shift();
	}
}


/*!
	This method releases memory used by NEXKSB structures. It is called by InitiateSIQSStructures,
	after the creation of even-indexed triples.

*/
void QuadraticSieve::NEXKSB_Delete(){
	this->nx_env->k = 0;
	this->nx_env->n = 0;
	this->nx_env->m = 0;
	this->nx_env->h = 0;
	this->nx_env->finished = 0;
	if (this->nx_env->subset){
		free(this->nx_env->subset);
		this->nx_env->subset = NULL;
	}
}

/*!
	This auxiliary method is used to create an instance of the relation struct and init
	its member variables, except for the large prime array. If any of the allocations
	failed, it returns NULL.

*/
qs_relation* QuadraticSieve::InitRelation() {
	qs_relation* result = NULL;
	result = (qs_relation*) AllocationMachine((void**) &result,sizeof(qs_relation));
	if (result == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in InitRelation",sizeof(qs_relation),40,"There was an error in InitRelation.");		
		return NULL;
	}
	mpz_init(result->argument);
        mpz_init(result->factorized_value);
        mpz_init(result->remaining);
        mpz_init(result->left_side_value);
        mpz_init(result->right_side_value);
	result->lprime1 = 1;
	result->lprime2 = 1;
        result->inited = (ARG_INITED | FAC_INITED | REM_INITED | LSV_INITED | RSV_INITED );
        result->index_of_used_polynomial = NO_INDEX;
        result->maximal_sieving_element_index = -1;    
	//printf("Div allocate: %d.\n\n\n\n",div_allocate);
        int max_factor = 2*div_allocate-1;
        result->currently_assigned_sieving_elements = max_factor+1;
        result->elements = NULL;
        result->elements = (sieving_element*) AllocationMachine((void **) &(result->elements),result->currently_assigned_sieving_elements*sizeof(sieving_element));    
        if (result->elements == NULL) {
		mpz_clear(result->argument);
	        mpz_clear(result->factorized_value);
	        mpz_clear(result->remaining);
        	mpz_clear(result->left_side_value);
	        mpz_clear(result->right_side_value);
		CLog::LogError(WHEREARG,"Quadratic sieve","in InitRelation",result->currently_assigned_sieving_elements*sizeof(sieving_element),40,"There was an error in InitRelation.");		
         	return NULL;
        }
	result->lps = NULL;
	result->max_lp_index  = -1;
	result->assigned_lps  =  0;
	result->components = 1;
	return result;
}

/*!
	This method resets the indicated relation to a relation having space for large
	prime array - therefore ready to accept combinations with other partial relations.
	We call such relation "a partial cycle".
	
	Return values: - CALLED_TOO_EARLY if relation is NULL or not ready,
		       - NOT_ENOUGH_MEMORY if allocation of space for large primes failed,
		       - ConstRC::Ok if everything went well.

*/
int  QuadraticSieve::RelationToPartialCycle(qs_relation* aRelation) {
	if (aRelation == NULL) {
		return CALLED_TOO_EARLY;
	}
	int expected = (ARG_INITED | FAC_INITED | REM_INITED | LSV_INITED | RSV_INITED );
	if ((expected & (aRelation->inited)) != expected) {
		return CALLED_TOO_EARLY;
	}
	mpz_set_ui(aRelation->left_side_value,1);
	mpz_set_ui(aRelation->right_side_value,1);
	aRelation->lprime1 = 1;
	aRelation->lprime2 = 1;
	int allocate = DEFAULT_LPS_ARRAY_ALLOCATE_SINGLE;
	if (this->using_variations == EDoubleLargePrimeV) {
		allocate = DEFAULT_LPS_ARRAY_ALLOCATE_DOUBLE;
	}
	aRelation->lps = (lp_element*) AllocationMachine((void**) &aRelation->lps,allocate*sizeof(lp_element));
	if (aRelation->lps == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in RelationToPartialCycle",allocate*sizeof(lp_element),40,"There was an error in RelationToPC.");		
		return NOT_ENOUGH_MEMORY;
	}
	aRelation->assigned_lps = allocate;
	aRelation->max_lp_index = -1;
	aRelation->components = 0;
	return ConstRC::Ok;
}

/*!
	This quicksorts the divisors in the indicated relation.
*/
void QuadraticSieve::SortRelation(qs_relation* aRelation) {
	if (aRelation == NULL) return;
	sieving_element* elements = aRelation->elements;
	if (elements == NULL) return;
	qsort(elements,aRelation->maximal_sieving_element_index+1,sizeof(sieving_element),&CompareDivisors);
}

/*!
	This is a destructor for the indicated relation, trying to free all memory
	used by its fields.

*/
void QuadraticSieve::ClearRelation(qs_relation* aRelation) {
	if (aRelation == NULL) return;
	if (aRelation->inited & ARG_INITED){
		mpz_clear(aRelation->argument);
		aRelation->inited &= (~ARG_INITED);
	}
	if (aRelation->inited & FAC_INITED){
		mpz_clear(aRelation->factorized_value);
		aRelation->inited &= (~FAC_INITED);
	}
	if (aRelation->inited & REM_INITED){
		mpz_clear(aRelation->remaining);
		aRelation->inited &= (~REM_INITED);
	}
	if (aRelation->inited & LSV_INITED){
		mpz_clear(aRelation->left_side_value);
		aRelation->inited &= (~LSV_INITED);
	}
	if (aRelation->inited & RSV_INITED){
		mpz_clear(aRelation->right_side_value);
		aRelation->inited &= (~RSV_INITED);
	}
	if (aRelation->currently_assigned_sieving_elements){
		free(aRelation->elements);
		aRelation->elements = NULL;
		aRelation->maximal_sieving_element_index = -1;
		aRelation->currently_assigned_sieving_elements = 0;
	}
	if (aRelation->lps) {
		//printf("Clearing lps, address = %p, max_lp_index = %d, assigned_lps = %d.\n\n",aRelation->lps,aRelation->max_lp_index,aRelation->assigned_lps);
		free(aRelation->lps);
		aRelation->lps = NULL;
		aRelation->max_lp_index = -1;
		aRelation->assigned_lps = 0;
	}
}

/*!
	A method which walks through the whole hashtable 1 and searches for
	entries equal to their own ancestor - that is, roots of the components.
	This method saves the roots into array this->root_list and returns number
	of roots found, or NOT_ENOUGH_MEMORY, if the allocation of root_list failed.

*/
int  QuadraticSieve::CountRoots() {
	int i, roots = 0;
	int max = cycle_table_1->AllocatedSize-1;
	for (i=0; i<= max; i++) {
		if (cycle_table_1->Hashtable[i].GetAncestor() == i) {
			roots++;
			this->rl_max_index++;
			if (this->rl_max_index == rl_allocated) {
				// We must reallocate the array
				this->root_list = (hashtable_entry_type_1*) AllocationMachine((void**) &this->root_list,(this->rl_allocated+DEFAULT_ROOT_LIST)*sizeof(hashtable_entry_type_1));
				if (this->root_list == NULL) {
					// Error Message
					return NOT_ENOUGH_MEMORY;
				}
				this->rl_allocated += DEFAULT_ROOT_LIST;
			}
			this->root_list[this->rl_max_index].value = cycle_table_1->Hashtable[i].GetValue();	
			this->root_list[this->rl_max_index].ancestor = i;
		}
	}
	return roots;
}
/*!
	This method is called by the SortSmoothValuesLogarithmic method, after a
	x-partial value has been found. It performs insertion of this value into the
	cycle counting hashtable (see Lenstra, Manasse) by finding the appropriate
	large primes, inserting the edge corresponding to the relation and joining
	the components.

*/
void QuadraticSieve::AddToCycleCountStructures(qs_relation* aRelation) {

	if (!factoring_structures_inited)
	{
		SetFactoringStructures();
	}
	long d_1  = 0;
	long d_2  = 0;
	long r_1  = 0;
	long r_2  = 0;

	bool present;
	bool enlarged;
	cycle_table_1->FindPrimeInHashtable(aRelation->lprime1, 0,present,d_1,enlarged);
	if (!present) {
		// This prime has not been in hashtable yet.					
		cycle_table_1->Hashtable[d_1].SetValue(aRelation->lprime1);
		cycle_table_1->Hashtable[d_1].SetAncestor(d_1);
		cycle_table_1->Added();
		this->vertices++;
		this->components++;
	}
	cycle_table_1->FindPrimeInHashtable(aRelation->lprime2,0,present,d_2,enlarged);
	if (!present) {
		// This prime has not been in hashtable yet.					
		cycle_table_1->Hashtable[d_2].SetValue(aRelation->lprime2);
		cycle_table_1->Hashtable[d_2].SetAncestor(d_2);
		cycle_table_1->Added();
		this->vertices++;
		this->components++;
	}
	// adding lprime1 or lprime2 could have resulted in rearranging,
	// thus we need to reset
	this->cycle_table_1->FindPrimeInHashtable(aRelation->lprime1,0,present,d_1,enlarged);
	this->edges++;
	// Finding roots of components
	r_1 = FindAndResetRoot(d_1);
	r_2 = FindAndResetRoot(d_2);
	if (r_1 != r_2) {	
		// The component with the larger prime at root is joined to the other one
		if (cycle_table_1->Hashtable[r_1].GetValue() > cycle_table_1->Hashtable[r_2].GetValue()) {
			cycle_table_1->Hashtable[r_1].SetAncestor(r_2);	
		} else {
			cycle_table_1->Hashtable[r_2].SetAncestor(r_1);	
		}
		this->components--;
	}
}

/*!
	A method called from AddToCycleCountStructures, whose aim is to find
	the root of a component including element at aIndex, and to reset all
	intermediate roots to a new, correct value (see Lenstra, Manasse). Returns
	the index of the root in hashtable 1.

*/
int QuadraticSieve::FindAndResetRoot(int aIndex) {
	int ancestor = cycle_table_1->Hashtable[aIndex].GetAncestor();
	int index = aIndex;
	int intermediate_index = 0;
	int root;
//	printf("Hledame predka pro tento prvek. Prvocislo %d, index = %d, index predka = %d\n",value,aIndex,ancestor);

	while (ancestor != index) {
		// Nekontrolujeme rozsah!!!
		intermediate_roots[intermediate_index] = index;
		intermediate_index++;
		index = ancestor;
		ancestor = cycle_table_1->Hashtable[index].GetAncestor();
		if (intermediate_index > this->max_ir_index) {
			printf("FindAndResetRoot: overflow.\n");
			return 0;
		}
		//printf("index = %d, ancestor = %d, iterace: %d\n",index,ancestor,intermediate_index);
	}
	root = index;
	int i, max = intermediate_index;
	for (i=0; i< max; i++) {
		cycle_table_1->Hashtable[intermediate_roots[i]].SetAncestor(root);
                //printf("Prenastavujeme koren indexu %d na %d.\n",intermediate_roots[i],root);
	}
	return root;
}

/*!
	This extensive method constructs the fundamental cycles in the graph of x-partial
	relations and creates the corresponding combined relations. To achieve this, the
	following operations are executed:
	
	Our first aim is to construct "images" of the cycles: that is, arrays of indices
	telling which edge adds up to which cycle, without actually performing the addition.
	This method of construction will later enable us to construct the actual cycles
	in "one pass", that is, in one reading of the partials from file; this advantage
	is somewhat compensated by larger memory consumption. 
	
	The "images" of cycles are constructed in the following way: the file of usable
	partials is iteratively re-read, and some of its entries are used to fill in a
	hashtable, as described in (Lenstra, Manasse); others are written back, so that
	they could be used in the next iteration. So, the process begins by reading relatively
	large file of all usable partials, and writing a smaller file of usable, but so far
	unused partials; then, the file of unused partials is considered a source of partials,
	and the process is repeated again and again, until expected amount of fundamental
	cycles has been constructed.
	
	Having finished the construction of "images", the process continues by reformatting
	them to a suitable form. At first, all the arrays of indices are sorted in ascending
	order; this will enable us to construct the cycles later by reading the file of
	usable partials only once. Then, all of the arrays are scanned for occurence of
	"duplicite paths". "Duplicite paths" emerge if the corresponding cycle does not
	go through the root of the appropriate component, for example, if two 2-partials
	complementing each other have been found. This is an extremely rare situation
	(currently seen only once in 10 000 cycles), which manifests itself in the corresponding
	array of indices by a double entry (that is, one x-partial is added twice into the
	cycle). Such double entry is useless, as it contributes to the resulting combined
	relation by vector of binary zeros. However, its inclusion means performing of some
	multiple precision operations, and so it is advantageous to remove double entries
	entirely. This is done by the RemoveDuplicitePaths() method.
	
	Finally, the "images" of cycles, cleaned of duplicite paths and sorted in ascending
	order, are ready for processing. The file of usable x-partial relations is opened for
	reading once more and the x-partial relations are read one-by-one. As each relation
	is read, all of the constructed cycles are checked for its occurence; sorting of
	indices in "images" makes sure that each relation has to be checked for only once.
	If a cycle requiring this relation is found, the relation is combined with the 
	corresponding partial cycle to another partial relation.
	
	Finally, the constructed cycles are subject to process of "smoothing", where the
	collected squares of large primes are included into the right side value of the
	corresponding relation. If an error has occured in the previous cycle construction,
	it is usually detected in this phase of processing.
	
	If info mode is enabled, this method also calls FindCycleSizes() in order to inform
	the user about sizes of cycles found. 
	
	Return codes: - NOT_ENOUGH_MEMORY if any of numerous allocations fails.
		      - COULD_NOT_OPEN_FILE in case of problems with file reading.
		      - anything thrown by called methods.
		      - ConstRC::Ok if everything went well.

*/
int  QuadraticSieve::ConstructCycles() {
	if (this->info_mode) {
		printf("   ConstructCycles: ");
	}
	int combinationResult = ConstRC::Ok;
	int index = this->so_far_smooths_found;
	int result = ConstRC::Ok;
	qs_relation* candidate = InitRelation();	
	// We create new hashtable in order to insert values again. The number of vertices
	// has been recounted in RereadRelations
	this->cycle_table_2 = new lp_hashtable_common<CRelationHashEntry>();	
	this->cycle_table_2->SetupHashtable(this->vertices);
	ResetHashtable_2();
	long i, d_1, d_2;
	int constructed_cycles = 0;	
	int iteration = 0;
	int used = 0;
	int deferred = 0;
	int current_relation = 0;
	int current_index    = 0;
	CRelationHashEntry* last_entry = NULL;
	bool found_1, found_2, depth_1, depth_2;
	bool print = false;
	qs_relation** partial_cycles = NULL;

	// We put all roots found before into the hashtable
	for (i=0; i<=this->rl_max_index; i++) {
		bool enlarged;
		cycle_table_2->FindPrimeInHashtable(this->root_list[i].value,0,found_1,d_1,enlarged);		
		this->cycle_table_2->Hashtable[d_1].SetValue(this->root_list[i].value);
		this->cycle_table_2->Hashtable[d_1].SetRootFingerprint(0);
		this->cycle_table_2->Hashtable[d_1].SetAncestor(d_1);
		this->cycle_table_2->Hashtable[d_1].SetRelation(ConstHT::ROOT_FLAG);
		this->cycle_table_2->Added();
	}
	int max = this->independent_cycles-1;
        cycle_construction_entry* cycle_array = NULL;
	cycle_array = (cycle_construction_entry*) AllocationMachine((void**) &cycle_array,(this->independent_cycles)*sizeof(cycle_construction_entry));
	if (cycle_array == NULL) {
		// Error Message !!
		ClearRelation(candidate);
  	        free(candidate); candidate = NULL;
		return NOT_ENOUGH_MEMORY;
	}
	for (i=0; i<= max; i++) {
		cycle_array[i].array = NULL;
		cycle_array[i].array = (int*) AllocationMachine((void**) &cycle_array[i].array,DEFAULT_CYCLE_ARRAY_ENTRIES*sizeof(int));
		if (cycle_array[i].array == NULL) {
			// delete all before
			ClearRelation(candidate);
			return NOT_ENOUGH_MEMORY;
		}
		cycle_array[i].allocated = DEFAULT_CYCLE_ARRAY_ENTRIES;
		cycle_array[i].current_max_index = -1;
	}
	// We will copy the file containing all usable partials, named "partials", to file 
	// named "usable_partials". This file will be used once more in construction of the
	// "real" cycles.
	if (OpenFile(this->partial_values_file_name,READ) == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error opening partials file.");				
		result = COULD_NOT_OPEN_FILE;
		goto deallocate;
	}
	if (OpenFile(this->usable_partial_values_file_name,WRITE) == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error opening usable partials file.");				
		result = COULD_NOT_OPEN_FILE;
		goto deallocate;
	}
	while(ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
		PrintRelationToFile(this->usable_partials_file,candidate,candidate->index);
	}
	result = CloseFile(this->partial_values_file_name);
	if (result != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error closing partials file.");				
		goto deallocate;
	}
	result = CloseFile(this->usable_partial_values_file_name);
	if (result != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error closing	usable partials file.");				
		goto deallocate;
	}
	// Now, the fundamental iterative part begins. At each pass, entries are read from file and possibly added to the graph
	
	iteration = 1; // we consider putting roots into hashtable as the iteration #0
	
	while (constructed_cycles != this->independent_cycles) {		
		printf("\n Constructing cycles: starting iter %d, so far cycles %d ...\n",iteration,constructed_cycles); fflush(stdout);
		if (OpenFile(this->partial_values_file_name,READ) == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error opening partials file.");				
			break;
		}
		if (OpenFile(this->temporary_file_name,WRITE) == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error opening temporary file.");				
			break;
		}
		used = 0;
		deferred = 0;		
		while (ReadRelationFromFile(this->partials_file,candidate) == ConstRC::Ok) {
			print = false;
			// we read each relation, put the corresponding primes into the hashtable, etc.
			// We will check whether any of the corresponding primes is in the hashtable
			// at previous depth			
			depth_1 = false;
			depth_2 = false;
			bool enlarged;
			cycle_table_2->FindPrimeInHashtable(candidate->lprime1,0,found_1,d_1,enlarged);
			cycle_table_2->FindPrimeInHashtable(candidate->lprime2,0,found_2,d_2,enlarged);
			if (enlarged) cycle_table_2->FindPrimeInHashtable(candidate->lprime1,0,found_1,d_1,enlarged);
			// printf("Processing partial #%d.\n",candidate->index);
			// printf("Prvnim ze dvou prvocisel je %d. Found v tabulce = %d, d_1 = %d, relation = %d, bez parity mask %d. Informace o druhem prvocisle: hodnota %d, found v tabulce = %d, d_2 = %d, relation = %d, bez parity mask %d.\n",candidate->lprime1,found_1,d_1,cycle_table_2[d_1].relation,cycle_table_2[d_1].relation & (~PARITY_MASK),candidate->lprime2,found_2,d_2,cycle_table_2[d_2].relation,cycle_table_2[d_2].relation & (~PARITY_MASK));
			/*if ((candidate->lprime1 == 843371) || (candidate->lprime1 == 3503329) || (candidate->lprime1 == 1195849)) {
				print = true;
			}
			if ((candidate->lprime2 == 843371) || (candidate->lprime2 == 3503329) || (candidate->lprime2 == 1195849)) {
				printf("Processing partial #%d.\n",candidate->index);
				printf("Prvnim ze dvou prvocisel je %d. Found v tabulce = %d, d_1 = %d, relation = %d, bez parity mask %d. Informace o druhem prvocisle: hodnota %d, found v tabulce = %d, d_2 = %d, relation = %d, bez parity mask %d.\n",candidate->lprime2,found_2,d_2,cycle_table_2[d_2].relation,cycle_table_2[d_2].relation & (~PARITY_MASK),candidate->lprime1,found_1,d_1,cycle_table_2[d_1].relation,cycle_table_2[d_1].relation & (~PARITY_MASK));
				print = true;
			}*/
			if (found_1) {
				if ((cycle_table_2->Hashtable[d_1].GetRelation() & (ConstHT::PARITY_MASK_NFS)) == (unsigned int) (((iteration-1)%2) << 31)) {
					depth_1 = true;
				}  
			}
			if (found_2) {
				if ((cycle_table_2->Hashtable[d_2].GetRelation() & (ConstHT::PARITY_MASK_NFS)) == (unsigned int) (((iteration-1)%2) << 31)) {
					depth_2 = true;
				} 
			}
			// If the currently read relation has neither vertex in the current graph at suitable
			// depth, we defer that edge
			if (((!found_1) && (!found_2)) || ((!depth_1) && (!depth_2))) {
			   // either both are out          // or some of them is in, but one or both at wrong depth
			   	//if (print) printf("Deferred edge\n");
				PrintRelationToFile(this->temporary_file,candidate,candidate->index);
				deferred++;
				// go to next edge
				continue;
			}  
			used++;
			// precisely one of the values is present in the table
			if (found_1 ^ found_2) {
				if (found_1) {					
					//if (print) printf("Adding edge to table 1 (#%d).\n",candidate->index);
					cycle_table_2->Hashtable[d_2].SetValue(candidate->lprime2);
					cycle_table_2->Hashtable[d_2].SetRootFingerprint(0);
					cycle_table_2->Hashtable[d_2].SetAncestor(d_1);
					cycle_table_2->Hashtable[d_2].SetRelation(candidate->index);
					cycle_table_2->Added();
					if (iteration % 2) {
						cycle_table_2->Hashtable[d_2].SetRelation(cycle_table_2->Hashtable[d_2].GetRelation() | ConstHT::PARITY_MASK_NFS);
					}
				}
				if (found_2) {
					//if (print) printf("Adding edge to table 2 (#%d).\n",candidate->index);
					cycle_table_2->Hashtable[d_1].SetValue(candidate->lprime1);
					cycle_table_2->Hashtable[d_1].SetRootFingerprint(0);
					cycle_table_2->Hashtable[d_1].SetAncestor(d_2);
					cycle_table_2->Hashtable[d_1].SetRelation(candidate->index);
					cycle_table_2->Added();
					if (iteration % 2) {
						cycle_table_2->Hashtable[d_1].SetRelation(cycle_table_2->Hashtable[d_1].GetRelation() | ConstHT::PARITY_MASK_NFS);
					}
				}
			} else {
				// we have found a cycle
				// This is the edge which finishes the cycle
				PutToConstructionEntry(&cycle_array[constructed_cycles],candidate->index);
				// We construct the way back to the root (each root is marked with ConstHT::ROOT_FLAG as relation) from d_1 and d_2		if 	
				/*if (constructed_cycles == 15343) {
					printf("============ 15343 =========== \n");
					printf("Cycle finished by edge #%d. Indices of primes %d and %d in hashtable are %d and %d.\n",candidate->index,candidate->lprime1,candidate->lprime2,d_1,d_2);
				}*/
				last_entry = &cycle_table_2->Hashtable[d_1];
				while ((last_entry->GetRelation() & ConstHT::ROOT_FLAG) != ConstHT::ROOT_FLAG) {
					// by (& ConstHT::ROOT_FLAG) bit operation we mask out the parity bit
					PutToConstructionEntry(&cycle_array[constructed_cycles],last_entry->GetRelation() & ConstHT::ROOT_FLAG);					
					/*if (constructed_cycles == 15343) {
						printf("Way #1 to edge: #%d, whose ancestor is %d.\n",last_entry->relation & ConstHT::ROOT_FLAG,last_entry->ancestor);
					}	*/
					last_entry = &cycle_table_2->Hashtable[last_entry->GetAncestor()];
				}
				last_entry = &cycle_table_2->Hashtable[d_2];
				while ((last_entry->GetRelation() & ConstHT::ROOT_FLAG) != ConstHT::ROOT_FLAG) {
					// by (& ConstHT::ROOT_FLAG) bit operation we mask out the parity bit
					PutToConstructionEntry(&cycle_array[constructed_cycles],last_entry->GetRelation() & ConstHT::ROOT_FLAG);					
					last_entry = &cycle_table_2->Hashtable[last_entry->GetAncestor()];
					/*if (constructed_cycles == 15343) {
						printf("Way #2 to edge: #%d, whose ancestor is %d.\n",last_entry->relation & ConstHT::ROOT_FLAG,last_entry->ancestor);
					}*/	
				}
				/*if (constructed_cycles == 15343) {
					printf("====================\n");
				}*/
				constructed_cycles++;
			}
		}
		//printf(" end: used E %d, deferred E %d, now cycles %d.\n",used,deferred,constructed_cycles);
		
		if (this->info_mode) {
			printf(" %d ",constructed_cycles);
			fflush(stdout);
		}
		result = CloseFile(this->partial_values_file_name);
		if (result != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error closing partials file.");				
			break;
		}
		result = CloseFile(this->temporary_file_name);
		if (result != ConstRC::Ok) {
			CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error closing temporary file.");				
			break;	
       		}
		remove(this->partial_values_file_name.c_str());
		rename(this->temporary_file_name.c_str(),this->partial_values_file_name.c_str());
		iteration++;
	} 
	if (this->info_mode) {
		printf("(%d iterations)\n",iteration);

	}
	// Sort cycle definitions
	for (i=0; i<= max; i++) {
		qsort(cycle_array[i].array,cycle_array[i].current_max_index+1,sizeof(int),&CompareCCE);
		RemoveDuplicitePaths(&cycle_array[i]);
		/*if (i==15343) {
			int j, max1 = cycle_array[i].current_max_index;
			printf("Cycle %d has %d entries: ",i,max1+1);	
			for (j=0; j<=max1; j++) {
				printf("%d ",cycle_array[i].array[j]);
			}
			printf("\n");
		}*/
		cycle_array[i].running_index = 0;
	}
	if (this->info_mode) {
		FindCycleSizes(cycle_array,this->independent_cycles-1);
	}
	// Now, we actually construct the cycles according to the information gathered in previous code
	// At first, we create space for them
	partial_cycles = (qs_relation**) AllocationMachine((void**) &partial_cycles,(this->independent_cycles)*sizeof(qs_relation*));
	if (partial_cycles == NULL) {
		result = NOT_ENOUGH_MEMORY;
		goto deallocate;
	}	
	for (i=0; i<=max; i++) {
		partial_cycles[i] = NULL;
		cycle_array[i].running_index = 0;
	}
	for (i=0; i<=max; i++){
		partial_cycles[i] = InitRelation();
		if (partial_cycles[i] == NULL) {		
			goto deallocate;
		}
		result = RelationToPartialCycle(partial_cycles[i]);
		if (result != ConstRC::Ok) {
			goto deallocate;		
		}
		partial_cycles[i]->index = i;
	}
	// Now, we read the relations from file usable_partials one-by-one and check whether they can be used in each of the
	// generated cycles.
	if (OpenFile(this->usable_partial_values_file_name,READ) == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error opening usable partials file.");				
		result = COULD_NOT_OPEN_FILE;
		goto deallocate;
	}

	while(ReadRelationFromFile(this->usable_partials_file,candidate) == ConstRC::Ok) {
		current_relation = candidate->index;
		for (i=0; i<=max; i++) {
			if ((current_index = cycle_array[i].running_index) != (int) ConstHT::FINISHED) {
				if (current_relation == cycle_array[i].array[current_index]) {
					// We add the relation to the corresponding partial cycle
					/*if (i==15343) {
						printf("\n==========15343==========\n");
						printf("\nCycle: \n");						
						PrintRelation(partial_cycles[i]);
						printf("\nAdded_relation:\n");
						PrintRelation(candidate);
					}*/	
#if 0
					if ((debug_mode) && ((i>= 280) && (i<= 300)))
					{					
						printf("\nCycle before addition: \n");	
						PrintRelation(partial_cycles[i]);
						printf("\nAdded_relation:\n");
						PrintRelation(candidate);
					}
#endif
							
					int ret = CombineTwoRelations(partial_cycles[i],partial_cycles[i],candidate,CYCLE);
					if (ret != ConstRC::Ok)
					{
						combinationResult = ret;
						break;
					}
					partial_cycles[i]->components++;
#if 0
					if ((debug_mode) && ((i>= 280) && (i<= 300)))
					{
						printf("\nResult of addition: \n");		
						PrintRelation(partial_cycles[i]);
					}					
#endif
					/*if (i==15343) {						
						printf("\nResult: \n");						
						PrintRelation(partial_cycles[i]);
						
					}*/
					
					// and shift the index in definition
					if ((++(cycle_array[i].running_index)) > cycle_array[i].current_max_index) {
						cycle_array[i].running_index = ConstHT::FINISHED;
					}
				}
			}
		}
	}
	result = CloseFile(this->usable_partial_values_file_name);
	if (result != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error closing	usable partials file.");				
		goto deallocate;
	}
	if (combinationResult != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error during combination of relations.");				
		goto deallocate;
	}
	// We shift the cycles into the right sides
	for (i=0; i<= max; i++) {		
		result = SmoothenCycle(partial_cycles[i]);
		if (result != ConstRC::Ok) {
			goto deallocate;
		}
	}
	// Now, we add the cycles to the smooths file.
	result = CloseFile(this->smooth_values_file_name);
	if (result != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",result,40,"There was an error closing	smooths file.");				
		goto deallocate;
	}
	if (OpenFile(this->smooth_values_file_name,APPEND) == NULL) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",0,40,"There was an error opening smooths file.");				
		result = COULD_NOT_OPEN_FILE;
		goto deallocate;
	}	
	for (i=0; i<=max; i++){
		PrintRelationToFile(this->smooth_values_file,partial_cycles[i],index++);
	}
	result = CloseFile(this->smooth_values_file_name);
	if (result != ConstRC::Ok) {
		CLog::LogError(WHEREARG,"Quadratic sieve","in ConstructCycles",result,40,"There was an error closing	smooths file.");				
		goto deallocate;
	}

deallocate:
	if (result != ConstRC::Ok) {
		for (i=0; i<=max; i++){	
			if (partial_cycles[i] != NULL) {		
				ClearRelation(partial_cycles[i]);				
			}
		}
	}	
	// Deallocate used temporary structures
	if (partial_cycles != NULL) {
		for (i=0; i< this->independent_cycles; i++) {						
			ClearRelation(partial_cycles[i]);
			if (partial_cycles[i] != NULL) {
				free(partial_cycles[i]);
			}
			partial_cycles[i] = NULL;
		}
		free(partial_cycles);
		partial_cycles = NULL;
	}
	for (i=0; i<= max; i++) {
		if (cycle_array[i].array != NULL) {
			free(cycle_array[i].array);
			cycle_array[i].array = NULL;
			cycle_array[i].allocated = 0;
			cycle_array[i].current_max_index = -1;
		}
	}
	free(cycle_array);
	cycle_array = NULL;
	ClearRelation(candidate);
        free(candidate); candidate = NULL;
	this->iRelationProcessingFlags |= CYCLES_CONSTRUCTED_FLAG;
	return result;	
}

/*!
	This method takes care of cleaning the "images" of constructed fundamental cycles from double
	edges, whose occurence means that a duplicite path has emerged. For details, see documentation
	on ConstructCycles() method.

 * During construction of cycles, it might happen that the two paths constructing
 * the cycle (one going from the root to one of the primes, and the other from
 * the second prime back to the root) have some part in common. Edges which are
 * in common are duplicite and add nothing new to the cycle (their product is 0
 * mod 2 with regard to all exponents); on the other hand, they complicate
 * one-pass cycle construction. Therefore we remove them entirely here. 
 * This method requires the array of edge indices in aEntry to be sorted.
 * From the logic of the construction algorithm, an edge may be at most twice
 * in the array; therefore, we do not care about triple or quadruple occurrences.
 */
void QuadraticSieve::RemoveDuplicitePaths(cycle_construction_entry* aEntry) {
	if (aEntry == NULL) return;
	int i = 0, max = aEntry->current_max_index, j;
	while (1) {
		if ((i+1) > max) break;
		if (aEntry->array[i] == aEntry->array[i+1]) {
			// We skip both of these entries
			j = i+2;
			while (j<= max) {
				aEntry->array[j-2] = aEntry->array[j];
				j++;
			}
			max = (aEntry->current_max_index -= 2);
		} else {
			i++;
		}
	}
}

/*!
	This method iterates through the constructed cycles and shifts their aggregated squares
	of large primes to the right side value. At this step, an inconsistency of a generated cycle
	might be detected; in such situation, the method aborts by returning GENERAL_ERROR code.
	
	Return codes: ConstRC::Ok if everything went well, GENERAL_ERROR in case of inconsistency.

*/
int  QuadraticSieve::SmoothenCycle(qs_relation* aElement) {
	if (aElement == NULL) 			{ return CALLED_TOO_EARLY; }
	if (!(aElement->inited & RSV_INITED))    { return CALLED_TOO_EARLY; }
	if (aElement->lps == NULL)              { return CALLED_TOO_EARLY; }
	int i,j,max=aElement->max_lp_index,exponent;
	for (i=0; i<=max; i++) {
		exponent = aElement->lps[i].exponent;
		if (exponent%2) {
			CLog::LogError(WHEREARG,"Quadratic sieve","in SmoothenCycle",i,0,"There was an error in generated cycle.");		
			PrintRelation(aElement);
			return GENERAL_ERROR;
		}
		exponent /= 2;
		for (j=0; j< exponent; j++) {
			mpz_mul_ui(aElement->right_side_value,aElement->right_side_value,aElement->lps[i].prime);
		}		
	}
	mpz_mod(aElement->right_side_value,aElement->right_side_value,this->kN);
	free(aElement->lps);
	aElement->lps = NULL;
	aElement->max_lp_index = -1;
	aElement->assigned_lps = 0;
        int res1 = AssertRelationConsistency(aElement,"Problem",__LINE__);
	if (res1 != ConstRC::Ok)
	{
		ABNORMAL_EXIT(1);
	}
	return ConstRC::Ok;
}

/*!
	This is an auxiliary method used by ConstructCycles() during creation of cycle "images".
	Returns NOT_ENOUGH_MEMORY or ConstRC::Ok, depending on whether the inner allocation failed.

*/
int  QuadraticSieve::PutToConstructionEntry(cycle_construction_entry *aCycle,int aRelationIndex) {
	aCycle->current_max_index++;
	if (aCycle->current_max_index == aCycle->allocated) {
		// We need to reallocate the array for entries
		aCycle->array = (int*) AllocationMachine((void**) &aCycle->array,(aCycle->allocated + DEFAULT_CYCLE_ARRAY_ENTRIES)*sizeof(int));
		if (aCycle->array == NULL) {
			CLog::LogError(WHEREARG,"Quadratic sieve","in PutToConstructionEntry",aCycle->allocated+DEFAULT_CYCLE_ARRAY_ENTRIES,40,"There was an error reallocating array.");				
			aCycle->allocated = 0;
			aCycle->current_max_index = -1;
			return NOT_ENOUGH_MEMORY;
		} else {
			aCycle->allocated += DEFAULT_CYCLE_ARRAY_ENTRIES;
		}
	}
	// Now, the allocated space suffices
	aCycle->array[aCycle->current_max_index] = aRelationIndex;
	return ConstRC::Ok;
}

/*!
	This is an auxiliary method for printing out current properties of the x-partial graph. It is
	called once before and once after removal of large singletons, to show the user how the removal
	of singletons "clears" the graph.

*/
void QuadraticSieve::PrintGraphInformation() {
	printf("   Graph properties: Vertices %d, Components %d, Edges %d, IndCycles %d.\n",this->vertices,this->components,this->edges,this->edges+this->components-this->vertices);
}

/*!
	An auxiliary method used by ConstructCycles() to inform the user about sizes of constructed
	fundamental cycles (by "size" we mean number of edges which take part in the cycle after
	removal of duplicite path).

*/
void QuadraticSieve::FindCycleSizes(cycle_construction_entry *aCycles, int aMaxIndex) {
	int i;
	int max = 0;
	for (i=0; i<= aMaxIndex; i++) {
		if ((aCycles[i].current_max_index+1) > max) {
			max = aCycles[i].current_max_index+1;
		}
	}
	int* sizes = NULL;
	sizes = (int*) AllocationMachine((void**) &sizes,(max+1)*sizeof(int));
	for (i=0; i<= max; i++) {
		sizes[i] = 0;
	}
	for (i=0; i<= aMaxIndex; i++) {
		sizes[aCycles[i].current_max_index+1]++;
	}
	printf("   Cycle sizes:\n");
	for (i=0; i<=max; i++) {
		if (sizes[i] != 0) {
			printf("     %4d: %7d.\n",i,sizes[i]);
		}
	}
	free(sizes);
	sizes = NULL;
}

// This function is here because Win32 executables do not like mpz_out_str with
// statically linked libc. (In my opinion: not even with dynamically linked,
// at least I had no success).
void QuadraticSieve::PrintMpzToFile(FILE* aFile, int base, mpz_t aValue) {		
		this->auxCharForMpz = mpz_get_str(this->auxCharForMpz,10,aValue);		
		fprintf(aFile,"%s",this->auxCharForMpz);
}

// Info on profiling with MSVC:
// http://www.tc.cornell.edu/CTC-Main/Services/Education/Topics/Optimization/Profiling/2.+MS+Command+Line.htm

int  QuadraticSieve::CopyADivisors(main_sieving_type* aArray) const
{
	if ((this->QSMode == EMpqs) || (this->QSMode == EOldMpqs))
	{
		return 0;
	}
	int i, max = this->s;
	for (i=0; i< max; i++)
	{
		aArray[i] = divisors_of_a[i];
	}
	return this->s;
}


float QuadraticSieve::GetSecondsInSieving()
{
	//printf("Finish sieving clock: %ld, start sieving clock: %ld.\n",iFinishSievingClock ,iStartSievingClock);
	float diff =(float) (iFinishSievingClock - iStartSievingClock);
	diff /= CLOCKS_PER_SEC;
	return diff;
}

Batch	QuadraticSieve::CreateNewJobs(unsigned int aBatchSize)
{
	// To create a new batch of jobs, we need:
	// - correctly set parameters like N, M, Factor Base
	if ((!mpz_cmp_ui(this->N,0)))
	{
		throw new AlgorithmException("Creating a batch of new jobs: N not specified",__FILE__,__LINE__);
	}
	if ((!mpz_cmp_ui(this->kN,0)))
	{
		// We must determine 'k', 'kN'.
		CalculateKRoot();
	}
	if (!integer_M)
	{
		throw new AlgorithmException("Creating a batch of new jobs: M not specified",__FILE__,__LINE__);
	}
	if (!factor_base)
	{
		throw new AlgorithmException("Creating a batch of new jobs: factor base NULL",__FILE__,__LINE__);
	}
	Batch batch;	
	int indexStart = iLowerMachineBoundIndex;
	int indexEnd   = iUpperMachineBoundIndex;
	main_sieving_type smallestMSD = factor_base[indexStart];	
	main_sieving_type greatestMSD = factor_base[indexEnd];

#define PRINT_ALOUD
#ifdef PRINT_ALOUD
	printf("QuadraticSieve::CreateNewJobs.\nIndexStart = %d, prime = %d\nIndexEnd = %d, prime = %d\n",indexStart, smallestMSD, indexEnd, greatestMSD);
#endif	

	int amount     = machine_specific_divisors_number;	
	if (amount <= 0)
	{
		// No jobs for parallel processing.
		return batch;
	}
	int currentSize = iJobsForNodes.size();
	if (currentSize > 0)
	{
		const QSParameters& params = iJobsForNodes[currentSize-1];
		int amountLast = params.Machine_Specific_Divisors_Count();
		if (amount != amountLast)
		{
			throw new AlgorithmException("Creating a batch of new jobs: mismatch of MSD size",__FILE__,__LINE__,amount,amountLast);
		}
		main_sieving_type lastPrime = params.Get_MSD(amount-1);
		// We will try to find the index of the lastPrime in FB
		int index = GetTrueIndexInFB(&lastPrime);
		if (index == -1)
		{
			throw new AlgorithmException("Could not find the last MSD prime in the FB",__FILE__,__LINE__,lastPrime,0);
		} else
		{
			indexStart = index + 1;
		}		
	}		
#ifdef PRINT_ALOUD
	printf("QuadraticSieve::CreateNewJobs. Creating %d jobs now.\n",aBatchSize);
	printf("\t\tMachine specific generation mode: %d \n", this->Get_MachineSpecificGeneration());
#endif	
	// generate aBatchSize different combinations of machine specific
	// divisors, if possible
	
	int i, jar[3];
	jar[0]=indexStart-3;//only last 'amount' members of jar are used
	jar[1]=jar[0]+1;
	jar[2]=jar[1]+1;
	for (i=0; i < aBatchSize; i++)
	{

#ifdef PRINT_ALOUD
		string message = "   Created a new set of MSDs: ";
#endif
		// Now we know that we have at least "amount" primes left in the MSD
		// interval
		// We can create a new job info
		QSParameters* params = new QSParameters(*this);		
		params->Set_Machine_Specific_Divisors_Given(TRUE);			
		for (int j=1; j<= amount; j++)
		{
			params->Add_MSD(factor_base[jar[3-j]]);
#ifdef PRINT_ALOUD
			char   number[20];
			memset(number,0,20);
			sprintf(number,"%d ",factor_base[jar[3-j]]);
			message += number;
#endif
		}		
#ifdef PRINT_ALOUD
		message += "\t\t(total MSDs: ";
		char   number[20];
		memset(number,0,20);
		sprintf(number,"%d",params->Machine_Specific_Divisors_Count());
		message += number;
		message += ")";
#endif
		params->Set_MSD_Interval_Given(TRUE);
		params->Set_ThresholdOptimizer_Level(iOptimizationPolicy);
		//params->Machine_Specific_Divisors_Count(amount);
		params->Set_MSD_Lower_Bound(smallestMSD);
		params->Set_MSD_Upper_Bound(greatestMSD);
		params->Set_DistributionMachineType(ENode);
		params->Set_InfoMode(GetInfoMode());
		params->Set_DebugMode(GetDebugMode());

		params->Set_UsingVariations(Get_UsingVariations());
#ifdef PRINT_ALOUD
		printf("%s\n",message.c_str());
#endif
		iJobsForNodes.push_back(*params);		
		batch.push_back(new QSParameters(*params));

		switch (this->Get_MachineSpecificGeneration())
		{
		case MSGOneDiffer:
		case MSGTwoDiffer:

		int Usable;
		do {
			if (jar[2]<indexEnd) {
				jar[2]++;
			} else 
			if ((jar[1]<indexEnd-1)&&(amount>1)) {
				jar[1]++;
				jar[2]=jar[1]+1;
			} else 
			if ((jar[0]<indexEnd-2)&&(amount>2)) {
				jar[0]++;
				jar[1]=jar[0]+1;
				jar[2]=jar[1]+1;	
			} else
			{
				// out of machine specific divisors, return
				return batch;
			};
			// In case we allow only MachineSpecificDivisor sets
			// with at least two non-common divisors, we need to
			// check
			Usable =2;
			if (MSGTwoDiffer == this->Get_MachineSpecificGeneration())
			for(int tempi=0;(tempi<iJobsForNodes.size())&&(Usable>0);tempi++)
				{
					Usable =2;
					for(int tempj=0;tempj<amount;tempj++)
						for(int tempk=0;tempk<amount;tempk++)
							if (iJobsForNodes[tempi].Get_MSD(tempj)==factor_base[jar[tempk]])
								Usable--;
				};
		} while ((MSGTwoDiffer==this->Get_MachineSpecificGeneration())&&
				(Usable<=0));
		break;
		case MSGThreeDiffer:
			for (int tempi=0;tempi<3;tempi++)
				jar[tempi]+=amount;
			if (jar[2]>=indexEnd) return batch;
		}	
	}
	return batch;
}



int QuadraticSieve::SetupParameters(const JobParameters& aParameters)
{
 	try
  	{
		const QSParameters& qp = dynamic_cast<const QSParameters&>(aParameters);
		SetupParameters(qp);
	} 
	catch (std::bad_cast e)
  	{    		
		CLog::LogError(WHEREARG,"QuadraticSieve","SetupParameters (JobParameters)",0,0,"Unknown subtype of JobParameters, doing nothing");
		return GENERAL_ERROR;
  	}	
}

int QuadraticSieve::SetupParameters(const QSParameters& aParameters)
{	
	// Must be first, for loading from job.xml to work
	Set_DirectoryName(aParameters.Get_DirectoryName().c_str());		
	if (iSerializator)
	{
		iSerializator->Set_DirectoryName(aParameters.Get_DirectoryName().c_str());
	}
	SetDataCompressionLevel(aParameters.Get_CompressionLevel());	
	int loadFlags = aParameters.Get_LoadFlags();
	printf("Setup Parameters: load flags %x\n",loadFlags);
	if (loadFlags & JOB_LOAD_FLAG)
	{
		// Load data from job.xml
		if (iSerializator)
		{
			QSParameters& editableParams = const_cast<QSParameters&>(aParameters);
			int ret = iSerializator->DeserializeJob(editableParams);
			if (ret != ConstRC::Ok)
			{
				return ret;
			}
		}
	} 
	if (loadFlags & LIST_LOAD_FLAG)
	{
		return GENERAL_ERROR;
	}
	Set_QSMode(aParameters.Get_QSMode());
	
	mpz_t x1;
	mpz_init(x1);
			
	if (aParameters.Get_FrontendPipe().length() > 0)
		SetFrontendPipe(aParameters.Get_FrontendPipe());
	Set_MachineSpecificGeneration(aParameters.Get_MachineSpecificGeneration());

	aParameters.Get_Modulus(x1);		
	Set_N(x1);		
		
	mpz_set_ui(x1,aParameters.Get_SievingIntervalLength());
	Set_M(x1);
	mpz_clear(x1);

	Set_UpperFBBound(aParameters.Get_UpperIntBound());
	Set_DistributionMachineType(aParameters.Get_DistributionMachineType());	
	Set_CommitInterval(aParameters.Get_CommitInterval());
	SetAssertionMode(aParameters.Is_AssertionMode());	
	SetInfoMode(aParameters.Is_InfoMode());
	SetDebugMode(aParameters.Is_DebugMode());
	Set_CheckMode(aParameters.Get_CheckMode());
	Set_LTMode(aParameters.Is_LtMode());
	Set_SievingIntervalMode(aParameters.Get_SievingIntervalMode());
	Set_CheckLevel(aParameters.Get_CheckLevel());
	Set_MemblockSize(aParameters.Get_MemblockSize());
	Set_TimeMessageInterval(aParameters.Get_TimeMessageInterval());
	Set_DIMode(aParameters.Is_DiMode());
	SetMatrixType(aParameters.Get_LinalgType());
	Set_WPList(aParameters.Is_WpList());
	Set_UsingVariations(aParameters.Get_UsingVariations());
	SetRankCalculationMode(aParameters.Is_RankCalculationMode());
	Set_ValDistMode(aParameters.Is_ValueDistributionMode());
	Set_PrintingSort(aParameters.Get_PrintSort());
	Set_ExtK(aParameters.Is_ExtendedK());
	Set_SingletonMode(aParameters.Get_SingletonMode());
	Set_PowerMode(aParameters.Get_PowerMode());
	Set_FactoringAlgorithm(aParameters.Get_FactorAlg());
	Set_NrPolynomials(aParameters.Get_MaxIter());
	Set_UseRootShort(aParameters.Is_RootShort());	
	if (iSerializator)
	{
		iSerializator->Serialize(EQsJob);
	}
	SetOptimizationPolicy(aParameters.Get_ThresholdOptimizer_Level());
	this->iLoadFlags = loadFlags;
	return InitEnvironment();
}

int QuadraticSieve::DeleteParameters(const JobParameters *aParameters) {

	try {
                const QSParameters *qp = dynamic_cast<const QSParameters *>(aParameters);
                delete qp;
        }
        catch (const std::bad_cast & e) {
                CLog::LogError(WHEREARG,"QuadraticSieve","DeleteParameters",0,0,"Unknown subtype of JobParameters, doing nothing.");
                return ConstRC::GeneralError;
        }

        return ConstRC::Ok;
}

void	QuadraticSieve::RunNodeInstance()
{
	Set_TimeMessageInterval(10);
	Set_WPList(TRUE);
	Set_UsingVariations(EDoubleLargePrimeV);	
	SetInfoMode(TRUE);
	Set_DistributionMachineType(ENode);
	iNodeInstanceRunning = TRUE;
	pthread_mutex_init(&iDistributedMutex, NULL);
	iDistributedMutexInited = TRUE;
	RunMPQS();
}

bool	QuadraticSieve::NodeInstanceRunning() const
{
	return iNodeInstanceRunning;
}

bool	QuadraticSieve::HasDataToSend() const
{
	printf("iRelationsSinceLastDataTransmission: %d, already sent: %d\n",iRelationsSinceLastDataTransmission,iRelationsSent);
	return iRelationsSinceLastDataTransmission;
}

unsigned int QuadraticSieve::DataToSend() const
{
	return iRelationsSinceLastDataTransmission;
}

const char* QuadraticSieve::DataUnit() const
{
	return "relations";
}

void    QuadraticSieve::ClearFreshData()
{
	iRelationsSent+=iRelationsSinceLastDataTransmission;
	iRelationsSinceLastDataTransmission = 0;
}

void	QuadraticSieve::LockDataMutex()
{
	if (ENone != iDistributionMachineType)
	{
// 		printf("LockDataMutex invoked.\n");
		pthread_mutex_lock(&iDistributedMutex);
// 		printf("LockDataMutex successful.\n");
	}
}

void	QuadraticSieve::UnlockDataMutex()
{
	if (ENone != iDistributionMachineType)
	{
// 		printf("UnlockDataMutex invoked.\n");
		pthread_mutex_unlock(&iDistributedMutex);
// 		printf("UnlockDataMutex successful.\n");
	}
}


/// Getters and setters for distribution-related parameters
bool  	QuadraticSieve::Get_IsDistributed() const 
{ 
	return this->iDistributionMachineType;
}

distribution    QuadraticSieve::Get_DistributionMachineType() const 
{ 
	return this->iDistributionMachineType;
}

int		QuadraticSieve::Get_CommitInterval() const 
{ 
	return this->iCommitInterval;
}

void		QuadraticSieve::Set_DistributionMachineType(distribution aValue) 
{ 
	this->iDistributionMachineType = aValue;
}

void		QuadraticSieve::Set_CommitInterval(int aValue) 
{ 
	this->iCommitInterval = aValue;
}

const main_sieving_type* QuadraticSieve::Get_MachineSpecificDivisors() const
{
	return machine_specific_divisors;
}

unsigned int  	   QuadraticSieve::Get_MSDCount() const
{
	return machine_specific_divisors_number;
}

machine_specific_generation_type QuadraticSieve::Get_MachineSpecificGeneration() const
{
	return iMachineSpecificGeneration;
}


main_sieving_type  QuadraticSieve::Get_LowerMachineBound() const
{
	return this->iLowerMachineBound;
}

main_sieving_type  QuadraticSieve::Get_UpperMachineBound() const
{
	return this->iUpperMachineBound;
}

// A method that sets msd's from an external source (like job.xml), typical for nodes of a distributed computation.
void 	QuadraticSieve::Set_MachineSpecificDivisors(main_sieving_type* aDivisors, int aLength)
{
	int i, max = aLength;
	if (max > 3)
	{
		printf("Set Machine specific divisors: too many MSDs. %d.\n",aLength);
	}
	for (i=0; i< max; i++)
	{
		machine_specific_divisors[i] 		= aDivisors[i];
		machine_specific_divisor_indices[i]	= -1;
	}
	this->iMSDSpecifiedExternally = TRUE;
}

void	QuadraticSieve::Set_MSDCount(int aCount)
{
	machine_specific_divisors_number = aCount;
}

void QuadraticSieve::Set_MachineSpecificGeneration(machine_specific_generation_type aValue)
{
	iMachineSpecificGeneration = aValue;
}


void	QuadraticSieve::Set_LowerMachineBound(main_sieving_type aPrime)
{
	this->iLowerMachineBound = aPrime;
}

void	QuadraticSieve::Set_UpperMachineBound(main_sieving_type aPrime)
{
	this->iUpperMachineBound = aPrime;
}

void 	QuadraticSieve::WaitForNodeSetup()
{
	printf("Wait for node setup not finished yet.");
	/*this->iConnector = new FilesystemConnector(iSharedURL, iSharedURL, iMachineName);
	FilesystemMessage* message = new FilesystemMessage(iSharedURL,"Waiting",EReady, 0, 41, iMachineName, "center", "");
	iConnector->Send(message);
	delete message; // Frees the memory (does not delete the file)*/
}


int     QuadraticSieve::RunCenter()
{
	printf("QuadraticSieve:: RunCenter not ready.");
	/*mpz_set_ui(this->start_time_main_process,time(NULL));
	int ret = InitEnvironment();
		if (ret == PRIME_NUMBER) { return PRIME_NUMBER; }
		if (ret != ConstRC::Ok) { return ERROR_INIT_ENVIRONMENT; }

	// Creates connectors, then periodically checks them.
	AbstractConnector* cnt = new FilesystemConnector(iSharedURL, iSharedURL, iMachineName);
	RegisterCenterConnector(cnt);
	iCenterRunInterrupted = FALSE;
	this->iIsInterruptable = TRUE;
	int wakeCounter = 0;
	do 
	{
		// Defined as a macro in definitions.h
		SLEEP(1); // Sleep for 1 second		
		printf("\rCenter woke up ... %7d sec   ",wakeCounter);
		wakeCounter++;
		fflush(stdout);		
	} while (!iCenterRunInterrupted);*/

	/*if (iCenterRunInterrupted)
	{
		printf("Center run interrupted from an external source\n");
	}*/
	return ConstRC::Ok;
}

const char* QuadraticSieve::AlgorithmName() const
{
	return "MPQS/SIQS";
}

JobParameters* QuadraticSieve::CreateNewParameters() const
{
	return new QSParameters();
}

const char*	QuadraticSieve::Get_SharedURL() const
{
	return iSharedURL.c_str();
}

const char*	QuadraticSieve::Get_MachineName() const
{
	return iMachineName.c_str();
}

int 	QuadraticSieve::GetOptimizationPolicy() const
{
	return iOptimizationPolicy;
}

void	QuadraticSieve::SetOptimizationPolicy(int aValue)
{
	iOptimizationPolicy = aValue;
}

ThresholdOptimizer&	QuadraticSieve::GetOptimizer() const
{
	return *iOptimizer;
}

