#include "lattice_sieving_phase.h"

#include <cstring>
#include <math.h>
#include "const_parameter_names.h"
#include "const_element_flags.h"
#include "const_sieving_phase.h"
#include "adv_factor_base_helper.h"
#include "supporting_factor_alg_creator.h"
#include "abstract_rel_processing_phase.h"
#include "integer_class.h"
#include "log.h"
#include "nfs_utils.h"


#include "test_suite.h"

//#define OLD_FACTORIZATION
//#define OLD_MIDDLEPRIMES_FACTORIZATION
//#define FACTOR_DEBUG

using namespace std;

typedef map<main_sieving_type, int>::const_iterator SpecialPrimeIt;

CLatticeSievingPhase::CLatticeSievingPhase() : MSerializable(ConstFN::SIEVING_PHASE), ASievingPhase()
{
	Set_AutoSerializable(ConstSPhase::AUTO_SERIALIZABLE);
	Set_InnerState((int)ESPhaseNotStarted);

	this->iTimes = new CTimeContainer(ConstSPhase::INFO_TIME_COUNT);
	this->iStats = new CStatContainer(ConstSPhase::INFO_STAT_COUNT);

	this->iRelationBuffer = NULL;
	this->iRelationBufferMaxCount = 0;
	this->iRelationWorkingBuffer = NULL;
	this->iRelationWorkingBufferMaxCount = 0;

	this->iRelation = NULL;

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iSievingBlocks[i] = NULL;
		this->iInitBlockValues[i] = 0;
		this->iInitBlockValueChange[i] = 0;
		this->iRSerialInitBlockValueChange[i] = 0;
		this->iLargishPrimesThreshold[i] = 0;
#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeSievingMiddle[i] = 0;
		this->iTimeSievingMiddleStart[i] = 0;
		this->iTimeSievingLargish[i] = 0;
		this->iTimeSievingLargishStart[i] = 0;
		this->iTimeFactorMiddle[i] = 0;
		this->iTimeFactorMiddleStart[i] = 0;
		this->iTimeFactorLargish[i] = 0;
		this->iTimeFactorLargishStart[i] = 0;
		this->iTimeFactorDivideSpecial[i] = 0;
		this->iTimeFactorDivideSpecialStart[i] = 0;
#endif
	}

	this->iUseCyclesForRelConstruction = false;

	this->iUseHighVariations = false;

	this->iLPHashtable = NULL;

	this->iHighVarHashtable = NULL;

	mpz_init(this->iAuxMpz1);
	mpz_init(this->iAuxMpz2);

	this->iIteration = 0;
	this->iWorkingIteration = 0;

	this->iRSerialRootIndex = 0;
	this->iRSerialSpecialQ = 0;

	this->iSpecialQ = 0;
	this->iRootIndex = 0;
	this->iSpecialR = 0;

	this->iLinesInBlock = 0;
	this->iLinesInBlockSet = 0;

	this->iBlockSetSize = 0;
	this->iBlockCount = 0;

	this->iFactorBlockShift = 0;
	this->iFactorBlockMask = 0;
	this->iFactorBlockSize = 0;
	this->iFactorBlock = NULL;

	this->iRSerialAccordingIters = false;
	this->iRSerialAccordingTime = false;
	this->iRSerialAccordingRelations = false;

	int rc;
	if ((rc = pthread_mutex_init(&(this->iIterationMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CLatticeSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iMaxIterationMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CLatticeSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iTimeMessageMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CLatticeSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iWorkingIterationMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CLatticeSievingPhase",rc,0,"Cannot initialize mutex.");
	}
	if ((rc = pthread_mutex_init(&(this->iRelationBufferMutex), NULL)) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CLatticeSievingPhase",rc,0,"Cannot initialize mutex.");
	}
}
CLatticeSievingPhase::~CLatticeSievingPhase()
{
	Reset();

	if (this->iTimes != NULL)
	{
		delete this->iTimes;
		this->iTimes = NULL;
	}
	if (this->iStats != NULL)
	{
		delete this->iStats;
		this->iStats = NULL;
	}
	
	if (this->iRelationBuffer != NULL)
	{
		int max = this->iRelationBuffer->size();
		for (int i = 0; i < max; i++)
		{
			if (this->iRelationBuffer->at(i) != NULL)
			{
				delete this->iRelationBuffer->at(i);
				this->iRelationBuffer->at(i) = NULL;
			}
		}
		this->iRelationBufferMaxCount = 0;
		this->iRelationBuffer->clear();

		delete this->iRelationBuffer;
		this->iRelationBuffer = NULL;
	}

	if (this->iRelationWorkingBuffer != NULL)
	{
		int max = this->iRelationWorkingBuffer->size();
		for (int i = 0; i < max; i++)
		{
			if (this->iRelationWorkingBuffer->at(i) != NULL)
			{
				delete this->iRelationWorkingBuffer->at(i);
				this->iRelationWorkingBuffer->at(i) = NULL;
			}
		}
		this->iRelationWorkingBufferMaxCount = 0;
		this->iRelationWorkingBuffer->clear();

		delete this->iRelationWorkingBuffer;
		this->iRelationWorkingBuffer = NULL;
	}
	DisposeFunctorField();
	DisposeGMP();
	DisposeMutexes();
	DisposeAcceptedReturnCodes();
}
int CLatticeSievingPhase::Reset()
{
	if (this->iTimes != NULL)
	{
		this->iTimes->ResetAll();
	}
	if (this->iStats != NULL)
	{
		this->iStats->ResetAll();
	}
	if (this->iRelation != NULL)
	{
		delete this->iRelation;
		this->iRelation = NULL;
	}

	this->iUseCyclesForRelConstruction = false;

	this->iUseHighVariations = false;

	this->iSpecialQ = 0;
	this->iRootIndex = 0;

	Set_Iteration(0);
	Set_MaxIteration(0);
	Set_WorkingIteration(0);
	Set_TimeMessage(0);

	this->iRSerialAccordingIters = false;
	this->iRSerialAccordingTime = false;
	this->iRSerialAccordingRelations = false;

	this->iIntermediateRoots.clear();

	if (this->iLPHashtable != NULL)
	{
		delete this->iLPHashtable;
		this->iLPHashtable = NULL;
	}

	if (this->iHighVarHashtable != NULL) 
	{
		delete this->iHighVarHashtable;
		this->iHighVarHashtable = NULL;
	}

	// we use just one instance of the same factoring algorithm - TODO better
	if (this->iVariationsInfo[0].FactorAlg == this->iVariationsInfo[1].FactorAlg)
	{
		this->iVariationsInfo[1].FactorAlg = NULL;
	}

	int max, i;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iOptimizer[i].Reset();
		
		this->iAdvFB[i].Reset();
		
		this->iFBs[i].Reset();
		
		this->iNumberFields[i].Reset();

		this->iHashtables[i].DeleteHashtables();

		if (this->iSievingBlocks[i] != NULL)
		{
			free(this->iSievingBlocks[i]);
			this->iSievingBlocks[i] = NULL;
		}

		this->iInitBlockValues[i] = 0;
		this->iInitBlockValueChange[i] = 0;
		this->iRSerialInitBlockValueChange[i] = 0;

		this->iVariationsInfo[i].Reset();

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeSievingMiddle[i] = 0;
		this->iTimeSievingMiddleStart[i] = 0;
		this->iTimeSievingLargish[i] = 0;
		this->iTimeSievingLargishStart[i] = 0;
		this->iTimeFactorMiddle[i] = 0;
		this->iTimeFactorMiddleStart[i] = 0;
		this->iTimeFactorLargish[i] = 0;
		this->iTimeFactorLargishStart[i] = 0;
		this->iTimeFactorDivideSpecial[i] = 0;
		this->iTimeFactorDivideSpecialStart[i] = 0;
#endif
	}

	if (this->iRelationBuffer != NULL)
	{
		max = this->iRelationBuffer->size();
		for (i = 0; i < max; i++)
		{
			if (this->iRelationBuffer->at(i) != NULL)
			{
				this->iRelationBuffer->at(i)->Reset();
			}
		}
		this->iRelationBufferMaxCount = 0;
	}

	if (this->iRelationWorkingBuffer != NULL)
	{
		max = this->iRelationWorkingBuffer->size();
		for (i = 0; i < max; i++)
		{
			if (this->iRelationWorkingBuffer->at(i) != NULL)
			{
				this->iRelationWorkingBuffer->at(i)->Reset();
			}
		}
		this->iRelationWorkingBufferMaxCount = 0;
	}

	this->iLinesInBlock = 0;
	this->iLinesInBlockSet = 0;

	this->iFactorBlockShift = 0;
	this->iFactorBlockMask = 0;
	this->iFactorBlockSize = 0;

	if (this->iFactorBlock != NULL)
	{
		free(this->iFactorBlock);
		this->iFactorBlock = NULL;
	}

	ASievingPhase::Reset();

	Set_InnerState((int)ESPhaseNotStarted);
	return ConstRC::Ok;
}
int CLatticeSievingPhase::DisposeGMP()
{
	mpz_clear(this->iAuxMpz1);
	mpz_clear(this->iAuxMpz2);

	return ASievingPhase::DisposeGMP();
}
int CLatticeSievingPhase::DisposeMutexes()
{
	int rc;

	if ((rc = pthread_mutex_destroy(&(this->iIterationMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iMaxIterationMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iTimeMessageMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iWorkingIterationMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}
	if ((rc = pthread_mutex_destroy(&(this->iRelationBufferMutex))) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DisposeMutexes",rc,0,"Cannot destroy mutex.");
	}

	return ASievingPhase::DisposeMutexes();
}
int CLatticeSievingPhase::FillFunctorField()
{
	if (Get_FunctorsInit() == false)
	{
		Set_FunctorsInit(true);
		this->iPhaseFunctors.reserve(ConstSPhase::INNER_STATE_COUNT);
		this->iPhaseFunctors.clear();
		this->iPhaseFunctors.push_back(NULL); // ESPhaseNotStarted
		this->iPhaseFunctors.push_back(NULL); // ESPhaseInitialization
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CLatticeSievingPhase>(this, &CLatticeSievingPhase::InitFactorBases)); // ESPhaseInitFB
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CLatticeSievingPhase>(this, &CLatticeSievingPhase::InitAdvFactorBases)); // ESPhaseInitAdvFB
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CLatticeSievingPhase>(this, &CLatticeSievingPhase::RunSieve)); // ESPhaseSieving
		this->iPhaseFunctors.push_back(new CPhaseFunctor<CLatticeSievingPhase>(this, &CLatticeSievingPhase::SaveResult)); // ESPhaseSavingResult
		this->iPhaseFunctors.push_back(NULL); // ESPhaseFinished
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::CleanUp(AFactorAlgParameters* aParameters, int aPhaseNBR, cleaning_types aType, bool aError)
{
	if (aParameters == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CleanUp",0,0,"Parameters are NULL.");
		return ConstRC::BadArgument;
	}

	string identifier;
	string directory = "";
	string file;
	string auxStr;

	int phaseNBR;
	if (aParameters->GetIntegerPhaseParameter(ConstPN::SIEVING_PHASE_NUMBER, phaseNBR, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CleanUp",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}

	if (aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_IDENTIFIER, identifier, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","CleanUp",0,0,"Didn't find serial identifier.");
		return ConstRC::GeneralError;
	}
	aParameters->GetStringPhaseParameter(ConstPN::SERIALIZATION_DIRECTORY, directory, (int)EMain);

	switch (aType)
	{
		case ECleanAllIfNotError:
			if (aError)
			{
				break;
			}
		case ECleanAll:
		case ECleanAllExcCF:
			// Smooth relation full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::SMOOTH_RELATIONS_FILENAME, 
					ConstPN::SMOOTH_RELATIONS_DIRECTORY, ConstPN::SMOOTH_RELATIONS_COMPRESSION, directory,
					ConstFN::SMOOTH_RELATIONS, identifier, phaseNBR, TXT_FILE_EXTENSION);
			}
			remove(file.c_str());

			// Partial relation full filename
			if (aParameters->ContainsPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::PARTIAL_REL_FILENAME, 
					ConstPN::PARTIAL_REL_DIRECTORY, ConstPN::PARTIAL_REL_COMPRESSION, directory,
					ConstFN::PARTIAL_REL, identifier, phaseNBR, TXT_FILE_EXTENSION);
			}
			remove(file.c_str());

		case ECleanAllExcCFRF:
			// FB full filenames
			for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				auxStr = Utils::ConvertToString(i + 1);
				if (aParameters->ContainsPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, Get_PhaseNbr()))
				{
					aParameters->GetStringPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, file, Get_PhaseNbr());
					MSerializable::AdaptPathSeparator(file);
				}
				else
				{
					aParameters->CreateFullFileName(file, ConstPN::FACTOR_BASE_FILENAME + auxStr, 
						ConstPN::FACTOR_BASE_DIRECTORY + auxStr, ConstPN::FACTOR_BASE_COMPRESSION + auxStr, Get_SerialMainDirectory(),
						ConstFN::FACTOR_BASE + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
				}
				remove(file.c_str());
			}

			// Running Serialization full filename - default is same as iSerialPhaseFullFileName
			if (aParameters->ContainsPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, phaseNBR))
			{
				aParameters->GetStringPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, file, phaseNBR);
				MSerializable::AdaptPathSeparator(file);
			}
			else
			{
				aParameters->CreateFullFileName(file, ConstPN::RSERIAL_SIEVING_FILENAME, 
					ConstPN::RSERIAL_SIEVING_DIRECTORY, ConstPN::RSERIAL_SIEVING_COMPRESSION, directory,
					ConstFN::SIEVING_PHASE, identifier, phaseNBR);
			}
			remove(file.c_str());

			break;
		default:
			break;
	}

	return ASievingPhase::CleanUp(aParameters, phaseNBR, aType, aError);
}
int CLatticeSievingPhase::InitParameters()
{
	Set_InnerState((int)ESPhaseInitialization);

	AFactorAlgParameters * param = Get_Parameters();

	if (param == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","InitParameters",0,0,"Parameters aren't set.");
		return ConstRC::GeneralError;
	}

	// generally we don't know order of the sieving phase
	int auxInt;
	if (param->GetIntegerPhaseParameter(ConstPN::SIEVING_PHASE_NUMBER, auxInt, (int)EMain) != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","InitParameters",0,0,"Didn't find phase number.");
		return ConstRC::GeneralError;
	}
	Set_PhaseNbr(auxInt);

	// now we initialize members from parents's classes
	int rc = ASievingPhase::InitParameters();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int blockSize = Get_BlockSize();
	int i;
	long auxLong;
	int varType = (int)ConstSPhase::VARIATIONS_TYPE;
	int algType = (int)ConstSPhase::SUPPORTING_FACTOR_ALG;
	ASupportingFactorAlg * alg = NULL;
	bool deleteAlg = true;
	string auxStr;
	long bound = 0;
	double ratio = 1.0;
	int policy = ConstSPhase::THRESHOLD_OPTIMIZER_POLICY;
	Polynomial * poly = NULL; // do not delete - it is saved in NumberField
	
	// we try to find value for all 
	param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND, bound, bound, (int)EMain);
	param->GetIntegerPhaseParameter(ConstPN::THRESHOLD_OPT_POLICY, policy, policy, Get_PhaseNbr());
	param->GetDoublePhaseParameter(ConstPN::THRESHOLD_OPT_RATIO, ratio, ratio, Get_PhaseNbr());
	param->GetIntegerPhaseParameter(ConstPN::VARIATIONS_TYPE, varType, varType, (int)EMain);
	param->GetIntegerPhaseParameter(ConstPN::SUPPORTING_FACTOR_ALG, algType, algType, Get_PhaseNbr());
	param->GetIntegerPhaseParameter(ConstPN::FACTOR_BLOCK_SIZE, blockSize, auxInt, Get_PhaseNbr());
	Utils::GetClosestPowerOf2(auxLong, auxInt);
	this->iFactorBlockSize = auxLong;
	rc = CFactorAlgCreator::CreateFactorAlg((supporting_factor_alg_types)algType, alg);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	variations_types maxType = (variations_types)varType;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = Utils::ConvertToString(i + 1);
		rc = param->GetPolynomialPhaseParameter(ConstPN::SIEVING_POLY + auxStr, poly, (int)EMain);
		if (rc != ConstRC::Ok)
		{
			CLog::LogError(WHEREARG,"CLatticeSievingPhase","InitParameters",0,0,"Didn't find the sieving polynomial.");
			return rc;
		}
		this->iNumberFields[i].SetSievingPoly(poly);
		poly = NULL;

		// we try to find individual value
		param->GetLongPhaseParameter(ConstPN::FB_UPPER_BOUND + auxStr, bound, bound, (int)EMain);
		this->iFBs[i].UpperFBBound = bound;

		param->GetDoublePhaseParameter(ConstPN::THRESHOLD_OPT_RATIO + auxStr, ratio, ratio, Get_PhaseNbr());
		this->iOptimizer[i].Set_DesiredRatio(ratio);

		param->GetIntegerPhaseParameter(ConstPN::THRESHOLD_OPT_POLICY + auxStr, policy, policy, Get_PhaseNbr());
		this->iOptimizer[i].Set_OptimizationPolicy(policy);

		param->GetIntegerPhaseParameter(ConstPN::LARGISH_PRIMES_THRESHOLD + auxStr, blockSize, auxInt, (int)EMain);
		this->iLargishPrimesThreshold[i] = auxInt;

		param->GetIntegerPhaseParameter(ConstPN::PARTIAL_FACTOR_MULTIPLIER + auxStr, ConstSPhase::PARTIAL_FACTOR_MULTIPLIER, auxInt, (int)EMain);
		this->iPartialFactorMultiplier[i] = auxInt;

		param->GetIntegerPhaseParameter(ConstPN::VARIATIONS_TYPE + auxStr, varType, auxInt, (int)EMain);
		this->iVariationsInfo[i].Type = (variations_types)auxInt;
		maxType = max(this->iVariationsInfo[i].Type, maxType);

		if (this->iVariationsInfo[i].Type > ELargePrimeV)
		{
			// we want only one instance of the same factorization algorithm
			param->GetIntegerPhaseParameter(ConstPN::SUPPORTING_FACTOR_ALG + auxStr, algType, auxInt, Get_PhaseNbr());
			if (auxInt != algType)
			{
				rc = CFactorAlgCreator::CreateFactorAlg((supporting_factor_alg_types)auxInt, this->iVariationsInfo[i].FactorAlg);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			else
			{
				this->iVariationsInfo[i].FactorAlg = alg;
				deleteAlg = false;
			}	
		}
	}

	if (maxType == ELargePrimeV)
	{
		// we can use cycle construction method for creating smooth relations
		this->iUseCyclesForRelConstruction = true;
	}

	if ((maxType == EDoubleLargePrimeV) || (maxType == ETrippleLargePrimeV)) this->iUseHighVariations = true;

	if (deleteAlg)
	{
		delete alg;
		alg = NULL;
	}
	else
	{
		// factorization algorithm will be deleted in Reset method
		alg = NULL;
	}

	if (bound == 0)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","InitParameters",0,0,"Didn't find the factor base upper bound.");
		return ConstRC::GeneralError;
	}

	rc = param->GetLongPhaseParameter(ConstPN::SIEVING_REGION_WIDTH, this->iLatticeInfo.I, (int)EMain);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","InitParameters",0,0,"Didn't find the sieving interval length.");
		return rc;
	}
	// we support only limited length of the sieving interval
	if (this->iLatticeInfo.I > ConstSPhase::LATTICE_MAX_HALF_INTERVAL_LENGTH)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","InitParameters",0,this->iLatticeInfo.I,"Sieving interval is too long. It will be reduced.");
		this->iLatticeInfo.I = ConstSPhase::LATTICE_MAX_HALF_INTERVAL_LENGTH;
	}
	// we get I/2
	this->iLogI = Utils::GetClosestPowerOf2(this->iLatticeInfo.I, this->iLatticeInfo.I << 1);
	this->iModI = this->iLatticeInfo.I - 1;

	// determine special Q range
	if (param->GetIntegerPhaseParameter(ConstPN::SIEVING_REGION_HEIGHT, iSpecialQCount, (int)EMain) != ConstRC::Ok) iSpecialQCount = (int)SIEVING_REGION_NO_HEIGHT_BOUND;
	if (param->GetIntegerPhaseParameter(ConstPN::SIEVING_REGION_START, iFirstSpecialQIndex, (int)EMain) != ConstRC::Ok) iFirstSpecialQIndex = (int)SIEVING_REGION_DEFAULT_START;

	// number of smooths - if given would override usual check enough
	if (param->GetIntegerPhaseParameter(ConstPN::NUMBER_OF_SMOOTHS, iAssignedNumberOfSmooths, (int)EMain) != ConstRC::Ok) iAssignedNumberOfSmooths = (int)NUMBER_OF_SMOOTHS_NOT_GIVEN;

	param->GetDoublePhaseParameter(ConstPN::POLY_SKEWNESS, ConstSPhase::POLY_SKEWNESS, this->iSkewness, (int)EMain);

	param->GetIntegerPhaseParameter(ConstPN::THRESHOLD_OPT_INT, ConstSPhase::THRESHOLD_OPTIMIZER_INT, this->iThresholdOptInterval, Get_PhaseNbr());

	param->GetIntegerPhaseParameter(ConstPN::MAX_ITERATIONS, ConstSPhase::MAX_ITERATIONS, this->iMaxIteration, Get_PhaseNbr());

	param->GetIntegerPhaseParameter(ConstPN::TIME_MESSAGE_INTERVAL, ConstSPhase::TIME_MESSAGE_INTERVAL, auxInt, Get_PhaseNbr());
	Set_TimeMessage(auxInt);

	// Running serialization setup

	if (param->ContainsPhaseParameter(ConstPN::RSERIALIZATION_ITER_INTERVAL, Get_PhaseNbr()))
	{
		this->iRSerialAccordingIters = true;
		param->GetIntegerPhaseParameter(ConstPN::RSERIALIZATION_ITER_INTERVAL, this->iRSerialIterInterval, Get_PhaseNbr());
	}
	if (param->ContainsPhaseParameter(ConstPN::RSERIALIZATION_TIME_INTERVAL, Get_PhaseNbr()))
	{
		this->iRSerialAccordingTime = true;
		param->GetLongPhaseParameter(ConstPN::RSERIALIZATION_TIME_INTERVAL, this->iRSerialTimeInterval, Get_PhaseNbr());
	}
	if (param->ContainsPhaseParameter(ConstPN::RSERIALIZATION_REL_COUNT, Get_PhaseNbr()))
	{
		this->iRSerialAccordingRelations = true;
		param->GetIntegerPhaseParameter(ConstPN::RSERIALIZATION_REL_COUNT, this->iRSerialRelationsCount, Get_PhaseNbr());
	}
	// Running Serialization full filename - default is same as iSerialPhaseFullFileName
	if (param->ContainsPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::RSERIAL_SIEVING_FULL_FILENAME, this->iRunningSerialFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iRunningSerialFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iRunningSerialFullFileName, ConstPN::RSERIAL_SIEVING_FILENAME, 
			ConstPN::RSERIAL_SIEVING_DIRECTORY, ConstPN::RSERIAL_SIEVING_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::SIEVING_PHASE, Get_SerialIdentifier(), Get_PhaseNbr());
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		auxStr = Utils::ConvertToString(i + 1);
		// FB full filename
		if (param->ContainsPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, Get_PhaseNbr()))
		{
			param->GetStringPhaseParameter(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, this->iFBs[i].FBFullFileName, Get_PhaseNbr());
			MSerializable::AdaptPathSeparator(this->iFBs[i].FBFullFileName);
		}
		else
		{
			param->CreateFullFileName(this->iFBs[i].FBFullFileName, ConstPN::FACTOR_BASE_FILENAME + auxStr, 
				ConstPN::FACTOR_BASE_DIRECTORY + auxStr, ConstPN::FACTOR_BASE_COMPRESSION + auxStr, Get_SerialMainDirectory(),
				ConstFN::FACTOR_BASE + auxStr, Get_SerialIdentifier(), Get_PhaseNbr());
		}
	}

	// Smooth relation full filename
	if (param->ContainsPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, this->iSmoothRelFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iSmoothRelFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iSmoothRelFullFileName, ConstPN::SMOOTH_RELATIONS_FILENAME, 
			ConstPN::SMOOTH_RELATIONS_DIRECTORY, ConstPN::SMOOTH_RELATIONS_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::SMOOTH_RELATIONS, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}
	// Partial relation full filename
	if (param->ContainsPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, Get_PhaseNbr()))
	{
		param->GetStringPhaseParameter(ConstPN::PARTIAL_REL_FULL_FILENAME, this->iPartialRelFullFileName, Get_PhaseNbr());
		MSerializable::AdaptPathSeparator(this->iPartialRelFullFileName);
	}
	else
	{
		param->CreateFullFileName(this->iPartialRelFullFileName, ConstPN::PARTIAL_REL_FILENAME, 
			ConstPN::PARTIAL_REL_DIRECTORY, ConstPN::PARTIAL_REL_COMPRESSION, Get_SerialMainDirectory(),
			ConstFN::PARTIAL_REL, Get_SerialIdentifier(), Get_PhaseNbr(), TXT_FILE_EXTENSION);
	}

	// should we deserialize or serialize?
	bool deserialize = false;
	param->GetBoolPhaseParameter(ConstPN::DESERIALIZE, false, deserialize, Get_PhaseNbr());
	CBaseParameters bparam;
	bparam.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (Get_InfoMode())
	{
		CLog::Log(10000, "0182000", "Lattice Sieving Phase is running ...\n\n");
	}
#endif

	if (deserialize)
	{
		rc = Deserialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}
	else
	{
		rc = Serialize(bparam);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

int CLatticeSievingPhase::DetermineRequestedRelations() 
{
	long min_requested;

	// reserve
	this->iRequestedRelations = this->iRelationReserve;

	// quadratic characters
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
		{
			this->iRequestedRelations += ConstSPhase::QUAD_CHARS_COUNT;
		}
	}

	// we give a lower bound to the estimates to make the sieving progress more fluent
	min_requested = this->iRequestedRelations;

	// bases
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		min_requested += this->iFBs[i].FBMaxIndex + 1;
	}

	// special Q
	min_requested += Get_WorkingIteration();

	if (this->iUseHighVariations && this->iHighVarHashtable != NULL && this->iHighVarHashtable->Filled)
	{
		relation_index_type relCount = this->iStats->Get_Total(ESSFF) + this->iStats->Get_Total(ESSPartialR);
		// we allocate a temporary array of bytes, with a single bit per each found relation.	
		// when we found singleton primeideal, we indicate its relation as singleton in this array,
		// so when we iterate throught relations we can quickly decide if it has singleton primeideal.
		// 0 - nonsingleton, 1 - singleton
		// extra care during iteration because of bit values!
		detection_type * detectionField = NULL;
		int detectionFieldSize = (relCount >> ConstHT::BITS_IN_BYTE_SHIFT) + 1;
		detectionField = (detection_type*) NFSUtils::AllocationMachine((void**) &detectionField, detectionFieldSize * sizeof(detection_type));
		if (detectionField == NULL) 
		{
			CLog::LogError(WHEREARG, "CLatticeSievingPhase", "DetermineRequestedRelations", detectionFieldSize, detectionFieldSize * sizeof(detection_type),  "Not enough memory to allocate data space.");
			return ConstRC::NotEnoughMemory;
		}
		memset(detectionField, ConstHT::NONSINGLETON, detectionFieldSize * sizeof(detection_type));

		relation_index_type primeIdealHFCount = 0;
		relation_index_type singletonCount = 0;
		this->iHighVarHashtable->FindSingletonRelations(detectionField, singletonCount, primeIdealHFCount);
		free(detectionField); // TODO: persistent

		this->iRequestedRelations += primeIdealHFCount + singletonCount;
	}

	if (this->iRequestedRelations < min_requested) this->iRequestedRelations = min_requested;

	return ConstRC::Ok;
}

int CLatticeSievingPhase::RunSieve()
{
	Set_InnerState(ESPhaseSieving);

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTRunSieve);
#endif

	int rc = PrepareForSieving();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int i;
	int max;
	long * roots = NULL;
	// we use last part for generating special q
	Polynomial * mainPoly = this->iNumberFields[CNFSRelation::DEFAULT_NFS_PARTS_COUNT - 1].SievingPoly;
	roots = (long*) Utils::AllocationMachine((void**) &roots, mainPoly->GetDegree() * sizeof(long));
	if (roots == NULL) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","RunSieve",0,0,"Could not allocate roots.");
		return ConstRC::NotEnoughMemory;
	}

	INT64 modLC;
	mpz_t prime;
	mpz_init_set_si(prime, this->iSpecialQ);

	bool enough = false;

	bool special_Q_range_end = false;
	int special_Q_count = 0;

	int iter = Get_Iteration() - 1;

	float skewnessSqr = this->iSkewness * this->iSkewness;

	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iRunningSerialFullFileName);
	param.SetParameter(ConstPN::RUNNING_SERIALIZATION, Utils::ConvertToString(true));
	rc = StartRunningSerialization(param);
	if (rc != ConstRC::Ok)
	{
		free(roots);
		roots = NULL;
		mpz_clear(prime);
		return rc;
	}
	if (this->iRSerialAccordingTime)
	{
		this->iTimes->Start(ESTRSerial);
	}

	if (Get_InfoMode())
	{
		PrintHeader();
	}

	do
	{
		// special q and roots
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTRootUpdate);
#endif
		modLC = 0;
		max = mainPoly->RootsModPrimeGeneral(roots, this->iSpecialQ);
		if (max > 0)
		{
			modLC = mpz_mod_ui(this->iAuxMpz1, *mainPoly->Leading2(), this->iSpecialQ);
		}
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTRootUpdate);
#endif

		if (modLC != 0)
		{
			for (; this->iRootIndex < max; this->iRootIndex++)
			{
				iter++;

				Set_WorkingIteration(iter);

				if ((Get_InfoMode() && Get_TimeMessage()) != 0)
				{
					if ((iter % Get_TimeMessage()) == 0)
					{
						PrintValues();

						if (iter % (10*Get_TimeMessage()) == 0)
						{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
							this->iTimes->End(ESTRunSieve);
							this->iTimes->Start(ESTRunSieve);
#endif

							PrintEffectiveness();
							PrintHeader();
						
							this->iStats->ResetAllIntermediate();
							this->iTimes->ResetAllIntermediate();
						}
					}
				}

				// running serialization
				if (this->iRSerialAccordingTime)
				{
					this->iTimes->End(ESTRSerial);
				}
				if (this->iRelationWorkingBufferMaxCount > ConstSPhase::MAX_RELATION_BUFFER_SIZE || 
					(this->iRSerialAccordingRelations && this->iRelationWorkingBufferMaxCount > this->iRSerialRelationsCount) ||
					(this->iRSerialAccordingIters && ((iter % this->iRSerialIterInterval) == 0)) ||
					(this->iRSerialAccordingTime && this->iTimes->Get_TotalTime(ESTRSerial) > this->iRSerialTimeInterval))
				{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
					this->iTimes->End(ESTRunSieve);
					this->iTimes->Start(ESTRunSieve);
#endif
					rc = RunRunningSerialization();
					if (rc != ConstRC::Ok)
					{
						free(roots);
						roots = NULL;
						mpz_clear(prime);
						return rc;
					}
					if (this->iRSerialAccordingTime)
					{
						this->iTimes->Reset(ESTRSerial);
					}
				}
				if (this->iRSerialAccordingTime)
				{
					this->iTimes->Start(ESTRSerial);
				}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
				this->iTimes->Start(ESTRootUpdate);
#endif

				this->iSpecialR = roots[this->iRootIndex];
				if (this->iSpecialR > 0)
				{
					this->iSpecialR = this->iSpecialQ - this->iSpecialR;
				}

				// compute reduced base of main lattice
				ReduceBase(this->iLatticeInfo.a0, this->iLatticeInfo.b0, this->iLatticeInfo.a1, 
					this->iLatticeInfo.b1, this->iSpecialQ, 0, this->iSpecialR, 1, skewnessSqr);

				// prepare adv. factor bases
				for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
				{
					CAdvFBHelper::FillLatticeFB(this->iAdvFB[i], this->iLatticeInfo);
					//ControlLatticeElementInfo(this->iAdvFB[i]->FactorBase, this->iFBs[i].FBMaxIndex);
				}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
				this->iTimes->End(ESTRootUpdate);
#endif

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
				if (Get_InfoMode())
				{
					CLog::Log(0, "0182001", "LatticeSiever: Calling SieveRegion()\n");
				}
#endif

				rc = SieveRegion();
				if (rc != ConstRC::Ok)
				{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
					if (Get_InfoMode())
					{
						CLog::Log(0, "0182002", "LatticeSiever: The result of SieveRegion() was not OK = %d.\n", rc);
					}
#endif
					free(roots);
					roots = NULL;
					mpz_clear(prime);
					return rc;
				}				

				if (Get_MaxIteration() != ConstSPhase::UNLIMITED_ITERATIONS)
				{
					if (iter >= Get_MaxIteration()) 
					{
//#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
//						CLog::Log(5000, "0182900", "\n    Total time spent by sieving itself is %12.4f sec.\n", this->iTimes->Get_TotalTime(ESTLineSieving));
//#endif
						break; 
					}
				}

				DetermineRequestedRelations();
				if ((enough = (CheckEnough() == ConstRC::EnoughSmooths))) break;
			}

			if (Get_MaxIteration() != ConstSPhase::UNLIMITED_ITERATIONS)
			{
				CLog::Log(50, "", "Recalculating requested count of relations.");
				if (iter >= Get_MaxIteration()) 
				{
					break; 
				}
			}
		}

		// next special q
		mpz_nextprime(prime, prime);
		this->iSpecialQ = mpz_get_si(prime);
		this->iRootIndex = 0;

		// check special Q range end
		if (iSpecialQCount != (int)SIEVING_REGION_NO_HEIGHT_BOUND) {
			if (++special_Q_count == iSpecialQCount) special_Q_range_end = true;
		}
	} 
	while ((!enough) && (!special_Q_range_end));

	// set correct iteration
	iter++;
	Set_WorkingIteration(iter);

	if (this->iRSerialAccordingTime)
	{
		this->iTimes->End(ESTRSerial);
	}

	free(roots);
	roots = NULL;
	mpz_clear(prime);
	
	rc = EndRunningSerialization();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTRunSieve);
#endif
	if (Get_InfoMode())
	{
		PrintValues();
		PrintEffectiveness();
	}

	// The sieving ended before enough relations have been found.
	if (!enough) {		
		if (special_Q_range_end) {
                        CLog::Log(50, "", "Special Q range was exhausted.");
                        fprintf(stdout, "\n\tSpecial Q range was exhausted.\n");
                        return ConstRC::SievingRegionExhausted;
                }
		return ConstRC::NotEnoughSmooths;
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::SieveRegion()
{
	int rc;
	int k, i, j;
	int block_max;

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTLineSieving);
#endif

	// Fill hashtable
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		//this->iHashtables[i].PrintHashtables();
		this->iInitBlockValues[i] = SIEVING_PHASE_MASK - 
				ComputeExpectedLog(this->iNumberFields[i], this->iVariationsInfo[i]);
	}

	mpz_set_si(this->iAuxMpz1, this->iSpecialQ);
	this->iInitBlockValues[CNFSRelation::DEFAULT_NFS_PARTS_COUNT - 1] += mpz_sizeinbase(this->iAuxMpz1, ConstSPhase::LOGARITHM_BASE);

	if (this->iWorkingIteration % this->iThresholdOptInterval == 0)
	{
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			this->iOptimizer[i].AnalyzeAndOptimize(this->iInitBlockValueChange[i]);
		}
	}

	for (k = 0, i = 0; k < this->iBlockCount; k+= this->iBlockSetSize, i++) 
	{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182003", "\n Block set sieve: positive k = %d start.\n", k);
#endif
#endif
		block_max = this->iBlockSetSize;		
		if ((this->iBlockCount - k) < block_max)
		{
			block_max = this->iBlockCount - k;
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTBlockSieving);
#endif

		// Fill hashtable
		for (j = 0; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++)
		{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
			this->iTimes->Start(ESTFirstFillHT + j);
#endif

			rc = FillHashtable(k, this->iHashtables[j], this->iAdvFB[j]);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
			this->iTimes->End(ESTFirstFillHT + j);
#endif
		}

		// The sieving itself
		rc = SieveBlock(i, 0, block_max); // last two values have no purpose...
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTBlockSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182004", "\n Block set sieve: positive k = %d end.\n", k);
#endif
#endif
	} // end of for cycle

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTLineSieving);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	Utils::Wait();
#endif

	return ConstRC::Ok;
}
/*!
	The hashtable principle:
	
	We have iBlockSetSize hashtables available.
*/
int CLatticeSievingPhase::SieveBlock(int aBlockSetIndex, int aBlockMin, int aBlockMax)
{
	int rc;
	for (int i = aBlockMin; i < aBlockMax; i++) 
	{			
		rc = Sieving(i);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

		// Now the factorizable candidates are found and factorized
		rc = FindingAndFactorizing(aBlockSetIndex, i);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}
	}
	
	return ConstRC::Ok;
}
int CLatticeSievingPhase::SieveWithMiddlePrimes(lattice_update_t* aStart, lattice_update_t* aEnd, log_type* aBlock) 
{
	register main_sieving_type offset;
	register log_type          log;
	register main_sieving_type b0;
	register main_sieving_type b1;
	register main_sieving_type a;
	register main_sieving_type c;
	register main_sieving_type modI = this->iModI;
	register main_sieving_type i;
	register int blocksize = Get_BlockSize();
	log_type *aBlockPtr;
	main_sieving_type a_plus_c;

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	bool assertionMode = Get_AssertionMode();
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	long sum = 0;
	long hits = 0;
#endif
	
	lattice_update_t* running = aStart;
	do
	{
		offset = running->h;	
		running->start = offset;

		if (offset >= 0) 
		{
			log = running->logp;
			b0 = running->b0;
			b1 = running->b1 + ConstSPhase::LATTICE_FB_B1_CHANGE;
			a = running->a;
			c = running->c;

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

			if (!log)
			{
				CLog::Log(50, "0182005", "Log is zero for p = %d, strange.\n", p);
			}

			CLog::Log(50, "0182006", "SievingWithMiddlePrimes: p = %d, start = %d, curr = %p, end = %p.\n",running->p,running->z, (void*)running, (void*)aEnd);
#endif
#endif

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
			if (assertionMode)
			{
				if (offset >= 0)
				{								
					// TODO control
				}

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
				if (running->element->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME &&
					running->element->r == DEBUG_SIEVING_PHASE_CRITICAL_ROOT)
				{
					
				}
#endif
			}
#endif


			aBlockPtr = aBlock+offset;
			a_plus_c = a+c;
			while (offset < blocksize)
			{				
				*aBlockPtr += log;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
				sum += log;
				hits++;
#endif
				i = offset & modI;

				if (i >= b0) {
					if (i < b1) {
						offset += a_plus_c;
						aBlockPtr += a_plus_c;
					}
					else {
						offset += a;
						aBlockPtr += a;
					}
				}
				else {
					if (i < b1) {
						offset += c;
						aBlockPtr += c;
					}
				}
			}

			// The new index for the next block
			running->h = offset - blocksize;
		} // else the root is invalid
	} 
	while (running++ != aEnd); // Will be broken by internal check.

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	if (hits != 0)
	{
		CLog::Log(50, "0182007", "Sieve mid primes: added total %d logs in %d hits (%d on avg).\n",sum,hits,sum/hits);
	}
#endif
#endif

	return ConstRC::Ok;
}
int CLatticeSievingPhase::SieveWithLargishPrimes(lattice_script_t* aScript, log_type* aBlock)
{
	register lattice_script_t* script = aScript;
	register lattice_update2_t *updates = NULL;
	register int j;
	register int max;

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	register bool assertionMode = Get_AssertionMode();
#endif


#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	int total = 0;
	int listIndex = 0;
	int runningIndex = 0;

	long sum = 0;
	int hits = 0;
#endif
	while (script)
	{	
		max = script->num_used;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
		total += max;
		listIndex++;
#endif
		updates = script->list;
		for (j = 0; j < max; j++, updates++)
		{
/*			if ((j & 7) == 0) //if (j % 8 == 0) 
			{
				PREFETCH(updates + 16);
			} */

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
			if (assertionMode) 
			{
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
				if (updates->element->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME &&
					updates->element->r == DEBUG_SIEVING_PHASE_CRITICAL_ROOT)
				{
					PrintUpdateInfo(*updates);
				}
#endif
				//TestUpdate(*updates);
			}
#endif
			aBlock[updates->h] += updates->logp;
		}
		script = script->next;
	}
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

	if (hits) 
	{
		CLog::Log(50, "0182008", "Sieve largish primes: added total %d logs in %d hits (avg %d).\n",sum,hits,sum/hits);
	} 
	else 
	{
		CLog::Log(50, "0182009", "Sieve largish primes: No largish prime availables\n");
	}

#endif
#endif

	return ConstRC::Ok;
}
int CLatticeSievingPhase::FillHashtable(int aFirstBlockIndex, CLatticeHashtables & aHashtables, CAdvLFBInfo & aLFB)
{
	if (aLFB.LargishPrimes == NULL)
	{
		// In case that all primes are middle (no largish exist).
		return ConstRC::Ok;
	}

	int rc;
	int blocksize = Get_BlockSize();
	register int logBlockSize = Get_LogBlockSize();
	register int modBlockSize = Get_ModBlockSize();
	register main_sieving_type h; 
	register int i; 
	register int b0;
	register int b1;
	register main_sieving_type a;
	register main_sieving_type c;
	register int min = aFirstBlockIndex * blocksize;
	register int max = min + blocksize * this->iBlockSetSize;
	register main_sieving_type modI = this->iModI;
	
	aHashtables.ClearHashtables();

// #define FILL_HASHTABLE_STATS 1
#ifdef FILL_HASHTABLE_STATS
	int plusa=0;
	int plusc=0;
	int primes=0;
#endif

	// The upper bound of the FB is denoted by a zero block;
	lattice_update_t* running = aLFB.LargishPrimes;
	lattice_update_t* end = aLFB.LastLargishPrimes;
	do
	{		
#ifdef FILL_HASHTABLE_STATS
		primes++;
#endif
		h = running->h;
		// if element is valid ... and in right block set
		if (min <= h && h < max)
		{
			b0 = running->b0;
			b1 = running->b1 + ConstSPhase::LATTICE_FB_B1_CHANGE;
			a = running->a;
			c = running->c;
			int ac = a+c;
			while (h < max)
			{
				// Add to the update table
				rc = aHashtables.AddToTable(running, h & modBlockSize, (h >> logBlockSize) - aFirstBlockIndex);			
				if (rc != ConstRC::Ok)
				{
					return rc;
				}

				// compute new h
				i = h & modI;
				// >= b0
				if (i >= b0)
				{
					if (i < b1) {
#ifdef FILL_HASHTABLE_STATS
						plusa++;
						plusc++;
#endif
						h += ac; 
					} else {
#ifdef FILL_HASHTABLE_STATS
						plusa++;
#endif
						h += a; 
					}
				} else {
					// < b1
					if (i < b1)
					{
#ifdef FILL_HASHTABLE_STATS
						plusc++;
#endif
						h += c; 
					}
				}
			}
			running->h = h;
		}

		// max <= h ---> continue
	}
	while (running++ != end);

#ifdef FILL_HASHTABLE_STATS
		CLog::Log(50000, "0182008", "Primes: %d, plusa: %d, plusc: %d).\n",primes, plusa, plusc);
#endif
	return ConstRC::Ok;
}
int CLatticeSievingPhase::FindingAndFactorizing(int aBlockSetIndex, int aInnerIndex)
{
	// FOR DEBUG
	//if (aInnerIndex > 999)
	//{
	//	ControlBlock(aBlockSetIndex, aInnerIndex);
	//}
	//return ConstRC::Ok;

#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
	//for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	//{
	//	WriteDumpSieveFile(this->iSievingBlocks[i]);
	//}
#endif

	int blockSize = Get_BlockSize();

	// reset factor block hashtable
	memset(this->iFactorBlock, 0, this->iFactorBlockSize * sizeof(log_type));
	int candCount = 0;
	int index;
	int col;
	int row;
	int halfI = (this->iLatticeInfo.I >> 1);
	int result;

	// we start with last block
	log_type * block = this->iSievingBlocks[1]; // TODO - without magic numbers
	log_walking_type* cast  = (log_walking_type*)block;
	log_walking_type  mask  = SIEVING_PHASE_CHECK_MASK;
	int size = sizeof(log_walking_type);
	if (size == SIEVING_PHASE_64BITS_SIZE)
	{
		mask <<= ((size / 2) * 8);
		mask += SIEVING_PHASE_CHECK_MASK;
	}
	int rc;
	bool checkAll = Get_CheckAllMode();
	int iters = blockSize / size;
	int offset = 0;
	int i, j;
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
	int mask_hits  = 0;
	int candidates = 0;
	int sizes[SIEVING_PHASE_SIZES_COUNT];
	memset(sizes, 0, SIEVING_PHASE_SIZES_COUNT * sizeof(int));
#endif


	for (i = 0; i < iters; i++)
	{
		if ((cast[i] & mask)  || checkAll) 
		{ 
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
			mask_hits++;
#endif
			offset = size * i;	
			for (j = 0; j < size; j++, offset++)
			{
				
				if ((block[offset] & SIEVING_PHASE_MASK) || checkAll)
				{	
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

					if ((offset + j) >= blockSize)
					{
						CLog::Log(50, "0182010", "Size: %d, i = %d, offset = %d, j = %d, offset+j = %d, pointer = 0x%p, iters = %d.\n",size,i,offset,j,offset+j,&(block[offset+j]),iters);
						Utils::Wait();
					}
#endif
#endif

					if (this->iSievingBlocks[0][offset] & SIEVING_PHASE_MASK || checkAll)
					{
#ifdef OLD_FACTORIZATION
						rc = Factorize(offset, aBlockSetIndex, aInnerIndex);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
#else

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

						CLog::Log(50, "0182011", "FF: Offset = %d, block set index = %d, inner index = %d\n", offset, aBlockSetIndex, aInnerIndex);
#endif
#endif
						// row = 1 + lines in offset + lines in block * innerindex + lines in blockset * blocksetindex
						row = 1 + (offset >> this->iLogI) + this->iLinesInBlock * aInnerIndex + this->iLinesInBlockSet * aBlockSetIndex;
						// col = get this lines offset and subtract half of line
						col = (offset & this->iModI) - halfI;

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
						this->iStats->Increase(ESSTested);
#endif
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

						CLog::Log(50, "0182012", "FF: i = %d, j = %d\n", col, row);
#endif
#endif
						
						result = 0;
						Utils::GCD(abs(col), abs(row), result);

						if (result != 1)
						{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
							this->iStats->Increase(ESSRejected);
							this->iStats->Increase(ESSNotCoprime);
#endif
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50

							CLog::Log(50, "0182013", "FF: Rejected - GCD = %d\n", result);
#endif
#endif
							continue;
						}

						this->iFactorsInfo.SaveOffset(offset);
						this->iFactorsInfo.Rows[candCount] = row;
						this->iFactorsInfo.Cols[candCount] = col;
						candCount++;
						index = offset >> this->iFactorBlockShift; //23
						//index = offset & this->iFactorBlockMask; //34
						if (this->iFactorBlock[index] != 0)
						{
							this->iFactorBlock[index] = ConstSPhase::FACTOR_BLOCK_COLLISION;
						}
						else
						{
							this->iFactorBlock[index] = candCount;
						}
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
						candidates++;	
#endif
					}
				} 
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
				else 
				{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
					if ((offset + j) >= blockSize)
					{
						CLog::Log(50, "0182014", "Size: %d, i = %d, offset = %d, j = %d, offset+j = %d, pointer = 0x%p, iters = %d.\n",size,i,offset,j,offset+j,&(block[offset+j]),iters);
						Utils::Wait();
					}
#endif
				}
#endif
			}
		} 
	}
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182015", "Had %d mask hits and %d candidates for factorization. Size statistics:\n",mask_hits,candidates);
#endif
#endif

	rc = ConstRC::Ok;
#ifndef OLD_FACTORIZATION
	if (candCount > 0)
	{
		rc = Factorize2(aBlockSetIndex, aInnerIndex);
		this->iFactorsInfo.Clear();
	}
#endif

	return rc;
}
int CLatticeSievingPhase::Sieving(int aIndex)
{
	int rc;
	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182016", "Sieving: index = %d. Memset the block to 0x%x, size %d.\n",aIndex,this->iInitBlockValues[i] - this->iInitBlockValueChange[i],Get_BlockSize()*sizeof(log_type));
#endif
#endif

		// iInitAlgebraic1BlockValue = iNfsInstance->ComputeExpectedAlgebraicLog(0); was called in SieveLine		
		if (memset(this->iSievingBlocks[i], this->iInitBlockValues[i] - this->iInitBlockValueChange[i], Get_BlockSize() * sizeof(log_type)) == NULL)
		{
			CLog::LogError(WHEREARG,"CLatticeSievingPhase","Sieving",0,Get_BlockSize() * sizeof(log_type),"Memset failed!");
			return ConstRC::NotEnoughMemory;
		}

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182017", "Calling Sieve With Middle Primes For FB nbr. %d.\n",i);
#endif
#endif

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstSieving + i);
#endif
#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeSievingMiddleStart[i] = clock();
#endif
		// Then, we sieve with the middle-sized primes.
		rc = SieveWithMiddlePrimes(this->iAdvFB[i].MiddlePrimes, this->iAdvFB[i].LastMiddlePrimes, 
			this->iSievingBlocks[i]);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
		WriteDumpSieveFile(this->iSievingBlocks[i]);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182018", "Sieve With Middle Primes For FB nbr. %d Finished. Calling Sieve With Largish Primes.\n", i);
#endif
#endif

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeSievingMiddle[i] += (clock() - this->iTimeSievingMiddleStart[i]);
		this->iTimeSievingLargishStart[i] = clock();
#endif

		// Then, we will sieve with the largish primes	
		rc = SieveWithLargishPrimes(this->iHashtables[i].Hashtables[aIndex].script, this->iSievingBlocks[i]);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}

#ifdef DEBUG_SIEVING_PHASE_WRITE_DUMP_FILE
		WriteDumpSieveFile(this->iSievingBlocks[i]);
#endif

#ifdef DEBUG_SIEVING_PHASE_BLOCK_SIEVE	
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182019", "Sieve With Largish Primes For FB nbr. %d Finished.\n", i);
#endif
#endif

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeSievingLargish[i] += (clock() - this->iTimeSievingLargishStart[i]);
#endif

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstSieving + i);
#endif

	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::CheckEnough() 
{
	long rel_count;

	rel_count = iStats->Get_Total(ESSFF);
	if (iUseCyclesForRelConstruction) rel_count += this->iStats->Get_Total(ESSEdges) + this->iStats->Get_Total(ESSComponents) - this->iStats->Get_Total(ESSVertices);

	if (iAssignedNumberOfSmooths != (int)NUMBER_OF_SMOOTHS_NOT_GIVEN) return iAssignedNumberOfSmooths > rel_count ? ConstRC::NotEnoughSmooths : ConstRC::EnoughSmooths;

	if ((!iUseCyclesForRelConstruction) && iUseHighVariations) rel_count += iStats->Get_Total(ESSPartialR);

	return this->iRequestedRelations > rel_count ? ConstRC::NotEnoughSmooths : ConstRC::EnoughSmooths;
}

/*!
 * \brief One-time method for allocation of the factor bases.
 *
 * There are 2 factor bases used during the NFS: one consisting of integers and the
 * other of prime ideals. We must allocate some space for them.
 * 
 * In both factor bases, we allocate some extra space. It is caused by the fact that
 * we want to sieve with use of a zero-byte block at the end, which will serve as a 
 * limit value during sieving ("read until you meet a zero"); we want the total 
 * allocated block to be multiple of 4 long.
 *
 * We memset the blocks to zero, in order to have a sieving limit. 
 * The integral factor base contains all primes in span [2,F], and moreover -1.
 * We want it to include also all primes contained in prime ideals; therefore, we
 * reject any other progress if its upper bound is lower than that of algebraic
 * factor base.
 *
 * We also find and save all free relations.
 */
int CLatticeSievingPhase::InitFactorBases() 
{
	Set_InnerState(ESPhaseInitFB);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182020", "  Init integral domain ...\n");
	}
#endif

	int i;
	int rc;

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = this->iNumberFields[i].InitIntegralDomain(this->iFBs[i].UpperFBBound);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182021", "  Init factor bases ...\n");
	}
#endif

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTInitFB);
#endif

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = this->iFBs[i].InitFB();
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		rc = this->iFBs[i].FillFB(this->iNumberFields[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		this->iNumberFields[i].FillSpecialPrimes();
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTInitFB);
#endif

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182022", "    Serializing factor bases ...\n");
	}
#endif

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// serialize factor bases 
	// and set FB filenames for other phases
	string auxStr;
	CBaseParameters paramFB;
	AFactorAlgParameters * param = Get_Parameters();
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		paramFB.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iFBs[i].FBFullFileName);
		rc = this->iFBs[i].Serialize(paramFB);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		auxStr = Utils::ConvertToString(i + 1);
		param->SetParameterForOthers(ConstPN::FACTOR_BASE_FULL_FILENAME + auxStr, this->iFBs[i].FBFullFileName, Get_PhaseNbr());
	}
#endif

	// write some info
	if (Get_InfoMode())
	{
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			this->iNumberFields[i].PrintInfo(i, this->iFBs[i].FactorBase);
#ifdef DEBUG_SIEVING_PHASE_WRITE_LC_DIVIDING_PRIMES
			this->iFBs[i].PrintLCPrimes();
#endif
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
		CLog::Log(5000, "0182901", "    Init FB took: %12.4f sec\n", this->iTimes->Get_TotalTime(ESTInitFB));
#endif
#endif
	}

	bool useFreeRelations = ConstSPhase::USE_FREE_RELATIONS;
	Get_Parameters()->GetBoolPhaseParameter(ConstPN::USE_FREE_RELATIONS, ConstSPhase::USE_FREE_RELATIONS, useFreeRelations, Get_PhaseNbr());

	if (useFreeRelations)
	{
		rc = FindFreeRelations();
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	if (this->iUseCyclesForRelConstruction)
	{
		// Now, we prepare structures for counting cycles in the LPV	
		rc = PrepareCycleTable();
	}

	return rc;
}
int CLatticeSievingPhase::InitAdvFactorBases()
{
	Set_InnerState(ESPhaseInitAdvFB);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182023", "  Initializing the advanced FB ...\n");
	}
#endif

	// we serialize this phase
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();
	CBaseParameters param;
	param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	int rc = Serialize(param);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = CAdvFBHelper::InitAndPrepareLatticeFB(this->iAdvFB[i], this->iFBs[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		//lattice_type2 * f = NULL;
		//f = (lattice_type2*)Utils::AllocationMachine((void**)&f, sizeof(lattice_type2)*this->iAdvFB[i].Count);
		//memset(f, 0, sizeof(lattice_type2)*this->iAdvFB[i].Count); 
		//for (int j = 0; j < this->iAdvFB[i].Count; j++)
		//{
		//	f[j].p = this->iAdvFB[i].FactorBase[j].p;
		//	f[j].r = this->iAdvFB[i].FactorBase[j].r;
		//	f[j].logp = this->iAdvFB[i].FactorBase[j].logp;
		//	f[j].flags = this->iAdvFB[i].FactorBase[j].flags;
		//}

		//Utils::WriteDataToFile("lattice-mbase-100-adv-fb" + Utils::ConvertToString(i + 1), f, this->iAdvFB[i].Count, sizeof(lattice_type2));
		//free(f);
		//f = NULL;

		rc = this->iAdvFB[i].Init(this->iLargishPrimesThreshold[i]);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}

/*!
	This method find all free relations and save them to the smooth relation file.
	We suppose that this file doesn't exist - it shouldn't, so we delete it if there is some one.

	The free relations is a factorization of the ideal (p), where p is prime, for which 
	f_{1} mod p and f_{2} mod p fully factorize to linear factors.

	For finding we use FBs, where we are looking for prime ideals with same prime. 
	If we find d (degree of sieving poly) of them, than we know that f mod p is fully factorized (in both bases).
*/
int CLatticeSievingPhase::FindFreeRelations()
{

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182024", "  Finding free relation ...\n");
	}
#endif

	int k;
	int degrees[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	fb_element * fb[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	fb_element * end[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];
	for (k = 0; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
	{
		degrees[k] = this->iNumberFields[k].SievingPoly->GetDegree();
		fb[k] = this->iFBs[k].FactorBase;
		end[k] = this->iFBs[k].FactorBase + this->iFBs[k].FBMaxIndex;
	}

	CNFSRelation relation(EDoNotUseVariations);
	
	int rc;
	int i;
	int count = 0;
	main_sieving_type prime = 0;
	bool good = false;
	int lastWrittenIndex = -1;
	unsigned int flag = 0;

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// prepare file - file will be deleted if there is some one
	FILE * smoothsFILE = fopen(this->iSmoothRelFullFileName.c_str(), FILE_MODE_WRITE);
	if (smoothsFILE == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","FindFreeRelations",0,0,"Could not open file with smooth relations.");
		return ConstRC::CouldNotOpenFile;
	}
#endif

	while (fb[0] < end[0])
	{
		prime = fb[0]->prime;
		count = 1;
		fb[0]++;
		while (fb[0]->prime == prime && prime != fb[0]->root)
		{
			fb[0]++;
			count++;
		}

		// we find factorization of ideal (p) in FB1
		if (count == degrees[0])
		{
			good = true;
			for (k = 1; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
			{
				// find position before prime ideals above p
				while (fb[k] < end[k] && fb[k]->prime < prime)
				{
					fb[k]++;
				}
				count = 0;
				// count number of prime ideals above p in FBk
				while (fb[k]->prime == prime && prime != fb[k]->root)
				{
					fb[k]++;
					count++;
				}
				if (count != degrees[k])
				{
					good = false;
					break;
				}
			}

			if (good)
			{
				// setup relation - we want just ideal (p)
				relation.Reset();
				// we need to set 0 relation count - we never use leading coeff
				relation.ZeroizeNumberOfRelations();
				relation.A = prime;
				relation.B = 0;

				for (k = 0; k < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; k++)
				{
					fb[k]--; // go back 
					relation.Get_Parts(k)->Poly->Set(0, prime);
					mpz_ui_pow_ui(relation.Get_Parts(k)->Norm, (unsigned long)prime, degrees[k]);
					mpz_set_ui(relation.Get_Parts(k)->NormRemaining, 1);
					if (fb[k]->flags & ConstEF::SPECIAL_PRIME)
					{
						// we note that this is special prime and we save its residual which has to be 1
						flag = ConstEF::SPECIAL_PRIME | (unsigned int)1;
						// TODO check
						//if (fb[k]->flags & ConstEF::DIVIDES_L)
						//{
						//	// if this prime divides the leading coeff than we save exact valuation which is still 1
						//	flag |= ConstEF::DIVIDES_L | ((unsigned int)1 << ConstEF::F_VALUATION_SHIFT);
						//}
					}
					else
					{
						flag = 0;
					}

					// set prime ideals divisors - we go through found prime ideals with same prime
					lastWrittenIndex = -1;
					for (i = - degrees[k] + 1; i <= 0; i++)
					{
						rc = relation.Get_Parts(k)->Main.AddDivisor(prime, 1, fb[k][i].root, flag, lastWrittenIndex);
						if (rc != ConstRC::Ok)
						{
#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
							fclose(smoothsFILE);
							smoothsFILE = NULL;
#endif
							return rc;
						}
					}

					relation.Type |= (CRelationPart::SMOOTH << (ARelation::PART_TYPE_SHIFT * k));
				}

				relation.Type |= RELATION_FREE;
				relation.Index = this->iStats->Get_Total(ESSFF);
				this->iStats->Increase(ESSFF);

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
				this->iStats->Increase(ESSTested);
				this->iStats->Increase(ESSAccepted);
#endif

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
				rc = relation.PrintToFile(smoothsFILE);
				if (rc != ConstRC::Ok)
				{
					fclose(smoothsFILE);
					smoothsFILE = NULL;
					return rc;
				}
#endif
			}
		}
	}

#ifndef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	fclose(smoothsFILE);
	smoothsFILE = NULL;
#endif

	return ConstRC::Ok;
}
int CLatticeSievingPhase::PrepareCycleTable()
{
	if (this->iUseCyclesForRelConstruction) 
	{	
		if (CNFSRelation::DEFAULT_NFS_PARTS_COUNT != ConstSPhase::ELEMENT_MASK_COUNT)
		{
			CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareCycleTable",0,0,"Relation size mismatch.");
			return ConstRC::GeneralError;
		}
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
		if (Get_InfoMode())
		{
			CLog::Log(100, "0182025", "  Preparing cycle table ...\n");
		}
#endif
		
		this->iLPHashtable = new CLPType1Hashtable();
		this->iLPHashtable->SetupHashtableFB(max(this->iFBs[0].FBMaxIndex, this->iFBs[1].FBMaxIndex));

		// Special case: PRIME=1, root of its own component.
		this->iStats->Increase(ESSVertices);
		this->iStats->Increase(ESSComponents);
		//this->vertices = 1;
		//this->components = 1;
		//this->edges = 0;

		this->iIntermediateRoots.reserve(ConstSPhase::INTERMEDIATE_ROOTS_COUNT);
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::AddToCycleCountStructures(CNFSRelation* aRelation) 
{
	int oldCycles = IndependentCyclesCount();
	int rc;
	long roots[ConstSPhase::ELEMENT_MASK_COUNT];
	long index[ConstSPhase::ELEMENT_MASK_COUNT];
	bool present = false;
	int  searched;
	int  root_fingerprint;

	bool enlarged = false;
	bool recalculate = false;
	for (int j = 0; j < ConstSPhase::ELEMENT_MASK_COUNT*2; j++)
	{
		int i = j % ConstSPhase::ELEMENT_MASK_COUNT;
		if (aRelation->Type & (CRelationPart::SMOOTH << (ARelation::PART_TYPE_SHIFT * i)))
		{
			searched = 1;	
			root_fingerprint = 0;
		} 
		else 
		{
			searched = aRelation->Get_Parts(i)->Large.Elements[0].prime;
			root_fingerprint = aRelation->Get_Parts(i)->Large.Elements[0].root;
			root_fingerprint |= ConstSPhase::ELEMENT_MASK[i];
		}

		rc = this->iLPHashtable->FindPrimeInHashtable(searched, root_fingerprint, present, index[i], enlarged);
		if (enlarged) recalculate = true; 
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
		if (!present) 
		{
			// This prime has not been in hashtable yet.					
			this->iLPHashtable->Hashtable[index[i]].SetValue(searched);
			this->iLPHashtable->Hashtable[index[i]].SetRootFingerprint(root_fingerprint);
			this->iLPHashtable->Hashtable[index[i]].SetAncestor(index[i]);
			this->iLPHashtable->Added();
			this->iStats->Increase(ESSVertices);
			this->iStats->Increase(ESSComponents);
		} 

		if (ConstSPhase::ELEMENT_MASK_COUNT-1==j && !recalculate) break;
	}

	for (int i = 0; i < ConstSPhase::ELEMENT_MASK_COUNT; i++)
	{
		// Finding roots of components
		roots[i] = this->iLPHashtable->FindAndResetRoot(index[i], this->iIntermediateRoots);
	}

	this->iStats->Increase(ESSEdges);

	if (roots[0] != roots[1]) 
	{			
//		std::cout << ":Spojuji:"<<iLPHashtable->Hashtable[roots[0]].GetValue() << ":"<<iLPHashtable->Hashtable[roots[1]].GetValue()<<":";
		// The component with the larger prime at root is joined to the other one
		if (this->iLPHashtable->Hashtable[roots[0]].GetValue() > this->iLPHashtable->Hashtable[roots[1]].GetValue()) 
		{
			this->iLPHashtable->Hashtable[roots[0]].SetAncestor(roots[1]);	
		} 
		else 
		{
			this->iLPHashtable->Hashtable[roots[1]].SetAncestor(roots[0]);	
		}
		this->iStats->Decrease(ESSComponents);
	}

	if (IndependentCyclesCount() != oldCycles) 
	{
		aRelation->Type |= RELATION_FINISHING_CYCLE;
	}

	return ConstRC::Ok;
}


int CLatticeSievingPhase::IndependentCyclesCount()
{
	return this->iStats->Get_Total(ESSEdges) + this->iStats->Get_Total(ESSComponents) - this->iStats->Get_Total(ESSVertices);
}
int CLatticeSievingPhase::PrepareForSieving()
{
	// We calculate the maximal partial factor (in case of Single and Double LPV).
	int maxx = 0;
	int i;

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iVariationsInfo[i].Prepare(this->iFBs[i].UpperFBBound, 
			this->iPartialFactorMultiplier[i], ConstSPhase::LOGARITHM_BASE);
		maxx = max(maxx, (int)this->iVariationsInfo[i].Type);
	}

	int rc;
	int blockSize = Get_BlockSize();
	this->iRelation = new CNFSRelation(maxx);
	rc = this->iFactorsInfo.Init(blockSize, maxx);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	this->iRelationBuffer = new vector<CNFSRelation*>;
	this->iRelationBuffer->reserve(ConstSPhase::RELATION_BUFFER_RESERVE);
	this->iRelationWorkingBuffer = new vector<CNFSRelation*>;
	this->iRelationWorkingBuffer->reserve(ConstSPhase::RELATION_BUFFER_RESERVE);

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iSievingBlocks[i] = (log_type*) Utils::AllocationMachine((void**) &this->iSievingBlocks[i], blockSize * sizeof(log_type));
		if (this->iSievingBlocks[i] == NULL)
		{
			CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",0,blockSize*sizeof(log_type),"Could not allocate space for internal sieving block");
			return ConstRC::NotEnoughMemory;
		} 
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
		else 
		{
			if (Get_InfoMode())
			{
				CLog::Log(5000, "", "\n");
				CLog::Log(5000, "0182902", "  Sieving block nbr %d is allocated to %d bytes.", i, (int)(blockSize * sizeof(log_type)));
			}
		}
#endif
		memset(this->iSievingBlocks[i], 0, blockSize * sizeof(log_type));
	}

	this->iFactorBlock = (log_type*) Utils::AllocationMachine((void**) &this->iFactorBlock, this->iFactorBlockSize * sizeof(log_type));
	if (this->iFactorBlock == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",0,this->iFactorBlockSize*sizeof(log_type),"Could not allocate space for factor block");
		return ConstRC::NotEnoughMemory;
	} 
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
	else 
	{
		if (Get_InfoMode())
		{
			CLog::Log(5000, "", "\n");
			CLog::Log(5000, "0182903", "  Factor block hashtable is allocated to %d bytes.\n", (int)(this->iFactorBlockSize * sizeof(log_type)));
		}
	}
#endif

	memset(this->iFactorBlock, 0, this->iFactorBlockSize * sizeof(log_type));
	this->iFactorBlockMask = this->iFactorBlockSize - 1;
	long auxLong;
	this->iFactorBlockShift = Get_LogBlockSize() - Utils::GetClosestPowerOf2(auxLong, this->iFactorBlockSize);

	if (this->iLatticeInfo.I <= 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",this->iLatticeInfo.I,0,"Value of line upper bound must be positive!");
		return ConstRC::InconsistentState;
	}
	if (this->iLatticeInfo.I >> 1 > ConstSPhase::LATTICE_MAX_HALF_INTERVAL_LENGTH)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",this->iLatticeInfo.I,ConstSPhase::LATTICE_MAX_HALF_INTERVAL_LENGTH,"Value of line upper bound too big - must not exceed a particular value!");
		return ConstRC::InconsistentState;
	}

	if (this->iLatticeInfo.I >= blockSize)
	{
		this->iLatticeInfo.J = this->iLatticeInfo.I >> 1;
	}
	else
	{
		// IJ has to be exactly divisible by block size
		int n = blockSize >> this->iLogI;
		this->iLatticeInfo.J = this->iLatticeInfo.I >> 1;
		this->iLatticeInfo.J = this->iLatticeInfo.J + ((-this->iLatticeInfo.J) % n);
	}

	mpz_set_si(this->iAuxMpz1, this->iLatticeInfo.I);
	mpz_mul_si(this->iAuxMpz2, this->iAuxMpz1, this->iLatticeInfo.J);
	mpz_divexact_ui(this->iAuxMpz1, this->iAuxMpz2, blockSize);
	this->iBlockCount = mpz_get_si(this->iAuxMpz1);

	int hashtableSize = Get_HashtableSize();

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
	if (Get_InfoMode())
	{
		CLog::Log(5000, "", "\n");
		CLog::Log(5000, "0182904", "  Sieving region: I = %d, J = %d.\n", this->iLatticeInfo.I, this->iLatticeInfo.J);
		CLog::Log(5000, "0182905", "  Number of blocks: %d\n", this->iBlockCount);
		CLog::Log(5000, "0182906", "  Block size: %d\n", blockSize);
		CLog::Log(5000, "0182907", "  Hashtable size: %d\n", hashtableSize);
		CLog::Log(5000, "0182908", "  Factor block size: %d\n", this->iFactorBlockSize);
	}
#endif

	long upper = 0;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		upper = CIntegerClass::Max(upper, this->iFBs[i].UpperFBBound);
	}
	if (upper <= 0)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",upper,0,"Value of factor base upper bound must be positive.");
		return ConstRC::InconsistentState;
	}
	if (upper > ConstSPhase::MAX_FACTOR_BASE_UPPER_BOUND)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",upper,ConstSPhase::MAX_FACTOR_BASE_UPPER_BOUND,"Value of factor base upper bound too big - must not exceed a particular value!");
		return ConstRC::InconsistentState;
	}

	mpf_t tt, one, prime, r;
	mpf_init(tt);
	mpf_init(one);
	mpf_init(prime);
	mpf_init(r);
	lattice_type2 * element;

	int hsize = 0;
	int sumhsize = 0;
	int tts[CNFSRelation::DEFAULT_NFS_PARTS_COUNT];

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		tts[i] = 0;
		if (this->iAdvFB[i].LargishFB != NULL)
		{
			mpf_set_ui(one, 1);
			mpf_set_ui(tt, 0);

			element = this->iAdvFB[i].LargishFB;
			while (element->p) // WARNING field must end with zero element
			{
				mpf_set_si(prime, element->p);
				mpf_div(r, one, prime);
				mpf_add(tt, tt, r);
				element++;
			}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
			//if (Get_InfoMode())
			//{
				//CLog::Log(5000, "", "\n\n");
				//CLog::Log(5000, "0182909", "INDEX %d\n", i + 1);
				//CLog::Log(5000, "0182910", "  Middle/Largish prime threshold: %d\n", this->iLargishPrimesThreshold[i]);

				//CLog::Log(5000, "0182911", "  Sum of 1/p: %Ff\n", tt);
			//}
#endif

			// probable number of large primes in sieving block
			mpf_mul_ui(tt, tt, blockSize);

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
			//if (Get_InfoMode())
			//{
				//CLog::Log(5000, "0182912", "  Prob. number of primes: %Ff\n", tt);
			//}
#endif
			
			tts[i] = mpf_get_si(tt) * ConstSPhase::PROPAB_PRIMES_SUM_RATIO;
			// size of hashtable with given number of primes (not exact - complex structure is used)
			hsize = sizeof(lattice_update2_t) * mpf_get_si(tt);
			sumhsize += hsize;
#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
			if (Get_InfoMode())
			{
				CLog::Log(5000, "0182913", "  Prob. number of primes - less - 95: %d\n", tts[i]);
				CLog::Log(5000, "0182914", "  Size of update2: %d\n", (int)sizeof(lattice_update2_t));
				CLog::Log(5000, "0182915", "  Hashtable block size: %d\n", hsize);
				CLog::Log(5000, "0182916", "  Max hashtable block set size: %d\n", hashtableSize / hsize);
				CLog::Log(5000, "0182917", "  Number of Middle primes: %d\n", this->iAdvFB[i].LastMiddlePrimes - this->iAdvFB[i].MiddlePrimes + 1);
				CLog::Log(5000, "0182918", "  Number of Largish primes: %d\n", this->iAdvFB[i].LastLargishPrimes - this->iAdvFB[i].LargishPrimes + 1);
			}
#endif
#endif
		}
	}

	if (sumhsize != 0)
	{
		this->iBlockSetSize = hashtableSize / (sumhsize * ConstSPhase::PROPAB_PRIMES_SUM_RATIO);// this->iBlockCount;
	}
	else
	{
		this->iBlockSetSize = 1; // special case
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
	if (Get_InfoMode())
	{
		CLog::Log(5000, "", "\n");
		CLog::Log(5000, "0182919", "  Block set size: %d\n", this->iBlockSetSize);
		CLog::Log(5000, "0182920", "  Block set count: %d\n", this->iBlockCount / this->iBlockSetSize);
		CLog::Log(5000, "0182920", "  blockSize: %d\n", blockSize);
		CLog::Log(5000, "0182920", "  Get_LogBlockSize: %d\n", Get_LogBlockSize());
		CLog::Log(5000, "0182920", "  Get_ModBlockSize: %d\n", Get_ModBlockSize());
	}
#endif

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
	if (Get_InfoMode()) CLog::Log(5000, "0172901", "\n  Relation reserve is set to %d relations.", this->iRelationReserve);
#endif

	mpf_clear(tt);
	mpf_clear(one);
	mpf_clear(r);
	mpf_clear(prime);

	if (blockSize < this->iLatticeInfo.I)
	{
		// cannot be
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","PrepareForSieving",0,0,"Block size must be bigger than sieving interval I!");
		return ConstRC::InconsistentState;
	}
	else
	{
		this->iLinesInBlock = blockSize / this->iLatticeInfo.I;
		this->iLinesInBlockSet = this->iLinesInBlock * this->iBlockSetSize;
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		// iBlockSetSize is number of blocks in block set - it is also the count of hashtables
		rc = this->iHashtables[i].SetupHashtables(this->iBlockSetSize, tts[i]);
		if (rc != ConstRC::Ok)	
		{
			return rc;
		}
		this->iHashtables[i].SetBlockSize(blockSize, Get_LogBlockSize(), Get_ModBlockSize());

		// we set also the change limits for ThresholdOptimizers
		this->iOptimizer[i].Set_MinChange(ConstSPhase::OPTIMIZER_MIN_CHANGE);
		this->iOptimizer[i].Set_MaxChange(ConstSPhase::OPTIMIZER_MAX_CHANGE);
	}

	if (this->iSpecialQ == 0)
	{
		mpz_set_si(this->iAuxMpz1, upper);
		for (i = 0; i < iFirstSpecialQIndex; i++) mpz_nextprime(this->iAuxMpz1, this->iAuxMpz1);
		this->iSpecialQ = mpz_get_si(this->iAuxMpz1);
		this->iRootIndex = 0;
	}

	DetermineRequestedRelations();

	if (this->iUseHighVariations && this->iHighVarHashtable == NULL) 
	{
		this->iHighVarHashtable = new CFrequencyHashtable();
		this->iHighVarHashtable->SetupHashtableProb(this->iRequestedRelations << 2);
	}

	return rc;
}
/*!
	Gauss base reduction (like LLL for 2-dim lattice) - we use weighted scalar product
	more Cohen - A Course in Computational Algebraic Number Theory - algorithm 1.3.14

	Special case by Jens Franke - from GGNFS.
*/
void CLatticeSievingPhase::ReduceBase(main_sieving_type & outA0, main_sieving_type & outB0, 
							   main_sieving_type & outA1, main_sieving_type & outB1, 
							   main_sieving_type inA0, main_sieving_type inB0, 
							   main_sieving_type inA1, main_sieving_type inB1, float aSigma)
{
	float a0sq, a1sq, s;
	main_sieving_type k;

	a0sq = ((float) inA0) * inA0;
	a0sq += aSigma * ((float) inB0) * inB0;
	a1sq = ((float) inA1) * inA1;
	a1sq += aSigma * ((float) inB1) * inB1;

	while (true) 
	{
		s = ((float) inA0) * inA1;
		s += aSigma * ((float) inB0) * inB1;

		if (a0sq < a1sq) 
		{
			k = (main_sieving_type)floor(s / a0sq + 0.5);
			if (k == 0)
			{
				break;
			}

			inA1 -= k * inA0;
			inB1 -= k * inB0;
			a1sq = ((float) inA1) * inA1;
			a1sq += aSigma * ((float) inB1) * inB1;
		} 
		else 
		{
			k = (main_sieving_type)floor(s / a1sq + 0.5);
			if (k == 0)
			{
				break;
			}

			inA0 -= k * inA1;
			inB0 -= k * inB1;
			a0sq = ((float) inA0) * inA0;
			a0sq += aSigma * ((float) inB0) * inB0;
		}
	}

	if (inB0 < 0) 
	{
		inB0 = -inB0;
		inA0 = -inA0;
	}
	if (inB1 < 0) 
	{
		inB1 = -inB1;
		inA1 = -inA1;
	}
	if (a0sq > a1sq) 
	{
		outA0 = inA0;
		outB0 = inB0;
		outA1 = inA1;
		outB1 = inB1;
	} 
	else 
	{
		outA0 = inA1;
		outB0 = inB1;
		outA1 = inA0;
		outB1 = inB0;
	}
}

int CLatticeSievingPhase::Factorize2(int aBlockSetIndex, int aInnerIndex)
{
	int rc;
	int i;

	rc = FactorizePrepareRelations();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// Now, we have factored the relation with all factor base primes. But we
	// need to find potential large primes too.
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstFactorization + i);
#endif

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeFactorMiddleStart[i] = clock();
#endif

		rc = FactorizeWithMiddlePrimes(i);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeFactorMiddle[i] += (clock() - this->iTimeFactorMiddleStart[i]);
		this->iTimeFactorLargishStart[i] = clock();
#endif

		rc = FactorizeWithLargishPrimes(i, aInnerIndex);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeFactorLargish[i] += (clock() - this->iTimeFactorLargishStart[i]);
#endif

		rc = FactorizeDivide(i, aInnerIndex, aBlockSetIndex);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstFactorization + i);
#endif
	}

	rc = FactorizeProcessRelations();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::FactorizeWithMiddlePrimes(int aIndex)
{
#ifdef OLD_MIDDLEPRIMES_FACTORIZATION
	INT64 aux = 0;
	register int i;
	register int rc;
	register int countMax = this->iFactorsInfo.Count;
	register int prime;
	register int cr;
	CFactorInfo * factors = this->iFactorsInfo.Get_FactorInfo(aIndex);
	register int * rows = this->iFactorsInfo.Rows;
	register int * cols = this->iFactorsInfo.Cols;
	register int length = factors->Length;
	register int count = this->iFactorsInfo.Get_CurrentCount();
	CAdvLFBInfo * info = this->iAdvFB + aIndex;
	register int * primes = factors->Get_Primes(0); // get all primes
	register int * roots = factors->Get_Roots(0); // get all roots
	register int * rowPrimes;
	register int * rowRoots;

	lattice_type2* running = info->FactorBase;
	if (running->p == -1)
	{
		// skip -1
		running++;
	}
	lattice_type2* last = info->LastMiddleFB;	

	/* Now, the tiny primes and middle primes*/
	while (running <= last)
	{
		prime = running->p;

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			if (prime == DEBUG_SIEVING_PHASE_CRITICAL_PRIME && 
				running->r == DEBUG_SIEVING_PHASE_CRITICAL_ROOT)
			{
				Utils::Wait();
			}
#endif

		// we control if this prime is not special or divide L
		// then we control if the sublattice exists (must be cr >= 0 see CAdvFBHelper::FillLatticeFB) 
		if ((running->flags & ConstEF::LATTICE_NOT_FOR_SIEVING_MINI_FLAG) == 0 &&
			(cr = running->cr) >= 0)
		{
			for (i = 0, rowPrimes = primes, rowRoots = roots; i < count; i++, rowPrimes += length, rowRoots += length)
			{
				aux = rows[i] % prime;
				aux *= cr;
				if ((aux - cols[i]) % prime == 0)
				{
					(*rowPrimes)++; // primes in i-th row - first cell is count of primes in the row
					if (*rowPrimes < length)
					{
						*(rowPrimes + *(rowPrimes)) = prime;
						*(rowRoots + *(rowPrimes)) = running->r;
					}
					else
					{
						// problem = we have to extend the field - allocate it again and refill it
						rc = factors->ResizeFactorization(countMax);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
						//new length
						length = factors->Length;
						primes = factors->Get_Primes(0);
						roots = factors->Get_Roots(0);
						rowPrimes = primes + i * length;
						rowRoots = roots + i * length;
						*(rowPrimes + *(rowPrimes)) = prime;
						*(rowRoots + *(rowPrimes)) = running->r;
					}
					// very slow
					//rc = factors->SaveFactor(countMax, i, prime, running->r);
				}
			}
		}

		running++;
	}

	return ConstRC::Ok;

#else

	register int rc;
	register main_sieving_type offset;
	register main_sieving_type b0;
	register main_sieving_type b1;
	register main_sieving_type a;
	register main_sieving_type c;
	register main_sieving_type modI = this->iModI;
	register main_sieving_type i;
	register int index;
	register int blocksize = Get_BlockSize();
	register int prime;
	register int cr;
	register int aux;

	register int countMax = this->iFactorsInfo.Count;
	CFactorInfo * factors = this->iFactorsInfo.Get_FactorInfo(aIndex);
	register int count = this->iFactorsInfo.Get_CurrentCount();
	CAdvLFBInfo * info = this->iAdvFB + aIndex;
	register int * rows = this->iFactorsInfo.Rows;
	register int * cols = this->iFactorsInfo.Cols;

	lattice_type2* runningTiny = info->FactorBase;
	if (runningTiny->p == -1)
	{
		// skip -1
		runningTiny++;
	}
	lattice_type2* lastTiny = info->MiddleFB - 1;	

	/* Now, the tiny primes and middle primes*/
	while (runningTiny <= lastTiny)
	{
		prime = runningTiny->p;

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			if (prime == DEBUG_SIEVING_PHASE_CRITICAL_PRIME && 
				runningTiny->r == DEBUG_SIEVING_PHASE_CRITICAL_ROOT)
			{
				Utils::Wait();
			}
#endif

		// we control if this prime is not special or divide L
		// then we control if the sublattice exists (must be cr >= 0 see CAdvFBHelper::FillLatticeFB) 
		if ((runningTiny->flags & ConstEF::LATTICE_NOT_FOR_SIEVING_MINI_FLAG) == 0 &&
			(cr = runningTiny->cr) >= 0)
		{
			for (i = 0; i < count; i++)
			{
				aux = rows[i] % prime;
				aux *= cr;
				if ((aux - cols[i]) % prime == 0)
				{
#ifdef FACTOR_DEBUG
					CLog::Log(MAXIMAL_INT_VALUE, "0182026", "FactorizeWithMiddlePrimes: save factor - (%d,%d) - for [%d,%d]\n", prime, runningTiny->r, cols[i], rows[i]);
#endif
					rc = factors->SaveFactor(countMax, i, prime, runningTiny->r);
					if (rc != ConstRC::Ok)
					{
						return rc;
					}
				}
			}
		}

		runningTiny++;
	}

	lattice_update_t* running = info->MiddlePrimes;
	lattice_update_t* last = info->LastMiddlePrimes;

	/* Now, the middle primes*/
	while (running <= last)
	{
		offset = running->start;	

		if (offset >= 0) 
		{
			b0 = running->b0;
			b1 = running->b1 + ConstSPhase::LATTICE_FB_B1_CHANGE;
			a = running->a;
			c = running->c;

			while (offset < blocksize)
			{		
				if ((index = this->iFactorBlock[offset >> this->iFactorBlockShift]) > 0)
				//if ((index = this->iFactorBlock[offset & this->iFactorBlockMask]) > 0)
				{
					if ((index != ConstSPhase::FACTOR_BLOCK_COLLISION &&
						*(factors->Get_Roots(--index)) == offset) ||
						(index == ConstSPhase::FACTOR_BLOCK_COLLISION &&
						(index = this->iFactorsInfo.FindOffset(offset)) != ConstRC::NotFound))
					{
#ifdef FACTOR_DEBUG
						CLog::Log(MAXIMAL_INT_VALUE, "0182027", "FactorizeWithMiddlePrimes: save factor - (%d,%d) - for offset %d\n", running->element->p, running->element->r, offset);
#endif
						rc = factors->SaveFactor(countMax, index, running->element->p, running->element->r);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
					}
				}

				i = offset & modI;
				if (i >= b0)
				{
					offset += a; 
				}
				if (i < b1)
				{
					offset += c; 
				}
			} 

		} // else the root is invalid
		running++;
	} 

	return ConstRC::Ok;

#endif
}
int CLatticeSievingPhase::FactorizeWithLargishPrimes(int aIndex, int aHashtableIndex)
{
	register lattice_script_t* script = this->iHashtables[aIndex].Hashtables[aHashtableIndex].script;
	register lattice_update2_t *updates = NULL;
	register int rc;
	register int k;
	register int max;
	register int index;
	register int count = this->iFactorsInfo.Count;
	register main_sieving_type h;
	CFactorInfo * factors = this->iFactorsInfo.Get_FactorInfo(aIndex);

	while (script)
	{	
		max = script->num_used;
		updates = script->list;
		for (k = 0; k < max; k++, updates++)
		{
			if ((k & 7) == 0) //if (k % 8 == 0) 
			{
				PREFETCH(updates + 16);
			}

			h = updates->h;

#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
			if (updates->element->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME && 
				updates->element->r == DEBUG_SIEVING_PHASE_CRITICAL_ROOT)
			{
				Utils::Wait();
			}
#endif

			if ((index = this->iFactorBlock[h >> this->iFactorBlockShift]) > 0)
			//if ((index = this->iFactorBlock[h & this->iFactorBlockMask]) > 0)
			{
				if ((index != ConstSPhase::FACTOR_BLOCK_COLLISION &&
					*(factors->Get_Roots(--index)) == h) ||
					(index == ConstSPhase::FACTOR_BLOCK_COLLISION &&
					(index = this->iFactorsInfo.FindOffset(h)) != ConstRC::NotFound))
				{
#ifdef FACTOR_DEBUG
						CLog::Log(MAXIMAL_INT_VALUE, "0182028", "FactorizeWithLargishPrimes: save factor - (%d,%d) - for offset %d\n", updates->element->p, updates->element->r, h);
#endif
					rc = factors->SaveFactor(count, index, updates->element->p, updates->element->r);
					if (rc != ConstRC::Ok)
					{
						return rc;
					}
				}
			}
		}
		script = script->next;
	}

	return ConstRC::Ok;
}
/*!
	Prepare aux relations for new factorization.
	Relations will be cleared.
	New As, Bs and norms will be computed.
	If a % this->iSpecialQ == 0 && b % this->iSpecialQ == 0 then we divide by special q.
*/
int CLatticeSievingPhase::FactorizePrepareRelations()
{
	int rc;
	this->iFactorsInfo.ClearRelations();
	int count = this->iFactorsInfo.Get_CurrentCount();
	CNFSRelation ** curr = this->iFactorsInfo.Relations;

	int row, col;
	algebraic_coeff_type a, b;
	for (int i = 0; i < count; i++, curr++)
	{
		col = this->iFactorsInfo.Cols[i];
		row = this->iFactorsInfo.Rows[i];

		// TODO a and b could be bigger than MAXINT
		a = ((algebraic_coeff_type) col) * ((algebraic_coeff_type) this->iLatticeInfo.a0) + ((algebraic_coeff_type) row) * ((algebraic_coeff_type) this->iLatticeInfo.a1);
		b = ((algebraic_coeff_type) col) * ((algebraic_coeff_type) this->iLatticeInfo.b0) + ((algebraic_coeff_type) row) * ((algebraic_coeff_type) this->iLatticeInfo.b1);

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182029", "FactorizePrepareRelations: i = %d, j = %d\n", col, row);
		CLog::Log(50, "0182030", "FactorizePrepareRelations: a = %d, b = %d\n", a, b);
#endif
#endif

		if (a % this->iSpecialQ == 0 && b % this->iSpecialQ == 0)
		{
			// this doesn't change factorization - except divisibility of special q
			// but this doesn't occur often so we will always check if the norm is divisible by special q
			do
			{
				a = a / this->iSpecialQ;
				b = b / this->iSpecialQ;
			}
			while (a % this->iSpecialQ == 0 && b % this->iSpecialQ == 0);

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
			CLog::Log(50, "0182031", "FactorizePrepareRelations: after reduction - a = %d, b = %d\n", a, b);
#endif
#endif
		}

		rc = (*curr)->Setup(a, b, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iNumberFields);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::FactorizeDivide(int aIndex, int aInnerIndex, int aBlockSetIndex)
{
	CFactorInfo * factors = this->iFactorsInfo.Get_FactorInfo(aIndex);

	bool assertionMode = false;

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	assertionMode = Get_AssertionMode();
#endif

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182032", "FactorizeDivide: index = %d, block set index = %d, inner index = %d\n", aIndex, aInnerIndex, aBlockSetIndex);
#endif
#endif

	CNumberFieldInfo * nf = this->iNumberFields + aIndex;
	CAdvLFBInfo * fb = this->iAdvFB + aIndex;
	bool good;
	bool minusOne = fb->FactorBase->p == -1;
	bool divideByQ = aIndex == 1; // TODO better - global variable
	INT64 rem;
	int lastWrittenIndex = -1;
	int rc;
	int j, maxBadPrimes = fb->BadPrimes.size();
	int maxSpecialPrimes = nf->Domain->SpecialPrimesCount;
	int maxSpecialIdeals = nf->SpecialIdeals.size();
	int count = this->iFactorsInfo.Get_CurrentCount(); // number of candidates
	int exponent;
	int prime;
	unsigned int flag;
	log_type flags;
	long valuation;
	int * primes = NULL;
	int * roots = NULL;
	int length; // number of factors
	CNFSRelation ** curr = this->iFactorsInfo.Relations;
	CDivisors *divisors;
	CRelationPart *part;
	mpz_t* remaining;
	SpecialPrimeIt it;
	PrimeIdeal * ideal = NULL;

	for (int i = 0; i < count; i++, curr++)
	{
#ifdef FACTOR_DEBUG
		CLog::Log(MAXIMAL_INT_VALUE, "0182033", "FactorizeDivide: i = %d\n", i);
		CLog::Log(MAXIMAL_INT_VALUE, "0182034", "FactorizeDivide: [a,b] = [%d,%d]\n", (*curr)->A, (*curr)->B);
#endif
		primes = factors->Get_Primes(i);
		roots = factors->Get_Roots(i);
		length = *primes;

		part = (*curr)->Get_Parts(aIndex);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		this->iStats->Increase(ESSAverageSizeIP_log2 + aIndex, mpz_sizeinbase(part->Norm, 2));
		this->iStats->Increase(ESSAverageSizeIP_log10 + aIndex, mpz_sizeinbase(part->Norm, 10));
#endif

		if (aIndex > 0) 
		{
			// this is not first part factorization
			// we check if it is worth to continue - some part's remainder is not too large
			if ((*curr)->Type & ARelation::TOO_LARGE_MASK)
			{
#ifdef FACTOR_DEBUG
				CLog::Log(MAXIMAL_INT_VALUE, "0182035", "FactorizeDivide: too large - continue\n");
#endif
				// remaining factor is too large, we don't have to continue with this relation 
				continue;
			}
		}

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeFactorDivideSpecialStart[aIndex] = clock();
#endif

		lastWrittenIndex = -1;
		divisors = &part->Main;
		remaining = &part->NormRemaining;

		// if it is integral part we control if norm is negative
		if (minusOne)
		{
			if (mpz_cmp_si(*remaining, 0) < 0) 
			{
				prime = -1;
				exponent = 1;

#ifdef FACTOR_DEBUG
				CLog::Log(MAXIMAL_INT_VALUE, "0182036", "FactorizeDivide: reducing %Zd by (%d,%d)^%d.\n", *remaining, -1, 0, 1);
#endif
				rc = divisors->AddDivisor(prime, exponent, 0, 0, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok) 
				{ 
					return rc; 
				}              		
				mpz_abs(*remaining, *remaining);
			}
		}

		if (divideByQ)
		{
#ifdef FACTOR_DEBUG
			CLog::Log(MAXIMAL_INT_VALUE, "0182037", "FactorizeDivide: Q - reducing %Zd by (%d,%d).\n", *remaining, this->iSpecialQ, this->iSpecialR);
#endif

			exponent = 0;

			while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, this->iSpecialQ))
			{
				exponent++;
				mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
			this->iStats->Increase(ESSDivrem);
#endif

#ifdef FACTOR_DEBUG
			CLog::Log(MAXIMAL_INT_VALUE, "0182037", "FactorizeDivide: Q - reduced %Zd, exponent %d.\n", *remaining, exponent);
#endif

			rc = divisors->AddDivisor(this->iSpecialQ, exponent, this->iSpecialR, 0, lastWrittenIndex, assertionMode);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		primes++;
		roots++;
		for (j = 0; j < length; j++, primes++, roots++)
		{
#ifdef FACTOR_DEBUG
				CLog::Log(MAXIMAL_INT_VALUE, "0182038", "FactorizeDivide: reducing %Zd by (%d,%d).\n", *remaining, *primes, *roots);
#endif

			exponent = 0;
			
			while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, *primes))
			{
				exponent++;
				mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
			this->iStats->Increase(ESSDivrem);
#endif

			if (exponent) 	
			{
#ifdef FACTOR_DEBUG
				CLog::Log(MAXIMAL_INT_VALUE, "0182038", "FactorizeDivide: reduced %Zd, exponent %d.\n", *remaining, exponent);
#endif
				rc = divisors->AddDivisor(*primes, exponent, *roots, 0, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok) 
				{ 
					return rc; 
				}            
			} 
			else
			{
				CLog::Log(MAXIMAL_INT_VALUE, "0182039", "FactorizeDivide: error - reduction of %Zd by %d impossible (r = %d): exponent %d\n",*remaining,*primes,*roots,exponent);
			}
		}

		// divide with bad primes - primes for which we cannot construct sublattice - (b0 * r - a0) % p == 0
		if (maxBadPrimes > 0)
		{
			for (j = 0; j < maxBadPrimes; j++)
			{
#ifdef FACTOR_DEBUG
						CLog::Log(MAXIMAL_INT_VALUE, "0182040", "FactorizeDivide: B - reducing %Zd by (%d,%d).\n", *remaining, fb->BadPrimes.at(j).p, fb->BadPrimes.at(j).r);
#endif

				prime = fb->BadPrimes.at(j).p;
				rem = fb->BadPrimes.at(j).r;
				if (((*curr)->A - (rem = ((*curr)->B % prime) * rem)) % prime == 0)
				{
					exponent = 0;
					
					while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
						CLog::Log(MAXIMAL_INT_VALUE, "0182040", "FactorizeDivide: B - reduced %Zd, exponent %d.\n", *remaining, exponent);
#endif
						rc = divisors->AddDivisor(prime, exponent, fb->BadPrimes.at(j).r, 0, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok) 
						{ 
							return rc; 
						}            
					} 
					else
					{
#ifdef FACTOR_DEBUG
						CLog::Log(MAXIMAL_INT_VALUE, "0182041", "\nFactorizeDivide: reduction of %Zd by %d impossible: exponent %d",*remaining,prime,exponent);
#endif
					}
				}
			}
		} // end if - bad primes

		// factorize with special ideals
		if (maxSpecialPrimes > 0)
		{
			good = false;
			// firstly - we find out if the remaining is low enough
			// and save exponents for other computing...
			for (j = 0; j < maxSpecialPrimes; j++)
			{
				exponent = 0;
				prime = nf->Domain->SpecialPrimes[j];
				while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
				{
					exponent++;
					good = true;
					mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
				}
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
				nf->SpecialPrimes[prime] = exponent;

				if (exponent) {
					// add divisor for rational part
					if (nf->SievingPoly->GetDegree() == 1) {
						if ((rc = divisors->AddDivisor(prime, exponent, 0, ConstEF::SPECIAL_PRIME, lastWrittenIndex, assertionMode)) != ConstRC::Ok) return rc;
					}
				}
			}

			// algebraic part
			if (nf->SievingPoly->GetDegree() > 1) {
				// is it worth to continue?
				if (good && this->iVariationsInfo[aIndex].Control(*remaining))
				{
					// find base of ideal (c_{d}a + b\bar{\theta})
					if (nf->SievingPolyMonic)
					{
						nf->Domain->ABaseOfPrincipal(nf->IdealBase, (*curr)->A, (*curr)->B);
					}
					else
					{
						// non-monic poly
						mpz_mul_si(this->iAuxMpz2, *nf->SievingPoly->Leading2(), (*curr)->A);
						nf->Domain->ABaseOfPrincipal(nf->IdealBase, this->iAuxMpz2, (*curr)->B);
					}
	 
					// we find out exact factorization
					for (j = 0; j < maxSpecialIdeals; j++)
					{
						ideal = nf->SpecialIdeals[j];
						prime = ideal->Prime;
						it = nf->SpecialPrimes.find(prime);
						flags = fb->FactorBase[ideal->Index].flags;
						if ((*it).second > 0 || flags & ConstEF::DIVIDES_L_MINI_FLAG)
						{
							rc = nf->Domain->ValuationOfPrime(&valuation, ideal, nf->IdealBase);
							if (rc != ConstRC::Ok)
							{
								return rc;
							}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
							//CLog::Log(50, "0182042", "Valuation of ideal with prime %d is %ld.\n",prime,valuation);
#endif
							exponent = valuation;
							if (exponent > 0) 
							{
								// we note that this is special prime and we save its residual
								flag = ConstEF::SPECIAL_PRIME | ideal->Residual;
								if (flags & ConstEF::DIVIDES_L_MINI_FLAG)
								{
									valuation = (*it).second;
									nf->SpecialPrimes[prime] = 0;
									// if this prime divides the leading coeff than we save exact valuation of F(a,-b)
									flag |= ConstEF::DIVIDES_L | ((unsigned int)valuation << ConstEF::F_VALUATION_SHIFT);
									// we also save this prime valuation of the leading coeff
									flag |= ((flags & ConstEF::L_VALUATION_MINI_MASK) << ConstEF::L_VALUATION_SHIFT);
								}
								else
								{
									// norm of prime ideal is p^{residual} so if valuation > 0 than F(a,-b) is divisible by p^{valuation*residual}
									valuation *= ideal->Residual;

									nf->SpecialPrimes[prime] = (*it).second - valuation;
								}

#ifdef FACTOR_DEBUG
								CLog::Log(MAXIMAL_INT_VALUE, "0182043", "FactorizeDivide: reducing %Zd by (%d,%d)^%d.\n", *remaining, prime, ideal->Cp, exponent);
#endif
	
								rc = divisors->AddDivisor(prime, exponent, ideal->Cp, flag, lastWrittenIndex, assertionMode);
								if (rc != ConstRC::Ok)
								{
									return rc;
								}
						
							} // end if - exponent
	
						} // end if - used this special ideal

					} // end of for cycle - special ideals
			
				} // end if - continue?

			} // end if - algebraic part

		} // end if - special primes

#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
		this->iTimeFactorDivideSpecial[aIndex] += (clock() - this->iTimeFactorDivideSpecialStart[aIndex]);
#endif

		// setup optimizer - good/bad
		if (this->iVariationsInfo[aIndex].Control(*remaining) == false)
		{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
			CLog::Log(50, "0182044", "FactorizeDivide: Rejected - too big = %Zd\n", *remaining);
#endif
#endif
			// remainder is too big
			this->iOptimizer[aIndex].IncrementRejected();
			// save that is too big
			(*curr)->Type |= (CRelationPart::TOO_LARGE << (ARelation::PART_TYPE_SHIFT * aIndex));
		}
		else
		{
			// this part is good
			this->iOptimizer[aIndex].IncrementAccepted();
		}

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
		if (assertionMode)
		{
			if (mpz_cmp_ui(*remaining, 1) != 0)
			{
				// We will try all primes
				lattice_type2 * fb = this->iAdvFB[aIndex].FactorBase;
				if (minusOne)
				{
					fb++;
				}
				while (fb != NULL)
				{
					prime = fb->p;
					if (!prime) 
					{
						break;
					}
					if (mpz_mod_ui(this->iAuxMpz1, *remaining, prime) == 0)
					{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
						CLog::Log(50, "0182045", "\nFactorizeDivide: remaining %Zd is divisible by prime %d\n",*remaining,prime);
#endif
						// Search
						(*curr)->PrintToScreen();
						Utils::Wait();
						this->iHashtables[aIndex].SearchForPrimeInHashtable(prime, fb->r);
					}
					fb++;
				}			
			}
		}
#endif
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::FactorizeProcessRelations()
{
	int rc;
	unsigned int type;
	int j;
	int sievingStat;
	int count = this->iFactorsInfo.Get_CurrentCount();
	CNFSRelation ** curr = this->iFactorsInfo.Relations;
	for (int i = 0; i < count; i++, curr++)
	{
		// firstly check if some part's remainder is not too large
		if ((*curr)->Type & ARelation::TOO_LARGE_MASK)
		{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
			this->iStats->Increase(ESSRejected);
			this->iStats->Increase(ESSRemainderTooLarge);
#endif
			continue;
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTPartialFactorization);
#endif

		rc = (*curr)->DetermineType(CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iVariationsInfo);

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTPartialFactorization);
#endif

		if (rc == ConstRC::PrimeTooLarge) {
			this->iStats->Increase(ESSPrimeTooLarge);
			rc = ConstRC::Impropriate;
		};

		if (rc == ConstRC::FactorTooLarge) {
			this->iStats->Increase(ESSFactorTooLarge);
			rc = ConstRC::Impropriate;
		};

		if (rc == ConstRC::Ok)
		{
			type = (*curr)->Type;
			if ((type & ARelation::PARTIAL_MASK) == 0)
			{
				(*curr)->Index = this->iStats->Get_Total(ESSFF);
				this->iStats->Increase(ESSFF);
			}
			else
			{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
				if (this->iUseCyclesForRelConstruction)
				{
					if ((type & ARelation::P1_LARGE) && (type & ARelation::P2_SMOOTH))
					{
						this->iStats->Increase(ESSPF);
					}
					else if ((type & ARelation::P1_SMOOTH) && (type & ARelation::P2_LARGE))
					{
						this->iStats->Increase(ESSFP);
					}
					else
					{
						this->iStats->Increase(ESSPP);
					}
				}
				else
				{
					sievingStat = 0;
					for (j = 0; j < ARelation::MAX_SUPPORTED_PARTS; j++)
					{
						switch (type & ARelation::PART_TYPE_MASK)
						{
							case CRelationPart::LARGE:
								sievingStat += 1;
								break;
							case CRelationPart::DOUBLE_LARGE:
								sievingStat += 2;
								break;
							case CRelationPart::TRIPPLE_LARGE:
								sievingStat += 3;
								break;
						}
						type = type >> ARelation::PART_TYPE_SHIFT;
					}

					this->iStats->Increase(ESSPartialR + sievingStat);
				}
#endif
				(*curr)->Index = this->iStats->Get_Total(ESSPartialR);
				this->iStats->Increase(ESSPartialR);
				if (this->iUseCyclesForRelConstruction &&
					IndependentCyclesCount() < ConstSPhase::MAX_INDEPENDENT_CYCLES_COUNT)  
				{
					rc = AddToCycleCountStructures((*curr));	
					if (rc != ConstRC::Ok)
					{
						return rc;
					}
				}
			}

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
			this->iStats->Increase(ESSAccepted);
#endif
			// add prime ideals of relation to hashtable if high variation is used
			if (iUseHighVariations) 
			{
				for (j = 0; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++) 
				{
					rc = iHighVarHashtable->AddToTable((*curr)->Get_Parts(j), ConstSPhase::ELEMENT_MASK[j], this->iStats->Get_Total(ESSFF) + this->iStats->Get_Total(ESSPartialR));
					if (rc != ConstRC::Ok)
					{
						return rc;
					}
				}
			}

			if (this->iRelationWorkingBufferMaxCount < this->iRelationWorkingBuffer->size())
			{
				rc = (*curr)->Copy2(*this->iRelationWorkingBuffer->at(this->iRelationWorkingBufferMaxCount));
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
			else
			{
				this->iRelationWorkingBuffer->push_back(new CNFSRelation(*(*curr)));
			}
			this->iRelationWorkingBufferMaxCount++;
		}
		else if (rc == ConstRC::Impropriate)
		{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
			this->iStats->Increase(ESSRejected);
#endif
		}
		else
		{
			// error
			return rc;
		}
	}

	return ConstRC::Ok;
}

int CLatticeSievingPhase::AddToCountStructures(CNFSRelation *aRelation) {
	unsigned int type = aRelation->Type;
	int rc;

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182053", "Factorize: Good - Type = %u\n", type);
#endif
#endif
	if ((type & ARelation::PARTIAL_MASK) == 0)
	{
		aRelation->Index = this->iStats->Get_Total(ESSFF);
		this->iStats->Increase(ESSFF);
	}
	else
	{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		if (this->iUseCyclesForRelConstruction)
		{
			if ((type & ARelation::P1_LARGE) && (type & ARelation::P2_SMOOTH))
			{
				this->iStats->Increase(ESSPF);
			}
			else if ((type & ARelation::P1_SMOOTH) && (type & ARelation::P2_LARGE))
			{
				this->iStats->Increase(ESSFP);
			}
			else
			{
				this->iStats->Increase(ESSPP);
			}
		}
		else
		{
			int sievingStat = 0;
			for (int j = 0; j < ARelation::MAX_SUPPORTED_PARTS; j++)
			{
				switch (type & ARelation::PART_TYPE_MASK)
				{
					case CRelationPart::LARGE:
						sievingStat += 1;
						break;
					case CRelationPart::DOUBLE_LARGE:
						sievingStat += 2;
						break;
					case CRelationPart::TRIPPLE_LARGE:
						sievingStat += 3;
						break;
				}
				type = type >> ARelation::PART_TYPE_SHIFT;
			}

			this->iStats->Increase(ESSPartialR + sievingStat);
		}
#endif
		aRelation->Index = this->iStats->Get_Total(ESSPartialR);
		this->iStats->Increase(ESSPartialR);
		if (this->iUseCyclesForRelConstruction &&
			IndependentCyclesCount() < ConstSPhase::MAX_INDEPENDENT_CYCLES_COUNT)  
		{
			if ((rc = AddToCycleCountStructures(aRelation)) != ConstRC::Ok) return rc;
		}
	}

	// add prime ideals of relation to hashtable if high variation is used
	if (iUseHighVariations) 
	{
		for (int j = 0; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++) 
		{
			rc = iHighVarHashtable->AddToTable(aRelation->Get_Parts(j), ConstSPhase::ELEMENT_MASK[j], this->iStats->Get_Total(ESSFF) + this->iStats->Get_Total(ESSPartialR));
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
	this->iStats->Increase(ESSAccepted);
#endif

        return ConstRC::Ok;
}

int CLatticeSievingPhase::Factorize(main_sieving_type aOffset, int aBlockSetIndex, int aInnerIndex)
{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182046", "Factorize: Offset = %d, block set index = %d, inner index = %d\n", aOffset, aBlockSetIndex, aInnerIndex);
#endif
#endif
	int rc;
	int blockSize = Get_BlockSize();
	this->iRelation->Reset();

	// aBlockSetIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize + aOffset
	mpz_set_si(this->iAuxMpz1, aBlockSetIndex);
	mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, this->iBlockSetSize);
	mpz_add_ui(this->iAuxMpz1, this->iAuxMpz1, aInnerIndex);
	mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, blockSize);
	mpz_add_ui(this->iAuxMpz1, this->iAuxMpz1, aOffset);

	mpz_set_si(this->iAuxMpz2, this->iModI);
	mpz_and(this->iAuxMpz2, this->iAuxMpz2, this->iAuxMpz1);
	int col = mpz_get_si(this->iAuxMpz2) - (this->iLatticeInfo.I >> 1);

	mpz_fdiv_q_2exp(this->iAuxMpz2, this->iAuxMpz1, this->iLogI);
	int row = mpz_get_si(this->iAuxMpz2) + 1;

#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
	this->iStats->Increase(ESSTested);
#endif

	main_sieving_type result = 0;
	Utils::GCD(abs(col), abs(row), result);

	if (result != 1)
	{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		this->iStats->Increase(ESSRejected);
		this->iStats->Increase(ESSNotCoprime);
#endif
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182047", "Factorize: Rejected - GCD = %d\n", result);
#endif
#endif
		return ConstRC::Ok;
	}

	// TODO a and b could be bigger than MAXINT
	algebraic_coeff_type a = ((algebraic_coeff_type) col) * ((algebraic_coeff_type) this->iLatticeInfo.a0) + ((algebraic_coeff_type) row) * ((algebraic_coeff_type) this->iLatticeInfo.a1);
	algebraic_coeff_type b = ((algebraic_coeff_type) col) * ((algebraic_coeff_type) this->iLatticeInfo.b0) + ((algebraic_coeff_type) row) * ((algebraic_coeff_type) this->iLatticeInfo.b1);

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182048", "Factorize: i = %d, j = %d\n", col, row);
		CLog::Log(50, "0182049", "Factorize: a = %d, b = %d\n", a, b);
#endif
#endif

	bool divisible = true;
	if (a % this->iSpecialQ == 0 && b % this->iSpecialQ == 0)
	{
		// this doesn't change factorization - except divisibility of special q
		do
		{
			a = a / this->iSpecialQ;
			b = b / this->iSpecialQ;
		}
		while (a % this->iSpecialQ == 0 && b % this->iSpecialQ == 0);

		divisible = false;
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182050", "Factorize: after reduction - a = %d, b = %d\n", a, b);
#endif
#endif
	}

	this->iRelation->Setup(a, b, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iNumberFields);

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182051", "Factorize: [a,b] = [%d,%d]\n", this->iRelation->A, this->iRelation->B);
#endif
#endif
	

	int i;
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iStats->Increase(ESSAverageSizeIP_log2 + i, mpz_sizeinbase(this->iRelation->Get_Parts(i)->Norm, 2));
		this->iStats->Increase(ESSAverageSizeIP_log10 + i, mpz_sizeinbase(this->iRelation->Get_Parts(i)->Norm, 10));
	}
#endif

	// Now, we have factored the relation with all factor base primes. But we
	// need to find potential large primes too.
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->Start(ESTFirstFactorization + i);
#endif
		if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
		{
			rc = FactorizeAlgebraic(col, row, aOffset, &this->iHashtables[i].Hashtables[aInnerIndex], 
					divisible && i != 0, i); // TODO 
		}
		else
		{
			rc = FactorizeIntegral(col, row, aOffset, &this->iHashtables[i].Hashtables[aInnerIndex], i);
		}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
		this->iTimes->End(ESTFirstFactorization + i);
#endif
		if (rc != ConstRC::Ok)
		{
			return rc;
		}

		if (this->iVariationsInfo[i].Control(this->iRelation->Get_Parts(i)->NormRemaining) == false)
		{
			this->iOptimizer[i].IncrementRejected();

			// remaining factor is too big, we don't have to continue. 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
			this->iStats->Increase(ESSRejected);
#endif
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
			CLog::Log(50, "0182052", "Factorize: Rejected - too big = %Zd\n", this->iRelation->Get_Parts(i)->NormRemaining);
#endif
#endif
			return ConstRC::Ok;
		}
		// this part is good
		this->iOptimizer[i].IncrementAccepted();	
	}

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->Start(ESTPartialFactorization);
#endif

	rc = this->iRelation->DetermineType(CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iVariationsInfo);

#ifdef DEBUG_SIEVING_PHASE_TIME_MESSAGE_ON
	this->iTimes->End(ESTPartialFactorization);
#endif

	if (rc == ConstRC::Ok)
	{
		if ((rc = AddToCountStructures(iRelation)) != ConstRC::Ok) return rc;

		if (this->iRelationWorkingBufferMaxCount < this->iRelationWorkingBuffer->size())
		{
			rc = this->iRelation->Copy2(*this->iRelationWorkingBuffer->at(this->iRelationWorkingBufferMaxCount));
		}
		else
		{
			this->iRelationWorkingBuffer->push_back(new CNFSRelation(*this->iRelation));
		}
		this->iRelationWorkingBufferMaxCount++;

		return rc;
	}
	else if (rc == ConstRC::Impropriate)
	{
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_ON
		this->iStats->Increase(ESSRejected);
#endif
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182054", "Factorize: Rejected - after factorization\n");
#endif
#endif
		return ConstRC::Ok;
	}
	else
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::FactorizeIntegral(int aI, int aJ, main_sieving_type aOffset, lattice_hashtable_t * aHashtable, int aIndex)
{
	bool assertionMode = false;

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	assertionMode = Get_AssertionMode();
#endif

	int lastWrittenIndex = -1;
	int rc;
	int exponent;
	main_sieving_type prime;
	main_sieving_type cr;
	INT64 rem = 0;	

	CDivisors * divisors = &this->iRelation->Get_Parts(aIndex)->Main;
	mpz_t* remaining = &this->iRelation->Get_Parts(aIndex)->NormRemaining;

	lattice_type2* running = this->iAdvFB[aIndex].FactorBase + 1; // skip -1
	lattice_type2* moddivision_end = this->iAdvFB[aIndex].LastMiddleFB;	

	// The integral part
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182055", "FactorizeIntegral: remaining is %Zd.\n",*remaining);
#endif
#endif

	// At first, we will take care of -1.
	if (mpz_cmp_si(*remaining, 0) < 0) 
	{
		prime = -1;
		exponent = 1;
		rc = divisors->AddDivisor(prime, exponent, 0, 0, lastWrittenIndex, assertionMode);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}              		
		mpz_mul_si(*remaining, *remaining, -1);

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182056", "FactorizeIntegral: was negative, added -1 to list.\n");
#endif
#endif
	}

	// now we check prime = 2
	exponent = mpz_scan1(*remaining, 0);
	if (exponent)
	{
		prime = 2;
		rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
		if (rc != ConstRC::Ok) 
		{ 
			return rc; 
		}       

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182057", "FactorizeIntegral: Reduced %Zd by 2^%d.\n", this->iRelation->Get_Parts(aIndex)->Norm, exponent);
		//Utils::Wait();
#endif
#endif

		mpz_fdiv_q_2exp(*remaining, *remaining, exponent);				
	}
	running++; // next prime after 2

	/* Now, the tiny primes and middle primes*/
	while (running <= moddivision_end)
	{
		prime = running->p;
		if ((running->flags & ConstEF::LATTICE_NOT_FOR_SIEVING_MINI_FLAG) == 0 &&
			(cr = running->cr) >= 0)
		{
			// Compare the required 
			rem = aJ % prime;
			rem *= cr;
			if ((rem - aI) % prime == 0)
			{
				exponent = 0;
			
				while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
				{
					exponent++;
					mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
				} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif

				if (exponent) 	
				{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
					CLog::Log(50, "0182058", "FactorizeIntegral: reducing %Zd by %d^%d.\n", *remaining, prime, exponent);
#endif
#endif
					rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
					if (rc != ConstRC::Ok) 
					{ 
						return rc; 
					}            
				} 
				else
				{
					CLog::Log(MAXIMAL_INT_VALUE, "0182059", "\nINT FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, i = %d, j = %d, rem = %d, condition = %d\n",*remaining,prime,exponent,aOffset,aI,aJ,rem, (rem - aI) % prime);
				}		
			}
		}
////		else if ((this->iRelation->A - (rem = (this->iRelation->B % prime) * running->r)) % prime == 0)
////		{
////			// TODO stejna kontrola jako u ALG - zkontrolovat!!!
////			exponent = 0;
////			
////			while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
////			{
////				exponent++;
////				mpz_swap(*remaining, this->iAuxMpz1);
////#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
////				this->iStats->Increase(ESSDivrem);
////#endif
////			} 
////#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
////			this->iStats->Increase(ESSDivrem);
////#endif
////
////			if (exponent) 	
////			{
////				rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
////				if (rc != ConstRC::Ok) 
////				{ 
////					return rc; 
////				}            
////			} 
////			else
////			{
////				CLog::Log(MAXIMAL_INT_VALUE, "0182060", "\nINT FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, i = %d, j = %d\n",*remaining,prime,exponent,aOffset,aI,aJ);
////			}
////		}
		running++;
	}

	int k, max;
	// Now, the largish primes. We will walk through the current update table ...
	if (aHashtable)
	{
		lattice_update2_t *updates;
		lattice_script_t* script = aHashtable->script;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (k = 0; k < max; k++, updates++)
			{
				if (updates->h == aOffset)
				{
					prime = updates->element->p;
					exponent = 0;			
					while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
						CLog::Log(50, "0182061", "FactorizeIntegral: reducing %Zd by %d^%d.\n", *remaining, prime, exponent);
#endif
#endif
						rc = divisors->AddDivisor(prime, exponent, updates->element->r, 0, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok) 
						{ 
							return rc; 
						}            
					} 
					else 
					{
						CLog::Log(MAXIMAL_INT_VALUE, "0182062", "\nINT FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, updates[k].h = %d.\n",*remaining,prime,exponent,aOffset,updates->h);
					}			
				}
			}
			script = script->next;
		}
	}

	max = this->iAdvFB[aIndex].BadPrimes.size();
	if (max > 0)
	{
		int root;
		for (k = 0; k < max; k++)
		{
			prime = this->iAdvFB[aIndex].BadPrimes.at(k).p;
			if (prime == 2)
			{
				continue;
			}
			root = this->iAdvFB[aIndex].BadPrimes.at(k).r;
			if ((this->iRelation->A - (rem = (this->iRelation->B % prime) * root)) % prime == 0)
			{
				exponent = 0;
				
				while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
				{
					exponent++;
					mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
				} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif

				if (exponent) 	
				{
					rc = divisors->AddDivisor(prime, exponent, root, 0, lastWrittenIndex, assertionMode);
					if (rc != ConstRC::Ok) 
					{ 
						return rc; 
					}            
				} 
				else
				{
					CLog::Log(MAXIMAL_INT_VALUE, "0182063", "\nINT FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, i = %d, j = %d\n",*remaining,prime,exponent,aOffset,aI,aJ);
				}
			}
		}
	}

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (assertionMode)
	{
		if (mpz_cmp_ui(*remaining, 1) != 0)
		{
			// We will try all primes
			lattice_type2 * fb = this->iAdvFB[aIndex].FactorBase;
			fb++; // skip -1
			while (fb != NULL)
			{
				prime = fb->p;
				if (!prime) 
				{
					break;
				}
				if (mpz_mod_ui(this->iAuxMpz1, *remaining, prime) == 0)
				{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
					CLog::Log(50, "0182064", "\nFactorizeIntegral: remaining %Zd is divisible by prime %d\n",*remaining,prime);
#endif
					// Search
					this->iRelation->PrintToScreen();
					Utils::Wait();
					this->iHashtables[aIndex].SearchForPrimeInHashtable(prime, fb->r);
				}
				fb++;
			}			
		}
	}
#endif

	return ConstRC::Ok;
}
int CLatticeSievingPhase::FactorizeAlgebraic(int aI, int aJ, main_sieving_type aOffset, lattice_hashtable_t *aHashtable, bool aDivisible, int aIndex) 
{
	int rc;
	int lastWrittenIndex = -1;

	bool assertionMode = false;
#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	assertionMode = Get_AssertionMode();
#endif

	CNumberFieldInfo * nf = this->iNumberFields + aIndex;
	CDivisors * divisors = &this->iRelation->Get_Parts(aIndex)->Main;
	mpz_t * remaining = &this->iRelation->Get_Parts(aIndex)->NormRemaining;

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	CLog::Log(50, "0182065", "FactorAlgebraic: remaining is %Zd.\n",*remaining);
#endif
#endif

	lattice_type2* running = this->iAdvFB[aIndex].FactorBase;
	lattice_type2* moddivision_end = this->iAdvFB[aIndex].LastMiddleFB;

	INT64 aux;
	main_sieving_type exponent = 0, prime, rem;
	unsigned int flag;
	log_type flags;
	long valuation;

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	if (assertionMode)
	{
		if (aDivisible && mpz_divisible_ui_p(*remaining, this->iSpecialQ) == 0)
		{
			CLog::Log(50, "0182066", "\nALG FactorAlgebraic Error: reduction of %Zd by special q %d impossible! a = %d, b = %d\n", *remaining, this->iSpecialQ, this->iRelation->A, this->iRelation->B);
		}
	}
#endif
#endif

	if (aDivisible)
	{
		exponent = 1;
		mpz_divexact_ui(*remaining, *remaining, this->iSpecialQ);
	
		while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, this->iSpecialQ))
		{
			exponent++;
			mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
			this->iStats->Increase(ESSDivrem);
#endif
		} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
		this->iStats->Increase(ESSDivrem);
#endif

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		CLog::Log(50, "0182067", "FactorAlgebraic: Q - reducing %Zd by (%d,%d)^%d.\n", *remaining, this->iSpecialQ, this->iSpecialR, exponent);
#endif
#endif

		rc = divisors->AddDivisor(this->iSpecialQ, exponent, this->iSpecialR, 0, lastWrittenIndex, assertionMode);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	/* Now, the tiny primes and middle primes*/
	while (running <= moddivision_end)
	{
		prime = running->p;
		exponent = 0;
		//if (!(running->flags & ConstEF::SPECIAL_PRIME_MINI_FLAG) && !(running->flags & ConstEF::DIVIDES_L_MINI_FLAG)) 
		if (!(running->flags & ConstEF::LATTICE_NOT_FOR_SIEVING_MINI_FLAG))
		{
			if (running->cr >= 0) // TODO why?? --> because it is prime for which temp == 0 in CAdvFBHelper::FillLatticeFB
			{
				// Compare the required 
				aux = aJ % prime;
				aux *= running->cr;
				if ((aux - aI) % prime == 0)
				{
					while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					// Save the divisor if appropriate
					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
						CLog::Log(50, "0182068", "FactorAlgebraic: reducing %Zd by %d^%d.\n",*remaining,prime,exponent);
#endif
#endif
						rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
					} 
					else  // some problem has occurred
					{
						CLog::Log(MAXIMAL_INT_VALUE, "0182069", "\nALG FactorAlgebraic Error: reduction of %Zd by %d impossible: exponent %d, aOffset = %d.\n", *remaining, prime, exponent, aOffset);
					}	

				} // enf if (rem - aI) % prime == 0
			}
			// bad primes
//			else if ((this->iRelation->A - (aux = this->iRelation->B % prime) * running->r) % prime == 0)
//			{
//				while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
//				{
//					exponent++;
//					mpz_swap(*remaining, this->iAuxMpz1);
//#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
//					this->iStats->Increase(ESSDivrem);
//#endif
//				} 
//#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
//				this->iStats->Increase(ESSDivrem);
//#endif
//
//				// Save the divisor if appropriate
//				if (exponent) 	
//				{
//
//					rc = divisors->AddDivisor(prime, exponent, running->r, 0, lastWrittenIndex, assertionMode);
//					if (rc != ConstRC::Ok)
//					{
//						return rc;
//					}
//				} 
//			}

		} // end if - normal prime ideal
		else if ((running->flags & ConstEF::LATTICE_SPECIAL_PRIME_MINI_FLAG) == 0 && 
			(running->flags & ConstEF::LATTICE_DIVIDES_L_MINI_FLAG) && prime != running->r)
		{
			// we do two ideals at once

			// ideal (p) valuation of (c_{d}a + \bar{\theta}b)
			if (this->iRelation->B % prime == 0)
			{
				rem = this->iRelation->B;
				do
				{
					// TODO better
					exponent++;
					rem /= prime;
				}
				while (rem % prime == 0);
			}

			// p - valuation of F(a, -b)
			valuation = 0;
			if (mpz_divisible_ui_p(*remaining, prime))
			{
				while(!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
				{
					valuation++;
					mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
				} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			}

			// we have prime ideal (p, \bar{\theta} - 0) and ideal (p,p)
			flags = running->flags;
			flag = nf->SievingPoly->GetDegree();
			rem = valuation + (flags & ConstEF::L_VALUATION_MINI_MASK) * (flag - 1) - exponent * flag;
			
			flag = ConstEF::DIVIDES_L | flag;
			flag |= ((flags & ConstEF::L_VALUATION_MINI_MASK) << ConstEF::L_VALUATION_SHIFT);
			if (exponent == 0)
			{
				flag |= (valuation << ConstEF::F_VALUATION_SHIFT);
			}

			if (rem > 0)
			{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
				CLog::Log(50, "0182070", "FactorAlgebraic: L - reducing %Zd by (%d,%d)^%d.\n", *remaining, prime, running->r, rem);
#endif
#endif
				// save (p, \bar{\theta} - 0)
				rc = divisors->AddDivisor(prime, rem, running->r, flag, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}

			if (exponent > 0)
			{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
				CLog::Log(50, "0182071", "FactorAlgebraic: L - reducing %Zd by (%d,%d)^%d.\n", *remaining, prime, running->r, exponent);
#endif
#endif
				// save (p, p) with p-valuation of F(a, -b)
				flag |= (valuation << ConstEF::F_VALUATION_SHIFT);
				rc = divisors->AddDivisor(prime, exponent, prime, flag, lastWrittenIndex, assertionMode);
				if (rc != ConstRC::Ok)
				{
					return rc;
				}
			}
		}

		running++;
	} // while (running <= moddivison_end

	int k, max;
	// Now, the largish primes. We will walk through the current update table ...
	if (aHashtable)
	{
		lattice_update2_t *updates;
		lattice_script_t* script = aHashtable->script;
		while (script)
		{
			max = script->num_used;
			updates = script->list;
			for (k = 0; k < max; k++, updates++)
			{
#ifdef DEBUG_SIEVING_PHASE_ONE_PRIME
				if (updates->element->p == DEBUG_SIEVING_PHASE_CRITICAL_PRIME)
				{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
					CLog::Log(50, "0182072", "Factor algebraic script message:\n");						
#endif
					PrintUpdateInfo(*updates);
					Utils::Wait();
				}
#endif
				if (updates->h == aOffset)
				{
					prime = updates->element->p;
					exponent = 0;			
					while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
					{
						exponent++;
						mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
						this->iStats->Increase(ESSDivrem);
#endif
					} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif

					if (exponent) 	
					{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
						CLog::Log(50, "0182073", "FactorAlgebraic: reducing %Zd by (%d,%d)^%d.\n", *remaining, prime, updates->element->r, exponent);
#endif
#endif
						rc = divisors->AddDivisor(prime, exponent, updates->element->r, 0, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok) 
						{ 
							return rc; 
						}            		
					} 
					else 
					{							
						CLog::Log(MAXIMAL_INT_VALUE, "0182074", "\nFactorAlgebraicError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, updates[k]= %d.\n", *remaining, prime, exponent, aOffset, updates->h);
						this->iRelation->PrintToScreen();
						Utils::Wait();
					}			
				} 
			}
			script = script->next;
		}
	}

	max = this->iAdvFB[aIndex].BadPrimes.size();
	if (max > 0)
	{
		for (k = 0; k < max; k++)
		{
			prime = this->iAdvFB[aIndex].BadPrimes.at(k).p;
			rem = this->iAdvFB[aIndex].BadPrimes.at(k).r;
			if ((this->iRelation->A - (aux = (this->iRelation->B % prime) * rem)) % prime == 0)
			{
				exponent = 0;
				
				while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
				{
					exponent++;
					mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
					this->iStats->Increase(ESSDivrem);
#endif
				} 
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif

				if (exponent) 	
				{
#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
					CLog::Log(50, "0182075", "FactorAlgebraic: B - reducing %Zd by (%d,%d)^%d.\n", *remaining, prime, this->iAdvFB[aIndex].BadPrimes.at(k).r, exponent);
#endif
#endif
					rc = divisors->AddDivisor(prime, exponent, this->iAdvFB[aIndex].BadPrimes.at(k).r, 0, lastWrittenIndex, assertionMode);
					if (rc != ConstRC::Ok) 
					{ 
						return rc; 
					}            
				} 
				else
				{
					CLog::Log(MAXIMAL_INT_VALUE, "0182076", "\nALG FactorError: reduction of %Zd by %d impossible: exponent %d, aOffset = %d, i = %d, j = %d\n",*remaining,prime,exponent,aOffset,aI,aJ);
				}
			}
		}
	}

	// factorize with special ideals
	int size = nf->Domain->SpecialPrimesCount;
	if (size > 0)
	{
		// firstly - we find out if the remaining is low enough
		// and save exponents for other computing...
		int i;
		bool good = false;
		for (i = 0; i < size; i++)
		{
			exponent = 0;
			prime = nf->Domain->SpecialPrimes[i];
			while (!mpz_fdiv_qr_ui(this->iAuxMpz1, this->iAuxMpz2, *remaining, prime))
			{
				exponent++;
				good = true;
				mpz_swap(*remaining, this->iAuxMpz1);
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
				this->iStats->Increase(ESSDivrem);
#endif
			}
#ifdef DEBUG_SIEVING_PHASE_INFO_STATS_DIVREM_ON
			this->iStats->Increase(ESSDivrem);
#endif
			nf->SpecialPrimes[prime] = exponent;
		}

		// is it worth to continue?
		if (good && this->iVariationsInfo[aIndex].Control(*remaining))
		{
			// find base of ideal (c_{d}a + b\bar{\theta})
			if (nf->SievingPolyMonic)
			{
				nf->Domain->ABaseOfPrincipal(nf->IdealBase, this->iRelation->A, this->iRelation->B);
			}
			else
			{
				// non-monic poly
				mpz_mul_si(this->iAuxMpz2, *nf->SievingPoly->Leading2(), this->iRelation->A);
				nf->Domain->ABaseOfPrincipal(nf->IdealBase, this->iAuxMpz2, this->iRelation->B);
			}
 
			SpecialPrimeIt it;
			PrimeIdeal * ideal = NULL;
			// we find out exact factorization
			size = nf->SpecialIdeals.size();
			for (i = 0; i < size; i++)
			{
				ideal = nf->SpecialIdeals[i];
				prime = ideal->Prime;
				it = nf->SpecialPrimes.find(prime);
				flags = this->iAdvFB[aIndex].FactorBase[ideal->Index].flags;
				if ((*it).second > 0 || flags & ConstEF::DIVIDES_L_MINI_FLAG)
				{
					rc = nf->Domain->ValuationOfPrime(&valuation, ideal, nf->IdealBase);
					if (rc != ConstRC::Ok)
					{
						return rc;
					}

					exponent = valuation;
					if (exponent > 0) 
					{
						// we note that this is special prime and we save its residual
						flag = ConstEF::SPECIAL_PRIME | ideal->Residual;
						if (flags & ConstEF::DIVIDES_L_MINI_FLAG)
						{
							valuation = (*it).second;
							nf->SpecialPrimes[prime] = 0;
							// if this prime divides the leading coeff than we save exact valuation of F(a,-b)
							flag |= ConstEF::DIVIDES_L | ((unsigned int)valuation << ConstEF::F_VALUATION_SHIFT);
							// we also save this prime valuation of the leading coeff
							flag |= ((flags & ConstEF::L_VALUATION_MINI_MASK) << ConstEF::L_VALUATION_SHIFT);
						}
						else
						{
							// norm of prime ideal is p^{residual} so if valuation > 0 than F(a,-b) is divisible by p^{valuation*residual}
							valuation *= ideal->Residual;

							nf->SpecialPrimes[prime] = (*it).second - valuation;
						}

#ifdef FACTOR_DEBUG
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
						CLog::Log(50, "0182077", "FactorAlgebraic: S - reducing %Zd by (%d,%d)^%d.\n", *remaining, prime, ideal->Cp, exponent);
#endif
#endif

						rc = divisors->AddDivisor(prime, exponent, ideal->Cp, flag, lastWrittenIndex, assertionMode);
						if (rc != ConstRC::Ok)
						{
							return rc;
						}
					
					} // end if - exponent

				} // end if - used this special idesl

			} // end of for cycle - special ideals
		
		} // end if - continue?

	} // end if - special primes

#ifdef DEBUG_SIEVING_PHASE_ASSERTION_MODE_ON
	if (assertionMode)
	{
		if (mpz_cmp_ui(*remaining, 1) != 0)
		{
			// We will try all primes
			lattice_type2 * fb = this->iAdvFB[aIndex].FactorBase;
			while (fb != NULL)
			{
				prime = fb->p;
				if (!prime) 
				{
					break;
				}
				if (mpz_mod_ui(this->iAuxMpz1, *remaining, prime) == 0)
				{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
					CLog::Log(50, "0182078", "\nFactorizeAlgebraic: remaining %Zd is divisible by prime %d\n",*remaining,prime);
#endif
					// Search
					this->iRelation->PrintToScreen();
					Utils::Wait();
					this->iHashtables[aIndex].SearchForPrimeInHashtable(prime, fb->r);
				}
				fb++;
			}			
		}
	}
#endif

	return ConstRC::Ok;
}
int CLatticeSievingPhase::ComputeExpectedLog(CNumberFieldInfo & aNF, CVariationsInfo & aVarInfo)
{
	int rough = 0;

	// we want to try 128 = 16 * 8 values across sieving region
	int stepI = this->iLatticeInfo.I >> 4;
	int stepJ = this->iLatticeInfo.J >> 3;
	int halfI = this->iLatticeInfo.I >> 1;
	mpz_t a, b, res, stepA, stepB;
	mpz_init(res);
	mpz_init(a);
	mpz_init(b);
	mpz_init(stepA);
	mpz_init(stepB);

	// step for a
	mpz_set_si(this->iAuxMpz1, this->iLatticeInfo.a0);
	mpz_mul_si(stepA, this->iAuxMpz1, stepI);

	// step for b
	mpz_set_si(this->iAuxMpz2, this->iLatticeInfo.b0);
	mpz_mul_si(stepB, this->iAuxMpz2, stepI);

	int i, j;
	for (j = 1; j < this->iLatticeInfo.J; j += stepJ)
	{
		// start value of a for this row j
		mpz_set_si(this->iAuxMpz1, this->iLatticeInfo.a0);
		mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, -halfI);
		mpz_set_si(this->iAuxMpz2, this->iLatticeInfo.a1);
		mpz_mul_si(this->iAuxMpz2, this->iAuxMpz2, j);
		mpz_add(a, this->iAuxMpz1, this->iAuxMpz2);

		// start value of b for this row j
		mpz_set_si(this->iAuxMpz1, this->iLatticeInfo.b0);
		mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, -halfI);
		mpz_set_si(this->iAuxMpz2, this->iLatticeInfo.b1);
		mpz_mul_si(this->iAuxMpz2, this->iAuxMpz2, j);
		mpz_add(b, this->iAuxMpz1, this->iAuxMpz2);

		for (i = -halfI; i < halfI; i += stepI, mpz_add(a, a, stepA), mpz_add(b, b, stepB))
		{
			aNF.Domain->Norm(aNF.SievingPoly, res, a, b);
			rough += mpz_sizeinbase(res, ConstSPhase::LOGARITHM_BASE);
		}
	}

	mpz_clear(res);
	mpz_clear(a);
	mpz_clear(b);
	mpz_clear(stepB);
	mpz_clear(stepA);

	rough = (rough >> 7) + 1;
	rough -= aNF.ExpectedErrorFactor;
	if (aVarInfo.Type != EDoNotUseVariations) 
	{
		// Space for maximal partial factor
		rough -= aVarInfo.SizeOfMaxRemainder;
	} 
	
	return rough;
}
int CLatticeSievingPhase::SaveResult()
{
	Set_InnerState(ESPhaseSavingResult);

	AFactorAlgParameters * param = Get_Parameters();

	string auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSFF));
	param->SetParameterForOthers(ConstPN::SMOOTH_REL_COUNT, auxStr, Get_PhaseNbr());

	auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSEdges));
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_GRAPH_EDGES, auxStr, Get_PhaseNbr());

	auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSComponents));
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_GRAPH_COMPONENTS, auxStr, Get_PhaseNbr());

	auxStr = Utils::ConvertToString(this->iStats->Get_ConfirmTotal(ESSVertices));
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_GRAPH_VERTICES, auxStr, Get_PhaseNbr());

	param->SetParameterForOthers(ConstPN::SMOOTH_RELATIONS_FULL_FILENAME, this->iSmoothRelFullFileName, Get_PhaseNbr());
	
	param->SetParameterForOthers(ConstPN::PARTIAL_REL_FULL_FILENAME, this->iPartialRelFullFileName, Get_PhaseNbr());

	// we need to contruct new FB with all 'special q'
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182079", "    Construct new FB ... \n");
	}
#endif

	// we use last part for lattice sieving
	int index = CNFSRelation::DEFAULT_NFS_PARTS_COUNT - 1;
	int rc = this->iFBs[index].RefillFB(this->iSpecialQ + 1, this->iNumberFields[index]);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SaveResult", 0, 0, "Refill FB failed.");
		return rc;
	}

	CBaseParameters paramBase;
	paramBase.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iFBs[index].FBFullFileName);
	rc = this->iFBs[index].Serialize(paramBase);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SaveResult", 0, 0, "Serialize FB failed.");
		return rc;
	}	

	// save new FB upper bound
	auxStr = Utils::ConvertToString(this->iFBs[index].UpperFBBound);
	param->SetParameterForOthers(ConstPN::FB_UPPER_BOUND + Utils::ConvertToString(index), auxStr, (int)EMain);

	// we serialize this phase
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();
	paramBase.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_SerialPhaseFullFileName());
	rc = Serialize(paramBase);
	if (rc != ConstRC::Ok)
	{
		CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SaveResult", 0, 0, "Serialize phase failed.");
		return rc;
	}

	Set_InnerState(ESPhaseFinished);

	return ConstRC::Ok;
}
/*!
	This is an auxiliary method for printing out the intermediate results
	of the sieving step.

*/
void CLatticeSievingPhase::PrintHeader() 
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (this->iUseCyclesForRelConstruction)
	{
		CLog::Log(10000, "", "\n\n");
		CLog::Log(10000, "0182921", " __________________________________________________________________________________ \n");
		CLog::Log(10000, "0182922", "|   LINE  |  SMOOTH |  PARTS  |  CYCLES |   EDGES |  COMPON |  VERTEX |  PERCENT   |\n");
	}
	else
	{
		CLog::Log(10000, "", "\n\n");
		CLog::Log(10000, "0182923", " __________________________________________________________________________________ \n");
		CLog::Log(10000, "0182924", "|   LINE  |  SMOOTH | 1-PARTS | 2-PARTS | 3-PARTS | 4-PARTS | X-PARTS |  PERCENT   |\n");
	}
#endif
}
/*!
	This is an auxiliary method for printing out the intermediate results
	of the sieving step.

*/
void CLatticeSievingPhase::PrintValues() 
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 10000
	if (this->iUseCyclesForRelConstruction)
	{
		int independentCycles = this->iStats->Get_Total(ESSEdges) + 
			this->iStats->Get_Total(ESSComponents) - this->iStats->Get_Total(ESSVertices);	
		int collected = this->iStats->Get_Total(ESSFF) + independentCycles;
		float ratio = (float) collected * (float) 100;
		if (iAssignedNumberOfSmooths == (int)NUMBER_OF_SMOOTHS_NOT_GIVEN) ratio = (float) ratio / (float) this->iRequestedRelations;
		else ratio = (float) ratio / (float) iAssignedNumberOfSmooths;
		CLog::Log(10000, "0182925", "\r| %7d | %7d | %7d | %7d | %7d | %7d | %7d |   %6.2f %% |", 
			Get_WorkingIteration(),
			this->iStats->Get_Total(ESSFF),
			this->iStats->Get_Total(ESSPartialR),
			independentCycles,				
			this->iStats->Get_Total(ESSEdges),			
			this->iStats->Get_Total(ESSComponents),				
			this->iStats->Get_Total(ESSVertices),					
			ratio);
		fflush(stdout);
	}
	else
	{
		int collected = this->iStats->Get_Total(ESSFF);
		if (iUseHighVariations && (iAssignedNumberOfSmooths == (int)NUMBER_OF_SMOOTHS_NOT_GIVEN)) collected += this->iStats->Get_Total(ESSPartialR);
		float ratio = (float) collected * (float) 100;
		if (iAssignedNumberOfSmooths == (int)NUMBER_OF_SMOOTHS_NOT_GIVEN) ratio = (float) ratio / (float) this->iRequestedRelations;
		else ratio = (float) ratio / (float) iAssignedNumberOfSmooths;
		CLog::Log(10000, "0182926", "\r| %7d | %7d | %7d | %7d | %7d | %7d | %7d |   %6.2f %% |", 
			Get_WorkingIteration(),
			this->iStats->Get_Total(ESSSmooth),
			this->iStats->Get_Total(ESS1Partial),	
			this->iStats->Get_Total(ESS2Partial),					
			this->iStats->Get_Total(ESS3Partial),				
			this->iStats->Get_Total(ESS4Partial),				
			this->iStats->Get_Total(ESSXPartial),					
			ratio);
		fflush(stdout);
	}
#endif
}
void CLatticeSievingPhase::PrintEffectiveness()
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 5000
	CLog::Log(5000, "", "\n");
	CLog::Log(5000, "0182927", "So far, %d iterations have been run.\n", Get_WorkingIteration());
	if (this->iAssignedNumberOfSmooths != (int)NUMBER_OF_SMOOTHS_NOT_GIVEN)
	{
		CLog::Log(5000, "0182981", " Assigned relations (smooth): %d\n", this->iAssignedNumberOfSmooths);
	}
	else if (this->iUseHighVariations)
	{
		CLog::Log(5000, "0182982", " Requested relations (all): %d\n", this->iRequestedRelations);
	
	}
	else
	{
		CLog::Log(5000, "0182983", " Requested relations (smooth): %d\n", this->iRequestedRelations);
	}
	float seconds = this->iTimes->Get_TotalTime(ESTLineSieving);
	float abs;
	float secondsBlock;
	int i;
	abs = this->iTimes->Get_TotalTime(ESTRunSieve); 
	CLog::Log(5000, "0182928", "  Overall time spent in RunSieve is %12.4f sec \n", abs);
	if (abs > 0)
	{
		secondsBlock = this->iTimes->Get_TotalTime(ESTRootUpdate); 
		CLog::Log(5000, "0182929", "    From this in LatticeInit is %12.4f sec (%6.4f %%)\n", secondsBlock, 100 * (secondsBlock / abs));
		CLog::Log(5000, "0182930", "    From this in SieveRegion is %12.4f sec (%6.4f %%)\n", seconds, 100 * (seconds / abs));
	}		
	if (seconds > 0)
	{
		secondsBlock = this->iTimes->Get_TotalTime(ESTBlockSieving);
		CLog::Log(5000, "0182931", "      From this in SieveBlock is %12.4f sec (%6.4f %%)\n", secondsBlock, 100 * (secondsBlock / seconds));
		if (secondsBlock > 0)
		{
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				abs = this->iTimes->Get_TotalTime(ESTFirstFillHT + i); 
				CLog::Log(5000, "0182932", "        From this in Fill HT #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
			}
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				abs = this->iTimes->Get_TotalTime(ESTFirstSieving + i); 
				CLog::Log(5000, "0182933", "        From this in Sieving #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
				abs = this->iTimeSievingMiddle[i] / (double) CLOCKS_PER_SEC;
				CLog::Log(5000, "0182976", "          Sieving Middle     #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
				abs = this->iTimeSievingLargish[i] / (double) CLOCKS_PER_SEC;
				CLog::Log(5000, "0182977", "          Sieving Largish    #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
#endif

			}
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				abs = this->iTimes->Get_TotalTime(ESTFirstFactorization + i); 
				CLog::Log(5000, "0182934", "        From this in Factor  #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
#ifdef DEBUG_SIEVING_PHASE_TECHNICAL_TIMES_ON
				abs = this->iTimeFactorMiddle[i] / (double) CLOCKS_PER_SEC;
				CLog::Log(5000, "0182978", "          Factor Middle      #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
				abs = this->iTimeFactorLargish[i] / (double) CLOCKS_PER_SEC;
				CLog::Log(5000, "0182979", "          Factor Largish     #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
				abs = this->iTimeFactorDivideSpecial[i] / (double) CLOCKS_PER_SEC;
				CLog::Log(5000, "0182980", "          Factor Divide Spec #%d is %12.4f sec (%6.4f %%)\n", i + 1, abs, 100 * (abs / secondsBlock));
#endif
			}
			abs = this->iTimes->Get_TotalTime(ESTPartialFactorization); 
			CLog::Log(5000, "0182935", "        From this in Par Factor is %12.4f sec (%6.4f %%)\n", abs, 100 * (abs / secondsBlock));
		}
	}	 

	CLog::Log(5000, "", "\n");
	CLog::Log(5000, "0182936", "EFFECTIVENESS: \n\n");
	CLog::Log(5000, "0182937", "  Global: \n");
	int succ = 0;
	int tested = this->iStats->Get_Total(ESSTested);
	CLog::Log(5000, "0182938", "    Total tested      :  %10d.\n", tested);
	int value = 0;

	if (this->iUseCyclesForRelConstruction)
	{		
		value = this->iStats->Get_Total(ESSFF);
		CLog::Log(5000, "0182939", "    Total full-full   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESSFP);
		CLog::Log(5000, "0182940", "    Total full-part   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESSPF);
		CLog::Log(5000, "0182941", "    Total part-full   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESSPP);
		CLog::Log(5000, "0182942", "    Total part-part   :  %10d.\n", value);
	}
	else
	{
		value = this->iStats->Get_Total(ESSSmooth);
		CLog::Log(5000, "0182943", "    Total smooth      :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESS1Partial);
		CLog::Log(5000, "0182944", "    Total 1-partial   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESS2Partial);
		CLog::Log(5000, "0182945", "    Total 2-partial   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESS3Partial);
		CLog::Log(5000, "0182946", "    Total 3-partial   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESS4Partial);
		CLog::Log(5000, "0182947", "    Total 4-partial   :  %10d.\n", value);
		succ += value;
		value = this->iStats->Get_Total(ESSXPartial);
		CLog::Log(5000, "0182948", "    Total X-partial   :  %10d.\n", value);
	}

	succ += value;
	if (tested > 0)
	{
		CLog::Log(5000, "0182949", "    Total successes   :  %10d (%6.4f %%).\n", succ, (((double) succ)*(double) 100)/(double) tested);
	}
	else
	{
		CLog::Log(5000, "0182950", "    Total successes could not be determined - 0 relations tested.\n");
	}
	int rejected = this->iStats->Get_Total(ESSRejected);
	int factortoolarge= this->iStats->Get_Total(ESSFactorTooLarge);
	int remaindertoolarge = this->iStats->Get_Total(ESSRemainderTooLarge);
	int notcoprime = this->iStats->Get_Total(ESSNotCoprime);
	int primetoolarge = this->iStats->Get_Total(ESSPrimeTooLarge);
	if (tested > 0)
	{
		CLog::Log(5000, "0182951", "    Total rejected    :  %10d (%6.4f %%).\n", rejected, (((double) rejected)*(double) 100)/(double) tested);
		CLog::Log(5000, "0182952", "      GCD > 1:  %10d (%6.4f %%).\n", notcoprime, (((double) notcoprime)*(double) 100)/(double) tested);
		CLog::Log(5000, "0182953", "      Expensi:  %10d (%6.4f %%).\n", (rejected - notcoprime), (((double) rejected - notcoprime)*(double) 100)/(double) tested);
		CLog::Log(5000, "01829534", "      Factor too large:  %10d (%6.4f %%).\n", (factortoolarge), (((double) factortoolarge  )*(double) 100)/(double) tested);
		CLog::Log(5000, "01829534", "      Remainder too large:  %10d (%6.4f %%).\n", (remaindertoolarge), (((double) remaindertoolarge)*(double) 100)/(double) tested);
		CLog::Log(5000, "01829534", "      Prime too large:  %10d (%6.4f %%).\n", (primetoolarge), (((double) primetoolarge)*(double) 100)/(double) tested);
	}
	CLog::Log(5000, "", "\n\n");
	seconds =  this->iTimes->Get_IntermediateTime(ESTLineSieving);
	tested = this->iStats->Get_Intermediate(ESSTested);
	CLog::Log(5000, "0182954", "  Local (%6.2f sec): \n", seconds);
	CLog::Log(5000, "0182955", "    Local tested      :  %10d (%6.2f per sec).\n", tested, (seconds > 0) ? (double) tested / seconds : 0);
	succ = 0;
	if (this->iUseCyclesForRelConstruction)
	{
		value = this->iStats->Get_Intermediate(ESSFF);
		CLog::Log(5000, "0182956", "    Local full-full   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESSFP);
		CLog::Log(5000, "0182957", "    Local full-part   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESSPF);
		CLog::Log(5000, "0182958", "    Local part-full   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESSPP);
		CLog::Log(5000, "0182959", "    Local part-part   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
	}
	else
	{
		value = this->iStats->Get_Intermediate(ESSSmooth);
		CLog::Log(5000, "0182960", "    Local smooth      :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESS1Partial);
		CLog::Log(5000, "0182961", "    Local 1-partial   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESS2Partial);
		CLog::Log(5000, "0182962", "    Local 2-partial   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESS3Partial);
		CLog::Log(5000, "0182963", "    Local 3-partial   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESS4Partial);
		CLog::Log(5000, "0182964", "    Local 4-partial   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
		succ += value;
		value = this->iStats->Get_Intermediate(ESSXPartial);
		CLog::Log(5000, "0182965", "    Local X-partial   :  %10d (%6.2f per sec).\n", value, (seconds > 0) ? (double) value / seconds : 0);
	}

	succ += value;
	if (tested > 0 && seconds > 0)
	{
		CLog::Log(5000, "0182966", "    Local successes   :  %10d (%6.4f %%).\n", succ, (((double) succ)*(double) 100)/(double) tested);
	}
	else
	{
		CLog::Log(5000, "0182967", "    Local successes could not be determined - 0 relations tested or 0 seconds.\n");
	}
	rejected = this->iStats->Get_Intermediate(ESSRejected);
	notcoprime = this->iStats->Get_Intermediate(ESSNotCoprime);
	if (tested > 0 && seconds > 0)
	{
		CLog::Log(5000, "0182968", "    Local rejected    :  %10d (%6.4f %%).\n", rejected, (((double) rejected) * (double) 100)/(double) tested);
		CLog::Log(5000, "0182969", "      GCD > 1:  %10d (%6.4f %%).\n", notcoprime, (((double) notcoprime)*(double) 100)/(double) tested);
		CLog::Log(5000, "0182970", "      Expensi:  %10d (%6.4f %%).\n", (rejected - notcoprime), (((double) rejected - notcoprime) * (double) 100)/(double) tested);
	}
	CLog::Log(5000, "0182971", "\n");

	if (tested - notcoprime > 0) 
	{
		tested -= notcoprime;
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			CLog::Log(5000, "0182972", "    Avg size - Part #%d:  %4d (bits), %4d (dec)\n", i + 1, this->iStats->Get_Intermediate(ESSAverageSizeIP_log2 + i) / tested, this->iStats->Get_Intermediate(ESSAverageSizeIP_log10 + i) / tested);
		}	
	} 
	else 
	{
		CLog::Log(5000, "0182973", "    Avgs sizes could not be determined - 0 relations tested\n");
	}

	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		CLog::Log(5000, "0182974", "    Part #%d: Expected value:  %4d; Change: %4d (base: %d)\n", i + 1, SIEVING_PHASE_MASK - this->iInitBlockValues[i], this->iInitBlockValueChange[i], ConstSPhase::LOGARITHM_BASE);
	}
	
	double totalEntries = ((double)this->iLatticeInfo.I * this->iLatticeInfo.J) * Get_TimeMessage();	
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		seconds = this->iTimes->Get_IntermediateTime(ESTFirstSieving + i) * 1000;
		CLog::Log(5000, "0182975", "    Sieving #%d speed :  %8.2f * 10^3 per sec\n", i + 1, totalEntries / seconds);
	}

	fflush(stdout);	
#endif
}
void CLatticeSievingPhase::ControlBlock(int aBlockIndex, int aInnerIndex)
{
	int blockSize = Get_BlockSize();

	// aBlockIndex * blockSize * this->iBlockSetSize + aInnerIndex * blockSize
	mpz_set_si(this->iAuxMpz1, aBlockIndex);
	mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, this->iBlockSetSize);
	mpz_add_ui(this->iAuxMpz1, this->iAuxMpz1, aInnerIndex);
	mpz_mul_si(this->iAuxMpz1, this->iAuxMpz1, blockSize);

	mpz_set_si(this->iAuxMpz2, this->iModI);
	mpz_and(this->iAuxMpz2, this->iAuxMpz2, this->iAuxMpz1);
	int startCol = - (this->iLatticeInfo.I >> 1);
	int maxCol = - startCol;

	mpz_fdiv_q_2exp(this->iAuxMpz2, this->iAuxMpz1, this->iLogI);
	int startRow = mpz_get_si(this->iAuxMpz2) + 1;
	int maxRow = (blockSize >> this->iLogI) + startRow;
	int i;

	algebraic_coeff_type a;
	algebraic_coeff_type b;

	int base;

	string aux = "e:\\Data\\data_q" + Utils::ConvertToString(this->iSpecialQ) + "block_" + Utils::ConvertToString(aInnerIndex + 1) + ".txt";

	FILE * data = fopen(aux.c_str(), FILE_MODE_WRITE);
	// col, row, Int Block, Alg Block, Int Value, Alg Value, Int Rem, Alg Rem, gcd

	int index = 0;
	for (int j = startRow; j < maxRow; j++)
	{
		a = ((algebraic_coeff_type) startCol) * ((algebraic_coeff_type) this->iLatticeInfo.a0) + ((algebraic_coeff_type) j) * ((algebraic_coeff_type) this->iLatticeInfo.a1);
		b = ((algebraic_coeff_type) startCol) * ((algebraic_coeff_type) this->iLatticeInfo.b0) + ((algebraic_coeff_type) j) * ((algebraic_coeff_type) this->iLatticeInfo.b1);
		
		for (i = startCol; i < maxCol; i++, a += this->iLatticeInfo.a0, b += this->iLatticeInfo.b0)
		{
			this->iRelation->Reset();
			this->iRelation->Setup(a, b, CNFSRelation::DEFAULT_NFS_PARTS_COUNT, this->iNumberFields);
			fprintf(data, "%d\t%d\t", i, j);
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				fprintf(data, "%d\t", (- (int)this->iInitBlockValues[i] + this->iInitBlockValueChange[i] + (int)this->iSievingBlocks[i][index]));
			}
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				base = mpz_sizeinbase(this->iRelation->Get_Parts(i)->Norm, ConstSPhase::LOGARITHM_BASE);
				fprintf(data, "%d\t", base);
			}
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{		
				if (this->iNumberFields[i].SievingPoly->GetDegree() > 1)
				{
					FactorizeAlgebraic(i, j, index, &this->iHashtables[i].Hashtables[aInnerIndex], true, i); 
				}
				else
				{
					FactorizeIntegral(i, j, index, &this->iHashtables[i].Hashtables[aInnerIndex], i);
				}
			}
			for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				base = mpz_sizeinbase(this->iRelation->Get_Parts(i)->NormRemaining, ConstSPhase::LOGARITHM_BASE);
				fprintf(data, "%d\t", base);
			}
			Utils::GCD(abs(i), abs(j), base);
			fprintf(data, "%d\t", base);
			fprintf(data, "\n");
			fflush(data);

			index++;
		}
	}

	fclose(data);
	data = NULL;
}
int CLatticeSievingPhase::PrepareRunningSerialization()
{
	this->iStats->ConfirmAll();
	this->iTimes->ConfirmAll();
	Set_Iteration(Get_WorkingIteration());

	this->iRSerialRootIndex = this->iRootIndex;
	this->iRSerialSpecialQ	= this->iSpecialQ;

	int i;
	for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		this->iRSerialInitBlockValueChange[i] = this->iInitBlockValueChange[i];
	}

	// buffer switch
	vector<CNFSRelation*>* aux = Get_RelationBuffer();
	int size = this->iRelationBufferMaxCount;
	Set_RelationBuffer(this->iRelationWorkingBuffer, this->iRelationWorkingBufferMaxCount);
	this->iRelationWorkingBuffer = aux;
	this->iRelationWorkingBufferMaxCount = 0;
	for (i = 0; i < size; i++)
	{
		this->iRelationWorkingBuffer->at(i)->Reset();
	}

	return ConstRC::Ok;
}
/*!
	Throughout the 1-partials processing, a hashtable containing large primes
	is used to track the state of the "partial relation graph". It serves us
	  to determine the current number of independent cycles found.

	As such, this hashtable can be (re)constructed from the list of partial relations
	collected so far. That is why we do not save it into any XML form; instead,
	we recreate it during deserialization from the partials file.

*/
int CLatticeSievingPhase::ReloadHashtable()
{	
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182080", "\tReconstructing hashtable for partial relations ...\n");
	}
#endif

	// all new
	this->iStats->Reset(ESSEdges);
	this->iStats->Reset(ESSVertices);
	this->iStats->Reset(ESSComponents);

	// create hashtable and add ONE to hashtable
	int rc = PrepareCycleTable();
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (MSerializable::ExistFile(this->iPartialRelFullFileName) == false)
	{
		return ConstRC::Ok;
	}

	CNFSRelation relation(CNFSRelation::MODE_CYCLE_CONSTRUCTION);

	FILE * partialsFILE = fopen(this->iPartialRelFullFileName.c_str(), FILE_MODE_READ);
	if (partialsFILE == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","ReloadHashtable",0,0,"Could not open file with partial relations.");
		return ConstRC::CouldNotOpenFile;
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
	if (Get_InfoMode())
	{
		CLog::Log(0, "0182081", "\tPartial relations file name: %s.\n", this->iPartialRelFullFileName.c_str());
	}
#endif

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	int index = 0;
#endif

	bool infoMode = Get_InfoMode();
	while (relation.ReadFromFile(partialsFILE) == ConstRC::Ok)
	{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
		if (infoMode)
		{
			if (!(index % 10))
			{
				CLog::Log(50, "0182082", "\r\tAdding %8d relations...", index);
				fflush(stdout);
			}
			index++;
		}
#endif
		rc = AddToCycleCountStructures(&relation);
		if (rc != ConstRC::Ok)
		{
			fclose(partialsFILE);
			partialsFILE = NULL;
			return rc;
		}
	}

	fclose(partialsFILE);
	partialsFILE = NULL;

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	if (infoMode)
	{
		CLog::Log(50, "0182083", "\r\tHashtable reconstructed. Altogether %d partials read, forming %d independent cycles.", index, IndependentCyclesCount());
	}
#endif

	return ConstRC::Ok;
}
/*!
	Throughout the X-partials processing, a hashtable containing all prime ideals
	is used to track the required number of relations.

	As such, this hashtable can be (re)constructed from the list of smooth and partial relations
	collected so far. That is why we do not save it into any XML form; instead,
	we recreate it during deserialization from the partials file.

*/
int CLatticeSievingPhase::ReloadFreqHashtable()
{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182088", "\tReconstructing frequency hashtable from all relations ...\n");
	}
#endif

	// create hashtable
	DetermineRequestedRelations();

	if (this->iHighVarHashtable == NULL) 
	{
		this->iHighVarHashtable = new CFrequencyHashtable();
		this->iHighVarHashtable->SetupHashtableProb(this->iRequestedRelations << 2);
	}

	CNFSRelation relation(CNFSRelation::MODE_CYCLE_CONSTRUCTION);
	FILE * relationsFILE = NULL;
	relation_index_type index = 0;
	int rc;
	int i, j;
	bool infoMode = Get_InfoMode();
	string fileName = this->iSmoothRelFullFileName;

	for (i = 0; i < 2; i++)
	{
		if (i = 1)
		{
			fileName = this->iPartialRelFullFileName;
		}

		if (MSerializable::ExistFile(fileName))
		{

			relationsFILE = fopen(fileName.c_str(), FILE_MODE_READ);
			if (relationsFILE == NULL)
			{
				CLog::LogError(WHEREARG,"CLatticeSievingPhase","ReloadFreqHashtable",0,0,"Could not open file with relations.");
				return ConstRC::CouldNotOpenFile;
			}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 0
			if (Get_InfoMode())
			{
				CLog::Log(0, "0182089", "\tRelations file name: %s.\n", fileName.c_str());
			}
#endif


			while (relation.ReadFromFile(relationsFILE) == ConstRC::Ok)
			{
#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
				if (infoMode)
				{
					if (!(index % 10))
					{
						CLog::Log(50, "0182090", "\r\tAdding %8d relations...", index);
						fflush(stdout);
					}
				}
#endif
				index++;
				for (j = 0; j < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; j++) 
				{
					rc = iHighVarHashtable->AddToTable(relation.Get_Parts(j), ConstSPhase::ELEMENT_MASK[j], index);
					if (rc != ConstRC::Ok)
					{
						fclose(relationsFILE);
						relationsFILE = NULL;
						return rc;
					}
				}
			}

			fclose(relationsFILE);
			relationsFILE = NULL;
		}
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 50
	if (infoMode)
	{
		CLog::Log(50, "0182091", "\r\tHashtable reconstructed. Altogether %d relation read.", index);
	}
#endif

	return ConstRC::Ok;
}

/*!
	TODO: If call from more threads then could be a problem with writing relations - same files...
*/
int CLatticeSievingPhase::Serialize(const CBaseParameters &aParam) const
{
#ifdef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// quick end
	return ConstRC::Ok;
#endif

	int rc = MSerializable::Serialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	if (Running())
	{
		AFactorAlgParameters * param = Get_Parameters();

		// for deserialization at new start
		param->SetPhaseParameter(ConstPN::DESERIALIZE, Utils::ConvertToString(true), Get_PhaseNbr());
		param->SetPhaseParameter(ConstPN::PHASE_FULL_FILENAME, 
			MSerializable::CreateFileNameAuto(aParam, ConstFN::SIEVING_PHASE), Get_PhaseNbr());

		CBaseParameters paramb;
		paramb.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, Get_ParametersFullFileName()); 
		rc = param->Serialize(paramb);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	// if it is running serialization, we serialize also found relations
	bool rserial = false;
	aParam.GetBoolParameter(ConstPN::RUNNING_SERIALIZATION, rserial);
	if (rserial)
	{
		vector<CNFSRelation*>* relations = Get_RelationBuffer();
		if (this->iRelationBufferMaxCount > 0 && relations != NULL)
		{
			FILE * partialsFILE = NULL;
			FILE * smoothsFILE = fopen(this->iSmoothRelFullFileName.c_str(), FILE_MODE_APPEND);
			if (smoothsFILE == NULL)
			{
				CLog::LogError(WHEREARG,"CLatticeSievingPhase","Serialize",0,0,"Could not open file with smooth relations.");
				return ConstRC::CouldNotOpenFile;
			}

			if (this->iVariationsInfo[0].Type != EDoNotUseVariations ||
				this->iVariationsInfo[1].Type != EDoNotUseVariations)
			{
				partialsFILE = fopen(this->iPartialRelFullFileName.c_str(), FILE_MODE_APPEND);
				if (partialsFILE == NULL)
				{
					fclose(smoothsFILE);
					smoothsFILE = NULL;
					CLog::LogError(WHEREARG,"CLatticeSievingPhase","Serialize",0,0,"Could not open file with partial relations.");
					return ConstRC::CouldNotOpenFile;
				}
			}

			CNFSRelation * relation = NULL;
			for (int i = 0; i < this->iRelationBufferMaxCount; i++)
			{
				relation = relations->at(i);
				if ((relation->Type & RELATION_FULL_ALGEBRAIC1) && (relation->Type & RELATION_FULL_ALGEBRAIC2))
				{
					// smooth relation
					rc = relation->PrintToFile(smoothsFILE);
					if (rc != ConstRC::Ok)
					{
						fclose(smoothsFILE);
						smoothsFILE = NULL;
						if (partialsFILE != NULL)
						{
							fclose(partialsFILE);
							partialsFILE = NULL;
						}
						return rc;
					}
				}
				else
				{
					// partial relation
					rc = relation->PrintToFile(partialsFILE);
					if (rc != ConstRC::Ok)
					{
						fclose(smoothsFILE);
						smoothsFILE = NULL;
						fclose(partialsFILE);
						partialsFILE = NULL;
						return rc;
					}
				}
			}

			fclose(smoothsFILE);
			smoothsFILE = NULL;

			if (partialsFILE != NULL)
			{
				fclose(partialsFILE);
				partialsFILE = NULL;
			}
		}
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::Serialize(const CBaseParameters &aParam, xmlTextWriterPtr &aWriter) const
{
#ifdef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// quick end
	return ConstRC::Ok;
#endif

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_SIEVING_PHASE_NAME, mainTagName);

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST mainTagName.c_str());
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","Serialize",rc,0,"Could not write the start element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, FALSE);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_INNER_STATE_NAME, Get_InnerState());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	CBaseParameters param;
	rc = this->iStats->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = this->iTimes->Serialize(param, aWriter);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_ITERATION_NAME, Get_Iteration());
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedLong(aWriter, XML_SERIAL_SPECIAL_Q_NAME, this->iSpecialQ);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}
	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_ROOT_INDEX_NAME, this->iRSerialRootIndex);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	// write info about individual parts - generaly we have only two
	rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_PARTS_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","Serialize",rc,0,"Could not write the elements tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
	{
		rc = SerializePart(i, aWriter);
		if (rc != ConstRC::Ok)
		{
			return rc;
		}
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_PARTS_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","Serialize",rc,0,"Could not end the elements tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // mainTagName
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","Serialize",rc,0,"Could not end the start element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter,FALSE);

	return rc;
}
/*!
	Serialize info about individual part of sieving phase which corresponded to relation part.
*/
int CLatticeSievingPhase::SerializePart(int aIndex, xmlTextWriterPtr &aWriter) const
{
#ifdef DEBUG_SIEVING_PHASE_SERIALIZATION_OFF
	// quick end
	return ConstRC::Ok;
#endif
	if (aIndex < 0 || aIndex >= CNFSRelation::DEFAULT_NFS_PARTS_COUNT)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","SerializePart",aIndex,0,"Index out of range.");
		return ConstRC::BadArgument;
	}

	int rc = xmlTextWriterStartElement(aWriter, BAD_CAST XML_SERIAL_ELEMENT_NAME);
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","SerializePart",rc,0,"Could not write the elements tag.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, false);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = CXMLHelper::SerializeSignedInt(aWriter, XML_SERIAL_THRESHOLD_CHANGE_NAME, this->iRSerialInitBlockValueChange[aIndex]);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	rc = xmlTextWriterEndElement(aWriter); // XML_SERIAL_ELEMENT_NAME
	if (rc < 0) 
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","SerializePart",rc,0,"Could not end the start element.");
		return ConstRC::GeneralError;
	}
	rc = CXMLHelper::NewLine(aWriter, false);

	return rc;
}
/*!
	Shouldn't be call from outside!!!
*/
int CLatticeSievingPhase::Deserialize(const CBaseParameters &aParam)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != ESPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
	if (Get_InfoMode())
	{
		CLog::Log(100, "0182084", "  Deserializing Sieving Phase...\n");
	}
#endif

	int rc = MSerializable::Deserialize(aParam);
	if (rc != ConstRC::Ok)
	{
		return rc;
	}

	int state;
	if ((state = Get_InnerState()) >= ESPhaseInitFB)
	{
		// for setup
		if (state == ESPhaseSieving)
		{
			Set_InnerState(ESPhaseInitAdvFB);
		}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
		if (Get_InfoMode())
		{
			CLog::Log(100, "0182085", "  Init integral domain ...\n");
		}
#endif

		int i;
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			rc = this->iNumberFields[i].InitIntegralDomain(this->iFBs[i].UpperFBBound);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
		if (Get_InfoMode())
		{
			CLog::Log(100, "0182086", "  Deserializing FBs ...\n");
		}
#endif

		CBaseParameters param;
		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			param.SetParameter(ConstPN::SERIALIZATION_FULL_FILENAME, this->iFBs[i].FBFullFileName);
			rc = this->iFBs[i].Deserialize(param);
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

#if LOG_MINIMAL_MESSAGE_PRIORITY <= 100
		if (Get_InfoMode())
		{
			CLog::Log(100, "0182087", "  Filling auxiliary fields ...\n");
		}
#endif

		for (i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
		{
			this->iFBs[i].FillAuxFields(this->iNumberFields[i]);
			this->iNumberFields[i].FillSpecialPrimes();
		}

		if (this->iUseCyclesForRelConstruction)
		{
			//partial relations hashtable
			rc = ReloadHashtable();
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}

		if (this->iHighVarHashtable)
		{
			rc = ReloadFreqHashtable();
			if (rc != ConstRC::Ok)
			{
				return rc;
			}
		}
	}

	return ConstRC::Ok;
}
/*!
	Shouldn't be call from outside!!!
*/
int CLatticeSievingPhase::Deserialize(const CBaseParameters &aParam, xmlTextReaderPtr &aReader)
{
	// we can deserialize only during initialization
	if (Get_InnerState() != ESPhaseInitialization)
	{
		return ConstRC::CannotBeDone;
	}

	string mainTagName;	
	aParam.GetStringParameter(ConstPN::SERIALIZATION_ELEMENT_NAME, XML_SERIAL_SIEVING_PHASE_NAME, mainTagName);

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","Deserialize",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name,BAD_CAST mainTagName.c_str()) == 0)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","Deserialize",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	CBaseParameters param;
	int auxInt = 0;
	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_INNER_STATE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize inner state...", auxInt, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_InnerState(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_STATISTIC_CONTAINER_NAME))
		{
			rc = this->iStats->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_TIME_CONTAINER_NAME))
		{
			rc = this->iTimes->Deserialize(param, aReader);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_SPECIAL_Q_NAME))
		{
			rc = CXMLHelper::DeserializeSignedLong(aReader, "Deserialize special q...", this->iSpecialQ, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iRSerialSpecialQ = this->iSpecialQ;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ROOT_INDEX_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize root index...", this->iRootIndex, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iRSerialRootIndex = this->iRootIndex;
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ITERATION_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize iteration...", auxInt, false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			Set_Iteration(auxInt);
			continue;
		}

		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_PARTS_NAME))
		{
			int depthParts = xmlTextReaderDepth(aReader);
			for (int i = 0; i < CNFSRelation::DEFAULT_NFS_PARTS_COUNT; i++)
			{
				if (CXMLHelper::MoveToNextNode(aReader, depthParts) != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
				rc = DeserializePart(i, aReader);
				if (rc != ConstRC::Ok)
				{
					rc = ConstRC::GeneralError;
					break;
				}
			}

			if (rc != ConstRC::Ok)
			{
				break;
			}

			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::DeserializePart(int aIndex, xmlTextReaderPtr &aReader)
{
	if (aIndex < 0 || aIndex >= CNFSRelation::DEFAULT_NFS_PARTS_COUNT)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DeserializePart",aIndex,0,"Index out of range.");
		return ConstRC::BadArgument;
	}

	int rc;
	xmlChar* name = xmlTextReaderName(aReader);

	if (name == NULL)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DeserializePart",0,0,"Deserialization main element name not found.");
		return ConstRC::GeneralError;  
	}

	if (xmlStrEqual(name, BAD_CAST XML_SERIAL_ELEMENT_NAME) == 0)
	{
		CLog::LogError(WHEREARG,"CLatticeSievingPhase","DeserializePart",0,0,"Main element name is wrong.");
		return ConstRC::GeneralError; 
	}

	int depth = xmlTextReaderDepth(aReader);

	while ((rc = CXMLHelper::MoveToNextNode(aReader, depth)) == ConstRC::Ok)
	{
		name = xmlTextReaderName(aReader);
	
		if (xmlStrEqual(name, BAD_CAST XML_SERIAL_THRESHOLD_CHANGE_NAME))
		{
			rc = CXMLHelper::DeserializeSignedInt(aReader, "Deserialize threshold change...", this->iRSerialInitBlockValueChange[aIndex], false);
			if (rc != ConstRC::Ok)
			{
				rc = ConstRC::GeneralError;
				break;
			}
			this->iInitBlockValueChange[aIndex] = this->iRSerialInitBlockValueChange[aIndex];
			continue;
		}
	}

	if (rc == ConstRC::GeneralError || rc == ConstRC::NotEnoughMemory)
	{
		return rc;
	}

	return ConstRC::Ok;
}
int CLatticeSievingPhase::Get_Iteration() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iIterationMutex));
	ret = this->iIteration;
	pthread_mutex_unlock(&(this->iIterationMutex));

	return ret;
}
int CLatticeSievingPhase::Set_Iteration(int aIteration)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iIterationMutex));
		this->iIteration = aIteration;
		pthread_mutex_unlock(&(this->iIterationMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CLatticeSievingPhase::Get_WorkingIteration() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iWorkingIterationMutex));
	ret = this->iWorkingIteration;
	pthread_mutex_unlock(&(this->iWorkingIterationMutex));

	return ret;
}
int CLatticeSievingPhase::Set_WorkingIteration(int aWorkingIteration)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iWorkingIterationMutex));
		this->iWorkingIteration = aWorkingIteration;
		pthread_mutex_unlock(&(this->iWorkingIterationMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CLatticeSievingPhase::Get_MaxIteration() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iMaxIterationMutex));
	ret = this->iMaxIteration;
	pthread_mutex_unlock(&(this->iMaxIterationMutex));

	return ret;
}
int CLatticeSievingPhase::Set_MaxIteration(int aMaxIteration)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iMaxIterationMutex));
		this->iMaxIteration = aMaxIteration;
		pthread_mutex_unlock(&(this->iMaxIterationMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
int CLatticeSievingPhase::Get_TimeMessage() const
{
	int ret = 0;
	pthread_mutex_lock(&(this->iTimeMessageMutex));
	ret = this->iTimeMessage;
	pthread_mutex_unlock(&(this->iTimeMessageMutex));

	return ret;
}
int CLatticeSievingPhase::Set_TimeMessage(int aTimeMessage)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iTimeMessageMutex));
		this->iTimeMessage = aTimeMessage;
		pthread_mutex_unlock(&(this->iTimeMessageMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}
vector<CNFSRelation*>* CLatticeSievingPhase::Get_RelationBuffer() const
{
	vector<CNFSRelation*>* ret = NULL;
	pthread_mutex_lock(&(this->iRelationBufferMutex));
	ret = this->iRelationBuffer;
	pthread_mutex_unlock(&(this->iRelationBufferMutex));

	return ret;
}
int CLatticeSievingPhase::Set_RelationBuffer(vector<CNFSRelation*>* aRelationBuffer, int aCount)
{
	if (Running())
	{
		pthread_mutex_lock(&(this->iRelationBufferMutex));
		this->iRelationBuffer = aRelationBuffer;
		this->iRelationBufferMaxCount = aCount;
		pthread_mutex_unlock(&(this->iRelationBufferMutex));

		return ConstRC::Ok;
	}

	return ConstRC::CannotBeDone;
}

/*
        Methods for distributed case.
*/

// initializes sieving phase and generates factor bases (called by center)
int CLatticeSievingPhase::InitDistributedSieving(AFactorAlgParameters *aParameters) {

        Set_Parameters(aParameters);

        if (InitParameters() != ConstRC::Ok) {
                CLog::LogError(WHEREARG, "CLatticeSievingPhase", "InitDistributedSieving", 0, 0, "Parameters initialization failed.");
                return ConstRC::GeneralError;
        }

        if (InitFactorBases() != ConstRC::Ok) {
                CLog::LogError(WHEREARG, "CLatticeSievingPhase", "InitDistributedSieving", 0, 0, "Factor bases initialization failed.");
                return ConstRC::GeneralError;
        }

        if (InitAdvFactorBases() != ConstRC::Ok) {
                CLog::LogError(WHEREARG, "CLatticeSievingPhase", "InitDistributedSieving", 0, 0, "Advanced factor bases initialization failed.");
                return ConstRC::GeneralError;
        }

	if (PrepareForSieving() != ConstRC::Ok) {
                CLog::LogError(WHEREARG, "CLatticeSievingPhase", "InitDistributedSieving", 0, 0, "Prepare for sieving failed.");
                return ConstRC::GeneralError;
	}

        DetermineRequestedRelations();

	iSpecialQWalkSteps = 0;

        return ConstRC::Ok;
}

// reads relations from given file, adds relations to count structures and appends to smooths or partials files
int CLatticeSievingPhase::SupplyRelations(const char *aPath, bool aSmooth, int &aRelsCount) {
        FILE *f_rel, *f_append;
        string path_append;
	int mode = 0;
	if (iUseCyclesForRelConstruction) mode = CNFSRelation::MODE_CYCLE_CONSTRUCTION;
	else {
		if (iUseHighVariations) mode = ARelation::MAX_LARGE_PRIMES;
	}
        CNFSRelation relation(mode);

        if (aSmooth) path_append = iSmoothRelFullFileName;
        else path_append = iPartialRelFullFileName;

        if ((f_rel = fopen(aPath, "r")) == NULL) {
                CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SupplyRelations", 0, 0, "Could not open relations file.");
                return ConstRC::CouldNotOpenFile;
        }
        if ((f_append = fopen(path_append.c_str(), "a")) == NULL) {
                CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SupplyRelations", 0, 0, "Could not open append file.");
                fclose(f_rel);
                return ConstRC::CouldNotOpenFile;
        }

        aRelsCount = 0;
        while (relation.ReadFromFile(f_rel) == ConstRC::Ok) {
//              if (CheckEnough() == ConstRC::EnoughSmooths) break;
                if (AddToCountStructures(&relation) != ConstRC::Ok) {
                        CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SupplyRelations", 0, 0, "Add to count structures failed.");
                        fclose(f_rel);
                        fclose(f_append);
                        return ConstRC::GeneralError;
                }
                if (relation.PrintToFile(f_append) != ConstRC::Ok) {
                        CLog::LogError(WHEREARG, "CLatticeSievingPhase", "SupplyRelations", 0, 0, "Print relation to file failed.");
                        fclose(f_rel);
                        fclose(f_append);
                        return ConstRC::GeneralError;
                }
                ++aRelsCount;
        }

        fclose(f_rel);
        fclose(f_append);

        return ConstRC::Ok;
}

void CLatticeSievingPhase::ConfirmAllStats() {

        iStats->ConfirmAll();

        return;
}

// increments special Q by aSteps-iSpecialQWalkSteps
// aSteps should be cumulative count of steps since start of sieving (iSpecialQWalkSteps is last cumulative count)
void CLatticeSievingPhase::WalkWithSpecialQ(long aSteps) {
	long steps, i;
	mpz_t prime;

	mpz_init_set_si(prime, this->iSpecialQ);

	steps = aSteps-iSpecialQWalkSteps;
	for (i = 0; i < steps; i++) mpz_nextprime(prime, prime);

	this->iSpecialQ = mpz_get_si(prime);

	mpz_clear(prime);

	iSpecialQWalkSteps = aSteps;

	this->Set_WorkingIteration(iSpecialQWalkSteps);

	return;
}

